/*
* Created on Mar 29, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package examples.bearingproblem;

import jsl.modeling.Experiment;
import jsl.modeling.JSLEvent;
import jsl.modeling.Model;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.random.distributions.Triangular;

/**
 *
 */
public class BearingProblemProposal extends SchedulingElement {

	public static final double FAILED = 0.0;
	public static final double WORKING = 1.0;
	public static final int BEARING_FAILED = 1;
	public static final int END_REPAIR = 2;
	
	protected RandomVariable myBearingLifeTime;
	protected RandomVariable myWalkingDelayTime;
	protected double myRepairTime;
	protected TimeWeighted myMillStatus;
	protected ResponseVariable myTotalCost;
	protected ResponseVariable myCostPerFailure;
	protected Counter myBearingCounter;
	
	protected int myNumBearingsUsed;
	protected int myNumRepairs;
	protected double myTotalDelayTime;
	protected double myTotalRepairTime;
	protected double myCostPerBearing = 30.0;// cost per bearing
	protected double myDownTimeCost = 10.0*60.0; // cost per hour
	protected double myRepairPersonCost = 24.0; // $/hour
	/**
	 * @param parent
	 */
	public BearingProblemProposal(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public BearingProblemProposal(ModelElement parent, String name) {
		super(parent, name);
		myBearingLifeTime = new RandomVariable(this, new Triangular(1000.0, 1200.0, 1900.0));
		myWalkingDelayTime = new RandomVariable(this, new Triangular(5./60., 5.0/60.0, 15.0/60.));
		myRepairTime = (40.0/60.0);
		myMillStatus = new TimeWeighted(this, WORKING, "Mill Status");
		myTotalCost = new ResponseVariable(this, "TotalCost");
		myCostPerFailure = new ResponseVariable(this, "AvgCostPerStoppage");
		myBearingCounter = new Counter(this, "NumberOfBearings");
	}

	protected void initialize(){
		myNumBearingsUsed = 0;
		myNumRepairs = 0;
		myTotalDelayTime = 0.0;
		myTotalRepairTime = 0.0;
		scheduleEvent(getMillTimeToFailure(), BEARING_FAILED);
	}
	
	protected double getMillTimeToFailure(){
		double min = myBearingLifeTime.getValue();;
		double t2 = myBearingLifeTime.getValue();
		if (t2 < min)
			min = t2;
		double t3 = myBearingLifeTime.getValue();
		if (t3 < min)
			min = t3;
		return(min);
	}
	
	protected void handleEvent(JSLEvent event){
		if (event.getType() == BEARING_FAILED){
			myMillStatus.setValue(FAILED);
			double w = myWalkingDelayTime.getValue();
			double rt = myRepairTime;
			myNumRepairs = myNumRepairs + 1;
			myNumBearingsUsed = myNumBearingsUsed + 3;
			myTotalDelayTime = myTotalDelayTime + w;
			myTotalRepairTime = myTotalRepairTime + rt;
			double c = myCostPerBearing*3.0 + (w+rt)*myDownTimeCost + rt*myRepairPersonCost;
			myCostPerFailure.setValue(c);
			myBearingCounter.increment(3);
			scheduleEvent(w+rt, END_REPAIR);
		}
		
		if (event.getType() == END_REPAIR){
			myMillStatus.setValue(WORKING);
			scheduleEvent(getMillTimeToFailure(), BEARING_FAILED);			
		}
	}
	
	protected void replicationEnded(){
		System.out.println("------------");
		System.out.println("Replication ended at time: " + getTime());
		System.out.println("num repairs = " + myNumRepairs);
		System.out.println("num bearings replaced = " + myNumBearingsUsed);
		System.out.println("bearing counter: " + myBearingCounter.getValue());
		System.out.println("------------");
		
		double bc = myNumBearingsUsed*myCostPerBearing;
		double dc = myTotalDelayTime*myDownTimeCost;
		double dtc = myTotalRepairTime*myDownTimeCost;
		double rpc = myTotalRepairTime*myRepairPersonCost;
		myTotalCost.setValue(bc+dc+dtc+rpc);
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// create the containing model
		Model m = Model.createModel();
		
		// create the model element and attach it to the main model
		new BearingProblemProposal(m, "BearingProblem");
				
		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
		// set the parameters of the experiment
		e.setNumberOfReplications(2);
		e.setLengthOfReplication(20000.0);
		
		// turn on the desired reporting
		e.turnOnExperimentReport();
		e.turnOnExperimentConsoleOutput();
		
		// tell the experiment to run
		e.runAll();	}

}
