/*
* Created on Apr 10, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package examples.elevator;

import java.util.HashMap;
import java.util.ListIterator;
import java.util.Map;

import jsl.modeling.JSLEvent;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.reporting.JSL;

/**
 * @author rossetti
 *
 */
public class Elevator extends SchedulingElement {

	protected double myInterFloorTravelTime = 0.25; // minutes
	
	protected Floor myHomeBase;
	
	protected Floor myCurrentFloor;
	
	protected Floor myNextFloor;
	
	protected ListIterator<Floor> myFloorIterator;
	
	protected int myCapacity = 6;
	
	public static enum Direction {GOING_UP, GOING_DOWN, NONE};
		
	protected Direction myStatus = Direction.NONE;
	
	protected ElevatorBank myElevatorBank;
	
	protected Map<Floor, Queue> myPassengers;
	
	protected TimeWeighted myNumPassengers;
	
	protected TimeWeighted myMovingEmptyIndicator;
	
	protected TimeWeighted myIdleIndicator;
	
	protected TimeWeighted myMovingWithPeopleIndicator;
	
	/**
	 * @param parent
	 */
	public Elevator(ElevatorBank elevatorModel) {
		this(elevatorModel, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public Elevator(ElevatorBank elevatorModel, String name) {
		super(elevatorModel, name);
		myElevatorBank = elevatorModel;
		myStatus = Direction.NONE;
		myNumPassengers = new TimeWeighted(this, 0.0, getName() + "NumPassengers");
		myMovingEmptyIndicator = new TimeWeighted(this, 0.0, getName() +"EmptyMoveIndicator");
		myIdleIndicator = new TimeWeighted(this, 1.0, getName() + "IdleIndicator");
		myMovingWithPeopleIndicator = new TimeWeighted(this, 0.0, getName()+"MovingWithPeopleIndicator");
		myPassengers = new HashMap<Floor, Queue>();
	}

	/**
	 * @return the apacity
	 */
	protected final int getCapacity() {
		return myCapacity;
	}

	/**
	 * @param myCapacity the myCapacity to set
	 */
	protected final void setCapacity(int capacity) {
		if (capacity <= 0)
			throw new IllegalArgumentException("The elevator capacity must be >=1");

		myCapacity = capacity;
	}

	/**
	 * @return the interFloorTravelTime
	 */
	protected final double getInterFloorTravelTime() {
		return myInterFloorTravelTime;
	}

	/**
	 * @param travelTime the InterFloorTravelTime to set
	 */
	protected final void setInterFloorTravelTime(double travelTime) {
		if (travelTime <=0.0)
			throw new IllegalArgumentException("The travel time must be > 0.0");
		myInterFloorTravelTime = travelTime;
	}

	public boolean isIdle(){
		return(myIdleIndicator.getValue() > 0.0);
	}
	
	public boolean hasPassengers(){
		return(myNumPassengers.getValue() > 0.0);
	}
	
	public boolean isGoingUp(){
		return(myStatus == Direction.GOING_UP);
	}
	
	public boolean isGoingDown(){
		return(myStatus == Direction.GOING_DOWN);
	}
	
	public String toString(){
		String s = "Elevator " + myStatus + " carrying " + myNumPassengers.getValue() + " passengers\n";
		for(Floor f: myPassengers.keySet()){
			Queue q = myPassengers.get(f);
			int nd = q.size();
			s = s + " #to drop off " + nd + " at " + f.getName() + "\n";
		}
		return(s);
	}
	
	protected void callElevator(Floor passengerFloor, ElevatorBank.Button button) {
		if (!isIdle())
			return;
		if (ElevatorBank.TRACE_ON){
			JSL.out.println("----------------------------------------");
			JSL.out.println(getTime() + " > The elevator is being called to floor " + passengerFloor.getName() + " to go " + button);
			JSL.out.println(this);
			JSL.out.println("----------------------------------------");			
		}
		
		// the elevator is idle, passenger needs pick up
		// determine the direction to go
		myIdleIndicator.setValue(false);

		if (myCurrentFloor.compareTo(passengerFloor) == 0){
			// passenger is at current floor
			if (button == ElevatorBank.Button.UP){
				myStatus = Direction.GOING_UP;
				Passenger p = (Passenger)myCurrentFloor.myUpQ.removeNext();
				Queue q = myPassengers.get(p.myDestinationFloor);
				q.enqueue(p);
				myNumPassengers.increment();
				moveToFloor(myFloorIterator.next());
			} else {
				myStatus = Direction.GOING_DOWN;
				Passenger p = (Passenger)myCurrentFloor.myDownQ.removeNext();
				Queue q = myPassengers.get(p.myDestinationFloor);
				q.enqueue(p);
				myNumPassengers.increment();
				moveToFloor(myFloorIterator.previous());				
			}
		} else if (myCurrentFloor.compareTo(passengerFloor) < 0){
				myStatus = Direction.GOING_UP;
				moveToFloor(myFloorIterator.next());
		} else {
			myStatus = Direction.GOING_DOWN;
			moveToFloor(myFloorIterator.previous());
		}		
	}
	
	protected void handleEvent(JSLEvent e){
		// arrival of elevator to a floor
		if (ElevatorBank.TRACE_ON){
			JSL.out.println("----------------------------------------");
			JSL.out.println(getTime() + " > \nEvent: Elevator arrived to floor " + myNextFloor.getName());
			JSL.out.println("Last Floor: " + myCurrentFloor.getName());
			JSL.out.println("Current Floor: " + myNextFloor.getName());
			JSL.out.println(myNextFloor);
			JSL.out.println(this);
			JSL.out.println("----------------------------------------");
		}

		myCurrentFloor = myNextFloor;
		
		unLoadPassengers();
		setDirection();
		pickUpPassengers();
		Floor f = getNextFloor();
		if (f != null) {
			moveToFloor(f);
		}
	}
	
	private void unLoadPassengers(){
		if (ElevatorBank.TRACE_ON){
			JSL.out.println(" > The elevator will unload any passengers");
		}
		
		Queue q = myPassengers.get(myCurrentFloor);
		while(!q.isEmpty()){
			Passenger p = (Passenger)q.removeNext();
			myNumPassengers.decrement();
			if (myCurrentFloor != myElevatorBank.getGroundFloor())
				p.beginFloorStay(myCurrentFloor);
		}
	}
	
	private void setDirection(){
		
		if (isGoingUp()){// going up
			if (!myFloorIterator.hasNext()) // no next floor
				myStatus = Direction.GOING_DOWN; // change direction
			return;
		}
		
		if (isGoingDown()){ // going down
			if (!myFloorIterator.hasPrevious()) // no next floor
				myStatus = Direction.GOING_UP; // change direction
			return;
		}
	}

	protected Floor getNextFloor() {
		Floor f = null;
		if (hasPassengers()){ // elevator has passengers
			// it only picks up people going in the same direction
			// continue movement in current direction
			if (ElevatorBank.TRACE_ON){
				JSL.out.println(" > The elevator is continuing " + myStatus + " because it has passengers.");
			}
			if (isGoingUp())
				return(myFloorIterator.next());
			if (isGoingDown())
				return(myFloorIterator.previous());
		} else { // no passengers were picked up
			// if going up, check for passengers going up, above current floor
			// if found return next floor in up direction
			if (isGoingUp()){
				if (myElevatorBank.checkForAnyWaitingPassengersAbove(myCurrentFloor)){
					if (ElevatorBank.TRACE_ON){
						JSL.out.println(" > The elevator is continuing " + myStatus + " because it has passengers waiting above it.");
					}
					return(myFloorIterator.next());
				}
					
			}
			
			// if going down, check for passengers going down, below current floor
			// if found return next floor in down direction
			if (isGoingDown()){
				if (myElevatorBank.checkForAnyWaitingPassengersBelow(myCurrentFloor)){
					if (ElevatorBank.TRACE_ON){
						JSL.out.println(" > The elevator is continuing " + myStatus + " because it has passengers waiting below it.");
					}
					return(myFloorIterator.previous());
				}
					
			}
			if (ElevatorBank.TRACE_ON){
				JSL.out.println(" > The elevator found no passengers going in same direction and it has no passengers");
				JSL.out.println(" > The elevator is checking the current floor for passengers going in the opposite direction");
			}			
			// none found going in same direction and elevator has no passengers
			// check current floor for any passengers going in opposite direction
			if (isGoingUp()){// elevator currently going up
				if (!myCurrentFloor.myDownQ.isEmpty()){// check for going down
					myStatus = Direction.GOING_DOWN; // change direction
					if (ElevatorBank.TRACE_ON){
						JSL.out.println(" > The elevator is picking up passengers on the current floor and changing direction to go "+ myStatus);
					}
					pickUpPassengers();
					return(myFloorIterator.previous());
				}
			}
			
			// check current floor for any passengers going in opposite direction
			if (isGoingDown()){// elevator currently going down
				if (!myCurrentFloor.myUpQ.isEmpty()){// check for going up passengers
					myStatus = Direction.GOING_UP; // change direction
					if (ElevatorBank.TRACE_ON){
						JSL.out.println(" > The elevator is picking up passengers on the current floor and changing direction to go "+ myStatus);
					}
					pickUpPassengers();
					return(myFloorIterator.previous());
				}
			}

			if (ElevatorBank.TRACE_ON){
				JSL.out.println(" > The elevator found no passengers at the current floor");
				JSL.out.println(" > The elevator must check for any passengers waiting below the current floor");
			}
			
			if (myElevatorBank.checkForAnyWaitingPassengersBelow(myCurrentFloor)){
				myStatus = Direction.GOING_DOWN; // change direction
				if (ElevatorBank.TRACE_ON){
					JSL.out.println(" > The elevator found people waiting on a lower floor and is changing direction to go "+ myStatus);
				}
				return(myFloorIterator.previous());						
			}
			
			if (ElevatorBank.TRACE_ON){
				JSL.out.println(" > The elevator found no passengers waiting below the floor");
				JSL.out.println(" > The elevator must check for any passengers waiting above the current floor");
			}
			
			if (myElevatorBank.checkForAnyWaitingPassengersAbove(myCurrentFloor)){
				myStatus = Direction.GOING_UP; // change direction
				if (ElevatorBank.TRACE_ON){
					JSL.out.println(" > The elevator found people waiting on a upper floor and is changing direction to go "+ myStatus);
				}
				return(myFloorIterator.next());											
			}
			
			// then elevator should go towards the home base
			if (ElevatorBank.TRACE_ON){
				JSL.out.println(" > The elevator found no passengers anywhere and has decided to return to its home base");
			}
			if (myCurrentFloor.compareTo(myHomeBase) == 0){
				makeElevatorIdle();
			} else if (myCurrentFloor.compareTo(myHomeBase) < 0){
					myStatus = Direction.GOING_UP;
					if (ElevatorBank.TRACE_ON){
						JSL.out.println(this);
						JSL.out.println("----------------------------------------");
					}
					return(myFloorIterator.next());
			} else {
				myStatus = Direction.GOING_DOWN;
				if (ElevatorBank.TRACE_ON){
					JSL.out.println(this);
					JSL.out.println("----------------------------------------");
				}
				return(myFloorIterator.previous());
			}
		}
		
		return(f);
	}
	
	private void makeElevatorIdle() {
		myMovingEmptyIndicator.setValue(false);
		myIdleIndicator.setValue(true);
		myMovingWithPeopleIndicator.setValue(false);
		myStatus = Direction.NONE;
		if (ElevatorBank.TRACE_ON){
			JSL.out.println("----------------------------------------");
			JSL.out.println(getTime() + " > The elevator has gone idle");
			JSL.out.println(this);
			JSL.out.println("----------------------------------------");
		}
	}

	private void pickUpPassengers(){
		if (isIdle())
			throw new IllegalStateException("Can't pick up passengers when idle");
		
		if (ElevatorBank.TRACE_ON){
			JSL.out.println(" > The elevator will attempt to pick up passengers");
		}

		Queue q = null;
		if (myStatus == Direction.GOING_UP)
			q = myCurrentFloor.myUpQ;
		else
			q = myCurrentFloor.myDownQ;
		// determine the number that can be picked up
		int n = myCapacity - (int)myNumPassengers.getValue();
		int m = q.size();
		n = Math.min(n, m);
		// n is the number that will be picked up
		// transfer them from the floor queue to the elevator
		for(int i=1;i<=n;i++){
			Passenger p = (Passenger)q.removeNext();
			Queue eq = myPassengers.get(p.myDestinationFloor);
			eq.enqueue(p);
			myNumPassengers.increment();
		}
		// determine how many that didn't get pick up
		int c = m - n;
		for(int i=1;i<=m;i++){
			if (i<=c)
				myCurrentFloor.myProportionNotPickedUp.setValue(1.0);
			else
				myCurrentFloor.myProportionNotPickedUp.setValue(0.0);
		}
	}

	private void moveToFloor(Floor f) {
		myNextFloor = f;
		if (hasPassengers()){
			myMovingWithPeopleIndicator.setValue(true);
			myMovingEmptyIndicator.setValue(false);
		}
		else {
			myMovingWithPeopleIndicator.setValue(false);
			myMovingEmptyIndicator.setValue(true);
		}
	
		if (myCurrentFloor == f)
			scheduleEvent(0.0);
		else
			scheduleEvent(myInterFloorTravelTime);
	}
	
	protected void addInternalPassengerQ(Floor floor){
		Queue q = new Queue(this, getName() + "-" + floor.getName() + "-Q");
		myPassengers.put(floor, q);
	}
	
	protected void setHomeBase(Floor f) {
		myHomeBase = f;
	}

	protected void initialize(){
		myCurrentFloor = myHomeBase;
		myFloorIterator = myElevatorBank.getFloorIterator(myCurrentFloor);
		myStatus = Direction.NONE;
	}
}
