/*
* Created on Apr 10, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package examples.elevator;

import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;

import jsl.modeling.Experiment;
import jsl.modeling.JSLEvent;
import jsl.modeling.Model;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.EventGenerator;
import jsl.modeling.elements.EventGeneratorListenerIfc;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.reporting.JSL;

/**
 * @author rossetti
 *
 */
public class ElevatorBank extends SchedulingElement {
	
	public static boolean TRACE_ON = false;
	
	public static enum Button {UP, DOWN};

	protected EventGenerator myGroundFloorPassengerGenerator;
	
	protected List<Floor> myFloors;
	
	protected Elevator myElevator;
	
	protected Floor myHomeBase;
	
	protected ResponseVariableAverageObserver myOverallQDelay;
	
	/**
	 * @param parent
	 */
	public ElevatorBank(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public ElevatorBank(ModelElement parent, String name) {
		super(parent, name);
		myGroundFloorPassengerGenerator = new EventGenerator(this, new GroundFloorArrival(), new Exponential(1.0), new Exponential(1.0), "Passenger Generator");
		myOverallQDelay = new ResponseVariableAverageObserver(this, "Delay Across Qs");
		myElevator = new Elevator(this, "Elevator");
		myFloors = new ArrayList<Floor>();
	}

	/**
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setMaximumNumberOfEventsForReplications(long)
	 */
	public void setMaximumNumberOfPassengerArrivals(long maxNumEvents) {
		myGroundFloorPassengerGenerator.setMaximumNumberOfEventsForReplications(maxNumEvents);
	}

	/**
	 * @param timeBtwEvents
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEventsForReplications(jsl.utilities.random.RandomIfc, long)
	 */
	public void setTimeBetweenPassengerArrivals(RandomIfc timeBtwEvents, long maxNumEvents) {
		myGroundFloorPassengerGenerator.setTimeBetweenEventsForReplications(timeBtwEvents, maxNumEvents);
	}

	/**
	 * @param timeBtwEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEventsForReplications(jsl.utilities.random.RandomIfc)
	 */
	public void setTimeBetweenPassengerArrivals(RandomIfc timeBtwEvents) {
		myGroundFloorPassengerGenerator.setTimeBetweenEventsForReplications(timeBtwEvents);
	}

	public void setElevatorCapacity(int c){
		myElevator.setCapacity(c);
	}
	
	public void setInterFloorTravelTime(double time){
		myElevator.setInterFloorTravelTime(time);
	}
	
	public Floor addFloor(){
		String s = "Floor-"+ (myFloors.size()+1);
		Floor f = new Floor(this, s);
		myFloors.add(f);
		f.addQueueTimeAggregate(myOverallQDelay);
		f.setFloorNumber(myFloors.indexOf(f));
		f.setElevator(myElevator);
		if (f.myNum == 0){
			myElevator.setHomeBase(f);
		}
		myElevator.addInternalPassengerQ(f);
		return(f);
	}
	
	public boolean checkForWaitingPassengersGoingUp(Floor f){
		// determine index of floor
		int index = myFloors.indexOf(f);
		// get an iterator positioned to return the floor as the next
		ListIterator<Floor> i = myFloors.listIterator(index);
		i.next();
		while(i.hasNext()){
			Floor fn = i.next();
			if (!fn.myUpQ.isEmpty())
				return(true);
		}
		return(false);
	}
	
	public boolean checkForWaitingPassengersGoingDown(Floor f){
		// determine index of floor
		int index = myFloors.indexOf(f);
		// get an iterator positioned to return the floor as the next
		ListIterator<Floor> i = myFloors.listIterator(index);
		while(i.hasPrevious()){
			Floor fn = i.previous();
			if (!fn.myDownQ.isEmpty())
				return(true);
		}		
		return(false);
	}
	
	public boolean checkForAnyWaitingPassengersBelow(Floor f){
		// determine index of floor
		int index = myFloors.indexOf(f);
		// get an iterator positioned to return the floor as the next
		ListIterator<Floor> i = myFloors.listIterator(index);
		while(i.hasPrevious()){
			Floor fn = i.previous();
			if (!fn.myDownQ.isEmpty() || !fn.myUpQ.isEmpty())
				return(true);
		}		
		
		return(false);
	}
	
	public boolean checkForAnyWaitingPassengersAbove(Floor f){
		// determine index of floor
		int index = myFloors.indexOf(f);
		// get an iterator positioned to return the floor as the next
		ListIterator<Floor> i = myFloors.listIterator(index);
		i.next();
		while(i.hasNext()){
			Floor fn = i.next();
			if (!fn.myUpQ.isEmpty() || !fn.myDownQ.isEmpty())
				return(true);
		}
		return(false);
	}
	
	public void setHomeBase(Floor f){
		myHomeBase = f;
		myElevator.setHomeBase(f);
	}
	
	public Floor getGroundFloor(){
		return(myFloors.get(0));
	}
	
	public Floor getTopFloor(){
		return(myFloors.get(myFloors.size()-1));
	}
	
	protected void initialize(){
		if (myFloors.size() <= 1)
			throw new IllegalStateException("The number of floors must be >= 2");
	}
	
	/** Returns an iterator position to return the homebase as the first
	 *  call to next()
	 * 
	 * @return
	 */
	protected ListIterator<Floor> getFloorIterator(Floor f){
		if (f == null)
			throw new IllegalArgumentException("The floor was null");
		
		if (!myFloors.contains(f))
			throw new IllegalArgumentException("The floor is not valid for this Elevator Bank");
		
		int n = myFloors.indexOf(f);
		return(myFloors.listIterator(n));
	}
	
	protected class GroundFloorArrival implements EventGeneratorListenerIfc {

		public void generate(EventGenerator generator, JSLEvent event) {
			// create the arriving customer
			Passenger p = new Passenger();
			// set the current floor
			Floor currentFloor = getGroundFloor();
			p.myCurrentFloor = currentFloor;
			// determine the destination floor
			p.determineDestination();
			// put passenger in up queue for the floor
			currentFloor.myUpQ.enqueue(p);
			// signal the elevator
			if (TRACE_ON){
				JSL.out.println("******************************************");
				JSL.out.println(getTime() + " > \nEvent: GroundFloorArrival");
				JSL.out.println(p);
				JSL.out.println(currentFloor);
				JSL.out.println("******************************************");					
			}
			myElevator.callElevator(currentFloor, Button.UP);
		}
		
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		System.out.println("Elevator Problem");
		// create the containing model
		Model m = Model.createModel();		
		// create the model element and attach it to the main model
		ElevatorBank em = new ElevatorBank(m);
//		em.setMaximumNumberOfPassengerArrivals(3);
//		ElevatorBank.TRACE_ON = true;
		
		Floor f1 = em.addFloor();
		Floor f2 = em.addFloor();
		Floor f3 = em.addFloor();
		Floor f4 = em.addFloor();
		Floor f5 = em.addFloor();
		
		f1.addDestinationFloor(f2, 0.25);
		f1.addDestinationFloor(f3, 0.25);
		f1.addDestinationFloor(f4, 0.25);
		f1.addDestinationFloor(f5);

		f2.addDestinationFloor(f1, 0.70);
		f2.addDestinationFloor(f3, 0.10);
		f2.addDestinationFloor(f4, 0.10);
		f2.addDestinationFloor(f5);

		f3.addDestinationFloor(f1, 0.70);
		f3.addDestinationFloor(f2, 0.10);
		f3.addDestinationFloor(f4, 0.10);
		f3.addDestinationFloor(f5);
		
		f4.addDestinationFloor(f1, 0.70);
		f4.addDestinationFloor(f2, 0.10);
		f4.addDestinationFloor(f3, 0.10);
		f4.addDestinationFloor(f5);
		
		f5.addDestinationFloor(f1, 0.70);
		f5.addDestinationFloor(f2, 0.10);
		f5.addDestinationFloor(f3, 0.10);
		f5.addDestinationFloor(f4);

		//em.setHomeBase(f3);
		
		// create the experiment to run the model
		Experiment e = new Experiment(m, "Elevator Experiments");
		
		// set the parameters of the experiment
		e.setNumberOfReplications(10);
		e.setLengthOfReplication(20.0*60.0);
//		e.setLengthOfWarmUp(5.0);

		// turn on the desired reporting
		e.turnOnReplicationConsoleOutput();
		e.turnOnExperimentReport();
//		e.turnOnExperimentConsoleOutput();
		
		// tell the experiment to run
		e.runAll();
		System.out.println("Done!");

	}
}
