/*
* Created on Apr 10, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package examples.elevator;

import jsl.modeling.JSLEvent;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.variable.Aggregate;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Uniform;
import jsl.utilities.reporting.JSL;

/**
 * @author rossetti
 *
 */
public class Floor extends SchedulingElement implements Comparable<Floor> {

	protected static final RandomIfc DEFAULT_FLOOR_TIME = new Uniform(15.0, 120.0);
	protected Queue myUpQ;
	protected Queue myDownQ;
	protected int myNum;
	protected RandomElement<Floor> myDestFloors;
	protected RandomVariable myFloorStayTime;
	protected Elevator myElevator;
	protected ElevatorBank myElevatorBank;
	protected ResponseVariable myProportionNotPickedUp;
	
	/**
	 * @param parent
	 */
	protected Floor(ElevatorBank elevatorModel) {
		this(elevatorModel, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	protected Floor(ElevatorBank elevatorBank, String name) {
		super(elevatorBank, name);
		myElevatorBank = elevatorBank;
		myUpQ = new Queue(this, getName() +" UpQ");
		myDownQ = new Queue(this, getName() +" DownQ");
		myDestFloors = new RandomElement<Floor>(this, getName() + "Destinations");
		myFloorStayTime = new RandomVariable(this, DEFAULT_FLOOR_TIME, getName() + "FloorStayTime");
		myProportionNotPickedUp = new ResponseVariable(this, getName() + "Proportion Not Picked Up");
	}

	public void setFloorStayTimeInitialRandomSource(RandomIfc r){
		myFloorStayTime.setInitialRandomSource(r);
	}
	
	public void addDestinationFloor(Floor f, double prob){
		myDestFloors.add(f, prob);
	}
	
	public void addDestinationFloor(Floor f){
		myDestFloors.addLast(f);
	}
	
	public Floor generateDestination(){
		return(myDestFloors.getRandomElement());
	}
	
	protected void addQueueTimeAggregate(Aggregate aggregate){
		myUpQ.subscribeToTimeInQueue(aggregate);
		myDownQ.subscribeToTimeInQueue(aggregate);
	}
	
	protected void scheduleFloorStay(Passenger p){
		p.myCurrentFloor = this;
		scheduleEvent(myFloorStayTime.getValue(), p);
	}
	
	protected void handleEvent(JSLEvent e){
		// end of floor stay for the passenger
		Passenger p = (Passenger)e.getMessage();
		p.determineDestination();
		if (ElevatorBank.TRACE_ON){
			JSL.out.println("####################################################################");
			JSL.out.println(getTime() + " > \nEvent: Passenger arrives to floor " + getName() + " after floor stay");
			JSL.out.println(p);
		}
		if (compareTo(p.myDestinationFloor) > 0){
			myDownQ.enqueue(p);
			myElevator.callElevator(this, ElevatorBank.Button.DOWN);
		}
		else {
			myUpQ.enqueue(p);
			myElevator.callElevator(this, ElevatorBank.Button.UP);
		}
		
		if (ElevatorBank.TRACE_ON){
			JSL.out.println(this);
			JSL.out.println("####################################################################");
		}
	}
		
	protected void setFloorNumber(int n){
		myNum = n;
	}
	
	protected void setElevator(Elevator e){
		myElevator = e;
	}

	public int compareTo(Floor f) {
		if (myNum < f.myNum)
			return (-1);
		if (myNum > f.myNum)
			return (1);
		
		return(0);
	}
	
	public String toString(){
		int nup = myUpQ.size();
		int nd = myDownQ.size();
		String s = getName() + " #going up= " + nup + " #going down = " + nd;
		return(s);
	}
}
