/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/


package examples.jobshopV1;

import java.util.*;

import jsl.modeling.elements.*;
import jsl.modeling.elements.variable.*;
import jsl.modeling.*;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.DEmpirical;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.random.distributions.Gamma;

/**
 *
 */
public class JobShop extends ModelElement {
    
    private EventGenerator myJobGenerator;
    private List<WorkStation> myWorkStations;
    private RandomVariable myJobType;
    private List<JobStep>[] myTypeSequence;
    private ResponseVariable[] mySystemTime;
    
    /** Creates a new instance of JobShop */
    public JobShop(ModelElement parent) {
    	super(parent);

        myJobGenerator = new EventGenerator(this, new JobArrival(), new Exponential(0.25), new Exponential(0.25), "Job Generator");
        
        myWorkStations = new ArrayList<WorkStation>();
        
        // create the workstations
        
        WorkStation w1 = new WorkStation(this, "w1", 3);
        myWorkStations.add(w1);
        
        WorkStation w2 = new WorkStation(this, "w2", 2);
        myWorkStations.add(w2);
 
        WorkStation w3 = new WorkStation(this, "w3", 4);
        myWorkStations.add(w3);

        WorkStation w4 = new WorkStation(this, "w4", 3);
        myWorkStations.add(w4);
        
        WorkStation w5 = new WorkStation(this, "w5", 1);
        myWorkStations.add(w5);
  
        // create the service time distributions
        
        Variable v13 = new RandomVariable(this, new Gamma(2.0, 0.5/2.0));

        Variable v11 = new RandomVariable(this, new Gamma(2.0, 0.6/2.0));

        Variable v12 = new RandomVariable(this, new Gamma(2.0, 0.85/2.0));

        Variable v15 = new RandomVariable(this, new Gamma(2.0, 0.5/2.0));

        Variable v24 = new RandomVariable(this, new Gamma(2.0, 1.1/2.0));

        Variable v21 = new RandomVariable(this, new Gamma(2.0, 0.8/2.0));

        Variable v23 = new RandomVariable(this, new Gamma(2.0, 0.75/2.0));
        
        Variable v32 = new RandomVariable(this, new Gamma(2.0, 1.2/2.0));
        
        Variable v35 = new RandomVariable(this, new Gamma(2.0, 0.25/2.0));
        
        Variable v31 = new RandomVariable(this, new Gamma(2.0, 0.7/2.0));
        
        Variable v34 = new RandomVariable(this, new Gamma(2.0, 0.9/2.0));
        
        Variable v33 = new RandomVariable(this, new Gamma(2.0, 1.0/2.0));
        
        
        // make a sequence for each job type
        myTypeSequence = new List[4];
        mySystemTime = new ResponseVariable[4];
        
        for(int i=1;i<=3;i++){
            myTypeSequence[i] = new ArrayList<JobStep>();
            mySystemTime[i] = new ResponseVariable(this, "System Time Type"+i);
        }
        
        // make and add a job step for each sequence
        
        myTypeSequence[1].add(new JobStep(w3,v13));
        myTypeSequence[1].add(new JobStep(w1,v11));
        myTypeSequence[1].add(new JobStep(w2,v12));
        myTypeSequence[1].add(new JobStep(w5,v15));
              
        myTypeSequence[2].add(new JobStep(w4,v24));
        myTypeSequence[2].add(new JobStep(w1,v21));
        myTypeSequence[2].add(new JobStep(w3,v23));
               
        myTypeSequence[3].add(new JobStep(w2,v32));
        myTypeSequence[3].add(new JobStep(w5,v35));
        myTypeSequence[3].add(new JobStep(w1,v31));
        myTypeSequence[3].add(new JobStep(w4,v34));
        myTypeSequence[3].add(new JobStep(w3,v33));
              
        DEmpirical d = new DEmpirical();
        d.addProbabilityPoint(1.0, 0.3);
        d.addProbabilityPoint(2.0, 0.5);
        d.addLastProbabilityPoint(3.0);
        myJobType = new RandomVariable(this, d);
    }
     
    private class JobArrival implements EventGeneratorListenerIfc {
        public void generate(EventGenerator generator, JSLEvent event) {
            int type = (int)myJobType.getValue();
             // construct a job
            Job job = new Job(type, getTime(), myTypeSequence[type], mySystemTime[type]);
             
            // send the job to next location
            job.nextJobStep();
        }        
    }
}
