/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package examples.jobshopV2;

import jsl.modeling.Experiment;
import jsl.modeling.Model;
import jsl.modeling.elements.processview.description.EntityProcessGenerator;
import jsl.modeling.elements.processview.description.ProcessDescription;
import jsl.modeling.elements.processview.description.commands.Delay;
import jsl.modeling.elements.processview.description.commands.Release;
import jsl.modeling.elements.processview.description.commands.Seize;
import jsl.modeling.elements.processview.description.commands.Terminate;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.resource.Resource;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.Variable;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.random.distributions.Gamma;

/**
 * @author rossetti
 *
 */
public class ProcessViewJobShop {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// TODO Auto-generated method stub
		Model m = buildModel();
		
		// create the experiment to run the model
		Experiment e = new Experiment(m, "Process View");
		
		// set the parameters of the experiment
		e.setNumberOfReplications(100);
		e.setLengthOfReplication(10000.0);
		e.setLengthOfWarmUp(5000.0);
	
		// turn on the desired reporting
		e.turnOnExperimentReport();
		e.turnOnExperimentConsoleOutput();
		
		// tell the experiment to run
		e.runAll();
		System.out.println("Done!");
	    
	}
	
	public static Model buildModel(){
		// create the containing model
		Model m = Model.createModel();

		// create the elements in the model
		
		Variable v1 = new Variable(m, 1.0, "amt requested");
		RandomVariable sts1w3 = new RandomVariable(m,new Gamma(2.0, 0.5/2.0),"ST S1 W3");
		RandomVariable sts1w1 = new RandomVariable(m,new Gamma(2.0, 0.6/2.0),"ST S1 W1");
		RandomVariable sts1w2 = new RandomVariable(m,new Gamma(2.0, 0.85/2.0),"ST S1 W2");
		RandomVariable sts1w5 = new RandomVariable(m,new Gamma(2.0, 0.5/2.0),"ST S1 W5");

		RandomVariable sts2w4 = new RandomVariable(m,new Gamma(2.0, 1.1/2.0), "ST S2 W4");
		RandomVariable sts2w1 = new RandomVariable(m,new Gamma(2.0, 0.8/2.0), "ST S2 W1");
		RandomVariable sts2w3 = new RandomVariable(m,new Gamma(2.0, 0.75/2.0), "ST S2 W3");

		RandomVariable sts3w2 = new RandomVariable(m,new Gamma(2.0, 1.2/2.0),"ST S3 W2");
		RandomVariable sts3w5 = new RandomVariable(m,new Gamma(2.0, 0.25/2.0),"ST S3 W5");
		RandomVariable sts3w1 = new RandomVariable(m,new Gamma(2.0, 0.7/2.0),"ST S3 W1");
		RandomVariable sts3w4 = new RandomVariable(m,new Gamma(2.0, 0.9/2.0),"ST S3 W4");
		RandomVariable sts3w3 = new RandomVariable(m,new Gamma(2.0, 1.0/2.0),"ST S3 W3");
	    
		Resource r1 = new Resource(m, 3, "W1");
		Queue q1 = new Queue(m, "W1 Queue");
		
		Resource r2 = new Resource(m, 2, "W2");
		Queue q2 = new Queue(m, "W2 Queue");
		
		Resource r3 = new Resource(m, 4, "W3");
		Queue q3 = new Queue(m, "W3 Queue");
		
		Resource r4 = new Resource(m, 3, "W4");
		Queue q4 = new Queue(m, "W4 Queue");
		
		Resource r5 = new Resource(m, 1, "W5");
		Queue q5 = new Queue(m, "W5 Queue");
		
		// create the process description
		ProcessDescription jt1 = new ProcessDescription(m, "Job Type 1");
		DistributionIfc tbajt1 = new Exponential((10.0/3.0)*0.25);
		
		// create the model element to generate type 1 jobs
		new EntityProcessGenerator(m, jt1, tbajt1, tbajt1);

		// create the commands used in the process description
		// workstation 3
		jt1.addProcessCommand(new Seize(m, v1, r3, q3));
		jt1.addProcessCommand(new Delay(m, sts1w3));
		jt1.addProcessCommand(new Release(m, r3, q3));
		// workstation 1
		jt1.addProcessCommand(new Seize(m, v1, r1, q1));
		jt1.addProcessCommand(new Delay(m, sts1w1));
		jt1.addProcessCommand(new Release(m, r1, q1));
		// workstation 2
		jt1.addProcessCommand(new Seize(m, v1, r2, q2));
		jt1.addProcessCommand(new Delay(m, sts1w2));
		jt1.addProcessCommand(new Release(m, r2, q2));
		// workstation 5
		jt1.addProcessCommand(new Seize(m, v1, r5, q5));
		jt1.addProcessCommand(new Delay(m, sts1w5));
		jt1.addProcessCommand(new Release(m, r5, q5));
		jt1.addProcessCommand(new Terminate(m,true,"Type 1"));
		
		// create the process description
		ProcessDescription jt2 = new ProcessDescription(m, "Job Type 2");
		DistributionIfc tbajt2 = new Exponential((10.0/5.0)*0.25);
		
		// create the model element to generate type 2 jobs
		new EntityProcessGenerator(m, jt2, tbajt2, tbajt2);

		// create the commands used in the process description
		// workstation 4
		jt2.addProcessCommand(new Seize(m, v1, r4, q4));
		jt2.addProcessCommand(new Delay(m, sts2w4));
		jt2.addProcessCommand(new Release(m, r4, q4));
		// workstation 1
		jt2.addProcessCommand(new Seize(m, v1, r1, q1));
		jt2.addProcessCommand(new Delay(m, sts2w1));
		jt2.addProcessCommand(new Release(m, r1, q1));
		// workstation 3
		jt2.addProcessCommand(new Seize(m, v1, r3, q3));
		jt2.addProcessCommand(new Delay(m, sts2w3));
		jt2.addProcessCommand(new Release(m, r3, q3));
		jt2.addProcessCommand(new Terminate(m,true,"Type 2"));

		// create the process description
		ProcessDescription jt3 = new ProcessDescription(m, "Job Type 3");
		DistributionIfc tbajt3 = new Exponential((10.0/2.0)*0.25);
		
		// create the model element to generate type 3 jobs
		new EntityProcessGenerator(m, jt3, tbajt3, tbajt3);

		// create the commands used in the process description
		// workstation 2
		jt3.addProcessCommand(new Seize(m, v1, r2, q2));
		jt3.addProcessCommand(new Delay(m, sts3w2));
		jt3.addProcessCommand(new Release(m, r2, q2));
		// workstation 5
		jt3.addProcessCommand(new Seize(m, v1, r5, q5));
		jt3.addProcessCommand(new Delay(m, sts3w5));
		jt3.addProcessCommand(new Release(m, r5, q5));
		// workstation 1
		jt3.addProcessCommand(new Seize(m, v1, r1, q1));
		jt3.addProcessCommand(new Delay(m, sts3w1));
		jt3.addProcessCommand(new Release(m, r1, q1));
		// workstation 4
		jt3.addProcessCommand(new Seize(m, v1, r4, q4));
		jt3.addProcessCommand(new Delay(m, sts3w4));
		jt3.addProcessCommand(new Release(m, r4, q4));	
		// workstation 3
		jt3.addProcessCommand(new Seize(m, v1, r3, q3));
		jt3.addProcessCommand(new Delay(m, sts3w3));
		jt3.addProcessCommand(new Release(m, r3, q3));	
		jt3.addProcessCommand(new Terminate(m,true,"Type 3"));
		
		return (m);
	}

}
