/*
* Created on Apr 23, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package examples.kanban;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.queue.QObject;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.resource.Request;
import jsl.modeling.elements.resource.Resource;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.distributions.Exponential;

/**
 * @author rossetti
 *
 */
public class KanbanStation extends SchedulingElement implements SupplierIfc, CustomerIfc {

	protected Queue myKBBox;
	protected Queue myOutputBuffer;
	protected Queue myPartQ;
	protected Resource myMachines;
	protected RandomVariable myST;
	protected int myInitialKB = 2;
	
	CustomerIfc myCustomer;
	SupplierIfc mySupplier;
	
	/**
	 * @param parent
	 */
	public KanbanStation(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public KanbanStation(ModelElement parent, String name) {
		super(parent, name);
		myKBBox = new Queue(this, getName()+"KBBox");
		myOutputBuffer = new Queue(this, getName()+"OBuffer");
		myPartQ = new Queue(this, getName()+"PartQ");
		myMachines = new Resource(this, getName() +"Machines");
		myST = new RandomVariable(this, new Exponential(1.0/3.0));
	}

	protected void initialize(){
		for(int i=1;i<=myInitialKB;i++){
			QObject part = new QObject();
			QObject kb = new QObject();
			part.setQueuedObject(kb);
			myOutputBuffer.enqueue(part);
		}
	}
	
	public void setCustomer(CustomerIfc c){
		myCustomer = c;
	}
	
	public void setSupplier(SupplierIfc s){
		mySupplier = s;
	}
	
	public void setNumberOfKanbans(int k){
		if (k <= 0)
			throw new IllegalArgumentException("The #kanbans must be > 0");
		myInitialKB = k;
	}
	

	public void sendPart() {
		if (!myOutputBuffer.isEmpty())
			removePartFromOutputBuffer();
	}

	private void removePartFromOutputBuffer(){
		QObject part = myOutputBuffer.removeNext();
		QObject kb = (QObject)part.getQueuedObject();
		part.setQueuedObject(null);
		myCustomer.receivePart(part);
		myKBBox.enqueue(kb);
		mySupplier.sendPart();
	}
	
	public boolean needsPart() {
		return !myKBBox.isEmpty();
	}

	public void receivePart(QObject part) {
		QObject kb = myKBBox.removeNext();
		part.setQueuedObject(kb);
		startProcessing(part);
	}

	private void startProcessing(QObject part) {
		Request r = Request.create();// a request for 1 unit
		r.setRequestingObject(part);
		//enqueue the request
		myPartQ.enqueue(r);
		// check if resource is idle
		if (myMachines.isIdle()){
			// at least 1 unit is available, check if the arriving request
			// is the next in the queue
			Request next = (Request)myPartQ.peekNext();
			if (r == next){
				// check if number idle is sufficient
				if (next.getAmountNeeded() <= myMachines.getNumberIdle()){
					// remove the request
					myPartQ.removeNext();
					// allocate the resource to the request
					myMachines.allocate(r);
					//	schedule end of service, include the request as the event's message
					scheduleEvent(myST, r); 
				}
			}
		}		
	}

	protected void handleEvent(JSLEvent e){
		// end of service at the machines
		// the request was attached
		Request r = (Request)e.getMessage();
		// get the attached part on the request
		QObject part = (QObject)r.getRequestingObject();
		// store the part in the output buffer
		storePart(part);
		// tell the resource to release the request
		myMachines.release(r);
		// check if machine can work on next request
		if (!myPartQ.isEmpty()) { // queue is not empty
			Request next= (Request)myPartQ.peekNext(); //check the next request
			if (next.getAmountNeeded() <= myMachines.getNumberIdle()){
				myPartQ.removeNext();
				myMachines.allocate(next);
				scheduleEvent(myST, next);
			}
		}
	}
	
	private void storePart(QObject part){
		// put the part in the output buffer
		myOutputBuffer.enqueue(part);
		// check if customer needs the part
		if (myCustomer.needsPart()){
			removePartFromOutputBuffer();
		}
	}
}
