/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package examples.montecarlo;

import jsl.utilities.math.FunctionIfc;
import jsl.utilities.rootfinding.BisectionRootFinder;
import jsl.utilities.rootfinding.Interval;
import jsl.utilities.rootfinding.RootFinder;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.Beta;
import jsl.utilities.random.distributions.Lognormal;
import jsl.utilities.random.distributions.Normal;
import jsl.utilities.statistic.*;

public class NORTAEvaluation implements FunctionIfc {

	private RootFinder myRootFinder = new BisectionRootFinder();
    
	protected Normal snd;
	protected Beta xd;
	protected Lognormal yd;
	protected int mySampleSize = 1000;
	protected int myReplications = 1000;
	protected StatisticXY statXY;
	protected Statistic stat;
	protected Statistic tStat;
	protected double[] x;
	protected double myRho;
	protected double myDesiredRho;
	
	public NORTAEvaluation(double rho) {
		setRho(rho);
		xd = new Beta(1.5162, 2.1784);
		yd = new Lognormal(0.289, 0.336);
		snd = new Normal();
		statXY = new StatisticXY();
		stat = new Statistic();
		tStat = new Statistic();
		x = new double[2];
	}
	
	public void setRho(double rho){
		if ((rho < -1.0) || (rho > 1.0))
			throw new IllegalArgumentException("rho must be in [-1,1]");
		
		myRho = rho;
	}

	public double estCorrelation(){
		stat.reset();
		System.out.println("Performing replication: ");
		for(int i=1;i<=myReplications;i++){// macro replications
			System.out.print(i + ",");
			if ( (i%20) == 0) System.out.println();
			
			long t = System.currentTimeMillis();
			
			double erho = sampleCorrelation();
			
			double et = System.currentTimeMillis() - t;
			tStat.collect(et/1000.0);
			
			stat.collect(erho);
			
		}
		
		System.out.println(stat);
		System.out.println(tStat);
		return(stat.getAverage());
	}
	
	public double estCorrelationAV(){

//		System.out.println("Performing replication: ");
		double xodd = 0.0;
		stat.reset();
		for(int i=1;i<=myReplications;i++){// macro replications
			
//			System.out.print(i + ",");
//			if ( (i%20) == 0) System.out.println();
			
//			long t = System.currentTimeMillis();
			
			double erho = sampleCorrelation();
			
//			double et = System.currentTimeMillis() - t;
//			tStat.collect(et/1000.0);
			
			if ( (i%2) == 0){
				double x = (xodd + erho)/2.0;
				stat.collect(x);
				snd.setAntithetic(false);
			} else {
				xodd = erho;
				snd.setAntithetic(true);
			}
			
		}
		
//		System.out.println(stat);
//		System.out.println(tStat);
		return(stat.getAverage());
	}
	
	public final double sampleCorrelation(){
		statXY.reset();
		for(int j=1;j<=mySampleSize;j++){
			generate();
			statXY.collectXY(x[0], x[1]);
		}
		return(statXY.getCorrelationXY());		
	}
	
	public double[] generate(){
		double z1 = snd.getValue();
		double z2 = myRho*z1 + Math.sqrt(1.0-myRho*myRho)*snd.getValue();
		double u1 = snd.cdf(z1);
		double u2 = snd.cdf(z2);
		x[0] = xd.invCDF(u1);
		x[1] = yd.invCDF(u2);
		return(x);
	}
	
	public final Statistic getCorrelationStatistic(){
		return(stat);
	}
	
	public final int getReplications() {
		return myReplications;
	}

	public final void setReplications(int numReps) {
		myReplications = numReps;
	}

	public final int getSampleSize() {
		return mySampleSize;
	}

	public final void setSampleSize(int sampleSize) {
		mySampleSize = sampleSize;
	}
	
	public double findMatchingRho(double desiredRho, Interval interval){
		myDesiredRho = desiredRho;
		myRootFinder.setInterval(this, interval);
		myRootFinder.evaluate();
		
		if (!myRootFinder.hasConverged())
			System.out.println("Unable to find a matching rho");
		
		double r = myRootFinder.getResult();

		return(r);
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		NORTAEvaluation ne = new NORTAEvaluation(0.8);
		ne.setReplications(10);
	    ne.setSampleSize(10000);
		System.out.println("Running....");
		ne.estCorrelation();
		System.out.println("Avg = " + ne.getCorrelationStatistic().getAverage());
		System.out.println("StdDev = " + ne.getCorrelationStatistic().getStandardDeviation());
/*
		double rhox = -1.0;
		System.out.println("rho = " + rhox);
		ne.setRho(rhox);
		ne.estCorrelationAV();
		System.out.println("Avg = " + ne.getCorrelationStatistic().getAverage());
		System.out.println("StdDev = " + ne.getCorrelationStatistic().getStandardDeviation());
		System.out.println("----");

		double rhoy = 1.0;
		System.out.println("rho = " + rhoy);
		ne.setRho(rhoy);
		ne.estCorrelationAV();
		System.out.println("Avg = " + ne.getCorrelationStatistic().getAverage());
		System.out.println("StdDev = " + ne.getCorrelationStatistic().getStandardDeviation());
		System.out.println("----");

		/*
		for (int i=1;i<=9; i++){
			double rho = 0.9 + i*0.01;
			System.out.println("rho = " + rho);
			ne.setRho(rho);
			ne.estCorrelationAV();
			System.out.println("Avg = " + ne.getCorrelationStatistic().getAverage());
			System.out.println("StdDev = " + ne.getCorrelationStatistic().getStandardDeviation());
			System.out.println("----");
		}
*/
		
//		ne.setDesiredRho(0.8);
//		System.out.println(ne.fx(-1.0));
//		System.out.println(ne.fx(1.0));
		
//		Interval range = new Interval(0.0,1.0);
//		ne.findMatchingRho(0.8, range); 
		System.out.println("Done....");
		
	}
	
	public double fx(double x) {
		setRho(x);
		double f = estCorrelationAV();
		return f - myDesiredRho;
	}

	public final void setDesiredRho(double desiredRho) {
		myDesiredRho = desiredRho;
	}
}
