/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package examples.multitellerbank;

import java.util.ArrayList;
import java.util.List;

import jsl.modeling.Experiment;
import jsl.modeling.JSLEvent;
import jsl.modeling.Model;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.EventGenerator;
import jsl.modeling.elements.EventGeneratorListenerIfc;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.resource.Entity;
import jsl.modeling.elements.resource.Request;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;

public class MultiTellerBank extends SchedulingElement {

	protected int myNumTellers;
	protected List<Teller> myTellers;
	protected EventGenerator myCustomerGenerator;
	protected CustomerArrival myArrivalListener = new CustomerArrival();
	
	protected ResponseVariable myTimeInQ;
	protected TimeWeighted myNumInQ;
	
	public MultiTellerBank(ModelElement parent) {
		this(parent, 4, new Exponential(1.0), new Exponential(4.5), null);
	}

	public MultiTellerBank(ModelElement parent, int numTellers) {
		this(parent, numTellers, new Exponential(1.0), new Exponential(4.5), null);
	}
	public MultiTellerBank(ModelElement parent, int numTellers, DistributionIfc timeBtwArrivals, DistributionIfc serviceTime, String name) {
		super(parent, name);
		
		if (numTellers < 1)
			throw new IllegalArgumentException("Number of tellers was < 1");
		AggregateTimeWeightedVariable s1 = new AggregateTimeWeightedVariable(this, "TW Total in Q");
		ResponseVariableAverageObserver s2 = new ResponseVariableAverageObserver(this, "Avg Time In Q");
		myTellers = new ArrayList<Teller>();
		myNumTellers = numTellers;
		for(int i=1; i<=numTellers; i++){
			Teller t = new Teller(this, this);
			t.setServiceDistributionInitialRandomSource(serviceTime);
			myTellers.add(t);
			t.getQueue().subscribe(s1);
			t.getQueue().subscribe(s2);
		}
		myCustomerGenerator = new EventGenerator(this, myArrivalListener, timeBtwArrivals, timeBtwArrivals);
		myCustomerGenerator.setGenerationEndingTimeForReplications(480.0);
		
		myTimeInQ = new ResponseVariable(this, "Time in Q");
		myNumInQ = new TimeWeighted(this, "Number in Q");
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.Variable#setValue(double)
	 */
	protected void recordTimeInQ(double value) {
		myTimeInQ.setValue(value);
	}
	
	protected void incrementNumInQ(){
		myNumInQ.increment();
	}
	
	protected void decrementNumInQ(){
		myNumInQ.decrement();
	}
	
	protected void departing(Entity entity, Teller teller){
		// a customer is departing, opportunity for jockeying
		
		// get number in queue at the teller the customer is departing from
		int n = teller.getNumInQueue();
//		System.out.println("Departing teller " + myTellers.indexOf(teller) + " num in q = " + n);
		// get the index of current teller
		int i = myTellers.indexOf(teller);
		int min = Integer.MAX_VALUE;
		// find the tellers to jockey from
		Teller fTeller = null;
		for(Teller t: myTellers){
			if (t != teller){
				if (t.getNumInQueue() > n + 1){// teller t has more in queue than teller
					int d = Math.abs(i - myTellers.indexOf(t)); // distance from current teller
					if(d < min){
						fTeller = t;
						min = d;
					}
				}				
			}
		}
		
		if (fTeller != null){ // a teller to jockey from was found
//			System.out.println("Jockeying from teller " + myTellers.indexOf(teller) + " num in q " + fTeller.getNumInQueue());
//			System.out.println("Jockeying from teller " + myTellers.indexOf(teller) + " to " + myTellers.indexOf(fTeller));
			Queue q = fTeller.getQueue(); // get the queue
//			int index = q.size() - 1; // e.g. if size() = 10, there are 10 elements, the last element is at index 9
//			Request r = (Request)q.remove(index); // remove the last request in this queue
			Request r = (Request)q.peekLast(); // get the reference to the last in the queue
			q.remove(r); // remove the object from the queue w/o collecting statistics
			Entity e = r.getEntity(); // get the customer associated with the request
			teller.receive(e); // tell the teller to process this customer
		}
	}

	private Teller selectTeller(){

		// check if there is an idle teller, then use it
		for(Teller t: myTellers){
			if (t.isIdle())
				return(t);
		}
		
		// no idle tellers, then pick teller with shortest queue
		Teller shortest = null;
		int min = Integer.MAX_VALUE;
		for(Teller t: myTellers){
			int  n = t.getNumInQueue();
			if (n < min){
				shortest = t;
				min = n;
			}
		}
		return(shortest);
		
	}
	
	private class CustomerArrival implements EventGeneratorListenerIfc {

		public void generate(EventGenerator generator, JSLEvent event) {
			Teller t = selectTeller();
			
			Entity entity = new Entity();
			incrementNumInQ();		
			t.receive(entity);
			
		}
		
	}
	
	
	public static void main(String[] args) {
		System.out.println("MultiTeller Bank Test");
        
//		JSL.LOGGER.setLevel(Level.FINE);
   
//    	testReplication();
		testExperiment();
        
		System.out.println("Done!");
        
	}
	
	public static void testExperiment(){
		// create the containing model
		Model m = Model.createModel();
		
		// create the model element and attach it to the main model
		MultiTellerBank bank = new MultiTellerBank(m, 4);
		
		m.turnOnLogReport();
		
		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
		// set the parameters of the experiment
		e.setNumberOfReplications(10);
//		e.setLengthOfReplication(480.0);
		
		// turn on the desired reporting
//		e.turnOnBatchReport();
		e.turnOnExperimentReport();
		e.turnOnReplicationReport();
		e.turnOnExperimentConsoleOutput();
		
		// tell the experiment to run
		e.runAll();
	}


	
}
