/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package examples.queueing;

import jsl.modeling.*;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.resource.Request;
import jsl.modeling.elements.resource.Resource;
import jsl.modeling.elements.resource.ResourceSet;
import jsl.modeling.elements.variable.*;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;

public class DriverLicenseBureauWithResourceSet extends SchedulingElement {

	private ResourceSet myResourceSet;
	private Queue myWaitingQ;
	private DistributionIfc myServiceDistribution;
	private DistributionIfc myArrivalDistribution;
	private RandomVariable myServiceRV;
	private RandomVariable myArrivalRV;
	private TimeWeighted myNS;

	private ArrivalListener myArrivalListener;
	private EndServiceListener myEndServiceListener;

	public DriverLicenseBureauWithResourceSet(ModelElement parent) {
		this(parent, 1, new Exponential(1.0), new Exponential(0.5));
	}

	public DriverLicenseBureauWithResourceSet(ModelElement parent, int numServers, DistributionIfc ad, DistributionIfc sd) {
		super(parent);

		setServiceDistributionInitialRandomSource(sd);
		setArrivalDistributionInitialRandomSource(ad);

        myWaitingQ = new Queue(this, "DriverLicenseQ");
		myResourceSet = new ResourceSet(this, "ResourceSet");
		myResourceSet.addResource(numServers, "Clerks");
		myNS = new TimeWeighted(this, 0.0, "NS");

		myArrivalListener = new ArrivalListener();
		myEndServiceListener = new EndServiceListener();
	}

	public void setServiceDistributionInitialRandomSource(DistributionIfc d) {

		if (d == null)
			throw new IllegalArgumentException("Service Time Distribution was null!");

		myServiceDistribution = d;

		if (myServiceRV == null) { // not made yet
			myServiceRV = new RandomVariable(this, myServiceDistribution, "Service RV");
		} else { // already had been made, and added to model
			// just change the distribution
			myServiceRV.setInitialRandomSource(myServiceDistribution);
		}

	}

	public void setArrivalDistributionInitialRandomSource(DistributionIfc d) {

		if (d == null)
			throw new IllegalArgumentException("Arrival Time Distribution was null!");

		myArrivalDistribution = d;

		if (myArrivalRV == null) { // not made yet
			myArrivalRV = new RandomVariable(this, myArrivalDistribution, "Arrival RV");
		} else { // already had been made, and added to model
			// just change the distribution
			myArrivalRV.setInitialRandomSource(myArrivalDistribution);
		}
	}
	
	protected void initialize() {
		super.initialize();

		// start the arrivals
		scheduleEvent(myArrivalListener, myArrivalRV.getValue(), "Arrival");
	}

	class ArrivalListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			myNS.increment(); // new customer arrived
			// create the customer as a request for the resource
			Request arrivingCustomer = Request.create();//wants a single unit
			// enqueue arriving customer
			myWaitingQ.enqueue(arrivingCustomer);	
			// ask the resource set to provide a resource for the request			
			Resource resource = myResourceSet.selectIdleResource(arrivingCustomer);
			// if not null, then an idle resource was returned
			if (resource != null){
				// at least 1 unit is available, check if the arriving customer
				// is the next in the queue
				Request nextCustomer = (Request)myWaitingQ.peekNext();
				if (arrivingCustomer == nextCustomer){
					myWaitingQ.removeNext();
					resource.allocate(arrivingCustomer);
					//	schedule end of service, include the customer as the event's message
					scheduleEvent(myEndServiceListener, myServiceRV.getValue(), "End Service", arrivingCustomer); 	
				}
			}
			//	always schedule the next arrival
			scheduleEvent(myArrivalListener, myArrivalRV.getValue(), "Arrival"); 
		}
	}

	class EndServiceListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			myNS.decrement(); // customer is departing
			Request departingCustomer = (Request) event.getMessage();
			Resource resource = departingCustomer.getResource();
			// tell the request to release the resource
			resource.release(departingCustomer);

			if (!myWaitingQ.isEmpty()) { // queue is not empty
				Request nextCustomer = (Request)myWaitingQ.peekNext(); //remove the next customer
				Resource nextResource = myResourceSet.selectIdleResource(nextCustomer);
				if (nextResource != null){
					myWaitingQ.removeNext();
					nextResource.allocate(nextCustomer);
					//	schedule end of service
					scheduleEvent(myEndServiceListener, myServiceRV.getValue(), "End Service", nextCustomer);
				}
			}
		}
	}

	public static void main(String[] args) {
		System.out.println("Driver License Bureau Test");

		// create the containing model
		Model m = Model.createModel();
		
		// create the model element and attach it to the main model
		new DriverLicenseBureauWithResourceSet(m);
		
		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
		// set the parameters of the experiment
		e.setNumberOfReplications(10);
		e.setLengthOfReplication(20000.0);
		e.setLengthOfWarmUp(5000.0);
		
		// turn on the desired reporting
		e.turnOnExperimentConsoleOutput();
		
		// tell the experiment to run
		e.runAll();    
		System.out.println("Done!");
        
	}
	
}
