/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package examples.queueing;

import jsl.modeling.*;
import jsl.modeling.elements.variable.*;
import jsl.modeling.elements.queue.*;
import jsl.modeling.elements.resource.*;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionFunctionIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;

public class GGcWSModelResourceProvider extends SchedulingElement {

	private ResourceProvider myResourceProvider;
	private DistributionIfc myServiceDistribution;
	private DistributionIfc myArrivalDistribution;
	private RandomVariable myServiceRV;
	private RandomVariable myArrivalRV;
	private RandomVariable myAmtRequestedRV;
	private ResponseVariable myWs;

	private ArrivalListener myArrivalListener;
	private EndServiceListener myEndServiceListener;
	private BeginService myBeginService;
	
	public GGcWSModelResourceProvider(ModelElement parent) {
		this(parent, new Exponential(1.0), new Exponential(0.5));
	}

	public GGcWSModelResourceProvider(ModelElement parent, DistributionIfc ad, DistributionIfc sd) {
		super(parent);
		setServiceDistributionInitialRandomSource(sd);
		setArrivalDistributionInitialRandomSource(ad);
		
//		DistributionIfc d = new DUniform(1, 3);
		DistributionIfc d = new Constant(1);
		
		myAmtRequestedRV = new RandomVariable(this, d);
		
		myResourceProvider = new ResourceProvider(this, "ResourceProvider");
		
		myResourceProvider.addResource(1, "Resource1");
//		myResourceProvider.addResource(2, "Resource2");
//		myResourceProvider.addResource(3, "Resource3");
		
		myWs = new ResponseVariable(this, "WS");

		myArrivalListener = new ArrivalListener();
		myBeginService = new BeginService();
		myEndServiceListener = new EndServiceListener();
	}

	public DistributionFunctionIfc getArrivalDistribution() {
		return (myArrivalDistribution);
	}

	public DistributionFunctionIfc getServiceDistribution() {
		return (myServiceDistribution);
	}

	public RandomVariable getArrivalRV() {
		return (myArrivalRV);
	}

	public RandomVariable getServiceRV() {
		return (myServiceRV);
	}

	public void setServiceDistributionInitialRandomSource(DistributionIfc d) {

		if (d == null)
			throw new IllegalArgumentException("Service Time Distribution was null!");

		myServiceDistribution = d;

		if (myServiceRV == null) { // not made yet
			myServiceRV = new RandomVariable(this, myServiceDistribution, "Service RV");
		} else { // already had been made, and added to model
			// just change the distribution
			myServiceRV.setInitialRandomSource(myServiceDistribution);
		}

	}

	public void setArrivalDistributionInitialRandomSource(DistributionIfc d) {

		if (d == null)
			throw new IllegalArgumentException("Arrival Time Distribution was null!");

		myArrivalDistribution = d;

		if (myArrivalRV == null) { // not made yet
			myArrivalRV = new RandomVariable(this, myArrivalDistribution, "Arrival RV");
		} else { // already had been made, and added to model
			// just change the distribution
			myArrivalRV.setInitialRandomSource(myArrivalDistribution);
		}
	}

	protected void initialize() {
		super.initialize();

		// start the arrivals
		scheduleArrival();
	}

	private void scheduleArrival() {
		scheduleEvent(myArrivalListener, myArrivalRV.getValue(), "Arrival");
	}

	private void scheduleEndService(QObject c) {
		scheduleEvent(myEndServiceListener, myServiceRV.getValue(), "End Service", c);
	}

	private class ArrivalListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			int amt = (int)myAmtRequestedRV.getValue();
			myResourceProvider.seize(amt, myBeginService);

			scheduleArrival();

		}
	}
	
	private class BeginService implements ResourceAllocationListenerIfc {

		public void allocated(Request request, Resource resource) {
			scheduleEndService(request);			
		}	
	}
	
	private class EndServiceListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			Request request = (Request) event.getMessage();
			double ws = getTime() - request.getTimeEnteredQueue();
			myWs.setValue(ws);

			Resource resource = request.getResource();
			// tell the request to release the resource
			resource.release(request);

		}
	}

}
