/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/


package examples.queuejockeying;

import java.util.Observer;

import jsl.modeling.ActionListenerIfc;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.resource.*;
import jsl.modeling.elements.queue.*;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;

/**
 */
public class Station extends SchedulingElement {

	protected Queue myQueue;
	protected Resource myResource;
	protected ResponseVariable mySystemTime;
	protected EndServiceListener myEndServiceListener;
	protected DistributionIfc myServiceDistribution;
	protected RandomVariable myServiceRV;
	protected FastFoodRestaurant myFastFoodRestaurant;
	
	public Station(ModelElement parent, FastFoodRestaurant restaurant) {
		this(parent, restaurant, null);
	}
	
	public Station(ModelElement parent, FastFoodRestaurant restaurant, String name) {
		super(parent, name);
		setServiceDistributionInitialRandomSource(Constant.ONE);
		myFastFoodRestaurant = restaurant;
		myQueue = new Queue(this, getName() + " Queue");
		myResource = new Resource(this, 1, getName() + " Resource");
		mySystemTime = new ResponseVariable(this, getName() + " System Time");
		myEndServiceListener = new EndServiceListener();
	}

	public final int getNumInQueue(){
		return(myQueue.size());
	}
	
	public final boolean isIdle(){
		return(myResource.isIdle());
	}
	
	protected final Request removeLastCustomer(){
		Request r = (Request)myQueue.peekLast();
		myQueue.remove(r, false);
		return(r);
	}
	
	protected final void attachNumberInQueueObserver(Observer observer){
		myQueue.addNumberInQueueObserver(observer);
	}
	
	protected final void attachTimeInQueueObserver(Observer observer){
		myQueue.addTimeInQueueObserver(observer);
	}

	public void setServiceDistributionInitialRandomSource(DistributionIfc d) {

		if (d == null)
			throw new IllegalArgumentException("Service Time Distribution was null!");

		myServiceDistribution = d;

		if (myServiceRV == null) { // not made yet
			myServiceRV = new RandomVariable(this, myServiceDistribution);
		} else { // already had been made, and added to model
			// just change the distribution
			myServiceRV.setInitialRandomSource(myServiceDistribution);
		}

	}
	
	protected void receive(Request customer) {
		if (customer == null)		
			throw new IllegalArgumentException("The custoemr must be non-null!");
		// a customer is arriving to the station, set the time requested
		customer.setTimeRequested(myServiceRV.getValue());
		// enqueue arriving customer
		myQueue.enqueue(customer);
		// check if arriving customer gets the resource
		if (myResource.isIdle()){
			Request r = (Request)myQueue.peekNext();
			if (r == customer){
				if (r.getInitialAmountRequested() <= myResource.getNumberIdle()){
					myQueue.removeNext();
					myResource.allocate(r);
					scheduleEvent(myEndServiceListener, r.getTimeRequested(), getName() + " End Service", r);
				}				
			}
		}
	}
	
	class EndServiceListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			Request departingCustomer = (Request) event.getMessage();
			// get the time 
			double ws = getTime() - departingCustomer.getTimeEnteredQueue();
			mySystemTime.setValue(ws);
			// tell the resource to release the request
			myResource.release(departingCustomer);
			if (!myQueue.isEmpty()) {
				Request r = (Request)myQueue.peekNext();
				if (r.getInitialAmountRequested() <= myResource.getNumberIdle()){
					myQueue.removeNext();
					myResource.allocate(r);
					scheduleEvent(myEndServiceListener, r.getTimeRequested(), getName() + " End Service", r);	
				}				
			}
			// customer is departing a station, check for jockey opportunity
			myFastFoodRestaurant.checkForJockey(Station.this);
		}
	}

}
