/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import jsl.modeling.conditions.*;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.observers.ExperimentConsoleObserver;
import jsl.observers.ReplicationConsoleObserver;
import jsl.observers.textfile.*;
import jsl.observers.variable.StatisticalObserver;
import jsl.utilities.reporting.JSL;
import jsl.utilities.statistic.AbstractStatistic;
import jsl.utilities.statistic.Statistic;

/** This class provides the information for running a simulation experiment.
 *  An experiment is a specification for the number of replications, the 
 *  warm up length, the ending condition (time or condition based), etc for 
 *  controlling the running of a replication.  
 * 
 *  The defaults include:
 *   - length of replication = Double.POSITIVE_INFINITY
 *   - length of warm up = 0.0 // no warm up
 *   - length of batch interval = 0.0 // no batch interval
 *   - replication initialization TRUE -> The system state is re-initialized prior
 *       to each replication
 *   - default batching option FALSE -> Do not batch replications within an experiment
 *   - save replications option FALSE -> Do not save the instances of each replication 
 *   - reset start stream option FALSE -> Do not reset the streams of the random variables
 *       to their starting points prior to running the replications within the experiment. 
 *       This implies that if the experiment is re-run on the same model in the same code invocation
 *       that an independent set of replications will be made.
 *   - reset next substream option TRUE -> The random variables in a replications within an experiment
 *       will start at the next sub-stream for each new replication
 * 
 */
public class Experiment extends IterativeProcess<Replication> {
 
	//  ===========================================
	//      OBJECT ATTRIBUTES
	//  ===========================================

	/** A reference to the model for this experiment
	 */
	private Model myModel;

	/** A reference to the scheduler
	 */
	private Scheduler myScheduler;

	/** The number of replications to run for this experiment
	 * 
	 */
	private int myNumReps;

	/** The current number of replications that have been run for this experiment
	 */
	private double myCurRepNum;

	/** A reference to the current replication being run by this
	 *  experiment.
	 */
	private Replication myReplication;

	/** The specified length of each planned replication for
	 *  this experiment.  The default is Double.POSITIVE_INFINITY.
	 */
	private double myLengthOfReplication = Double.POSITIVE_INFINITY;

	/** The length of time from the start of an individual replication to
	 * the warm up event for that replication.
	 */
	private double myLengthOfWarmUp = 0.0; // zero is no warmup

	/** The condition that can be checked to see if a replication
	 *  within the experiment is completed.
	 */
	private ConditionIfc myEndReplicationCondition;

	/** A flag to indicate whether or not each replication within
	 *  the experiment should be re-initialized at the beginning of
	 *  each replication. True means that it will be re-initialized.
	 */
	private boolean myRepInitOption;

	/** A flag to indicate whether or not the experiment will
	 *  have the model be batched for each replication
	 */
	private boolean myDefaultBatchingOption;

	/** A flag to detect whether user set default batching option
	 */
	private boolean myDefaultBatchingOptionFlag;

	/** A flag to indicate that the batching report should
	 * be on for replications
	 */
	 private boolean myBatchReportFlag;

	/** A time interval (in simulated time) that represents
	 *  the default time between batches
	 */
	private double myDefaultBatchInterval;

	/** Used to report across replication statistics
	 */
	private ExperimentReport myExpReport;
	
	/** Used to indicate whether the experiment report is on
	 */
	private boolean myExpReportOption;

	/** Used to report antithetic statistics
	 */
	private AntitheticExperimentReport myAVExpReport;
	
	/** Used to indicate whether the antithetic experiment report is on
	 */
	private boolean myAVExpReportOption;

	/** Used to report with replication statistics
	 */
	private ReplicationReport myRepReport;
	
	/** Used to indicate whether replication reporting is on
	 */
	private boolean myRepReportOption;

	/** Used to output replication information to the console
	 */
	private ReplicationConsoleObserver myRepConsoleObserver;
	
	/** Used to indicate whether replication console reporting is on
	 */
	private boolean myRepConsoleOption;

	/** Used to output experiment information to the console
	 */
	private ExperimentConsoleObserver myExpConsoleObserver;
	
	/** Used to indicate whether experiment console reporting is on
	 */
	private boolean myExpConsoleOption;

	/** The maximum allowable execution time "wall" clock time
	*  for an individual replication to complete processing in milliseconds
	*/
	protected long myMaxAllowedExecutionTimePR;

	/** The reset start stream option
	 *  This option indicates whether or not the random variables used
	 *  during the experiment will be reset to their starting stream
	 *  prior to running the first replication.  The default is FALSE.  This
	 *  ensures that the random variable's streams WILL NOT be reset prior
	 *  to running the experiment.  This will cause
	 *  different experiments or the same experiment run multiple times
	 *  that use the same random variables (via the same model) to continue
	 *  within their current stream.  Therefore the experiments will
	 *  be independent when invoked within the same program execution. To get
	 *  common random number (CRN), run the experiments in different program
	 *  executions OR set this option to true prior to running the experiment again
	 *  within the same program invocation.
	 */
	private boolean myResetStartStreamOption;

	/** The reset next sub stream option
	 *  This option indicates whether or not the random variables used
	 *  during the replication within the experiment will be reset to their next substream
	 *  after running each replication.  The default is TRUE.  This
	 *  ensures that the random variables will jump to the next
	 *  substream within their current stream at the end of a replication.  This will cause
	 *  the random variables in each subsequent replication to start in the same substream
	 *  in the underlying random number streams if the replication is repeatedly used and
	 *  the ResetStartStreamOption is set to false (which is the default). Otherwise, this
	 *  option really has no effect if there is only 1 replication in an experiment.
	 *  and then jump to the next substream (if this option is on).  
	 *  Having ResetNextSubStreamOption true assists in sychronizing the
	 *  random number draws from one replication to another aiding in
	 *  the implementation of common random numbers.  Each replication
	 *  within the same experiment is still independent.
	 */
	private boolean myResetNextSubStreamOption;

	/** Used to indicate whether monte carlo option
	 *  is on for the replications in the experiment
	 */
	private boolean myMonteCarloOption;
	
	/** Indicates whether or not replication statistics will be saved
	 *  by the replication.  This also forces the replication save step
	 *  option to be turned on. In other words, replication statistics
	 *  cannot be saved unless the replications are also saved. The 
	 *  default is false (no saving).
	 */
	private boolean mySaveRepStatOption;

	/** Used to indicate whether or not statistics related to the experiment
	 *  (across replications) should be saved.  The default is false
	 */
	private boolean mySaveStatisticsOption;
	
	/** Used to hold statistics associated with the experiment if the
	 *  save statistics option is true.
	 */
	private List<AbstractStatistic> myStatistics;
	
	/** Indicates whether or not antithetic replications should
	 *  be run.  The default is false.  If set (via the constructor)
	 *  the user must supply an even number of replications; otherwise
	 *  an exception will be thrown
	 */
	private boolean myAntitheticOption = false;
	
	/** Represents the design for this experiment
	 */
	private DesignIfc myDesign;

	//  ===========================================
	//        CONSTRUCTORS
	//  ===========================================

	/** Constructs an experiment and constructs a default model and scheduler
	 *  to be used during the experiment.  The getModel() method can be used to
	 *  access the model prior to running the experiment.
	 */
	public Experiment() {
		this(Model.createModel(), null, false, null);
	}

	/** Constructs an experiment and constructs a default model and scheduler
	 *  to be used during the experiment.  The getModel() method can be used to
	 *  access the model prior to running the experiment.
	 * @param antitheticOption Indicates whether or not antithetic replications are to be used
	 */
	public Experiment(boolean antitheticOption) {
		this(Model.createModel(), null, antitheticOption, null);
	}
	
	/** Constructs an experiment and constructs a default model and scheduler
	 *  to be used during the experiment.  The getModel() method can be used to
	 *  access the model prior to running the experiment.
	 * @param name The name of the experiment
	 */
	public Experiment(String name) {
		this(Model.createModel(), name, false, null);
	}

	/** Constructs an experiment and constructs a default model and scheduler
	 *  to be used during the experiment.  The getModel() method can be used to
	 *  access the model prior to running the experiment.
	 * @param name The name of the experiment
	 * @param antitheticOption Indicates whether or not antithetic replications are to be used
	 */
	public Experiment(String name, boolean antitheticOption) {
		this(Model.createModel(), name, antitheticOption, null);
	}
	
	/** Constructs an experiment given the model and uses a default scheduler
	 *  to be used during the experiment
	 * @param model A reference to the model for this experiment
	 */
	public Experiment(Model model) {
		this(model, null, false, null);
	}

	/** Constructs an experiment given the model and uses a default scheduler
	 *  to be used during the experiment
	 * @param model A reference to the model for this experiment
	 * @param antitheticOption Indicates whether or not antithetic replications are to be used
	 */
	public Experiment(Model model, boolean antitheticOption) {
		this(model, null, antitheticOption, null);
	}
	
	/** Constructs an experiment called name for the given
	 *  model and supplies a default scheduler
     * @param model A reference to the model for this experiment
     * @param name
	 */
	public Experiment(Model model, String name) {
		this(model, name, false, null);
	}

	/** Constructs an experiment called name for the given
	 *  model and supplies a default scheduler
	 * @param model A reference to the model for this experiment
     * @param name
     * @param antitheticOption Indicates whether or not antithetic replications are to be used
	 */
	public Experiment(Model model, String name, boolean antitheticOption) {
		this(model, name, antitheticOption, null);
	}
	
	/** Constructs an experiment called "name" with the given model and scheduler
	 * @param name The name of the experiment
	 * @param model A reference to the model for this experiment
	 * @param antitheticOption Indicates whether or not antithetic replications are to be used
	 * @param scheduler A reference to the scheduler for this replication
	 */
	public Experiment(Model model, String name, boolean antitheticOption, Scheduler scheduler) {
		super(name);
		setModel(model);
		myAntitheticOption = antitheticOption;
		setScheduler(scheduler);
		myNumReps = 1;
		myLengthOfReplication = Double.POSITIVE_INFINITY;
		myLengthOfWarmUp = 0.0; // zero means no warm up
		myDefaultBatchInterval = 0.0; // zero means no batch interval
		myRepInitOption = true;
		myMaxAllowedExecutionTimePR = 0;
		myDefaultBatchingOption = false;
		myDefaultBatchingOptionFlag = false;
		myBatchReportFlag = false;
		myResetStartStreamOption = false;
		myResetNextSubStreamOption = true;
		myMonteCarloOption = false;
		mySaveRepStatOption = false;
		setSaveStepOption(false);
		setDesign(new Design());
	}

	//  ===========================================
	//        PUBLIC METHODS
	//  ===========================================

	/** Gets the model for this experiment
	 * @return the model
	 */
	public Model getModel() {
		return (myModel);
	}

	/** Returns the number of replications to run
	 * @return the number as a double
	 */
	public final double getNumberOfReplications() {
		return (myNumReps);
	}

	/** Sets the desired number of replications for this experiment
	 *  If the number of replications is set to less than the current number of replications
	 *  during the experiment then this will be found at the end of the current replication
	 *  and cause the experiment to be ended.
	 * @param numReps The number of replications, must be > 0
	 */
	public final void setNumberOfReplications(int numReps) {
		if (numReps <= 0)
			throw new IllegalArgumentException("Number of replications <= 0");

		if (myAntitheticOption == true){
			if ((numReps % 2) != 0 )
				throw new IllegalArgumentException("Number of replications must be even if antithetic option is on.");
		}

		myNumReps = numReps;
		
	}

	/** Returns the current number of replications completed
	 * @return the number as a double
	 */
	public final double getCurrentReplicationNumber() {
		return (myCurRepNum);
	}

	/** Returns the last replication completed
	 * @return the replication
	 */
	public final Replication getCurrentReplication() {
		return (myReplication);
	}

	/** Returns the length of the replication as a double
	 * @return the length of the replication
	 */
	public final double getLengthOfReplication() {
		return (myLengthOfReplication);
	}

	/** Sets the length of each replication for this experiment.
	 *  The length of the replication must be > 0, if set
	 * 
	 * @param lengthOfReplication
	 */
	public final void setLengthOfReplication(double lengthOfReplication) {
		if (lengthOfReplication <= 0.0)
			throw new IllegalArgumentException("Simulation replication length < 0.0");
		myLengthOfReplication = lengthOfReplication;
	}

	/** Gets the length of the warm up for each replication with this experiment
	 * @return the the lenght of the warm up
	 */
	public final double getLengthOfWarmUp() {
		return (myLengthOfWarmUp);
	}

	/** Sets the length of the warm up for each replication within this experiment
	 *  The length of the warm up must be > 0.0, if set.
	 *  If the warm up is set greater than the run length then there will be no warm up
	 *
	 * @param lengthOfWarmUp length of the warm up, must be >= 0.0
	 */
	public final void setLengthOfWarmUp(double lengthOfWarmUp) {
		if (lengthOfWarmUp < 0.0)
			throw new IllegalArgumentException("Warmup time cannot be less than zero");

		myLengthOfWarmUp = lengthOfWarmUp;
	}

	/** Returns the setting for whether or not each replication will be reinitialized
	 *  prior to running.
	 * @return true means that each replication will be initialized
	 */
	public final boolean getReplicationInitializationOption() {
		return (myRepInitOption);
	}

	/** Sets the replication initialization option for the experiment
	 * If set to true, each replication within the experiment will
	 * be initialized prior to running.
	 * @param repInitOption true means to initialize
	 */
	public final void setReplicationInitializationOption(boolean repInitOption) {
		myRepInitOption = repInitOption;
	}
	
	/** Gets the monte carlo option flag
	 *  if this is true, then the model's montecarlo method
	 *  will be called after, beforeReplication
	 * @return
	 */
	public final boolean getMonteCarloOption() {
		return myMonteCarloOption;
	}

	/** Sets the monte carlo option flag
	 *  if this is true, then the model's montecarlo method
	 *  will be called after, beforeReplication
	 * @param b
	 */
	public final void setMonteCarloOption(boolean b) {
		myMonteCarloOption = b;
	}
	
	/** Returns the setting for whether each replication will be batched for
	 *  the experiment. 
	 *  If there is only one replication, then it will be assumed to be batched.
	 *  If there will be more than one replication, then each replication
	 *  will not be batched unless overridden.
	 * @return true means that each replication will be batched
	 */
	public final boolean getDefaultBatchingOption() {
		return (myDefaultBatchingOption);
	}

	/** Sets the default batching option for each replication. This controls
	 *  whether or not a standard sized batching interval will be used to 
	 *  collect statistics on time persistent variables during a replication.
	 * 
	 *  If there is only one replication, then the replication will be assumed to be batched.
	 *  If there will be more than one replication, then each replication
	 *  will not be batched unless overridden by the use of this method.
	 * 
	 *  This method only indicates how individual time weighted variables that
	 *  do not control their own batching will be controlled by default during
	 *  each replication within the experiment.
	 * 
	 * @param batchOption true means batching is on.
	 */
	public final void setDefaultBatchingOption(boolean batchOption) {
		myDefaultBatchingOption = batchOption;
		myDefaultBatchingOptionFlag = true;
	}

	/** Returns the default batching interval for all time weighted variables
	 *  when the experiment controls the interval setting
	 * during each replication.  Individual time weighted variables may elect
	 *  to have their own batch interval.
	 * @return a double representing the interval length
	 */
	public final double getDefaultBatchInterval() {
		return (myDefaultBatchInterval);
	}

	/** Sets the default batching  interval for the experiment
	 * If the default batching option is on, this value will be used
	 * for the batching interval for all time weighted variables that participate
	 * in batching. If the batching interval is set greater than the run length
	 * then no batching will occur.
	 * @param batchInterval a double > 0
	 */
	public final void setDefaultBatchInterval(double batchInterval) {
		if (batchInterval <= 0.0)
			throw new IllegalArgumentException("Batching interval must be > 0");
		myDefaultBatchInterval = batchInterval;
	}

	/** Returns the condition that will cause an individual replication to stop
	 * @return the condition as a ConditionIfc
	 */
	public final ConditionIfc getEndReplicationCondition() {
		return (myEndReplicationCondition);
	}

	/** Sets the condition for ending the execution of replications within the experiment.
	 *  This condition is checked prior to executing an individual event. If the condition is true,
	 *  the event processing terminates and the replication ends.
	 *  The condition is automatically cleared prior to executing any events at the beginning of a replication.
	 *
	 *  Throws an IllegalArgumentException if the supplied condition is null.
	 *  @param condition An object that implements the Condition interface
	 */
	public final void setEndReplicationCondition(ConditionIfc condition) {
		if (condition == null)
			throw new IllegalArgumentException("The Condition must be non-null.");
		myEndReplicationCondition = condition;
	}

	/** Gets the value of the save replication statistics option. True
	 *  indicates that replications and their statistics will be saved
	 *  in a list
	 *  
	 * @return Returns the option.
	 */
	public final boolean getSaveReplicationStatisticOption() {
		return mySaveRepStatOption;
	}

	/** Sets the value of the save replication statistics option. True
	 *  indicates that replications and their statistics will be saved
	 *  in a list
	 *  
	 * @param option The option to set.
	 */
	public final void setSaveReplicationStatisticOption(boolean option) {
		mySaveRepStatOption = option; // turn on the option
		setSaveStepOption(option);
	}

	/** Turns on the experiment report. This report yields a text file of
	 *  the across replication statistics
	 */
	public final void turnOnExperimentReport(){
		turnOnExperimentReport(null);
	}
	
	/** Turns on the experiment report. This report yields a text file of
	 *  the across replication statistics
	 *  @param name specifies a prefix name for the report 
	 */
	public final void turnOnExperimentReport(String name){
		myExpReportOption = true;
        
		if(myExpReport == null){ // has not been made, so make it
			try{
				String s;
				if (name != null)
					s = name + " " + getName() + " " + getModel().getName() + " Report";
				else
					s = getName() + " " + getModel().getName() + " Report";
				myExpReport = new ExperimentReport(s);
				addObserver(myExpReport);
			}
			catch(IOException e){
				String s = "I/O exception opening ExperimentReport for experiment: ID = " + getId() + " Name: " + getName();
				System.err.println(s);
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.SEVERE))
					JSL.LOGGER.severe(s);
				System.exit(1);
			}
		} else {
			// the report has already been made and named
			// the user may be trying to change the name of the report
			// from the name based on the input
			String s;
			if (name != null)
				s = name + " " + getName() + " " + getModel().getName() + " Report";
			else
				s = getName() + " " + getModel().getName() + " Report";
			// check if the new name and old name are the same
			if (s.equals(myExpReport.getName())){
				JSL.LOGGER.warning("The Experiment report " + s + " will be over written." );
			} 
			// remove the old observer and attach new observer
			deleteObserver(myExpReport);
			try {
				myExpReport = new ExperimentReport(s);
				addObserver(myExpReport);
			} catch (IOException e){
				String s1 = "I/O exception opening ExperimentReport for experiment: ID = " + getId() + " Name: " + getName();
				System.err.println(s1);
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.SEVERE))
					JSL.LOGGER.severe(s1);
				System.exit(1);
			}	
		}
	}
 
	/** Returns a reference to the default experiment report if it exists
	 * 
	 * @return
	 */
	public final ExperimentReport getDefaultExperimentReport(){
		return(myExpReport);
	}
	
	/** Turns off the experiment report
	 */
	public final void turnOffExperimentReport(){
		if (myExpReportOption == true){
			deleteObserver(myExpReport);
			myExpReportOption = false;
			myExpReport = null;
		}
	}


	/** Turns on the experiment report. This report yields a text file of
	 *  the across replication statistics
	 */
	public final void turnOnAntitheticExperimentReport(){
		turnOnAntitheticExperimentReport(null);
	}
	
	/** Turns on the antithetic experiment report. This report yields a text file of
	 *  the antithetic statistics.  This automatically turns on the antithetic option
	 *  If the number of specified replication is not divisible by 2, then an
	 *  exception will be thrown.
	 *  
	 *  @param name specifies a prefix name for the report 
	 */
	public final void turnOnAntitheticExperimentReport(String name){
		if ((myNumReps % 2) != 0 )
			throw new IllegalArgumentException("Number of replications must be even if antithetic option is on.");

		myAntitheticOption = true;
		
		myAVExpReportOption = true;
        
		if(myAVExpReport == null){ // has not been made, so make it
			try{
				String s;
				if (name != null)
					s = name + " " + getName() + " " + getModel().getName() + " Antithetic Report";
				else
					s = getName() + " " + getModel().getName() + " Antithetic Report";
				myAVExpReport = new AntitheticExperimentReport(s);
				addObserver(myAVExpReport);
			}
			catch(IOException e){
				String s = "I/O exception opening Antithetic Experiment Report for experiment: ID = " + getId() + " Name: " + getName();
				System.err.println(s);
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.SEVERE))
					JSL.LOGGER.severe(s);
				System.exit(1);
			}
		} else {
			// the report has already been made and named
			// the user may be trying to change the name of the report
			// form the name based on the input
			String s;
			if (name != null)
				s = name + " " + getName() + " " + getModel().getName() + " Antithetic Report";
			else
				s = getName() + " " + getModel().getName() + " Antithetic Report";
			// check if the new name and old name are the same
			if (s.equals(myAVExpReport.getName())){
				JSL.LOGGER.warning("The Antithetic Experiment report " + s + " will be over written." );
			} 
			// remove the old observer and attach new observer
			deleteObserver(myAVExpReport);
			try {
				myAVExpReport = new AntitheticExperimentReport(s);
				addObserver(myAVExpReport);
			} catch (IOException e){
				String s1 = "I/O exception opening Antithetic Experiment Report for experiment: ID = " + getId() + " Name: " + getName();
				System.err.println(s1);
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.SEVERE))
					JSL.LOGGER.severe(s1);
				System.exit(1);
			}	
		}
	}

	/** Returns a reference to the default experiment report if it exists
	 * 
	 * @return
	 */
	public final AntitheticExperimentReport getAntitheticExperimentReport(){
		return(myAVExpReport);
	}
	
	/** Turns off the experiment report
	 */
	public final void turnOffAntitheticExperimentReport(){
		if (myAVExpReportOption == true){
			deleteObserver(myAVExpReport);
			myAVExpReportOption = false;
			myAVExpReport = null;
		}
	}
	
	/** Turns on the experiment console output. This output sends experiment information and
	 *  the across replication statistics to the console.  No update on the replication progress is printed.
	 */
	public final void turnOnExperimentConsoleOutput(){
		turnOnExperimentConsoleOutput(false);
	}
	
	/** Turns on the experiment console output. This output sends experiment information and
	 *  the across replication statistics to the console
	 *  
	 *  @param flag Indicates whether or not updates concerning the replication are printed, false means no replication update
	 */
	public final void turnOnExperimentConsoleOutput(boolean flag){
		myExpConsoleOption = true;
        
		if(myExpConsoleObserver == null){ // has not been made, so make it
			myExpConsoleObserver = new ExperimentConsoleObserver(flag);
		}
		addObserver(myExpConsoleObserver);

	}
    
	/** Turns off the experiment console output
	 */
	public final void turnOffExperimentConsoleOutput(){
		if (myExpConsoleOption == true){
			deleteObserver(myExpConsoleObserver);
			myExpConsoleOption = false;
		}
	}
 
	/** Turns on the replication console output. This output sends replication information and
	 *  to the console
	 */
	public final void turnOnReplicationConsoleOutput(){
		myRepConsoleOption = true;
        
		if(myRepConsoleObserver == null){ // has not been made, so make it
			myRepConsoleObserver = new ReplicationConsoleObserver();
		}
		addObserver(myRepConsoleObserver);

	}
    
	/** Turns off the replication console output
	 */
	public final void turnOffReplicationConsoleOutput(){
		if (myRepConsoleOption == true){
			deleteObserver(myRepConsoleObserver);
			myRepConsoleOption = false;
		}
	}

	/** Turns on the replication report. This report yields a text file of
	 *  the within replication statistics
	 */
	public final void turnOnReplicationReport(){
		turnOnReplicationReport(null);
	}

	/** Turns on the replication report. This report yields a text file of
	 *  the within replication statistics
     *
     * @param name
     */
	public final void turnOnReplicationReport(String name){
		myRepReportOption = true;
        
		if(myRepReport == null){ // has not been made, so make it
			try{
				if (name != null)
					myRepReport = new ReplicationReport(name + " " + getName()+ " " + getModel().getName() + " ReplicationReport");
				else
					myRepReport = new ReplicationReport(getName()+ " " + getModel().getName() + " ReplicationReport");					
				addObserver(myRepReport);
			}
			catch(IOException e){
				System.out.println("I/O exception opening " + myRepReport.getName());
				System.exit(1);
			}
		} else {
			// the report has already been made and named
			// the user may be trying to change the name of the report
			// form the name based on the input
			String s;
			if (name != null)
				s = name + " " + getName() + " " + getModel().getName() + " ReplicationReport";
			else
				s = getName() + " " + getModel().getName() + " ReplicationReport";
			// check if the new name and old name are the same
			if (s.equals(myRepReport.getName())){
				JSL.LOGGER.warning("The Replication report " + s + " will be over written." );
			}
			// remove the old observer and attach new observer
			deleteObserver(myRepReport);
			try {
				myRepReport = new ReplicationReport(s);
				addObserver(myRepReport);
			} catch (IOException e){
				String s1 = "I/O exception opening ReplicationReport for experiment: ID = " + getId() + " Name: " + getName();
				System.err.println(s1);
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.SEVERE))
					JSL.LOGGER.severe(s1);
				System.exit(1);
			}	
		}

	}
    
	/** Returns a reference to the default replication report if it exists
	 * 
	 * @return
	 */
	public final ReplicationReport getDefaultReplicationReport(){
		return(myRepReport);
	}
	
	/** Turns off the replication report
	*/
	public final void turnOffReplicationReport(){
		if (myRepReportOption == true){
			deleteObserver(myRepReport);
			myRepReportOption = false;
		}
	}

	/** Turns on the batch report. By default this report is off
	 *  for each replication.  Turning this on tells each replication
	 *  to turn on their default batch report.
	 *  This will have no effect unless the batching
	 *  option is also true
	*/
	public final void turnOnBatchReport(){
		myBatchReportFlag = true;
	}
	
	/** Turns off the batch report.
	*/
	public final void turnOffBatchReport(){
		myBatchReportFlag = false;
	}
	
	/** This method tells the scheduler to turn on its
	 *  default trace report
	 */
	public final void turnOnDefaultTraceReport() {
		myScheduler.turnOnDefaultTraceReport();
	}

	/** This method tells the scheduler to turn off its default trace report
	 */
	public void turnOffDefaultTraceReport() {
		myScheduler.turnOffDefaultTraceReport();
	}

	/** Returns a reference to the underlying event scheduler
	 * 
	 * @return
	 */
	public final Scheduler getScheduler(){
		return(myScheduler);
	}
	
	/** Returns maximum (real) clock time allocated for the iterative process
	 * @return the number as long representing milliseconds
	 */
	public final long getMaximumAllowedExecutionTimePerReplication() {
		return (myMaxAllowedExecutionTimePR);
	}

	/** Set the maximum allotted (suggested) execution (real) clock for the 
	 *  entire iterative process. This is suggested because the execution time
	 *  requirement is only checked after the completion of an individual step
	 *  After it is discovered that cumulative time for executing the 
	 *  step has exceeded the maximum time, then the iterative process will
	 *  be ended (perhaps) not completing other steps.
	 * 
	 * @param milliseconds
	 */
	public final void setMaximumExecutionTimePerReplication(long milliseconds) {
		if (milliseconds <= 0.0)
			throw new IllegalArgumentException("The maximum number of execution time (clock time) must be > 0.0");
		myMaxAllowedExecutionTimePR = milliseconds;
	}

	/** Gets the reset next substream option. The reset next sub stream option
	 *  This option indicates whether or not the random variables used
	 *  during the replication within the experiment will be reset to their next substream
	 *  after running each replication.  The default is TRUE.  This
	 *  ensures that the random variables will jump to the next
	 *  substream within their current stream at the end of a replication.  This will cause
	 *  the random variables in each subsequent replication to start in the same substream
	 *  in the underlying random number streams if the replication is repeatedly used and
	 *  the ResetStartStreamOption is set to false (which is the default). Otherwise, this
	 *  option really has no effect if there is only 1 replication in an experiment.
	 *  and then jump to the next substream (if this option is on).  
	 *  Having ResetNextSubStreamOption true assists in synchronizing the
	 *  random number draws from one replication to another aiding in
	 *  the implementation of common random numbers.  Each replication
	 *  within the same experiment is still independent.
	 *  
	 * @return true means the option is on
	 */
	public final boolean getResetNextSubStreamOption() {
		return myResetNextSubStreamOption;
	}
 
	/** Sets the reset next substream option. The reset next sub stream option
	 *  This option indicates whether or not the random variables used
	 *  during the replication within the experiment will be reset to their next substream
	 *  after running each replication.  The default is TRUE.  This
	 *  ensures that the random variables will jump to the next
	 *  substream within their current stream at the end of a replication.  This will cause
	 *  the random variables in each subsequent replication to start in the same substream
	 *  in the underlying random number streams if the replication is repeatedly used and
	 *  the ResetStartStreamOption is set to false (which is the default). Otherwise, this
	 *  option really has no effect if there is only 1 replication in an experiment.
	 *  and then jump to the next substream (if this option is on).  
	 *  Having ResetNextSubStreamOption true assists in synchronizing the
	 *  random number draws from one replication to another aiding in
	 *  the implementation of common random numbers.  Each replication
	 *  within the same experiment is still independent.
	 *  
	 * @param b
	 */
	public final void setResetNextSubStreamOption(boolean b) {
		myResetNextSubStreamOption = b;
	}
	
	/** Gets the reset start stream option. The reset start stream option
	 *  This option indicates whether or not the random variables used
	 *  during the experiment will be reset to their starting stream
	 *  prior to running the first replication.  The default is FALSE.  This
	 *  ensures that the random variable's streams WILL NOT be reset prior
	 *  to running the experiment.  This will cause
	 *  different experiments or the same experiment run multiple times
	 *  that use the same random variables (via the same model) to continue
	 *  within their current stream.  Therefore the experiments will
	 *  be independent when invoked within the same program execution. To get
	 *  common random number (CRN), run the experiments in different program
	 *  executions OR set this option to true prior to running the experiment again
	 *  within the same program invocation.
	 * @return true means the option is on
	 */
	public final boolean getResetStartStreamOption() {
		return myResetStartStreamOption;
	}

	/** Sets the reset start stream option.  The reset start stream option
	 *  This option indicates whether or not the random variables used
	 *  during the experiment will be reset to their starting stream
	 *  prior to running the first replication.  The default is FALSE.  This
	 *  ensures that the random variable's streams WILL NOT be reset prior
	 *  to running the experiment.  This will cause
	 *  different experiments or the same experiment run multiple times
	 *  that use the same random variables (via the same model) to continue
	 *  within their current stream.  Therefore the experiments will
	 *  be independent when invoked within the same program execution. To get
	 *  common random number (CRN), run the experiments in different program
	 *  executions OR set this option to true prior to running the experiment again
	 *  within the same program invocation.
	 * @param b
	 */
	public final void setResetStartStreamOption(boolean b) {
		myResetStartStreamOption = b;
	}
	
	/** True indicates that the save statistics option is on for the experiment.
	 * 
	 * @return Returns the option.
	 */
	public final boolean getSaveExperimentStatisticsOption() {
		return mySaveStatisticsOption;
	}

	/** Indicates whether or not the across replication statistics 
	 *  will be saved to a list by the experiment. True indicates that the statistics will be saved.
	 *  
	 * @param option The option to set.
	 */
	public final void setSaveExperimentStatisticsOption(boolean option) {
		mySaveStatisticsOption = option;
	}
	
	/** Sets the design for this experiment. This method
	 *  must be called prior to initializing the experiment for
	 *  the experiment to use the design.
	 * 
	 * @param design
	 */
	public final void setDesign(DesignIfc design){
		if (design == null)
			throw new IllegalArgumentException("The design must be non-null.");
		myDesign = design;		
	}
	
	/** Returns an iterator to the replications within the experiment.  Unless
	 *  the save steps option is true or the save replication statistics option
	 *  is true this iterator will not have any elements
	 * 
	 * @return
	 */
	public final Iterator<Replication> getReplications(){
		Iterator<Replication> i = mySteps.iterator();
		return(i);
	}

	/** Returns a list that holds a copy of the statistics collected across
	 *  the replications.  This method will return an empty list if the save statistic
	 *  option is not turned on.  The list holds copies, not references to the 
	 *  original collected statistics.  The list will hold only across replication statistics
	 *  
	 * @return Returns the statistics.
	 */
	public final List<AbstractStatistic> getAcrossReplicationStatistics() {
		if (myStatistics == null)
			myStatistics = new ArrayList<AbstractStatistic>();
		
		return myStatistics;
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
		
    /**
     *
     */
    protected void initializeIterations() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t initialize_()");		

		if (myAntitheticOption == true){
			if ((myNumReps % 2) != 0 )
				throw new IllegalArgumentException("Number of replications must be even if antithetic option is on.");
			// don't allow replications to allow rv's to advance to next substream
			// this will be controlled via the antithetic option
			setResetNextSubStreamOption(false);
		}

		myCurRepNum = 0; // reset to zero in case experiment is reused
		
		if (myDefaultBatchingOptionFlag == false){ // the user has not set the batching option
			if (myNumReps == 1){// if there is only 1 replication
				// the batching option should be true
				myDefaultBatchingOption = true;
			}			
		}
		
		myDesign.setUp(this, myModel);
		/* tell all the random variables to follow the stream options for the experiment
		 * ResetStartStreamOption = false, means don't reset to start of the streams prior
		 * to the experiment, 
		 * ResetNextSubStreamOption = true means jump to next substream in the stream
		 */
		myModel.setAllRVResetStartStreamOptions(getResetStartStreamOption());
		myModel.setAllRVResetNextSubStreamOptions(getResetNextSubStreamOption());
		myModel.beforeExperiment_();
		super.initializeIterations(); 
	}

    /**
     * 
     */
    protected void endIterations() {
		super.endIterations(); 
		myDefaultBatchingOptionFlag = false;
		myModel.afterExperiment_();
		
		if (mySaveStatisticsOption == true)
			saveStatistics();

	}

	private final void saveStatistics(){
		if (myStatistics == null)
			myStatistics = new ArrayList<AbstractStatistic>();
		
		List<ResponseVariable> rsList = myModel.getResponseVariables();
		
		for(ResponseVariable rs: rsList){
			StatisticalObserver so = rs.getStatisticalObserver();
			Statistic across = (Statistic)so.getAcrossReplicationStatistic();
			Statistic stat = Statistic.newInstance(across);
			myStatistics.add(stat);
		}		
	}
	
	protected final boolean hasNext(){
		return(myCurRepNum < myNumReps);
	}
	
	protected final Replication next(){

		if (!hasNext())
			return null;
		
		if (myAntitheticOption == true){
			if ((myNumReps % 2) != 0 )
				throw new IllegalArgumentException("Number of replications must be even if antithetic option is on.");
		}

		if (myEndReplicationCondition == null)
			myEndReplicationCondition =	new Condition("End Replication Condition");

		if ((myReplication == null) || (mySaveStepOption == true)){
			myReplication = new Replication(myModel, myScheduler, myEndReplicationCondition);
			if (mySaveStepOption == true)
				myReplication.setSaveStatisticsOption(true);
		}
		
		myReplication.setExperiment(this);
		
		myCurRepNum = myCurRepNum + 1;

		myReplication.setReplicationNumber(myCurRepNum);
		myReplication.setLengthOfReplication(myLengthOfReplication);
		myReplication.setLengthOfWarmUp(myLengthOfWarmUp);
		myReplication.setReplicationInitializationOption(myRepInitOption);
		myReplication.setEndCondition(myEndReplicationCondition);
		myReplication.setMonteCarloOption(myMonteCarloOption);
		myReplication.setDefaultBatchingOption(myDefaultBatchingOption);
		
		if (myAntitheticOption == true){
			if ((myCurRepNum % 2) == 0 ){// even replication number
				myModel.resetStartSubStream(); // return to beginning of substream
				myModel.turnOnAntithetic();
			} else {// odd replication number
				if (myCurRepNum > 1){
					myModel.turnOffAntithetic();
					myModel.resetNextSubstream();
				}
			}
		} else { // no antithetic option
			myReplication.setResetNextSubStreamOption(myResetNextSubStreamOption);
		}
		
		// check if user has turned on batch reporting and batching option
		if ((myBatchReportFlag == true) && (myDefaultBatchingOption == true)){
			myReplication.turnOnBatchReport(getName());		
		}
		
		// by default the batch interval is zero, if not zero it has been set
		if (myDefaultBatchInterval > 0.0){ // interval was set for experiment
			// set the replication's default interval to what was set for the experiment
			myReplication.setDefaultBatchInterval(myDefaultBatchInterval);			
		} else { // interval was not set for experiment
			// since all parameters of the replication have been set
			// set its batch interval to the recommended approximation
			// this is set regardless of batching option, 
			// the batching option controls whether or not batching will occur
			myReplication.setDefaultBatchInterval(myReplication.getApproximateBatchInterval());
		}
			
		if (myMaxAllowedExecutionTimePR > 0) // user has set it for each replication
			myReplication.setMaximumExecutionTime(myMaxAllowedExecutionTimePR);
						
		return(myReplication);
	}
	
	protected final void runStep(){
		myCurrentStep = next();
		myReplication = (Replication)myCurrentStep;
		myReplication.runAll();
		if (mySaveStepOption == true)
			mySteps.add(myReplication);
	}

	//  ===========================================
	//      PRIVATE METHODS
	//  ===========================================

	/** Sets the model for this experiment
	 * 
	 * @param model, if model is null, then a default model is created
	 */
	private final void setModel(Model model) {
		if (model == null)
			model = Model.createModel();
		myModel = model;
	}

	/** Sets the scheduler for this experiment
	 * 
	 * @param scheduler, if scheduler is null, then a default scheduler is created
	 */
	private final void setScheduler(Scheduler scheduler) {
		if (scheduler == null)
			scheduler = new Scheduler();
		myScheduler = scheduler;
	}
}
