/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling;

import java.util.Iterator;

/** This class provides an abstract template class (example) for
 *  providing experiments to a simulation. The user of this class must implement
 *  the next() and hasNext() methods of the Iterator interface to provide
 *  the next experiment to an instance of the Simulation class that is running
 *  multiple experiments.  A fully functional experiment must be provided and
 *  ready to be run (e.g. have runAll() called).
 *
 */
abstract public class ExperimentProvider implements Iterator<Experiment> {

	/** can be used by subclasses to represent the 
	 *  next experiment provided
	 */
	protected Experiment myExperiment;
	
	/** can be used by subclasses to create
	 *  experiments that use this model
	 */
	protected Model myModel;

	/** Creates an ExperimentProvider that will
	 *  provide experiments.  Creates a default
	 *  instance of Model to be used within the provider
	 */
	public ExperimentProvider() {
		this(Model.createModel());
	}
		
	/** Creates an ExperimentProvider that will
	 *  provide experiments using the provided Model
	 * 
     *
     * @param model
     */
	public ExperimentProvider(Model model) {
		setModel(model);
	}

	private final void setModel(Model model){
		if (model == null)
			throw new IllegalArgumentException("The model must be non-null.");
		myModel = model;
	}
	
	/** Should return true if there are more experiments
     * @return
     * @see java.util.Iterator#hasNext()
	 */
	abstract public boolean hasNext();

	/** Should return an Object that is an Experiment
	 *  The next experiment provided
     * @return
     * @see java.util.Iterator#next()
	 */
	abstract public Experiment next();

	/** Can be overridden to provide remove functionality
	 *  Currently is a no-op
	 * @see java.util.Iterator#remove()
	 */
	public void remove() {

	}

}
