/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling;

import java.util.*;

import jsl.utilities.reporting.JSL;

/** An experiment runner executes in an iterative manner the experiments
 *  that are provided to it via an iterator across experiments, 
 *  Iterator<Experiment>.  It is a sub-class of IterativeProcess where each
 *  step in its process is an entire Experiment.
 *  
 */
public class ExperimentRunner extends IterativeProcess<Experiment>{
	
	//  ===========================================
	//      OBJECT ATTRIBUTES
	//  ===========================================
	
	/** An interface to a class that provides experiments to be run
	 */
	private Iterator<Experiment> myExperimentProvider;

	/** A reference to the current experiment being run by this
	 *  simulation.
	 */
	private Experiment myExperiment;

	/** Constructs a simulation called "Simulation"
	 * @param experimentProvider A reference to the provider of experiments for this simulation
	 */
	public ExperimentRunner(Iterator<Experiment> experimentProvider) {
		this(null, experimentProvider);
	}

	/** Constructs an simulation called "name" 
	 * @param name The name of the experiment
	 * @param experimentProvider A reference to the provider of experiments for this simulation
	 */
	public ExperimentRunner(String name, Iterator<Experiment> experimentProvider) {
		super(name);
		setExperimentProvider(experimentProvider);
		setSaveStepOption(true);
	}
	
	/** Gets the experiment provider for this simulation
	 * @return
	 */
	public final Iterator<Experiment> getExperimentProvider() {
		return myExperimentProvider;
	}

	/** Sets the experiment provider for this simulation
	 * @param iterator a reference to an ExperimentProviderIfc
	 */
	public final void setExperimentProvider(Iterator<Experiment> iterator) {
		if (iterator == null)
			throw new IllegalArgumentException("The experiment provider was null.");
		myExperimentProvider = iterator;
	}

	/** Gets the current experiment that is running
	 *  May be null if no experiment
	 * @return
	 */
	public final Experiment getExperiment(){
		return(myExperiment);
	}
	
	/** Returns a reference to the underlying event scheduler
	 *  for the current experiment being run
	 *  
	 * @return
	 */
	public final Scheduler getScheduler(){
		return(myExperiment.getScheduler());
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
	
	protected final boolean hasNext(){
		return(myExperimentProvider.hasNext());
	}
	
	protected final Experiment next(){
		return(myExperimentProvider.next());
	}
	
	protected final void runStep(){
		myCurrentStep = next();
		myExperiment = (Experiment)myCurrentStep;
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("Simulation is running experiment " + myExperiment.getName() +
		                " with " + myExperiment.getNumberOfReplications() + " replications.");
		myExperiment.runAll();
		if (mySaveStepOption == true)
			mySteps.add(myExperiment);
	}

}
