/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling;

import java.util.*;
import java.io.IOException;

import jsl.modeling.conditions.*;
import jsl.observers.textfile.IPLogReport;
import jsl.utilities.IdentityIfc;
import jsl.utilities.reporting.JSL;

/** A IterativeProcess is an abstract base class for modeling the execution of
 *  a series of steps until some condition is met (see the setCondition() method).  
 *  An iterative process will stop when one of the following occurs:
 *  
 *  a) It has no more steps to execute
 *  b) A user supplied ConditionIfc has become true after a step
 *  c) A user supplied (real) time limit has been (approximately) reached
 *  d) Calling the end() method
 *  
 *  An iterative process follows a well defined state transition pattern. An
 *  iterative process has the following states: (Created, Initialized, StepCompleted, Ended)
 *  An iterative process is first created and placed in the Created state.
 *  
 *  From the Created state it can transition to Initialized (via initialize()) or
 *  Ended (via end()). From the Ended state it can transition to Initialized (via initialize())
 *  From the Initialized state it can transition to StepCompleted (via runNext()) or to
 *  Ended (via runAll() or end()).  From the StepCompleted state it can transition back to 
 *  StepCompleted (via runNext()) or to Ended (via runAll(), runNext(), or end())
 *  
 *  Thus, a common usage of an IterativeProcess is to call initialize() after construction and
 *  call runNext() to run each step of the process as needed.  An IterativeProcess can also
 *  run to completion by calling runAll().  Subclasses must provide the following 
 *  abstract protected methods: 
 *  
 *  boolean hasNext():  checks if there is a next step
 *  T next(): gets an object that represents the next step
 *  void runStep(): causes the next step to execute
 *  
 *  Sub-classes should also consider overriding initializeIterations() and endIterations() to
 *  provide specific initialization logic and logic for after the process ends.  When overriding these
 *  methods care should be taken to properly call super.initializeIterations() and super.endIterations()
 *  otherwise non-determinant behavior will occur.
 *  
 *  Concrete examples of IterativeProcess include Replication, Experiment, and Simulation.  
 *  A Replication is a series of events that are executed (i.e. each event is a step).  An
 *  Experiment is a series of Replications that are executed (i.e. each replication is a step).
 *  A Simulation is a series of Experiments that are executed (i.e. each experiment is a step in 
 *  a simulation).  Thus, the pattern is general enough to allow IterativeProcesses to run
 *  other IterativeProcesses within each step.
 *  
 * @param <T>
 */
abstract public class IterativeProcess<T> extends Observable implements IdentityIfc, IterativeProcessIfc<T> {

	//  ===========================================
	//      CLASS ATTRIBUTES
	//  ===========================================

	/** A counter to count the number of created to assign "unique" ids
	 */
	private static long myIdCounter_;

	/** The id of this object
	 */
	protected long myId;
	
	/** The name of this step
	*/
	protected String myName;

	/** The default value interpreted as milliseconds for how long in
	* real clock time a iterative process will be approximately
	* allowed to execute.
	*/
	public static final long DEFAULT_MAX_EXECUTION_TIME = 1000;

	/** indicates whether the iterative process has completed running all steps
	*/
	public static final int COMPLETED = 1;

	/** indicates whether the iterative process has exceeded its maximum
	*  execution time
	*/
	public static final int TIMED_OUT = 2;

	/** indicates whether the iterative process ended due the end condition
	*  being met
	*/
	public static final int END_CONDITION_MET = 3;

	/** indicates whether the iterative process ended due to not one of the above
	*/
	public static final int UNFINISHED = 4;

	//  ===========================================
	//      OBJECT ATTRIBUTES
	//  ===========================================
	
	/** The wall clock time in milliseconds that the iterative process started
	*/
	protected long myBeginExecutionTime;

	/** The wall clock time in milliseconds that the iterative process ended
	*/
	protected long myEndExecutionTime;

	/** The maximum allowable execution time "wall" clock time
	*  for the iterative process to complete processing in milliseconds
	*/
	protected long myMaxAllowedExecutionTime;

	/** A flag to indicate whether the iterative process is done
	*  A iterative process can be done if:
	*  1) it ran all of its steps
	*  2) it was canceled by a client prior to completing all of its steps
	*  3) it exceeded its maximum allowable execution time before completing
	*     all of its steps.
	*/
	protected boolean myDoneFlag;
	
	/** Indicates that the iterative process is running
	 * 
	 */
	protected boolean myRunningFlag = false;

	/** indicates how the iterative process ended
	*/
	protected int myEndIndicator;
	
	/** A reference to an object related to the current step of the process
	 *  It can be passed to observers
	 */
	protected T myCurrentStep;
	
	/** A collection of all the steps that have been run by
	 *  this iterative process.
	 */
	protected List<T> mySteps;

	/** A flag to indicate whether or not the iterative process will
	 *  save the steps as they are created and then run in a Collection
	 *  The default is FALSE (they will be not be saved).
	 */
	protected boolean mySaveStepOption;

	/** Used to log state changes
	 */
	protected IPLogReport myIPLogReport;
	
	/** Used to indicate whether logging is on
	 */
	private boolean myIPLogReportOption;

	/** The condition that can be checked to see if the step
	*  is completed.
	*/
	protected ConditionIfc myEndCondition;

	/** A reference to the current state of the iterative process
	*/
	protected IterativeState myState;

	/** A reference to the created state for the iterative process
	*  A iterative process is in the created state when it is initially
	*  constructed and can then only transition to the initialized state
	*/
	protected Created myCreatedState = new Created();

	/** A reference to the initialized state of the iterative process
	*  A iterative process is in the initialized state after the initialize() method
	*  is called from a proper state.
	*/
	protected Initialized myInitializedState = new Initialized();

	/** A reference to the step completed state of the iterative process
	 *  A iterative process is in the step completed state after the runNext method
	 *  is called from a proper state
	*/
	protected StepCompleted myStepCompletedState = new StepCompleted();

	/** A reference to the ended state of the iterative process
	*  A iterative process is in the ended state after the process is told to end
	*/
	protected Ended myEndedState = new Ended();

	/** A Timer used to perform timed task
	 *  e.g. displaying information about the IterativeProcess
	 */
	protected Timer myTimer;

	/** The time between task invocations for the Timer
	 *  and its TimerTask
	 */
	protected long myTBConsoleUpdates;

	/** The task that the timer uses
	*/
	protected TimerTask myTimerTask;
	
	/** A flag that indicates whether or not the iterative process
	 *  has already been initialized, false means it has not been
	 *  initialized, true means that it has been initialized
	 */
	protected boolean myInitFlag;
	
	/** Counts the number of steps executed since the last time the process
	 *  was initialized
	 * 
	 */
	protected long myStepCounter = 0;

	//  ===========================================
	//        CONSTRUCTORS
	//  ===========================================

	/** Constructs an iterative process with default name
	 * 
	 */
	public IterativeProcess() {
		this(null);
	}

	/** Constructs an iterative process with the given name
	 * 
	 * @param name
	 */
	public IterativeProcess(String name) {
		super();
		setId();
		setName(name);
		myEndIndicator = UNFINISHED;
		myMaxAllowedExecutionTime = 0;
		mySteps = new ArrayList<T>();
		mySaveStepOption = false;
		myInitFlag = false;
		// assume that it has not run to completion
		setState(myCreatedState);
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINER))
			JSL.LOGGER.finer("Class name \t" + this.getClass().getName() + "\t Name: " + getName() + "\t ID: \t " + getId() + "\t Constructed");		
	}
	
	//  ===========================================
	//        PUBLIC METHODS
	//  ===========================================

	/** Gets the name.
	* @return The name of object.
	*/
	public String getName() {
		return myName;
	}

	/** Sets the name 
	* @param str The name as a string.
	*/
	public void setName(String str) {
		if (str == null){
			String s = this.getClass().getName();
			int k = s.lastIndexOf(".");
			if (k != -1)
				s = s.substring(k+1);
			myName = s + " " + getId();
		}
		else
			myName = str;
	}

	/** Returns the id for this iterative process
	 * 
	 * @return
	 */
	public long getId() {
		return(myId);
	}

	/** Can be overridden to supply a unique id to the object
	 *  By default a simple static counter is used to assign a number
	 *  as instances are created
	 * 
	 */
	protected void setId() {
		myIdCounter_ = myIdCounter_ + 1;
		myId = myIdCounter_;
	}

	/** Gets a string for the iterative process.  Currently the name.
	 * @return Yields the name of the iterative process.
	 */
	public String toString() {
		String s;

		s = getName();

		return (s);
	}

	/** A flag to indicate whether the iterative process is done
	*  A iterative process can be done if:
	*  1) it ran all of its steps
	*  2) it was ended by a client prior to completing all of its steps
	*  3) it ended because it exceeded its maximum allowable execution time before completing
	*     all of its steps.
	*  4) its end condition was satisfied
	* 
	* @return true if done
	*/
	public final boolean isDone() {
		return (myDoneFlag);
	}
	
	/** Returns system time in milliseconds that the iterative process started
	 * @return the number as a long
	 */
	public final long getBeginExecutionTime() {
		return (myBeginExecutionTime);
	}

	/** Gets the clock time in milliseconds since
	 *  the simulation was initialized
	 * 
	 * @return a long representing the elapsed time
	 */
	public final long getElapsedExecutionTime(){
		if (myBeginExecutionTime > 0)
			return(System.currentTimeMillis() - myBeginExecutionTime);
		else
			return(0);
	}
	
	/** Returns system time in milliseconds that the iterative process ended
	 * @return the number as a long
	 */
	public final long getEndExecutionTime() {
		return (myEndExecutionTime);
	}

	/** Returns maximum (real) clock time allocated for the iterative process
	 * @return the number as long representing milliseconds
	 */
	public final long getMaximumAllowedExecutionTime() {
		return (myMaxAllowedExecutionTime);
	}

	/** Set the maximum allotted (suggested) execution (real) clock for the 
	 *  entire iterative process. This is suggested because the execution time
	 *  requirement is only checked after the completion of an individual step
	 *  After it is discovered that cumulative time for executing the 
	 *  step has exceeded the maximum time, then the iterative process will
	 *  be ended (perhaps) not completing other steps.
	 * 
	 * @param milliseconds
	 */
	public final void setMaximumExecutionTime(long milliseconds) {
		if (milliseconds <= 0.0)
			throw new IllegalArgumentException("The maximum number of execution time (clock time) must be > 0.0");
		myMaxAllowedExecutionTime = milliseconds;
	}

	/** Returns the condition that will cause the iterative process to stop
	*  running its steps
	* @return the condition as a ConditionIfc
	*/
	public ConditionIfc getEndCondition() {
		return (myEndCondition);
	}

	/** Sets the condition for ending the running of the iterative process's steps
	*  The condition is automatically set to false when the iterative process is 
	*  initialized.
	* 
	*  This condition is checked prior to running each step
	*  Steps cannot be started if the condition is true. 
	*
	*  In a multiple step process, if the condition becomes true additional steps
	*  will not be run and the iterative process is ended.
	* 
    *  An example application is to set a condition that terminates the iterative process after
	*  a certain condition has been met across the steps.
	*
	*  Throws an IllegalArgumentException if the supplied condition is null.
	*  @param condition An object that implements the Condition interface
	*/
	public final void setEndCondition(ConditionIfc condition) {
		if (condition == null)
			throw new IllegalArgumentException("The Condition must be non-null.");
		myEndCondition = condition;
	}
	
	/** Gets the save step option, true means 
	 *  that the steps will be saved by the iterative process into a Collection
	 *  after running
	 * @return flag indicating the option
	 */
	public final boolean getSaveStepOption() {
		return mySaveStepOption;
	}

	/** Sets the save step option, true means 
	 *  that the steps will be saved by the iterative process into a Collection
	 *  after running
     * @param option
	 */
	public final void setSaveStepOption(boolean option) {
		mySaveStepOption = option;
	}

	/** Returns an iterator to the saved steps
	 * 
	 * @return an iterator
	 */
	public final Iterator<T> getStepIterator(){
		return(mySteps.iterator());
	}
	
	/** Returns an unmodifiable list view of the steps that have
	 *  been saved
	 * 
	 * @return
	 */
	public final List<T> getStepList(){
		return(Collections.unmodifiableList(mySteps));
	}
	
	/** Returns the current step for the iterative process
	 *  the step that is or has just completed processing
	 *  Note: Sub-classes are responsible for properly setting the current step
	 *  when a step is run via runStep()
	 *  
	 * @return
	 */
	public final T getCurrentStep(){
		return myCurrentStep;
	}

	/** Returns the number of steps completed since the iterative process
	 *  was last initialized
	 * 
	 * @return
	 */
	public final long getNumberStepsCompleted(){
		return myStepCounter;
	}
	
	/** Turns on the log report. This report yields a text file of
	 *  the state changes for the iterative process
	 */
	public final void turnOnLogReport(){
		turnOnLogReport(null);
	}
	
	/** Turns on the log report. This report yields a text file of
	 *  the state changes for the iterative process
	 *  @param name specifies a prefix name for the report 
	 */
	public void turnOnLogReport(String name){
		myIPLogReportOption = true;
        
		if(myIPLogReport == null){ // has not been made, so make it
			try{
				String s;
				if (name != null)
					s = name + " " + getName() + " LogReport";
				else
					s = getName() + " LogReport";
				myIPLogReport = new IPLogReport(s);
				addObserver(myIPLogReport);
			}
			catch(IOException e){
				String s = "I/O exception opening IterativeProcess = " + getId() + " Name: " + getName();
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.SEVERE))
					JSL.LOGGER.severe(s);
				System.err.println(s);
				System.exit(1);
			}
		}
	}
    
	/** Turns off log reporting.
	 */
	public void turnOffLogReport(){
		if (myIPLogReportOption == true){
			deleteObserver(myIPLogReport);
			myIPLogReportOption = false;
			myIPLogReport = null;
		}
	}
    
	/** Returns a reference to the current log report.
	 * @return The log report
	 */
	public final IPLogReport getLogReport(){
		return(myIPLogReport);
	}

	/** This method will cause the a timer to start
	 *  allowing a TimerTask to be scheduled.  This method
	 *  causes a print to console task to be run
	 *  which prints the elapsed time since initializing
	 *  the iterative process.
	 * @param milliseconds
	*/
	public void turnOnTimer(long milliseconds) {
		turnOnTimer(milliseconds, new ShowElapsedTimeTask());
	}

	/** This method will cause the a timer to start
	 *  allowing a TimerTask to be scheduled.  An example
	 *  use of this is to periodically print some information
	 *  to the console about the iterative process
	 *  to be printed to the console periodically according to
	 *  the parameter given
     * @param milliseconds
     * @param timerTask
	*/
	public final void turnOnTimer(long milliseconds, TimerTask timerTask) {
		if (timerTask == null)
			throw new IllegalArgumentException("The TimerTask must be non-null.");
		myTimerTask = timerTask;
		myTBConsoleUpdates = milliseconds;
	}

	/** Checks if the iterative process is in the initialized state
	*  After the iterative process has been initialized this method will return
	*  true
	*  @return true if initialized
	*/
	public final boolean isInitialized() {
		return (myState == myInitializedState);
	}
	
	/** An iterative process is running if it is been told to run (i.e. runNext())
	 *  but has not yet been told to end().
	 * 
	 * @return
	 */
	public final boolean isRunning(){
		return(myRunningFlag);
	}

	/** Checks if the iterative process is in the completed step state
	*  After the iterative process has successfully completed a step this method will return
	*  true
	*  @return true if the iterative process completed the step
	*/
	public final boolean isStepCompleted() {
		return (myState == myStepCompletedState);
	}

	/** Checks if the iterative process is in the ended state
	*  After the iterative process has been ended this method will return
	*  true
	*  @return true if ended
	*/
	public final boolean isEnded() {
		return (myState == myEndedState);
	}

	/** The iterative process may end by a variety of means, this method
	*  checks if the iterative process ended because it ran all of its steps
	*
     *
     * @return
     */
	public final boolean isCompleted() {
		return (myEndIndicator == COMPLETED);
	}

	/** The iterative process may end by a variety of means, this method
	*  checks if the iterative process ended because it timed out
	*
     *
     * @return
     */
	public final boolean isTimedOut() {
		return (myEndIndicator == TIMED_OUT);
	}

	/** The iterative process may end by a variety of means, this method
	*  checks if the iterative process ended because its end condition was met
	*
     *
     * @return
     */
	public final boolean isEndConditionMet() {
		return (myEndIndicator == END_CONDITION_MET);
	}

	/** The iterative process may end by a variety of means, this method
	*  checks if the iterative process ended but was unfinished, not all steps completed
	*
     *
     * @return 
     */
	public final boolean isUnfinished() {
		return (myEndIndicator == UNFINISHED);
	}

	/** Initializes the iterative process prior to running any steps
	*  This must be done prior to calling runNext();
	*/
	public final void initialize() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t initialize()");		
		myState.initialize();
	}

	/** Runs the next step in the iterative process 
	*/
	public final void runNext() {
		myRunningFlag = true;
		myState.runNext();
	}

	/** Runs all of the steps 
	 *  The iterative process is first initialized, then each step is run, and
	 *  finally the iterative process is ended.
	 *
	*/
	public final void runAll() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t runAll()");		
		runAll_();
	}

	/** Tells the iterative process to end.
	 *
	*/
	public final void end() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t end()");		
		myState.end();
	}
	
	//  ===========================================
	//      ABSTRACT PROTECTED METHODS
	//  ===========================================
	
	/** This method should check to see if another step is necessary
	 *  for the iterative process. True means that the process has
	 *  another step to be executed.  False, means that no more steps
	 *  can be executed.
	 *  
	 * @return
	 */
	abstract protected boolean hasNext();
	
	/** This method should return the next step to be executed
	 *  in the iterative process or null if no more steps can
	 *  be executed. It should advance the current step to
	 *  the next step if it is available
	 * 
	 * @return
	 */
	abstract protected T next();
	
	/** This method tells the iterative process to execute
	 *  the current step.  Typical usage is to call this
	 *  after calling next() to advance to the next step.
	 * 
	 */
	abstract protected void runStep();
		
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
	
	/** Can be overwritten by subclasses to have output when
	 *   the console flag is on
	 * 
	 */
	protected void consoleOutput(){	
	}
	
	/** The method is used to set the state and notify observers
	*  of the change
     *
     * @param state
     */
	protected final void setState(IterativeState state) {
		myState = state;
		setChanged();
		notifyObservers(myCurrentStep);
	}
	
    /**
     *
     */
    protected void initializeIterations(){
		myDoneFlag = false;
		myStepCounter = 0;
		
		if (mySteps != null)
			mySteps.clear();
		
		if (myEndCondition == null)
			myEndCondition = new Condition("End Condition");
		else
			myEndCondition.clear();
	
		myBeginExecutionTime = System.currentTimeMillis();
		setState(myInitializedState);
		myInitFlag = true;
		if (myTBConsoleUpdates > 0){
			myTimer = new Timer();
			myTimer.schedule(myTimerTask, myTBConsoleUpdates, myTBConsoleUpdates);
		}

	}

    /**
     *
     */
    protected void runNext_() {
	
		if (!myEndCondition.isSet()) { //condition not met
			if (hasNext()) { // has another step
				if (myMaxAllowedExecutionTime > 0){
					// user has specified a max execution time
					// check if past the time
					long runtime = System.currentTimeMillis() - myBeginExecutionTime;
					if (runtime < myMaxAllowedExecutionTime) {
						// has not exceeded allotted execution time
						runStep();
						setState(myStepCompletedState);
						myStepCounter++;
					} else { // exceeded allotted time
						myEndIndicator = TIMED_OUT;
						myDoneFlag = true;
					}					
				} else { //no need to check the execution time
					// just run the step
					runStep();
					setState(myStepCompletedState);	
					myStepCounter++;
				}
			} else { // all steps done
				myEndIndicator = COMPLETED;
				myDoneFlag = true;
			}
		} else { // the end condition was met
			myDoneFlag = true;
			myEndIndicator = END_CONDITION_MET;
		}
	}
		
	/** Runs all of the steps. Each step is run, and
	*  finally the iterative process is ended.
	*
	*/
	protected final void runAll_() {
		if (myInitFlag == false)
			initialize();
		while (!isDone()) {
			runNext();
		}
		end();
	}
	
    /**
     *
     */
    protected void endIterations() {
		myRunningFlag = false;
		myDoneFlag = true;
		myEndExecutionTime = System.currentTimeMillis();
		setState(myEndedState);
		
		if (myTimer != null){
			myTimer.cancel();
			System.out.println("Elapsed Clock Time: " + getElapsedExecutionTime()/1000.0 + " seconds.");
			System.out.println(getName() + " ended.");
			myTimer = null;
		}
		myInitFlag = false;
	}
	
	//  ===========================================
	//      INNER CLASSES
	//  ===========================================

	private class IterativeState {

		private String myName;

		public IterativeState(String name) {
			myName = name;
		}

		public void initialize() {
			throw new IllegalStateException("Tried to initialize from an illegal state: " + myState);
		}

		public void runNext() {
			throw new IllegalStateException("Tried to run the next step from an illegal state: " + myState);
		}

		public void runAll() {
			throw new IllegalStateException("Tried to run all the steps from an illegal state: " + myState);
		}

		public void end() {
			throw new IllegalStateException("Tried to end from an illegal state: " + myState);
		}

		public String toString() {
			return (myName);
		}
	}

	private class Created extends IterativeState {

		public Created() {
			super("CreatedState");
		}

		public void initialize() {
			initializeIterations();
		}

		public void end() {
			endIterations();
		}

	}

	private class Initialized extends IterativeState {

		public Initialized() {
			super("InitializedState");
		}

		public void runNext() {
			runNext_();
		}

		public void runAll() {
			runAll_();
		}

		public void end() {
			endIterations();
		}

	}

	private class StepCompleted extends IterativeState {

		public StepCompleted() {
			super("CompletedState");
		}

		public void runNext() {
			runNext_();
		}

		public void runAll() {
			runAll_();
		}

		public void end() {
			endIterations();
		}
	}

	private class Ended extends IterativeState {

		public Ended() {
			super("EndedState");
		}

		public void initialize() {
			initializeIterations();
		}
	}
	
	private class ShowElapsedTimeTask extends TimerTask {

		/* (non-Javadoc)
		 * @see java.util.TimerTask#run()
		 */
		public void run() {
			System.out.println("Elapsed Clock Time: " + getElapsedExecutionTime()/1000 + " seconds.");
		}
		
	}

}
