package jsl.modeling;

import java.util.Iterator;
import java.util.List;
import java.util.TimerTask;

import jsl.modeling.conditions.ConditionIfc;
import jsl.observers.textfile.IPLogReport;

/**
 *
 * @author rossetti
 * @param <T>
 */
public interface IterativeProcessIfc<T> {

	/** Gets the name.
	 * @return The name of object.
	 */
	public String getName();

	/** Sets the name 
	 * @param str The name as a string.
	 */
	public void setName(String str);

	/** Returns the id for this iterative process
	 * 
	 * @return
	 */
	public long getId();

	/** Gets a string for the iterative process.  Currently the name.
	 * @return Yields the name of the iterative process.
	 */
	public String toString();

	/** A flag to indicate whether the iterative process is done
	 *  A iterative process can be done if:
	 *  1) it ran all of its steps
	 *  2) it was ended by a client prior to completing all of its steps
	 *  3) it ended because it exceeded its maximum allowable execution time before completing
	 *     all of its steps.
	 *  4) its end condition was satisfied
	 * 
	 * @return true if done
	 */
	public boolean isDone();

	/** Returns system time in milliseconds that the iterative process started
	 * @return the number as a long
	 */
	public long getBeginExecutionTime();

	/** Gets the clock time in milliseconds since
	 *  the simulation was initialized
	 * 
	 * @return a long representing the elapsed time
	 */
	public long getElapsedExecutionTime();

	/** Returns system time in milliseconds that the iterative process ended
	 * @return the number as a long
	 */
	public long getEndExecutionTime();

	/** Returns maximum (real) clock time allocated for the iterative process
	 * @return the number as long representing milliseconds
	 */
	public long getMaximumAllowedExecutionTime();

	/** Set the maximum allotted (suggested) execution (real) clock for the 
	 *  entire iterative process. This is suggested because the execution time
	 *  requirement is only checked after the completion of an individual step
	 *  After it is discovered that cumulative time for executing the 
	 *  step has exceeded the maximum time, then the iterative process will
	 *  be ended (perhaps) not completing other steps.
	 * 
	 * @param milliseconds
	 */
	public void setMaximumExecutionTime(long milliseconds);

	/** Returns the condition that will cause the iterative process to stop
	 *  running its steps
	 * @return the condition as a ConditionIfc
	 */
	public ConditionIfc getEndCondition();

	/** Sets the condition for ending the running of the iterative process's steps
	 *  The condition is automatically set to false when the iterative process is 
	 *  initialized.
	 * 
	 *  This condition is checked prior to running each step
	 *  Steps cannot be started if the condition is true. 
	 *
	 *  In a multiple step process, if the condition becomes true additional steps
	 *  will not be run and the iterative process is ended.
	 * 
	 *  An example application is to set a condition that terminates the iterative process after
	 *  a certain condition has been met across the steps.
	 *
	 *  Throws an IllegalArgumentException if the supplied condition is null.
	 *  @param condition An object that implements the Condition interface
	 */
	public void setEndCondition(ConditionIfc condition);

	/** Gets the save step option, true means 
	 *  that the steps will be saved by the iterative process into a Collection
	 *  after running
	 * @return flag indicating the option
	 */
	public boolean getSaveStepOption();

	/** Sets the save step option, true means 
	 *  that the steps will be saved by the iterative process into a Collection
	 *  after running
     * @param option
	 */
	public void setSaveStepOption(boolean option);

	/** Returns an iterator to the saved steps
	 * 
	 * @return an iterator
	 */
	public Iterator<T> getStepIterator();

	/** Returns an unmodifiable list view of the steps that have
	 *  been saved
	 * 
	 * @return
	 */
	public List<T> getStepList();

	/** Returns the current step for the iterative process
	 *  the step that is or has just completed processing
	 *  Note: Sub-classes are responsible for properly setting the current step
	 *  when a step is run 
	 *  
	 * @return
	 */
	public T getCurrentStep();

	/** Returns the number of steps completed since the iterative process
	 *  was last initialized
	 * 
	 * @return
	 */
	public long getNumberStepsCompleted();

	/** Turns on the log report. This report yields a text file of
	 *  the state changes for the iterative process
	 */
	public void turnOnLogReport();

	/** Turns on the log report. This report yields a text file of
	 *  the state changes for the iterative process
	 *  @param name specifies a prefix name for the report 
	 */
	public void turnOnLogReport(String name);

	/** Turns off log reporting.
	 */
	public void turnOffLogReport();

	/** Returns a reference to the current log report.
	 * @return The log report
	 */
	public IPLogReport getLogReport();

	/** This method will cause the a timer to start
	 *  allowing a TimerTask to be scheduled.  This method
	 *  causes a print to console task to be run
	 *  which prints the elapsed time since initializing
	 *  the iterative process.
	 * @param milliseconds
	 */
	public void turnOnTimer(long milliseconds);

	/** This method will cause the a timer to start
	 *  allowing a TimerTask to be scheduled.  An example
	 *  use of this is to periodically print some information
	 *  to the console about the iterative process
	 *  to be printed to the console periodically according to
	 *  the parameter given
     * @param milliseconds
     * @param timerTask
	 */
	public void turnOnTimer(long milliseconds, TimerTask timerTask);

	/** Checks if the iterative process is in the initialized state
	 *  After the iterative process has been initialized this method will return
	 *  true
	 *  @return true if initialized
	 */
	public boolean isInitialized();

	/** An iterative process is running if it is been told to run (i.e. runNext())
	 *  but has not yet been told to end().
	 * 
	 * @return
	 */
	public boolean isRunning();

	/** Checks if the iterative process is in the completed step state
	 *  After the iterative process has successfully completed a step this method will return
	 *  true
	 *  @return true if the iterative process completed the step
	 */
	public boolean isStepCompleted();

	/** Checks if the iterative process is in the ended state
	 *  After the iterative process has been ended this method will return
	 *  true
	 *  @return true if ended
	 */
	public boolean isEnded();

	/** The iterative process may end by a variety of means, this method
	 *  checks if the iterative process ended because it ran all of its steps
	 *
     *
     * @return
     */
	public boolean isCompleted();

	/** The iterative process may end by a variety of means, this method
	 *  checks if the iterative process ended because it timed out
	 *
     *
     * @return
     */
	public boolean isTimedOut();

	/** The iterative process may end by a variety of means, this method
	 *  checks if the iterative process ended because its end condition was met
	 *
     *
     * @return
     */
	public boolean isEndConditionMet();

	/** The iterative process may end by a variety of means, this method
	 *  checks if the iterative process ended but was unfinished, not all steps completed
	 *
     *
     * @return 
     */
	public boolean isUnfinished();

	/** Initializes the iterative process prior to running any steps
	 *  This must be done prior to calling runNext();
	 */
	public void initialize();

	/** Runs the next step in the iterative process 
	 */
	public void runNext();

	/** Runs all of the steps 
	 *  The iterative process is first initialized, then each step is run, and
	 *  finally the iterative process is ended.
	 *
	 */
	public void runAll();

	/** Tells the iterative process to end.
	 *
	 */
	public void end();

}