/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling;

import java.lang.Comparable;

import java.text.DecimalFormat;
import jsl.utilities.IdentityIfc;

/** This class represents a simulated event. This file contains the JSLEvent class.  It allows
 *	for the simulation of durations of simulated time. These events are placed on the 
 *  scheduler and ordered by time, priority, and order of creation.  JSLEvents
 *  are the mechanism by which ModelElements notify each other of work to be done.
 *
 *  The JSLEvent is part of a delegation event mechanism patterned closely to the AWT delegation
 *  mechanism with a few enhancements
 */
public class JSLEvent implements Comparable<JSLEvent>, IdentityIfc {
    //  ===========================================
    //      CLASS AND OBJECT ATTRIBUTES
    //  ===========================================
    
    private Object myMessage;
    private EventListenerIfc myListener;
    private ModelElement myDispatcher;
    
    private String myName;
    private double myTime;
    private int myPriority;
    private long myId;
    private int myType;
    private boolean myCancelledFlag;
    private boolean myScheduledFlag;
    
    /*  --Class Constants---------------- */
    /** Represents the default priority for events within the scheduler
     * DEFAULT_PRIORITY = 10.  Lower priority goes first.  All integer
     * priority numbers can be used to set the priority of an event.
     */
    public static final int DEFAULT_PRIORITY = 10;
    
    /**
     *  The default time for an event is zero.  It is just a number
     *  that can be use to map over to event type
     */
    public static final int DEFAULT_TYPE = 0;
    
    //  ===========================================
    //        CONSTRUCTORS
    //  ===========================================
    
    /**
     * Constructs an instance of an event.  This constructor has package scope
     * because only the Scheduler class can make events.  The methods
     *
     *   setScheduledFlag()
     *   setCancelledFlag()
     *   setDispatcher()
     *   setName()
     *   setListener()
     *   setTime()
     *   setPriority()
     *   setMessage()
     *   setId()
     *
     * must all be appropriately called by Scheduler to ensure a valid state
     * for the event.
     *
     * @returns an unprepared JSLEvent
     */
    JSLEvent(){
        myCancelledFlag = false;
        myScheduledFlag = false;
    }
    
    //  ===========================================
    //        PUBLIC INTERFACE
    //  ===========================================
    
    /*  --Mutators----------------------- */
    
    /** Allows the setting of an object that is sent with the event.
     * This object can be used to pass information between the
     * dispatcher and the listener
     * @param obj Represents the Object acting as a message
     */
    public final void setMessage(Object obj){
        myMessage = obj;
    }
    
    /** Sets the name of the event.  Useful for tracing events.
     * @param name A String representing the name of the event
     */
    public final void setName(String name){
        myName = name;
    }
    
    /** Sets the type of the event.  Useful for tracing events.
     *  or distinguishing between events.
     * @param type An int representing the type of the event
     */
    public final void setType(int type){
        myType = type;
    }
    
    /** Sets the listener for the event.  The listener of the event
     * is an object that implements the EventListenerIfc interface.
     * Typically, an ActionListenerIfc is used.  If null is passed in
     * a JSLEventException occurs.  Every event must have
     * a listener.  The user may change the listener after the event
     * has been scheduled.
     * @param listener A instance of a class that implements the EventListenerIfc (e.g. ActionListenerIfc)
     * must be non-null or a JSLException occurs
     */
    public final void setListener(EventListenerIfc listener){
        if (listener == null)//no listener
            throw new JSLEventException("No listener provided for the event");
        myListener = listener;
    }
    
    /** Sets the dispatcher for the event.  The dispatcher of the event
     * is a ModelElement that will cause the listener to be called at
     * the appropriate simulated time. If null is passed in
     * a JSLEventException occurs.  Every event must have
     * a dispatcher  The user may change the dispatcher after the event
     * has been scheduled.
     * @param dispatcher A instance of a ModelElement
     * must be non-null or a JSLException occurs
     */
    public final void setDispatcher(ModelElement dispatcher){
        if (dispatcher == null)
            throw new IllegalArgumentException("Dispatching ModelElement was null!");
        
        myDispatcher = dispatcher;
    }
    
    /** Indicates to the scheduler that the event should be canceled or not.
     * True implies that the event is to be canceled.  It is up to the
     * scheduler to handle the cancellation.  If the event is canceled, it's
     * execute method will not be called when it becomes the current
     * event.  Thus, an event can be canceled or uncanceled at any
     * simulated time prior to when the event is scheduled to occur.
     *
     * @param b A boolean (true is canceled, false is not canceled)
     *
     */
    public final void setCanceledFlag(boolean b){
        myCancelledFlag = b;
    }
    
    /*  --Accessors---------------------- */
    
    /** Returns the object representing the message sent with the event
     *
     * @return Object representing the message sent with the event
     */
    public final Object getMessage(){
        return(myMessage);
    }
    
    /** Gets the EventListenerIfc
     * @return The object representing the EventListenerIfc
     */
    public final EventListenerIfc getListener(){
        return(myListener);
    }
    
    /** Gets the name of the event
     * @return The name of the event
     */
    public final String getName(){
        return(myName);
    }
    
    /** Gets the time that the event is schedule for.
     * @return A double representing simulated time
     */
    public final double getTime(){
        return(myTime);
    }
    
    /** Gets the priority of the event
     * @return An int representing the priority.  Lower is better.
     */
    public final int getPriority(){
        return(myPriority);
    }
    
    /** Gets the the id assigned to the event by the scheduler.  No two
     * events have the same id.
     * @return A long representing the id
     */
    public final long getId(){
        return(myId);
    }
    
    /** Gets the type of the event.  Useful for tracing events.
     *  or distinguishing between events.
     * @return an int representing the type of the event.
     */
    public final int getType(){
        return(myType);
    }
    
    /** Gets a flag indicating whether the event is to be canceled or not.
     * True implies that the event is to be canceled.  It is up to the
     * scheduler to handle the cancellation.  If the event is canceled, it's
     * execute method will not be called when it becomes the current
     * event.  Thus, an event can be canceled or uncanceled at any
     * simulated time prior to when the event is scheduled to occur.
     *
     * @return b A boolean (true is canceled, false is not canceled)
     *
     */
    public final boolean getCanceledFlag(){
        return(myCancelledFlag);
    }
    
    /** Gets the ModelElement that will act as the dispatcher for the event.
     * @return a ModelElement that will act as the dispatcher for the event
     */
    public final ModelElement getDispatcher(){
        return(myDispatcher);
    }
    
    /** Gets a flag indicating whether the event is in the event calendar, i.e.
     * it has been scheduled.
     *
     * True implies that the event has been scheduled.  It is an error to attempt
     * to schedule an event that has already been scheduled.
     *
     * @return A boolean (true is scheduled, false is not scheduled)
     */
    public final boolean isScheduled(){
        return(myScheduledFlag);
    }
        
    /** Provides a string representation for the event.  Useful for tracing
     *
     * @return A String representing the event
     */
    public String toString(){
    	StringBuilder sb = new StringBuilder();       
        DecimalFormat df = new DecimalFormat("0.###");
        sb.append(df.format(myTime) + "> " + " Event = " + myName + "\t");
        sb.append("ID = " + myId + "\t");
        sb.append("Priority = " + myPriority + "\t");
        sb.append("Type = " + myType + "\t");
        sb.append("Listener = " + myListener + "\t");
        sb.append("Dispatcher = " + myDispatcher);
         return(sb.toString());
    }
       
    //  ===========================================
    //      PROTECTED METHODS
    //  ===========================================
    
    /** Sets the id of the event, package scope because only the Scheduler should be setting the id
     * @param id Provided by the Scheduler class
     */
    protected final void setId(long id){
        myId = id;
    }
    
    /** Sets the scheduled flag of the event, package scope because only the Scheduler should
     * indicate if the event is scheduled
     * @param flag Provided by the Scheduler class (true is scheduled)
     */
    protected final void setScheduledFlag(boolean flag){
        myScheduledFlag = flag;
    }
    
    /** Sets the scheduled time of the event, protected scope because only the Scheduler should
     * indicate when the event is scheduled
     * @param t Provided by the Scheduler class.  The time of the event.
     *
     */
    protected final void setTime(double t){
        if (isScheduled())//already scheduled
            throw new JSLEventException("Tried to change time of already scheduled event!");
        myTime = t;
    }
    
    /** Sets the scheduled priority of the event, package scope because only the Scheduler should
     * indicate the priority of the event
     * @param p Provided by the Scheduler class
     */
    protected final void setPriority(int p){
        if (isScheduled())//already scheduled
            throw new JSLEventException("Tried to change priority of already scheduled event!");
        
        myPriority = p;
    }
    
    /** Called by the Scheduler class to tell the dispatcher to
     * call the listener at the appropriate time.
     */
    protected void execute(){
        myDispatcher.dispatchEvent(this);
    }
     
    /** Returns a negative integer, zero, or a positive integer
     * if this object is less than, equal to, or greater than the
     * specified object.
     *
     * Natural ordering: time, then priority, then order of creation
     *
     * Lower time, lower priority, lower order of creation goes first
     *
     * Throws ClassCastException if the specified object's type
     * prevents it from begin compared to this object.
     *
     * Throws RuntimeException if the id's of the objects are the same,
     * but the references are not when compared with equals.
     *
     * Note:  This class may have a natural ordering that is inconsistent
     * with equals.
     * @param event The event to compare this event to
     * @return Returns a negative integer, zero, or a positive integer
     * if this object is less than, equal to, or greater than the
     * specified object.
     */
    public final int compareTo(JSLEvent event) {
        
        // compare time first
        
        if (myTime < event.getTime())
            return (-1);
        
        if (myTime > event.getTime())
            return (1);
        
        // times are equal, check priorities
        
        if (myPriority < event.getPriority())
            return (-1);
        
        if (myPriority > event.getPriority())
            return (1);
        
        // time and priorities are equal, compare ids
        
        if (myId < event.getId()) // lower id, implies created earlier
            return (-1);
        
        if (myId > event.getId())
            return (1);
        
        // if the id's are equal then the object references must be equal
        // if this is not the case there is a problem
        
        if (this.equals(event))
            return(0);
        else
            throw new RuntimeException("Id's were equal, but references were not, in JSLEvent compareTo");
        
    }
      
}



