/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling;

import java.io.IOException;
import java.util.*;

import jsl.modeling.elements.spatial.SpatialModel;
import jsl.modeling.elements.variable.*;
import jsl.observers.textfile.*;
//import jsl.spatial.spatial2D.SpatialModel2D;
import jsl.utilities.reporting.JSL;
import jsl.modeling.elements.processview.description.EntityType;
import jsl.utilities.random.rng.RNGStreams;
import jsl.utilities.random.rng.RngIfc;

/**
 *
 * This class serves as the base (container) model element for all
 * model elements in the simulation.  
 */
public class Model extends ModelElement{
    
    //  ===========================================
    //      CLASS ATTRIBUTES
    //  ===========================================
	
	/** A constant for the default batch interval
	 */
	public static final double DEFAULT_BATCH_INTERVAL = 100;

	/** A constant for the default batch priority
	 */
	public static final int DEFAULT_BATCH_PRIORITY = 2;    

	//  ===========================================
	//      OBJECT ATTRIBUTES
	//  ===========================================

    /** This indicates whether or not the model is performing the default
     *  batching of its elements.
     */
    private boolean myDefaultBatchingOption = false;

	/** The action listener that reacts to the timed update event.
	 */
	protected BatchActionListener myBatchActionListener;

	/** A reference to the Batching event.
	 */
	protected JSLEvent myBatchEvent;

	/** The priority for the batching events.
	 */
	protected int myBatchEventPriority = DEFAULT_BATCH_PRIORITY;

	/** The time interval between batching events.
	 *  The default is zero for no batching
	 */
	protected double myBatchInterval = 0.0;
	
	/** A list of the TimeWeighted response variables.  These may
	 *  opt in to batching via the model
	 */
	protected List<BatchIfc> myBatchVariables;

	/** A list of all the response variables within the model
	 */
	protected List<ResponseVariable> myResponseVariables;
	
	/** A list of all the Counters within the model
	 */
	protected List<Counter> myCounters;
	
	/** A list of all the Variables within the model
	 */
	protected List<Variable> myVariables;
	
	/** A list of all random variables within the model
	 */
	protected List<RandomVariable> myRandomVariables;
	
	/** A reference to the default entity type
	 */
	private EntityType myDefaultEntityType;
	
	/** A Map that holds all the model elements in the order
	 *  in which they are created
	 */
	private Map<String,ModelElement> myModelElementMap;
	
	/** A flag that indicates whether model log reporting is on
	 */
	private boolean myLogReportFlag;
	
	/** A reference to the model's logging report
	 */
	private LogReport myLogReport;
    
    /** A reference to the currently running replication
     */
    private Replication myReplication;
    
	/** A reference to the action listener to handle
	 *  the end of the replication event
	 */
    private EndReplicationActionListener myEndReplicationActionListener;
    
    /** A reference to the end replication event
     */
    private JSLEvent myEndReplicationEvent;
    
    /** Indicates whether or not the model should automatically remove
     *  any elements that have been marked for removal prior to each 
     *  replication.  The default is false;
     * 
     */
    private boolean myAutoRemoveMarkedElementsOption = false;
    
    /** If a model element is added to the model after a replication has started
     *  this is likely to be a conceptual error. A warning message will be printed
     *  to the console in this case.  The user can turn off this warning message 
     *  by setting this flag to false.  The default is true.
     * 
     */
    private boolean myElementAddedAfterReplicationStartedWarningOption = true;
            
    //  ===========================================
    //        CONSTRUCTORS
    //  ===========================================
    
    /**
     *
     * @return
     */
    public static final Model createModel(){
    	return(Model.createModel(null));
    }
    
    /**
     *
     * @param name
     * @return
     */
    public static final Model createModel(String name){
    	return(new Model(name));
    }
    
    /** Constructs a model called "name"  
	* @param name The name of the model
    */
    public Model(String name){
        super(name);
        myModelElementMap = new LinkedHashMap<String,ModelElement>();
        myBatchVariables = new ArrayList<BatchIfc>();
        myResponseVariables = new ArrayList<ResponseVariable>();
        myCounters = new ArrayList<Counter>();
        myRandomVariables = new ArrayList<RandomVariable>();
        myVariables = new ArrayList<Variable>();
        myLengthOfWarmUp = 0.0; // zero means no warm up
        myEndReplicationActionListener = new EndReplicationActionListener();
		setModel(this);
		setParentModelElement(null);
		addToModelElementMap(this);
		myDefaultEntityType = new EntityType(this, "DEFAULT_ENTITY_TYPE");
    }
    
    //  ===========================================
    //        PUBLIC INTERFACE
    //  ===========================================

	/** Returns whether or not default batching is on for the model
	 *  True means that it is ON
	 * @return Returns the defaultBatchingOption.
	 */
	public final boolean isDefaultBatchingOptionOn() {
		return myDefaultBatchingOption;
	}
	
	/** Sets the batch event priority.
	 * @param priority The batch event priority, lower means earlier
	 */
	public final void setBatchEventPriority(int priority) {
		myBatchEventPriority = priority;
	}

	/** Gets the batch event priority
	 * @return The batch event priority
	 */
	public final int getBatchEventPriority() {
		return (myBatchEventPriority);
	}

	/** Gets the current batch interval length
	 * @return The batch interval as time
	 */
	public final double getBatchInterval() {
		return (myBatchInterval);
	}

	/** Checks if a batching event has been scheduled for this model element
	 * @return True means that it has been scheduled.
	 */
	public final boolean isBatchEventScheduled() {
		if (myBatchEvent == null)
			return (false);
		else
			return (myBatchEvent.isScheduled());
	}

	/** Checks if a batching event has been scheduled for this model element
	 * @return True means that it has been scheduled.
	 */
	public final boolean isEndReplicationEventScheduled() {
		if (myEndReplicationEvent == null)
			return (false);
		else
			return (myEndReplicationEvent.isScheduled());
	}
	
	/** Gets the current replication
	 * @return
	 */
	public final Replication getReplication() {
		return (myReplication);
	}
 
	/** Returns the variable associated with the name or null if named
	 *  element is not in the model
	 *  @param name The name of the Variable model element
	 *  @return the associated variable, may be null if
	 *   provided name does not exist in the model
	 */
	public final Variable getVariable(String name){
		if (name == null)
			throw new IllegalArgumentException("The name must be non-null.");

		if(!myModelElementMap.containsKey(name))
			return(null);
		
		ModelElement v = myModelElementMap.get(name);
		
		if (v instanceof Variable)
			return((Variable)v);
		else
			return(null);
	}

	/** Returns the random variable associated with the name or null if named
	 *  element is not in the model
	 *  @param name The name of the RandomVariable model element	
	 *  @return the associated random variable, may be null if
	 *   provided name does not exist in the model
	 */
	public final RandomVariable getRandomVariable(String name){
		if (name == null)
			throw new IllegalArgumentException("The name must be non-null.");

		if(!myModelElementMap.containsKey(name))
			return(null);
		
		ModelElement v = myModelElementMap.get(name);
		
		if (v instanceof RandomVariable)
			return((RandomVariable)v);
		else
			return(null);
	}
	
	/** Returns the response variable associated with the name or null if named
	 *  element is not in the model
	 *  @param name The name of the ResponseVariable model element
	 *  
	 *  @return the associated within replication statistic information, may be null if
	 *   provided name does not exist in the model
	 */
	public final WithinReplicationStatisticIfc getResponseVariable(String name){
		if (name == null)
			throw new IllegalArgumentException("The name must be non-null.");

		if(!myModelElementMap.containsKey(name))
			return(null);
		
		ModelElement v = myModelElementMap.get(name);
		
		if (v instanceof ResponseVariable)
			return((WithinReplicationStatisticIfc)v);
		else
			return(null);
	}
	
	/** Gets the AcrossReplicationStatisticIfc for the provided name. If there is no
	 *  response variable with the provided name then null is returned.
	 *  
	 *  @param name name of response variable in the model, must not be null
	 *  @return the associated across replication statistic information, may be null if
	 *   provided name does not exist in the model
	 */
	public final AcrossReplicationStatisticIfc getAcrossReplicationResponseVariable(String name){
        if (name == null)
              throw new IllegalArgumentException("The name must be non-null.");

        if(!myModelElementMap.containsKey(name))
              return(null);
       
        ModelElement v = myModelElementMap.get(name);
       
        if (v instanceof ResponseVariable)
              return((AcrossReplicationStatisticIfc)v);
        else
              return(null);
	}
	
	/** Returns the Counter associated with the name or null if named
	 *  element is not in the model
	 *  @param name The name of the Counter model element
	 *  @return the associated counter, may be null if
	 *   provided name does not exist in the model
	 */
	public final CounterActionIfc getCounter(String name){
		if (name == null)
			throw new IllegalArgumentException("The name must be non-null.");

		if(!myModelElementMap.containsKey(name))
			return(null);
		
		ModelElement v = myModelElementMap.get(name);
		
		if (v instanceof Counter)
			return((CounterActionIfc)v);
		else
			return(null);
	}

	/** Checks to see if the model element has been registered with the Model
	 *  using it's uniquely assigned name.
	 *  
	 * @param modelElementName
	 * @return
	 */
	public final boolean containsModelElement(String modelElementName){
		return(myModelElementMap.containsKey(modelElementName));
	}

	/** Returns the model element associated with the name.  In some sense,
	 *  this model violates encapsulation for ModelElements, but since the client
	 *  knows the unique name of the ModelElement, we assume that the client
	 *  knows what he/she is doing.
	 *  
	 *  @param name The name of the model element, the name must not be null
	 *  @return The named ModelElement if it exists in the model by the supplied name, null otherwise
	 */
	public final ModelElement getModelElement(String name){
		if (name == null)
			throw new IllegalArgumentException("The name must be non-null.");

		if(!myModelElementMap.containsKey(name))
			return(null);
		
		ModelElement v = myModelElementMap.get(name);
		
		return(v);
	}
	
	/** Returns the model element that has the provided unique id or null if not found
	 * 
	 * @param id
	 * @return
	 */
	public final ModelElement getModelElement(int id){
		
		for(Map.Entry<String,ModelElement> entry: myModelElementMap.entrySet()){
			ModelElement me = entry.getValue();
			if (me.getId() == id)
				return me;
		}
		
		return null;
	}
	
	/** Gets all instances of the Variable class within the model
	 *  hierarchy. Only Variable objects (not its subclasses)
	 * @param c a Collection to hold the instances of Variable found
	 */
	public final void getVariables(Collection<Variable> c){
 		getAllVariables(c);
 	}

	/** Returns an unmodifiable collection holding all the Variables
	 *  within the Model.  The list is unchangeable, but the elements in
	 *  the list can be still be accessed and changed as necessary
	 * 
	 * @return An unmodifiable List of the variables
	 */
	public final List<Variable> getVariables(){
		return(Collections.unmodifiableList(myVariables));
	}
	
	/** Gets all instances of the RandomVariable class within the model
	 *  hierarchy. All instances of RandomVariable will be returned including
	 *  any instances of subclasses of RandomVariable
	 * @param c a Collection to hold the instances of RandomVariable found
	 */
	public final void getRandomVariables(Collection<RandomVariable> c){
		getAllRandomVariables(c);
	}
 
	/** Returns an unmodifiable collection holding all the RandomVariables
	 *  within the Model.  The list is unchangeable, but the elements in
	 *  the list can be still be accessed and changed as necessary
	 * 
	 * @return An unmodifiable List of the random variables
	 */
	public final List<RandomVariable> getRandomVariables(){
		return(Collections.unmodifiableList(myRandomVariables));
	}
	
	/** Causes RandomVariables that have been added to the model
	 *  to immediately turn on their antithetic generating streams.
	 */
	public final void turnOnAntithetic(){
		for(RandomVariable rv: myRandomVariables)
			rv.setAntithetic(true);
	}

	/** Causes RandomVariables that have been added to the model
	 *  to immediately turn off their antithetic generating streams.
	 */
	public final void turnOffAntithetic(){
		for(RandomVariable rv: myRandomVariables)
			rv.setAntithetic(false);
	}

	/** Causes RandomVariables that have been added to the model
	 *  to immediately reset their random number streams to the 
	 *  next substream in their stream.
	 */
	public final void resetNextSubstream(){
		for(RandomVariable rv: myRandomVariables)
			rv.resetNextSubstream();
	}

	/** Causes RandomVariables that have been added to the model
	 *  to immediately reset their random number streams to the 
	 *  beginning of their starting stream.
	 */
	public final void resetStartStream(){
		for(RandomVariable rv: myRandomVariables)
			rv.resetStartStream();
	}

	/** Causes RandomVariables that have been added to the model
	 *  to immediately reset their random number streams to the 
	 *  beginning of their current sub stream.
	 */
	public final void resetStartSubStream(){
		for(RandomVariable rv: myRandomVariables)
			rv.resetStartSubstream();
	}

	/** Fills up the provided collection with all of the response variables
	 *  that are contained by any model elements within the model.
	 *  In other words, any response variables (and subclasses, e.g. TimeWeighted
	 *  that are in the model element hierarchy below the model.
	 * @param c The collection to be filled.
	 */
	public final void getResponseVariables(Collection<ResponseVariable> c) {
		getAllResponseVariables(c);
	}
 
	/** Returns an unmodifiable collection holding all the ResponseVariables
	 *  within the Model.  The list is unchangeable, but the elements in
	 *  the list can be still be accessed and changed as necessary
	 * 
	 * @return An unmodifiable List of the variables
	 */
	public final List<ResponseVariable> getResponseVariables(){
		return(Collections.unmodifiableList(myResponseVariables));
	}

	/** Fills up the provided collection with all of the Counters
	 *  that are contained by any model elements within the model.
	 *  In other words, any Counters
	 *  that are in the model element hierarchy below the model.
	 * @param c The collection to be filled.
	 */
	public final void getCounters(Collection<Counter> c) {
		getAllCounters(c);
	}
	
	/** Returns an unmodifiable collection holding all the Counters
	 *  within the Model.  The list is unchangeable, but the elements in
	 *  the list can be still be accessed and changed as necessary
	 * 
	 * @return An unmodifiable List of the Counters
	 */
	public final List<Counter> getCounters(){
		return(Collections.unmodifiableList(myCounters));
	}
	
	/** Adds the supplied observer to the ModelElement with the given name.  While
	 *  any model element can be observed the client needs a reference to tell the model
	 *  element to add an observer. This method provides way to add observers to any ModelElement
	 *  regardless of where it is in the model hierarchy, provided the client knows the unique
	 *  name of the ModelElement.
	 *  
	 *  @param name The name of the model element
	 *  @param observer The observer of the named model element
	 *  @return  If this method returns false, then no such named model element exists in the Model
	 */
	public final boolean addModelElementObserver(String name, Observer observer){
		if (name == null)
			throw new IllegalArgumentException("The name must be non-null.");

		if (observer == null)
			throw new IllegalArgumentException("The observer must be non-null.");

		if(!myModelElementMap.containsKey(name))
			return(false);
		
		ModelElement me = myModelElementMap.get(name);
		
		me.addObserver(observer);

		return(true);
	}

	/** Removes the supplied observer from the ModelElement with the given name.  While
	 *  any model element can be observed the client needs a reference to tell the model
	 *  element to remove an observer. This method provides way to remove observers from any ModelElement
	 *  regardless of where it is in the model hierarchy, provided the client knows the unique
	 *  name of the ModelElement.
	 *  
	 *  @param name The name of the model element
	 *  @param observer The observer of the named model element
	 *  @return  If this method returns false, then no such named model element exists in the Model
	 */
	public final boolean deleteModelElementObserver(String name, Observer observer){
		if (name == null)
			throw new IllegalArgumentException("The name must be non-null.");

		if (observer == null)
			throw new IllegalArgumentException("The observer must be non-null.");

		if(!myModelElementMap.containsKey(name))
			return(false);
		
		ModelElement me = myModelElementMap.get(name);
		
		me.deleteObserver(observer);

		return(true);
	}

	/** Turns on the JSL log report. This report yields a text file of model
	 * (based) events:
	 *  beforeExperiment
	 *  initialize
	 *  beforeReplication
	 *  warm up
	 *  afterReplication
	 * 	afterExperiment
	 * for the model.
	 * 
	 *  WARNING: This log report only logs the actions for model elements 
	 *  that were added to the model prior to the call of this method.  
	 *  ModelElements added after this method is called will not be logged on the 
	 *  report. Because every model element is observed, this method carries 
	 *  a significant execution overhead.
	 * 
	 *  The JSL.LOGGER logs actions for all model elements
	 */
	public final void turnOnLogReport(){
		turnOnLogReport(null);
	}
	
	/** Turns on the JSL log report. This report yields a text file of model
	 * (based) events:
	 *  beforeExperiment
	 *  initialize
	 *  beforeReplication
	 *  warm up
	 *  afterReplication
	 * 	afterExperiment
	 * for the model.
	 * 
	 *  WARNING: This log report only logs the actions for model elements 
	 *  that were added to the model prior to the call of this method.  
	 *  ModelElements added after this method is called will not be logged on the 
	 *  report. Because every model element is observed, this method carries 
	 *  a significant execution overhead.
	 * 
	 *  The JSL.LOGGER logs actions for all model elements
     *
     * @param name
     */
	public final void turnOnLogReport(String name){
		myLogReportFlag = true;
        
		if(myLogReport == null){ // has not been made, so make it
			try{
				if (name != null)
					myLogReport = new LogReport(name + " " + getName()+ " LogReport");
				else
					myLogReport = new LogReport(getName()+ " LogReport");					
				addObserverToModelElements(myLogReport);
			}
			catch(IOException e){
				System.out.println("I/O exception opening " + myLogReport.getName());
				System.exit(1);
			}
		}
	}
    
	/** Turns off log reporting.
	 */
	public final void turnOffLogReport(){
		if (myLogReportFlag == true){
			deleteObserverFromModelElements(myLogReport);
			myLogReportFlag = false;
		}
	}
    
	/** Returns a reference to the current log report.
	 * @return The log report
	 */
	public final LogReport getLogReport(){
		return(myLogReport);
	}
	
	/** Returns a reference to the default entity type
	 * @return
	 */
	public final EntityType getDefaultEntityType() {
		return myDefaultEntityType;
	}
	
	/** Returns an Iterator to all the ModelElements registered
	 *  inside this model
	 * 
	 * @return
	 */
	public final Iterator<ModelElement> getModelElementIterator(){
		return(myModelElementMap.values().iterator());
	}
	
	/** Returns the model hierarchy as a string of the form
	 * 
	 *  parent name: \t name parent id: \t id \t child name: \t name \t child id: \t id \n
	 * 
	 *  Useful for debugging etc.
	 *  
	 * @return
	 */
	public final String getModelHierarchyAsString(){
		StringBuffer s = new StringBuffer();
		
		for(ModelElement m: myModelElementMap.values()){
			ModelElement parent = m.getParentModelElement();
			if (parent != null)
				s.append("model: \t " + m.getModel().getName() + "\t parent type: \t" + parent.getClass().getSimpleName() + "\t parent name: \t" + parent.getName() + "\t parent id: \t" + parent.getId() + 
						"\t child type: \t" + m.getClass().getSimpleName() + "\t child name: \t" + m.getName() + "\t child id: \t" + m.getId() + "\n");
			else
				s.append("model: \t " + m.getModel().getName() + "\t parent type: \t" + "null" + "\t parent: \t" + "null" + "\t child: \t" + m.getName() + "\t child id: \t" + m.getId()  + "\n");				
		}
		return(s.toString());
	}
	
	/** This method can be used to ensure that all model elements within the model use the same spatial
	 *  model
	 * 
	 * @param model
	 */
	public final void setSpatialModelForAllElements(SpatialModel model){
		
		//set the model's spatial model
		setSpatialModel(model);
		// iterate through all elements and set their spatial model
		for(ModelElement m: myModelElementMap.values()){
			m.setSpatialModel(model);
		}
	}

	/** Places all model elements in the model that are marked for removal into
	 *  the supplied collection
	 * 
	 * @param c
	 */
	public final void getAllElementsMarkedForRemoval(Collection<ModelElement> c){
		if (c == null)
			throw new IllegalArgumentException("The supplied collection was null");
		getAllElementsNeedingRemoval(c);
	}

	/** Indicates whether or not the model will automatically remove elements that
	 *  are marked for removal prior to each replication.  The default is false.
	 *  
	 * @return the option
	 */
	public final boolean getAutoRemoveMarkedElementsOption() {
		return myAutoRemoveMarkedElementsOption;
	}

	/** Sets the option for the model to automatically remove elements that
	 *  are marked for removal prior to each replication.  False means elements
	 *  will not be automatically removed.  True means that they will automatically
	 *  be removed.
	 * @param flag
	 */
	public final void setAutoRemoveMarkedElementsOption(boolean flag) {
		myAutoRemoveMarkedElementsOption = flag;
	}

    /** If a model element is added to the model after a replication has started
     *  this is likely to be a conceptual error. A warning message will be printed
     *  to the console in this case.  The user can turn off this warning message 
     *  by setting this flag to false.  The default is true.
     *  
	 * @return the option
	 */
	public final boolean getElementAddedAfterReplicationStartedWarningOption() {
		return myElementAddedAfterReplicationStartedWarningOption;
	}

    /** If a model element is added to the model after a replication has started
     *  this is likely to be a conceptual error. A warning message will be printed
     *  to the console in this case.  The user can turn off this warning message 
     *  by setting this flag to false.  The default is true.
     *  
	 * @param flag
	 */
	public final void setElementAddedAfterReplicationStartedWarningOption(boolean flag) {
		myElementAddedAfterReplicationStartedWarningOption = flag;
	}

    //  ===========================================
    //      PROTECTED METHODS
    //  ===========================================

	/** Sets the reset start stream option for all random variables in the model
	 *  to the supplied value, true is the default behavior.  This method
	 *  is used by an experiment prior to beforeExperiment_() being called
	 *  Thus, any random variables must already have been created and attached to 
	 *  the model elements
	 * @param option The option, true means to reset prior to each experiment
	 */
	protected final void setAllRVResetStartStreamOptions(boolean option) {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINER))
			JSL.LOGGER.finer("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t setAllRVResetStartStreamOptions()\t OPTION:\t" + option);		

		for(RandomVariable rv: myRandomVariables){
			rv.setResetStartStreamOption(option);
		}
	}

	/** Sets the reset next sub stream option for all random variables in the model
	 *  to the supplied value, true is the default behavior. True implies that
	 *  the substreams will be advanced at the end of the replication. This method
	 *  is used by an experiment prior to beforeExperiment_() being called
	 *  Thus, any random variables must already have been created and attached to 
	 *  the model elements
	 * @param option The option, true means to reset prior to each replication
	 */
	protected final void setAllRVResetNextSubStreamOptions(boolean option) {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINER))
			JSL.LOGGER.finer("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t setAllRVResetNextSubStreamOptions()\t OPTION:\t" + option);		
		
		for(RandomVariable rv: myRandomVariables){
			rv.setResetNextSubStreamOption(option);
		}
	}
    
	/** Sets the batch interval size.
	 * @param batchInterval The batch interval size in time units must be >=0, if it is larger than
	 * run length it will not occur
	 */
	protected final void setBatchInterval(double batchInterval) {
		if (batchInterval < 0.0)
			throw new IllegalArgumentException("The batch interval cannot be less than zero");
		myBatchInterval = batchInterval;
		
		if (myBatchInterval == 0.0)
			myDefaultBatchingOption = true;
		else
			myDefaultBatchingOption = false;
		
		if (isDefaultBatchingOptionOn()){
			for(ResponseVariable r: myResponseVariables){
				if (r.isDefaultBatchingOn()){
					r.setBatchingOption(true);
				}
			}
		}
	}
	
	/** Sets the current replication
	 * @param replication
	 */
	protected final void setReplication(Replication replication) {
		if (replication == null)
			throw new IllegalArgumentException("The replication must be non-null.");
		myReplication = replication;
	}

	protected void beforeExperiment() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t beforeExperiment()");		
	}
    
	/** Used to remove any flagged model elements from the model prior to a replication
	 * 
	 *
	 */
	protected void removeModelElementsMarkedForRemovalPriorToReplication(){
//		System.out.println("In Model's removeModelElementsMarkedForRemovalPriorToReplication()");
//		System.out.println("------ Hierarchy before remove");
//		System.out.println(getModelHierarchyAsString());

		List<ModelElement> toRemove = new ArrayList<ModelElement>();
		// find all that need to be removed
		getAllElementsNeedingRemoval(toRemove);
		
		// now tell them to remove themselves from the model as well as their children
		for(ModelElement m: toRemove){
			m.removeFromModel();
		}
		
		toRemove.clear();
//		System.out.println("--------- Hierarchy after remove");
//		System.out.println(getModelHierarchyAsString());
	}
	
	/** Checks if there exists at least one model element that was added
	 *  during the replication.  True means that at least one exists.
	 * 
	 * @return
	 */
	public final boolean checkForModelElementsAddedWhileReplicationWasRunning(){
		boolean found = false;
		
		for(ModelElement m: myModelElementMap.values()){
			if (m.getAddedWhileReplicationWasRunningFlag())
				return true;
		}
		return found;
	}
	
	/** Returns as a String the names of the model elements returned by
	 * getAllElementsNeedingRemoval()
	 * 
	 * @return
	 */
	public final String getModelElementsMarkedForRemovalPriorToReplicationAsString(){
		List<ModelElement> toRemove = new ArrayList<ModelElement>();
		// find all that need to be removed
		getAllElementsNeedingRemoval(toRemove);
		
		if (toRemove.isEmpty()){
			toRemove = null;
			return null;
		}
		
		StringBuilder sb = new StringBuilder();
		for(ModelElement m: toRemove){
			sb.append(m.getName() + "\n");
		}
		return(sb.toString());
	}
	
	protected void beforeReplication() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t beforeReplication() \t REPLICATION: \t " + myReplication.getReplicationNumber());		
		
		// if the replication is finite then schedule the ending event
		double length = myReplication.getLengthOfReplication();
		if  (!Double.isInfinite(length)){ // run length is finite
			if (length > 0.0) // finite and > 0, there is a rep length so schedule it
				scheduleEndOfReplication();
		}
		
		// if a batch interval has been set then schedule the first batching event
		if (getBatchInterval() > 0) {
			// the batch interval is > 0, ==> element wants a batch event
			// schedule the batch event
			myBatchActionListener = new BatchActionListener();
			myBatchEvent = getScheduler().scheduleEvent(this, myBatchActionListener, getBatchInterval(), 
			                             getName() + " Batch", myBatchEventPriority, null);			
		}
	}   
	
	protected void initialize() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t initialize()");		
	}

	protected void montecarlo() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t montecarlo()");		
	}

	protected void warmUp() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t warmUp()\t TIME: \t" + getTime());
		
		if (isBatchEventScheduled()) {
			/* warmup time is now essentially time zero, cancel any previously
			 scheduled batch event, and restart batching
			 */
			getScheduler().cancel(myBatchEvent);
			myBatchEvent = getScheduler().scheduleEvent(this, myBatchActionListener, getBatchInterval(),
			                             getName() + " Batch ", myBatchEventPriority, null);
		}
	}

	protected void afterReplication() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t afterReplication() \t REPLICATION: \t" + myReplication.getReplicationNumber());		
	}

	protected void afterExperiment() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t afterExperiment()");		
	}	
	
	/** Adds the model element to the Model's model element map
	 *  This method is called from the model element's constructor
	 * 
	 * @param modelElement
	 */
	protected final void addToModelElementMap(ModelElement modelElement){
		if (modelElement == null)
			throw new IllegalArgumentException("The ModelElement must be non-null.");
		
		String key = modelElement.getName();

		if (key == null)
			throw new IllegalArgumentException("The ModelElement Unique key must be non-null.");

		if(myModelElementMap.containsKey(key)){
			String s = "A ModleElement with the name: " + modelElement.getName() + " has already been added to the Model.\n";
			s = s + "Every model element must have a unique name";
			throw new IllegalArgumentException(s);
		}
		
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINER)){
			JSL.LOGGER.finer("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t addToModelElementMap()");		
			JSL.LOGGER.finer("Adding " + key + " to Model " + getName() + "'s ModelElementMap." );	
		}
		
		myModelElementMap.put(key, modelElement);
		
		if (modelElement instanceof ResponseVariable)
			myResponseVariables.add((ResponseVariable)modelElement);
		
		if (modelElement instanceof Counter)
			myCounters.add((Counter)modelElement);
		
		if (modelElement instanceof BatchIfc)
			myBatchVariables.add((BatchIfc)modelElement);
		
		if (modelElement instanceof RandomVariable)
			myRandomVariables.add((RandomVariable)modelElement);
		
		if (modelElement instanceof Variable)
			if (Variable.class == modelElement.getClass())
				myVariables.add((Variable)modelElement);
		
		Replication r = getReplication();
		if (r != null){
			if (r.isRunning()){
				if (getElementAddedAfterReplicationStartedWarningOption()){
					System.out.println("WARNING: The model element: " + modelElement.getName() + " was added after the replication started");
					System.out.println("WARNING: To not see this warning message call setElementAddedAfterReplicationStartedWarningOption(false) on Model");					
				}
				modelElement.setAddedWhileReplicationWasRunningFlag(true);
			}
		}
	}
	
	/** Removes the given model element from the Model's model element map.
	 *  Any child model elements of the supplied model element are also
	 *  removed from the map, until all elements below the given model element
	 *  are removed.
	 * 
	 * @param modelElement
	 */
	protected final void removeFromModelElementMap(ModelElement modelElement){
		if (modelElement == null)
			throw new IllegalArgumentException("The ModelElement must be non-null.");
		
		String key = modelElement.getName();
		
		if (myModelElementMap.containsKey(key)){
			//	remove the associated model element from the map, if there
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINER)){
				JSL.LOGGER.finer("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
					+ "\t METHOD: \t removeFromModelElementMap()");		
				JSL.LOGGER.finer("Removing " + key + " from Model " + getName() + "'s ModelElementMap." );
			}
			myModelElementMap.remove(key);
			
			if (modelElement instanceof ResponseVariable)
				myResponseVariables.remove((ResponseVariable)modelElement);
			
			if (modelElement instanceof Counter)
				myCounters.remove((Counter)modelElement);
			
			if (modelElement instanceof BatchIfc)
				myBatchVariables.remove((BatchIfc)modelElement);
			
			if (modelElement instanceof RandomVariable)
				myRandomVariables.remove((RandomVariable)modelElement);
			
			if (modelElement instanceof Variable)
				if (Variable.class == modelElement.getClass())
					myVariables.remove((Variable)modelElement);
			
			// remove any of the modelElement's children and so forth from the map
			Iterator<ModelElement> i = modelElement.getChildModelElementIterator();
			ModelElement m;
			while (i.hasNext()) {
				m = (ModelElement) i.next();
				removeFromModelElementMap(m);
			}
		}
	}
	
	/** Used by ModelElement to assist with changing the model element's parent.
	 *  This method ensures that the model has model element and its children
	 *  registered in its map after the change.
	 * 
	 * @param element
	 */
	protected final void addModelElementAndChildrenToModelElementMap(ModelElement element){
		if (element == null)
			throw new IllegalArgumentException("The ModelElement must be non-null.");
		
		String key = element.getName();

		if (key == null)
			throw new IllegalArgumentException("The ModelElement Unique key must be non-null.");

		if(myModelElementMap.containsKey(key))
			throw new IllegalArgumentException("The ModelElement has been already added to the Map.");
		
		// make sure that the element knows its model
		element.setModel(this);		
		
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINER)){
			JSL.LOGGER.finer("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t addModelElementAndChildrenToModelElementMap()");		
			JSL.LOGGER.finer("Adding " + key + " to Model " + getName() + "'s ModelElementMap." );
		}
		
		// put the element in the map using its key
		myModelElementMap.put(key, element);
		
		// recursively add any of the modelElement's children and so forth to the map
		Iterator<ModelElement> i = element.getChildModelElementIterator();
		ModelElement m;
		while (i.hasNext()) {
			m = (ModelElement) i.next();
			addModelElementAndChildrenToModelElementMap(m);
		}
		
	}

	/** The batch method is called during each replication when the batching event occurs
	 *  This method ensures that each contained model element that participates in the global
	 *  batching event will be called
	 */
	protected void batch() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
			JSL.LOGGER.finest("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t batch() \t TIME:\t" + getTime() + " \t BATCH OPTION:\t " + isDefaultBatchingOptionOn());

		// if we get here the batch event had been scheduled, so batch all the variables
		// that have their default batch option on
		for(BatchIfc b: myBatchVariables){
			b.batch();
		}
		
	}

    //  ===========================================
    //      PRIVATE METHODS
    //  ===========================================
    
	/** Used to schedule the end of a replication for the Model
	 * 
	 */
	private final void scheduleEndOfReplication(){
		double length = myReplication.getLengthOfReplication();
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t scheduleEndOfReplication() \t SCHEDULED TIME \t" + length);		
		myEndReplicationEvent = getScheduler().scheduleEvent(this, myEndReplicationActionListener,
				length, "End Replication", JSLEvent.DEFAULT_PRIORITY+1, null);
	}
    
    //  ===========================================
    //      INNER CLASSES
    //  ===========================================

	private class BatchActionListener implements ActionListenerIfc {
		/** The action to perform at a scheduled batching event
		 * @param event The event that triggered the TimedUpdate action.
		 */
		public void action(JSLEvent event) {
			batch();
			getScheduler().reschedule(event, getBatchInterval());
		}
	}
	
    private class EndReplicationActionListener implements ActionListenerIfc{
        public void action(JSLEvent event){
 		      stopReplication();
        }
    }
    
}
