/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling;

import java.util.Date;

/** This class sets project information 
 */
public class Project {
    //  ===========================================
    //      CLASS AND OBJECT ATTRIBUTES
    //  ===========================================
    
    private String myAnalyst;
    private String myTitle;
    private Date myDate;
        
    //  ===========================================
    //        CONSTRUCTORS
    //  ===========================================
    
    public Project(){
    		this(null, null, new Date());
    }
    
    /** Creates an instance of Project with the given name
     * @param title The title or name of the project
     *
     */
    public Project(String title){
        this(null, title, new Date());
    }
    
    /** Creates an instance of project with the given name
     * @param title The title or name of the project
     *
     * @param analyst The analyst for the project.
     */
    public Project(String analyst, String title){
        this(analyst, title, new Date());
    }
    
    /** Creates an instance of project with the given name
     * @param analyst The analyst for the project.
     * @param title The title or name of the project.
     * @param date The date the project was created.
     */
    public Project(String analyst, String title, Date date){
        setAnalystName(analyst);
        setTitle(title);
        setDate(date);
     }
    
	//  ===========================================
    //        PUBLIC INTERFACE
    //  ===========================================
    
    /** Runs one replication of the simulation with the specified run length
     * 
     * @param replicationLength
     */
    public void runSimulation(double replicationLength){
    		runSimulation(1, replicationLength, 0.0);
    }
    
    /** Runs one replication of the simulation with the specified run length, and warmup length
     * 
     * @param replicationLength
     * @param warmupLength
     */
    public void runSimulation(double replicationLength, double warmUpLength){
    		runSimulation(1, replicationLength, warmUpLength);
    }
    
    /** Runs the simulation with the specified number of replications, run length
     * 
     * @param numberOfReps
     * @param replicationLength
     */
    public void runSimulation(int numberOfReps, double replicationLength){
    		runSimulation(numberOfReps, replicationLength, 0.0);
    }
    
    /** Runs the simulation with the specified number of replications, run length, and warmup length
     * 
     * @param numberOfReps
     * @param replicationLength
     * @param warmupLength
     */
    public void runSimulation(int numberOfReps, double replicationLength, double warmupLength){
    		Model m = buildModel(Model.createModel());
        Experiment e = new Experiment(m);
   		e.setLengthOfReplication(replicationLength);
    		e.setNumberOfReplications(numberOfReps);
    		e.setLengthOfWarmUp(warmupLength);
    		e.runAll();
    }
    
    /** Sets the analyst for the project
     * @param analyst The analyst for the project.
     */
    public void setAnalystName(String analyst){
        myAnalyst = analyst;
    }
    
    /** Sets the title for the project
     * @param title The title for the project.
     */
    public void setTitle(String title){
        myTitle = title;
    }
    
    /** Sets the creation date for the project
     * @param date The creation date for the project.
     */
    public void setDate(Date date){
        myDate = date;
    }
    
    /*  --Accessors---------------------- */
    
    /** Gets the analyst for the project
     * @return The analyst for the project.
     */
    public final String getAnalystName(){
        return(myAnalyst);
    }
    
    /** Gets the title for the project
     * @return The title for the project.
     */
    public final String getTitle(){
        return(myTitle);
    }
    
    /** Gets the creation date for the project
     * @return The creation date for the project.
     */
    public final Date getDate(){
        return(myDate);
    }
 
	//  ===========================================
    //        PUBLIC INTERFACE
    //  ===========================================

    /** The user must override this method to build a valid 
     *  Model
     *  
     * @param The model to attache model elements to
     */
    protected Model buildModel(Model model) {
 
        return(model);
	}
}
