/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.TimerTask;

import jsl.modeling.conditions.*;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.utilities.reporting.JSL;
import jsl.utilities.statistic.*;
import jsl.observers.textfile.*;
import jsl.observers.variable.StatisticalObserver;

/** 
 * The Replication class represents an individual replication within an
 * experiment.  That is, a setting of the experiment that can be repeated.
 * 
 */
public class Replication extends IterativeProcess<JSLEvent> {

	//  ===========================================
	//      CLASS ATTRIBUTES
	//  ===========================================

	/** A constant for the default batch interval for a replication
	 *  If there is no run length specified and the user turns on
	 *  default batching, then the time interval between batches will
	 *  be equal to this value.  The default value is 10.0
	 */
	public static final double DEFAULT_BATCH_INTERVAL = 10.0;

	/** A constant for the default batching.  This value is used in the calculation of
	 *  the approximate batching interval if batching is turned on and there is
	 *  a finite run length.
	 *  
	 *  If the run length is finite then the batch interval is approximated
	 *  as follows
	 * 
	 *		t = length of replication - length of warmup; // actual observation length
	 *		n =	myBatchingConstant*BatchStatistic.MIN_NUM_BATCHES*BatchStatistic.MAX_BATCH_MULTIPLE;
	 *		batch interval = t / n;
	 *
	 * myBatchConstant = DEFAULT_BATCH_CONSTANT = 4.0 unless changed by the user.
	 */
	public static final double DEFAULT_BATCH_CONSTANT = 4.0;

	//  ===========================================
	//      OBJECT ATTRIBUTES
	//  ===========================================

	/** A reference to the experiment for this replication
	 */
	private Experiment myExperiment;

	/** A reference to the model for this replication
	 */
	private Model myModel;

	/** A reference to the scheduler
	 */
	private Scheduler myScheduler;
	
	/** The number (index) of this replication in the experiment
	 */
	private double myRepNum;

	/** The specified length of each planned replication for
	 *  this experiment.  The default is Double.POSITIVE_INFINITY.
	 */
	private double myLengthOfReplication;

	/** The actual length of the replication.  This may be 
	 *  different than planned due to how the replication ends
	 */
	private double myEndingLengthOfReplication;

	/** The length of time from the start of the replication to
	 * the warm up event for the replication.
	 */
	private double myLengthOfWarmUp; // zero is no warmup

	/** A flag to indicate whether or not the model 
	 *  should be re-initialized at the beginning of
	 *  this replication. True means that it will be re-initialized.
	 */
	private boolean myRepInitOption;

	/** A flag to indicate whether or not the replication will
	 *  have the model be batched. Turning on this option will
	 *  cause the model to use a batching interval, either user defined
	 *  or provided by default.
	 */
	private boolean myDefaultBatchingOption;

	/** A time interval (in simulated time) that represents
	 *  the default time between batches
	 */
	private double myDefaultBatchInterval;

	/** Used in approximating a batch interval size
	 */
	private double myBatchingConstant;
	
	/** Used to indicate whether replication reporting is on
	 */
	private boolean myBatchReportOption;

	/** Used to report with replication statistics
	 */
	private BatchReport myBatchReport;

	/** The reset start stream option
	 *  This option indicates whether or not the random variables used
	 *  during the replication will be reset to their starting stream
	 *  prior to running the replication.  The default is FALSE.
	 *  This ensures that the random variables WILL NOT be reset
	 *  to the start of their current stream prior to running the same replication again
     *  on the SAME model or a different replication on the SAME model.
     *  Thus, reusing the replication or creating
     *  a new replication that uses the same model, will cause the 
     *  random variables to continue within their same stream. The results
     *  from running replications in the SAME program invocation will thus
     *  be statistically independent.
	 */
	private boolean myResetStartStreamOption;

	/** The reset next sub stream option
	 *  This option indicates whether or not the random variables used
	 *  during the replication will be reset to their next substream
	 *  after running each replication.  The default is TRUE.  This
	 *  ensures that the random variables will be jump to the next
	 *  substream within their current stream at the end of a replication.  This will cause
	 *  the random variables in each subsequent replication to start in the same substream
	 *  in the underlying random number streams if the replication is repeatedly used and
	 *  the ResetStartStreamOption is set to false (which is the default). Otherwise, this
	 *  option really has no effect if a replication is only run once.
	 *  If the replication is reused and ResetStartStreamOption is set to false
	 *  the replication will continue in the same stream, but at the beginning
	 *  of the next substream.  If ResetStartStreamOption is set to true, the
	 *  replication will first return to the beginning of its current stream
	 *  and then jump to the next substream (if this option is on).  
	 *  Having ResetNextSubStreamOption true assists in sychronizing the
	 *  random number draws from one replication to another aiding in
	 *  the implementation of common random numbers.  Each replication
	 *  within the same experiment is still independent.
	 */
	private boolean myResetNextSubStreamOption;

	/** Used to indicate whether monte carlo option
	 *  is on for the replication
	 */
	private boolean myMonteCarloOption;
	
	/** Used to indicate whether or not statistics related to the replication
	 *  should be saved.  The default is false
	 */
	private boolean mySaveStatisticsOption = false;
	
	/** Used to hold statistics associated with the replication if the
	 *  save statistics option is true.
	 */
	private List<AbstractStatistic> myStatistics;
	
	/** Constructs a replication with the a default new model with a default scheduler and
	 *  ending condition
	 *  
	 *  The getModel() method can be used to
	 *  access the model prior to running the replication.
	 *  
	 *  The defaults include:
	 *   length of replication = Double.POSITIVE_INFINITY
	 *   length of warmup = 0.0 // no warm up
	 *   length of batch interval = 0.0 // no batch interval
	 *   replication initialization TRUE
	 *   default batching option TRUE
	 */
	public Replication() {
		this(Model.createModel(), new Scheduler(), new Condition("End Replication Condition"));
	}
	
	/** Constructs a replication with the given model with a default scheduler and
	 *  ending condition
	 *  The defaults include:
	 *   length of replication = Double.POSITIVE_INFINITY
	 *   length of warmup = 0.0 // no warm up
	 *   length of batch interval = 0.0 // no batch interval
	 *   replication initialization TRUE
	 *   default batching option TRUE
	 * @param model A reference to the model for this replication
	 */
	public Replication(Model model) {
		this(model, new Scheduler(), new Condition("End Replication Condition"));
	}

	/** Constructs a replication with the given model the supplied scheduler and
	 *  given ending condition
	 *  The defaults include:
	 *   length of replication = Double.POSITIVE_INFINITY
	 *   length of warmup = 0.0 // no warm up
	 *   length of batch interval = 0.0 // no batch interval
	 *   replication initialization TRUE
	 *   default batching option TRUE
	 * @param model A reference to the model for this replication
	 */
	public Replication(Model model, ConditionIfc condition) {
		this(model, new Scheduler(), condition);
	}

	/** Constructs a replication with the given model.
	 *  The defaults include:
	 *   length of replication = Double.POSITIVE_INFINITY
	 *   length of warmup = 0.0 // no warm up
	 *   length of batch interval = 0.0 // no batch interval
	 *   replication initialization TRUE
	 *   default batching option TRUE
	 *   save statistics option FALSE
	 * @param model A reference to the model for this replication
	 * @param scheduler A reference to the scheduler for this replication
	 */
	public Replication(Model model,	Scheduler scheduler, ConditionIfc condition) {
		super();
		setModel(model);
		setScheduler(scheduler);
		setEndCondition(condition);
		myLengthOfReplication = Double.POSITIVE_INFINITY;
		myLengthOfWarmUp = 0.0; // zero means no warm up
		myDefaultBatchInterval = 0.0; // zero means no batch interval
		myRepInitOption = true;
		myDefaultBatchingOption = true;
		myResetStartStreamOption = false;
		myResetNextSubStreamOption = true;
		myMonteCarloOption = false;
		mySaveStatisticsOption = false;
		myBatchingConstant = DEFAULT_BATCH_CONSTANT;
	}

	//  ===========================================
	//        PUBLIC METHODS
	//  ===========================================

	/** Gets the model for this replication
	 * @return the model
	 */
	public final Model getModel() {
		return (myModel);
	}

	/** Gets the replication number for this replication
	 * @return the replication number
	 */
	public final double getReplicationNumber() {
		return myRepNum;
	}

	/** Sets the replication number for this replication
	 * @param the replication number > 0
	 */
	public final void setReplicationNumber(double repNum) {
		if (repNum <= 0.0)
			throw new IllegalArgumentException("Replication number was <= 0.0");

		myRepNum = repNum;
	}

	/** Returns the length of the replication as a double
	 * @return the length of the replication
	 */
	public final double getLengthOfReplication() {
		return (myLengthOfReplication);
	}

	/** Sets the length of the replication
	 *  The length of the replication must be > 0, if set
	 * 
	 * @param lengthOfReplication
	 */
	public final void setLengthOfReplication(double lengthOfReplication) {
		if (lengthOfReplication <= 0.0)
			throw new IllegalArgumentException("Simulation replication length < 0.0");
		myLengthOfReplication = lengthOfReplication;
	}

	/** Gets the length of the warm up for this replication
	 * @return the the length of the warm up
	 */
	public final double getLengthOfWarmUp() {
		return (myLengthOfWarmUp);
	}

	/** Sets the length of the warm up for each replication within this experiment
	 *  The length of the warm up must be > 0.0, if set.
	 *  If the warm up is set greater than the run length then there will be no warm up
	 *
	 * @param lengthOfWarmUp length of the warm up, must be >= 0.0
	 */
	public final void setLengthOfWarmUp(double lengthOfWarmUp) {
		if (lengthOfWarmUp < 0.0)
			throw new IllegalArgumentException("Warmup time cannot be less than zero");

		myLengthOfWarmUp = lengthOfWarmUp;
	}

	/** Returns the setting for whether or not the model will be initialized
	 *  prior to running this replication
	 * @return true means that each replication will be initialized
	 */
	public final boolean getReplicationInitializationOption() {
		return (myRepInitOption);
	}

	/** Returns the setting for whether this replication will cause
	 *  the model to be batched 
	 * @return true means that the model will be batched
	 */
	public final boolean getDefaultBatchingOption() {
		return (myDefaultBatchingOption);
	}

	/** Sets the default batching option for this replication. This controls
	 *  whether or not a standard sized batching interval will be used to 
	 *  collect statistics on time persistent variables during this replication.
	 * 
	 *  This method only indicates how individual time weighted variables that
	 *  do not control their own batching will be controlled by default during
	 *  each replication within the experiment.
	 * 
	 * @param batchOption true means batching is on.
	 */
	public final void setDefaultBatchingOption(boolean batchOption) {
		myDefaultBatchingOption = batchOption;
	}
	
	/** Returns the default batching interval for all time weighted variables
	 *  for this replication.  Individual time weighted variables may elect
	 *  to have their own batch interval.
	 * @return a double representing the interval length
	 */
	public final double getDefaultBatchInterval() {
		return (myDefaultBatchInterval);
	}

	/** Sets the default batching interval for this replication
	 * If the default batching option is on, this value will be used
	 * for the batching interval for all time weighted variables that participate
	 * in batching. If the batching interval is set greater than the run length
	 * then no batching will occur.
	 * @param batchInterval a double > 0
	 */
	public final void setDefaultBatchInterval(double batchInterval) {
		if (batchInterval <= 0.0)
			throw new IllegalArgumentException("Batching interval must be > 0");
		myDefaultBatchInterval = batchInterval;
	}
	
	/** Set batch constant used in approximating batch interval size
	 *  
	 * @param bc a real number greater than zero
	 */
	public final void setBatchingConstant(double bc){
		if (bc <= 0.0)
			throw new IllegalArgumentException("Batching contant must be > 0");
		myBatchingConstant = bc;
	}
	
	/** This method returns a suggested batching interval based on the
	 *  length of the run, the warm up period, default number of batches, and
	 *  maximum batch multiple.
	 * 
	 *  if the run length is finite then the batch interval is approximated
	 *  as follows
	 * 
	 *		t = length of replication - length of warmup; // actual observation length
	 *		n =	myBatchingConstant*BatchStatistic.MIN_NUM_BATCHES*BatchStatistic.MAX_BATCH_MULTIPLE;
	 *		batch interval = t / n;
	 *  
	 *  if the run length is infinite then the batch interval is set to 
	 *  DEFAULT_BATCH_INTERVAL
	 * 
	 * @return a double representing an approximate batch interval
	 */
	public final double getApproximateBatchInterval() {
		double deltaT = 0.0;

		if (Double.isInfinite(getLengthOfReplication()))
			// runlength is infinite
			deltaT = DEFAULT_BATCH_INTERVAL;
		else { // runlength is finite
			double t = getLengthOfReplication();
			t = t - getLengthOfWarmUp(); // actual observation length
			double n =	myBatchingConstant*BatchStatistic.MIN_NUM_BATCHES*BatchStatistic.MAX_BATCH_MULTIPLE;
			deltaT = t / n;
		}
		return (deltaT);
	}
	
	/** Gets the monte carlo option flag
	 *  if this is true, then the model's montecarlo method
	 *  will be called after, beforeReplication
	 * @return
	 */
	public final boolean getMonteCarloOption() {
		return myMonteCarloOption;
	}

	/** Sets the monte carlo option flag
	 *  if this is true, then the model's montecarlo method
	 *  will be called after, beforeReplication
	 * @param b
	 */
	public final void setMonteCarloOption(boolean b) {
		myMonteCarloOption = b;
	}

	/** True indicates that the save statistics option is on for the replication
	 * 
	 * @return Returns the option.
	 */
	public final boolean getSaveStatisticsOption() {
		return mySaveStatisticsOption;
	}

	/** Indicates whether or not the within replication and batch statistics (if they exist)
	 *  will be saved to a list by the replication. True indicates that the statistics will be saved.
	 *  
	 * @param option The option to set.
	 */
	public final void setSaveStatisticsOption(boolean option) {
		mySaveStatisticsOption = option;
	}

	/** Returns a list that holds a copy of the statistics collected during
	 *  the replication.  This method will return an empty list if the save statistic
	 *  option is not turned on.  The list holds copies, not references to the 
	 *  original collected statistics.  The list will hold both batch (if batching is on) and 
	 *  within replication statistics.
	 *  
	 * @return Returns the statistics.
	 */
	public final List<AbstractStatistic> getReplicationStatistics() {
		if (myStatistics == null)
			myStatistics = new ArrayList<AbstractStatistic>();
		
		return myStatistics;
	}

	/** Turns on the batch report. This report yields a text file of
	 *  the batch statistics within the replication 
	 */
	public final void turnOnBatchReport() {
		turnOnBatchReport(null);
	}
	
	/** Turns on the batch report. This report yields a text file of
	 *  the batch statistics within the replication 
	 *  @param name, specifies a prefix name for the report 
	 */
	public final void turnOnBatchReport(String name) {
		myBatchReportOption = true;

		if (myBatchReport == null) { // has not been made, so make it
			try {
				String s;
				if (name != null)
					s = name + " " + getName() + " BatchReport";
				else
					s = getName() + " BatchReport";
				myBatchReport = new BatchReport(s);
				addObserver(myBatchReport);
			} catch (IOException e) {
				String s = "I/O exception opening BatchReport for replication: ID = " + getId() + " Name: " + getName();
				System.err.println(s);
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.SEVERE))
					JSL.LOGGER.severe(s);
				System.exit(1);
			}
		} else {
			// the report has already been made and named
			// the user may be trying to change the name of the report
			// form the name based on the input
			String s;
			if (name != null)
				s = name + " " + getName() + " BatchReport";
			else
				s = getName() + " BatchReport";
			// check if the new name and old name are the same
			if (s.equals(myBatchReport.getName())){
				JSL.LOGGER.warning("The Batch report " + s + " will be over written." );
			}
			// names not the same, remove the old observer and attach new observer
			deleteObserver(myBatchReport);
			try {
				myBatchReport = new BatchReport(s);
				addObserver(myBatchReport);
			} catch (IOException e){
				String s1 = "I/O exception opening BatchReport for replication: ID = " + getId() + " Name: " + getName();
				System.err.println(s1);
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.SEVERE))
					JSL.LOGGER.severe(s1);
				System.exit(1);
			}	
		}
	}

	/** Turns off the replication's batch report
	*/
	public void turnOffBatchReport() {
		if (myBatchReportOption == true) {
			deleteObserver(myBatchReport);
			myBatchReportOption = false;
			myBatchReport = null;
		}
	}

	/** Returns the default batch report if it exists
	 * 
	 * @return
	 */
	public BatchReport getDefaultBatchReport(){
		return(myBatchReport);
	}
	
	/** This method tells the scheduler to turn on its
	 *  default trace report
	 */
	public final void turnOnDefaultTraceReport() {
		myScheduler.turnOnDefaultTraceReport();
	}

	/** This method tells the scheduler to turn off its default trace report
	 */
	public final void turnOffDefaultTraceReport() {
		myScheduler.turnOffDefaultTraceReport();
	}
	
	/** Gets the current simulated time
	 * @return
	 */
	public final double getTime() {
		if (myScheduler != null)
			return(myScheduler.getTime());
		else
			return(0.0);
	}

	/** Gets the actual simulated time that the replication ended
	 * @return
	 */
	public final double getEndingLengthOfReplication() {
		return myEndingLengthOfReplication;
	}

	/** Gets the reset next substream option
	 *  
	 * @return true means the option is on
	 */
	public final boolean getResetNextSubStreamOption() {
		return myResetNextSubStreamOption;
	}
 
	/** Gets the reset start stream option
	 * @return true means the option is on
	 */
	public final boolean getResetStartStreamOption() {
		return myResetStartStreamOption;
	}

	/** Sets the reset next substream option
	 * @param b
	 */
	public final void setResetNextSubStreamOption(boolean b) {
		myResetNextSubStreamOption = b;
	}

	/** Sets the reset start stream option
	 * @param b
	 */
	public final void setResetStartStreamOption(boolean b) {
		myResetStartStreamOption = b;
	}

	/** Gets the experiment associated with this replication
	 *  
	 * @return the experiment or null if not associated with an experiment
	 */
	public final Experiment getExperiment() {
		return (myExperiment);
	}

	/** Returns a string with statistics concerning the
	 *  scheduler's performance during the replication.
	 * 
	 * @return
	 */
	public final String getSchedulerStatistics(){
		if (myScheduler == null)
			return(null);
		else
		    return(myScheduler.toString());
	}
	
	/** Returns a reference to the underlying event scheduler
	 * 
	 * @return
	 */
	public final Scheduler getScheduler(){
		return(myScheduler);
	}

	/** This method will cause the a timer to start
	 *  allowing a TimerTask to be scheduled.  This method
	 *  causes a print to console task to be run
	 *  which prints the elapsed time since initializing
	 *  the replication and the current simulation time
	 * @param milliseconds
	*/
	public void turnOnTimer(long milliseconds) {
		turnOnTimer(milliseconds, new ShowElapsedSimTimeTask());
	}
	
	/** Sets the replication initialization option for the replication
	 * If set to true, this replication will initialize the model
	 * @param repInitOption true means to initialize
	 */
	protected final void setReplicationInitializationOption(boolean repInitOption) {
		myRepInitOption = repInitOption;
	}

	/** Sets the experiment for this replication
	 * 
	 * @param exp, must be non-null or an IllegalArgumentException is thrown
	 */
	protected final void setExperiment(Experiment exp) {
		if (exp == null)
			throw new IllegalArgumentException("The experiment was null");
		myExperiment = exp;
	}

	protected final void initializeIterations() {
		super.initializeIterations();
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINER))
			JSL.LOGGER.finer("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t initialize_()");		
		
		//remove any model elements marked for removal, if automatic removal is on
		if (myModel.getAutoRemoveMarkedElementsOption())
			myModel.removeModelElementsMarkedForRemovalPriorToReplication();
		else{
			// automatic removal is not on, check for any model elements added during a replication
			if (myModel.checkForModelElementsAddedWhileReplicationWasRunning()){
				JSL.LOGGER.warning("Automatic model element removal is turned off and there were model elements added during a replication.");
			}
		}
		
		if (myStatistics != null)
			myStatistics.clear();

		if (myExperiment == null) {
			// no experiment for this replication, running alone
			// must tell the model to prepare for experiment
			// running a single replication
			setReplicationNumber(1.0);
			// have the replication tell the model to set all random variables to the desired options
			// before the experiment, since only 1 replication must tell model to do it's beforeExperiment action directly
			// random variables will thus have their before experiment actions performed
			myModel.setAllRVResetStartStreamOptions(getResetStartStreamOption());
			myModel.beforeExperiment_();
			// if myExperiment != null then this has already 
			// been done by the experiment
		}

		if (Double.isInfinite(getLengthOfReplication())) {
			// run length is infinite
			if (getMaximumAllowedExecutionTime() == 0) {
				// no execution time set
				setDefaultBatchingOption(false);
				String s =
					"There was no replication length specified. \n" +
				    "There was no maximum execution time specified. \n" +
				    "The end condition has also been cleared. \n" +
				    "The user is responsible for ensuring that the end condition is set \n" +
				    "sometime during the replication; otherwise, the replication will not\n" +
				    "terminate. The default time based batching has been turned off. \n" +
				 	"To turn on the time based batching use setDefaultBatchingOption(true) prior to calling runAll()\n" +
				 	"To set the maximimum execution time use setMaximumExecutionTime(long milliseconds)\n" +
				 	"To set the run length use setReplicationLength(double runLength).\n";
				JSL.LOGGER.warning(s);
			}
		}

		myModel.setReplication(this);
		myModel.setLengthOfWarmUp(getLengthOfWarmUp());
		
		// have the model control the rv's substream option to the desired option for the replication
		myModel.setAllRVResetNextSubStreamOptions(getResetNextSubStreamOption());
		
		if (myDefaultBatchingOption == true) {
			if (getDefaultBatchInterval() == 0.0)  // has not be set yet
				setDefaultBatchInterval(getApproximateBatchInterval());
			
			myModel.setBatchInterval(getDefaultBatchInterval());
		}

		// initialize the scheduler
		if (myRepInitOption == true) {
			myScheduler.setLengthOfReplication(getLengthOfReplication());
			myScheduler.initialize();
		}

		// tell the model to do its before replication action and all sub model elements
		myModel.beforeReplication_();

		if (myRepInitOption == true) {
			// initialize the model and all model elements with initialize option on
			myModel.initialize_();
		}
	
		if (getMonteCarloOption()){
			// since monte carlo option was turned on, assume everyone wants to listen
			myModel.setMonteCarloOptionForModelElements(true);
			myModel.montecarlo_();
		}

	}

	protected final void endIterations() {
		myEndingLengthOfReplication = myScheduler.getTime();
		myModel.replicationEnded_();
		myModel.afterReplication_();
		
		if (myExperiment == null) {
			// no experiment for this replication, running alone
			// must tell the model that experiment is complete
			myModel.afterExperiment_();
			// if myExperiment != null then this will
			// be done by the experiment
		}
		
		if (getSaveStatisticsOption())
			saveStatistics();
		
		super.endIterations();
		
		if (myTimer != null){
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
				JSL.LOGGER.fine("Final simulated time = " + myScheduler.getTime());
		}
		
	}

	private final void saveStatistics(){
		if (myStatistics == null)
			myStatistics = new ArrayList<AbstractStatistic>();
		
		List<ResponseVariable> rsList = myModel.getResponseVariables();
		
		for(ResponseVariable rs: rsList){
			StatisticalObserver so = rs.getStatisticalObserver();
			Statistic within = (Statistic)so.getWithinReplicationStatistic();
			Statistic stat = Statistic.newInstance(within);
			myStatistics.add(stat);

			BatchStatistic batch = (BatchStatistic)so.getAcrossBatchStatistic();
			if (batch != null){
				BatchStatistic bstat = BatchStatistic.newInstance(batch);
				myStatistics.add(bstat);
			}
		}		
	}
	
	protected final boolean hasNext() {
		return (!myScheduler.isDone());
	}

	protected final JSLEvent next() {
		return (myScheduler.peekNext());
	}

	protected final void runStep() {
		myCurrentStep = next();
		try {
			myScheduler.executeNextEvent();						
		} catch (RuntimeException e){
			System.out.println("######################################");
			System.out.println("An exception occurred near this event:");
			System.out.println(myCurrentStep);
			System.out.println("######################################");
			System.out.println();
			throw e;
		}
		if (mySaveStepOption == true){
			// only if save option is on, remember the event
			mySteps.add(myCurrentStep);			
		}
	}

	//  ===========================================
	//      PRIVATE METHODS
	//  ===========================================

	/** Sets the model for this replication
	 * 
	 * @param model, must be non-null or an IllegalArgumentException is thrown
	 */
	private final void setModel(Model model) {
		if (model == null)
			throw new IllegalArgumentException("The model was null");
		myModel = model;
	}

	/** Sets the scheduler for this replication
	 * 
	 * @param model, must be non-null or an IllegalArgumentException is thrown
	 */
	private final void setScheduler(Scheduler scheduler) {
		if (scheduler == null)
			throw new IllegalArgumentException("The scheduler was null");
		myScheduler = scheduler;
		ModelElement.setScheduler(myScheduler);
	}

	private class ShowElapsedSimTimeTask extends TimerTask {

		/* (non-Javadoc)
		 * @see java.util.TimerTask#run()
		 */
		public void run() {
			System.out.println("Elapsed Clock Time: " + getElapsedExecutionTime()/1000 + " seconds. Simulation time > " + getTime());
		}
		
	}
}
