/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling;

import java.util.*;
import java.io.*;

import jsl.utilities.reporting.JSL;
import jsl.observers.scheduler.SchedulerTraceReport;
import jsl.calendar.*;
import jsl.utilities.IdentityIfc;

/** The Scheduler manages the event calendar. 
 */
public class Scheduler extends Observable implements IdentityIfc {

	//  ===========================================
	//      CLASS ATTRIBUTES
	//  ===========================================
	public final static int INITIALIZED = 1;
	public final static int EVENT_EXECUTED = 2;
	public final static int STOPPED = 3;
	
	/** A counter to count the number of schedulers created to assign "unique" ids
	 */
	private static int myIdCounter_;

	//  ===========================================
	//      OBJECT ATTRIBUTES
	//  ===========================================

	/** The name of this scheduler
	*/
	protected String myName;

	/** The id of this object
	 */
	private int myId;

	/** The reference to the event calendar
	 */
	private CalendarIfc myEventCalendar;

	/** The current simulated time
	 */
	private double myCurrentTime;

	/** A reference to the event that just executed
	 */
	private JSLEvent myCurrentEvent;

	/** A flag to indicate that the scheduler is
	 *  done processing events, true is done
	 */
	private boolean myDoneFlag;

	/** A flag to idicate that the scheduler
	 *  is to be traced
	 */
	private boolean myTraceReportFlag;

	/** A reference to a tracing report
	 */
	private SchedulerTraceReport myTraceReport;

	/** A counter that tracks the number of events
	 *  processed.
	 */
	private long myEventCount = 0;

	private int myObserverState;
	
	/** The specified length of each planned replication for
	 *  this experiment.  The default is Double.POSITIVE_INFINITY.
	 */
	private double myLengthOfReplication;

	//  ===========================================
	//        CONSTRUCTORS
	//  ===========================================

	public Scheduler() {
		//		this(null, new TreeSetEventCalendar());
//		this(null, new SkewHeapEventCalendar()); // a little faster than TreeSetEventCalendar()
		this(null, new PriorityQueueEventCalendar()); // a little faster than SkewHeapEventCalendar()
	}

	public Scheduler(CalendarIfc c) {
		this(null, c);
	}

	public Scheduler(String name, CalendarIfc c) {
		myIdCounter_ = myIdCounter_ + 1;
		myId = myIdCounter_;
		setName(name);
		myEventCalendar = c;
		myDoneFlag = false;
		myObserverState = 0;
		myLengthOfReplication = Double.POSITIVE_INFINITY;
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINER))
			JSL.LOGGER.finer("Class name \t" + this.getClass().getName() + "\t Name: " + getName() + "\t ID: \t " + getId() + "\t Constructed");		
	}

	//  ===========================================
	//        PUBLIC INTERFACE
	//  ===========================================

	/** Gets the name.
	* @return The name of object.
	*/
	public final String getName() {
		return myName;
	}

	/** Sets the name 
	* @param str The name as a string.
	*/
	public final void setName(String str) {
		if (str == null) {
			String s = this.getClass().getName();
			int k = s.lastIndexOf(".");
			if (k != -1)
				s = s.substring(k + 1);
			myName = s + myId;
		} else
			myName = str;
	}

	/** Returns the id for this iterative process
	 * 
	 * @return
	 */
	public long getId() {
		return (myId);
	}

	/** A flag to indicate whether the scheduler is done, i.e.
	 *  that there are no more events or that it has
	 *  been told to stop
	* @return true if done
	*/
	public final boolean isDone() {
		if (myDoneFlag == true)
			return (true);
		if (isEmpty())
			return (true);
		return (myDoneFlag);
	}

	/** This method asks the scheduler to stop executing events.
	 * After the current event executes, the scheduler will drop out
	 * of the executeEvents() method
	 */
	public final void stop() {
		myDoneFlag = true;
		myObserverState = STOPPED;
		setChanged();
		notifyObservers();
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t stop() \t TIME: \t" + getTime());		
	}

	/** Can be used by observers to check what state occurred
	 * 
	 * @return
	 */
	public final int getObserverState(){
		return(myObserverState);
	}
	
	/** This method creates an instance of the default SchedulerTraceReport
	 * and tells it to observe the scheduler.
	*/
	public final void turnOnDefaultTraceReport() {
		turnOnDefaultTraceReport(null);
	}

	/** This method creates an instance of the default SchedulerTraceReport
	 * and tells it to observe the scheduler.
	 *  @param name, specifies a prefix name for the report 
	*/
	public void turnOnDefaultTraceReport(String name) {

		if (myTraceReportFlag == false) {
			try {
				String s;
				if (name != null)
					s = name + " " + getName() + " TraceReport";
				else
					s = getName() + " TraceReport";
				myTraceReport = new SchedulerTraceReport(s);
				addObserver(myTraceReport);
				myTraceReportFlag = true;
			} catch (IOException e) {
				String s = "I/O exception opening trace report: ID = " + getId() + " Name: " + getName();
				System.err.println(s);
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.SEVERE))
					JSL.LOGGER.severe(s);
			}
		}
	}

	/** Removes the default scheduler SchedulerTraceReport as an observer
	 * of the Scheduler
	 */
	public void turnOffDefaultTraceReport() {
		if (myTraceReportFlag == true) {
			deleteObserver(myTraceReport);
			myTraceReportFlag = false;
			myTraceReport = null;
		}
	}

	/** This method allows a previously *executed* event to be reused
	 * The event must have already been removed from the calendar through the
	 * natural execute event mechanism and have been executed.
	 * 
	 * NOTE: If an event is cancelled it remains in the scheduler until
	 * its originally scheduled event time. Then and only then does it 
	 * become unscheduled at which time the event can be rescheduled.  If a
	 * client needs to reschedule a cancelled event prior to the originally
	 * scheduled event time, then just use schedule to make a new event.
	 *
	 * @param event The event that needs rescheduling
	 * @param time represents the inter-event time, i.e. the interval from the current time to when the
	 *        event will need to occur
	 */
	public final void reschedule(JSLEvent event, double time) {
		if (event.isScheduled())
			throw new IllegalArgumentException("Attempted to reschedule an already scheduled event.");
		event.setCanceledFlag(false);
		event.setTime(getTime() + time);
		schedule(event);
	}

	/** Creates an event and schedules it onto the event calendar
	 * @param dispatcher represents the Model Element that dispatches the event
	 * @param listener represents an ActionListener that will handle the change of state logic
	 * @param time represents the inter-event time, i.e. the interval from the current time to when the
	 *        event will need to occur
	 * @param priority is used to influence the ordering of events
	 * @param message is a generic Object that may represent data to be transmitted with the event
	 * @return a valid JSLEvent
	 */
	public final JSLEvent scheduleEvent(ModelElement dispatcher, ActionListenerIfc listener,
		                                double time, String name, int priority, Object message) {
		
		// create the event
		JSLEvent event = new JSLEvent();
		
		// prepare the event
		event.setDispatcher(dispatcher);
		event.setName(name);
		event.setListener(listener);
		event.setTime(getTime() + time);
		event.setPriority(priority);
		event.setMessage(message);

		schedule(event);
		return (event);
	}

	/** Tells the scheduler to cancel the provided event. The scheduler
	 * simply tells the event calendar to cancel the event.
	 * 
	 * The event must have been scheduled otherwise an IllegalArgumentException is thrown.
	 * 
	 * @param e A reference to the event to be cancelled.
	 */
	public final void cancel(JSLEvent e) {
		if (!e.isScheduled())
			throw new IllegalArgumentException("Attempted to cancel an unscheduled event.");
		myEventCalendar.cancel(e);
	}

	/** The scheduler tells the event calendar to cancel all events.
	 */
	public final void cancelAllEvents() {
		myEventCalendar.cancelAll();
	}

	/** Allows a check of the event calendar to see if it is empty,
	 * i.e. it does not have any more events. Note:  Depending on how
	 * the event calendar handles cancelled events, cancelled events may
	 * or may not count.
	 * @return True is empty, False is not empty
	 */
	public final boolean isEmpty() {
		return myEventCalendar.isEmpty();
	}

	/** Returns the current simulated time as a double
	 * @return Simulated time as a double
	 *
	 */
	public final double getTime() {
		return (myCurrentTime);
	}

	/** Currently returns only the name of the class.  Future versions may
	 * provide additional diagnostics on the scheduler's state.
	 * @return A string representing the scheduler.
	 */
	public String toString() {
		String s = "Name " + getName() + "\n" +
		           "Total number of events scheduled " + getTotalNumberEventsScheduled() + "\n";
		return (s);
	}
			
	/** Gets the total number of events scheduled
	 * @return
	 */
	public final double getTotalNumberEventsScheduled() {
		return myEventCount;
	}

	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	/** Sets the length of the replication
	 *  The length of the replication must be > 0, if set
	 * 
	 * @param lengthOfReplication
	 */
	protected final void setLengthOfReplication(double lengthOfReplication) {
		if (lengthOfReplication <= 0.0)
			throw new IllegalArgumentException("Simulation replication length < 0.0");
		myLengthOfReplication = lengthOfReplication;
	}

	/** Tells the event calendar to cancel all the events.
	 * Resets the simulation time to 0.0.
	 * Resets the event identification counter.
	 */
	protected final void initialize() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINE))
			JSL.LOGGER.fine("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
				+ "\t METHOD: \t initialize()");		
		myDoneFlag = false;
		myCurrentTime = 0.0;
		cancelAllEvents();
		resetEventIdCounter();
		myObserverState = INITIALIZED;
		setChanged();
		notifyObservers();
	}

	/** Peeks at the next event to *potentially* be executed
	 *  The event my be labeled for cancellation.
	 * 
	 * @return
	 */
	public final JSLEvent peekNext(){
		return(myEventCalendar.peekNext());
	}
	
	/** Processes the next event in the event calendar
	 *  if it exists or it has not been canceled.
	 *
	 * @return a reference to the executed event if available
	 */
	protected final JSLEvent executeNextEvent() {
		myCurrentEvent = myEventCalendar.nextEvent();

		if (myCurrentEvent == null) {
			myDoneFlag = true; // no more events
		} else {
			// the event is no longer scheduled
			myCurrentEvent.setScheduledFlag(false);

			if (myCurrentEvent.getCanceledFlag() == false) {
				// event was not cancelled
				// update the current simulation time
				setTime(myCurrentEvent.getTime());

				beforeEvent(myCurrentEvent);
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
					JSL.LOGGER.finest("Scheduler executing \t ID: \t "
						+ myCurrentEvent.getId()
						+ "\t NAME: \t "
						+ myCurrentEvent.getName()
						+ " at time "
						+ getTime());
				myCurrentEvent.execute();
			}
		}

		return (myCurrentEvent);

	}

	//  ===========================================
	//      PRIVATE METHODS
	//  ===========================================

	private void schedule(JSLEvent e) {
		if (e.isScheduled()) //already scheduled
			throw new JSLEventException("Event has already been scheduled!");

		if (e.getTime() < myCurrentTime)
			throw new JSLEventException("Event to be Scheduled before Current Time!");

		if (e.getTime() <= myLengthOfReplication) {
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("CLASSNAME \t" + this.getClass().getName() + "\t NAME: \t " + getName() + "\t ID: \t " + getId() 
					+ "\t METHOD: \t schedule() \t EVENT TIME: \t " + e.getTime());		
			myEventCount = myEventCount + 1;
			e.setId(myEventCount);
			myEventCalendar.add(e);
			e.setScheduledFlag(true);
		} 
	}

	private final void setTime(double t) {
		myCurrentTime = t;
	}

	private final void beforeEvent(JSLEvent e) {
		myObserverState = EVENT_EXECUTED;
		setChanged();
		notifyObservers(e);
	}

	private final void resetEventIdCounter() {
		myEventCount = 0;
	}
	
}
