/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.conditions;

import jsl.utilities.reporting.JSL;
import jsl.utilities.IdentityIfc;

public class Condition implements ConditionIfc, IdentityIfc {
    
    protected boolean myState = false;
    protected String myName;
    private static int myCounter_;
    
    private int myId;
    
    //  ===========================================
    //        CONSTRUCTORS
    //  ===========================================
    
    /** Creates a model element with name "null"
     */
    public Condition(){
        this(null);
    }
    
    /** Creates a model element with the given name.
     * @param name The name of the model element
     */
    public Condition(String name){
        myCounter_ = myCounter_ + 1;
        myId = myCounter_;
        setName(name);
        if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINER))
        		JSL.LOGGER.finer("Constructed \t Class " + this.getClass().getName() + "\t Name: " + getName() + "\t ID: " + getId());
    }
        
	/** Gets the name.
	* @return The name of object.
	*/
	public final String getName() {
		return myName;
	}

	/** Sets the name 
	* @param str The name as a string.
	*/
	public final void setName(String str) {
		if (str == null){
			String s = this.getClass().getName();
			int k = s.lastIndexOf(".");
			if (k != -1)
				s = s.substring(k+1);
			myName = s;
		}
		else
			myName = str;
	}

	/** Returns the id for this condition
	 * 
	 * @return
	 */
	public long getId(){
		return(myId);
	}
        
    /** converts to a string
     *  @return a string representing the condition
     */
	public String toString(){
		String s = "ID: " + getId() + " NAME: " + getName() + " isSet() = " + isSet();
		return(s);
	}
    
    /** Checks whether the condition has been set.
     * @return True if the condition has been set, false otherwise
     */
    public boolean isSet(){
        return(myState);
    }
    
    /** Determines the current value for the condition
     *  This method should be overwritten to provide the logic necessary
     *  to properly determine whether the condition is true or false
     */
    public void evaluate(){
        
    }
    
    /** Clears the condition, i.e. ensures that isSet() returns false
     */
    public void clear(){
        myState = false;
    }
    
    /** Sets the condition, i.e. ensures that isSet() returns true
     */
    public void set(){
        myState = true;
    }
    
}
