/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling.conditions;

import jsl.modeling.elements.variable.*;
import jsl.utilities.statistic.*;
import jsl.observers.variable.StatisticalObserver;

/**
 *
 */
public class WithinReplicationHalfWidthStoppingCondition extends Condition {
    
    private ResponseVariable myRV;
    protected double myConfidenceCoeff;
    protected double myRelativePrecision = 0.0;
    protected double myDesiredHalfWidth = 1.0 ;
    protected boolean myCheckingOption = HALF_WIDTH;
    public final static boolean RELATIVE_PRECISION = false;
    public final static boolean HALF_WIDTH = true;
    public final static double DEFAULT_CONFIDENCE_COEFF = 1.965;
    
    /** Creates a new instance of WithinReplicationHalfWidthStoppingCondition */
    public WithinReplicationHalfWidthStoppingCondition(ResponseVariable rv, double confidenceCoeff, boolean checkingOption) {
        super();
        setResponseVariable(rv);
        setConfidenceCoefficient(confidenceCoeff);
        setCheckingOption(checkingOption);
    }
    
    public WithinReplicationHalfWidthStoppingCondition(ResponseVariable rv, double confidenceCoeff) {
        this(rv, confidenceCoeff, HALF_WIDTH);
    }
    
    public WithinReplicationHalfWidthStoppingCondition(ResponseVariable rv) {
        this(rv, DEFAULT_CONFIDENCE_COEFF, HALF_WIDTH);
    }
    
    public void setResponseVariable(ResponseVariable rv){
        if (rv == null)
            throw new IllegalArgumentException("ResponseVariable must not be null.");
        myRV = rv;
    }
    
    public void setCheckingOption(boolean checkingOption){
        myCheckingOption = checkingOption;
    }
    
    public void setConfidenceCoefficient(double confidenceCoeff){
        if (confidenceCoeff <=0)
            throw new IllegalArgumentException("Confidence Coefficent must be > 0.");
        
        myConfidenceCoeff = confidenceCoeff;
    }
    
    public void setDesiredHalfWidth(double desiredHalfWidth){
        if (desiredHalfWidth <=0)
            throw new IllegalArgumentException("Desired half-width must be > 0.");
        
        myDesiredHalfWidth = desiredHalfWidth;
    }
    
    public void setRelativePrecision(double desiredRelativePrecision){
        if (desiredRelativePrecision <= 0)
            throw new IllegalArgumentException("Relative Precision must be > 0.");
        
        myRelativePrecision = desiredRelativePrecision;
    }
    
    public double getConfidenceCoefficient(){
        return(myConfidenceCoeff);
    }
    
    public double getDesiredHalfWidth(){
        return(myDesiredHalfWidth);
    }
    
    public double getDesiredRelativePrecision(){
        return(myRelativePrecision);
    }
    
    public void evaluate(){
        StatisticalObserver so = myRV.getStatisticalObserver();
        StatisticAccessorIfc stat = so.getAcrossBatchStatistic();
        if (stat != null){
            double se = stat.getStandardError();
            double hw = se*myConfidenceCoeff;
//            System.out.println("hw = " + hw + " E = " + myDesiredHalfWidth);
            if (myCheckingOption == HALF_WIDTH){
                if (hw <= myDesiredHalfWidth)
                    set();
            } else {
                double xbar = stat.getAverage();
                if (hw < myRelativePrecision*xbar)
                    set();
            }
        }        
    }
}
