/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.processview.description.EntityType;
import jsl.modeling.elements.processview.description.EntityTypeSetter;
import jsl.modeling.elements.resource.Entity;
import jsl.utilities.random.*;

/**
 *
 */
public class EntityGenerator extends EntitySendingAgent implements EventGeneratorIfc{

	/** The underlying event generator
	 * 
	 */
	protected EventGenerator myEventGenerator;
	
	/** The underlying event generator listener
	 * 
	 */
	protected EntityCreationListener myEntityCreationListener;

	/** A reference to an object that can determine the type of the
	 *  entity being generated
	 */
	protected EntityTypeSetter myEntityTypeSetter;
	
	/** used to identify the type of entity that is being generated
	 *  the default is the EntityType.DEFAULT type
	 */
	protected EntityType myEntityType;

	// ===========================================
	// CONSTRUCTORS
	// ===========================================

	/**
	 * Creates an EntityGenerator
	 * 
	 */
	public EntityGenerator(ModelElement parent) {
		this(parent, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * Creates an EntityGenerator
	 * 
	 */
	public EntityGenerator(ModelElement parent, String name) {
		this(parent, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}
	
	/**
	 * Creates an EntityGenerator 
	 * @param timeUntilFirst
	 *            A RandomIfc object that supplies the time until the first event
	 * 
	 */
	public EntityGenerator(ModelElement parent, RandomIfc timeUntilFirst) {
		this(parent, timeUntilFirst, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}
	
	/**
	 * Creates an EntityGenerator 
	 * @param timeUntilFirst
	 *            A RandomIfc object that supplies the time until the first event
	 * 
	 */
	public EntityGenerator(ModelElement parent, RandomIfc timeUntilFirst, String name) {
		this(parent, timeUntilFirst, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * Creates an EntityGenerator.
	 * @param timeUntilFirst
	 *            A RandomIfc object that supplies the time until the first event
	 * @param timeUntilNext
	 *            A RandomIfc object that supplies the time between events
	 * 
	 */
	public EntityGenerator(ModelElement parent, RandomIfc timeUntilFirst,
			RandomIfc timeUntilNext) {
		this(parent, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}
	
	/**
	 * Creates an EntityGenerator.
	 * @param timeUntilFirst
	 *            A RandomIfc object that supplies the time until the first event
	 * @param timeUntilNext
	 *            A RandomIfc object that supplies the time between events
	 * 
	 */
	public EntityGenerator(ModelElement parent, RandomIfc timeUntilFirst,
			RandomIfc timeUntilNext, String name) {
		this(parent, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * Creates an EntityGenerator 
	 * @param timeUntilFirst
	 *            A RandomIfc object that supplies the time until the first
	 *            event.
	 * @param timeUntilNext
	 *            A RandomIfc object that supplies the time between events.
	 * @param maxNum
	 *            A RandomIfc object that supplies the maximum number of events
	 *            to generate. This is typically a constant. When the generator
	 *            is created, this variable is used to set the number of events
	 *            to generate. The number is rounded to the closest long. Each
	 *            time an event is to be scheduled the maximum number of events
	 *            is checked. If the maximum has been reached, then the
	 *            generator is turned off. If this variable is random, then it
	 *            is only used once to set the number of events for the
	 *            generator.
	 * 
	 */
	public EntityGenerator(ModelElement parent, RandomIfc timeUntilFirst,
			RandomIfc timeUntilNext, long maxNum, String name) {
		this(parent, timeUntilFirst, timeUntilNext, maxNum, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * Creates an EntityGenerator that generates entities
	 * 
	 * @param timeUntilFirst
	 *            A RandomIfc object that supplies the time until the first
	 *            event.
	 * @param timeUntilNext
	 *            A RandomIfc object that supplies the time between events.
	 * @param maxNum
	 *            A RandomIfc object that supplies the maximum number of events
	 *            to generate. This is typically a constant. When the generator
	 *            is created, this variable is used to set the number of events
	 *            to generate. The number is rounded to the closest long. Each
	 *            time an event is to be scheduled the maximum number of events
	 *            is checked. If the maximum has been reached, then the
	 *            generator is turned off. If this variable is random, then it
	 *            is only used once to set the number of events for the
	 *            generator.
	 * @param timeUntilLast
	 *            A RandomIfc object that supplies a time to stop generating
	 *            events. This is typically a constant. When the generator is
	 *            created, this variable is used to set the ending time of the
	 *            generator. Each time an event is to be scheduled the ending
	 *            time is checked. If the time of the next event is past this
	 *            time, then the generator is turned off and the event won't be
	 *            scheduled. If this variable is random, then it is only used
	 *            once to set the ending time of the generator.
	 * @param name
	 *            The name of the generator.
	 */
	 public EntityGenerator(ModelElement parent, RandomIfc timeUntilFirst,
	 		RandomIfc timeUntilNext, long maxNum, double timeUntilLast, String name) {
		super(parent, name);
		myEntityCreationListener = new EntityCreationListener();
		
		myEventGenerator = new EventGenerator(this, myEntityCreationListener, timeUntilFirst,
				timeUntilNext, maxNum, timeUntilLast, getName() + " Event Generator");
		myEntityType = getModel().getDefaultEntityType();
	 }
	
	/** The supplied setter will be used to determine the type
	 *  of the entity being generated.   In this manner,
	 *  the entity generator could generate entities types from a random variable.
	 * 
	 * @param entityTypeVariable
	 */
	public final void setEntityTypeSetter(EntityTypeSetter entityTypeSetter){
		if (entityTypeSetter == null)
			throw new IllegalArgumentException("EntityTypeSetter must be non-null!");
		myEntityTypeSetter = entityTypeSetter;
	}

	/** The variable used to set the type
	 *  of the entity being generated, null if no variable is to be used
	 * 
	 * @return A reference to a Variable that can be used to set the type
	 *  of the entity being generated, null if no variable is to be used
	 */
	public final EntityTypeSetter getEntityTypeSetter(){
		return(myEntityTypeSetter);
	}
	
	/**
	 * @return Returns the type.
	 */
	public final EntityType getEntityType() {
		return myEntityType;
	}
	
	/**
	 * @param type The type to set.
	 */
	public final void setEntityType(EntityType entityType) {
		if (entityType == null)
			throw new IllegalArgumentException("EntityType name was null!");    
		myEntityType = entityType;
	}	
	
	/**
	 * @return
	 */
	public final double getEndingTime() {
		return myEventGenerator.getEndingTime();
	}
	
	/**
	 * @return
	 */
	public final long getMaximumNumberOfEvents() {
		return myEventGenerator.getMaximumNumberOfEvents();
	}
	
	/**
	 * @return
	 */
	public final boolean isGeneratorDone() {
		return myEventGenerator.isGeneratorDone();
	}
	
	/**
	 * @return
	 * @see jsl.modeling.elements.EventGenerator#getStartOnInitializeFlag()
	 */
	public final boolean getStartOnInitializeFlag() {
		return myEventGenerator.getStartOnInitializeFlag();
	}

	/**
	 * @param flag
	 * @see jsl.modeling.elements.EventGenerator#setStartOnInitializeFlag(boolean)
	 */
	public final void setStartOnInitializeFlag(boolean flag) {
		myEventGenerator.setStartOnInitializeFlag(flag);
	}

	/**
	 * 
	 * @see jsl.modeling.elements.EventGenerator#turnOnGenerator()
	 */
	public final void turnOnGenerator() {
		myEventGenerator.turnOnGenerator();
	}

	/**
	 * @param t
	 * @see jsl.modeling.elements.EventGenerator#turnOnGenerator(double)
	 */
	public final void turnOnGenerator(double t) {
		myEventGenerator.turnOnGenerator(t);
	}

	/**
	 * @param r
	 * @see jsl.modeling.elements.EventGenerator#turnOnGenerator(jsl.utilities.random.RandomIfc)
	 */
	public void turnOnGenerator(RandomIfc r) {
		myEventGenerator.turnOnGenerator(r);
	}

	/**
	 * 
	 */
	public final void turnOffGenerator() {
		myEventGenerator.turnOffGenerator();
	}
	
	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getGenerationEndingTimeForReplications()
	 */
	public double getGenerationEndingTimeForReplications() {
		return myEventGenerator.getGenerationEndingTimeForReplications();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getNumberOfEventsGenerated()
	 */
	public long getNumberOfEventsGenerated() {
		return myEventGenerator.getNumberOfEventsGenerated();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getTimeUntilFirstEventForReplications()
	 */
	public RandomIfc getTimeUntilFirstEventForReplications() {
		return myEventGenerator.getTimeUntilFirstEventForReplications();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGenerator#getTimeBetweenEvents()
	 */
	public final RandomIfc getTimeBetweenEvents() {
		return myEventGenerator.getTimeBetweenEvents();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#isSuspended()
	 */
	public boolean isSuspended() {
		return myEventGenerator.isSuspended();
	}

	/**
	 * 
	 * @see jsl.modeling.elements.EventGeneratorIfc#resume()
	 */
	public void resume() {
		myEventGenerator.resume();
	}

	/**
	 * @param endingTime
	 * @see jsl.modeling.elements.EventGeneratorIfc#setEndingTime(double)
	 */
	public void setEndingTime(double endingTime) {
		myEventGenerator.setEndingTime(endingTime);
	}

	/**
	 * @param endingTime
	 * @see jsl.modeling.elements.EventGeneratorIfc#setGenerationEndingTimeForReplications(double)
	 */
	public void setGenerationEndingTimeForReplications(double endingTime) {
		myEventGenerator.setGenerationEndingTimeForReplications(endingTime);
	}

	/**
	 * @param maxNum
	 * @see jsl.modeling.elements.EventGeneratorIfc#setMaximumNumberOfEvents(long)
	 */
	public void setMaximumNumberOfEvents(long maxNum) {
		myEventGenerator.setMaximumNumberOfEvents(maxNum);
	}

	/**
	 * @param timeBtwEvents
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEvents(jsl.utilities.random.RandomIfc, long)
	 */
	public void setTimeBetweenEvents(RandomIfc timeBtwEvents, long maxNumEvents) {
		myEventGenerator.setTimeBetweenEvents(timeBtwEvents, maxNumEvents);
	}

	/**
	 * @param timeUntilNext
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEvents(jsl.utilities.random.RandomIfc)
	 */
	public void setTimeBetweenEvents(RandomIfc timeUntilNext) {
		myEventGenerator.setTimeBetweenEvents(timeUntilNext);
	}

	/**
	 * @param timeBtwEvents
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEventsForReplications(jsl.utilities.random.RandomIfc, long)
	 */
	public void setTimeBetweenEventsForReplications(RandomIfc timeBtwEvents, long maxNumEvents) {
		myEventGenerator.setTimeBetweenEventsForReplications(timeBtwEvents, maxNumEvents);
	}

	/**
	 * @param timeUntilFirst
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeUntilFirstEventForReplications(jsl.utilities.random.RandomIfc)
	 */
	public void setTimeUntilFirstEventForReplications(RandomIfc timeUntilFirst) {
		myEventGenerator.setTimeUntilFirstEventForReplications(timeUntilFirst);
	}

	/**
	 * 
	 * @see jsl.modeling.elements.EventGeneratorIfc#suspend()
	 */
	public void suspend() {
		myEventGenerator.suspend();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getMaximumNumberOfEventsForReplications()
	 */
	public long getMaximumNumberOfEventsForReplications() {
		return myEventGenerator.getMaximumNumberOfEventsForReplications();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getTimeBetweenEventsForReplications()
	 */
	public RandomIfc getTimeBetweenEventsForReplications() {
		return myEventGenerator.getTimeBetweenEventsForReplications();
	}

	/**
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setMaximumNumberOfEventsForReplications(long)
	 */
	public void setMaximumNumberOfEventsForReplications(long maxNumEvents) {
		myEventGenerator.setMaximumNumberOfEventsForReplications(maxNumEvents);
	}

	/**
	 * @param timeBtwEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEventsForReplications(jsl.utilities.random.RandomIfc)
	 */
	public void setTimeBetweenEventsForReplications(RandomIfc timeBtwEvents) {
		myEventGenerator.setTimeBetweenEventsForReplications(timeBtwEvents);
	}

	/** Tells the entity process generator to create a suitable entity
	 *  This is primarily used by the generator's listener.
	 * 
	 * @return The created entity
	 */
	protected Entity createEntity(){
		Entity e;
		if (myEntityTypeSetter == null)
			e = new Entity(getEntityType());
		else {
			e = new Entity();
			myEntityTypeSetter.setType(e);
		}
		return(e);	
	}
	
	protected void generateAction(EventGenerator generator, JSLEvent event){
		// make the entity
		Entity entity = createEntity();
		// send the entity
		send(entity);	
	}
	
	private class EntityCreationListener implements EventGeneratorListenerIfc {

		/* (non-Javadoc)
		 * @see jsl.modeling.elements.EventGeneratorListenerIfc#generate(jsl.modeling.elements.EventGenerator, jsl.modeling.JSLEvent)
		 */
		public void generate(EventGenerator generator, JSLEvent event) {
			generateAction(generator, event);
		}	
	}
}
