/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements;

import java.util.ArrayList;

import jsl.modeling.*;
import jsl.modeling.elements.resource.Entity;

/**
 *
 */
public class EntitySendingAgent extends SchedulingElement implements EntitySendingIfc{

	protected EntityReceivingIfc myDefaultEntityReceiver;
	protected EntityReceiverIteratorIfc myEntityReceiverIterator;
	
	/**
	 * @param parent
	 */
	public EntitySendingAgent(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public EntitySendingAgent(ModelElement parent, String name) {
		super(parent, name);
		myDefaultEntityReceiver = new EntityReceivingAgent(this, getName() + " Default Entity Receiver");
		myEntityReceiverIterator = new EntityReceiverListIterator(new ArrayList<EntityReceivingIfc>());
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.SendEntityIfc#send(jsl.modeling.elements.resource.Entity)
	 */
	public void send(Entity entity) {
		if (entity == null)		
			throw new IllegalArgumentException("Entity must be non-null!");
		
		// figure out where to send the entity
		EntityReceivingIfc receiver = null;
		if (entity.hasNextEntityReceiver() == true){ // first ask the entity
			receiver = entity.nextEntityReceiver();
		} else if (myEntityReceiverIterator.hasNextEntityReceiver()) {
			 // entity doesn't know, use the sender agent's receiver selection rule
			receiver = myEntityReceiverIterator.nextEntityReceiver();
		} else {// entity doesn't know and sender's rule doesn't have a receiver
			receiver = myDefaultEntityReceiver;
		}
		
		receiver.receive(entity, this);	
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.SendEntityIfc#send(jsl.modeling.elements.resource.Entity, jsl.modeling.elements.ReceiveEntityIfc)
	 */
	public void send(Entity entity, EntityReceivingIfc receiver) {
		if (entity == null)		
			throw new IllegalArgumentException("Entity must be non-null!");

		if (receiver == null)		
			throw new IllegalArgumentException("ReceiveEntityIfc receiver must be non-null!");

		receiver.receive(entity, this);
	}
	
	/** Sets the default receiver for this sending agent
	 * @param nextReceiver The nextReceiver to set.
	 */
	public final void setDefaultEntityReceiver(EntityReceivingIfc nextReceiver) {
		if(nextReceiver == null)
			throw new IllegalArgumentException("ReceiveEntityAgent nextReceiver must be non-null!");
		
		myDefaultEntityReceiver = nextReceiver;
	}

	/**
	 * @param myNextReceiver The myNextReceiver to set.
	 */
	public final void setEntityReceiverIterator(EntityReceiverIteratorIfc iterator) {
		if(iterator == null)
			throw new IllegalArgumentException("EntityReceiverSelectionRuleIfc iterator must be non-null!");
		
		myEntityReceiverIterator = iterator;
	}
	
}
