/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import jsl.modeling.ModelElement;
import jsl.utilities.random.robj.DUniformList;
import jsl.utilities.random.robj.RList;
import jsl.utilities.random.robj.RListIfc;
import jsl.utilities.reporting.JSL;

/** Allows for random selection from a list of objects as supplied by the RList<T>
 * 
 *  Note: The underlying streams for the random selection process are controlled 
 *   by the replication, see RandomVariable
 *   
 *  Note: If the user calls setRandomList() or calls any mutator methods on the list, 
 *  during the replication the list will be changed; however, the supplied initial list
 *  will be used to reinitialize the list at the beginning of each replication. If 
 *  the user does not want this behavior, then overwrite the initialize() method or turn 
 *  off automatic initialization via the ModelElement interface.
 *  
 *  If the user does not supply an initial list then there will not be any elements to
 *  select from during the replication unless setRandomList() or the mutator methods are called
 *  prior to or during the replication.
 *  
 *  Warning: If the user changes the supplied initial list during or between replications, then
 *  each replication may not start under the same conditions.
 *  
 *  Warning: There may not be elements in the list, thus a call to getRandomElement() may return null
 * 
 * The default RList is DUniformList, equally likely selection over the elements in the list
 * 
 */
public class RandomList<T> extends ModelElement implements RListIfc<T> {

	protected RList<T> myRList;
	protected List<T> myInitialList;
	private boolean myResetStartStreamOption;
	private boolean myResetNextSubStreamOption;
	
	/**
	 * @param parent
	 */
	public RandomList(ModelElement parent) {
		this(parent, null, null);
	}
	
	/**
	 * @param parent
	 */
	public RandomList(ModelElement parent, List<T> initialList) {
		this(parent, initialList, null);
	}


	/** The initial list is used to initialize the list prior to 
	 *  each replication
	 * 
	 * @param parent
	 * @param initialList
	 * @param name
	 */
	public RandomList(ModelElement parent, List<T> initialList, String name) {
		super(parent, name);

        setWarmUpOption(false); // do not need to respond to warm events
        setResetStartStreamOption(true);
        setResetNextSubStreamOption(true);
        myRList = new DUniformList<T>();
        setInitialList(initialList);
	}

	protected void initializeList(List<T> list){
		myRList.clear();
		if (list != null)
			myRList.addAll(list);
	}

	protected void initialize(){
		initializeList(myInitialList);
	}
	
	protected void removedFromModel(){
		super.removedFromModel();
		if (myInitialList != null)
			myInitialList.clear();
		myInitialList = null;
		myRList.clear();
		myRList = null;
	}
	
	/** before any replications reset the underlying random number generator to the
     *  starting stream
     *
     */
    protected void beforeExperiment(){
        super.beforeExperiment();   
        
        if (getResetStartStreamOption())
        		resetStartStream();

    }
    
    /** after each replication reset the underlying random number generator to the next
     *  sub-stream
     */
    protected void afterReplication(){
        super.afterReplication();
        
        if (getResetNextSubStreamOption())
        		resetNextSubstream();

    }   

    //  ===========================================
    //        PUBLIC INTERFACE
    //  ===========================================
 
	/** Allows the user to change the random list after it has 
	 *  been initialized by the initial list.  The supplied
	 *  list will be used directly instead of the current list
	 * @param list
	 */
	public void setRandomList(RList<T> list){
		if (list == null)
			throw new IllegalArgumentException("The list was null");
		myRList = list.newInstance();
	}
	
	/** Sets the initial list to be used to initialize 
	 *  the list at the beginning of each replication
	 * 
	 * @param list
	 */
	public void setInitialList(List<T> list){
		myInitialList = list;
	}
	
	/** Gets the current Reset Start Stream Option
	 * @return
	 */
	public final boolean getResetStartStreamOption() {
		return myResetStartStreamOption;
	}

	/** Sets the reset start stream option, true
	 *  means that it will be reset to the starting stream
	 * @param b
	 */
	public final void setResetStartStreamOption(boolean b) {
		myResetStartStreamOption = b;
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
			JSL.LOGGER.finest("Setting \t ID: " + getId() + "\t NAME: " + getName() + "\t reset start stream option " + b);
	}
 
	/** Gets the current reset next sub-stream option
	 *  true means, that it is set to jump to the next sub-stream after 
	 *  each replication
	 * @return
	 */
	public final boolean getResetNextSubStreamOption() {
		return myResetNextSubStreamOption;
	}

	/** Sets the current reset next sub-stream option
	 *  true means, that it is set to jump to the next sub-stream after 
	 *  each replication
	 * @param b
	 */
	public final void setResetNextSubStreamOption(boolean b) {
		myResetNextSubStreamOption = b;
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
			JSL.LOGGER.finest("Setting \t ID: " + getId() + "\t NAME: " + getName() + "\t reset next substream option " + b);
	}
	    
    /**
	 * @param index
	 * @param element
	 * @see jsl.utilities.random.robj.RList#add(int, java.lang.Object)
	 */
	public void add(int index, T element) {
		myRList.add(index, element);
	}

	/**
	 * @param obj
	 * @return
	 * @see jsl.utilities.random.robj.RList#add(java.lang.Object)
	 */
	public boolean add(T obj) {
		return myRList.add(obj);
	}

	/**
	 * @param c
	 * @return
	 * @see jsl.utilities.random.robj.RList#addAll(java.util.Collection)
	 */
	public boolean addAll(Collection<? extends T> c) {
		return myRList.addAll(c);
	}

	/**
	 * @param index
	 * @param c
	 * @return
	 * @see jsl.utilities.random.robj.RList#addAll(int, java.util.Collection)
	 */
	public boolean addAll(int index, Collection<? extends T> c) {
		return myRList.addAll(index, c);
	}

	/**
	 * 
	 * @see jsl.utilities.random.robj.RList#clear()
	 */
	public void clear() {
		myRList.clear();
	}

	/**
	 * @param arg0
	 * @return
	 * @see jsl.utilities.random.robj.RList#contains(java.lang.Object)
	 */
	public boolean contains(Object arg0) {
		return myRList.contains(arg0);
	}

	/**
	 * @param arg0
	 * @return
	 * @see jsl.utilities.random.robj.RList#containsAll(java.util.Collection)
	 */
	public boolean containsAll(Collection<?> arg0) {
		return myRList.containsAll(arg0);
	}

	/**
	 * @param index
	 * @return
	 * @see jsl.utilities.random.robj.RList#get(int)
	 */
	public T get(int index) {
		return myRList.get(index);
	}

	/**
	 * @return
	 * @see jsl.utilities.random.robj.RList#getList()
	 */
	public List<T> getList() {
		return myRList.getList();
	}

	/**
	 * @return
	 * @see jsl.utilities.random.robj.RList#getRandomElement()
	 */
	public T getRandomElement() {
		return myRList.getRandomElement();
	}

	/**
	 * @param arg0
	 * @return
	 * @see jsl.utilities.random.robj.RList#indexOf(java.lang.Object)
	 */
	public int indexOf(Object arg0) {
		return myRList.indexOf(arg0);
	}

	/**
	 * @return
	 * @see jsl.utilities.random.robj.RList#isEmpty()
	 */
	public boolean isEmpty() {
		return myRList.isEmpty();
	}

	/**
	 * @return
	 * @see jsl.utilities.random.robj.RList#iterator()
	 */
	public Iterator<T> iterator() {
		return myRList.iterator();
	}

	/**
	 * @param o
	 * @return
	 * @see jsl.utilities.random.robj.RList#lastIndexOf(java.lang.Object)
	 */
	public int lastIndexOf(Object o) {
		return myRList.lastIndexOf(o);
	}

	/**
	 * @return
	 * @see jsl.utilities.random.robj.RList#listIterator()
	 */
	public ListIterator<T> listIterator() {
		return myRList.listIterator();
	}

	/**
	 * @param index
	 * @return
	 * @see jsl.utilities.random.robj.RList#listIterator(int)
	 */
	public ListIterator<T> listIterator(int index) {
		return myRList.listIterator(index);
	}

	/**
	 * @param index
	 * @return
	 * @see jsl.utilities.random.robj.RList#remove(int)
	 */
	public T remove(int index) {
		return myRList.remove(index);
	}

	/**
	 * @param obj
	 * @return
	 * @see jsl.utilities.random.robj.RList#remove(java.lang.Object)
	 */
	public boolean remove(Object obj) {
		return myRList.remove(obj);
	}

	/**
	 * @param c
	 * @return
	 * @see jsl.utilities.random.robj.RList#removeAll(java.util.Collection)
	 */
	public boolean removeAll(Collection<?> c) {
		return myRList.removeAll(c);
	}

	/**
	 * 
	 * @see jsl.utilities.random.robj.RList#resetNextSubstream()
	 */
	public void resetNextSubstream() {
		myRList.resetNextSubstream();
	}

	/**
	 * 
	 * @see jsl.utilities.random.robj.RList#resetStartStream()
	 */
	public void resetStartStream() {
		myRList.resetStartStream();
	}

	/**
	 * 
	 * @see jsl.utilities.random.robj.RList#resetStartSubstream()
	 */
	public void resetStartSubstream() {
		myRList.resetStartSubstream();
	}

	/**
	 * @param c
	 * @return
	 * @see jsl.utilities.random.robj.RList#retainAll(java.util.Collection)
	 */
	public boolean retainAll(Collection<?> c) {
		return myRList.retainAll(c);
	}

	/**
	 * @param index
	 * @param element
	 * @return
	 * @see jsl.utilities.random.robj.RList#set(int, java.lang.Object)
	 */
	public T set(int index, T element) {
		return myRList.set(index, element);
	}

	/**
	 * @param flag
	 * @see jsl.utilities.random.robj.RList#setAntithetic(boolean)
	 */
	public void setAntithetic(boolean flag) {
		myRList.setAntithetic(flag);
	}

	/**
	 * @return
	 * @see jsl.utilities.random.robj.RList#size()
	 */
	public int size() {
		return myRList.size();
	}

	/**
	 * @param fromIndex
	 * @param toIndex
	 * @return
	 * @see jsl.utilities.random.robj.RList#subList(int, int)
	 */
	public List<T> subList(int fromIndex, int toIndex) {
		return myRList.subList(fromIndex, toIndex);
	}

	/**
	 * @return
	 * @see jsl.utilities.random.robj.RList#toArray()
	 */
	public Object[] toArray() {
		return myRList.toArray();
	}

	/**
	 * @param <T>
	 * @param a
	 * @return
	 * @see jsl.utilities.random.robj.RList#toArray(T[])
	 */
	public <T> T[] toArray(T[] a) {
		return myRList.toArray(a);
	}    

}
