/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements;

/**  Objects are often supplied to other objects to be processed.  The action
 *   may take some simulated time to complete and may involve many other objects than 
 *   just the receiver. After the processing is completed the sender may need to know
 *   that the action has completed.  This interface should be implemented by classes that
 *   perform the action and notify listeners that the action is completed
 *   
 *   Clients should use the addActionCompletedListener() method to register listeners for 
 *   the action's completion.  Typically this is done when the implementor is constructed.
 *   
 *   The startAction(QObject qObj) method is used to start the action.  
 *   The startAction(QObject qObj, ActionCompletedListenerIfc listener) method can be use
 *   by clients to register the listener at the time of the call.  
 *   
 *   Implementors must ensure that multiple notifications to the same listener do not occur.
 *   
 *   Implementors should quarantee that the listeners are notified in the order in which
 *   the listeners are registered.
 * 
 *
 */
public interface StartActionIfc {

	/** Tells the action to start on the supplied object.  To make this method useful
	 *  the ActionCompletedListeners must be notified when the action is completed. The
	 *  notification does not have to happen within the startAction() method.  The implication
	 *  is that the action may be completed at some future simulated time and that
	 *  listeners will be notified at that time.
	 * 
	 * @param obj
	 */
	public void startAction(Object obj);
	
	/** The supplied listener must not be null.  The implication is that the listener
	 *  will be used to call back when the action is completed.
	 * 
	 * @param obj
	 * @param listener
	 */
	public void startAction(Object obj, ActionCompletedListenerIfc listener);
	
	/** Registers the listener to be notified when actions are completed
	 * 
	 * @param listener
	 */
	public void addActionCompletedListener(ActionCompletedListenerIfc listener);
	
	/** Unregisters the listener to not receive notifications of action completions.
	 * 
	 * @param listener
	 */
	public void deleteActionCompletedListener(ActionCompletedListenerIfc listener);
	
	/** Notifies any registered action listeners that an action occured that may be 
	 *  associated with the supplied QObject
	 * 
	 * @param obj  An object for a message (may be null)
	 */
	public void notifyActionListeners(Object obj);
	
}
