/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements;

import java.util.*;

import jsl.modeling.*;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Constant;

/**
 *
 */
public class TimedAction extends SchedulingElement {

	private boolean myScheduleOnInitFlag;
	private boolean myRepeatActionFlag;
	private RandomVariable myActionTimeRV;
	private List<TimedActionListenerIfc> myTimedActions;
	private JSLEvent myTimedActionEvent;

	/**
	 * @param parent
	 */
	public TimedAction(ModelElement parent) {
		this(parent, Constant.POSITIVE_INFINITY, null);
	}
	
	/**
	 * @param parent
	 * @param actionTime
	 */
	public TimedAction(ModelElement parent, RandomIfc actionTimeCDF) {
		this(parent, actionTimeCDF, null);
	}

	/**
	 * @param parent
	 * @param actionTime
	 * @param name
	 */
	public TimedAction(ModelElement parent, RandomIfc actionTimeCDF, String name) {
		super(parent, name);
		myScheduleOnInitFlag = false;
		myRepeatActionFlag = false;
		setTimeBetweenActionsInitialRandomSource(actionTimeCDF);
		myTimedActions = new ArrayList<TimedActionListenerIfc>();
	}

	/**
	 * @return Returns the actionTime.
	 */
	public final RandomIfc getTimeBetweenActions() {
		return myActionTimeRV.getRandomSource();
	}
	
	/**
	 * @return Returns the actionTime.
	 */
	public final RandomIfc getTimeBetweenActionsForReplications() {
		return myActionTimeRV.getInitialRandomSource();
	}
	
	/** Sets the time between actions for all replications
	 * 
	 * @param actionTime The actionTime to set.
	 */
	public final void setTimeBetweenActionsInitialRandomSource(RandomIfc actionTime) {
		if (actionTime == null )
			throw new IllegalArgumentException("The action time must not be null.");
		
		if (myActionTimeRV == null)
			myActionTimeRV = new RandomVariable(this, actionTime);
		else
			myActionTimeRV.setInitialRandomSource(actionTime);
		
	}
	
	/** Sets the time between actions for the current replication
	 * 
	 * @param actionTime The actionTime to set.
	 */
	public final void setTimeBetweenActions(RandomIfc actionTime) {
		if (actionTime == null )
			throw new IllegalArgumentException("The action time must not be null.");
		
		if (myActionTimeRV == null)
			myActionTimeRV = new RandomVariable(this, actionTime);
		else
			myActionTimeRV.setRandomSource(actionTime);
		
	}
	
	/** This flag indicates whether or not the action will automatically
	 *  scheduled when initialize() is called.  By default this option is
	 *  false.
	 * 
	 * @return
	 */
	public boolean getScheduleActionOnInitializeFlag() {
		return myScheduleOnInitFlag;
	}

	/** Sets the flag that indicates whether or not the action will automatically
	 *  scheduled when initialize() is called.
	 * 
	 * @param flag
	 */
	public void setScheduleActionOnInitializeFlag(boolean flag) {
		myScheduleOnInitFlag = flag;
	}

	/** Causes the event associated with the action to be canceled
	 *  if it had been scheduled.  The next scheduled action will
	 *  not occur.
	 * 
	 */
	public final void cancelAction(){
		if (myTimedActionEvent != null)
			cancelEvent(myTimedActionEvent);
	}
	
	/** True means the action will repeat.  By default the
	 *  action will not repeat.
	 * @return Returns the repeatActionFlag.
	 */
	public final boolean getRepeatActionFlag() {
		return myRepeatActionFlag;
	}
	
	/** True means the action will repeat. By default the
	 *  action will not repeat.
	 * @param repeatActionFlag The repeatActionFlag to set.
	 */
	public final void setRepeatActionFlag(boolean repeatActionFlag) {
		myRepeatActionFlag = repeatActionFlag;
	}
	
	/** Adds a listener to react to the action event
	 * @param arg0
	 * @return
	 */
	public boolean add(TimedActionListenerIfc action) {
		return myTimedActions.add(action);
	}
	
	/** Removes the listener so that it does not react to the action event
	 * @param arg0
	 * @return
	 */
	public boolean remove(TimedActionListenerIfc action) {
		return myTimedActions.remove(action);
	}

	/** If the action has not been previously scheduled this
	 *  method causes the action to be scheduled according to 
	 *  it's action time distribution
	 * 
	 */
	public final void scheduleAction(){
		if (myTimedActionEvent == null){
			double t = myActionTimeRV.getValue();
			if (Double.isInfinite(t))
				return;
			myTimedActionEvent = scheduleEvent(t);
		}
	}
	
	protected void initialize() {
		if (myScheduleOnInitFlag){
			scheduleAction();
		}
	}
	
	protected final void handleEvent(JSLEvent event){
		notifyTimedActionListeners(event);
		if (myRepeatActionFlag == true)
			rescheduleEvent(event, myActionTimeRV.getValue());	
		else
			myTimedActionEvent = null;
	}
	
	protected final void notifyTimedActionListeners(JSLEvent event){
		for(TimedActionListenerIfc a: myTimedActions)
			a.action(this, event);
	}
	
	// ===========================================
	// INNER CLASSES
	// ===========================================

}
