/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.operation;

import java.util.LinkedList;
import java.util.List;

import jsl.modeling.IllegalStateException;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
 
/**
 *
 */
public class Activity extends Operation {

	private List<OperationListenerIfc> myInterruptedListeners;
	private List<OperationListenerIfc> myCancelledListeners;
	
	private RandomVariable myActivityTimeRV;
	
	/**
	 * @param parent
	 */
	public Activity(ModelElement parent) {
		this(parent, new Constant(0), null);
	}

	public Activity(ModelElement parent, DistributionIfc distribution){
		this(parent, distribution, null);
	}
	
	/**
	 * @param parent
	 * @param name
	 */
	public Activity(ModelElement parent, DistributionIfc distribution, String name) {
		super(parent, name);
		myActivityTimeRV = new RandomVariable(this, distribution);
		myInterruptedListeners = new LinkedList<OperationListenerIfc>();
		myCancelledListeners = new LinkedList<OperationListenerIfc>();
	}
	
	public final void setActivityTimeDistributionInitialRandomSource(DistributionIfc distribution){
		if (distribution == null)
	    		throw new IllegalArgumentException("Attempted to set the activity time distribution to null!");
		
		myActivityTimeRV.setInitialRandomSource(distribution);
		
	}
	
	public final void addCancelListener(OperationListenerIfc listener){
		if (listener == null)
	    		throw new IllegalArgumentException("Attempted to add a null listener");
		myCancelledListeners.add(listener);
	}	

	public final void addInterruptListener(OperationListenerIfc listener){
		if (listener == null)
			throw new IllegalArgumentException("Attempted to add a null listener");
		myInterruptedListeners.add(listener);
	}

	public final void start(Task task){

		if (task == null)
			throw new IllegalArgumentException("The supplied task was null.");

		if (task.getStatus() != Task.Status.READY)
			throw new IllegalStateException("Tried to start a task when it is not READY." + getName() + " Task " + task.getName());

		task.setCurrentOperation(this);
		
		task.setStatus(Task.Status.STARTED);
		
		notifyOperationStartedListeners(task);
		
		scheduleActivityCompletion(task);
		
	}
	
	public final void cancel(Task task){
		
		if (task.getCurrentOperation() != this)
			throw new IllegalArgumentException("The task is not experiencing this activity.");

		if (task.getStatus() != Task.Status.TIME_DELAYED)
			throw new IllegalStateException("Tried to cancel a task when it is not time delayed." + getName() + " Task " + task.getName());

		JSLEvent e = task.getTimedEvent();
		if (e == null)
			throw new IllegalStateException("Tried to cancel a task when it is not scheduled." + getName() + " Task " + task.getName());
		
		cancelEvent(e);
		
		task.setStatus(Task.Status.CANCELLED);
		
		notifyActivityCancelledListeners(task);
		
		task.setStatus(Task.Status.READY);

	}
	
	public final void interrupt(Task task){
		
		if (task.getCurrentOperation() != this)
			throw new IllegalArgumentException("The task is not experiencing this activity.");

		if (task.getStatus() != Task.Status.TIME_DELAYED)
			throw new IllegalStateException("Tried to interrupt a task when it is not time delayed." + getName() + " Task " + task.getName());

		JSLEvent e = task.getTimedEvent();
		if (e == null)
			throw new IllegalStateException("Tried to interrupt a task when it is not scheduled." + getName() + " Task " + task.getName());
		
		cancelEvent(e);
		
		task.setInterruptTime();

		task.setStatus(Task.Status.INTERRUPTED);

		notifyActivityCancelledListeners(task);
		
	}
	
	public final void restart(Task task){
		
		if (task.getCurrentOperation() != this)
			throw new IllegalArgumentException("The task is not experiencing this activity.");

		if (task.getStatus() != Task.Status.INTERRUPTED)
			throw new IllegalStateException("Tried to restart a task when it is not interrupted." + getName() + " Task " + task.getName());
		
		scheduleActivityCompletion(task);
	}
	
	protected final void scheduleActivityCompletion(Task task){
		
		double time = 0.0;
		if (task.getUseDurationFlag() == true){
			time = task.getDurationTime();
		} else {
			time = myActivityTimeRV.getValue();
			task.setDurationTime(time);
		}
		
		JSLEvent e = scheduleEvent(time , 0);
		e.setMessage(task);
		task.setTimedEvent(e);
		task.setStatus(Task.Status.TIME_DELAYED);

	}
	
	protected void handleEvent(JSLEvent e){
		Task task = (Task)e.getMessage();
		task.setStatus(Task.Status.COMPLETED);
		notifyOperationCompletedListeners(task);
		task.setStatus(Task.Status.READY);
	}
	
	protected final void notifyActivityCancelledListeners(Task task){		
		for(OperationListenerIfc a: myCancelledListeners){
			a.update(this, task);
		}
	}
	
	protected final void notifyActivityInterruptedListeners(Task task){
		for(OperationListenerIfc a: myInterruptedListeners){
			a.update(this, task);
		}
	}

	protected final void removeCancelListener(OperationListenerIfc listener){
		if (listener == null)
	    		throw new IllegalArgumentException("Attempted to remove a null listener");
		myCancelledListeners.remove(listener);
	}

	protected final void removeInterruptListener(OperationListenerIfc listener){
		if (listener == null)
			throw new IllegalArgumentException("Attempted to remove a null listener");
		myInterruptedListeners.remove(listener);
	}
	

}
