/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.operation;

import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;

import jsl.modeling.IllegalStateException;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.operation.OperationIfc;
import jsl.modeling.elements.operation.OperationListenerIfc;

/**
 *
 */
public class OperationList extends Operation {

	protected List<OperationIfc> myOperationList;
	protected OperationCompletedListener myOperationCompletedListener;
	/**
	 * @param parent
	 */
	public OperationList(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public OperationList(ModelElement parent, String name) {
		super(parent, name);
		myOperationList = new ArrayList<OperationIfc>();
		myOperationCompletedListener = new OperationCompletedListener();
	}
	
	public final void addOperation(OperationIfc operation){
		if (operation == null)
	    	throw new IllegalArgumentException("Attempted to add a null operation");
		myOperationList.add(operation);
		operation.addCompletionListener(myOperationCompletedListener);
	}

	public void start(Task task){

		if (task == null)
			throw new IllegalArgumentException("The supplied task was null.");

		if (task.getStatus() != Task.Status.READY)
			throw new IllegalStateException("Tried to start a task when it is not READY." + getName() + " Task " + task.getName());
	
		if (task.getOperationIterator() == null){
			// no iterator on this task, not running in another operation list
			ListIterator<OperationIfc> iterator = myOperationList.listIterator();
			task.setOperationIterator(iterator);
			if (iterator.hasNext()){
				OperationIfc operation = iterator.next();
				task.setCurrentOperation(operation);
				operation.start(task);
			} else
				operationListCompleted(task);	
		} else
			throw new IllegalStateException("Tried to start an already started task through the operation list.");			

	}

	protected final void notifyOperationListCompletedListeners(Task task){
		for(OperationListenerIfc a: myCompletedListeners){
			a.update(this, task);
		}
	}

	protected void operationCompleted(OperationIfc completedOperation, Task task){

		ListIterator<OperationIfc> iterator = task.getOperationIterator();
		
		if (iterator.hasNext()){
			OperationIfc operation = iterator.next();
			task.setCurrentOperation(operation);
			if (task.getStatus() == Task.Status.COMPLETED)
				task.setStatus(Task.Status.READY);
			operation.start(task);
		} else
			operationListCompleted(task);	
		
	}
	
	protected void operationListCompleted(Task task){
		
		notifyOperationListCompletedListeners(task);
		task.setCurrentOperation(null);
		task.setOperationIterator(null);
		//myStartCount = myStartCount - 1;
	}
	
	protected class OperationCompletedListener implements OperationListenerIfc {

		public void update(OperationIfc operation, Task task) {
			operationCompleted(operation, task);
		}
		
	}


}
