package jsl.modeling.elements.operation;
/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
import java.util.LinkedList;
import java.util.List;

import jsl.modeling.ActionListenerIfc;
import jsl.modeling.IllegalStateException;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.operation.OperationListenerIfc;
import jsl.modeling.elements.resource.Request;
import jsl.modeling.elements.resource.Resource;
import jsl.modeling.elements.resource.ResourceAllocationListenerIfc;
import jsl.modeling.elements.resource.ResourceProvider;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;


public class Process extends Operation {

	//  ===========================================
	//        CLASS ATTRIBUTES
	//  ===========================================

	private List<OperationListenerIfc> mySeizedResourceListeners;
	private List<OperationListenerIfc> myInterruptedListeners;
	private List<OperationListenerIfc> myCancelledListeners;
	private List<OperationListenerIfc> myQueuedListeners;

	private ResourceProvider myResourceProvider;
	
	private RandomVariable myProcessTimeRV;
	private RandomVariable myAmtRequestedRV;
	
	private ResponseVariable myTotalProcessTime;
	
	private ResourceAllocationListener myResourceAllocationListener;
	private ProcessCompletedListener myProcessCompletedListener;

	//  ===========================================
	//        CONSTRUCTORS
	//  ===========================================
	
	public Process(ModelElement parent, ResourceProvider resourceProvider) {
		this(parent, resourceProvider, null, new Constant(1), null);
	}
	
	public Process(ModelElement parent, ResourceProvider resourceProvider, DistributionIfc processTimeDF) {
		this(parent, resourceProvider, processTimeDF, new Constant(1), null);
	}

	public Process(ModelElement parent, ResourceProvider resourceProvider, DistributionIfc processTimeDF, DistributionIfc amtRequestedDF) {
		this(parent, resourceProvider, processTimeDF, amtRequestedDF, null);
	}

	public Process(ModelElement parent, ResourceProvider resourceProvider, DistributionIfc processTimeDF, int amtRequested) {
		this(parent, resourceProvider, processTimeDF, new Constant(amtRequested), null);
	}
	
	public Process(ModelElement parent, ResourceProvider resourceProvider, DistributionIfc processTimeDF, String name) {
		this(parent, resourceProvider, processTimeDF, new Constant(1), name);
	}
	
	public Process(ModelElement parent, ResourceProvider resourceProvider, DistributionIfc processTimeDF, DistributionIfc amtRequestedDF, String name) {
		super(parent, name);
		myResourceAllocationListener = new ResourceAllocationListener();
		myProcessCompletedListener = new ProcessCompletedListener();
						
		setResourceProvider(resourceProvider);

		if (processTimeDF == null)
			myProcessTimeRV = new RandomVariable(this, Constant.ZERO);
		else
			myProcessTimeRV = new RandomVariable(this, processTimeDF);

		if (amtRequestedDF == null)
			myAmtRequestedRV = new RandomVariable(this, Constant.ONE);
		else
			myAmtRequestedRV = new RandomVariable(this, amtRequestedDF);
				
		myTotalProcessTime = new ResponseVariable(this, getName() + " TotalTimeInProcess");
		
		mySeizedResourceListeners = new LinkedList<OperationListenerIfc>() ;
		myInterruptedListeners = new LinkedList<OperationListenerIfc>();
		myCancelledListeners = new LinkedList<OperationListenerIfc>();
		myQueuedListeners = new LinkedList<OperationListenerIfc>();
	}

	//  ===========================================
	//        PUBLIC METHODS
	//  ===========================================

	public final void start(Task task){

		if (task == null)
			throw new IllegalArgumentException("The supplied task was null.");

		if (task.getStatus() != Task.Status.READY)
			throw new IllegalStateException("Tried to start a task when it is not READY." + getName() + " Task " + task.getName());

		task.setCurrentOperation(this);
		
		task.setStatus(Task.Status.STARTED);
		
		notifyOperationStartedListeners(task);
			
		int amt = 1;
		if (task.getUseAmountFlag() == true){
			amt = task.getAmount();
		} else{
			amt = getAmountRequested();
			task.setAmount(amt);
		}
		
		Request r = myResourceProvider.seize(amt, getResourceAllocationListener(), task);
		
		task.setRequest(r);
		
		if (r.isQueued()){
			task.setStatus(Task.Status.QUEUED);
			notifyProcessQueuedListeners(task);
		}

	}

	public final void cancel(Task task){
		
		if (task.getCurrentOperation() != this)
			throw new IllegalArgumentException("The task is not experiencing this process.");

		if (task.getStatus() == Task.Status.QUEUED){
			cancelQueuedTask(task);
		} else if (task.getStatus() == Task.Status.USING_RESOURCE){
			cancelUsingResourceTask(task);
		} else if (task.getStatus() == Task.Status.INTERRUPTED){
			cancelInterruptedTask(task);
		} else {
			throw new IllegalStateException("Tried to cancel a task when it is not queued or time delayed using a resource." + getName() + " Task " + task.getName());			
		}
		
	}
	
	public final void interrupt(Task task){
		
		if (task.getCurrentOperation() != this)
			throw new IllegalArgumentException("The task is not experiencing this process.");

		if (task.getStatus() != Task.Status.USING_RESOURCE)
			throw new IllegalStateException("Tried to interrupt a task when it is not time delayed using a resource." + getName() + " Task " + task.getName());			
		
		interruptTask(task);
		
	}
	
	public final void restart(Task task){
		
		if (task.getCurrentOperation() != this)
			throw new IllegalArgumentException("The task is not experiencing this process.");

		if (task.getStatus() != Task.Status.INTERRUPTED)
			throw new IllegalStateException("Tried to restart a task when it is not interrupted." + getName() + " Task " + task.getName());			

		// determine the processing time
		double time = 0.0;
		if (task.getUseDurationFlag() == true){
			time = task.getDurationTime();
		} else {
			time = getProcessTime();
			task.setDurationTime(time);
		}

		// schedule the end of the processing time and attach the request
		JSLEvent e = scheduleEvent(myProcessCompletedListener, time, task.getRequest());
		
		// have the task remember the event if it needs to be cancelled
		task.setTimedEvent(e);
		task.setStatus(Task.Status.USING_RESOURCE);
	}
	
	public void setAmountRequestedDistributionInitialRandomSource(DistributionIfc d){
		myAmtRequestedRV.setInitialRandomSource(d);
	}
	
	public void setAmountRequestedToConstant(int amount){
		myAmtRequestedRV.setInitialRandomSource(new Constant(amount));
	}
	
	public int getAmountRequested(){
		return((int)myAmtRequestedRV.getValue());
	}

	public final void addCancelListener(OperationListenerIfc listener){
		if (listener == null)
	    		throw new IllegalArgumentException("Attempted to add a null listener");
		myCancelledListeners.add(listener);
	}	

	public final void addSeizedResourceListener(OperationListenerIfc listener){
		if (listener == null)
	    		throw new IllegalArgumentException("Attempted to add a null listener");
		mySeizedResourceListeners.add(listener);
	}

	public final void addInterruptListener(OperationListenerIfc listener){
		if (listener == null)
			throw new IllegalArgumentException("Attempted to add a null listener");
		myInterruptedListeners.add(listener);
	}

	//  ===========================================
	//        PROTECTED METHODS
	//  ===========================================

	protected void cancelQueuedTask(Task task){
		
		Request r = task.getRequest();
		myResourceProvider.cancelRequest(r);
		task.setStatus(Task.Status.CANCELLED);
		notifyProcessCancelledListeners(task);
		task.setStatus(Task.Status.READY);
		
	}
	
	protected void cancelUsingResourceTask(Task task){
		
		JSLEvent e = task.getTimedEvent();
		if (e == null)
			throw new IllegalStateException("Tried to cancel a task when it is not scheduled." + getName() + " Task " + task.getName());
		
		cancelEvent(e);
		
		Request r = task.getRequest();
		r.releaseResource();

		task.setStatus(Task.Status.CANCELLED);
		
		notifyProcessCancelledListeners(task);
		
		task.setStatus(Task.Status.READY);
		
	}
	
	protected void interruptTask(Task task){

		JSLEvent e = task.getTimedEvent();
		if (e == null)
			throw new IllegalStateException("Tried to cancel a task when it is not scheduled." + getName() + " Task " + task.getName());
		
		cancelEvent(e);

		task.setStatus(Task.Status.INTERRUPTED);
		task.setInterruptTime();
		notifyProcessInterruptedListeners(task);
		
	}
	
	protected void cancelInterruptedTask(Task task){
		
		Request r = task.getRequest();
		r.releaseResource();

		task.setStatus(Task.Status.CANCELLED);
		
		notifyProcessCancelledListeners(task);
		
		task.setStatus(Task.Status.READY);

	}
	
	protected void setProcessTimeDistributionInitialRandomSource(DistributionIfc distribution) {

		if (distribution == null)
    			throw new IllegalArgumentException("Attempted to set the process time distribution to null!");

		myProcessTimeRV.setInitialRandomSource(distribution);

	}
	
	protected double getProcessTime(){
		return(myProcessTimeRV.getValue());
	}

	protected final ResourceProvider getResourceProvider() {
		return myResourceProvider;
	}

	protected final void setResourceProvider(ResourceProvider resourceProvider) {
		if (resourceProvider == null)
			throw new IllegalArgumentException("Attempted to set the ResourceProvider to null!");
		myResourceProvider = resourceProvider;
	}
		
	protected final void notifyProcessQueuedListeners(Task task){
		for(OperationListenerIfc a: myQueuedListeners){
			a.update(this, task);
		}
	}	
	
	protected final void notifyProcessSeizedResourceListeners(Task task){
		for(OperationListenerIfc a: mySeizedResourceListeners){
			a.update(this, task);
		}
	}
	
	protected final void notifyProcessCancelledListeners(Task task){		
		for(OperationListenerIfc a: myCancelledListeners){
			a.update(this, task);
		}
	}
	
	protected final void notifyProcessInterruptedListeners(Task task){
		for(OperationListenerIfc a: myInterruptedListeners){
			a.update(this, task);
		}
	}
	
	protected final ResourceAllocationListener getResourceAllocationListener() {
		return myResourceAllocationListener;
	}

	protected final void removeCancelListener(OperationListenerIfc listener){
		if (listener == null)
	    		throw new IllegalArgumentException("Attempted to remove a null listener");
		myCancelledListeners.remove(listener);
	}

	protected final void removeSeizedResourceListener(OperationListenerIfc listener){
		if (listener == null)
	    		throw new IllegalArgumentException("Attempted to remove a null listener");
		mySeizedResourceListeners.remove(listener);
	}

	protected final void removeInterruptListener(OperationListenerIfc listener){
		if (listener == null)
			throw new IllegalArgumentException("Attempted to remove a null listener");
		myInterruptedListeners.remove(listener);
	}
	
	private class ResourceAllocationListener implements ResourceAllocationListenerIfc {

		public void allocated(Request request, Resource resource) {
			// the processing needs to be started now that the resource has been provided
			Task task = (Task)request.getRequestingObject();
			task.setRequest(request);
			
			task.setStatus(Task.Status.SEIZED_RESOURCE);
			// notify any listeners that the task was allocated a resource
			notifyProcessSeizedResourceListeners(task);
			// determine the processing time
			double time = 0.0;
			if (task.getUseDurationFlag() == true){
				time = task.getDurationTime();
			} else {
				time = getProcessTime();
				task.setDurationTime(time);
			}

			// schedule the end of the processing time and attach the request
			JSLEvent e = scheduleEvent(myProcessCompletedListener, time, request);
			
			// have the task remember the event if it needs to be cancelled
			task.setTimedEvent(e);
			task.setStatus(Task.Status.USING_RESOURCE);

		}	
	}

	private class ProcessCompletedListener implements ActionListenerIfc {

		public void action(JSLEvent event) {
			// get the attached request
			Request request = (Request) event.getMessage();
			
			// tally waiting time in process
			myTotalProcessTime.setValue(getTime() - request.getTimeEnteredQueue());

			request.releaseResource();
			Task task = (Task)request.getRequestingObject();
			task.setStatus(Task.Status.COMPLETED);
			notifyOperationCompletedListeners(task);
			task.setStatus(Task.Status.READY);
			
		}
		
	}
}
