package jsl.modeling.elements.operation;
/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
import java.util.ListIterator;

import jsl.modeling.JSLEvent;
import jsl.modeling.Model;
import jsl.modeling.elements.processview.description.EntityType;
import jsl.modeling.elements.resource.Entity;
import jsl.modeling.elements.resource.Request;
import jsl.modeling.elements.spatial.CoordinateIfc;


public class Task extends Entity{

	/**
	 * READY: Represents that the task is ready to be started
	 * STARTED: Represents that the task has started an operation
	 * TIME_DELAYED: Represents that the task is experiencing a time delay
	 * QUEUED: Represents that the task is in a queue
	 * CANCELLED: Represents that the task had started a time delay but had been cancelled
	 * INTERRUPTED:  Represents that the task had started a time delay but had been interrupted
	 * SEIZED_RESOURCE: Represents that the task has seized a resource
	 * USING_RESOURCE: Represents that the task is using a resource within a time delay
	 * COMPLETED:  Represents that the task has completed a time delay.
	 * 
	 */
	public static enum Status {READY, STARTED, TIME_DELAYED, QUEUED, CANCELLED, INTERRUPTED, SEIZED_RESOURCE, USING_RESOURCE, COMPLETED};

	//  ===========================================
	//        CLASS ATTRIBUTES
	//  ===========================================

	/** used to track the operation the task is executing
	 *  when using operation lists
	 * 
	 */
	private ListIterator<OperationIfc> myOperationIterator;
	
	/** The current operation that the task is executing
	 * 
	 */
	private OperationIfc myCurrentOperation;

	/** The request associated with the current operation
	 * 
	 */
	private Request myRequest;

	/** Used to remember the event associated with the task
	 *  when it is in an activity or delay
	 */
	private JSLEvent myTimedEvent;
	
	/** If the task is interrupted while experiencing a time delay
	 *  this field will hold the time that the interrupt occured
	 */
	private double myInterruptTime;
	
	/** Indicates the status of the task
	 */
	private Status myStatus;
	
	/** Can be used to indicate a location in 2D space where
	 *  the task is originating
	 */
	private CoordinateIfc myOrigin;
	
	/** Can be used to indicate a location in 2D space where the 
	 *  task is going to.
	 */
	private CoordinateIfc myDestination;

	//  ===========================================
	//        CONSTRUCTORS
	//  ===========================================

	/**
	 * 
	 */
	public Task() {
		this(null, null);
	}

	/**
	 * @param name
	 */
	public Task(String name) {
		this(name, null);
	}
	
	/**
	 * @param entityType
	 */
	public Task(EntityType entityType) {
		this(null, entityType);
	}

	/**
	 * @param name
	 * @param entityType
	 */
	public Task(String name, EntityType entityType) {
		super(name, entityType);
		myStatus = Status.READY;
	}

	//  ===========================================
	//        PUBLIC METHODS
	//  ===========================================

	public final OperationIfc getCurrentOperation() {
		return myCurrentOperation;
	}

	public final ListIterator<OperationIfc> getOperationIterator() {
		return myOperationIterator;
	}

	public final Request getRequest() {
		return myRequest;
	}

	public final JSLEvent getTimedEvent() {
		return myTimedEvent;
	}
	
	public final double getInterruptTime() {
		return myInterruptTime;
	}

	public final void setInterruptTime() {
		myInterruptTime = Model.getTime();
	}

	public final double getTimeLeftAfterInterrupt() {
		return myTimedEvent.getTime() - myInterruptTime;
	}
	
	public final void setStatus(Status status){
		myStatus = status;
	}
	
	public final Status getStatus(){
		return myStatus;
	}

	//  ===========================================
	//        PROTECTED METHODS
	//  ===========================================

	protected final void setTimedEvent(JSLEvent event) {
		myTimedEvent = event;
	}

	protected final void setRequest(Request request) {
		myRequest = request;
	}

	protected final void setOperationIterator(ListIterator<OperationIfc> operationIterator) {
		myOperationIterator = operationIterator;
	}

	protected final void setCurrentOperation(OperationIfc currentOperation) {
		myCurrentOperation = currentOperation;
	}

	public final CoordinateIfc getOrigin() {
		return myOrigin;
	}

	public final void setOrigin(CoordinateIfc origin) {
		myOrigin = origin;
	}

	public final CoordinateIfc getDestination() {
		return myDestination;
	}

	public final void setDestination(CoordinateIfc destination) {
		myDestination = destination;
	}

}
