/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.processview.description;

import java.util.*;

import jsl.modeling.elements.*;
import jsl.modeling.*;

/**
 *
 */
public class EntityType extends ModelElement {

	/** A reference to the set of AttributeTypes
	 */
	private List<AttributeType> myAttributeTypes;
	
	/** A hash set to keep track of what names have been
	 *  applied to AttributeTypes
	 */
	private HashSet<String> myAttributeTypeNames;
	
	/** A list to hold the entity's list of receivers to visit
	 * 
	 */
	private List<EntityReceivingIfc> myReceiverSequence;

	public EntityType(ModelElement parent){
		this(parent, null, null);
	}
		
	public EntityType(ModelElement parent, String name){
		this(parent, name, null);
	}
	
	/**
	 * @param name
	 * @param receiverSequence, a List of objects that implement ReceiveEntityIfc
	 */
	public EntityType(ModelElement parent, String name, List<EntityReceivingIfc> receiverSequence) {
		super(parent, name);
		if (receiverSequence != null)
	    	setReceiverSequence(receiverSequence);
	    myAttributeTypes = new ArrayList<AttributeType>();
		myAttributeTypeNames = new HashSet<String>();
	}

	/** Defines an attribute type with the given name and
	 *  adds it to the available types for this entity type
	 * 
	 * @param name, the name of the attribute type, must be non-null
	 *  and unique to this entity type
	 */
	public final AttributeType defineAttributeType(String name){
		if(name == null)
			throw new IllegalArgumentException("Name of attribute type must be non-null!");
		
		if (!myAttributeTypeNames.add(name))
			throw new IllegalArgumentException("Name of attribute type must be unique for this process description!");
		
		AttributeType at = new AttributeType(name);
				
		myAttributeTypes.add(at);
		return(at);
	}
	
	/** Returns an iterator to the list of attribute types
	 * 
	 * @return
	 */
	public final Iterator<AttributeType> getAttributeTypesIterator(){
		return(myAttributeTypes.iterator());
	}
	
	/** Adds a receiver to the receiver sequence
	 * 
	 * @param receiver
	 */
	public final void addEntityReceiver(EntityReceivingIfc receiver){
		if(receiver == null)
			throw new IllegalArgumentException("ReceiveEntityIfc receiver must be non-null!");
		if (myReceiverSequence ==  null)
			myReceiverSequence = new ArrayList<EntityReceivingIfc>();
			
		myReceiverSequence.add(receiver);
	}
	
	/** Gets an EntityReceiverIteratorIfc to the sequence of receivers
	 *  positioned at the beginning of the sequence
	 * 
	 * @return, the iterator or null if no receiver list
	 */
	public final EntityReceiverIteratorIfc getEntityReceiverIterator(){
		if (myReceiverSequence == null)
			return null;
		else
			return(new EntityReceiverListIterator(myReceiverSequence));
	}
	
	protected final void setReceiverSequence(List<EntityReceivingIfc> receiverSequence){
		if(receiverSequence == null)
			throw new IllegalArgumentException("List receiverSequence must be non-null!");
		myReceiverSequence = receiverSequence;
	}

}
