/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling.elements.processview.description;

import jsl.modeling.IllegalStateException;
import java.util.*;

import jsl.modeling.elements.resource.Entity;
import jsl.utilities.reporting.JSL;

public class ProcessExecutor {

	/**
	 * A reference to the process description that is being executed
	 */
	private ProcessDescription myProcessDescription;

	/**
	 * A reference to the list of commands for the process description that is being
	 * executed
	 */
	private List<ProcessCommand> myCommandList;

	/**
	 * The index (zero based) to the current command to be executed in 
	 * the process description
	 */
	private int myCurrentCommandIndex;

	/** 
	 *  A reference to the entity that is currently involved
	 *  in the execution of the process description for this executor
	 */
	private Entity myCurrentEntity;
	
	/**
	 * A reference to the current state of process executor
	 */
	private ProcessExecutorState myState;

	/** Indicates that the process executor is jumping to 
	 *  another command
	 */
	private boolean myJumpToFlag;

	/**
	 * A reference to the created state for the iterative process A process
	 * executor is in the created state when it is initially constructed and can
	 * then only transition to the initialized state
	 */
	private Created myCreatedState = new Created();

	/**
	 * A reference to the initialized state of the process executor A process
	 * executor is in the initialized state after the initialize() method is
	 * called from a proper state.
	 */
	private Initialized myInitializedState = new Initialized();

	/**
	 * A reference to suspended state of the process executor A process executor
	 * is in the suspended state after the suspend method is called from a
	 * proper state
	 */
	private Suspended mySuspendedState = new Suspended();

	/**
	 * A reference to Executing state of the process executor A process executor
	 * is in the Executing state after the resume method is called from a proper
	 * state
	 */
	private Executing myExecutingState = new Executing();

	/**
	 * A reference to the terminated state of the process executor A process
	 * executor is in the terminated state after the terminate method is called
	 */
	private Terminated myTerminatedState = new Terminated();

	/** Holds the listeners that are notified prior to execution
	 * 
	 */
	private List<ProcessExecutorListenerIfc> myBeforeExecutionListeners;
	
	/** Holds the listeners that are notified after the execution terminates
	 * 
	 */
	private List<ProcessExecutorListenerIfc> myAfterExecutionListeners;

	/** Creates a process executor given the supplied
	 *  process description and entity
	 * 
	 * @param processDescription
	 * @param entity
	 */
	protected ProcessExecutor(ProcessDescription processDescription, Entity entity) {
		if (processDescription == null)
			throw new IllegalArgumentException(
					"ProcessDescription must be non-null!");

		if (entity == null)
			throw new IllegalArgumentException(
					"Entity must be non-null!");

		myBeforeExecutionListeners  = new LinkedList<ProcessExecutorListenerIfc>();
		myAfterExecutionListeners  = new LinkedList<ProcessExecutorListenerIfc>();

		myProcessDescription = processDescription;
		myCommandList = myProcessDescription.getProcessCommands();
		myCurrentCommandIndex = -1;
		myState = myCreatedState;
		myCurrentEntity = entity;
		myCurrentEntity.setProcessExecutor(this);
	}

	/** Returns the total number of commands
	 *  in the associated process description
	 * @return The number of commands in the associated process description
	*/
	public final int getNumberOfCommands(){
		return(myCommandList.size());
	}
	
	/** Returns an integer representing the location of the current command 
	 *  within process description, zero based
	 * 
	 * @return A integer representing the location of the command
	 */
	public final int getCurrentCommandIndex(){
		return(myCurrentCommandIndex);
	}
	
	/** Gets the entity currently executing
	 *  
	 * @return the entity
	 */
	public final Entity getCurrentEntity(){
		return(myCurrentEntity);
	}
	
	/** Adds a listener to be called prior to the execution
	 * 
	 * @param listener
	 */
	public final void addBeforeExecutionListener(ProcessExecutorListenerIfc listener){
		if (listener == null)
	    		throw new IllegalArgumentException("Attempted to add a null listener");
		if (myBeforeExecutionListeners.contains(listener))
			throw new IllegalArgumentException("Supplied listener was already added.");
		myBeforeExecutionListeners.add(listener);
	}

	/** Removes the listener that is called prior to the execution 
	 * 
	 * @param listener
	 */
	public final void removeBeforeExecutionListener(ProcessExecutorListenerIfc listener){
		if (listener == null)
	    		throw new IllegalArgumentException("Attempted to remove a null listener");
		myBeforeExecutionListeners.remove(listener);
	}

	/** Adds a listener to be called after the execution 
	 * 
	 * @param listener
	 */
	public final void addAfterExecutionListener(ProcessExecutorListenerIfc listener){
		if (listener == null)
	    		throw new IllegalArgumentException("Attempted to add a null listener");
		if (myAfterExecutionListeners.contains(listener))
			throw new IllegalArgumentException("Supplied listener was already added.");
		myAfterExecutionListeners.add(listener);
	}

	/** Removes the listener that is called after the execution 
	 * 
	 * @param listener
	 */
	public final void removeAfterExecutionListener(ProcessExecutorListenerIfc listener){
		if (listener == null)
	    		throw new IllegalArgumentException("Attempted to remove a null listener");
		myAfterExecutionListeners.remove(listener);
	}
	
	/** After being created the process executor must be intialized
	 * before being started
	 */
	public final void initialize(){
		myCurrentCommandIndex = -1;
		myState.initialize();
	}
	
	/** Starts the process executor executing at the first command 
	 * in the associated process description
	 */
	public final void start() {
		start(0);
	}
	
	/** Starts the process executor executing at the command indicated
	 *  by the commandIndex.  Note: It is up to the user to ensure that
	 *  the command is an appropriate location to start executing
	 * 
	 * @param commandIndex represents the index in the sequence of commands,
	 *        index = 0, represents the first command
	 * @throws IndexOutOfBoundsException - if the index is out of range (index < 0 || index >= Number of commands).
	 */
	public void start(int commandIndex){
		if(commandIndex < 0 || commandIndex >= myCommandList.size())
			throw new IndexOutOfBoundsException("Index was out of range, zero based referencing");
		notifyBeforeExecutionListeners();
		myState.start(commandIndex);
	}

	/** Resumes the execution of the process executor at the next command
	 *  after the previously executed command
	 */
	public final void resume(){
		resume(getNextCommandIndex());
	}
	
    /** Resumes the execution of the process executor at the command indicated
	 *  by the commandIndex.  Note: It is up to the user to ensure that
	 *  the indicated command is an appropriate location to resume executing
	 *  A command index out of range will cause the process executor to terminate
     * 
     * @param commandIndex
     */
	public void resume(int commandIndex){
		myState.resume(commandIndex);
	}

	/** Cause the execution of the process executor to jump to the command indicated
	 *  by the commandIndex.  Note: It is up to the user to ensure that
	 *  the indicated command is an appropriate location to continue executing.
	 *  This method is only valid in executing state.
	 *  A command index out of range will cause the process executor to terminate
	 * 
	 * @param commandIndex
	 */
	public void jumpTo(int commandIndex){
		myState.jumpTo(commandIndex);
	}
	
	/**  Suspends the execution of the process executor at the current command
	 *
	 */
	public void suspend(){
		myState.suspend();
	}

	/** Terminates the execution of the process executor 
	 *
	 */
	public void terminate(){
		myState.terminate();
		myProcessDescription.processExecutorTerminated(this);
		notifyAfterExecutionListeners();
	}
	
	/** Creates a sub-process executor on this executor that will
	 *  execute the given process description
	 * 
	 * @param processDescription
	 * @return A process executor to execute the sub-process
	 */
	public ProcessExecutor createSubProcessExecutor(ProcessDescription processDescription){
		if (processDescription == null)
			throw new IllegalArgumentException(
					"ProcessDescription must be non-null!");
					
		if (processDescription == myProcessDescription)
			throw new IllegalArgumentException(
					"ProcessDescription must be different from current process description.");

		return(new SubProcessExecutor(this, processDescription, myCurrentEntity));
	}
	
	/**
	 * Checks if the process executor is in the created state. After the process
	 * executor has been created this method will return true
	 * 
	 * @return true if created
	 */
	public final boolean isCreated() {
		return (myState == myCreatedState);
	}

	/**
	 * Checks if the process executor is in the initialized state. After the
	 * process executor has been initialized this method will return true
	 * 
	 * @return true if initialized
	 */
	public final boolean isInitialized() {
		return (myState == myInitializedState);
	}

	/**
	 * Checks if the process executor is in the terminated state After the
	 * process executor has been terminated this method will return true
	 * 
	 * @return true if terminated
	 */
	public final boolean isTerminated() {
		return (myState == myTerminatedState);
	}

	/**
	 * Checks if the process executor is in the suspended state After the
	 * process executor has been suspended this method will return true
	 * 
	 * @return true if suspended
	 */
	public final boolean isSuspended() {
		return (myState == mySuspendedState);
	}

	/**
	 * Checks if the process executor is in the Executing state After the
	 * process executor has been resumed this method will return true
	 * 
	 * @return true if executing
	 */
	public final boolean isExecuting() {
		return (myState == myExecutingState);
	}
	
	/** Gets the command associated with the current command index
	 *  This method returns null if the index is not valid (out of range).
	 * 
	 * @return the command at the current command index
	 */
	protected final ProcessCommand getCurrentCommand() {
		int i = getCurrentCommandIndex();	
		if (i < 0 || i >= myCommandList.size())
			return(null);
		else
			return ((ProcessCommand)myCommandList.get(i));
	}
	
	/** Sets the index to the current command.  The indexing is zero based,
	 *  i.e. 0=first command
	 *  An index < 0 or > number of commands, implies no valid command, i.e.
	 *  the current command will be null
	*/
	protected final void setCurrentCommandIndex(int index){	
		if(index == myCurrentCommandIndex)
			throw new IllegalArgumentException("Tried to set current command index to last command index.\n" +
				" causing infinite loop! Current index is = " + index);
				
		myCurrentCommandIndex = index;
	}
	
	/** The main method that executes commands within the ProcessExecutor
	 */
	protected final void execute() {

		while (!isTerminated() && !isSuspended()) {
			ProcessCommand c = getCurrentCommand();
			if (c != null){
				JSL.LOGGER.finest("ProcessExecutor Executing " + myProcessDescription.getName() +
				                  " Command Index = " + myCurrentCommandIndex + 
				                  " Command = " + c.getName()); 
				c.execute(this);
				if (isExecuting()){// still in the executing state
				  // either we are going to execute the next command or we are going to another command
				  if(myJumpToFlag == true){
				  	// next command determined by jumpTo
				  	// reset flag for next command
				  	myJumpToFlag = false;
				  } else { // not jumping, need to execute the next command
					setCurrentCommandIndex(getNextCommandIndex());
				  }
				}
			} else
				terminate();		
		}
	}
	
	/** Used to get the command index that is next to be executed
	 * 
	 * @return
	 */
	protected final int getNextCommandIndex(){
		return(getCurrentCommandIndex() + 1);
	}
	
	/** Used to set the state of the executor
	 * 
	 * @param state
	 */
	protected final void setState(ProcessExecutorState state){
		if (state == null)
			throw new IllegalArgumentException("ProcessExecutorState must be non-null!");
		myState = state;
	}
	
	/** Notifies the before execution listeners
	 */
	protected final void notifyBeforeExecutionListeners(){
		for(ProcessExecutorListenerIfc a: myBeforeExecutionListeners){
			a.update(this);
		}
	}

	/** Notifies the before execution listeners
	 */
	protected final void notifyAfterExecutionListeners(){
		for(ProcessExecutorListenerIfc a: myAfterExecutionListeners){
			a.update(this);
		}
	}
	
	private class ProcessExecutorState {

		private String myName;

		public ProcessExecutorState(String name) {
			myName = name;
		}

		public void initialize() {
			throw new IllegalStateException(
					"Tried to initialize from an illegal state: " + myState);
		}
		
		public void start(int commandIndex) {
			throw new IllegalStateException(
					"Tried to start from an illegal state: " + myState);
		}
		
		public void suspend() {
			throw new IllegalStateException(
					"Tried to suspend from an illegal state: " + myState);
		}

		public void jumpTo(int commandIndex) {
			throw new IllegalStateException(
					"Tried to jumpTo a ProcessCommand from an illegal state: " + myState);
		}
				
		public void resume(int commandIndex) {
			throw new IllegalStateException(
					"Tried to resume at a ProcessCommand from an illegal state: " + myState);
		}
		
		public void terminate() {
			throw new IllegalStateException(
					"Tried to terminate from an illegal state: " + myState);
		}

		public String toString() {
			return (myName);
		}
	}

	private final class Created extends ProcessExecutorState {

		public Created() {
			super("CreatedState");
		}

		public void initialize() {
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("ProcessExecutor Initializing " + myProcessDescription.getName()); 

			// Go from created state to initialized, set state to initialized
			                  
			setState(myInitializedState);
			
		}
	}

	private final class Initialized extends ProcessExecutorState {

		public Initialized() {
			super("InitializedState");
		}

		public void start(int commandIndex){
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("ProcessExecutor Starting " + myProcessDescription.getName()); 

			// we are in the initialized state
			// we want to start executing at the supplied command index
			
			setCurrentCommandIndex(commandIndex);
			
			// set the state to executing
			setState(myExecutingState);
			
			// tell the executor to begin executing
			execute();
			
		}
		
		public void terminate() {
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("ProcessExecutor Terminating " + myProcessDescription.getName());

			setState(myTerminatedState);
		}
	}

	private final class Executing extends ProcessExecutorState {

		public Executing() {
			super("ExecutingState");
		}

		public void jumpTo(int commandIndex) {
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("ProcessExecutor " + myProcessDescription.getName() + 
						" Jumping to command index = " + commandIndex + 
						" Command = " + (ProcessCommand)myCommandList.get(commandIndex));
							  
			// we are in the executing state and want to
			// continue executing the indicated index

			setCurrentCommandIndex(commandIndex);
			
			// set the flag to indicate that we are jumping
			myJumpToFlag = true;
			
			// no need to set the state to executing, since we're
			// all ready in the executing state
			
		}
				
		public void suspend() {
			
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("ProcessExecutor Suspending " + myProcessDescription.getName());

			setState(mySuspendedState);
		}
		
		public void terminate() {
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("ProcessExecutor Terminating " + myProcessDescription.getName());
							  
			setState(myTerminatedState);
		}	
	}

	private final class Suspended extends ProcessExecutorState {

		public Suspended() {
			super("SuspendedState");
		}

		public void resume(int commandIndex) {
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("ProcessExecutor Resuming " + myProcessDescription.getName());			
			// we are in the suspended state
			// we want to continue executing at the indicated index
			
			setCurrentCommandIndex(commandIndex);
			
			// set the state to executing
			setState(myExecutingState);
			
			// tell the executor to continue executing
			execute();

		}
		
		public void terminate() {
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("ProcessExecutor Terminating " + myProcessDescription.getName());
							  
			setState(myTerminatedState);
		}

	}

	private final class Terminated extends ProcessExecutorState {

		public Terminated() {
			super("TerminatedState");
		}

		public void initialize() {
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("ProcessExecutor Initializing " + myProcessDescription.getName());

			setState(myInitializedState);
		}
	}

}