/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.queue;

import jsl.utilities.reporting.JSL;
import jsl.utilities.*;
import jsl.modeling.IllegalStateException;
import jsl.modeling.Model;
import jsl.modeling.ModelElement;
import jsl.modeling.State;
import jsl.modeling.StateAccessorIfc;

/** QObject can be used as a base class for objects that need
 *  to be placed in queues on a regular basis.  In addition, a QObject
 *  is created each time that an arbitrary object is placed in a Queue.
 *  A QObject "wraps" the arbitrary object, so that statistics can be
 *  collected on arbitrary objects placed in Queues.  A QObject can 
 *  be in one and only one Queue at a time.  An arbitrary object (not subclassed
 *  from QObject) can be placed in multiple Queues because a new QObject is used each time
 *  the object is placed in a Queue.  If a QObject is placed in a Queue
 *  no wrapping QObject is created, thereby saving an additional object creation (with
 *  the restriction that it can be in only one Queue at a time).
 *
 */
public class QObject implements IdentityIfc, Comparable<QObject> {
	
	/** incremented to give a running total of the 
	 *  number of model QObject created
	 */
	private static long myCounter_;
	
	/** The id of the QObject, currently if
	 *  the QObject is the ith QObject created
	 *  then the id is equal to i
	 */
	private long myId;

	/** The name of the QObject
	 */
	private String myName;
	
	/** A state representing that the QObject was created
	 */
	private State myCreatedState;

	/** A state representing when the QObject was queued
	 */
	private State myQueuedState;
	
	/** A priority for use in queueing
	 */
	private int myPriority;

	/** The current queue that the QObject is in, null if not in a queue
	 */
	private Queue myQueue;

	/** A reference to an object that can be attached to the QObject
	 *  when queued
	 */
	private Object myQueuedObject;

	/** can be used to time stamp the qObject
	 */
	protected double myTimeStamp;
		
	//  ===========================================
	//        CONSTRUCTORS
	//  ===========================================

	/** Creates a QObject with name "null"
     *  and the creation time set to the current simulation time
	 */
	public QObject() {
		this(null);
	}

	/** Creates an QObject with the given name 
     *  and the creation time set to the current simulation time
	 * @param name The name of the QObject
	 */
	public QObject(String name) {
		initialize(name);

		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
			JSL.LOGGER.finest("Constructed \t Class " + this.getClass().getName() + "\t Name: " + getName() + "\t ID: " + getId());

	}

	//  ===========================================
	//        PUBLIC INTERFACE
	//  ===========================================

	/*  --Accessors---------------------- */

	public String toString(){
		return("ID= "+getId() + " Name= " + getName());
	}
	
	/** Gets this QObject's name.
	 * @return The name of the QObject.
	 */
	public final String getName() {
		return myName;
	}

	/** Gets a uniquely assigned integer identifier for this QObject.
	 * This identifier is assigned when the QObject is
	 * created.  It may vary if the order of creation changes.
	 * @return The identifier for the entity.
	 */
	public final long getId() {
		return (myId);
	}
	
    /** Gets the time that the QObject was created
     * @return A double representing simulated creation time
     */
    public final double getCreateTime(){
        return(myCreatedState.getTimeStateEntered());
    }
    
    /** Returns the total elapsed time since the object was created
     * 
     * @return
     */
    public final double getTimeSinceCreation(){
    	return(getTime()-getCreateTime());
    }

	/** Gets the queueing priority associated with this QObject
	 * @return The priority as an int
	 */    
	public final int getPriority() {
		return(myPriority);
	}

    /** Returns the queue that the QObject was last enqueued within
     *  
     * @return The Queue, or null if no queue
    */ 
	public final Queue getQueue(){
		return(myQueue);
	}

	/** Gets the time the QObject was LAST enqueued
	 * @return A double representing the time the QObject was enqueued
	 */    
	public final double getTimeEnteredQueue() {
		return(myQueuedState.getTimeStateEntered());
	}

	/** Gets the time the QObject LAST exited a queue
	 * @return A double representing the time the QObject last exited a QObject
	 */    
	public final double getTimeExitedQueue() {
		return(myQueuedState.getTimeStateExited());
	}

	/** This method can be used to get direct access to the State
	 *  that represents when the object was queued.  This allows
	 *  access to the total time in the queued state as well as other 
	 *  statistical accumulation of state statistics
	 *  
	 * @return Returns the QueuedState.
	 */
	public final StateAccessorIfc getQueuedState() {
		return myQueuedState;
	}	
	
	/** If there is no attached queued object then this method returns null
	 * @return A reference to the attached object
	 */    
	public final Object getQueuedObject(){
		return(myQueuedObject);
	}

	/** Checks if the QObject is queued
	 *  
	 * @return true if it is queued
	 */
	public final boolean isQueued(){
		return(myQueuedState.isEntered());
	}
	
	/* mutators --------------------------*/
	
	/** Sets the priority to the supplied value
	 *  If the QObject is queued, the queue's changePriority()
	 *  method is called (possibly causing a reordering of the queue)
	 *  which may cause significant reordering overhead
	 *  otherwise the priority is directly changed
	 *  Changing this value only changes how the QObjects are compared
	 *  and may or may not change how they are ordered in the queue,
	 *  depending on the queue discipline used
	 *  
	 * @param priority An integer representing the priority of the QObject
	 */
	public final void setPriority(int priority){
		if (isQueued())
			myQueue.changePriority(this, priority);
		else
			setPriority_(priority);
	}
	
	/** Sets the name of this QObject
	 * @param str The name as a string.
	 */
	public final void setName(String str) {
		myName = str;
	}
	
	//  ===========================================
	//        PROTECTED METHODS
	//  ===========================================

	/** Causes the QObject to look like new,
	 *  gets a new name, number, priority is reset to 1, 
	 *  states are initialized, and starts in created state.  
	 *  As if newly, created.  Useful if reusing QObjects
	 * 
	 * @param name
	 */
	protected final void initialize(String name){
		myCounter_ = myCounter_ + 1;
		myId = myCounter_;
		setName(name);
		myPriority = 1;
		myQueue = null;
		myQueuedObject = null;
		
		if (myCreatedState == null)
			myCreatedState = new State(getName()+"Created");
		else
			myCreatedState.initialize();
		myCreatedState.enter();
			
		if (myQueuedState == null)
			myQueuedState = new State(getName()+"Queued");
		else
			myQueuedState.initialize();

	}

	/** Indicates that the QObject enters the queue
	 * 
	 */
	protected final void enterQueue(Queue queue, int priority, Object obj) {
		if (queue == null)
			throw new IllegalArgumentException("The Queue must be non-null");
		if (myQueuedState.isEntered())
			throw new IllegalStateException("The QObject was already queued!");
		myQueuedState.enter();
		setQueue(queue);
		setPriority_(priority);
		setQueuedObject(obj);
	}
	
	/** Indicates that the QObject exited the queue
	 * 
	 */
	protected final void exitQueue() {
		if (!myQueuedState.isEntered())
			throw new IllegalStateException("The QObject was not in a queue!");
		myQueuedState.exit();
		setQueue(null);
	}   
    
	/** Gets the current simulated time
	 * @return
	 */
	protected static final double getTime() {
		return(ModelElement.getTime());
	}
		
    /** Sets the queueing priority for this QObject
     *  Changing the priority while the object is in a queue
     *  has no effect on the ordering of the queue.  This priority is
     *  only used to determine the ordering in the queue when the item
     *  enters the queue.
     * 
     * @param priority, lower priority implies earlier ranking in the queue
     */
	protected final void setPriority_(int priority){
		myPriority = priority;
	}

    /** Sets the queue that the QObject is enqueued
     *  
     * @param The Queue that the object is enqueued
     */ 
	protected void setQueue(Queue queue){
		myQueue = queue;
    }
 
    /** Sets an object that can be queued with this QObjectIfc object
     *  
     * @param The attached object
     */ 
	public final void setQueuedObject(Object obj){
		myQueuedObject = obj;
    }
    
    /** Used to make the QObject not have any references,
     *  e.g. to a Queue and to an Object that was queued
     *
     */
    protected void setNulls(){
    		myQueuedObject = null;
    		myQueue = null;
    }
    
    /** Causes all references to objects from this QObject
     *  to be set to null and all internal objects to be set to null
     *  including State information
     *  
     *  Meant primarily to facilitate garbage collection.  After
     *  this call, the object should not be used.
     * 
     */
    public void nullify(){
    	setNulls();
        myName = null;
     	myCreatedState = null;
    	myQueuedState = null;
    }
    
	//  ===========================================
	//        Comparable Interface
	//  ===========================================
	
	/** Returns a negative integer, zero, or a positive integer
	 * if this object is less than, equal to, or greater than the
	 * specified object.
	 *
	 * Throws ClassCastException if the specified object's type
	 * prevents it from begin compared to this object.
	 *
	 * Throws RuntimeException if the id's of the objects are the same,
	 * but the references are not when compared with equals.
	 *
	 * Note:  This class may have a natural ordering that is inconsistent
	 * with equals.
	 * @param obj The object to compare to
	 * @return Returns a negative integer, zero, or a positive integer
	 * if this object is less than, equal to, or greater than the
	 * specified object.
	 */    
	public final int compareTo(QObject obj) {
		QObject qObj = (QObject )obj;
        
		// compare the priorities
        
		if (myPriority < qObj.getPriority())
			return (-1);
        
		if (myPriority > qObj.getPriority())
			return (1);
        
		// priorities are equal, compare time stamps
        
		if (getTimeEnteredQueue() < qObj.getTimeEnteredQueue())
			return (-1);
        
		if (getTimeEnteredQueue() > qObj.getTimeEnteredQueue())
			return (1);
        
		// time stamps are equal, compare ids
        
		if (myId < qObj.getId()) // lower id, implies created earlier
			return (-1);
        
		if (myId > qObj.getId())
			return (1);
        
		// if the id's are equal then the object references must be equal
		// if this is not the case there is a problem
        
		if (this.equals(obj))
			return(0);
		else
			throw new RuntimeException("Id's were equal, but references were not, in QObject compareTo");
        
	}

	/**
	 * @return Returns the TimeStamp.
	 */
	public final double getTimeStamp() {
		return myTimeStamp;
	}

	/**
	 * @param timeStamp The timeStamp to set.
	 */
	public final void setTimeStamp(double timeStamp) {
		if (timeStamp < Model.getTime())
			throw new IllegalArgumentException("The time stamp was less than the current time.");
		myTimeStamp = timeStamp;
	}

	/**
	 * Sets timestamp to current simulated time
	 */
	public final void setTimeStamp() {
		myTimeStamp = Model.getTime();
	}

	/** Returns the amount of time since
	 *  the last time stamp
	 * 
	 * @return The amt of time since last time stamp
	 */
	public final double getElapsedTime() {
		return(Model.getTime()-myTimeStamp);
	}

}
