/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling.elements.queue;

import java.util.*;

import jsl.modeling.elements.variable.Aggregate;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.AveragePerTimeWeightedVariable;
import jsl.modeling.elements.variable.ResponseStatisticsIfc;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.modeling.ModelElement;
import jsl.utilities.misc.Predicate;


/** The Queue class provides the ability to hold entities (QObjects)
 * within the model.  Any object can be added to a Queue.  When
 * an object is added to a Queue, the object is wrapped by a
 * QObject which provides statistical collection.  In this way,
 * objects that queue do not need additional behavior
 */
public class Queue extends ModelElement implements Iterable<QObject> {
 
	//  ===========================================
	//      CLASS ATTRIBUTES
	//  ===========================================
  
	/** ENQUEUED indicates that something was just enqueued
	 *  DEQUEUED indicates that something was just dequeued
	 * 
	 */
	public static enum Status {ENQUEUED, DEQUEUED};
	
    /** A QueueDiscipline to allow first in, first out behavior.
     */
    public static final FIFODiscipline FIFO = new FIFODiscipline();
    
    /** A QueueDiscipline for last in, first out behavior.
     */
    public static final LIFODiscipline LIFO = new LIFODiscipline();
    
    /** A QueueDiscipline for random selection from the queue. This queue
     *  discipline uses a distribution which can have its stream controlled
     *  This reference provides a shared discipline.  To ensure that queues
     *  using the random discipline use different streams, create additional
     *  instances of RandomDiscipline for each Queue rather than sharing this
     *  instance
     */
    public static final RandomDiscipline RANDOM = new RandomDiscipline();
    
    /** A QueueDiscipline for ranked ordering of the queue.
     */
    public static final RankedQDiscipline RANKED = new RankedQDiscipline();
    
    //  ===========================================
    //      OBJECT ATTRIBUTES
    //  ===========================================
    
    /** The list of items in the queue.
     */
    protected List<QObject> myList;
    
    /** The current QueueDiscipline for this Queue.
     */
    protected QueueDiscipline myDiscipline;
 
    /** The initial QueueDiscipline for this Queue.
     */
    protected QueueDiscipline myInitialDiscipline;
    
    /** Tracks the number in queue.
     */
    protected TimeWeighted myNumInQ;
    
    /** Tracks the time in queue.
     */
    protected ResponseVariable myTimeInQ;
    
    /** Allows a 1-1 mapping between QObjects and the objects entered into the queue
     */
    protected Map<Object,QObject> myQObjectMap;
    
    /** turns on map association between QObjects and the objects entered in the queue
     */
    private boolean myMapTrackingFlag;
    
    /** Holds the listeners for this queue's enqueue and removeNext method use
     */
    protected List<QueueListenerIfc> myQueueListeners;
    
    /** Indicates whether something was just enqueued or dequeued
     */
    protected Status myStatus;
        
    //  ===========================================
    //        CONSTRUCTORS
    //  ===========================================
    
    /** Constructs a Queue.  The default will be a FIFO queue
     */
    public Queue(ModelElement parent){
        this(parent, null, FIFO, false);
    }
    
    /** Constructs a Queue with the given name.  The queue
     * will be a FIFO by default.
     * @param name The name of the queue
     */
    public Queue(ModelElement parent, String name){
        this(parent, name, FIFO, false);
    }
    
    /** Constructs a Queue that follows the
     * given queue discipline.
     * @param discipline The queueing discipline to be followed
     */
    public Queue(ModelElement parent, QueueDiscipline discipline) {
        this(parent, null, discipline, false);
    }
    
    /** Constructs a Queue with the given name that follows the
     * given queue discipline.
     * @param name The name of the queue
     * @param discipline The queueing discipline to be followed
     */
    public Queue(ModelElement parent, String name, QueueDiscipline discipline) {
        this(parent, name, discipline, false);
    }
    
    /** Constructs a Queue with the given name that follows the
     * given queue discipline.
     * @param name The name of the queue
     * @param discipline The queueing discipline to be followed
     * @param mapTrackingFlag, true turns on tracking
     */
    public Queue(ModelElement parent, String name, QueueDiscipline discipline, boolean mapTrackingFlag) {
        super(parent, name);
		myList = new LinkedList<QObject>();       
		setEntityMapTrackingOption(mapTrackingFlag);
               
        changeDiscipline(discipline);
        myInitialDiscipline = myDiscipline;
        
        myNumInQ = new TimeWeighted(this, 0.0, getName() + " NumberInQ");
        myTimeInQ = new ResponseVariable(this, getName() + " TimeInQ");
        
    }
    
    //  ===========================================
    //      PROTECTED METHODS
    //  ===========================================
    
    /** can be called to initialize the queue
     * The default behavior is to have the queue cleared 
     *
     **/
    protected void initialize(){
        super.initialize();
        if (myDiscipline != myInitialDiscipline)
        	changeDiscipline(myInitialDiscipline);
        myDiscipline.initialize();
        clear();
    }
    
    protected void beforeExperiment(){
    	super.beforeExperiment();
    	myDiscipline.beforeExperiment();
    }
    
    protected void afterReplication(){
    	super.afterReplication();
    	myDiscipline.afterReplication();
    }
 
    protected void removedFromModel(){
        super.removedFromModel();
        myList.clear();
        myList = null;
        myDiscipline = null;
        myInitialDiscipline = null;
        myNumInQ = null;
        myTimeInQ = null;
        if (myQObjectMap != null)
            myQObjectMap.clear();
        myQObjectMap = null;
        if (myQueueListeners != null)
            myQueueListeners.clear();
        myQueueListeners = null;
        myStatus = null;
    }

    //  ===========================================
    //      PUBLIC METHODS
    //  ===========================================
    
	/** Adds the supplied listener to this queue
     * @param listener Must not be null, cannot already be added
     */
	public final boolean addQueueListener(QueueListenerIfc listener) {
		if (listener == null)
	           throw new IllegalArgumentException("The listener was null.");
		
		if (myQueueListeners == null)
			myQueueListeners = new ArrayList<QueueListenerIfc>();
		
		if (myQueueListeners.contains(listener))
	           throw new IllegalArgumentException("The queue already has the supplied listener.");
		 			
		return myQueueListeners.add(listener);
	}

	/** Removes the supplied listener from this queue
     * @param listener Must not be null
     */
	public boolean removeQueueListener(QueueListenerIfc listener) {
		if (listener == null)
	           throw new IllegalArgumentException("The listener was null.");
		if (myQueueListeners == null)
			return(false);		
		
		return myQueueListeners.remove(listener);
	}
	
	/** Allows an observer to be attached to the time in queue response variable
	 * 
	 * @param observer
	 */
	public final void addTimeInQueueObserver(Observer observer){
		myTimeInQ.addObserver(observer);
	}

	/** Allows an observer to be removed from the time in queue response variable
	 * 
	 * @param observer
	 */
	public final void removeTimeInQueueObserver(Observer observer){
		myTimeInQ.deleteObserver(observer);
	}

	/** Allows an observer to be attached to the number in queue time weighted variable
	 * 
	 * @param observer
	 */
	public final void addNumberInQueueObserver(Observer observer){
		myNumInQ.addObserver(observer);
	}

	/** Allows an observer to be removed from the number in queue time weighted variable
	 * 
	 * @param observer
	 */
	public final void removeNumberInQueueObserver(Observer observer){
		myNumInQ.deleteObserver(observer);
	}

	/** Causes the supplied AggregateTimeWeightedVariable to
	 *  be subscribed to the number in queue variable
	 * 
	 * @param aggregate
	 */
	public void subscribe(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myNumInQ);
	}

	/** Causes the supplied AggregateTimeWeightedVariable to
	 *  be unsubscribed from the number in queue variable
	 * 
	 * @param aggregate
	 */
	public void unsubscribe(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myNumInQ);
	}

	/** Causes the supplied AveragePerTimeWeightedVariable to
	 *  be subscribed to the number in queue variable
	 * 
	 * @param aggregate
	 */
	public void subscribe(AveragePerTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myNumInQ);		
	}

	/** Causes the supplied AveragePerTimeWeightedVariable to
	 *  be unsubscribed from the number in queue variable
	 * 
	 * @param aggregate
	 */
	public void unsubscribe(AveragePerTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myNumInQ);		
	}

	/** Causes the supplied ResponseVariableAverageObserver to
	 *  be subscribed to the number in queue variable
	 * 
	 * @param aggregate
	 */
	public void subscribe(ResponseVariableAverageObserver aggregate){
		aggregate.subscribeTo(myTimeInQ);		
	}

	/** Causes the supplied ResponseVariableAverageObserver to
	 *  be unsubscribed from the number in queue variable
	 * 
	 * @param aggregate
	 */
	public void unsubscribe(ResponseVariableAverageObserver aggregate){
		aggregate.unsubscribeFrom(myTimeInQ);		
	}

	/** Allows an Aggregate to subscribe to the time in queue variable
	 * 
	 * @param aggregate
	 */
	public final void subscribeToTimeInQueue(Aggregate aggregate){
		aggregate.subscribeTo(myTimeInQ);
	}

	/** Allows an Aggregate to unsubscribe from the time in queue variable
	 * 
	 * @param aggregate
	 */
	public final void unsubscribeFromTimeInQueue(Aggregate aggregate){
		aggregate.unsubscribeFrom(myTimeInQ);
	}

	/** Allows an Aggregate to subscribe to the number in queue variable
	 * 
	 * @param aggregate
	 */
	public final void subscribeToNumberInQueue(Aggregate aggregate){
		aggregate.subscribeTo(myNumInQ);
	}

	/** Allows an Aggregate to unsubscribe from the number in queue variable
	 * 
	 * @param aggregate
	 */
	public final void unsubscribeFromNumberInQueue(Aggregate aggregate){
		aggregate.unsubscribeFrom(myNumInQ);
	}

	/** Gets whether or not the last action was enqueue or dequeueing an object
	 * 
	 * @return
	 */
	public final Status getStatus(){
		return myStatus;
	}
	
	/** Sets the queue's discipline to the given discipline. Throws
     * an IllegalArgumentException if the discipline is null.
     * @param discipline An interface to a queue discipline
     */
    public final void changeDiscipline(QueueDiscipline discipline){
        if (discipline == null)
            throw new IllegalArgumentException("The discipline must be non-null");
        
        discipline.switchFrom(myList, myDiscipline);
        
        myDiscipline = discipline;
    }
    
    /** Changes the priority of the supplied QObject. May cause the queue
     *  to reorder using its discipline
     * 
     * @param qObject
     * @param priority
     */
    public final void changePriority(QObject qObject, int priority){
    	myDiscipline.changePriority(myList, qObject, priority);
    }
    
    /** Gets the initial queue discipline
     * 
     * @return
     */
    public final QueueDiscipline getInitialDiscipline() {
		return myInitialDiscipline;
	}

	/** Sets the initial queue discipline
	 * 
	 * @param discipline
	 */
    public final void setInitialDiscipline(QueueDiscipline discipline) {
		if (discipline == null)
			throw new IllegalArgumentException("The QueueDisciplineIfct must be non-null");
		myInitialDiscipline = discipline;
	}

    /** Creates and enqueue's a QObject with default priority 1
     * 
     * @return The created QObject
     */
    public final QObject enqueue(){
    	return(enqueue(1));
    }
    
    /** Creates and enqueue's a QObject with the given priority
     * 
     * @param priority
     * @return The created QObject
     */
    public final QObject enqueue(int priority){
		QObject qObj = new QObject(null);
		enqueue(qObj, priority, null);
		return(qObj);
    }
    
	/** Places the object in the queue, with the default priority of 1
     * Returns a reference to the QObject that wraps the supplied Object
     *
     *  Automatically, updates the number in queue response variable.
     * @param obj - the object to enqueue
     * @return a reference to the QObject object that wraps the removed item
     */
    public final QObject enqueue(Object obj){
        return(enqueue(obj, 1));
    }
    
    /** Places the object in the queue, with the specified priority
     * Returns a reference to the QObject that wraps the supplied Object
     *
     *  Automatically, updates the number in queue response variable.
     *
     * @param obj - the object to enqueue
     * @param priority - the priority for ordering the object, lower has more priority
     * @return a reference to the QObject object that wraps the enqueued item
     */
    public final QObject enqueue(Object obj, int priority){
    		QObject qObj = new QObject(null);
    		enqueue(qObj, priority, obj);
    		return(qObj);
    }

	/** Places the QObject in the queue, with the default priority of 1
	 *  Automatically, updates the number in queue response variable.
	 *
	 * @param QObject - the QObject to enqueue
	 * @param priority - the priority for ordering the object, lower has more priority
	 */
	public final void enqueue(QObject queueingObject){
		enqueue(queueingObject, 1, queueingObject.getQueuedObject());
	}

	/** Places the QObject in the queue, with the default priority of 1
	 *  Automatically, updates the number in queue response variable.
	 *
	 * @param QObject - the QObject to enqueue
	 * @param priority - the priority for ordering the object, lower has more priority
	 */
	public final void enqueue(QObject queueingObject, int priority){
		enqueue(queueingObject, priority, queueingObject.getQueuedObject());
	}
    
	/** Places the QObject in the queue, with the specified priority
	 *  Automatically, updates the number in queue response variable.
	 *
	 * @param QObject - the QObject to enqueue
	 * @param priority - the priority for ordering the object, lower has more priority
	 * @param Object an Object to be "wrapped" and queued while the QObject is queued
	 */
	public void enqueue(QObject qObject, int priority, Object obj){
		if (qObject == null)
			throw new IllegalArgumentException("The QObject must be non-null");
		
		qObject.enterQueue(this, priority, obj);
		
		myDiscipline.add(myList, qObject);
		if (myMapTrackingFlag == true){
			if (obj == null)
				throw new IllegalArgumentException("Attached object was null and map tracking was on.!");
			myQObjectMap.put(obj, qObject);
		}
		myNumInQ.setValue(myList.size());
        myStatus = Status.ENQUEUED;
        notifyQueueListeners(qObject);
	}

    /** Places all the objects in the specified collection in the queue
     *  using the queue's discipline.  Throws an IllegalArgumentException if the Collection is null
     *
     *  Automatically, updates the number in queue response variable.
     *
     *  @param c the Collection c of items to check
     *
     */
    public final void enqueueAll(Collection<?> c){
        if (c == null)
            throw new IllegalArgumentException("The Collection c must be non-null");
        
        for (Iterator<?> i=c.iterator(); i.hasNext(); ){
            enqueue(i.next());
        }
    }
    
    /** Returns a reference to the QObject representing
     * the item that is next to be removed from the queue according
     * to the queue discipline that was specified.
     *
     * @return a reference to the QObject object that wraps the next item to be removed, or null if the queue is empty
     */
    public final QObject peekNext(){
        return(myDiscipline.peekNext(myList));
    }
    
    /** Removes the next item from the queue according to the queue discipline
     * that was specified.  Returns a reference to the QObject representing
     * the item that was removed
     *
     * Automatically, collects the time in queue for the item and includes it in
     * the time in queue response variable.
     *
     *  Automatically, updates the number in queue response variable.
     *
     * @return a reference to the QObject object that wraps the removed item, or null if the queue is empty
     */
    public final QObject removeNext(){
        QObject qObj = myDiscipline.removeNext(myList);
        if (qObj != null){
        		qObj.exitQueue();
        		double timeInQ = getTime() - qObj.getTimeEnteredQueue();
            myTimeInQ.setValue(timeInQ);
            myNumInQ.setValue(myList.size());
            if (myMapTrackingFlag == true){
                Object obj = qObj.getQueuedObject();
                myQObjectMap.remove(obj);
            }
        }
        myStatus = Status.DEQUEUED;
        notifyQueueListeners(qObj);
        return(qObj);
    }
    
    /** Returns true if this queue contains the specified element.
     * More formally, returns true if and only if this list contains at least
     * one element e such that (o==null ? e==null : o.equals(e)).
     *
     * Throws an IllegalArgumentException if QObject qObj is null.
     *
     * @param qObj The object to be removed
     * @return True if the queue contains the specified element.
     */
    public final boolean contains(QObject qObj){
        if (qObj == null)
            throw new IllegalArgumentException("The QObject qObj must be non-null");
        return(myList.contains(qObj));
    }
    
    /** Returns true if this queue contains all of the elements in the specified collection
     * WARNING: The collection should contain references to QObject's otherwise it will
     * certainly return false.
     *
     * Throws an IllegalArguementException if the Collection is null
     *
     * @param c Collection c of items to check
     * @return True if the queue contains all of the elements.
     */
    public final boolean contains(Collection<QObject> c){
        if (c == null)
            throw new IllegalArgumentException("The Collection c must be non-null");
        return(myList.containsAll(c));
    }
    
    /** Returns the index in this queue of the first occurrence of the specified element,
     * or -1 if the queue does not contain this element.
     * More formally, returns the lowest index i such that
     * (o==null ? get(i)==null : o.equals(get(i))), or -1 if there is no such index.
     *
     * Throws an IllegalArgumentException if QObject qObj is null.
     *
     * @param qObj The object to be found
     * @return The index (zero based) of the element or -1 if not found.
     */
    public final int indexOf(QObject qObj){
        if (qObj == null)
            throw new IllegalArgumentException("The QObject qObj must be non-null");
        return(myList.indexOf(qObj));
    }
    
    
    /** Returns the index in this queue of the last occurrence of the specified element,
     * or -1 if the queue does not contain this element.
     * More formally, returns the lowest index i such that
     * (o==null ? get(i)==null : o.equals(get(i))), or -1 if there is no such index.
     *
     * Throws an IllegalArgumentException if QObject qObj is null.
     *
     * @param qObj The object to be found
     * @return The (zero based) index or -1 if not found.
     */
    public final int lastIndexOf(QObject qObj){
        if (qObj == null)
            throw new IllegalArgumentException("The QObject qObj must be non-null");
        return(myList.lastIndexOf(qObj));
    }
    
    
    /** If the QObject, object map tracking is turned on
     * return the QObject associated with this object
     * Once the QObject is found, the user may use the queue methods
     * to operate on it
     *
     * return Null if object is not in Queue or if map tracking is not on.
     *
     * Throws an IllegalArgumentException if Object obj is null.
     *
     * @param obj The object for which the QObject must be found
     * @return The index (zero based) of the element or -1 if not found.
     */
    public final QObject getQObject(Object obj){
        if (obj == null)
            throw new IllegalArgumentException("The Object obj must be non-null");
        
        QObject qObj = null;
        if (myMapTrackingFlag == true){
            qObj = (QObject)myQObjectMap.get(obj);
        }
        
        return(qObj);
    }
    
    /** Finds all the QObjects in the Queue that satisfy the condition and adds them
     *  to the foundItems collection
     * 
     * @param condition
     * @param foundItems
     * @return yields true if at least one was found, false otherwise
     */
    public final boolean find(Predicate<QObject> condition, Collection<QObject> foundItems){
    	boolean found = false;
    	for(QObject qo: myList){
    		if (condition.test(qo)){
    			found = true;
    			foundItems.add(qo);
    		}
    	}
    	return(found);
    }
    
    /** Finds the first QObject whose getQueuedObject().equals(object)
     * 
     * @param object
     * @return null if no QObject is found
     */
    public final QObject find(Object object){
    	for(QObject qo: myList){
    		if (qo.getQueuedObject().equals(object)){
    			return(qo);
    		}
    	}
    	return(null);
    }

    /** Finds all QObjects whose getQueuedObject().equals(object)
     * 
     * @param foundQObjects 
     * @param object
     * @return returns true if at least one match is found
     */
    public final boolean find(Collection<QObject> foundQObjects, Object object){
    	boolean found = false;
    	for(QObject qo: myList){
    		if (qo.getQueuedObject().equals(object)){
    			foundQObjects.add(qo);
    			found = true;
    		}
    	} 
    	return(found);
    }
    
    /** Finds and removes all the QObjects in the Queue that satisfy the condition and adds them
     *  to the deletedItems collection.  Waiting time statistics are automatically collected
     * 
     * @param condition The condition to check
     * @param deletedItems Holds the items that were removed from the Queue
     * @return yields true if at least one was deleted, false otherwise
     */
    public final boolean remove(Predicate<QObject> condition, Collection<QObject> deletedItems){
    	return(remove(condition, deletedItems, true));
    }
    
    /** Finds and removes all the QObjects in the Queue that satisfy the condition and adds them
     *  to the deletedItems collection
     * 
     * @param condition The condition to check
     * @param deletedItems Holds the items that were removed from the Queue
     * @param waitstats indicates whether or not waiting time statistics should be collected
     * @return yields true if at least one was deleted, false otherwise
     */
    public final boolean remove(Predicate<QObject> condition, Collection<QObject> deletedItems, boolean waitStats){
    	boolean found = false;
    	for(int i=0;i<myList.size();i++){
    		QObject qo = myList.get(i);
    		if (condition.test(qo)){
    			found = true;
    			deletedItems.add(qo);
    			remove(qo, waitStats);
    		}    		
    	}
    	return(found);
    }
 
    /** Removes the first occurrence in the queue of the specified element
     *  Automatically collects waiting time statistics and number in queue
     *  statistics.  If the queue does not contain the element then it
     *  is unchanged and false is returned
     *  
     * Throws an IllegalArgumentException if QObject qObj is null.
     * 
     * @param qObj
     * @return true if the item was removed
     */
    public final boolean remove(QObject qObj){
    		return(remove(qObj, true));
    }
    
    /** Removes the first occurrence in the queue of the specified element
     *  Automatically collects waiting time statistics and number in queue
     *  statistics.  If the queue does not contain the element then it
     *  is unchanged and false is returned
     *
     * Throws an IllegalArgumentException if QObject qObj is null.
     *
     * @param qObj The object to be removed
     * @param waitStats Indicates whether waiting time statistics should
     *        be collected on the removed item, true means collect statistics
     * @return True if the item was removed.
     */
    public final boolean remove(QObject qObj, boolean waitStats){
	    	if (qObj == null)
	    		throw new IllegalArgumentException("The QObject qObj must be non-null");
	    	
	    	if (myList.remove(qObj)){
	    		qObj.exitQueue();
	    		myNumInQ.setValue(myList.size());
	    		if (myMapTrackingFlag == true){
	    			Object obj = qObj.getQueuedObject();
	    			myQObjectMap.remove(obj);
	    		}
	    		if (waitStats){
	        		double timeInQ = getTime() - qObj.getTimeEnteredQueue();
	        		myTimeInQ.setValue(timeInQ);
	    		}
	    		return(true);
	    	}
	    	else
	    		return(false);
    }
 
    /** Removes the element at the specified position in this queue.
     * Shifts any subsequent elements to the left (subtracts one from their indices).
     * Returns the element that was removed from the list.
     *
     * Automatically, collects the time in queue for the item and includes it in
     * the time in queue response variable.
     *
     * Automatically, updates the number in queue response variable.
     *
     * Throws an IndexOutOfBoundsException if the specified index is out of range (index < 0 || index >= size()).
     *
     * @param index - the index of the element to be removed.
     * @return the element previously at the specified position
     */
    public final QObject remove(int index){
    	return(remove(index, true));
    }
    
    /** Removes the element at the specified position in this queue.
     * Shifts any subsequent elements to the left (subtracts one from their indices).
     * Returns the element that was removed from the list.
     *
     * Automatically, collects number in queue statistics. If waitStats flag is
     * true, then automatically collects the time in queue for the item and includes it in
     * the time in queue response variable.
     *
     * Throws an IndexOutOfBoundsException if the specified index is out of range (index < 0 || index >= size()).
     *
     * @param index - the index of the element to be removed.
     * @param waitStatis - true means collect waiting time statistics, false means do not
     * @return the element previously at the specified position
     */
    public final QObject remove(int index, boolean waitStats){
        QObject qObj = (QObject)myList.remove(index);
        if (myMapTrackingFlag == true){
            Object obj = qObj.getQueuedObject();
            myQObjectMap.remove(obj);
        }
        qObj.exitQueue();
		if (waitStats){
    		double timeInQ = getTime() - qObj.getTimeEnteredQueue();
    		myTimeInQ.setValue(timeInQ);
		}
        myNumInQ.setValue(myList.size());
        return(qObj);
    }
    
    /** Removes the QObject at the front of the queue
     *  Uses remove(int index) where index = 0 
     *  
     * @return The first QObject in the queue or null if the list is empty
     */
    public final QObject removeFirst(){
		if (myList.isEmpty())
			return(null);
		else
			return(remove(0));   	
    }
    
    /** Removes the QObject at the last index in the queue.
     *  Uses remove(int index) where index is the size of the list - 1
     *  
     * @return The last QObject in the queue or null if the list is empty
     */
    public final QObject removeLast(){
    		if (myList.isEmpty())
    			return(null);
    		else
    			return(remove(myList.size() -1));
    }

    /** Returns the QObject at the front of the queue
     *  Depending on the queue discipline this may not be the next QObject
     *  
     * @return The first QObject in the queue or null if the list is empty
     */
    public final QObject peekFirst(){
		if (myList.isEmpty())
			return(null);
		else
			return(myList.get(0));   	
    }
    
    /** Returns the QObject at the last index in the queue.
     *  
     * @return The last QObject in the queue or null if the list is empty
     */
    public final QObject peekLast(){
    		if (myList.isEmpty())
    			return(null);
    		else
    			return(myList.get(myList.size() -1));
    }
 
    /** Returns the QObject at the supplied index in the queue.
     *  
     * Throws an IndexOutOfBoundsException if the specified index is out of range (index < 0 || index >= size()).
     * @param index the index to inspect
     * @return The QObject at index in the queue or null if the list is empty
     */
    public final QObject peekAt(int index){
    		if (myList.isEmpty())
    			return(null);
    		else
    			return(myList.get(index));
    }
 
    /** Removes from this queue all the elements that are contained in the specified collection
     * The collection should contain references to objects of type QObject
     * that had been enqueued in this queue; otherwise, nothing will be removed.
     *
     * Automatically, updates the number in queue variable
     * and time in queue statistics on removed items
     *
     * Throws an IllegalArguementException if the Collection is null
     *
     * @param c The collection containing the QObject's to remove
     * @return true if the queue changed as a result of the call
     */
    public final boolean removeAll(Collection<QObject> c){
    	return(removeAll(c, true));
    }
    
    /** Removes from this queue all the elements that are contained in the specified collection
     * The collection should contain references to objects of type QObject
     * that had been enqueued in this queue; otherwise, nothing will be removed.
     *
     * Automatically, updates the number in queue variable
     * If statFlag is true it automatically collects time in queue statistics on removed items
     *
     * Throws an IllegalArguementException if the Collection is null
     *
     * @param c The collection containing the QObject's to remove
     * @param statFlag true means collect statistics, false means do not
     * @return true if the queue changed as a result of the call
     */
    public final boolean removeAll(Collection<QObject> c, boolean statFlag){
        if (c == null)
            throw new IllegalArgumentException("The Collection c must be non-null");
        
        boolean removedFlag = false;
        for(QObject qObj: c){
         	removedFlag = remove(qObj, statFlag);
        }        
        return(removedFlag);
    }
 
    /** Removes from this queue all the elements that are presented by iterating through this iterator
     * The iterator should be based on a collection that contains references to objects of type QObject
     * that had been enqueued in this queue; otherwise, nothing will be removed.
     *
     * Automatically, updates the number in queue variable
     * and time in queue statistics on removed items
     *
     * Throws an IllegalArguementException if the Iterator is null
     *
     * @param c The iterator over the collection containing the QObject's to remove
     * @return true if the queue changed as a result of the call
     */
    public final boolean removeAll(Iterator<QObject> c){
    	return(removeAll(c, true));
    }
    
    /** Removes from this queue all the elements that are presented by iterating through this iterator
     * The iterator should be based on a collection that contains references to objects of type QObject
     * that had been enqueued in this queue; otherwise, nothing will be removed.
     *
     * Automatically, updates the number in queue variable
     * If statFlag is true it automatically collects time in queue statistics on removed items
     *
     * Throws an IllegalArguementException if the Iterator is null
     *
     * @param c The iterator over the collection containing the QObject's to remove
     * @param statFlag true means collect statistics, false means do not
     * @return true if the queue changed as a result of the call
     */
    public final boolean removeAll(Iterator<QObject> c, boolean statFlag){
        if (c == null)
            throw new IllegalArgumentException("The iterator must be non-null");
        
        boolean removedFlag = false;
        while(c.hasNext()){
        	QObject qo = c.next();
        	removedFlag = remove(qo, statFlag);
        }
        
        return(removedFlag);
    }
    
    /** Gets a TimeWeighted response variable for the number in queue.
     * @return The TimeWeighted response variable.
     */
//    public final TimeWeighted getNumberInQueueResponseVariable(){
 //       return(myNumInQ); // return the response variable for now, consider cloning
 //   }
    
    /** Gets the ResponseVariable for the waiting times in the queue.
     * @return The waiting time response variable.
     */
 //   public final ResponseVariable getTimeInQueueResponseVariable(){
 //       return(myTimeInQ); // return the response variable for now, consider cloning
 //   }
    
    /** Turns on the tracing to a text file of the times in queue.
     */
    public final void turnOnTimeInQTrace(){
        myTimeInQ.turnOnTrace();
    }
    
    /** Turns on the tracing to a text file the number in queue for each
     * state change.
     */
    public final void turnOnNumberInQTrace(){
        myNumInQ.turnOnTrace();
    }
    
    /** Turns off the tracing of the times in queue.
     */
    public final void turnOffTimeInQTrace(){
        myTimeInQ.turnOffTrace();
    }
    
    /** Turns off the tracing of the number in queue.
     */
    public final void turnOffNumberInQTrace(){
        myNumInQ.turnOffTrace();
    }
    
    /** Removes all of the elements from this collection
     *
     * WARNING: This method DOES NOT record the time in queue for the cleared items
     * if the user wants this functionality, it can be accomplished using the
     * remove(int index) method, while looping through the items to remove
     * No listeners are notified of the queue change.
     * 
     * This method simply clears the underlying data structure that holds the objects
     */
    public final void clear() {
    	for(QObject qObj: myList){
    		qObj.exitQueue();
     	}
        myList.clear();
        myNumInQ.setValue(myList.size());
        if (myMapTrackingFlag == true)
            myQObjectMap.clear();
    }
    
    /** Returns an iterator (as specified by Collection ) over the elements in the queue in proper sequence.
     * The elements will be ordered according to the state of the queue given
     * the specified queue discipline.
     *
     * WARNING:  The remove() method is not supported by this iterator.  A call to remove() with this iterator
     * will result in an UnsupportedOperationException
     *
     * @return an iterator over the elements in the queue
     */
    public final Iterator<QObject> iterator(){
    		return(new QueueListIterator());
    }
    
    /** Returns an iterator (as specified by Collection ) over the elements in the queue in proper sequence.
     * The elements will be ordered according to the state of the queue given
     * the specified queue discipline.
     *
     * WARNING:  The add(), remove(), and set() methods are not supported by this iterator.  Calls to these
     * methods will result in an UnsupportedOperationException
     *
     * @return an iterator over the elements in the queue
     */
    public final ListIterator<QObject> listIterator(){
        return(new QueueListIterator());
    }
    
    /** Gets the size (number of elements) of the queue.
     * @return The number of items in the queue.
     */
    public final int size(){
        return(myList.size());
    }
    
    /** Returns whether or not the queue is empty.
     * @return True if the queue is empty.
     */
    public final boolean isEmpty(){
        return(myList.isEmpty());
    }
    
    /** Returns true if the queue is not empty
     * 
     * @return
     */
    public final boolean isNotEmpty(){
    	return(!isEmpty());
    }
    
    /** The queue as a string.
     * @return The queue as a string.
     */
    public String toString(){
        return(myList.toString());
    }
 		   	
	/** Gets the entity tracking option, true means on
	 * 
	 * @return
	 */
	public final boolean getEntityTrackingOption(){
		return(myMapTrackingFlag);
	}
	
	/** Sets the option for tracking entities in a map, while
	 *  they are in the queue, true means tracking is on.
	 *  If the user attempts to turn tracking on when the queue is not empty
	 *  then an IllegalArgumentException is thrown.
	 * 
	 */
	public final void setEntityMapTrackingOption(boolean b){
		if (b == false){// wants to turn it off
			myMapTrackingFlag = false;
			if (myQObjectMap != null)
				myQObjectMap.clear();
		} else { // wants to turn it on
			if (!isEmpty()) // if queue is not empty throw an exception
				throw new IllegalArgumentException("Cannot turn on entity map tracking when the queue is not empty.");
			// queue is empty
			myMapTrackingFlag = true;
			if (myQObjectMap == null) // if not made yet, make the map
				myQObjectMap = new LinkedHashMap<Object,QObject>();
		}
	}

	/** Turns off the ability to track entities in the queue
	 *  and clears the map tracking if it had been on
	 */
	public final void turnOffEntityMapTracking() {
		myMapTrackingFlag = false;
		if (myQObjectMap != null)
			myQObjectMap.clear();
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
	
	/** Notifies any listeners that the queue changed
	 * @param qObject The qObject associated with the notification
	 */
	protected void notifyQueueListeners(QObject qObject){
		if (myQueueListeners == null)
			return;
		
		for(QueueListenerIfc ql: myQueueListeners){
			ql.update(this, qObject);
		}
	}
	
	private class QueueListIterator implements ListIterator<QObject> {

		protected ListIterator<QObject> myIterator;
		
		protected QueueListIterator(){
			myIterator = myList.listIterator();
		}
		
		public void add(QObject o) {
			throw new UnsupportedOperationException("The method add() is not supported for Queue iteration");
		}

		public boolean hasNext() {
			return myIterator.hasNext();
		}

		public boolean hasPrevious() {
			return myIterator.hasPrevious();
		}

		public QObject next() {
			return myIterator.next();
		}

		public int nextIndex() {
			return myIterator.nextIndex();
		}

		public QObject previous() {
			return myIterator.previous();
		}

		public int previousIndex() {
			return myIterator.previousIndex();
		}

		public void remove() {
			throw new UnsupportedOperationException("The method remove() is not supported for Queue iteration");
		}

		public void set(QObject o) {
			throw new UnsupportedOperationException("The method set() is not supported for Queue iteration");
		}

	}

	/** Get the number in queue response statistics
	 * 
	 * @return the NumInQ
	 */
	public final ResponseStatisticsIfc getNumInQResponseStatistics() {
		return myNumInQ;
	}

	/** Get the time in queue response statistics
	 * 
	 * @return the TimeInQ
	 */
	public final ResponseStatisticsIfc getTimeInQResponseStatistics() {
		return myTimeInQ;
	}


}
