/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling.elements.queue;

import java.util.List;

import jsl.utilities.random.distributions.DUniform;
import jsl.utilities.random.rng.RandomStreamIfc;

/** The RandomDiscipline provides a mechanism to the Queue object
 * to provide a random selection of objects from the queue
 * rule.
 */
public class RandomDiscipline extends QueueDiscipline implements RandomStreamIfc {

    /** indicates whether or not the random picking
     *  distribution has it stream reset to the default
     *  stream, or not prior to each experiment.  Resetting
     *  allows each experiment to use the same underlying random numbers
     *  i.e. common random numbers, this is the default
     * 
     *  Setting it to true indicates that it does reset
     */
    protected boolean myResetStartStreamOption;

	/** indicates whether or not the random picking
	 *  distribution has it stream reset to the next substream
	 *  stream, or not, prior to each replication.  Resetting
	 *  allows each replication to better ensure that each
	 *  replication will be start at the same place in the 
	 *  substreams, thereby, improving sychronization when using
	 *  common random numbers.  
	 * 
	 *  Setting it to true indicates that it does jump to
	 *  the next substream, true is the default
	 */
	protected boolean myResetNextSubStreamOption;

    private DUniform myDistribution;
    private int myNext;
    
    /** Creates a new instance of RandomDiscipline */
    public RandomDiscipline() {
       setResetStartStreamOption(true);
       setResetNextSubStreamOption(true);
       myDistribution = new DUniform();
    }
    
    /** Adds the specified element to the proper location in the
     * supplied list.
     *
     * @param list The list to which the object is being added
     * @param qObject The element to be added to the supplied list
     */
    protected void add(List<QObject> list, QObject qObject) {
        list.add(qObject);
    }
    
    /** Returns a reference to the next QObjectIfc to be removed
     * from the queue according to a ramdom discipline.  The item is
     * not removed from the list.
     * @param list The list to be peeked into
     * @return The QObjectIfc that is next, or null if the list is empty
     */
    protected QObject peekNext(List<QObject> list) {
        
        if (list.isEmpty())
            return(null);
        
        if (list.size() == 1)
            myNext = 0;
        else {
            myDistribution.setRange(0, list.size()-1);
            myNext = (int)myDistribution.getValue();
        }
        return((QObject)list.get(myNext));// randomly pick it from the range available
    }
    
    /** Removes the next item from the supplied list according to
     * a random discipline.
     * @param list The list for which the next item will be removed
     * @return A reference to the QObjectIfc item that was removed or null if the list is empty
     */
    protected QObject removeNext(List<QObject> list) {
        
        if (list.isEmpty())
            return(null);
        
        peekNext(list); // sets the next randomly
        
        return((QObject)list.remove(myNext)); // now returns the next
    }

    /** Provides a "hook" method to be called when switching from one discpline to another
     *  The implementor should use this method to ensure that the underlying queue is in a state
     *  that allows it to be managed by this queue discipline.
     *
     *  Since random can be applied from any queue state this method performs no operation.
     *
     * @param list The list for which the next item will be removed
     * @param currentDiscipline The queueing discpline that is currently managing the queue
     */
    protected void switchFrom(List<QObject> list, QueueDiscipline currentDiscipline){
    }

	/**
	 * 
	 * @see jsl.utilities.random.rng.RandomStreamIfc#resetNextSubstream()
	 */
	public void resetNextSubstream() {
		myDistribution.resetNextSubstream();
	}

	/**
	 * 
	 * @see jsl.utilities.random.rng.RandomStreamIfc#resetStartStream()
	 */
	public void resetStartStream() {
		myDistribution.resetStartStream();
	}

	/**
	 * 
	 * @see jsl.utilities.random.rng.RandomStreamIfc#resetStartSubstream()
	 */
	public void resetStartSubstream() {
		myDistribution.resetStartSubstream();
	}

	/**
	 * @param flag
	 * @see jsl.utilities.random.rng.RandomStreamIfc#setAntithetic(boolean)
	 */
	public void setAntithetic(boolean flag) {
		myDistribution.setAntithetic(flag);
	}

	/** Gets the current Reset Start Stream Option
	 * @return
	 */
    public final boolean getResetStartStreamOption() {
		return myResetStartStreamOption;
	}

	/** Sets the reset start stream option, true
	 *  means that it will be reset to the starting stream
	 * @param b
	 */
    public final void setResetStartStreamOption(boolean b) {
		myResetStartStreamOption = b;
	}
 
	/** Gets the current reset next substream option
	 *  true means, that it is set to jump to the next substream after 
	 *  each replication
	 * @return
	 */
    public final boolean getResetNextSubStreamOption() {
		return myResetNextSubStreamOption;
	}

	/** Sets the current reset next substream option
	 *  true means, that it is set to jump to the next substream after 
	 *  each replication
	 * @param b
	 */
    public final void setResetNextSubStreamOption(boolean b) {
		myResetNextSubStreamOption = b;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.queue.QueueDisciplineIfc#initialize()
	 */
    protected void initialize() {
	}

    protected void beforeExperiment(){
         if (getResetStartStreamOption())
         	myDistribution.resetStartStream();
    }
    
    /** after each replication reset the underlying random number generator to the next
     *  substream
     */
    protected void afterReplication(){
        if (getResetNextSubStreamOption())
        	myDistribution.resetNextSubstream();
    } 
}
