/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.resource;

import jsl.modeling.elements.EntityReceiverIteratorIfc;
import jsl.modeling.elements.EntityReceivingIfc;
import jsl.modeling.elements.queue.*;
import jsl.modeling.elements.processview.description.*;

import java.util.*;

/**
 *
 */
public class Entity extends QObject implements EntityReceiverIteratorIfc {
	
	/** used to identify the type of entity
	 */
	private EntityType myEntityType;
	
	/** A reference to the entity's current process executor
	 */
	private ProcessExecutor myProcessExecutor;

	/**
	 *  An entity can make many requests for resources
	 */
	private Map<Resource, Request> myRequests;
	
	/** A map to hold the attached named attributes
	 * 
	 */
	private Map<String, AttributeIfc> myAttributes;
	
	/** Used to determine the next receiver for the entity
	 * 
	 */
	private EntityReceiverIteratorIfc myEntityReceiverIterator;

	/**  Can be used to attach information/objects to an entity
	 */
	private Object myMessage;
	
	/** Used to indicate whether the task will supply
	 *  the duration of an activity, delay, etc.
	 * 
	 */
	private boolean myUseDurationFlag = false;
	
	/** The amount of the current duration, this
	 *  can be changed as the task enters Operations
	 */
	private double myDurationTime;

	/** Used to indicate whether the task will supply
	 *  the amount associated with a resource usage
	 * 
	 */
	private boolean myUseAmountFlag = false;
	
	/** The amount of the current use of a resource if set, this
	 *  can be changed as the task enters Operations
	 */
	private int myAmount;

	//  ===========================================
	//        CONSTRUCTORS
	//  ===========================================

	/** Creates a Entity with name = null
	 *  and type = null
	 */
	public Entity() {
		this(null, null);
	}

	/** Creates a Entity with given type
     *  and the creation time set to the current simulation time
     *  The name of the entity is the same as its type
	 */
	public Entity(EntityType entityType) {
		this(entityType.getName(), entityType);
	}
	
	/** Creates an Entity with the given name and EntityType = null
     *  and the creation time set to the current simulation time
 	 * @param name The name of the entity
	 */
	public Entity(String name) {
		this(name, null);
	}

	/** Creates an Entity with the given name 
     *  and the creation time set to the current simulation time
	 * @param name The name of the entity
	 */
	public Entity(String name, EntityType entityType) {
		super(name);
		setType(entityType);		
	}
	
	//  ===========================================
	//        PUBLIC INTERFACE
	//  ===========================================
	
	/**
	 * @return Returns the type.
	 */
	public final EntityType getType() {
		return myEntityType;
	}
	
	/** Gets a reference to the process executor that is 
	 *  currently executing with the entity.
	 * @return A reference to the process executor.
	 */
	public final ProcessExecutor getProcessExecutor() {
		return (myProcessExecutor);
	}
	
 	/* mutators --------------------------*/

	/** Changes the type of the entity.
	 *  Side effects: 
	 *  1. ALL previously supplied attributes will be cleared, even user defined attributes.
	 *	   The attributes of the entity will be assigned based on the newly supplied entity type.  
	 *  2. The entity type may be set to null, clearing all attributes and setting the
	 *     entity's EntityReceiverIteratorIfc to null
	 * 
	 * @param type The type to set.
	 */
	public final void setType(EntityType entityType) {
	    
	    if (entityType == null){// making the entity type null
			if (myAttributes != null) // if they exist, we don't need attributes anymore
				myAttributes.clear();
			myEntityReceiverIterator = null; // if there is no type, there is no receiver and thus no iterator
		} else { // entity type is not null
			if (myEntityType != entityType){ // changing the entity type
				if (myAttributes != null) // if they exist, we don't need attributes anymore
					myAttributes.clear();
					
				// note both of these code sequences assume
				// that the returned iterator is on an already built list
				// i.e. if someone changes the list after the entity is created
				// then these iterators reflect the original list
				for (Iterator<AttributeType> i = entityType.getAttributeTypesIterator(); i.hasNext();){
					AttributeType at = i.next();
					Attribute a = new Attribute(at.getName());
					addAttribute(a.getName(), a);
				}
				myEntityReceiverIterator = entityType.getEntityReceiverIterator();			
			}
		}
		myEntityType = entityType;
	}
	
	/** Adds the named attribute to the entity as a valid attribute
	 *  Arguments must be non-null, or exceptions will be thrown.
	 *  The named attribute must not already have been added to the entity
	 *  or an exception will be thrown.
	 *  
	 * @param attributeName
	 * @param attribute
	 */
	public final void addAttribute(String attributeName, AttributeIfc attribute){
		if (attributeName == null)
			throw new IllegalArgumentException("Attribute name was null!");    

		if (attribute == null)
			throw new IllegalArgumentException("Attribute was null!");    
		
		if (myAttributes == null)
			myAttributes = new LinkedHashMap<String, AttributeIfc>();
		
		if(myAttributes.containsKey(attributeName))
			throw new IllegalArgumentException("The Entity already has an attribute named: " + attributeName);

		myAttributes.put(attributeName, attribute);
	}
	
	/** Returns whether or not the attribute has been named for
	 *  this entity
	 * 
	 * @param attributeName
	 * @return true if already added, false otherwise
	 */
	public final boolean containsAttribute(String attributeName){
		if (myAttributes == null)
			return(false);
		
		if (attributeName == null)
			return(false);
		else
			return(myAttributes.containsKey(attributeName));
	}
	
	/** Allows the user to set the value of the named attribute
	 *  to the given value
	 * 
	 * @param attributeName, the attribute name must be present for the entity or an exception is thrown
	 * @param value
	 */
	public final void setAttributeValue(String attributeName, double value){
		if (myAttributes == null)
			throw new IllegalArgumentException("The Entity does not have an attribute named: " + attributeName);

		if(!myAttributes.containsKey(attributeName))
			throw new IllegalArgumentException("The Entity does not have an attribute named: " + attributeName);

		AttributeIfc attribute = (AttributeIfc)myAttributes.get(attributeName);
		attribute.setValue(value);
	}

	/** Get's the value of the attribute using the GetValueIfc
	 * 
	 * @param attributeName
	 * @return
	 */
	public final double getAttributeValue(String attributeName){
		if (myAttributes == null)
			throw new IllegalArgumentException("The Entity does not have an attribute named: " + attributeName);
		
		if(!myAttributes.containsKey(attributeName))
			throw new IllegalArgumentException("The Entity does not have an attribute named: " + attributeName);

		AttributeIfc attribute = (AttributeIfc)myAttributes.get(attributeName);
		return(attribute.getValue());
	}
	
	/** This is used by the ProcessExecutor to tell the entity
	 *  what process executor it is currently executing within.
	 * 
	 * @param processExecutor The ProcessExecutor
	 */
	public final void setProcessExecutor(ProcessExecutor processExecutor){
		if (processExecutor == null)
			throw new IllegalArgumentException("ProcessExecutor was null!");    
		myProcessExecutor = processExecutor;
	}
		
	/**Sometimes the user will need to get the request that an entity 
	 * made on a particular resource.  This method returns the 
	 * request for the supplied resource.
	 * 
	 * @param resource, The resource to find the request for
	 * @return, the Request associated with the supplied request or null
	 */
	public final Request getRequest(Resource resource){
		if (myRequests == null)
			return (null);
		else
			return((Request)myRequests.get(resource));
	}
	
	/** Gets the next receiver for this entity
	 *  Returns null if no next receiver
	 *  
	 * @return, the receiver or null if none
	 */
	public final EntityReceivingIfc nextEntityReceiver(){
		if (myEntityReceiverIterator == null)
			return(null);
		else
			return(myEntityReceiverIterator.nextEntityReceiver());
	}
	
	/** Gets the previous receiver for this entity
	 *  Returns null if no previous receiver
	 *  
	 * @return, the receiver or null if none
	 */
	public final EntityReceivingIfc previousEntityReceiver() {
		if (myEntityReceiverIterator == null)
			return(null);
		else
			return myEntityReceiverIterator.previousEntityReceiver();
	}

	/** Checks if there is a next entity receiver for this entity
	 *  Returns false if no next receiver
	 *  
	 * @return, true if there is one, false otherwise
	 */
	public final boolean hasNextEntityReceiver() {		
		if (myEntityReceiverIterator == null)
			return(false);
		else
			return myEntityReceiverIterator.hasNextEntityReceiver();
	}

	/** Checks if there is a previous entity receiver for this entity
	 *  Returns false if no previous receiver
	 *  
	 * @return, true if there is one, false otherwise
	 */
	public final boolean hasPreviousEntityReceiver() {
		if (myEntityReceiverIterator == null)
			return(false);
		else
			return myEntityReceiverIterator.hasPreviousEntityReceiver();
	}

	// ===========================================
	// PROTECTED METHODS
	// ===========================================	

	/** Whenever a request for a particular resource is made by an entity,
	 *  the entity keeps track of the request it has made for each resource.  
	 *  This method removes the association between the entity, the request and the resource. 
	 *  This method is protected and used within Request.
	 * 
	 * @param resource, The Resource to be associated with the entity's request
	 */
	protected final Request removeRequest(Resource resource){
		if (myRequests == null)
			return (null);
		else
			return((Request)myRequests.remove(resource));
	}

	/** Whenever a request for a particular resource is made by an entity,
	 *  the entity keeps track of the request for each resource.  
	 *  This method records the association between the entity, the request and the resource. 
	 *  This method is protected and used within Request.
	 *  
	 *  If a request has already been made by this entity for the resource then an exception will
	 *  be thrown
	 * 
	 * @param resource, The Resource to be associated with the entity's request
	 * @param request, The Request to be associated with the entity's resource
	 */
	protected final void addRequest(Resource resource, Request request){
		if (myRequests == null)
			myRequests = new HashMap<Resource, Request>();
		if (resource == null)
			throw new IllegalArgumentException("The resource was null!");    
		if (request == null)
			throw new IllegalArgumentException("The request was null!");    
		if (myRequests.containsKey(resource))
			throw new IllegalArgumentException("The entity already has a request associated with the supplied resource.");
		myRequests.put(resource, request);
	}	
	
	/**
	 * @return Returns the Message.
	 */
	public final Object getMessage() {
		return myMessage;
	}
	
	/**
	 * @param myMessage The Message to set.
	 */
	public final void setMessage(Object message) {
		myMessage = message;
	}
	
	/** Returns the duration time.  This can be used by clients
	 *  to set a duration time and be used within activities, processes,
	 *  time events, etc.
	 * 
	 * @return
	 */
	public final double getDurationTime() {
		return myDurationTime;
	}

	/** Sets the duration time.  This can be used by clients
	 *  to set a duration time and be used within activities, processes,
	 *  time events, etc.
	 * 
	 * @param durationTime
	 */
	public final void setDurationTime(double durationTime) {
		myDurationTime = durationTime;
	}

	/** This can be used by clients to indicate whether they
	 *  should use getDurationTime().  By default it is false
	 *  (do not use the duration time).
	 * 
	 * @return
	 */
	public final boolean getUseDurationFlag() {
		return myUseDurationFlag;
	}

	/** This can be used by clients to indicate whether they
	 *  should used getDurationTime().  By default it is false
	 *  (do not use the duration time).
	 */
	public final void setUseDurationFlag(boolean flag) {
		myUseDurationFlag = flag;
	}

	/** This can be used by clients to set an amount, generally
	 *  used when seizing resources.
	 * 
	 * @return
	 */
	public final int getAmount() {
		return myAmount;
	}

	/** This can be used by clients to set an amount, generally
	 *  used when seizing resources.
	 * 
	 * @param amount must be > 0
	 */
	public final void setAmount(int amount) {
		if (amount <= 0)
			throw new IllegalArgumentException("The amount was <= zero.");			
		myAmount = amount;
	}
	
	/** This can be used by clients to indicate whether they
	 *  should used getAmount().  By default it is false
	 *  (do not use the getAmount()).
	 * 
	 * @return
	 */
	public final boolean getUseAmountFlag() {
		return myUseAmountFlag;
	}

	/** This can be used by clients to indicate whether they
	 *  should used getAmount().  By default it is false
	 *  (do not use the getAmount()).
	 *  
	 * @param flag
	 */
	public final void setUseAmountFlag(boolean flag) {
		myUseAmountFlag = flag;
	}

}
