/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.resource;

import java.util.LinkedList;
import java.util.List;

import jsl.modeling.elements.queue.QObject;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.IllegalStateException;
import jsl.modeling.*;

/**
 * 
 */
public class Request extends QObject {

    //  ===========================================
    //      CLASS AND OBJECT ATTRIBUTES
    //  ===========================================
	
	/** The default size of the Request cache
	 */
	public static final int DEFAULT_PRIORITY = JSLEvent.DEFAULT_PRIORITY;

	/** The default size of the Request cache
	 */
	public final static int DEFAULT_CACHE_LIMIT = 50;

	/** This class caches used Requests for
	 *  reuse, perhaps gaining some efficiency
	 */
	private static int myRequestCacheSizeLimit = DEFAULT_CACHE_LIMIT;

	/** A reference to the list for cached Requests
	 */
	private static List<Request> myRequestCache = new LinkedList<Request>();
	
	// ===========================================
	// CLASS AND OBJECT ATTRIBUTES
	// ===========================================

	/** An event associated with the request.  Can be used to cancel
	 *  an event scheduled for the service completion of the request
	 */
	private JSLEvent myEvent;
	
	/** A reference to the object that made the request
	 *  Can be used to call back on the requester
	 */
	private Object myRequestingObject;
	
	/**
	 * The entity making the request.
	 */
	private Entity myEntity;

	/**
	 * The current resource that the request is allocated to, null if not in a
	 * resource
	 */
	private Resource myResource;

	/** Called whenever units of a resource are allocated to a request
	 * 
	 */
	private ResourceAllocationListenerIfc myResourceAllocationListener;

	/** Called whenever units of a resource are released by this request
	 * 
	 */
	private ResourceReleaseListenerIfc myResourceReleaseListener;

	/** Indicates whether or not the request allows the
	 *  amount requested to be partially allocated
	 */
	private boolean myAllowPartialFillingFlag;
	
	/**
	 * The amount of resource requested
	 */
	private int myInitialAmountRequested;

	/**
	 * The amount of resource allocated to the request
	 */
	private int myAmountAllocated;
	
	/** The amount of the resource remaining to be released
	 * 
	 */
	private int myAmtRemainingToRelease;
	
	/** Indicates if the amount allocated has reached
	 *  the amount requested.  Then the request is completely
	 *  allocated and this flag is true.
	 */
	private boolean mySatisfiedFlag;
	
	/** Indicates that this request has completely
	 *  release any allocate units of a resource that it has
	 */
	private boolean myCompletedFlag;

	/** The amount of time requested
	 */
	private double myTimeRequested;

	/**
	 * A state representing when no units of a resource
	 * have been allocated to the request
	 */
	private State myUnallocatedState;

	/**
	 * A state representing when units of a resource
	 * are allocated to the request
	 */
	private State myPartiallyAllocatedState;

	/**
	 * A state representing when units of a resource
	 * are allocated to the request
	 */
	private State myFullyAllocatedState;

	/**
	 * A state representing when units of a resource
	 * have been partially released by a request
	 */
	private State myPartiallyReleasedState;

	/**
	 * A state representing when all allocated units of a resource
	 * have been released by the request
	 */
	private State myFullyReleasedState;

	// ===========================================
	// CONSTRUCTORS
	// ===========================================

	/** This constructor is used to make a request for 1 unit of a Resource
	 *   This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 */
	protected Request(){
		this(1, Request.DEFAULT_PRIORITY, null, false, null);
	}
	
	/** This constructor is used to make a request for units of a Resource
	 *   This constructor simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 */
	protected Request(int amtRequested){
		this(amtRequested, Request.DEFAULT_PRIORITY, null, false, null);
	}
	
	/** This constructor is used to make a request for units of a Resource
	 *   This constructor simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param partialFillFlag indicates whether the request allows partial filling
	 */
	protected Request(int amtRequested, boolean partialFillFlag){
		this(amtRequested, Request.DEFAULT_PRIORITY, null, partialFillFlag, null);
	}
	
	/** This constructor is used to make a request 1 unit of a Resource
	 *  by a particular entity.  This constructor simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param entity, the entity making the request
	 */
	protected Request(Entity entity){
		this(1, Request.DEFAULT_PRIORITY, entity, false, null);
	}
	
	/** This constructor is used to make a request for units of a Resource
	 *   This constructor simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param priority, a priority for determining the order of the queue
	 */
	protected Request(int amtRequested, int priority){
		this(amtRequested, priority, null, false, null);
	}
	
	/** This constructor is used to make a request for units of a Resource
	 *   This constructor simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param priority, a priority for determining the order of the queue
	 * @param partialFillFlag indicates whether the request allows partial filling
	 */
	protected Request(int amtRequested, int priority, boolean partialFillFlag){
		this(amtRequested, priority, null, partialFillFlag, null);
	}
	
	/** This constructor is used to make a request for units of a Resource
	 *  by a particular entity.  This constructor simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param entity, the entity making the request
	 */
	protected Request(int amtRequested, Entity entity){
		this(amtRequested, Request.DEFAULT_PRIORITY, entity, false, null);
	}

	/** This constructor is used to make a request for units of a Resource
	 *  by a particular entity.  This constructor simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param entity, the entity making the request
	 * @param priority, a priority for determining the order of the queue
	 */
	public Request(int amtRequested, int priority, Entity entity){
		this(amtRequested, priority, entity, false, null);
	}

	/** This constructor is used to make a request for units of a Resource
	 *  by a particular entity.  This constructor simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param entity, the entity making the request
	 * @param partialFillFlag indicates whether the request allows partial filling
	 * @param priority, a priority for determining the order of the queue
	 */
	public Request(int amtRequested, int priority, Entity entity, boolean partialFillFlag){
		this(amtRequested, priority, entity, partialFillFlag, null);
	}
	
	/** This constructor is used to make a request for units of a Resource
	 *  by a particular entity.  This constructor simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param entity, the entity making the request
	 * @param priority, a priority for determining the order of the queue
	 * @param partialFillFlag indicates whether the request allows partial filling
	 * @param obj, a object that the entity can bring with it as it waits for and/or uses the resource
	 */
	protected Request(int amtRequested, int priority, Entity entity, boolean partialFillFlag, Object obj) {
		super();
		initialize(amtRequested, priority, entity, partialFillFlag, obj);
	}

	/** Initializes the request as "new". Allows reuse
	 *  Default: 
	 *  	Object = null
	 *  	partial fill flag = false
	 *  	entity = null
	 *  	priority = Request.DEFAULT_PRIORITY
	 *  	amtRequested = 1
	 */
	protected final void initialize(){
		initialize(1, Request.DEFAULT_PRIORITY, null, false, null);
	}
	
	/** Initializes the request as "new". Allows reuse
	 *  Default: 
	 *  	Object = null
	 *  	partial fill flag = false
	 *  	entity = null
	 *  	priority = Request.DEFAULT_PRIORITY
	 *  
	 * @param amtRequested
	 */
	protected final void initialize(int amtRequested){
		initialize(amtRequested, Request.DEFAULT_PRIORITY, null, false, null);
	}
	
	/** Initializes the request as "new". Allows reuse
	 *  Default: 
	 *  	Object = null
	 *  	partial fill flag = false
	 *  	entity = null
	 *  
	 * @param amtRequested
	 * @param priority
	 */
	protected final void initialize(int amtRequested, int priority){
		initialize(amtRequested, priority, null, false, null);
	}
	
	/** Initializes the request as "new". Allows reuse
	 *  Default: 
	 *  	Object = null
	 *  	partial fill flag = false
	 *  
	 * @param amtRequested
	 * @param priority
	 * @param entity
	 */
	protected final void initialize(int amtRequested, int priority, Entity entity){
		initialize(amtRequested, priority, entity, false, null);
	}
	
	/** Initializes the request as "new". Allows reuse
	 *  Default: Object = null
	 *  
	 * @param amtRequested
	 * @param priority
	 * @param entity
	 * @param partialFillFlag
	 */
	protected final void initialize(int amtRequested, int priority, Entity entity, boolean partialFillFlag){
		initialize(amtRequested, priority, entity, partialFillFlag, null);
	}
	
	/** Initializes the request as "new". Allows reuse
	 * 
	 * @param amtRequested
	 * @param priority
	 * @param entity
	 * @param partialFillFlag
	 * @param obj
	 */
	protected final void initialize(int amtRequested, int priority, Entity entity, boolean partialFillFlag, Object obj) {

		setInitialAmountRequested(amtRequested);
		myAllowPartialFillingFlag = partialFillFlag;
		mySatisfiedFlag = false;
		myCompletedFlag = false;
		myTimeRequested = Double.NaN;
		setPriority_(priority);
		setEntity(entity);
		setQueuedObject(obj);

		if (myUnallocatedState == null){
			myUnallocatedState = new State(getName() + " Unallocated");
			myPartiallyAllocatedState = new State(getName() + " PartiallyAllocated");
			myFullyAllocatedState = new State(getName() + " FullyAllocated");
			myPartiallyReleasedState = new State(getName() + " PartiallyReleased");
			myFullyReleasedState = new State(getName() + " FullyReleased");
		}
		else {
			myUnallocatedState.initialize();
			myPartiallyAllocatedState.initialize();
			myFullyAllocatedState.initialize();
			myPartiallyReleasedState.initialize();
			myFullyReleasedState.initialize();
		}
		
		myUnallocatedState.enter();
	}
 
	// ===========================================
	// PUBLIC INTERFACE
	// ===========================================

	/** This method is used to make a request for 1 unit of a Resource
	 *   This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @return A reference to the Request that was made
	 */
	public static final Request create(){
		return(create(1, Request.DEFAULT_PRIORITY, null, false, null));
	}
	
	/** This method is used to make a request for units of a Resource
	 *   This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @return A reference to the Request that was made
	 */
	public static final Request create(int amtRequested){
		return(create(amtRequested, Request.DEFAULT_PRIORITY, null, false, null));
	}

	/** This method is used to make a request for units of a Resource
	 *   This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param partialFillFlag indicates whether the request allows partial filling
	 * @return A reference to the Request that was made
	 */
	public static final Request create(int amtRequested, boolean partialFillFlag){
		return(create(amtRequested, Request.DEFAULT_PRIORITY, null, partialFillFlag, null));
	}
	
	/** This method is used to make a request 1 unit of a Resource
	 *  by a particular entity.  This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param entity, the entity making the request
	 * @return A reference to the Request that was made
	 */
	public static final Request create(Entity entity){
		return(create(1, Request.DEFAULT_PRIORITY, entity, false, null));
	}

	/** This method is used to make a request for units of a Resource
	 *   This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param priority, a priority for determining the order of the queue
	 * @return A reference to the Request that was made
	 */
	public static final Request create(int amtRequested, int priority){
		return(create(amtRequested, priority, null, false, null));
	}

	/** This method is used to make a request for units of a Resource
	 *   This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param priority, a priority for determining the order of the queue
	 * @param partialFillFlag indicates whether the request allows partial filling
	 * @return A reference to the Request that was made
	 */
	public static final Request create(int amtRequested, int priority, boolean partialFillFlag){
		return(create(amtRequested, priority, null, partialFillFlag, null));
	}

	/** This method is used to make a request for units of a Resource
	 *  by a particular entity.  This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param entity, the entity making the request
	 * @return A reference to the Request that was made
	 */
	public static final Request create(int amtRequested, Entity entity){
		return(create(amtRequested, Request.DEFAULT_PRIORITY, entity, false, null));
	}

	/** This method is used to make a request for units of a Resource
	 *  by a particular entity.  This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param entity, the entity making the request
	 * @param priority, a priority for determining the order of the queue
	 * @return A reference to the Request that was made
	 */
	public static final Request create(int amtRequested, int priority, Entity entity){
		return(create(amtRequested, priority, entity, false, null));
	}

	/** This method is used to make a request for units of a Resource
	 *  by a particular entity.  This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param entity, the entity making the request
	 * @param partialFillFlag indicates whether the request allows partial filling
	 * @param priority, a priority for determining the order of the queue
	 * @return A reference to the Request that was made
	 */
	public static final Request create(int amtRequested, int priority, Entity entity, boolean partialFillFlag){
		return(create(amtRequested, priority, entity, partialFillFlag, null));
	}
	
	/** This method is used to make a request for units of a Resource
	 *  by a particular entity.  This method simply creates a Request that 
	 *  can then be used to ask the resource to allocate units of the resource.
	 * 
	 * @param amountRequested, the amount requested, must be > 0
	 * @param entity, the entity making the request
	 * @param priority, a priority for determining the order of the queue
	 * @param partialFillFlag indicates whether the request allows partial filling
	 * @param obj, a object that the entity can bring with it as it waits for and/or uses the resource
	 * @return A reference to the Request that was made
	 */
	public static final Request create(int amtRequested, int priority, Entity entity, boolean partialFillFlag, Object obj){
		Request request;

		if (myRequestCache.isEmpty())
			request = new Request(amtRequested, priority, entity, partialFillFlag, obj);
		else {// cache not empty get request from cache
			request = (Request) myRequestCache.remove(0);
			// prepare the request
			request.initialize(amtRequested, priority, entity, partialFillFlag, obj);
		}

		return (request);		
	}
	
	/**
	 * 
	 * @param request
	 */
	public final static void dispose(Request request) {
		if (!request.isCompleted() || !request.isUnallocated())
			throw new IllegalStateException("Cannot dispose a request that is not completed or is allocated");
		
		if (myRequestCache.size() < myRequestCacheSizeLimit) { // can add element to cache
			request.setNulls();
			myRequestCache.add(request);
		}	
	}

	public String toString(){
		return(super.toString()+ " AmtNeeded= " + getAmountNeeded());
	}
	
	/** Returns the amount needed by the request
	 * 
	 */
	public final int getAmountNeeded(){
		return(myInitialAmountRequested - myAmountAllocated);
	}
	
	/** Gets the original amount requested.
	 * 
	 * @return The original amount requested
	 */
	public final int getInitialAmountRequested() {
		return (myInitialAmountRequested);
	}

	/** Gets the amount allocated by the resource to the request
	 * 
	 * @return The amount of the resource given to the request
	 */
	public final int getAmountAllocated() {
		return (myAmountAllocated);
	}

	/** Gets the amount remaining to release
	 * 
	 * @return The amount of the resource remaining to release
	 */
	public final int getAmountRemainingToRelease() {
		return (myAmtRemainingToRelease);
	}
	
	/** Indicates that the request allows the amount
	 *  requested to be partially filled
	 *  
	 * @return Returns the myAllowPartialFillingFlag.
	 */
	public final boolean getAllowPartialFillingFlag() {
		return myAllowPartialFillingFlag;
	}
	
	/** Gets the time the Request was last enqueued
	 * 
	 * @return A double representing the time the Request was enqueued
	 */
	public final double getTimeEnteredResource() {
		return (myPartiallyAllocatedState.getTimeStateEntered());
	}

	/** Gets the time the Request was last exited a queue
	 * @return A double representing the time the Request last exited a queue
	 */
	public final double getTimeExitedResource() {
		return (myPartiallyAllocatedState.getTimeStateExited());
	}

	/** Gets the resource that the request will or is currently using
	 * 
	 * @return The resource or null if not currently using a resource
	 */
	public final Resource getResource() {
		return (myResource);
	}

	/** A convenience method.  If the request has a resource, the resource
	 *  will be told to release().  Rather than first calling getResource()
	 *  and then telling the resource to release, the client can just call
	 *  this method.
	 *
	 */
	public final void releaseResource(){
		if (myResource != null)
			myResource.release(this);
	}

	/** Causes the request to be reinitialized with the given parameters
	 *  This can only be done if the request is completed (isCompleted()) or
	 *  if the request is unallocated (isUnallocated()); otherwise, an
	 *  IllegalStateException will be thrown. The amount requested will be 1.
	 *  
	 *  NOTES:
	 *   - The request is placed in the Unallocated state and becomes unsatisfied
	 *   - The time requested is set to Double.NaN
	 *  - Does not automatically reset the accumulated time spent in each state
	 *  {Unallocated, Partially Allocated, Fully Allocated, Partially Released, Completed}
	 *  - Does not change the attached queued object if attached
	 *  - Does not change the partial filling option
	 *  - The entity making the request is set to NULL
	 *  - The priority of the request is not changed
	 *  
	 * @param amtRequested The amount needed for the request
	 */
	public final void reinitialize() {
		reinitialize(1, getPriority(), null, getAllowPartialFillingFlag(), getQueuedObject(), false);
	}
	
	/** Causes the request to be reinitialized with the given parameters
	 *  This can only be done if the request is completed (isCompleted()) or
	 *  if the request is unallocated (isUnallocated()); otherwise, an
	 *  IllegalStateException will be thrown.
	 *  
	 *  NOTES:
	 *   - The request is placed in the Unallocated state and becomes unsatisfied
	 *   - The time requested is set to Double.NaN
	 *  - Does not automatically reset the accumulated time spent in each state
	 *  {Unallocated, Partially Allocated, Fully Allocated, Partially Released, Completed}
	 *  - Does not change the partial filling option
	 *  - The entity making the request is set to NULL
	 *  - The priority of the request is not changed
	 *  
	 * @param amtRequested The amount needed for the request
	 * @param obj An object attached to the request
	 */
	public final void reinitialize(int amtRequested, Object obj) {
		reinitialize(amtRequested, getPriority(), null, getAllowPartialFillingFlag(), obj, false);
	}
	
	/** Causes the request to be reinitialized with the given parameters
	 *  This can only be done if the request is completed (isCompleted()) or
	 *  if the request is unallocated (isUnallocated()); otherwise, an
	 *  IllegalStateException will be thrown.
	 *  
	 *  NOTES:
	 *   - The request is placed in the Unallocated state and becomes unsatisfied
	 *   - The time requested is set to Double.NaN
	 *  - Does not automatically reset the accumulated time spent in each state
	 *  {Unallocated, Partially Allocated, Fully Allocated, Partially Released, Completed}
	 *  - Does not change the attached queued object if attached
	 *  - Does not change the partial filling option
	 *  - The entity making the request is set to NULL
	 *  - The priority of the request is not changed
	 *  
	 * @param amtRequested The amount needed for the request
	 */
	public final void reinitialize(int amtRequested) {
		reinitialize(amtRequested, getPriority(), null, getAllowPartialFillingFlag(), getQueuedObject(), false);
	}
	
	/** Causes the request to be reinitialized with the given parameters
	 *  This can only be done if the request is completed (isCompleted()) or
	 *  if the request is unallocated (isUnallocated()); otherwise, an
	 *  IllegalStateException will be thrown.
	 *  
	 *  NOTES:
	 *   - The request is placed in the Unallocated state and becomes unsatisfied
	 *   - The time requested is set to Double.NaN
	 *  - Does not automatically reset the accumulated time spent in each state
	 *  {Unallocated, Partially Allocated, Fully Allocated, Partially Released, Completed}
	 *  - Does not change the attached queued object if attached
	 *  - Does not change the partial filling option
	 *  - The entity making the request is set to NULL
	 *  
	 * @param amtRequested The amount needed for the request
	 * @param priority The priority of the request
	 */
	public final void reinitialize(int amtRequested, int priority) {
		reinitialize(amtRequested, priority, null, getAllowPartialFillingFlag(), getQueuedObject(), false);
	}
	
	/** Causes the request to be reinitialized with the given parameters
	 *  This can only be done if the request is completed (isCompleted()) or
	 *  if the request is unallocated (isUnallocated()); otherwise, an
	 *  IllegalStateException will be thrown.
	 *  
	 *  NOTES:
	 *   - The request is placed in the Unallocated state and becomes unsatisfied
	 *   - The time requested is set to Double.NaN
	 *  - Does not automatically reset the accumulated time spent in each state
	 *  {Unallocated, Partially Allocated, Fully Allocated, Partially Released, Completed}
	 *  - Does not change the attached queued object if attached
	 *  - Does not change the partial filling option
	 *  
	 * @param amtRequested The amount needed for the request
	 * @param priority The priority of the request
	 * @param entity The entity associated with the request
	 */
	public final void reinitialize(int amtRequested, int priority, Entity entity) {
		reinitialize(amtRequested, priority, entity, getAllowPartialFillingFlag(), getQueuedObject(), false);
	}
	
	/** Causes the request to be reinitialized with the given parameters
	 *  This can only be done if the request is completed (isCompleted()) or
	 *  if the request is unallocated (isUnallocated()); otherwise, an
	 *  IllegalStateException will be thrown.
	 *  
	 *  NOTES:
	 *   - The request is placed in the Unallocated state and becomes unsatisfied
	 *   - The time requested is set to Double.NaN
	 *  - Does not automatically reset the accumulated time spent in each state
	 *  {Unallocated, Partially Allocated, Fully Allocated, Partially Released, Completed}
	 *  - Does not change the attached queued object if attached
	 *  
	 * @param amtRequested The amount needed for the request
	 * @param priority The priority of the request
	 * @param entity The entity associated with the request
	 * @param partialFillFlag Whether or not the request can be filled incrementally
	 */
	public final void reinitialize(int amtRequested, int priority, Entity entity, 
			boolean partialFillFlag) {
		reinitialize(amtRequested, priority, entity, partialFillFlag, getQueuedObject(), false);
	}
	
	/** Causes the request to be reinitialized with the given parameters
	 *  This can only be done if the request is completed (isCompleted()) or
	 *  if the request is unallocated (isUnallocated()); otherwise, an
	 *  IllegalStateException will be thrown.
	 *  
	 *  NOTES:
	 *   - The request is placed in the Unallocated state and becomes unsatisfied
	 *   - The time requested is set to Double.NaN
	 *   - Does not automatically reset the accumulated time spent in each state
	 *  {Unallocated, Partially Allocated, Fully Allocated, Partially Released, Completed}
	 *  
	 * @param amtRequested The amount needed for the request
	 * @param priority The priority of the request
	 * @param entity The entity associated with the request
	 * @param partialFillFlag Whether or not the request can be filled incrementally
	 * @param obj An object attached to the request
	 */
	public final void reinitialize(int amtRequested, int priority, Entity entity, 
			boolean partialFillFlag, Object obj) {
		reinitialize(amtRequested, priority, entity, partialFillFlag, obj, false);
	}
	
	/** Causes the request to be reinitialized with the given parameters
	 *  This can only be done if the request is completed (isCompleted()) or
	 *  if the request is unallocated (isUnallocated()); otherwise, an
	 *  IllegalStateException will be thrown.
	 * 
	 *  NOTES:
	 *   - The request is placed in the Unallocated state and becomes unsatisfied
	 *   - The time requested is set to Double.NaN
	 *   
	 * @param amtRequested The amount needed for the request
	 * @param priority The priority of the request
	 * @param entity The entity associated with the request
	 * @param partialFillFlag Whether or not the request can be filled incrementally
	 * @param obj An object attached to the request
	 * @param resetStateAccumulation Whether or not the request's accumulated time spent
	 *  in each state {Unallocated, Partially Allocated, Fully Allocated, Partially Released, Completed}
	 *  will be cleared (initialize)
	 */
	public final void reinitialize(int amtRequested, int priority, Entity entity, 
			boolean partialFillFlag, Object obj, boolean resetStateAccumulation) {
		
		if (isCompleted() || isUnallocated()){
			setInitialAmountRequested(amtRequested);
			myAllowPartialFillingFlag = partialFillFlag;
			mySatisfiedFlag = false;
			myCompletedFlag = false;
			myTimeRequested = Double.NaN;
			setPriority(priority);
			setEntity(entity);
			setQueuedObject(obj);
			
			if (isFullyReleased()){//if fully released, transition to unallocated state
				myFullyReleasedState.exit();
				myUnallocatedState.enter();
			}// else must be unallocated already
			
			if (resetStateAccumulation)
				resetAccumulatedStateTime();			
		} else
			throw new IllegalStateException("Cannot reinitialize a request that is not completed or that has been allocated.");
		

	}

	/** The request's accumulated time spent
	 *  in each state {Unallocated, Partially Allocated, Fully Allocated, 
	 *  Partially Released, Completed} will be cleared (initialize)
	 */
	public final void resetAccumulatedStateTime(){
		myUnallocatedState.initialize();
		myPartiallyAllocatedState.initialize();
		myFullyAllocatedState.initialize();
		myPartiallyReleasedState.initialize();
		myFullyReleasedState.initialize();		
	}
	
	/** Gets the entity that made the request
	 * 
	 * @return The entity that made the request
	 */
	public final Entity getEntity() {
		return (myEntity);
	}

	/** Gets the amount of time requested for the request
	 * 
	 * @return The amount of time requested
	 */
	public final double getTimeRequested() {
		return (myTimeRequested);
	}
	
	/** Can be used to return an event associated with the request
	 *  Can be null
	 * @return
	 */
	public final JSLEvent getEvent() {
		return myEvent;
	}

	/** Can be used to set an event associated with the request.
	 *  The event can be null.
	 * 
	 * @param event
	 */
	public final void setEvent(JSLEvent event) {
		myEvent = event;
	}

	/** Indicates whether or not the request has been allocated
	 *  its full amount, i.e. whether it has been satisfied
	 *  
	 *  The request is satisfied if the amount requested equals
	 *  the amount allocated, i.e. it has been allocated all that it
	 *  has requested.
	 * 
	 * @return true if the request is satisfied
	 */
	public final boolean isSatisfied(){
		return(mySatisfiedFlag);
	}

	/** Indicates whether or not the request has been completed.
	 *  The request is completed if the amount released equals
	 *  the amount allocated, i.e. it has released all that it
	 *  was allocated
	 * 
	 * @return true if the request is completed
	 */
	public final boolean isCompleted(){
		return(myCompletedFlag);
	}

	/** Indicates whether the request has not been allocated any
	 *  units from a resource
	 *  
	 * @return
	 */
	public final boolean isUnallocated() {
		return myUnallocatedState.isEntered();
	}

	/** Indicates whether the request has been partially allocated
	 *  to a resource
	 *  
	 * @return
	 */
	public final boolean isPartiallyAllocated() {
		return myPartiallyAllocatedState.isEntered();
	}
	
	/** Indicates whether the request has partially released some
	 *  of a resource
	 *  
	 * @return
	 */
	public final boolean isPartiallyReleased() {
		return myPartiallyReleasedState.isEntered();
	}
	
	/** Checks if the request has been fully released (i.e. completed)
	 * 
	 * @return
	 */
	public final boolean isFullyReleased(){
		return (myFullyReleasedState.isEntered());
	}
	
	/* mutators */
	
	/** A ResourceAllocationListener can be attached to the request and provide
	 *  behavior right after any units of a resource are allocated to the request
	 *  
	 * @param allocationListener
	 */
	public final void setResourceAllocationListener(ResourceAllocationListenerIfc allocationListener) {
		myResourceAllocationListener = allocationListener;
	}

	/** A ResourceReleaseListener can be attached to the request and provide
	 *  behavior right after the request has release units of a resource.
	 *  
	 * @param releaseListener
	 */
	public final void setResourceReleaseListener(ResourceReleaseListenerIfc releaseListener) {
		myResourceReleaseListener = releaseListener;
	}	
	
	/** The ModelElement that made the request.  This may be null if
	 *  the requesting element was not set.  The client is responsible
	 *  for casting the returned ModelElement to a more useful form.
	 *  
	 * @return Returns the ModelElement that made the request
	 */
	public final Object getRequestingObject() {
		return myRequestingObject;
	}
	
	/** Sets the  Object that made the request.  This may be null if
	 *  the requesting object was not set.  The client is responsible
	 *  for casting the returned Object to a more useful form.
	 * 
	 * @param requestingObject The requesting object
	 */
	public final void setRequestingObject(Object requestingObject) {
		myRequestingObject = requestingObject;
	}
	
	/** Sets the amount of time requested
	 * 
	 */
	public final void setTimeRequested(double timeRequested) {

		if (timeRequested < 0)
			throw new IllegalArgumentException("Time requested was less than zero!");

		myTimeRequested = timeRequested;
	}

	// ===========================================
	// PROTECTED METHODS
	// ===========================================
	

	protected final void setInitialAmountRequested(int amountRequested) {

		if (amountRequested <= 0)
			throw new IllegalArgumentException("Amount requested was less or equal to zero!");

		myInitialAmountRequested = amountRequested;
		myAmountAllocated = 0;
	}

	/** Allocates the amount indicated to the request from the given
	 *  resource
	 * 
	 * @param value
	 * @param resource
	 */
	protected final void allocateAmount(int value, Resource resource) {

		if (myCompletedFlag == true)
			throw new IllegalStateException("The request has already been completed!");			
			
		if ((myResource != null) && (myResource != resource))
			throw new IllegalArgumentException("The request was already using another resource!");		

		if (value <= 0)
			throw new IllegalArgumentException("Amount allocated was <= zero!");

		if (myAmountAllocated + value > myInitialAmountRequested )
			throw new IllegalArgumentException("Amount allocated was > the amount requested!");
	
		if ((myAllowPartialFillingFlag == false) && (value != myInitialAmountRequested))
			throw new IllegalArgumentException("No partial filling, the amount allocated was != the amount requested!");
				
		myResource = resource;

		myAmountAllocated = myAmountAllocated + value;
		
		myUnallocatedState.exit();
		
		myPartiallyAllocatedState.enter();
		
		if (myAmountAllocated == myInitialAmountRequested){
			myAmtRemainingToRelease = myInitialAmountRequested;
			mySatisfiedFlag = true;
			myFullyAllocatedState.enter();
			myPartiallyAllocatedState.exit();
			if ((myEntity != null) && (myResource != null))
				myEntity.addRequest(myResource, this);
		}

		if (myResourceAllocationListener != null)
			myResourceAllocationListener.allocated(this, myResource);
					
	}

	protected final void releaseAmount(int value) {
		
		if (mySatisfiedFlag == false)
			throw new IllegalStateException("The request was never satisfied. It was never allocated its full amount!");		
			
		if (myResource == null)
			throw new IllegalArgumentException("The request was not using a resource!");		

		if (value < 0)
			throw new IllegalArgumentException("Amount released was less zero!");

		if (value > myAmtRemainingToRelease)
			throw new IllegalArgumentException("Amount released was > amount remaining!");

		myAmtRemainingToRelease = myAmtRemainingToRelease - value;
		
		myFullyAllocatedState.exit();
		myPartiallyReleasedState.enter();

		if (myAmtRemainingToRelease == 0){
			myCompletedFlag = true;
			myPartiallyReleasedState.exit();
			myFullyReleasedState.enter();	
			if ((myEntity != null) && (myResource != null))
				myEntity.removeRequest(myResource);	
			myResource = null;
		}

		if (myResourceReleaseListener != null)
			myResourceReleaseListener.released(this, myResource);
		
	}
	
	protected final void setQueue(Queue queue){
		super.setQueue(queue);
	}
	
	protected final void setEntity(Entity entity) {
		myEntity = entity;
	}

	protected void setNulls(){
		super.setNulls();
		myResource = null;
		myEntity = null;	
		myEvent = null;
		myRequestingObject = null;
		myResourceReleaseListener = null;
		myResourceAllocationListener = null;
	}
	
	public void nullify(){
		super.nullify();
		setNulls();
		myUnallocatedState = null;
		myPartiallyAllocatedState = null;
		myFullyAllocatedState = null;
		myPartiallyReleasedState = null;
		myFullyReleasedState = null;
	}
	
    //  ===========================================
    //      STATIC PUBLIC METHODS
    //  ===========================================
	
	/** Gets the limit for the size of the Request cache
	 * 
	 * @return
	 */
	public static final int getRequestCacheSizeLimit(){
		return myRequestCacheSizeLimit;
	}
	
	/** Gets the current size of the Request cache
	 * 
	 * @return
	 */
	public static final int getRequestCacheSize(){
		return(myRequestCache.size());
	}
	   
	/** Sets the size of the Request cache, if the size is < 1
	 *  then the default size is used.
	 * 
	 * @param size
	 */
	public static final void setRequestCacheSizeLimit(int size) {
		if (size <= 0)
			myRequestCacheSizeLimit = DEFAULT_CACHE_LIMIT;
		else
			myRequestCacheSizeLimit = size;
	}
	
	//  ===========================================
	//      STATIC PRIVATE METHODS
	//  ===========================================


	

}
