/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.resource;

import java.util.*;

import jsl.modeling.IllegalStateException;
import jsl.modeling.ModelElement;
import jsl.modeling.ModelElementState;
import jsl.modeling.SchedulingElement;
import jsl.modeling.StateAccessorIfc;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.modeling.elements.resource.Request;
import jsl.modeling.elements.resource.ResourceSet;

/** A Resource has a given capacity of units that can be allocated to a Request.  When all
 *  the units of a resource have been allocated, it is considered busy (as a whole); otherwise,
 *  it is considered to be idle.  That is, if at least one unit of the resource is not allocated,
 *  then the resource is idle.  A resource cannot allocate more units than its currently 
 *  available capacity.  Resources are intimately tied to the Request class.  If units of a
 *  resource are allocated, they are associated with the request.  The only way to return the 
 *  units of the resource is to use the instance of Request and the release() methods of Resource.
 *  An instance of a Request can only be using the units of one resource at a time.  A request 
 *  might not have all of its requested units allocated, see allocate(Request, int)
 *
 */
public class Resource extends SchedulingElement implements ResourceAllocationListenerIfc {

    //  ===========================================
    //      CLASS AND OBJECT ATTRIBUTES
    //  ===========================================

	/** Holds references to objects that want to know when the 
	 *  resource changes state
	 */
	private List<ResourceStateListenerIfc> myResourceStateListeners;

	/** The list of resource sets currently holding this resource
	 */
	private Set<ResourceSet> myResourceSets;

	/**
	 *  The initial capacity of the resource unit
	 *  The maximum number of individual units that 
	 *  can be available at any time t
	 */
	private int myInitialCapacity;
	
	/**
	 *  The current number of idle units for the resource
	 */
	private TimeWeighted myNumIdle;

	/**
	 *  The current number busy for the resource
	 */
	private TimeWeighted myNumBusy;

	/** The busy state, keeps track of when all units are busy
	 * 
	 */
	private ModelElementState myBusyState;	
	
	/** The idle state, keeps track of when there are idle units
	 *  i.e. if any unit is idle then the resource as a whole is 
	 *  considered idle
	 */
	private ModelElementState myIdleState;
	
	/**
	 *  A list used to hold the requests that are in service
	 */
	private List<Request> myUsingResourceList;

	/** Constructs a resource with the given parent
	 * @param parent
	 */
	public Resource(ModelElement parent) {
		this(parent, 1, null);
	}

	/** Constructs a resource with the given parent
	 * @param parent
	 */
	public Resource(ModelElement parent, int capacity) {
		this(parent, capacity, null);
	}
	
    /** Constructs a Resource with the given name.  
     *  Initial capacity is 1
     * @param name The name of the resource
     */
    public Resource(ModelElement parent, String name){
        this(parent, 1, name);
    }
    
	/** Constructs a Resource with the given parent, capacity, and name
	 * @param parent
	 * @param capacity
	 * @param name
	 */
	public Resource(ModelElement parent, int capacity, String name) {
		super(parent, name);
        setInitialCapacity(capacity);
        myBusyState = new ModelElementState(this, getName() + " Busy State");
        myIdleState = new ModelElementState(this, getName() + " Idle State");
        setInitialNumberIdle(getInitialCapacity());
        myUsingResourceList = new ArrayList<Request>(getInitialCapacity());
        myResourceStateListeners = new LinkedList<ResourceStateListenerIfc>() ;
 		myResourceSets = new HashSet<ResourceSet>();
	}

    //  ===========================================
    //      PUBLIC METHODS
    //  ===========================================

	/** Checks if the resource is idle.  It is idle if
	 *  the number of idle units > 0
	 * 
	 * @return true if idle, false otherwise
	 */
	public final boolean isIdle(){
		// if the number idle = 0, then resource is busy
		// if the number idle > 0, then the resource is idle
		return(getNumberIdle() > 0);
	}
	
	/** Checks to see if the resource does not have any idle units
	 *  i.e. if getNumberIdle() == 0, then the resource is busy
	 * 
	 * @return
	 */
	public final boolean isBusy(){
		return(!isIdle());
	}
	
	/** Gets the initial capacity of the resource
	 * 
	 * @return
	 */
	public final int getInitialCapacity() {
		return myInitialCapacity;
	}

	public String toString(){
		String s = getName() + " number idle= " + getNumberIdle() + " capacity= " + getInitialCapacity();
		return(s);
	}
	
	/** Returns a StateAccessorIfc so that tabulated state information can
	 *  be retreived on the busy state of the resource
	 * 
	 * @return
	 */
	public final StateAccessorIfc getBusyState(){
		return(myBusyState);
	}
	
	/** Returns a StateAccessorIfc so that tabulated state information can
	 *  be retreived on the idle state of the resource
	 * 
	 * @return
	 */
	public final StateAccessorIfc getIdleState(){
		return(myIdleState);
	}
	
    /** Sets the initial capacity of the resource. This only changes it for
     *  when the resource is initialized. 
     * 
     * @param capacity
     */
	public void setInitialCapacity(int capacity){
	    	if (capacity <= 0)
	    		throw new IllegalArgumentException("Attempted to set resource capacity less or equal to zero!");
	    	myInitialCapacity = capacity;
    }
	
	/** Returns the current number of idle units
	 * 
	 * @return
	 */
	public final double getNumberIdle() {
		return myNumIdle.getValue();
	}

	/** Returns the current number of busy units
	 * 
	 * @return
	 */
	public final double getNumberBusy() {
		return myNumBusy.getValue();
	}
	
	/** Allows an observer to be attached to the number busy time weighted variable
	 * 
	 * @param observer
	 */
	public final void addNumberBusyObserver(Observer observer){
		myNumBusy.addObserver(observer);
	}

	/** Allows an observer to be attached to the number idle time weighted variable
	 * 
	 * @param observer
	 */
	public final void addNumberIdleObserver(Observer observer){
		myNumIdle.addObserver(observer);
	}
	
	/** Returns an unmodifiable collection holding all the requests
	 *  currently using the resource.  The list is unchangeable, but the elements in
	 *  the list can be still be accessed and changed as necessary
	 * 
	 * @return An unmodifiable List of the requests
	 */
	public final List<Request> getRequestsUsingResource(){
		return(Collections.unmodifiableList(myUsingResourceList));
	}

	/** Returns an unmodifiable collection holding all the ResourceSets
	 *  currently holding the resource.  The list is unchangeable, but the elements in
	 *  the list can be still be accessed and changed as necessary
	 * 
	 * @return An unmodifiable Set holding the ResourceSets
	 */
	public final Set<ResourceSet> getResourceSetsHoldingResource(){
		return(Collections.unmodifiableSet(myResourceSets));
	}

	/** Adds the ResourceStateListener to the resource.  The order 
	 *  that the listeners are added, implies the order in which they 
	 *  will be called.  Because a listener may change the state of the resource
	 *  each listener should check the state of the resource after being called.
	 * 
	 * @param listener
	 */
	public final void addStateListener(ResourceStateListenerIfc listener){
		if (listener == null)
			throw new IllegalArgumentException("The state listener was null");
		
		myResourceStateListeners.add(listener);
		
	}

	/** Removes the ResourceStateListener.
	 * 
	 * @param listener
	 */
	public final boolean removeStateListener(ResourceStateListenerIfc listener){		
		return(myResourceStateListeners.remove(listener));		
	}
	
	/** Returns an iterator to the list of state change listeners.  Could be 
	 *  used to reorder the list
	 * 
	 * @return
	 */
	public final ListIterator<ResourceStateListenerIfc> getStateListenerIterator(){
		return(myResourceStateListeners.listIterator());
	}
	
	/** Allocates 1 idle unit of a resource and creates a request
	 * 
	 *  Conditions for using this method include:
	 * 
	 * 	1) The resource has idle units available, i.e. isIdle() returns true
	 *  2) The amount requested <= the number of idle units
	 *  3) See requirements on allocate(Request request, int amtToAllocate)
	 *  
	 *  The request will be completely allocated after this method, i.e.
	 *  all of the amount requested by the request will be given to the request
	 *  by the resource.
	 * 
	 * @return
	 */
	public final Request allocate(){
		return(allocate(1));
	}
	
	/** Allocates the number of idle units requested of a resource and creates a request
	 * 
	 *  Conditions for using this method include:
	 * 
	 * 	1) The resource has idle units available, i.e. isIdle() returns true
	 *  2) The amount requested <= the number of idle units
	 *  3) See requirements on allocate(Request request, int amtToAllocate)
	 *  
	 *  The request will be completely allocated after this method, i.e.
	 *  all of the amount requested by the request will be given to the request
	 *  by the resource.

	 * @param amtRequested
	 * @return
	 */
	public final Request allocate(int amtRequested){
		Request r = Request.create(amtRequested);
		allocate(r);
		return(r);
	}
	
	/** Allocates idle units of a resource to a request.  
	 * 
	 *  Conditions for using this method include:
	 * 
	 * 	1) The resource has idle units available, i.e. isIdle() returns true
	 *  2) The amount requested <= the number of idle units
	 *  3) The supplied request must not be null.
	 *  
	 *  The request will be completely allocated after this method, i.e.
	 *  all of the amount requested by the request will be given to the request
	 *  by the resource.
	 * 
	 * @param request
	 * @param amtToAllocate
	 */
	public final void allocate(Request request){
		if (request == null)
			throw new IllegalArgumentException("The supplied request was null!");
		allocate(request, request.getAmountNeeded());
	}
	
	/** Allocates idle units of a resource to a request.  
	 * 
	 *  Conditions for using this method include:
	 * 
	 * 	1) The resource has idle units available, i.e. isIdle() returns true
	 *  2) The amount requested <= the number of idle units
	 *  3) The supplied request must not be null.
	 *  4) The amount to allocate must be <= the number of idle units
	 *  5) If units have been allocated to the request previously, then the
	 *     resource associated with the request must match this resource
	 *  6) The request must not have already been satisfied i.e. isSatisfied() returns false
	 *  
	 *  The request may or may not be fully allocated after this method, i.e. isAllocated()
	 *  may not be true.
	 * 
	 * @param request
	 * @param amtToAllocate
	 */
	public final void allocate(Request request, int amtToAllocate){
		if (request == null)
			throw new IllegalArgumentException("The supplied request was null!");

		if (amtToAllocate <= 0)
			throw new IllegalArgumentException("The amount to allocate was <= 0 !");
		
		if (amtToAllocate > getNumberIdle())
			throw new IllegalArgumentException("The amount was for more than the number of idle units for this resource.");

		if (request.getAmountNeeded() > getInitialCapacity())
			throw new IllegalArgumentException("The amount needed was > the capacity of the resource!");
		
		if (request.isSatisfied() == true)
			throw new IllegalStateException("The supplied request has already been satisfied!");

		if ((request.isPartiallyAllocated() == true) && (request.getResource() != this))
			throw new IllegalStateException("The supplied request is not valid for this resource!");

		// amtToAllocate is > 0 and <= number idle

		// allocate units of this resource to the request
		request.allocateAmount(amtToAllocate, this);

		// consider the request as using the resource
		myUsingResourceList.add(request);
						
		// the number of idle units should be reduced by the amount allocated
		myNumIdle.decrement(amtToAllocate);
		
		// the number of busy units should be increased by the amount allocated
		myNumBusy.increment(amtToAllocate);
		
		// update the state, notify listeners that the resource is busy
		if (isBusy()){
			myIdleState.exit();
			myBusyState.enter();
			notifyResourceSetsOfBusyness();
			notifyResourceStateListeners(request);
		}
		
		// notify any outside observes that there was an update on this resource
		notifyUpdateObservers();
	
	}
	
	/** Releases all the units allocated to the request. Conditions
	 *  on the use of this method:
	 *  
	 *  1) The supplied request must not be null.
	 *  2) The supplied request must have been created based on the use of this resource, 
	 *     i.e. it's getResource() method must return this resource and isUsingResource() is true
	 *  3) The release amount must be > 0 and <= request.getAmountAllocated() 
	 * 
	 * @param request
	 * @param releaseAmount
	 */
	public final void release(Request request){
		release(request, request.getAmountAllocated());
	}
	
	/** Releases the amount of units specified for the resource. Conditions
	 *  on the use of this method:
	 *  
	 *  1) The supplied request must not be null.
	 *  2) The supplied request must have been created based on the use of this resource, 
	 *     i.e. it's getResource() method must return this resource and isUsingResource() is true
	 *  3) The release amount must be > 0 and <= request.getAmountAllocated() 
	 * 
	 * @param request
	 * @param releaseAmount
	 */
	public final void release(Request request, int releaseAmount){

		if (request == null)
			throw new IllegalArgumentException("The supplied request was null!");
		
		if (request.getResource() != this)
			throw new IllegalArgumentException("The supplied request was is not valid for this resource!");

		if (!request.isSatisfied())
			throw new IllegalStateException("Tried to release a request that has not been satisfied.!");

		if (releaseAmount <= 0)
			throw new IllegalArgumentException("The amount being release is <=0!");

		if (releaseAmount > request.getAmountAllocated())
			throw new IllegalArgumentException("The amount being release is greater than the amount allocated to the request!");

		request.releaseAmount(releaseAmount);
		
		if (request.isCompleted()){
			// the request is completed and needs to be removed from service
			myUsingResourceList.remove(request);
		}

		// the number of idle units should be increased by the amount released
		myNumIdle.increment(releaseAmount);
		
		// the number of busy units should be decreased by the amount released
		myNumBusy.decrement(releaseAmount);

		if (isIdle()){
			myIdleState.enter();
			myBusyState.exit();
			notifyResourceSetsOfIdleness();
			notifyResourceStateListeners(request);
		}	
		
		// notify any outside observes that there was an update on this resource
		notifyUpdateObservers();
	}


	/** This method can be overridden by subclasses to provide behavior when
	 *  the resource is allocated
	 * 
	 * @param request
	 * @param resource
	 */
	public void allocated(Request request, Resource resource) {
		
	}
	
    //  ===========================================
    //      PROTECTED METHODS
    //  ===========================================

	/** can be called to initialize the resource
	 * The default behavior is to set the number idle equal to the capacity
	 *
	 **/
	protected void initialize(){
		super.initialize();
		setInitialNumberIdle(getInitialCapacity());

		myUsingResourceList.clear();	
		
		if (isIdle()){
			notifyResourceSetsOfIdleness();
			myIdleState.enter();
		}
		else{
			notifyResourceSetsOfBusyness();
			myBusyState.enter();
		}
			

	}
	       
	/** Sets the number available for the resource unit
	 * 
	 * @param value
	 */
	protected void setInitialNumberIdle(int value){
		if (value < 0)
			throw new IllegalArgumentException("Invalid argument. Attempted to set resource number idle less than Zero!");
		
		if (value > myInitialCapacity)
			throw new IllegalArgumentException("Invalid argument. Attempted to set resource number idle greater than capacity!");

        if (myNumIdle == null){
			myNumIdle = new TimeWeighted(this, (double)value, getName()+ " Number Idle Units");
			myNumBusy = new TimeWeighted(this, getInitialCapacity()- myNumIdle.getInitialValue(), getName()+ " Number Busy Units");
        }
        else {
			myNumIdle.setInitialValue((double)value);
			myNumBusy.setInitialValue(getInitialCapacity()-myNumIdle.getInitialValue());
        }
        
	}   
	
	protected final boolean addResourceSet(ResourceSet set){
		return(myResourceSets.add(set));
	}
	
	protected final boolean removeResourceSet(ResourceSet set){
		return(myResourceSets.remove(set));
	}

	protected final void notifyResourceSetsOfIdleness(){		
		for(ResourceSet set: myResourceSets){
			set.resourceBecameIdle(this);
		}	
	}

	protected final void notifyResourceSetsOfBusyness(){		
		for(ResourceSet set: myResourceSets){
			set.resourceBecameBusy(this);
		}	
	}
	
	protected final void notifyResourceStateListeners(Request request){
		
		for(ResourceStateListenerIfc listener: myResourceStateListeners){
			listener.resourceStateChange(this, request);
		}
		
	}

}
