/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.resource;

import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.queue.QueueDiscipline;

public class ResourceProvider extends SchedulingElement {

	/** The resource set used by this provider to ask for
	 *  and and return idle resources
	 */
	protected ResourceSet myResourceSet;
	
	/** A queue to hold waiting requests when an idle
	 *  resource is not immediately available. 
	 */
	protected Queue myRequestQ;
	
	/** Can be used to supply a rule for how the requests 
	 *  are selected for allocation
	 */
	protected RequestSelectionRuleIfc myInitialRequestSelectionRule;

	/** Can be used to supply a rule for how the requests 
	 *  are selected for allocation
	 */
	protected RequestSelectionRuleIfc myRequestSelectionRule;

	/** Creates a ResourceProviderr that uses a FIFO queue discipline. An
	 *  empty ResourceSet is created and must be filled
	 * 
	 * @param parent
	 */
	public ResourceProvider(ModelElement parent) {
		this(parent, null, null, null);
	}

	/** Creates a ResourceProviderr that uses a FIFO queue discipline. An
	 *  empty ResourceSet is created and must be filled
	 * 
	 * @param parent
	 * @param name
	 */
	public ResourceProvider(ModelElement parent, String name) {
		this(parent, name, null, null);
	}
	
	/** Creates a ResourceProvider that uses the supplied set and FIFO queue discipline
	 * 
	 * @param parent
	 * @param name
	 * @param set
	 */
	public ResourceProvider(ModelElement parent, String name, ResourceSet set) {
		this(parent, name, set, null);
	}

	/** Creates a ResourceProvider that uses the supplied set and FIFO queue discipline
	 * 
	 * @param parent
	 * @param set
	 */
	public ResourceProvider(ModelElement parent, ResourceSet set) {
		this(parent, null, set, null);
	}
	
	/** Creates a ResourceProvider that uses the supplied set and queue discipline
	 * 
	 * @param parent
	 * @param set
	 * @param discipline
	 */
	public ResourceProvider(ModelElement parent, ResourceSet set, QueueDiscipline discipline) {
		this(parent, null, set, discipline);
	}
	
	/** Creates a ResourceProvider that uses the supplied set and queue discipline
	 * 
	 * @param parent
	 * @param name
	 * @param set
	 * @param discipline
	 */
	public ResourceProvider(ModelElement parent, String name, ResourceSet set, QueueDiscipline discipline) {
		super(parent, name);
		
		if (discipline == null)
			discipline = Queue.FIFO;
		
		myRequestQ = new Queue(this, getName() + " ResourceProviderQ", discipline);

		setResourceSet(set);
		
	}

	/** Returns a reference to the resource set used within the provider
	 *  
	 * @return
	 */
	public final ResourceSet getResourceSet() {
		return myResourceSet;
	}


	/** Adds a resource with unit capacity to the
	 *  underlying resource set
	 *
	 * @return
	 */
	public final Resource addResource(){
		return(addResource(1, null));
	}
	
	/** Adds a resource with the given capacity to the
	 *  underlying resource set
	 * 
	 * @param capacity
	 * @return
	 */
	public final Resource addResource(int capacity){
		return(addResource(capacity, null));
	}
	
	/** Adds a resource with the given capacity and name to the
	 *  underlying resource set
	 * 
	 * @param capacity
	 * @param name
	 * @return
	 */
	public Resource addResource(int capacity, String name){
		return(myResourceSet.addResource(capacity, name));
	}
	
	/** Adds the given resource to the underlying set of the provider
	 * 
	 * @param resource
	 */
	public final void addResource(Resource resource){
		myResourceSet.addResource(resource);
	}
	
	/** Returns the current number of idle resources within the underlying
	 *  resource set
	 * 
	 * @return
	 */
	public final int getNumberOfIdleResources(){
		return(myResourceSet.getNumberOfIdleResources());
	}
	
	/** This will change the queue discipline of the underlying Queue
	 * 
	 * @param discipline
	 */
	public final void changeQueueDiscipline(QueueDiscipline discipline){
		myRequestQ.changeDiscipline(discipline);
	}
	
	/** Returns the initial discipline for the queue
	 * 
	 * @return
	 */
	public final QueueDiscipline getQueueInitialDiscipline() {
		return myRequestQ.getInitialDiscipline();
	}

	/** Sets the initial queue discipline
	 * 
	 * @param discipline
	 */
	public final void setQueueInitialDiscipline(QueueDiscipline discipline) {
		myRequestQ.setInitialDiscipline(discipline);
	}

	/** Returns the current number of requests in the queue
	 * 
	 * @return
	 */
	public final int getNumberInQueue(){
		return(myRequestQ.size());
	}
	
	/** If the request is unallocated and in the queue, this removes it.  It does this
	 *  without collecting statistics on the time in queue for the request.
	 * 
	 * @param request Should not be null
	 */
	public final void cancelRequest(Request request){
		if (request.isUnallocated())
			myRequestQ.remove(request, false);
	}

	/** The client asks the ResourceProvider to seize 1 unit of a resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 * @return
	 */
	public final Request seize(ResourceAllocationListenerIfc allocationListener){
		return(seize(1, allocationListener, null, null, Request.DEFAULT_PRIORITY, false, null));
	}

	/** The client asks the ResourceProvider to seize 1 unit of a resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 *  @param entity The entity to be associated with this request during the seize
	 * @return
	 */
	public final Request seize(ResourceAllocationListenerIfc allocationListener, Entity entity){
		return(seize(1, allocationListener, entity, null, Request.DEFAULT_PRIORITY, false, null));
	}	
	
	
	/** The client asks the ResourceProvider to seize 1 unit of a resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 * @param requestingObject This object will be attached to the request and can be "called back" as the requester as needed
	 * @return
	 */
	public final Request seize(ResourceAllocationListenerIfc allocationListener, Object requestingObject){
		return(seize(1, allocationListener, null, null, Request.DEFAULT_PRIORITY, false, requestingObject));
	}
	
	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener){
		return(seize(amount, allocationListener, null, null, Request.DEFAULT_PRIORITY, false, null));
	}

	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 *  @param priority The priority of the request
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, int priority){
		return(seize(amount, allocationListener, null, null, priority, false, null));
	}
	
	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 *  @param entity The entity to be associated with this request during the seize
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, Entity entity){
		return(seize(amount, allocationListener, entity, null, Request.DEFAULT_PRIORITY, false, null));
	}

	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 *  @param entity The entity to be associated with this request during the seize
	 *  @param requestingObj An object to be attached to the request
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, 
			Entity entity, Object requestingObj){
		return(seize(amount, allocationListener, entity, null, Request.DEFAULT_PRIORITY, false, requestingObj));
	}
	
	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 *  @param entity The entity to be associated with this request during the seize
	 *  @param priority The priority of the request
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, Entity entity, int priority){
		return(seize(amount, allocationListener, entity, null, priority, false, null));
	}
	
	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 * @param requestingObject This object will be attached to the request and can be "called back" as the requester as needed
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, Object requestingObject){
		return(seize(amount, allocationListener, null, null, Request.DEFAULT_PRIORITY, false, requestingObject));
	}
	
	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 * @param releaseListener This is attached to the request so that it can react when the resource
	 *  is released
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, 
			ResourceReleaseListenerIfc releaseListener){
		return(seize(amount, allocationListener, null, releaseListener, Request.DEFAULT_PRIORITY, false, null));
	}

	/** The client asks the ResourceProvider to seize 1 unit of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 * @param releaseListener This is attached to the request so that it can react when the resource
	 *  is released
	 * @return
	 */
	public final Request seize(ResourceAllocationListenerIfc allocationListener, 
			ResourceReleaseListenerIfc releaseListener){
		return(seize(1, allocationListener, null, releaseListener, Request.DEFAULT_PRIORITY, false, null));
	}	
	
	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 * @param releaseListener This is attached to the request so that it can react when the resource
	 *  is released
	 * @param requestingObject This object will be attached to the request and can be "called back" as the requester as needed
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, 
			ResourceReleaseListenerIfc releaseListener, Object requestingObject){
		return(seize(amount, allocationListener, null, releaseListener, Request.DEFAULT_PRIORITY, false, requestingObject));
	}
	
	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 * @param releaseListener This is attached to the request so that it can react when the resource
	 *  is released
	 * @param priority Provides for a ranking in the queue for the request if the queue is ranked
	 * @param requestingObject This object will be attached to the request and can be "called back" as the requester as needed
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, 
			ResourceReleaseListenerIfc releaseListener, int priority, Object requestingObject){
		return(seize(amount, allocationListener, null, releaseListener, priority, false, requestingObject));
	}

	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 * @param releaseListener This is attached to the request so that it can react when the resource
	 *  is released
	 * @param priority Provides for a ranking in the queue for the request if the queue is ranked
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, 
			ResourceReleaseListenerIfc releaseListener, int priority){
		return(seize(amount, allocationListener, null, releaseListener, priority, false, null));
	}

	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 * @param releaseListener This is attached to the request so that it can react when the resource
	 *  is released
	 * @param priority Provides for a ranking in the queue for the request if the queue is ranked
	 * @param allowPartialFill Indicates whether the request allows partial filling
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, 
			ResourceReleaseListenerIfc releaseListener, int priority, boolean allowPartialFill){
		return(seize(amount, allocationListener, null, releaseListener, priority, allowPartialFill, null));
	}

	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 *  @param entity The entity that is to be associated with the request for the seize
	 * @param releaseListener This is attached to the request so that it can react when the resource
	 *  is released
	 * @param priority Provides for a ranking in the queue for the request if the queue is ranked
	 * @param allowPartialFill Indicates whether the request allows partial filling
	 * @param requestingObject This object will be attached to the request and can be "called back" as the requester as needed
	 * @return
	 */
	public final Request seize(int amount, ResourceAllocationListenerIfc allocationListener, 
			ResourceReleaseListenerIfc releaseListener, int priority,
			boolean allowPartialFill, Object requestingObject){
		return(seize(amount, allocationListener, null, releaseListener, priority, allowPartialFill, null));
	}
	
	/** The client asks the ResourceProvider to seize the given amount of the resource.  
	 *  The request immediately enters the providers queue. If there is an idle resource
	 *  available that can be used to allocate to the request then the request is processed.
	 *  If the request can be satisfied in full it is removed from the queue and the
	 *  units of the resource are allocated to it.  If it cannot be allocated in full, it is
	 *  checked to see if it allows partial filling, if so the available units of the resource
	 *  are allocated to it, but it remains in the queue.
	 *  
	 *  NOTE:  It is up to the ResourceAllocationListenerIfc to check if the request is
	 *  still in the queue or if it has been satisfied before proceeding.
	 *  
	 *  The amount requested must not exceed the maximum capacity of the underlying resource
	 *  set.
	 * 
	 * @param amount The amount requested
	 * @param allocationListener This is attached to the request so that it can react when units
	 *  of the resource are allocated to it.
	 *  @param entity The entity that is to be associated with the request for the seize
	 * @param releaseListener This is attached to the request so that it can react when the resource
	 *  is released
	 * @param priority Provides for a ranking in the queue for the request if the queue is ranked
	 * @param allowPartialFill Indicates whether the request allows partial filling
	 * @param requestingObject This object will be attached to the request and can be "called back" as the requester as needed
	 * @return
	 */
	public Request seize(int amount, ResourceAllocationListenerIfc allocationListener, Entity entity,
			ResourceReleaseListenerIfc releaseListener, int priority,
			boolean allowPartialFill, Object requestingObject){

		if (allocationListener == null)
			throw new IllegalArgumentException("The supplied ResourceAllocationListenerIfc was null!");

		if (amount > myResourceSet.getMaxCapacity())
			throw new IllegalArgumentException("The amount requested was > the maximum capacity of any resource in the set!");

		Request request = Request.create(amount, priority, entity, allowPartialFill);
		request.setResourceAllocationListener(allocationListener);
		request.setResourceReleaseListener(releaseListener);
		if (requestingObject != null)
			request.setRequestingObject(requestingObject);
		
		// always enqueue the request
		myRequestQ.enqueue(request);
		
		// select the next request for the resource
		Request nextRequest = selectNextRequest();
		
		// if the selected request is the same as the entering request
		// we can try to provide a resource for it
		Resource resource = null;
		if (nextRequest == request){
			// check if there is an idle resource
			resource = myResourceSet.selectIdleResource(nextRequest);
			
			if (resource != null){ // an idle resource has been selected for the request	
				// can the resource fill the entire request				
				if (nextRequest.getAmountNeeded() <= resource.getNumberIdle()){
					if(nextRequest == myRequestQ.peekNext())
						myRequestQ.removeNext();
					else 
						myRequestQ.remove(myRequestQ.indexOf(nextRequest));
					resource.allocate(nextRequest);					
				} else {// can't fill full request
					// leave the request in the queue, but allocate units if it allows partial filling
					if (nextRequest.getAllowPartialFillingFlag()){// request allows partial filling
						// allocate whatever it can to the request
						resource.allocate(nextRequest, (int)resource.getNumberIdle());
					}
				}
			}
		}
		
		return(request);
	}

	/** Returns a reference to the request selection rule. May be null.
	 * 
	 * @return
	 */
	public final RequestSelectionRuleIfc getRequestSelectionRule() {
		return myRequestSelectionRule;
	}

	/** A request selection rule can be supplied to provide alternative behavior
	 *  within the selectNextRequest() method. A request selection rule, provides
	 *  a mechanism to select the next request from the queue of waiting requests
	 * 
	 * @param rule
	 */
	public final void setRequestSelectionRule(RequestSelectionRuleIfc rule) {
		myRequestSelectionRule = rule;
	}
	
	/** The rule to use when this provider is initialized
	 * 
	 * @return
	 */
	public final RequestSelectionRuleIfc getInitialRequestSelectionRule() {
		return myInitialRequestSelectionRule;
	}

	/** The rule to use when this provider is initialized
	 * 
	 */
	public final void setInitialRequestSelectionRule(RequestSelectionRuleIfc rule) {
		myInitialRequestSelectionRule = rule;
	}

	/** Selects a candidate request from the queue for allocation
	 *  to one of the resources.  The selection process does not remove
	 *  the request from the queue.
	 *  
	 * @return The request that was selected to for a resource
	 */	
	protected Request selectNextRequest(){
		Request r = null;
		if (myRequestSelectionRule != null)
			r = myRequestSelectionRule.selectNextRequest(myRequestQ, this);
		else
			r = (Request)myRequestQ.peekNext();	
		return(r);
	}
	
	protected void initialize(){
		setRequestSelectionRule(getInitialRequestSelectionRule());
	}
	
	protected final void setResourceSet(ResourceSet set){

		if ((set == null) && (myResourceSet == null))
			myResourceSet = new ResourceSet(this, getName() + " ResourceSet");
		else
			myResourceSet = set;
		
		myResourceSet.addResourceProvider(this);

	}

	/** This is called when the resource set has an idle resource returned to it
	 * 
	 */
	protected void resourceReleased(Resource releasedResource) {

		// there are requests waiting for a resource
		boolean done = false;
		while ((!done) && (!myRequestQ.isEmpty())) {
			// select the next request for the resource
			Request request = selectNextRequest();

			// check if there is an idle resource
			Resource resource = myResourceSet.selectIdleResource(request);
			if (resource == null) {
				done = true;
			} else { // an idle resource has been selected for the request
				// can the resource fill the entire request
				if (request.getAmountNeeded() <= resource.getNumberIdle()) {
					if (request == myRequestQ.peekNext())
						myRequestQ.removeNext();
					else
						myRequestQ.remove(myRequestQ.indexOf(request));
					resource.allocate(request);
				} else {// can't fill full request
					// leave the request in the queue, but allocate units if it
					// allows partial filling
					if (request.getAllowPartialFillingFlag()) {// request allows partial filling
						// allocate whatever it can to the request
						resource.allocate(request, (int) resource.getNumberIdle());
					}
				}
			}
		}
	}	
}
