/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.resource;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.TimeWeighted;

/** A ResourceSet holds a set of resources so that they can be selected
 *  for allocation via a ResouceSelectionRuleIfc.  A ResouceSet can have
 *  many resources and a resource can be in many resource sets.  The
 *  selectIdleResource(Request r) method uses the ResourceSelectionRuleIfc (if provided)
 *  to select the next idle resource for a request.  The default is to cycle
 *  through the resources to select the resources in the order in which they were
 *  released.
 */
public class ResourceSet extends ModelElement {

	protected List<Resource> myResources;

	protected List<Resource> myIdleResources;
	
	protected List<ResourceProvider> myResourceProviders;
	
	protected ResourceSelectionRuleIfc myResourceSelectionRule;
	
	protected TimeWeighted myNumIdleResources;

	private int myMaxCapacity;
	
	/** Creates a resource set with the given model element as a parent 
	 *  and assigned a default name.
	 * 
	 * @param parent
	 */
	public ResourceSet(ModelElement parent) {
		this(parent, null);
	}

	/** Creates a resource set with the given model element as a parent and the 
	 *  given name.  
	 * 
	 * @param parent
	 * @param name
	 */
	public ResourceSet(ModelElement parent, String name) {
		super(parent, name);
		myResources = new LinkedList<Resource>();
		myIdleResources = new LinkedList<Resource>();
		myResourceProviders = new LinkedList<ResourceProvider>();
		myNumIdleResources = new TimeWeighted(this, getName() + " Num Idle Resources");
	}

	/**
	 * @return Returns the MaxCapacity.
	 */
	public final int getMaxCapacity() {
		return myMaxCapacity;
	}
	
	/** Creates and adds a resource to the set
	 * 
	 * @param resource
	 * @return
	 */
	public final Resource addResource() {		
		return(addResource(1));
	}
	
	/** Creates and adds a resource to the set
	 * 
	 * @param capacity
	 * @return
	 */
	public final Resource addResource(int capacity) {		
		Resource resource = new Resource(this, capacity);
		addResource(resource);
		return(resource);
	}

	/** Creates and adds a resource to the set
	 * 
	 * @param capacity
	 * @param name
	 * @return
	 */
	public final Resource addResource(int capacity, String name) {		
		Resource resource = new Resource(this, capacity, name);
		addResource(resource);
		return(resource);
	}

	/** Adds a resource to the set.  
	 * 
	 * @param resource
	 * @return
	 */
	public final void addResource(Resource resource) {
		if (resource == null)
			throw new IllegalArgumentException("The supplied resource was null!");
		int capacity = resource.getInitialCapacity();
		if (capacity > myMaxCapacity)
			myMaxCapacity = capacity;	
		myResources.add(resource);
		resource.addResourceSet(this);
	}	
	
	/** Returns an iterator to the resources in this set
	 * 
	 * @return
	 */
	public final ListIterator<Resource> getResourceIterator(){
		return(myResources.listIterator());
	}

	/** Returns the next idle resource to be used for allocating
	 *  to requests, null if none are found that can satisfy the request. 
	 *  
	 *  The default selection method is to cycle through the resources in the 
	 *  order in which they were RELEASED.
	 *  
	 *  The idle resource set starts in the order the resouces were added to the
	 *  resouce set and assumes that the first resource listed was the resource that
	 *  was allocated the longest time in the past, and so on.  For example,
	 *  if the set of idle resources has the following list:
	 *  
	 *  Resource B
	 *  Resource A
	 *  Resource C
	 *  
	 *  Then C was the most recently used, resource A the next most recently used, and 
	 *  resource B the oldest used.  Resource B will be next allocated, if it can
	 *  satisfy the amount of the request. If not, Resource A will be checked, and so forth
	 *  until a resource that can satisfy the request is found. Thus, it is entirely
	 *  possible that B will not be the next recommended resource. The selected resource
	 *  is added to the end of the list after it is released.
	 *  
	 *  Note that the the order can vary with this list depending on the order in which the resources are 
	 *  released.  To provide alternative behavior either override
	 *  this method, or provide a ResourceSelectionRule
	 * 
	 * @param request
	 * @return
	 */
	public Resource selectIdleResource(Request request){

		if (request == null)
			throw new IllegalArgumentException("The supplied request was null!");
		
		if (request.getInitialAmountRequested() > getMaxCapacity())
			throw new IllegalArgumentException("The amount requested was > the maximum capacity of any resource in the set!");
			
		if (myIdleResources.isEmpty())
			return (null);
		
		if (myResourceSelectionRule != null){
			return(myResourceSelectionRule.selectResource(myIdleResources, request));
		} else {
			// the default is the first resource in the idle resource list
			// that can satisfy all of the request
			
			Resource found = null;
			
			if (!myIdleResources.isEmpty()){
				Iterator<Resource> iter = myIdleResources.iterator();
				Resource r = null;
				int amt = request.getAmountNeeded();
				while(iter.hasNext()){
					r = iter.next();
					if (amt <= r.getNumberIdle()){
						found = r;
						break;
					}			
				}
			}
			return(found);		
		}

	}

	/** Returns the current number of idle resources in the set
	 * 
	 * @return
	 */
	public final int getNumberOfIdleResources(){
		return(myIdleResources.size());
	}
	
	public String toString(){
		return(myResources.toString());
	}
	
    /** Returns the current resource selection rule or null if none
     * 
     * @return
     */
	public final ResourceSelectionRuleIfc getResourceSelectionRule() {
		return myResourceSelectionRule;
	}

	/** Sets the resource selection rule.  The supplied rule is responsible for
	 *  both recommending an idle resource (not removing it) and for returning an idle
	 *  resource back to the list.  This allows the rule to maintain the list in an
	 *  order if necessary.
	 * 
	 * @param rule
	 */
	public final void setResourceSelectionRule(ResourceSelectionRuleIfc rule) {
		myResourceSelectionRule = rule;
	}

	protected boolean resourceBecameBusy(Resource resource){

		if (resource == null)
			throw new IllegalArgumentException("The supplied resource was null!");

		if (!resource.isBusy())
			throw new IllegalArgumentException("The supplied resource was not busy!");
		
		if (!myResources.contains(resource))
			throw new IllegalArgumentException("The supplied resource is not a member of this resource set!");

		if (myIdleResources.isEmpty())
			return (false);

		if (!myIdleResources.contains(resource))
			throw new IllegalArgumentException("The supplied resource is not a member of this idle resource set!");

		boolean flag = myIdleResources.remove(resource);
		
		myNumIdleResources.setValue(myIdleResources.size());

		return(flag);

	}
	
	protected void resourceBecameIdle(Resource resource){
		
		if (resource == null)
			throw new IllegalArgumentException("The supplied resource was null!");

		if (!resource.isIdle())
			throw new IllegalArgumentException("The supplied resource is not idle!");

		if (!myResources.contains(resource))
			throw new IllegalArgumentException("The supplied resource is not a member of this resource set!");

		// add it only if it is not already in the idle list
		if (!myIdleResources.contains(resource)){
			if (myResourceSelectionRule != null)
				myResourceSelectionRule.addIdleResource(myIdleResources, resource);
			else
				myIdleResources.add(resource);
			
			myNumIdleResources.setValue(myIdleResources.size());

		}
		
		// notify any resource providers that a idle resource has been released
		for(ResourceProvider tp: myResourceProviders){
			tp.resourceReleased(resource);
		}

	}

	protected final boolean addResourceProvider(ResourceProvider arg0) {
		return myResourceProviders.add(arg0);
	}
	 
	protected void initialize(){
		
		for(Resource r: myResources){
			if (r.isIdle()){
				// add it only if it is not already in the idle list
				if (!myIdleResources.contains(r)){
					if (myResourceSelectionRule != null)
						myResourceSelectionRule.addIdleResource(myIdleResources, r);
					else
						myIdleResources.add(r);
				}			
			}
		}
		
		myNumIdleResources.setInitialValue(myIdleResources.size());

	}
}
