/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.spatial;

import jsl.modeling.ActionListenerIfc;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.reporting.JSL;

/** An AbstractMover implements the MoverIfc and is a model element (SpatialModelElement).
 *  It represents an object that can move through a spatial model.  It can begin a trip and make
 *  movements towards the destination associated with the trip.  Each trip can be broken up
 *  into a series of movements.  Each movement can have a different length and velocity as long
 *  as the object eventually completes the trip. 
 *  
 *  It has default behaviors that can be modified by some helper classes.
 *  
 *  A MovementControllerIfc can be supplied to control the movement within a trip.
 *  A CollisionDetectorIfc can be supplied to detect collisions while moving.
 *  A CollisionHandlerIfc can be supplied to handle (react) to collisions detected while moving
 *  A OutsideSpatialModelHandlerIfc can be supplied to handle the situation if the the object
 *  attempts to go outside the valid frame of reference associated with its spatail model
 *
 */
abstract public class AbstractMover extends SpatialModelElement implements MoverIfc {

	// ===========================================
	// CLASS CONSTANTS
	// ===========================================

	/**
	 * Indicates that a AbstractMover started a trip
	 */
	public static final int TRIP_STARTED = JSL.getNextEnumConstant();

	/**
	 * Indicates that a AbstractMover ended a trip
	 */
	public static final int TRIP_ENDED = JSL.getNextEnumConstant();;

	/**
	 * Indicates that a AbstractMover started a movement
	 */
	public static final int MOVE_STARTED = JSL.getNextEnumConstant();

	/**
	 * Indicates that a AbstractMover ended a movement
	 */
	public static final int MOVE_ENDED = JSL.getNextEnumConstant();

	// ===========================================
	// CLASS ATTRIBUTES
	// ===========================================
	
	/**
	 * Controls the movement of this element
	 */
	protected MovementControllerIfc myMovementController;

	/**
	 * Handles the occurs of collisions
	 */
	protected CollisionHandlerIfc myCollisionHandler;

	/**
	 * Detects the occurence of collisions
	 */
	protected CollisionDetectorIfc myCollisionDetector;

	/**
	 * Detects the occurence of collisions
	 */
	protected OutsideSpatialModelHandlerIfc myOSMHandler;

	/**
	 * Holds the spatial element of the destination for the current trip
	 * (movement) if given
	 */
	protected SpatialElementIfc mySEDestination;

	/** A temporary vector to hold the position.
	 */
	protected Vector3D myPosition;

	/**
	 * Holds the x,y coordinates of the destination for the current trip
	 * (movement)
	 */
	protected Vector3D myDestination;

	/**
	 * Holds the x,y coordinates of the end position associated with the current
	 * movement
	 */
	protected Vector3D myNextPosition;

	/**
	 * This Vector2D is used to hold the x and y components of the current
	 * direction of travel
	 */
	protected Vector3D myDirection;

	/**
	 * Used to indicate the position of the moving element at a projected time t
	 */
	protected Vector3D myFuturePosition;

	/** Represents the inherent velocity of the mover.
	 *  Used to set the movement velocity if no movement controller is present
	 */
	protected RandomVariable myVelocity;
	
	/**
	 * Represents the velocity (speed) in the current direction of travel for
	 * the current movement
	 */
	protected double myMovementVelocity = 1.0;

	/**
	 * Represents the distance in the current direction of travel for the
	 * current movement
	 */
	protected double myMovementDistance;

	/**
	 * The time that the current movement started
	 */
	protected double myMovementStartTime;

	/**
	 * The length of time expected for the current movement
	 */
	protected double myMovementTime;

	/**
	 * The event that represents the currently scheduled movement
	 */
	protected JSLEvent myCurrentMovementEvent;

	/**
	 * ActionListener that handles the end of movement event
	 */
	protected EndMovementListener myEndMovementListener;

	/**
	 * A flag that indicates that the object is moving
	 */
	protected boolean myMovingFlag = false;

	/**
	 * A flag that indicates that the object is on a trip
	 */
	protected boolean myTripFlag = false;

	/**
	 * Used and reused to hold information about the next collision
	 */
	protected Collision myCollision;

	// ===========================================
	// CONSTRUCTORS
	// ===========================================

	/** Creates a AbstractMover with the default position within its spatial model.
	 *  The spatial model of the parent is used as the spatial model of this object.  If the parent
	 *  does not have a spatial model (i.e. getSpatialModel() == null), then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 */
	public AbstractMover(ModelElement parent){
		this (parent, null, null, null);
	}
	
	/** Creates a AbstractMover with the default position within its spatial model.
	 *  The spatial model of the parent is used as the spatial model of this object.  If the parent
	 *  does not have a spatial model (i.e. getSpatialModel() == null), then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param name
	 */
	public AbstractMover(ModelElement parent, String name){
		this (parent, name, null, null);
	}
	
	/** Creates a AbstractMover at the supplied coordinate within its spatial model.
	 *  The spatial model of the parent is used as the spatial model of this object.  If the parent
	 *  does not have a spatial model (i.e. getSpatialModel() == null), then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param coordinate
	 */
	public AbstractMover(ModelElement parent, CoordinateIfc coordinate){
		this (parent, null, null, coordinate);
	}

	/** Creates a AbstractMover at the same coordinates as the supplied SpatialElementIfc within its spatial model.
	 *  The spatial model of the parent is used as the spatial model of this object.  If the parent
	 *  does not have a spatial model (i.e. getSpatialModel() == null), then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param position
	 */
	public AbstractMover(ModelElement parent, SpatialElementIfc element){
		this (parent, null, null, element.getCurrentPosition());
	}

	/** Creates a AbstractMover at the same coordinates as the supplied SpatialElementIfc within its spatial model.
	 *  The spatial model of the parent is used as the spatial model of this object.  If the parent
	 *  does not have a spatial model (i.e. getSpatialModel() == null), then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param position
	 * @param name
	 */
	public AbstractMover(ModelElement parent, SpatialElementIfc element, String name){
		this (parent, name, null, element.getCurrentPosition());
	}
	
	/** Creates a AbstractMover with the default position within the given spatial model.
	 *  If the supplied spatial model is null the spatial model of the parent is used as the spatial model of this object.  If the parent
	 *  does not have a spatial model (i.e. getSpatialModel() == null), then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param name
	 * @param spatialModel
	 */
	public AbstractMover(ModelElement parent, String name, SpatialModel spatialModel){
		this (parent, name, spatialModel, null);
	}

	/** Creates a AbstractMover with the default position within the given spatial model.
	 *  If the supplied spatial model is null the spatial model of the parent is used as the spatial model of this object.  If the parent
	 *  does not have a spatial model (i.e. getSpatialModel() == null), then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param spatialModel
	 */
	public AbstractMover(ModelElement parent, SpatialModel spatialModel){
		this (parent, null, spatialModel, null);
	}
	
	/** Creates a AbstractMover at the supplied coordinates within the given spatial model.
	 *  If the supplied spatial model is null the spatial model of the parent is used as the spatial model of this object.  If the parent
	 *  does not have a spatial model (i.e. getSpatialModel() == null), then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param name
	 * @param spatialModel
	 * @param coordinate Must be valid for the spatial model
	 */
	public AbstractMover(ModelElement parent, String name, SpatialModel spatialModel, CoordinateIfc coordinate){
		super(parent, name, spatialModel, coordinate);
		
		myVelocity = new RandomVariable(this, Constant.ONE);
		myPosition = new Vector3D();
		myDirection = new Vector3D();
		myDestination = new Vector3D();
		myNextPosition = new Vector3D();
		myFuturePosition = new Vector3D();
		myCollision = new Collision(this);
		myEndMovementListener = new EndMovementListener();

	}
	
	// ===========================================
	// PUBLIC METHODS
	// ===========================================

	// ===========================================
	// MoverIfc methods
	// ===========================================
	

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getFuturePosition(double)
	 */
	public final CoordinateIfc getFuturePosition(double time) {
		if (time < 0.0)
			throw new IllegalArgumentException("The time was less than zero.");
		// clear the future position
		myFuturePosition.setCoordinates(0.0, 0.0, 0.0);
		// set the direction of travel, direction is a unit vector
		myFuturePosition.setCoordinates(myDirection);
		// set the distance, v*t, in the direction of travel
		myFuturePosition.multiply(getMovementVelocity() * time);
		// add the current position to get the future position
		CoordinateIfc c = getCurrentPosition();
		myFuturePosition.add(c);
		return (myFuturePosition);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#isMoving()
	 */
	public final boolean isMoving() {
		return (myMovingFlag);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#isOnTrip()
	 */
	public final boolean isOnTrip() {
		return (myTripFlag);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getMovementDistance()
	 */
	public final double getMovementDistance() {
		return myMovementDistance;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getMovementTime()
	 */
	public final double getMovementTime() {
		return myMovementTime;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getMovementStartTime()
	 */
	public final double getMovementStartTime() {
		return myMovementStartTime;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getMovementVelocity()
	 */
	public final double getMovementVelocity() {
		return myMovementVelocity;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getCollisionDetector()
	 */
	public final CollisionDetectorIfc getCollisionDetector() {
		return myCollisionDetector;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setCollisionDetector(jsl.modeling.elements.spatial.CollisionDetectorIfc)
	 */
	public final void setCollisionDetector(CollisionDetectorIfc collisionDetector) {
		myCollisionDetector = collisionDetector;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getCollisionHandler()
	 */
	public final CollisionHandlerIfc getCollisionHandler() {
		return myCollisionHandler;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setCollisionHandler(jsl.modeling.elements.spatial.CollisionHandlerIfc)
	 */
	public final void setCollisionHandler(CollisionHandlerIfc collisionHandler) {
		myCollisionHandler = collisionHandler;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getMovementController()
	 */
	public final MovementControllerIfc getMovementController() {
		return myMovementController;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setMovementController(jsl.modeling.elements.spatial.MovementControllerIfc)
	 */
	public final void setMovementController(
			MovementControllerIfc movementController) {
		myMovementController = movementController;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setVelocity(jsl.utilities.random.DistributionIfc)
	 */
	public final void setVelocityInitialRandomSource(RandomIfc source){
		myVelocity.setInitialRandomSource(source);
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setVelocityRandomSource(jsl.utilities.random.RandomIfc)
	 */
	public void setVelocityRandomSource(RandomIfc source) {
		myVelocity.setRandomSource(source);	
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getOSMHandler()
	 */
	public final OutsideSpatialModelHandlerIfc getOSMHandler() {
		return myOSMHandler;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setOSMHandler(jsl.modeling.elements.spatial.OutsideSpatialModelHandlerIfc)
	 */
	public final void setOSMHandler(OutsideSpatialModelHandlerIfc OSMHandler) {
		myOSMHandler = OSMHandler;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setMovement(double, jsl.modeling.elements.spatial.CoordinateIfc)
	 */
	public final void setMovement(double velocity, CoordinateIfc position) {
		if (velocity <= 0.0)
			throw new IllegalArgumentException("The velocity was <= 0");

		// remember the next position
		myNextPosition.setCoordinates(position);

		// set the velocity of the move
		myMovementVelocity = velocity;
		
		// get the current position as a vector
		CoordinateIfc c = getCurrentPosition();
		myPosition.setCoordinates(c);
		
		// determine the distance associated with the move
		myMovementDistance = distanceTo(myNextPosition);

		// the vector from A to B can be obtained by
		// substracting the position vector a of point A from
		// the position vector b of point B
		// set direction vector equal to next position's coordinates
		myDirection.setCoordinates(myNextPosition);
		
		// subtract the current position, resulting in a vector from A to B
		myDirection.subtract(myPosition);
		
		// normalize the vector for direction
		myDirection.normalize();
		
		// record the time that the movement will start
		myMovementStartTime = getTime();
		
		// compute the total time associated with the movement
		myMovementTime = myMovementDistance / myMovementVelocity;
	}

	public String toString() {
		String s1 = super.toString();

		String s2 = null;

		s2 = s2 + "Trip Destination Position: " + myDestination + "\n"
				+ "Element started trip: " + myTripFlag + "\n"
				+ "Element started moving: " + myMovingFlag + "\n"
				+ "Movement Direction: " + myDirection + "\n"
				+ "Movement Speed: " + myMovementVelocity + "\n"
				+ "Movement Distance: " + myMovementDistance + "\n"
				+ "Movement Start Time: " + myMovementStartTime + "\n"
				+ "Movement Time: " + myMovementTime + "\n"
				+ "End Movement Position: " + myNextPosition + "\n";

		return (s1 + s2);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getDestination()
	 */
	public CoordinateIfc getDestination() {
		return myDestination;
	}
	
	// ===========================================
	// PROTECTED METHODS
	// ===========================================
		

	/**
	 * Causes the element to travel from its current position to the coordinates
	 * specified. This starts a trip. A trip is a series of movements to move
	 * from the current position to the specified coordinates. At the beginning
	 * of a trip, trip start observers are notified. If the trip is broken down
	 * into movements, then each movement can have its own velocity, distance,
	 * and direction (as long as the final movement ends at the destination). If
	 * the destination is the same as the current position of the element then
	 * no trip is started and no movement occurs, i.e. nothing occurs
	 * 
	 * If the specified coordinates are not in the element's associated spatial
	 * model then outsideSpatialModelHandler() is called. The default action is
	 * to throw an exception but this can be overridden in
	 * outsideSpatialModelHandler().
	 * 
	 * @param destination
	 */
	protected final void moveTo(CoordinateIfc destination) {
		if (isOnTrip())
			throw new IllegalArgumentException(
					"The element is already on a trip.");

		if(!getSpatialModel().isValid(destination))
			outsideSpatialModelHandler(destination);
		else {
			// check to see if the destination is different than the current position
			if (!isPositionEqualTo(destination)){
				myDestination.setCoordinates(destination);
				if (!isTripCompleted()) {
					startTrip_();
				}				
			} else {// told to travel to same location, start and end the trip 
				myTripFlag = true;
				beforeTripStarts();
				notifyObservers(TRIP_STARTED);
				myTripFlag = false;
				afterTripEnds();
				notifyObservers(TRIP_ENDED);
				startNextTrip();				
			}
		}
	}
	
	/** Causes the element to move to the coordinates of the specified
	 *  SpatialElementIfc
	 * 
	 * @param destination
	 */
	protected final void moveTo(SpatialElementIfc destination){
		moveTo(destination.getCurrentPosition());
	}

	protected void initialize() {
		super.initialize();
		myTripFlag = false;
		myMovingFlag = false;
		myDirection.setCoordinates(0.0, 0.0, 0.0);
		myDestination.setCoordinates(0.0, 0.0, 0.0);
		myFuturePosition.setCoordinates(0.0, 0.0, 0.0);
		myCollision.clear();
	}

	/**
	 * Can be used by subclasses to invoke logic prior to the start of a trip
	 */
	protected void beforeTripStarts() {
	}

	/**
	 * Can be used by subclasses to invoke logic after a trip ends
	 */
	protected void afterTripEnds() {
	}

	/** Should be used by subclasses to initiate (schedule) the next
	 *  trip after ending the current trip.  This is called immediately
	 *  after afterTripEnds()
	 */
	protected void startNextTrip() {
	}

	/**
	 * Can be used by subclasses to invoke logic prior to the start of a
	 * movement
	 */
	protected void beforeMovementStarts() {
	}

	/**
	 * Can be used by subclasses to invoke logic after a movement ends
	 */
	protected void afterMovementEnds() {
	}

	/**
	 * This method is called if the destination is outside of the element's
	 * current spatial model. By default it throws an IllegalArgumentException
	 * unless the user supplies an object that implements the
	 * OutsideSpatialModelHandlerIfc. Alternatively subclasses can override
	 * this method and provide the behavior directly
	 * 
	 * @param coordinate
	 */
	protected void outsideSpatialModelHandler(CoordinateIfc coordinate) {
		if (myOSMHandler != null)
			myOSMHandler.handleOutsideSpatialModel(this, coordinate);
		else
			throw new IllegalArgumentException("The coordinate (x1= " + coordinate.getX1()
					+ ",x2= " + coordinate.getX2() + ",x3= " + coordinate.getX3() +") is not contained in the spatial model.");
	}

	/**
	 * This method is called at the end of a movement to update the position.
	 * The setCurrentPosition() method is used to set the current position to the next
	 * position attribute Overriding methods can specify other ways to set the
	 * position as long as they are consistent with the process implemented for
	 * the movements.
	 */
	protected void updatePosition() {
		setCurrentPosition(myNextPosition);
	}
	
	/**
	 * This method can be overridden by subclasses to compute the characteristcs
	 * of a movement. Alternatively the user can supply an object that
	 * implements the MovementControllerIfc By default the movement goes
	 * directly to the destination at the default mover velocity. It is important
	 * that subclasses utilize the setMovement() method to properly set the
	 * characteristics of the movement.
	 */
	protected void computeMovement(){
		setMovement(myVelocity.getValue(), getDestination());		
	}
	
	/**
	 * This method can be overridden by subclasses to provide collision
	 * detection. Alternatively the user can supply an object that implements
	 * the CollisionDetectorIfc By default there is no collision detection.
	 * When there is no collision detector this method always returns false.
	 * 
	 * @return False if no collision, true if collided
	 */
	protected boolean checkForCollision() {
		// clear the collision for reuse
		myCollision.clear();

		if (myCollisionDetector != null)
			return (myCollisionDetector.checkForCollision(myCollision));
		else
			return (false);
	}

	/**
	 * This method can be overridden by subclasses to provide collision
	 * handling. Alternatively the user can supply an object that implements the
	 * CollisionHandlerIfc By default there is no collision handling.
	 */
	protected void handleCollision() {
		CollisionHandlerIfc handler = myCollision.getCollisionHandler();

		if (handler != null) // first check the collision for a handler
			handler.handleCollision(myCollision);
		else { // no handler on the collision, see if moving element has a
				// handler
			if (myCollisionHandler != null)
				myCollisionHandler.handleCollision(myCollision);
			// no handler, do nothing
		}
	}

	// ===========================================
	// PRIVATE METHODS
	// ===========================================

	/**
	 * Calls the beforeTripStarts() method and notifies trip started observers
	 */
	private final void startTrip_() {
		myTripFlag = true;
		beforeTripStarts();
		notifyObservers(TRIP_STARTED);
		startMovement_();
	}

	/**
	 * Calls the afterTripEnds() method and notifies trip end observers
	 */
	private final void endTrip_() {
		myTripFlag = false;
		afterTripEnds();
		notifyObservers(TRIP_ENDED);
		startNextTrip();
	}

	/**
	 * Used to start a movement towards a destination computes the
	 * characteristics of the movement (distance, velocity, direction, end
	 * position, etc) check for collisions and handles if necessary, then
	 * schedules the movement. The beforeMovementStarts observers are called
	 * after scheduling the movement.
	 */
	private final void startMovement_() {

		computeMovement_();

		// check for collision using collision detector
		// if collision detected call collision handler
		if (checkForCollision())
			handleCollision();
		else {
			// if no collision, then just schedule the movement
			myCurrentMovementEvent = scheduleEvent(myEndMovementListener,
					myMovementTime, "EndMovement");
			myMovingFlag = true;
			beforeMovementStarts();
			notifyObservers(MOVE_STARTED);
		}
	}

	/** Used internally to compute the movement either by
	 *  using a supplied MovementController or by calling
	 *  computeMovement()
	 */
	private final void computeMovement_() {
		if (myMovementController != null)
			myMovementController.controlMovement(this);
		else {
			computeMovement();
		}
	}

	/**
	 * After a movement is completed this is used to continue movement towards
	 * the destination if the trip is not completed.
	 */
	private final void continueMovement_() {
		computeMovement_();

		// check for collision using collision detector
		// if collision detected call collision handler
		if (checkForCollision())
			handleCollision();
		else {
			// if no collision, then just schedule the movement
			rescheduleEvent(myCurrentMovementEvent, myMovementTime);
			myMovingFlag = true;
			beforeMovementStarts();
			notifyObservers(MOVE_STARTED);
		}
	}

	/**
	 * Represents the end of a scheduled movement. This method 1) updates the
	 * position of the element 2) notifys movement ended observers 4) checks if
	 * the trip is not completed continues the movements, notifying movement
	 * started observers else notifies trip ended observers
	 */
	private final void endMovement_() {
		updatePosition();
		myMovingFlag = false;
		afterMovementEnds();
		notifyObservers(MOVE_ENDED);

		if (!isTripCompleted()) {
			continueMovement_();
		} else
			endTrip_();
	}

	/**
	 * Checks to if the current position is the same as trip's destination
	 * 
	 * @return true if completed
	 */
	private final boolean isTripCompleted() {
		return (isPositionEqualTo(myDestination));
	}

	// ===========================================
	// INNER CLASSES
	// ===========================================

	private class EndMovementListener implements ActionListenerIfc {
		/**
		 * The warmup action.
		 * 
		 * @param event
		 *            The event that triggered the warmup.
		 */
		public void action(JSLEvent event) {
			endMovement_();
		}
	}

}
