/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.spatial;

import java.util.Observable;
import java.util.Observer;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.resource.Resource;
import jsl.modeling.elements.spatial.transporter.EmptyMoveCompletionIfc;
import jsl.modeling.elements.spatial.transporter.TransportCompletionIfc;
import jsl.modeling.elements.spatial.transporter.Transporter;
import jsl.utilities.random.RandomIfc;

/**
 *
 */
abstract public class MobileResource extends Resource implements MoverIfc {

	/** Used to model the movements
	 */
	private Transporter myTransporter;
	
	/** Listens for empty move completions
	 */
	private EmptyMoveCompletedListener myEmptyMoveCompletedListener = new EmptyMoveCompletedListener();
	
	/** Listens for transport completed events
	 */
	private TransportCompletedListener myTransportCompletedListener = new TransportCompletedListener();

	// ===========================================
	// CONSTRUCTORS
	// ===========================================

	/** Creates a MobileResource with capacity 1 at the default position in the spatial model. 
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 */
	public MobileResource(ModelElement parent){
		this (parent, 1, null, null, null);
	}

	/** Creates a MobileResource with capacity 1 at the default position in the spatial model. 
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param name
	 */
	public MobileResource(ModelElement parent, String name){
		this (parent, 1, name, null, null);
	}
	
	/** Creates a MobileResource with the given capacity at the default position in the spatial model. 
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 */
	public MobileResource(ModelElement parent, int capacity){
		this (parent, capacity, null, null, null);
	}
	
	/** Creates a MobileResource with the given capacity at the default position in the spatial model. 
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 * @param name
	 */
	public MobileResource(ModelElement parent, int capacity, String name){
		this (parent, capacity, name, null, null);
	}

	/** Creates a MobileResource with capacity 1 at the given coordinates.  
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param name
	 * @param position
	 */
	public MobileResource(ModelElement parent, String name, CoordinateIfc position){
		this (parent, 1, name, null, position);
	}
	
	/** Creates a MobileResource with capacity 1 at the given coordinates.  
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 * @param name
	 * @param position
	 */
	public MobileResource(ModelElement parent, int capacity, String name, CoordinateIfc position){
		this (parent, capacity, name, null, position);
	}

	/** Creates a MobileResource with capacity 1 and (x,y) position.  
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param name
	 * @param x
	 * @param y
	 */
	public MobileResource(ModelElement parent, String name, double x, double y){
		this (parent, 1, name, null, x, y, 0.0);
	}
	
	/** Creates a MobileResource with the given capacity and (x,y) position.  
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 * @param name
	 * @param x
	 * @param y
	 */
	public MobileResource(ModelElement parent, int capacity, String name, double x, double y){
		this (parent, capacity, name, null, x, y, 0.0);
	}	

	/** Creates a MobileResource with the given capacity at the coordinates of the given spatial element
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param position
	 */
	public MobileResource(ModelElement parent, SpatialElementIfc position){
		this (parent, 1, null, null, position.getCurrentPosition());
	}

	/** Creates a MobileResource with the given capacity at the coordinates of the given spatial element
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param position
	 */
	public MobileResource(ModelElement parent, SpatialElementIfc position, String name){
		this (parent, 1, name, null, position.getCurrentPosition());
	}
	
	/** Creates a MobileResource with the given capacity at the coordinates of the given spatial element
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 * @param position
	 */
	public MobileResource(ModelElement parent, int capacity, SpatialElementIfc position){
		this (parent, capacity, null, null, position.getCurrentPosition());
	}
	
	/** Creates a MobileResource with the given capacity at the given coordinates.  
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 * @param position
	 */
	public MobileResource(ModelElement parent, int capacity, CoordinateIfc position){
		this (parent, capacity, null, null, position);
	}
	
	/** Creates a MobileResource with the given capacity and (x,y) position.  
	 *  The spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 * @param x
	 * @param y
	 */
	public MobileResource(ModelElement parent, int capacity, double x, double y){
		this (parent, capacity, null, null, x, y, 0.0);
	}

	/** Creates a MobileResource with the given capacity at the default coordinates in the supplied spatial model.  
	 *  The supplied is used unless it is null, then the spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 * @param spatialModel
	 */
	public MobileResource(ModelElement parent, int capacity, SpatialModel spatialModel){
		this (parent, capacity, null, spatialModel, null);
	}

	/** Creates a MobileResource with the given capacity at the default coordinates in the supplied spatial model.  
	 *  The supplied is used unless it is null, then the spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 * @param name
	 * @param spatialModel
	 */
	public MobileResource(ModelElement parent, int capacity, String name, SpatialModel spatialModel){
		this (parent, capacity, name, spatialModel, null);
	}

	/** Creates a MobileResource with the given capacity at the (x,y,z) coordinates in the supplied spatial model.  
	 *  The supplied is used unless it is null, then the spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 * @param name
	 * @param spatialModel
	 * @param x
	 * @param y
	 * @param z
	 */
	public MobileResource(ModelElement parent, int capacity, String name, SpatialModel spatialModel, double x, double y, double z){
		this(parent, capacity, name, spatialModel, new Vector3D(x,y,z));
	}

	/** Creates a MobileResource with the given capacity at the supplied coordinates in the supplied spatial model.  
	 *  The supplied is used unless it is null, then the spatial model of the parent is used as the spatial model.  If the parent
	 *  does not have a spatial model, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param capacity
	 * @param name
	 * @param spatialModel
	 * @param coordinate
	 */
	public MobileResource(ModelElement parent, int capacity, String name, SpatialModel spatialModel, CoordinateIfc coordinate){
		super(parent, capacity, name);

		if (spatialModel == null){
			spatialModel = parent.getSpatialModel();
			if (spatialModel == null){
				throw new IllegalArgumentException("No spatial model is available!");
			} 
		}
		
		setSpatialModel(spatialModel); // set the model element's spatial model

		myTransporter = new Transporter(this, getName()+"Mover", spatialModel, coordinate);
		// allow the Moveable Resource to observe state changes in underlying transporter
		// so that they can be passed along to observers of the MobileResource
		myTransporter.addObserver(new TransporterObserver());
	
	}
	
	// ===========================================
	// PUBLIC METHODS
	// ===========================================


	// ===========================================
	// MoverIfc methods
	// ===========================================

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getCollisionDetector()
	 */
	public final CollisionDetectorIfc getCollisionDetector() {
		return myTransporter.getCollisionDetector();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getCollisionHandler()
	 */
	public final CollisionHandlerIfc getCollisionHandler() {
		return myTransporter.getCollisionHandler();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getDestination()
	 */
	public final CoordinateIfc getDestination() {
		return myTransporter.getDestination();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getFuturePosition(double)
	 */
	public final CoordinateIfc getFuturePosition(double time) {
		return myTransporter.getFuturePosition(time);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getMovementController()
	 */
	public final MovementControllerIfc getMovementController() {
		return myTransporter.getMovementController();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getMovementDistance()
	 */
	public final double getMovementDistance() {
		return myTransporter.getMovementDistance();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getMovementStartTime()
	 */
	public final double getMovementStartTime() {
		return myTransporter.getMovementStartTime();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getMovementTime()
	 */
	public final double getMovementTime() {
		return myTransporter.getMovementTime();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getMovementVelocity()
	 */
	public final double getMovementVelocity() {
		return myTransporter.getMovementVelocity();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#getOSMHandler()
	 */
	public final OutsideSpatialModelHandlerIfc getOSMHandler() {
		return myTransporter.getOSMHandler();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#isMoving()
	 */
	public final boolean isMoving() {
		return myTransporter.isMoving();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#isOnTrip()
	 */
	public final boolean isOnTrip() {
		return myTransporter.isOnTrip();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setCollisionDetector(jsl.modeling.elements.spatial.CollisionDetectorIfc)
	 */
	public final void setCollisionDetector(CollisionDetectorIfc collisionDetector) {
		myTransporter.setCollisionDetector(collisionDetector);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setCollisionHandler(jsl.modeling.elements.spatial.CollisionHandlerIfc)
	 */
	public final void setCollisionHandler(CollisionHandlerIfc collisionHandler) {
		myTransporter.setCollisionHandler(collisionHandler);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setMovement(double, jsl.modeling.elements.spatial.CoordinateIfc)
	 */
	public final void setMovement(double velocity, CoordinateIfc position) {
		myTransporter.setMovement(velocity, position);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setMovementController(jsl.modeling.elements.spatial.MovementControllerIfc)
	 */
	public final void setMovementController(MovementControllerIfc movementController) {
		myTransporter.setMovementController(movementController);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setOSMHandler(jsl.modeling.elements.spatial.OutsideSpatialModelHandlerIfc)
	 */
	public final void setOSMHandler(OutsideSpatialModelHandlerIfc OSMHandler) {
		myTransporter.setOSMHandler(OSMHandler);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.MoverIfc#setVelocity(jsl.utilities.random.DistributionIfc)
	 */
	public final void setVelocityInitialRandomSource(RandomIfc source) {
		myTransporter.setVelocityInitialRandomSource(source);
	}
	
	/**
	 * @param source
	 * @see jsl.modeling.elements.spatial.AbstractMover#setVelocityRandomSource(jsl.utilities.random.RandomIfc)
	 */
	public void setVelocityRandomSource(RandomIfc source) {
		myTransporter.setVelocityRandomSource(source);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#attachPositionObserver(java.util.Observer)
	 */
	public final void attachPositionObserver(Observer observer) {
		myTransporter.attachPositionObserver(observer);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#changeSpatialModel(jsl.modeling.elements.spatial.SpatialModel, jsl.modeling.elements.spatial.CoordinateIfc)
	 */
	public final void changeSpatialModel(SpatialModel spatialModel, CoordinateIfc coordinate) {
		myTransporter.changeSpatialModel(spatialModel, coordinate);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#distanceTo(jsl.modeling.elements.spatial.CoordinateIfc)
	 */
	public final double distanceTo(CoordinateIfc coordinate) {
		return myTransporter.distanceTo(coordinate);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#distanceTo(jsl.modeling.elements.spatial.SpatialElementIfc)
	 */
	public final double distanceTo(SpatialElementIfc element) {
		return myTransporter.distanceTo(element);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#getCurrentPosition()
	 */
	public final CoordinateIfc getCurrentPosition() {
		return myTransporter.getCurrentPosition();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#getInitialPosition()
	 */
	public final CoordinateIfc getInitialPosition() {
		return myTransporter.getInitialPosition();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialModelElement#setInitialPosition(jsl.modeling.elements.spatial.CoordinateIfc)
	 */
	public void setInitialPosition(CoordinateIfc coordinate) {
		myTransporter.setInitialPosition(coordinate);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#getInitialSpatialModel()
	 */
	public final SpatialModel getInitialSpatialModel() {
		return myTransporter.getInitialSpatialModel();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#getModelElement()
	 */
	public final ModelElement getModelElement() {
		return myTransporter.getModelElement();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#getPreviousPosition()
	 */
	public final CoordinateIfc getPreviousPosition() {
		return myTransporter.getPreviousPosition();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#initializeSpatialElement()
	 */
	public final void initializeSpatialElement() {
		myTransporter.initializeSpatialElement();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#isPositionEqualTo(jsl.modeling.elements.spatial.CoordinateIfc)
	 */
	public final boolean isPositionEqualTo(CoordinateIfc coordinate) {
		return myTransporter.isPositionEqualTo(coordinate);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#isPositionEqualTo(jsl.modeling.elements.spatial.SpatialElementIfc)
	 */
	public final boolean isPositionEqualTo(SpatialElementIfc element) {
		return myTransporter.isPositionEqualTo(element);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#removePositionObserver(java.util.Observer)
	 */
	public final void removePositionObserver(Observer observer) {
		myTransporter.removePositionObserver(observer);
	}
	
	// ===========================================
	// Protected methods
	// ===========================================

	protected void initialize(){
		super.initialize();
		// need to put the transporter back in the allocated state
		// transporter can only be freed, moveEmpty(), or transport()
		myTransporter.allocate();
	}
	
	/** Subclasses can use this to access the underlying Transporter
	 * 
	 * @return Returns the Transporter.
	 */
	protected final Transporter getTransporter() {
		return myTransporter;
	}
	
	/** Tells the resource to move empty to the supplied coordinates
	 * @param coordinate
	 */
	protected void moveEmpty(CoordinateIfc coordinate) {
		myTransporter.moveEmpty(coordinate, myEmptyMoveCompletedListener);
	}
	
	/** Tells the resource to move empty to the coordinates of the supplied spatial element
	 * @param element
	 */
	protected void moveEmpty(SpatialElementIfc element) {
		myTransporter.moveEmpty(element.getCurrentPosition(), myEmptyMoveCompletedListener);
	}
	
	/** Tells the resource to transport to the supplied coordinates
	 * @param coordinate
	 */
	protected void transport(CoordinateIfc coordinate) {
		myTransporter.transport(coordinate, myTransportCompletedListener);
	}

	/** Tells the resource to transport to the coordinates of the supplied spatial element
	 * @param element
	 */
	protected void transport(SpatialElementIfc element) {
		myTransporter.transport(element.getCurrentPosition(), myTransportCompletedListener);
	}
	
	/** Subclasses must override this method to react to the end of 
	 *  an empty move 
	 *
	 */
	abstract protected void emptyMoveCompleted();
	
	/** Subclasses must override this method to react to the end of
	 *  a transport
	 */
	abstract protected void transportCompleted();

	private class EmptyMoveCompletedListener implements EmptyMoveCompletionIfc {

		public void emptyMoveComplete(Transporter transporter) {
			emptyMoveCompleted();
		}
	}
	
	private class TransportCompletedListener implements TransportCompletionIfc {

		public void transportComplete(Transporter transporter) {
			transportCompleted();
		}
	}
	
	private class TransporterObserver implements Observer{

		public void update(Observable arg0, Object arg1) {
			
			// get the state of the transporter
			int state = myTransporter.getObserverState();
			// pass the notification along to observers of the MobileResource
			//if (state != ModelElement.INITIALIZED)
			if (state >= AbstractMover.TRIP_STARTED) //TODO this may be a problem
				MobileResource.this.notifyObservers(state);
			
		}
		
	}


}
