/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.spatial;

import java.awt.Shape;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Collections;

import jsl.utilities.math.JSLMath;

/**
 *
 */
public class RectangularGridSpatialModel2D extends SpatialModel {

	/** The upper left corner point for the grid
	 */
	private Point2D myUpperLeftCornerPt;
	
	/** The lower left corner point for the grid
	 */
	private Point2D myLowerLeftCornerPt;
	
	/** The upper right corner point for the grid
	 */
	private Point2D myUpperRightCornerPt;
	
	/** The lower right corner point for the grid
	 */
	private Point2D myLowerRightCornerPt;
	
	/** The line at the top of the grid
	 */
	private Line2D myTopLine;
	
	/** The line at the bottom of the grid
	 */
	private Line2D myBottomLine;
	
	/** The line at the right side of the grid
	 */
	private Line2D myRightLine;
	
	/** The line at the left side of the grid
	 */
	private Line2D myLeftLine;
	
	/** The width of the grid in user dimensions
	 */
	private double myWidth;
	
	/** The height of the grid in user dimensions
	 */
	private double myHeight;
	
	/** The number of rows in the grid, the grid
	 *  is zero based (the first row is the 0th row)
	 */
	private int myNumRows;
	
	/** The number of columns in the grid, the grid
	 *  is zero based (the first column is the 0th column)
	 */
	private int myNumCols;
	
	/** The width of the cells in the grid
	 */
	private double myCellWidth;
	
	/** The height of the cells in the grid
	 */
	private double myCellHeight;
	
	/** An 2-d array of points forming the grid,
	 *  point[0][0] = left upper corner point
	 */
	private Point2D[][] myPoints;
	
	/** An 2-d array of the horizontal line segments in the grid
	 *  hline[0][0] = point[0][0] --> point[0][1] and so forth
	 */
	private Line2D[][] myHorzLines;
	
	/** An 2-d array of the vertical line segments in the grid
	 *  vline[0][0] = point[0][0] --> point[1][0] and so forth
	 */
	private Line2D[][] myVertLines;
	
	/** An 2-d array of the cells forming the grid
	 *  cell[0][0] = upper left most cell with left corner point point[0][0]
	 */
	private RectangularCell2D[][] myCells;
		
	/** A List of the cells in the grid formed row by row 
	 */
	private List<RectangularCell2D> myCellList;

	/** A referecne to the rectangle that forms the outer edge of 
	 *  the grid
	 */
	private Rectangle2D myOuterRectangle;
	
	/** A reference to the path that forms the grid including all line segments
	 */
	private GeneralPath myPath;

	/** Creates a grid in the 2D plane.  The grid is based on the standard
	 *  user coordinate system, with (x,y) = (0,0) being the upper left most
	 *  corner point, with the x-axis going from left to right and the
	 *  y-axis going from the top down
	 *  Default values
	 *  name = use default assigned name
	 *  x = 0.0 The x coordinate of the upper left most corner point
	 *  y = 0.0 The y coordinate of the upper left most corner point
	 *  width = Double.MAX_VALUE The width (along the x-axis) of the grid
	 *  height = Double.MAX_VALUE The height (along the y-axis) of the grid
	 *  numRows = 1 The number of rows in the grid (0-based)
	 *  numCols = 1 The number of columns in the grid (0-based)
	 * 
	 */
	public RectangularGridSpatialModel2D() {
		this(null, 0.0, 0.0, Double.MAX_VALUE, Double.MAX_VALUE, 1, 1);
	}

	/** Creates a grid in the 2D plane.  The grid is based on the standard
	 *  user coordinate system, with (x,y) = (0,0) being the upper left most
	 *  corner point, with the x-axis going from left to right and the
	 *  y-axis going from the top down
	 *  Default values
	 *  x = 0.0 The x coordinate of the upper left most corner point
	 *  y = 0.0 The y coordinate of the upper left most corner point
	 *  width = Double.MAX_VALUE The width (along the x-axis) of the grid
	 *  height = Double.MAX_VALUE The height (along the y-axis) of the grid
	 *  numRows = 1 The number of rows in the grid (0-based)
	 *  numCols = 1 The number of columns in the grid (0-based)
	 * 
	 * @param name The name of the spatial model
	 */
	public RectangularGridSpatialModel2D(String name) {
		this(name, 0.0, 0.0, Double.MAX_VALUE, Double.MAX_VALUE, 1, 1);
	}

	/** Creates a grid in the 2D plane.  The grid is based on the standard
	 *  user coordinate system, with (x,y) = (0,0) being the upper left most
	 *  corner point, with the x-axis going from left to right and the
	 *  y-axis going from the top down

	 *  Default values
	 *  name = use default assigned name
	 *  x = 0.0 The x coordinate of the upper left most corner point
	 *  y = 0.0 The y coordinate of the upper left most corner point

	 * @param width The width (along the x-axis) of the grid
	 * @param height The height (along the y-axis) of the grid
	 * @param numRows The number of rows in the grid (0-based)
	 * @param numCols The number of columns in the grid (0-based)
	 */
	public RectangularGridSpatialModel2D(double width, double height, int numRows, int numCols) {
		this(null, 0.0, 0.0, width, height, numRows, numCols);
	}
	
	/** Creates a grid in the 2D plane.  The grid is based on the standard
	 *  user coordinate system, with (x,y) = (0,0) being the upper left most
	 *  corner point, with the x-axis going from left to right and the
	 *  y-axis going from the top down
	 * 
	 * @param x The x coordinate of the upper left most corner point
	 * @param y The y coordinate of the upper left most corner point
	 * @param width The width (along the x-axis) of the grid
	 * @param height The height (along the y-axis) of the grid
	 * @param numRows The number of rows in the grid (0-based)
	 * @param numCols The number of columns in the grid (0-based)
	 */
	public RectangularGridSpatialModel2D(double x, double y, double width, double height, int numRows, int numCols) {
		this(null, x, y, width, height, numRows, numCols);
	}

	/** Creates a grid in the 2D plane.  The grid is based on the standard
	 *  user coordinate system, with (x,y) = (0,0) being the upper left most
	 *  corner point, with the x-axis going from left to right and the
	 *  y-axis going from the top down
	 * 
	 * @param name The name of the spatial model
	 * @param x The x coordinate of the upper left most corner point
	 * @param y The y coordinate of the upper left most corner point
	 * @param width The width (along the x-axis) of the grid
	 * @param height The height (along the y-axis) of the grid
	 * @param numRows The number of rows in the grid (0-based)
	 * @param numCols The number of columns in the grid (0-based)
	 */
	public RectangularGridSpatialModel2D(String name, double x, double y, double width, double height, int numRows, int numCols) {
		super(name);
		setWidth(width);
		setHeight(height);
		setGrid(x, y, numRows, numCols);
		myOuterRectangle = new Rectangle2D.Double(x, y, width, height);
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.spatial2D.SpatialModel2D#contains(double, double)
	 */
	public final boolean contains(double x, double y) {
		return myOuterRectangle.contains(x,y);
	}
	
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.spatial2D.SpatialModel2D#getShape()
	 */
	public final Shape getShape() {
		return myPath;
	}
	
	/**
	 * @return Returns the delta.
	 */
	public final double getDeltaX() {
		return myCellWidth;
	}

	/**
	 * @return Returns the delta.
	 */
	public final double getDeltaY() {
		return myCellHeight;
	}

	/**
	 * @return Returns the number of columns.
	 */
	public final int getNumColumns() {
		return myNumCols;
	}
	
	/**
	 * @return Returns the number of rows
	 */
	public final int getNumRows() {
		return myNumRows;
	}

	/**
	 * @return Returns the height.
	 */
	public final double getHeight() {
		return myHeight;
	}
		
	/**
	 * @return Returns the width.
	 */
	public final double getWidth() {
		return myWidth;
	}

	public String toString(){
		StringBuffer s = new StringBuffer();

		s.append("Grid \n");
		s.append("width = " + myWidth + " height = " + myHeight + "\n");
		s.append("ULPT = " + myUpperLeftCornerPt + "\n");
		s.append("URPT = " + myUpperRightCornerPt + "\n");
		s.append("LRPT = " + myLowerRightCornerPt + "\n");
		s.append("LLPT = " + myLowerLeftCornerPt + "\n");
		s.append("TopLine = " + myTopLine.getP1() + "---->" + myTopLine.getP2() + "\n");
		s.append("RightLine = " + myRightLine.getP1() + "---->" + myRightLine.getP2() + "\n");
		s.append("BottomLine = " + myBottomLine.getP1() + "---->" + myBottomLine.getP2() + "\n");
		s.append("LeftLine = " + myLeftLine.getP1() + "---->" + myLeftLine.getP2() + "\n");

		s.append("Points \n");
		for (int i=0;i <= myNumRows; i++){
			for(int j=0; j <= myNumCols; j++){
				s.append("Point[i=" + i + "][j=" + j + "]= " + myPoints[i][j] + "\n");
			}
		}

		s.append("Horizontal lines \n");
		for(int i = 0; i < myNumRows+1; i++){
			for(int j = 0; j < myNumCols; j++){
				s.append(myHorzLines[i][j].getP1() + "---->" + myHorzLines[i][j].getP2() + "\n");
			}			
		}

		s.append("Vertical lines \n");
		for(int j = 0; j < myNumCols+1; j++){
			for(int i = 0; i < myNumRows; i++){
				s.append(myVertLines[i][j].getP1() + "---->" + myVertLines[i][j].getP2() + "\n");
			}			
		}

		s.append("Cells \n");

		for(int i = 0; i < myNumRows; i++){
			for(int j = 0; j < myNumCols; j++){
				s.append(myCells[i][j] + "\n");		
			}			
		}

		return(s.toString());
	}

	/** Returns the cell that the element is in or null
	 * 
	 * @param element
	 * @return
	 */
	public final RectangularCell2D getCell(SpatialElement element){
		if (contains(element)){
			CoordinateIfc c = element.getCurrentPosition();
			double x = c.getX1();
			double y = c.getX2();
			return(getCell(x, y));
		}
		else
			return(null);
	}

	/** The cell that contains this x,y coordinate or null if no cell
	 * 
	 * @param x
	 * @param y
	 * @return
	 */
	public final RectangularCell2D getCell(double x, double y){
		if (!contains(x,y))
			return null;
		int col = (int)(x/myCellWidth);
		int row = (int)(y/myCellHeight);
		return getCell(row, col);
	}
	
	/** The cell at this row, col
	 * 
	 * @param row
	 * @param col
	 * @return
	 */public final RectangularCell2D getCell(int row, int col){
	
		if ((row < 0) || (row >= myNumRows))
			return null;

		if ((col < 0) || (col >= myNumCols))
			return null;

		return(myCells[row][col]);
	}
	
	/** An iterator to the cells in the grid.  The cells are
	 *  accesses by rows (row, col): (0,0), then (0,1), etc
	 *  0th row first,
	 * @return
	 */
	public final Iterator<RectangularCell2D> getCellIterator(){
		return(myCellList.iterator());
	}

	/** Returns an array containing the 1st Moore neighborhood for the cell in the grid
	 * 
	 * @param row
	 * @param col
	 * @return
	 */
	public final RectangularCell2D[][] getMooreNeighborhood(RectangularCell2D coreCell){
		RectangularCell2D[][] neighborhood = new RectangularCell2D[3][3];
		getMooreNeighborhood(coreCell, neighborhood);
		return(neighborhood);
	}

	/** Returns an array containing the 1st Moore neighborhood for the cell at
	 *  row, col in the grid
	 * 
	 * @param row
	 * @param col
	 * @return
	 */
	public final RectangularCell2D[][] getMooreNeighborhood(int row, int col){
		RectangularCell2D[][] neighborhood = new RectangularCell2D[3][3];
		getMooreNeighborhood(row, col, neighborhood);
		return(neighborhood);
	}
	
	/** Fills the supplied array with the 1st order Moore neighborhood for the
	 * given cell at row, col of the grid
	 * 
	 * @param row
	 * @param col
	 * @param neighborhood
	 */
	public final void getMooreNeighborhood(int row, int col, RectangularCell2D[][] neighborhood){
		getMooreNeighborhood(getCell(row,col), neighborhood);
	}

	/** Returns a default set of coordinates to be used to initialize the location of spatial elements
	 *  if necessary.
	 * 
	 * @return
	 */
	public CoordinateIfc getDefaultCoordinate(){
		return(new Vector3D());
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialModel#getCoordinate(double, double, double)
	 */
	@Override
	public CoordinateIfc getCoordinate(double x1, double x2, double x3) {
		if (!myOuterRectangle.contains(x1, x2))
			throw new IllegalArgumentException("The coordinate is not valid for this spatial model!");
			
		return new Vector3D(x1, x2, x3);
	}

	/** This only checks whether 1st and 2nd (x,y) coordinates are valid, since
	 *  this is only rectangluar 2D grid. The coordinate is valid if it is within
	 *  the boundary defined by the rectangle for the grid
	 *  
	 * @see jsl.spatial.SpatialModel#isValid(jsl.spatial.Coordinate)
	 */
	public boolean isValid(CoordinateIfc coordinate) {
		double x = coordinate.getX1();
		double y = coordinate.getX2();
		return myOuterRectangle.contains(x,y);
	}

	/** This is the Euclidean 2D distance between the coordinates.
	 * 
	 * @see jsl.spatial.SpatialModel#distance(jsl.spatial.Coordinate, jsl.spatial.Coordinate)
	 */
	public double distance(CoordinateIfc fromCoordinate, CoordinateIfc toCoordinate) {
		if (!isValid(fromCoordinate))
			throw new IllegalArgumentException("The from coordinate is not valid for this spatial model!");
		
		if (!isValid(toCoordinate))
			throw new IllegalArgumentException("The to coordinate is not valid for this spatial model!");
		
		double x1 = fromCoordinate.getX1();
		double y1 = fromCoordinate.getX2();
		double x2 = toCoordinate.getX1();
		double y2 = toCoordinate.getX2();
		double dx = x1-x2;
		double dy = y1-y2;
		double d = Math.sqrt(dx*dx + dy*dy);
		return d;
	}
	
	/** Only checks 1st and 2nd coordinates, because of 2D grid
	 * 
	 * @see jsl.spatial.SpatialModel#comparePositions(jsl.spatial.Coordinate, jsl.spatial.Coordinate)
	 */
	public boolean comparePositions(CoordinateIfc coordinate1, CoordinateIfc coordinate2) {
		if (!isValid(coordinate1))
			throw new IllegalArgumentException("The coordinate 1 is not valid for this spatial model!");
		
		if (!isValid(coordinate2))
			throw new IllegalArgumentException("The coordinate 2 is not valid for this spatial model!");

		double x1 = coordinate1.getX1();
		double y1 = coordinate1.getX2();
		double x2 = coordinate2.getX1();
		double y2 = coordinate2.getX2();
		boolean b1 = JSLMath.equal(x1, x2, myDefaultPositionPrecision);
		boolean b2 = JSLMath.equal(y1, y2, myDefaultPositionPrecision);
		return(b1 && b2);
	}
	
	/** Fills the supplied array with the 1st order Moore neighborhood for the
	 * given core cell. 
	 * 
	 * set the top row of the neighborhood 
	 * neighborhood[0][0] = getCell(i-1, j-1) 
	 * neighborhood[0][1] = getCell(i-1, j)
	 * neighborhood[0][2] = getCell(i-1, j+1)
	 * 
	 * set the middle row of the neighborhood 
	 * neighborhood[1][0] = getCell(i, j-1) 
	 * neighborhood[1][1] = getCell(i, j) 
	 * neighborhood[1][2] = getCell(i, j+1)
	 * 
	 * set the bottom row of the neighborhood
	 * neighborhood[2][0] = getCell(i+1,j-1) 
	 * neighborhood[2][1] = getCell(i+1, j)
	 * neighborhood[2][2] = getCell(i+1, j+1)
	 * 
	 * @param coreCell
	 * @param neighborhood
	 */
	public final void getMooreNeighborhood(RectangularCell2D coreCell, RectangularCell2D[][] neighborhood){
		if (neighborhood == null)
			throw new IllegalArgumentException("The neighborhood array was null");
		
		if (neighborhood.length < 3)
			throw new IllegalArgumentException("Row size of the array must be 3.");

		for (int i=0; i<neighborhood.length; i++){
			if (neighborhood[i].length < 3)
				throw new IllegalArgumentException("Column size of the array must be 3.");				
		}
		
		if (coreCell == null)
			throw new IllegalArgumentException("The core cell was null");
		
		if (coreCell.getParentRectangularGrid2D() != this)
			throw new IllegalArgumentException("The core cell is not part of this grid.");
		
		// nullify the neighborhood
		for (int i=0;i<3;i++){
			for(int j=0;j<3;j++){
				neighborhood[i][j] = null;
			}
		}
		
		// get the core cell's indices
		int i = coreCell.getRowIndex();
		int j = coreCell.getColumnIndex();

		// set the top row of the neighborhood
		neighborhood[0][0] = getCell(i-1, j-1); 
		neighborhood[0][1] = getCell(i-1, j);
		neighborhood[0][2] = getCell(i-1, j+1);
		
		// set the middle row of the neighborhood
		neighborhood[1][0] = getCell(i, j-1); 
		neighborhood[1][1] = getCell(i, j);
		neighborhood[1][2] = getCell(i, j+1);

		// set the bottom row of the neighborhood
		neighborhood[2][0] = getCell(i+1, j-1); 
		neighborhood[2][1] = getCell(i+1, j);
		neighborhood[2][2] = getCell(i+1, j+1);

	}

	//  ===========================================
	//     PROTECTED METHODS
	//  ===========================================
	
	/** Adds the spatial element to the spatial model
	 * @param element The spatial element to be added
	 */
	protected void addSpatialElement(SpatialElement element){
		super.addSpatialElement(element);
		CoordinateIfc c = element.getCurrentPosition();
		double x = c.getX1();
		double y = c.getX2();

		RectangularCell2D cell = getCell(x,y);
		cell.addSpatialElement(element);
	}
	
	/** Removes the spatial element from the spatial model
	 * @param element
	 */
	protected boolean removeSpatialElement(SpatialElement element){
		boolean found = super.removeSpatialElement(element);
		if (found){
			CoordinateIfc c = element.getCurrentPosition();
			double x = c.getX1();
			double y = c.getX2();
			RectangularCell2D cell = getCell(x,y);
			cell.removeSpatialElement(element);
		}
		return(found);
	}
	
	/** When the spatial element's position is changed then this
	 *  method is called to ensure that the spatial model properly
	 *  tracks the elements position
	 * @param element The SpatialElement2D who's position changed
	 */
	protected final void updatePosition(){
		SpatialElementIfc element = this.getUpdatingSpatialElement();
		
		CoordinateIfc p = element.getPreviousPosition();
		double xp = p.getX1();
		double yp = p.getX2();

		CoordinateIfc c = element.getCurrentPosition();
		double xc = c.getX1();
		double yc = c.getX2();

		// The element's position has just changed
		// get the previous position's cell
		
		RectangularCell2D previousCell = getCell(xp, yp);
		// get the cell it is supposed to be in now
		RectangularCell2D nextCell = getCell(xc, yc);
		if (previousCell != nextCell) {// change in cell
			// remove the element from the previous cell
			previousCell.removeSpatialElement(element);
			// add it to the next cell
			nextCell.addSpatialElement(element);
		}

	}
	
	/**
	 * @param height The height to set.
	 */
	protected final void setHeight(double height) {
		if (height <= 0)
			throw new IllegalArgumentException("Height must be > 0");
		myHeight = height;
	}
	
	/**
	 * @param width The width to set.
	 */
	protected final void setWidth(double width) {
		if (width <= 0)
			throw new IllegalArgumentException("Width must be > 0");
		myWidth = width;
	}
	
	/** Sets up the rectangular grid into points, lines, and rectangle
	 *  Dividing the grid into the appropriate number of rows and columns.
	 * 
	 * @param startX
	 * @param startY
	 * @param numRows
	 * @param numCols
	 */
	protected final void setGrid(double startX, double startY, int numRows, int numCols){
		if (numRows < 1)
			throw new IllegalArgumentException("The number of rows must be >=1");
		
		if (numCols < 1)
			throw new IllegalArgumentException("The number of rows must be >=1");
		
		if (startX < 0)
			throw new IllegalArgumentException("x must be >= 0");
		
		if (startY < 0)
			throw new IllegalArgumentException("y must be >= 0");
		
		myNumRows = numRows;
		myNumCols = numCols;
		myCellWidth = myWidth/numCols;
		myCellHeight = myHeight/numRows;
		
		// make the points
		myPoints = new Point2D.Double[myNumRows+1][myNumCols+1];	
		double x = 0.0;
		double y = startY;
		for (int i=0;i <= myNumRows; i++){
			x = startX;
			for(int j=0; j <= myNumCols; j++){
				myPoints[i][j] = new Point2D.Double(x,y);
				x = x + myCellWidth;
			}
			y = y + myCellHeight;
		}
		
		// set the outer corner points
		myUpperLeftCornerPt = myPoints[0][0];
		myUpperRightCornerPt = myPoints[0][myNumCols];
		myLowerLeftCornerPt = myPoints[myNumRows][0];
		myLowerRightCornerPt = myPoints[myNumRows][myNumCols];	

		// set the outer lines
		myTopLine = new Line2D.Double(myUpperLeftCornerPt, myUpperRightCornerPt);
		myBottomLine = new Line2D.Double(myLowerLeftCornerPt, myLowerRightCornerPt);
		myLeftLine = new Line2D.Double(myUpperLeftCornerPt, myLowerLeftCornerPt);
		myRightLine = new Line2D.Double(myUpperRightCornerPt, myLowerRightCornerPt);
		
		// make the general path to be built from the points/lines
		myPath = new GeneralPath();
		// set the intial point on the path
		myPath.moveTo((float)myUpperLeftCornerPt.getX(), (float)myUpperLeftCornerPt.getY());
		
		// make the horizontal line segments
		myHorzLines = new Line2D[myNumRows+1][myNumCols];

		for(int i = 0; i < myNumRows+1; i++){
			for(int j = 0; j < myNumCols; j++){
				myHorzLines[i][j] = new Line2D.Double(myPoints[i][j], myPoints[i][j+1]);
				myPath.append(myHorzLines[i][j], false);
			}			
		}

		// make the vertical line segments
		myVertLines = new Line2D[myNumRows][myNumCols+1];

		for(int j = 0; j < myNumCols+1; j++){
			for(int i = 0; i < myNumRows; i++){
				myVertLines[i][j] = new Line2D.Double(myPoints[i][j], myPoints[i+1][j]);
				myPath.append(myVertLines[i][j], false);
			}			
		}
		
		myPath.closePath();

		// make the cells from the points and lines

		myCells = new RectangularCell2D[myNumRows][myNumCols];
		myCellList = new ArrayList<RectangularCell2D>(myNumRows*myNumCols);

		for(int i = 0; i < myNumRows; i++){
			for(int j = 0; j < myNumCols; j++){
				myCells[i][j] = new RectangularCell2D(i,j);
				myCellList.add(myCells[i][j]);
			}			
		}		
	}

	//  ===========================================
	//     INNER CLASSES
	//  ===========================================

	public class RectangularCell2D {
				
		private int myRow;
		
		private int myCol;
		
		private double myWidth;
		
		private double myHeight;
						
		private Rectangle2D myRectangle;		
						
		/* A cell must be inside a grid, this provides a reference to the
		 * cell's containing grid
		 */
		private RectangularGridSpatialModel2D myParentRectangularGrid2D;

		private List<SpatialElementIfc> mySpatialElements;

		private RectangularCell2D(int row, int col){
			myParentRectangularGrid2D = RectangularGridSpatialModel2D.this;
			myRow = row;
			myCol = col;
			myWidth = myCellWidth;
			myHeight = myCellHeight;
			// create and set the rectangle
			myRectangle = new Rectangle2D.Double(myPoints[row][col].getX(), myPoints[row][col].getY(), myWidth, myHeight);
			mySpatialElements = new LinkedList<SpatialElementIfc>();
		}
		
		/** Returns an unmodifiable view of the spatial elements in this cell
		 *  The list is unmodfiable, i.e. you can't change the list, but you can
		 *  still change the elements in the list. WARNING:  Don't change the elements 
	     *  unless you really know what you are doing.
		 * @return
		 */
		public final List<SpatialElementIfc> getUnmodifiableSpatialElements(){
			return (Collections.unmodifiableList(mySpatialElements));
		}
		
		/** Returns an array containing the cells associated with the 1st Moore
		 *  neighborhood for the cell.
		 * 
		 * @param neighborhood
		 */
		public final void getMooreNeighborhood(RectangularCell2D[][] neighborhood){
			myParentRectangularGrid2D.getMooreNeighborhood(this, neighborhood);
		}
 						
		/** Converts the cell to a string
		 * @return
		 */
		public final String toString(){
			StringBuffer s = new StringBuffer();
			
			s.append("Cell column = " + myCol + "\n");
			s.append("Cell row = " + myRow + "\n");
			s.append("width = " + myWidth + "\n");
			s.append("height = " + myHeight + "\n");
			s.append("maximum x = " + getMaxX() + "\n");
			s.append("minimum x = " + getMinX() + "\n");
			s.append("maximum y = " + getMaxY() + "\n");
			s.append("minimum y = " + getMinY() + "\n");
			s.append("center x = " + getCenterX() + "\n");
			s.append("center y = " + getCenterY() + "\n");

			return(s.toString());
		}
		
		/** Returns the RectangularGridSpatialModel2D that contains this cell
		 * 
		 * @return
		 */
		public final RectangularGridSpatialModel2D getParentRectangularGrid2D(){
			return myParentRectangularGrid2D;
		}
				
		/**
		 * @return Returns the column index
		 */
		public final int getColumnIndex() {
			return myCol;
		}
		
		/**
		 * @return Returns the height of the cell
		 */
		public final double getHeight() {
			return myHeight;
		}
						
		/**
		 * @return Returns the cell's Row Index.
		 */
		public final int getRowIndex() {
			return myRow;
		}
				
		/**
		 * @return Returns the cell's Width.
		 */
		public final double getWidth() {
			return myWidth;
		}
		
		/** Checks if x and y are in this cell
		 * @param x
		 * @param y
		 * @return
		 */
		public final boolean contains(double x, double y) {
			return myRectangle.contains(x, y);
		}
		
		/** The x-coordinate of the upper left corner of the 
		 *  rectangle for the cell
		 * 
		 * @return
		 */
		public final double getX(){
			return(myRectangle.getX());
		}

		/** The y-coordinate of the upper left corner of the 
		 *  rectangle for the cell
		 * 
		 * @return
		 */
		public final double getY(){
			return(myRectangle.getY());
		}
		
		/** The x-coordinate of the center of the cell
		 * @return
		 */
		public final double getCenterX() {
			return myRectangle.getCenterX();
		}
				
		/** The y-coordinate of the center of the cell
		 * @return
		 */
		public final double getCenterY() {
			return myRectangle.getCenterY();
		}
		
		/** The x-coordinate of the maximum x still within the cell
		 * @return
		 */
		public final double getMaxX() {
			return myRectangle.getMaxX();
		}
		
		/** The y-coordinate of the maximum y still within the cell
		 * @return
		 */
		public final double getMaxY() {
			return myRectangle.getMaxY();
		}
		
		/** The x-coordinate of the minimum x still within the cell
		 * @return
		 */
		public final double getMinX() {
			return myRectangle.getMinX();
		}
		
		/** The y-coordinate of the minimum y still within the cell
		 * @return
		 */
		public final double getMinY() {
			return myRectangle.getMinY();
		}
		
		/** Add the spatial element to the cell
		 * 
		 * @param element
		 */
		protected final void addSpatialElement(SpatialElementIfc element){
			mySpatialElements.add(element);
		}

		/** Removes the spatial element from the cell
		 * 
		 * @param element
		 * @return
		 */
		protected final boolean removeSpatialElement(SpatialElementIfc element){
			return(mySpatialElements.remove(element));
		}		
	}

}
