/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.spatial;

import java.util.Observable;
import java.util.Observer;

import jsl.modeling.ModelElement;
import jsl.utilities.reporting.JSL;

/** A SpatialElement represents an object that is within a spatial model. 
 *  A spatial element can be in one and only one spatial model at a time.  
 *  A spatial model may contain many spatial elements.  Spatial elements can 
 *  be added to and removed from a spatial model.  A spatial element has a 
 *  position within a spatial model.  The position should allow the distance 
 *  between spatial elements to be computed within a spatial model.  If the position 
 *  of the spatial element is changed then the spatial element should notify
 *  any position observers and should indicate to its spatial model that
 *  the position has changed.
 *  
 *  A spatial element should start in one spatial model.  The starting spatial
 *  model is the element's initial model. A spatial element may move to other spatial
 *  models during a simulation and thus should always have a current spatial model.
 *  A spatial element may be related to a ModelElement.
 * 
 *
 */
public class SpatialElement extends Observable implements SpatialElementIfc {

	/** incremented to give a running total of the 
	 *  number of elements created
	 */
	private static int myCounter_;

	//  ===========================================
	//      CLASS CONSTANTS
	//  ===========================================
	
	/** An enum to indicate that an element changed
	 *  its position within a spatial model
	 */	
	public static final int CHANGED_POSITION = JSL.getNextEnumConstant();

	//  ===========================================
	//      CLASS ATTRIBUTES
	//  ===========================================

	/** Holds the initial position in the spatial model
	 */
	private Vector3D myInitialPosition;

	/** Holds the previous position in the spatial model
	 */
	private Vector3D myPreviousPosition;

	/** Holds the current position in in the spatial model
	 */
	private Vector3D myCurrentPosition;

	/** A reference to a spatial model to allow location
	 *  within the spatial context
	 */
	private SpatialModel mySpatialModel;

	/** A reference to a spatial model to allow location
	 *  within the spatial context
	 */
	private SpatialModel myInitialSpatialModel;

	/** Is used by observers to check what state the element is in
	 *  when notified of a change
	 */
	private int myObserverState;

	/** A name for the element
	 */
	private String myName;

	/** A unique id given to the element
	 */
	private long myId;

	/** If the element is related to a JSL ModelElement then
	 *  this refers to the element
	 */
	private ModelElement myModelElement;

	//  ===========================================
	//      CONSTRUCTORS
	//  ===========================================

	/** Creates a spatial element with the given spatial model.  The CoordinateIfc will be the default
	 *  within the supplied spatial model. 
	 * 
	 * @param spatialModel
	 */
	public SpatialElement(SpatialModel spatialModel) {
		this(spatialModel, null, null);
	}

	/** Creates a spatial element with the given parent and spatial model.  TThe CoordinateIfc must be valid
	 *  within the supplied spatial model. 
	 * 
	 * @param spatialModel
	 * @param name
	 */
	public SpatialElement(SpatialModel spatialModel, CoordinateIfc coordinate) {
		this(spatialModel, coordinate, null);
	}
	
	/** Creates a spatial element with the given parent and spatial model.  TThe coordinate will be the default
	 *  within the supplied spatial model.  
	 * 
	 * @param spatialModel
	 * @param name
	 */
	public SpatialElement(SpatialModel spatialModel, String name) {
		this(spatialModel, null, name);
	}

	/** Creates a spatial element at the given (x, y, 0.0) coordinates
	 * 
	 * @param spatialModel
	 * @param x
	 * @param y
	 */
	public SpatialElement(SpatialModel spatialModel, double x, double y){
		this(spatialModel, new Vector3D(x,y,0.0), null);
	}

	/** Creates a spatial element at the given (x, y, 0.0) coordinates
	 * 
	 * @param spatialModel
	 * @param x
	 * @param y
	 */
	public SpatialElement(SpatialModel spatialModel, double x, double y, String name){
		this(spatialModel, new Vector3D(x,y,0.0), name);
	}
	
	/** Creates a spatial element at the given (x,y,z) coordinates
	 * 
	 * @param spatialModel
	 * @param x
	 * @param y
	 * @param z
	 * @param name
	 */
	public SpatialElement(SpatialModel spatialModel, double x, double y, double z){
		this(spatialModel, new Vector3D(x,y,z), null);
	}
	
	/** Creates a spatial element at the given (x,y,z) coordinates
	 * 
	 * @param spatialModel
	 * @param x
	 * @param y
	 * @param z
	 * @param name
	 */
	public SpatialElement(SpatialModel spatialModel, double x, double y, double z, String name){
		this(spatialModel, new Vector3D(x,y,z), name);
	}
	
	/** Creates a spatial element with the given parent and spatial model.  The coordinate must be valid
	 *  within the supplied spatial model. 
	 * 
	 * @param spatialModel
	 * @param coordinate
	 * @param name
	 */
	public SpatialElement(SpatialModel spatialModel, CoordinateIfc coordinate, String name) {

		if (spatialModel == null)
			throw new IllegalArgumentException("No spatial model was provided!");
		
		if (coordinate == null)
			coordinate = spatialModel.getDefaultCoordinate();

		if (!spatialModel.isValid(coordinate))
			throw new IllegalArgumentException("The supplied coordinate is not valid for the given spatial model.");

		// if we get here we know that the coordinate is valid for the spatial model
		
		myCounter_ = myCounter_ +1;
		myId = myCounter_;
		setName(name);

		// set initial spatial model 
		setInitialSpatialModel(spatialModel);	// this just sets the initial model it doesn't do anything else
		
		// set spatial model and add the element to the spatial model
		setSpatialModel(spatialModel);  // this just sets the current spatial model
		
		// set up the positions
		myInitialPosition = Vector3D.newInstance(coordinate);
		myPreviousPosition = Vector3D.newInstance(coordinate);
		myCurrentPosition = Vector3D.newInstance(coordinate);
		spatialModel.addSpatialElement(this);

	}

	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================

	/** Gets this element's name.
	 * @return The name of the element.
	 */
	public final String getName() {
		return myName;
	}

	/** Gets a uniquely assigned integer identifier for this element.
	 * This identifier is assigned when the element is
	 * created.  It may vary if the order of creation changes.
	 * @return The identifier for the element.
	 */
	public final long getId() {
		return (myId);
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#getCurrentPosition()
	 */
	public final CoordinateIfc getCurrentPosition() {
		return myCurrentPosition;
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#getInitialPosition()
	 */
	public final CoordinateIfc getInitialPosition() {
		return myInitialPosition;
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#getPreviousPosition()
	 */
	public final CoordinateIfc getPreviousPosition() {
		return myPreviousPosition;
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#getObserverState()
	 */
	public final int getObserverState() {
		return myObserverState;
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#distanceTo(jsl.spatial.temp.CoordinateIfc)
	 */
	public final double distanceTo(CoordinateIfc coordinate) {
		return(mySpatialModel.distance(getCurrentPosition(), coordinate));
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#isPositionEqualTo(jsl.spatial.temp.CoordinateIfc)
	 */
	public final boolean isPositionEqualTo(CoordinateIfc coordinate) {
		return(mySpatialModel.comparePositions(getCurrentPosition(), coordinate));
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#distanceTo(jsl.spatial.temp.SpatialElementIfc)
	 */
	public final double distanceTo(SpatialElementIfc element) {
		return(mySpatialModel.distance(this, element));
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#isPositionEqualTo(jsl.spatial.temp.SpatialElementIfc)
	 */
	public final boolean isPositionEqualTo(SpatialElementIfc element) {
		return(mySpatialModel.comparePositions(this, element));
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#getSpatialModel()
	 */
	public final SpatialModel getSpatialModel() {
		return mySpatialModel;
	}
	
	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#getInitialSpatialModel()
	 */
	public final SpatialModel getInitialSpatialModel(){
		return(myInitialSpatialModel);
	}
	
	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#changeSpatialModel(jsl.spatial.temp.SpatialModel, jsl.spatial.temp.CoordinateIfc)
	 */
	public final void changeSpatialModel(SpatialModel spatialModel, CoordinateIfc coordinate){
		
		if (spatialModel == null)
			throw new IllegalArgumentException("The suppled spatial model was equal to null.");
				
		// check if the coordinate is valid within the supplied spatial model
		if (!spatialModel.isValid(coordinate))
			throw new IllegalArgumentException("The spatial model is not valid for the supplied coordinate.");

		if (spatialModel == mySpatialModel)// no need to change spatial model
			return;		

		// remove the element from its current spatial model
		mySpatialModel.removeSpatialElement(this);
		
		//set the spatial model to new spatial model
		setSpatialModel(spatialModel);
		
		// adds the element
		spatialModel.addSpatialElement(this);
		
		// directly updates the positions, coordinate should be valid
		myPreviousPosition.setCoordinates(coordinate);
		myCurrentPosition.setCoordinates(coordinate);
		
		// notify any position update listeners for the spatial model
		mySpatialModel.updatePosition_(this);
		
		// notify any position change listeners for this spatial element
		notifyObservers(CHANGED_POSITION);
				
	}
	
	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#getModelElement()
	 */
	public ModelElement getModelElement() {
		return myModelElement;
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#initializeSpatialElement()
	 */
	public void initializeSpatialElement(){
		
		if (myInitialSpatialModel != mySpatialModel){
			// There has been a change from the initial spatial model
			// return the element to it's initial spatial model, at the initial position
			changeSpatialModel(myInitialSpatialModel, getInitialPosition());
		} else { // no change in spatial model, but position might have changed
			if (!isPositionEqualTo(getInitialPosition())){
				// there has been a change in position from the initial position
				// set current/previous to coordinates of initial position
				setPreviousPosition(getInitialPosition());
				setCurrentPosition(getInitialPosition());
			}
		}
		
	}

	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#attachPositionObserver(java.util.Observer)
	 */
	public final void attachPositionObserver(Observer observer){
		addObserver(observer);
	}
	
	/* (non-Javadoc)
	 * @see jsl.spatial.temp.SpatialElementIfc#removePositionObserver(java.util.Observer)
	 */
	public final void removePositionObserver(Observer observer){
		deleteObserver(observer);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#setInitialPosition(jsl.modeling.elements.spatial.CoordinateIfc)
	 */
	public final void setInitialPosition(CoordinateIfc initialPosition) {
		if (!mySpatialModel.isValid(initialPosition))
			throw new IllegalArgumentException("The supplied coordinate is not valid for the give spatial model.");

		myInitialPosition.setCoordinates(initialPosition);
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	/** Sets the spatial model for the spatial element.  This method
	 *  is used by addSpatialElement() within SpatialModel
	 *  
	 * @param 
	 */
	protected void setSpatialModel(SpatialModel spatialModel){
		mySpatialModel = spatialModel;
	}

	/** Sets the model element associated with this spatial element, may be null
	 * 
	 * @param modelElement
	 */
	protected void setModelElement(ModelElement modelElement) {
		myModelElement = modelElement;
	}

	/**
	 * @param currentPosition The currentPosition to set.
	 */
	protected final void setCurrentPosition(CoordinateIfc currentPosition) {
		if (!mySpatialModel.isValid(currentPosition))
			throw new IllegalArgumentException("The supplied coordinate is not valid for the give spatial model.");

		setPreviousPosition(getCurrentPosition());
		
		myCurrentPosition.setCoordinates(currentPosition);
		
		mySpatialModel.updatePosition_(this);
		
		notifyObservers(CHANGED_POSITION);
	}

	/**
	 * @param previousPosition The previousPosition to set.
	 */
	protected final void setPreviousPosition(CoordinateIfc previousPosition) {
		myPreviousPosition.setCoordinates(previousPosition);
	}

	/** Sets the name of this element
	 * @param str The name as a string.
	 */
	protected void setName(String str) {

		if (str == null) { // no name is being passed, construct a default name
			String s = this.getClass().getName();
			int k = s.lastIndexOf(".");
			if (k != -1)
				s = s.substring(k + 1);
			str = s + "-" + getId();
		}
		
		myName = str;		

	}

	/** Sets the initial spatial model for the position. Since a element can move
	 *  between spatial models during a simulation, there must be a way to remember
	 *  which spatial model the element started with prior to re-running a simulation
	 *  This method sets the initial spatial model.  This only has an effect when
	 *  the element is initialized.  Calling this method during a simulation has
	 *  no effect until the element is initialized.  The current spatial model is not
	 *  changed by this method.  When the element is initialized, its current spatial
	 *  model is set to the supplied spatial model.
	 * 
	 * @param spatialModel
	 */
	protected final void setInitialSpatialModel(SpatialModel spatialModel){
		if (spatialModel == null)
			throw new IllegalArgumentException("The suppled spatial model was equal to null.");
		myInitialSpatialModel = spatialModel;
	}

	/** Can be called by sub-classes to set an indicator of the state
	 *  of the position prior to notifying observers
	 *  
	 * @param observerState The observerState to set.
	 */
	protected final void setObserverState(int observerState) {
		myObserverState = observerState;
	}

	/**
	 * @param observerState The observerState to set.
	 */
	protected final void notifyObservers(int observerState) {
		setObserverState(observerState);
		setChanged();
		notifyObservers();
	}

}
