/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.spatial;

import java.util.Observer;

import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;

/** SpatialModelElement represents a ModelElement within a SpatialModel. A SpatialModelElement
 *  does not have the ability to move within the SpatialModel.  Sub-classes
 *  may implement this behavior.  A SpatialModelElement has a fixed location within the SpatialModel
 *
 */
public class SpatialModelElement extends SchedulingElement implements SpatialElementIfc {

	private SpatialElement mySpatialElement;
		
	/** Creates a Location2D with (0.0, 0.0) position.  
	 *  The SpatialModel2D of the parent is used as the SpatialModel2D.  If the parent
	 *  does not have a SpatialModel2D, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 */
	public SpatialModelElement(ModelElement parent){
		this (parent, null, null, null);
	}
	
	/** Creates a Location2D with (0.0, 0.0) position.  
	 *  The SpatialModel2D of the parent is used as the SpatialModel2D.  If the parent
	 *  does not have a SpatialModel2D, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param name
	 */
	public SpatialModelElement(ModelElement parent, String name){
		this (parent, name, null, null);
	}

	/** Creates a Location2D with at the coordinates of the supplied position.  
	 *  The SpatialModel2D of the parent is used as the SpatialModel2D.  If the parent
	 *  does not have a SpatialModel2D, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param position
	 */
	public SpatialModelElement(ModelElement parent, CoordinateIfc position){
		this (parent, null, null, position);
	}

	/** Creates a Location2D with at the coordinates of the supplied position.  
	 *  The SpatialModel2D of the parent is used as the SpatialModel2D.  If the parent
	 *  does not have a SpatialModel2D, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param position
	 */
	public SpatialModelElement(ModelElement parent, SpatialElementIfc position){
		this (parent, null, null, position.getCurrentPosition());
	}
	
	/** Creates a Location2D with the given parent and SpatialModel2D.  The default position
	 *  is (0.0, 0.0).  If the SpatialModel2D
	 *  is null, the SpatialModel2D of the parent is used as the SpatialModel2D.  If the parent
	 *  does not have a SpatialModel2D, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param name
	 * @param spatialModel
	 */
	public SpatialModelElement(ModelElement parent, String name, SpatialModel spatialModel){
		this (parent, name, spatialModel, null);
	}

	/** Creates a Location2D with the given parent and SpatialModel2D.  The default position
	 *  is (0.0, 0.0).  If the SpatialModel2D
	 *  is null, the SpatialModel2D of the parent is used as the SpatialModel2D.  If the parent
	 *  does not have a SpatialModel2D, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param spatialModel
	 */
	public SpatialModelElement(ModelElement parent, SpatialModel spatialModel){
		this (parent, null, spatialModel, null);
	}

	/** Creates a Location2D with the given parent and SpatialModel2D.  The default position
	 *  is (0.0, 0.0).  If the SpatialModel2D
	 *  is null, the SpatialModel2D of the parent is used as the SpatialModel2D.  If the parent
	 *  does not have a SpatialModel2D, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param spatialModel
	 * @param element
	 */
	public SpatialModelElement(ModelElement parent, SpatialModel spatialModel, SpatialElementIfc element){
		this (parent, null, spatialModel, element.getCurrentPosition());
	}
	
	/** Creates a Location2D with the given parent and SpatialModel2D.  If the SpatialModel2D
	 *  is null, the SpatialModel2D of the parent is used as the SpatialModel2D.  If the parent
	 *  does not have a SpatialModel2D, then an IllegalArgumentException is thrown
	 * 
	 * @param parent
	 * @param name
	 * @param spatialModel
	 * @param coordinate
	 */
	public SpatialModelElement(ModelElement parent, String name, SpatialModel spatialModel, CoordinateIfc coordinate){
		super(parent, name);
		
		if (spatialModel == null){
			spatialModel = parent.getSpatialModel();
			if (spatialModel == null){
				throw new IllegalArgumentException("No spatial model is available!");
			} 
		}
		
		setSpatialModel(spatialModel); // set the model element's spatial model
		setSpatialElement(new SpatialElement(spatialModel, coordinate, getName()));
		
	}

	// ===========================================
	// PUBLIC METHODS
	// ===========================================

	public final CoordinateIfc getCurrentPosition() {
		return mySpatialElement.getCurrentPosition();
	}

	public final CoordinateIfc getInitialPosition() {
		return mySpatialElement.getInitialPosition();
	}

	public final CoordinateIfc getPreviousPosition() {
		return mySpatialElement.getPreviousPosition();
	}

	public final double distanceTo(CoordinateIfc coordinate) {
		return mySpatialElement.distanceTo(coordinate);
	}

	public final boolean isPositionEqualTo(CoordinateIfc coordinate) {
		return mySpatialElement.isPositionEqualTo(coordinate);
	}

	public final double distanceTo(SpatialElementIfc element) {
		return mySpatialElement.distanceTo(element);
	}

	public final boolean isPositionEqualTo(SpatialElementIfc element) {
		return mySpatialElement.isPositionEqualTo(element);
	}

	public final SpatialModel getInitialSpatialModel() {
		return mySpatialElement.getInitialSpatialModel();
	}

	public final void changeSpatialModel(SpatialModel spatialModel, CoordinateIfc coordinate) {
		mySpatialElement.changeSpatialModel(spatialModel, coordinate);
	}

	public final ModelElement getModelElement() {
		return mySpatialElement.getModelElement();
	}

	public void attachPositionObserver(Observer observer) {
		mySpatialElement.attachPositionObserver(observer);
	}

	public void removePositionObserver(Observer observer) {
		mySpatialElement.removePositionObserver(observer);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.spatial.SpatialElementIfc#setInitialPosition(jsl.modeling.elements.spatial.CoordinateIfc)
	 */
	public void setInitialPosition(CoordinateIfc coordinate) {
		mySpatialElement.setInitialPosition(coordinate);
	}

	public void initializeSpatialElement() {
		mySpatialElement.initializeSpatialElement();
	}

	// ===========================================
	// PROTECTED METHODS
	// ===========================================

	protected void initialize(){
		super.initialize();
		mySpatialElement.initializeSpatialElement();
	}

	/** Returns the spatial element associated with this spatial model element
	 * 
	 * @return
	 */
	protected final SpatialElement getSpatialElement(){
		return(mySpatialElement);
	}
	
	/** Sets the underlying SpatialElement
	 * 
	 * @param spatialElement
	 */
	protected final void setSpatialElement(SpatialElement spatialElement) {
		if (spatialElement == null) 
			throw new IllegalArgumentException("The supplied spatial element was null!");
		mySpatialElement = spatialElement;
		mySpatialElement.setModelElement(this);
	}

	protected void setModelElement(ModelElement modelElement) {
		mySpatialElement.setModelElement(modelElement);
	}
	
	protected final void setCurrentPosition(CoordinateIfc currentPosition) {
		mySpatialElement.setCurrentPosition(currentPosition);
	}
}
