/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.variable;

import jsl.modeling.ModelElement;
import jsl.utilities.statistic.StatisticAccessorIfc;

/**
 * @author rossetti
 *
 */
public class AggregateTimeWeightedVariable extends Aggregate {

	/** This is used to remember the aggregate value
	 *  when any of its aggregatable's change
	 * 
	 */
	protected TimeWeighted myAggTW;

	/**
	 * @param parent
	 */
	public AggregateTimeWeightedVariable(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public AggregateTimeWeightedVariable(ModelElement parent, String name) {
		super(parent, name);
		myAggTW = new TimeWeighted(this, "Delegate TimeWeighted- " + getName());
		// the aggregate variable does not get initialized by the model
		// it must be initialized after all its observed variables are initialized
		myAggTW.setInitializationOption(false);
		// the aggregate variable does not get warmed up by the model
		// it must be warmed up after all its observed variables are warmed up
		myAggTW.setWarmUpOption(false);
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.variable.ResponseVariable#getDefaultReportingOption()
	 */
	public boolean getDefaultReportingOption() {
		return myAggTW.getDefaultReportingOption();
	}

	/**
	 * @param flag
	 * @see jsl.modeling.elements.variable.ResponseVariable#setDefaultReportingOption(boolean)
	 */
	public void setDefaultReportingOption(boolean flag) {
		myAggTW.setDefaultReportingOption(flag);
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.variable.PreviousValueIfc#getPreviousValue()
	 */
	public double getPreviousValue() {
		return myAggTW.getPreviousValue();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.variable.ResponseVariable#getValue()
	 */
	public double getValue() {
		return myAggTW.getValue();
	}
	
	/**
	 * @return
	 * @see jsl.modeling.elements.variable.ResponseVariable#getAcrossReplicationAverage()
	 */
	public double getAcrossReplicationAverage() {
		return myAggTW.getAcrossReplicationAverage();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.variable.ResponseVariable#getAcrossReplicationStatistic()
	 */
	public StatisticAccessorIfc getAcrossReplicationStatistic() {
		return myAggTW.getAcrossReplicationStatistic();
	}

	/**
	 * @param name
	 * @see jsl.modeling.elements.variable.ResponseVariable#setAcrossReplicationStatisticName(java.lang.String)
	 */
	public void setAcrossReplicationStatisticName(String name) {
		myAggTW.setAcrossReplicationStatisticName(name);
	}
	
	protected void variableAddedBeforeReplication(Aggregatable variable) {
//		System.out.println("################################");
//		System.out.println("A variable is being added.");
//		System.out.println("Aggegate name: " + getName());
//		System.out.println("Aggegatable's name: " + variable.getName());
//		System.out.println(variable.getName() + "  value = " + variable.getValue());

		// a new variable has just been attached to the aggregate
		// either before the start of the replication
		
		myAggTW.assignInitialValue(sumValues());
	
//		System.out.println(getName() + " previous value = " + getPreviousValue());
//		System.out.println(getName() + " new value = " + getValue());
	
		notifyAggregatesOfValueChange();
		
//		System.out.println("################################");
	}

	protected void variableAddedDuringReplication(Aggregatable variable) {
//		System.out.println("################################");
//		System.out.println("A variable is being added.");
//		System.out.println("Aggegate name: " + getName());
//		System.out.println("Aggegatable's name: " + variable.getName());
//		System.out.println(variable.getName() + "  value = " + variable.getValue());
		
		// a new variable has just been attached to the aggregate
		// either before or after the start of a replication
		
		
		// still need to make the variable have the correct value
		myAggTW.setValue(sumValues());
		
//		System.out.println(getName() + " previous value = " + getPreviousValue());
//		System.out.println(getName() + " new value = " + getValue());
		
		notifyAggregatesOfValueChange();
//		System.out.println("################################");
	}
	

	protected void variableRemovedBeforeReplication(Aggregatable variable) {
		// a new variable has just been removed from the aggregate
		// either before or after the start of a replication
		// still need to make the variable have the correct value
		myAggTW.assignInitialValue(sumValues());
		notifyAggregatesOfValueChange();
	}

	protected void variableRemovedDuringReplication(Aggregatable variable) {
		// a new variable has just been removed from the aggregate
		// either before or after the start of a replication
		// still need to make the variable have the correct value
		myAggTW.setValue(sumValues());
		notifyAggregatesOfValueChange();
	}
	
	protected void valueChangedBeforeReplication(Aggregatable variable) {
		double sum = getValue(); // current value

//		System.out.println("---------------------------");
//		System.out.println("In Aggregate's valueChangedBeforeReplication(Aggregatable variable)");
//		System.out.println("Time> " + getTime());
//		System.out.println("Aggegate name: " + getName());
//		System.out.println("Aggegatable's name: " + variable.getName());
		
//		System.out.println(getName() + " current value = " + getValue());
//		System.out.println(variable.getName() + " prev value = " + variable.getPreviousValue());
//		System.out.println(variable.getName() + " new value = " + variable.getValue());
		
		sum = sum - variable.getPreviousValue(); // subtract off the old value
//		System.out.println("after subtracting prev value: sum = " + sum);

		sum = sum + variable.getValue(); // add on the new value
//		System.out.println("after adding the new value: sum = " + sum);
//		System.out.println();
		if (sum < 0)
			throw new IllegalStateException("The sum was negative");
		
		myAggTW.assignInitialValue(sumValues());
//		myAggTW.setValue(sumValues());
		
		notifyAggregatesOfValueChange();
		
//		System.out.println(getName() + " final value = " + getValue());
//		System.out.println("---------------------------");
	}

	protected void valueChangedDuringReplication(Aggregatable variable) {
		double sum = getValue(); // current value

//		System.out.println("---------------------------");
//		System.out.println("In Aggregate's valueChanged()");
//		System.out.println("Time> " + getTime());
//		System.out.println("Aggegate name: " + getName());
//		System.out.println("Aggegatable's name: " + variable.getName());
		
//		System.out.println(getName() + " current value = " + sum);
//		System.out.println(variable.getName() + " prev value = " + variable.getPreviousValue());
//		System.out.println(variable.getName() + " new value = " + variable.getValue());
		
		sum = sum - variable.getPreviousValue(); // subtract off the old value
//		System.out.println("after subtracting prev value: sum = " + sum);

		sum = sum + variable.getValue(); // add on the new value
//		System.out.println("after adding the new value: sum = " + sum);
//		System.out.println();
		if (sum < 0){
//			System.out.println("Replication: " + this.getModel().getReplication().getReplicationNumber());
//			System.out.println("Time> " + getTime());
			throw new IllegalStateException("The sum was negative");
		}
		myAggTW.setValue(sum);
		notifyAggregatesOfValueChange();
//		System.out.println(getName() + " final value = " + getValue());
//		System.out.println("---------------------------");
	}

	@Override
	protected void initializeAggregate() {
		myAggTW.initialize_();// does not call initialize() because initialize is off
		myAggTW.initialize();
	}

	@Override
	protected void warmUpAggregate() {
		myAggTW.warmUp_();// already calls warmUp()
	}
	
	protected void removedFromModel(){
		super.removedFromModel();
		myAggTW = null;
	}
}
