/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.variable;

import java.util.ArrayList;
import java.util.Collection;

import jsl.modeling.ModelElement;
import jsl.observers.variable.CounterObserver;
import jsl.utilities.reporting.JSL;
import jsl.utilities.statistic.StatisticAccessorIfc;

/**
 *
 */
public class Counter extends Aggregatable implements AcrossReplicationStatisticIfc, CounterActionIfc, DefaultReportingOptionIfc {

	/** Can be used the reports to indicate whether or not the response should appear
	 *  The default is true
	 * 
	 */
	protected boolean myDefaultReportingOption = true;

	/** An "enum" for the counter state.
	 */
	public static final int COUNTER_LIMIT_REACHED = JSL.getNextEnumConstant();

	/** The current value of the counter
	 */
	protected long myValue;
	
    /** The previous value of the counter
     */
    protected long myPrevValue;
    
    /** The time that the counter changed
     */
    protected double myTimeOfChange = 0.0;
    
    /** The previous time that the counter changed
     */
    protected double myPrevTimeOfChange = 0.0;

	/** Sets a limit for the counter to fire an action
	 */
	protected long myLimit;
	
	/** Indicates that the limit has not yet been reached
	 */
	protected boolean myCountLimitFlag;

	/** A collection containing the counter actions for the counter
	 */
	protected Collection<CounterActionListenerIfc> myCounterActions;

	/** Remembers the count at the previous timed update
	 */
	protected long myCountAtPreviousTimedUpdate = 0;
	
	/** Remembers the total during a timed update
	 */
	protected long myTotalDuringTimedUpdate = 0;
    
    /** By default every Counter has a CounterObserver attached
     *  to collect across replication statistics on the final
     *  value of the counter
     */
	protected CounterObserver myCounterObserver;
    
    /** If used, represents a stoping action for the Counter
     *  Causes the Counter to stop when its limit is reached
     */
	protected StoppingAction myStoppingAction;
	    
	/**
	 * @param parent
	 */
	public Counter(ModelElement parent) {
		this(parent, Long.MAX_VALUE, null);
	}
	
	/**
	 * @param parent
	 * @param limit
	 */
	public Counter(ModelElement parent, long limit) {
		this(parent, limit, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public Counter(ModelElement parent, String name) {
		this(parent, Long.MAX_VALUE, name);
	}

	/**
	 * @param parent
	 * @param limit
	 * @param name
	 */
	public Counter(ModelElement parent, long limit, String name) {
		super(parent, name);
	    myValue = 0;
	    myPrevValue = 0;
		myLimit = limit;
		myCountLimitFlag = false;
		myCounterObserver = new CounterObserver(getName());
	    addObserver(myCounterObserver);        
		myCounterActions = new ArrayList<CounterActionListenerIfc>();
	}
    
    /** Increments the value of the variable by 1 at the current time.
     */
    public final void increment(){
        increment(1);
    }
    
    /** Increments the value of the variable by the amount supplied.
     * Throws an IllegalArgumentException if the value is negative.
     *
     * @param value The amount to increment by. Must be non-negative.
     */
    public final void increment(long value){
        if (value < 0)
            throw new IllegalArgumentException("Invalid argument. Attempted an negative increment.");
        setValue(myValue+value);
		notifyAggregatesOfValueChange();
        notifyUpdateObservers();
    } 
        
    /** This method simply returns the value of the variable.  
     * @return The value of the variable.
     */
    public final double getValue() {
        return(myValue);
    }
 
    /**
     *  Gets the previous number generated from the underlying DistributionIfc
     *  without advancing in the random number sequence
     * @return a double representing the last number generated
     */
    public final double getPreviousValue(){
        return(myPrevValue);
    }

	/**
	 * @return the PrevTimeOfChange
	 */
	public double getPreviousTimeOfChange() {
		return myPrevTimeOfChange;
	}

	/**
	 * @return the TimeOfChange
	 */
	public double getTimeOfChange() {
		return myTimeOfChange;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.CounterActionIfc#getCounterActionLimit()
	 */
	public final long getCounterActionLimit() {
		return myLimit;
	}
	
	/**
	 * @param limit The limit to set.
	 */
	public final void setCounterActionLimit(long limit) {
		if (limit <= 0)
			throw new IllegalArgumentException("The limit must be > 0");

		myLimit = limit;
	}

	/**
	 * @return Returns the totalDuringTimedUpdate.
	 */
	public final long getTotalDuringTimedUpdate() {
		return myTotalDuringTimedUpdate;
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.CounterActionIfc#checkForCounterLimitReachedState()
	 */
	public final boolean checkForCounterLimitReachedState() {
		return(this.checkObserverState(COUNTER_LIMIT_REACHED));
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.CounterActionIfc#addStoppingAction()
	 */
	public final void addStoppingAction(){
		if (myStoppingAction == null){
			myStoppingAction = new StoppingAction();
			addCounterActionListener(myStoppingAction);
		}
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.CounterActionIfc#addCounterActionListener(jsl.modeling.elements.variable.CounterActionListenerIfc)
	 */
	public final boolean addCounterActionListener(CounterActionListenerIfc action) {
		if (myCounterActions.contains(action))
			throw new IllegalArgumentException("The action has already been added to the counter's action list");
		
		return myCounterActions.add(action);
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.CounterActionIfc#removeCounterActionListener(jsl.modeling.elements.variable.CounterActionListenerIfc)
	 */
	public final boolean removeCounterActionListener(CounterActionListenerIfc action) {
		return myCounterActions.remove(action);
	}
	
	/** Returns a reference to the counter's observer
	 * 
	 * @return
	 */
	public final CounterObserver getCounterObserver(){
		return(myCounterObserver);
	}

    /** Returns a StatisticAccessorIfc for the across replication statistics
     *  that have been collected on this Counter
     * 
     * @return
     */
    public final StatisticAccessorIfc getAcrossReplicationStatistic(){
    	return (myCounterObserver.getAcrossReplicationStatistic());
    }
 
    /** A convenience method to get the across replication average from
     *  the underlying CounterObserver For other statistics use
     *  getAcrossReplicationStatistic()
     * 
     * @return
     */
    public final double getAcrossReplicationAverage(){
    	return(myCounterObserver.getAcrossReplicationStatistic().getAverage());
    }
    
    /** Resets the counter to zero, clears the counter limit flag
     *  If timed updates are on, then count since the last timed update
     *  will be zero.  Notifies update observers of the state change
     */
    public final void resetCounter(){
    	resetCounter(true);
    }
    
    /** Resets the counter to zero, clears the counter limit flag
     *  If timed updates are on, then count since the last timed update
     *  will be zero
     *  @param notifyUpdateObservers If true, any update observers will be notified
     *   otherwise they will not be notified
     */
	public final void resetCounter(boolean notifyUpdateObservers){
        myPrevValue = 0;
        myPrevTimeOfChange = getTime();
        
        myValue = 0;
        myTimeOfChange = myPrevTimeOfChange;

    	myCountLimitFlag = false;
    	myCountAtPreviousTimedUpdate = 0;
    	
    	if (notifyUpdateObservers)
    		notifyUpdateObservers();
    }
 
	/**
	 * @param name
	 * @see jsl.observers.variable.CounterObserver#setAcrossReplicationStatisticName(java.lang.String)
	 */
	public void setAcrossReplicationStatisticName(String name) {
		myCounterObserver.setAcrossReplicationStatisticName(name);
	}
	

    /** Sets the default reporting option. True means the 
     *  response will appear on default reports
     * 
     * @param flag
     */
    public void setDefaultReportingOption(boolean flag){
    	myDefaultReportingOption = flag;
    }
    
    /** Returns the default reporting option.  True means that
     *  the response should appear on the default reports
     * 
     * @return
     */
    public boolean getDefaultReportingOption(){
    	return myDefaultReportingOption;
    }
    
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
	
	protected void beforeReplication(){
	}
	
    /** Resets the counter and notifies any initialization observers
     * 
     */
	protected void initialize() { 
        resetCounter(false);
        //if there are aggregates let them know about the initialization
        notifyAggregatesOfInitialization();
    }
    
	/** Sets the value of the counter, ensures a check against the limit
	 *  Does not notify any update observers
	 * 
     * @param value The value to assign to the counter
     */
    protected void setValue(long value){
        if (value < 0)
        	throw new IllegalArgumentException("Tried to set the counter to a value < 0");
        
        // remember the old values
        myPrevValue = myValue;
        myPrevTimeOfChange = myTimeOfChange;
        
        // record the new value and time
        myValue = value;
        myTimeOfChange = getTime();

        if(myCountLimitFlag == false){
         	if (myValue >= myLimit){
               	myCountLimitFlag = true;
        		notifyCounterLimitReachedObservers();
        		notifyCounterActions();
         	}
        }
    }
 
	/** Initialize the value to zero at the warm up time
	 *  and notify warm up observers
	 */
	protected void warmUp(){
		resetCounter(false);
    	//if there are aggregates let them know about the warmUp
    	notifyAggregatesOfWarmUp();
	}
    
	protected void timedUpdate() {		
		myTotalDuringTimedUpdate = myValue - myCountAtPreviousTimedUpdate;
		myCountAtPreviousTimedUpdate = myValue;
	}

	/** The method is used to notify observers that this model element
	 * has triggered its counter limit
	 */
	protected final void notifyCounterLimitReachedObservers() {
		setObserverState(COUNTER_LIMIT_REACHED);
		setChanged();
		notifyObservers();
	}

	protected final void notifyCounterActions(){
		if (!myCounterActions.isEmpty()) {
			for(CounterActionListenerIfc a: myCounterActions)
				a.action(this);
		}
	}
  
	/** The method is used to notify observers that this model element
	 * has been updated.
	 */
	protected void notifyUpdateObservers() {
		super.notifyUpdateObservers();
	}
	
	protected void removedFromModel(){
		super.removedFromModel();
		if (myCounterActions != null){
			myCounterActions.clear();
			myCounterActions = null;
		}
		myCounterObserver = null;
		myStoppingAction = null;
	}
	
	private class StoppingAction implements CounterActionListenerIfc {

		/* (non-Javadoc)
		 * @see jsl.modeling.elements.variable.CounterActionListenerIfc#action(jsl.modeling.elements.variable.Counter)
		 */
		public void action(AcrossReplicationStatisticIfc counter) {
			stopReplication();		
		}
		
	}

}
