/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling.elements.variable;

import jsl.observers.variable.*;
import jsl.utilities.statistic.*;
import jsl.modeling.*;

/**
 *
 */
public class ResponseVariable extends Variable implements ResponseStatisticsIfc, DefaultReportingOptionIfc {
	
	/** Can be used by the reports to indicate whether or not the response should appear
	 *  The default is true
	 * 
	 */
	protected boolean myDefaultReportingOption = true;

	/** Specifies whether or not the response variable will
	 *  allow the model to batch it (i.e. participates in default
	 *  batching). True means that it will participate in default batching
	 *  It is true by default, allowing the model to batch it if automatic batching
	 *  is turned on.  By turning this option to false,
	 *  the response variable opts out of model control. The response variable
	 *  can still do their own batching.
	 */
	protected boolean myDefaultBatchingOption = true;

	/** This variable indicated whether or not the response variable
	 *  has batching turned on, either by default or by its own choosing.
	 *  In other words, whether it will be batched. This if false by default, and
	 *  will be set to true if:
	 *  1) the response variable participates in default batching or
	 *  2) the user sets this directly
	 */
	protected boolean myBatchingOption = false;
	
    /** A reference to the statistical observer for this
     *  ResponseVariable
     */
    protected StatisticalObserver myStatisticalObserver;

    /** Creates a ResponseVariable with the given parent with initial value 0.0 over the range
     *  [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element.
     * @param initialValue The initial value of the variable.
     */
    public ResponseVariable(ModelElement parent){
        this(parent, 0.0, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, null);
    }
    
    /** Creates a ResponseVariable with the given name and initial value over the range
     *  [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element.
     * @param initialValue The initial value of the variable.
     */
    public ResponseVariable(ModelElement parent, double initialValue){
        this(parent, initialValue, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, null);
    }
    
    /** Creates a ResponseVariable with the given name and initial value, 0.0, over the range
     *  [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element.
     * @param name The name of the variable.
     */
    public ResponseVariable(ModelElement parent, String name){
        this(parent, 0.0, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, name);
    }
 
    /** Creates a ResponseVariable with the given name and initial value over the supplied range
     *   The default range is [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element
     * @param initialValue The initial value of the variable.
     * @param name The name of the variable.
     */
   public ResponseVariable(ModelElement parent, double initialValue, String name){
	   this(parent, initialValue, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, name);
   }
 
   /** Creates a ResponseVariable with the given name and initial value over the supplied range
    *   [lowerLimit, Double.POSITIVE_INFINITY]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    * @param name The name of the variable.
    */
   public ResponseVariable(ModelElement parent, double initialValue, double lowerLimit, String name){
	   this(parent, initialValue, lowerLimit, Double.POSITIVE_INFINITY, name);
   }
 
   /** Creates a ResponseVariable with the initial value over the supplied range
    *   [lowerLimit, Double.POSITIVE_INFINITY]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    */
   public ResponseVariable(ModelElement parent, double initialValue, double lowerLimit){
	   this(parent, initialValue, lowerLimit, Double.POSITIVE_INFINITY, null);
   }
 
   /** Creates a ResponseVariable with the initial value over the supplied range
    *   [lowerLimit, upperLimit]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    * @param upperLimit, the upper limit on the range for the variable
    */
   public ResponseVariable(ModelElement parent, double initialValue, double lowerLimit, double upperLimit){
	   this(parent, initialValue, lowerLimit, upperLimit, null);	   
   }
   
    /** Creates a ResponseVariable with the given name and initial value over the supplied range
    *   [lowerLimit, upperLimit]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.  Must be within the range.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    * @param upperLimit, the upper limit on the range for the variable
    * @param name The name of the variable.
    */
   public ResponseVariable(ModelElement parent, double initialValue, double lowerLimit, double upperLimit, String name){   
	   super(parent, initialValue, lowerLimit, upperLimit, name);
	   setStatisticalObserver();
   }
 
	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.BatchStatisticsIfc#isDefaultBatchingOn()
	 */
	public final boolean isDefaultBatchingOn() {
		return (myDefaultBatchingOption);
	}

	/** Sets the default batching option flag for this response variable
	 * 
	 * @param batchOption True means that it will participate in default batching
	 */
	protected final void setDefaultBatchingOption(boolean batchingOption) {
		myDefaultBatchingOption = batchingOption;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.BatchStatisticsIfc#isBatchingOn()
	 */
	public final boolean isBatchingOn() {
		return (myBatchingOption);
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.BatchStatisticsIfc#setBatchingOption(boolean)
	 */
	public void setBatchingOption(boolean option){
		myBatchingOption = option;
	}

    /** Sets the default reporting option. True means the 
     *  response will appear on default reports
     * 
     * @param flag
     */
    public void setDefaultReportingOption(boolean flag){
    	myDefaultReportingOption = flag;
    }
    
    /** Returns the default reporting option.  True means that
     *  the response should appear on the default reports
     * 
     * @return
     */
    public boolean getDefaultReportingOption(){
    	return myDefaultReportingOption;
    }
    
    /** Every Variable must implement the getValue method.  This method simply
     * returns the value of the variable.
     * @return The value of the variable.
     */
    public double getValue(){
        return (myValue);
    }
        
    /** Sets the name of the underlying statistical observer
     *  Throws an IllegalArgumentException if the name is null.
     * @param name The name as a String
     */
    public final void setStatisticalObserverName(String name){
        if (name == null)
            throw new IllegalArgumentException("Name was null");
        myStatisticalObserver.setName(getName() + "_" + name);
    }
    
   /** Gets the statistical observer for this response variable
     *
     * @return A StatisticalObserver representing the statistics for this variable
     */
    public StatisticalObserver getStatisticalObserver(){
        return(myStatisticalObserver);
    }
    
    /* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.WithinReplicationStatisticsIfc#getWithinReplicationStatistic()
	 */
    public StatisticAccessorIfc getWithinReplicationStatistic(){
    	return (myStatisticalObserver.getWithinReplicationStatistic());
    }
    
    /* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.WithinReplicationStatisticsIfc#getWithinReplicationAverage()
	 */
    public double getWithinReplicationAverage(){
    	return(myStatisticalObserver.getWithinReplicationStatistic().getWeightedAverage());
    }
 
    /* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.BatchStatisticsIfc#getReplicationBatchStatistic()
	 */
    public StatisticAccessorIfc getReplicationBatchStatistic(){
    	return (myStatisticalObserver.getAcrossBatchStatistic());
    }
 
    /* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.BatchStatisticsIfc#getReplicationAcrossBatchAverage()
	 */
    public double getReplicationAcrossBatchAverage(){
    	return(myStatisticalObserver.getAcrossBatchStatistic().getAverage());
    }
    
    /** Returns a StatisticAccessorIfc for the across replication statistics
     *  that have been collected on this ResponseVariable
     * 
     * @return
     */
    public StatisticAccessorIfc getAcrossReplicationStatistic(){
    	return (myStatisticalObserver.getAcrossReplicationStatistic());
    }
 
    /** A convenience method to get the across replication average from
     *  the underlying StatisticalObserver For other statistics use
     *  getAcrossReplicationStatistic()
     * 
     * @return
     */
    public double getAcrossReplicationAverage(){
    	return(myStatisticalObserver.getAcrossReplicationStatistic().getAverage());
    }
    
    /* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.BatchStatisticsIfc#setBatchingParameters(int, int, int)
	 */
    public final void setBatchingParameters(int minNumBatches, int minBatchSize, int maxNBMultiple){
        myStatisticalObserver.setBatchingParameters(minNumBatches, minBatchSize, maxNBMultiple);
    }
 
    /** Tells the statistical observer to cause a warm up (reset of statistical collection)
     *  when the supplied number of observations has been reached
     * 
     * @param countLimit
     */
    public final void turnOnCountBasedWarmUp(long countLimit){
    	myStatisticalObserver.setCountBasedWarmUpLimit(countLimit);
    }

    /** Tells the statistical observer to stop the replication when the 
     *  supplied number of observations has been reached
     * 
     * @param countLimit
     */
    public void turnOnCountBasedStopping(long countLimit){
    	myStatisticalObserver.setCountBasedStopLimit(countLimit);
    }

    /* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.BatchStatisticsIfc#setAcrossBatchStatisticName(java.lang.String)
	 */
	public void setAcrossBatchStatisticName(String name) {
		myStatisticalObserver.setAcrossBatchStatisticName(name);
	}

	/**
	 * @param name
	 * @see jsl.observers.variable.StatisticalObserver#setAcrossReplicationStatisticName(java.lang.String)
	 */
	public void setAcrossReplicationStatisticName(String name) {
		myStatisticalObserver.setAcrossReplicationStatisticName(name);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.WithinReplicationStatisticsIfc#setWithinReplicationStatisticName(java.lang.String)
	 */
	public void setWithinReplicationStatisticName(String name) {
		myStatisticalObserver.setWithinReplicationStatisticName(name);
	}

	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	/** If the default batching option is on for the replication
     *  this method will ensure that the response variable's
     *  batching option is turned on, prior to the replication running
     * 
     */    
    protected void beforeReplication(){
		Replication r = getModel().getReplication();
    	
		if (r.getDefaultBatchingOption()){
			// the default batching option has been turned on for the model
			// turn it on for this response variable
			setBatchingOption(true);
		}   	
    }
    
    /** must be overridden by subclasses to set the
     *  protected variable myStatisticalObserver for appropriate observation
     */
    protected void setStatisticalObserver(){
    	String name = getStringLabel();
    	if (name == null)
    		name = getName();
        myStatisticalObserver = new StatisticalObserver(name);
        addObserver(myStatisticalObserver);        
    }
    
    protected void removedFromModel(){
    	super.removedFromModel();
    	myStatisticalObserver = null;
    }
}
