/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling.elements.variable;

import jsl.modeling.ActionListenerIfc;
import jsl.modeling.JSLEvent;
import jsl.modeling.Model;
import jsl.modeling.ModelElement;
import jsl.observers.variable.StatisticalObserver;
import jsl.observers.variable.TimeWeightedStatisticalObserver;
import jsl.utilities.statistic.StatisticAccessorIfc;

/**
 *
 */
public class TimeWeighted extends ResponseVariable implements BatchIfc {

	/** The action listener that reacts to the timed update event.
	 */
	protected BatchActionListener myBatchActionListener;

	/** A reference to the Batching event.
	 */
	protected JSLEvent myBatchEvent;

	/** The priority for the batching events.
	 */
	protected int myBatchEventPriority = Model.DEFAULT_BATCH_PRIORITY;

	/** The time interval between batching events.
	 *  The default is zero for no batching
	 */
	protected double myBatchInterval = 0.0;
		
    //  ===========================================
    //      CONSTRUCTORS
    //  ===========================================
 
    /** Creates a TimeWeighted with the given parent with initial value 0.0 over the range
     *  [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element.
     * @param initialValue The initial value of the variable.
     */
    public TimeWeighted(ModelElement parent){
        this(parent, 0.0, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, null);
    }
    
    /** Creates a TimeWeighted with the given name and initial value over the range
     *  [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element.
     * @param initialValue The initial value of the variable.
     */
    public TimeWeighted(ModelElement parent, double initialValue){
        this(parent, initialValue, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, null);
    }
    
    /** Creates a TimeWeighted with the given name and initial value, 0.0, over the range
     *  [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element.
     * @param name The name of the variable.
     */
    public TimeWeighted(ModelElement parent, String name){
        this(parent, 0.0, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, name);
    }
 
    /** Creates a TimeWeighted with the given name and initial value over the supplied range
     *   The default range is [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element
     * @param initialValue The initial value of the variable.
     * @param name The name of the variable.
     */
   public TimeWeighted(ModelElement parent, double initialValue, String name){
	   this(parent, initialValue, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, name);
   }
 
   /** Creates a TimeWeighted with the given name and initial value over the supplied range
    *   [lowerLimit, Double.POSITIVE_INFINITY]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    * @param name The name of the variable.
    */
   public TimeWeighted(ModelElement parent, double initialValue, double lowerLimit, String name){
	   this(parent, initialValue, lowerLimit, Double.POSITIVE_INFINITY, name);
   }
 
   /** Creates a TimeWeighted with the initial value over the supplied range
    *   [lowerLimit, Double.POSITIVE_INFINITY]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    */
   public TimeWeighted(ModelElement parent, double initialValue, double lowerLimit){
	   this(parent, initialValue, lowerLimit, Double.POSITIVE_INFINITY, null);
   }
 
   /** Creates a TimeWeighted with the initial value over the supplied range
    *   [lowerLimit, upperLimit]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    * @param upperLimit, the upper limit on the range for the variable
    */
   public TimeWeighted(ModelElement parent, double initialValue, double lowerLimit, double upperLimit){
	   this(parent, initialValue, lowerLimit, upperLimit, null);	   
   }
   
    /** Creates a TimeWeighted with the given name and initial value over the supplied range
    *   [lowerLimit, upperLimit]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.  Must be within the range.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    * @param upperLimit, the upper limit on the range for the variable
    * @param name The name of the variable.
    */
   public TimeWeighted(ModelElement parent, double initialValue, double lowerLimit, double upperLimit, String name){   
	   super(parent, initialValue, lowerLimit, upperLimit, name);
   }
 
	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================
  
    /** Increments the value of the variable by 1 at the current time.
     */
    public final void increment(){
        increment(1.0);
    }
    
    /** Increments the value of the variable by the amount supplied.
     * Throws an IllegalArgumentException if the value is negative.
     *
     * @param value The amount to increment by. Must be non-negative.
     */
    public final void increment(double value){
        if (value < 0)
            throw new IllegalArgumentException("Invalid argument. Attempted an negative increment.");
        
        setValue(myValue+value);
    }
    
    /** Decrements the value of the variable by 1 at the current time.
     */
    public final void decrement(){
        decrement(1.0);
    }
    
    /** Decrements the value of the variable by the amount supplied.
     * Throws an IllegalArgumentException if the value is negative.
     *
     * @param value The amount to decrement by. Must be non-negative.
     */
    public final void decrement(double value){
        if (value < 0)
            throw new IllegalArgumentException("Invalid argument. Attempted an negative decrement.");
        
        setValue(myValue-value);
    }
 
    /** Sets the value of the variable to the current value at the current time
     *  and notifies any batch observers. This method is part of the BatchIfc.  This
     *  method only does anything if the default batching option is on
     */    
    public void batch(){
    		if (isDefaultBatchingOn()){
    			batch_();
    		}
    }
 
	/** Sets the batch event priority
	 * @param value The batch event priority, lower means earlier
	 */
	public final void setBatchEventPriority(int priority) {
		myBatchEventPriority = priority;
	}

	/** Gets the batch event priority
	 * @return The batch event priority
	 */
	public final int getBatchEventPriority() {
		return (myBatchEventPriority);
	}

	/** Sets the batch interval size.  If the batch interval is zero then default batching
	 *  is turned on; otherwise, it is turned off
	 *  
	 * @param value The batch interval size in time units must be >=0, if it is larger than
	 * run length it will not occur
	 */
	public final void setBatchInterval(double batchInterval) {
		if (batchInterval < 0.0)
			throw new IllegalArgumentException("The batch interval cannot be less than zero");
		myBatchInterval = batchInterval;
		
		if (myBatchInterval > 0){
			// user is setting a batch interval for this TimeWeighted variable
			// variable should not participate in default batching, and batch
			// option is on
			setDefaultBatchingOption(false);
			myBatchingOption = true;
		} else {// interval was 0.0, allow default batching, set batching option to false
			// model will turn on batching option if default batching is indicated
			setDefaultBatchingOption(true);
			myBatchingOption = false;
		}
	}

	/** Indicates whether or not the time weighted variable
	 *  has batching turned on.  Setting this to true, requires
	 *  that the client will ensure that the batch event occurs.
	 *  
	 *  For example, the Model uses this and ensures that batch() is called
	 *  according to the default time based batching.  If the client
	 *  sets this time weighted variable's batch interval using
	 *  setBatchInterval() then the batching option is automatically
	 *  turned on.
	 *  
	 *  This method should be called by Model; otherwise, the use of setBatchInterval() 
	 *  is the recommended method of turning on the batching option
	 * @param option
	 */
	public final void setBatchingOption(boolean option){
		myBatchingOption = option;
	}
	
	/** Gets the current batch interval length
	 * @return The batch interval as time
	 */
	public final double getBatchInterval() {
		return (myBatchInterval);
	}

	/** Checks if a batching event has been scheduled for this model element
	 * @return True means that it has been scheduled.
	 */
	public final boolean isBatchEventScheduled() {
		if (myBatchEvent == null)
			return (false);
		else
			return (myBatchEvent.isScheduled());
	}

	/** Gets the statistical observer for this response variable. The
	 *  statistical observer can be used to get the within, batch, and across
	 *  replication statistics.
	 *  
	 *  The within replication statistics are accumulated up to and including
	 *  the current time.
	 *  
	 *  Note: Causes an observation of the current
	 *  value of the variable to occur at the current time.  This creates 1 extra
	 *  observation, but it has no effect in the computation of the time weighted
	 *  average
	 *  
	 * @return A StatisticalObserver representing the statistics for this variable
	 */
	public StatisticalObserver getStatisticalObserver(){
		return(getStatisticalObserver(true));
	}

	/** Gets the statistical observer for this response variable. The
	 *  statistical observer can be used to get the within, batch, and across
	 *  replication statistics.
	 *  
	 *  Note: The parameter value true, causes an observation of the current
	 *  value of the variable to occur at the current time.  This creates 1 extra
	 *  observation, but it has no effect in the computation of the time weighted
	 *  average
	 *
	 * @param toCurrenTime (if true up to an including the current time, if false
	 * up to and including the time of the last variable change)
	 * @return A StatisticalObserver representing the statistics for this variable
	 */
	public StatisticalObserver getStatisticalObserver(boolean toCurrentTime){
		if (toCurrentTime)
			setValue(getValue());
		return(myStatisticalObserver);
	}

	/** Returns a StatisticAccessorIfc for the within replication statistics
	 *  that have been collected on this ResponseVariable, up to and including
	 *  the current time
	 *  
	 *  Note: Causes an observation of the current
	 *  value of the variable to occur at the current time.  This creates 1 extra
	 *  observation, but it has no effect in the computation of the time weighted
	 *  average
	 *  
	 * @return
	 */
	public StatisticAccessorIfc getWithinReplicationStatistic(){
		return (getStatisticalObserver(true).getWithinReplicationStatistic());
	}

	/** Returns a StatisticAccessorIfc for the within replication statistics
	 *  that have been collected on this ResponseVariable
	 * 
	 * 	Note: The parameter value true, causes an observation of the current
	 *  value of the variable to occur at the current time.  This creates 1 extra
	 *  observation, but it has no effect in the computation of the time weighted
	 *  average
	 *  
	 * @param toCurrenTime (if true up to an including the current time, if false
	 * up to and including the time of the last variable change)
	 * @return
	 */
	public StatisticAccessorIfc getWithinReplicationStatistic(boolean toCurrentTime){
		return (getStatisticalObserver(toCurrentTime).getWithinReplicationStatistic());
	}
	
	/** A convenience method to get the within replication average from
	 *  the underlying StatisticalObserver.  For other statistics use
	 *  getWithinReplicationStatistic(),  up to and including
	 *  the current time
	 *  
	 *  Note: Causes an observation of the current
	 *  value of the variable to occur at the current time.  This creates 1 extra
	 *  observation, but it has no effect in the computation of the time weighted
	 *  average
	 *   
	 * @return
	 */
	public double getWithinReplicationAverage(){
		return(getStatisticalObserver(true).getWithinReplicationStatistic().getWeightedAverage());
	}

	/** A convenience method to get the within replication average from
	 *  the underlying StatisticalObserver.  For other statistics use
	 *  getWithinReplicationStatistic(), , up to and including
	 *  the current time
	 *  
	 *  Note: The parameter value true, causes an observation of the current
	 *  value of the variable to occur at the current time.  This creates 1 extra
	 *  observation, but it has no effect in the computation of the time weighted
	 *  average
	 *  
	 * @param toCurrenTime (if true up to an including the current time, if false
	 * up to and including the time of the last variable change)
	 * @return
	 */
	public double getWithinReplicationAverage(boolean toCurrentTime){
		return(getStatisticalObserver(toCurrentTime).getWithinReplicationStatistic().getWeightedAverage());
	}
		
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
	
    /** must be overridden by subclasses to set the
     *  protected variable myStatisticalObserver for appropriate observation
     */
    protected void setStatisticalObserver(){
    	String name = getStringLabel();
    	if (name == null)
    		name = getName();
        myStatisticalObserver = new TimeWeightedStatisticalObserver(name);
        addObserver(myStatisticalObserver);
    }
    
    /** Sets the weight, the current time - the time of the last change
     */
    protected final void setWeight(){
        myWeight = getTime() - myTimeOfChange;
        if (myWeight < 0)
        	myWeight = 0.0;
    }
 
	protected final void beforeReplication() {
		super.beforeReplication();		
				
		// if a batch interval has been set then schedule the first batching event
		if (getBatchInterval() > 0) {
			// the batch interval is > 0, ==> element wants a batch event
			// schedule the batchevent
			myBatchActionListener = new BatchActionListener();
			myBatchEvent = getScheduler().scheduleEvent(this, myBatchActionListener, getBatchInterval(), 
			                             getName() + " Batch", myBatchEventPriority, null);
		}
	}    
    
	/** Initialize the value to the current value at this time
	 */
	protected void initialize(){
//		System.out.println("In TimeWeighted initialize() " + getName());
		super.initialize();
		// this is so at least two changes are recorded on the variable
		// to properly account for variables that have zero area throughout the replication
		setValue(getValue()); 
	}
    
    /** Schedules the batch events for after the warm up period
     *  Sets the value of the variable to the current value at the current time
     *  Resets the time of last change to the current time
     */
    protected void warmUp(){
        super.warmUp();
        // make it think that it changed at the warm up time to the same value
        myTimeOfChange = getTime();
		// this is so at least two changes are recorded on the variable
		// to properly account for variables that have zero area throughout the replication
        setValue(getValue());
    }
        
    /** Sets the value of the variable to the current value at the current time
     *  to collect state to end of replication
     */    
 //   protected void afterReplication(){
  //      super.afterReplication();
        // this allows time weighted to be collected all the way to end of simulation
        // due to order in which model elements are called
	//	setValue(getValue());
   // } 
  
    /** Sets the value of the variable to the current value at the current time
     *  to collect state to end of replication
     */    
    protected void replicationEnded(){
        super.replicationEnded();
        // this allows time weighted to be collected all the way to end of simulation
        // due to order in which model elements are called
		setValue(getValue());
    } 
    
    /** Actually does the batching for this element
     */
    protected void batch_(){
    		if (isBatchingOn()){
    	        setValue(getValue());
    	        notifyBatchObservers();    	   			
    		}
    }
 
    protected void removedFromModel(){
	super.removedFromModel();
	if (myBatchEvent != null){
		if (myBatchEvent.isScheduled())
			myBatchEvent.setCanceledFlag(true);
		myBatchEvent = null;
		myBatchActionListener = null;
	}
    }

    //  ===========================================
    //      INNER CLASSES
    //  ===========================================

	private class BatchActionListener implements ActionListenerIfc {
		/** The action to perform at a scheduled batching event
		 * @param event The event that triggered the TimedUpdate action.
		 */
		public void action(JSLEvent event) {
			batch_();
			getScheduler().reschedule(event, getBatchInterval());
		}
	}
}
