/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.modeling.elements.variable;

import java.io.IOException;

import jsl.modeling.ModelElement;
import jsl.observers.variable.*;
import jsl.utilities.reporting.JSL;

/**
 *
 */
public class Variable extends Aggregatable implements VariableIfc {
    
    //  ===========================================
    //      CLASS AND OBJECT ATTRIBUTES
    //  ===========================================
    	
	/** Represents the lowest possible value allowed for this variable
	 *  Attempts to set the value of the variable to < myLowerLimit
	 *  will throw an exception
	 */
	protected double myLowerLimit = Double.NEGATIVE_INFINITY;
	
	/** Represents the lowest possible value allowed for this variable
	 *  Attempts to set the value of the variable to > myUpperLimit
	 *  will throw an exception
	 */
	protected double myUpperLimit = Double.POSITIVE_INFINITY;
	
    /** The value of the variable.
     */
    protected double myValue;

    /** Holds the initial value of the variable.
     */
    protected double myInitialValue;
    
    /** The previous value of the variable.
     */
    protected double myPrevValue;
    
    /** The next value of the variable.
     */
    protected double myNextValue;
    
    /** The time that the variable changed
     */
    protected double myTimeOfChange = 0.0;
    
    /** The previous time that the variable changed
     */
    protected double myPrevTimeOfChange = 0.0;
    
    /** The weight associated with the change
     */
    protected double myWeight = 1.0;
    
    /** Holds a reference to an observer that
     *  will trace the variable's changes
     */
    protected VariableTraceTextReport myVariableTraceTextReport;
    
    /** Indicates whether or not text file tracing is on/off
     */
    private boolean myTraceFlag = false;
    
    /** Indicates whether the variable will notify update
     *  observers when its value changes.  If set to false
     *  the variable will not notify update observers (until
     *  it is changed back to true). The default is true
     * 
     */
    private boolean mySetValueUpdateObserversFlag = true;
    
    /** Creates a Variable with the given parent with initial value 0.0 over the range
     *  [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element.
     * @param initialValue The initial value of the variable.
     */
    public Variable(ModelElement parent){
        this(parent, 0.0, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, null);
    }
    
    /** Creates a Variable with the given name and initial value over the range
     *  [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element.
     * @param initialValue The initial value of the variable.
     */
    public Variable(ModelElement parent, double initialValue){
        this(parent, initialValue, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, null);
    }
    
    /** Creates a Variable with the given name and initial value, 0.0, over the range
     *  [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element.
     * @param name The name of the variable.
     */
    public Variable(ModelElement parent, String name){
        this(parent, 0.0, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, name);
    }
 
    /** Creates a Variable with the given name and initial value over the supplied range
     *   The default range is [Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY]
     *   
     * @param parent, the variable's parent model element
     * @param initialValue The initial value of the variable.
     * @param name The name of the variable.
     */
   public Variable(ModelElement parent, double initialValue, String name){
	   this(parent, initialValue, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, name);
   }
 
   /** Creates a Variable with the given name and initial value over the supplied range
    *   [lowerLimit, Double.POSITIVE_INFINITY]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    * @param name The name of the variable.
    */
   public Variable(ModelElement parent, double initialValue, double lowerLimit, String name){
	   this(parent, initialValue, lowerLimit, Double.POSITIVE_INFINITY, name);
   }
 
   /** Creates a Variable with the initial value over the supplied range
    *   [lowerLimit, Double.POSITIVE_INFINITY]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    */
   public Variable(ModelElement parent, double initialValue, double lowerLimit){
	   this(parent, initialValue, lowerLimit, Double.POSITIVE_INFINITY, null);
   }
 
   /** Creates a Variable with the initial value over the supplied range
    *   [lowerLimit, upperLimit]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    * @param upperLimit, the upper limit on the range for the variable
    */
   public Variable(ModelElement parent, double initialValue, double lowerLimit, double upperLimit){
	   this(parent, initialValue, lowerLimit, upperLimit, null);	   
   }
   
    /** Creates a Variable with the given name and initial value over the supplied range
    *   [lowerLimit, upperLimit]
    *   
    * @param parent, the variable's parent model element
    * @param initialValue The initial value of the variable.  Must be within the range.
    * @param lowerLimit, the lower limit on the range for the variable, must be < upperLimit
    * @param upperLimit, the upper limit on the range for the variable
    * @param name The name of the variable.
    */
   public Variable(ModelElement parent, double initialValue, double lowerLimit, double upperLimit, String name){
       super(parent, name);
       if (lowerLimit >= upperLimit)
			throw new IllegalArgumentException("Invalid argument. lower limit must be < upper limit.");
   	   
       myLowerLimit = lowerLimit;
       myUpperLimit = upperLimit;
       
       setInitialValue(initialValue);
       myValue = myInitialValue;
       myPrevValue = myInitialValue;
   }
   
    /** Gets the initial value of the variable
     * @return The initial value.
     */
    public final double getInitialValue(){
        return(myInitialValue);
    }
    
    /** Sets the initial value of the variable. Only relevant prior to each
     * replication.  Changing during a replication has no effect until the next replication.
     * @param value The initial value for the variable.
     */
    public final void setInitialValue(double value){
    	if ((value < myLowerLimit) || (value > myUpperLimit))
    		throw new IllegalArgumentException("Invalid argument. supplied value was not in range, [" + myLowerLimit + "," + myUpperLimit + "]");
   		
        myInitialValue = value;
    }
    
    /** The lower limit for the range of this random variable
	 * @return Returns the lowerLimit.
	 */
	public final double getLowerLimit() {
		return myLowerLimit;
	}

	/** The upper limit for the range of this random variable
	 * @return Returns the upperLimit.
	 */
	public final double getUpperLimit() {
		return myUpperLimit;
	}

	/** Maps true to 1.0 and false to 0.0
	 * 
	 * @param value
	 */
	public final void setValue(boolean value){
		if (value)
			setValue(1.0);
		else
			setValue(0.0);
	}
	
	/** Sets the value of the variable and notifies any observers
     *  of the change
     *  
     * @param value The observation of the variable.
     */
    public void setValue(double value){
//        System.out.println(getName() + " is being set to the value --> " + myValue);
        assignValue(value); 
//        System.out.println(getName() + " is notifying its aggregates");
        notifyAggregatesOfValueChange();
//        System.out.println(getName() + " is notifying its update observers");
        if (mySetValueUpdateObserversFlag == true)
        	notifyUpdateObservers(); 
//        System.out.println(getName() + " is done setting its value");
   }
    
    /** Every Variable must implement the getValue method.  By default
     *  this method simply returns the value of the variable.  
     * @return The value of the variable.
     */
    public double getValue() {
        return(myValue);
    }
    
    /**
     *  Gets the previous number value that was assigned before the current value
     *  
     * @return a double representing that was assigned before the current value
     */
    public final double getPreviousValue(){
        return(myPrevValue);
    }
    
    /** Gets the weight associated with the last value observed.
     * @return The weight for the value.
     */
    public final double getWeight(){
        return (myWeight);
    }
    
    /** Sets the weight
     */
    protected void setWeight(){
        myWeight = 1.0;
    }
     
    /** Gets the time associated with the last value observed.
     * @return The time of the last observation
     */
    public final double getTimeOfChange(){
        return(myTimeOfChange);
    }
    
    /** Gets the time associated with previous variable change.
     * @return The time of the previous change.
     */
    public final double getPreviousTimeOfChange(){
        return(myPrevTimeOfChange);
    }
    
    /** Turns on the automatic tracing of this variable to a text file.
     */
    public final void turnOnTrace(){
    	turnOnTrace(null);
    }

    /** Turns on the automatic tracing of this variable to a text file
     * 
     * @param name Used to name the text file
     */
    public final void turnOnTrace(String name){
        
        setTraceObserver(name);
        
        // trace report has been made, add trace report to observers only if trace had been off
        if (myTraceFlag == false)
            addObserver(myVariableTraceTextReport);
        
        myTraceFlag = true;
    }
    
    /** Turns off the automatic tracing of this variable to a text file.
     */
    public final void turnOffTrace(){
        // if trace had been on, remove the report
        if (myTraceFlag == true){
            if (myVariableTraceTextReport != null)
                deleteObserver(myVariableTraceTextReport);
        }
        myTraceFlag = false;
    }
    
    /** Assigns the value of the variable to the supplied value. Ensures
     *  that time of change is current time and previous value and previous
     *  time of change are the same as the current value and current time
     * @param value
     */
    protected void assignInitialValue(double value){
//        System.out.println(getName() + " is assigning the initial value = " + value);
        myValue = value;
        myTimeOfChange = getTime();
        myPrevValue = myValue;
        myPrevTimeOfChange = myTimeOfChange;   	
//        System.out.println(getName() + " was assigned the value --> " + myValue + " in assignInitialValue()");
    }
    
    /** Properly assigns the value of the variable and remembers previous value
     *  without notifying any update observers
     * 
     * @param value
     */
    protected void assignValue(double value){
		if ((value < myLowerLimit) || (value > myUpperLimit))
			throw new IllegalArgumentException("Invalid argument. supplied value was not in range, [" + myLowerLimit + "," + myUpperLimit + "]");

		// record the new weight
        setWeight();
                
        // remember the old values
        myPrevValue = myValue;
        myPrevTimeOfChange = myTimeOfChange;
        
        // record the new value and time
        myValue = value;
        myTimeOfChange = getTime();   
        
//        System.out.println(getName() + " was assigned the value --> " + myValue);
    }
    
    /** Can be overridden by subclasses to set the
     *  protected variable myVariableTraceTextReport for appropriate
     *  text tracing
     */
    protected void setTraceObserver(String name){
        // lazy initialization, make trace report when trace is turned on.
        if (myVariableTraceTextReport == null){
            try{
            	if (name != null)
            		myVariableTraceTextReport = new VariableTraceTextReport(name + " " + getName()+ " Trace");
            	else
            		myVariableTraceTextReport = new VariableTraceTextReport(getName()+ " Trace");	
            }
            catch(IOException e){
            	String s = "I/O exception opening " + getName();
				if (JSL.LOGGER.isLoggable(java.util.logging.Level.SEVERE))
					JSL.LOGGER.severe(s);
				System.out.println(s);
                System.exit(1);

            }
        }
    }
  
    protected void beforeExperiment(){
    	// always assign the value of the variable to
    	// the provided initial value prior to the start
    	// of an experiment (the first of any replications)
        assignInitialValue(myInitialValue);
    }
    
    protected void initialize() { 
//		System.out.println("In Variable initialize() " + getName());
    	// initialize may or may not get called depending on whether
    	// the variable participates, if it does make sure that
    	// every replication starts with the same initial value
        assignInitialValue(myInitialValue);
        //if there are aggregates let them know about the initialization
        notifyAggregatesOfInitialization();
    }
    
    protected void warmUp(){
    	//if there are aggregates let them know about the warmUp
    	notifyAggregatesOfWarmUp();
    }
    
    protected void removedFromModel(){
    	super.removedFromModel();
    	myVariableTraceTextReport = null;
    }
    
    /** Gets the value of the flag that indicates whether or not observers
     *  will be notified when the value changes.
     * 
     * @return
     */
    public final boolean getSetValueUpdateNotificationFlag(){
    	return mySetValueUpdateObserversFlag;
    }
     
    /** Turns off or on the notification of update observers for changes
     *  to the value of the variable.  For example, this is useful to save the execution
     *  time associated with the statistical collection, etc on the variable
     *  Once it is off, it remains off until turned on. Thus multiple executions
     *  of the same model will remember this setting
     * 
     * @param flag
     *
     */
    public final void setValueUpdateNotificationFlag(boolean flag){
    	mySetValueUpdateObserversFlag = flag;
    }
    
}
