/*
* Created on Sep 9, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.variable.nhpp;

/**
 * @author rossetti
 *
 */
public class ConstantRateSegment implements RateSegmentIfc {

	protected double rate = 0; // the rate for the interval
	protected double crWidth = 0; // the width of the interval on the cumulative rate scale (crWidth = crUL - crLL)
	protected double crLL = 0; // the lower limit of the interval on cumulative rate scale
	protected double crUL = 0; // the upper limit of the interval on the cumulative rate scale
	protected double tWidth = 0; // the width of the interval on the time scale (tWidth = tUL - tLL)
	protected double tLL = 0; // the lower limit of the interval on the time scale
	protected double tUL = 0; // the upper limit of the interval on the time scale
	
	public ConstantRateSegment(double tLower, double duration, double rate, double crLower){

		if (tLower < 0.0)
		     throw new IllegalArgumentException("The lower time limit must be >= 0");

		if (crLower < 0.0)
		     throw new IllegalArgumentException("The lower rate limit must be >= 0");

		if (rate < 0.0)
		     throw new IllegalArgumentException("The rate must be > 0");
		
		if (rate >= Double.POSITIVE_INFINITY)
		     throw new IllegalArgumentException("The rate must be < infinity");

		if (duration <= 0.0)
		     throw new IllegalArgumentException("The duration must be > 0");
		
		if (Double.isInfinite(duration))
		     throw new IllegalArgumentException("The duration cannot be infinite.");
		
		tLL = tLower;
		tUL = tLL + duration;
		tWidth = duration;
		crLL = crLower;
		this.rate = rate;
		crUL = crLL + rate*tWidth;
		crWidth = crUL - crLL;		
	}

	public boolean contains(double time){
		return (tLL <= time) && (time < tUL);
	}
	
	/** Returns the rate for the interval
	 * 
	 * @return
	 */
	public double getRate(){
		return(rate);
	}
	
	/** Gets the rate for the time within the interval
	 * 
	 */
	public double getRate(double time){
		return rate;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.nhpp.RateSegmentIfc#getRateAtLowerTimeLimit()
	 */
	public double getRateAtLowerTimeLimit() {
		return rate;
	}

	/** The rate at the upper time limit is undefined Double.NaN.  The rate for the
	 *  segment is constant throughout the interval but undefined at the end of the interval
	 * 
	 */
	public double getRateAtUpperTimeLimit() {
		return Double.NaN;
	}
	
	/** The lower time limit
	 * 
	 * @return
	 */
	public double getLowerTimeLimit(){
		return(tLL);
	}
	
	/** The upper time limit
	 * 
	 * @return
	 */
	public double getUpperTimeLimit(){
		return(tUL);
	}
	
	/** The width of the interval
	 * 
	 * @return
	 */
	public double getTimeWidth(){
		return (tWidth);
	}
	
	/** The lower limit on the cumulative rate axis
	 * 
	 * @return
	 */
	public double getCumulativeRateLowerLimit(){
		return (crLL);
	}
	
	/** The upper limit on the cumulative rate axis
	 * 
	 * @return
	 */
	public double getCumulativeRateUpperLimit(){
		return(crUL);
	}
	
	/** The cumulative rate interval width
	 * 
	 * @return
	 */
	public double getCumulativeRateIntervalWidth(){
		return(crWidth);
	}
	
	/** Returns the value of the cumulative rate function for the interval
	 * given a value of time within that interval 
	 * 
	 * @param time
	 * @return
	 */
	public double getCumulativeRate(double time){
		double t = time - tLL;
		double mt = crLL + rate*t;
		return(mt);
	}

	/** Returns the inverse of the cumulative rate function given the interval
	 *  and a cumulative rate value within that interval.  Returns a time
	 * 
	 * @param cumRate
	 * @return
	 */
	public double getInverseCumulativeRate(double cumRate){
		if (rate == 0.0)
			return(Double.NaN);
		
		double t = cumRate - crLL;
		double inverse = tLL + t/rate;
		return(inverse);
	}
	
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append("rate = " + rate + " [" + tLL + "," + tUL + ")" + " width = " + tWidth );
		sb.append(" [" + crLL + "," + crUL + "]" + " cr width = " + crWidth + "\n");
		return(sb.toString());
	}

}
