/*
* Created on Sep 9, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.variable.nhpp;

import jsl.utilities.math.JSLMath;

/**
 * @author rossetti
 *
 */
public class LinearRateSegment implements RateSegmentIfc {

	protected double slope = 0; // the slope for the interval
	protected double rLL = 0; // the rate at the lower limit of the interval
	protected double rUL = 0; // the rate at the upper limit of the interval
	protected double crWidth = 0; // the width of the interval on the cumulative rate scale (crWidth = crUL - crLL)
	protected double crLL = 0; // the lower limit of the interval on cumulative rate scale
	protected double crUL = 0; // the upper limit of the interval on the cumulative rate scale
	protected double tWidth = 0; // the width of the interval on the time scale (tWidth = tUL - tLL)
	protected double tLL = 0; // the lower limit of the interval on the time scale
	protected double tUL = 0; // the upper limit of the interval on the time scale

	protected LinearRateSegment(double timeLL, double rateLL, double timeUL, double rateUL, double cumRate){
		tLL = timeLL;
		rLL = rateLL;
		tUL = timeUL;
		rUL = rateUL;
		tWidth = tUL - tLL;
		slope = (rUL - rLL)/tWidth;
		crLL = cumRate;
		crUL = crLL + (0.5)*(rUL + rLL)*(tUL - tLL);
		crWidth = crUL - crLL;
	}
	
	public boolean contains(double time){
		return (tLL <= time) && (time <= tUL);
	}
	
	public double getSlope(){
		return slope;
	}
	
	public double getRate(double time){
		if (JSLMath.equal(slope, 0.0)){
			return rLL;
		} else {
			return (rLL + slope*(time - tLL));
		}
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.nhpp.RateSegmentIfc#getRateAtLowerTimeLimit()
	 */
	public double getRateAtLowerTimeLimit() {
		return rLL;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.nhpp.RateSegmentIfc#getRateAtUpperTimeLimit()
	 */
	public double getRateAtUpperTimeLimit() {
		return rUL;
	}

	/** The lower time limit
	 * 
	 * @return
	 */
	public double getLowerTimeLimit(){
		return(tLL);
	}
	
	/** The upper time limit
	 * 
	 * @return
	 */
	public double getUpperTimeLimit(){
		return(tUL);
	}
	
	/** The width of the interval
	 * 
	 * @return
	 */
	public double getTimeWidth(){
		return (tWidth);
	}
	
	/** The lower limit on the cumulative rate axis
	 * 
	 * @return
	 */
	public double getCumulativeRateLowerLimit(){
		return (crLL);
	}
	
	/** The upper limit on the cumulative rate axis
	 * 
	 * @return
	 */
	public double getCumulativeRateUpperLimit(){
		return(crUL);
	}
	
	/** The cumulative rate interval width
	 * 
	 * @return
	 */
	public double getCumulativeRateIntervalWidth(){
		return(crWidth);
	}
	
	/** Returns the value of the cumulative rate function for the interval
	 * given a value of time within that interval 
	 * 
	 * @param time
	 * @return
	 */
	public double getCumulativeRate(double time){
		if (JSLMath.equal(slope, 0.0)){
			return (crLL + rLL*(time - tLL));
		} else {
			return (crLL + rLL*(time - tLL) + (0.5)*slope*(time - tLL)*(time - tLL));
		}
	}

	/** Returns the inverse of the cumulative rate function given the interval
	 *  and a cumulative rate value within that interval.  Returns a time
	 * 
	 * @param cumRate
	 * @return
	 */
	public double getInverseCumulativeRate(double cumRate){

		if (JSLMath.equal(slope, 0.0)){
			if (JSLMath.equal(rLL, 0.0))
				return Double.NaN;
			else {
				return (tLL + (cumRate - crLL)/rLL);					
			}
		} else {
			double n = 2.0*(cumRate - crLL);
			double d = rLL + Math.sqrt(rLL*rLL + 2.0*slope*(cumRate - crLL));
			return (tLL + n/d);
		}

	}
	
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append(" [" + rLL + "," + rUL + "]" + " slope = " + slope );
		sb.append(" [" + tLL + "," + tUL + "]" + " width = " + tWidth );
		sb.append(" [" + crLL + "," + crUL + "]" + " cr width = " + crWidth + "\n");
		return(sb.toString());
	}

}
