/*
* Created on Aug 30, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.variable.nhpp;

import jsl.utilities.random.distributions.Exponential;


/**
 * @author rossetti
 *
 */
public class PiecewiseConstantRateFunction extends PiecewiseRateFunction {
		
	/**
	 * 
	 */
	public PiecewiseConstantRateFunction(double duration, double rate) {
		super();
		
		// create the first segment
		
		ConstantRateSegment first = new ConstantRateSegment(0.0, duration, rate, 0.0);
		
		myRateSegments.add(first);
		
		if (rate > myMaxRate)
			myMaxRate = rate;
		
		if (rate < myMinRate)
			myMinRate = rate;

	}

	/** Allows the construction of the piecewise rate function starting at time zero.
	 *  The user supplies the arrival rate and the duration for that arrival
	 *  rate, by consecutive calls to addInterval().  
	 * 
	 * @param rate must be > 0, and less than Double.POSITIVE_INFINITY
	 * @param duration must be > 0
	 */
	public void addRateSegment(double duration, double rate){

		int k = myRateSegments.size();
		RateSegmentIfc prev = myRateSegments.get(k-1);
	
		ConstantRateSegment next = new ConstantRateSegment(prev.getUpperTimeLimit(), duration, rate, prev.getCumulativeRateUpperLimit());
		
		myRateSegments.add(next);
		
		if (rate > myMaxRate)
			myMaxRate = rate;
		
		if (rate < myMinRate)
			myMinRate = rate;

	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.variable.nhpp.RateFunctionIfc#contains(double)
	 */
	public boolean contains(double time){
		return (getTimeRangeLowerLimit() <= time) && (time < getTimeRangeUpperLimit());
	}
	
	/** Searches for the interval that the supplied time
	 *  falls within.  Returns -1 if no interval is found
	 *  
	 *  Interval indexing starts at index 0 (i.e. 0 is the first interval, 
	 *  silly Java zero based indexing)
	 * 
	 * @param time
	 * @return
	 */
	public final int findTimeInterval(double time){
		
		int k = -1;
		for(RateSegmentIfc i: myRateSegments){
			k = k + 1;
			if (time < i.getUpperTimeLimit())
				return (k);
		}
		return(-1);
	}
	
	public static void main(String[] args) {
		
//		test1();
		
//		test2();
		
		test3();
	}
	
	public static void test1(){
		PiecewiseConstantRateFunction f = new PiecewiseConstantRateFunction(15.0, 1.0);
		
		f.addRateSegment(20.0, 2.0);
		f.addRateSegment(15.0, 1.0);
		
		System.out.println("-----");
		System.out.println("intervals");
		System.out.println(f);
		
		// check the rate function
		System.out.println("-----");
		System.out.println("rate function");
		double mt = f.getTimeRangeUpperLimit();
		for(double t= 0.0;t<mt;t=t+1)
		 System.out.println("rate(" + t + ")= " + f.getRate(t));
		System.out.println("-----");
		System.out.println("cumulative rate function");
		// check the cumulative rate function

		for(double t= 0.0;t<mt;t=t+1)
		 System.out.println("cum rate(" + t + ")= " + f.getCumulativeRate(t));
		System.out.println("-----");
		System.out.println("inverse cumulative rate function");
		// check the cumulative rate function
		double mr = f.getCumulativeRateRangeUpperLimit();
		for(double r= 0.0;r<=mr;r=r+1)
		 System.out.println("inv cum rate(" + r + ")= " + f.getInverseCumulativeRate(r));		
	}
	
	public static void test2(){
		PiecewiseRateFunction f = new PiecewiseConstantRateFunction(15.0, 1.0);
		
		f.addRateSegment(20.0, 0.0);
		f.addRateSegment(15.0, 1.0);
		f.addRateSegment(20.0, 0.0);
		f.addRateSegment(15.0, 1.0);
		
		System.out.println("-----");
		System.out.println("intervals");
		System.out.println(f);
		
		// check the rate function
		System.out.println("-----");
		System.out.println("rate function");
		double mt = f.getTimeRangeUpperLimit();
		for(double t= 0.0;t<mt;t=t+1)
		 System.out.println("rate(" + t + ")= " + f.getRate(t));
		System.out.println("-----");
		System.out.println("cumulative rate function");
		// check the cumulative rate function

		for(double t= 0.0;t<mt;t=t+1)
		 System.out.println("cum rate(" + t + ")= " + f.getCumulativeRate(t));
		System.out.println("-----");
		System.out.println("inverse cumulative rate function");
		// check the cumulative rate function
		double mr = f.getCumulativeRateRangeUpperLimit();
		for(double r= 0.0;r<=mr;r=r+1)
		 System.out.println("inv cum rate(" + r + ")= " + f.getInverseCumulativeRate(r));		
	}

	
	public static void test3(){
		PiecewiseConstantRateFunction f = new PiecewiseConstantRateFunction(15.0, 1.0);
		
		f.addRateSegment(20.0, 2.0);
		f.addRateSegment(15.0, 1.0);
		
		System.out.println("-----");
		System.out.println("intervals");
		System.out.println(f);
		
		// check the rate function
		System.out.println("-----");
		System.out.println("Generate non-stationary Poisson process.");
		double a = 0.0;
		double y = 0.0;
		double tau = f.getCumulativeRateRangeUpperLimit();
		Exponential e = new Exponential();
		int n = 0;
		y = y + e.getValue();
		while (y < tau ){
			a = f.getInverseCumulativeRate(y);
			n++;
			System.out.println("a[" + n + "] = "+ a);
			y = y + e.getValue();
		}
	}
}
