/*
* Created on Aug 30, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.modeling.elements.variable.nhpp;

/**
 * @author rossetti
 *
 */
public class PiecewiseLinearRateFunction extends PiecewiseRateFunction {
			
	public PiecewiseLinearRateFunction(double duration, double secondRate){
		this(0.0, duration, secondRate);
	}
	
	/** Constructs a PiecewiseLinearRateFunction with the first rate
	 *  specified for time zero
	 * 
	 * @param firstRate The rate at time zero
	 * @param duration The duration until the second rate
	 * @param secondRate The rate after the supplied duratioin
	 */
	public PiecewiseLinearRateFunction(double firstRate, double duration, double secondRate) {
		super();
		
		if (firstRate < 0.0)
		     throw new IllegalArgumentException("The rate must be >= 0");
		
		if (firstRate >= Double.POSITIVE_INFINITY)
		     throw new IllegalArgumentException("The rate must be < infinity");

		if (firstRate > myMaxRate)
			myMaxRate = firstRate;
		
		if (firstRate < myMinRate)
			myMinRate = firstRate;

		if (secondRate < 0.0)
		     throw new IllegalArgumentException("The rate must be >= 0");
		
		if (secondRate >= Double.POSITIVE_INFINITY)
		     throw new IllegalArgumentException("The rate must be < infinity");

		if (secondRate > myMaxRate)
			myMaxRate = secondRate;
		
		if (secondRate < myMinRate)
			myMinRate = secondRate;

		if (duration <= 0.0)
		     throw new IllegalArgumentException("The duration must be > 0");
		
		if (Double.isInfinite(duration))
		     throw new IllegalArgumentException("The duration cannot be infinite.");

		LinearRateSegment first = new LinearRateSegment(0.0, firstRate, duration, secondRate, 0.0);
		
		myRateSegments.add(first);

	}

	/** Allows the construction of the piecewise linear rate function starting at time zero.
	 *  The user supplies the knot points on the piecewise linear function by supplying
	 *  the rate at the end of the supplied duration via consecutive calls to addInterval().  
	 * 
	 * @param duration must be > 0 and less than Double.POSITIVE_INFINITY
	 * @param rate must be >= 0, and less than Double.POSITIVE_INFINITY
	 */
	public final void addRateSegment(double duration, double rate){
		
		if (rate < 0.0)
		     throw new IllegalArgumentException("The rate must be > 0");
		
		if (rate >= Double.POSITIVE_INFINITY)
		     throw new IllegalArgumentException("The rate must be < infinity");
		
		if (duration <= 0.0)
		     throw new IllegalArgumentException("The duration must be > 0");
		
		if (Double.isInfinite(duration))
		     throw new IllegalArgumentException("The duration cannot be infinite.");
		
		if (rate > myMaxRate)
			myMaxRate = rate;
		
		if (rate < myMinRate)
			myMinRate = rate;

		int k = myRateSegments.size();
		// get the last interval
		RateSegmentIfc last = myRateSegments.get(k-1);
	
		RateSegmentIfc next = new LinearRateSegment(last.getUpperTimeLimit(), last.getRateAtUpperTimeLimit(),
				last.getUpperTimeLimit() + duration, rate, last.getCumulativeRateUpperLimit());
		
		myRateSegments.add(next);
		
	}
		
	public static void main(String[] args) {
		
		test1();
		
		
	}
	
	public static void test1(){
		PiecewiseRateFunction f = new PiecewiseLinearRateFunction(0.5, 200.0, 0.5);
		
		f.addRateSegment(400.0, 0.9);
		f.addRateSegment(400.0, 0.9);
		f.addRateSegment(200.0, 1.2);
		f.addRateSegment(300.0, 0.9);
		f.addRateSegment(500.0, 0.5);

		System.out.println("-----");
		System.out.println("intervals");
		System.out.println(f);
		
		// check the rate function
		System.out.println("-----");
		System.out.println("rate function");
		double mt = f.getTimeRangeUpperLimit();
		for(double t= 0.0;t<mt;t=t+1)
		 System.out.println("rate(" + t + ")= " + f.getRate(t));
		System.out.println("-----");
		System.out.println("cumulative rate function");
		// check the cumulative rate function

		for(double t= 0.0;t<mt;t=t+1)
		 System.out.println("cum rate(" + t + ")= " + f.getCumulativeRate(t));
		System.out.println("-----");
		System.out.println("inverse cumulative rate function");
		// check the cumulative rate function
		double mr = f.getCumulativeRateRangeUpperLimit();
		for(double r= 0.0;r<=mr;r=r+1)
		 System.out.println("inv cum rate(" + r + ")= " + f.getInverseCumulativeRate(r));		
	}

}
