/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.observers.database.access;

import java.sql.*;
import java.util.*;

import jsl.utilities.statistic.*;
import jsl.modeling.*;
import jsl.modeling.elements.variable.*;
import jsl.observers.variable.CounterObserver;
import jsl.observers.variable.StatisticalObserver;

/**
 *
 */
public class AccessDBReplicationReport extends AccessDBReport {
    
    /** Creates a new instance of AccessDBReplicationReport **/
    public AccessDBReplicationReport(String name){
        super(name);
    }
        
    protected void afterReplication(ModelElement m, Object arg){

    	Model model = m.getModel();
        Replication r = model.getReplication();
		Experiment e = r.getExperiment();
        double nreps = r.getReplicationNumber();
        
        List<ResponseVariable> rvs = model.getResponseVariables();
        
        for(ResponseVariable rv: rvs){
            String type;
            if (rv instanceof TimeWeighted)
                type = "Time-Persistent";
            else
                type = "Observational";
            
            StatisticalObserver so = rv.getStatisticalObserver();
            StatisticAccessorIfc stat = so.getWithinReplicationStatistic();
			insertRow(stat, e, (long)rv.getId(), nreps, type);        	
        }
       
		List<Counter> counters = r.getModel().getCounters();
		for(Counter c: counters){
			CounterObserver co = c.getCounterObserver();
			StatisticAccessorIfc tus = co.getAcrossTimedUpdateStatistic();
			if (tus != null)
				insertRow(tus, e, (long)c.getId(), nreps, "Observational");			
		}
		
		if (!counters.isEmpty()){
			for(Counter c: counters){
				insertCounterRow(c, e, nreps);
			}
		}

    }
    
    protected void afterExperiment(ModelElement m, Object arg){
        close();    
    }
 
    public void insertCounterRow(Counter c, Experiment e, double nreps){
    	
		String sqlStmt = "INSERT INTO tblWithinReplicationStatistics (ID, SetupNum, RepNum, Name, StatType, Count) VALUES (" + c.getId() + ",";
		
		if (e != null){
    		sqlStmt = sqlStmt + e.getId() + ",";
    	} else {
    		sqlStmt = sqlStmt + "" + ",";
    	}
		String type = "Counter";
		
		sqlStmt = sqlStmt + nreps + ",'" + c.getName() + "','" + type + "'," + c.getValue()+ ")";;

        try {
            myInsertRowStmt.executeUpdate(sqlStmt);
        } catch(SQLException ex) {
            System.out.println("\n--- SQLException caught ---\n");
            System.out.println("in insertCounterRow()");
            System.out.println("The SQL Statement was:");
            System.out.println(sqlStmt);
            while (ex != null) {
                System.out.println("Message:   "
                + ex.getMessage());
                System.out.println("SQLState:  "
                + ex.getSQLState());
                System.out.println("ErrorCode: "
                + ex.getErrorCode());
                ex = ex.getNextException();
                System.out.println("");
            }
        }
    }

	public String makeInsertSQLStatement(StatisticAccessorIfc stat, long setupnum, long id, double nreps, String type){
		StringBuilder sb = new StringBuilder();
		sb.append("INSERT INTO tblWithinReplicationStatistics ");
		sb.append("(ID, SetupNum, RepNum, Name, StatType, Count, Average, Variance, StandardDeviation, StandardError,");
		sb.append("Minimum, Maximum, Sum, WeightedSum, WeightedSumOfSquares, SumOfWeights, WeightedAverage,");
		sb.append("DeviationSumOfSquares, LastValue, LastWeight, Kurtosis, Skewness, Lag1Correlation, Lag1Covariance, VonNeumannLag1TestStat, NumMissingObservations) ");
		sb.append(" VALUES (" + id + "," + setupnum + "," + nreps + ",'" + stat.getName() + "','" + type + "',");
		makeCSVStatistic(stat, sb);
		sb.append(")");
		return(sb.toString());
	}
	
	public String makeInsertSQLStatement(StatisticAccessorIfc stat, long id, double nreps, String type){
		StringBuilder sb = new StringBuilder();
		sb.append("INSERT INTO tblWithinReplicationStatistics VALUES (");
		sb.append("(ID, SetupNum, RepNum, Name, StatType, Count, Average, Variance, StandardDeviation, StandardError,");
		sb.append("Minimum, Maximum, Sum, WeightedSum, WeightedSumOfSquares, SumOfWeights, WeightedAverage,");
		sb.append("DeviationSumOfSquares, LastValue, LastWeight, Kurtosis, Skewness, Lag1Correlation, Lag1Covariance, VonNeumannLag1TestStat, NumMissingObservations) ");
		sb.append(" VALUES (" + id + "," + "" + "," + nreps + ",'" + stat.getName() + "','" + type + "',");
		makeCSVStatistic(stat, sb);
		sb.append(")");
		return(sb.toString());
	}
	  
	public void insertRow(StatisticAccessorIfc stat, Experiment e, long id, double nreps, String type) {
        
		String sqlStmt;
        
		if (e == null)
			sqlStmt = makeInsertSQLStatement(stat, id, nreps, type);
		else
			sqlStmt = makeInsertSQLStatement(stat, e.getId(), id, nreps, type);        
        
        try {
            myInsertRowStmt.executeUpdate(sqlStmt);
        } catch(SQLException ex) {
            System.out.println("\n--- SQLException caught ---\n");
            System.out.println("in insertRow()");
            System.out.println("The SQL Statement was:");
            System.out.println(sqlStmt);
            while (ex != null) {
                System.out.println("Message:   "
                + ex.getMessage());
                System.out.println("SQLState:  "
                + ex.getSQLState());
                System.out.println("ErrorCode: "
                + ex.getErrorCode());
                ex = ex.getNextException();
                System.out.println("");
            }
            
        }
    }
}
