/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.observers.textfile;

import java.io.IOException;
import java.util.*;

import jsl.modeling.elements.variable.ResponseVariable;
import jsl.observers.variable.StatisticalObserver;
import jsl.utilities.statistic.Statistic;
import jsl.utilities.statistic.StatisticAccessorIfc;
import jsl.modeling.*;

/**
 */
public class AntitheticExperimentReport extends StatisticalReport implements Observer {

	protected List<ResponseVariable> myResponseVariables;
	protected double[] myOddAvg;
	protected List<Statistic> myAcrossPairStats;
	
	/**
	 * @param name
	 * @throws IOException
	 */
	public AntitheticExperimentReport(String name) throws IOException {
		this(null, name);
	}
	
	/**
	 * @param name
	 * @throws IOException
	 */
	public AntitheticExperimentReport(String directory, String name) throws IOException {
		super(directory, name);
		myResponseVariables = new ArrayList<ResponseVariable>();
		myAcrossPairStats = new ArrayList<Statistic>();
	}

	/** Returns an unmodifiable list of all the across pair antithetic statistics
	 *  associated with the response variables in the model
	 * 
	 * @return
	 */
	public final List<Statistic> getAntitheticStatistics(){
		return(Collections.unmodifiableList(myAcrossPairStats));
	}
	
	/* (non-Javadoc)
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	public void update(Observable experiment, Object arg1) {
		Experiment e = (Experiment)experiment;
				
		if (e.isInitialized()){
			Model m = e.getModel();
			myResponseVariables.clear();
			m.getResponseVariables(myResponseVariables);
			myOddAvg = new double[myResponseVariables.size()];
			myAcrossPairStats.clear();
			for(ResponseVariable rv: myResponseVariables){
				StatisticalObserver so = rv.getStatisticalObserver();
				Statistic stat = new Statistic(so.getName() + " AV Stat");
				myAcrossPairStats.add(stat);
			}
			addLines("Experiment: " + e.getName());
			addLines("Time of Report: " + new Date());
		}
 		
		if (e.isStepCompleted()){// indicates that a replication has just ended
			Replication r = e.getCurrentReplication();
			int repNum = (int)r.getReplicationNumber();
			if ((repNum % 2) != 0){// odd replication
				// save the odd averages in the array
				int i=0;
				for(ResponseVariable rv: myResponseVariables){
					StatisticalObserver so = rv.getStatisticalObserver();
					StatisticAccessorIfc stat = so.getWithinReplicationStatistic();
					myOddAvg[i] = stat.getWeightedAverage();
					i++;
				}
			} else {// even replication
				// collect the pair average
				int i=0;
				for(ResponseVariable rv: myResponseVariables){
					StatisticalObserver so = rv.getStatisticalObserver();
					StatisticAccessorIfc stat = so.getWithinReplicationStatistic();
					double y = stat.getWeightedAverage();
					double value = (myOddAvg[i] + y)/2.0;
					Statistic pairStat = myAcrossPairStats.get(i);
					pairStat.collect(value);
					i++;
				}
			}
		}
		
		if (e.isEnded()){
			addLines("Number of replications: " + e.getNumberOfReplications());
			addLines("Run length for each replication: " + e.getLengthOfReplication());
			addLines("Warm up for each replication: " + e.getLengthOfWarmUp());
			long et = e.getMaximumAllowedExecutionTime();
			if (et == 0)
				addLines("Maximum allowed experiment execution time not specified.");
			else
				addLines("Maximum allowed experiment execution time: " + et + " milliseconds.");
			et = e.getMaximumAllowedExecutionTimePerReplication();
			if (et == 0)
				addLines("Maximum allowed execution time per replication not specified.");
			else
				addLines("Maximum allowed execution time per replication: " + et + " milliseconds.");
			if (e.getDefaultBatchingOption())
				addLines("Batch interval size: " + e.getDefaultBatchInterval());
			skipLine();		
			long t = e.getEndExecutionTime() - e.getBeginExecutionTime();
			addLines("The total time was approximately " + t + " milliseconds");

			if (e.isCompleted()){
				addLines("The experiment ran all replications.");
			}
 			
			if (e.isTimedOut()){
				addLines("The experiment timed out.");
			}
 			
			if (e.isEndConditionMet()){
				addLines("The experiment ended because its end condition was met.");
				addLines(e.getEndCondition());
			}
 			
			if (e.isUnfinished()){
				addLines("The experiment was ended early.");
			}

			skipLine();
			skipLine();
			addLines("----------------------------------------------------------");
			addLines("Antithetic Experiment Statistics");
			addLines("----------------------------------------------------------");
			skipLine();
			
			
		    for(Statistic stat: myAcrossPairStats){
		    		addLines(stat);
		    }

			close();
		}

	}

}
