/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.observers.textfile;

import java.util.*;
import java.io.IOException;

import jsl.modeling.*;

import jsl.modeling.elements.variable.*;
import jsl.observers.variable.CounterObserver;
import jsl.utilities.statistic.StatisticAccessorIfc;

public class BatchReport extends StatisticalReport implements Observer {

	//  ===========================================
	//      CLASS AND OBJECT ATTRIBUTES
	//  ===========================================

	//  ===========================================
	//        CONSTRUCTORS
	//  ===========================================
 
	public BatchReport(String name) throws IOException {
		this(null, name);
	}

	public BatchReport(String directory, String name) throws IOException {
		super(directory, name);
	}
	
	public void update(java.util.Observable observable, java.lang.Object obj) {
		Replication r = (Replication) observable;

		if (r.isInitialized()) {
			addLines("Batch Report: " + r.getName());
			addLines("Time of Report: " + new Date());
			skipLine();
		}

		if (r.isEnded()) {
			addLines("Planned Run length for replication: " + r.getLengthOfReplication());
			addLines("Warm up for replication: " + r.getLengthOfWarmUp());
			addLines("The actual run length for replication: " + r.getEndingLengthOfReplication());
			long et = r.getMaximumAllowedExecutionTime();
			if (et == 0)
				addLines("Maximum allowed execution time not specified.");
			else
				addLines("Maximum allowed execution time: " + et + " milliseconds.");
			skipLine();
			long t = r.getEndExecutionTime() - r.getBeginExecutionTime();
			addLines("The total execution time was approximately " + t/1000.0 + " seconds");
				
			if (r.getDefaultBatchingOption())
				addLines("Batch interval size: " + r.getDefaultBatchInterval());

			if (r.isCompleted()) {
				addLines("The replication ran all its events.");
			}

			if (r.isTimedOut()) {
				addLines("The replication timed out.");
			}

			if (r.isEndConditionMet()) {
				addLines("The replication ended because its end condition was met.");
				addLines(r.getEndCondition());
			}

			if (r.isUnfinished()) {
				addLines("The replication was ended early.");
			}

			List<ResponseVariable> rvs = r.getModel().getResponseVariables();

			if (!rvs.isEmpty()){
				skipLine();
				skipLine();
			}
			
			for(ResponseVariable rv: rvs){
				if (rv.getDefaultReportingOption())
					addLines(rv.getStatisticalObserver());				
			}

			List<Counter> counters = r.getModel().getCounters();

			if (!counters.isEmpty()){
				skipLine();
				skipLine();
				addLines("----------------------------------------------------------");
				addLines("Counter statistics:");
				addLines("----------------------------------------------------------");
				skipLine();
							
				for(Counter c: counters){
					if (c.getDefaultReportingOption()){
				        String s;
						s = "Name " + c.getName() + "\n";
				        s = s + "Final Counter Value " + c.getValue() + "\n";
						addLines(s);						
					}
				}

				for(Counter c: counters){
					if (c.getDefaultReportingOption()){
						CounterObserver co = c.getCounterObserver();
						StatisticAccessorIfc tus = co.getAcrossTimedUpdateStatistic();
						if (tus != null){
							addLines(tus);
						}						
					}		
				}				
			}
			
			writeOutModelElementObservers();
			
			skipLine();
			skipLine();
			addLines("----------------------------------------------------------");
			addLines("Scheduler statistics");
			addLines("----------------------------------------------------------");
			skipLine();
			addLines(r.getSchedulerStatistics());
			skipLine();
	
		}
	}
}
