/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.observers.textfile;

import java.io.IOException;
import java.util.*;

import jsl.utilities.statistic.*;
import jsl.modeling.*;
import jsl.modeling.elements.variable.*;
import jsl.observers.variable.CounterObserver;
import jsl.observers.variable.StatisticalObserver;

/**
 *
 */
public class CSVReplicationReport extends CSVReport {
  
	public CSVReplicationReport(String name) throws IOException {
		this(null, name);
	}

	public CSVReplicationReport(String directory, String name) throws IOException {
		super(directory, name);
	}
	
    protected void afterReplication(ModelElement m, Object arg){
    	Model model = m.getModel();
        Replication r = model.getReplication();
		Experiment e = r.getExperiment();
        double nreps = r.getReplicationNumber();
        
        List<ResponseVariable> rvs = model.getResponseVariables();
        
        for(ResponseVariable rv: rvs){
        	if (rv.getDefaultReportingOption()){
                String type;
                if (rv instanceof TimeWeighted)
                    type = "Time-Persistent";
                else
                    type = "Observational";
                
                StatisticalObserver so = rv.getStatisticalObserver();
                StatisticAccessorIfc stat = so.getWithinReplicationStatistic();
    			writeRow(stat, e, (long)rv.getId(), nreps, type);         		
        	}
        }
        
		List<Counter> counters = r.getModel().getCounters();
		for(Counter c: counters){
			if (c.getDefaultReportingOption()){
				CounterObserver co = c.getCounterObserver();
				StatisticAccessorIfc tus = co.getAcrossTimedUpdateStatistic();
				if (tus != null)
					writeRow(tus, e, (long)c.getId(), nreps, "Observational");					
			}	
		}
		
		if (!counters.isEmpty()){
			for(Counter c: counters){
				if (c.getDefaultReportingOption()){
					String s = c.getId() + ",";
					if (e != null)
						s = s + e.getId() + "," + e.getName() + ",";
					else
						s = s + "" + "," + "" + ",";
					s = s + nreps + ",\"" + c.getName() + "\",\"" + "Counter" + "\"," + c.getValue();
					myWriter.println(s);					
				}
			}
		}
    }

	public void writeRow(StatisticAccessorIfc stat, Experiment e, long id, double nreps, String type) {
	    
		String row;
	    
		if (e == null)
			row = makeRepRow(stat, id, nreps, type);
		else
			row = makeExpRow(stat, e, id, nreps, type);  
		
		myWriter.println(row);
	    
	}

	public String makeExpRow(StatisticAccessorIfc stat, Experiment e, long id, double nreps, String type) {
		
		StringBuilder sb = new StringBuilder();
		sb.append(id);
		sb.append(",");
		sb.append(e.getId());
		sb.append(",");
		sb.append(e.getName());
		sb.append(",");
		sb.append(nreps);
		sb.append(",\"");
		sb.append(stat.getName());
		sb.append("\",\"");
		sb.append(type);
		sb.append("\",");
		sb.append(stat.getCount());
		sb.append(",");
		sb.append(stat.getAverage());
		sb.append(",");
		sb.append(stat.getStandardDeviation());
		sb.append(",");
		sb.append(stat.getStandardError());
		sb.append(",");
		sb.append(stat.getHalfWidth());
		sb.append(",");
		sb.append(stat.getConfidenceCoefficient());
		sb.append(",");
		sb.append(stat.getMin());
		sb.append(",");
		sb.append(stat.getMax());
		sb.append(",");
		sb.append(stat.getSum());
		sb.append(",");
		sb.append(stat.getVariance());
		sb.append(",");
		sb.append(stat.getWeightedAverage());
		sb.append(",");
		sb.append(stat.getWeightedSum());
		sb.append(",");
		sb.append(stat.getSumOfWeights());
		sb.append(",");
		sb.append(stat.getWeightedSumOfSquares());
		sb.append(",");
		sb.append(stat.getDeviationSumOfSquares());
		sb.append(",");
		sb.append(stat.getLastValue());
		sb.append(",");
		sb.append(stat.getLastWeight());
		sb.append(",");
		sb.append(stat.getKurtosis());
		sb.append(",");
		sb.append(stat.getSkewness());
		sb.append(",");
		sb.append(stat.getLag1Covariance());
		sb.append(",");
		sb.append(stat.getLag1Correlation());
		sb.append(",");
		sb.append(stat.getVonNeumannLag1TestStatistic());
		sb.append(",");
		sb.append(stat.getNumberMissing());
		return(sb.toString());		

	}

	public String makeRepRow(StatisticAccessorIfc stat, long id, double nreps, String type) {
		
		StringBuilder sb = new StringBuilder();
		sb.append(id);
		sb.append(",");
		sb.append("");
		sb.append(",");
		sb.append("");
		sb.append(",");
		sb.append(nreps);
		sb.append(",\"");
		sb.append(stat.getName());
		sb.append("\",\"");
		sb.append(type);
		sb.append("\",");
		sb.append(stat.getCount());
		sb.append(",");
		sb.append(stat.getAverage());
		sb.append(",");
		sb.append(stat.getStandardDeviation());
		sb.append(",");
		sb.append(stat.getStandardError());
		sb.append(",");
		sb.append(stat.getHalfWidth());
		sb.append(",");
		sb.append(stat.getConfidenceCoefficient());
		sb.append(",");
		sb.append(stat.getMin());
		sb.append(",");
		sb.append(stat.getMax());
		sb.append(",");
		sb.append(stat.getSum());
		sb.append(",");
		sb.append(stat.getVariance());
		sb.append(",");
		sb.append(stat.getWeightedAverage());
		sb.append(",");
		sb.append(stat.getWeightedSum());
		sb.append(",");
		sb.append(stat.getSumOfWeights());
		sb.append(",");
		sb.append(stat.getWeightedSumOfSquares());
		sb.append(",");
		sb.append(stat.getDeviationSumOfSquares());
		sb.append(",");
		sb.append(stat.getLastValue());
		sb.append(",");
		sb.append(stat.getLastWeight());
		sb.append(",");
		sb.append(stat.getKurtosis());
		sb.append(",");
		sb.append(stat.getSkewness());
		sb.append(",");
		sb.append(stat.getLag1Covariance());
		sb.append(",");
		sb.append(stat.getLag1Correlation());
		sb.append(",");
		sb.append(stat.getVonNeumannLag1TestStatistic());
		sb.append(",");
		sb.append(stat.getNumberMissing());
		return(sb.toString());
	}
 
}
