/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.observers.textfile;

import java.util.Collection;
import java.util.ArrayList;
import java.util.List;
import java.util.Observer;
import java.io.IOException;

import jsl.modeling.*;

import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.observers.variable.CounterObserver;
import jsl.observers.variable.StatisticalObserver;
import jsl.utilities.statistic.StatisticAccessorIfc;
import jsl.utilities.reporting.TextReport;

public class MReplicationReport extends TextReport implements Observer {

    //  ===========================================
    //      CLASS AND OBJECT ATTRIBUTES
    //  ===========================================
    
    protected Collection<ResponseVariable> myResponseVariables;
    protected Model myModel;
    
    //  ===========================================
    //        CONSTRUCTORS
    //  ===========================================
 
    public MReplicationReport(String name) throws IOException{
        this(null, name);
    }
    
    public MReplicationReport(String directory, String name) throws IOException{
        super(directory, name);
        myResponseVariables = new ArrayList<ResponseVariable>();
    }
    
    public void update(java.util.Observable observable, java.lang.Object obj) {
        String s;
        myModel = (Model)observable;
        
        if (myModel.checkForBeforeExperiment()){
            addHeaderLines();
            addLines(myModel);
            myModel.getResponseVariables(myResponseVariables);
        }
        
        if (myModel.checkForBeforeReplication()){
        	Replication r = myModel.getReplication();
            s = "Starting replication " + r.getReplicationNumber() +"\n";
            s = s + "Planned replication length " + r.getLengthOfReplication() + "\n";
            s = s + "Warmup time " + r.getLengthOfWarmUp() + "\n";
            addLines(s);
        }
        
        if (myModel.checkForAfterReplication()){
			Replication r = myModel.getReplication();
            s = "Ending replication " + r.getReplicationNumber() +"\n";
            s = s + "Ending time " + ModelElement.getTime() + "\n";
            
            addLines(s);
            
			List<ResponseVariable> rvs = r.getModel().getResponseVariables();
			
			if (!rvs.isEmpty()){
				skipLine();
				skipLine();
				addLines("----------------------------------------------------------");
				addLines("Within Replication statistics:");
				addLines("----------------------------------------------------------");
				skipLine();
				
				for(ResponseVariable rv: rvs){
					if (rv.getDefaultReportingOption()){
						StatisticalObserver so = rv.getStatisticalObserver();
						addLines(so.getWithinReplicationStatistic());				
					}
				}				
			}

			List<Counter> counters = r.getModel().getCounters();

			if (!counters.isEmpty()){
				skipLine();
				skipLine();
				addLines("----------------------------------------------------------");
				addLines("Counter statistics:");
				addLines("----------------------------------------------------------");
				skipLine();
							
				for(Counter c: counters){
					if (c.getDefaultReportingOption()){
						s = "Name " + c.getName() + "\n";
				        s = s + "Final Counter Value " + c.getValue() + "\n";
						addLines(s);
					}	
				}

				for(Counter c: counters){
					if (c.getDefaultReportingOption()){
						CounterObserver co = c.getCounterObserver();
						StatisticAccessorIfc tus = co.getAcrossTimedUpdateStatistic();
						if (tus != null){
							addLines(tus);
						}							
					}
				}				
			}

        }
    }    
}

