/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.observers.textfile;

import java.util.*;
import java.io.IOException;

import jsl.modeling.*;

import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.observers.variable.CounterObserver;
import jsl.observers.variable.StatisticalObserver;
import jsl.utilities.statistic.StatisticAccessorIfc;

public class ReplicationReport extends StatisticalReport implements Observer {

	// ===========================================
	// CLASS AND OBJECT ATTRIBUTES
	// ===========================================

	// ===========================================
	// CONSTRUCTORS
	// ===========================================

	public ReplicationReport(String name) throws IOException {
		this(null, name);
	}

	public ReplicationReport(String directory, String name) throws IOException {
		super(directory, name);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	public void update(Observable experiment, Object arg1) {
		Experiment e = (Experiment) experiment;

		if (e.isStepCompleted()) {
			skipLine();
			Replication r = e.getCurrentReplication();
			addLines("Experiment Name: " + e.getName());
			addLines("Time of Report: " + new Date());
			skipLine();
			addLines("Replication Name: " + r.getName());
			addLines("Completed Replication: " + r.getReplicationNumber() + " of "
					+ e.getNumberOfReplications());
			addLines("Planned replication run length: "
					+ r.getLengthOfReplication());
			addLines("Warm up for replication: " + r.getLengthOfWarmUp());
			long et = r.getMaximumAllowedExecutionTime();
			if (et == 0)
				addLines("Maximum allowed execution time not specified.");
			else
				addLines("Maximum allowed execution time: " + et
						+ " milliseconds.");
			if (r.getDefaultBatchingOption())
				addLines("Batch interval size: " + r.getDefaultBatchInterval());
			addLines("Actual simulation run length "
					+ r.getEndingLengthOfReplication());
			skipLine();
			long t = r.getEndExecutionTime() - r.getBeginExecutionTime();
			addLines("The total time was approximately " + t + " milliseconds");

			if (r.isCompleted()) {
				addLines("The replication ran to completion.");
			}

			if (r.isTimedOut()) {
				addLines("The replication timed out.");
			}

			if (r.isEndConditionMet()) {
				addLines("The replication ended because its end condition was met.");
				addLines(r.getEndCondition());
			}

			if (r.isUnfinished()) {
				addLines("The replication was ended early.");
			}

			List<ResponseVariable> rvs = r.getModel().getResponseVariables();
			
			if (!rvs.isEmpty()){
				skipLine();
				skipLine();
				addLines("----------------------------------------------------------");
				addLines("Within Replication statistics:");
				addLines("----------------------------------------------------------");
				skipLine();
				
				for(ResponseVariable rv: rvs){
					if (rv.getDefaultReportingOption()){
						StatisticalObserver so = rv.getStatisticalObserver();
						addLines(so.getWithinReplicationStatistic());				
					}
				}				
			}

			List<Counter> counters = r.getModel().getCounters();

			if (!counters.isEmpty()){
				skipLine();
				skipLine();
				addLines("----------------------------------------------------------");
				addLines("Counter statistics:");
				addLines("----------------------------------------------------------");
				skipLine();
							
				for(Counter c: counters){
					if (c.getDefaultReportingOption()){
				        String s;
						s = "Name " + c.getName() + "\n";
				        s = s + "Final Counter Value " + c.getValue() + "\n";
						addLines(s);
					}	
				}

				for(Counter c: counters){
					if (c.getDefaultReportingOption()){
						CounterObserver co = c.getCounterObserver();
						StatisticAccessorIfc tus = co.getAcrossTimedUpdateStatistic();
						if (tus != null){
							addLines(tus);
						}							
					}
				}				
			}
			
			writeOutModelElementObservers();

			skipLine();
			skipLine();
			addLines("----------------------------------------------------------");
			addLines("Scheduler statistics");
			addLines("----------------------------------------------------------");
			skipLine();
			addLines(r.getSchedulerStatistics());
			skipLine();
		}
	}
}
