/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.observers.variable;

import jsl.utilities.statistic.*;
import jsl.utilities.reporting.*;
import jsl.modeling.elements.variable.*;
import jsl.modeling.*;
import jsl.observers.ModelElementObserver;

/**
 *
 */
public class StatisticalObserver extends ModelElementObserver {

	/** indicates the count when the warm up should occur **/
	private long myCountWarmUpLimit = 0;

	/** indicates the count when the simulation should stop **/
	private long myCountStopLimit = 0;

	/** indicates that the count based warm up has happened **/
	private boolean myCountWarmUpActionFlag = false;

	/** indicates that the count based stop has happened **/
	private boolean myCountStopActionFlag = false;

	/** Minimum batch size used when batching
	 */
	protected int myMinBatchSize = BatchStatistic.MIN_NUM_OBS_PER_BATCH;
	
	/** Minimum number of batches used when batching
	 */	
	protected int myMinNumBatches = BatchStatistic.MIN_NUM_BATCHES;
	
	/** Maximum number of batches used when batching
	 */	
	protected int myMaxNumBatchesMultiple = BatchStatistic.MAX_BATCH_MULTIPLE;

	/** Collects within replication statistics
	 */
	protected Statistic myWithinRepStat;

	/** Collects across replication statistics
	 */
	protected Statistic myAcrossRepStat;

	/** Collects within replication batch statistics
	 */
	protected BatchStatistic myBatchStatistic;

	/** The response variable being observed
	 */
	protected ResponseVariable myResponseVariable;
	
	/** The model associated with the response variable
	 */
	protected Model myModel;
	
	/** Creates a new instance of StatisticalObserver */
	public StatisticalObserver() {
		this(null);
	}

	/** Creates a new instance of StatisticalObserver */
	public StatisticalObserver(String name) {
		super(name);
		myWithinRepStat = new Statistic("Within Rep Stat " + name);
		myAcrossRepStat = new Statistic("Across Rep Stat " + name);
	}

	/** Gets the statistics that have been accumulated within the replication
	 *  for this variable.
	 *
	 * @return A StatisticAccessorIfc representing the within replication statistics.
	 */
	public StatisticAccessorIfc getWithinReplicationStatistic() {
		return (myWithinRepStat);
	}

	/** A convenience method to set the name of the underlying Statistic
	 *  used to collect within replication statistics
	 * 
	 * @param name
	 */
	public void setWithinReplicationStatisticName(String name){
		myWithinRepStat.setName(name);
	}
	
	/** Gets the statistics that have been accumulated across all replications
	 *  for this variable.
	 *
	 * @return A StatisticAccessorIfc representing the across replication statistics.
	 */
	public StatisticAccessorIfc getAcrossReplicationStatistic() {
		return (myAcrossRepStat);
	}

	/** A convenience method to set the name of the underlying Statistic
	 *  for tabulating across replication statistics
	 * 
	 * @param name
	 */
	public void setAcrossReplicationStatisticName(String name){
		myAcrossRepStat.setName(name);
	}
	
	/** Gets the statistics that have been accumulated across the
	 *  batches for this ResponseVariable.  This may be null if batching was not enabled
	 *  for the observed response variable
	 *
	 * @return A StatisticAccessorIfc representing the batching statistics.
	 */
	public StatisticAccessorIfc getAcrossBatchStatistic() {
		return (myBatchStatistic);
	}

	/** A convenience method to set the name of the underlying Statistic
	 *  for tabulating across batch statistics.  If batching is not on
	 *  this method does nothing.  If the ResponseVariable that is being observed
	 *  has not initialized yet, then this method does nothing.
	 * 
	 * @param name
	 */
	public void setAcrossBatchStatisticName(String name){
		if (myBatchStatistic != null)
			myBatchStatistic.setName(name);
	}
	
	/** Sets the parameters for batch statistics
	 * @param minNumBatches The minimum number of batches, must be >= 2
	 * @param minBatchSize The minimum number of observations per batch, must be >= 2
	 * @param maxNBMultiple The maximum number of batches as a multiple of the minimum number of batches.
	 * For example, if minNB = 20 and maxNBMultiple = 2 then the maximum number of batches allowed will be 40.
	 * maxNBMultiple must be >= 2.
	 */
	public final void setBatchingParameters(int minNumBatches, int minBatchSize, int maxNBMultiple) {
		if (minNumBatches <= 1)
			throw new IllegalArgumentException("Number of batches must be >= 2");
		myMinNumBatches = minNumBatches;

		if (minBatchSize <= 1)
			throw new IllegalArgumentException("Batch size must be >= 2");
		myMinBatchSize = minBatchSize;

		if (maxNBMultiple <= 1)
			throw new IllegalArgumentException("Maximum number of batches multiple must be >= 2");
		myMaxNumBatchesMultiple = maxNBMultiple;
	}

	/** Sets the count limit for determining when the warm up should occur
	 * If countLimit is <= than current count when set
	 * then the action will occur at next update.
	 * @param countLimit must be >=0, zero implies no count limit
	 */
	public void setCountBasedWarmUpLimit(long countLimit) {
		if (countLimit < 0)
			throw new IllegalArgumentException("Count Limit must be >= 0.");

		myCountWarmUpLimit = countLimit;
	}

	/** Sets the count limit for determining when the count based stopping should occur
	 * If countLimit is <= than current count when set
	 * then the action will occur at next update.
	 * @param countLimit must be >=0, zero implies no count limit
	 */
	public void setCountBasedStopLimit(long countLimit) {
		if (countLimit < 0)
			throw new IllegalArgumentException("Count Limit must be >= 0.");

		myCountStopLimit = countLimit;
	}

	/** Resets the within replication and batch statistics
	 *
	 */
	public void resetStatistics() {
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
			JSL.LOGGER.finest("Calling \t ID: " + getId() + "\t NAME: " + getName() + "\t myWithinRepStat.reset()");
		myWithinRepStat.reset();
		if (myBatchStatistic != null) {
			if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
				JSL.LOGGER.finest("Calling \t ID: " + getId() + "\t NAME: " + getName() + "\t myBatchStatistic.reset()");
			myBatchStatistic.reset();
		}
	}

	public String toString(){

		StringBuffer sb = new StringBuffer();
		
		if (myBatchStatistic != null){
			sb.append("-------------------------------------------------\n");
			sb.append("Batch Statistic:\n");
			sb.append("-------------------------------------------------\n");
			sb.append(myBatchStatistic);
			sb.append("-------------------------------------------------\n");
			sb.append("\n");
		}

		if (myAcrossRepStat.getCount() >= 2.0){
			sb.append("-------------------------------------------------\n");
			sb.append("Across Replication Statistic:\n");
			sb.append("-------------------------------------------------\n");
			sb.append(myAcrossRepStat);
			sb.append("-------------------------------------------------\n");
			sb.append("\n");
		}

		if (myWithinRepStat != null){
			sb.append("-------------------------------------------------\n");
			sb.append("Within Replication Statistic:\n");
			sb.append("-------------------------------------------------\n");
			sb.append(myWithinRepStat);
			sb.append("-------------------------------------------------\n");
			sb.append("\n");
		}		
		
		return(sb.toString());
	}
	
	protected void beforeExperiment(ModelElement m, Object arg) {
		myResponseVariable = (ResponseVariable) m;
		myModel = myResponseVariable.getModel();
		if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
			JSL.LOGGER.finest("Calling \t ID: " + getId() + "\t NAME: " + getName() + "\t myAcrossRepStat.reset()");
		myAcrossRepStat.reset();
		resetStatistics();
	}

	protected void beforeReplication(ModelElement m, Object arg) {
		myResponseVariable = (ResponseVariable) m;
		myCountStopActionFlag = false;
		myCountWarmUpActionFlag = false;
		resetStatistics();
	}
	
	protected void initialize(ModelElement m, Object arg) {
		myResponseVariable = (ResponseVariable) m;
		if (myResponseVariable.isBatchingOn()) {
			// if batching option is on, create the statistic for collecting on the batch means
			if (myBatchStatistic == null)
				myBatchStatistic = new BatchStatistic(myMinNumBatches, myMinBatchSize,
					myMaxNumBatchesMultiple, "BatchStatistic " + getName());
		}
	}

	protected void warmUp(ModelElement m, Object arg) {
		resetStatistics();
	}

	protected void update(ModelElement m, Object arg) {
		myResponseVariable = (ResponseVariable) m;
		if (myBatchStatistic != null) {
			myBatchStatistic.collect(myWithinRepStat.getLastValue());
		}

		myWithinRepStat.collect(myResponseVariable.getValue());

		if (myCountWarmUpActionFlag == false) { // the count based warm up hasn't occurred yet
			if (myCountWarmUpLimit > 0) { // a limit has been set, so check it
				if (myWithinRepStat.getCount() >= (double) myCountWarmUpLimit) {
					myCountWarmUpActionFlag = true;
					if (JSL.LOGGER.isLoggable(java.util.logging.Level.FINEST))
						JSL.LOGGER.finest("Count based warm up at time " + ModelElement.getTime());
					resetStatistics();
				}
			}
		}

		if (myCountStopActionFlag == false) { // the count based stopping hasn't occurred yet
			if (myCountStopLimit > 0) { // a limit has been set, so check it
				if (myWithinRepStat.getCount() >= (double) myCountStopLimit) {
					myCountStopActionFlag = true;
					m.stopReplication();
				}
			}
		}
	}

	protected void afterReplication(ModelElement m, Object arg) {
		myAcrossRepStat.collect(myWithinRepStat.getWeightedAverage());
	}

	protected void afterExperiment(ModelElement m, Object arg) {
//		JSL.out.println(toString());
	}
	
	protected void removedFromModel(ModelElement m, Object arg) {
            super.removedFromModel(m, arg);
		myWithinRepStat = null;
		myAcrossRepStat = null;
		myBatchStatistic = null;
		myResponseVariable = null;
		myModel = null;
	}

}
