/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.observers.variable;

import jsl.utilities.statistic.*;
import jsl.modeling.elements.variable.*;
import jsl.modeling.*;
import jsl.observers.ModelElementObserver;
import jsl.modeling.conditions.ConditionIfc;

/**
 *
 */
public class WithinReplicationBatchRelativePrecisionChecker extends ModelElementObserver {

	/** The confidence coefficient for within replication
	 *  half-width checking. The default is Statistic.DEFAULT_CONFIDENCE_COEFF
	 * 
	 */
	protected double myConfidenceCoeff = Statistic.DEFAULT_CONFIDENCE_COEFF;
	
	/** The desired relative precision for stopping
	 */
	protected double myDesiredRelativePrecision = 1.0;
	
	/** Creates a new instance of StatisticalObserver */
	public WithinReplicationBatchRelativePrecisionChecker(double desiredPrecision) {
		this(desiredPrecision, null);
	}

	/** Creates a new instance of StatisticalObserver */
	public WithinReplicationBatchRelativePrecisionChecker(double desiredPrecision, String name) {
		super(name);
		setDesiredRelativePrecision(desiredPrecision);
	}

	/** Sets the confidence interval coefficient used in computing the half-width
	 * 
	 * @param confidenceCoeff
	 */
	public final void setConfidenceCoefficient(double confidenceCoeff) {
		if (confidenceCoeff <= 0)
			throw new IllegalArgumentException("Confidence Coefficent must be > 0.");

		myConfidenceCoeff = confidenceCoeff;
	}

	/** Sets the desired relative precision
	 * 
	 * @param desiredHalfWidth
	 */
	public final void setDesiredRelativePrecision(double desiredPrecision) {
		if (desiredPrecision <= 0)
			throw new IllegalArgumentException("Desired relative precision must be > 0.");

		myDesiredRelativePrecision = desiredPrecision;
	}

	/** Gets the current confidence coefficient
	 * 
	 * @return
	 */
	public final double getConfidenceCoefficient() {
		return (myConfidenceCoeff);
	}

	/** Gets the current desired half-width
	 * 
	 * @return
	 */
	public final double getDesiredRelativePrecision() {
		return (myDesiredRelativePrecision);
	}

	protected void update(ModelElement m, Object arg) {
		ResponseVariable x = (ResponseVariable)m;
		
		if (ModelElement.getTime() > x.getWarmUpEventTime()){
			StatisticAccessorIfc stat = x.getReplicationBatchStatistic();
			if (stat != null){
				double se = stat.getStandardError();
				double hw = se*myConfidenceCoeff;
				double xbar = stat.getAverage();
				if (hw <= myDesiredRelativePrecision*xbar){
					Replication r = x.getCurrentReplication();
					ConditionIfc c = r.getEndCondition();
					c.set();
					c.setName("Batch Statistic relative precision conditon met for " + x.getName());
				}				
			}	
		}
	}

}
