/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/


package jsl.utilities.random.distributions;

import java.lang.IllegalArgumentException;

import jsl.utilities.random.rng.RngIfc;
import jsl.utilities.random.rng.RngStream;

/** An Bernouli provides an implementation of the Bernouli
 *  distribuiton with success probability (p)
 *  P(X=1) = p
 *  P(X=0) = 1-p
*/
public class Bernoulli extends Distribution {
    // private data members
    
    private double myProbSuccess;
    private double myProbFailure;
    
    /**
     * Constructs a probability distribution with the default
     * random number generator, the default is Bernouli(0.5)
     *
     * @returns a valid Distribution
     */
    public Bernoulli(){
        this(0.5, new RngStream());
    }
    
    /** Constructs a probability distribution with the default
     * random number generator,
     *
     * @param parameter should be an array with parameter[0]=p
     */     
    public Bernoulli(double[] parameters){
         this(parameters[0], new RngStream());
     }

    /** Constructs a probability distribution with the default
     * random number generator,
     *
     * @param prob is the success probability
     */     
     public Bernoulli(double prob){
     	this(prob, new RngStream());
     }

    /** Constructs a probability distribution with the default
     * random number generator,
     *
     * @param prob is the success probability
     * @param a class that implements the RngIfc
     */     
     public Bernoulli(double prob, RngIfc rng){
     	super(rng);
        setProbabilityOfSuccess(prob);
     }
     
     // public methods
     
 	/** Returns a new instance of the random source with the same parameters
 	 *  but an independent generator
 	 * 
 	 * @return
 	 */
 	public final Bernoulli newInstance(){
 		return (new Bernoulli(getParameters()));
 	}
 	
     /** Sets the success probability
      * @param prob The success probability
      */     
     public final void setProbabilityOfSuccess(double prob){
         if ((prob < 0.0) || (prob > 1.0))
             throw new IllegalArgumentException("Probability must be [0,1]");
         myProbSuccess = prob;
         myProbFailure = 1.0 - myProbSuccess;
     }
     
     /** Gets the success probability
      * @return The success probability
      */     
     public final double getProbabilityOfSuccess(){
         return (myProbSuccess);
     }
     
     /** Returns the next random number 1 or 0
      * @return The random number
      */     
     public final double getValue(){
    	 if (myProbSuccess == 1.0)
    		 return (1.0);
    	 else if (myProbSuccess == 0.0)
    		 return (0.0);
    	 else {
    		 double prob = myRNG.randU01();
    		 return(invCDF(prob));
    	 }
     }
     
     /** Returns a randomly generated boolean according to the Bernoulli distribution
      * 
      * @return
      */
     public final boolean nextBoolean(){
    	 if (getValue() == 0.0)
    		 return(false);
    	 else
    		 return(true);
     }
     
     /** Returns the P(X<=x)
      * @param xx The value we want the cumulative probability up to
      * @return The cumulative probability
      */
    public final double cdf(double xx){
         int x = (int)xx;
         
         if (x < 0)
             return 0.0;
         else if (x >= 0 && x < 1)
             return myProbFailure;
         else //if (x >= 1)
             return 1.0;
     }
     
     /** Returns the value of x such that p = Pr{X <= x} where F represents the
      * cumulative distribution function
      *
      * @param prob a double representing the probability
      * @return a double representing the bernouli variate
      */
     public final double invCDF(double prob){
         if ((prob < 0.0) || (prob > 1.0))
             throw new IllegalArgumentException("Probability must be [0,1]");
         
         if (prob <= myProbSuccess)
             return(1.0);
         else
             return(0.0);
     }
     
    /** Returns the mean of the distributioin if defined
     * @return double  the mean or expected value for the distribution
     */
     public final double getMean(){
         return myProbSuccess;
     }
     
	/** Returns the f(x) where f represents the probability
	 * density function for the distribution.  
	 *
	 * @param x a double representing the value to be evaluated
	 * @return f(x)
	 */
	 public final double pdf(double x){
     	return(pmf((int)x));
     }
     
     /** Returns the f(x) where f represents the probability
      *  mass function for the distribution.  
      *
      * @param x a double representing the value to be evaluated
      * @return f(x)
      */
     public final double pmf(int x){
         if (x == 0)
             return myProbFailure;
         else if (x == 1)
             return myProbSuccess;
         else
             return 0.0;
     }
     
    /** Returns the variance of the random variate if defined
     * @return double  the variance of the random variable
     */
     public final double getVariance(){
         return myProbSuccess*myProbFailure;
     }
     
     /** Sets the parameters for the distribution
      *
      * @param parameters an array of doubles representing the parameters for
      * the distribution
      */
     public final void setParameters(double[] parameters) {
         setProbabilityOfSuccess(parameters[0]);
     }
     
     /** Gets the parameters for the distribution
      *
      * @return Returns an array of the parameters for the distribution
      */
     public final double[] getParameters() {
         double[] param = new double[1];
         param[0] = myProbSuccess;
         return(param);       
     }
     
}
