/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/


package jsl.utilities.random.distributions;

import java.lang.IllegalArgumentException;
import java.lang.NullPointerException;

import jsl.utilities.random.AbstractRandom;
import jsl.utilities.random.rng.RngIfc;
import jsl.utilities.random.rng.RngStream;

/** An Distribution provides a skeletal implementation for
 * classes that must implement the DistributionIfc.
 * This class is an abstract class.  Subclasses must provide concrete
 * implementations for the methods,
 * pdf, cdf, getMean, getVariance, nextRandom,etc.  A Distribution uses
 * a random number generator (RngIfc) to generate random variates via its
 * getValue() method.
 * 
 * Remarks:
 *  
 * 1) Each Distribution that is created, instantiates its own reference
 *    to an object that implements the RngIfc.  By default this is an instance
 *    of RngStream.  The user can provide their own RngIfc using the 
 *    setRandomNumberGenerator() method.
 * 2) Whenever an instance of a RngStream is made, each instance starts
 *    at a different stream within the sequence of random numbers provided
 *    by RngStream.  Thus, since each distribution has its own RngStream they
 *    will all be independent because they use different streams within the 
 *    generator.
 * 3) Resetting the start stream on the underlying RngStream, places the distribution's
 *    RngStream back to its initial point in the stream when it was created
 * 4) Resetting to the next substream, places the RngStream at the next logical
 *    stream in its sequence from the original starting point.
 */
public abstract class Distribution extends AbstractRandom implements DistributionIfc {
    
	//  ===========================================
	//      CLASS ATTRIBUTES
	//  ===========================================


    //  ===========================================
    //      OBJECT ATTRIBUTES
    //  ===========================================


    /** myRNG provides a reference to the underlying random number generator
     */
    protected RngIfc myRNG;
            
    public Distribution(){
    	this(new RngStream(), null);
    }
  
    /**
     * Constructs a probability distribution using the supplied RngIfc
     * 
     * @param a class that implements the RngIfc
     * @returns a valid Distribution
     */
    public Distribution(RngIfc rng){
    	this(rng, null);
    }
    
    /**
     * Constructs a probability distribution using the supplied RngIfc
     * 
     * @param rng a class that implements the RngIfc
     * @param name a String name
     * @returns a valid Distribution
     */
    public Distribution(RngIfc rng, String name){
    	super(name);
        setRandomNumberGenerator(rng);
     }    
    
    //  ===========================================
    //        PUBLIC INTERFACE
    //  ===========================================
   

    /** Sets the parameters for the distribution
     *
     * @param parameters an array of doubles representing the parameters for 
     * the distribution
     */
    abstract public void setParameters(double[] parameters);

    /** Gets the parameters for the distribution
     *
     * @return Returns an array of the parameters for the distribution
     */
    abstract public double[] getParameters();

    /** Returns the f(x) where f represents the probability
     * density function for the distribution.  Note this is not
     * a probability.
     *
     * @param x a double representing the value to be evaluated
     * @return Returns the value of the density function at x
     */
    abstract public double pdf(double x);
    
    /** Returns the F(x) = Pr{X <= x} where F represents the
     * cumulative distribution function
     *
     * @param x a double representing the upper limit
     * @return a double representing the probability
     */
    abstract public double cdf(double x);
    
    /** Returns the Pr{x1<=X<=x2} for the distribution throws an
     * IllegalArgumentException if x1 > x2
     *
     * @param x1 a double representing the lower limit
     * @param x2 a double representing the upper limit
     * @return cdf(x2)-cdf(x1)
     */
    public double cdf(double x1, double x2){
        if ( x1 > x2){
            String msg = "x1 = " + x1 + " > x2 = " + x2 + " in cdf(x1,x2)";
            throw new IllegalArgumentException(msg);
        }
        
        return (cdf(x2)-cdf(x1));
    }
    
    /** Computes the complementary cumulative probability
     * distribution function for given value of x
     * @param x The value to be evaluated
     * @return The probability, 1-P{X<=x}
     */      
    public double complementaryCDF(double x){
    	return (1.0 - cdf(x));
    }
    
    /** Returns the mean of the distribution if defined
     * @return double  the mean or expected value for the distribution
     */
    abstract public double getMean();
    
    /** Returns the variance of the distribution if defined
     * @return double  the variance of the random variable
     */
    abstract public double getVariance();
    
    /** Returns the standard deviation for the probability distribution
     * as the square root of the variance if it exists
     *
     * @return sqrt(getVariance())
     */
    public final double getStandardDeviation(){
        return Math.sqrt(getVariance());
    }

	/**
	 * 
	 */
	public void resetNextSubstream() {
		myRNG.resetNextSubstream();
	}
	
	/**
	 * 
	 */
	public void resetStartStream() {
		myRNG.resetStartStream();
	}
	
	/**
	 * 
	 */
	public void resetStartSubstream() {
		myRNG.resetStartSubstream();
	}
	
	/**
	 * @param flag
	 */
	public void setAntithetic(boolean flag) {
		myRNG.setAntithetic(flag);
	}
    	
	/** Returns the distributions underlying random number generator
	 * 
	 * @return
	 */
	public RngIfc getRandomNumberGenerator(){
		return(myRNG);
	}
	
    /** Sets the underlying random number generator for the distribution
     * Throws a NullPointerException if rng is null
     * @param rng the reference to the random number generator
     */
    public void setRandomNumberGenerator(RngIfc rng){
        if (rng == null)
            throw new NullPointerException("RngIfc rng must be non-null");
        myRNG = rng;
    }

    /** Returns a random number that is randomly distributed
     * according to the approporiate distribution
     *
     * @return the next random number according the appropriate distribution
     * using the underlying random number generator as the source.
     */
    public abstract double getValue();
    
     /** Provides the inverse cumulative distribution function for the distribution
     * @param p The probability to be evaluated for the inverse, p must be [0,1] or
     * an IllegalArgumentException is thrown
     * @return The inverse cdf evaluated at prob
     */  
    public abstract double invCDF(double p);
    
    /** Returns a string describing the distribution
     *
     *
     * @return a String representation of the distribution
     */
    public String toString(){
    	String s = this.getClass().getName();
    	int k = s.lastIndexOf(".");
    	if (k != -1)
    		s = s.substring(k + 1);
    	
    	s =  s + " Mean = " + getMean() + " Variance = " + getVariance();
    	return(s);
    }

	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
   
    
}
