package jsl.utilities.random.distributions;

import jsl.utilities.random.ParametersIfc;

public interface DistributionFunctionIfc extends ParametersIfc {

	/**
	 * Returns the f(x) where f represents the probability
	 * density function for the distribution.  Note this is not
	 * a probability.
	 *
	 * @param x a double representing the value to be evaluated
	 * @returns f(x)
	 */
	public abstract double pdf(double x);

	/**
	 * Returns the F(x) = Pr{X <= x} where F represents the
	 * cumulative distribution function
	 *
	 * @param x a double representing the upper limit
	 * @returns a double representing the probability
	 */
	public abstract double cdf(double x);

	/**
	 * Returns the Pr{x1<=X<=x2} for the distribution
	 *
	 * @param x1 a double representing the lower limit
	 * @param x2 a double representing the upper limit
	 * @returns cdf(x2)-cdf(x1)
	 * @throws IllegalArgumentException if x1 > x2
	 */
	public abstract double cdf(double x1, double x2);

	/** Computes the complementary cumulative probability
     * distribution function for given value of x
     * @param x The value to be evaluated
     * @return The probability, 1-P{X<=x}
     */      
        public abstract double complementaryCDF(double x);
    
	/** Returns the mean of the distribution if defined
	 * @return double  the mean or expected value for the distribution
	 */
	public abstract double getMean();

	/** Returns the variance of the random variate if defined
	 * @return double  the variance of the random variable
	 */
	public abstract double getVariance();

	/**
	 * Returns the standard deviation for the probability distribution
	 * as the square root of the variance if it exists
	 *
	 * @returns sqrt(getVariance())
	 */
	public abstract double getStandardDeviation();

	/** Provides the inverse cumulative distribution function for the distribution
	 * 
	 * While closed form solutions for the inverse cdf may not exist, numerical search
	 * methods can be used to solve F(X) = U.
	 * 
	 * If the inverse cumulative distribution function for a distribution does
	 * not exist or if it is extremely difficult to obtain, the implementor of
	 * the DistributionIfc may choose to simply throw a NoSuchMethodException inside
	 * the invCDF() method.  This is not an ideal solution to this problem and simply 
	 * "passes the buck" to the user of the class.
	 * 
	 * @param p The probability to be evaluated for the inverse, p must be [0,1] or
	 * an IllegalArgumentException is thrown
	 * @return The inverse cdf evaluated at the supplied probability
	 */
	public abstract double invCDF(double p);

}