/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 *
 * Copyright (c) Manuel D. Rossetti (rossetti@uark.edu)
 *
 * Contact:
 *	Manuel D. Rossetti, Ph.D., P.E.
 *	Department of Industrial Engineering
 *	University of Arkansas
 *	4207 Bell Engineering Center
 *	Fayetteville, AR 72701
 *	Phone: (479) 575-6756
 *	Email: rossetti@uark.edu
 *	Web: www.uark.edu/~rossetti
 *
 * This file is part of the JSL (a Java Simulation Library). The JSL is a framework
 * of Java classes that permit the easy development and execution of discrete event
 * simulation programs.
 *
 * The JSL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * The JSL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the JSL (see file COPYING in the distribution);
 * if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA, or see www.fsf.org
 *
 */
package jsl.utilities.random.rng;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

/** The purpose of this class is to allow for the creation and management
 *  of a large number of streams from the random number generator
 *
 *  Clients can use this class when they want to ensure that the
 *  random source for Distributions etc utilize specific streams
 *
 *  For example, when creating a distribution the client can supply
 *  an object that implements the RngIfc
 *
 *  RNGStreams streams = new RNGStreams(10);
 *
 *  Normal n = new Normal(0.0, 1.0, streams.get(0));
 *  Exponential e = new Exponential(4.0, streams.get(0));
 *
 *  Thus, n and e share the same stream
 *
 *  A globally available set of streams can also be made by using
 *  the static method createDefaultStreams(int n)
 *  Access to this set of streams is available with static methods
 *  getDefaultStreams() and getDefaultStream(int index)
 *
 *  Usage:
 *
 *  RNGStreams.createDefaultStreams(10);
 *  
 *  Normal n = new Normal(0.0, 1.0, RNGStreams.getStream(0));
 *
 * @author rossetti
 */
public class RNGStreams {

    /** A globally static set of random number streams
     *  for use in creating random sources
     */
    private static RNGStreams myRNGStreams;

    /** The number of streams that will be created when
     *  using Streams() by default
     */
    public static final int DEFAULT_NUM_STREAMS = 100;

    /** Holds the streams
     */
    protected List<RngIfc> myStreams;

    /** Creates DEFAULT_NUM_STREAMS
     *
     */
    public RNGStreams() {
        this(DEFAULT_NUM_STREAMS);
    }

    /** Makes n streams
     *
     * @param n number of streams to make, must be > 0
     */
    public RNGStreams(int n) {
        if (n <= 0) {
            throw new IllegalArgumentException("The number of streams must be >= 0");
        }

        myStreams = new ArrayList<RngIfc>(n);
        for (int i = 1; i <= n; i++) {
            myStreams.add(new RngStream());
        }

    }

    /** Creates a globally available list of default random number streams
     * @param n
     * @return
     */
    public static final void createDefaultStreams(int n) {
        if (myRNGStreams == null) {
            myRNGStreams = new RNGStreams(n);
        }
    }

    /** Gets a globally available list of default random number
     *  streams.  Will create a default set of DEFAULT_NUM_STREAMS streams,
     *  if createDefaultStreams(int n) has not yet been called
     *
     * @return
     */
    public static final RNGStreams getDefaultStreams() {
        if (myRNGStreams == null) {
            myRNGStreams = new RNGStreams();
        }
        return myRNGStreams;
    }

    /** Returns the default stream at the supplied index
     *  Will create a default set of DEFAULT_NUM_STREAMS streams,
     *  if createDefaultStreams(int n) or getDefaultStreams() has not yet been called
     *
     * @param index
     * @return
     */
    public static final RngIfc getDefaultStream(int index) {
        if (myRNGStreams == null) {
            myRNGStreams = new RNGStreams();
        }
        return myRNGStreams.get(index);
    }

    /** Adds n streams to the end of the list of streams
     *
     * @param n
     */
    public void addStreams(int n) {
        if (n <= 0) {
            throw new IllegalArgumentException("The number of streams must be > 0");
        }

        for (int i = 1; i <= n; i++) {
            myStreams.add(new RngStream());
        }
    }

    /** Returns the number of streams
     *
     * @return
     */
    public int getNumberOfStreams() {
        return myStreams.size();
    }

    /** Returns a ListIterator starting at the supplied index
     *
     * @param index
     * @return
     */
    public ListIterator<RngIfc> listIterator(int index) {
        return myStreams.listIterator(index);
    }

    /** Returns a ListIterator over the streams
     *
     * @return
     */
    public ListIterator<RngIfc> listIterator() {
        return myStreams.listIterator();
    }

    /**Returns the index of the last occurrence of the
     * specified element in this Streams, or -1 if this Streams
     * does not contain the element.
     *
     * @param o
     * @return
     */
    public int lastIndexOf(Object o) {
        return myStreams.lastIndexOf(o);
    }

    /** Returns an iterator over the streams
     *
     * @return
     */
    public Iterator<RngIfc> iterator() {
        return myStreams.iterator();
    }

    /** Checks to see if there are any streams
     *
     * @return
     */
    public boolean isEmpty() {
        return myStreams.isEmpty();
    }

    /** Returns the index of the first occurrence
     *  of the specified element in this Streams, or -1 if this Streams
     *  does not contain the element
     *
     * @param o
     * @return
     */
    public int indexOf(Object o) {
        return myStreams.indexOf(o);
    }

    /** Gets the stream at the designated index
     *  0 is the first stream
     *  size() - 1 is the last stream
     *
     * @param index
     * @return
     */
    public RngIfc get(int index) {
        return myStreams.get(index);
    }

    /** Returns true if this Streams contains all of the elements in the specified collection.
     *
     * @param c
     * @return
     */
    public boolean containsAll(Collection<?> c) {
        return myStreams.containsAll(c);
    }

    /**  Returns true if this Streams contains the specified stream
     *
     * @param o
     * @return
     */
    public boolean contains(Object o) {
        return myStreams.contains(o);
    }

    /** Causes all the streams to turn on antithetic variate generation
     */
    public final void turnOnAntithetic() {
        for (RngIfc rv : myStreams) {
            rv.setAntithetic(true);
        }
    }

    /** Causes all the streams to turn off antithetic variate generation
     */
    public final void turnOffAntithetic() {
        for (RngIfc rv : myStreams) {
            rv.setAntithetic(false);
        }
    }

    /** Causes all the streams to immediately reset to the next substream
     */
    public final void resetNextSubstream() {
        for (RngIfc rv : myStreams) {
            rv.resetNextSubstream();
        }
    }

    /** Causes all of the streams to immediately reset to the
     *  beginning of their starting stream.
     */
    public final void resetStartStream() {
        for (RngIfc rv : myStreams) {
            rv.resetStartStream();
        }
    }

    /** Causes all the streams to immediately reset the
     *  beginning of their current sub stream.
     */
    public final void resetStartSubStream() {
        for (RngIfc rv : myStreams) {
            rv.resetStartSubstream();
        }
    }
}
