/*
 * Created on Mar 1, 2007
 * Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
 *
 * Contact:
 *	Manuel D. Rossetti, Ph.D., P.E.
 *	Department of Industrial Engineering
 *	University of Arkansas
 *	4207 Bell Engineering Center
 *	Fayetteville, AR 72701
 *	Phone: (479) 575-6756
 *	Email: rossetti@uark.edu
 *	Web: www.uark.edu/~rossetti
 *
 * This file is part of the JSL (a Java Simulation Library). The JSL is a framework
 * of Java classes that permit the easy development and execution of discrete event
 * simulation programs.
 *
 * The JSL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * The JSL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the JSL (see file COPYING in the distribution);
 * if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA, or see www.fsf.org
 *
 */
package jsl.utilities.random.robj;

import java.util.HashMap;

import jsl.utilities.ControllableIfc;
import jsl.utilities.Controls;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.SampleIfc;
import jsl.utilities.random.rng.RngIfc;
import jsl.utilities.random.rng.RngStream;

/**
 * @author rossetti
 *
 */
public class DPopulation implements RandomIfc, SampleIfc {

    //  ===========================================
    //      CLASS ATTRIBUTES
    //  ===========================================
    /** A counter to count the number of created to assign "unique" ids
     */
    private static long myIdCounter_;

    //  ===========================================
    //      OBJECT ATTRIBUTES
    //  ===========================================
    /** The id of this object
     */
    protected long myId;

    /** Holds the name of the statistic for reporting purposes.
     */
    protected String myName;

    private RngIfc myRNG;

    private double[] myElements;

    /**
     *
     * @param elements
     */
    public DPopulation(double[] elements) {
        this(new RngStream(), elements, null);
    }

    /**
     * @param myRng
     * @param myElements
     */
    public DPopulation(RngIfc rng, double[] elements) {
        this(new RngStream(), elements, null);
    }

    /**
     * @param myRng
     * @param myElements
     */
    public DPopulation(RngIfc rng, double[] elements, String name) {
        setId();
        setName(name);
        setRandomNumberGenerator(rng);
        setParameters(elements);
    }

    //  ===========================================
    //        PUBLIC INTERFACE
    //  ===========================================
    /** Gets the name.
     * @return The name of object.
     */
    public final String getName() {
        return myName;
    }

    /** Sets the name
     * @param str The name as a string.
     */
    public final void setName(String str) {
        if (str == null) {
            String s = this.getClass().getName();
            int k = s.lastIndexOf(".");
            if (k != -1) {
                s = s.substring(k + 1);
            }
            myName = s;
        } else {
            myName = str;
        }
    }

    /** Returns the id for this collector
     *
     * @return
     */
    public final long getId() {
        return (myId);
    }

    /** Returns a new instance of the random source with the same parameters
     *  but an independent generator
     *
     * @return
     */
    public final DPopulation newInstance() {
        return (new DPopulation(getParameters()));
    }

    /** Creates a new array that contains a random sample with replacement
     *  from the existing population.
     *
     * @param sampleSize
     * @return
     */
    public final double[] getSample(int sampleSize) {
        double[] x = new double[sampleSize];
        for (int i = 0; i < sampleSize; i++) {
            x[i] = getValue();
        }
        return (x);
    }

    /** Fills the supplied array with a random sample with replacement
     *  from the existing population.
     *
     * @param sampleSize
     * @return
     */
    public final void getSample(double[] values) {
        if (values == null) {
            throw new IllegalArgumentException("The supplied array was null");
        }
        for (int i = 0; i < values.length; i++) {
            values[i] = getValue();
        }
    }

    /** Creates a new array that contains a random sample without replacement
     *  from the existing population.
     *
     * @param sampleSize
     * @return
     */
    public final double[] getSampleWithoutReplacement(int sampleSize) {
        Permutation.sampleWithoutReplacement(myElements, sampleSize, myRNG);
        double[] x = new double[sampleSize];
        System.arraycopy(myElements, 0, x, 0, x.length);
        return (x);
    }

    /** Creates a new array that contains a random permutation of the population
     *
     * @return
     */
    public final double[] getPermutation() {
        return (getSampleWithoutReplacement(myElements.length));
    }

    /** Causes the population to form a new permutation,
     *  The ordering of the elements in the population will be changed.
     */
    public final void permute() {
        Permutation.permutation(myElements, myRNG);
    }

    /** Returns the value at the supplied index
     *
     * @param index, must be > 0 and less than size() - 1
     * @return
     */
    public final double get(int index) {
        return (myElements[index]);
    }

    /** Sets the element at the supplied index to the supplied value
     *
     * @param index
     * @param x
     */
    public final void set(int index, double value) {
        myElements[index] = value;
    }

    /** Returns the number of elements in the population
     *
     * @return the size of the population
     */
    public final int size() {
        return (myElements.length);
    }

    /** Gets a copy of the population array, in its current state
     * @return
     */
    public double[] getParameters() {
        double x[] = new double[myElements.length];
        System.arraycopy(myElements, 0, x, 0, myElements.length);
        return x;
    }

    /** Copies the values from the supplied array to the population array
     *
     */
    public void setParameters(double[] elements) {
        if (elements == null) {
            throw new IllegalArgumentException("The element array was null");
        }
        myElements = new double[elements.length];
        System.arraycopy(elements, 0, myElements, 0, elements.length);
    }

    protected class DPopControls extends Controls {

        protected DPopControls(ControllableIfc controllable) {
            super(controllable);
            myDoubleArrayControls = new HashMap<String, double[]>();
            myDoubleArrayControls.put("parameters", getParameters());
        }
    }

    public Controls makeControls() {
        return new DPopControls(this);
    }

    public void setControls(Controls controls) {
        if (controls == null) {
            throw new IllegalArgumentException("The supplied controls were null!");
        }

        if (controls.getControllable() != this) {
            throw new IllegalArgumentException("The supplied controls do not belong to this distribution");
        }
        setParameters(controls.getDoubleArrayControl("parameters"));
    }

    /** Returns a randomly selected element from the population.  All
     *  elements are equally likely.
     * @return
     */
    public final double getValue() {
        return myElements[getRandomIndex()];
    }

    /** Returns a random index into the population (assuming elements numbered starting at zero)
     *
     * @return
     */
    protected final int getRandomIndex() {
        return (myRNG.randInt(0, myElements.length - 1));
    }

    /* (non-Javadoc)
     * @see jsl.utilities.random.RandomStreamIfc#resetNextSubstream()
     */
    public final void resetNextSubstream() {
        myRNG.resetNextSubstream();
    }


    /* (non-Javadoc)
     * @see jsl.utilities.random.RandomStreamIfc#resetStartStream()
     */
    public final void resetStartStream() {
        myRNG.resetStartStream();
    }

    /* (non-Javadoc)
     * @see jsl.utilities.random.RandomStreamIfc#resetStartSubstream()
     */
    public final void resetStartSubstream() {
        myRNG.resetStartSubstream();
    }


    /* (non-Javadoc)
     * @see jsl.utilities.random.RandomStreamIfc#setAntithetic(boolean)
     */
    public final void setAntithetic(boolean flag) {
        myRNG.setAntithetic(flag);
    }

    /** Returns the underlying random number generator
     *
     * @return
     */
    public final RngIfc getRandomNumberGenerator() {
        return (myRNG);
    }

    /** Sets the underlying random number generator 
     * Throws a NullPointerException if rng is null
     * @param rng the reference to the random number generator
     */
    public final void setRandomNumberGenerator(RngIfc rng) {
        if (rng == null) {
            throw new NullPointerException("RngIfc rng must be non-null");
        }
        myRNG = rng;
    }

    public String toString() {
        return (toString(myElements));
    }

    public static String toString(double[] x) {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < x.length; i++) {
            sb.append("Element(" + i + ") = " + x[i] + "\n");
        }
        return (sb.toString());
    }

    public static void main(String[] args) {

        double[] y = new double[10];
        for (int i = 0; i < 10; i++) {
            y[i] = i + 1;
        }

        DPopulation p = new DPopulation(y);
        System.out.println(p);

        p.permute();
        System.out.println(p);

        System.out.println("Permuting y");
        Permutation.permutation(y);
        System.out.println(DPopulation.toString(y));

        double[] x = p.getSample(5);
        System.out.println("Sampling x from population");
        System.out.println(DPopulation.toString(x));

    }

    //  ===========================================
    //      PROTECTED METHODS
    //  ===========================================
    protected void setId() {
        myIdCounter_ = myIdCounter_ + 1;
        myId = myIdCounter_;
    }
}
