/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jsl.utilities.random.robj;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import jsl.utilities.random.rng.RngIfc;
import jsl.utilities.random.rng.RngStream;


abstract public class RList<T> implements RListIfc<T> {

	protected List<T> myElements;
	
	protected RngIfc myRNG;
	
	public RList(){
		myElements = new ArrayList<T>();
		myRNG = new RngStream();
	}

	/** The object cannot be null, but it can be added more than once
	 *  See how List handles multiple instances of the same object 
	 *  
	 * @param obj
	 */
	public boolean add(T obj){
		if (obj == null)
			throw new IllegalArgumentException("The object was null");
		return myElements.add(obj);
	}
	
	public boolean remove(Object obj){
		return myElements.remove(obj);
	}
	
	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#getRandomElement()
	 */
	abstract public T getRandomElement();
	
	abstract public RList<T> newInstance(); 
	
	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#contains(java.lang.Object)
	 */
	public boolean contains(Object arg0) {
		return myElements.contains(arg0);
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#containsAll(java.util.Collection)
	 */
	public boolean containsAll(Collection<?> arg0) {
		return myElements.containsAll(arg0);
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#indexOf(java.lang.Object)
	 */
	public int indexOf(Object arg0) {
		return myElements.indexOf(arg0);
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#isEmpty()
	 */
	public boolean isEmpty() {
		return myElements.isEmpty();
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#size()
	 */
	public int size() {
		return myElements.size();
	}
	
	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#getList()
	 */
	public List<T> getList(){
		return(Collections.unmodifiableList(myElements));
	}

	/**
	 * @param index
	 * @param element
	 * @see java.util.List#add(int, java.lang.Object)
	 */
	public void add(int index, T element) {
		if (element == null)
			throw new IllegalArgumentException("The object was null");
		myElements.add(index, element);
	}

	/**
	 * @param c
	 * @return
	 * @see java.util.List#addAll(java.util.Collection)
	 */
	public boolean addAll(Collection<? extends T> c) {
		if (c == null)
			throw new IllegalArgumentException("The collection was null");
		if (c.contains(null))
			throw new IllegalArgumentException("The an object in the collection was null");			
		return myElements.addAll(c);
	}

	/**
	 * @param index
	 * @param c
	 * @return
	 * @see java.util.List#addAll(int, java.util.Collection)
	 */
	public boolean addAll(int index, Collection<? extends T> c) {
		if (c == null)
			throw new IllegalArgumentException("The collection was null");
		if (c.contains(null))
			throw new IllegalArgumentException("The an object in the collection was null");	
		return myElements.addAll(index, c);
	}

	/**
	 * 
	 * @see java.util.List#clear()
	 */
	public void clear() {
		myElements.clear();
	}

	/**
	 * @param index
	 * @return
	 * @see java.util.List#get(int)
	 */
	public T get(int index) {
		return myElements.get(index);
	}

	/**
	 * @return
	 * @see java.util.List#iterator()
	 */
	public Iterator<T> iterator() {
		return myElements.iterator();
	}

	/**
	 * @param o
	 * @return
	 * @see java.util.List#lastIndexOf(java.lang.Object)
	 */
	public int lastIndexOf(Object o) {
		return myElements.lastIndexOf(o);
	}

	/**
	 * @return
	 * @see java.util.List#listIterator()
	 */
	public ListIterator<T> listIterator() {
		return myElements.listIterator();
	}

	/**
	 * @param index
	 * @return
	 * @see java.util.List#listIterator(int)
	 */
	public ListIterator<T> listIterator(int index) {
		return myElements.listIterator(index);
	}

	/**
	 * @param index
	 * @return
	 * @see java.util.List#remove(int)
	 */
	public T remove(int index) {
		return myElements.remove(index);
	}

	/**
	 * @param c
	 * @return
	 * @see java.util.List#removeAll(java.util.Collection)
	 */
	public boolean removeAll(Collection<?> c) {
		return myElements.removeAll(c);
	}

	/**
	 * @param c
	 * @return
	 * @see java.util.List#retainAll(java.util.Collection)
	 */
	public boolean retainAll(Collection<?> c) {
		return myElements.retainAll(c);
	}

	/**
	 * @param index
	 * @param element
	 * @return
	 * @see java.util.List#set(int, java.lang.Object)
	 */
	public T set(int index, T element) {
		if (element == null)
			throw new IllegalArgumentException("The object was null");
		return myElements.set(index, element);
	}

	/**
	 * @param fromIndex
	 * @param toIndex
	 * @return
	 * @see java.util.List#subList(int, int)
	 */
	public List<T> subList(int fromIndex, int toIndex) {
		return myElements.subList(fromIndex, toIndex);
	}

	/**
	 * @return
	 * @see java.util.List#toArray()
	 */
	public Object[] toArray() {
		return myElements.toArray();
	}

	/**
	 * @param <T>
	 * @param a
	 * @return
	 * @see java.util.List#toArray(T[])
	 */
	public <T> T[] toArray(T[] a) {
		return myElements.toArray(a);
	}
	
	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#resetNextSubstream()
	 */
	public void resetNextSubstream() {
		myRNG.resetNextSubstream();
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#resetStartStream()
	 */
	public void resetStartStream() {
		myRNG.resetStartStream();
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#resetStartSubstream()
	 */
	public void resetStartSubstream() {
		myRNG.resetStartSubstream();
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomListIfc#setAntithetic(boolean)
	 */
	public void setAntithetic(boolean flag) {
		myRNG.setAntithetic(flag);
	}
	
}
