/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.utilities.statistic;

/**
 *
 */
abstract public class AbstractStatistic extends AbstractCollector implements StatisticAccessorIfc {

	//  ===========================================
	//      CLASS ATTRIBUTES
	//  ===========================================
	
	/** the default confidence coefficient based on Z table
	 *  resulting in an approximate confidence interval
	 */
	public final static double DEFAULT_CONFIDENCE_COEFF = 1.965;
	    
    //  ===========================================
    //      OBJECT ATTRIBUTES
    //  ===========================================

    /** Holds the confidence coefficient for the statistic
     */
    protected double myConfidenceCoeff;
    
    /** Used to count the number of missing data points presented
     *  When a data point having the value of Double.NaN is presented
     *  it is excluded from the summary statistics and the number
     *  of missing points is noted. Implementers of subclasses are
     *  responsible for properly collecting this value and 
     *  resetting this value.
     * 
     */
    protected double myNumMissing = 0.0;

	/** 
	 * 
	 */
	public AbstractStatistic() {
		this(null);
	}
	
	/**
	 * 
	 */
	public AbstractStatistic(String name) {
		super(name);
        myConfidenceCoeff = DEFAULT_CONFIDENCE_COEFF;
	}
	
    //  ===========================================
    //        PUBLIC INTERFACE
    //  ===========================================

    /** Sets the confidence coefficient for the statistic
     * @param confidenceCoeff, must be > 0.0
     */
    public void setConfidenceCoefficient(double confidenceCoeff){
        if (confidenceCoeff <=0)
            throw new IllegalArgumentException("Confidence Coefficent must be > 0.");
        
        myConfidenceCoeff = confidenceCoeff;
    }
 	
    /** Collect weighted statistics on the supplied value using the
     * supplied weight
     * @param x a double representing the observation
     * @param weight a double to be used to weight the observation
     */
    abstract public void collect(double x, double weight);
    
	//  ===========================================
	//      StatisticAccessorIfc METHODS
	//  ===========================================
 
    /** Gets the confidence interval coefficient.  The default is given by
     *  AbstractStatistic.DEFAULT_CONFIDENCE_COEFF = 1.965, which is a 95% confidence level
     *  assuming normally distributed data.
     * @return A double representing the coefficient
     */
    public double getConfidenceCoefficient(){
        return(myConfidenceCoeff);
    }

    /* (non-Javadoc)
     * @see jsl.utilities.statistic.StatisticAccessorIfc#getNumberMissing()
     */
    public double getNumberMissing(){
    	return(myNumMissing);
    }
    
    /* (non-Javadoc)
     * @see jsl.utilities.statistic.StatisticAccessorIfc#getStatistics(double[])
     */
    public void getStatistics(double[] statistics){
    	if (statistics.length != 23)
    		throw new IllegalArgumentException("The supplied array was not of size 23");
 
        statistics[0] = getCount();
        statistics[1] = getAverage();
        statistics[2] = getStandardDeviation();
        statistics[3] = getStandardError();
        statistics[4] = getHalfWidth();
        statistics[5] = getConfidenceCoefficient();
        statistics[6] = getMin();
        statistics[7] = getMax();
        statistics[8] = getSum();
        statistics[9] = getVariance();
        statistics[10] = getWeightedAverage();
        statistics[11] = getWeightedSum();
        statistics[12] = getSumOfWeights();
        statistics[13] = getWeightedSumOfSquares();
        statistics[14] = getDeviationSumOfSquares();
        statistics[15] = getLastValue();
        statistics[16] = getLastWeight();
        statistics[17] = getKurtosis();
        statistics[18] = getSkewness();
        statistics[19] = getLag1Covariance();
        statistics[20] = getLag1Correlation();
        statistics[21] = getVonNeumannLag1TestStatistic();
        statistics[22] = getNumberMissing();
    	
    }
    
    /* (non-Javadoc)
     * @see jsl.utilities.statistic.StatisticAccessorIfc#getStatistics()
     */
    public double[] getStatistics(){
    	double[] x = new double[23];
    	getStatistics(x);
    	return(x);
    }

    /**  Fills the supplied StringBuilder with a comma separated
     *   list of the statistics as per @see jsl.utilities.statistic.StatisticAccessorIfc#getStatistics(double[])
     * 
     * @param sb
     */
    public void makeCSVStatistic(StringBuilder sb){
    	double[] stats = getStatistics();
    	for(int i=0; i< stats.length; i++){
    		if (Double.isNaN(stats[i]) || Double.isInfinite(stats[i]))
    			sb.append("");
    		else
    			sb.append(stats[i]);
    		if (i < stats.length - 1)
    			sb.append(",");
    	}
    }
}
