/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/


package jsl.utilities.statistic;

import java.lang.String;
import java.lang.Math;
import java.util.Arrays;

/** The Statistic class allows the collection of summary statistics
 * on data via the collect() methods.  Both weighted and unweighted
 * statistical summaries are supported.  The primary statistical
 * summary is for the statistical moments.
 */
public class Statistic extends AbstractStatistic {
	
	//  ===========================================
	//      CLASS ATTRIBUTES
	//  ===========================================
	
	//  ===========================================
	//      OBJECT ATTRIBUTES
	//  ===========================================
	
	/** Holds the minimum of the observed data.
	 */
	protected double min;
	
	/** Holds the maximum of the observed data
	 */
	protected double max;
	
	/** Holds the weighted sum of squares of the data.
	 */
	protected double wsumsq;
	
	/** Holds the number of observations observed
	 */
	protected double num;
	
	/** Holds the weighted sum of the data.
	 */
	protected double wsum;
	
	/** Holds the sum of the weights observed.
	 */
	protected double sumw;
	
	/** Holds the last value observed
	 */
	protected double myValue;
	
	/** Holds the last weight observed
	 */
	protected double myWeight;
	    
    /** Holds the sum the lag-1 data, i.e. from the second data point on
     * variable for collecting lag1 covariance
     */
    protected double sumxx = 0.0;

	/** Holds the first observed data point, needed for von-neuman statistic
	 */
	protected double firstx = 0.0;

	/** Holds the first 4 statistical moments
	 */
	protected double[] moments;
	
	/** Holds sum = sum + j*x
	 */
	protected double myJsum;
	
	//  ===========================================
	//        CONSTRUCTORS
	//  ===========================================
	
	/** Creates a Statistic with name "null"
	 */
	public Statistic(){
		this(null, null);
	}

	/** Creates a Statistic with the given name
	 * @param name A String representing the name of the statistic
	 */
	public Statistic(String name){
		this(name, null);
	}

	/** Creates a Statistic \based on the provided array
	 * @param an array of values to collect statistics on
	 */
	public Statistic(double[] values){
		this(null, values);
	}

	/** Creates a Statistic with the given name based on the provided array
	 * @param name A String representing the name of the statistic
	 * @param an array of values to collect statistics on
	 */
	public Statistic(String name, double[] values){
		super(name);
		moments = new double[5];
		reset();      
		if (values != null){
			collect(values);
		}
	}
	
	/** Returns a statistic that summarizes the passed in array of values
	 * 
	 * @param x the values to compute statistics for
	 * @return
	 */
	public final static Statistic collectStatistics(double[] x){	
		Statistic s = new Statistic();
		s.collect(x);
		return(s);
	}
	
	/** Returns a statistic that summarizes the passed in arrays
	 *  The lengths of the arrays must be the same.
	 *  
	 * @param x the values
	 * @param w the weights
	 * @return
	 */
	public final static Statistic collectStatistics(double[] x, double[] w){
		if (x.length != w.length)
			throw new IllegalArgumentException("The supplied arrays are not of equal length");
		
		Statistic s = new Statistic();
		s.collect(x, w);
		
		return(s);
	}
	
	/** Creates a instance of Statistic that is a copy of the supplied Statistic
	 *  All internal state is the same except for the id of the returned Statistic
	 * 
	 * @param stat
	 * @return
	 */
	public final static Statistic newInstance(Statistic stat){
		Statistic s = new Statistic();
		s.myNumMissing = stat.myNumMissing;
		s.firstx = stat.firstx;
		s.max = stat.max;
		s.min = stat.min;
		s.myConfidenceCoeff = stat.myConfidenceCoeff;
		s.myJsum = stat.myJsum;
		s.myValue = stat.myValue;
		s.myName = stat.myName;
		s.myWeight = stat.myWeight;
		s.num = stat.num;
		s.sumw = stat.sumw;
		s.wsum = stat.wsum;
		s.wsumsq = stat.wsumsq;
		s.sumxx = stat.sumxx;		
		for(int i=0;i<s.moments.length;i++){
			s.moments[i] = stat.moments[i];
		}
		return(s);
	}
	
	/** Creates a instance of Statistic that is a copy of this Statistic
	 *  All internal state is the same except for the id of the returned Statistic
	 * 
	 * @return
	 */
	public final Statistic newInstance(){
		Statistic s = new Statistic();
		s.myNumMissing = myNumMissing;
		s.firstx = firstx;
		s.max = max;
		s.min = min;
		s.myConfidenceCoeff = myConfidenceCoeff;
		s.myJsum = myJsum;
		s.myValue = myValue;
		s.myName = myName;
		s.myWeight = myWeight;
		s.num = num;
		s.sumw = sumw;
		s.wsum = wsum;
		s.wsumsq = wsumsq;
		s.sumxx = sumxx;	
		for(int i=0;i<s.moments.length;i++){
			s.moments[i] = moments[i];
		}
		return(s);
	}
	
    /** Returns the index associated with the minimum element in the array
     *  For ties, this returns the first found
     * 
     * @param x
     * @return
     */
	public final static int getIndexOfMin(double[] x){
    	int index = 0;
    	double min = Double.MAX_VALUE;
    	for(int i=0; i< x.length;i++){
    		if (x[i] < min){
    			min = x[i];
    			index = i;
    		}
    	}
    	return(index);
    }
 
    /** Returns the index associated with the maximum element in the array
     *  For ties, this returns the first found
     * 
     * @param x
     * @return
     */
	public final static int getIndexOfMax(double[] x){
    	int index = 0;
    	double max = Double.MIN_VALUE;
    	for(int i=0; i< x.length;i++){
    		if (x[i] > max){
    			max = x[i];
    			index = i;
    		}
    	}
    	return(index);
    }
	
	/** Returns the median of the data.  The array is sorted
	 * 
	 * @param data
	 * @return
	 */
	public final static double getMedian(double[] data){
		Arrays.sort(data);
		int size=data.length;
		double median=-1;
		if(size%2==0){//even
			int firstIndex=(size/2)-1;
			int secondIndex=firstIndex+1;
			double firstValue=data[firstIndex];
			double secondValue=data[secondIndex];
			median=(firstValue+secondValue)/2;
		}
		else{//odd
			int index=(int)Math.ceil(size/2);
			median=data[index];
		}
		return median;
	}
	
	//  ===========================================
	//        PUBLIC
	//  ===========================================
	
	/*  --Accessors---------------------- */
	
	/** Gets the count of the number of the observations.
	 * @return A double representing the count
	 */
	public final double getCount(){
		return (moments[0]);
	}
	
	/** Gets the unweighted sum of the observations.
	 * @return A double representing the unweighted sum
	 */
	public final double getSum(){
		return (moments[1]*moments[0]);
	}
	
	/** Gets the weighted sum of observations observed.
	 * @return A double representing the weighted sum
	 */
	public final double getWeightedSum(){
		return (wsum);
	}
	
	/** Gets the weighted sum of squares of the observations.
	 * @return A double representing the weighted sum of squares
	 */
	public final double getWeightedSumOfSquares(){
		return (wsumsq);
	}
	
	/** Gets the unweighted average of the observations.
	 * @return A double representing the average or Double.NaN if no observations.
	 */
	public final double getAverage(){
		if (moments[0] < 1.0)
			return Double.NaN;
		
		return (moments[1]);
	}
	
	/** Gets the sum of the observed weights.
	 * @return A double representing the sum of the weights
	 */
	public final double getSumOfWeights(){
		return(sumw);
	}
	
	/** Gets the weighted average of the observations.
	 * @return A double representing the weighted average or Double.NaN if no observations.
	 */
	public final double getWeightedAverage(){
		if (sumw <= 0.0)
			return Double.NaN;
		
		return(wsum/sumw);
	}
	
	/** Gets the sum of squares of the deviations from the average
	 *  This is the numerator in the classic sample variance formula
	 * @return A double representing the sum of squares of the deviations from the average
	 */
	public final double getDeviationSumOfSquares(){
		return(moments[2]*moments[0]);
	}
	
	/** Gets the unweighted sample variance of the observations.
	 * @return A double representing the sample variance or Double.NaN if 1 or less observations.
	 */
	public final double getVariance(){
		if (moments[0] < 2)
			return Double.NaN;
		
		return (getDeviationSumOfSquares()/(moments[0] - 1.0));
	}
	
	/** Gets the unweighted sample standard deviation of the observations. Simply
	 *  the square root of getVariance()
	 * @return A double representing the sample standard deviation or Double.NaN if 1 or less observations.
	 */
	public final double getStandardDeviation(){
		return (Math.sqrt(getVariance()));
	}
	
	/** Gets the minimum of the observations.
	 * @return A double representing the minimum
	 */
	public final double getMin(){
		return (min);
	}
	
	/** Gets the maximum of the observations.
	 * @return A double representing the maximum
	 */
	public final double getMax(){
		return (max);
	}
	
	/** Gets the last observed data point
	 * @return A double representing the last observations
	 */
	public final double getLastValue(){
		return(myValue);
	}
	
	/** Gets the last observed weight
	 * @return A double representing the last weight
	 */
	public final double getLastWeight(){
		return(myWeight);
	}
	
	/** Gets the kurtosis of the data
	 * @return A double representing the kurtosis
	 */
	public final double getKurtosis(){
		if (moments[0] < 4)
			return(Double.NaN);
		
		double n = moments[0];
		double n1 = n - 1.0;
		double v = getVariance();
		double d = (n - 1.0)*(n - 2.0)*(n - 3.0)*v*v;
        double t = n*(n+1.0)*n*moments[4] - 3.0*n1*n1*n1*v*v;
		double k = (double)t/(double)d;
		return(k);
	}
	
	/** Gets the skewness of the data
	 * @return A double representing the skewness
	 */
	public final double getSkewness(){
		if (moments[0] < 3)
			return(Double.NaN);
		
		double n = moments[0];
		double v = getVariance();
		double s = Math.sqrt(v);
		double d = (n - 1.0)*(n - 2.0)*v*s;
		double t = n*n*moments[3];

		double k = (double)t/(double)d;
		return(k);
	}
	
	/** Gets the confidence interval half-width.  Simply the
	 * sample standard error times the confidence coefficient
	 * @return A double representing the half-width or Double.NaN if < 1 observation
	 */
	public final double getHalfWidth(){
		if (moments[0] < 1.0)
			return (Double.NaN);
		
		return (getConfidenceCoefficient()*getStandardError());
	}
	
	/** Gets the standard error of the observations.  Simply the
	 * sample standard deviation divided by the square root of the number of observations
	 * @return A double representing the standard error or Double.NaN if < 1 observation
	 */
	public final double getStandardError(){
		if (moments[0] < 1.0)
			return (Double.NaN);
		
		return (getStandardDeviation()/Math.sqrt(moments[0]));
	}
	
	/** Gets the lag-1 sample covariance of the unweighted observations.
	 *  Note: See Box, Jenkins, Reinsel, Time Series Analysis, 3rd edition, Prentice-Hall, pg 31
	 * @return A double representing the sample covariance or Double.NaN if <=2 observations
	 */
    public final double getLag1Covariance(){
        if (num > 2.0){
            double c1 = sumxx - (num+1.0)*moments[1]*moments[1] + moments[1]*(firstx + myValue);
            return (c1/num);
        }
        else
            return(Double.NaN);
    }
	
	/** Gets the lag-1 sample correlation of the unweighted observations.
	 *  Note: See Box, Jenkins, Reinsel, Time Series Analysis, 3rd edition, Prentice-Hall, pg 31
	 * @return A double representing the sample correlation or Double.NaN if <=2 observations
	 */
	public final double getLag1Correlation(){
		if (num > 2.0)
			return(getLag1Covariance()/moments[2]);
		else
			return(Double.NaN);
	}
	
	/** Gets the Von Neumann Lag 1 test statistic for checking the hypothesis that the data are uncorrelated
	 *  Note: See Handbook of Simulation, Jerry Banks editor, McGraw-Hill, pg 253.
	 * @return A double representing the Von Neumann test statistic
	 */
	public final double getVonNeumannLag1TestStatistic(){
		if (num > 2.0){
			double r1 = getLag1Correlation();
			double t = (firstx - moments[1])*(firstx - moments[1]) + (myValue - moments[1])*(myValue - moments[1]);
			double b = 2.0*num*moments[2];
			double v = Math.sqrt((num*num - 1.0)/(num - 2.0))*(r1 + (t/b));
			return (v);
		}
		else
			return(Double.NaN);
	}
	
	/** Returns the observation weighted sum of the data
	 *  i.e. sum = sum + j*x
	 *  where j is the observation number and x is jth observation
	 * 
	 * @return
	 */
	public final double getObsWeightedSum(){
		return(myJsum);
	}
	
	/** Collect weighted statistics on the supplied value using the
	 * supplied weight
	 * @param x a double representing the observation
	 * @param weight a double to be used to weight the observation
	 */
	public void collect(double x, double weight){
		if (Double.isNaN(x)){
			myNumMissing++;
			return;
		}
		
		if (getSaveDataOption())
			saveData(x, weight);
		
		double n, n1, n2, delta, d2, d3, r1;
		
		// update moments
		num = num + 1;
		sumw = sumw + weight;
		wsum = wsum + x*weight;
		wsumsq = wsumsq + x*x*weight;
		myJsum = myJsum + num*x;
		
		n = moments[0];
		n1 = n + 1.0;
		n2 = n*n;
		delta = (moments[1] - x)/n1;
		d2 = delta*delta;
		d3 = delta*d2;
		r1 = n/n1;
		moments[4] = (1.0 + n*n2)*d2*d2 + 6.0*moments[2]*d2 + 4.0*moments[3]*delta + moments[4];
		moments[4] *= r1;
		
		moments[3] = (1.0 - n2)*d3 + 3.0*moments[2]*delta + moments[3];
		moments[3] *= r1;
		
		moments[2] = (1.0 + n)*d2 + moments[2];
		moments[2] *= r1;
		
		moments[1] = moments[1] - delta;
		moments[0] = n1;
		
		// to collect lag 1 cov, we need x(1)
		
		if (num == 1.0)
			firstx = x;
		
        // to collect lag 1 cov, we must provide new x and previous x
        // to collect lag 1 cov, we must sum x(i) and x(i+1)
        if (num >= 2.0)
            sumxx = sumxx + x*myValue;
		
		// update min, max, current value, current weight
		if (x > max)
			max = x;
		if (x < min)
			min = x;
		
		myValue = x;
		myWeight = weight;
		
	}
	
	/** Resets the statistical accumulators as if no data had been
	 *  collected.
	 */
	public void reset(){
		myNumMissing = 0.0;
		min = Double.POSITIVE_INFINITY;
		max = Double.NEGATIVE_INFINITY;
		wsumsq = 0.0;
		num = 0.0;
		wsum = 0.0;
		sumw = 0.0;
		myJsum = 0.0;
		sumxx = 0.0;
		
		for (int i=0; i< moments.length; i++)
			moments[i] = 0.0;
		
		clearSavedData();
	}
	
	//  Want String conversion for output?
	
	/** Returns a String representation of the Statistic
	 * @return A String with basic summary statistics
	 */
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append("ID " + getId() + "\n");
		sb.append("Name " + getName() + "\n");
		sb.append("Number " + getCount() + "\n");
		sb.append("Average " + getAverage() + "\n");
		sb.append("Standard Deviation " + getStandardDeviation() + "\n");
		sb.append("Standard Error " + getStandardError() + "\n");
		sb.append("Half-width " + getHalfWidth() + "\n");
		sb.append("Confidence Coefficient " + getConfidenceCoefficient() + "\n");
		sb.append("Minimum " + getMin() + "\n");
		sb.append("Maximum " + getMax() + "\n");
		sb.append("Sum " + getSum() + "\n");
		sb.append("Variance " + getVariance() + "\n");
		sb.append("Weighted Average " + getWeightedAverage() + "\n");
		sb.append("Weighted Sum " + getWeightedSum() + "\n");
		sb.append("Sum of Weights " + getSumOfWeights() + "\n");
		sb.append("Weighted Sum of Squares " + getWeightedSumOfSquares() + "\n");
		sb.append("Deviation Sum of Squares " + getDeviationSumOfSquares() + "\n");
		sb.append("Last value collected " + getLastValue() + "\n");
		sb.append("Last weighted collected " + getLastWeight() + "\n");
		sb.append("Kurtosis " + getKurtosis() + "\n");
		sb.append("Skewness " + getSkewness() + "\n");
		sb.append("Lag 1 Covariance " + getLag1Covariance() + "\n");
		sb.append("Lag 1 Correlation " + getLag1Correlation() + "\n");
		sb.append("Von Neumann Lag 1 Test Statistic " + getVonNeumannLag1TestStatistic() + "\n");
		sb.append("Number of missing observations " + getNumberMissing() + "\n");
		return(sb.toString());
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
	
}