/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jsl.utilities.statistic;

/**
 *
 */
public class StatisticXY {

    // variables for collecting statistics
    protected double avgx = 0.0;
    protected double avgy = 0.0;
    protected double varx = 0.0;
    protected double vary = 0.0;
    protected double sumxy = 0.0;
    protected double covxy = 0.0;
    protected double nxy = 0.0;
    
	/**
	 * 
	 */
	public StatisticXY() {
		this(null);
	}
	
	public StatisticXY(StatisticXY stat){
		super();
		initialize(stat);
	}

	public static StatisticXY newInstance(StatisticXY stat){
		StatisticXY s = new StatisticXY();
		s.initialize(stat);
		return(s);
	}
	
	public final void initialize(StatisticXY stat){
		if (stat != null){
			avgx = stat.avgx;
			avgy = stat.avgy;
			covxy = stat.covxy;
			nxy = stat.nxy;
			varx = stat.varx;
			vary = stat.vary;			
		}		
	}
	
	public final double getAverageX(){
		return(avgx);
	}
	
	public final double getAverageY(){
		return(avgy);
	}
	
	public final double getVarianceX(){
		return(varx);
	}
	
	public final double getVarianceY(){
		return(vary);
	}
	
	public final double getSumXY(){
		return(sumxy);
	}
	
	public final double getCovarianceXY(){
        if (nxy > 1.0)
            return(covxy);
         else
             return(Double.NaN);
	}
	
	public final double getCount(){
		return(nxy);
	}
	
    public final double getCorrelationXY(){
        if (nxy > 1.0){
           return((covxy)/(Math.sqrt(varx*vary))); // matches Pearson correlation
        }
        else
            return(Double.NaN);
    }
    
    public void reset(){
        avgx = 0.0;
        avgy = 0.0;
        varx = 0.0;
        vary = 0.0;
        sumxy = 0.0;
        covxy = 0.0;
        nxy = 0.0;
    }
    
    public final void collectXY(double x, double y){
        double deltax, deltay, n1, r1;
        // compute for n+1
        n1 = nxy + 1.0;
        r1 = nxy/n1;
        deltax = (avgx - x)/n1;
        deltay = (avgy - y)/n1;
        sumxy = sumxy + x*y;
        covxy = r1*covxy + nxy*deltax*deltay;
        avgx = avgx - deltax;
        avgy = avgy - deltay;
        varx = nxy*deltax*deltax + r1*varx;
        vary = nxy*deltay*deltay + r1*vary;
        nxy = nxy + 1.0; 
    }

    public final double getRatioXY(){
    	if (avgy == 0.0){
    		if (avgx > 0.0)
    			return(Double.POSITIVE_INFINITY);
    		else if (avgx < 0.0)
    			return(Double.NEGATIVE_INFINITY);
    		else
    			return(Double.NaN);
    	} else {
    		return(avgx/avgy);
    	}
    }
    
    public final double getRatioXYVariance(){
    	double mu = getRatioXY();
    	double vx = getVarianceX();
    	double vy = getVarianceY();
    	double cxy = getCovarianceXY();
    	double var = vx - 2.0*mu*cxy + mu*mu*vy;
    	return(var);
    }
    
    public final double getRatioXYStdError(){
    	if (nxy >= 1.0){ 
    		if (avgy != 0.0){
    	   		double sd = Math.sqrt(getRatioXYVariance());
    			double se = sd/(avgy*nxy);
    			return(se);
    		}else
    			return(Double.NaN);
    	} else
    		return(Double.NaN);
    }
    
	public final String toString() {
		
		StringBuffer sb = new StringBuffer();
		sb.append("---------------\n");
		sb.append("X Y Statistics: \n");
		sb.append("---------------\n");
		sb.append("nxy: " + nxy + "\n");
		sb.append("avgx: " + avgx + "\n");
		sb.append("avgy: " + avgy + "\n");
		sb.append("varx: " + varx + "\n");
		sb.append("vary: " + vary + "\n");
		sb.append("covxy: " + covxy + "\n");
		sb.append("corrxy: " + getCorrelationXY() + "\n");
		sb.append("ratioxy: " + getRatioXY() + "\n");
		sb.append("var ratioxy: " + getRatioXYVariance() + "\n");
		sb.append("se ratioxy: " + getRatioXYStdError() + "\n");

		return (sb.toString());
	}
	
	public static void main(String args[]) {
		
		double[] x = {0.0, 30.0, 10.0, 25.0, 12.0, 14.0, 5.0, 40.0, 42.0, 32.0, 16.0};
		double[] y = {0.0, 4.0, 2.0, 3.0, 2.0, 3.0, 1.0, 6.0, 7.0, 5.0, 3.0};
		
		StatisticXY stat = new StatisticXY();
		
		for(int i= 0; i<x.length; i++){
			stat.collectXY(x[i], y[i]);
		}
		
		System.out.println(stat);
	}
}
