/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.adc;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.component.Component;
import jsl.modeling.elements.operation.Activity;
import jsl.modeling.elements.operation.Operation;
import jsl.modeling.elements.operation.OperationIfc;
import jsl.modeling.elements.operation.OperationListenerIfc;
import jsl.modeling.elements.operation.Task;
import jsl.modeling.elements.processview.description.commands.Delay;
import jsl.modeling.elements.resource.Request;
import jsl.modeling.elements.resource.Resource;
import jsl.modeling.elements.resource.ResourceAllocationListenerIfc;
import jsl.modeling.elements.resource.ResourceProvider;
import jsl.modeling.elements.spatial.RectangularGridSpatialModel2D;
import jsl.modeling.elements.spatial.SpatialResource;
import jsl.modeling.elements.spatial.ResourceLocation;
import jsl.modeling.elements.spatial.SpatialModel;
import jsl.modeling.elements.spatial.SpatialModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.modeling.elements.variable.Variable;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jslx.adc.adcsystem.*;
import jslx.adc.hmsystem.HMSensor;
import jslx.adc.hmsystem.HMSystem;
import jslx.adc.hmsystem.SensorUpdateModelIfc;




public class AirBase extends SpatialModelElement implements SensorUpdateModelIfc {
	
	// --------  ADC System components
	protected ADCSystem myNetwork;
	
	// --------------------------------------------
	
	protected ResourceProvider myRepairStationProvider; // a queue is already provided
	
	protected List<ResourceLocation> myEquipmentLocations;

	protected List<SpatialResource> myEquipment;

	protected ResourceProvider myWorkerProvider;  // a queue is already provided
	
	protected List<ResourceLocation> myWorkerLocations;

	protected BeginRepairStationService myBeginRepairStationServiceListener = new BeginRepairStationService();

	protected RepairOperation myRepairOperation;
	
	protected EndRepairOperationListener myEndRepairOperationListener = new EndRepairOperationListener();
	
	
	protected TimeWeighted myNumPlanesInRepair;
	
	//This is what I added.  BH 
	protected DistributionIfc myDistribution;
	
	//protected Operation myInspectionDelay;
	
	protected ActivityCompletionListener myCompletionListener;
	
	protected Activity myInspectionDelay;


	
	public AirBase(ModelElement parent){
		this(parent, null, null);
	}
	
	public AirBase(ModelElement parent, String name){
		this(parent, null, null);
	}
	
	public AirBase(ModelElement parent, SpatialModel airfield) {
		this(parent, airfield, null);
	}

	public AirBase(ModelElement parent, SpatialModel airfield, String name) {
		super(parent, name, airfield);
		
//		setAirField(airfield);
		
		myNumPlanesInRepair = new TimeWeighted(this, "NumPlanesInRepair");
		
		myRepairStationProvider = new ResourceProvider(this,"RepairStationProvider");
		
		myEquipmentLocations = new LinkedList<ResourceLocation>();
		
		myEquipment = new LinkedList<SpatialResource>();

		myWorkerProvider = new ResourceProvider(this,"WorkerProvider");
		
		myWorkerLocations = new LinkedList<ResourceLocation>();
		
		// create the repair operation
		myRepairOperation = new RepairOperation(this,"RepairOperation");
		
		// set the listener so actions can happen when repair operation is completed
		myRepairOperation.addCompletionListener(myEndRepairOperationListener);
		
		myDistribution = new Exponential(5);
		
		
		myInspectionDelay = new Activity(this, myDistribution);
		myCompletionListener = new ActivityCompletionListener();
		
		myInspectionDelay.addCompletionListener(myCompletionListener);
	}
	
	// ===========================================
	// PUBLIC METHODS
	// ===========================================
	
	public final SpatialModel getAirField() {
		return (getSpatialModel());
	}

	public final SpatialResource addRepairStation(double x, double y){
		return(addRepairStation(null, x, y));
	}
	
	public final SpatialResource addRepairStation(String name, double x, double y){
		SpatialResource rs = new SpatialResource(this, name, x, y);
		myRepairStationProvider.addResource(rs);
		return(rs);
	}
	
	public final Worker addWorker(double x, double y){
		return(addWorker(null, x, y));
	}
	
	public final Worker addWorker(ResourceLocation location){
		return(addWorker(null, location));
	}
	
	public final Worker addWorker(String name, ResourceLocation location){
		return(addWorker(name, location.getCurrentPosition().getX1(), location.getCurrentPosition().getX2()));
	}
	
	public final Worker addWorker(String name, double x, double y){
		Worker w = new Worker(this, name, x, y);
		myWorkerProvider.addResource(w);
		return(w);
	}
	
	public final ResourceLocation addWorkerLocation(String name, double x, double y){
		ResourceLocation location = new ResourceLocation(this, name, x, y);
		myWorkerLocations.add(location);
		return(location);
	}

	public final ResourceLocation addEquipmentLocation(double x, double y){
		return(addEquipmentLocation(null, x, y));
	}
	
	public final ResourceLocation addEquipmentLocation(String name, double x, double y){
		ResourceLocation location = new ResourceLocation(this, name, x, y);
		myEquipmentLocations.add(location);
		return(location);
	}

	public SpatialResource addEquipment(ResourceLocation location){
		return(addEquipment(null, location));
	}

	public SpatialResource addEquipment(String name, ResourceLocation location){
		SpatialResource resource = new SpatialResource(this, name, location);
		location.addInitialResource(resource);
		myEquipment.add(resource);
		return(resource);
	}
	
	public ADCSystem addNetwork(SpatialModel sm){
		myNetwork = new ADCSystem("ADC Network", sm);
		return(myNetwork);
		
	}
	
	public boolean getAircraftStatus(Aircraft a){
		return a.isFailed();
		
	}
	
	// ===========================================
	// PROTECTED METHODS
	// ===========================================

	protected final void setAirField(SpatialModel airfield) {
		if (airfield == null)
			airfield = new RectangularGridSpatialModel2D();

		setSpatialModel(airfield);
	}

	protected final List<SpatialResource> getEquipmentList(){
		return(myEquipment);
	}
	
	protected final List<ResourceLocation> getEquipmentLocationList() {
		
		if (myNetwork != null){
			return(myNetwork.getEquipmentLocationList());
		} else {
			return (myEquipmentLocations);
		}
		
	}

	protected final ResourceProvider getWorkerProvider() {
		return myWorkerProvider;
	}
	
	protected final ListIterator<Resource> getWorkerListIterator(){
		return(myWorkerProvider.getResourceSet().getResourceIterator());
	}
	
	/**
	 * @return Returns the myNumPlanesInRepair.
	 */
	public final TimeWeighted getNumPlanesInRepair() {
		return myNumPlanesInRepair;
	}

	protected void serviceAircraft(Aircraft aircraft){
		// the aircraft has just completed a sortie and is
		// coming in for service
		myNumPlanesInRepair.increment();
		// create a repair task
		RepairTask task = new RepairTask();
		
		// set the attributes of the task
		task.setAirBase(this);
		task.setAircraft(aircraft);

		// must allocate one of the repair stations to the aircraft
		// or have it wait in a queue

		System.out.println("t> " + getTime() + " Seizing repair station");
		myRepairStationProvider.seize(myBeginRepairStationServiceListener, task);

	}
	
	protected void scheduleAircraftSortie(Aircraft aircraft){
		
		// tell the aircraft to begin its next sortie
		aircraft.beginSortie();	
		myNumPlanesInRepair.decrement();
	}
	
	// ===========================================
	// PRIVATE CLASSES
	// ===========================================

	private class BeginRepairStationService implements ResourceAllocationListenerIfc {
		
		public void allocated(Request request, Resource resource) {
			// a repairstation has been allocated to a plane	
			System.out.println("t> " + getTime() + " Repair station allocated");
			
			RepairTask inspect = (RepairTask)request.getEntity();
			inspect.setRepairStation((SpatialResource)resource);
			myInspectionDelay.start(inspect);
			
		}	
	}
	
	private class ActivityCompletionListener implements OperationListenerIfc {
		
		Aircraft myAircraft;
		
		public void update(OperationIfc operation, Task task){
			
			System.out.println("t> " + getTime() + " Inspection Complete");
			
//			 get the repair task			
			RepairTask myTask = (RepairTask)task;
			myAircraft = myTask.getAircraft();
			
			
			if(myAircraft.isFailed()){
				
				System.out.println("Aircraft " + myAircraft.getName() + " Failed");
				// remember the seized repair station
				//myTask.setRepairStation((SpatialResource)resource);
				
				// tell the operation to start, this implies that the completion listener will be called
				// its critical that the RepairOperation class notify its completion listeners
				System.out.println("t> " + getTime() + " Repair operation starting.");
				myRepairOperation.start(task);
			}
			else{	
				
				System.out.println("Aircraft " + myAircraft.getName() + " Not Failed");
				// the repair operation has completed a repair task
				SpatialResource repairStation = myTask.getRepairStation();
				
				System.out.println("t> " + getTime() + " Releasing repair station");
				// release the repair station using the original request
				Request req = task.getRequest(repairStation);
				repairStation.release(req);
				
				// the aircraft is now repaired
				Aircraft aircraft = myTask.getAircraft();
				
				System.out.println("t> " + getTime() + " Aircraft " + aircraft.getName() + " repaired");
				aircraft.setRepairedIndicator();
				
				scheduleAircraftSortie(aircraft);
			}
			
			
		}
		
	}
	
	private class EndRepairOperationListener implements OperationListenerIfc {

		public void update(OperationIfc operation, Task t) {

			System.out.println("t> " + getTime() + " Repair operation ended");
			// cast the task to repair task
			RepairTask task = (RepairTask)t;
			
			// the repair operation has completed a repair task
			SpatialResource repairStation = task.getRepairStation();
			
			System.out.println("t> " + getTime() + " Releasing repair station");
			// release the repair station using the original request
			Request request = task.getRequest(repairStation);
			repairStation.release(request);
			
			// the aircraft is now repaired
			Aircraft aircraft = task.getAircraft();
			
			System.out.println("t> " + getTime() + " Aircraft " + aircraft.getName() + " repaired");
			aircraft.setRepairedIndicator();
			
			scheduleAircraftSortie(aircraft);
			
		}

	}

	public void sensorUpdate(HMSystem system, HMSensor sensor) {
		// Put what the airbase is supposed to do for a sensor update here
		System.out.println("----> The Airbase was notified of a sensor update.");
		
		if (sensor.isComponentFailed()){// sensor is reporting a failure
			Component c = sensor.getComponent();
			Aircraft plane = (Aircraft)c.getAssembly();
			// we now have the plane that is associated with the sensor
			// that is reporting a failure
		}

	}
		

}
