/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.adc;

import jsl.modeling.ActionListenerIfc;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.component.ComponentAssembly;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.random.distributions.Bernoulli;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;

public class Aircraft extends ComponentAssembly {

	protected RandomVariable mySortieTimeRV;
		
	protected RandomVariable myFailureIndicatorRV;
	
	protected TimeWeighted myStatus;
	
	protected ActionListenerIfc mySortieCompletionAction;
	
	protected AirBase myAirbase;

	public Aircraft(ModelElement parent, AirBase airfield) {
		this(parent, airfield, null);
	}

	public Aircraft(ModelElement parent, AirBase airfield, String name) {
		super(parent, name);

		if (airfield == null)
			throw new IllegalArgumentException("The Airfield was null");

		myAirbase = airfield;
		
		DistributionIfc st = new Exponential(2);
		DistributionIfc status = new Bernoulli(0.3);
		
		mySortieTimeRV = new RandomVariable(this, st, getName() + "SortieTime");
		
		myFailureIndicatorRV = new RandomVariable(this, status, getName() + "FailureIndicator");
		
		myStatus = new TimeWeighted(this, getName() + "Status Variable");
		// assume that it starts not failed
		myStatus.setInitialValue(1.0);
		
		mySortieCompletionAction = new SortieCompletionAction();
		
	}

	public final boolean isFailed(){
		return (myStatus.getValue() == 0.0);
	}
	
	protected void initialize() {
		// assume that it starts on a sortie
		beginSortie();	
	}
	
	protected void setRepairedIndicator(){
		myStatus.setValue(1.0);
	}
	
	protected void beginSortie(){		
		if (!isFailed()){
			System.out.println("t> " + getTime() + " " + getName() + " Sortie started.");
			scheduleEvent(mySortieCompletionAction, mySortieTimeRV.getValue());	
		}
	}
	
	protected void sortieCompletionAction(JSLEvent event){
		// the plane has completed its sortie/mission
		// indicate its failure status
		myStatus.setValue(myFailureIndicatorRV.getValue());

		System.out.println("t> " + getTime() + " " + getName() + " Sortie completed.");
		
		// ask the airfield to service the aircraft			
		myAirbase.serviceAircraft(this);
		
	}
	
	// ===========================================
	// INNER CLASSES
	// ===========================================

	private class SortieCompletionAction implements ActionListenerIfc {
		public void action(JSLEvent event) {
			sortieCompletionAction(event);				
		}
	}
}
