/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.adc;


import jsl.modeling.ModelElement;
import jsl.modeling.elements.component.Component;
import jsl.modeling.elements.spatial.RectangularGridSpatialModel2D;
import jsl.modeling.elements.spatial.SpatialResource;
import jsl.modeling.elements.spatial.ResourceLocation;
import jsl.modeling.elements.spatial.SpatialModel;
import jsl.utilities.random.distributions.Exponential;
import jslx.adc.adcsystem.*;
import jslx.adc.hmsystem.BernoulliDetection;
import jslx.adc.hmsystem.DetectionModelIfc;
import jslx.adc.hmsystem.FalseAlarmModelAbstract;
import jslx.adc.hmsystem.HMSensor;
import jslx.adc.hmsystem.HMSystem;
import jslx.adc.hmsystem.RPFalseAlarmModel;


public class TIMEModelHMSDemo extends ModelElement {
	
	protected SpatialModel myAirField;
	
	protected AirBase myAirBase;
	
	protected ComponentBasedAircraft myPlane1;

	
	
	public TIMEModelHMSDemo(ModelElement parent) {
		this(parent, null);
	}

	public TIMEModelHMSDemo(ModelElement parent, String name) {
		super(parent, name);
		
		// create the specfications of the spatial representation for the airfield
		myAirField = new RectangularGridSpatialModel2D();
		
		// create the spatial elements used in the spatial model
		// and set their spatial attributes
		
		// create the airbase
		myAirBase= new AirBase(this, myAirField);
		
		// create the sensor and set up its detection model and false alarm model
		HMSensor hmsensor = new HMSensor(this);
		
		DetectionModelIfc dm = new BernoulliDetection(hmsensor, 0.9);
		hmsensor.setDetectionModel(dm);
		
		FalseAlarmModelAbstract fm = new RPFalseAlarmModel(hmsensor, new Exponential(9.0));
		hmsensor.setFalseAlarmModel(fm);
				
		// create the HMSystem and add the sensor
		HMSystem hms = new HMSystem(this);
		hms.addSensor(hmsensor);
		hms.attachSensorUpdateModel(myAirBase);
				
		// create the planes and attach them to the airbase
		myPlane1 = new ComponentBasedAircraft(this, myAirBase, "Plane 1");
		Component c1 = new Component(this, new Exponential(0.1), new Exponential(.80), new Exponential(.50));
		myPlane1.addComponent(c1);
		
		// place the sensor on the component
		hmsensor.setComponent(c1);
		
		new Aircraft(this, myAirBase, "Plane 2");
		new Aircraft(this, myAirBase, "Plane 3");
		
		// create the repair stations and add them to the airbase
		SpatialResource rs1 = myAirBase.addRepairStation("Repair Station 1", 10,100);
		SpatialResource rs2 = myAirBase.addRepairStation("Repair Station 2", 10,300);
		SpatialResource rs3 = myAirBase.addRepairStation("Repair Station 3",10,500);
		

		// create the resources and add them to the airbase 
		
		Worker w = myAirBase.addWorker("Worker 1", 200,5);
		Worker w2 = myAirBase.addWorker("Worker 2", 400,5);
		Worker w3 = myAirBase.addWorker("Worker 3", 600,5);
		Worker w4 = myAirBase.addWorker("Worker 4", 800,5);
		Worker w5 = myAirBase.addWorker("Worker 5", 1000,5);
		ResourceLocation r1 = myAirBase.addEquipmentLocation(1000,300);
		ResourceLocation r2 = myAirBase.addEquipmentLocation(1000,500);
		ResourceLocation r3 = myAirBase.addEquipmentLocation(1000,700);
		ResourceLocation r4 = myAirBase.addEquipmentLocation(1000,900);
		SpatialResource e = myAirBase.addEquipment("Power Unit 1", r1);
		SpatialResource f = myAirBase.addEquipment("Power Unit 2", r2);
		SpatialResource g = myAirBase.addEquipment("Power Unit 3", r3);
		SpatialResource h = myAirBase.addEquipment("Power Unit 4", r4);
		
		
		// create the ADC system
		ADCSystem n = myAirBase.addNetwork(myAirField);
		n.addReader("Repair Station 1",rs1);
		n.addReader("Repair Station 2",rs2);
		n.addReader("Repair Station 3",rs3);
		n.addReader("Worker Rest Location 1", w.getResourceLocation());
		n.addReader("Worker Rest Location 2", w2.getResourceLocation());		
		n.addReader("Worker Rest Location 3", w3.getResourceLocation());
		n.addReader("Worker Rest Location 4", w4.getResourceLocation());
		n.addReader("Worker Rest Location 5", w5.getResourceLocation());
		n.addReader("Equipment Location 1", r1);
		n.addReader("Equipment Location 2", r2);
		n.addReader("Equipment Location 3", r3);
		n.addReader("Equipment Location 4", r4);
		n.addRFIDTag(w);
		n.addRFIDTag(w2);
		n.addRFIDTag(w3);
		n.addRFIDTag(w4);
		RFIDTag tag = n.addRFIDTag(e);
		n.addRFIDTag(f);
		n.addRFIDTag(g);
		n.addRFIDTag(h);

		
	/*	AnimationTextFileGenerator atfg = null;
		try{
			atfg = new AnimationTextFileGenerator("animation");
		}
		catch(IOException ept){
			System.out.println("I/O exception opening animation file");
			System.exit(1);
		}
		
		MoveableResourceAnimationObserver obs1 = new MoveableResourceAnimationObserver(atfg);
//		MovingElementAnimationObserver obs1 = new MovingElementAnimationObserver(atfg);
		Resource2DAnimationObserver obs2 = new Resource2DAnimationObserver(atfg);
	
		VariableAnimationObserver obs3 = new VariableAnimationObserver(atfg);
		
		w.addObserver(obs1);
		e.addObserver(obs2);
		
		myAirBase.getNumPlanesInRepair().addObserver(obs3);

	*/}

}
