/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.adc;

import java.util.List;
import java.util.ListIterator;

import jsl.modeling.IllegalStateException;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.resource.Request;
import jsl.modeling.elements.spatial.CoordinateIfc;
import jsl.modeling.elements.spatial.EuclideanStepBasedMovementController;
import jsl.modeling.elements.spatial.MobileResource;
import jsl.modeling.elements.spatial.SpatialResource;
import jsl.modeling.elements.spatial.ResourceLocation;
import jsl.modeling.elements.spatial.SpatialElementIfc;
import jsl.modeling.elements.spatial.Vector3D;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.DistributionFunctionIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.modeling.elements.operation.*;
import jsl.observers.Mover2DObserver;

/**
 *
 */
public class Worker extends MobileResource{

	public static enum MoveType {SEARCHING_FOR_EQUIPMENT, MOVING_TO_START_REPAIR, 
		MOVING_TO_REPAIR_WITH_EQUIPMENT, MOVING_TO_RETURN_EQUIPMENT, MOVING_TO_REST_AREA};
	
	/** Holds the current type of move
	 */
	protected MoveType myMoveType;

	/** Holds the current task being worked on by the worker
	 */
	protected RepairTask myCurrentTask;
	
	/** If the worker has found equipment and is using it
	 *  then this variable should be set
	 */
	protected SpatialResource myEquipment;
	
	/** The request associated with the allocated equipment
	 */
	protected Request myEquipmentRequest;
	
	/** If the worker is searching for a piece of equipment this should be used
	 */
	protected ResourceLocation myCandidateEquipmentLocation;
	
	/** This remembers where the worker got the equipment so that it can be returned
	 */
	protected ResourceLocation myEquipmentLocation;
	
	protected ResourceLocation myInitialWorkerLocation;
	
	/** Holds the list of locations to search
	 * 
	 */
	protected List<ResourceLocation> mySearchList;
	
	/** An iterator for moving around on the search list
	 * 
	 */
	protected ListIterator<ResourceLocation> mySearchListIterator;
	
	/** Observes the underlying transporter for individual moves and updates
	 *  the position of the equipment that the worker is using
	 */
	protected UpdateEquipmentPositionObserver myUpdateEquipmentPositionObserver = new UpdateEquipmentPositionObserver();
	
	protected DistributionIfc myDistribution;
	
	protected Operation myRepairDelay;
	
	protected ActivityCompletionListener myCompletionListener;
	
	// ===========================================
	// CONSTRUCTORS
	// ===========================================

	/** Creates a Worker with the given parent at location (0.0, 0.0). The SpatialModel2D of 
	 *  the parent is used as the SpatialModel2D.  If the parent does not have a SpatialModel2D, 
	 *  then an IllegalArgumentException is thrown.
	 * 
	 * @param parent
	 */
	public Worker(ModelElement parent){
		this(parent, null, null);
	}
	
	/** Creates a Worker with the given parent at location (0.0, 0.0). The SpatialModel2D of 
	 *  the parent is used as the SpatialModel2D.  If the parent does not have a SpatialModel2D, 
	 *  then an IllegalArgumentException is thrown.
	 * 
	 * @param parent
	 * @param name
	 */
	public Worker(ModelElement parent, String name){
		this(parent, name, null);
	}
	
	/** Creates a Worker with the given parent at location (x, y). The SpatialModel2D of 
	 *  the parent is used as the SpatialModel2D.  If the parent does not have a SpatialModel2D, 
	 *  then an IllegalArgumentException is thrown.
	 * 
	 * @param parent
	 * @param x
	 * @param y
	 */
	public Worker(ModelElement parent, double x, double y){
		this(parent, null, x, y, 0.0);
	}

	/** Creates a Worker at the same (x,y) coordinates as the supplied position.  The SpatialModel2D of 
	 *  the parent is used as the SpatialModel2D.  If the parent does not have a SpatialModel2D, 
	 *  then an IllegalArgumentException is thrown.	 
	 *   
	 * @param parent
	 * @param position
	 */
	public Worker(ModelElement parent, CoordinateIfc position){
		this (parent, null, position);
	}

	/** Creates a Worker at the same (x,y) coordinates as the supplied position.  The SpatialModel2D of 
	 *  the parent is used as the SpatialModel2D.  If the parent does not have a SpatialModel2D, 
	 *  then an IllegalArgumentException is thrown.	 
	 *   
	 * @param parent
	 * @param position
	 */
	public Worker(ModelElement parent, SpatialElementIfc position){
		this (parent, null, position.getCurrentPosition());
	}
	
	/** Creates a Worker with the given parent at location (x, y). The SpatialModel2D of 
	 *  the parent is used as the SpatialModel2D.  If the parent does not have a SpatialModel2D, 
	 *  then an IllegalArgumentException is thrown.
	 * 
	 * @param parent
	 * @param name
	 * @param x
	 * @param y
	 * @param z
	 */
	public Worker(ModelElement parent, String name, double x, double y){
		this(parent, name, new Vector3D(x,y,0.0));
	}
	
	/** Creates a Worker with the given parent at location (x, y). The SpatialModel2D of 
	 *  the parent is used as the SpatialModel2D.  If the parent does not have a SpatialModel2D, 
	 *  then an IllegalArgumentException is thrown.
	 * 
	 * @param parent
	 * @param name
	 * @param x
	 * @param y
	 * @param z
	 */
	public Worker(ModelElement parent, String name, double x, double y, double z){
		this(parent, name, new Vector3D(x,y,z));
	}
	
	/** Creates a Worker with the given parent at location (x, y). The SpatialModel2D of 
	 *  the parent is used as the SpatialModel2D.  If the parent does not have a SpatialModel2D, 
	 *  then an IllegalArgumentException is thrown.
	 * 
	 * @param parent
	 * @param name
	 * @param coordinate
	 */
	public Worker(ModelElement parent, String name, CoordinateIfc coordinate){
		super(parent, 1, name, null, coordinate);
	
		getTransporter().addObserver(myUpdateEquipmentPositionObserver);
		
		EuclideanStepBasedMovementController c = new EuclideanStepBasedMovementController(this);
		setMovementController(c);
		myDistribution = new Exponential(5);
		myRepairDelay = new Activity(this, myDistribution);
		myCompletionListener = new ActivityCompletionListener();
		
		myRepairDelay.addCompletionListener(myCompletionListener);
		
		myInitialWorkerLocation= new ResourceLocation(this, getName() + " Worker Rest Location", coordinate);
		System.out.println(getName() + " Location" + myInitialWorkerLocation.getCurrentPosition());
		
		//System.out.println("Worker Created.");
	
	}
	
	// ===========================================
	// PUBLIC METHODS
	// ===========================================

	/** Tells the worker to start performing the given repair task. If the
	 *  worker is not at the repair station the worker moves to the 
	 *  repair station.
	 * 
	 * @param rtask
	 */
	protected final void performTask(RepairTask rtask){
		// what does the worker to with the task
		// the worker needs to search for the piece of equipment
		// searching may take simulated time so remember the task
		myCurrentTask = rtask;
		
		// setup the iterator for searching for equipment
		setupSearchList(rtask);
		
		// check if at the repair station for the task
		SpatialResource repairStation = myCurrentTask.getRepairStation();
		
		if (this.isPositionEqualTo(repairStation)){// checks if worker location is the same as repairstation
			// worker is at repair station, begin search for equipment
			searchForEquipment();
		} else { // move empty to the repairStation's location
			System.out.println("t> " + getTime() + " " + getName() + " is moving to start repair.");
			myMoveType = MoveType.MOVING_TO_START_REPAIR;
			moveEmpty(repairStation);
		}

	}
	
	/** Sets up the list to find the equipment locations, this is called
	 *  within performTask() using the current task
	 * 
	 * @param base
	 */
	protected void setupSearchList(RepairTask rtask){
		System.out.println("t> " + getTime() + " Setting up the equipment location search list.");
		AirBase base = rtask.getAirBase();
		// simple search in order of the list, this can be varied by sorting, etc
		mySearchList = base.getEquipmentLocationList();
		//System.out.println(mySearchList);
		mySearchListIterator = mySearchList.listIterator();	
		
	}
	
	/** This method should provide the next equipment location with the search
	 *  or null if no locations are available
	 * @return
	 */
	protected ResourceLocation findNextEquipmentLocation(){
		if (mySearchListIterator != null){
			if (mySearchListIterator.hasNext())
				return(mySearchListIterator.next());	
			else
				return(null);
		}
		else
			return(null);
	}
	
	/** Executes the basic logic for searching for equipment. 
	 *  1) uses findNextEquipmentLocation()
	 *  2) if the location is not null, checks if worker is at the location
	 *     if so equipmentLocationFound() is called, else the worker begins
	 *     moving to the desired location using moveEmpty()
	 *  3) if the search location is null, equipmentLocationNotFound() is called
	 *     (which can be overridden to provide behavior to continue or stop the search)
	 */
	protected void searchForEquipment(){
		System.out.println("t> " + getTime() + " " + getName() + " is searching for equipment.");
		// determine location to search at		
		myCandidateEquipmentLocation = findNextEquipmentLocation();
		if (myCandidateEquipmentLocation != null){
			//System.out.println("t> " + getTime() + " Equipment location was not null in searchForEquipment()");
			// is location of equipment the same as worker
			if(isPositionEqualTo(myCandidateEquipmentLocation)){
				System.out.println("t> " + getTime() + " isPosition() true for candidate location " + myCandidateEquipmentLocation.getName());
				// equipment location is the same location as the worker
				equipmentLocationFound(myCandidateEquipmentLocation);
			} else {// location of equipment is somewhere else
				// move to get equipment
				myMoveType = MoveType.SEARCHING_FOR_EQUIPMENT;
				System.out.println("t> " + getTime() + " " + getName() + " is moving in search of equipment " );
				moveEmpty(myCandidateEquipmentLocation);
			}
		} else {
			System.out.println("t> " + getTime() + " Equipment location was null in searchForEquipment()");
			equipmentLocationNotFound();
		}
	}
	
	/** Can be overridden by subclasses to provide behavior if no valid
	 *  equipment location is found while searching for equipment. The default
	 *  throws a RunTimeException, NoValidLocationFoundException()
	 */
	protected void equipmentLocationNotFound(){

		if (mySearchListIterator == null)
			throw new NoValidLocationFoundException("No equipment location found. Iterator was null.");
		
		if (mySearchList == null)
			throw new NoValidLocationFoundException("No equipment location list found. List was null.");
		
		System.out.println("t> " + getTime() + " Equipment location not found.");
		System.out.println("t> " + getTime() + " resetting the search list.");
		
		// reset up the search list
		setupSearchList(myCurrentTask);
		
		// find out if the iterator has locations, if not big problem, throw exception
		if (!mySearchListIterator.hasNext()){
			throw new NoValidLocationFoundException("No equipment location found");
		} else { // reset iterator has locations
			// find out if the current location is the same as the next in search
			if(myCandidateEquipmentLocation == mySearchList.get(0)){
				// advance the iterator by 1 location
				mySearchListIterator.next();
				// check if another location exists
				if (!mySearchListIterator.hasNext())
					throw new NoValidLocationFoundException("No equipment location found");
				else
					searchForEquipment();
			} else {// continue the search
				searchForEquipment();
			}
		}
	}
	
	/** Uses the supplied location to find if it has any idle equipment 
	 *  that the worker can use.  If no idle equipment is at the location, then
	 *  noIdleEquipmentAtCurrentLocation() is called
	 * 
	 * @param location
	 */
	protected void equipmentLocationFound(ResourceLocation location){
		System.out.println("t> " + getTime() + " Equipment location found.");

		SpatialResource equipment = location.findIdleResource2D();
		if (equipment != null){ // idle equipment found
			myEquipmentRequest = Request.create();
			equipment.allocate(myEquipmentRequest);
			myEquipment = equipment;
			myEquipmentLocation = location;
			myMoveType = MoveType.MOVING_TO_REPAIR_WITH_EQUIPMENT;
			System.out.println("t> " + getTime() + " " + getName() + " is moving equipment to repair station.");
			transport(myCurrentTask.getRepairStation());
		} else { // no idle equipment at this resource location, continue search
			System.out.println("t> " + getTime() + " No idle equipment found at location.");
			noIdleEquipmentAtCurrentLocation(location);
		}
		
	}
	
	/** The default behavior is to continue the search by calling searchForEquipment()
	 *  but this could be overridden in subclasses
	 * 
	 * @param location
	 */
	protected void noIdleEquipmentAtCurrentLocation(ResourceLocation location){
		searchForEquipment();
	}
	
	/** This method is called after the worker completes an empty move while
	 *  searching for equipment.  The worker will either get the equipment
	 *  using equipmentLocationFound() or continue the search
	 *
	 */
	protected void searchingEmptyMoveCompleted(){
		//System.out.println("searchingEmptyMoveCompleted Worker is at X: " + getX() + " Y: " + getY());
		//System.out.println("candidate location  " + myCandidateEquipmentLocation + " X: " + myCandidateEquipmentLocation.getX() + " Y: " + myCandidateEquipmentLocation.getY());
			
		
		if(isPositionEqualTo(myCandidateEquipmentLocation)){
			// equipment is at same location as the worker
			System.out.println("Worker Postion " + getCurrentPosition());
			equipmentLocationFound(myCandidateEquipmentLocation);
		} else {// location of equipment is somewhere else
			// continue search for equipment
			searchForEquipment();
		}
		
	}
	
	/** This method is called after the worker completes an empty move
	 *  while going to the repair station to start a new task
	 *
	 */
	protected void movingToStartRepairCompleted(){
		// at the repair station, need to get equipment
		searchForEquipment();
	}
	
	/** This method is called after the worker completes an empty move
	 *  while returning to the designated rest area
	 */
	protected void movingToRestingAreaCompleted(){
		
		// get the request on the worker
		Request r = myCurrentTask.getRequest(this);
		System.out.println("t> " + getTime() + " " + getName() + " is releasing request: " + r);
		//System.out.println("X: " + getX() + " Y: " + getY());
		release(r);
		// how will operation know its complete???
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.resource.MoveableResource#emptyMoveCompleted()
	 */
	protected void emptyMoveCompleted() {
		
		switch (myMoveType){
		case SEARCHING_FOR_EQUIPMENT:
			System.out.println("t> " + getTime() + " " + getName() + " completed move searching for equipment");
			searchingEmptyMoveCompleted();
			break;
		case MOVING_TO_START_REPAIR:
			System.out.println("t> " + getTime() + " " + getName() + " moved to start repair");
			movingToStartRepairCompleted();
			break;
		case MOVING_TO_REST_AREA:
			System.out.println("t> " + getTime() + " " + getName() + " moved to rested area.");
			movingToRestingAreaCompleted();
			break;
		default:
			throw new IllegalStateException("Invalid state for empty move complete.");			
		}
		
	}

	/** This method is called after the worker completes a transport move 
	 *  (transporting the equipment to the repair station).
	 */
	protected void movingToRepairWithEquipmentCompleted(){
		
		SpatialResource repairStation = myCurrentTask.getRepairStation();

		System.out.println("t> " + getTime() + " " + getName() + " started the repair.");
		myRepairDelay.start(myCurrentTask);
		
		
	}
	
	/** This method is called after the worker completes a transport move
	 *  (transporting the equipment back to the location that the worker got it)
	 *  The worker releases the equipment an then moves to the rest area
	 */
	protected void movingToReturnEquipmentCompleted(){
		System.out.println("t> " + getTime() + " " + getName() + " is releasing equipment.");

		myEquipment.release(myEquipmentRequest);
		myEquipment = null;
		myEquipmentRequest = null;
		myEquipmentLocation = null;
		myMoveType = MoveType.MOVING_TO_REST_AREA;
		System.out.println("t> " + getTime() + " " + getName() + " is moving to rest area.");
		//moveEmpty(getInitialX(),getInitialY());

		moveEmpty(myInitialWorkerLocation);
		
		//System.out.println("X: " + myInitialWorkerLocation.getInitialX() + " Y: " + myInitialWorkerLocation.getInitialY());
	}
	
	/** This method should be called when the worker is done using the equipment
	 *  and needs to return the equipment to its location
	 */
	protected void returnEquipment(){
		// how is the Worker told to return the equipment, i.e. how is returnEquipment() called?
		myMoveType = MoveType.MOVING_TO_RETURN_EQUIPMENT;
		System.out.println("t> " + getTime() + " " + getName() + " is returning equipment.");
		transport(myEquipmentLocation);
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.resource.MoveableResource#transportCompleted()
	 */
	protected void transportCompleted() {

		switch (myMoveType){
		case MOVING_TO_REPAIR_WITH_EQUIPMENT:
			System.out.println("t> " + getTime() + " " + getName() + " moved to repair with equipment.");
			movingToRepairWithEquipmentCompleted();
			break;
		case MOVING_TO_RETURN_EQUIPMENT:
			System.out.println("t> " + getTime() + " " + getName() + " returned equipment");
			movingToReturnEquipmentCompleted();
			break;
		default:
			throw new IllegalStateException("Invalid state for transport complete.");			
		}
		
	}
	
	public void setActivityDelay(DistributionIfc distribution){
		myDistribution = distribution;
	}
	
	public DistributionFunctionIfc getActivityDelay(){
		return myDistribution;
	}
		

	
	protected void setInitialLocation(ResourceLocation location){
		myInitialWorkerLocation = location;
	}
	
	public ResourceLocation getResourceLocation(){
		return myInitialWorkerLocation;
	}
	
	

	/** This inner class ensures that the equipment's position is updated when
	 *  the worker is moving with the equipment
	 * 
	 * @author rossetti
	 *
	 */
	private class UpdateEquipmentPositionObserver extends Mover2DObserver {

		protected void moveEnded(ModelElement sm, Object arg) {
			if (myEquipment != null){
				myEquipment.setPosition(Worker.this);
			}
		}
	}
	
	private class ActivityCompletionListener implements OperationListenerIfc {
		
		public void update(OperationIfc operation, Task task){
			
			RepairOperation ro = myCurrentTask.getRepairOperation();
			ro.notifyOperationCompletedListeners(myCurrentTask);
			returnEquipment();
			
		}
		
	}
}

