/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.adc.adcsystem;

import java.util.*;

import jsl.utilities.IdentityIfc;
import jsl.modeling.elements.spatial.CoordinateIfc;
import jsl.modeling.elements.spatial.SpatialResource;
import jsl.modeling.elements.spatial.ResourceLocation;
import jsl.modeling.elements.spatial.SpatialElementIfc;
import jsl.modeling.elements.spatial.SpatialModel;


/**
 */

public class ADCSystem implements IdentityIfc{
	
	/**
	 * Assigns a name to the network that is being used.
	 */

	private String mySystemName;
	
	/**
	 * This creates a counter that gives each ADCSystem a unique identification.  
	 */
	
	private static long ADCSystem_Counter;
	
	private long ADCSystemID;
	
	/**
	 * This is a list of readers that are part of the ADCSystem.  This is set in the addReader() method.
	 */
	
	private List<RFIDReader> myReaderList;

	/**
	 * This list represents the RFID Tags that belong to this ADCSystem.  This is set in the addRFIDTag() method.
	 */
	
	private List<RFIDTag> myTagList;

	/**
	 * This is a list of RFID Tags that are attached to Resource2D objects.  This is set in the addRFIDTag() method.
	 */
	
	private List<RFIDTag> myEquipmentTagList;

	/**
	 * This is the spatial model that the ADCSystem belongs to.  This is used to create spatial elements for readers
	 * and other objects that have a spatial element.
	 */
	
	private SpatialModel mySpatialModel;
	

	/**
	 * This constructor is used when an ADCSystem is created without a reference for a name.  
	 */
	public ADCSystem(SpatialModel sm){
		
		this(null, sm);
		
		
	}
		
	/**
	 * This constructor is called when an ADCSystem is created with a String representing the name of the system.
	 * 
	 * @param name
	 */
	public ADCSystem(String name,SpatialModel sm){
		
		ADCSystem_Counter++;
		ADCSystemID = ADCSystem_Counter;
		mySystemName = name;
		myReaderList = new ArrayList<RFIDReader>();
		myTagList = new ArrayList<RFIDTag>();
		myEquipmentTagList = new ArrayList<RFIDTag>();
		mySpatialModel = sm;
		
	}
	
	/**
	 * This method adds a tag to the the airbase.  This instance is called when a name for the tag is not given.
	 * @param spo
	 * @return
	 */
	public RFIDTag addRFIDTag(SpatialElementIfc spo){
		
		return addRFIDTag(null, spo);
		
	}
		
	/**
	 * This method adds a tag to the the airbase.  It also attaches the tag as an observer of an object.
	 * The tag is added to the network's list of tags.  This is used when checking for tags in reader range.
	 * The tag is added to the equipment tag list if it is attached to a Resource2D.
	 * @param name
	 * @param spo
	 * @return
	 */
	public RFIDTag addRFIDTag(String name, SpatialElementIfc spo){
		RFIDTag s = new RFIDTag(name, spo);
		myTagList.add(s);
		s.setADCSystem(this);
		spo.attachPositionObserver(s);
		if (spo instanceof SpatialResource){
			System.out.println("Tag Added!");
			myEquipmentTagList.add(s);
		}
		return(s);
	}
	
	
	/**
	 * This method adds a reader to the network and spatial model.  The reader is added to the network's
	 * reader list.
	 * @param x
	 * @param y
	 * @param name
	 * @return
	 */
	public RFIDReader addReader(String name, SpatialElementIfc po){
		RFIDReader r = new RFIDReader(name,this, mySpatialModel,po); //TODO how is mySpatialModel set???
		System.out.println(r.getX1() + ", " + r.getX2());
		myReaderList.add(r);
		return(r);
	}
	
	/**
	 * This method adds a reader to the network and spatial model.  This method is called when a name is not given
	 * to the reader.  
	 * @param po
	 */
	public void addReader(SpatialElementIfc po){
		addReader(null, po);
	}
	
	/**
	 * This update method is called any time a tag is updated to a new position.
	 *
	 */
	protected void tagUpdate(){
		for(RFIDReader r: myReaderList){
			r.checkForTags();
		}
	}
	
	/**
	 * This method is used to get the list of RFID tags that belong to the ADCSystem.
	 * @return
	 */
	protected List<RFIDTag> getRFIDTagList(){
		return myTagList;
	}
	
	/**
	 * This method is used to obtain all RFID tag locations in the system.
	 */
	public void getAllSensorLocations(){
		for(RFIDTag t: myTagList){
			System.out.println("Sensor: " + t.getID()+ " Location: " + t.getX1() + ", " + t.getX2());
		}
	}
	
	/**
	 * This method compiles a list of Resource2D in the model by checking the readers to find all tags 
	 * attached to them.  
	 * @return
	 */
	
	protected List<SpatialResource> getEquipmentList(){
		System.out.println("----- Getting Equipment List:");
		List<SpatialResource>  rList = new ArrayList<SpatialResource>();
		for(RFIDReader rr: myReaderList){
			System.out.println("Checking reader " + rr.getName() + " X= " + rr.getX1() + " Y= " + rr.getX2());
			rr.checkForTags();
			List<RFIDTag> InRange = rr.getTagsInRange();
			System.out.println("Tags in Range");
			System.out.println(InRange);
			for(RFIDTag rt: InRange){
				if(rt.getSpatialElement() instanceof SpatialResource){
					SpatialResource equip = (SpatialResource)rt.getSpatialElement();
					if(!rList.contains(equip))
						rList.add(equip);
				}
			}
		}
		System.out.println("----- Done Getting Equipment List");
		System.out.println("----- Equipment List");
		System.out.println(rList);
		return rList;
	}

	/**
	 * This method compiles a list of resource locations for every Resource2D in the model by checking
	 * the readers to find all tags attached to them.  
	 * @return
	 */
	
	public List<ResourceLocation> getEquipmentLocationList(){
		List<ResourceLocation> rList = new ArrayList<ResourceLocation>();
		for(RFIDReader rr: myReaderList){
			rr.checkForTags();
			List<RFIDTag> InRange = rr.getTagsInRange();
			for(RFIDTag rt: InRange){
				if(rt.getSpatialElement() instanceof SpatialResource){
					SpatialResource equip = (SpatialResource)rt.getSpatialElement();
					ResourceLocation location = equip.getResourceLocation();
					if(!rList.contains(location)){
						if(equip.isPositionEqualTo(location))
							rList.add(location);
					}	
				}
				if (rList.isEmpty()){
					for(RFIDTag r: myEquipmentTagList){
						SpatialResource eq = (SpatialResource)r.getSpatialElement();
						ResourceLocation loc = eq.getResourceLocation();
						if(!rList.contains(loc)){
							if(eq.isPositionEqualTo(loc))
								rList.add(loc);
						}
					}
					return rList;
				}
			}
		}
		return rList;
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.IdentityIfc#getName()
	 */
	public String getName() {
		// TODO Auto-generated method stub
		return mySystemName;
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.IdentityIfc#getId()
	 */
	public long getId() {
		// TODO Auto-generated method stub
		return ADCSystemID;
	}	
	
}