/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.adc.adcsystem;

import jsl.utilities.IdentityIfc;
import jsl.utilities.random.distributions.Bernoulli;

import java.util.*;

import jsl.modeling.elements.spatial.CoordinateIfc;
import jsl.modeling.elements.spatial.SpatialElementIfc;
import jsl.modeling.elements.spatial.SpatialModel;


/**
 */
public class RFIDReader implements IdentityIfc{  
	
	/**
	 * Sets a defualt range for all readers in the model.  
	 */
	
	//public static final double DEFAULT_RANGE = 5.0;
	
	//private static double myDefaultRange = DEFAULT_RANGE;
	
	public static final double DEFAULT_ERROR = 0.05;
	
	private static double myDefaultErrorRate = DEFAULT_ERROR;
	
	/**
	 * Defines the reading range of the reader.
	 */
	
	private RangeFunctionIfc myRangeFunction;
	
	private double myErrorRate;
	
	/**
	 * Allows for all RFID tags that are within the reader's range to be put in a list.
	 */
	
	private List<RFIDTag> myTagsInRange;
	
	/**
	 * This is to allow for each reader in the model to have a unique identification.
	*/
	
	private static long myID;
	
	/**
	 * This assigns a name to the reader.  This name can represent the area that
	 */
	
	private String myName;
	
	/**
	 * This is a reference to the ADCSystem that the reader belongs to.
	 */
	
	private ADCSystem myADCSystem;
	
	/**
	 * This is a referece to the object that the reader is located at.  This object can be anything that implements
	 * the Position2DIfc.  
	 */
	
	private SpatialElementIfc mySpatialElement;
	
	protected SpatialModel mySpatialModel;
	
	/**
	 * This constructor creates an RFID reader that is placed at the location of the Position2DIfc.  There is a
	 * reference to the ADC system and the spatial model.  Range is set as the default range value.  
	 * @param n
	 * @param sm
	 * @param po
	 */
	public RFIDReader(ADCSystem n, SpatialModel sm, SpatialElementIfc po){
		this("Zone 1",null,myDefaultErrorRate,n, sm, po);		
	}
		
	/**
	 * This constructor creates an RFID reader that is placed at the location of the Position2DIfc.  There is a
	 * reference to the ADC system and the spatial model.  Range is set as the default range value.  This constructor
	 * contains a reference to the name of the area the reader resides in.
	 * @param name
	 * @param n
	 * @param sm
	 * @param po
	 */
	public RFIDReader(String name, ADCSystem n, SpatialModel sm, SpatialElementIfc po){
		this(name,null,myDefaultErrorRate,n, sm, po);		
	}
	
	/**
	 * This constructor creates an RFID reader that is placed at the location of the Position2DIfc.  There is a
	 * reference to the ADC system and the spatial model.  Range can be set as a double in this constructor.  
	 * @param r
	 * @param n
	 * @param sm
	 * @param po
	 */
	public RFIDReader(RangeFunctionIfc r, ADCSystem n, SpatialModel sm, SpatialElementIfc po){
		this("Zone 1",r,myDefaultErrorRate,n, sm, po);	
	}
	
	/**
	 * This constructor creates an RFID reader that is placed at the location of the Position2DIfc.  There is a
	 * reference to the ADC system and the spatial model.  Range can be set as a double in this constructor.  This 
	 * constructor contains a reference to the name of the area the reader resides in.
	 * @param name
	 * @param r
	 * @param n
	 * @param sm
	 * @param po
	 */
	public RFIDReader(String name, RangeFunctionIfc r, double e, ADCSystem n, SpatialModel sm, SpatialElementIfc po){
		myID++;
		myName = name;
		mySpatialElement = po;
		mySpatialModel = sm;
		setRangeFunction(r);
		setNetwork(n);
		myTagsInRange = new ArrayList<RFIDTag>();
		setErrorRate(e);
		
	}

	private void setErrorRate(double e) {
		myErrorRate = e;
		
	}

	public final long getId(){
		
		return(myID);
	}
	
	public final String getName(){
		return(myName);
	}	
	
	/**
	 * This method sets the range as a double.  If a negative value is passed in, then an illegal argument exception
	 * occurs.
	 * @param r
	 */
	public void setRangeFunction(RangeFunctionIfc r){
		if(r == null){
			myRangeFunction = new CircularRangeFunction(this,10);
			
		}else{
			myRangeFunction = r;
		}
	}
	
	/**
	 * This method returns the reader's spatial element X and Y values.
	 * @return
	 */
	
	public CoordinateIfc getCurrentLocation(){
		return mySpatialElement.getCurrentPosition();
	}
	
	public double getX1(){
		return mySpatialElement.getCurrentPosition().getX1();
	}
	
	public double getX2(){
		return mySpatialElement.getCurrentPosition().getX2();
	}
	
	public double getX3(){
		return mySpatialElement.getCurrentPosition().getX3();
	}
	
	
	
	/**
	 * This method checks for tags that are within the reader's range.  This is done by calling up the network's 
	 *  tag list and checking for RFID tags in the network.  A Bernoulli random variable is created to simulate the
	 *  error rate of the reader.  
	 */
	
	public void checkForTags(){
		List<RFIDTag> TagList = myADCSystem.getRFIDTagList();
		int j = 0;
		Bernoulli b = new Bernoulli(1 - myErrorRate);
		myTagsInRange.clear();
		for(RFIDTag t: TagList){
			if(myRangeFunction.isTagInRange(t)){
				if(b.getValue()== 1){
					//System.out.println("1");
					if (!myTagsInRange.contains(t))
						myTagsInRange.add(t);
						j++;
				}
				//else
					//System.out.println("0");
				
			}
		}
	}
	
	
	
	/**
	 *  This method creates a list of tags that are in the range of the reader.  This method is only consistent
	 *  with the last instance of the checkForTags method.
	 */
	public List<RFIDTag> getTagsInRange(){
		return myTagsInRange;	
	}
	
	/**
	 * This method checks the distance between the tag and reader to determine whether the tag is within
	 * the reader's range.
	 * @param tag
	 * @return
	 */
	/*public boolean isTagInRange(RFIDTag tag){
		double xDistance = Math.abs(getX() - tag.getX());
		double yDistance = Math.abs(getY() - tag.getY());
		if(Math.sqrt((xDistance * xDistance) + (yDistance * yDistance))<= myRange)
			return(true);
		else
			return(false);
	}*/
	
	protected void setNetwork(ADCSystem n){
		myADCSystem = n;
	}	
}
