/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.adc.adcsystem;

import java.util.*;

import jsl.modeling.elements.spatial.CoordinateIfc;
import jsl.modeling.elements.spatial.SpatialElementIfc;
import jsl.utilities.IdentityIfc;

/**
 */
public class RFIDTag implements IdentityIfc, Observer{
	
	/**
	 * Assigns an Electronic Product Code (EPC) to the tag.
	 */
	
	private String tagName;
	/**
	 * If no EPC is given, the tag is given an identification number that is unique.
	 */
	private static int tagID;
	
	/**
	 * Specifies the size of the tag in millimeters.
	 */
	
	private double length;
	private double width;
	
	/**
	 * Specifies the cost of the tag.
	 */

	private double cost;
	
	/**
	 * Specifies the weight of the tag in kilograms.
	 */

	private double weight;
	
	/**
	 * Assigns the tag to an object that implements the Position2DIfc.
	 */

	private SpatialElementIfc mySpatialElement;
	
	private ADCSystem myNetwork;	
	
	/**
	 * This creates an instance of a tag which has been attached to an object.
	 * @param obs
	 */
	protected RFIDTag(SpatialElementIfc obs){
		this(null,0,0,0,0,obs);		
	}

	/**
	 * This constructor creates a tag which has been attached to an object and has been given a name.
	 * @param name
	 * @param obs
	 */
	protected RFIDTag(String name, SpatialElementIfc obs){
		this(name,0,0,0,0,obs);	
	}
	
	/**
	 * This constructor creates a tag which has been attached to an object and has been given a name.  For this 
	 * tag, length, width, cost, and weight have also been entered as part of this tag.
	 * @param name
	 * @param l
	 * @param w
	 * @param c
	 * @param wt
	 * @param obs
	 */
	protected RFIDTag(String name, double l, double w, double c, double wt, SpatialElementIfc obs){
		tagID++;
		setInformation(name,l,w,c,wt);
		setSpatialElement(obs);
		
				
	}
	
	/**
	 * The update method is called everytime the object that the tag is attached to changes position.  This
	 * method calls the tagUpdate() method in the network class.
	 */
	
	public void update(Observable observable, Object arg){
				
		myNetwork.tagUpdate();
		
	}
	
	protected void setSpatialElement(SpatialElementIfc se){
		if (se == null){
			throw new IllegalArgumentException("Spatial Element cannot be null!");
		}
		mySpatialElement = se;
		
	}
	
	protected void setADCSystem(ADCSystem n){
		myNetwork = n;
	}
	
	protected SpatialElementIfc getSpatialElement(){
		return mySpatialElement;
	}
	
	public CoordinateIfc getCurrentPosition(){
		
		return mySpatialElement.getCurrentPosition();
	}
	
	public double getX1(){
		return mySpatialElement.getCurrentPosition().getX1();
	}
	
	public double getX2(){
		return mySpatialElement.getCurrentPosition().getX2();
	}
	
	public double getX3(){
		return mySpatialElement.getCurrentPosition().getX3();
	}
	
	protected void setTagName(String s){
		tagName = s;
	}
	
	protected void setWeight(double w){
		weight = w;
	}
	
	protected double getWeight(){
		return weight;
	}
	
	protected double getLength(){
		return length;
	}
	
	protected void setLength(double l){
		length = l;
	}
	
	protected double getWidth(){
		return width;
	}
	
	protected void setWidth(double w){
		width = w;
	}
	
	protected double getCost(){
		return cost;
	}
	
	protected void setCost(double c){
		cost = c;
	}
	
	protected void setInformation(String epc, double length, double width, double cost, double weight){
		setTagName(epc);
		setLength(length);
		setWidth(width);
		setCost(cost);
		setWeight(weight);
	}
	
	public long getId(){
		long l = (long)tagID;
		return l;
	}
	
	public String getName(){
		String s = "Tag: " + tagID + " " + tagName;
		return s;
	}
	
	protected int getID(){
		return tagID;
	}
	
	public String toString(){
		String s = getName() + " X= " + getX1() + " Y = " + getX2() + "\n";
		s = s + "Name: " + tagName + "\n";
		s = s + "Length: " + length + "\n";
		s = s + "Width: " + width + "\n";
		s = s + "Cost: " + cost + "\n";
		s = s + "Weight: " + weight + "\n";
		return(s);
	}
}
