/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.adc.hmsystem;

import jsl.modeling.ActionListenerIfc;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.component.Component;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.RandomIfc;

public abstract class FalseAlarmModelAbstract extends SchedulingElement {

	protected RandomVariable myTimeToFalseAlarmRV;
	
	protected FalseAlarmListener myFalseAlarmListener = new FalseAlarmListener();
	
	protected JSLEvent myFalseAlarmEvent;

	public FalseAlarmModelAbstract(ModelElement parent, RandomIfc timeToFalseAlarm) {
		this(parent, timeToFalseAlarm, null);
	}

	public FalseAlarmModelAbstract(ModelElement parent, RandomIfc timeToFalseAlarm, String name) {
		super(parent, name);
		setTimeToFalseAlarmInitialRandomSource(timeToFalseAlarm);
	}

	/** Gets the distribution associated with the time to a false alarm
	 * 
	 * @return
	 */
	public final RandomIfc getTimeToFalseAlarmInitialRandomSource(){
		return(myTimeToFalseAlarmRV.getInitialRandomSource());
	}
	
	/** Sets the distribution associated with the time to a false alarm
	 * 
	 * @param timeToFalseAlarm
	 */
	public final void setTimeToFalseAlarmInitialRandomSource(RandomIfc timeToFalseAlarm){
		
		if (timeToFalseAlarm == null) 
			throw new IllegalArgumentException("The time to false alarm cdf was null.");
		
		if (myTimeToFalseAlarmRV == null){
			myTimeToFalseAlarmRV = new RandomVariable(this, timeToFalseAlarm);
		} else
			myTimeToFalseAlarmRV.setInitialRandomSource(timeToFalseAlarm);
			
	}
	
	/** Automatically called when the sensor indicates that the component 
	 *  transitioned to the available state
	 * 
	 * @param sensor
	 * @param c
	 */
	abstract protected void componentAvailable(HMSensor sensor, Component c);

	/** Automatically called when the sensor indicates that the component 
	 *  transitioned to the unavailable state
	 * 
	 * @param sensor
	 * @param c
	 */
	abstract protected void componentUnavailable(HMSensor sensor, Component c);

	/** Automatically called when the sensor indicates that the component 
	 *  transitioned to the repairing state
	 * 
	 * @param sensor
	 * @param c
	 */
	abstract protected void componentStartedRepair(HMSensor sensor, Component c);

	/** Automatically called when the sensor indicates that the component 
	 *  transitioned to the failed state
	 * 
	 * @param sensor
	 * @param c
	 */
	abstract protected void componentFailed(HMSensor sensor, Component c);
	
	/** Automatically called when the sensor indicates that the component 
	 *  transitioned to the operating state
	 * 
	 * @param sensor
	 * @param c
	 */
	abstract protected void componentStartedOperating(HMSensor sensor, Component c);

	/** This method is called when the false alarm occurs
	 * 
	 * @param sensor
	 */
	protected void falseAlarmAction(HMSensor sensor){
		sensor.falseAlarmModelAction();
	}
	
	private class FalseAlarmListener implements ActionListenerIfc {
		public void action(JSLEvent evt) {
			falseAlarmAction((HMSensor)evt.getMessage());
			// once the alarm event occurs, nullify it to indicate that no alarm is pending
			myFalseAlarmEvent = null;
		}
	}
}
