/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.adc.hmsystem;

import jsl.modeling.ModelElement;
import jsl.modeling.StateAccessorIfc;
import jsl.modeling.elements.component.Component;
import jsl.modeling.elements.component.ComponentStateChangeListener;
import jsl.utilities.reporting.JSL;

public class HMSensor extends ComponentStateChangeListener {

	//  ===========================================
	//      CLASS CONSTANTS
	//  ===========================================

	/** Used to represent that the component the sensor is observing 
	 *  has been made available for use
	 * 
	 */
	public final static int AVAILABLE = JSL.getNextEnumConstant();
	
	/** Used to represent that the component has signaled a repair
	 * 
	 */
	public final static int IN_REPAIR = JSL.getNextEnumConstant();

	/** Used to represent that the component the sensor is observing 
	 *  has been deactivated or made unavailable
	 * 
	 */
	public final static int UNAVAILABLE = JSL.getNextEnumConstant();

	/** Indicates the working state
	 * 
	 */
	public final static int OPERATING = JSL.getNextEnumConstant();
	
	/** Indicates the failed state
	 * 
	 */
	public final static int FAILED = JSL.getNextEnumConstant();
	
	//  ===========================================
	//      ATTRIBUTES
	//  ===========================================

	/** A reference to the HMSystem that contains this sensor.
	 *  A sensor can be in only 1 HMSystem
	 */
	protected HMSystem myHMSystem;
	
	/** A false alarm model can be supplied to produce false alarms by the sensor
	 */
	protected FalseAlarmModelAbstract myFalseAlarmModel;
	
	/** A detection model can be supplied to determine whether the sensor
	 *  actually detects a failure by the component
	 */
	protected DetectionModelIfc myDetectionModel;
	
	/** Represents the last known true state of the component
	 */
	protected int myTrueComponentState;
	
	/** Represents the state of the component that the sensor reports
	 */
	protected int myComponentState;
	
	/** Represents the previous state of the component that the sensor reports
	 */
	protected int myPrevComponentState;
	
	/** The component to which the sensor is attached
	 */
	protected Component myComponent;
	
	/** Creates a HMSensor with the given model element as a parent
	 * 
	 * @param parent
	 */
	public HMSensor(ModelElement parent) {
		this(parent, null, null, null, null);
	}

	/** Creates a HMSensor with the given model element as a parent
	 *  and a given name with no detection model or false alarm model
	 *  
	 * @param parent
	 * @param name
	 *
	 */
	public HMSensor(ModelElement parent, String name) {
		this(parent, null, null, null, name);
	}

	/** Creates a HMSensor with the given model element as a parent
	 *  and a default name, false alarm model, and detection model
	 * 
	 * @param parent
	 * @param component
	 * @param detectionModel
	 */
	public HMSensor(ModelElement parent, Component component) {
		this(parent, component, null, null, null);
	}
	
	/** Creates a HMSensor with the given model element as a parent
	 *  and a default name, false alarm model, and detection model
	 * 
	 * @param parent
	 * @param component
	 * @param detectionModel
	 */
	public HMSensor(ModelElement parent, Component component, DetectionModelIfc detectionModel) {
		this(parent, component, detectionModel, null, null);
	}

	/** Creates a HMSensor with the given model element as a parent
	 *  and a default name, false alarm model, and detection model
	 * 
	 * @param parent
	 * @param component
	 * @param falseAlarmModel
	 */
	public HMSensor(ModelElement parent, Component component, FalseAlarmModelAbstract falseAlarmModel) {
		this(parent, component, null, falseAlarmModel, null);
	}
	
	/** Creates a HMSensor with the given model element as a parent
	 *  and a default name, false alarm model, and detection model
	 * 
	 * @param parent
	 * @param component
	 * @param detectionModel
	 * @param falseAlarmModel
	 */
	public HMSensor(ModelElement parent, Component component, DetectionModelIfc detectionModel, FalseAlarmModelAbstract falseAlarmModel) {
		this(parent, component, detectionModel, falseAlarmModel, null);
	}
	
	/** Creates a HMSensor with the given model element as a parent
	 *  and a given name, false alarm model, and detection model
	 * 
	 * @param parent
	 * @param component
	 * @param detectionModel
	 * @param falseAlarmModel
	 * @param name
	 */
	public HMSensor(ModelElement parent, Component component, DetectionModelIfc detectionModel, FalseAlarmModelAbstract falseAlarmModel, String name) {
		super(parent, name);
		if (component != null)
			setComponent(component);
		setFalseAlarmModel(falseAlarmModel);
		setDetectionModel(detectionModel);
	}
	
	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================

	/** Gets the failure DetectionModel. May be null
	 * 
	 * @return Returns the DetectionModel.
	 */
	public final DetectionModelIfc getDetectionModel() {
		return myDetectionModel;
	}

	/** A detection model is invoked when the component that the sensor is observing
	 *  indicates a failure.  The detection model must determine whether or not the
	 *  sensor actually detects the failure or reports that the component is working
	 *  when in fact it is not working (i.e. a false positive).  If no detection model
	 *  is supplied, there is perfect detection (unless the user provides for something by 
	 *  sub-classing and overriding the failure() method
	 *  
	 * @param detectionModel The detectionModel to set.
	 */
	public final void setDetectionModel(DetectionModelIfc detectionModel) {
		myDetectionModel = detectionModel;
	}

	/** Gets the FalseAlarmModel.  May be null
	 * 
	 * @return Returns the FalseAlarmModel.
	 */
	public final FalseAlarmModelAbstract getFalseAlarmModel() {
		return myFalseAlarmModel;
	}

	/** The FalseAlarmModel represents a process for the sensor that
	 *  can generate false alarms (i.e. indicates that the component that the
	 *  sensor is observing is failed when in fact it is working. If no false alarm
	 *  model is supplied, there are no false alarms unless the user provides for something
	 *  by sub-classing
	 *  
	 * @param falseAlarmModel The FalseAlarmModel to set.
	 */
	public final void setFalseAlarmModel(FalseAlarmModelAbstract falseAlarmModel) {
		this.myFalseAlarmModel = falseAlarmModel;
	}

	/** Returns the component that the sensor is monitoring
	 * 
	 * @return
	 */
	public final Component getComponent(){
		return(myComponent);
	}
	
	/** Gets the last known state that the sensor reported for the component
	 * 
	 * @return Returns the componentState.
	 */
	public final int getComponentState() {
		return myComponentState;
	}
	
	/** Checks whether or not the sensor last reported that the component was failed
	 * 
	 * @return
	 */
	public final boolean isComponentFailed(){
		return(myComponentState == FAILED);
	}

	/** Checks whether or not the sensor last reported that the component was operating
	 * 
	 * @return
	 */
	public final boolean isComponentOperating(){
		return(myComponentState == OPERATING);
	}

	/** Checks whether or not the sensor last reported that the component was in repair
	 * 
	 * @return
	 */
	public final boolean isComponentInRepair(){
		return(myComponentState == IN_REPAIR);
	}

	/** Checks whether or not the sensor last reported that the component was unavailable
	 * 
	 * @return
	 */
	public final boolean isComponentUnavailable(){
		return(myComponentState == UNAVAILABLE);
	}
	
	/** Gets the previous reported state of the component
	 * 
	 * @return Returns the previous componentState.
	 */
	public final int getPreviousComponentState() {
		return myPrevComponentState;
	}
	
	/** Sets the sensor's component to the supplied component and attaches
	 *  the sensor as a state change listener on the component. If the sensor
	 *  is already attached to a different component it is removed from that
	 *  component and attached to the supplied component.  A sensor can be 
	 *  attached to one and only one component at a time.
	 * 
	 * @param c
	 */
	public final void setComponent(Component c){
		if (c == null)
			throw new IllegalArgumentException("The component was null.");
		
		if (myComponent != null){
			myComponent.removeStateChangeListener(this);	
		} 
		myComponent = c;
		myComponent.attachStateChangeListener(this);
	}
	
	public String toString(){
		StringBuffer sb = new StringBuffer();
		sb.append("Sensor: " + getName());
		sb.append("\n");
		if (myComponent != null){
			sb.append("Component: " + myComponent.getName());
			sb.append("\n");		
			if(myComponent.getAssembly() != null){
				sb.append("Component Assembly: " + myComponent.getAssembly().getName());
				sb.append("\n");			
			}
		}
		sb.append("True component state: " + stateToString(getTrueComponentState()));
		sb.append("\n");
		sb.append("Current Reported component state: " + stateToString(getComponentState()));
		sb.append("\n");
		sb.append("Previous Reported component state: " + stateToString(getPreviousComponentState()));
		sb.append("\n");
		
		return(sb.toString());
	}
	
	/** Converts the integer representation of the state of the component to a string
	 * 
	 * @param state
	 * @return
	 */
	public String stateToString(int state){
		String s = null;
		if (state == AVAILABLE){
			s = "AVAILABLE";
		} else if (state == IN_REPAIR){
			s = "IN_REPAIR";
		} else if (state == UNAVAILABLE){
			s = "UNAVAILABLE";
		} else if (state == OPERATING){
			s = "OPERATING";
		} else if (state == FAILED){
			s = "FAILED";
		}
		return(s);
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	/** Gets the last known true state of the observed component
	 * @return Returns the trueComponentState.
	 */
	protected final int getTrueComponentState() {
		return myTrueComponentState;
	}

	/** Sets the true state of the component
	 * @param state The trueComponentState to set.
	 */
	protected final void setTrueComponentState(int state) {
		myTrueComponentState = state;
	}

	/** Sets the state of the component that the sensor reports
	 * @param state The state to set.
	 */
	protected final void setComponentState(int state) {
		myPrevComponentState = myComponentState;
		myComponentState = state;
		if (myHMSystem != null)
			myHMSystem.sensorUpdate_(this);
		notifyObservers(myComponentState);
	}

	/** This method should be called by the false alarm model
	 *  when the false alarm occurs to notify the sensor that
	 *  a false alarm has occurred
	 * 
	 */
	protected final void falseAlarmModelAction(){
		System.out.println("---------------");
		System.out.println("False Alarm Model: sensor detected failure.");
		System.out.println("---------------");
		setComponentState(HMSensor.FAILED);
	}
	
	/**
	 * @return Returns the HMSystem.
	 */
	protected final HMSystem getHMSystem() {
		return myHMSystem;
	}

	/**
	 * @param hms The HMSystem to set.
	 */
	protected final void setHMSystem(HMSystem hms) {
		myHMSystem = hms;
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.component.ComponentStateChangeListenerIfc#stateChange(jsl.modeling.elements.component.Component)
	 */
	public void stateChange(Component c){
		
		StateAccessorIfc pState = c.getPreviousComponentState();
		StateAccessorIfc state = c.getComponentState();
		System.out.println("t = " + getTime() + " > Transitioning from state " + pState + " to state " + state);
		
		super.stateChange(c);
		
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.component.ComponentStateChangeListener#componentAvailable(jsl.modeling.elements.component.Component)
	 */
	protected void componentAvailable(Component c){

		myTrueComponentState = AVAILABLE;
		setComponentState(myTrueComponentState);

		if (myFalseAlarmModel != null)
			myFalseAlarmModel.componentAvailable(this, c);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.component.ComponentStateChangeListener#componentStartedOperating(jsl.modeling.elements.component.Component)
	 */
	protected void componentStartedOperating(Component c) {
		
		myTrueComponentState = OPERATING;
		setComponentState(myTrueComponentState);

		if (myFalseAlarmModel != null)
			myFalseAlarmModel.componentStartedOperating(this, c);
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.component.ComponentStateChangeListener#componentUnavailable(jsl.modeling.elements.component.Component)
	 */
	protected void componentUnavailable(Component c) {
		
		myTrueComponentState = UNAVAILABLE;
		setComponentState(myTrueComponentState);

		if (myFalseAlarmModel != null)
			myFalseAlarmModel.componentUnavailable(this, c);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.component.ComponentStateChangeListener#componentFailed(jsl.modeling.elements.component.Component)
	 */
	protected void componentFailed(Component c) {
		
		myTrueComponentState = FAILED;

		if (myFalseAlarmModel != null)
			myFalseAlarmModel.componentFailed(this, c);

		if (myDetectionModel != null)
			setComponentState(myDetectionModel.componentFailure(this, c));
		else
			setComponentState(myTrueComponentState);

	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.component.ComponentStateChangeListener#componentStartedRepair(jsl.modeling.elements.component.Component)
	 */
	protected void componentStartedRepair(Component c) {		
		
		myTrueComponentState = IN_REPAIR;
		setComponentState(myTrueComponentState);
		
		if (myFalseAlarmModel != null)
			myFalseAlarmModel.componentStartedRepair(this, c);
	}
	
	//  ===========================================
	//      PRIVATE METHODS
	//  ===========================================


}
