/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.adc.hmsystem;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;


import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.component.Component;

public class HMSystem extends SchedulingElement {

	private List<HMSensor> mySensors;
	
	private Set<HMSensor> mySensorsIndicatingFailure;
	
	protected List<SensorUpdateModelIfc> mySensorUpdateModels;
	
	/** Creates an HMSystem w/o an update model
	 * 
	 * @param parent
	 */
	public HMSystem(ModelElement parent) {
		this(parent, null);
	}
	
	/** Creates an HMSystem with the supplied update model
	 * 
	 * @param parent
	 * @param name
	 */
	public HMSystem(ModelElement parent, String name) {
		super(parent, name);
		mySensors = new LinkedList<HMSensor>();
		mySensorsIndicatingFailure = new HashSet<HMSensor>();
		mySensorUpdateModels = new LinkedList<SensorUpdateModelIfc>();
	}

	/** Creates a new sensor and adds it to the HMSystem
	 *  The user is responsible for setting the sensor's component or other parameters
	 * @return
	 */
	public final HMSensor addSensor(){
		HMSensor s =new HMSensor(this);
		addSensor(s);
		return(s);
	}

	/** Creates a new sensor with the given parameters, attaches it to the component, 
	 *  and adds it to the HMSystem
	 * 
	 * @param component
	 * @param detectionModel
	 * @return
	 */
	public final HMSensor addSensor(Component component){
		HMSensor s =new HMSensor(this, component);
		addSensor(s);
		return(s);
	}
	
	/** Creates a new sensor with the given parameters, attaches it to the component, 
	 *  and adds it to the HMSystem
	 * 
	 * @param component
	 * @param detectionModel
	 * @return
	 */
	public final HMSensor addSensor(Component component, DetectionModelIfc detectionModel){
		HMSensor s =new HMSensor(this, component, detectionModel);
		addSensor(s);
		return(s);
	}
	
	/** Creates a new sensor with the given parameters, attaches it to the component, 
	 *  and adds it to the HMSystem
	 * 
	 * @param component
	 * @param detectionModel
	 * @param falseAlarmModel
	 * @return
	 */
	public final HMSensor addSensor(Component component, DetectionModelIfc detectionModel, FalseAlarmModelAbstract falseAlarmModel){
		HMSensor s =new HMSensor(this, component, detectionModel, falseAlarmModel);
		addSensor(s);
		return(s);
	}
	
	/** Creates a new sensor with the given parameters, attaches it to the component, 
	 *  and adds it to the HMSystem
	 * 
	 * @param component
	 * @param detectionModel
	 * @param falseAlarmModel
	 * @param name
	 * @return
	 */
	public final HMSensor addSensor(Component component, DetectionModelIfc detectionModel, FalseAlarmModelAbstract falseAlarmModel, String name){
		HMSensor s =new HMSensor(this, component, detectionModel, falseAlarmModel, name);
		addSensor(s);
		return(s);
	}
	
	/** Adds the supplied sensor to the HMSystem
	 * 
	 * @param sensor
	 */
	public final void addSensor(HMSensor sensor) {
		
		if (sensor == null)
			throw new IllegalArgumentException("The supplied HMSensor was null!");
		
		if (sensor.getHMSystem() != null)
			throw new IllegalArgumentException("The supplied HMSensor has already been added to an HMSystem!");		
		
		sensor.setHMSystem(this);
		mySensors.add(sensor);
	}

	/** Checks if the supplied sensor is already part of the HMSystem
	 * 
	 * @param sensor
	 * @return
	 */
	public final boolean contains(HMSensor sensor) {
		return mySensors.contains(sensor);
	}

	/** Removes the sensor from the HMSystem.  Does not remove the sensor
	 *  from the component (if any) that it is observing
	 * 
	 * @param sensor
	 */
	public final void removeSensor(HMSensor sensor) {
		if (sensor == null)
			throw new IllegalArgumentException("The supplied HMSensor was null!");
		
		if (!contains(sensor))
			throw new IllegalArgumentException("The supplied HMSensor is not in the system!");
		
		sensor.setHMSystem(null);
		mySensors.remove(sensor);
	}

	/** The sensor update model can be used to plug in behavior for how the
	 *  HMSystem will work when a sensor update occurs.  If a sensor update
	 *  model is supplied it will be called automatically whenever a sensor
	 *  notifies the system of an update.  This occurs in the 
	 *  sensorUpdate() method.  Alternatively the user can subclass, HMSystem
	 *  and override the sensorUpdate() method to provide update behavior
	 *  
	 * @param sensorUpdateModel The sensorUpdateModel to attach.
	 */
	public final void attachSensorUpdateModel(SensorUpdateModelIfc sensorUpdateModel) {
		if (sensorUpdateModel == null)
			throw new IllegalArgumentException("The SensorUpdateModel was null!");
		
		if (mySensorUpdateModels.contains(sensorUpdateModel))
			throw new IllegalArgumentException("The SensorUpdateModel is already attached to the HMSystem!");
		
		mySensorUpdateModels.add(sensorUpdateModel);
		
	}

	/** Returns whether or not the update model has been attached to this system
	 *  
	 * @param sensorUpdateModel The sensorUpdateModel to check
	 * @return
	 */
	public final boolean containsSensorUpdateModel(SensorUpdateModelIfc sensorUpdateModel) {
		if (sensorUpdateModel == null)
			return(false);
		
		return(mySensorUpdateModels.contains(sensorUpdateModel));
	}
	
	/** The sensor update model can be used to plug in behavior for how the
	 *  HMSystem will work when a sensor update occurs.  If a sensor update
	 *  model is supplied it will be called automatically whenever a sensor
	 *  notifies the system of an update.  This occurs in the 
	 *  sensorUpdate() method.  Alternatively the user can subclass, HMSystem
	 *  and override the sensorUpdate() method to provide update behavior
	 *  
	 * @param sensorUpdateModel The sensorUpdateModel to remove.
	 */
	public final void removeSensorUpdateModel(SensorUpdateModelIfc sensorUpdateModel) {
		if (sensorUpdateModel == null)
			throw new IllegalArgumentException("The SensorUpdateModel was null!");
		
		if (!mySensorUpdateModels.contains(sensorUpdateModel))
			throw new IllegalArgumentException("The SensorUpdateModel is notattached to the HMSystem!");
		
		mySensorUpdateModels.remove(sensorUpdateModel);
		
	}
	
	/** Returns an unmodifiable list of the sensors in the HMSystem
	 * 
	 * @return
	 */
	public final List<HMSensor> getSensors(){
		return(Collections.unmodifiableList(mySensors));
	}

	/** Returns an unmodifiable list of the sensors in the HMSystem
	 * 
	 * @return
	 */
	public final Set<HMSensor> getSensorsIndicatingFailure(){
		return(Collections.unmodifiableSet(mySensorsIndicatingFailure));
	}
	
	/** The user can subclass, HMSystem
	 *  and override the sensorUpdate() method to provide update behavior
	 * 
	 * @param sensor
	 */
	protected void sensorUpdate(HMSensor sensor){
		System.out.println();
		System.out.println("HMSystem Update");
		System.out.println("Sensor " + sensor);
		System.out.println();
	}
	
	/** This method is called when a sensor changes state.  It keeps
	 *  track of the sensors that are reporting failures, calls 
	 *  sensorUpdate() which can be overridden to provide update behavior
	 *  This method also calls any attached SensorUpdateModelIfc object
	 *  of the update.
	 * 
	 * @param sensor
	 */
	protected void sensorUpdate_(HMSensor sensor){//TODO could be final, could be package scope
		if (sensor.isComponentFailed())
			mySensorsIndicatingFailure.add(sensor);
		else
			mySensorsIndicatingFailure.remove(sensor);
		
		sensorUpdate(sensor);
		
		for(SensorUpdateModelIfc m:mySensorUpdateModels){
			m.sensorUpdate(this, sensor);
		}		
	}

}
