/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.adc.hmsystem;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.component.Component;
import jsl.utilities.random.distributions.DistributionIfc;

public class RPFalseAlarmModel extends FalseAlarmModelAbstract {

	public RPFalseAlarmModel(ModelElement parent, DistributionIfc timeToFalseAlarmCDF) {
		this(parent, timeToFalseAlarmCDF, null);
	}
	
	public RPFalseAlarmModel(ModelElement parent, DistributionIfc timeToFalseAlarmCDF, String name) {
		super(parent, timeToFalseAlarmCDF, name);
	}

	/* (non-Javadoc)
	 * @see afrltime4v2.hmsystem.FalseAlarmModelAbstract#componentAvailable(afrltime4v2.hmsystem.HMSensor, jsl.modeling.elements.component.Component)
	 */
	protected void componentAvailable(HMSensor sensor, Component c) {
		// the sensor has indicated that the component has transitioned
		// if this occurs before a previously scheduled false alarm event
		// then the false alarm should be cancelled
		if (myFalseAlarmEvent != null){
			// a false alarm event was previously scheduled so cancel it
			cancelEvent(myFalseAlarmEvent);
		}

	}

	/* (non-Javadoc)
	 * @see afrltime4v2.hmsystem.FalseAlarmModelAbstract#componentUnavailable(afrltime4v2.hmsystem.HMSensor, jsl.modeling.elements.component.Component)
	 */
	protected void componentUnavailable(HMSensor sensor, Component c) {
		// the sensor has indicated that the component has transitioned
		// if this occurs before a previously scheduled false alarm event
		// then the false alarm should be cancelled
		if (myFalseAlarmEvent != null){
			// a false alarm event was previously scheduled so cancel it
			cancelEvent(myFalseAlarmEvent);
		}
	}

	/* (non-Javadoc)
	 * @see afrltime4v2.hmsystem.FalseAlarmModelAbstract#componentStartedRepair(afrltime4v2.hmsystem.HMSensor, jsl.modeling.elements.component.Component)
	 */
	protected void componentStartedRepair(HMSensor sensor, Component c) {
		// the sensor has indicated that the component has transitioned
		// if this occurs before a previously scheduled false alarm event
		// then the false alarm should be cancelled
		if (myFalseAlarmEvent != null){
			// a false alarm event was previously scheduled so cancel it
			cancelEvent(myFalseAlarmEvent);
		}
	}

	/* (non-Javadoc)
	 * @see afrltime4v2.hmsystem.FalseAlarmModelAbstract#componentFailed(afrltime4v2.hmsystem.HMSensor, jsl.modeling.elements.component.Component)
	 */
	protected void componentFailed(HMSensor sensor, Component c) {
		// the sensor has indicated that the component has transitioned
		// if this occurs before a previously scheduled false alarm event
		// then the false alarm should be cancelled
		if (myFalseAlarmEvent != null){
			// a false alarm event was previously scheduled so cancel it
			cancelEvent(myFalseAlarmEvent);
		}
	}

	/* (non-Javadoc)
	 * @see afrltime4v2.hmsystem.FalseAlarmModelAbstract#componentStartedOperating(afrltime4v2.hmsystem.HMSensor, jsl.modeling.elements.component.Component)
	 */
	protected void componentStartedOperating(HMSensor sensor, Component c) {
		// the sensor has indicated that the component has transitioned
		// if this occurs before a previously scheduled false alarm event
		// then the false alarm should be cancelled
		if (myFalseAlarmEvent != null){
			// a false alarm event was previously scheduled but the component is starting an operation
			cancelEvent(myFalseAlarmEvent);
		}
		
		// the component has started operating, determine if a false alarm will occur
		// during the period of operation
		double oTime = c.getOperationTime();
		double ttfa = myTimeToFalseAlarmRV.getValue();
		if (Double.isNaN(oTime)){
			// the component has been placed in the operating state for an undetermined length of time
			// go ahead and schedule the false alarm
			myFalseAlarmEvent = scheduleEvent(myFalseAlarmListener, ttfa);
			myFalseAlarmEvent.setMessage(sensor);
			// if the component transitions before the false alarm, the false alarm event will be cancelled
		} else {
			if (ttfa < oTime) {// only need to schedule the false alarm if it will occur before the end of the operating time
				myFalseAlarmEvent = scheduleEvent(myFalseAlarmListener, ttfa);
				myFalseAlarmEvent.setMessage(sensor);
			}
		}

	}
		
}
