package jslx.forecasting;


/**
 * Implements the intermittent forecaster named average demand technique
 * @author vvarghe
 *
 */
public class AverageDemand extends IntermittentForecastTechnique {
	
	private SimpleExponentialSmoothing myAtSES;
	
	public AverageDemand() {
		this(0.1, 3, null);
	}

	public AverageDemand(String name) {
		this(0.1, 3, name);
	}

	public AverageDemand(double alpha) {
		this(alpha, 3, null);
	}	
	
	public AverageDemand(double alpha, String name) {
		this(alpha, 3, name);
	}

	public AverageDemand(double alpha, int reqdnzdem, String name) {
		super(alpha, reqdnzdem, name);
		myAtSES = new SimpleExponentialSmoothing(alpha, name+"At_SES");
		double [] a = {alpha};
		setSmoothingParameter(a);
	}

	public AverageDemand(double [] alpha, String name) {
		this(alpha, 3, name);	
	}
	
	public AverageDemand(double[] alpha, int reqdnzdem, String name) {
		super(alpha, reqdnzdem, name);
		myAtSES = new SimpleExponentialSmoothing(alpha[0], name+"At_SES");
		setSmoothingParameter(alpha);
	}

	public final void setSmoothingParameter(double [] alpha){
		if ((alpha[0] < 0) || (alpha[0] > 1))
			throw new IllegalArgumentException("The smoothing parameter must be in [0,1]");

		myMapParameterLabelToValue.put("Smoothing Constant Alpha", alpha[0]);
		myAlpha = alpha;
        if(myAtSES==null)
            myAtSES = new SimpleExponentialSmoothing(alpha[0], getName()+"At_SES");
        
        myAtSES.setSmoothingParameter(alpha[0]);
	}	

	protected void estimateAverageDemand(){
		myAtSES.collect(myZt/myPt);
		myNextPeriodForecast = myAtSES.getNextPeriodForecast();
	}
	
	//overrides the method in the super class
	public void initializeHistory(double[] data, double[] forecasts,
			double[] parameters) {
		
		//TODO How to handle the cases where the data is not sufficient to get estimates	
		double [] para =  estimateInitialForecastParameter(data);
		setInitialForecastParameter(para);
		myQ = para[0];
		myPt = para[1];
		myZt = para[2];
        double [] at = {myZt/myPt};
		myAtSES.initializeHistory(at, null, null);
        myNextPeriodForecast = myAtSES.getNextPeriodForecast();
        setInitFlag(true);
	}

	protected void updateForecast(int i, double x) {
		if(myXt==0){
        	myQ++;
        }else{                
        	myPt=myQ;
            myZt=myXt;
            myQ=1;
        }
        //System.out.println("Zt = "+myZt+", myPt = "+myPt);
		estimateAverageDemand(); 
	}

	public final double[] getParameters() {
		double[] d = new double[1];
		d[0] = myAlpha[0];
		return d;
	}

	public final void setParameters(double[] parameters) {
		if (parameters == null)
			throw new IllegalArgumentException("The parameters array was null");
		
		if (parameters.length == 0)
			throw new IllegalArgumentException("parameters length was 0");

		setSmoothingParameter(parameters);
	}
	
	/**
	 * Creates a shallow copy of an AverageDemand forecast technique object, retaining only the forecast parameters.
	 */
	public ForecastTechnique copyInstance(){		
        return new AverageDemand(this.myAlpha,this.myReqdNumNZDemandsBeforeInitialization,this.getName());
    }
	
	@Override
	protected void clearHistory() {
		myNumNZDemandsBeforeInitialization = 0;
		myPt = Double.NaN;		
		myZt = Double.NaN;		
		myQ = Double.NaN;		
		myData = new double[1];
		myAtSES.reset();
	}
	
}
