/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.forecasting;

import jsl.utilities.statistic.Statistic;

/**
 * @author rossetti
 *
 */
public class CumulativeAverage extends ForecastTechnique {

	/**
	 * 
	 */
	public CumulativeAverage() {
		this(null);
	}

	/**
	 * @param name
	 */
	public CumulativeAverage(String name) {
		super(name);

	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#checkForSufficientHistory()
	 */
	protected void checkForSufficientHistory() {
		if (!isInitialized()){
			// not already initialized
			// need to check if enough data is available for initialization
			if (myTimeIndex == 1){
				// not initialized, and this is the first data point
				// make forecast for next period the same as first data point
				myNextPeriodForecast = myXt;
				// initialize it, so when 2nd point is collected
				// it is ready
				
				//TODO vijith added
				double [] hist = {myXt}; 
				double [] para = this.estimateInitialForecastParameter(hist);
				setInitialForecastParameter(para);
				
				setInitFlag(true);
			}
		}
	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#clearHistory()
	 */
	protected void clearHistory() {
		// Nothing to clear

	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#getForecast(int)
	 */
	public double getForecast(int n) {
		return myNextPeriodForecast;
	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#getForecasts(int)
	 */
	public double[] getForecasts(int n) {
		if (n < 1)	
			throw new IllegalArgumentException("The supplied n was < 1");
		
		double[] f = new double[n+1];
		f[0] = Double.NaN;
		for (int i=1;i<=n;i++)
			f[i] = myNextPeriodForecast;
		return f;
	}

	/** No history is saved for this technique.  Thus,
	 *  null is returned by this method
	 *  
	 */
	public double[] getHistory() {
		// Nothing to return
		return null;
	}

    /** The data is used to initialize the average
     * 
     */
	public void initializeHistory(double[] data, double[] forecasts,
			double[] parameters) {
		if (data == null)
			throw new IllegalArgumentException("The supplied data array was null");

		myStat.reset();
		myStat.collect(data);
		myNextPeriodForecast = myStat.getAverage();
		// initialize it
		
		//TODO Vijith added this
		double [] para = {myNextPeriodForecast};
		setInitialForecastParameter(para);				
		
		setInitFlag(true);
	}
    
    public double [] estimateInitialForecastParameter(double [] history){
    	Statistic stat = new Statistic();
	    stat.collect(history);
	    
        double mean = stat.getAverage();

        double [] init_stat_para = {mean};
        
		return init_stat_para;
		
	} 

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#updateForecast(int, double)
	 */
	protected void updateForecast(int i, double x) {
		myNextPeriodForecast = myStat.getAverage();
	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#updateHistory(int, double)
	 */
	protected void updateHistory(int i, double x) {
		// Nothing to update

	}

    /** There are no parameters. This method returns null
     * 
     */
	public double[] getParameters() {
		// Nothing to return
		return null;
	}

	/** There are no parameters to set so this
	 *  method does nothing
	 * 
	 */
	public void setParameters(double[] parameters) {
		// Nothing to set
	}
	
	/**
	 * Creates a shallow copy of a CumulativeAverage forecast technique object.
	 */
	public ForecastTechnique copyInstance(){
        return new CumulativeAverage(this.getName());
    }

}
