package jslx.forecasting;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * Implements a metaforecaster based on multinomial logistic regression
 * @author vvarghe
 *
 */
public class DefaultMetaForecaster extends AbstractMetaForecaster {

	public double myCoefficients [][];
    public int numForecastTechniques;
    public int numNonBaseForecastTechniques;
    public int numDemandAtrributeTypes;
    public int numCategories;
    public int numPredictors;
    public int CoeffDim1;
    public int CoeffDim2;
    public int numAdded;   
    
    public SortedSet<Pair> mySortedSet;       
    public Map <ForecastTechnique, Double> myRankedTable ;
	
    //TODO consider a more generic constructor where we can specify numforecasttechniques and numdemandattributetypes
	public DefaultMetaForecaster() {
		super();
		int numforecasttechniques = 7;
		int numdemandattributetypes =6; 
		mySortedSet = new TreeSet<Pair>();
		myRankedTable = new LinkedHashMap<ForecastTechnique, Double>();
		numForecastTechniques = numforecasttechniques;
		numNonBaseForecastTechniques = numForecastTechniques -1;
		numDemandAtrributeTypes = numdemandattributetypes;
		numCategories = numForecastTechniques;
		numPredictors = numDemandAtrributeTypes;
		CoeffDim1 = numPredictors + 1;
		CoeffDim2 = numCategories - 1;
		myCoefficients = new double[CoeffDim1][CoeffDim2];	
                       
	}
	
	public void addForecastTechnique(ForecastTechnique ft, double [] predictorcoefficients){
		// this will make sure that not more than the specified number of non base forecast techniques
		// will be added by the user
		if(myForecastTechniques.size() == numNonBaseForecastTechniques)
			throw new IllegalArgumentException("The number of forecast techniques " +
					"(that are not base technique) that can be added has reached its limit");

		// this will make sure that the a valid array will be provided by the user to set the 
		// coefficients of for each of the non base forecast technique 
		if(predictorcoefficients.length != CoeffDim1)
			throw new IllegalArgumentException("Invalid size of array of predictor-coefficients");
		
		for(int i = 0; i <predictorcoefficients.length; i++){
			myCoefficients[i][myForecastTechniques.size()] = predictorcoefficients[i];
			//System.out.println("myCoefficients["+i+"]["+myForecastTechniques.size()+"] = "+predictorcoefficients[i]);
		}
		
		addForecastTechnique(ft);
		
		numAdded++;
				
	}
	
	public void addBaseForecastTechnique(ForecastTechnique ft){
		addForecastTechnique(ft);
		setBaseForecastTechnique(ft);		
	}

	public ForecastTechnique getReccommendedForecastTechnique(double[] demandattributevector) {
		if(demandattributevector.length != numDemandAtrributeTypes)
			throw new IllegalArgumentException("Invalid size of the input vector");
		
		if(this.myBaseForecastTechnique == null)
			throw new IllegalArgumentException("Base forecast technique not set");
		
		if(this.numCategories - 1 != numAdded)
			throw new IllegalArgumentException("Not all non base forecast technique are set");
		
        this.setProbability(demandattributevector);  
        
        for(Pair p: mySortedSet){
			ForecastTechnique ft = p.myForecastTechnique;
			myRankedForecastTechniques.add(ft);
			myRankedTable.put(p.myForecastTechnique, (Double)p.myProbability);
		}
        return mySortedSet.last().myForecastTechnique;
	}
	
	public ForecastTechnique getReccommendedForecastTechnique(IterativeDataSource ds){
		double [] demandattributevector = new double[6];
		IntermittentDemandStatistic i = new IntermittentDemandStatistic();
		while(ds.hasNext()){
			double d = ds.next();
			i.collect(d);
		}
		
		demandattributevector [0] = i.getProbSuccessAfterFailure();
		demandattributevector [1] = i.getProbSuccessAfterSuccess();
		demandattributevector [2] = i.getDemandStat().getAverage();
		demandattributevector [3] = i.getDemandStat().getVariance();
		demandattributevector [4] = i.getDemandStat().getLag1Correlation();
		demandattributevector [5] = i.getIntervalBetweenNonZeroDemandStat().getLag1Correlation();	
		
		return getReccommendedForecastTechnique(demandattributevector);
	}
        
    public void setProbability(double[] demandattributevector){		
    	mySortedSet = new TreeSet<Pair>();
		myRankedTable = new LinkedHashMap<ForecastTechnique, Double>();
		myRankedForecastTechniques = new LinkedHashSet<ForecastTechnique>();
        
        double [] xb = new double[numNonBaseForecastTechniques];
        double sum = 0;
        for(int i = 0; i < numNonBaseForecastTechniques; i++){
        	double x = myCoefficients[0][i];
        	for(int j = 0; j < numPredictors; j++){
        		x = x + demandattributevector[j]*myCoefficients[j+1][i];
        	}
    		xb[i] = x;
        	sum = sum + Math.exp(x);  	        
        }      	
        
        int k = 0;
        for(ForecastTechnique ft: myForecastTechniques){
        	if(k <  numNonBaseForecastTechniques){
            	double prob = Math.exp(xb[k])/(1.0 + sum);
            	Pair p = new Pair(ft,prob);
            	mySortedSet.add(p);
            	k++;        		
        	}
        }
        double prob = 1.0 / (1.0 + sum);
    	Pair p = new Pair(myBaseForecastTechnique,prob);
    	mySortedSet.add(p);
    }

	public Set<ForecastTechnique> getRankedForecastTechniques() {
		if(mySortedSet.size() != numForecastTechniques)
			throw new NullPointerException("Sorted set is incomplete");
		
		return myRankedForecastTechniques;
	}
    
    public Map<ForecastTechnique, Double> getRankedTable(){
		if(mySortedSet.size() != numForecastTechniques)
			throw new NullPointerException("Sorted set is empty");
		
    	return myRankedTable;
    	
    }
    
    public Object[][] getMetaForecastResults(double[] demandattributevector){
        if(demandattributevector.length != numDemandAtrributeTypes)
            throw new IllegalArgumentException("Invalid size of the input vector");
        if(this.myBaseForecastTechnique == null)
            throw new IllegalArgumentException("Base forecast technique not set");
      
        if(this.numCategories - 1 != numAdded)
            throw new IllegalArgumentException("Not all non base forecast technique are set");
  
        this.setProbability(demandattributevector);
        
        Object[][] pairs=new Object[mySortedSet.size()][];
        int i=0;
        Pair[] prs=mySortedSet.toArray(new Pair[mySortedSet.size()]);
        
        Arrays.sort(prs,Collections.reverseOrder());
        for(Pair p:prs){
	        ForecastTechnique ft=p.myForecastTechnique;
	        double prob=p.myProbability;
	        pairs[i++]=new Object[]{ft,prob};
        }
        return pairs;
    }
        
    public class Pair implements Comparable<Pair>{            
        ForecastTechnique myForecastTechnique;
        double myProbability;            
        
        public Pair(ForecastTechnique ft, double prob){
        	System.out.println(ft.getName());
        	System.out.println(ft.getMapParameterLabelToValue());
        	System.out.println("Probability: "+prob);
        	System.out.println("********************************************");
        	
            myForecastTechnique = ft;
            myProbability = prob;            
        }

        public int compareTo(DefaultMetaForecaster.Pair o) {                
            if(myProbability > o.myProbability){
                return 1;
            
            }else if(myProbability < o.myProbability){
                return -1;                
            }else{
                return 0;
            }
        }
        
    }
	
	public static void main(String [] s){
		DefaultMetaForecaster mf = new DefaultMetaForecaster();
		
		ForecastTechnique ftech1 = new SimpleExponentialSmoothing(0.1);
	    ForecastTechnique ftech2 = new MovingAverage(19);
	    ForecastTechnique ftech3 = new SyntetosBoylan(0.1);
	    ForecastTechnique ftech4 = new MovingAverage(9);
	    ForecastTechnique ftech5 = new SyntetosBoylan(0.2);
	    ForecastTechnique ftech6 = new CumulativeAverage();
	    ForecastTechnique ftech7 = new SimpleExponentialSmoothing(0.2);
	    
	    //coeff = {intercept, p11, p01, phi, mean, variance, phi_mibt}
	    double [] ft1 = {-4.81	,	-12.34	,	34.09	,	-60.27	,	9.49	,	-0.99	,	-45.97};
        double [] ft2 = {-3.41	,	0.00	,	65.60	,	-56.25	,	-15.32	,	1.27	,	-29.05};
        double [] ft3 = {-13.31	,	-14.09	,	77.86	,	-64.76	,	0.00	,	0.00	,	-33.7102};
        double [] ft4 = {0.00	,	21.97	,	102.90	,	-23.31	,	-85.42	,	6.92	,	0.00};
        double [] ft5 = {-5.51	,	-55.20	,	87.76	,	-30.33	,	9.49	,	-0.79	,	-24.82};
        double [] ft6 = {-6.53	,	-9.94	,	56.13	,	-77.14	,	5.58	,	-0.48	,	-62.39};
	    
	    mf.addForecastTechnique(ftech1, ft1);
	    mf.addForecastTechnique(ftech2, ft2);
	    mf.addForecastTechnique(ftech3, ft3);		
	    mf.addForecastTechnique(ftech4, ft4);
	    mf.addForecastTechnique(ftech5, ft5);	
	    mf.addForecastTechnique(ftech6, ft6);
	    mf.addBaseForecastTechnique(ftech7);        
		
		double [] dav1 = new double[6];
		/*dav1 [0] = 0.658153;
		dav1 [1] = 0.51132;
		dav1 [2] = 0.03466;
		dav1 [3] = 1.82248;
		dav1 [4] = 5.8798;
		dav1 [5] = 0.006258;*/
		
		//p11, p01, phi, mean, variance, phi_mibt, 
		dav1 [0] = 0.165358;
		dav1 [1] = 0.140067;
		dav1 [2] = 0.00835;
		dav1 [3] = 0.32774;
		dav1 [4] = 1.0664;
		dav1 [5] = -0.000472;
		
		//System.out.println(mf.getRankedTable());
		
		ForecastTechnique ft = mf.getReccommendedForecastTechnique(dav1);
		
		System.out.println("***********");
        
		Map <ForecastTechnique, Double> map = mf.getRankedTable();
		for(ForecastTechnique f: map.keySet()){			
			System.out.println(f.getName()+" -- "+map.get(f));			
		}
		
		System.out.println("***********");
        
		for(ForecastTechnique f: mf.getRankedForecastTechniques()){			
			System.out.println(f.getName()+" -- "+f.getMapParameterLabelToValue());			
		}
		
		System.out.println("Best Technique");
		
        System.out.println(ft.getName());
        System.out.println(ft.getMapParameterLabelToValue());
		
	}

}
