/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E.
*	Department of Industrial Engineering
*	University of Arkansas
*	4207 Bell Engineering Center
*	Fayetteville, AR 72701
*	Phone: (479) 575-6756
*	Email: rossetti@uark.edu
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution);
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA, or see www.fsf.org
*
*/
package jslx.forecasting;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;

import jsl.utilities.IdentityIfc;
import jsl.utilities.random.ParametersIfc;
import jsl.utilities.reporting.JSL;
import jsl.utilities.statistic.Statistic;
import jsl.utilities.statistic.StatisticAccessorIfc;
import jslx.forecasting.metrics.ForecastMetric;

/**
 * @author rossetti
 *
 */
public abstract class ForecastTechnique extends Observable implements IdentityIfc, ParametersIfc, Observer  {

	//  ===========================================
	//      CLASS ATTRIBUTES
	//  ===========================================

	/** A counter to count the number of created to assign "unique" ids
	 */
	protected static long myIdCounter_;

	/** A mnemonic to indicate to observers that the forecast is in the uninitialized state
	  */
//	public static final int UNINITIALIZED = JSL.getNextEnumConstant();

	/** An "enum" to indicate that the forecast is in the initialized state
	 */
//	public static final int INITIALIZED = JSL.getNextEnumConstant();

	/** An "enum" to indicate that reason for notification of observers is that the
	 *  forecast technique has observed new data
	 */
	public static final int NEW_DATA = JSL.getNextEnumConstant();

	/** An "enum" to indicate that reason for notification of observers is that the
	 *  forecast technique has been reset
	 */
	public static final int RESET = JSL.getNextEnumConstant();

    //  ===========================================
    //      OBJECT ATTRIBUTES
    //  ===========================================

	/** Keeps track of the current state for observers
	 */
	private int myObserverState = RESET;

	/** Keeps track of the previous type of state change for observers
	 */
	private int myPreviousObserverState = RESET;

	/** The id of this object
	 */
	private long myId;

   /** Holds the name of the technique for reporting purposes.
     */
    private String myName;

    /** The latest observed value
     *
     */
	protected double myXt = Double.NaN;

	/** The last period's observation
	 *
	 */
	protected double myPrevX = Double.NaN;

    /** The last generated forecast.  The forecast for next period
     *
     */
    protected double myNextPeriodForecast = Double.NaN;

    /** The forecast for the current period
     *
     */
    protected double myCurrentPeriodForecast = Double.NaN;

    /** The last computed error
     *
     */
    protected double myError = Double.NaN;

    /** Flag to indicate whether or not the forecaster
     *  has been initialized
     *
     */
    private boolean myInitFlag = false;

    /** The current time index
     *
     */
    protected int myTimeIndex = 0;

    /** Used to collect statistics on the 1-step ahead
     *  forecast errors
     *
     */
    protected Statistic myErrorStat;

    /** The statistics on the observed values
     *
     */
    protected Statistic myStat;

    protected Map<String, Double> myMapParameterLabelToValue;

    /**
     * List of attached ForecastMetrics
     */
	private List<ForecastMetric> attachedMetrics;

	protected ForecastMetric myMAD;
	protected ForecastMetric myMSE;
	protected ForecastMetric myMAPE;
	/*protected ForecastMetric myMeanUAPE;
	protected ForecastMetric myMedianUAPE;
	protected ForecastMetric myCABasedMRAE;
	protected ForecastMetric myNFBasedMRAE;
	protected ForecastMetric myCABasedMdRAE;
	protected ForecastMetric myNFBasedMdRAE;
	protected ForecastMetric myCABasedGMRAE;
	protected ForecastMetric myNFBasedGMRAE;*/
	protected Statistic myComputationTime;

	//TODO
	protected double[] myInitialParameters;

	/**
	 *
	 */
	public ForecastTechnique() {
		this(null);
	}

	/**
	 *
	 */
	public ForecastTechnique(String name) {
        setId();
        setName(name);
        myStat = new Statistic("Summary Statistics");
        myErrorStat = new Statistic("1-Step Ahead Error");
        myMapParameterLabelToValue = new HashMap <String, Double>();
        attachedMetrics=new ArrayList<ForecastMetric>();

        myMAD = ForecastMetric.createMeanAbsoluteError();
        myMSE = ForecastMetric.createMeanSquareError();
        myMAPE = ForecastMetric.createMeanAbsolutePercentageError();
        myComputationTime = new Statistic("ComputationTime");
	}

    //  ===========================================
    //        PUBLIC INTERFACE
    //  ===========================================

	/** Gets the name.
	* @return The name of object.
	*/
	public final String getName() {
		return myName;
	}

	/** Sets the name
	* @param str The name as a string.
	*/
	public final void setName(String str) {
		if (str == null){
			myName = this.getClass().getSimpleName();
		}
		else
			myName = str;
	}

	/** Returns the id for this iterative process
	 *
	 * @return
	 */
	public final long getId(){
		return(myId);
	}

	/** Returns true if the forecaster has been initialized
	 *
	 * @return
	 */
	public final boolean isInitialized(){
		return myInitFlag;
	}

	/* (non-Javadoc)
	 * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
	 */
	public void update(Observable arg0, Object arg1) {
		DataSource ds = (DataSource)arg0;
		int state = ds.getObserverState();

		if (state == DataSource.NEW_VALUE){
			collect(ds.getValue());
		}
		else {
			if (ds instanceof IterativeDataSource){
				if (state == IterativeDataSource.RESET)
					reset();
			}
		}
	}

	/** Returns the value of the observation for the current period
	 *  This is the last observed value
	 *
	 * @return
	 */
	public double getObservation(){
		return myXt;
	}

	/** Get the last forecast made, i.e. for the next period. In a current period t
	 *  this method returns the forecast for period t+1
	 *  A new forecast becomes available after a new
	 *  observation is collected if the forecaster has been initialized
	 *  This method will return Double.NaN if the forecaster has not be initialized
	 *
	 * @return
	 */
	public double getNextPeriodForecast(){
		return myNextPeriodForecast;
	}

	/** Get forecast for the current period
	 *  A new forecast for the next period becomes available after a new
	 *  observation is collected. After the new forecast is made (for the next period)
	 *  this method returns the forecast for the the current period)
	 *  This method will return Double.NaN if the forecaster has not be initialized
	 *
	 *
	 * @return
	 */
	public double getCurrentPeriodForecast(){
		return myCurrentPeriodForecast;
	}

	/** Returns the observation from the previous period.  If the forecast has just
	 *  observed the observation for period t, then this method returns x(t-1)
	 *
	 * @return
	 */
	public double getPrevObservation(){
		return myPrevX;
	}

	/** Gets the last computed forecast error.  A new error becomes
	 *  available after a new observation is collected
	 *
	 * @return
	 */
	public double getError(){
		return myError;
	}

	/** Clears the forecaster, making it as if it has not observed
	 *  any data values.  The history is reset (resetHistory())
	 *  The forecaster also becomes not initialized.  The error
	 *  statistics are also reset.
	 *
	 */
	public void reset(){
		reset(true);
	}

	/** Clears the forecaster, making it as if it has not observed
	 *  any data values.  The history is reset (resetHistory())
	 *  The forecaster also becomes not initialized.
	 *
	 * @param clearErrorStats, if true the error statistics are also cleared
	 */
	public void reset(boolean clearErrorStats){
		myTimeIndex = 0;
	    myNextPeriodForecast = Double.NaN;
	    myError = Double.NaN;
	    myCurrentPeriodForecast = Double.NaN;
	    myXt = Double.NaN;
	    myPrevX = Double.NaN;
	    setInitFlag(false);
	    clearHistory();
	    if (clearErrorStats)
	    	resetErrorStatistics();
	    setChanged();
	    notifyObservers(RESET);
	    //TODO can we implement methods to reset the forecast metric
        myMAD = ForecastMetric.createMeanAbsoluteError();
        myMSE = ForecastMetric.createMeanSquareError();
        myMAPE = ForecastMetric.createMeanAbsolutePercentageError();
        myComputationTime.reset();
	}


	/** Resets all the error collection without clearing/resetting the forecaster
	 *
	 */
	public void resetErrorStatistics(){
	    myStat.reset();
	    myErrorStat.reset();
	}

	/** This method causes the observation to be collected.
	 *  If the forecast has been initialize, then
	 *  the error and new forecast is generated
	 *  The history is also updated
	 *
	 * @param x
	 */
	public void collect(double x){
		double t1 = System.nanoTime();
		// we are observing the actual data for period t
		myStat.collect(x);
	    // increment the period
		myTimeIndex++;
		// remember the forecast from the last period for this current period
	    myCurrentPeriodForecast = myNextPeriodForecast;
	    // remember the value of the last observation, before updating it
	    myPrevX = myXt;
		// x is the new observation for the current period t, remember it
		myXt = x;
		// update the history with the new data
		updateHistory(myTimeIndex, x);
		if (isInitialized()){
			computeError(myTimeIndex, x);
			updateForecast(myTimeIndex, x);
		}
		// check if sufficient data is available
		checkForSufficientHistory();
		// notify observers of the new data collection
		notifyObservers(NEW_DATA);
		double t2 = System.nanoTime();

		myComputationTime.collect(t2 - t1);
	}

	/** Checks to see if the technique is in the given observer state.
	 *  This method can be used by observers that are interested in reacting to the
	 *  action associated with this state for the technique.
	 *
	 *  UNINITIALIZED
	 *  INITIALIZED
	 *
	 * @return True means that this technique's is in the given state.
	 */
	public final boolean checkObserverState(int observerState) {
		return(myObserverState == observerState);
	}

	/**  Returns an integer representing the state of the technique
	 *   This can be used by Observers to find out which action occurred for the technique
	 *
	 * @return The state of the technique
	 */
	public final int getObserverState() {
		return (myObserverState);
	}

	/**  Returns an integer representing the previous state of the technique
	 *   This can be used by Observers to find out which action occurred prior to the current state change for the technique
	 *
	 * @return The state of the technique
	 */
	public final int getPreviousObserverState() {
		return (myPreviousObserverState);
	}

	/** Get the forecast for n periods ahead.
	 *  n=1 should return getForecast()
	 *  This will return Double.NaN if the forecast has not be initialized
	 *
	 * @param n, n > 0 or an IllegalArguementException will be thrown
	 * @return
	 */
	abstract public double getForecast(int n);

	/** Get the next n periods of forecast in an array
	 *
	 * @param n
	 * @return
	 */
	abstract public double[] getForecasts(int n);

	/** Gets the observation values that the forecaster considers part of its history
	 *  This is not necessarily every observation
	 *
	 * @return
	 */
	abstract public double[] getHistory();

	/** Every forecast technique may have different ways to be initialized
	 *  The purpose of this method is to allow the initialization of the forecasting
	 *  technique.  The supplied arrays can be interpreted in any manner
	 *  by subclasses; however, the intention is to have data[] represent
	 *  the past data to use for the initialization, forecasts[] to be used
	 *  for past forecasts, and parameters[] to be used for other parameters
	 *  needed for initialization.  It is up to sub-classes to utilize the
	 *  elements of the arrays appropriately (if at all).  It is up to sub-classes
	 *  to appropriately take into account the use of this method, prior to making
	 *  the 1st forecast (from start up or after a reset).
	 *
	 *  This method need not be called.  Sub-classes should provide a default
	 *  initialization procedure that ensures initialization after a certain
	 *  amount of data has been observed.
	 *
	 * @param data
	 * @param forecasts
	 * @param parameters
	 */
	abstract public void initializeHistory(double[] data, double[] forecasts, double[] parameters);

	/** The current time index all the count of the values observed. Should return
	 *  the current period number, if zero then no periods have been
	 *  observed
	 *
	 * @return
	 */
	public int getTimeIndex(){
		return myTimeIndex;
	}

	/** Returns the statistics on the observed error
	 *
	 * @return
	 */
	public StatisticAccessorIfc getErrorStatistics(){
		return myErrorStat;
	}

	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	protected void setId(){
		myIdCounter_ = myIdCounter_ + 1;
		myId = myIdCounter_;
	}

	protected void setInitFlag(boolean flag){
		myInitFlag = flag;
	}

	/** Used to notify observers that this technique has entered the given state.
	 *  Valid values for observerState include:
	 *
	 *  NEW_DATA
	 *  RESET
	 *
	 * @param observerState
	 */
	protected final void notifyObservers(int observerState) {
		setObserverState(observerState);
		setChanged();
		notifyObservers();
		notifyMetrics();
	}

	/**
	 * Notifies all of the attached metrics that a new forecast has been made.
	 */
	protected void notifyMetrics() {
           //System.out.println("attached metrics");
		for(ForecastMetric m:attachedMetrics)
			m.collectForecastDataForPeriod(this);

		myMAD.collectForecastDataForPeriod(this);
		myMSE.collectForecastDataForPeriod(this);
		myMAPE.collectForecastDataForPeriod(this);

	}

	/** Should be called after a new observation has arrived
	 *  to compute the error for the period.  This method
	 *  also tabulates the error statistics
	 *
	 * @param i, the index of the currently observed data point
	 * @param x the observed data for the current period
	 */
	protected void computeError(int i, double x){
		myError = myCurrentPeriodForecast - x;
		myErrorStat.collect(myError);
	}

	/**
	 * @param myObserverState The myObserverState to set.
	 */
	protected final void setObserverState(int observerState) {
		myPreviousObserverState = myObserverState;
		myObserverState = observerState;
	}

	/** The purpose of this method is to compute a new forecast
	 *  after new data has arrived.
	 *
	 * @param i, the index of the currently observed data point
	 * @param x the observed data for the current period
	 */
	abstract protected void updateForecast(int i, double x);

	/** The purpose of this method is to update the history
	 *  for the forecasting technique
	 *
	 * @param i, the index of the currently observed data point
	 * @param x, represents the newly observed data point
	 */
	abstract protected void updateHistory(int i, double x);

	/** The purpose of this method is to clear (reset) the history associated with
	 *  the forecasting technique.  It is called from
	 *
	 *  reset(boolean clearErrorStats)
	 *
	 *  in order to handle any specific sub-class behavior to clear the history
	 *
	 */
	//TODO how to handle when we try to reset the forecast technique
	//when initializing the ft with user specified history
	abstract protected void clearHistory();

	/** The purpose of this method is to check if sufficient history
	 *  has been obtained and if so initialize the technique if it has
	 *  not already been initialized.
	 *
	 */
	abstract protected void checkForSufficientHistory();

	//TODO
	public abstract double [] estimateInitialForecastParameter(double [] history);

	public void setInitialForecastParameter(double[] parameter) {
		myInitialParameters = parameter;
	}

	public double[] getInitialForecastEstimate() {
		return myInitialParameters;
	}

	public Map<String, Double> getMapParameterLabelToValue(){
		return myMapParameterLabelToValue;
	}

	/**
	 * Attaches a forecast metric to the forecast technique
	 *
	 * @param metric the metric to be attached
	 */
	public void attachMetric(ForecastMetric metric){
		attachedMetrics.add(metric);
	}

	public void removeAllMetrics(){
        attachedMetrics.clear();
    }

    public List<ForecastMetric> getAttachedMetrics(){
        return attachedMetrics;
    }

    public double getMeanDemand(){
    	return myStat.getAverage();
    }

    public double getDemandVariance(){
    	return myStat.getVariance();
    }

    public double getMAD(){
    	myMAD.computeMetricValue();

    	return myMAD.getMetricValue();

    }

    public double getMSE(){
    	myMSE.computeMetricValue();

    	return myMSE.getMetricValue();

    }

    public double getMAPE(){
    	myMAPE.computeMetricValue();

    	return myMAPE.getMetricValue();

    }

	public String toString(){
		StringBuilder sb = new StringBuilder();

		int t = getTimeIndex();
		double px = getPrevObservation();
		double x = getObservation();
		double pf = getCurrentPeriodForecast();
		double f = getNextPeriodForecast();
		double e = getError();

		sb.append("ID " + getId() + "\n");
		sb.append("Name " + getName() + "\n");
		sb.append("t \t x(t-1) \t x(t) \t f(t) \t e(t)  \t f(t+1)\n");
		sb.append(t + "\t " + px + "\t " + x + "\t " + pf + "\t " + e + "\t " + f +"\n");
		sb.append(myErrorStat);

		return(sb.toString());
	}

	public String toShortString(){
	        StringBuilder sb = new StringBuilder();
	        double[] params=getParameters();
	        if(params!=null){
	            int dontAddCommaIndex=params.length-1;
	            for(int i=0;i<params.length;i++){
	                if(i!=dontAddCommaIndex)
	                    sb.append(params[i]+",");
	                else
	                    sb.append(params[i]);
	            }
	        }
	        sb.append(")");
	        return getName()+"("+sb.toString();
	}

	public abstract ForecastTechnique copyInstance();

	public double getTimeElapsed() {
		return myComputationTime.getSum();
	}

	/*//TODO Vijith added this and overridden by
	public abstract ForecastTechnique copyInstance(double [] initpara);*/

}
