package jslx.forecasting;

import jsl.utilities.random.rng.AR1CorrelatedRng;
import jslx.utilities.random.arta.ARTACorrelationFinder;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.Distribution;
import jsl.utilities.random.distributions.NegativeBinomial;
import jsl.utilities.random.distributions.Poisson;
import jsl.utilities.random.distributions.ShiftedDistribution;
import jsl.utilities.random.rng.RngStream;
import jslx.forecasting.demandgeneration.TwoSMarkovChainGeneration;

/**
 * Implements the intermittent forecaster named MCARTA
 * @author vvarghe
 *
 */
public class MCARTA extends ForecastTechnique {

    /*
     * The default values for the control parameters
     * bootstrap size = 10000
     * required number of non-zero demand to initialize = 3
     * update the statistical parameters after initialization = true
     * turn on the arta matching = false
     * 
     */
    public static final double DEFAULT_BOOTSTRAP_SIZE = 10000;
    public static final double DEFAULT_REQD_NUM_TRANS_TO_INITIALIZE = 3;
    public static final double DEFAULT_UPDATE_FREQ = 1;
    public static final double DEFAULT_ARTA_MATCHING_ON_FREQ = Double.POSITIVE_INFINITY;
    public static final double [] DEFAULT_CONTROL_PARA = {DEFAULT_BOOTSTRAP_SIZE, DEFAULT_REQD_NUM_TRANS_TO_INITIALIZE,
    														DEFAULT_UPDATE_FREQ, DEFAULT_ARTA_MATCHING_ON_FREQ};
    
    /*
     * The control parameters
     * bootstrap size
     * required number of non-zero demand to initialize
     * update the statistical parameters after initialization
     * turn on the arta matching
     * 
     */    
    protected double myBootstrapSize;
    protected double myReqdNumTransactionsBeforeInit = 3;
    protected double myUpdatableFreq;
    protected boolean myUpdatable;
    protected double myARTAMatchingOnFreq;
    protected boolean myARTAMatchingOn;
    protected int myTimeIndexOfInitialization;
    protected double myParameterEstimationScope;

    /*
     * The statistical parameters
     * lag 1 correlation coeff of nz demand
     * mean nz demand
     * variance of nz demandn
     * prob of nz demand after nz demand
     * prob of z demand after nz demand
     * 
     */ 
    protected double myNonZeroDemandLag1CorrCoeff;
    protected double myMeanNonZeroDemand;
    protected double myNonZeroDemandVariance;   
    protected double myProb11;
    protected double myProb01;
    
    protected double myInitNonZeroDemandLag1CorrCoeff;
    protected double myInitMeanNonZeroDemand;
    protected double myInitNonZeroDemandVariance;   
    protected double myInitProb11;
    protected double myInitProb01;
    
    /**
     * It computes the estimates of the parameters that are required
     * for the algorithm
     *
     */
    protected boolean myValidEstimateOfStatisticalParameterFromHistory;

    /**
     * It computes the estimates of the parameters that are required
     * for the algorithm
     *
     */
    protected IntermittentDemandStatistic myIntermittentDemandStatistic;

    /**
     * the data until required history are collected
     *
     * the array keep on increasing
     *
     */
    protected double[] myData;
    
    /**
     * It has value 0 if Xt = z and 1 if Xt = nz
     *
     */
    protected int myStateOfXt;

    /**
     * The following are the bootstrap processes
     * MCARTA generator bootsrap the values from the bootstrap process
     *
     */
    private AR1CorrelatedRng myAR1CorrelatedRandomNumberGenerator;
    private TwoSMarkovChainGeneration myTSMCGeneration;
    private NegativeBinomial myNegativeBinomialRandomAmount;
    private Poisson myPoissonRandomAmount;
    private Constant myConstantAmount;
    private ShiftedDistribution myRandomAmount;
    private MCARTAGenerator myMCARTAGenerator;
    private boolean myMedianEstimatesOn;


    /*
     * In this case the MCARTA object will wait until it has valid theta estimates before initializing. 
     * After initializing, it is ready to do forecasting.
     *
     */
    public MCARTA() {
        this(DEFAULT_CONTROL_PARA);
    }
    public MCARTA(double [] controlparameters) {
        this(controlparameters, "DefaultMCARTA");
    }
    public MCARTA(double [] controlparameters, String name) {
        super(name);
        double bootstrapsize = controlparameters[0];
        double reqdtrans = controlparameters[1];
        double updatefreq = controlparameters[2];
        double artamatchingfreq = controlparameters[3];

        double [] para = {bootstrapsize, reqdtrans, updatefreq, artamatchingfreq};
        this.setParameters(para);

        myValidEstimateOfStatisticalParameterFromHistory = false;
        myIntermittentDemandStatistic = new IntermittentDemandStatistic();
        myData = new double[1];

        myAR1CorrelatedRandomNumberGenerator = new AR1CorrelatedRng();
        myNegativeBinomialRandomAmount = new NegativeBinomial();
        myPoissonRandomAmount = new Poisson();
        myConstantAmount = new Constant(0.0);
        myRandomAmount = new ShiftedDistribution(myPoissonRandomAmount, 1.0);
        myTSMCGeneration = new TwoSMarkovChainGeneration();
        myMCARTAGenerator = new MCARTAGenerator(myRandomAmount, myTSMCGeneration, bootstrapsize);
        
        myParameterEstimationScope = Double.POSITIVE_INFINITY;
        myMedianEstimatesOn = false;
        
        myInitNonZeroDemandLag1CorrCoeff = 0;
        myInitMeanNonZeroDemand = 1;
        myInitNonZeroDemandVariance = 1;
        myInitProb11 = 0;
        myInitProb01 = 0;
        
        double [] d = {0}; 
        double [] init_stat_para = {myInitNonZeroDemandLag1CorrCoeff, myInitMeanNonZeroDemand, 
        		myInitNonZeroDemandVariance, myInitProb11, myInitProb01, d[d.length - 1]};
        this.setInitialForecastParameter(init_stat_para);
        
    }

    /*
     * If the user want to specify the parameters. In this case the MCARTA object will
     * have valid theta estimates before initializing. After initializing, it is ready to do forecasting.
     *
     */
    public MCARTA(double [] control_para, double [] init_stat_para) {
        this(control_para, init_stat_para, "DefaultMCARTA");
    }
    public MCARTA(double [] control_para, double [] init_stat_para, String name) {
        this(control_para, name);
        
        double [] data = {init_stat_para[5]};
        double [] forecasts = new double[1];
        
        myInitNonZeroDemandLag1CorrCoeff = init_stat_para[0];
        myInitMeanNonZeroDemand = init_stat_para[1];
        myInitNonZeroDemandVariance = init_stat_para[2];
        myInitProb11 = init_stat_para[3];
        myInitProb01 = init_stat_para[4];
        this.setInitialForecastParameter(init_stat_para);
             
        initializeHistory(data, forecasts, init_stat_para);
    }

    /*
     * If the given demand array cannot give valid theta estimate then the mcarta is not valid
     *
     */
    public static MCARTA buildMCARTA(double [] demandarray){
        return buildMCARTA(DEFAULT_CONTROL_PARA, demandarray);
    }
    public static MCARTA buildMCARTA(double [] control_para, double [] demandarray){
        return buildMCARTA(control_para, demandarray, null);
    }
    public static MCARTA buildMCARTA(double [] control_para, double [] demandarray,
            String name){
        MCARTA mcarta = new MCARTA(control_para, name);
        
        mcarta.myIntermittentDemandStatistic.collect(demandarray);

        if(mcarta.checkValidEstimateFromHistoryForForecastParameters()){
            double phi = mcarta.myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation();
            double nzmean = mcarta.myIntermittentDemandStatistic.getNonZeroDemandStat().getAverage();
            double nzvar = mcarta.myIntermittentDemandStatistic.getNonZeroDemandStat().getVariance();
            double p11 = mcarta.myIntermittentDemandStatistic.getProbSuccessAfterSuccess();
            double p01 = mcarta.myIntermittentDemandStatistic.getProbSuccessAfterFailure();

            double [] data = new double[1];
            double [] forecasts = new double[1];
            double [] init_stat_para = {phi, nzmean, nzvar, p11, p01, demandarray[demandarray.length - 1]};

            mcarta.myInitNonZeroDemandLag1CorrCoeff = init_stat_para[0];
            mcarta.myInitMeanNonZeroDemand = init_stat_para[1];
            mcarta.myInitNonZeroDemandVariance = init_stat_para[2];
            mcarta.myInitProb11 = init_stat_para[3];
            mcarta.myInitProb01 = init_stat_para[4];
            mcarta.setInitialForecastParameter(init_stat_para);
                 
            mcarta.initializeHistory(data, forecasts, init_stat_para);
            
        }else throw new IllegalArgumentException("Demand array is not sufficient enough to estimate statistical parameter");
        
        return mcarta;
    }
    
    //TODO add some exception handling
    public double [] estimateInitialForecastParameter(double [] history){
    	IntermittentDemandStatistic stat = new IntermittentDemandStatistic();
    	
	    for(int i = 0; i<history.length; i++){	
	    	stat.collect(history[i]);
	    }
        double phi = stat.getNonZeroDemandStat().getLag1Correlation();
        double nzmean = stat.getNonZeroDemandStat().getAverage();
        double nzvar = stat.getNonZeroDemandStat().getVariance();
        double p11 = stat.getProbSuccessAfterSuccess();
        double p01 = stat.getProbSuccessAfterFailure();

        double [] init_stat_para = {phi, nzmean, nzvar, p11, p01, history[history.length - 1]};
        
		return init_stat_para;
		
	} 

    protected void clearHistory() {
        myIntermittentDemandStatistic = new IntermittentDemandStatistic();
        
        if(!Double.isInfinite(myParameterEstimationScope)){
            this.myIntermittentDemandStatistic.setSaveDataOption(true);
        }
        
        myData = new double[1];
        myValidEstimateOfStatisticalParameterFromHistory = false;
        myStateOfXt = 1;

        myAR1CorrelatedRandomNumberGenerator = new AR1CorrelatedRng();
        myNegativeBinomialRandomAmount = new NegativeBinomial();
        myPoissonRandomAmount = new Poisson();
        myConstantAmount = new Constant(0.0);
        myRandomAmount = new ShiftedDistribution(myPoissonRandomAmount, 1.0);
        myTSMCGeneration = new TwoSMarkovChainGeneration();
        myMCARTAGenerator = new MCARTAGenerator(myRandomAmount, myTSMCGeneration, myBootstrapSize);   
        
        double [] d = {0};
        double [] stat_para = {myInitNonZeroDemandLag1CorrCoeff, myInitMeanNonZeroDemand, 
        		myInitNonZeroDemandVariance, myInitProb11,  myInitProb01, d[d.length - 1]};  
        setInitialForecastParameter(stat_para);
        
        this.setForecastParameters(stat_para);
        
    }

    /*
     * Getters and setters for control para.
     *
     */
    private void setReqdNumTransactionsBeforeInit(double reqdtrans){
        myReqdNumTransactionsBeforeInit = reqdtrans;
    }
    public double getReqdNumTransactionsBeforeInit() {
        return myReqdNumTransactionsBeforeInit;
    }
    private void setBootstrapSize(double size){
        myBootstrapSize = size;
    }
    public double getBootstrapSize(){
        return myBootstrapSize;
    }
    public double[] getParameters() {
        double[] d = new double[4];
        d[0] = myBootstrapSize;
        d[1] = myReqdNumTransactionsBeforeInit;
        d[2] = myUpdatableFreq;
        d[3] = myARTAMatchingOnFreq;
        return d;
    }
    public void setParameters(double[] parameters) {
        if (parameters == null)
            throw new IllegalArgumentException("The parameters array was null");
        if (parameters.length == 0)
            throw new IllegalArgumentException("parameters length was 0");
        
        setBootstrapSize((int)parameters[0]);
        setReqdNumTransactionsBeforeInit((int)parameters[1]);
        if(parameters[2]==Double.POSITIVE_INFINITY){
        	this.setUpdatable(false);
        }else{
        	this.setUpdatable(true);        	
        }
    	this.myUpdatableFreq = parameters[2];

        if(parameters[3]==Double.POSITIVE_INFINITY){
        	this.setARTAMatchingOn(false);
        }else{
        	this.setARTAMatchingOn(true);        	
        }
    	this.myARTAMatchingOnFreq = parameters[3];
    	
        myMapParameterLabelToValue.put("BootstrapSize", myBootstrapSize);
        myMapParameterLabelToValue.put("ReqdNumTransactionsBeforeInit", myReqdNumTransactionsBeforeInit);
        myMapParameterLabelToValue.put("UpdateFrequency", myUpdatableFreq);
        myMapParameterLabelToValue.put("ARTAMatchingOnFrequency", myARTAMatchingOnFreq);
    }    
    public void setMedianEstimates(boolean updatable){
    	myMedianEstimatesOn = updatable;
    	this.myMCARTAGenerator.setMedianEstimates(updatable);
    }   
    public void setParemeterEstimationScope(double scope){
        myIntermittentDemandStatistic.setSaveDataOption(true);
    	myParameterEstimationScope = scope;
    }    
    private void setUpdatable(boolean updatable){
        myUpdatable = updatable;
    }
    public boolean isUpdatable() {
        return myUpdatable;
    }
    public boolean isARTAMatchingOn() {
        return myARTAMatchingOn;
    }
    private void setARTAMatchingOn(boolean myARTAMatchingOn) {
        this.myARTAMatchingOn = myARTAMatchingOn;
    }

    /*
     * Getters and setters forecast para.
     *
     */
    private void setNonZeroDemandLag1CorrCoeff(double myNonZeroDemandLag1CorrCoeff) {
        this.myNonZeroDemandLag1CorrCoeff = myNonZeroDemandLag1CorrCoeff;
    }
    private void setMeanNonZeroDemand(double myMeanNonZeroDemand) {
        this.myMeanNonZeroDemand = myMeanNonZeroDemand;
    }
    private void setNonZeroDemandVariance(double myNonZeroDemandVariance) {
        this.myNonZeroDemandVariance = myNonZeroDemandVariance;
    }
    public void setProb11(double myProb11) {
        this.myProb11 = myProb11;
    }
    private void setProb01(double myProb01) {
        this.myProb01 = myProb01;
    }
    public double getNonZeroDemandLag1CorrCoeff() {
        return myNonZeroDemandLag1CorrCoeff;
    }
    public double getMeanNonZeroDemand() {
        return myMeanNonZeroDemand;
    }
    public double getNonZeroDemandVariance() {
        return myNonZeroDemandVariance;
    }
    public double getProb11() {
        return myProb11;
    }
    public double getProb01() {
        return myProb01;
    }
    public double[] getForecastParameters() {
        double[] d = new double[5];
        d[0] = myNonZeroDemandLag1CorrCoeff;
        d[1] = myMeanNonZeroDemand;
        d[2] = myNonZeroDemandVariance;
        d[3] = myProb11;
        d[4] = myProb01;
        return d;
    }
    public double[] getForecastParametersEstimatedFromHistory() {
        double[] d = new double[5];
        d[0] = myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation();
        d[1] = myIntermittentDemandStatistic.getNonZeroDemandStat().getAverage();
        d[2] = myIntermittentDemandStatistic.getNonZeroDemandStat().getVariance();
        d[3] = myIntermittentDemandStatistic.getProbSuccessAfterSuccess();
        d[4] = myIntermittentDemandStatistic.getProbSuccessAfterFailure();
        return d;
    }

    /*
     * Whenever the stat. para. are set (via initalization or updating)
     * the bootsrap processes are set too.
     *
     */
    private void setForecastParameters(double [] statpara){
        myNonZeroDemandLag1CorrCoeff = statpara[0];
        myMeanNonZeroDemand = statpara[1];
        myNonZeroDemandVariance = statpara[2];
        myProb11 = statpara[3];
        myProb01 = statpara[4];
        setBootstrapProcess();
    }

    /*
     * Set the bootsrap process. nzphi1 is updated only if there is a
     * stat. significant change in its value. this is controlled quality control process
     *
     */
    private void setBootstrapProcess(){
        double [] prob = {myProb11, myProb01};
        myTSMCGeneration.setParameters(prob);
        myMCARTAGenerator.setStateOfXt(myStateOfXt);
        int distID;
        
        double [] constant = {0.0}; 
        if(myMeanNonZeroDemand == 1.0){
        	myConstantAmount.setParameters(constant);
            myRandomAmount.setDistribution(myConstantAmount,1.0);
            myMCARTAGenerator.setRandomAmount(myRandomAmount);        	
        }else{
            
            double v2mratio = myNonZeroDemandVariance/(myMeanNonZeroDemand - 1.0);
            if(v2mratio > 1.1){
                distID = 1;
            }else{
                distID = 2;
            }            
            
            switch (distID) {
                case 1:
                    double probnz = (myMeanNonZeroDemand - 1.0)/myNonZeroDemandVariance;
                    double numsuccess = ((myMeanNonZeroDemand - 1.0)*probnz)/(1.0-probnz);
                    myNegativeBinomialRandomAmount.setParameters(probnz, numsuccess);
                    myNegativeBinomialRandomAmount.setRandomNumberGenerator(new RngStream());
                    myNonZeroDemandLag1CorrCoeff = controlMatchingProblem(myNegativeBinomialRandomAmount, myNonZeroDemandLag1CorrCoeff);
                    myAR1CorrelatedRandomNumberGenerator.setLag1Correlation(myNonZeroDemandLag1CorrCoeff);
                    myNegativeBinomialRandomAmount.setRandomNumberGenerator(myAR1CorrelatedRandomNumberGenerator);
                    myRandomAmount.setDistribution(myNegativeBinomialRandomAmount,1.0);
                    myMCARTAGenerator.setRandomAmount(myRandomAmount);
                    break;
                case 2:
                    myPoissonRandomAmount.setMean((myMeanNonZeroDemand - 1.0));
                    myPoissonRandomAmount.setRandomNumberGenerator(new RngStream());
                    myNonZeroDemandLag1CorrCoeff = controlMatchingProblem(myPoissonRandomAmount, myNonZeroDemandLag1CorrCoeff);
                    myAR1CorrelatedRandomNumberGenerator.setLag1Correlation(myNonZeroDemandLag1CorrCoeff);
                    myPoissonRandomAmount.setRandomNumberGenerator(myAR1CorrelatedRandomNumberGenerator);
                    myRandomAmount.setDistribution(myPoissonRandomAmount,1.0);
                    myMCARTAGenerator.setRandomAmount(myRandomAmount);
                    break;
                default:
                    throw new IllegalArgumentException("Illegal");
            }
        	
        }
    }

    /*
     * This control the ARTA matching problem of non-zero lag 1 corr coeff
     *
     */
    private double controlMatchingProblem(Distribution d, double nzphi1){
        if(this.isARTAMatchingOn()){
            if(this.isUpdatable()){
            	if((myTimeIndex - myTimeIndexOfInitialization)%myARTAMatchingOnFreq == 0){                	
                    int n = 1000;
    				int r = 1;
    				boolean flag = false;
    				ARTACorrelationFinder b = new ARTACorrelationFinder(d, nzphi1, n, r, flag);
    				boolean found;
					try {
						found = b.checkMatchingCorrelation(nzphi1);
					} catch (Exception e) {
						found = false;
					}
    				if (found == true){
                        //b.setDesiredPrecision(0.005);
    					//nzphi1 = b.findMatchingCorrelation();
                        nzphi1 = b.recommendInitialPoint(50);
    				}            		
            	}
            }else{
                if(!this.isInitialized()){
                    int n = 1000;
                    int r = 1;
                    boolean flag = false;
                    ARTACorrelationFinder b = new ARTACorrelationFinder(d, nzphi1, n, r, flag);
                    boolean found;
					try {
						found = b.checkMatchingCorrelation(nzphi1);
					} catch (Exception e) {
						found = false;
					}
    				if (found == true){
                        //b.setDesiredPrecision(0.005);
                        //nzphi1 = b.findMatchingCorrelation();
                        nzphi1 = b.recommendInitialPoint(50);
                    }
                }
            }
        }
        return nzphi1;
    }

    /*
     * whether the current demand is zero or non-zero
     *
     */
    public int getStateOfXt() {
        return myStateOfXt;
    }

    protected void checkForSufficientHistory() {
        if (!isInitialized()){
            if(checkValidEstimateFromHistoryForForecastParameters()){
                double [] stat_para = new double[5];

                stat_para[0] = myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation();
                stat_para[1] = myIntermittentDemandStatistic.getNonZeroDemandStat().getAverage();
                stat_para[2] = myIntermittentDemandStatistic.getNonZeroDemandStat().getVariance();
                stat_para[3] = myIntermittentDemandStatistic.getProbSuccessAfterSuccess();
                stat_para[4] = myIntermittentDemandStatistic.getProbSuccessAfterFailure();
                
                this.setForecastParameters(stat_para);
                
                myInitNonZeroDemandLag1CorrCoeff = stat_para[0];
                myInitMeanNonZeroDemand = stat_para[1];
                myInitNonZeroDemandVariance = stat_para[2];
                myInitProb11 = stat_para[3];
                myInitProb01 = stat_para[4];

                this.setInitialForecastParameter(stat_para);
                
                myNextPeriodForecast = bootstrap();
                myTimeIndexOfInitialization = myTimeIndex;
                setInitFlag(true);
            }
        }
    }

    /*
     * check whether para estimated from the history is valid
     *
     */
    private boolean checkValidEstimateFromHistoryForForecastParameters(){
        /*
         * Here we myValidThetaEstimate is initialized as false
         * Once myValidThetaEstimate is true we dont want to check
         * the criterions for myValidThetaEstimate to be true.
         * Hence the additional if loop
         *
         */
        if(!this.myValidEstimateOfStatisticalParameterFromHistory){
        	/*
             * Check whether the required num of transactions has arrived
             * check whether the mu is not NAN
             * check whether the mu is not Infinite
             *
             */
            if(isMUInitialized(myIntermittentDemandStatistic)){
                /*
                 * check whether the sigma is not NAN
                 * check whether the sigma is not Infinite
                 *
                 */
                if(isSigmaInitialized(myIntermittentDemandStatistic)){
                    /*
                     * check whether the phi is not NAN
                     * check whether the phi is not Infinite
                     *
                     */
                    if(isPhi1Initialized(myIntermittentDemandStatistic)){
                        /*
                         * check whether the transition probabilities are not NAN
                         * check whether the transition probabilities are not Infinite
                         *
                         */
                        if(isTransProbInitialized(myIntermittentDemandStatistic)){
                            myValidEstimateOfStatisticalParameterFromHistory =  true;
                        }
                    }
                }
            }
        }
        return myValidEstimateOfStatisticalParameterFromHistory;
    }

    /*
     * returns forecast for the next n horizons
     *
     */
    public double getForecast(int n) {
            return myNextPeriodForecast;
    }

    public double[] getForecasts(int n) {
            if (n < 1)
                    throw new IllegalArgumentException("The supplied n was < 1");
            double[] f = new double[n+1];
            double[] f0 = bootstrap(n);
            f[0] = Double.NaN;
            for (int i=1;i<=n;i++)
                    f[i] = f0[i-1];
            return f;
    }

    public double[] getHistory() {
            if (!isInitialized()){
                    return myData;
            }else{
                    double [] tempData = new double [1];
                    tempData[0] = myXt;
                    myData = tempData;
                    return myData;
            }
    }

    public void initializeHistory(double[] data, double[] forecasts,
                    double[] stat_para) {
    	
    	//TODO
        myData = new double[1];        
        myXt = stat_para[5];
        myData [0] = myXt;
        myTimeIndex++;
        myStat.collect(myXt);
        if(myXt == 0){
            myStateOfXt = 0;        	
        }else{
            myStateOfXt = 1;        	
        }
        myTimeIndexOfInitialization = myTimeIndex;
        
        myIntermittentDemandStatistic = new IntermittentDemandStatistic();
        
        if(!Double.isInfinite(myParameterEstimationScope)){
            this.myIntermittentDemandStatistic.setSaveDataOption(true);
        }
        
        for(int i = 0; i<data.length; i++){
            myIntermittentDemandStatistic.collect(data[i]);            
        }       

        this.setForecastParameters(stat_para);
        
        myInitNonZeroDemandLag1CorrCoeff = stat_para[0];
        myInitMeanNonZeroDemand = stat_para[1];
        myInitNonZeroDemandVariance = stat_para[2];
        myInitProb11 = stat_para[3];
        myInitProb01 = stat_para[4];

        this.setInitialForecastParameter(stat_para);
        
        myNextPeriodForecast = bootstrap();
        
        setInitFlag(true);
    }

    protected void updateForecast(int i, double x) {
        myNextPeriodForecast = bootstrap();
        
    }

    /*
     * bootstrap the value from the bootstrap process which is inside the mcarta generator
     *
     */
    protected double bootstrap(){
        return myMCARTAGenerator.getValue();
    }
    
    /*
     * bootstrap the value for n horizons from the bootstrap process which is inside the mcarta generator
     *
     */
    protected double[] bootstrap(int n){
        return myMCARTAGenerator.getValue(n);
    }

    protected void updateHistory(int i, double x) {
        myIntermittentDemandStatistic.collect(x);  
        if(myXt == 0){
            myStateOfXt = 0;
        }else if(myXt > 0 && (!Double.isInfinite(myXt))){ // check the second condn
            myStateOfXt = 1;
        }
        if(!isInitialized()){
            double [] tempData = new double [i];
            System.arraycopy(myData, 0, tempData, 0, myData.length);
            //index begins at 0
            tempData[i-1] = x;
            myData = tempData;
        }else{
        	//only if it is updatable and if there is suff history to estimate the forecast para
        	//if the user is specifying the para and if it is updatable and if there is suff history
        	// the user specified para will be updated
        	//TODO controls the frequency of updatability
        	if((myTimeIndex - myTimeIndexOfInitialization)%myUpdatableFreq == 0){
                updateForecastParameters();        		
        	}        	
        }
    }

    /*
     * updates the stat. control para. if the mcarta stat para are updatable
     * further updating is done only when there is a valid estimates for the stat. para.
     *
     */
    private void updateForecastParameters(){
        if(myUpdatable){
            if(checkValidEstimateFromHistoryForForecastParameters()){
            	/*
            	 * If the whole data used to estimate
            	 */
            	if(Double.isInfinite(myParameterEstimationScope)){ 
        			setStatParaFromWholeHistory();            		
            	}else{
            		double [] array = myIntermittentDemandStatistic.getDemandsData();
            		double [] scopearray = new double [(int)myParameterEstimationScope];
            		/*
            		 * If a portion of data used to estimate
            		 * 
            		 */
            		if(array.length >= myParameterEstimationScope){
            			System.arraycopy(array, (int)(array.length - this.myParameterEstimationScope), 
            					scopearray, 0, (int)myParameterEstimationScope);
            			IntermittentDemandStatistic stat = new IntermittentDemandStatistic();
            			stat.collect(scopearray);
            			boolean b = false;
                        if(isMUInitialized(stat)){
                           if(isSigmaInitialized(stat)){
                                if(isPhi1Initialized(stat)){
                                    if(isTransProbInitialized(stat)){
                                        b =  true;
                                    }
                                }
                            }
                        }
                		/*
                		 * If a portion of data used to estimate and we have valid estimates
                		 * and we have enpugh scope to begin estimate
                		 */
                        if(b){              	
                            double [] stat_para = new double[5];
                            double phi = stat.getNonZeroDemandStat().getLag1Correlation();
                            if(phi < 0){        	
                            	phi = Math.max(phi, -1.0);        	
                            }
                            if(phi > 0){        	
                            	phi = Math.min(phi, 1.0);        	
                            }        
                            stat_para[0] = phi;
                            stat_para[0] = stat.getNonZeroDemandStat().getLag1Correlation();
                            stat_para[1] = stat.getNonZeroDemandStat().getAverage();
                            stat_para[2] = stat.getNonZeroDemandStat().getVariance();
                            stat_para[3] = stat.getProbSuccessAfterSuccess();
                            stat_para[4] = stat.getProbSuccessAfterFailure();
                            this.setForecastParameters(stat_para);                        	
                        }else{
                    		/*
                    		 * If a portion of data used to estimate and we dont
                    		 * have valid estimate
                    		 */
                			setStatParaFromWholeHistory();                        	
                        }            			
            		}else{ 
                		/*
                		 * If a portion of data used to estimate and we have valid estimates
                		 * but we dont have enough scope to begin estimate
                		 */
            			setStatParaFromWholeHistory();
            		}
            		
            	}
            }  
    	}        
    }
    
    /*
     * Set the statistical parameters of MCARTA based on the whole history
     * 
     *
     */
    private void setStatParaFromWholeHistory(){              	
        double [] stat_para = new double[5];
        double phi = myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation();
        if(phi < 0){        	
        	phi = Math.max(phi, -1.0);        	
        }
        if(phi > 0){        	
        	phi = Math.min(phi, 1.0);        	
        }        
        stat_para[0] = phi;
        stat_para[1] = myIntermittentDemandStatistic.getNonZeroDemandStat().getAverage();
        stat_para[2] = myIntermittentDemandStatistic.getNonZeroDemandStat().getVariance();
        stat_para[3] = myIntermittentDemandStatistic.getProbSuccessAfterSuccess();
        stat_para[4] = myIntermittentDemandStatistic.getProbSuccessAfterFailure();
        this.setForecastParameters(stat_para);    	
    }

    /*
     * whether nzmean can be intialized from the history
     *
     */
    public boolean isMUInitialized(IntermittentDemandStatistic stat){
        if(stat.getNonZeroDemandStat().getCount() >= (int)myReqdNumTransactionsBeforeInit){
                return true;
        }else{
                return false;
        }
    }
    
    /*
     * whether nzvariance can be intialized from the history
     *
     */
    public boolean isSigmaInitialized(IntermittentDemandStatistic stat){
            boolean o;
            if(Double.isNaN(stat.getNonZeroDemandStat().getStandardDeviation())
                            ||(stat.getNonZeroDemandStat().getStandardDeviation()<=0)
                            ||Double.isInfinite(stat.getNonZeroDemandStat().getStandardDeviation())){
                    o = false;
            }else{
                    o = true;
            }
            return o;
    }
    
    /*
     * whether nzphi1 can be intialized from the history
     *
     */
    public boolean isPhi1Initialized(IntermittentDemandStatistic stat){
            boolean o;
            if(Double.isNaN(stat.getNonZeroDemandStat().getLag1Correlation())
                            ||Double.isInfinite(stat.getNonZeroDemandStat().getLag1Correlation())){
                    o = false;
            }else{
                    o = true;
            }
            return o;
    }
    
    /*
     * whether trans prob can be intialized from the history
     *
     */
    public boolean isTransProbInitialized(IntermittentDemandStatistic stat){
            boolean o;
            if(Double.isNaN(stat.getProbSuccessAfterFailure())
                            ||Double.isInfinite(stat.getProbSuccessAfterFailure())
                            ||Double.isNaN(stat.getProbSuccessAfterSuccess())
                            ||Double.isInfinite(stat.getProbSuccessAfterSuccess())
                            ){
                    o = false;
            }else{
                    o = true;
            }
            return o;
    }
	
    //TODO when MCARTA is constructed based on user specified forecast parameters or user specified demand array
    public ForecastTechnique copyInstance() {
    	
    	MCARTA mcarta = new MCARTA(this.getParameters(), this.getName()); 
        
        mcarta.setForecastParameters(this.getInitialForecastEstimate());
        
        if(!Double.isInfinite(this.myParameterEstimationScope)){
            mcarta.setParemeterEstimationScope(this.myParameterEstimationScope);
        }
        
        mcarta.setMedianEstimates(this.myMedianEstimatesOn);
        
		// TODO Auto-generated method stub
		return mcarta;
	}

}
