package jslx.forecasting;

import jsl.utilities.random.distributions.Distribution;
import jsl.utilities.random.distributions.NegativeBinomial;
import jsl.utilities.random.distributions.Poisson;
import jsl.utilities.random.distributions.ShiftedDistribution;
import jsl.utilities.random.rng.AR1CorrelatedRng;
import jslx.utilities.random.arta.ARTACorrelationFinder;
import jsl.utilities.random.rng.RngStream;
import jsl.utilities.statistic.Statistic;
import jslx.forecasting.demandgeneration.TwoSMarkovChainGeneration;
import jslx.forecasting.metrics.ForecastMetric;

/**
 * Implements the intermittent forecaster named MCARTA
 * @author vvarghe
 *
 */
public class MCARTAold extends ForecastTechnique {

    /*
     * The default values for the control parameters
     * bootstrap size = 10000
     * required number of non-zero demand to initialize = 3
     * update the statistical parameters after initialization = true
     * turn on the arta matching = false
     * 
     */
    public static final int DEFAULT_BOOTSTRAP_SIZE = 10000;
    public static final int DEFAULT_REQD_NUM_TRANS_TO_INITIALIZE = 3;
    public static final int [] DEFAULT_CONTROL_PARA = {DEFAULT_BOOTSTRAP_SIZE, DEFAULT_REQD_NUM_TRANS_TO_INITIALIZE};
    public static final boolean DEFAULT_UPDATABLE_FLAG_PARAMETER = true;
    public static final boolean DEFAULT_ARTA_MATCHING_ON_FLAG_PARAMETER = false;
    public static final boolean [] DEFAULT_CONTROL_FLAG_PARAMETERS = {DEFAULT_UPDATABLE_FLAG_PARAMETER, 
                                                            DEFAULT_ARTA_MATCHING_ON_FLAG_PARAMETER};
    /*
     * The control parameters
     * bootstrap size
     * required number of non-zero demand to initialize
     * update the statistical parameters after initialization
     * turn on the arta matching
     * 
     */    
    protected int myBootstrapSize;
    protected int myReqdNumTransactionsBeforeInit = 3;
    protected boolean myUpdatable;
    protected boolean myARTAMatchingOn;
    //TODO
    protected boolean myQCBasedTriggerARTAMatchingOn;
    protected Statistic myLag1CorrCoeffStat;
    //TODO
    protected boolean myErrorSignalBasedTriggerUpdating;    
    protected double mySignalZ; 
    protected ForecastMetric myMAD;

    /*
     * The statistical parameters
     * lag 1 correlation coeff of nz demand
     * mean nz demand
     * variance of nz demandn
     * prob of nz demand after nz demand
     * prob of z demand after nz demand
     * 
     */ 
    protected double myNonZeroDemandLag1CorrCoeff; 
    protected double myUnMatchedNonZeroDemandLag1CorrCoeff;
    protected double myMeanNonZeroDemand;
    protected double myNonZeroDemandVariance;   
    protected double myProb11;
    protected double myProb01;
    
    /**
     * It computes the estimates of the parameters that are required
     * for the algorithm
     *
     */
    protected boolean myValidEstimateOfStatisticalParameterFromHistory;

    /**
     * It computes the estimates of the parameters that are required
     * for the algorithm
     *
     */
    protected IntermittentDemandStatistic myIntermittentDemandStatistic;

    /**
     * the data until required non zero demands are collected
     *
     * the array keep on increasing
     *
     */
    protected double[] myData;
    
    /**
     * It has value 0 if Xt = z and 1 if Xt = nz
     *
     */
    protected int myStateOfXt;

    /**
     * The following are the bootstrap processes
     * MCARTA generator bootsrap the values from the bootstrap process
     *
     */
    private AR1CorrelatedRng myAR1CorrelatedRandomNumberGenerator;
    private TwoSMarkovChainGeneration myTSMCGeneration;
    private NegativeBinomial myNegativeBinomialRandomAmount;
    private Poisson myPoissonRandomAmount;
    private ShiftedDistribution myRandomAmount;
    private MCARTAGenerator myMCARTAGenerator;


    /*
     * In this case the MCARTA object will wait until it has valid theta estimates before initializing. 
     * After initializing, it is ready to do forecasting.
     *
     */
    public MCARTAold() {
        this(DEFAULT_CONTROL_PARA);
    }
    public MCARTAold(int [] controlparameters) {
        this(controlparameters, DEFAULT_CONTROL_FLAG_PARAMETERS);
    }
    public MCARTAold(int [] controlparameters, boolean [] controlflag) {
        this(controlparameters, controlflag, null);
    }
    public MCARTAold(int [] controlparameters, boolean [] controlflag, String name) {
        super(name);
        int bootstrapsize = controlparameters[0];
        int reqdtrans = controlparameters[1];

        double [] para = {bootstrapsize, reqdtrans};
        this.setParameters(para);
        setControlFlags(controlflag);

        myValidEstimateOfStatisticalParameterFromHistory = false;
        myIntermittentDemandStatistic = new IntermittentDemandStatistic();
        myData = new double[1];

        myAR1CorrelatedRandomNumberGenerator = new AR1CorrelatedRng();
        myNegativeBinomialRandomAmount = new NegativeBinomial();
        myPoissonRandomAmount = new Poisson();
        myRandomAmount = new ShiftedDistribution(myPoissonRandomAmount, 1.0);
        myTSMCGeneration = new TwoSMarkovChainGeneration();
        myMCARTAGenerator = new MCARTAGenerator(myRandomAmount, myTSMCGeneration, bootstrapsize);
        

		myMAD = ForecastMetric.createMeanAbsoluteError();
		this.attachMetric(myMAD);
		
		myLag1CorrCoeffStat = new Statistic("Lag1CorrCoeffStat");

    }

    /*
     * If the user want to specify the parameters. In this case the MCARTA object will
     * have valid theta estimates before initializing. After initializing, it is ready to do forecasting.
     *
     */
    public MCARTAold(double [] initparameters) {
        this(DEFAULT_CONTROL_PARA, initparameters);
    }
    public MCARTAold(int [] control_para, double [] initparameters) {
        this(control_para, initparameters, DEFAULT_CONTROL_FLAG_PARAMETERS);
    }
    public MCARTAold(int [] control_para, double [] init_stat_para, boolean [] controlflag) {
        this(control_para, init_stat_para, controlflag, null);
    }
    public MCARTAold(int [] control_para, double [] init_stat_para, boolean [] controlflag, String name) {
        
    	this(control_para, controlflag, name);
        double [] data = new double[2];
        double [] forecasts = new double[1];
        //TODO
        // we make an arbitrary rule here
        myData = new double[2];
        
        data [0] = 0;
        myPrevX = data [0];
        myData [0] = myPrevX;
        myTimeIndex++;
        myStat.collect(myPrevX);
        
        data [1] = 0;
        myXt = data [1];
        myData [0] = myXt;
        myTimeIndex++;
        myStat.collect(myXt);
        myStateOfXt = 0;
        
        initializeHistory(data, forecasts, init_stat_para);
    }

    /*
     * If the given demand array cannot give valid theta estimate then the mcarta is not valid
     *
     */
    public static MCARTAold buildMCARTA(double [] demandarray){
        return buildMCARTA(DEFAULT_CONTROL_PARA, demandarray);
    }
    public static MCARTAold buildMCARTA(int [] control_para, double [] demandarray){
        return buildMCARTA(control_para, demandarray, DEFAULT_CONTROL_FLAG_PARAMETERS);
    }
    public static MCARTAold buildMCARTA(int [] control_para, double [] demandarray, boolean [] controlflag){
        return buildMCARTA(control_para, demandarray, controlflag, null);
    }
    public static MCARTAold buildMCARTA(int [] control_para, double [] demandarray, boolean [] controlflag,
            String name){
        MCARTAold mcarta = new MCARTAold(control_para, controlflag, name);
        for(int i = 0; i<=demandarray.length; i++){
            mcarta.myStat.collect(demandarray[i]);
            mcarta.myTimeIndex++;
            mcarta.myCurrentPeriodForecast = mcarta.myNextPeriodForecast;
            mcarta.myPrevX = mcarta.myXt;
            mcarta.myXt = demandarray[i];

            mcarta.myIntermittentDemandStatistic.collect(demandarray[i]);
            if(mcarta.myXt == 0){
                mcarta.myStateOfXt = 0;
            }else if(mcarta.myXt > 0 && (!Double.isInfinite(mcarta.myXt))){
                mcarta.myStateOfXt = 1;
            }

            double [] tempData = new double [i];
            System.arraycopy(mcarta.myData, 0, tempData, 0, mcarta.myData.length);
            tempData[i-1] = demandarray[i];
            mcarta.myData = tempData;
        }

        if(mcarta.checkValidStatControlParaEstimatedFromHistory()){
            double phi = mcarta.myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation();
            double nzmean = mcarta.myIntermittentDemandStatistic.getNonZeroDemandStat().getAverage();
            double nzvar = mcarta.myIntermittentDemandStatistic.getNonZeroDemandStat().getVariance();
            double p11 = mcarta.myIntermittentDemandStatistic.getProbSuccessAfterSuccess();
            double p01 = mcarta.myIntermittentDemandStatistic.getProbSuccessAfterFailure();

            double [] data = new double[1];
            double [] forecasts = new double[1];
            double [] init_stat_para = {phi, nzmean, nzvar, p11, p01};
            mcarta.initializeHistory(data, forecasts, init_stat_para);
        }else throw new IllegalArgumentException("Demand array is not sufficient enough to estimate statistical parameter");
        
        return mcarta;
    }

    protected void clearHistory() {
        myIntermittentDemandStatistic = new IntermittentDemandStatistic();
        myData = new double[1];
        myValidEstimateOfStatisticalParameterFromHistory = false;
        myStateOfXt = 1;

        myAR1CorrelatedRandomNumberGenerator = new AR1CorrelatedRng();
        myNegativeBinomialRandomAmount = new NegativeBinomial();
        myPoissonRandomAmount = new Poisson();
        myRandomAmount = new ShiftedDistribution(myPoissonRandomAmount, 1.0);
        myTSMCGeneration = new TwoSMarkovChainGeneration();
        myMCARTAGenerator = new MCARTAGenerator(myRandomAmount, myTSMCGeneration, myBootstrapSize);
        
        

    }

    /*
     * Getters and setters for control para.
     *
     */
    private void setReqdNumTransactionsBeforeInit(int reqdtrans){
        myReqdNumTransactionsBeforeInit = reqdtrans;
    }
    public int getReqdNumTransactionsBeforeInit() {
        return myReqdNumTransactionsBeforeInit;
    }
    private void setBootstrapSize(int size){
        myBootstrapSize = size;
        //myMCARTAGenerator.setBSSize(size);
    }
    public int getBootstrapSize(){
        return myBootstrapSize;
    }
    public double[] getParameters() {
        double[] d = new double[2];
        d[0] = myBootstrapSize;
        d[1] = myReqdNumTransactionsBeforeInit;
        return d;
    }
    public void setParameters(double[] parameters) {
        if (parameters == null)
            throw new IllegalArgumentException("The parameters array was null");
        if (parameters.length == 0)
            throw new IllegalArgumentException("parameters length was 0");
        setControlParameters(parameters);
    }
    private void setControlParameters(double[] parameters){
        setBootstrapSize((int)parameters[0]);
        setReqdNumTransactionsBeforeInit((int)parameters[1]);
        myMapParameterLabelToValue.put("BootstrapSize", (double)myBootstrapSize);
        myMapParameterLabelToValue.put("ReqdNumTransactionsBeforeInit", (double)myReqdNumTransactionsBeforeInit);
    }
    public double[] getControlParameters(){
        double[] d = new double[2];
        d[0] = myBootstrapSize;
        d[1] = myReqdNumTransactionsBeforeInit;
        return d;
    }
    private void setControlFlags(boolean [] controlflag){
        setUpdatable(controlflag[0]);
        setARTAMatchingOn(controlflag[1]);
    }
    private void setUpdatable(boolean updatable){
        myUpdatable = updatable;
    }
    public boolean isUpdatable() {
        return myUpdatable;
    }
    public boolean isARTAMatchingOn() {
        return myARTAMatchingOn;
    }
    private void setARTAMatchingOn(boolean myARTAMatchingOn) {
        this.myARTAMatchingOn = myARTAMatchingOn;
    }

    /*
     * Getters and setters for stat. para.
     *
     */
    private void setNonZeroDemandLag1CorrCoeff(double myNonZeroDemandLag1CorrCoeff) {
        this.myNonZeroDemandLag1CorrCoeff = myNonZeroDemandLag1CorrCoeff;
    }
    private void setMeanNonZeroDemand(double myMeanNonZeroDemand) {
        this.myMeanNonZeroDemand = myMeanNonZeroDemand;
    }
    private void setNonZeroDemandVariance(double myNonZeroDemandVariance) {
        this.myNonZeroDemandVariance = myNonZeroDemandVariance;
    }
    public void setProb11(double myProb11) {
        this.myProb11 = myProb11;
    }
    private void setProb01(double myProb01) {
        this.myProb01 = myProb01;
    }
    public double getNonZeroDemandLag1CorrCoeff() {
        return myNonZeroDemandLag1CorrCoeff;
    }
    public double getMeanNonZeroDemand() {
        return myMeanNonZeroDemand;
    }
    public double getNonZeroDemandVariance() {
        return myNonZeroDemandVariance;
    }
    public double getProb11() {
        return myProb11;
    }
    public double getProb01() {
        return myProb01;
    }
    public double[] getStatistcalParameters() {
        double[] d = new double[5];
        d[0] = myNonZeroDemandLag1CorrCoeff;
        d[1] = myMeanNonZeroDemand;
        d[2] = myNonZeroDemandVariance;
        d[3] = myProb11;
        d[4] = myProb01;
        return d;
    }
    public double[] getStatisticalParametersEstimatedFromHistory() {
        double[] d = new double[5];
        d[0] = myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation();
        d[1] = myIntermittentDemandStatistic.getNonZeroDemandStat().getAverage();
        d[2] = myIntermittentDemandStatistic.getNonZeroDemandStat().getVariance();
        d[3] = myIntermittentDemandStatistic.getProbSuccessAfterSuccess();
        d[4] = myIntermittentDemandStatistic.getProbSuccessAfterFailure();
        return d;
    }

    /*
     * Whenever the stat. para. are set (via initalization or updating)
     * the bootsrap processes are set too.
     *
     */
    private void setStatisticalParameters(double [] statpara){
        myNonZeroDemandLag1CorrCoeff = statpara[0];
        myMeanNonZeroDemand = statpara[1];
        myNonZeroDemandVariance = statpara[2];
        myProb11 = statpara[3];
        myProb01 = statpara[4];
        
        setBootstrapProcess();
    }

    /*
     * Set the bootsrap process. nzphi1 is updated only if there is a
     * stat. significant change in its value. this is controlled quality control process
     *
     */
    private void setBootstrapProcess(){
        double [] prob = {myProb11, myProb01};
        myTSMCGeneration.setParameters(prob);
        myMCARTAGenerator.setStateOfXt(myStateOfXt);
        int distID;
        double cv = myNonZeroDemandVariance/myMeanNonZeroDemand;
        if(cv > 1.1){
            distID = 1;
        }else{
            distID = 2;
        }
        switch (distID) {
            case 1:
                double probnz = myMeanNonZeroDemand/myNonZeroDemandVariance;
                int numsuccess = (int)Math.ceil((myMeanNonZeroDemand*probnz)/(1.0-probnz));
                myNegativeBinomialRandomAmount.setParameters(probnz, numsuccess);
                myNegativeBinomialRandomAmount.setRandomNumberGenerator(new RngStream());
                if(doesARTADetectChange()){
                    myNonZeroDemandLag1CorrCoeff = controlMatchingProblem(myNegativeBinomialRandomAmount, myNonZeroDemandLag1CorrCoeff);
                }
                myAR1CorrelatedRandomNumberGenerator.setLag1Correlation(myNonZeroDemandLag1CorrCoeff);
                myNegativeBinomialRandomAmount.setRandomNumberGenerator(myAR1CorrelatedRandomNumberGenerator);
                myRandomAmount.setDistribution(myNegativeBinomialRandomAmount,1.0);
                myMCARTAGenerator.setRandomAmount(myRandomAmount);
                break;
            case 2:
                myPoissonRandomAmount.setMean(myMeanNonZeroDemand);
                myPoissonRandomAmount.setRandomNumberGenerator(new RngStream());
                if(doesARTADetectChange()){
                    myNonZeroDemandLag1CorrCoeff = controlMatchingProblem(myPoissonRandomAmount, myNonZeroDemandLag1CorrCoeff);
                }
                myAR1CorrelatedRandomNumberGenerator.setLag1Correlation(myNonZeroDemandLag1CorrCoeff);
                myPoissonRandomAmount.setRandomNumberGenerator(myAR1CorrelatedRandomNumberGenerator);
                myRandomAmount.setDistribution(myPoissonRandomAmount,1.0);
                myMCARTAGenerator.setRandomAmount(myRandomAmount);
                break;
            default:
                throw new IllegalArgumentException("Illegal");
        }
    }

    /*
     * nzphi1 is updated only if there is a
     * stat. significant change in its value. this is controlled quality control process.
     * This QC process needs to be implemented in this method.
     * It also requires a boolean flag to turn on and off this check
     *
     */
    private boolean doesARTADetectChange(){
    	//TODO add the QC based ARTA triggering
    	
    	double previousunmatched = myUnMatchedNonZeroDemandLag1CorrCoeff;
    	double sigma = myLag1CorrCoeffStat.getStandardDeviation();
    	
    	boolean b1 = (previousunmatched - 3*sigma) < myNonZeroDemandLag1CorrCoeff;
    	boolean b2 = (previousunmatched + 3*sigma) > myNonZeroDemandLag1CorrCoeff;
    	
    	if(b1 && b2){
    		return true;
    	}else{        	
            return true;    		
    	}
    	
    }

    /*
     * this control the matching problem of nzphi1
     *
     */
    private double controlMatchingProblem(Distribution d, double nzphi1){
        if(this.isARTAMatchingOn()){
        	myUnMatchedNonZeroDemandLag1CorrCoeff = nzphi1;
            if(this.isUpdatable()){
                int n = 10000;
				int r = 10;
				boolean flag = false;
				ARTACorrelationFinder b = new ARTACorrelationFinder(d, nzphi1, n, r, flag);
				boolean found = b.checkMatchingCorrelation(nzphi1);
				if (found == true){
                    b.setDesiredPrecision(0.005);
                    nzphi1 = b.findMatchingCorrelation();
				}
            }else{
                if(!this.isInitialized()){
                    int n = 10000;
                    int r = 10;
                    boolean flag = false;
                    ARTACorrelationFinder b = new ARTACorrelationFinder(d, nzphi1, n, r, flag);
                    boolean found = b.checkMatchingCorrelation(nzphi1);
                    if (found == true){
                        b.setDesiredPrecision(0.005);
                        nzphi1 = b.findMatchingCorrelation();
                    }
                }
            }
        }
        return nzphi1;
    }

    public int getStateOfXt() {
        return myStateOfXt;
    }

    protected void checkForSufficientHistory() {
        if (!isInitialized()){
            if(checkValidStatControlParaEstimatedFromHistory()){
                double [] stat_para = new double[5];

                stat_para[0] = myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation();
                stat_para[1] = myIntermittentDemandStatistic.getNonZeroDemandStat().getAverage();
                stat_para[2] = myIntermittentDemandStatistic.getNonZeroDemandStat().getVariance();
                stat_para[3] = myIntermittentDemandStatistic.getProbSuccessAfterSuccess();
                stat_para[4] = myIntermittentDemandStatistic.getProbSuccessAfterFailure();

                this.setStatisticalParameters(stat_para);

                myNextPeriodForecast = bootstrap();
                setInitFlag(true);
            }
        }
    }

    private boolean checkValidStatControlParaEstimatedFromHistory(){
        /*
         * Here we myValidThetaEstimate is initialized as false
         * Once myValidThetaEstimate is true we dont want to check
         * the criterions for myValidThetaEstimate to be true.
         * Hence the additional if loop
         *
         */
        if(!this.myValidEstimateOfStatisticalParameterFromHistory){
            /*
             * Check whether the required num of transactions has arrived
             * check whether the mu is not NAN
             * check whether the mu is not Infinite
             *
             */
            if(isMUInitialized()){
                /*
                 * check whether the sigma is not NAN
                 * check whether the sigma is not Infinite
                 *
                 */
                if(isSigmaInitialized()){
                    /*
                     * check whether the phi is not NAN
                     * check whether the phi is not Infinite
                     *
                     */
                    if(isPhi1Initialized()){
                        /*
                         * check whether the transition probabilities are not NAN
                         * check whether the transition probabilities are not Infinite
                         *
                         */
                        if(isTransProbInitialized()){
                            myValidEstimateOfStatisticalParameterFromHistory =  true;
                        }
                    }
                }
            }
        }
        return myValidEstimateOfStatisticalParameterFromHistory;
    }

    public double getForecast(int n) {
            return myNextPeriodForecast;
    }

    public double[] getForecasts(int n) {
            if (n < 1)
                    throw new IllegalArgumentException("The supplied n was < 1");
            double[] f = new double[n+1];
            double[] f0 = bootstrap(n);
            f[0] = Double.NaN;
            for (int i=1;i<=n;i++)
                    f[i] = f0[i-1];
            return f;
    }

    public double[] getHistory() {
            if (!isInitialized()){
                    return myData;
            }else{
                    double [] tempData = new double [1];
                    tempData[0] = myXt;
                    myData = tempData;
                    return myData;
            }
    }

    public void initializeHistory(double[] data, double[] forecasts,
                    double[] stat_para) {

        this.setStatisticalParameters(stat_para);

        myNextPeriodForecast = bootstrap();
        
        mySignalZ = 0.0;
        
        setInitFlag(true);
    }

    protected void updateForecast(int i, double x) {
        myNextPeriodForecast = bootstrap();
        
    }

    /*
     * bootstrap the value from the bootstrap process which is inside the mcarta generator
     *
     */
    protected double bootstrap(){
        return myMCARTAGenerator.getValue();
    }
    protected double[] bootstrap(int n){
        return myMCARTAGenerator.getValue(n);
    }

    protected void updateHistory(int i, double x) {
        myIntermittentDemandStatistic.collect(x);
        if(myXt == 0){
            myStateOfXt = 0;
        }else if(myXt > 0 && (!Double.isInfinite(myXt))){ // check the second condn
            myStateOfXt = 1;
        }
        if(!isInitialized()){
            double [] tempData = new double [i];
            System.arraycopy(myData, 0, tempData, 0, myData.length);
            //index begins at 0
            tempData[i-1] = x;
            myData = tempData;
        }else{
        	if(isPhi1Initialized()){
        		myLag1CorrCoeffStat.collect(this.myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation());
        	}
            updateStatControlParameters();
        }
    }

    /*
     * updates the stat. control para. if the mcarta stat para are updatable
     * further updating is done only when there is a valid estimates for the stat. para.
     *
     */
    private void updateStatControlParameters(){
    	//TODO also add the logic for error signal based triggering of this updating
    	// for the error based updating to be triggered we have to flag this true as well as updatable also true
    	boolean errorsbasedupdating = false;
    	if(errorsbasedupdating){
        	double omega = 0.1;
        	mySignalZ = omega * myError + (1 - omega) * mySignalZ;
    		myMAD.computeMetricValue();
        	double mad = myMAD.getMetricValue();
        	
        	if(mySignalZ > - mad && mySignalZ < mad){
                if(myUpdatable){
                    if(checkValidStatControlParaEstimatedFromHistory()){
                        double [] stat_para = new double[5];
                        stat_para[0] = myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation();
                        stat_para[1] = myIntermittentDemandStatistic.getNonZeroDemandStat().getAverage();
                        stat_para[2] = myIntermittentDemandStatistic.getNonZeroDemandStat().getVariance();
                        stat_para[3] = myIntermittentDemandStatistic.getProbSuccessAfterSuccess();
                        stat_para[4] = myIntermittentDemandStatistic.getProbSuccessAfterFailure();
                        this.setStatisticalParameters(stat_para);
                    }
                }
        		
        	}
    		
    	}else{
            if(myUpdatable){
                if(checkValidStatControlParaEstimatedFromHistory()){
                    double [] stat_para = new double[5];
                    stat_para[0] = myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation();
                    stat_para[1] = myIntermittentDemandStatistic.getNonZeroDemandStat().getAverage();
                    stat_para[2] = myIntermittentDemandStatistic.getNonZeroDemandStat().getVariance();
                    stat_para[3] = myIntermittentDemandStatistic.getProbSuccessAfterSuccess();
                    stat_para[4] = myIntermittentDemandStatistic.getProbSuccessAfterFailure();
                    this.setStatisticalParameters(stat_para);
                }
            }    		
    	}
        
    }

    /*
     * whether nzmean can be intialized from the history
     *
     */
    public boolean isMUInitialized(){
        if(this.myIntermittentDemandStatistic.getNonZeroDemandStat().getCount()>=myReqdNumTransactionsBeforeInit){
                return true;
        }else{
                return false;
        }
    }
    /*
     * whether nzvariance can be intialized from the history
     *
     */
    public boolean isSigmaInitialized(){
            boolean o;
            if(Double.isNaN(this.myIntermittentDemandStatistic.getNonZeroDemandStat().getStandardDeviation())
                            ||(this.myIntermittentDemandStatistic.getNonZeroDemandStat().getStandardDeviation()<=0)
                            ||Double.isInfinite(this.myIntermittentDemandStatistic.getNonZeroDemandStat().getStandardDeviation())){
                    o = false;
            }else{
                    o = true;
            }
            return o;
    }
    /*
     * whether nzphi1 can be intialized from the history
     *
     */
    public boolean isPhi1Initialized(){
            boolean o;
            if(Double.isNaN(this.myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation())
                            ||Double.isInfinite(this.myIntermittentDemandStatistic.getNonZeroDemandStat().getLag1Correlation())){
                    o = false;
            }else{
                    o = true;
            }
            return o;
    }
    /*
     * whether trans prob can be intialized from the history
     *
     */
    public boolean isTransProbInitialized(){
            boolean o;
            if(Double.isNaN(this.myIntermittentDemandStatistic.getProbSuccessAfterFailure())
                            ||Double.isInfinite(this.myIntermittentDemandStatistic.getProbSuccessAfterFailure())
                            ||Double.isNaN(this.myIntermittentDemandStatistic.getProbSuccessAfterSuccess())
                            ||Double.isInfinite(this.myIntermittentDemandStatistic.getProbSuccessAfterSuccess())
                            ){
                    o = false;
            }else{
                    o = true;
            }
            return o;
    }
	
    public ForecastTechnique copyInstance() {
		// TODO Auto-generated method stub
		return null;
	}
	@Override
	public double[] estimateInitialForecastParameter(double[] history) {
		// TODO Auto-generated method stub
		return null;
	}

}
