/**
 * 
 */
package jslx.forecasting;

import jsl.utilities.statistic.Statistic;

/** Implements the moving average forecasting technique
 * 
 *  f(t+1) = f(t) + (x(t) - x(t-N))/N
 *  
 *  N = moving average order
 *  
 *
 */
public class MovingAverage extends ForecastTechnique {

	/** The window size
	 */
	protected int myN = 1;
	
	/** the data within the moving average window
	 *  data[0] is the oldest
	 *  data[N-1] is the newest, where N is the window size
	 */
	protected double[] myData;

	/** The oldest data point
	 */
	protected double myOldestX;
	
	/** Creates a no named forecaster for MovingAverage with 
	 *  default window of 1
	 * 
	 */
	public MovingAverage() {
		this(1, null);
	}

	/** Creates a no named forecaster for MovingAverage with 
	 *  default window of 1
	 * 
	 * @param name
	 */
	public MovingAverage(String name) {
		this(1, name);
	}

	/** Creates a no named forecaster for MovingAverage with 
	 *  the provided window size
	 *  
	 * @param windowSize
	 */
	public MovingAverage(int windowSize) {
		this(windowSize, null);
	}	
	
	/** Creates a no named forecaster for MovingAverage with 
	 *  the provided window size
	 *  
	 * @param windowSize
	 * @param name
	 */
	public MovingAverage(int windowSize, String name) {
		super(name);
		setWindowSize(windowSize);	
	}
	
	/**
	 * Creates a shallow copy of a MovingAverage forecast technique object, retaining only the forecast parameters.
	 */
	public ForecastTechnique copyInstance(){
        return new MovingAverage(this.myN,this.getName());
    }
	
	public int getWindowSize(){
		return myN;
	}
	
	protected void setWindowSize(int windowSize) {
		if (windowSize < 1.0)
			throw new IllegalArgumentException("The window size must be >=1");
		myN = windowSize;
		myMapParameterLabelToValue.put("Window Size", (double)myN);
		myData = new double[myN];
	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#getForecast(int)
	 */
	public double getForecast(int n) {
		return myNextPeriodForecast;
	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#getForecasts(int)
	 */
	public double[] getForecasts(int n) {
		if (n < 1)	
			throw new IllegalArgumentException("The supplied n was < 1");
		
		double[] f = new double[n+1];
		f[0] = Double.NaN;
		for (int i=1;i<=n;i++)
			f[i] = myNextPeriodForecast;
		return f;
	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#getHistory()
	 */
	public double[] getHistory() {
		double[] h = new double[myN];
		System.arraycopy(myData, 0, h, 0, myN);
		return h;
	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#clearHistory()
	 */
	protected void clearHistory() {
		myData = new double[myN];
	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#initializeHistory(double[], double[], double[])
	 */
	public void initializeHistory(double[] data, double[] forecasts,
			double[] parameters) {
		
            //TODO
            /*if (data.length <= myN)
                throw new IllegalArgumentException("The supplied data array is not the moving average window size");*/

            /*if (data.length != myN)
                throw new IllegalArgumentException("The supplied data array is not the moving average window size");*/

            //System.arraycopy(data, 0, myData, 0, myN);
                
            System.arraycopy(data, data.length - myN, myData, 0, myN);
                       
            /*double avg = 0.0;
            for(int j=0;j<myN;j++)
                    avg = avg - (avg - myData[j])/(j + 1.0);

            myNextPeriodForecast = avg;
            // initialize it
            setInitFlag(true);*/

            //TODO Vijith
            double [] para = this.estimateInitialForecastParameter(data);
            setInitialForecastParameter(para);

            myNextPeriodForecast = para[0];
            // initialize it
            setInitFlag(true);

	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#updateForecast(int, double)
	 */
	protected void updateForecast(int i, double x) {
		// the new forecast is based on the current forecast, newest observation and the oldest observation
		myNextPeriodForecast = myCurrentPeriodForecast + ((myXt - myOldestX)/myN);
	}

	/* (non-Javadoc)
	 * @see forecasting.ForecastTechnique#updateHistory(int, double)
	 */
	protected void updateHistory(int i, double x) {
		// save the oldest
		myOldestX = myData[0];
		
		if (i <= myN){
                        if(this.isInitialized()){
                            //this is possible when we initialize the forecast tehnique based on demand array
                            // the array is full
                            // shift the array down
                            for(int j=1;j <= myN - 1; j++){
                                    myData[j-1] = myData[j];
                            }
                            // capture the newest data point
                            myData[myN-1] = x;                            
                        }else{
                            // fill up the array to the window size
                            myData[i-1] = x;                            
                        }
		} else {
			// the array is full
			// shift the array down
			for(int j=1;j <= myN - 1; j++){
				myData[j-1] = myData[j];
			}
			// capture the newest data point
			myData[myN-1] = x;
		}
	}

	protected void checkForSufficientHistory(){
		
		if (!isInitialized()){
			// not already initialized
			// need to check if enough data is available for initialization
			if (myTimeIndex == myN){
				// not initialized, and this is the myNth data point
				// make forecast for next period the average
				/*double avg = 0.0;
				for(int j=0;j<myN;j++)
					avg = avg - (avg - myData[j])/(j + 1.0);

				myNextPeriodForecast = avg;
				// initialize it, so when 2nd point is collected
				// it is ready
				setInitFlag(true);*/
				
				//TODO Vijith
				double [] para = this.estimateInitialForecastParameter(myData);
				setInitialForecastParameter(para);
				
				myNextPeriodForecast = para[0];
				// initialize it
				setInitFlag(true);
				
				
			}
		}
	}
	
	/** In the returned array element [0] is the window size
	 * 
	 */
	public final double[] getParameters() {
		double[] d = new double[1];
		d[0] = myN;
		return d;
	}

	/** Element [0] in the supplied array is interpreted
	 *  as the window size
	 * 
	 */
	public final void setParameters(double[] parameters) {
		if (parameters == null)
			throw new IllegalArgumentException("The parameters array was null");
		
		if (parameters.length == 0)
			throw new IllegalArgumentException("parameters length was 0");

		setWindowSize((int)parameters[0]);
	}

	@Override
	public double[] estimateInitialForecastParameter(double[] history) {
		
		if (history.length < myN)
			throw new IllegalArgumentException("The supplied data array is not the moving average window size");		

		double sum = 0.0;
		
                //reads only the last myN demands
		for(int j = 0; j < myN; j++)
			sum = sum + history[history.length - 1 - j];
		
		double [] para = new double[1];
                
		para [0] =  sum/myN;
                
		return para;
	}

}
