/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.forecasting;

import jsl.utilities.statistic.Statistic;

/** Implements the simple exponential smoothing forecasting technique
 * 
 *  new forecast = alpha*latest observation + (1-alpha)*last forecast
 *  
 *  f(t) = alpha*x(t-1) + (1-alpha)f(t-1)
 *       = f(t) - alpha*(f(t-1) - x(t-1)
 *       = f(t) - alpha*e(t-1)
 *       
 *       t = 1, 2, ...
 *  
 *
 */
public class SimpleExponentialSmoothing extends ForecastTechnique {

	/** The smoothing parameter
	 * 
	 */
	protected double myAlpha;
	
	/** Creates a no named forecaster for exponential smoothing with 
	 *  default smoothing parameter 0.1
	 * 
	 */
	public SimpleExponentialSmoothing() {
		this(0.1, null);
	}

	/** Creates a named forecaster for exponential smoothing with 
	 *  default smoothing parameter 0.1
	 * @param name
	 */
	public SimpleExponentialSmoothing(String name) {
		this(0.1, name);
	}

	/** Creates a no named forecaster for exponential smoothing with 
	 *  default smoothing parameter 0.1
	 *  
	 * @param alpha
	 */
	public SimpleExponentialSmoothing(double alpha) {
		this(alpha, null);
	}	
	
	/** Creates a no named forecaster for exponential smoothing with 
	 *  default smoothing parameter 0.1
	 *  
	 * @param alpha
	 * @param name
	 */
	public SimpleExponentialSmoothing(double alpha, String name) {
		super(name);
		setSmoothingParameter(alpha);	
	}	
	
	/** Returns the smoothing parameter
	 * 
	 * @return
	 */
	public final double getSmoothingParameter(){
		return myAlpha;
	}
	
	/** Sets the smoothing parameter. 
	 * 
	 * @param alpha, must be in [0,1]
	 */
	public final void setSmoothingParameter(double alpha){
		if ((alpha < 0) || (alpha > 1))
			throw new IllegalArgumentException("The smoothing parameter must be in [0,1]");
		myAlpha = alpha;
		myMapParameterLabelToValue.put("Smoothing Constant Alpha", myAlpha);
	}

	/* (non-Javadoc)
	 * @see forecasting.Forecaster#getForecast(int)
	 */
	public double getForecast(int n) {
		return myNextPeriodForecast;
	}

	/* (non-Javadoc)
	 * @see forecasting.Forecaster#getForecasts(int)
	 */
	public double[] getForecasts(int n) {
		if (n < 1)	
			throw new IllegalArgumentException("The supplied n was < 1");
		
		double[] f = new double[n+1];
		f[0] = Double.NaN;
		for (int i=1;i<=n;i++)
			f[i] = myNextPeriodForecast;
		return f;
	}

	/** In the returned array element [0] is the smoothing
	 *  parameter
	 * 
	 */
	public final double[] getParameters() {
		double[] d = new double[1];
		d[0] = myAlpha;
		return d;
	}

	/** Element [0] in the supplied array is interpreted
	 *  as the smoothing parameter
	 * 
	 */
	public final void setParameters(double[] parameters) {
		if (parameters == null)
			throw new IllegalArgumentException("The parameters array was null");
		
		if (parameters.length == 0)
			throw new IllegalArgumentException("parameters length was 0");

		setSmoothingParameter(parameters[0]);
	}

	/** element [0] is the last observed value
	 * 
	 */
	public double[] getHistory() {
		double[] d = new double[1];
		d[0] = myXt;
		return d;
	}

	/* (non-Javadoc)
	 * @see forecasting.Forecaster#clearHistory()
	 */
	protected void clearHistory() {
		// there is nothing to clear for SES
	}

	/* (non-Javadoc)
	 * @see forecasting.Forecaster#updateForecast()
	 */
	protected void updateForecast(int i, double x) {
		// the new forecast is based on the current forecast and the new data
		myNextPeriodForecast = myAlpha*x + (1.0 - myAlpha)*myCurrentPeriodForecast;	
	}

	protected void updateHistory(int i, double x) {
       // myXt has already been captured
		
	}	
	
	protected void checkForSufficientHistory(){
		if (!isInitialized()){
			// not already initialized
			// need to check if enough data is available for initialization
			if (myTimeIndex == 1){
				// not initialized, and this is the first data point
				// make forecast for next period the same as first data point
				myNextPeriodForecast = myXt;
				// initialize it, so when 2nd point is collected
				// it is ready
				
				//TODO vijith added
				double [] hist = {myXt}; 
				double [] para = this.estimateInitialForecastParameter(hist);
				setInitialForecastParameter(para);	
				
				setInitFlag(true);
			}
		}
	}

	/** data[0] is interpreted as x(0)
	 *  forecasts[0] is interpreted as f(0)
	 *  
	 * 
	 */
	public void initializeHistory(double[] data, double[] forecasts,
			double[] parameters) {
		if (data == null)
			throw new IllegalArgumentException("The data array was null");
		/*if (forecasts == null)
			throw new IllegalArgumentException("The data array was null");*/
		
		myXt = data[data.length - 1];
		
		//TODO Vijith added this
		double [] p = {myXt};
		double [] para = this.estimateInitialForecastParameter(p);
		setInitialForecastParameter(para);
		
		//myCurrentPeriodForecast = forecasts[0]; // commented by Dr.Rossetti
		//myNextPeriodForecast = forecasts[forecasts.length - 1];
		myNextPeriodForecast = para[0];
		
		setInitFlag(true);
	}
	
	public double [] estimateInitialForecastParameter(double [] history){	
		double [] para = new double[1];
		
		//TODO ?
		//para [0] = history[history.length - 1];
		
		Statistic stat = new Statistic();
		
		stat.collect(history);
		
		para [0] =  stat.getAverage();
		
		return para;		
	}
	
	/**
	 * Creates a shallow copy of a Simple Exponential Smoothing forecast technique object, retaining only the 
	 * forecast parameters.
	 */
	public ForecastTechnique copyInstance(){
        return new SimpleExponentialSmoothing(this.myAlpha,this.getName());
    }

	/*@Override
	public ForecastTechnique copyInstance(double[] history) {
		//TODO I think there must be a better way to name.
		SimpleExponentialSmoothing a = new SimpleExponentialSmoothing(this.myAlpha,this.getName());
		
		double [] data = {history[history.length - 1]};
		double[] forecasts = {history[history.length - 1]};
		double[] parameters = new double[1];
		
		a.initializeHistory(data, forecasts, parameters);
		
		return a;
	}*/
}
