package jslx.forecasting;

/**
 * Implements the intermittent forecaster named syntetos boylan technique
 * @author vvarghe
 *
 */
public class SyntetosBoylan extends IntermittentForecastTechnique {
	
	private SimpleExponentialSmoothing myPtSES;
	private SimpleExponentialSmoothing myZtSES;
	
	public SyntetosBoylan() {
		this(0.1, 3, null);
	}

	public SyntetosBoylan(String name) {
		this(0.1, 3, name);
	}

	public SyntetosBoylan(double alpha) {
		this(alpha, 3, null);
	}	
	
	public SyntetosBoylan(double alpha, String name) {
		this(alpha, 3, name);
	}

	public SyntetosBoylan(double alpha, int reqdnzdem, String name) {
		super(alpha, reqdnzdem, name);
		myPtSES = new SimpleExponentialSmoothing(alpha, name+"Pt_SES");
		myZtSES = new SimpleExponentialSmoothing(alpha, name+"Zt_SES");
		double [] a = {alpha, alpha};
		setSmoothingParameter(a);
	}

	public SyntetosBoylan(double [] alpha, String name) {
		this(alpha, 3, name);	
	}
	
	public SyntetosBoylan(double[] alpha, int reqdnzdem, String name) {
		super(alpha, reqdnzdem, name);
		myPtSES = new SimpleExponentialSmoothing(alpha[0], name+"Pt_SES");
		myZtSES = new SimpleExponentialSmoothing(alpha[1], name+"Zt_SES");
		setSmoothingParameter(alpha);
	}
	
	/**
	 * Creates a shallow copy of a Simple Exponential Smoothing forecast technique object, retaining only the 
	 * forecast parameters.
	 */
	@Override
    public ForecastTechnique copyInstance(){
        return new SyntetosBoylan(this.myAlpha,this.myReqdNumNZDemandsBeforeInitialization,this.getName());
    }

	@Override
	protected void clearHistory() {
		myNumNZDemandsBeforeInitialization = 0;
		myPt = Double.NaN;		
		myZt = Double.NaN;		
		myQ = Double.NaN;		
		myData = new double[1];
		this.myPtSES.reset();
		this.myZtSES.reset();
	}
	
	@Override
	public final void setSmoothingParameter(double [] alpha){
		if ((alpha[0] < 0) || (alpha[0] > 1))
			throw new IllegalArgumentException("The smoothing parameter for Pt must be in [0,1]");
		if ((alpha[1] < 0) || (alpha[1] > 1))
			throw new IllegalArgumentException("The smoothing parameter for Zt must be in [0,1]");		
		myAlpha = alpha;
		myMapParameterLabelToValue.put("Smoothing Constant Time Between", myAlpha[0]);
		myMapParameterLabelToValue.put("Smoothing Constant Demand Size", myAlpha[1]);
        if(myPtSES==null)
            myPtSES = new SimpleExponentialSmoothing(alpha[0],getName()+"Pt_SES");
        if(myZtSES==null)
            myZtSES = new SimpleExponentialSmoothing(alpha[1],getName()+"Zt_SES");
        myPtSES.setSmoothingParameter(alpha[0]);
        myZtSES.setSmoothingParameter(alpha[1]);
	}

	public void initializeHistory(double[] data, double[] forecasts,
			double[] parameters) {
		
		
		//TODO How to handle the cases where the data is not sufficient to get estimates	
		double [] para =  estimateInitialForecastParameter(data);
		setInitialForecastParameter(para);
		myQ = para[0];
		myPt = para[1];
		double [] pt = {myPt};
		myPtSES.initializeHistory(pt, null, null);
        myZt = para[2];
        double [] zt = {myZt};
		myZtSES.initializeHistory(zt, null, null);
        estimateAverageDemand();
        setInitFlag(true);

	}

	@Override
	protected void estimateAverageDemand(){
		myNextPeriodForecast = (1-(myAlpha[0]/2))*(myZtSES.getNextPeriodForecast()/myPtSES.getNextPeriodForecast());
	}

	@Override
	protected void updateForecast(int i, double x) {
		if(myXt==0){
        	myQ++;
        }else{                
        	myPtSES.collect(myQ);
            myZtSES.collect(myXt);
            
            myPt=myPtSES.getNextPeriodForecast();
            myZt=myZtSES.getNextPeriodForecast();
            //System.out.println("Zt = "+myZt+", myPt = "+myPt);
            //myPt=myPt+myAlpha[0]*(myQ-myPt);
            //myZt=myZt+myAlpha[1]*(myXt-myZt);
            myQ=1;
        }
        //System.out.println("Zt = "+myZt+", myPt = "+myPt);
		estimateAverageDemand();
	}

	public final double[] getParameters() {
		double[] d = new double[2];
		d[0] = myAlpha[0];
		d[1] = myAlpha[1];
		return d;
	}

	public final void setParameters(double[] parameters) {
		if (parameters == null)
			throw new IllegalArgumentException("The parameters array was null");
		
		if (parameters.length == 0)
			throw new IllegalArgumentException("parameters length was 0");

		setSmoothingParameter(parameters);
	}
}
