/**
 * 
 */
package jslx.forecasting.demandgeneration;

import jsl.utilities.random.distributions.Binomial;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.random.GetValueIfc;
import jsl.utilities.random.distributions.Lognormal;
import jsl.utilities.random.distributions.NegativeBinomial;
import jsl.utilities.random.ParametersIfc;
import jsl.utilities.random.distributions.Poisson;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.ShiftedDistribution;
import jsl.utilities.random.rng.AR1CorrelatedRng;
import jsl.utilities.random.rng.RandomStreamIfc;
import jsl.utilities.statistic.Statistic;
import jslx.forecasting.DataSource;
import jslx.forecasting.IntermittentDemandStatistic;

/**
 * @author vvarghe
 *
 *The demand generation is based on Compound Poisson Process with occurence process N(t) and
 *the amount process A_i. The CPP, Y = Sum of A_i from i = 1 to N(t).
 *
 *The r.v A represents the non bucketed demand amount process and they are bucketed at a 
 *bucket size = t.
 *
 */
public class DemandGenerator extends DataSource implements ParametersIfc, GetValueIfc, RandomStreamIfc {
	
	/**
	 * Parameters
	 *
	 */
	private double myIntervalSize;
	private double myProbofZeroDemand;
	private double myDesiredBucketedMeanNZDemand;
	private double myDesiredBucketedVarianceNZDemand;
	
	 
	

	/**
	 * Distribution of the Interval between NZDemands
	 *
	 */
	private RandomIfc myDistIntervalBetweenNZDemands;
	
	/**
	 * Distribution of the NZDemand Amount
	 *
	 */
	private RandomIfc myDistNZDemandAmount;
			
	/**
	 * Bucket size
	 *
	 */	
	private double myDemandInterval;
	
	private double myCurrentTime;
	private double myPreviousTime;
	private double myNextDemandTime;

	private Statistic myNonBucketedNZDemand;
	private Statistic myNonBucketedDemandInterval;
	
	/** Compute the lambda of occur dist; mean and var of amount process
	 * i.e. Occurence process N(t)~Poisson(lambda*t)
	 * 
	 * The bucketed process that we desire is X(t)
	 * We are given the information of X(t)/X(t)>0
     * Let Y be r.v. X(t)/X(t)>0
     * E[Y] = E[N(t)]*E[A]*(1/P(X(t)>0))
     * E[Y^2]= (E[N(t)]*E[A^2] + E2[N(t)]*E2[A])*(1/P(X(t)>0))
     *  
     * @param interval bucket period
     * @param p0 probability of zero demand
     * @param phiT probability of zero demand
     * @param bucketed_nzmean E[Y]
     * @param bucketed_nzvar Var(Y))
     * 
     * From this we can find the lambda for dist of occurence process of N(t)
     * Also we can find the amount process A
     * 
     */
    public static double[] getCPP(double interval, double p0, double phiT, double bucketed_nzmean, double bucketed_nzvar){
		double lambda = -1 * (1/interval) * Math.log(p0);		
		double exp_numevents = lambda*interval; // E[N(t)]	
		double p1 = 1 - p0; // P(X(t)>0) = 1 - P(N(t)=0)		
		double nzmean = bucketed_nzmean * p1 * (1.0/exp_numevents); // E[A]	
		
		double nzvar = (bucketed_nzvar * p1 * (1/(lambda * interval))) 
					- (lambda * interval * nzmean * nzmean * (1 - (1.0/p1))) -(nzmean * nzmean);
		double [] cpp = {lambda, nzmean, nzvar};		
		return cpp;		
	}
	

	/** Constructs a DemandGenerator with an underlying CompoundPoissonProcess
	 * i.e. Occurence process N(t)~Poisson(lambda*t)
	 * 
	 * In this case, we assume Amount Process A~Binomial(n,p)
	 * 
	 * The bucketed process that we desire is X(t)
	 * We are given the information of X(t)/X(t)>0
     * Let Y be r.v. X(t)/X(t)>0
     * E[Y] = E[N(t)]*E[A]*(1/P(X(t)>0))
     * E[Y^2]= (E[N(t)]*E[A^2] + E2[N(t)]*E2[A])*(1/P(X(t)>0))
     *  
     * @param interval bucket period
     * @param p0 probability of zero demand
     * @param phiT probability of zero demand
     * @param bucketed_nzmean E[Y]
     * @param bucketed_nzvar Var(Y))
     * 
     * Here it has to be noted that a shifted binomial distribution is used
     * E[A'] = E[A] +1 where A' is shifted distribution shifted to the left by 1.0.
     * 
     * The computed E[A] is adjusted before setting the shifted distribution A'
     * 
     */
    public static DemandGenerator getBinomialAmountCPP(double interval, double p0, double phiT, double bucketed_nzmean, double bucketed_nzvar){
    	
    	double [] cpp = getCPP(interval, p0, phiT, bucketed_nzmean, bucketed_nzvar);
    	
		double lambda = cpp[0];
		AR1CorrelatedRng rng1 = new AR1CorrelatedRng(phiT);
		Exponential dist_interval_between_transactions = new Exponential(1.0/lambda, rng1);
		//System.out.print(dist_interval_between_transactions);
		//System.out.println(" -- Exponential("+dist_interval_between_transactions.getParameters()[0]+")");
		
		double nzmean = cpp[1] - 1.0; // E[A]
		
		double nzvar = cpp[2]; // Var(A)
		
		if(nzvar/nzmean >= 0.9)
			throw new IllegalArgumentException("Binomial distribution cannot be set");
		
		double a = (nzvar/(nzmean*nzmean)) - (1/nzmean);
		
		Binomial b = new Binomial();
		
		double p = 1 - (nzvar/nzmean);
		int n = (int)(Math.floor(nzmean/p)+1.0);
		b.setParameters(p, n);
		
		ShiftedDistribution distamount = new ShiftedDistribution(b, 1.0);
		
		/*if(a <= -1.0){
			distamount.setParameters(nzmean, 1);			
		}else if (a < 0){
			double k0 = Math.floor(a) + 1.0;
			int k = (int) k0;
			
			double q = 1 + (a*(1+k)) + Math.sqrt((-1*a*k*(1+k))-k);
			Bernoulli bn = new Bernoulli(q);
			
			double p = nzmean/(k+1-q);
			
			if(bn.getValue()==1.0){
				distamount.setParameters(p, k);
			}else{
				distamount.setParameters(p, k+1);				
			} throw new IllegalArgumentException("Binomial distribution cannot be set");		
			
		}*/
		
		//System.out.print(distamount);
		//System.out.println(" -- ShiftedBinomial("+distamount.getParameters()[1]+","+distamount.getParameters()[0]+" "+1.0+")");
		//System.out.println("**************************************************");
		
		DemandGenerator dg = new DemandGenerator(dist_interval_between_transactions, distamount, interval, "DG");
		
		return dg;	
		
	}
    
    /** Constructs a DemandGenerator with an underlying CompoundPoissonProcess
	 * i.e. Occurence process N(t)~Poisson(lambda*t)
	 * 
	 * In this case, we assume Amount Process A~Binomial(n,p)
	 * 
	 * The bucketed process that we desire is X(t)
	 * We are given the information of X(t)/X(t)>0
     * Let Y be r.v. X(t)/X(t)>0
     * E[Y] = E[N(t)]*E[A]*(1/P(X(t)>0))
     * E[Y^2]= (E[N(t)]*E[A^2] + E2[N(t)]*E2[A])*(1/P(X(t)>0))
     *  
     * @param interval bucket period
     * @param p0 probability of zero demand
     * @param phiT probability of zero demand
     * @param bucketed_nzmean E[Y]
     * @param bucketed_nzvar Var(Y))
     * 
     * Here it has to be noted that a shifted poisson distribution is used
     * E[A'] = E[A] +1 where A' is shifted distribution shifted to the left by 1.0.
     * 
     * The computed E[A] is adjusted before setting the shifted distribution A'
     * 
     */
    public static DemandGenerator getPoissonAmountCPP(double interval, double p0, double phiT, double bucketed_nzmean, double bucketed_nzvar){
    	
    	double [] cpp = getCPP(interval, p0, phiT, bucketed_nzmean, bucketed_nzvar);
    	
		double lambda = cpp[0];
		AR1CorrelatedRng rng1 = new AR1CorrelatedRng(phiT);
		Exponential dist_interval_between_transactions = new Exponential(1.0/lambda, rng1);
		System.out.print(dist_interval_between_transactions);
		System.out.println(" -- Exponential("+dist_interval_between_transactions.getParameters()[0]+")");
		
		double nzmean = cpp[1] - 1.0; // E[A]
		
		double nzvar = cpp[2]; // Var(A)
		
		if((nzvar/nzmean < 0.9) && (nzvar/nzmean > 1.1))
			throw new IllegalArgumentException("Poisson distribution cannot be set");
		
		Poisson p = new Poisson(nzmean);
		ShiftedDistribution distamount = new ShiftedDistribution(p, 1.0);
		
		System.out.print(distamount);
		System.out.println(" -- ShiftedPoisson("+p.getParameters()[0]+", "+1.0+")");
		System.out.println("**************************************************");	
		
		DemandGenerator dg = new DemandGenerator(dist_interval_between_transactions, distamount, interval, "DG");
		
		return dg;	
		
	}
    
    /** Constructs a DemandGenerator with an underlying CompoundPoissonProcess
	 * i.e. Occurence process N(t)~Poisson(lambda*t)
	 * 
	 * In this case, we assume Amount Process A~Binomial(n,p)
	 * 
	 * The bucketed process that we desire is X(t)
	 * We are given the information of X(t)/X(t)>0
     * Let Y be r.v. X(t)/X(t)>0
     * E[Y] = E[N(t)]*E[A]*(1/P(X(t)>0))
     * E[Y^2]= (E[N(t)]*E[A^2] + E2[N(t)]*E2[A])*(1/P(X(t)>0))
     *  
     * @param interval bucket period
     * @param p0 probability of zero demand
     * @param phiT probability of zero demand
     * @param bucketed_nzmean E[Y]
     * @param bucketed_nzvar Var(Y))
     * 
     * Here it has to be noted that a shifted negative binomial distribution is used
     * E[A'] = E[A] +1 where A' is shifted distribution shifted to the left by 1.0.
     * 
     * The computed E[A] is adjusted before setting the shifted distribution A'
     * 
     */
    public static DemandGenerator getNegativeBinomialAmountCPP(double interval, double p0, double phiT, double bucketed_nzmean, double bucketed_nzvar){
    	
    	double [] cpp = getCPP(interval, p0, phiT, bucketed_nzmean, bucketed_nzvar);
    	
		double lambda = cpp[0];
		AR1CorrelatedRng rng1 = new AR1CorrelatedRng(phiT);
		Exponential dist_interval_between_transactions = new Exponential(1.0/lambda, rng1);
		//System.out.print(dist_interval_between_transactions);
		//System.out.println("  -- Exponential("+dist_interval_between_transactions.getParameters()[0]+")");
		
		double nzmean = cpp[1] - 1.0; // E[A]
		
		double nzvar = cpp[2]; // Var(A)
		
		if(nzvar/nzmean <= 1.1)
			throw new IllegalArgumentException("NegativeBinomial distribution cannot be set");
		
		double p = nzmean/nzvar;
		
		int r = (int)(Math.floor(nzmean*p*(1/(1-p))) + 1.0);
		
		NegativeBinomial n = new NegativeBinomial(p, r);
		ShiftedDistribution distamount = new ShiftedDistribution(n, 1.0);
		
		//System.out.print(distamount);	
		//System.out.println(" -- ShiftedNegativeBinomial("+distamount.getParameters()[1]+","+distamount.getParameters()[0]+", "+1.0+"");
		//System.out.println("**************************************************");
		
		DemandGenerator dg = new DemandGenerator(dist_interval_between_transactions, distamount, interval, "DG");
		
		return dg;	
		
	}    
    
    /** This chooses which distribution is to be used based on the computed coefficient of variation
     * 
     */
    public static DemandGenerator getAxsaterBasedAmountDistCPP(double interval, double p0, double phiT, double bucketed_nzmean, double bucketed_nzvar){
    	
    	double [] cpp = getCPP(interval, p0, phiT, bucketed_nzmean, bucketed_nzvar);
		
		double nzmean = cpp[1] - 1.0; // E[A]
		
		double nzvar = cpp[2]; // Var(A)
		
		DemandGenerator dg;
		
		if(nzvar/nzmean < 0.9){
			dg = getBinomialAmountCPP(interval, p0, phiT, bucketed_nzmean, bucketed_nzvar);			
		}else if (nzvar/nzmean <= 1.1){
			dg = getPoissonAmountCPP(interval, p0, phiT, bucketed_nzmean, bucketed_nzvar);			
		}else {
			dg = getNegativeBinomialAmountCPP(interval, p0, phiT, bucketed_nzmean, bucketed_nzvar);			
		}
		
		return dg;	
		
	}
    
    /** Constructs a DemandGenerator with an underlying CompoundPoissonProcess
	 * i.e. Occurence process N(t)~Poisson(lambda*t)
	 * 
	 * In this case, we assume Amount Process A~Binomial(n,p)
	 * 
	 * The bucketed process that we desire is X(t)
	 * We are given the information of X(t)/X(t)>0
     * Let Y be r.v. X(t)/X(t)>0
     * E[Y] = E[N(t)]*E[A]*(1/P(X(t)>0))
     * E[Y^2]= (E[N(t)]*E[A^2] + E2[N(t)]*E2[A])*(1/P(X(t)>0))
     *  
     * @param interval bucket period
     * @param p0 probability of zero demand
     * @param phiT probability of zero demand
     * @param bucketed_nzmean E[Y]
     * @param bucketed_nzvar Var(Y))
     * 
     * Here it has to be noted that a shifted lognormal distribution is used
     * E[A'] = E[A] +1 where A' is shifted distribution shifted to the left by 1.0.
     * 
     * The computed E[A] is adjusted before setting the shifted distribution A'
     * 
     */
    public static DemandGenerator getLogNormalAmountCPP(double interval, double p0, double phiT, double bucketed_nzmean, double bucketed_nzvar){

    	
    	double [] cpp = getCPP(interval, p0, phiT, bucketed_nzmean, bucketed_nzvar);
    	
		double lambda = cpp[0];
		AR1CorrelatedRng rng1 = new AR1CorrelatedRng(phiT);
		Exponential dist_interval_between_transactions = new Exponential(1.0/lambda, rng1);
		//System.out.println(dist_interval_between_transactions);
		
		double nzmean = cpp[1] - 1.0; // E[A]
		
		double nzvar = cpp[2]; // Var(A)
		
		double para_mu = Math.log(nzmean) - 0.5 * Math.log(1 + (nzvar/(nzmean*nzmean)));
		double para_sigma2 = Math.log(1 + (nzvar/(nzmean*nzmean)));
		
		Lognormal l = new Lognormal (para_mu, para_sigma2);
		
		ShiftedDistribution distamount = new ShiftedDistribution(l, 1.0);
		
		//System.out.println(distamount);	
		
		DemandGenerator dg = new DemandGenerator(dist_interval_between_transactions, distamount, interval, "DG");
		
		return dg;	
		
	}
	
	/**
	 * 
	 */
	public DemandGenerator() {
		this(new Exponential(), new Constant(2.0), 1.0, null);
	}

	/**
	 * @param name
	 */
	public DemandGenerator(RandomIfc distInterval, RandomIfc distAmount, double interval, String name) {
		super(name);
		setDistribution(distInterval, distAmount);
		myNonBucketedNZDemand = new Statistic("NonBucketedNZDemand");
		myNonBucketedDemandInterval = new Statistic("NonBucketedDemandInterval");
		myCurrentTime = 0.0;
		myPreviousTime = 0.0 - interval;
		myDemandInterval = interval;
		myNextDemandTime = distInterval.getValue();	
		myNonBucketedDemandInterval.collect(myNextDemandTime);
		
	}
	
	public void setDistribution(RandomIfc interval, RandomIfc amt){
		myDistIntervalBetweenNZDemands = interval;
		myDistNZDemandAmount = amt;
	}



    /** 
     *  
     * @return The value.
     */
    public double getValue(){
    	myCurrentTime = myCurrentTime + myDemandInterval;
    	myPreviousTime = myPreviousTime + myDemandInterval;
    	
    	double x = 0.0;
    	
    	while(myNextDemandTime <= myCurrentTime){
    		double y = myDistNZDemandAmount.getValue();
    		if(y > Math.floor(y)){
    			y = Math.floor(y)+1.0;    			
    		}
    		myNonBucketedNZDemand.collect(y);
    		x = x + y;
    		double intvl = myDistIntervalBetweenNZDemands.getValue();
    		this.myNonBucketedDemandInterval.collect(intvl);
    		myNextDemandTime = myNextDemandTime + intvl;    		
    	}    	
    	setValue(x);    	
        return x;
    }
	
	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomIfc#getSample(int)
	 */
	public double[] getSample(int sampleSize) {		
		double [] sample = new double[sampleSize];		
		for(int i = 0; i<sampleSize; i++){
			sample[i] = getValue();			
		}
		return sample;
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomIfc#getSample(double[])
	 */
	public void getSample(double[] values) {		
		for(int i = 0; i<values.length; i++){
			values[i] = getValue();			
		}
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.ParametersIfc#getParameters()
	 */
	public double[] getParameters() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.ParametersIfc#setParameters(double[])
	 */
	public void setParameters(double[] parameters) {
		// TODO Auto-generated method stub

	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomStreamIfc#resetNextSubstream()
	 */
	public void resetNextSubstream() {
		this.myDistIntervalBetweenNZDemands.resetNextSubstream();
		this.myDistNZDemandAmount.resetNextSubstream();

	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomStreamIfc#resetStartStream()
	 */
	public void resetStartStream() {
		this.myDistIntervalBetweenNZDemands.resetStartStream();
		this.myDistNZDemandAmount.resetStartStream();

	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomStreamIfc#resetStartSubstream()
	 */
	public void resetStartSubstream() {
		this.myDistIntervalBetweenNZDemands.resetStartSubstream();
		this.myDistNZDemandAmount.resetStartSubstream();

	}

	/* (non-Javadoc)
	 * @see jsl.utilities.random.RandomStreamIfc#setAntithetic(boolean)
	 */
	public void setAntithetic(boolean flag) {
		this.myDistIntervalBetweenNZDemands.setAntithetic(flag);
		this.myDistNZDemandAmount.setAntithetic(flag);

	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		/*double ipBucketSize = 1.0;
        double ipProbZDemandBucketed = 0.1353;
        double ipL1CorrCoeffIntvlNonBucketed = 0.0;
        double ipMeanNZDemandNonBucketed = 6.9391;
        double ipVariNZDemandNonBucketed = 18.9268;*/
		
		double ipBucketSize = 1.0;
        double ipProbZDemandBucketed = 0.1353;
        double ipL1CorrCoeffIntvlNonBucketed = -0.8;
        double ipMeanNZDemandNonBucketed = 6.9391;
        double ipVariNZDemandNonBucketed = 18.9268;
        
        int myNumReplication = 100;
        int myReplicationLength = 1000;
    	
        Statistic statMeanDemandBucketed = new Statistic("MeanDemandBucketed");
        Statistic statVariDemandBucketed = new Statistic("VariDemandBucketed");
        Statistic statL1CorrCoeffDemandBucketed = new Statistic("L1CorrCoeffDemandBucketed");	            
        Statistic statMeanNZDemandBucketed = new Statistic("MeanNZDemandBucketed");
        Statistic statVariNZDemandBucketed = new Statistic("VariNZDemandBucketed");
        Statistic statL1CorrCoeffNZDemandBucketed = new Statistic("L1CorrCoeffNZDemandBucketed");	            
        Statistic statMeanIntvlBucketed = new Statistic("MeanIntvlBucketed");
        Statistic statVariIntvlBucketed = new Statistic("VariIntvlBucketed");
        Statistic statL1CorrCoeffIntvlBucketed = new Statistic("L1CorrCoeffIntvlBucketed");	            
        Statistic statProbZDemandBucketed = new Statistic("ProbZDemandBucketed");	            
        Statistic statMeanNZDemandNonBucketed = new Statistic("MeanNZDemandNonBucketed");
        Statistic statVariNZDemandNonBucketed = new Statistic("VariNZDemandNonBucketed");
        Statistic statL1CorrCoeffNZDemandNonBucketed = new Statistic("L1CorrCoeffNZDemandNonBucketed");	            
        Statistic statMeanIntvlNonBucketed = new Statistic("MeanIntvlNonBucketed");
        Statistic statVariIntvlNonBucketed = new Statistic("VariIntvlNonBucketed");
        Statistic statL1CorrCoeffIntvlNonBucketed = new Statistic("L1CorrCoeffIntvlNonBucketed");            
        
                    

        double [] cpp = DemandGenerator.getCPP(ipBucketSize, ipProbZDemandBucketed, ipL1CorrCoeffIntvlNonBucketed, ipMeanNZDemandNonBucketed, ipVariNZDemandNonBucketed);             
        System.out.println("IP ProbZDemandBucketed: "+ipProbZDemandBucketed);
        System.out.println("IP L1CorrCoeffIntvlNonBucketed: "+ipL1CorrCoeffIntvlNonBucketed);
        System.out.println("IP MeanNZDemandBucketed: "+ipMeanNZDemandNonBucketed); 
        System.out.println("IP VariNZDemandBucketed: "+ipVariNZDemandNonBucketed);       
        System.out.println("IP -> Lambda: "+cpp[0]);
        System.out.println("IP -> MeanNZDemandNonBucketed: "+cpp[1]);
        System.out.println("IP -> VariNZDemandNonBucketed: "+cpp[2]);     
        
        for(int j = 1; j <= myNumReplication; j++){
        	
        	System.out.println("Replication: "+j);
        	DemandGenerator dg = getAxsaterBasedAmountDistCPP(ipBucketSize, ipProbZDemandBucketed, ipL1CorrCoeffIntvlNonBucketed, ipMeanNZDemandNonBucketed, ipVariNZDemandNonBucketed);
            IntermittentDemandStatistic stat = new IntermittentDemandStatistic();
            
            for(int k = 1; k <= myReplicationLength; k++){
                    double x = dg.getValue();
                    stat.collect(x);
            }
            
            double myMeanDemandBucketed = stat.getDemandStat().getAverage();
            double myVariDemandBucketed = stat.getDemandStat().getVariance();
            double myL1CorrCoeffDemandBucketed = stat.getDemandStat().getLag1Correlation();                
            double myMeanNZDemandBucketed = stat.getNonZeroDemandStat().getAverage();
            double myVariNZDemandBucketed = stat.getNonZeroDemandStat().getVariance();
            double myL1CorrCoeffNZDemandBucketed = stat.getNonZeroDemandStat().getLag1Correlation();                
            double myProbZDemandBucketed = stat.getProbFailure();                            
            double myMeanIntvlBucketed = stat.getIntervalBetweenNonZeroDemandStat().getAverage();
            double myVariIntvlBucketed = stat.getIntervalBetweenNonZeroDemandStat().getVariance();
            double myL1CorrCoeffIntvlBucketed = stat.getIntervalBetweenNonZeroDemandStat().getLag1Correlation(); 
            double myMeanNZDemandNonBucketed = dg.myNonBucketedNZDemand.getAverage();
            double myVariNZDemandNonBucketed = dg.myNonBucketedNZDemand.getVariance();
            double myL1CorrCoeffNZDemandNonBucketed = dg.myNonBucketedNZDemand.getLag1Correlation();                
            double myMeanIntvlNonBucketed = dg.myNonBucketedDemandInterval.getAverage();
            double myVariIntvlNonBucketed = dg.myNonBucketedDemandInterval.getVariance();            
            double myL1CorrCoeffIntvlNonBucketed = dg.myNonBucketedDemandInterval.getLag1Correlation();
            
            
            statMeanDemandBucketed.collect(myMeanDemandBucketed);
            statVariDemandBucketed.collect(myVariDemandBucketed);
            statL1CorrCoeffDemandBucketed.collect(myL1CorrCoeffDemandBucketed);                
            statMeanNZDemandBucketed.collect(myMeanNZDemandBucketed);
            statVariNZDemandBucketed.collect(myVariNZDemandBucketed);
            statL1CorrCoeffNZDemandBucketed.collect(myL1CorrCoeffNZDemandBucketed);                
            statMeanIntvlBucketed.collect(myMeanIntvlBucketed);
            statVariIntvlBucketed.collect(myVariIntvlBucketed);
            statL1CorrCoeffIntvlBucketed.collect(myL1CorrCoeffIntvlBucketed);                
            statProbZDemandBucketed.collect(myProbZDemandBucketed);                
            statMeanNZDemandNonBucketed.collect(myMeanNZDemandNonBucketed);
            statVariNZDemandNonBucketed.collect(myVariNZDemandNonBucketed);
            statL1CorrCoeffNZDemandNonBucketed.collect(myL1CorrCoeffNZDemandNonBucketed);                
            statMeanIntvlNonBucketed.collect(myMeanIntvlNonBucketed);
            statVariIntvlNonBucketed.collect(myVariIntvlNonBucketed);
            statL1CorrCoeffIntvlNonBucketed.collect(myL1CorrCoeffIntvlNonBucketed);
            
        } 
        
        System.out.println(statMeanDemandBucketed);
        System.out.println(statVariDemandBucketed);
        System.out.println(statL1CorrCoeffDemandBucketed);	            
        System.out.println(statMeanNZDemandBucketed);
        System.out.println(statVariNZDemandBucketed);
        System.out.println(statL1CorrCoeffNZDemandBucketed);	            
        System.out.println(statMeanIntvlBucketed);
        System.out.println(statVariIntvlBucketed);
        System.out.println(statL1CorrCoeffIntvlBucketed);	            
        System.out.println(statProbZDemandBucketed);	            
        System.out.println(statMeanNZDemandNonBucketed);
        System.out.println(statVariNZDemandNonBucketed);
        System.out.println(statL1CorrCoeffNZDemandNonBucketed);            
        System.out.println(statMeanIntvlNonBucketed);
        System.out.println(statVariIntvlNonBucketed);
        System.out.println(statL1CorrCoeffIntvlNonBucketed);
        
	}


	public double getProbofZeroDemand() {
		return myProbofZeroDemand;
	}


	public double getNonBucketedNZDemandMean() {
		return myNonBucketedNZDemand.getAverage();
	}
	public double getNonBucketedNZDemandVariance() {
		return myNonBucketedNZDemand.getVariance();
	}

	public double getNonBucketedNZDemandLag1Correlation() {
		return myNonBucketedNZDemand.getLag1Correlation();
	}

	public double getNonBucketedDemandIntervalMean() {
		return myNonBucketedNZDemand.getAverage();
	}
	public double getNonBucketedDemandIntervalVariance() {
		return myNonBucketedNZDemand.getVariance();
	}

	public double getNonBucketedDemandIntervalLag1Correlation() {
		return myNonBucketedNZDemand.getLag1Correlation();
	}
}
