package jslx.forecasting.metrics;

import jsl.utilities.statistic.Statistic;
import jslx.forecasting.ForecastTechnique;
import jslx.forecasting.metrics.calc.StatCalculatorIfc;

/**An abstract class for forecasting metrics that are based on forecast error.
 * 
 * x_t= observation for time period t
 * f_t= forecast for time period t
 * e_t= forecast error for time period t
 * 
 * e_t= f_t-x_t
 * 
 * @author hmedal
 *
 */
public class ErrorBasedMetric extends ForecastMetric{
	
	/**
	 * Computes the metric value given the error
	 */
	protected MathFunctionIfc errorFunction;
	
	/**
	 * An object that calculates the forecast metric value for the collected data given a statistic.
	 */
	protected StatCalculatorIfc calculator;
	
	/**
	 * The statistic object that keeps track of statistics on forecast metric values for each period
	 */
	protected Statistic stat;
	
	/**
	 * Creates a no named forecast metric based on forecast error with a user specified metric calculator and 
	 * error function.
	 * 
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public ErrorBasedMetric(StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		this(null,calculator,errorFunction);
	}
	
	/**
	 * Creates a named forecast metric based on forecast error with a user specified metric calculator and 
	 * error function.
	 * 
	 * @param name the name of the forecast metric
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public ErrorBasedMetric(String name,StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		super(name);
		stat=new Statistic(name+" statistic");
		if(calculator.requiresSaved())
			stat.setSaveDataOption(true);
		this.calculator=calculator;
		this.errorFunction=errorFunction;
		setName(name);
	}
	
	/**Collects the error for a single forecast.
	* 
	 * @param ft the forecast technique used to compute the metric value
	 */
	@Override
	public void collectForecastDataForPeriod(ForecastTechnique ft) {
		double error=ft.getError();
		ft.getPrevObservation();
		if(!Double.isNaN(error))
			collectMetricForPeriod(error);
	}
	
	/** Sets the name 
	* @param str The name as a string.
	*/
	public void setName(String str) {
		if (str == null){
			myName= calculator.getName()+errorFunction.getName()+"Error";
		}
		else
			myName = str;
	}
	
	/**
	 * Computes the metric value using the StatCalculatorIfc object.
	 */
	@Override
	public void computeMetricValue() {
		metricValue=calculator.computeMetricValue(stat);
	}
	
	/** Calculates the metric value for a particular period, given the forecast error for that period.
	 * 
	 * @param error the error for a period
	 * @return the metric value
	 */
	public double collectMetricForPeriod(double error){
		double errorFunctionvalue=errorFunction.getValue(error);
		stat.collect(errorFunctionvalue);
		return errorFunctionvalue;
	}

	@Override
	public String getAbbreviatedName() {
		return calculator.getAbbreviatedName()+errorFunction.getAbbreviatedName()+"E";
	}
        
        @Override
        public ForecastMetric copyInstance(){
            return new ErrorBasedMetric(this.calculator,this.errorFunction);
        }

    @Override
    public void reset() {
        stat.reset();
        metricValue=0;
    }
        
        
}
