package jslx.forecasting.metrics;

import jsl.utilities.IdentityIfc;
import jslx.forecasting.CumulativeAverage;
import jslx.forecasting.ForecastTechnique;
import jslx.forecasting.NaiveForecastTechnique;
import jslx.forecasting.metrics.calc.GeometricMean;
import jslx.forecasting.metrics.calc.Mean;
import jslx.forecasting.metrics.calc.Median;

/**An abstract class for a forecasting metric, or a number that represents the performance of a forecasting
 * technique.
 * 
 * @author hmedal
 *
 */
public abstract class ForecastMetric implements IdentityIfc{
	
	/** A counter to count the number of created to assign "unique" ids
	 */
	protected static long myIdCounter_;
	
	/** The id of this object
	 */
	private long myId;

   /** Holds the name of the technique for reporting purposes.
     */
    protected String myName;
    
	/** The value of the metric for the entire dataset
	*/
	protected double metricValue;
	
	/** Creates a no name ForecastMetric
	 * 
	 */
	public ForecastMetric() {
            this(null);
	}
	
	/** Creates a named ForecastMetric
	 * 
	 */
	public ForecastMetric(String name) {
        setId();
	}

    //  ===========================================
    //        PUBLIC INTERFACE
    //  ===========================================
	
	/** Compute the metric value for all of the data collected before the moment that the method is called.	 
	*/
	public abstract void computeMetricValue();
	
	/** Collects the data needed for the metric (e.g. error, relative error, etc.) for a particular time period.
	 * Called from the update method.  Should not be called directly.
	 * 
	 * @param ft the ForecastingTechnique that generates forecasts
	 */
	public abstract void collectForecastDataForPeriod(ForecastTechnique ft);
	
	/**
	 * 
	 * @return the value of the forecast metric for the entire dataset
	 */
	public double getMetricValue(){
		return metricValue;
	}
	
	public long getId() {
		return myId;
	}
	
	/** Gets the name.
	* @return The name of object.
	*/
	public String getName() {
		return myName;
	}

	/** Sets the name 
	* @param str The name as a string.
	*/
	public void setName(String str) {
		if (str == null){
			myName= this.getClass().getSimpleName();
		}
		else
			myName = str;
	}
	
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append("ID " + getId() + "\n");
		sb.append("Name " + getName() + "\n");
		sb.append("Value "+ getMetricValue());
		return(sb.toString());
	}
	
	/** A method to get the abbreviated name of the forecast metric.  For example, "MAPE" is the abbreviated
	 * name of "Mean Absolute Percentage Error".
	 * 
	 * @return a string representing the abbreviated name of the metric
	 */
	public abstract String getAbbreviatedName();
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
   
	protected void setId(){
		myIdCounter_ = myIdCounter_ + 1;
		myId = myIdCounter_;
	}
	
	public static ForecastMetric createMeanSquareError(){
		return new ErrorBasedMetric(new Mean(),new Square());
	}
	
	public static ForecastMetric createMeanAbsoluteError(){
		return new ErrorBasedMetric(new Mean(),new Absolute());
	}
	
	public static ForecastMetric createMeanAbsolutePercentageError(){
		return new PercentageErrorBasedMetric(new Mean(),new Absolute());
	}
	
	public static ForecastMetric createMeanUAPE(){
		return new UnbiasedPercentageErrorBasedMetric(new Mean(),new Absolute());
	}
	
	public static ForecastMetric createMedianUAPE(){
		return new UnbiasedPercentageErrorBasedMetric(new Median(),new Absolute());
	}
	
	public static ForecastMetric createMeanRAE(){
		return new RelativeErrorBasedMetric (new Mean(), new Absolute());
	}
	
	public static ForecastMetric createMeanRAE(ForecastTechnique ft){
		return new RelativeErrorBasedMetric (ft, new Mean(), new Absolute());
	}
	
	public static ForecastMetric createRelMeanAE(){
		return new RelativeForecastMetric (ForecastMetric.createMeanAbsoluteError());
	}
	
	public static ForecastMetric createRelMeanAE(ForecastTechnique ft){
		return new RelativeForecastMetric (ForecastMetric.createMeanAbsoluteError(), ft);
	}
	
	public static ForecastMetric createMeanASE(){
		return new ScaledErrorBasedMetric (new Mean(), new Absolute());
	}
	
	public static void main(String[] args) {
		//ForecastMetric f=new AbsoluteError();
		//System.out.println(f.getName());
	}
        
    public abstract ForecastMetric copyInstance();

    public abstract void reset();
}
