package jslx.forecasting.metrics;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import jslx.forecasting.CumulativeAverage;
import jslx.forecasting.DataSourceArray;
import jslx.forecasting.ForecastTechnique;
import jslx.forecasting.metrics.calc.GeometricMean;
import jslx.forecasting.metrics.calc.Mean;
import jslx.forecasting.metrics.calc.Median;
import jslx.forecasting.metrics.calc.RootMean;
import jslx.forecasting.metrics.calc.RootMedian;
import jslx.forecasting.metrics.calc.StatCalculatorIfc;

public class ForecastMetricTester {
	
	private static final double ERROR_TOLERANCE_DEFAULT=.001;
	private double errorTolerance;
	private Map<String,Double> metricValues;
	private double[] observations;
	
	public ForecastMetricTester(String dataFileName){
		metricValues=new HashMap<String,Double>();
		readData1();
		errorTolerance=ERROR_TOLERANCE_DEFAULT;			
	}
	
	void readData1(){
		observations=new double[]{0,0,1,0,0,1,1,0,0,0,3,0,0,0,0,0,0,0,0,0,0,2};
		metricValues.put("GeometricMeanRelativeAbsoluteError",	0.554064993);
		metricValues.put("MeanAbsoluteError",	0.600658043);
		metricValues.put("MeanAbsolutePercentageError",	84.47619048);
		metricValues.put("MeanRelativeAbsoluteError",	0.642980829);
		metricValues.put("MeanSquareError",	0.694879957);
		metricValues.put("MeanUnbiasedAbsolutePercentageError",	187.3484848);
		metricValues.put("MedianAbsoluteError",	0.4);
		metricValues.put("MedianAbsolutePercentageError",	85.71428571);
		metricValues.put("MedianRelativeAbsoluteError",	0.8);
		metricValues.put("MedianUnbiasedAbsolutePercentageError",	200.0);
		metricValues.put("RootMeanSquareError",	0.8335946);
		metricValues.put("RootMeanSquarePercentageError",	85.19552009);
		metricValues.put("RootMedianSquarePercentageError",	85.71428571);
		metricValues.put("SymmetricMeanAbsolutePercentageError",	187.3484848);
		metricValues.put("SymmetricMedianAbsolutePercentageError",	200.0);
		metricValues.put("TheilU",	0.507374574);


	}
	
	public boolean containsMetric(String metricName){
		//System.out.println(metricName);
		//System.out.println(metricValues.keySet());
		metricName=metricName.toLowerCase();
		//System.out.println("mv "+metricValues);
		return metricValues.containsKey(metricName);
	}
	
	public boolean validateMetricValue(ForecastMetric metric){
		String metricName=metric.getName();
		metricName=metricName.toLowerCase();
		if(!containsMetric(metricName))
			throw new IllegalArgumentException("metric with name '"+metricName+"' not found in file");
		double value=metric.getMetricValue();
		double testValue=metricValues.get(metricName);
		//System.out.println(value+"\t"+testValue);
		if(Math.abs(value-testValue)<=errorTolerance)
			return true;
		else
			return false;
	}
	
	public double getMetricValue(String name){
		if(!containsMetric(name))
			throw new IllegalArgumentException("metric with name '"+name+"' not found in file");
		return metricValues.get(name);
	}
	
	public void computeAllMetricValues(ForecastTechnique ft,ArrayList<ForecastMetric> metrics){
		DataSourceArray data=new DataSourceArray(observations);
		data.addObserver(ft);
		for(ForecastMetric metric:metrics)
			ft.attachMetric(metric);
		data.iterateAll();
		for(ForecastMetric metric:metrics)
			metric.computeMetricValue();
	}
	
	private  ArrayList<ForecastMetric> createAllMetrics(){
		StatCalculatorIfc[] calculators={new Mean(),new GeometricMean(),new Median(),new RootMean(),
				new RootMedian()
		};
		MathFunctionIfc[] mathFunctions={new Absolute(),new Square()};
		ArrayList<ForecastMetric> metrics=new ArrayList<ForecastMetric>();
		for(StatCalculatorIfc stat:calculators){
			for(MathFunctionIfc mf:mathFunctions){
				try {
					StatCalculatorIfc newStat=stat.getClass().newInstance();
					MathFunctionIfc math=mf.getClass().newInstance();
					metrics.add(new ErrorBasedMetric(newStat,math));
					metrics.add(new PercentageErrorBasedMetric(newStat,math));
					metrics.add(new SymmetricPercentageErrorBasedMetric(newStat,math));
					metrics.add(new UnbiasedPercentageErrorBasedMetric(newStat,math));
					metrics.add(new RelativeErrorBasedMetric(newStat,math));
				} catch (InstantiationException e) {
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				}
				
			}
		}
		metrics.add(new TheilU());
		return metrics;
	}
	
	public static void main(String[] args) {
		String filename="C:/Documents and Settings/hmedal/My Documents/metricTestingCumAvg.txt";
		ForecastMetricTester tester=new ForecastMetricTester(filename);
		ArrayList<ForecastMetric> metrics=tester.createAllMetrics();
		CumulativeAverage ca=new CumulativeAverage();
		tester.computeAllMetricValues(ca,metrics);
		for(ForecastMetric metric:metrics){
			String name=metric.getName();
			//System.out.println(name);
			if(tester.containsMetric(name)){
				boolean validated=tester.validateMetricValue(metric);
				System.out.println(metric.getName()+"\t"+metric.getMetricValue()+"\t"+validated);
			}
		}
	}
}
