package jslx.forecasting.metrics;

import jsl.utilities.statistic.Statistic;
import jslx.forecasting.ForecastTechnique;
import jslx.forecasting.metrics.calc.StatCalculatorIfc;

/**An abstract class for forecast metrics based on percentage error.
 * 
 * @author hmedal
 *
 */
public class PercentageErrorBasedMetric extends ForecastMetric {
	
	/**
	 * Computes the metric value given the error
	 */
	protected MathFunctionIfc errorFunction;
	
	/**
	 * An object that calculates the forecast metric value for the collected data given a statistic.
	 */
	protected StatCalculatorIfc calculator;
	
	/**
	 * The statistic object that keeps track of statistics on forecast metric values for each period
	 * 
	 */
	protected Statistic stat;
	
        /**
	 * Creates a no named forecast metric based on forecast percentage error
	 *
	 */
	public PercentageErrorBasedMetric(){
		this(null,null,null);
	}
        
	/**
	 * Creates a no named forecast metric based on forecast percentage error
	 *
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public PercentageErrorBasedMetric(StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		this(null,calculator,errorFunction);
	}
	
	/**
	 * Creates a named forecast metric based on forecast percentage error
	 * 
	 * @param name the name of the forecast metric
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public PercentageErrorBasedMetric(String name,StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		super(name);
		stat=new Statistic(name+" statistic");
		this.calculator=calculator;
		this.errorFunction=errorFunction;
		setName(name);
		if(calculator.requiresSaved())
			stat.setSaveDataOption(true);
	}
	
	@Override
	public void collectForecastDataForPeriod(ForecastTechnique ft) {
		double error=ft.getError();
		if(!Double.isNaN(error)){
			double observation=ft.getObservation();
			if(observation!=0){
				double percentageError=100*(error/observation);
				collectMetricForPeriod(percentageError);
			}
		}
	}
	
	/**
	 * Computes the metric value using the StatCalculatorIfc object.
	 */
	@Override
	public void computeMetricValue() {
		metricValue=calculator.computeMetricValue(stat);
	}
	
	/**
	 * Given the percentage error for a period, calculates the metric value for a period.
	 * 
	 * @param percentageError the percentage error for a particular period
	 */
	public double collectMetricForPeriod(double percentageError){
		double errorFunctionvalue=errorFunction.getValue(percentageError);
		stat.collect(errorFunctionvalue);
		return errorFunctionvalue;
	}

	@Override
	public String getAbbreviatedName() {
		return calculator.getAbbreviatedName()+errorFunction.getAbbreviatedName()+"PE";
	}
	
	/** Sets the name 
	* @param str The name as a string.
	*/
	public void setName(String str) {
		if (str == null){
			myName= calculator.getName()+errorFunction.getName()+"PercentageError";
		}
		else
			myName = str;
	}
        
        @Override
        public ForecastMetric copyInstance(){
            return new PercentageErrorBasedMetric(this.calculator,this.errorFunction);
        }
        
        @Override
    public void reset() {
        stat.reset();
        metricValue=0;
    }
}
