package jslx.forecasting.metrics;

import jsl.utilities.statistic.Statistic;
import jslx.forecasting.ForecastTechnique;
import jslx.forecasting.NaiveForecastTechnique;
import jslx.forecasting.metrics.calc.StatCalculatorIfc;

/** An abstract class for forecast techniques based on error relative to another benchmark technique.
 * Allows the user to specify the benchmark technique.  The default is Naive1 (random walk).  
 * See Hyndman and Koehler 2005 for more details
 * 
 * @author hmedal
 *
 */
public class RelativeErrorBasedMetric extends ForecastMetric {
	
	/**
	 * Computes the metric value given the error
	 */
	protected MathFunctionIfc errorFunction;
	
	/**
	 * An object that calculates the forecast metric value for the collected data given a statistic.
	 */
	protected StatCalculatorIfc calculator;
	
	/** The technique to compare against
	 * 
	 */
	protected ForecastTechnique benchmarkTechnique;
	
	/**
	 * The statistic object that keeps track of statistics on forecast metric values for each period
	 */
	protected Statistic stat;
	
	/**
	 * Constructs a no named relative error based forecast metric with random walk as the benchmark technique.
	 * 
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public RelativeErrorBasedMetric(StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		this(null,new NaiveForecastTechnique(),calculator,errorFunction);
	}
	
	/**
	 * Constructs a named relative error based forecast metric with random walk as the benchmark technique.
	 * 
	 * @param name the name of the forecast metric
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public RelativeErrorBasedMetric(String name,StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		this(name,new NaiveForecastTechnique(),calculator,errorFunction);
	}
	
	/**
	 * Constructs a no named relative error based forecast metric with user specified benchmark technique and 
	 * calculator.
	 * 
	 * @param benchmarkTechnique the technique to compare against
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public RelativeErrorBasedMetric(ForecastTechnique benchmarkTechnique,StatCalculatorIfc calculator,
			MathFunctionIfc errorFunction){
		this(null,benchmarkTechnique,calculator,errorFunction);
	}
	
	/**
	 * Constructs a named relative error based forecast metric with user specified benchmark technique and 
	 * calculator.
	 * 
	 * @param name the name of the forecast metric
	 * @param benchmarkTechnique the technique to compare against
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public RelativeErrorBasedMetric(String name,ForecastTechnique benchmarkTechnique,StatCalculatorIfc calculator,
			MathFunctionIfc errorFunction) {
		super(name);
		this.benchmarkTechnique=benchmarkTechnique;
		stat=new Statistic();
		this.calculator=calculator;
		this.errorFunction=errorFunction;
		setName(name);
		if(calculator.requiresSaved())
			stat.setSaveDataOption(true);
	}
	
	/** Sets the name 
	* @param str The name as a string.
	*/
	public void setName(String str) {
		if (str == null){
			myName=calculator.getName()+"Relative"+errorFunction.getName()+"Error";
                        if(this.benchmarkTechnique.getClass()!=NaiveForecastTechnique.class)
                            myName+="_"+benchmarkTechnique.getName();
		}
		else
			myName = str;
	}
	
	@Override
	public String getAbbreviatedName() {
		return calculator.getAbbreviatedName()+"R"+errorFunction.getAbbreviatedName()+"E";
	}
	
	/**
	 * Collects relative error for a period when a forecast technique is passed.
	 * 
	 * @param ft the forecast technique used to compute the metric value
	 */
	@Override
	public void collectForecastDataForPeriod(ForecastTechnique ft) {
		double observation=ft.getObservation();
		double error=ft.getError();
		if(!Double.isNaN(error)){
			benchmarkTechnique.collect(observation);
			double benchMarkError=benchmarkTechnique.getError();
			if(!Double.isNaN(benchMarkError)){
				if(benchMarkError!=0){
					double relativeError=error/benchMarkError;
					collectMetricForPeriod(relativeError);
				}
			}
		}
	}
	
	/**
	 * Computes the metric value using the StatCalculatorIfc object.
	 */
	@Override
	public void computeMetricValue() {
		metricValue=calculator.computeMetricValue(stat);
	}
	
	/**
	 * Given the relative error for a period, calculates the metric value for a period.
	 * 
	 * @param relativeError the relative error for a particular period
	 */
	public double collectMetricForPeriod(double relativeError){
		double errorFunctionvalue=errorFunction.getValue(relativeError);
		stat.collect(errorFunctionvalue);
		return errorFunctionvalue;
	}
	
        @Override
        public ForecastMetric copyInstance(){
            return new RelativeErrorBasedMetric(this.calculator,this.errorFunction);
        }
        
        @Override
    public void reset() {
        stat.reset();
        metricValue=0;
        benchmarkTechnique.reset();
    }
}
