package jslx.forecasting.metrics;

import jslx.forecasting.ForecastTechnique;
import jslx.forecasting.NaiveForecastTechnique;

/**
 * A class for relative forecast metrics.  A relative forecast metric is computed by dividing the metric value
 * associated with a technique and by the metric value associated with a benchmark technique.  This class 
 * allows the user to choose any ForecastMetric used to compute the relative metric.  In addition, 
 * the user may choose any ForecastTechnique as the benchmark.
 *   
 * @author hmedal
 *
 */
public class RelativeForecastMetric extends ForecastMetric{
	
	/**
	 * The metric to use in evaluating a forecasting technique against a benchmark forecast technique.
	 */
	private ForecastMetric metric;
	
	/**
	 * The object use to collect the values of the metric for the benchmark.  Must be the same metric as the 
	 * 'metric' field.
	 */
	private ForecastMetric benchmarkMetric;
	
	/**
	 * The technique to test against.
	 */
	private ForecastTechnique benchmarkTechnique;
	
	/**
	 * Creates a no named relative forecast metric with a user specified benchmark technique.
	 * @param metric the metric to use in evaluating a forecasting technique against a benchmark 
	 * forecast technique.
	 */
	public RelativeForecastMetric(ForecastMetric metric){
		this(null,metric,new NaiveForecastTechnique());
	}
	
	/**
	 * Creates a named relative forecast metric with a user specified benchmark technique.
	 * @param name the name of the relative forecast metric
	 * @param metric the metric to use in evaluating a forecasting technique against a benchmark 
	 * forecast technique.
	 */
	public RelativeForecastMetric(String name,ForecastMetric metric){
		this(name,metric,new NaiveForecastTechnique());
	}
	
	/**
	 * Creates a no named relative forecast metric with a user specified benchmark technique.
	 * @param metric the metric to use in evaluating a forecasting technique against a benchmark 
	 * forecast technique.
	 * @param benchmarkTechnique the technique to test against.
	 */
	public RelativeForecastMetric(ForecastMetric metric,ForecastTechnique benchmarkTechnique){
		this(null,metric,benchmarkTechnique);
	}
	
	/**
	 * Creates a named relative forecast metric with a user specified benchmark technique.
	 * @param name name of the relative forecast metric
	 * @param metric the metric to use in evaluating a forecasting technique against a benchmark 
	 * forecast technique.
	 * @param benchmarkTechnique the technique to test against.
	 */
	public RelativeForecastMetric(String name,ForecastMetric metric,ForecastTechnique benchmarkTechnique){
		super(name);
		this.metric=metric;
		//Class c=metric.getClass();
        benchmarkMetric=metric.copyInstance();
		this.benchmarkTechnique=benchmarkTechnique;
		setName(name);
	}

	@Override
	public void collectForecastDataForPeriod(ForecastTechnique ft) {
		benchmarkTechnique.collect(ft.getObservation());
		benchmarkMetric.collectForecastDataForPeriod(benchmarkTechnique);
		metric.collectForecastDataForPeriod(ft);
	}

	@Override
	public void computeMetricValue() {
		metric.computeMetricValue();
		double currentMetricValue=metric.getMetricValue();		
		benchmarkMetric.computeMetricValue();
		double benchmarkMetricValue=benchmarkMetric.getMetricValue();
		metricValue = currentMetricValue/benchmarkMetricValue;
	}
	
	/** Sets the name 
	* @param str The name as a string.
	*/
	public void setName(String str) {
		if (str == null){
			myName= "Relative"+benchmarkMetric.getName();
			if(!benchmarkTechnique.getClass().equals(NaiveForecastTechnique.class))
				myName+="_"+benchmarkTechnique.getName();
		}
		else
			myName = str;
	}
	
	@Override
	public String getAbbreviatedName() {
		return "Rel"+benchmarkMetric.getAbbreviatedName();
	}
	
	@Override
    public ForecastMetric copyInstance(){
        return new RelativeForecastMetric(metric.copyInstance(),benchmarkTechnique.copyInstance());
    }
        
    @Override
    public void reset() {
        metric.reset();
        benchmarkMetric.reset();
        benchmarkTechnique.reset();
        metricValue=0;
    }
}
