package jslx.forecasting.metrics;

import jsl.utilities.statistic.Statistic;
import jslx.forecasting.ForecastTechnique;
import jslx.forecasting.NaiveForecastTechnique;
import jslx.forecasting.metrics.calc.StatCalculatorIfc;

/** 
 * 
 * @author vvarghe
 *
 */
public class ScaledErrorBasedMetric extends ForecastMetric {
	
	/**
	 * Computes the metric value given the error
	 */
	protected MathFunctionIfc errorFunction;
	
	/**
	 * An object that calculates the forecast metric value for the collected data given a statistic.
	 */
	protected StatCalculatorIfc calculator;
	
	/**
	 * The object use to collect the values of the metric for the benchmark.  Must be the same metric as the 
	 * 'metric' field.
	 */
	private ForecastMetric benchmarkMetric;
	
	/** The technique to compare against
	 * 
	 */
	protected ForecastTechnique benchmarkTechnique;
	
	/**
	 * The statistic object that keeps track of statistics on forecast metric values for each period
	 */
	protected Statistic stat;
	
	public ScaledErrorBasedMetric(StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		this(null,calculator,errorFunction);
	}
	
	public ScaledErrorBasedMetric(String name,StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		super(name);
		benchmarkTechnique = new NaiveForecastTechnique();
		benchmarkMetric = ForecastMetric.createMeanAbsoluteError();
		stat = new Statistic();
		this.calculator =calculator;
		this.errorFunction=errorFunction;
		setName(name);
		if(calculator.requiresSaved())
			stat.setSaveDataOption(true);
	}
	
	/** Sets the name 
	* @param str The name as a string.
	*/
	public void setName(String str) {
		if (str == null){
			myName=calculator.getName()+errorFunction.getName()+"ScaledError";
		}
		else
			myName = str;
	}
	
	@Override
	public String getAbbreviatedName() {
		return calculator.getAbbreviatedName()+"S"+errorFunction.getAbbreviatedName()+"E";
	}
	
	@Override
	public void collectForecastDataForPeriod(ForecastTechnique ft) {
		double observation=ft.getObservation();
		double error=ft.getError();		
		if(!Double.isNaN(error)){
			benchmarkTechnique.collect(observation);
			benchmarkMetric.collectForecastDataForPeriod(benchmarkTechnique);
			benchmarkMetric.computeMetricValue();
			double benchmarkMetricValue = benchmarkMetric.getMetricValue();			
			if(!Double.isNaN(benchmarkMetricValue)){
				if(benchmarkMetricValue!=0){
					double qt = error/benchmarkMetricValue;
					collectMetricForPeriod(qt);
				}
			}
		}
		
	}
	
	/**
	 * Computes the metric value using the StatCalculatorIfc object.
	 */
	@Override
	public void computeMetricValue() {
		metricValue = calculator.computeMetricValue(stat);
	}
	
	public double collectMetricForPeriod(double scalederror){
		double errorFunctionvalue=errorFunction.getValue(scalederror);
		stat.collect(errorFunctionvalue);
		return errorFunctionvalue;
	}
	
    @Override
    public ForecastMetric copyInstance(){
        return new ScaledErrorBasedMetric(this.calculator,this.errorFunction);
    }
        
        @Override
    public void reset() {
        stat.reset();
        metricValue=0;
        benchmarkTechnique.reset();
        benchmarkMetric.reset();
    }
}
