package jslx.forecasting.metrics;

import jsl.utilities.statistic.Statistic;
import jslx.forecasting.ForecastTechnique;
import jslx.forecasting.metrics.calc.StatCalculatorIfc;

/** Implements the forecast metric obtained by taking the arithmetic mean of the symmetric absolute 
 * percentage errors. 
 * See Hyndman and Koehler 2005 for more details
 * 
 * @author hmedal
 *
 */
public class SymmetricPercentageErrorBasedMetric extends ForecastMetric {
	
	/**
	 * Computes the metric value given the symmetric percentage error
	 */
	protected MathFunctionIfc errorFunction;
	
	/**
	 * An object that calculates the forecast metric value for the collected data given a statistic.
	 */
	protected StatCalculatorIfc calculator;
	
	/**
	 * The statistic object that keeps track of statistics on the symmetric absolute percentage errors
	 */
	private Statistic stat;
	
	/**
	 * Creates a no named forecast metric based on forecast error with a user specified metric calculator and 
	 * error function.
	 * 
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public SymmetricPercentageErrorBasedMetric(StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		this(null,calculator,errorFunction);
	}
	
	/**
	 * Creates a named forecast metric based on forecast error with a user specified metric calculator and 
	 * error function.
	 * 
	 * @param name the name of the forecast metric
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public SymmetricPercentageErrorBasedMetric(String name,StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		super(name);
		stat=new Statistic(name+" statistic");
		this.calculator=calculator;
		this.errorFunction=errorFunction;
		setName(name);
		if(calculator.requiresSaved())
			stat.setSaveDataOption(true);
	}
	
	/**
	 * Gives the symmetric percentage error for a period.
	 * 
	 * @param ft the forecast technique used to calculate the metric
	 */
	@Override
	public void collectForecastDataForPeriod(ForecastTechnique ft) {
		double observation=ft.getObservation();
		double forecast=ft.getCurrentPeriodForecast();
		double difference=observation-forecast;
		double sum=observation+forecast;
		if(sum!=0)
			collectMetricForPeriod((200*difference)/sum);
	}
	
	/**
	 * Given the symmetric percentage error for a period, calculates the metric value for that period.
	 * 
	 * @param percentageError the percentage error for a particular period
	 */
	public double collectMetricForPeriod(double symmetricPercentageError){
		double errorFunctionvalue=errorFunction.getValue(symmetricPercentageError);
		stat.collect(errorFunctionvalue);
		return errorFunctionvalue;
	}
	
	/**
	 * Computes the metric value using the StatCalculatorIfc object.
	 */
	@Override
	public void computeMetricValue() {
		metricValue=calculator.computeMetricValue(stat);
	}
	
	@Override
	public String getAbbreviatedName() {
		return "s"+calculator.getAbbreviatedName()+errorFunction.getAbbreviatedName()+"PE";
	}
	
	/** Sets the name 
	* @param str The name as a string.
	*/
	public void setName(String str) {
		if (str == null){
			myName= "Symmetric"+calculator.getName()+errorFunction.getName()+"PercentageError";
		}
		else
			myName = str;
	}
        
        @Override
        public ForecastMetric copyInstance(){
            return new SymmetricPercentageErrorBasedMetric(this.calculator,this.errorFunction);
        }
        
        @Override
    public void reset() {
        stat.reset();
        metricValue=0;
    }
}
