package jslx.forecasting.metrics;

import jslx.forecasting.ForecastTechnique;

/** Implements the Theil's U Statistic.  
 * See Makridakis et al. 1998 for more details
 * 
 * @author hmedal
 *
 */
public class TheilU extends ForecastMetric{
	
	/**
	 * the sum of the numerator of the root argument in Theil's U function
	 */
	private double numeratorSum;
	
	/**
	 * the sum of the denominator of the root argument in Theil's U function
	 */
	private double denominatorSum;
	
	/**
	 * Creates a no named forecast metric that implements Theil's U statistic.
	 * 
	 */
	public TheilU(){
		this(null);
	}
	
	/**
	 * Creates a named forecast metric that implements Theil's U statistic.
	 * 
	 * @param name the name of the metric
	 */
	public TheilU(String name){
		super(name);
		setName(name);
	}
	
	/**
	 * Keeps track of the numerator and denominator sums in Theil's U statistic.
	 * 
	 * @param ft the forecast technique that is used to calculate the metric
	 */
	@Override
	public void collectForecastDataForPeriod(ForecastTechnique ft) {
		double currentPeriodForecast=ft.getNextPeriodForecast();
		double currentObservation=ft.getObservation();
		double previousObservation=ft.getPrevObservation();
		if(!Double.isNaN(currentPeriodForecast)&&!Double.isNaN(currentObservation)&&
				!Double.isNaN(previousObservation)){
			if(previousObservation!=0){
				numeratorSum+=Math.pow((currentPeriodForecast-currentObservation)/previousObservation, 2);
				denominatorSum+=Math.pow((currentObservation-previousObservation)/previousObservation, 2);
			}
		}
	}
	
	/**
	 * Computes Theil's U statistic for the collected data
	 */
	@Override
	public void computeMetricValue() {
		metricValue=Math.sqrt(numeratorSum/denominatorSum);
	}

	@Override
	public String getAbbreviatedName() {
		return "TheilU";
	}
	
	/** Sets the name 
	* @param str The name as a string.
	*/
	public void setName(String str) {
		if (str == null){
			myName= "Theil's_U_Statistic";
		}
		else
			myName = str;
	}
        
        @Override
        public ForecastMetric copyInstance(){
            return new TheilU();
        }
        
        @Override
    public void reset() {
        metricValue=0;
    }
}
