package jslx.forecasting.metrics;

import jsl.utilities.statistic.Statistic;
import jslx.forecasting.ForecastTechnique;
import jslx.forecasting.metrics.calc.StatCalculatorIfc;

/**
 * An abstract class representing forecast techniques based on unbiased absolute percentage error.
 * See Collopy and Armstrong 2000 for more details
 * @author hmedal
 *
 */
public class UnbiasedPercentageErrorBasedMetric extends ForecastMetric{
	
	/**
	 * Computes the metric value given the error
	 */
	protected MathFunctionIfc errorFunction;
	
	/**
	 * An object that calculates the forecast metric value for the collected data given a statistic.
	 */
	protected StatCalculatorIfc calculator;
	
	/**
	 * The statistic object that keeps track of statistics on forecast metric values for each period
	 */
	protected Statistic stat;
	
	/**
	 * Creates a no named forecast metric based on forecast error with a user specified metric calculator and 
	 * error function.
	 * 
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public UnbiasedPercentageErrorBasedMetric(StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		this(null,calculator,errorFunction);
	}
	
	/**
	 * Creates a named forecast metric based on forecast error with a user specified metric calculator and 
	 * error function.
	 * 
	 * @param name the name of the forecast metric
	 * @param calculator the calculator used to calculate the forecast metric given a set a of data
	 * @param errorFunction the function to manipulate the supplied error value
	 */
	public UnbiasedPercentageErrorBasedMetric(String name,StatCalculatorIfc calculator,MathFunctionIfc errorFunction){
		super(name);
		stat=new Statistic(name+" statistic");
		this.calculator=calculator;
		this.errorFunction=errorFunction;
		setName(name);
		if(calculator.requiresSaved())
			stat.setSaveDataOption(true);
	}
	
	/**Collects the unbiased percentage error for a single forecast.
	* 
	 * @param ft the forecast technique used to compute the metric value
	 */
	@Override
	public void collectForecastDataForPeriod(ForecastTechnique ft) {
		double error=ft.getError();
		double observation=ft.getObservation();
		double denominator=(ft.getCurrentPeriodForecast()+observation)/2;
		double unbiasedAbsolutePercentageError=100*(error/denominator);
		if(!Double.isNaN(unbiasedAbsolutePercentageError))
			collectMetricForPeriod(unbiasedAbsolutePercentageError);
	}
	
	/** Calculates the metric value for a particular period, given the forecast error for that period.
	 * 
	 * @param error the error for a period
	 * @return the metric value
	 */
	public double collectMetricForPeriod(double unbiasedPercentageError) {
		double errorFunctionvalue=errorFunction.getValue(unbiasedPercentageError);
		stat.collect(errorFunctionvalue);
		return errorFunctionvalue;
	}
	
	/**
	 * Computes the metric value using the StatCalculatorIfc object.
	 */
	@Override
	public void computeMetricValue() {
		metricValue=calculator.computeMetricValue(stat);
	}
	
	@Override
	public String getAbbreviatedName() {
		return calculator.getAbbreviatedName()+"U"+errorFunction.getAbbreviatedName()+"PE";
	}
	
	/** Sets the name 
	* @param str The name as a string.
	*/
	public void setName(String str) {
		if (str == null){
			myName= calculator.getName()+"Unbiased"+errorFunction.getName()+"PercentageError";
		}
		else
			myName = str;
	}
        
        @Override
        public ForecastMetric copyInstance(){
            return new UnbiasedPercentageErrorBasedMetric(this.calculator,this.errorFunction);
        }
        
        @Override
    public void reset() {
        stat.reset();
        metricValue=0;
    }
}
