/*
 * RQInventoryModel.java
 *
 * Created on May 28, 2003, 1:20 PM
 */

package jslx.inventory;

import jsl.utilities.random.distributions.LossFunctionDistributionIfc;
import jsl.utilities.random.distributions.Normal;
import jsl.utilities.random.distributions.Poisson;


/**
 *
 * @author  rossetti
 */
public class RQInventoryModel {
    
    protected double myReorderPt;
    
    protected double myReorderQty;
    
    protected LossFunctionDistributionIfc myLTDDist;
    
    protected double myUnitCost;
    
    protected double myCostPerOrder;
    
    protected double myHoldingCostRatePerPeriod;
        
    protected double myBackOrderCostPerUnitPerPeriod;
    
    protected double myCostPerStockout;
    
    protected double myLeadTime;
    protected double myPeriodLength;
    
    protected double myDesiredFillRate;
    
    public RQInventoryModel(){
        myUnitCost = 1.0;
        myCostPerOrder = 1.0;
        myHoldingCostRatePerPeriod = 1.0;
        myBackOrderCostPerUnitPerPeriod = 1.0;
        myCostPerStockout = 1.0;
        myLeadTime = 1.0;
        myPeriodLength = 1.0;
    }
    
    /** Creates a new instance of RQInventoryModel */
    public RQInventoryModel(double reorderPt, double reorderQty, LossFunctionDistributionIfc distribution) {
        setRQPolicy(reorderPt, reorderQty);
        setLeadTimeDemandDistribution(distribution);
        myUnitCost = 1.0;
        myCostPerOrder = 1.0;
        myHoldingCostRatePerPeriod = 1.0;
        myBackOrderCostPerUnitPerPeriod = 1.0;
        myCostPerStockout = 1.0;
        myLeadTime = 1.0;
        myPeriodLength = 1.0;
    }
    
    public void setRQPolicy(double reorderPt, double reorderQty){
        setReorderQty(reorderQty);
        setReorderPoint(reorderPt);
    }
    
    public void setReorderQty(double reorderQty){
         if ((reorderQty <= 0.0))
            throw new IllegalArgumentException("Reorder Quantity must be > 0)");
         
         myReorderQty = reorderQty;
    }
    
    public void setReorderPoint(double reorderPt){
        if (reorderPt < -myReorderQty) // see page 189 of Zipkin(2000)
            throw new IllegalArgumentException("Reorder Point must be >= -Reorder Quantity)");
         
         myReorderPt = reorderPt;
    }
    
    public double getReorderPoint(){
        return(myReorderPt);
    }
    
    public double getReorderQty(){
        return(myReorderQty);
    }
    
    public void setDesiredFillRate(double fillrate){
        if ((0.0 < fillrate ) && (fillrate < 1.0))
            throw new IllegalArgumentException("Fill rate must be between 0 and 1)");
         
         myDesiredFillRate = fillrate;
    }
    
    public double getDesiredFillRate(){
        return(myDesiredFillRate);
    }
    
    public void setLeadTime(double leadTime){
        myLeadTime = leadTime;
    }
    
    public void setPeriodLength(double periodLength){
        myPeriodLength = periodLength;
    }
    
    public void setLeadTimeDemandDistribution(LossFunctionDistributionIfc distribution){
        if (distribution == null)
            throw new NullPointerException("LossFunctionDistributionIfc distribution must be non-null");
        myLTDDist = distribution;
    }
    
    public double getDemandPerPeriod(){
        double mltd = myLTDDist.getMean(); // mean demand during lead time
        return((mltd*myPeriodLength)/myLeadTime);
    }
    
    public double getOrderFrequency(){
        return(getDemandPerPeriod()/myReorderQty);
    }
    
    public double getFillRate() {
        double g1r = myLTDDist.firstOrderLossFunction(myReorderPt);
        double g1rQ = myLTDDist.firstOrderLossFunction(myReorderPt + myReorderQty);
        return(1.0-((g1r-g1rQ)/myReorderQty));
    }
    
    public double getProbStockout(){
        return(1.0-getFillRate());
    }
    
    public double getExpectedStockoutsPerPeriod(){
        double d = getDemandPerPeriod();
        return(d*getProbStockout());
    }
    
    public double getExpectedBackOrders() {
        double g2r = myLTDDist.secondOrderLossFunction(myReorderPt);
        double g2rQ = myLTDDist.secondOrderLossFunction(myReorderPt + myReorderQty);
        return(((g2r-g2rQ)/myReorderQty));
    }
    
    public double getExpectedOnHandInventory() {
        double mltd = myLTDDist.getMean(); // mean demand during lead time
        double b = getExpectedBackOrders(); // expected number of back orders
        double invL = 0.5*(myReorderQty+1.0) + myReorderPt - mltd + b;
        return(invL);
    }
    
    public double getExpectedCustomerWaitTime(){
        double d = getDemandPerPeriod();
        double b = getExpectedBackOrders(); // expected number of back orders
        return(b/d);
    }
    
    // costs
    
    public void setCostPerOrder(double costPerOrder){
        myCostPerOrder = costPerOrder;
    }
    
    public double getCostPerOrder(){
        return(myCostPerOrder);
    }
    
    public void setUnitCost(double unitCost){
        myUnitCost = unitCost;
    }
    
    public double getUnitCost(){
        return(myUnitCost);
    }
        
    public void setHoldingCostRatePerPeriod(double holdingCostRatePerPeriod){
        myHoldingCostRatePerPeriod = holdingCostRatePerPeriod;
    }
    
    public double getHoldingCostRatePerPeriod(){
        return(myHoldingCostRatePerPeriod);
    }
    
    public double getHoldingCostPerUnitPerPeriod(){
        return(myUnitCost*myHoldingCostRatePerPeriod);
    }
    
    public void setBackOrderCostPerUnitPerPeriod(double backOrderCostPerUnitPerPeriod){
        myBackOrderCostPerUnitPerPeriod = backOrderCostPerUnitPerPeriod;
    }
    
    public double getBackOrderCostPerUnitPerPeriod(){
        return(myBackOrderCostPerUnitPerPeriod);
    }
    
    public void setCostPerStockout(double costPerStockout){
        myCostPerStockout = costPerStockout;
    }
    
    public double getCostPerStockout(){
        return(myCostPerStockout);
    }
    
    public double getOrderCostPerPeriod(){
        return(myCostPerOrder*getOrderFrequency());
    }
    
    public double getHoldingCostPerPeriod(){
        return(getHoldingCostPerUnitPerPeriod()*getExpectedOnHandInventory());
    }
    
    public double getBackOrderCostPerPeriod(){
        return(myBackOrderCostPerUnitPerPeriod*getExpectedBackOrders());
    }
    
    public double getStockoutCostPerPeriod(){
        return(myCostPerStockout*getExpectedStockoutsPerPeriod());
    }
    
    public double getBackOrderModelTotalCostPerPeriod(){
        return(getOrderCostPerPeriod() + getHoldingCostPerPeriod() + getBackOrderCostPerPeriod());
    }
    
    public double getStockoutModelTotalCostPerPeriod(){
        return(getOrderCostPerPeriod() + getHoldingCostPerPeriod() + getStockoutCostPerPeriod());
    }
    
    public double getApproxEOQ(){
        double d = getDemandPerPeriod(); 
        double a = getCostPerOrder();
        double h = getHoldingCostPerUnitPerPeriod();
        return(Math.sqrt((2.0*d*a)/h));
    }
    
    public double getApproxOptimalReorderPtViaBackOrderModel(){
        double q = getApproxEOQ();
        double h = getHoldingCostPerUnitPerPeriod();
        double b = getBackOrderCostPerUnitPerPeriod();
        return(myLTDDist.invCDF(b/(b + h)));
    }

    public double getApproxOptimalReorderPtViaStockoutModel(){
        double d = getDemandPerPeriod(); 
        double kd = getCostPerStockout()*d;
        double q = getApproxEOQ();
        double h = getHoldingCostPerUnitPerPeriod();
        return(myLTDDist.invCDF(kd/(kd+ h*q)));
    }

    public String toString(){
        String s ="Reorder point = " + myReorderPt + "\n" +
        "Reorder Quantity = " + myReorderQty + "\n" +
        "Fill rate = " + getFillRate() + "\n" +
        "Probability of Stockout = " + getProbStockout() + "\n" +
        "Expected backorders = " + getExpectedBackOrders() + "\n" +
        "Expected customer waittime = " + getExpectedCustomerWaitTime() + "\n" +
        "Expected on hand inventory = " + getExpectedOnHandInventory() + "\n" +
        "Order Frequency = " + getOrderFrequency() + "\n" +
        "Order cost per period = " + getOrderCostPerPeriod() + "\n" +
        "Holding cost per period = " + getHoldingCostPerPeriod() + "\n" +
        "Backorder cost per period = " + getBackOrderCostPerPeriod() + "\n" +
        "Stockout cost per period = " + getStockoutCostPerPeriod() + "\n" +
        "Backorder model total cost per period = " + getBackOrderModelTotalCostPerPeriod() + "\n" +
        "Stockout model total cost per period = = " + getStockoutModelTotalCostPerPeriod() + "\n" +
        "Approximate EOQ = " + getApproxEOQ() + "\n" +
        "Approximate optimal reorder point via backorder model = " + getApproxOptimalReorderPtViaBackOrderModel() + "\n" +     
        "Approximate optimal reorder point via stockout model = " + getApproxOptimalReorderPtViaStockoutModel();    
        
        return(s);
    }
    
    //  ===========================================
    //      TEST METHODS
    //  ===========================================
    //  For individual classes use main() as test
    
    public static void main( String[] args ){
        System.out.println("Started!");
        
        Poisson d = new Poisson(1.726);  // mean demand during replenishment       
       
        RQInventoryModel m = new RQInventoryModel(3, 4, d); // example on pg 85 of HS
  //      m.setUnitCost(150.0);
 //       m.setHoldingCostRatePerPeriod(1.0);
  //      m.setCostPerOrder(65.138);
 //       m.setBackOrderCostPerUnitPerPeriod(116.5);
  //      m.setCostPerStockout(40.0);
   //     m.setLeadTime(30.0); // 45.0 days
   //     m.setPeriodLength(365.0); // 365 days per year
        
        System.out.println(m);
        
        System.out.println("Done!");
    }
}
