/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.inventory.intermittent;

import java.util.HashMap;

import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.variable.AcrossReplicationStatisticIfc;
import jsl.modeling.elements.variable.ResponseStatisticsIfc;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.utilities.ControllableIfc;
import jsl.utilities.Controls;
import jsl.utilities.ResponseMakerIfc;
import jsl.utilities.Responses;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Constant;
import jslx.forecasting.CumulativeAverage;
import jslx.forecasting.ForecastTechnique;
import jslx.supplychain.afosr.forecasting.DemandAndForecastArrivalListenerDefault;
import jslx.supplychain.afosr.forecasting.ForecastManager;
import jslx.supplychain.inventorylayer.BackLogQueue;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryDemandArrivalListenerIfc;
import jslx.supplychain.inventorylayer.InventoryPolicyReorderPointReorderQuantity;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;

/**
 * @author rossetti
 *
 */
public class ForecastBasedSLSIRQInventoryModel extends SchedulingElement {

    protected ItemType myItemType;
	
    protected DemandGenerator myDemandGenerator;
	
    protected Inventory myInventory;
	
    protected InventoryPolicyReorderPointReorderQuantity myInventoryPolicy;
    
    protected LeadTimeDemandFiller myLeadTimeDemandFiller;
    
    protected BackLogQueue myBLPolicy;
    
    protected ResponseVariable myLeadTimeDemand;
    
    protected double myDemandAggregationLevel;
    private ForecastTechnique myForecaster;
    protected DemandAndForecastArrivalListenerDefault myDefaultDemandForecastArrivalListener;

    protected InventoryDemandArrivalListenerIfc myCaptureLTDListener = new LeadTimeDemandCapturer();

    /**
     * @param parent
     */
    public ForecastBasedSLSIRQInventoryModel(ModelElement parent) {
            this(parent, null);
    }
    public ForecastBasedSLSIRQInventoryModel(ModelElement parent, String name) {
        this(parent, name, new CumulativeAverage(), 1.0);

    }

    /**
     * @param parent
     * @param name
     */
    public ForecastBasedSLSIRQInventoryModel(ModelElement parent, String name, 
            ForecastTechnique ft, double agglevel) {
        super(parent, name);
        // define the type
        myItemType = new ItemType(this, "Type-A");

        // define the filler, default lead time of 1 time unit
        myLeadTimeDemandFiller = new LeadTimeDemandFiller(this);
        myLeadTimeDemandFiller.addLeadTime(myItemType, Constant.ONE);

        // define the inventory and it's policy
        myInventoryPolicy = new InventoryPolicyReorderPointReorderQuantity(this);
        myInventory = new Inventory(this, myItemType, myInventoryPolicy);
        myBLPolicy = new BackLogQueue(myInventory);

        // tell the inventory to use the lead time demand filler
        myInventory.setDemandFiller(myLeadTimeDemandFiller);

        // define the demand generator
        myDemandGenerator = new DemandGenerator(this, myItemType, Constant.ONE, Constant.ONE);

        // tell the generator to send demand to the inventory
        myDemandGenerator.setDemandFiller(myInventory);

        myLeadTimeDemand = new ResponseVariable(this, "LTD Response");

        myInventory.setReplenishmentDemandArrivalListener(myCaptureLTDListener);        
        
        myDemandAggregationLevel = agglevel;
        //attach the demand arrival listener to the inventory
	myDefaultDemandForecastArrivalListener = new DemandAndForecastArrivalListenerDefault(this, 
            myInventory, null, this+"_Dem_Frcst_Arr_Listener", myDemandAggregationLevel);		
	
        myForecaster = ft;        
        myDefaultDemandForecastArrivalListener.setDemandForecaster(new ForecastManager(this, 
            myForecaster, myForecaster.getName()+"_FM"));
    }

    protected void initialize() {
        myForecaster.reset();
    }

    protected void afterReplication() {
        System.out.println("");
        System.out.println("Forecast Estimate:"+this.getForecast());
        System.out.println("MAD:"+this.getMAD());
        System.out.println("MSE:"+this.getMSE());
        System.out.println("EBO ***********************");
        System.out.println("EBO:"+
                this.getAmtBackLoggedResponseStatistics().getReplicationBatchStatistic());
        System.out.println("FR ***********************");
        System.out.println("Fillrate:"+
                this.getFirstFillRateResponseStatistics().getReplicationBatchStatistic());
    }
    
    /**
     * @param value
     * @see jslx.supplychain.inventorylayer.Inventory#setInitialOnHand(int)
     */
    public final void setInitialOnHand(int value) {
        myInventory.setInitialOnHand(value);
    }

    /**
     * @param reorderPoint
     * @param reorderQty
     * @see jslx.supplychain.inventorylayer.InventoryPolicyReorderPointReorderQuantity#setInitialPolicyParameters(int, int)
     */
    public final void setInitialPolicyParameters(int reorderPoint, int reorderQty) {
        myInventoryPolicy.setInitialPolicyParameters(reorderPoint, reorderQty);
    }

    /**
     * @param distribution
     * @see jslx.supplychain.inventorylayer.LeadTimeDemandFiller#addLeadTime(jslx.supplychain.inventorylayer.ItemType, jsl.utilities.random.RandomIfc)
     */
    public final void setLeadTime(RandomIfc distribution) {
        myLeadTimeDemandFiller.addLeadTime(myItemType, distribution);
    }

    /**
     * @param distribution
     * @see jslx.supplychain.inventorylayer.DemandGenerator#setAmountDistributionInitialRandomSource(jsl.utilities.random.RandomIfc)
     */
    public final void setAmountDistribution(RandomIfc distribution) {
        myDemandGenerator.setAmountDistributionInitialRandomSource(distribution);
    }

    /**
     * @param timeBtwEvents
     * @see jsl.modeling.elements.EventGenerator#setTimeBetweenEventsForReplications(jsl.utilities.random.RandomIfc)
     */
    public final void setTimeBetweenEventsForReplications(RandomIfc timeBtwEvents) {
        myDemandGenerator.setTimeBetweenEventsForReplications(timeBtwEvents);
    }

    /**
     * @param timeUntilFirst
     * @see jsl.modeling.elements.EventGenerator#setTimeUntilFirstEventForReplications(jsl.utilities.random.RandomIfc)
     */
    public final void setTimeUntilFirstEventForReplications(RandomIfc timeUntilFirst) {
        myDemandGenerator.setTimeUntilFirstEventForReplications(timeUntilFirst);
    }

    /**
     * @return
     * @see jslx.supplychain.inventorylayer.BackLogPolicyAbstract#getAmtBackLoggedResponseStatistics()
     */
    public final ResponseStatisticsIfc getAmtBackLoggedResponseStatistics() {
        return myBLPolicy.getAmtBackLoggedResponseStatistics();
    }

    /**
     * @return
     * @see jslx.supplychain.inventorylayer.BackLogQueue#getNumInBackLogQResponseStatistics()
     */
    public final ResponseStatisticsIfc getNumInBackLogQResponseStatistics() {
        return myBLPolicy.getNumInBackLogQResponseStatistics();
    }

    /**
     * @return
     * @see jslx.supplychain.inventorylayer.BackLogQueue#getTimeInBackLogQResponseStatistics()
     */
    public final ResponseStatisticsIfc getTimeInBackLogQResponseStatistics() {
        return myBLPolicy.getTimeInBackLogQResponseStatistics();
    }

    /**
     * @return
     * @see jslx.supplychain.inventorylayer.Inventory#getFirstFillRateResponseStatistics()
     */
    public final ResponseStatisticsIfc getFirstFillRateResponseStatistics() {
        return myInventory.getFirstFillRateResponseStatistics();
    }

    /**
     * @return
     * @see jslx.supplychain.inventorylayer.Inventory#getOnHandResponseStatistics()
     */
    public final ResponseStatisticsIfc getOnHandResponseStatistics() {
        return myInventory.getOnHandResponseStatistics();
    }

    /**
     * @return
     * @see jslx.supplychain.inventorylayer.Inventory#getOnOrderResponseStatistics()
     */
    public final ResponseStatisticsIfc getOnOrderResponseStatistics() {
        return myInventory.getOnOrderResponseStatistics();
    }

    /**
     * @return
     * @see jslx.supplychain.inventorylayer.Inventory#getReplenishmentRequestCounterStatistics()
     */
    public final AcrossReplicationStatisticIfc getReplenishmentRequestCounterStatistics() {
        return myInventory.getReplenishmentRequestCounterStatistics();
    }

    /**
     * @return
     * @see jslx.supplychain.inventorylayer.Inventory#getStockOutIndicatorResponseStatistics()
     */
    public final ResponseStatisticsIfc getStockOutIndicatorResponseStatistics() {
        return myInventory.getStockOutIndicatorResponseStatistics();
    }
    
    public double getForecast(){
        return this.myForecaster.getNextPeriodForecast();
    
    }
    
    public double getMAD(){
        return this.myForecaster.getMAD();
    
    }
    
    public double getMSE(){
        return this.myForecaster.getMSE();
    
    }

    class SLSIRQControls extends Controls {
        protected SLSIRQControls(ControllableIfc controllable) {
            super(controllable);
            //create the necessary maps
            myIntegerControls = new HashMap<String, Integer>();
            myRandomIfcControls = new HashMap<String, RandomIfc>();
            // fill the maps
            int r = myInventoryPolicy.getReorderPoint();
            int q = myInventoryPolicy.getReorderQty();
            int i = myInventory.getInitialOnHand();
            myIntegerControls.put("Reorder Point", r);
            myIntegerControls.put("Reorder Qty", q);
            myIntegerControls.put("Initial On Hand", i);
            myRandomIfcControls.put("DemandAmt", myDemandGenerator.getAmountDistribution());
            myRandomIfcControls.put("LeadTime", myLeadTimeDemandFiller.getLeadTime(myItemType));
            myRandomIfcControls.put("TimeBtwDemand", myDemandGenerator.getTimeBetweenEventsForReplications());
            myRandomIfcControls.put("TimeUntilFirst", myDemandGenerator.getTimeUntilFirstEventForReplications());
        }
    }

    /* (non-Javadoc)
     * @see jsl.modeling.ModelElement#getControls()
     */
    public Controls makeControls() {
        return new SLSIRQControls(this);
    }

    /* (non-Javadoc)
     * @see jsl.modeling.ModelElement#useControls()
     */
    protected void useControls() {	
        int r = getControls().getIntegerControl("Reorder Point");
        int q = getControls().getIntegerControl("Reorder Qty");
        int i = getControls().getIntegerControl("Initial On Hand");
        setInitialPolicyParameters(r, q);
        setInitialOnHand(i);
        setAmountDistribution(getControls().getRandomIfcControl("DemandAmt"));
        setTimeBetweenEventsForReplications(getControls().getRandomIfcControl("TimeBtwDemand"));
        setTimeUntilFirstEventForReplications(getControls().getRandomIfcControl("TimeUntilFirst"));
        setLeadTime(getControls().getRandomIfcControl("LeadTime"));	
    }

    class SLSIRQResponse extends Responses {
        protected SLSIRQResponse(ResponseMakerIfc responseMaker) {
            super(responseMaker);
            //create the maps
            myAcrossReplicationResponses = new HashMap<String, AcrossReplicationStatisticIfc>();
            // fill the maps
            myAcrossReplicationResponses.put("First Fill Rate", getFirstFillRateResponseStatistics());
            myAcrossReplicationResponses.put("On Hand", getOnHandResponseStatistics());
            myAcrossReplicationResponses.put("On Order", getOnOrderResponseStatistics());
            myAcrossReplicationResponses.put("Amt Backlogged", getAmtBackLoggedResponseStatistics());
            myAcrossReplicationResponses.put("Number of Replenishments", getReplenishmentRequestCounterStatistics());
            myAcrossReplicationResponses.put("Number Backlogged", getNumInBackLogQResponseStatistics());
            myAcrossReplicationResponses.put("Backlog Waiting Time", getTimeInBackLogQResponseStatistics());	
        }
    }

    public Responses makeResponses(){
        return new SLSIRQResponse(this);		
    }

    class LeadTimeDemandCapturer implements InventoryDemandArrivalListenerIfc {

        public void demandArrived(Inventory inv, Demand demand) {
            long currentTotal = inv.getTotalArrivedDemand();
            long prevTotal = demand.getTotalArrivedDemand();
            double ltd = currentTotal - prevTotal;
            myLeadTimeDemand.setValue(ltd);
        }

    }

    public final void turnOnLeadTimeDemandTrace(String name) {
        myLeadTimeDemand.turnOnTrace(name);
    }

    public final void turnOnLeadTimeDemandTrace() {
        myLeadTimeDemand.turnOnTrace();
    }

    public final void turnOffLeadTimeDemandTrace() {
        myLeadTimeDemand.turnOffTrace();
    }

}
