/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package jslx.inventory.intermittent;

import jsl.modeling.ActionListenerIfc;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Constant;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.ItemType;

/** The generator has two states an "low" period in which
 *  the demand generated is low and a "high" period in which
 *  the demand generated is high.  Distributions
 *  must be provided to determine the time that the
 *  generator is in the low state and high state.
 *  There is also a time between potential demand
 *  generations.  Whenever the time between potential
 *  demand generation "clock" ticks, the appropriate
 *  amount of demand is generated given the curent state
 *  the generator.
 * 
 *  The initial random source for amount of demand
 *  generated in the low period by default is Constant.ONE
 *
 *  The initial random source for the length of the
 *  low period by default is Constant.ONE
 *
 *  The initial random source for amount of demand
 *  generated in the high period by default is Constant.ONE
 *
 *  The initial random source for the length of the
 *  high period by default is Constant.ONE
 *
 * @author rossetti
 */
public class ILDemandGenerator extends DemandGenerator {

    /** Indicates the set of states
     *
     */
    public static enum State {Low, High};

    /** Indicates the current state
     *  Default starting state is State.Low
     *
     */
    protected State myState = State.Low;

    /** Used to set the initial state
     *  at the beginning of each replication
     *
     */
    protected State myInitialState = State.Low;

    /** Governs the amount of demand generated
     *  when the generator is in the low period
     *
     */
    protected RandomVariable myLPeriodAmt;

    /** Governs the length of the low period
     *
     */
    protected RandomVariable myLPeriodLength;

    /** Governs the length of the high period
     *
     */
    protected RandomVariable myHPeriodLength;

    /** Listens for end of current state to transtion
     *  to the next state
     *
     */
    protected StateEndListener myStateEndListener = new StateEndListener();

    /** Represents the end of a low or high state
     *  Used when rescheduling the transition event
     */
    protected JSLEvent myStateEvent;

    /** Tabulates the percentage of time in the high state
     *
     */
    protected TimeWeighted myHPeriodTW;

    /**
	 * @param parent
	 * @param demandFillerFinder an object that finds a DemandFillerIfc
	 */
	public ILDemandGenerator(ModelElement parent, ItemType type) {
		this(parent, type, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * @param parent
	 * @param demandFillerFinder an object that finds a DemandFillerIfc
	 * @param name
	 */
	public ILDemandGenerator(ModelElement parent, ItemType type, String name) {
		this(parent, type, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * @param parent
	 * @param demandFillerFinder an object that finds a DemandFillerIfc
	 * @param name
	 * @param timeUntilFirst
	 */
	public ILDemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilNext) {
		this(parent, type, null, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * @param parent
	 * @param demandFillerFinder an object that finds a DemandFillerIfc
	 * @param name
	 * @param timeUntilFirst
	 */
	public ILDemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilNext, String name) {
		this(parent, type, null, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * @param parent
	 * @param name
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 */
	public ILDemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilFirst, RandomIfc timeUntilNext) {
		this(parent, type, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * @param parent
	 * @param name
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 */
	public ILDemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilFirst,
			RandomIfc timeUntilNext, String name) {
		this(parent, type, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * @param parent
	 * @param name
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 */
	public ILDemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilFirst,
			RandomIfc timeUntilNext, long maxNum, String name) {
		this(parent, type, timeUntilFirst, timeUntilNext, maxNum, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * @param parent
	 * @param name
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 * @param timeUntilLast
	 */
	public ILDemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilFirst,
			RandomIfc timeUntilNext, long maxNum, double timeUntilLast, String name) {
		super(parent, type, timeUntilFirst, timeUntilNext, maxNum, timeUntilLast, name);

        myLPeriodLength = new RandomVariable(this, Constant.ONE);
        myLPeriodAmt = new RandomVariable(this, Constant.ONE);
        
        myHPeriodLength = new RandomVariable(this, Constant.ONE);
        // the amount distribution is used to generate
        // demand during the high period
        myAmtRV = new RandomVariable(this, Constant.ONE);
        myHPeriodTW = new TimeWeighted(this, "Proportion time in High state");

	}

    /** Gets the current state
     *
     * @return
     */
    public final State getState(){
        return myState;
    }

    /** Gets the initial state
     *
     * @return
     */
    public final State getInitialState(){
        return myInitialState;
    }

    public RandomIfc getHighPeriodLengthInitialRandomSource() {
        return myHPeriodLength.getInitialRandomSource();
    }

    public RandomIfc getLowPeriodAmountInitialRandomSource() {
        return myLPeriodAmt.getInitialRandomSource();
    }

    public RandomIfc getLowPeriodLengthInitialRandomSource() {
        return myLPeriodLength.getInitialRandomSource();
    }

    public void setHighPeriodLengthInitialRandomSource(RandomIfc source) {
        myHPeriodLength.setInitialRandomSource(source);
    }

    public void setLowPeriodAmountInitialRandomSource(RandomIfc source) {
        myLPeriodAmt.setInitialRandomSource(source);
    }

    public void setLowPeriodLengthInitialRandomSource(RandomIfc source) {
        myLPeriodLength.setInitialRandomSource(source);
    }

    /** Sets the initial state
     *
     * @param state
     */
    public final void setInitialState(State state){
        myInitialState = state;
    }

    protected void setState(State state){
        myState = state;
        if (myState == State.High)
            myHPeriodTW.setValue(1.0);
        else
            myHPeriodTW.setValue(0.0);
    }

    @Override
    protected int generateAmount() {
        if (myState == State.High)
          return (int)myAmtRV.getValue();
        else
          return (int)myLPeriodAmt.getValue();
    }
    
    @Override
    protected void initialize() {
        super.initialize();
        // determine the initial state and schedule the period
        setState(getInitialState());
        if (myState == State.High){
            // current state is high, schedule its end
            myStateEvent = scheduleEvent(myStateEndListener, myHPeriodLength);
        } else {
            // current state is low, schedule its end
            myStateEvent = scheduleEvent(myStateEndListener, myLPeriodLength);
        }
        
    }

    protected class StateEndListener implements ActionListenerIfc {

        public void action(JSLEvent evt) {
            // state is ending, set next state and schedule its ending
            if (myState == State.High){
                setState(State.Low);
                rescheduleEvent(myStateEvent, myLPeriodLength);
            }
            else {
                setState(State.High);
                rescheduleEvent(myStateEvent, myHPeriodLength);
            }
        }

    }


}
