/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.inventory.intermittent;

import java.util.HashMap;

import jsl.modeling.Experiment;
import jsl.modeling.Model;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.variable.AcrossReplicationStatisticIfc;
import jsl.modeling.elements.variable.ResponseStatisticsIfc;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.utilities.ControllableIfc;
import jsl.utilities.Controls;
import jsl.utilities.ResponseMakerIfc;
import jsl.utilities.Responses;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.random.distributions.ShiftedGeometric;
import jslx.utilities.random.sp.BatchOnOffProcess;
import jslx.supplychain.inventorylayer.BackLogQueue;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryDemandArrivalListenerIfc;
import jslx.supplychain.inventorylayer.InventoryPolicyReorderPointReorderQuantity;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;

/**
 * @author rossetti
 *
 */
public class LIRQInventoryModel extends SchedulingElement {

	protected ItemType myItemType;
	
	protected ILDemandGenerator myDemandGenerator;
	
	protected Inventory myInventory;
	
	protected InventoryPolicyReorderPointReorderQuantity myInventoryPolicy;
    
	protected LeadTimeDemandFiller myLeadTimeDemandFiller;
    
	protected BackLogQueue myBLPolicy;
    
    protected ResponseVariable myLeadTimeDemand;

    protected InventoryDemandArrivalListenerIfc myCaptureLTDListener = new LeadTimeDemandCapturer();

	/**
	 * @param parent
	 */
	public LIRQInventoryModel(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public LIRQInventoryModel(ModelElement parent, String name) {
		super(parent, name);
		// define the type
		myItemType = new ItemType(this, "Type-A");

		// define the filler, default lead time of 1 time unit
		myLeadTimeDemandFiller = new LeadTimeDemandFiller(this);
		myLeadTimeDemandFiller.addLeadTime(myItemType, Constant.ONE);
		
		// define the inventory and it's policy
		myInventoryPolicy = new InventoryPolicyReorderPointReorderQuantity(this);
		myInventory = new Inventory(this, myItemType, myInventoryPolicy);
		myBLPolicy = new BackLogQueue(myInventory);
		
		// tell the inventory to use the lead time demand filler
	    myInventory.setDemandFiller(myLeadTimeDemandFiller);
	    
		// define the demand generator
		myDemandGenerator = new ILDemandGenerator(this, myItemType, Constant.ONE, Constant.ONE);
		
		// tell the generator to send demand to the inventory
		myDemandGenerator.setDemandFiller(myInventory);

        myLeadTimeDemand = new ResponseVariable(this, "LTD Response");

        myInventory.setReplenishmentDemandArrivalListener(myCaptureLTDListener);
	}

	/**
	 * @param value
	 * @see jslx.supplychain.inventorylayer.Inventory#setInitialOnHand(int)
	 */
	public final void setInitialOnHand(int value) {
		myInventory.setInitialOnHand(value);
	}

	/**
	 * @param reorderPoint
	 * @param reorderQty
	 * @see jslx.supplychain.inventorylayer.InventoryPolicyReorderPointReorderQuantity#setInitialPolicyParameters(int, int)
	 */
	public final void setInitialPolicyParameters(int reorderPoint, int reorderQty) {
		myInventoryPolicy.setInitialPolicyParameters(reorderPoint, reorderQty);
	}

	/**
	 * @param distribution
	 * @see jslx.supplychain.inventorylayer.LeadTimeDemandFiller#addLeadTime(jslx.supplychain.inventorylayer.ItemType, jsl.utilities.random.RandomIfc)
	 */
	public final void setLeadTime(RandomIfc distribution) {
		myLeadTimeDemandFiller.addLeadTime(myItemType, distribution);
	}

	/**
	 * @param distribution
	 * @see jslx.supplychain.inventorylayer.DemandGenerator#setAmountDistributionInitialRandomSource(jsl.utilities.random.RandomIfc)
	 */
	public final void setAmountDistributionInitialRandomSource(RandomIfc distribution) {
		myDemandGenerator.setAmountDistributionInitialRandomSource(distribution);
	}

	/**
	 * @param timeBtwEvents
	 * @see jsl.modeling.elements.EventGenerator#setTimeBetweenEventsForReplications(jsl.utilities.random.RandomIfc)
	 */
	public final void setTimeBetweenEventsForReplications(RandomIfc timeBtwEvents) {
		myDemandGenerator.setTimeBetweenEventsForReplications(timeBtwEvents);
	}

	/**
	 * @param timeUntilFirst
	 * @see jsl.modeling.elements.EventGenerator#setTimeUntilFirstEventForReplications(jsl.utilities.random.RandomIfc)
	 */
	public final void setTimeUntilFirstEventForReplications(RandomIfc timeUntilFirst) {
		myDemandGenerator.setTimeUntilFirstEventForReplications(timeUntilFirst);
	}

    public void setLowPeriodLengthInitialRandomSource(RandomIfc source) {
        myDemandGenerator.setLowPeriodLengthInitialRandomSource(source);
    }

    public void setLowPeriodAmountInitialRandomSource(RandomIfc source) {
        myDemandGenerator.setLowPeriodAmountInitialRandomSource(source);
    }

    public void setHighPeriodLengthInitialRandomSource(RandomIfc source) {
        myDemandGenerator.setHighPeriodLengthInitialRandomSource(source);
    }

    public RandomIfc getLowPeriodLengthInitialRandomSource() {
        return myDemandGenerator.getLowPeriodLengthInitialRandomSource();
    }

    public RandomIfc getLowPeriodAmountInitialRandomSource() {
        return myDemandGenerator.getLowPeriodAmountInitialRandomSource();
    }

    public RandomIfc getHighPeriodLengthInitialRandomSource() {
        return myDemandGenerator.getHighPeriodLengthInitialRandomSource();
    }
	
	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.BackLogPolicyAbstract#getAmtBackLoggedResponseStatistics()
	 */
	public final ResponseStatisticsIfc getAmtBackLoggedResponseStatistics() {
		return myBLPolicy.getAmtBackLoggedResponseStatistics();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.BackLogQueue#getNumInBackLogQResponseStatistics()
	 */
	public final ResponseStatisticsIfc getNumInBackLogQResponseStatistics() {
		return myBLPolicy.getNumInBackLogQResponseStatistics();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.BackLogQueue#getTimeInBackLogQResponseStatistics()
	 */
	public final ResponseStatisticsIfc getTimeInBackLogQResponseStatistics() {
		return myBLPolicy.getTimeInBackLogQResponseStatistics();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.Inventory#getFirstFillRateResponseStatistics()
	 */
	public final ResponseStatisticsIfc getFirstFillRateResponseStatistics() {
		return myInventory.getFirstFillRateResponseStatistics();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.Inventory#getOnHandResponseStatistics()
	 */
	public final ResponseStatisticsIfc getOnHandResponseStatistics() {
		return myInventory.getOnHandResponseStatistics();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.Inventory#getOnOrderResponseStatistics()
	 */
	public final ResponseStatisticsIfc getOnOrderResponseStatistics() {
		return myInventory.getOnOrderResponseStatistics();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.Inventory#getReplenishmentRequestCounterStatistics()
	 */
	public final AcrossReplicationStatisticIfc getReplenishmentRequestCounterStatistics() {
		return myInventory.getReplenishmentRequestCounterStatistics();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.Inventory#getStockOutIndicatorResponseStatistics()
	 */
	public final ResponseStatisticsIfc getStockOutIndicatorResponseStatistics() {
		return myInventory.getStockOutIndicatorResponseStatistics();
	}
	
	class LumpyIntermittentRQControls extends Controls {
		protected LumpyIntermittentRQControls(ControllableIfc controllable) {
			super(controllable);
			//create the necessary maps
			myIntegerControls = new HashMap<String, Integer>();
			myRandomIfcControls = new HashMap<String, RandomIfc>();

			// fill the maps

            // start with the inventory and policy
			int r = myInventoryPolicy.getReorderPoint();
			int q = myInventoryPolicy.getReorderQty();
			int i = myInventory.getInitialOnHand();
			myIntegerControls.put("Reorder Point", r);
			myIntegerControls.put("Reorder Qty", q);
			myIntegerControls.put("Initial On Hand", i);
            
			// lead time controls
			myRandomIfcControls.put("LeadTime", myLeadTimeDemandFiller.getLeadTime(myItemType));
           
            // demand generator controls
			myRandomIfcControls.put("TimeBtwDemand", myDemandGenerator.getTimeBetweenEventsForReplications());
			myRandomIfcControls.put("TimeUntilFirst", myDemandGenerator.getTimeUntilFirstEventForReplications());
            myRandomIfcControls.put("LowPeriodAmt", myDemandGenerator.getLowPeriodAmountInitialRandomSource());
            myRandomIfcControls.put("LowPeriodLength", myDemandGenerator.getLowPeriodLengthInitialRandomSource());
            myRandomIfcControls.put("DemandAmt", myDemandGenerator.getAmountDistribution());
            myRandomIfcControls.put("HighPeriodLength", myDemandGenerator.getHighPeriodLengthInitialRandomSource());
		}
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.ModelElement#getControls()
	 */
	public Controls makeControls() {
		return new LumpyIntermittentRQControls(this);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.ModelElement#useControls()
	 */
	protected void useControls() {	
		int r = getControls().getIntegerControl("Reorder Point");
		int q = getControls().getIntegerControl("Reorder Qty");
		int i = getControls().getIntegerControl("Initial On Hand");
		setInitialPolicyParameters(r, q);
		setInitialOnHand(i);

        setLeadTime(getControls().getRandomIfcControl("LeadTime"));

		setTimeBetweenEventsForReplications(getControls().getRandomIfcControl("TimeBtwDemand"));
		setTimeUntilFirstEventForReplications(getControls().getRandomIfcControl("TimeUntilFirst"));

        setAmountDistributionInitialRandomSource(getControls().getRandomIfcControl("DemandAmt"));
        setHighPeriodLengthInitialRandomSource(getControls().getRandomIfcControl("HighPeriodLength"));
        
        setLowPeriodAmountInitialRandomSource(getControls().getRandomIfcControl("LowPeriodAmt"));
        setLowPeriodLengthInitialRandomSource(getControls().getRandomIfcControl("LowPeriodLength"));

	}

	class RQInventoryResponse extends Responses {
		protected RQInventoryResponse(ResponseMakerIfc responseMaker) {
			super(responseMaker);
			//create the maps
			myAcrossReplicationResponses = new HashMap<String, AcrossReplicationStatisticIfc>();
			// fill the maps
			myAcrossReplicationResponses.put("First Fill Rate", getFirstFillRateResponseStatistics());
            myAcrossReplicationResponses.put("Stock out Indicator", getStockOutIndicatorResponseStatistics());
            myAcrossReplicationResponses.put("Number of Replenishments", getReplenishmentRequestCounterStatistics());
			myAcrossReplicationResponses.put("On Hand", getOnHandResponseStatistics());
			myAcrossReplicationResponses.put("On Order", getOnOrderResponseStatistics());
			myAcrossReplicationResponses.put("Amt Backlogged", getAmtBackLoggedResponseStatistics());
			myAcrossReplicationResponses.put("Number Backlogged", getNumInBackLogQResponseStatistics());
			myAcrossReplicationResponses.put("Backlog Waiting Time", getTimeInBackLogQResponseStatistics());	
		}
	}
	
	public Responses makeResponses(){
		return new RQInventoryResponse(this);
	}

    class LeadTimeDemandCapturer implements InventoryDemandArrivalListenerIfc {

        public void demandArrived(Inventory inv, Demand demand) {
            long currentTotal = inv.getTotalArrivedDemand();
            long prevTotal = demand.getTotalArrivedDemand();
            double ltd = currentTotal - prevTotal;
            myLeadTimeDemand.setValue(ltd);
        }

    }

    public final void turnOnLeadTimeDemandTrace(String name) {
        myLeadTimeDemand.turnOnTrace(name);
    }

    public final void turnOnLeadTimeDemandTrace() {
        myLeadTimeDemand.turnOnTrace();
    }

    public final void turnOffLeadTimeDemandTrace() {
        myLeadTimeDemand.turnOffTrace();
    }

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		System.out.println("LumpyIntermittentRQInventoryModel Test");
        
//		runExperiment1();
 
//		runExperiment2();
		
		runExperiment3();

		System.out.println("Done!");
 
	}

	public static void runExperiment1() {
		// create the containing model
		Model m = Model.createModel();
		
		LIRQInventoryModel i = new LIRQInventoryModel(m);
		
//		System.out.println(m.getModelHierarchyAsString());
//		StringBuilder sb = new StringBuilder();
//		m.getInitializationOrderAsString(sb);
//		System.out.println(sb);
		
//		Controls c = i.getControls();

		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
		// set the parameters of the experiment
		e.setNumberOfReplications(30);
		e.setLengthOfReplication(3360.0);
		e.setLengthOfWarmUp(360.0);
//		e.turnOnExperimentReport();
//		e.turnOnReplicationReport();
		e.turnOnExperimentConsoleOutput();

		e.runAll();

	}
	
	public static void runExperiment2() {
		// create the containing model
		Model m = Model.createModel();

		LIRQInventoryModel i = new LIRQInventoryModel(m);

        DistributionIfc tbd = new Exponential(1.0/3.6);
      //  DistributionIfc tbd = new Exponential(1.0/7.2);

		DistributionIfc lt = new Constant(0.5);
		int rpt = -1;
		int qty = 2;
		int level = 0;

		i.setTimeUntilFirstEventForReplications(tbd);
		i.setTimeBetweenEventsForReplications(tbd);
		i.setLeadTime(lt);
		i.setInitialPolicyParameters(rpt, qty);
		i.setInitialOnHand(level);

        //i.turnOnLeadTimeDemandTrace("LeadTimeDemand");

//		System.out.println(m.getModelHierarchyAsString());
//		StringBuilder sb = new StringBuilder();
//		m.getInitializationOrderAsString(sb);
//		System.out.println(sb);
		
//		Controls c = i.getControls();

		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
//		e.turnOnDefaultTraceReport();
		
		// set the parameters of the experiment
		e.setNumberOfReplications(3);
		e.setLengthOfReplication(3360.0);
		e.setLengthOfWarmUp(360.0);
//		e.turnOnExperimentReport();
//		e.turnOnReplicationReport();
		e.turnOnExperimentConsoleOutput();

		e.runAll();

	}
	
	public static void runExperiment3() {
		// create the containing model
		Model m = Model.createModel();

		LIRQInventoryModel i = new LIRQInventoryModel(m);

		RandomIfc tbd = Constant.ONE;
		RandomIfc lt = Constant.ONE;

		i.setTimeUntilFirstEventForReplications(tbd);
		i.setTimeBetweenEventsForReplications(tbd);

		i.setLeadTime(lt);

        i.setAmountDistributionInitialRandomSource(new ShiftedGeometric(0.5));
        i.setHighPeriodLengthInitialRandomSource(new ShiftedGeometric(0.5));
        i.setLowPeriodAmountInitialRandomSource(Constant.ZERO);
        i.setLowPeriodLengthInitialRandomSource(new ShiftedGeometric(0.5));
        		
		int rpt = -1;
		int qty = 2;
		int level = 0;

		i.setInitialPolicyParameters(rpt, qty);
		i.setInitialOnHand(level);
		
//		System.out.println(m.getModelHierarchyAsString());
//		StringBuilder sb = new StringBuilder();
//		m.getInitializationOrderAsString(sb);
//		System.out.println(sb);
		
//		Controls c = i.getControls();

		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
//		e.turnOnDefaultTraceReport();
		
		// set the parameters of the experiment
		e.setNumberOfReplications(30);
		e.setLengthOfReplication(3360.0);
		e.setLengthOfWarmUp(360.0);
//		e.turnOnExperimentReport();
//		e.turnOnReplicationReport();
		e.turnOnExperimentConsoleOutput();

		e.runAll();

	}
}
