package jslx.inventory.rqmodel;

import jsl.utilities.random.distributions.LossFunctionDistributionIfc;

public abstract class RQItemAbstract implements ItemIfc{
	
	/**item number
	 * 
	 */
	protected int myItemNumber;
	
	/**unit cost
	 * 
	 */
	protected int myUnitCost;
	
	/**re-order pt
	 * 
	 */
	protected double myReorderPt;
	
	/**reorder qty
	 * 
	 */
	protected double myReorderQty;
	
	/**lead time for the item
	 * 
	 */
	protected double myLeadTime;
	
	/**ordering cost
	 * 
	 */
	protected double myCostPerOrder;
	
	/**Holding rate
	 * 
	 */
	protected double myHoldingCostRatePerPeriod;
	
	/**backorder cost/unit
	 * 
	 */
	protected double myBackorderCostPerUnit;
	
	/**cost of stockout
	 * 
	 */
	protected double myCostPerStockout;
	
	/**period length
	 * 
	 */
	protected double myPeriodLength;

	/**Demand rate
	 * 
	 */
	protected double myDemandRate;
	
	/** Mean lead time demand
	 * 
	 */
	protected double myMLTD;
	
	/** variance of lead time demand
	 * 
	 */
	protected double myVLTD;
	
	/**lead time distribution for the item
	 * 
	 */
	protected LossFunctionDistributionIfc myLTDCDF;
	
	/**should be overriden by subclasses to setup the lead time demand distr
	 * 
	 */
	abstract protected void setLeadTimeDemandDistribution();
	
	public void setPolicy(double r, double Q){
		if(Q<=0)
			throw new IllegalArgumentException("The quantity must be > 0");
		myReorderQty = Q;
		if(r<-myReorderQty)
			throw new IllegalArgumentException("Reorder point must be >= -Reorder Qty");
		myReorderPt = r;
	}
	
	public void setLeadTime(double time){
		if(time<=0)
			throw new IllegalArgumentException("lead time must be > 0");
		myLeadTime = time;
	}
	
	public void setPeriodLength(double period){
		myPeriodLength = period;
	}
	
	public void setItemNumber(int n){
		myItemNumber = n;
	}
	
	public int getItemNumber(){
		return myItemNumber;
	}
	
	public double getUnitCost(){
		return myUnitCost;
	}
	
	public double getDemandRate(){
		return myDemandRate;
	}
	
	public double getLeadTime(){
		return myLeadTime;
	}
	
	public double getReorderPoint(){
		return myReorderPt;
	}
	
	public double getReorderQty(){
		return myReorderQty;
	}
	
	public LossFunctionDistributionIfc getLeadTimeDemandDistribution(){
		return myLTDCDF;
	}
	
	public double getDemandPerPeriod(){
		double mltd = myLTDCDF.getMean();
		return((mltd*myPeriodLength)/myLeadTime);
	}
	
	public double getOrderFrequency(){
		return(getDemandPerPeriod()/myReorderQty);
	}
	
	public double getFillRate(){
		double g1r = myLTDCDF.firstOrderLossFunction(myReorderPt);
		double g1rQ = myLTDCDF.firstOrderLossFunction(myReorderPt+myReorderQty);
		return(1.0-((g1r-g1rQ)/myReorderQty));
	}
	
	public double getProbabilityOfStockOut(){
		return(1.0-getFillRate());
	}
	
	public double getExpectedStockOutsPerPeriod(){
		return(getDemandPerPeriod()*getProbabilityOfStockOut());
	}
	
	public double getExpectedBackOrders(){
		double g2r = myLTDCDF.secondOrderLossFunction(myReorderPt);
		double g2rQ = myLTDCDF.secondOrderLossFunction(myReorderPt+myReorderQty);
		return(((g2r-g2rQ)/myReorderQty));
	}
	
	public double getFirstOrderLossFunction(double x){
		return myLTDCDF.firstOrderLossFunction(x);
	}
	
	public double getSecondOrderLossFunction(double x){
		return myLTDCDF.secondOrderLossFunction(x);
	}
	
	public double getExpectedOnHandInventory(){
		//double mltd = getMeanLeadTimeDemand();
		double mltd = myMLTD;
		double BO = getExpectedBackOrders();
		double InvLevel = 0.5*(myReorderQty+1)+myReorderPt-mltd+BO;
		return InvLevel;
	}
	
	public double getExpectedCustomerWaitTime(){
		return(getExpectedBackOrders()/getDemandRate());
	}
	
	public double getMeanLeadTimeDemand(){
		return myLTDCDF.getMean();
	}
	
	public double getVarianceOfLeadTimeDemand(){
		return myLTDCDF.getVariance();
	}
	
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append("------------------------------------------------------\n");
		sb.append("(r,Q) Item results\n");
		sb.append("Item number: " + getItemNumber() + "\n");
		sb.append("Re-order point: " + getReorderPoint() + "\n");
		sb.append("Re-order quantity: " + getReorderQty() + "\n");
		sb.append("Unit cost: " + getUnitCost() + "\n");
		sb.append("Lead time: " + getLeadTime() + "\n");
		sb.append("Demand rate: " + getDemandRate() + "\n");
		sb.append("Mean lead time demand: " + myLTDCDF.getMean() + "\n");
		sb.append("Variance of lead time demand: " + myLTDCDF.getVariance() + "\n");
		sb.append("Expected On Hand: " + getExpectedOnHandInventory() + "\n");
		sb.append("Expected Number of backorders: " + getExpectedBackOrders() + "\n");
		sb.append("Fill rate: " + getFillRate() + "\n");
		sb.append("Probability of Stock out: " + getProbabilityOfStockOut() + "\n");
		sb.append("Expected customer waiting time: " + getExpectedCustomerWaitTime() + "\n");
		return(sb.toString());
	}

}