/*
* Created on Jun 21, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.inventory.varimetric;

import jsl.utilities.random.distributions.LossFunctionDistributionIfc;

/**
 * @author rossetti
 *
 */
public abstract class BaseStockItemAbstract implements BaseStockItemIfc {
	
	/** The item number for this item
	 */
	public int myItemNumber;

	/** The unit cost for this item (at the depot)
	 */
	public double myUnitCost;
	
	/** The depot stocking level
	 */
	protected int myStockLevel;
	
	/** The lead time for the item. For a depot item, this is the
	 *  depot repair time.  For a base item, this is determined
	 *  by the depot characteristics
	 */
	protected double myLeadTime;

	/** The demand rate for this item at its location
	 */
	protected double myDemandRate;

	/** The lead time demand distribution for the item
	 */
	protected LossFunctionDistributionIfc  myLTDCDF;
	
	/** Should be overridden by subclasses to setup the lead time
	 *  demand distribution
	 *  
	 */
	abstract protected void setLeadTimeDemandDistribution();

	/** Sets the stocking level for the item
	 * 
	 * @param level Must be >= 0
	 */
	public void setStockLevel(int level) {
		if (level < 0)
			throw new IllegalArgumentException("The level must be >= 0");
		myStockLevel = level;
	}

	/** Sets the unit cost for the item
	 * 
	 * @param cost Must be > 0
	 */
	public void setUnitCost(double cost) {
		if (cost <= 0)
			throw new IllegalArgumentException("The cost must be > 0");
		myUnitCost = cost;	
	}

	/** Sets the item number for the item
	 * 
	 * @param n 
	 */
	protected void setItemNumber(int n) {
		myItemNumber = n;
	}

	/** Sets the lead time for the item, the time that it
	 *  takes to resupply the item
	 * 
	 * @param time Must be > 0
	 */
	protected void setLeadTime(double time) {
		if (time <= 0)
			throw new IllegalArgumentException("The lead time must be > 0");
		myLeadTime = time;
	}

	/* (non-Javadoc)
	 * @see varimetric2.BaseStockItemIfc#getItemNumber()
	 */
	public int getItemNumber() {
		return myItemNumber;
	}

	/* (non-Javadoc)
	 * @see varimetric2.BaseStockItemIfc#getDemandRate()
	 */
	public double getDemandRate() {
		return myDemandRate;
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.BaseStockItemIfc#getLeadTime()
	 */
	public double getLeadTime() {
		return myLeadTime;
	}

	/* (non-Javadoc)
	 * @see varimetric2.BaseStockItemIfc#getStockingCost()
	 */
	public double getStockingCost() {
		return myUnitCost*myStockLevel;
	}

	/* (non-Javadoc)
	 * @see varimetric2.BaseStockItemIfc#getStockLevel()
	 */
	public int getStockLevel() {
		return myStockLevel;
	}

	/* (non-Javadoc)
	 * @see varimetric2.BaseStockItemIfc#getUnitCost()
	 */
	public double getUnitCost() {
		return myUnitCost;
	}

	/* (non-Javadoc)
	 * @see varimetric7.BaseStockItemIfc#getLeadTimeDemandDistribution()
	 */
	public LossFunctionDistributionIfc getLeadTimeDemandDistribution(){
		return myLTDCDF;
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.BaseStockItemIfc#getEBackOrders()
	 */
	public double getEBackOrders() {
		return myLTDCDF.firstOrderLossFunction(getStockLevel());
	}

	/* (non-Javadoc)
	 * @see varimetric2.BaseStockItemIfc#getVBackOrders()
	 */
	public double getVBackOrders(){
		double ebo = getEBackOrders();
		double g2 = myLTDCDF.secondOrderLossFunction(getStockLevel());
		return(2.0*g2-ebo*(ebo-1.0));
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.BaseStockItemIfc#getEOnHand()
	 */
	public double getEOnHand() {
		return getStockLevel() - myLTDCDF.getMean() + getEBackOrders();
	}

	/* (non-Javadoc)
	 * @see varimetric2.BaseStockItemIfc#getPStockOut()
	 */
	public double getPStockOut() {
		return myLTDCDF.complementaryCDF(getStockLevel() - 1.0);
	}
	
	/** Computes the expected backorder waiting tiem for the item
	 * @return
	 */
	public double getEBOWaitTime(){
		double ebo = getEBackOrders();
		return(ebo/getDemandRate());
	}
	
	/** Computes the variance of the backorder waiting time
	 * @return
	 */
	public double getVBOWaitTime(){
		double ebo = getEBackOrders();
		double g2 = myLTDCDF.secondOrderLossFunction(getStockLevel());
		double r = getDemandRate();
		double vbow = (2.0*g2 - ebo*ebo)/(r*r);
		return(vbow);
	}
	
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append("------------------------------------------------------\n");
		sb.append("Base Stock Item Results\n");
		sb.append("Item number: " + getItemNumber() + "\n");
		sb.append("Stock level: " + getStockLevel() + "\n");
		sb.append("Unit cost: " + getUnitCost() + "\n");
		sb.append("Stock cost: " + getStockingCost() + "\n");
		sb.append("Lead time: " + getLeadTime() + "\n");
		sb.append("Demand rate: " + getDemandRate() + "\n");
		sb.append("Mean lead time demand: " + myLTDCDF.getMean() + "\n");
		sb.append("Variance of lead time demand: " + myLTDCDF.getVariance() + "\n");
		sb.append("Expected On Hand: " + getEOnHand() + "\n");
		sb.append("Expected Number of backorders: " + getEBackOrders() + "\n");
		sb.append("Variance of Number of backorders: " + getVBackOrders() + "\n");
		sb.append("Probability of Stock out: " + getPStockOut() + "\n");
		sb.append("Expected backorder waiting time: " + getEBOWaitTime() + "\n");
		sb.append("Variance of backorder waiting time: " + getVBOWaitTime() + "\n");
		return(sb.toString());
	}
}
