/*
* Created on Jun 29, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.inventory.varimetric;

import java.util.ArrayList;
import java.util.List;

import jsl.utilities.statistic.Statistic;

/**
 * @author rossetti
 *
 */
public class MAFItemData {

	protected int myItemNo;
	protected int myLL;
	protected int myUL;
	protected double myAlphaTol = 0.0001;
	protected double myCost=0.0;
	
	/** The maximum number of rows in the minimum expected total base backorder array
	 *  This is NOT the largest value of total number of units of the item in the system
	 */
	protected int myMaxRows;
	
	/** The maximum number of columns in the minimum expected total base backorder array
	 *  This is NOT the largest value of depot stock levels.  It is the number of elements
	 *  in the depot stock set that need to be evaluated.
	 */
	protected int myMaxCols;
	protected int mysMax;
	
	/** The set of depot levels to be evaluated
	 */
	protected int[] myDLevels;
	
	/** an array representing the alpha function 
	 * (min total expected base backorders for s items in system given depot stock level
	 * 
	 */
	protected double[][] myMinEBBO; 
	
	/** The value of depot stock associated with aphaHat
	 * 
	 */
	protected int[] myDStar;
	protected int myLastConvexIndex;
	
	/** The marginal analysis function
	 * 
	 */
	protected int myIncrement;
	protected double[] myAlphaHat;
	protected int[]  myScValues;
	protected double[]  myAlphaHatcValues;
	protected int  myTotalnumberOfConvexPoint;
	protected int[] mySs;


	

	
	/**
	 * @param dLevels an array holding the depot levels to evaluate
	 * @param alphaTol the tolerance in computing the minimum expected backorder array
	 * @param sMaxRows
	 */
	protected MAFItemData(int[] dLevels, double alphaTol, int sMaxLimit, VMItem vi) {
		if (dLevels == null)
			throw new IllegalArgumentException("The depot level array was null");
		if (alphaTol <= 0.0)
			throw new IllegalArgumentException("The tolerance must be > 0");
		if (sMaxLimit <=0)
			throw new IllegalArgumentException("The number of rows in the EBBO array must be > 0");
		
		myAlphaTol = alphaTol;
		
		myDLevels = dLevels;
		myMaxCols = myDLevels.length;

		//add this
		myLL=myDLevels[0];
		myUL=myDLevels[myDLevels.length-1];
		//add this
		mysMax=sMaxLimit;
	    myMaxRows = mysMax-myLL+1;
	
		myMinEBBO = new double[myMaxRows][myMaxCols];
		// initialize all elements to infinity
		for(int i=0;i<myMaxRows;i++)
			for(int j=0; j<myMaxCols;j++)
				myMinEBBO[i][j] = Double.MAX_VALUE;
		
		myDStar = new int[myMaxRows];
		myAlphaHat = new double[myMaxRows];
		myScValues = new int[myMaxRows];
		myAlphaHatcValues=new double[myMaxRows];
		mySs=new int[myMaxRows];
		myItemNo=vi.getItemNumber();
		myCost=vi.getUnitCost();
	//	System.out.println("NOW HERE");
		computeMarginalBOArray(vi);
	

		
		
	}
	
	public void initializemyMinEBBO(){
		for(int i=0;i<myMaxRows;i++)
			for(int j=0; j<myMaxCols;j++)
				myMinEBBO[i][j] = Double.MAX_VALUE;
	}

	public static int[] createDepotLevels(int upperLimit){
		return(createDepotLevels(0, upperLimit, 1));
	}
	
	public static int[] createDepotLevels(int lowerLimit, int upperLimit){
		return(createDepotLevels(lowerLimit, upperLimit, 1));
	}
	
	public static int[] createDepotLevels(int lowerLimit, int upperLimit, int increment){			
		if (lowerLimit < 0)
			throw new IllegalArgumentException("The minimum depot stock level must be >=0");
		if (upperLimit <= lowerLimit)
			throw new IllegalArgumentException("The maximum depot stock level must be > the lower limit");
		if (increment <= 0)
			throw new IllegalArgumentException("The search increment must be > 0");
		if (increment > (upperLimit - lowerLimit))
			throw new IllegalArgumentException("The step size must be <= (upper limit - lower limt");
		// determine the number of elements in the interval
		int n = (int)Math.floor((upperLimit - lowerLimit)/increment) + 1;
		int[] d = new int[n];
		int i = 0;
		for (int s=lowerLimit;s<=upperLimit; s=s+increment){
			d[i] = s;
			i++;
		}
		
		return(d);
	}
	
	public void displaymyDStar(){
		System.out.println("**********************************");
		System.out.println("---myDStar Array----------------");
		for(int i=0;i<myMaxRows;i++){
			System.out.print(myDStar[i]+"\t");
		}
	}
	
	public void displaymyAlphaHat(){
		System.out.println("**********************************");
		System.out.println("---myAlphaHat Array----------------");
		for(int i=0;i<myMaxRows;i++){
			System.out.print(myAlphaHat [i]+"\t");
		}
		System.out.println();
	}
	
	
	public void computeMarginalBOArray(VMItem item){
		//TOD need to check for valid item
		makeMinBaseBOArray(item);
		makeAlphaHatArray(myMinEBBO);
		computemyScValues();
	}
	
	public void displaymyMinEBO(){
		System.out.println("**********************************");
		System.out.println("---myMinEBO Matrix----------------");
		for(int i=0;i<myMaxRows;i++){
			for(int j=0;j<myMaxCols;j++){
				System.out.print(myMinEBBO[i][j]+"\t");
			}
			System.out.println();
		}
	}
	/*
	protected void makeMinBaseBOArray(VMItem item){
		clearMinBaseBOArray();
		int i = 0;
		for(int ds=myLL;ds<=myUL; ds++){
			item.setStockLevel(ds);
			item.setBaseStockLevelsToZero();
			int j = 0;
			for(int s=1;s<=myMaxRows;s++){
				VMBaseItem bi = item.getBaseWithLargestProbOfStockout();
				bi.setStockLevel(bi.getStockLevel()+1);
				double tebbo = item.getTotalBaseEBackOrders();
				if (tebbo >= myAlphaTol){
					myMinEBBO[j][i] = tebbo;
					j++;
				} else
					break;
			}
			i++;
		}
	}
	*/
	
	protected void makeMinBaseBOArray(VMItem item){
	//	System.out.println("***"+item.getNumBases()+"****");
		myIncrement=myDLevels[1]-myDLevels[0]; 
		List<VMBaseItem> mbis=item.getBaseItems();
	//	System.out.println("***"+mbis.size()+"****");
		int counter=0;
		//boolean done=false;
		//double tebbo=0.0;
		initializemyMinEBBO();
	//	System.out.println(myMinEBBO[1][1]+"****");
		for(int i=0;i<myMaxCols;i++){
			counter=myDLevels[i];
			item.setStockLevel(counter);
	//		System.out.println("***"+item.getStockLevel()+"****");
			item.setBaseStockLevelsToZero();
			myMinEBBO[counter-myLL][i] =item.getTotalBaseEBackOrders();
			for(int s=counter+1;s<=mysMax;s++){
				//done=false;
				VMBaseItem bi = item.getBaseWithLargestProbOfStockout();
				//System.out.println(bi.myDemandRate+"...");
				//System.out.println("llll "+bi.getStockLevel()+"...");
				bi.setStockLevel(bi.getStockLevel()+1);
				myMinEBBO[s-myLL][i] =item.getTotalBaseEBackOrders();
	//			System.out.println(myMinEBBO[s-myLL][i]+"*-+++***");
				
			}
		}
	}
	
	protected void makeAlphaHatArray(double [][] a){
		for(int j=0;j<myMaxRows; j++){
			int k = Statistic.getIndexOfMin(myMinEBBO[j]);
			myDStar[j]= k*myIncrement+myLL;
			myAlphaHat[j]=myMinEBBO[j][k];
			
		}
	}
	
	
	
	protected void initializeSIndex(){
		// this method can be avoid considering the relationship between 
		//index and the real s values
		//but the convex point computations should be carefully revised according to
		for(int i=0;i<mySs.length;i++)
			mySs[i]=i+myLL;
	}
	protected double computeSlope(int a, int b){
		return((myAlphaHat[a]-myAlphaHat[b])/(double)(b-a));
	}
	
	protected void displayConvexSs(){
		System.out.println("the Convex S values are :");
		for(int j=0;j<myTotalnumberOfConvexPoint;j++){
			System.out.print(myScValues[j]+"\t");
		}
		System.out.println();
		System.out.println("the Convex Alpha hat values are :");
		for(int j=0;j<myTotalnumberOfConvexPoint;j++){
			System.out.print(myAlphaHatcValues[j]+"\t");
		}
		System.out.println();
	}
		
	protected void computemyScValues(){
		initializeSIndex();
		int i=0;
		int kIndex=0;
		int kLastIndex=0;
		int counterIndex=kIndex+1;
		myScValues[i]=mySs[kIndex];
		myAlphaHatcValues[i]=myAlphaHat[kIndex];

		while(kLastIndex!=mySs.length-1){
		double slopemax=0;
		for(counterIndex=kIndex+1;counterIndex<=mySs.length-1;counterIndex++) {
			if(computeSlope(kIndex,counterIndex)>= slopemax) {
				slopemax=computeSlope(kIndex,counterIndex);
				kLastIndex=counterIndex;
			}
			
		}
		kIndex=kLastIndex;
		i++;
		myScValues[i]=mySs[kLastIndex];
		myAlphaHatcValues[i]=myAlphaHat[kLastIndex];
			myTotalnumberOfConvexPoint=i+1;
		
		}
		}


	
	
	protected void clearMinBaseBOArray(){
		for(int i=0;i<myMinEBBO.length;i++)
			for(int j=0; j<myMinEBBO[i].length;j++)
				myMinEBBO[i][j] = 0.0;			
	}
	
	protected double computeNextMarginalBenefit(){
		if(myLastConvexIndex<myTotalnumberOfConvexPoint-1){
			return((myAlphaHatcValues[myLastConvexIndex]-myAlphaHatcValues[myLastConvexIndex+1])/((myCost)*(double)(myScValues[myLastConvexIndex+1]-myScValues[myLastConvexIndex])));
		}
		else {
			return(Double.MIN_VALUE);
			
		}
	}
	
	
	protected void createMinEBBOArray(int sMaxLimit){
		if (sMaxLimit <=0)
			throw new IllegalArgumentException("The number of rows in the EBBO array must be > 0");
		
		myMaxRows = sMaxLimit;
		// create the array to hold the rows
		myMinEBBO = new double[sMaxLimit][];
		// create the ragged rows
		for (int n=0;n<sMaxLimit;n++)
			myMinEBBO[n] = new double[n+1];
		// initialize all elements to infinity
		for(int i=0;i<myMinEBBO.length;i++)
			for(int j=0; j<myMinEBBO[i].length;j++)
				myMinEBBO[i][j] = Double.MAX_VALUE;
	}
	


}
