/*
* Created on Jun 23, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.inventory.varimetric;

import jsl.utilities.math.JSLMath;
import jsl.utilities.random.distributions.NegativeBinomial;
import jsl.utilities.random.distributions.Poisson;

/**
 * @author rossetti
 *
 */
public class NegBinomialBaseStockModel extends BaseStockItemAbstract {
	
	/**
	 * @param mLTD mean of lead time demand
	 * @param vLTD variance of lead time demand
	 */
	public NegBinomialBaseStockModel(double mLTD, double vLTD, double leadTime) {
		setLeadTimeDemand(mLTD, vLTD, leadTime);
	}
	
	public void setStockLevel(int level){
		super.setStockLevel(level);
	}

	/** The demand rate for the item at the base
	 * 
	 * @param rate Must be > 0
	 */
	protected void setDemandRate(double rate) {
		if (rate < 0)
			throw new IllegalArgumentException("The rate must be >= 0");
		myDemandRate = rate;
	}

	/* (non-Javadoc)
	 * @see varimetric7.BaseStockItemAbstract#setLeadTimeDemandDistribution()
	 */
	@Override
	protected void setLeadTimeDemandDistribution() {
	}

	/** Takes in the mean and the variance of the leadtime demand
	 *  and forms the parameters (r, p) for the negative binomial
	 *  distribution
	 * 
	 * @param mu Mean of the lead time demand
	 * @param sigma2 Variance of the lead time demand
	 */
	public void setLeadTimeDemand(double mu, double sigma2, double leadTime){
		if (mu <= 0.0)
			throw new IllegalArgumentException("mu must be >0");
		if (sigma2 <=0)
			throw new IllegalArgumentException("sigma2 must be > 0");
		
		if(JSLMath.equal(mu, sigma2)){
			if (myLTDCDF == null)
				myLTDCDF = new Poisson(mu);	
			else {
				Poisson cdf = (Poisson)myLTDCDF;
				cdf.setMean(mu);
			}
		} else if (mu < sigma2){
			double p = mu/sigma2;
			double r = (p/(1.0-p))*mu;
			if (myLTDCDF == null)
				myLTDCDF = new NegativeBinomial(p, r);
			else {
				NegativeBinomial cdf = (NegativeBinomial)myLTDCDF;
				cdf.setParameters(p, r);
			}
		} else {
			if (mu > sigma2)
				throw new IllegalArgumentException("sigma2 must be > mu");
		}

		setLeadTime(leadTime);
		setDemandRate(mu/leadTime);

	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		System.out.println("Base stock model Test");
		
//		NegBinomialBaseStockModel b = new NegBinomialBaseStockModel(2.0, 4.0, 0.5);
		
		NegBinomialBaseStockModel b = new NegBinomialBaseStockModel(0.989003756085243, 1.0075709252697, 0.5);

		System.out.println(b);
		
		System.out.println("SO \t OH \t BO \t VBO \t EBW \t VBW");
		for (int s=0;s<=6;s++){
			b.setStockLevel(s);
			double so = b.getPStockOut();
			double oh = b.getEOnHand();
			double bo = b.getEBackOrders();
			double vbo = b.getVBackOrders();
			double ebw = b.getEBOWaitTime();
			double vbw = b.getVBOWaitTime();
			System.out.println(so + "\t" + oh + "\t" + bo + "\t" + vbo + "\t" + ebw + "\t" + vbw);
		}		
	}

}
