/*
* Created on Jun 21, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.inventory.varimetric;

import jsl.utilities.math.JSLMath;
import jsl.utilities.random.distributions.Gamma;
import jsl.utilities.random.distributions.LossFunctionDistributionIfc;
import jsl.utilities.random.distributions.NegativeBinomial;
import jsl.utilities.random.distributions.Poisson;
import jsl.utilities.random.distributions.Distribution;


/**
 * @author rossetti
 *
 */
public class VMBaseItem extends BaseStockItemAbstract implements VMBaseItemIfc {

	/** The depot item associated with this base item
	 */
	protected VMItem myDepotItem;
	
	/** Represents the items' base
	 */
	protected int myBaseNumber;
	
	/** The time that it takes to repair this item at the base
	 */
	protected double myRepairTime;
	
	/** The probability that the item can be repaired at the base
	 */
	protected double myRepairProb;
	
	/** The time that it takes to order and ship the item to the base
	 */
	protected double myShipTime;
		
	/** The expected number in resupply, depends on the 
	 *  depot
	 */
	protected double myENumInResupply;
	
	/** The variance of the number of items in resupply
	 */
	protected double myVNumInResupply;
	
	/** The expected number of items at the depot
	 */
	protected double myENumAtDepot;
	
	/** An array to hold the parameters of the leadtime demand distribution
	 */
	protected double[] myParameters = new double[2];
	
	/** The lead time demand distribution can be either Poisson
	 *  or negative binomial or gamma.  Must have each to switch as
	 *  parameters change
	 * 
	 */
	protected LossFunctionDistributionIfc myPoisson = new Poisson();

	/** The lead time demand distribution can be either Poisson
	 *  or negative binomial or gamma.  Must have each to switch as
	 *  parameters change
	 * 
	 */
	protected LossFunctionDistributionIfc myNegBinomial = new NegativeBinomial();

	/** The lead time demand distribution can be either Poisson
	 *  or negative binomial or gamma.  Must have each to switch as
	 *  parameters change
	 * 
	 */
	protected LossFunctionDistributionIfc myGamma = new Gamma();

    protected VMBaseItem(VMItem depotItem, int baseNum, double demandRate, double repairTime, double repairProb,
			double shipTime, int stockLevel, double cost){
    	myDepotItem = depotItem;
    	myBaseNumber = baseNum;
		setItemNumber(depotItem.getItemNumber());
    	setDemandRate_(demandRate);
    	setRepairTime_(repairTime);
    	setRepairProb_(repairProb);
    	setShipTime_(shipTime);
    	setStockLevel(stockLevel);
    	setUnitCost(cost);
    }
    
	/* (non-Javadoc)
	 * @see varimetric2.BaseItemIfc#getBaseNumber()
	 */
	public int getBaseNumber(){
		return(myBaseNumber);
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.BaseItemIfc#getRepairProb()
	 */
	public double getRepairProb() {
		return myRepairProb;
	}

	/* (non-Javadoc)
	 * @see varimetric2.BaseItemIfc#getRepairTime()
	 */
	public double getRepairTime() {
		return myRepairTime;
	}

	/* (non-Javadoc)
	 * @see varimetric2.BaseItemIfc#getShipTime()
	 */
	public double getShipTime() {
		return myShipTime;
	}

	/* (non-Javadoc)
	 * @see varimetric2.BaseItemIfc#getReplenishmentDemand()
	 */
	public double getReplenishmentDemandRate() {
		return (1.0-myRepairProb)*myDemandRate;
	}

	/* (non-Javadoc)
	 * @see varimetric4.BaseItemIfc#getENumInBaseRepair()
	 */
	public double getENumInBaseRepair() {
		return myRepairProb*myDemandRate*myRepairTime;
	}

	/* (non-Javadoc)
	 * @see varimetric4.BaseItemIfc#getENumInTransit()
	 */
	public double getENumInTransit() {
		return getReplenishmentDemandRate()*myShipTime;
	}
	
	/* (non-Javadoc)
	 * @see varimetric5.BaseItemIfc#getENumAtDepot()
	 */
	public double getENumAtDepot(){
		return myENumAtDepot;
	}

	/* (non-Javadoc)
	 * @see varimetric5.BaseItemIfc#getENumInResupply()
	 */
	public double getENumInResupply() {
		return myENumInResupply;
	}

	/* (non-Javadoc)
	 * @see varimetric5.BaseItemIfc#getVNumInResupply()
	 */
	public double getVNumInResupply() {
		return myVNumInResupply;
	}
	
	/** Changes the repair probability at the base.  This method
	 *  will cause the depot demand rate to be updated and subsequently
	 *  all the lead time demand distributions (depot item and all base items)
	 * 
	 * @param prob
	 */
	public void setRepairProb(double prob){
		setRepairProb_(prob);
		myDepotItem.updateDepotDemandRate();
	}

	/** Sets the probability that this item will be repaired
	 *  at the base
	 *  
	 * @param prob Must be contained [0,1]
	 */
	protected void setRepairProb_(double prob) {
		if ((prob < 0.0) || (prob > 1.0))
			throw new IllegalArgumentException("Probability must be [0,1]");
		myRepairProb = prob;
	}

	/** Sets the repair time for the item.  Updates the lead time
	 *  demand distribution for the item
	 * 
	 * @param time
	 */
	public void setRepairTime(double time){
		setRepairTime_(time);
		setLeadTimeDemandDistribution();
	}
	
	/** Sets the repair time for this item at the base
	 * 
	 * @param time Must be > 0
	 */
	protected void setRepairTime_(double time) {
		if (time <= 0)
			throw new IllegalArgumentException("The repair time must be > 0");
		myRepairTime = time;
	}
	
	/** Sets the ship time for the item.  Updates the lead time
	 *  demand distribution for the item
	 * 
	 * @param time
	 */
	public void setShipTime(double time){
		setShipTime_(time);
		setLeadTimeDemandDistribution();
	}

	/** Sets the repair time for this item at the base
	 * 
	 * @param time Must be > 0
	 */
	protected void setShipTime_(double time) {
		if (time <= 0)
			throw new IllegalArgumentException("The ship time must be > 0");
		myShipTime = time;
	}

	/** Changes the demand rate at the base.  This method
	 *  will cause the depot demand rate to be updated and subsequently
	 *  all the lead time demand distributions (depot item and all base items)
	 * 
	 * @param rate
	 */
	public void setDemandRate(double rate){
		setDemandRate_(rate);
		myDepotItem.updateDepotDemandRate();
	}
	
	/** The demand rate for the item at the base
	 * 
	 * @param rate Must be > 0
	 */
	protected void setDemandRate_(double rate) {
		if (rate < 0)
			throw new IllegalArgumentException("The rate must be >= 0");
		myDemandRate = rate;
	}
		
	protected void setENumInResupply(){
		double enir = getENumInBaseRepair(); // expected number in repair
		double enit = getENumInTransit(); // expected number in transit
		myENumAtDepot = getReplenishmentDemandRate()*myDepotItem.getEBOWaitTime();	
		myENumInResupply = enir+enit+myENumAtDepot;		
//		System.out.println("Setting the mean of the num in resupply to : " + myENumInResupply);
	}
	
	protected void setVNumInResupply(){
		double enir = getENumInBaseRepair();
		double enit = getENumInTransit(); 
		double rd = getReplenishmentDemandRate();		
		double p = rd/myDepotItem.getDemandRate();
		double v1 = p*(1.0-p)*myDepotItem.getEBackOrders();
		double v2 = p*p*myDepotItem.getVBackOrders();
		myVNumInResupply = enir + enit + v1 + v2;	
//		System.out.println("Setting the variance of the num in resupply to : " + myVNumInResupply);
	}

	protected void setLeadTime(){
		double rp = getRepairProb();
		double ert = rp*getRepairTime();
		double est = (1.0-rp)*getShipTime();
		double ewt = (1.0-rp)*myDepotItem.getEBOWaitTime();
		setLeadTime(ert+est+ewt);
	}
	
	protected void setLeadTimeDemandDistribution(){
		setENumInResupply();
		setVNumInResupply();	
		setLeadTime();
		double mu = getENumInResupply();
		double s2 = getVNumInResupply();
//		System.out.println("mu= "+mu);
//		System.out.println("sigma2= "+s2);
		// lead time demand for base item needs to be set
		setLeadTimeDemand(mu, s2);		
	}
	
	/** Takes in the mean and the variance of the leadtime demand
	 *  and forms the parameters (r, p) for the negative binomial
	 *  distribution
	 * 
	 * @param mu Mean of the lead time demand
	 * @param sigma2 Variance of the lead time demand
	 */
	protected void setLeadTimeDemand(double mu, double sigma2){
		if (mu <= 0.0)
			throw new IllegalArgumentException("mu must be >0");
		if (sigma2 <=0)
			throw new IllegalArgumentException("sigma2 must be > 0");
		
		double c = sigma2/mu;
		
		
		if(c > 0.9 && c <1.1){
			myLTDCDF = myPoisson;
			myParameters[0] = mu;
		} else if (c > 1.1){
			double p = mu/sigma2;
			double r = (p/(1.0-p))*mu;
//			System.out.println("Setting parameters of negative binomial");
//			System.out.println("mu= "+mu);
//			System.out.println("sigma2= "+sigma2);
//			System.out.println("p= " + p);
//			System.out.println("r= "+ r);
			myLTDCDF = myNegBinomial;
			myParameters[0] = p;
			myParameters[1] = r;			
		} else {
			double beta = sigma2/mu;
			double alpha = mu/beta;
			myLTDCDF = myGamma;
			myParameters[0] = alpha;
			myParameters[1] = beta;			
			
			//TODO should come up with an approximation for this case
			/*if (mu > sigma2)
				throw new IllegalArgumentException("sigma2 must be > mu");*/
		}
		
/*		if(JSLMath.equal(mu, sigma2)){
			myLTDCDF = myPoisson;
			myParameters[0] = mu;
		} else if (mu < sigma2){
			double p = mu/sigma2;
			double r = (p/(1.0-p))*mu;
//			System.out.println("Setting parameters of negative binomial");
//			System.out.println("mu= "+mu);
//			System.out.println("sigma2= "+sigma2);
//			System.out.println("p= " + p);
//			System.out.println("r= "+ r);
			myLTDCDF = myNegBinomial;
			myParameters[0] = p;
			myParameters[1] = r;			
		} else {
			double beta = sigma2/mu;
			double alpha = mu/beta;
			myLTDCDF = myGamma;
			myParameters[0] = alpha;
			myParameters[1] = beta;			
			
			//TODO should come up with an approximation for this case
			if (mu > sigma2)
				throw new IllegalArgumentException("sigma2 must be > mu");
		}*/
		
		// to force use of Poisson uncomment these two lines
	//	myLTDCDF = myPoisson;
	//	myParameters[0] = mu;
		
		Distribution cdf = (Distribution)myLTDCDF;
		cdf.setParameters(myParameters);
	}
	
	/** Computes the optimal base stocking level for the given criteria
	 * 
	 * @param theta
	 * @return
	 */
	protected int getOptimalBaseLevelForLagrangeBaseSubProblem(double theta){
		double gamma = 1.0 - theta*getUnitCost();
		if (gamma <= 0.0)
			return 0;
		else
			return((int)myLTDCDF.invCDF(gamma));
	}
	
	/** Sets the base stocking level to the optimal for the Lagrange base sub problem
	 * 
	 * @param theta
	 */
	protected void optimizeLagrangeBaseSubProblem(double theta){
		setStockLevel(getOptimalBaseLevelForLagrangeBaseSubProblem(theta));
	}
	
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append(super.toString());
		sb.append("Base Item data \n");
		sb.append("Base number: " + getBaseNumber() + "\n");
		sb.append("Repair Probability: " + getRepairProb() + "\n");
		sb.append("Repair Time: " + getRepairTime() + "\n");
		sb.append("Ship Time " + getShipTime() + "\n");
		sb.append("Replenishment demand: " + getReplenishmentDemandRate() + "\n");
		sb.append("Expected Number in Base Repair: " + getENumInBaseRepair() + "\n");
		sb.append("Expected Number in Transit: " + getENumInTransit() + "\n");
		sb.append("Expected Number in Resupply: " + getENumInResupply() + "\n");
		sb.append("Variance of Number in Resupply: " + getVNumInResupply() + "\n");
		
		return(sb.toString());
	}
}
