/*
* Created on Jun 21, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.inventory.varimetric;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import jsl.utilities.random.distributions.Poisson;

/**
 * @author rossetti
 *
 */
public class VMItem extends BaseStockItemAbstract implements VMItemIfc {

	/** Holds the base items for this item
	 */
	protected ArrayList<VMBaseItem> myBaseItems;
			
	/** Creates a VMItem with the given parameters.  The VMItem will not be in a
	 *  valid state until the first base item is added
	 * 
	 * @param itemNumber
	 * @param stockLevel
	 * @param cost
	 * @param leadTime
	 */
	public VMItem(int itemNumber, int stockLevel, double cost, double leadTime){
		myBaseItems = new ArrayList<VMBaseItem>();
		setItemNumber(itemNumber);
		setStockLevel(stockLevel);
		setUnitCost(cost);
		setLeadTime(leadTime);
	}
	
	/** Creates and returns a (independent) deep copy of the supplied item
	 * 
	 * @param original
	 * @return
	 */
	public static VMItem newInstance(VMItem original){
		VMItem clone = new VMItem(original.myItemNumber, original.myStockLevel, original.myUnitCost, original.myLeadTime);
		for(VMBaseItem bi: original.myBaseItems){
			clone.addBaseItem(bi.myBaseNumber, bi.myDemandRate, bi.myRepairTime, bi.myRepairProb, bi.myShipTime, bi.myStockLevel, bi.myUnitCost);
		}
		return(clone);
	}
	
	/** Adds a base item with a default base item number
	 * 
	 * @param baseDemandRate
	 * @param repairTime
	 * @param repairProb
	 * @param shipTime
	 * @param stockLevel
	 * @param cost
	 * @return
	 */
	public VMBaseItemIfc addBaseItem(double baseDemandRate, double repairTime, double repairProb,
			double shipTime, int stockLevel, double cost){
		return(addBaseItem(myBaseItems.size()+1, baseDemandRate, repairTime, repairProb, shipTime, stockLevel, cost));
	}
	
	/** Changes the stock level to the supplied value.  This will cause an update to all
	 *  base item performance measures due to update of their lead time demand distributions.
	 *  
	 * @param The new stock level
	 */
	public void setStockLevel(int stockLevel){
		super.setStockLevel(stockLevel);
		updateBaseItemLeadTimeDemandDistributions();
	}
	
	/** Adds a base item
	 * 
	 * @param baseNum
	 * @param baseDemandRate
	 * @param repairTime
	 * @param repairProb
	 * @param shipTime
	 * @param stockLevel
	 * @param cost
	 * @return
	 */
	public VMBaseItemIfc addBaseItem(int baseNum, double baseDemandRate, double repairTime, double repairProb,
			double shipTime, int stockLevel, double cost){	
		if (myBaseItems.isEmpty())
			myLTDCDF = new Poisson();    
		// create the item
		VMBaseItem bi = new VMBaseItem(this, baseNum, baseDemandRate, repairTime, repairProb, shipTime, stockLevel, cost);
		// add it to the base items
//		System.out.println("Adding base item: " + bi.getBaseNumber());
		myBaseItems.add(bi);
		
		updateDepotDemandRate();

		return(bi);
	}
	
	/* (non-Javadoc)
	 * @see varimetric8.VMItemIfc#getTotalEBackOrders()
	 */
	public double getTotalEBackOrders() {
		return getEBackOrders() + getTotalBaseEBackOrders();
	}

	/* (non-Javadoc)
	 * @see varimetric8.VMItemIfc#getTotalEOnHand()
	 */
	public double getTotalEOnHand() {
		return getEOnHand() + getTotalBaseEOnHand();
	}

	/* (non-Javadoc)
	 * @see varimetric8.VMItemIfc#getTotalStockingCost()
	 */
	public double getTotalStockingCost() {
		return getStockingCost() + getTotalBaseStockingCost();
	}

	/* (non-Javadoc)
	 * @see varimetric8.VMItemIfc#getNumBases()
	 */
	public int getNumBases() {
		return myBaseItems.size();
	}

	/* (non-Javadoc)
	 * @see varimetric8.VMItemIfc#getBaseItems()
	 */
	public List<VMBaseItem> getBaseItems() {
		return Collections.unmodifiableList(myBaseItems);
	}

	/* (non-Javadoc)
	 * @see varimetric8.VMItemIfc#getTotalBaseEBackOrders()
	 */
	public double getTotalBaseEBackOrders() {
		double t = 0.0;
		for(VMBaseItem bi: myBaseItems)
			t = t + bi.getEBackOrders();
		return t;
	}

	/* (non-Javadoc)
	 * @see varimetric8.VMItemIfc#getTotalBaseEOnHand()
	 */
	public double getTotalBaseEOnHand() {
		double t = 0.0;
		for(VMBaseItem bi: myBaseItems)
			t = t + bi.getEOnHand();
		return t;
	}

	/* (non-Javadoc)
	 * @see varimetric8.VMItemIfc#getTotalBaseStockingCost()
	 */
	public double getTotalBaseStockingCost() {
		double t = 0.0;
		for(VMBaseItem bi: myBaseItems)
			t = t + bi.getStockingCost();
		return t;
	}

	
	/** Returns the stock levels as an array with s[0] being the depot
	 *  and s[j] being the level for each base
	 * 
	 * @return
	 */
	public int[] getStockLevels(){
		int[] s= new int[getNumBases()+1]; // holds the found stock levels, s[0] is depot
		s[0] = getStockLevel();
		int j = 1;
		for(VMBaseItem bi: myBaseItems){
			s[j] = bi.getStockLevel();
			j++;
		}
		return(s);
	}
	
	/** Sets the stocking levels at the depot and bases according to the supplied
	 *  array.  s[0] is the depot, s[j] j>=1 to the number of bases are the base
	 *  stock levels.  Throws an IllegalArgumentException if the array size is
	 *  not equal to getNumBase() + 1
	 * 
	 * @param s
	 */
	public void setStockLevels(int[] s){
		if (s.length != getNumBases() + 1)
			throw new IllegalArgumentException("The number of stock levels was not " + (getNumBases() + 1));
		setStockLevel(s[0]);
		int j = 1;
		for(VMBaseItem bi: myBaseItems){
			bi.setStockLevel(s[j]);
			j++;
		}
	}
	
	public void setBaseStockLevelsToZero(){
		for(VMBaseItem bi: myBaseItems)
			bi.setStockLevel(0);		
	}
	
	public VMBaseItem getBaseWithLargestProbOfStockout(){
		double max = Double.NEGATIVE_INFINITY;
		VMBaseItem fb = null;
		//VMBaseItem fb = myBaseItems.get(0);
		for(VMBaseItem bi: myBaseItems){
			double pso = bi.getPStockOut();
			if (pso > max){
				max = pso;
				fb = bi;
			}
		}
		return(fb);
	}

	public double getMinUnitCostAcrossBases(){
		double min = Double.MAX_VALUE;
		for(VMBaseItem bi: myBaseItems){
			double uc = bi.getUnitCost();
			if (uc < min){
				min = uc;;
			}
		}
		return(min);		
	}
	
	public double getMaxUnitCostAcrossBases(){
		double max = Double.MIN_VALUE;
		for(VMBaseItem bi: myBaseItems){
			double uc = bi.getUnitCost();
			if (uc > max){
				max = uc;;
			}
		}
		return(max);		
	}
	
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append("******************************************\n");
		sb.append(super.toString());
		sb.append("Aggregate Results\n");
		sb.append("Total Stocking Cost: " + getTotalStockingCost() + "\n");
		sb.append("Total Expected BackOrders " + getTotalEBackOrders() + "\n");
		sb.append("Total Expected Onhand: " + getTotalEOnHand() + "\n");
		sb.append("Number of bases: " + getNumBases() + "\n");
		sb.append("Total Base Stocking Cost: " + getTotalBaseStockingCost() + "\n"); // this is now correct
		sb.append("Total Expected Base Backorders: " + getTotalBaseEBackOrders() + "\n");
		sb.append("Total Expected Base Onhand: " + getTotalBaseEOnHand() + "\n");
		sb.append("\nIndividual base results\n");
		for(VMBaseItem bi:myBaseItems){
			sb.append("------------------------------------------------------\n");
			sb.append("Base number: " + bi.getBaseNumber() + "\n");
			sb.append(bi.toString());
		}
		sb.append("******************************************\n");
		return(sb.toString());
	}
	
	/** Causes the depot demand rate to be recomputed and subsequently
	 *  the lead time demand for the item and any of its base items
	 */
	protected void updateDepotDemandRate(){
		// recompute the depot demand rate for the item
		myDemandRate = 0.0;
		for(VMBaseItem b: myBaseItems)
			myDemandRate = myDemandRate + b.getReplenishmentDemandRate();
		// set up the depot's lead time demand distribution
		setLeadTimeDemandDistribution();
		// set up the lead time demand distributions for the base items
		updateBaseItemLeadTimeDemandDistributions();	
	}
	
	/* (non-Javadoc)
	 * @see varimetric8.BaseStockItemAbstract#setLeadTimeDemandDistribution()
	 */
	protected void setLeadTimeDemandDistribution(){
		double rate = getDemandRate();
		double lt = getLeadTime();
		Poisson p = (Poisson)myLTDCDF;
		p.setMean(rate*lt);		
	}
	
	/** Tells all the base items to update their lead time demand distributions
	 */
	protected void updateBaseItemLeadTimeDemandDistributions(){
		for(VMBaseItem b: myBaseItems)
			b.setLeadTimeDemandDistribution();
	}
	
	protected int[] optimizeLagrangeItemSubProblem(double theta, int minS, int maxS, int delta){
		if (theta < 0.0)
			throw new IllegalArgumentException("The value of theta must be > 0.0");
		if (minS < 0)
			throw new IllegalArgumentException("The minimum depot stock level must be >=0");
		if (maxS <= minS)
			throw new IllegalArgumentException("The maximum depot stock level must be > minS");
		if (delta <= 0)
			throw new IllegalArgumentException("The search increment must be > 0");
		if (delta > (maxS - minS))
			throw new IllegalArgumentException("The step size must be <= (maxS-minS");
		
		double wstar = Double.MAX_VALUE;

		int[] s= new int[getNumBases()+1]; // holds the found stock levels, s[0] is depot
		
		for(int rho=minS; rho<=maxS;rho=rho+delta){
			setStockLevel(rho);
			optimizeBases(theta);
			double w = getTotalBaseEBackOrders()+ theta*getTotalStockingCost();
			if (w < wstar){
				wstar = w;
				s[0] = rho;
				int j = 1;
				for(VMBaseItem bi: myBaseItems){
					s[j] = bi.getStockLevel();
					j++;
				}
			}
		}
		// optimal should be found set item at optimal
		setStockLevel(s[0]);
		int j = 1;
		for(VMBaseItem bi: myBaseItems){
			bi.setStockLevel(s[j]);
			j++;
		}	
		return(s);
	}
	
	protected double getMaxValueOfLagrangeMultiplierForItem(){
		return(1.0/getMaxUnitCostAcrossBases());
	}
	
	protected void optimizeBases(double theta){
		for(VMBaseItem bi: myBaseItems)
			bi.optimizeLagrangeBaseSubProblem(theta);
	}

	
}
