/*
* Created on Jun 30, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.inventory.varimetric;

import java.util.List;
import java.util.logging.Logger;


import jsl.utilities.math.FunctionIfc;
import jsl.utilities.random.distributions.LossFunctionDistributionIfc;

/**
 * @author rossetti
 *
 */
abstract public class VMLagrangeAlgorithm {

	public static Logger LOGGER = Logger.getLogger(VariMetricModel.class.getName());
	protected int myDefaultSearchIncrement = 1;
	protected double myDefaultThetaLowerTolerance =0.0001;
	protected double myDefaultThetaUpperTolerance = 0.0001;
	protected double myFinalTheta = Double.NaN;
	protected int myDefaultMaxNumIterations = 100;
	protected double myBudget;
	protected double myMinTheta;
	protected double myMaxTheta;
	protected double myGapAtMinTheta;
	protected double myGapAtMaxTheta;
	protected int myMaxNumEvaluations;
	protected VariMetricModel myModel;
	protected CostConstraint myCostConstraint = new CostConstraint();
	
	private boolean myConvergedWithinSearchLimit = false;

	public void setLagrangeMultiplierInterval(double minTheta, double maxTheta) {
		if (minTheta <= 0.0)
			throw new IllegalArgumentException("The minimum value of theta must be > 0.0");
		if (maxTheta <= minTheta)
			throw new IllegalArgumentException("The maximum value of theta must be > the minimum value");

		myMinTheta = minTheta + myDefaultThetaLowerTolerance;
		myMaxTheta = maxTheta - myDefaultThetaUpperTolerance;
	}

	public void setNumberOfEvaluations(int numPoints) {
		if (numPoints <= 0)
			throw new IllegalArgumentException("The number of search points must be > 0");
		myMaxNumEvaluations = numPoints;
	}

	/** Gets the difference between the total cost and the budget
	 *  Negative means that we are under budget (within the budget)
	 *  Positive means that we are over budget
	 * 
	 * @return
	 */
	public double getBudgetGap() {
		return(myModel.getTotalStockingCost() - myBudget);
	}

	/** Checks whether or not the budget constraint is satisfied
	 * 
	 * @return
	 */
	public boolean checkBudgetConstraint() {		
		return(getBudgetGap() <= 0.0);
	}

	/** Gets the maximum allowed value of the Lagrange multiplier for the given model
	 * 
	 * @return
	 */
	public double getRecommendedMaxAllowedValueOfLagrangeMultiplier(){
		// find the minimum of the maximum theta for each item
		double min = Double.MAX_VALUE;
		List<VMItem> items = myModel.getItems();
		for(VMItem vi: items){
			double maxTheta = vi.getMaxValueOfLagrangeMultiplierForItem();
			if (maxTheta < min)
				min = maxTheta;
		}
		return(min);
	}

	public void evaluateCostConstraint(double budget, double minTheta, double maxTheta, int numPoints){
		if (budget < 0.0)
			throw new IllegalArgumentException("The budget must be > 0.0");
		if (minTheta < 0.0)
			throw new IllegalArgumentException("The minimum value of theta must be > 0.0");
		if (maxTheta <= minTheta)
			throw new IllegalArgumentException("The maximum value of theta must be > the minimum value");
		if (numPoints <= 0)
			throw new IllegalArgumentException("The number of search points must be > 0");
		
		double delta = (maxTheta - minTheta)/numPoints;
		double theta = 0.0;
		for (theta = minTheta; theta<maxTheta;theta = theta + delta){
			System.out.println("Theta "+theta);
			double tsc = getTotalStockingCost(theta);
			double gap = tsc - budget;
			System.out.println("theta = " + theta + " \t Cost(theta) = " + tsc + " \t budget gap = " + gap);
		}
	
	}
	
	public double getFinalLagrangeMultiplierValue(){
		return myFinalTheta;
	}
	
	public boolean optimize(double budget) {
		setBudget(budget);
		return(optimize());
	}

	public boolean optimize() {
		// check simple budget feasibility
		double minUC = myModel.getMinUnitCostAcrossAllItems();
		/*if (myBudget < minUC)
			throw new IllegalArgumentException("The budget is less than the cost of the item with the minimum cost");*/
		if (myBudget < minUC){
			String st = "The supplied budget was < minimum unit cost for any item. No items can be stocked. \n";
			st = st + "Setting all stocking levels to zero for optimal solution\n";
			st = st + "budget = " + myBudget  + " \t  minimum unit cost  = " + minUC;
			LOGGER.warning(st);
			myModel.setAllStockingLevelToZero();
			return(true);
		}
		// stock levels are not all zero, try optimizing Lagrange multiplier
		if (hasRoot()){
			myConvergedWithinSearchLimit = true;
			searchLagrangeMultiplierInterval();
		}else {
			if (Math.abs(myGapAtMinTheta) < Math.abs(myGapAtMaxTheta)){
				// evaluate again at minTheta
				myCostConstraint.fx(myMinTheta);
				myFinalTheta = myMinTheta;
			} else {
				// already evaluated last at maxTheta
				myFinalTheta = myMaxTheta;
			}		
			String st = "There was no root in the Lagrange multiplier search interval (" + myMinTheta + "," + myMaxTheta + ")\n";
			st = st + "Budget Gaps at end points: f(" + myMinTheta + ") = " + myGapAtMinTheta + " f(" + myMaxTheta + ") = " + myGapAtMaxTheta + "\n";
			st = st + "Picking the multiplier with the smallest absolute budget gap, i.e. closest to budget value. theta = " + myFinalTheta;
			LOGGER.warning(st);
		}
		
		return (checkBudgetConstraint());
	}
	
	public boolean isConvergedWithinSearchLimit(){
		
		return myConvergedWithinSearchLimit;
		
	}
	
	abstract protected void searchLagrangeMultiplierInterval();

	protected void setVariMetricModel(VariMetricModel model) {
		if (model == null)
			throw new IllegalArgumentException("The supplied model was null");
		myModel = model;
	}

	protected void setBudget(double budget) {
		if (budget < 0)
			throw new IllegalArgumentException("The supplied budget was < 0.0");
		myBudget = budget;
	}

	/** True indicates that the problem has a Lagrangian root in the supplied search
	 *  interval.  The model is changed by this method.
	 * 
	 * @return
	 */
	protected boolean hasRoot() {		
		myGapAtMinTheta = myCostConstraint.fx(myMinTheta);
		myGapAtMaxTheta = myCostConstraint.fx(myMaxTheta);	
		//System.out.println("myGapAtMinTheta: "+myGapAtMinTheta+" myGapAtMaxTheta: "+myGapAtMaxTheta);
		return(myGapAtMinTheta * myGapAtMaxTheta <= 0);	
	}
	
	/** Tries to optimize the Lagrange item sub problem using a good range 
	 *  for the depot stock levels
	 * 
	 * @param item
	 * @param theta
	 * @return
	 */
	protected int[] optimizeLagrangeItemSubProblem(VMItem item, double theta){
		LossFunctionDistributionIfc cdf = item.getLeadTimeDemandDistribution();
		//System.out.println("myItemNumber "+item.myItemNumber);
		double mu = cdf.getMean();
		//System.out.println("mu "+mu);
		double s = cdf.getStandardDeviation();
		//System.out.println("s "+s);
		int fmu = (int)Math.floor(mu);
		//System.out.println("fmu "+fmu);
		int c2s = (int)Math.ceil(2.0*s);
		//System.out.println("c2s "+c2s);
		int sbl = (int)Math.floor(myBudget/item.getUnitCost());
		//System.out.println("myBudget "+myBudget);
		//System.out.println("item.getUnitCost() "+item.getUnitCost());		
		//System.out.println("sbl "+sbl);
		int maxS;
		int minS;		
		if (sbl <= fmu){
			maxS = Math.max(1, sbl);
			//maxS = sbl;
			minS = Math.max(0, maxS - c2s);
			//System.out.println("When (sbl <= fmu) maxS "+maxS+" minS "+minS);
		} else {
			maxS = Math.max(1, Math.min(sbl, fmu + c2s));
			//maxS = Math.min(sbl, fmu + c2s);
			minS = Math.max(0, fmu - c2s);
			//System.out.println("When (sbl > fmu) maxS "+maxS+" minS "+minS);
		}
		return(item.optimizeLagrangeItemSubProblem(theta, minS, maxS, myDefaultSearchIncrement));
	}
	
	/** Tries to optimize the Lagrange item sub-problem for the supplied item,
	 *  over the supplied range of depot stock levels
	 * 
	 * @param item
	 * @param theta
	 * @param minS
	 * @param maxS
	 * @param delta
	 * @return
	 */
	protected int[] optimizeLagrangeItemSubProblem(VMItem item, double theta, int minS, int maxS, int delta){
		return(item.optimizeLagrangeItemSubProblem(theta, minS, maxS, delta));
	}

	protected double getTotalStockingCost(double theta){
		System.out.println("Theta "+theta);
		List<VMItem> items = myModel.getItems();
		for(VMItem vi: items){
			optimizeLagrangeItemSubProblem(vi, theta);
		}
		return(myModel.getTotalStockingCost());
	}
	
	protected class CostConstraint implements FunctionIfc {
		public double fx(double theta) {
			double tsc = getTotalStockingCost(theta);
			double gap = tsc - myBudget;
			System.out.println("theta = " + theta + " \t Cost(theta) = " + tsc + " \t budget gap = " + gap);
			return gap;
		}
	}
}
