/*
* Created on Jun 30, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.inventory.varimetric;

import jsl.utilities.math.JSLMath;
import jsl.utilities.rootfinding.BisectionRootFinder;
import jsl.utilities.rootfinding.Interval;

/**
 * @author rossetti
 *
 */
public class VMLagrangeIterativeAlgorithm extends VMLagrangeAlgorithm {

	protected ThetaFinder myThetaFinder = new ThetaFinder();
	
	protected Interval myThetaInterval = new Interval(0.0, 1.0);

	protected double myDefaultBudgetPercentTolerance = 0.0001;

	protected double myBudgetTolerance;
	
	public VMLagrangeIterativeAlgorithm(VariMetricModel model, double budget, int numPoints, double budgetTolerance, double minTheta, double maxTheta){
		setVariMetricModel(model);
		setBudget(budget);
		setNumberOfEvaluations(numPoints);
		setBudgetTolerance(budgetTolerance);
		setLagrangeMultiplierInterval(minTheta, maxTheta);
	}

	public void setLagrangeMultiplierInterval(double minTheta, double maxTheta) {
		super.setLagrangeMultiplierInterval(minTheta, maxTheta);
		myThetaInterval.setInterval(minTheta, maxTheta);
	}
	protected void setBudgetTolerance(double tolerance){
		if (tolerance <= 0.0)
			throw new IllegalArgumentException("The budget tolerance must be > 0");
		myBudgetTolerance = tolerance;
	}
	
	protected void searchLagrangeMultiplierInterval(){
		myThetaFinder.setMaximumIterations(myMaxNumEvaluations);
		myThetaFinder.setInterval(myCostConstraint, myThetaInterval);
		myThetaFinder.evaluate();	
		double result = myThetaFinder.getResult();
		double theta = myThetaFinder.getFinalLagrangeMultiplierValue();
		if (!JSLMath.equal(result, theta)){
			// the last optimization was not at the recommended theta, run at theta
			myCostConstraint.fx(theta);
		}
		myFinalTheta = theta;
		if (!myThetaFinder.hasConverged()){
			String st = "The Lagrange Iterative Algorithm did not converge after " + myMaxNumEvaluations + " iterations with budget tolerance = " + myBudgetTolerance + ")\n";
			double tsc = myModel.getTotalStockingCost();
			double gap = tsc - myBudget;
			st = st + "theta = " + myFinalTheta + " \t Cost(theta) = " + tsc + " \t budget gap = " + gap + "\n";
			st = st + "Using best solution found over the iterations";
			LOGGER.warning(st);	
		}
	}

	protected class ThetaFinder extends BisectionRootFinder {

		protected double myGap;
		protected double myTheta = Double.NaN;
				
		protected double evaluateIteration(){
	        result = ( xPos + xNeg) * 0.5;
	        myGap = f.fx(result);
	        if ( myGap > 0 ){
	        	fPos = myGap;
	            xPos = result;
	        }
	        else {
	        	fNeg = myGap;
	            xNeg = result;
	        }
//			System.out.println("f(" + xPos + ") = " + fPos + " \t f(" + xNeg + ") = " + fNeg);
	        return relativePrecision( Math.abs( xPos - xNeg));
		}
		
		public boolean hasConverged(){
			return( super.hasConverged() || (Math.abs(myGap) < myBudgetTolerance));
		}

	    protected void finalizeIterations(){
			// pick the point that has the smallest constraint violation
			if ( fPos < Math.abs(fNeg)){
				myTheta = xPos;
			} else
				myTheta = xNeg;	    	
	    }
	    
	    protected double getFinalLagrangeMultiplierValue(){
	    	return (myTheta);
	    }
	}

}
