/*
* Created on Jun 21, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.inventory.varimetric;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jsl.modeling.ModelElement;
import jslx.supplychain.afosr.generation.BItem;
import jslx.supplychain.afosr.generation.DItem;
import jslx.supplychain.afosr.generation.NetworkConfiguration;
import jslx.supplychain.afosr.varimetricsimulator.SimVMModel;
import jslx.supplychain.inventorylayer.Inventory;

/**
 * @author rossetti
 *
 */
public class VariMetricModel implements VariMetricIfc {

	public static VariMetricModel getVarimetricModel(ModelElement parent, NetworkConfiguration nc){
		
		VariMetricModel vmodel = new VariMetricModel();
		
		//loop through the DItems and add the depot inventory
		for(DItem di: nc.myItems){
			int locationid = 0;// location id
	        String name = di.myName;// item name
	        int itemid = di.myItemID;// item id
	        double cost = di.myCost;// item cost
	        int costlevel = di.myCostLevel;// item cost level
	        int basestocklevel = di.myBaseStockLevel; // item stock level
	        double lambda = di.myLambda;// average demand
	        double repairdelay = di.myRepairDelay;// repair delay
	        double reparability = 1.0;// reparability
	        double shipdelay = 0.0;// ship delay
	        double dibudgetnw = di.getDItemBudget();// item budget at location
	        double dibudgetloc = di.getBudget();// total item level budget
	        int numbasesstocked = di.myNumStockedBases;// num stocked level
	                 
	        VMItem vm = new VMItem(itemid, basestocklevel, cost, repairdelay);
	        vmodel.addVMItem(vm);	        
	        vmodel.mapDItemToVMItem.put(di,vm);
	        
	        for(BItem bi: di.myBItems){
				int locationid_ = bi.myBaseID;// location id
		        String name_ = bi.myName;// item name
		        int itemid_ = bi.myItemID;// item id
		        double cost_ = bi.myCost;// item cost
		        int costlevel_ = bi.myCostLevel;// item cost level
		        int basestocklevel_ = bi.myBaseStockLevel; // item stock level
		        double lambda_ = bi.myLambda;// average demand
		        double repairdelay_ = bi.myRepairDelay;// repair delay
		        double reparability_ = bi.myReparability;// reparability
		        double shipdelay_ = bi.myShipDelay;// ship delay
		        double bibudgetnw_ = bi.getBudget();// item budget at location
		        double bibudgetloc_ = di.getBudget();// total item level budget
                int plevel = bi.myPatternLevel;
                double tp = bi.myHorizonSpan;
                double td = bi.myDeploymentSpan;
                double ts = bi.mySustainmentSpan;
                double tr = bi.myReconfigurationSpan;
                double lMIN = bi.myLambdaMin;
                double lMAX = bi.myLambdaMax;
		        int numbasesstocked_ = 0;// num stocked level	        
		        VMBaseItem vmb = (VMBaseItem)vm.addBaseItem(locationid_,lMIN,
		        		repairdelay_,reparability_,shipdelay_,basestocklevel_,cost_);
		        vmodel.mapBItemToVMBaseItem.put(bi,vmb);				
			}		
		} 		
		return vmodel;
	}
	
	protected ArrayList<VMItem> myItems;
	private Map <DItem, VMItem> mapDItemToVMItem;
	private Map <BItem, VMBaseItem> mapBItemToVMBaseItem;
		
	public VariMetricModel(){
		myItems = new ArrayList<VMItem>();
		mapDItemToVMItem = new HashMap <DItem, VMItem>();
		mapBItemToVMBaseItem = new HashMap <BItem, VMBaseItem>();
	}
	
	public Map<DItem, VMItem> getMapDItemToVMItem() {
		return mapDItemToVMItem;
	}

	public Map<BItem, VMBaseItem> getMapBItemToVMBaseItem() {
		return mapBItemToVMBaseItem;
	}

	/** Creates and returns a new (independent) deep copy of the supplied model
	 * 
	 * @param original
	 * @return
	 */
	public static VariMetricModel newInstance(VariMetricModel original){
		VariMetricModel clone = new VariMetricModel();
		clone.setMapDItemToVMItem(original.getMapDItemToVMItem());
		clone.setMapBItemToVMBaseItem(original.getMapBItemToVMBaseItem());
		for(VMItem item: original.myItems){
			clone.addVMItem(VMItem.newInstance(item));
		}
		return(clone);
	}
	
	private void setMapDItemToVMItem(Map<DItem, VMItem> mapDItemToVMItem) {
		this.mapDItemToVMItem = mapDItemToVMItem;
	}

	private void setMapBItemToVMBaseItem(Map<BItem, VMBaseItem> mapBItemToVMBaseItem) {
		this.mapBItemToVMBaseItem = mapBItemToVMBaseItem;
	}

	/** Creates and adds an item with the supplied parameters to the model
	 * 
	 * @param itemNumber
	 * @param stockLevel
	 * @param cost
	 * @param leadTime
	 * @return The created item
	 */
	public VMItem addVMItem(int itemNumber, int stockLevel, double cost, double leadTime){
		VMItem item = new VMItem(itemNumber, stockLevel, cost, leadTime);
		addVMItem(item);
		return(item);
	}
	
	/** Adds the supplied item to the model.  
	 * 
	 * @param item Must not be null and must not already be in the model
	 */
	public void addVMItem(VMItem item){
		if (item == null)
			throw new IllegalArgumentException("The item was null");
		if (myItems.contains(item))
			throw new IllegalArgumentException("The item is already in the model");
		myItems.add(item);
	}
	
	/** Returns an unmodifiable list of the items
	 * 
	 * @return
	 */
	public List<VMItem> getItems(){
		return(Collections.unmodifiableList(myItems));
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.VariMetricIfc#getTotalEBackOrders()
	 */
	public double getTotalEBackOrders() {
		return getTotalBaseEBackOrders() + getTotalDepotEBackOrders();
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.VariMetricIfc#getTotalBaseEBackOrders()
	 */
	public double getTotalBaseEBackOrders(){
		double t = 0.0;
		for(VMItemIfc i: myItems){
			t = t + i.getTotalBaseEBackOrders();
		}
		return(t);
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.VariMetricIfc#getTotalDepotEBackOrders()
	 */
	public double getTotalDepotEBackOrders(){
		double t = 0.0;
		for(VMItemIfc i: myItems){
			t = t + i.getTotalEBackOrders();
		}
		return(t);
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.VariMetricIfc#getTotalEOnHand()
	 */
	public double getTotalEOnHand() {
		return getTotalBaseEOnHand() + getTotalDepotEOnHand();
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.VariMetricIfc#getTotalBaseEOnHand()
	 */
	public double getTotalBaseEOnHand(){
		double t = 0.0;
		for(VMItemIfc i: myItems){
			t = t + i.getTotalBaseEOnHand();
		}
		return(t);
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.VariMetricIfc#getTotalDepotEOnHand()
	 */
	public double getTotalDepotEOnHand(){
		double t = 0.0;
		for(VMItemIfc i: myItems){
			t = t + i.getTotalEOnHand();
		}
		return(t);
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.VariMetricIfc#getTotalStockingCost()
	 */
	public double getTotalStockingCost() {
		return getTotalBaseStockingCost() + getTotalDepotStockingCost();
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.VariMetricIfc#getTotalBaseStockingCost()
	 */
	public double getTotalBaseStockingCost(){
		double t = 0.0;
		for(VMItemIfc i: myItems){
			t = t + i.getTotalBaseStockingCost();
		}
		return(t);
	}
	
	/* (non-Javadoc)
	 * @see varimetric2.VariMetricIfc#getTotalDepotStockingCost()
	 */
	public double getTotalDepotStockingCost(){
		double t = 0.0;
		for(VMItemIfc i: myItems){
			t = t + i.getStockingCost();
		}
		return(t);
	}

	/* (non-Javadoc)
	 * @see varimetric2.VariMetricIfc#getNumItems()
	 */
	public int getNumItems() {
		return myItems.size();
	}
	
	/** Sets all stocking levels to zero
	 * 
	 *
	 */
	public void setAllStockingLevelToZero(){
		for(VMItem vi: myItems){
			vi.setStockLevel(0);
			for(VMBaseItem b: vi.getBaseItems()){
				b.setStockLevel(0);
			}
		}				
	}
	
	public double getMinUnitCostAcrossAllItems(){
		double min = Double.MAX_VALUE;
		for(VMItem i:myItems){
			double uc = i.getMinUnitCostAcrossBases();
			if (uc < min)
				min = uc;
		}
		return(min);
	}
		
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append("Varimetric model results: \n");
		sb.append("==================================================== \n");
		sb.append("Across item Aggregate Results\n");
		sb.append("Total Stocking Cost: " + getTotalStockingCost() + "\n");
		sb.append("Total Expected BackOrders " + getTotalEBackOrders() + "\n");
		sb.append("Total Expected Onhand: " + getTotalEOnHand() + "\n");
		sb.append("Number of items: " + getNumItems() + "\n");
		sb.append("Depot level Aggregate Results across items\n");
		sb.append("Total Depot Stocking Cost: " + getTotalDepotStockingCost() + "\n");
		sb.append("Total Expected Depot Backorders: " + getTotalDepotEBackOrders() + "\n");
		sb.append("Total Expected Depot Onhand: " + getTotalDepotEOnHand() + "\n");
		sb.append("Base level Aggregate Results across items\n");
		sb.append("Total Base Stocking Cost: " + getTotalBaseStockingCost() + "\n");
		sb.append("Total Expected Base Backorders: " + getTotalBaseEBackOrders() + "\n");
		sb.append("Total Expected Base Onhand: " + getTotalBaseEOnHand() + "\n");
		sb.append("\nIndividual item results\n");
		for(VMItem i:myItems){
			sb.append("Item number: " + i.getItemNumber() + "\n");
			sb.append(i.toString());
		}
		sb.append("==================================================== \n");
		return(sb.toString());
	}

}
