/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

import java.util.Observer;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.queue.*;
import jsl.modeling.elements.variable.TimeWeighted;

/**
 *
 */
public abstract class BackLogPolicyAbstract extends ModelElement implements BackLogPolicyIfc {

	/** Tracks the amount of individual inventory units
	 *  required to fill back ordered demands
	 * 
	 */
	protected TimeWeighted myAmtBackLogged;
	
	/** Holds the backorderd demand
	 * 
	 */
	protected Queue myBackLogQ;
	
	/** Refers to the inventory that this policy is
	 *  controlling
	 * 
	 */
	protected Inventory myInventory;

	/** Creates a BackLogPolicy
	 * @param parent, the policy's parent model element
	 * @param inventory, the policy's inventory 
	 */
	public BackLogPolicyAbstract(ModelElement parent) {
		this(parent, Queue.FIFO, null);
	}

	/** Creates a BackLogPolicy
	 * @param parent, the policy's parent model element
	 * @param inventory, the policy's inventory 
	 * @param name, the policy's name
	 */
	public BackLogPolicyAbstract(ModelElement parent, String name) {
		this(parent, Queue.FIFO, name);
	}
	
	/** Creates a BackLogPolicy
	 * @param parent, the policy's parent model element
	 * @param inventory, the policy's inventory 
	 * @param backLogQ, the policy's back log queue
	 */
	public BackLogPolicyAbstract(ModelElement parent, QueueDiscipline discipline) {
		this(parent, discipline, null);
	}
	
	/** Creates a BackLogPolicy
	 * @param parent, the policy's parent model element
	 * @param inventory, the policy's inventory 
	 * @param backLogQ, the policy's back log queue
	 * @param name, the policy's name
	 */
	public BackLogPolicyAbstract(ModelElement parent, QueueDiscipline discipline, String name) {
		super(parent, name);
		setBackLoggedDemandQueue(new Queue(this, getName() + "-DemandBackLogQ", discipline));
		myAmtBackLogged = new TimeWeighted(this, getName() + "-AmtBackLogged");
	}

	/**
	 * @return
	 */
	public final double getAmountBackLogged() {
		return myAmtBackLogged.getValue();
	}
	
	/**
	 * @return
	 */
	public final int getNumberOfDemandsBackLogged() {
		return myBackLogQ.size();
	}

	/** Allows an observer to attached to the amount backlogged
	 * 
	 * @param observer
	 */
	public final void addAmountBackLoggedObserver(Observer observer){
		myAmtBackLogged.addObserver(observer);
	}

	/** Allows an observer to attached to the amount backlogged
	 * 
	 * @param observer
	 */
	public final void addBackLoggedQueueTimeObserver(Observer observer){
		myBackLogQ.addTimeInQueueObserver(observer);
	}

	/** Allows an observer to attached to the amount backlogged
	 * 
	 * @param observer
	 */
	public final void addNumInBackLoggedQueueObserver(Observer observer){
		myBackLogQ.addNumberInQueueObserver(observer);
	}
	
	/** Asks the backlogger to backlog the demand
	 * 
	 * @param demand Must not be null
	 */
	abstract protected void backlog(Demand demand);
	
	/** Asks the backlogger to try to fill
	 *  backlog demands
	 */
	abstract protected void fillBackLogs();
	
	/** This method should return a reference to the next
	 *  demand that needs to be filled.  It should not
	 *  remove the demand from the backlog queue
	 * 
	 * @return The next demand or null if no demand
	 */
	abstract protected Demand nextBackLog();
	
	/** Checks if there is next backlog
	 * @return
	 */
	abstract protected boolean hasNextBackLog();
	
	/** This method should remove the demand that 
	 *  was returned from the last call to nextBackLog()
	 */
	abstract protected void removeNextBackLog();
	
	/** Asks the inventory to determine how much of the
	 *  backlog can be filled.
	 * 
	 * @param d
	 * @return The amount that can be supplied
	 */
	protected final int processBackLog(Demand d){
		return(myInventory.processBackLog(d));
	}
		
	/** Asks the inventory to fill an individual demand
	 *  that has been backlogged, returns true if the
	 *  backlog has been completely filled
	 * 
	 * @param demand
	 * @return The amount that was filled
	 */
	protected final boolean fillBackLog(Demand demand){
		return(myInventory.fillBackLog(demand));
	}
	
	/**
	 * @return Returns the myBackLoggedDemands.
	 */
	protected final Queue getBackLoggedDemandQueue() {
		return myBackLogQ;
	}
	
	/**
	 * @param myBackLoggedDemands The myBackLoggedDemands to set.
	 */
	private final void setBackLoggedDemandQueue(Queue queue) {
        if (queue == null)
            throw new IllegalArgumentException("The Queue must be non-null");
		myBackLogQ = queue;
	}
	
	/**
	 * @return Returns the InventoryAbstract.
	 */
	protected final Inventory getInventory() {
		return myInventory;
	}
	
	/**
	 * @param inventory The inventory to set.
	 */
	protected final void setInventory(Inventory inventory) {
        if (inventory == null)
            throw new IllegalArgumentException("The Inventory must be non-null");
        myInventory = inventory;
	}
	

}
