/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.queue.Queue;


/** This class processes the backlogged demands in FIFO order
 *  If the demand at the head of the queue cannot be filled in full
 *  other backlogged demands continue to wait even if there is inventory
 *  for them to be filled.
 *  
 *
 */
public class DefaultBackLogPolicy extends BackLogPolicyAbstract {

	/** Creates a BackLogPolicy
	 * @param parent, the policy's parent model element
	 * @param inventory, the policy's inventory 
	 */
	public DefaultBackLogPolicy(ModelElement parent) {
		this(parent, null);
	}
			
	/** Creates a BackLogPolicy
	 * @param parent, the policy's parent model element
	 * @param inventory, the policy's inventory 
	 * @param backLogQ, the policy's back log queue
	 * @param name, the policy's name
	 */
	public DefaultBackLogPolicy(ModelElement parent, String name) {
	    super(parent, Queue.FIFO, name);
	}

	/* (non-Javadoc)
	 * @see jsl.examples.inventory.BackLogPolicyAbstract#backlog(jsl.examples.inventory.Demand)
	 */
	protected final void backlog(Demand demand) {
		// if the demand is already backlogged we don't backlog it again
		if (demand.getBackLoggedIndicatorFlag() == false){
			// demand has not yet been backlogged, set the indicator
			demand.setBackLoggedIndicatorFlag();
			// record the units back logged
			myAmtBackLogged.increment(demand.getRemainingDemand());
			// just put the demand in the queue
			myBackLogQ.enqueue(demand);	
		}
	}
	
	/* (non-Javadoc)
	 * @see jsl.examples.inventory.BackLogPolicyAbstract#getNextDemand(jsl.examples.inventory.Inventory, jsl.modeling.elements.queue.Queue)
	 */	
	protected final void fillBackLogs(){
		
		// process the back logs as long as there are backlogs
		// that can be filled
		boolean stopFlag = false;
		while(hasNextBackLog() && (stopFlag == false)){
			// get the next backlog to process
			Demand d = nextBackLog();
			// process the backlog, determine how much 
			// can be supplied to the demand by the inventory
			int canSupply = processBackLog(d);
			if (canSupply > 0){ // something can be supplied
				// go ahead and let this backlogged demand have
				// all of what can be supplied
				myAmtBackLogged.decrement(canSupply);
				boolean filled = fillBackLog(d);
				if (filled){ // if it is filled remove it from the backlog queue
					removeNextBackLog();
				} else {
					// the demand at the head of the list could not 
					// be supplied in full, stop processing further demands
					stopFlag = true;					
				}
			} else // nothing could be supplied to the first demand, must stop processing
				stopFlag = true;
		}
		
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.BackLogPolicyAbstract#nextBackLog()
	 */
	protected final Demand nextBackLog() {
		return((Demand)myBackLogQ.peekNext());
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.BackLogPolicyAbstract#hasNextBackLog()
	 */
	protected final boolean hasNextBackLog() {
		return !myBackLogQ.isEmpty();
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.BackLogPolicyAbstract#removeNextBackLog()
	 */
	protected final void removeNextBackLog() {
		myBackLogQ.removeNext();
	}
}
