/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

import java.util.LinkedList;
import java.util.List;

import jsl.modeling.Model;
import jsl.modeling.elements.queue.QObject;


/**
 * 
 */
public class Demand extends QObject {

	//  ===========================================
	//      CLASS ATTRIBUTES
	//  ===========================================

	/** The default size of the QObject cache
	 */
	public final static int DEFAULT_CACHE_LIMIT = 50;

	/** Caches used Demand for
	 *  reuse, perhaps gaining some efficiency
	 */
	private static int myDemandCacheSizeLimit = DEFAULT_CACHE_LIMIT;

	/** A reference to the list for caching Demands
	 */
	private static List<Demand> myDemandCache = new LinkedList<Demand>();;
	
    //  ===========================================
    //      OBJECT ATTRIBUTES
    //  ===========================================

	/**
	 * CREATED: Represents that the demand request is newly created
	 * FILL_ALL:  Represents that the request can be filled in full
	 * BACKLOG_PARTIAL:  Represents that the demand request should be backlogged for a partial fill
	 * BACKLOG_NO_PARTIAL:  Represents that the demand request should be backlooged with no partial filling
	 * LOST_PARTIAL:  Represents that the demand request should not be backlogged for a partial fill, 
	 *                i.e. that it will lose part of the request
	 * LOST_ALL:  Represents that the demand request should not be backlogged and no partial filling.
	 *            i.e. that it will lose all of the request
	 */
	public static enum RequestStatus {CREATED, FILL_ALL, BACKLOG_PARTIAL, BACKLOG_NO_PARTIAL, LOST_PARTIAL, LOST_ALL};
	
	/** Holds the current RequestStatus of the demand
	 * 
	 */
	private RequestStatus myRequestStatus;

	/** The amount of demand still needed
	 */
	private int myAmountNeeded;
	
	/** The amount of demand originally needed
	 */
	private int myOriginalAmtDemanded;
	
	/** The amount of demand that can be supplied
	 *  to the original request on the first fill opportunity
	 */
	private int myInitialRequestFillAmount;
	
	/** When a demand is backlogged, it may be processed
	 *  a number of times until it is completely filled. This
	 *  variable holds the amount that the demand will get
	 *  after it is processed by the backlog filling logic.
	 */
	private int myBacklogFillAmount;
	
	/** The amount of the original demand that has been
	 *  filled.  When this equals the original amount
	 *  the demand has been filled.
	 */
	private int myAmountFilled;
	
	/** A reference to the ultimate provider for this demand 
	 */
	private InventoryProviderIfc myInventoryProvider;
	
	/** A reference to the ultimate consumer of this demand
	 */
	private InventoryCustomerIfc myInventoryCustomer;

	/** A reference to the shipper of this demand
	 */
	private InventoryShipperIfc myInventoryShipper;
	
	/** whether or not the demand has been filled
	 *  true means filled
	 */
	private boolean myFilledStatus;
	
	/** How long it took to fill the demand
	 */
	private double myTimeToFill;

	/** Whether or not the demand is willing
	 *  to wait in a backlog (true means willing)
	 *  The default is true.
	 */
	private boolean myAllowBackLoggingFlag;
	
	/** Whether or not the demand requires all
	 *  original units demanded to be filled at one
	 *  time. (true means that it is allowed)
	 */
	private boolean myAllowPartialFillingFlag;
	
	/** Indicates whether all or some of the demand has been lost
	 */
	private boolean myLostDemandFlag;
		
	/** The type of item required by the demand
	 */
	private ItemType myItemType;
	
	/** Indicates that this demand is serving as a
	 *  replenishment for some inventory (true means it
	 *  is a replenishment)
	 */
	private boolean myReplenishmentFlag;
	
	/** Indicates whether or not the demand has been
	 *  backlogged.
	 */
	private boolean myBackLoggedIndicatorFlag;
	
	/**
	 * @param entityType
	 */
	public Demand(ItemType itemType) {
		this(itemType, 1, null);
	}

	/**
	 * @param entityType
	 */
	public Demand(ItemType itemType, int amtDemanded) {
		this(itemType, amtDemanded, null);
	}

	/** Creates a demand, with defaults:
	 *   - Allow backlogging
	 *   - Allow partial filling
	 *   
	 * @param name
	 * @param entityType
	 */
	public Demand(ItemType itemType, int amtDemanded, String name) {
		super(name);
		initializeDemand();
		setItemType(itemType);
		setOriginalAmountDemanded(amtDemanded);
	}

	/**
	 * Returns the state of the demand. The states are defined by the public
	 * static int variables declared for the Demand
	 * 
	 * @return
	 */
	public final RequestStatus getRequestStatus() {
		return (myRequestStatus);
	}

	/** Sets the demand state to the supplied constant
	 *  Throws illegal argument exception if the supplied state 
	 *  is not a valid demand state
	 * 
	 * @param status
	 */
	protected void setRequestStatus(RequestStatus status) {

		switch (status) {
			case FILL_ALL:
				myRequestStatus = RequestStatus.FILL_ALL;
				break;
			case BACKLOG_PARTIAL:
				myRequestStatus = RequestStatus.BACKLOG_PARTIAL;
				break;
			case BACKLOG_NO_PARTIAL:
				myRequestStatus = RequestStatus.BACKLOG_NO_PARTIAL;
				break;
			case LOST_PARTIAL:
				myRequestStatus = RequestStatus.LOST_PARTIAL;
				break;
			case LOST_ALL:
				myRequestStatus = RequestStatus.LOST_ALL;
				break;
			case CREATED:
				myRequestStatus = RequestStatus.CREATED;
				break;
			default:
				throw new IllegalArgumentException("The demand request's status was illegal: " + status);		
		}
	}
	
	/** The amount of the demand's request that can be filled immediately
	 * @return Returns the requestFillAmount.
	 */
	public final int getInitialRequestFillAmount() {
		return myInitialRequestFillAmount;
	}
	
	/** The amount of the demand's request that can be filled immediately 
	 * @param requestFillAmount The requestFillAmount to set.
	 */
	protected final void setInitialRequestFillAmount(int requestFillAmount) {
		if (requestFillAmount < 0)
			throw new IllegalArgumentException(
					"Invalid argument. Amount must be >= 0");
		myInitialRequestFillAmount = requestFillAmount;
	}

	/** The amount of the demand's request that can be filled
	 *  after the backlog processing
	 * @return Returns the backlogFillAmount.
	 */
	public final int getBackLogFillAmount() {
		return myBacklogFillAmount;
	}
	
	/** The amount of the demand's backlogged request that can be filled for this
	 *  backlog processing.
	 * @param backLogAmount The amount of the back log that will be filled
	 */
	protected final void setBackLogFillAmount(int backLogAmount) {
		if (backLogAmount < 0)
			throw new IllegalArgumentException("Invalid argument. Amount must be >= 0");
		if (backLogAmount > myAmountNeeded)
			throw new IllegalArgumentException("Invalid argument. Amount must be <= the amount needed");
		myBacklogFillAmount = backLogAmount;
	}
	
	/** Checks if demand has the supplied request status
	 * 
	 * @param status
	 * @return
	 */
	public final boolean isRequestStatus(RequestStatus status) {
		return(myRequestStatus == status);
	}

	/** Gets the amount of demand still remaining to be filled
	 *  When this reaches zero the demand will be "filled"
	 *  
	 * @return the amount remaining to be filled
	 */
	public final int getRemainingDemand() {
		return myAmountNeeded;
	}

	/** Gets the item type for the demand
	 * @return the ItemType for this demand
	 */
	public final ItemType getItemType(){
		return(myItemType);
	}
	
	/** Sets the item type for the demand
	 * @param the ItemType for this demand, if null
	 *  the default item type is used
	 */
	protected final void setItemType(ItemType itemType){
		if (itemType == null)
			throw new IllegalArgumentException("Invalid argument.ItemType itemType was null");

		myItemType = itemType;
	}
	
	/**
	 * @return Returns the inventoryProvider.
	 */
	public final InventoryProviderIfc getInventoryProvider() {
		return myInventoryProvider;
	}
	
	/**
	 * @param inventoryProvider The inventoryProvider to set.
	 */
	protected final void setInventoryProvider(InventoryProviderIfc inventoryProvider) {
		if (inventoryProvider == null)
			throw new IllegalArgumentException(
					"Invalid argument. InventoryProviderIfc inventoryProvider must be non-null.");

		myInventoryProvider = inventoryProvider;
	}
	
	/**
	 * @return Returns the myInventoryConsumer.
	 */
	public final InventoryCustomerIfc getInventoryCustomer() {
		return myInventoryCustomer;
	}
	
	/**
	 * @param inventoryConsumer The nventoryConsumer to set.
	 */
	protected final void setInventoryCustomer(InventoryCustomerIfc customer) {
		if (customer == null)
			throw new IllegalArgumentException(
					"Invalid argument. InventoryCustomerIfc customer must be non-null.");

		myInventoryCustomer = customer;
	}
	
	/**
	 * @return Returns the inventoryShipper.
	 */
	public final InventoryShipperIfc getInventoryShipper() {
		return myInventoryShipper;
	}
	
	/**
	 * @param inventoryShipper The inventoryShipper to set.
	 */
	protected final void setInventoryShipper(InventoryShipperIfc inventoryShipper) {
		if (inventoryShipper == null)
			throw new IllegalArgumentException(
					"Invalid argument. InventoryShipperIfc inventoryShipper must be non-null.");
		
		if (!inventoryShipper.canShip(this))
   			throw new IllegalArgumentException("The supplied shipper cannot handle the demand " + this);
		 
		myInventoryShipper = inventoryShipper;
		
	}
	
	/** The original amount of demand requested.  Once
	 *  this quantity is supplied to the demand then the 
	 *  demand is considered "filled"
	 * 
	 * @return the amount remaining to be filled
	 */
	public final int getOriginalAmountDemanded() {
		return myOriginalAmtDemanded;
	}

	/** Return the amount that has been supplied
	 *  to the orginal demand amount.  When this reaches
	 *  the original amount demanded then the demand
	 *  is considered "filled"
	 *  
	 * @return Returns the amountFilled.
	 */
	public final int getAmountFilled() {
		return myAmountFilled;
	}
	
	/** Returns the weight of the amount supplied
	 *  to this demand
	 * 
	 * @return, the weight as a double
	 */
	public final double getWeight(){
		double w = 0.0;
		if (getItemType() != null)
			w = getItemType().getWeight();
		return(w*getAmountFilled());
	}
	
	/** Returns the cube of the amount supplied
	 *  to this demand
	 * 
	 * @return, the cube as a double
	 */
	public final double getCube(){
		double c = 0.0;
		if (getItemType() != null)
			c = getItemType().getCube();
		return(c*getAmountFilled());		
	}
	
	/** True indicates that this demand can be backordered
	 * 
	 * @return Returns the allowBackLoggingFlag.
	 */
	public final boolean getAllowBackLoggingFlag() {
		return myAllowBackLoggingFlag;
	}
	
	/** True indicates that this demand can be backordered
	 * 
	 * @param allowBackLoggingFlag The allowBackLoggingFlag to set.
	 */
	public final void setAllowBackLoggingFlag(boolean allowBackLoggingFlag) {
		myAllowBackLoggingFlag = allowBackLoggingFlag;
	}
	
	/**True indicates that this demand allows partial filling
	 * 
	 * @return Returns the myAllowPartialFillingFlag.
	 */
	public final boolean getAllowPartialFillingFlag() {
		return myAllowPartialFillingFlag;
	}
	
	/** True indicates that this demand allows partial filling
	 * @param allowPartialFillingFlag The allowPartialFillingFlag to set.
	 */
	public final void setAllowPartialFillingFlag(boolean allowPartialFillingFlag) {
		myAllowPartialFillingFlag = allowPartialFillingFlag;
	}
	
	/**
	 * Returns true if all the original units demanded 
	 * have been supplied
	 * 
	 * @return
	 */
	public final boolean getFilledStatus() {
		return (myFilledStatus);
	}

	/** The amount of time taken to completely fill the
	 *  demand.
	 * @return Returns the amount of time or Double.NaN if not yet filled
	 */
	public final double getTotalTimeToFill() {
		return myTimeToFill;
	}

	/**
	 * If the amount supplied is < 0 an illegal argument exception is thrown. If
	 * the amount supplied is > amount demanded then only the amount needed will
	 * be used from that supplied. 
	 * 
	 * If the amount supplied is not equal to the amount originally demanded
	 * and partial order filling is not permitted then an illegal argument exception
	 * will be thrown.
	 * 
	 * Once filled status becomes true, this method should not be called
	 * 
	 * @param amtSupplied   The amount to supply to the demand.
	 * @return The amount of the amtSupplied that was used by the demand
	 * this becomes 0 when the demand is filled.
	 */
	public final int fillDemand(int amtSupplied) {
		if (amtSupplied < 0)
			throw new IllegalArgumentException(
					"Invalid argument. Amount filled must be >= 0");
		
		if (myAllowPartialFillingFlag == false){
			if (amtSupplied != myOriginalAmtDemanded)
				throw new IllegalArgumentException(
				"Invalid argument. Amount supplied must be equal to " +
				"the amount demanded if partial order filling is not allowed");
		}
		
		int amtUsed = 0;
		if (myFilledStatus == false) {
			if (amtSupplied > myAmountNeeded)
				amtUsed = myAmountNeeded;
			else
				amtUsed = amtSupplied;
			myAmountNeeded = myAmountNeeded - amtUsed;
			myAmountFilled = myAmountFilled + amtUsed;
			if (myAmountFilled == myOriginalAmtDemanded){
				setFillStatus();
			}
		}
		
		return (amtUsed);
	}
	
	/**
	 * @param amtDemanded
	 *            The amtDemanded to set, must be > 0
	 */
	protected final void setOriginalAmountDemanded(int amtDemanded) {
		if (amtDemanded <= 0)
			throw new IllegalArgumentException(
					"Invalid argument. Amount demanded must be > 0");

		myAmountNeeded = amtDemanded;
		myOriginalAmtDemanded = amtDemanded;
		myAmountFilled = 0;
	}
	
	/** True means that this demand will never be filled,
	 *  i.e. it is lost
	 * @return Returns the lostDemandFlag.
	 */
	public final boolean getLostDemandFlag() {
		return myLostDemandFlag;
	}
	
	/** True means that this demand will never be filled,
	 *  i.e. it is lost
	 * @param lostDemandFlag The lostDemandFlag to set.
	 */
	public final void setLostDemandFlag(boolean lostDemandFlag) {
		myLostDemandFlag = lostDemandFlag;
	}
	
	/** True means that the demand had been backlogged
	 * @return Returns the backLoggedIndicatorFlag.
	 */
	public final boolean getBackLoggedIndicatorFlag() {
		return myBackLoggedIndicatorFlag;
	}
	
	/** Sets the backlog indicator flag to true
	 *  True means that the demand had been backlogged
	 * @param flag The flag to set.
	 */
	public final void setBackLoggedIndicatorFlag() {
		myBackLoggedIndicatorFlag = true;
	}

	/**
	 * @return
	 */
	public final boolean getReplenishmentFlag() {
		return myReplenishmentFlag;
	}

	/**
	 * @param b
	 */
	public final void setReplenishmentFlag(boolean b) {
		myReplenishmentFlag = b;
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	/** Nullifies the object references within
	 *  the demand, its order, inventory, state change listeners, 
	 *  and item type are set to null to facilitate
	 *  recyling.
	 *  
	 *  Initializes all variables to their defaults.
	 *  
	 *  This method should only be used when the demand
	 *  will never be used again.
	 */
	protected final void initializeDemand(){
		myInventoryProvider = null;
		myInventoryCustomer = null;
		myItemType = null;
		myOriginalAmtDemanded = 0;
		myRequestStatus = RequestStatus.CREATED;
		myFilledStatus = false;
		myTimeToFill = Double.NaN;
		myAllowBackLoggingFlag = true;
		myAllowPartialFillingFlag = false;
		myLostDemandFlag = false;
		myReplenishmentFlag = false;
		myBackLoggedIndicatorFlag = false;
	}
	
	//  ===========================================
	//     PRIVATE METHODS
	//  ===========================================

	private final void setFillStatus(){
		myFilledStatus = true;
		myTimeToFill = Model.getTime() - getCreateTime();
	}
	
	//  ===========================================
	//      CLASS METHODS
	//  ===========================================

	public final static Demand create(ItemType itemType, int amtDemanded){
		return(getDemandFromCache(itemType, amtDemanded));
	}
	
	public final static void recycle(Demand demand){
		// initialize the demand aspects
		demand.initializeDemand();
		addDemandToCache(demand);
	}
	
	/** Gets the limit for the size of the Demand cache
	 * 
	 * @return
	 */
	public static final int getDemandCacheSizeLimit(){
		return myDemandCacheSizeLimit;
	}
	
	/** Gets the current size of the Demand cache
	 * 
	 * @return
	 */
	public static final int getDemandCacheSize(){
		return(myDemandCache.size());
	}
	   
	/** Sets the size of the Demand cache, if the size is < 1
	 *  then the default size is used.
	 * 
	 * @param size
	 */
	public static final void setDemandCacheSizeLimit(int size) {
		if (size <= 0)
			myDemandCacheSizeLimit = DEFAULT_CACHE_LIMIT;
		else
			myDemandCacheSizeLimit = size;
	}

	//  ===========================================
	//      PRIVATE STATIC METHODS
	//  ===========================================

	private static void addDemandToCache(Demand d) {
		if (myDemandCache.size() < myDemandCacheSizeLimit) { // can add element to cache
			myDemandCache.add(d);
		}
	}
	
	private static Demand getDemandFromCache(ItemType itemType, int amtDemanded) {
		// check if cache is empty, if so make new Demand
		Demand d;

		if (myDemandCache.isEmpty()){
			d = new Demand(itemType, amtDemanded);
		}
		else { // cache not empty get event from cache
			d = myDemandCache.remove(0);
			d.setItemType(itemType);
			d.setOriginalAmountDemanded(amtDemanded);
		}

		// initialize the QObject aspects
		d.initialize(null);

		return (d);
	}
}
