/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.*;
import jsl.modeling.elements.EventGenerator;
import jsl.modeling.elements.EventGeneratorIfc;
import jsl.modeling.elements.EventGeneratorListenerIfc;
import jsl.utilities.random.RandomIfc;

/**
 *
 */
public class DemandGenerator extends ModelElement implements InventoryCustomerIfc, EventGeneratorIfc {

	/** Indicated whether or not the DemandGenerator will
	 *  permit backlogging of its demand requests 
	 */
	private boolean myPermitBackLoggingFlag;
	
	/** Indicated whether or not the DemandGenerator will
	 *  permit partial filling of its demand requests 
	 */
	private boolean myPermitPartialFillingFlag;

	/** The underlying event generator
	 * 
	 */
	private EventGenerator myEventGenerator;

	/** The underlying event generator listener
	 * 
	 */
	private DemandCreationListener myDemandCreationListener;

	/** Used to determine the amount of demand at each creation
	 * 
	 */
	private RandomIfc myAmtDistribution;
	
	/** Used to determine the amount of demand at each creation
	 * 
	 */
	private RandomVariable myAmtRV;
	
	/** Fills the generated demand
	 * 
	 */
	private InventoryProviderIfc myInventoryProvider;
		
	/** used to identify the type of demand that is being generated
	 */
	private ItemType myItemType;
		
	/**
	 * @param parent
	 */
	public DemandGenerator(ModelElement parent) {
		this(parent, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public DemandGenerator(ModelElement parent, String name) {
		this(parent, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * @param parent
	 * @param name
	 * @param timeUntilFirst
	 */
	public DemandGenerator(ModelElement parent,   
			RandomIfc timeUntilFirst) {
		this(parent, timeUntilFirst, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}
	
	/**
	 * @param parent
	 * @param name
	 * @param timeUntilFirst
	 */
	public DemandGenerator(ModelElement parent,  
			RandomIfc timeUntilFirst, String name) {
		this(parent, timeUntilFirst, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * @param parent
	 * @param name
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 */
	public DemandGenerator(ModelElement parent,  
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext) {
		this(parent, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}
	
	/**
	 * @param parent
	 * @param name
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 */
	public DemandGenerator(ModelElement parent,  
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext, String name) {
		this(parent, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * @param parent
	 * @param name
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 */
	public DemandGenerator(ModelElement parent,  
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext,
			long maxNum, String name) {
		this(parent, timeUntilFirst, timeUntilNext, maxNum, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * @param parent
	 * @param name
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 * @param timeUntilLast
	 */
	public DemandGenerator(ModelElement parent,  
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext,
			long maxNum, double timeUntilLast, String name) {
		super(parent, name);
				
		myDemandCreationListener = new DemandCreationListener();
		
		myEventGenerator = new EventGenerator(this, myDemandCreationListener, timeUntilFirst,
				timeUntilNext, maxNum, timeUntilLast, getName() + " Demand Generator");

	}

	/**
	 * @return Returns the distribution governing the amount of demand generated
	 */
	public final RandomIfc getAmountDistribution() {
		return myAmtDistribution;
	}
	
	/**
	 * @param myAmtDistribution The myAmtDistribution to set.
	 */
	public final void setAmountDistributionInitialRandomSource(RandomIfc distribution) {
        if (distribution == null)
            throw new IllegalArgumentException("Invalid argument. RandomIfc was null");
        
        if (myAmtRV == null){
        	myAmtRV = new RandomVariable(this, distribution);
        } else
        	myAmtRV.setInitialRandomSource(distribution);
        
		myAmtDistribution = distribution;
	}
	
	/**
	 * @return Returns the InventoryProviderIfc
	 */
	public final InventoryProviderIfc getInventoryProvider() {
		return myInventoryProvider;
	}
	
	/** The type of item to generate 
	 * 
	 * @return Returns the itemType, may be null
	 */
	public final ItemType getItemType() {
		return myItemType;
	}
	
	/** The type and provider must not be null.  The provider must supply
	 *  the given type.
	 *  
	 *  1) type must not be null
	 *  2) provider must not be null
	 *  3) permitBL, permitPF, and type must be compatible with the provider
	 * 
	 * @param type	The type of inventory item to be generated
	 * @param permitBL  Whether or not generator makes demands that allow backlogging
	 * @param permitPF  Whether or not generator makes demands that allow partial filling
	 * @param provider The desired InventoryProvider for the generator
	 *
	 */
	public final void setDemandRequirements(ItemType type, boolean permitBL, boolean permitPF, InventoryProviderIfc provider) {
        if (type == null)
            throw new IllegalArgumentException("Invalid argument. ItemType was null");
        
		myItemType = type;

        if (provider == null)
            throw new IllegalArgumentException("Invalid argument. InventoryProviderIfc was null");
        
		myInventoryProvider = provider;

		myPermitBackLoggingFlag = permitBL;
		
		myPermitPartialFillingFlag = permitPF;

        // ask the provider if it provides inventory with the given requirements
        if (!provider.providesInventory(type, this))
            throw new IllegalArgumentException("Invalid argument. InventoryProviderIfc does not provide the supplied requirements.");
       		
	}
	
	/**
	 * @return
	 */
	public final double getEndingTime() {
		return myEventGenerator.getEndingTime();
	}
	
	/**
	 * @return
	 */
	public final long getMaxNumActions() {
		return myEventGenerator.getMaximumNumberOfEvents();
	}
	/**
	 * @return
	 */
	public final boolean isGeneratorDone() {
		return myEventGenerator.isGeneratorDone();
	}

	
	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getGenerationEndingTimeForReplications()
	 */
	public double getGenerationEndingTimeForReplications() {
		return myEventGenerator.getGenerationEndingTimeForReplications();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getMaximumNumberOfEvents()
	 */
	public long getMaximumNumberOfEvents() {
		return myEventGenerator.getMaximumNumberOfEvents();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getNumberOfEventsGenerated()
	 */
	public long getNumberOfEventsGenerated() {
		return myEventGenerator.getNumberOfEventsGenerated();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getTimeUntilFirstEventForReplications()
	 */
	public RandomIfc getTimeUntilFirstEventForReplications() {
		return myEventGenerator.getTimeUntilFirstEventForReplications();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#isSuspended()
	 */
	public boolean isSuspended() {
		return myEventGenerator.isSuspended();
	}

	/**
	 * 
	 * @see jsl.modeling.elements.EventGeneratorIfc#resume()
	 */
	public void resume() {
		myEventGenerator.resume();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGenerator#getStartOnInitializeFlag()
	 */
	public final boolean getStartOnInitializeFlag() {
		return myEventGenerator.getStartOnInitializeFlag();
	}

	/**
	 * @param flag
	 * @see jsl.modeling.elements.EventGenerator#setStartOnInitializeFlag(boolean)
	 */
	public final void setStartOnInitializeFlag(boolean flag) {
		myEventGenerator.setStartOnInitializeFlag(flag);
	}

	/**
	 * 
	 * @see jsl.modeling.elements.EventGenerator#turnOnGenerator()
	 */
	public final void turnOnGenerator() {
		myEventGenerator.turnOnGenerator();
	}

	/**
	 * @param t
	 * @see jsl.modeling.elements.EventGenerator#turnOnGenerator(double)
	 */
	public final void turnOnGenerator(double t) {
		myEventGenerator.turnOnGenerator(t);
	}

	/**
	 * @param r
	 * @see jsl.modeling.elements.EventGenerator#turnOnGenerator(jsl.utilities.random.RandomIfc)
	 */
	public void turnOnGenerator(RandomIfc r) {
		myEventGenerator.turnOnGenerator(r);
	}

	/**
	 * @param endingTime
	 * @see jsl.modeling.elements.EventGeneratorIfc#setEndingTime(double)
	 */
	public void setEndingTime(double endingTime) {
		myEventGenerator.setEndingTime(endingTime);
	}

	/**
	 * @param endingTime
	 * @see jsl.modeling.elements.EventGeneratorIfc#setGenerationEndingTimeForReplications(double)
	 */
	public void setGenerationEndingTimeForReplications(double endingTime) {
		myEventGenerator.setGenerationEndingTimeForReplications(endingTime);
	}

	/**
	 * @param maxNum
	 * @see jsl.modeling.elements.EventGeneratorIfc#setMaximumNumberOfEvents(long)
	 */
	public void setMaximumNumberOfEvents(long maxNum) {
		myEventGenerator.setMaximumNumberOfEvents(maxNum);
	}

	/**
	 * @param timeBtwEvents
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEvents(jsl.utilities.random.RandomIfc, long)
	 */
	public void setTimeBetweenEvents(RandomIfc timeBtwEvents, long maxNumEvents) {
		myEventGenerator.setTimeBetweenEvents(timeBtwEvents, maxNumEvents);
	}

	/**
	 * @param timeUntilNext
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEvents(jsl.utilities.random.RandomIfc)
	 */
	public void setTimeBetweenEvents(RandomIfc timeUntilNext) {
		myEventGenerator.setTimeBetweenEvents(timeUntilNext);
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGenerator#getTimeBetweenEvents()
	 */
	public final RandomIfc getTimeBetweenEvents() {
		return myEventGenerator.getTimeBetweenEvents();
	}

	/**
	 * @param timeBtwEvents
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEventsForReplications(jsl.utilities.random.RandomIfc, long)
	 */
	public void setTimeBetweenEventsForReplications(RandomIfc timeBtwEvents, long maxNumEvents) {
		myEventGenerator.setTimeBetweenEventsForReplications(timeBtwEvents, maxNumEvents);
	}

	/**
	 * @param timeUntilFirst
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeUntilFirstEventForReplications(jsl.utilities.random.RandomIfc)
	 */
	public void setTimeUntilFirstEventForReplications(RandomIfc timeUntilFirst) {
		myEventGenerator.setTimeUntilFirstEventForReplications(timeUntilFirst);
	}

	/**
	 * 
	 * @see jsl.modeling.elements.EventGeneratorIfc#suspend()
	 */
	public void suspend() {
		myEventGenerator.suspend();
	}

	/**
	 * 
	 */
	public final void turnOffGenerator() {
		myEventGenerator.turnOffGenerator();
	}
			
	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getMaximumNumberOfEventsForReplications()
	 */
	public long getMaximumNumberOfEventsForReplications() {
		return myEventGenerator.getMaximumNumberOfEventsForReplications();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getTimeBetweenEventsForReplications()
	 */
	public RandomIfc getTimeBetweenEventsForReplications() {
		return myEventGenerator.getTimeBetweenEventsForReplications();
	}

	/**
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setMaximumNumberOfEventsForReplications(long)
	 */
	public void setMaximumNumberOfEventsForReplications(long maxNumEvents) {
		myEventGenerator.setMaximumNumberOfEventsForReplications(maxNumEvents);
	}

	/**
	 * @param timeBtwEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEventsForReplications(jsl.utilities.random.RandomIfc)
	 */
	public void setTimeBetweenEventsForReplications(RandomIfc timeBtwEvents) {
		myEventGenerator.setTimeBetweenEventsForReplications(timeBtwEvents);
	}

	/** If there is no reference to a InventoryProviderIfc, then
	 *  no demand is created, i.e. nothing happens
	 * 
	 * @param generator
	 * @param event
	 */
	protected void generateAction(EventGenerator generator, JSLEvent event){

		if (myInventoryProvider != null){
			// get the amount of the demand
			int amt = 0;		
			if (myAmtRV == null)
				amt = 1;
			else
				amt = (int)myAmtRV.getValue();
		    
			InventoryIfc i = myInventoryProvider.getInventory(myItemType, this);
			
			if (i != null){
				
				Demand d = i.requestInventory(getItemType(), this, amt);
				
				if (d != null)
					i.fillInventory(d);
				
			}
		}
				
	}	
	
	private class DemandCreationListener implements EventGeneratorListenerIfc {

		/* (non-Javadoc)
		 * @see jsl.modeling.elements.EventGeneratorListenerIfc#generate(jsl.modeling.elements.EventGenerator, jsl.modeling.JSLEvent)
		 */
		public void generate(EventGenerator generator, JSLEvent event) {
			generateAction(generator, event);
		}	
	}



	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryCustomerIfc#receiveReplenishment(jsl.modeling.elements.supplychain.inventory.Demand)
	 */
	public void receiveReplenishment(Demand d) {
		Demand.recycle(d);	
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryCustomerIfc#permitReplenishmentBackLogging(jsl.modeling.elements.supplychain.inventory.ItemType)
	 */
	public final boolean permitReplenishmentBackLogging(ItemType type) {
		if (myItemType != type)
			return false;
		else
			return myPermitBackLoggingFlag;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryCustomerIfc#permitReplenishmentPartialFilling(jsl.modeling.elements.supplychain.inventory.ItemType)
	 */
	public final boolean permitReplenishmentPartialFilling(ItemType type) {
		if (myItemType != type)
			return false;
		else
			return myPermitPartialFillingFlag;
	}
}
