/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

import java.util.Observer;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.TimeWeighted;

/**
 */
public class Inventory extends ModelElement implements InventoryProviderIfc, InventoryCustomerIfc, InventoryIfc {

	
	/** A reference to the type of item for this
	 *  inventory provider
	 */
	private ItemType myItemType;

	/** Indicates whether or not this inventory allows partial filling
	 *  of demand requests
	 */
	private boolean myAllowPartialFillingFlag;
	
	/** Indicates whether or not this inventory allows backlogging
	 *  of demand requests
	 */
	private boolean myAllowBackLoggingFlag;

	/** Indicated whether or not the inventory will
	 *  permit backlogging of its replenishment requests 
	 */
	private boolean myPermitBackLoggingFlag = true;
	
	/** Indicated whether or not the inventory will
	 *  permit partial filling of its replenishment requests 
	 */
	private boolean myPermitPartialFillingFlag;
	
	/** Tracks whether or not a demand is filled
	 *  directly from onhand on first attempt
	 */
	private ResponseVariable myFirstFillRate;

	/** Tracks the amount of a lost demand
	 */
	private ResponseVariable myLostDemandAmount;

	/** Tracks the amount of inventory actual
	 *  on hand and available for supply
	 */
	private TimeWeighted myOnHand;
	
	/** Tracks when the net inventory is <= 0
	 *  net inventory = myOnHand - myAmtBackLogged
	 */
	private TimeWeighted myStockOutIndicator;

	/** Tracks the amount of individual inventory units
	 *  that have been ordered to replenish the stock
	 */
	private TimeWeighted myOnOrder;

	/** Controls how replenishment occurs
	 */
	private InventoryPolicyAbstract myInventoryPolicy;

	/** Controls how backlogs are handled
	 */
	private BackLogPolicyAbstract myBackLogPolicy;

	/** A reference to an object that will provide inventory
	 *  (replenish) this inventory
	 */
	private InventoryProviderIfc myInventoryReplenisher;
	
	/** Will try to find a object that will supply (replenish) the inventory
	 */
	private InventoryProviderNegotiator myInventoryReplenishmentNegotiator;
	
	/** A reference to an object that will ship the 
	 *  requested inventory back to the customer
	 */
	private InventoryShipperIfc myInventoryShipper;

	/** Create an instance of Inventory with the given name
	 * 
	 *  Partial filling is false by default.
	 *  No backlogging
	 *  
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param supplier, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 * @param initialLevel
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, 
			InventoryPolicyAbstract policy, InventoryShipperIfc shipper, int initialLevel){
		this(parent, type, replenisher, policy, shipper, initialLevel, null, false, null);
	}

	/** Create an instance of Inventory with the given name
	 * 
	 *  Partial filling is false by default.
	 *  No backlogging
	 *  
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param replenisher, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 * @param initialLevel
	 * @param name
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, 
			InventoryPolicyAbstract policy, InventoryShipperIfc shipper, int initialLevel, String name){
		this(parent, type, replenisher, policy, shipper, initialLevel, null, false, name);
	}

	/** Create an instance of Inventory with the given name
	 * 
	 *  Partial filling is false by default.
	 *  InitialLevel is 0 
	 *  No backlogging
	 *  
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param replenisher, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 * @param name
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, 
			InventoryPolicyAbstract policy, InventoryShipperIfc shipper, String name){
		this(parent, type, replenisher, policy, shipper, 0, null, false, name);
	}
	
	/** Create an instance of Inventory with
	 * 
	 * Partial filling is given by the flag
	 * InitialLevel is 0
	 * No backlogging
	 * 
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param replenisher, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 * @param partialFillFlag whether or not the inventory will allow partial filling of demand (true means yes)
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, 
			InventoryPolicyAbstract policy, InventoryShipperIfc shipper, boolean partialFillFlag){
		this(parent, type, replenisher, policy, shipper, 0, null, partialFillFlag, null);
	}
	
	/** Create an instance of Inventory 
	 * 
	 *  Partial filling is false by default.
	 *  InitialLevel is 0 
	 *  No backlogging
	 *  
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param replenisher, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, 
			InventoryPolicyAbstract policy, InventoryShipperIfc shipper){
		this(parent, type, replenisher, policy, shipper, 0, null, false, null);
	}

	/** Create an instance of Inventory with
	 *  given name.  
	 *  
	 *  Partial filling is false by default.
	 *  InitialLevel is 0
	 *  
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param replenisher, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 * @param backlogPolicy, the policy for backlogging demands
	 * @param name
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, 
			InventoryPolicyAbstract policy, InventoryShipperIfc shipper, BackLogPolicyAbstract backlogPolicy, String name){
		this(parent, type, replenisher, policy, shipper, 0, backlogPolicy, false, name);
	}
	
	/** Create an instance of Inventory 
	 * 
	 *  Partial filling is false by default.
	 *  InitialLevel is 0
	 *  
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param replenisher, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 * @param backlogPolicy, the policy for backlogging demands
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, 
			InventoryPolicyAbstract policy, InventoryShipperIfc shipper, BackLogPolicyAbstract backlogPolicy){
		this(parent, type, replenisher, policy, shipper, 0, backlogPolicy, false, null);
	}

	/** Create an instance of Inventory 
	 * 
	 *  Partial filling is false by default.
	 *  InitialLevel is 0
	 *  
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param replenisher, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 * @param backlogPolicy, the policy for backlogging demands
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, 
			InventoryPolicyAbstract policy, InventoryShipperIfc shipper, int initialLevel, 
			BackLogPolicyAbstract backlogPolicy,String name){
		this(parent, type, replenisher, policy, shipper, initialLevel, backlogPolicy, false, name);
	}
	
	/** Create an instance of Inventory with
	 *  given initial level.  
	 *  
	 *  Partial filling is false by default
	 *  
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param replenisher, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 * @param initialLevel, must be >= 0
	 * @param backlogPolicy, the policy for backlogging demands
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, InventoryPolicyAbstract policy, 
			InventoryShipperIfc shipper, int initialLevel, BackLogPolicyAbstract backlogPolicy){
		this(parent, type, replenisher, policy, shipper, initialLevel, backlogPolicy, false, null);
	}
	
	/** Create an instance of Inventory with
	 *  given initial level
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param replenisher, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 * @param initialLevel, must be >= 0
	 * @param backlogPolicy, the policy for backlogging demands
	 * @param partialFillFlag whether or not the inventory will allow partial filling of demand (true means yes)
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, InventoryPolicyAbstract policy, 
			InventoryShipperIfc shipper, int initialLevel, BackLogPolicyAbstract backlogPolicy, boolean partialFillFlag){
		this(parent, type, replenisher, policy, shipper, initialLevel, backlogPolicy, partialFillFlag, null);
	}

	/** Create an instance of Inventory with
	 *  given name and initial level
	 * @param parent
	 * @param type The ItemType provided by this Inventory
	 * @param replenisher, serves the replenshishment request
	 * @param policy, The inventory policy
	 * @param shipper, The inventory shipper for this inventory
	 * @param initialLevel, must be >= 0
	 * @param backlogPolicy, the policy for backlogging demands
	 * @param partialFillFlag whether or not the inventory will allow partial filling of demand (true means yes)
	 * @param name
	 */
	public Inventory(ModelElement parent, ItemType type, InventoryProviderIfc replenisher, 
			InventoryPolicyAbstract policy, InventoryShipperIfc shipper,
             int initialLevel, BackLogPolicyAbstract backlogPolicy, boolean partialFillFlag, String name) {
		super(parent, name);
		
		setItemType(type);
		setBackLogPolicy(backlogPolicy);
		myAllowPartialFillingFlag = partialFillFlag;			
		
		setInventoryReplenisher(replenisher);
		setInventoryShipper(shipper);
		
		setInventoryPolicy(policy);
		myStockOutIndicator = new TimeWeighted(this, getName() + "-StockoutFrequency");
		myOnHand = new TimeWeighted(this, initialLevel, 0, getName() + "-OnHand");
		myOnOrder = new TimeWeighted(this, getName() + "-OnOrder");
		myFirstFillRate = new ResponseVariable(this, getName() + "-FirstFillrate");
		myLostDemandAmount = new ResponseVariable(this, getName() + "-LostDemandAmount");
		myInventoryReplenishmentNegotiator = new InventoryProviderNegotiator();
	}

	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================

	public String toString(){
		String s = "time = " + getTime() + " on hand = " + this.getAmountOnHand();
		s = s + " on order = " + this.getAmountOnOrder();
		s = s + " back ordered = " + this.getAmountBackLogged();
		s = s + " inv pos = " + this.getInventoryPosition();
		return(s);
	}
	
	/** Gets the type of item associated with this inventory
	 *  This may be null, if not associated with a particular type
	 * @return
	 */
	public final ItemType getItemType() {
		return myItemType;
	}
			
    /** Gets the initial value of the onhand
     * @return The initial value.
     */
    public final int getInitialOnHand(){
        return((int)myOnHand.getInitialValue());
    }
 
    /** Sets the initial on hand of the inventory. Only relevant prior to each
     * replication.  Changing during a replication has no effect until the next replication.
     * @param value The initial value for the inventory.
     */
    public final void setInitialOnHand(int value){
        if (value < 0)
            throw new IllegalArgumentException("Invalid argument. Cannot have a negative on hand");

    	myOnHand.setInitialValue(value);
    	if (value == 0){
    		myStockOutIndicator.setInitialValue(1.0);
    	} else
    		myStockOutIndicator.setInitialValue(0.0);
    }

	/** Returns the inventory position of the InventoryProvider
	 *  InventoryPosition should be Onhand + OnOrder - BackLogged
	 * 
	 * @return The current inventory position of the ItemType
	 */
	public final int getInventoryPosition() {
		int ip = getAmountOnHand() + getAmountOnOrder() - getAmountBackLogged();
    	return(ip);
	}

	/** Returns the inventory position for this inventory
	 * @return
	 */
	public final int getAmountOnHand() {
		return ((int)myOnHand.getValue());
	}

	/** Allows an observer to be attached to the amount on hand.
	 * 
	 * @param observer
	 */
	public final void addOnHandObserver(Observer observer){
		myOnHand.addObserver(observer);
	}

	/** Attaches an AggregateTimeWeightedVariable to the on hand 
	 *  for this inventory.  Whenever the on hand gets updated
	 *  the aggregate variable is updated.
	 * 
	 * @param variable
	 */
	public final void attachAggregateTimeWeightedVariableToOnHand(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myOnHand);
	}
	
	/** Allows an observer to be attached to the first fill opportunities
	 * 
	 * @param observer
	 */
	public final void attachAggregateResponseToFillRate(ResponseVariableAverageObserver aggregate){
		aggregate.subscribeTo(myFirstFillRate);
	}

	/** Allows an observer to be attached to the stock out indicator
	 * 
	 * @param observer
	 */
	public final void addStockOutFrequencyObserver(Observer observer){
		myStockOutIndicator.addObserver(observer);
	}
	
	/** Returns the inventory position for this inventory
	 * @return
	 */
	public final int getAmountOnOrder() {
		return (int)myOnOrder.getValue();
	}

	/** Returns the inventory position for this inventory
	 * @return
	 */
	public final int getAmountBackLogged() {
		if (myBackLogPolicy == null)
			return(0);
		else
			return ((int)myBackLogPolicy.getAmountBackLogged());
	}
	
	/**
	 * @return
	 */
	public final InventoryProviderIfc getInventoryReplenisher() {
		return myInventoryReplenisher;
	}

	/** Sets the replenishment provider, if the provider is null, then
	 *  the default LeadTimeReplenishmentProvider will be used
	 *  
	 * @param provider
	 */
	public final void setInventoryReplenisher(InventoryProviderIfc replenisher) {
		if (replenisher == null)
			throw new IllegalArgumentException("Invalid argument. The supplied InventoryProviderIfc replenisher was null!");	

		myInventoryReplenisher = replenisher;
	}

	/** Sets the inventory shipper for this inventory, must not be null
	 * 
	 * @param shipper
	 */
	public void setInventoryShipper(InventoryShipperIfc shipper) {
		if (shipper == null)
			throw new IllegalArgumentException("Invalid argument. InventoryShipperIfc shipper must be non-null.");
		myInventoryShipper = shipper;
	}
	
	/** True means that this Inventory is willing to permit backlogging of
	 *  replenishment requests
	 * @param flag The flag to set.
	 */
	public final void setPermitReplenishmentBackLoggingFlag(boolean flag) {
		myPermitBackLoggingFlag = flag;
	}

	/** True means that this Inventory is willing to permit partial filling of
	 *  replenishment requests
	 * @param flag The flag to set.
	 */
	public final void setPermitReplenishmentPartialFillingFlag(boolean flag) {
		myPermitPartialFillingFlag = flag;
	}
	
	/**
	 * @param policy
	 */
	public final void setInventoryPolicy(InventoryPolicyAbstract policy) {
		if (policy == null)
			throw new IllegalArgumentException("Invalid argument. InventoryPolicyAbstract policy must be non-null.");
		myInventoryPolicy = policy;
		myInventoryPolicy.setInventory(this);
	}

	
	/** Asks the Inventory to provide inventory of the given
	 *  ItemType for the amount requested. 
	 *  
	 *  The Inventory should return null for the following cases:
	 *  
	 *  1) The ItemType cannot be provided by the InventoryProvider
	 *  2) The customer wants backlogging but the provider cannot backlog
	 *  3) The customer does not want backlogging but the provider backlogs
	 *  4) The customer wants partial filling but the provider cannot partial fill
	 *  5) The customer does not want partial filling but the provider partially fills
	 *  
	 *  Thus, the burden is on the customer to decide what to do if the request 
	 *  can not be satisfied to its requirements.
	 * @param type The ItemType needed
	 * @param customer The customer who needs the inventory
	 * @param amountNeeded  The amount of ItemType needed
	 * 
	 * @return A Demand representing the request or null if the request cannot be satisfied.
	 */
	public Demand requestInventory(ItemType type, InventoryCustomerIfc customer, int amountNeeded) {
		
		if (customer == null)
			throw new IllegalArgumentException("Invalid argument. InventoryCustomerIfc customer must be non-null.");

		if (type == null)
			throw new IllegalArgumentException("Invalid argument. ItemType type must be non-null.");

		// does the customer want what I can provide
		// ask the customer to compare my inventory requirements with its replenishment requirements
		// if they don't match then I can't provide a demand
		if (myItemType != type)
			return null;

		if (customer.permitReplenishmentBackLogging(type) != allowsBackLogging())
			return null;
		
		if (customer.permitReplenishmentPartialFilling(type) != allowsPartialFilling())
			return null;

		// make the demand to represent the request
		Demand d = Demand.create(type, amountNeeded);
		
		d.setAllowBackLoggingFlag(allowsBackLogging());
		d.setAllowPartialFillingFlag(allowsPartialFilling());
		
		setInventoryProvider(d);
		setInventoryCustomer(customer, d);
		
		processInitialRequest(d);
		
		return d;
	}

	/** Asks the InventoryProvider to fill the demand that had been
	 *  requested.  This method should throw an IllegalArgumentException
	 *  if the supplied demand is null.  The demand must also have
	 *  been supplied (returned) by this InventoryProvider. The demand
	 *  must also have been created at the current simulation time. If
	 *  simulation time elapses between the call to requestInventory()
	 *  and fillInventory() then the demand request is invalidated.
	 * 
	 * @param demand The demand to be filled by the inventory provider
	 */
	public final void fillInventory(Demand demand) {
		if (demand == null)
			throw new IllegalArgumentException("Invalid argument. Demand demand must be non-null.");
	
		if (demand.getCreateTime() < getTime())
			throw new IllegalArgumentException("Invalid argument. Demand request was create at time" +
					demand.getCreateTime() + " prior to curent time " + getTime() + " The demand " +
							"request is no longer valid");
		
		if (demand.getInventoryProvider() != this)
			throw new IllegalArgumentException("Invalid argument. Demand request was not created by this provider.");

//		System.out.println(" customer " + demand.getId() + " wanted " + demand.getDemand());
//		System.out.println(this);
		handleInitialRequest(demand);
	}

    /** Indicate to the inventory to replenish by the amount 
     *  associated with the demand supplied,
     *  demand.getAmountFilled() is used to know the amount of the replenishment
     * @param demand, demand must be non-null
     */
	public void receiveReplenishment(Demand demand) {
		if (demand == null)
			throw new IllegalArgumentException("Invalid argument. Demand demand must be non-null.");
		
		if (demand.getInventoryCustomer() != this) 
			throw new IllegalArgumentException("Invalid argument. This inventory did not request this replenishment.");
			
		int replenishmentQty = demand.getAmountFilled();
//		System.out.println("replenishment arrived " + replenishmentQty);
//		System.out.println(this);
		decrementOnOrder(replenishmentQty);
	    incrementOnHand(replenishmentQty);
	    if (allowsBackLogging())
	    		myBackLogPolicy.fillBackLogs();
	    checkInventory();
	    Demand.recycle(demand);
//		System.out.println(this);
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	/** Initializes the inventory prior to a replication
	 *  Does a check of the inventory to see if an replenishment
	 *  is needed.
	 */
	protected void initialize(){
		checkInventory();
	}
	
	/** Allows subclasses to set the inventory provider
	 *  on a demand
	 */
	protected final void setInventoryProvider(Demand d){
		if (d == null)
			throw new IllegalArgumentException("Invalid argument. Demand demand must be non-null.");
		d.setInventoryProvider(this);
	}
	
	/** Allows subclasses to set the inventory customer
	 *  on a demand
	 */
	protected final void setInventoryCustomer(InventoryCustomerIfc customer, Demand d){
		if (d == null)
			throw new IllegalArgumentException("Invalid argument. Demand demand must be non-null.");
		if (customer == null)
			throw new IllegalArgumentException("Invalid argument. InventoryCustomerIfc customer must be non-null.");
		d.setInventoryCustomer(customer);
	}
	
	/** Sets the type of item associated with this inventory
	 *  The type may be null, indicating that this inventory 
	 *  uses the default item type designation.
	 * @param type
	 */
	protected final void setItemType(ItemType type) {
		if (type == null){
	        throw new IllegalArgumentException("Invalid argument. ItemType type was null.");			
		}
		myItemType = type;
	}
	
	/** Prepares a replenishment demand for the quantity provided
	 *  increments the amount on order and asks the inventory's
	 *  replenishment provider to provide the demand.  Throws a
	 *  NoReplenishmentOptionException if the requested demand
	 *  cannot be filled by the InventoryProvider
	 * 
	 * @param replenishmentQty
	 */
	protected void requestReplenishment(int replenishmentQty) {
		
		// get the inventory's replenisher
		InventoryProviderIfc replenisher = getInventoryReplenisher();
		
		InventoryIfc inventory;		
		// check if provider can provide inventory with my requirements
		if (replenisher.providesInventory(myItemType, this)){
			// get the inventory
			inventory = replenisher.getInventory(myItemType, this);
		} else { // provider can't do the job
			// find a provider that will
			replenisher = myInventoryReplenishmentNegotiator.findInventoryProvider(this, myItemType, replenishmentQty);
			inventory = replenisher.getInventory(myItemType, this);
		}
		
		// if we get here then the request can be handled
		// and the provider or negotiator has supplied and InventoryIfc that can meet the requirements
		// of this "customer" , because the provider already checked the requirements
		
		// create a demand request on the inventory
		Demand d = inventory.requestInventory(myItemType, this, replenishmentQty);
		
		// the demand has only been created, no attempt to fill has been made
		// replenishmentQty is how much was requested
		// d.getOriginalAmountDemanded() the amount requested
		// d.getRequestFillAmount() amount of the demand's request that can be filled immediately
		
		// assume that the "customer" (this inventory) will go ahead and try to fill the demand
		
		// if back logging is permitted then the whole original amount demanded will be on order
		// if back logging is not permitted then the amount on order is only what can given by the supplier
		if (permitReplenishmentBackLogging(myItemType)){
			incrementOnOrder(d.getOriginalAmountDemanded());
		} else // no backlogging of replenishment permitted
			incrementOnOrder(d.getInitialRequestFillAmount());
		
		// ask the "supplier" to fill the demand
		
		inventory.fillInventory(d);
		
		// the demand may or may not have been filled in its entirety, i.e. the fill status might not be true
		// the demand might have been backlogged, backlogged with partial fill, not backlogged
		
		// d.getAmountFilled() is how much of the original amount demanded was filled
		// if back logging is permitted then the whole original amount demanded will be on order
		// if back logging is not permitted then the amount on order is only what was given by the supplier
		
/*		if (permitReplenishmentBackLogging(myItemType)){
			incrementOnOrder(d.getOriginalAmountDemanded());
		} else // no backlogging of replenishment permitted
			incrementOnOrder(d.getAmountFilled());
*/
	}
	
	protected final void checkInventory(){
		myInventoryPolicy.checkInventory();
	}
		
	/** Decrements the amount on order
	 * @param amount The amount to decrement by
	 */
	protected final void decrementOnOrder(int amount) {
		myOnOrder.decrement(amount);
		if (myOnOrder.getValue() < 0)
			throw new IllegalStateException("The amount on order was set to a negative value: " + myOnOrder.getValue());
	}
	
	/** Increments the amount on order
	 * 
	 * @param amount
	 */
	protected final void incrementOnOrder(int amount){
		myOnOrder.increment(amount);
	}

	/**
	 * @param amount The amount to decrement by
	 */
	protected final void decrementOnHand(int amt){
		myOnHand.decrement(amt);
		if (myOnHand.getValue() < 0)
			throw new IllegalStateException("The amount on hand was set to a negative value: " + myOnHand.getValue());

		if (myOnHand.getValue() < 1.0)
    			myStockOutIndicator.setValue(1.0);
   }
    
	/**
	 * @param amount The amount to increment by
	 */
	protected final void incrementOnHand(int amt){
		myOnHand.increment(amt);
		if (myOnHand.getValue() > 0.0)
    			myStockOutIndicator.setValue(0.0);
    }
	
	/** The purpose of this method is to set up the demand
	 *  request based on available supply.  This method should
	 *  indicate on the demand whether it will be:
	 *  
	 *  1) Filled immediately
	 *  2) Backlogged
	 *  3) Lost
	 *  
	 *  This method should be called from requestInventory()
	 *  
	 * @param demand The demand to be checked for supply
	 */
	protected void processInitialRequest(Demand demand) {
		
		int onHand = getAmountOnHand();
		int amtNeeded = demand.getRemainingDemand();
		
		// determine the amount that can be supplied
		int amtSupplied = 0;
		// check if amount on hand can satisfy all the demand
		if (onHand >= amtNeeded){
			amtSupplied = amtNeeded;
			// the demand can be filled immediately
			demand.setRequestStatus(Demand.RequestStatus.FILL_ALL);
		} else { // cannot fill all the demand, demand > onhand
			// check if demand allows backlogging
			if (demand.getAllowBackLoggingFlag() == false){
				// the demand is unwilling to wait
				// some or all of the request cannot be satisfied
				if (demand.getAllowPartialFillingFlag() == true){
					// part of the demand can be filled
					// since demand > onhand, we can supply all remaining on hand
					amtSupplied = onHand;
					//set state for unwilling to wait but can be partially filled
					demand.setRequestStatus(Demand.RequestStatus.LOST_PARTIAL);
				} else { // unwilling to wait and cannot be partially filled
					//set state for unwilling to wait and cannot be partially filled
					demand.setRequestStatus(Demand.RequestStatus.LOST_ALL);
				}
			} else {// willing to wait for the request to be filled
				if (demand.getAllowPartialFillingFlag() == true){
					// part of the demand can be filled
					// since demand > onhand, we can supply all remaining on hand
					amtSupplied = onHand;
					//set state for willing to wait and can be partially filled
					demand.setRequestStatus(Demand.RequestStatus.BACKLOG_PARTIAL);
				} else { // willing to wait and cannot be partially filled
					//set state for willing to wait and cannot be partially filled
					demand.setRequestStatus(Demand.RequestStatus.BACKLOG_NO_PARTIAL);
				}   					
			}	
		}
		// amtSupplied = 0 or onHand or amtNeeded
	
		//set the amount that can be supplied to this request
		demand.setInitialRequestFillAmount(amtSupplied);
//		System.out.println(" customer " + demand.getId() + " wanted " + demand.getDemand());
//		System.out.println(this);
	}

	/** Asks the Inventory to handle the initial demand request.
	 *  This method is called from fillInventory() to handle the first
	 *  request for inventory from a customer.
	 * 
	 * @param demand The demand to be filled by the inventory
	 */
	protected void handleInitialRequest(Demand demand){

		Demand.RequestStatus status = demand.getRequestStatus();
		switch (status) {
			case FILL_ALL:
				immediateInitialFill(demand);
				shipReplenishment(demand);
				checkInventory();
				break;
			case BACKLOG_PARTIAL:
				backLogPartialInitialFill(demand);
				checkInventory();
				break;
			case BACKLOG_NO_PARTIAL:
				backLogNoPartialInitialFill(demand);
				checkInventory();
				break;
			case LOST_PARTIAL:
				lostPartialInitialFill(demand);
				shipReplenishment(demand);
				checkInventory();
				break;
			case LOST_ALL:
				lostAllInitialFill(demand);
				break;
			default:
				throw new IllegalArgumentException("The demand request's status was illegal: " + status);		
		}

	}

	/** Ships the demand back to the customer
	 * 
	 * @param d
	 */
	protected final void shipReplenishment(Demand d){
		InventoryShipperIfc s = d.getInventoryShipper();
		
		// allow the demand to override which shipper if available
		if (s != null)
			s.shipReplenishment(d);
		else
			myInventoryShipper.shipReplenishment(d);
	}
	
	/** Used to perform an immediate fill for the demand
	 *  Records the first fill rate, decrements on hand, 
	 *  and fills the demand
	 * 
	 * @param d
	 */
	protected final void immediateInitialFill(Demand d) {
		// initial request can be filled immediately
		// set the fill indicator to indicate a successful first fill		
		myFirstFillRate.setValue(1.0);
		// get how much can be supplied to the request
		int amtSupplied = d.getInitialRequestFillAmount();
		// fill the amount supplied
		decrementOnHand(amtSupplied);
		d.fillDemand(amtSupplied);	
	}

	/**  Allows the inventory to back log a demand while
	 *   also partially filling it.
	 * 
	 * @param d
	 */
	protected final void backLogPartialInitialFill(Demand d) {
		// initial request needs back ordering, but can be partially filled
		// set the fill indicator to indicate unsuccessful first fill
		myFirstFillRate.setValue(0.0);  
		// get how much can be supplied to the request
		int amtSupplied = d.getInitialRequestFillAmount();
		// fill by the amount supplied
		decrementOnHand(amtSupplied);
		d.fillDemand(amtSupplied);
		// back log the demand
		myBackLogPolicy.backlog(d);
	}

	/** Allows the inventory to back log a demand with the
	 *  no partial filling option.
	 * 
	 * @param d
	 */
	protected final void backLogNoPartialInitialFill(Demand d) {
		// initial request needs back ordering, but cannot be partially filled
		// set the fill indicator to indicate unsuccessful first fill
		myFirstFillRate.setValue(0.0);
		// back log the demand the entire demand
		myBackLogPolicy.backlog(d);
	}

	/** Allows the inventory to partially fill a demand on 
	 *  the initial request, but lose part of the demand
	 * 
	 * @param d
	 */
	protected final void lostPartialInitialFill(Demand d) {
		// part of the initial request will be lost and part can be partially filled
		// set the fill indicator to indicate unsuccessful first fill
		myFirstFillRate.setValue(0.0);
		// get the amount that can be supplied
		int amtSupplied = d.getInitialRequestFillAmount();
		// supply the amount
		decrementOnHand(amtSupplied);
		d.fillDemand(amtSupplied);
		// indicate that part (or) all of the demand was lost
		d.setLostDemandFlag(true);
		// record how much was lost
		myLostDemandAmount.setValue(d.getRemainingDemand());
	}
	
	/** Used to indicate that all the demand was lost
	 *  on the initial fill attempt, updates the first fill rate
	 *  and sets the demand's lost demand flag.
	 * 
	 * @param d
	 */
	protected final void lostAllInitialFill(Demand d) {
		// the initial request will be lost 
		// set the fill indicator to indicate unsuccessful first fill
		myFirstFillRate.setValue(0.0);
		// indicate that part (or) all of the demand was lost
		d.setLostDemandFlag(true);
		// record how much was lost
		myLostDemandAmount.setValue(d.getRemainingDemand());
	}
	
	/** Similar to processInitialRequest() this method should
	 *  determine how much of the demand can be satisfied and set
	 *  the demand's setRequestFillAmount() method
	 * 
	 * @param demand
	 * @return The amount that can be supplied.
	 */
	protected final int processBackLog(Demand demand){
		int onHand = getAmountOnHand();
		int amtNeeded = demand.getRemainingDemand();
	    	int amtSupplied = 0;
	    	// check if amount on hand can satisfy all the demand
	    	if (onHand >= amtNeeded){
	    		amtSupplied = amtNeeded;
	    	} else { // cannot fill all the demand, demand > onhand
	    		// check if demand allows partial filling
				if (demand.getAllowPartialFillingFlag() == true){
					// part of the demand can be filled
					// since demand > onhand, we can supply all remaining on hand
					amtSupplied = onHand;				
				}
	    	}
	    	demand.setBackLogFillAmount(amtSupplied);
	    	return(amtSupplied);
	}
	
	/** Once a demand has been processed through processBackLog()
	 *  then this method can be used to have the inventory fill
	 *  the units that it can fill.  
	 * 
	 * @param demand
	 */
	protected final boolean fillBackLog(Demand demand){
		
		// get how much can be supplied to the request
		int amtSupplied = demand.getBackLogFillAmount();
		if (amtSupplied > 0){
			// fill by the amount supplied
			decrementOnHand(amtSupplied);
			demand.fillDemand(amtSupplied);	
		}
		
		if (demand.getFilledStatus() == true){
			shipReplenishment(demand);
			return true;
		} else
			return false;
	}

	/** Returns true if this InventoryProvider can provide
	 *  this type of item with the given customer requirements.
	 * 
	 * @param type The ItemType to check
	 * @param customer The customer that wants the inventory
	 * @return True if ItemType can be provided given the customer's requirements
	 */
	public boolean providesInventory(ItemType type, InventoryCustomerIfc customer) {
		if (customer == null)
	        throw new IllegalArgumentException("Invalid argument. InventoryCustomerIfc customer was null.");			

		// check if the desired requirements (inputs) match what this inventory can do
		if(myItemType != type)
			return false;
		
		if (allowsBackLogging() != customer.permitReplenishmentBackLogging(type))
			return false;
		
		if (allowsPartialFilling() != customer.permitReplenishmentPartialFilling(type))
			return false;
		
		// if we get here all requirements matched
		
		return true;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryProviderIfc#getInventory(jsl.modeling.elements.supplychain.inventory.ItemType, boolean, boolean)
	 */
	public final InventoryIfc getInventory(ItemType type, InventoryCustomerIfc customer) {
		if (providesInventory(type, customer))
			return this;
		else
			return null;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryIfc#providesItemType(jsl.modeling.elements.supplychain.inventory.ItemType)
	 */
	public final boolean providesItemType(ItemType type) {
		return myItemType == type;
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryCustomerIfc#checkReplenishmentRequirements(jsl.modeling.elements.supplychain.inventory.ItemType, boolean, boolean)
	 */
	public boolean checkReplenishmentRequirements(ItemType type, boolean permitBL, boolean permitPF) {
		if (myItemType != type)
			return false;
		
		if (myPermitBackLoggingFlag != permitBL)
			return false;
		
		if (myPermitPartialFillingFlag != permitPF)
			return false;

		return true;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryCustomerIfc#permitReplenishmentBackLogging(jsl.modeling.elements.supplychain.inventory.ItemType)
	 */
	public final boolean permitReplenishmentBackLogging(ItemType type) {
		if (myItemType != type)
			return false;
		else
			return myPermitBackLoggingFlag;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryCustomerIfc#permitReplenishmentPartialFilling(jsl.modeling.elements.supplychain.inventory.ItemType)
	 */
	public final boolean permitReplenishmentPartialFilling(ItemType type) {
		if (myItemType != type)
			return false;
		else
			return myPermitPartialFillingFlag;
	}
	
	/** Indicates whether or not the inventory allows backlogging for incoming demands
	 * @return
	 */
	public final boolean allowsBackLogging(){
		return (myAllowBackLoggingFlag);
	}
	
	/** Indicates whether or not the inventory allows partial filling for incoming demands
	 * @return
	 */
	public final boolean allowsPartialFilling(){
		return myAllowPartialFillingFlag;
	}
	
	/** Null implies that the inventory does not backlog demand.
	 *  Once the backlog policy has been set to a non-null value
	 *  it cannot be changed, i.e. once the backlog policy is not null
	 *  this method does nothing.
	 * @param backLogPolicy The backLogPolicy to set.
	 */
	private final void setBackLogPolicy(BackLogPolicyAbstract backLogPolicy) {
		
		// if the backlog policy has already been set, do not allow it to change
		if (myBackLogPolicy != null)
			return;		
		
		if (backLogPolicy != null){
			myAllowBackLoggingFlag = true;
			myBackLogPolicy = backLogPolicy;
			myBackLogPolicy.setInventory(this);
		}
	}

}
