/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

import java.util.*;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;

/**
 */
public class InventoryHoldingPoint extends ModelElement implements InventoryCustomerIfc,
				InventoryProviderIfc {
	
	/** A Map to hold the inventory in the IHP
	 *  key = ItemType, object = Inventory
	 */	
	private Map<ItemType, Inventory> myInventory;
			
	/** The element that supplies inventory to this IHP
	 */
	private InventoryProviderIfc myInventoryProvider;
	
	/** The shipper that sends replenishments to this IHP's
	 *  customers
	 */
	private InventoryShipperIfc myInventoryShipper;
	
	private AggregateTimeWeightedVariable myTotalOnHand;
	
	private ResponseVariableAverageObserver myAvgFirstFillRate;
	
	/** Creates an inventory holding point
	 * 
	 * @param parent
	 * @param provider The element that provides (supplies) inventory to this IHP
	 * @param shipper  The element that ships demand to customers of this IHP
	 * @param name
	 */
	public InventoryHoldingPoint(ModelElement parent, InventoryProviderIfc provider, InventoryShipperIfc shipper) {
		this(parent, provider, shipper, null);
	}


	/** Creates an inventory holding point
	 * 
	 * @param parent
	 * @param provider The element that provides (supplies) inventory to this IHP
	 * @param shipper  The element that ships demand to customers of this IHP
	 * @param name
	 */
	public InventoryHoldingPoint(ModelElement parent, InventoryProviderIfc provider, InventoryShipperIfc shipper, String name) {
		super(parent, name);
		setInventoryProvider(provider);
		setInventoryShipper(shipper);
		myInventory = new HashMap<ItemType, Inventory>();
		myTotalOnHand = new AggregateTimeWeightedVariable(this, getName() + " Total On Hand");
		
		myAvgFirstFillRate = new ResponseVariableAverageObserver(this, getName() + "Aggregate Fill Rate");
	}

	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================
	
	/** Adds an inventory to this IHP. The ItemType of the supplied
	 *  inventory must not have already been added to this IHP
	 * 
	 * @param inventory Must not be null
	 */
	public final void addInventory(Inventory inventory){

		if (inventory == null)
			throw new IllegalArgumentException("The Inventory inventory must be non-null.");
		
		ItemType type = inventory.getItemType();
		
		if (myInventory.containsKey(type))
			throw new IllegalArgumentException("The IHP already contains an inventory for type: " + type.getName());
		
		inventory.attachAggregateTimeWeightedVariableToOnHand(myTotalOnHand);
		inventory.attachAggregateResponseToFillRate(myAvgFirstFillRate);
		myInventory.put(type, inventory);		
	}

	/** Creates and adds an inventory using an RQ Inventory Policy to the IHP
	 *  using a DefaultBackLogPolicy and no Partial filling permitted
	 *  
	 * @param type
	 * @param initialLevel
	 * @param reorderPt
	 * @param reorderQty
	 * @return
	 */
	public final Inventory addRQInventoryItem(ItemType type, int initialLevel, int reorderPt, int reorderQty){
		
		if (type == null)
			throw new IllegalArgumentException("The ItemType type must be non-null.");
		
		InventoryPolicyAbstract policy = new RQInventoryPolicy(this, reorderPt, reorderQty);	
		DefaultBackLogPolicy bpolicy = new DefaultBackLogPolicy(this, getName() + " BackLog-" + type.getName());
					 	
		Inventory inventory = new Inventory(this, type, myInventoryProvider, policy, myInventoryShipper, 
				initialLevel, bpolicy, getName() + " Inv.-" + type.getName());	
		
		addInventory(inventory);	
		return(inventory);	
	}
	
	/** Creates and adds an inventory to the IHP
	 *  and no Partial filling permitted
	 *  
	 * @param type
	 * @param policy
	 * @param bpolicy The BackLogPolicy
	 * @param initialLevel
	 * @return
	 */
	public final Inventory addInventory(ItemType type, InventoryPolicyAbstract policy, 
			BackLogPolicyAbstract bpolicy, int initialLevel){
		
		if (type == null)
			throw new IllegalArgumentException("The ItemType type must be non-null.");

		if (policy == null)
			throw new IllegalArgumentException("The InventoryPolicyAbstract policy must be non-null.");

		if (bpolicy == null)
			throw new IllegalArgumentException("The BackLogPolicyAbstract bpolicy must be non-null.");

		Inventory inventory = new Inventory(this, type, myInventoryProvider, policy, myInventoryShipper, 
				initialLevel, bpolicy, getName() + " Inv.-" + type.getName());		
		
		addInventory(inventory);
		
		return(inventory);		
	}

	//  ===========================================
	//      InventoryCustomerIfc METHODS
	//  ===========================================
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryCustomerIfc#permitReplenishmentBackLogging(jsl.modeling.elements.supplychain.inventory.ItemType)
	 */
	public final boolean permitReplenishmentBackLogging(ItemType type) {
		Inventory i = myInventory.get(type);
		if (i == null)
			return false;
		else
			return(i.permitReplenishmentBackLogging(type));
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryCustomerIfc#permitReplenishmentPartialFilling(jsl.modeling.elements.supplychain.inventory.ItemType)
	 */
	public final boolean permitReplenishmentPartialFilling(ItemType type) {
		Inventory i = myInventory.get(type);
		if (i == null)
			return false;
		else
			return(i.permitReplenishmentPartialFilling(type));
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryCustomerIfc#receiveReplenishment(jsl.modeling.elements.supplychain.inventory.Demand)
	 */
	public final void receiveReplenishment(Demand demand) {
        if (demand == null)
            throw new IllegalArgumentException("Invalid argument. Demand was null.");
		ItemType t = demand.getItemType();
		Inventory i = myInventory.get(t);
		i.receiveReplenishment(demand);		
	}

	//  ===========================================
	//      InventoryProviderIfc METHODS
	//  ===========================================

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryProviderIfc#providesInventory(jsl.modeling.elements.supplychain.inventory.ItemType, boolean, boolean)
	 */
	public final boolean providesInventory(ItemType type, InventoryCustomerIfc customer) {
		Inventory i = myInventory.get(type);
		if (i != null)
			return(i.providesInventory(type, customer));
		else
			return false;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryProviderIfc#getInventory(jsl.modeling.elements.supplychain.inventory.ItemType, boolean, boolean)
	 */
	public final InventoryIfc getInventory(ItemType type, InventoryCustomerIfc customer) {
		Inventory i = myInventory.get(type);
		if (i != null)
			return(i.getInventory(type, customer));
		else
			return null;
	}

	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
	
	/**
	 * @param provider The inventory Provider to set.
	 */
	protected final void setInventoryProvider(InventoryProviderIfc provider) {
		if (provider == null)
			throw new IllegalArgumentException("The InventoryProviderIfc provider must be non-null.");
		myInventoryProvider = provider;
	}
	
	/**
	 * @param inventoryShipper The InventoryShipper to set.
	 */
	protected final void setInventoryShipper(InventoryShipperIfc shipper) {
		if (shipper== null)
			throw new IllegalArgumentException("The InventoryShipperIfc shipper must be non-null.");
		myInventoryShipper = shipper;
	}

	/**
	 * @return Returns the InventoryShipper.
	 */
	protected final InventoryShipperIfc getInventoryShipper() {
		return myInventoryShipper;
	}
	
	/**
	 * @return Returns the InventoryShipper.
	 */
	protected final InventoryProviderIfc getInventoryProvider() {
		return myInventoryProvider;
	}

}
