/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

/**
 */
public interface InventoryIfc {
	
	/** Indicates if the inventory provides this type
	 * @return
	 */
	public abstract boolean providesItemType(ItemType type);

	/** Gets the initial value of the onhand
	 * @return The initial value.
	 */
	public abstract int getInitialOnHand();

	/** Returns the inventory position of the InventoryProvider
	 *  InventoryPosition should be Onhand + OnOrder - BackLogged
	 * 
	 * @return The current inventory position of the ItemType
	 */
	public abstract int getInventoryPosition();

	/** Returns the inventory position for this inventory
	 * @return
	 */
	public abstract int getAmountOnHand();

	/** Returns the inventory position for this inventory
	 * @return
	 */
	public abstract int getAmountOnOrder();

	/** Returns the inventory position for this inventory
	 * @return
	 */
	public abstract int getAmountBackLogged();

	/** Asks the Inventory to provide inventory of the given
	 *  ItemType for the amount requested. 
	 *  
	 *  The Inventory should return null for the following cases:
	 *  
	 *  1) The ItemType cannot be provided by the InventoryProvider
	 *  2) The customer wants backlogging but the provider cannot backlog
	 *  3) The customer does not want backlogging but the provider backlogs
	 *  4) The customer wants partial filling but the provider cannot partial fill
	 *  5) The customer does not want partial filling but the provider partially fills
	 *  
	 *  Thus, the burden is on the customer to decide what to do if the request 
	 *  can not be satisfied to its requirements.
	 * @param type The ItemType needed
	 * @param customer The customer who needs the inventory
	 * @param amountNeeded  The amount of ItemType needed
	 * 
	 * @return A Demand representing the request or null if the request cannot be satisfied.
	 */
	public abstract Demand requestInventory(ItemType type,
			InventoryCustomerIfc customer, int amountNeeded);

	/** Asks the InventoryProvider to fill the demand that had been
	 *  requested.  This method should throw an IllegalArgumentException
	 *  if the supplied demand is null.  The demand must also have
	 *  been supplied (returned) by this InventoryProvider. The demand
	 *  must also have been created at the current simulation time. If
	 *  simulation time elapses between the call to requestInventory()
	 *  and fillInventory() then the demand request is invalidated.
	 * 
	 * @param demand The demand to be filled by the inventory provider
	 */
	public abstract void fillInventory(Demand demand);
}