/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.processview.description.EntityType;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionFunctionIfc;
import jsl.utilities.random.distributions.DistributionIfc;

/**
 */
public class ItemType extends EntityType {

	/** Holds the default lead time for the item type
	 *  Used if it does not have set provider when filling demand
	 *  replenishments
	 */
	private RandomVariable myLeadTime;
	
	/** Specifies the distribution for the default lead time
	 */
	private DistributionIfc myLeadTimeD;
	
	/** indicates the weight of the item
	 * 
	 */
	private double myWeight;
	
	/** indicates the cube (length x width x height) of the item
	 * 
	 */
	private double myCube;
	
	/**
	 * @param parent
	 */
	public ItemType(ModelElement parent) {
		this(parent, null, 1.0, 1.0);
	}
	
	/**
	 * @param parent
	 * @param name
	 */
	public ItemType(ModelElement parent, String name) {
		this(parent, name, 1.0, 1.0);
	}

	/**
	 * 
	 * @param parent
	 * @param weight
	 * @param cube
	 */
	public ItemType(ModelElement parent, double weight, double cube) {
		this(parent, null, weight, cube);
	}
	
	/**
	 * 
	 * @param parent
	 * @param name
	 * @param weight
	 * @param cube
	 */
	public ItemType(ModelElement parent, String name, double weight, double cube) {
		super(parent, name, null);
		setWeight(weight);
		setCube(cube);
		myLeadTimeD = new Constant(0);
		myLeadTime = new RandomVariable(this, myLeadTimeD);
	}

	/** The cube of the item
	 * @return
	 */
	public final double getCube() {
		return myCube;
	}

	/** The weight of the item
	 * @return
	 */
	public final double getWeight() {
		return myWeight;
	}

	/** Sets the cube of the item, must be >=0
	 * @param cube, the cube of the item
	 */
	public final void setCube(double cube) {
		if (cube < 0)
			throw new IllegalArgumentException("Cube must be >= 0");
		myCube = cube;
	}

	/** Sets the weight of the item, must be >=0
	 * @param weight, the weight of the item
	 */
	public final void setWeight(double weight) {
		if (weight <= 0)
			throw new IllegalArgumentException("Weight must be >= 0");
		myWeight = weight;
	}

	/**
	 * @param leadTimeD The leadtime distribution to set.
	 */
	public final void setLeadTimeDistributionInitialRandomSource(DistributionIfc leadTimeD) {
		if (leadTimeD == null)
			throw new IllegalArgumentException("DistributionIfc leadTimeD must be non-null");
		myLeadTimeD = leadTimeD;
		myLeadTime.setInitialRandomSource(leadTimeD);
	}
	
	/** Returns a value from the item type's default
	 *  leadtime distribution, a value of 0 indicates
	 *  no leadtime
	 * 
	 * @return
	 */
	public final double getLeadTime(){
		return(myLeadTime.getValue());
	}
	
	/**
	 * @return Returns the leadtime distribution
	 */
	public final DistributionFunctionIfc getLeadTimeDistribution() {
		return myLeadTimeD;
	}
}
