/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

import jsl.modeling.*;
import jsl.utilities.random.distributions.DistributionIfc;

/**
 */
public class LeadTimeReplenishmentProvider extends InventoryProviderAbstract implements InventoryIfc, InventoryShipperIfc {

    private TimeBasedItemTypeShipper myInventoryShipper;
    
	/**
	 * @param parent
	 */
	public LeadTimeReplenishmentProvider(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public LeadTimeReplenishmentProvider(ModelElement parent, String name) {
		super(parent, name);
		myInventoryShipper = new TimeBasedItemTypeShipper(this);
	}
	
	/** Asks the Inventory to provide inventory of the given
	 *  ItemType for the amount requested.  This provider has
	 *  unlimited supply of all item types and can handle any
	 *  demand request
	 * @param type The ItemType needed
	 * @param customer The customer who needs the inventory
	 * @param amountNeeded  The amount of ItemType needed
	 * 
	 * @return A Demand representing the request or null if the request cannot be satisfied.
	 */
	public final Demand requestInventory(ItemType type, InventoryCustomerIfc customer, int amountNeeded) {
		// make the demand to represent the request
		Demand d = Demand.create(type, amountNeeded);
		
		// set the requirements to whatever the customer wants, it really doesn't matter
		d.setAllowBackLoggingFlag(customer.permitReplenishmentBackLogging(type));
		d.setAllowPartialFillingFlag(customer.permitReplenishmentPartialFilling(type));
		
		// indicate that the demand will be filled immediately
		d.setRequestStatus(Demand.RequestStatus.FILL_ALL);
		// the customer will get all that they requested
		d.setInitialRequestFillAmount(amountNeeded);
		setInventoryProvider(d);
		setInventoryCustomer(customer, d);
		return d;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryProviderIfc#fillInventory(jsl.modeling.elements.supplychain.inventory.Demand)
	 */
	public final void fillInventory(Demand demand) {
		if (demand == null)
			throw new IllegalArgumentException("Invalid argument. Demand demand must be non-null.");
		
		// get how much can be supplied to the request
		int amtSupplied = demand.getInitialRequestFillAmount();
		// fill the amount supplied
		demand.fillDemand(amtSupplied);	
		shipReplenishment(demand);
	}

	/**
	 * @param type
	 * @param distribution
	 */
	public void addLeadTime(ItemType type, DistributionIfc distribution) {
		myInventoryShipper.addLeadTime(type, distribution);
	}
	
	/**
	 * @param d
	 */
	public final void shipReplenishment(Demand d) {
		myInventoryShipper.shipReplenishment(d);
	}

	public final boolean canShip(Demand d) {
		return true;
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryProviderIfc#providesInventory(jsl.modeling.elements.supplychain.inventory.ItemType, boolean, boolean)
	 */
	public final boolean providesInventory(ItemType type, InventoryCustomerIfc customer) {
		if (customer == null)
			throw new IllegalArgumentException("Invalid argument. InventoryCustomerIfc customer must be non-null.");

		return true;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryProviderIfc#getInventory(jsl.modeling.elements.supplychain.inventory.ItemType, boolean, boolean)
	 */
	public final InventoryIfc getInventory(ItemType type, InventoryCustomerIfc customer) {
		if (customer == null)
			throw new IllegalArgumentException("Invalid argument. InventoryCustomerIfc customer must be non-null.");

		return this;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryIfc#providesItemType(jsl.modeling.elements.supplychain.inventory.ItemType)
	 */
	public final boolean providesItemType(ItemType type) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryIfc#getInitialOnHand()
	 */
	public final int getInitialOnHand() {
		return Integer.MAX_VALUE;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryIfc#getInventoryPosition()
	 */
	public final int getInventoryPosition() {
		return Integer.MAX_VALUE;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryIfc#getAmountOnHand()
	 */
	public final int getAmountOnHand() {
		return Integer.MAX_VALUE;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryIfc#getAmountOnOrder()
	 */
	public final int getAmountOnOrder() {
		return Integer.MAX_VALUE;
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryIfc#getAmountBackLogged()
	 */
	public final int getAmountBackLogged() {
		return 0;
	}

}
