/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

import java.util.logging.Level;

import jsl.modeling.*;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.reporting.JSL;
import jslx.scinventory.*;

/**
 */
public class TestInventoryHoldingPoints {

	public static void main(String[] args) {

		System.out.println("Test InventoryElements");
        
		JSL.LOGGER.setLevel(Level.INFO);
   
//		testReplication();
		testExperiment();
        
		System.out.println("Done!");		
	}
	
	private static Model buildModel(){
		
		// create the containing model
		Model m = Model.createModel();
		
		// create the distributions
		DistributionIfc tbdA1 = new Exponential(1.0/3.6);
		DistributionIfc ltA1 = new Constant(0.5);
		int rptA1 = -1;
		int qtyA1 = 2;
		int levelA1 = 0;
		
		DistributionIfc tbdA2 = new Exponential(1.0/3.6);
		DistributionIfc ltA2 = new Constant(0.5);
		int rptA2 = -1;
		int qtyA2 = 2;
		int levelA2 = 0;

		DistributionIfc tbdB1 = new Exponential(1.0/3.6);
		DistributionIfc ltB1 = new Constant(0.5);
		int rptB1 = -1;
		int qtyB1 = 2;
		int levelB1 = 0;
		
		DistributionIfc tbdB2 = new Exponential(1.0/3.6);
		DistributionIfc ltB2 = new Constant(0.5);
		int rptB2 = -1;
		int qtyB2 = 2;
		int levelB2 = 0;
		
		// define the types
		ItemType typeA1 = new ItemType(m, "Type-A1");
		ItemType typeA2 = new ItemType(m, "Type-A2");
		ItemType typeB1 = new ItemType(m, "Type-B1");
		ItemType typeB2 = new ItemType(m, "Type-B2");
		
		// define the top level provider and leadtimes
		LeadTimeReplenishmentProvider supplier = new LeadTimeReplenishmentProvider(m);
		supplier.addLeadTime(typeA1,ltA1);
		supplier.addLeadTime(typeA2,ltA2);
		supplier.addLeadTime(typeB1,ltB1);
		supplier.addLeadTime(typeB2,ltB2);
		
		// define the IHP's, tell the IHP's to use the supplier for replenishment and for shipping
		InventoryHoldingPoint ihpA = new InventoryHoldingPoint(m, supplier, supplier);
		InventoryHoldingPoint ihpB = new InventoryHoldingPoint(m, supplier, supplier);
		
		// create the inventory at each IHP, assumes backlogging allowed, no partial filling
		ihpA.addRQInventoryItem(typeA1,0,rptA1,qtyA1);
		ihpA.addRQInventoryItem(typeA2,0,rptA2,qtyA2);
		ihpB.addRQInventoryItem(typeB1,0,rptB1,qtyB1);
		ihpB.addRQInventoryItem(typeB2,0,rptB2,qtyB2);
				
		// define the demand generator
		DemandGenerator dgA1 = new DemandGenerator(m, tbdA1, tbdA1);		
		dgA1.setDemandRequirements(typeA1, true, false, ihpA);
		DemandGenerator dgA2 = new DemandGenerator(m, tbdA2, tbdA2);		
		dgA2.setDemandRequirements(typeA2, true, false, ihpA);

		DemandGenerator dgB1 = new DemandGenerator(m, tbdB1, tbdB1);		
		dgB1.setDemandRequirements(typeB1, true, false, ihpB);
		DemandGenerator dgB2 = new DemandGenerator(m, tbdB2, tbdB2);		
		dgB2.setDemandRequirements(typeB2, true, false, ihpB);
		
		return(m);
	}
	
	public static void testExperiment(){
		
		Model m = buildModel();
		
//		m.turnOnLogReport();
		
		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
		// set the parameters of the experiment
		e.setNumberOfReplications(30);
		e.setLengthOfReplication(3360.0);
		e.setLengthOfWarmUp(360.0);
		
		// turn on the desired reporting
//		e.turnOnBatchReport(); 
		e.turnOnExperimentReport();
		e.turnOnReplicationReport();
		
		// tell the experiment to run
		e.runAll();
	}

}
