/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.scinventory;

import java.util.HashMap;
import java.util.Map;

import jsl.modeling.ActionListenerIfc;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.distributions.DistributionIfc;

/**
 *
 */
public class TimeBasedItemTypeShipper extends InventoryShipperAbstract {

	private ReplenishmentArrival myReplenishmentArrivalListener;

	/** A Map to hold the leadtimes by ItemType
	 *  key = ItemType, object = RandomVariable representing the leadtime
	 */	
	private Map<ItemType, RandomVariable> myLeadTimes;

	/**
	 * @param parent
	 */
	public TimeBasedItemTypeShipper(ModelElement parent) {
		this(parent, null);
		myReplenishmentArrivalListener = new ReplenishmentArrival();
	}

	/**
	 * @param parent
	 * @param name
	 */
	public TimeBasedItemTypeShipper(ModelElement parent, String name) {
		super(parent, name);
		myLeadTimes = new HashMap<ItemType, RandomVariable>();
	}
	
	public final void addLeadTime(ItemType type, DistributionIfc distribution){
		if (type == null)
			throw new IllegalArgumentException("The ItemType type must be non-null.");

		if (distribution == null)
			throw new IllegalArgumentException("The DistributionIfc distribution must be non-null.");
		
		if (myLeadTimes.containsKey(type)){ // item type has already been added
			// need only to set the distribution associated with the type
			RandomVariable rv = myLeadTimes.get(type);
			rv.setInitialRandomSource(distribution);			
		} else {
			// new item type being entered
			// create the random variable
			RandomVariable rv = new RandomVariable(this, distribution);
			myLeadTimes.put(type, rv);
		}
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryShipperIfc#canShip(jsl.modeling.elements.supplychain.inventory.Demand)
	 */
	public boolean canShip(Demand d) {
		
		// get the type of the demand
		ItemType type = d.getItemType();
		
		return myLeadTimes.containsKey(type);
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.supplychain.inventory.InventoryShipperIfc#shipReplenishment(jsl.modeling.elements.supplychain.inventory.Demand)
	 */
	public void shipReplenishment(Demand demand) {

		// get the type of the demand
		ItemType type = demand.getItemType();
		// get the associated leadtime random variable
		RandomVariable rv = myLeadTimes.get(type);
		if (rv == null){
			String s = "The ItemType " + type.getName() + " did not have a leadtime distribution ";
			throw new NoShipperOptionException(s);
		}

		// generate the leadtime
		double leadtime = rv.getValue();
		// schedule the replenishment for the demand (customer)
		scheduleEvent(myReplenishmentArrivalListener, leadtime, "Replenishment Arrival", demand);
	}

	private class ReplenishmentArrival implements ActionListenerIfc {
		public void action(JSLEvent event) {
			Demand demand = (Demand)event.getMessage();
			InventoryCustomerIfc customer = demand.getInventoryCustomer();
			customer.receiveReplenishment(demand);
		}
	}

}
