/*
* Created on Aug 27, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.supplychain.inventorylayer.DemandFillerIfc;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;

/** A MultiEchelonIHPNetwork models an abhorescent "tree" network of inventory holding points (IHPs).
 *  In this type of network, each IHP can have one parent IHP that fills its replenishment requests.  It
 *  can have many child IHPs to which it fills replenshments.  No checking is done to ensure
 *  that the IHPs are compatible in terms of their ability to supply certain ItemTypes.  Any demands
 *  sent to an IHP that cannot be handled by the IHP will be rejected.  The default rejection
 *  behavior for an IHP is to throw an exception.  To override this behavior the client must supply
 *  a DemandListenerRejectedIfc for the IHP or DemandGenerator.
 *  
 *  The top level of the network is modeled with a LeadTimeDemandFiller and
 *  is considered an external supplier to the network.  The external supplier (LeadTimeDemandFiller)
 *  has an infinite supply of the ItemTypes added to the network, which can be produced and delivered
 *  after a given leadtime, which may depend on the ItemType.  Since the top level supplier is "external"
 *  to the network the leadtime represents the time to produce the item.  If no item types are added 
 *  then all demands sent to the external supplier will be rejected.
 *  
 *  The generation of external demand to the IHPs is modeled through the use of instances of 
 *  the DemandGenerator class.  Each IHP may have zero or more DemandGenerators attached to it, and
 *  each DemandGenerator may be attached to one or more IHPs.  If no DemandGenerators are attached to any IHP (or to the
 *  network) then nothing will happen, since IHPs must have demands sent to them to begin processing.
 *  
 *  In addition, external DemandGenerators may be attached to the network by calling the 
 *  allowExternalDemandGenerators() method which allows the IHPs to ship/deliver to the external
 *  demand generators with zero delay.  The external demand generators must know where to
 *  send their demands (typically through a DemandFillerFinderIfc that they have access to).
 *  
 *  NOTE:  It is up to clients to provide transport between locations.  This class does not explicitly model
 *  the transport between locations.  The filled demands are automatically shipped and delivered after being filled.
 *  To model transport provide a DemandCarrierIfc to the IHP's
 *
 */
public class MultiEchelonIHPNetwork extends ModelElement {

	/** Models the lead time to produce the item type at the external supplier
	 */
	private LeadTimeDemandFiller myExternalSupplier;

	/** Collects statistics on the aggregate amount backordered
	 */
	private AggregateTimeWeightedVariable myTotalAmtBackOrdered;
	
	/** Collects statistics on the aggregate number waiting in backorder
	 */
	private AggregateTimeWeightedVariable myTotalNumBackOrdered;

	/** Collects statistics on the aggregate inventory
	 */
	private AggregateTimeWeightedVariable myTotalOnHand;

	/** Collects statistics on the aggregate on order
	 */
	private AggregateTimeWeightedVariable myTotalOnOrder;

	/** Collects statistics on the aggregate stockout
	 */
	private AggregateTimeWeightedVariable myTotalStockOutIndicator;

	/** Collects statistics on the aggregate first fill rate
	 */
	private ResponseVariableAverageObserver myAvgFirstFillRate;

	/** Collects statistics on the aggregate customer waiting time
	 */
	private ResponseVariableAverageObserver myAvgCustomerWaitTime;
	
	/** Counts all the replenishments made by the network
	 */
	private AggregateCounter myTotalNumOfReplenishments;
	
	/** Holds the inventory holding points and their children as a list
	 */
	private Map<InventoryHoldingPoint, List<InventoryHoldingPoint>> myIHPs;
	
	/** Holds all the IHPs that act as customers to other IHPs in the network
	 */
	private Set<InventoryHoldingPoint> myCustomerIHPs;
	
	/** Holds all the IHPs that are attached to the external supplier
	 */
	private Set<InventoryHoldingPoint> myIHPsAttachedToExternalSupplier;
	
	/** Holds the DemandGenerators for an IHP if they are attached
	 */
	private Map<InventoryHoldingPoint, List<DemandGenerator>> myDemandGenerators;
	
	/**
	 * @param parent
	 */
	public MultiEchelonIHPNetwork(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public MultiEchelonIHPNetwork(ModelElement parent, String name) {
		super(parent, name);
		myIHPs = new HashMap<InventoryHoldingPoint, List<InventoryHoldingPoint>>();
		myCustomerIHPs = new HashSet<InventoryHoldingPoint>();
		myIHPsAttachedToExternalSupplier = new HashSet<InventoryHoldingPoint>();
		myDemandGenerators = new HashMap<InventoryHoldingPoint, List<DemandGenerator>>();
		
		myTotalOnHand = new AggregateTimeWeightedVariable(this, getName() + " Total On Hand");
		myAvgFirstFillRate = new ResponseVariableAverageObserver(this, getName() + " Aggregate Fill Rate");
		myTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, getName() + " Aggregate Number of Stockouts");
		myTotalOnOrder = new AggregateTimeWeightedVariable(this, getName() + " Total On Order");
		myTotalNumOfReplenishments = new AggregateCounter(this, getName() + " NumberOfReplenishments");
		myTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Amt BackOrdered");
		myTotalNumBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Number BackOrdered");
		myAvgCustomerWaitTime = new ResponseVariableAverageObserver(this, getName() + "Avg Customer Wait");
		
		//create external LeadTimeDemandFiller and its own TimeBasedDemandShipper
		myExternalSupplier = new LeadTimeDemandFiller(this, getName() + "ExternalSupplier");

	}

	/** Adds an ItemType with a default name and given leadtime distribution
	 *  to the network
	 * 
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(RandomIfc leadtime){
		return(addItemType(null, leadtime));
	}
	
	/** Adds an ItemType and its lead time distribution to the network. The name
	 *  of the item type must be unique to the model
	 * 
	 * @param name
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(String name, RandomIfc leadtime){
		ItemType type = new ItemType(this, name);
		myExternalSupplier.addLeadTime(type, leadtime);
		return(type);
	}
	
	/** Causes an InventoryHoldingPoint to be created with a uniquely assigned name
	 *  and added to the network.  This only adds the IHP to the network it does
	 *  not cause the IHP to be connected to any other IHPs in the network. The client
	 *  can use the returned reference to the IHP to connect to other IHPs.
	 * 
	 * @return the created IHP
	 */
	public final InventoryHoldingPoint addInventoryHoldingPoint(){
		return(addInventoryHoldingPoint(null));
	}
	
	/** Causes an InventoryHoldingPoint to be created with the supplied name
	 *  and added to the network.  This only adds the IHP to the network it does
	 *  not cause the IHP to be connected to any other IHPs in the network. The client
	 *  can use the returned reference to the IHP to connect to other IHPs.
	 * 
	 * @param name must be unique in the Model
	 * @return the created IHP
	 */
	public final InventoryHoldingPoint addInventoryHoldingPoint(String name){
		// create the ihp and add it to the network so that it may have child ihps
		InventoryHoldingPoint ihp = new InventoryHoldingPoint(this, name);
		myIHPs.put(ihp, new ArrayList<InventoryHoldingPoint>());
		
		// attach aggregate response variable observers to each ihp, to get network performance
		ihp.subscribeToAggregateOnHandInventory(myTotalOnHand);
		ihp.subscribeToAggregateAmountOnOrder(myTotalOnOrder);
		ihp.subscribeToAggregateStockOutIndicator(myTotalStockOutIndicator);
		ihp.subscribeToAggregateAvgFirstFillRate(myAvgFirstFillRate);
		ihp.subscribeToAggregateReplenishmentDemandCounter(myTotalNumOfReplenishments);		
		ihp.subscribeToAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		ihp.subscribeToAggregateNumberBackOrdered(myTotalNumBackOrdered);
		ihp.subscribeToAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);

		return(ihp);
	}

	/** Checks if the supplied IHP is attached to the external supplier of the network
	 * 
	 * @param ihp
	 * @return true if attached
	 */
	public final boolean isAttachedToExternalSupplier(InventoryHoldingPoint ihp){
		return(myIHPsAttachedToExternalSupplier.contains(ihp));
	}
	
	/** Checks if the supplied IHP is attached as a customer in the network.
	 *  An IHP may only be attached to one IHP as a customer
	 * 
	 * @param ihp
	 * @return true if it a customer somewhere in the network
	 */
	public final boolean isAttachedAsCustomer(InventoryHoldingPoint ihp){
		return(myCustomerIHPs.contains(ihp));
	}
	
	/** Checks if the supplied IHP is attached as a customer for the supplied supplier.
	 *  An IHP may only be attached to one IHP as a customer
	 * 
	 * @param supplier
	 * @param customer
	 * @return true if it is a customer of the provided supplier
	 */
	public final boolean isCustomer(InventoryHoldingPoint supplier, InventoryHoldingPoint customer){
		if (!myIHPs.containsKey(supplier))
			return(false);
		
		if (!myIHPs.containsKey(customer))
			return(false);
				
		List<InventoryHoldingPoint> customers = myIHPs.get(supplier);
		return(customers.contains(customer) && (customer.getDemandFiller() == supplier));		
	}
	
	/** Gets the DemandFillerIfc that is serving as the external supplier to the network
	 *  This reference can be used to establish transport between the external supplier
	 *  and the IHP's that are attached to it through a DemandCarrierIfc
	 *  
	 * @return the externalSupplier
	 */
	public DemandFillerIfc getExternalSupplier() {
		return myExternalSupplier;
	}

	/** Attaches the IHP to the external supplier 
	 * 
	 * @param ihp Must not already be attached to the external supplier or to another IHP
	 * @param transportTime
	 */
	public final void attachIHPToExternalSupplier(InventoryHoldingPoint ihp){
		if (myIHPsAttachedToExternalSupplier.contains(ihp))
			throw new IllegalArgumentException("The supplied IHP is already attached to the external supplier.");
		
		if (myCustomerIHPs.contains(ihp))
			throw new IllegalArgumentException("The supplied IHP is already a customer to another IHP");
		
		myIHPsAttachedToExternalSupplier.add(ihp);

		// designate its demand filler as the external supplier
		ihp.setDemandFiller(myExternalSupplier);
	}

	/** Attaches the IHP (customer) to the given IHP (supplier) with the given transport delay
	 *  between them for deliveries
	 * 
	 * @param supplier Must be in the network
	 * @param customer Must be in the network and must not be a customer already on another IHP
	 */
	public final void attachIHPToSupplier(InventoryHoldingPoint supplier, InventoryHoldingPoint customer){
		if (!myIHPs.containsKey(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		if (!myIHPs.containsKey(customer))
			throw new IllegalArgumentException("The provided customer has not be added to the network");

		if (myCustomerIHPs.contains(customer))
			throw new IllegalArgumentException("The supplied customer IHP is already a customer IHP");
		
		//both in the network and customer is not already a customer, can add it to the supplier's list
		List<InventoryHoldingPoint> customers = myIHPs.get(supplier);
		customers.add(customer);
		// designate it as a customer
		myCustomerIHPs.add(customer);
		// ensure that it uses the supplier for its replensishments
		customer.setDemandFiller(supplier);
	
	}

	/** Creates and attaches a demand generator to the IHP
	 * 
	 * @param supplier
	 * @param type
	 * @param timeUntilNext
	 * @return
	 */
	public final DemandGenerator attachDemandGeneratorToIHP(InventoryHoldingPoint supplier,
			ItemType type, DistributionIfc timeUntilNext){
		return(attachDemandGeneratorToIHP(supplier, type, timeUntilNext, null));
	}
	
	/** Creates and attaches a demand generator to the IHP
	 * 
	 * @param supplier
	 * @param type
	 * @param timeUntilNext
	 * @param name
	 * @return
	 */
	public final DemandGenerator attachDemandGeneratorToIHP(InventoryHoldingPoint supplier,
			ItemType type, DistributionIfc timeUntilNext, String name){
		DemandGenerator dg = new DemandGenerator(this, type, timeUntilNext, name);
		attachDemandGeneratorToIHP(supplier, dg);
		return(dg);
	}

	/** Attaches the supplied demand generator as a customer on the given IHP (supplier)
	 *  
	 * @param supplier Must be in the network
	 * @param generator Must not have been already attached to the given supplier
	 */
	public final void attachDemandGeneratorToIHP(InventoryHoldingPoint supplier, DemandGenerator generator){
		if (!myIHPs.containsKey(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		List<DemandGenerator> generators = myDemandGenerators.get(supplier);
		if (generators == null){
			generators = new ArrayList<DemandGenerator>();
			myDemandGenerators.put(supplier, generators);
		}
		if (generators.contains(generator))
			throw new IllegalArgumentException("The provided generator is already attached to the supplier");
		
		generators.add(generator);
		generator.setDemandFiller(supplier);	
	}
	
	/** Returns a List copy of all the IHPs currently attached to the network
	 *  This method can be used to assist in building an instance of a class that implements 
	 *  the DemandFillerFinderIfc interface and needs a list of all possible
	 *  IHPs in the network. For example, this could be useful in developing an
	 *  external demand generator that can probabilistically send to any IHP.
	 *  
	 * @return
	 */
	public final List<InventoryHoldingPoint> getInventoryHoldingPoints(){
		return(new ArrayList<InventoryHoldingPoint>(myIHPs.keySet()));
	}
}
