/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AggregateCounter;

import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;



import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeOrderFiller;
import jslx.supplychain.inventorylayer.OrderGenerator;
import jslx.supplychain.inventorylayer.Warehouse;
import jslx.supplychain.inventorylayer.WarehouseAbstract;
import jslx.supplychain.transportlayer.TimeBasedOrderCarrier;

/** A TimeBasedShippingMultiEchelonInventoryOrderNetwork models an abhorescent "tree" network of
 *  warehouses (WarehouseAbstract). A warehouse is a location that can fill orders (OrderFillerIfc)
 *  and can send orders (OrderSenderIfc) for replenishment. Orders consist of many demands for item types.
 *  
 *  In this type of network, each warehouse can have one parent warehouse that fills its replenishment order.  It
 *  can have many child warehouses to which it fills replenshments.  No checking is done to ensure
 *  that the warehouses are compatible in terms of their ability to supply certain ItemTypes.  Any orders
 *  sent to a warehouse that cannot be handled by the warehouse will be rejected.  The default rejection
 *  behavior for the warehouse is to thrown an exception.  To override this behavior the client must supply
 *  an OrderListenerRejectedIfc for the warehouse or order generator.
 *  
 *  The top level of the network is modeled with a LeadTimeOrderFiller and
 *  is considered an external supplier to the network.  The external supplier (LeadTimeOrderFiller)
 *  has an infinite supply of the ItemTypes added to the network, which can be produced and delivered
 *  after a given leadtime, which may depend on the ItemType.  Since the top level supplier is "external"
 *  to the network the leadtime represents the time to produce the item.  If no item types are added 
 *  then all orders sent to the external supplier will be rejected.
 *  
 *  The time to transport from the external supplier may depend on the location of the customer (warehouse) 
 *  and is thus modelled as a separate transport time which may depend upon the customer and its location 
 *  relative to the external supplier.  If the time for the transport depends upon the link between the external
 *  supplier to the customer then a TimeBasedOrderShipper is used to model the transport time
 *  from the external supplier to the top level warehouses.
 *  
 *  The generation of external orders to the warehouses is modeled through the use of instances of 
 *  the OrderGenerator class.  Each warehouse may have zero or more OrderGenerators attached to it, and
 *  each OrderGenerator may be attached to one or more warehouses.  This may occur for any warehouse at any level
 *  of the network to represent external orders sent to that warehouse. During the attachment process, the
 *  client is responsible for specifying the time that it may take to transport the order to
 *  the OrderGenerator once it is filled. If no OrderGenerators are attached to any warehouse (or to the
 *  network) then nothing will happen, since warehouses must have orders sent to them to begin processing.
 *  
 *  In addition, external OrderGenerators may be attached to the network by calling the 
 *  allowExternalOrderGenerators() method which allows the warehouses to ship/deliver to the external
 *  order generators with zero delay.  The external order generators must know where to
 *  send their orders (typically through a OrderFillerFinderIfc that they have access to).
 *  
 *
 */
public class TimeBasedShippingMultiEchelonWarehouseNetwork extends ModelElement {

	/** Models the lead time to produce the item type at the external supplier
	 */
	private LeadTimeOrderFiller myExternalSupplier;
	
	/** Models the transport time (time to deliver) the item type from the external
	 *  supplier to the connected (top level) IHPs
	 */
	private TimeBasedOrderCarrier myExternalSupplierShipper;

	/** Collects statistics on the aggregate amount backordered
	 */
	private AggregateTimeWeightedVariable myTotalAmtBackOrdered;
	
	/** Collects statistics on the aggregate number waiting in backorder
	 */
	private AggregateTimeWeightedVariable myTotalNumBackOrdered;

	/** Collects statistics on the aggregate inventory
	 */
	private AggregateTimeWeightedVariable myTotalOnHand;

	/** Collects statistics on the aggregate on order
	 */
	private AggregateTimeWeightedVariable myTotalOnOrder;

	/** Collects statistics on the aggregate stockout
	 */
	private AggregateTimeWeightedVariable myTotalStockOutIndicator;

	/** Collects statistics on the aggregate first fill rate
	 */
	private ResponseVariableAverageObserver myAvgFirstFillRate;

	/** Collects statistics on the aggregate customer waiting time
	 */
	private ResponseVariableAverageObserver myAvgCustomerWaitTime;
	
	/** Counts all the replenishment demands made by the network
	 */
	private AggregateCounter myTotalNumOfReplenishmentDemands;
	
	/** Counts all the replenishment orders made by the network
	 */
	private AggregateCounter myTotalNumOfReplenishmentOrders;
	
	/** Collects statistics on the aggregate order waiting time
	 */
	private ResponseVariableAverageObserver myAvgOrderWaitTime;

	/** Collects statistics on the aggregate number of orders waiting
	 */
	private AggregateTimeWeightedVariable myTotalOrdersWaiting;

	/** Holds the inventory holding points and their children as a list
	 */
	private Map<WarehouseAbstract, List<WarehouseAbstract>> myWarehouses;
	
	/** Holds all the warehouses that act as customers to other warehouses in the network
	 */
	private Set<WarehouseAbstract> myCustomerWarehouses;
	
	/** Holds all the warehouses that are attached to the external supplier
	 */
	private Set<WarehouseAbstract> myWarehousesAttachedToExternalSupplier;
	
	/** Holds the OrderGenerators for a warehouse if they are attached
	 */
	private Map<WarehouseAbstract, List<OrderGenerator>> myOrderGenerators;

	/** Indicates whether the network will allow its warehouses to receive demand
	 *  from any external order generator.  Any external order generator
	 *  must be willing to accept a no delay shippment.
	 */
	private boolean myAllowExternalOrderGeneratorsFlag = false;
	
	/** Creates a MultiEchelonInventoryNetwork
	 * 
	 * @param parent
	 */
	public TimeBasedShippingMultiEchelonWarehouseNetwork(ModelElement parent) {
		this(parent, null);
	}
	
	/** Creates a MultiEchelonInventoryNetwork
	 * 
	 * @param parent
	 * @param name
	 */
	public TimeBasedShippingMultiEchelonWarehouseNetwork(ModelElement parent, String name) {
		super(parent, name);
		
		myWarehouses = new HashMap<WarehouseAbstract, List<WarehouseAbstract>>();
		myCustomerWarehouses = new HashSet<WarehouseAbstract>();
		myWarehousesAttachedToExternalSupplier = new HashSet<WarehouseAbstract>();
		myOrderGenerators = new HashMap<WarehouseAbstract, List<OrderGenerator>>();
		
		myTotalOnHand = new AggregateTimeWeightedVariable(this, getName() + " Total On Hand");
		myAvgFirstFillRate = new ResponseVariableAverageObserver(this, getName() + " Aggregate Fill Rate");
		myTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, getName() + " Aggregate Number of Stockouts");
		myTotalOnOrder = new AggregateTimeWeightedVariable(this, getName() + " Total On Order");
		myTotalNumOfReplenishmentDemands = new AggregateCounter(this, getName() + " NumberOfReplenishmentDemands");
		myTotalNumOfReplenishmentOrders = new AggregateCounter(this, getName() + " NumberOfReplenishmentOrders");
		myAvgOrderWaitTime = new ResponseVariableAverageObserver(this, getName() + " Aggregate Order Wait Time");
		myTotalOrdersWaiting = new AggregateTimeWeightedVariable(this, getName() + " Aggregate Num Orders Waiting");
		myAvgCustomerWaitTime = new ResponseVariableAverageObserver(this, getName() + "Avg Customer Wait");
		myTotalNumBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Number BackOrdered");
		myTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Amt BackOrdered");
		
		//create external LeadTimeDemandFiller and its own TimeBasedDemandShipper
		myExternalSupplier = new LeadTimeOrderFiller(this, getName() + "ExternalSupplier");
		myExternalSupplierShipper = new TimeBasedOrderCarrier(this, getName() + "ExternalShipper");
		myExternalSupplier.setOrderShipper(myExternalSupplierShipper);
	}
		
	/** Adds an ItemType with a default name and given leadtime distribution
	 *  to the network
	 * 
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(RandomIfc leadtime){
		return(addItemType(null, leadtime));
	}
	
	/** Adds an ItemType and its lead time distribution to the network. The name
	 *  of the item type must be unique to the model
	 * 
	 * @param name
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(String name, RandomIfc leadtime){
		ItemType type = new ItemType(this, name);
		myExternalSupplier.addLeadTime(type, leadtime);
		return(type);
	}
	
	/** Causes a Warehouse to be created with a uniquely assigned name
	 *  and added to the network.  This only adds the warehouse to the network it does
	 *  not cause the warehouse to be connected to any other warehouses in the network. The client
	 *  can use the returned reference to the warehouses to connect to other warehouses.
	 * 
	 * @return the created warehouse
	 */
	public final Warehouse addWarehouse(){
		Warehouse w = new Warehouse(this);
		addWarehouse(w);
		return(w);
	}
	
	/** Causes a Warehouse to be created with the supplied name
	 *  and added to the network.  This only adds the warehouse to the network it does
	 *  not cause the warehouse to be connected to any other warehouses in the network. The client
	 *  can use the returned reference to the warehouses to connect to other warehouses.
	 * 
	 * @param name must be unique in the Model
	 * @return the created IHP
	 */
	public final Warehouse addWarehouse(String name){
		// create the warehouse and add it to the network so that it may have child ihps
		Warehouse w = new Warehouse(this, name);
		addWarehouse(w);
		return(w);
	}

	public final void addWarehouse(WarehouseAbstract warehouse){
		
		myWarehouses.put(warehouse, new ArrayList<WarehouseAbstract>());
		// assign the warehouse its own time based shipper for shipping to its customers
		TimeBasedOrderCarrier shipper = new TimeBasedOrderCarrier(this, warehouse.getName() + "TimeBasedShipper");
		warehouse.setOrderShipper(shipper);
		if (myAllowExternalOrderGeneratorsFlag == true){
			shipper.setImmediateTransportFlag(true);
		}
		// attach aggregate response variable observers to each warehouse, to get network performance
				
		warehouse.subscribeToAggregateOnHandInventory(myTotalOnHand);
		warehouse.subscribeToAggregateAmountOnOrder(myTotalOnOrder);
		warehouse.subscribeToAggregateStockOutIndicator(myTotalStockOutIndicator);
		warehouse.subscribeToAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		warehouse.subscribeToAggregateNumberBackOrdered(myTotalNumBackOrdered);
		warehouse.subscribeToAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);

		warehouse.subscribeToAggregateAvgFirstFillRate(myAvgFirstFillRate);
		
		warehouse.subscribeToAggregateReplenishmentDemandCounter(myTotalNumOfReplenishmentDemands);
		warehouse.attachAggregateReplenishmentOrderCounter(myTotalNumOfReplenishmentOrders);
		
		warehouse.attachAggregateTimeWeightedVariableToNumberOfOrdersWaiting(myTotalOrdersWaiting);
		warehouse.attachAggregateResponseVariableToOrderWaitingTime(myAvgOrderWaitTime);
		
	}
	
	/** Checks if the supplied Warehouse is attached to the external supplier of the network
	 * 
	 * @param warehouse
	 * @return true if attached
	 */
	public final boolean isAttachedToExternalSuppler(WarehouseAbstract warehouse){
		return(myWarehousesAttachedToExternalSupplier.contains(warehouse));
	}
	
	/** Checks if the supplied Warehouse is attached as a customer in the network.
	 *  A Warehouse may only be attached to one warehouse as a customer
	 * 
	 * @param warehouse
	 * @return true if it a customer somewhere in the network
	 */
	public final boolean isAttachedAsCustomer(WarehouseAbstract warehouse){
		return(myCustomerWarehouses.contains(warehouse));
	}
	
	/** Checks if the supplied warehouse is attached as a customer for the supplied supplier.
	 *  An warehouse may only be attached to one IHP as a customer
	 * 
	 * @param supplier
	 * @param customer
	 * @return true if it is a customer of the provided supplier
	 */
	public final boolean isCustomer(WarehouseAbstract supplier, WarehouseAbstract customer){
		if (!myWarehouses.containsKey(supplier))
			return(false);
		
		if (!myWarehouses.containsKey(customer))
			return(false);
				
		List<WarehouseAbstract> customers = myWarehouses.get(supplier);
		return(customers.contains(customer) && (customer.getReplenishmentOrderFiller() == supplier));		
	}
	
	/** Attaches the warehouse to the external supplier with a no transport delay
	 * 
	 * @param ihp Must not already be attached to the external supplier or to another warehouse
	 */
	public final void attachWarehouseToExternalSupplier(WarehouseAbstract warehouse){
		attachWarehouseToExternalSupplier(warehouse, null);
	}

	/** Attaches the warehouse to the external supplier with the given transport delay
	 * 
	 * @param ihp Must not already be attached to the external supplier or to another warehouse
	 * @param transportTime
	 */
	public final void attachWarehouseToExternalSupplier(WarehouseAbstract warehouse, DistributionIfc transportTime){
		if (myWarehousesAttachedToExternalSupplier.contains(warehouse))
			throw new IllegalArgumentException("The supplied warehouse is already attached to the external supplier.");
		
		if (myCustomerWarehouses.contains(warehouse))
			throw new IllegalArgumentException("The supplied warehouse is already a customer to another IHP");
		
		myWarehousesAttachedToExternalSupplier.add(warehouse);
		if (transportTime != null)
			myExternalSupplierShipper.setTransportTime(warehouse, transportTime);
		else
			myExternalSupplierShipper.setImmediateTransportFlag(true);	

		// designate its order filler as the external supplier
		warehouse.setReplenishmentOrderFiller(myExternalSupplier);
	}

	/** Attaches the warehouse (customer) to the given warehouse (supplier) with no transport delay
	 *  between them for deliveries
	 * 
	 * @param supplier Must be in the network
	 * @param customer Must be in the network and must not be a customer already on another warehouse
	 */
	public final void attachWarehouseToSupplier(WarehouseAbstract supplier, WarehouseAbstract customer){
		attachWarehouseToSupplier(supplier, customer, null);
	}

	/** Attaches the warehouse (customer) to the given warehouse (supplier) with the given transport delay
	 *  between them for deliveries
	 * 
	 * @param supplier Must be in the network
	 * @param customer Must be in the network and must not be a customer already on another warehouse
	 * @param transportTime
	 */
	public final void attachWarehouseToSupplier(WarehouseAbstract supplier, WarehouseAbstract customer, 
			DistributionIfc transportTime){
		if (!myWarehouses.containsKey(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		if (!myWarehouses.containsKey(customer))
			throw new IllegalArgumentException("The provided customer has not be added to the network");

		if (myCustomerWarehouses.contains(customer))
			throw new IllegalArgumentException("The supplied customer IHP is already a customer IHP");
		
		//both in the network and customer is not already a customer, can add it to the supplier's list
		List<WarehouseAbstract> customers = myWarehouses.get(supplier);
		customers.add(customer);
		// designate it as a customer
		myCustomerWarehouses.add(customer);
		// ensure that it uses the supplier for its replensishments
		customer.setReplenishmentOrderFiller(supplier);
		
		if (transportTime != null){
			// ensure that it uses its supplier's shipper
			TimeBasedOrderCarrier shipper = (TimeBasedOrderCarrier)supplier.getOrderShipper();
			shipper.setTransportTime(customer, transportTime);
		} else {
			// directing an attached IHP to be shipped with no delay time
			// designate the time based shipper to allow no delay shipment
			TimeBasedOrderCarrier shipper = (TimeBasedOrderCarrier)supplier.getOrderShipper();
			shipper.setImmediateTransportFlag(true);
		}
	}	
	
	/** Attaches the supplied demand generator as customer on the given Warehouse (supplier)
	 *  with no transport delay
	 *  
	 * @param supplier Must be in the network
	 * @param generator Must not have been already attached to the given supplier
	 */
	public final void attachOrderGeneratorToWarehouse(WarehouseAbstract supplier, OrderGenerator generator){
		attachOrderGeneratorToWarehouse(supplier, generator, null);
	}

	/** Attaches the supplied demand generator as a customer on the given Warehouse (supplier)
	 *  with the given transport delay
	 *  
	 * @param supplier Must be in the network
	 * @param generator Must not have been already attached to the given supplier
	 * @param transportTime
	 */
	public final void attachOrderGeneratorToWarehouse(WarehouseAbstract supplier, OrderGenerator generator,
			DistributionIfc transportTime){
		if (!myWarehouses.containsKey(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		List<OrderGenerator> generators = myOrderGenerators.get(supplier);
		if (generators == null){
			generators = new ArrayList<OrderGenerator>();
			myOrderGenerators.put(supplier, generators);
		}
		if (generators.contains(generator))
			throw new IllegalArgumentException("The provided generator is already attached to the supplier");
		
		generators.add(generator);
		generator.setOrderFiller(supplier);
		if (transportTime != null){
			// ensure that it uses its supplier's shipper
			TimeBasedOrderCarrier shipper = (TimeBasedOrderCarrier)supplier.getOrderShipper();
			shipper.setTransportTime(generator, transportTime);
		} else {
			// directing an attached Warehouse to be shipped with no delay time
			// designate the time based shipper to allow no delay shipment
			TimeBasedOrderCarrier shipper = (TimeBasedOrderCarrier)supplier.getOrderShipper();
			shipper.setImmediateTransportFlag(true);
		}		
	}
	
	/** Causes the Warehouses in the network to be able to deliver completed demands from 
	 *  any external demand generator with a no delay transport time.  Any Warehouses attached
	 *  after this option has been set will also have this capability within their attached shipper
	 */
	public final void allowExternalOrderGenerators(){
		myAllowExternalOrderGeneratorsFlag = true;
		for (WarehouseAbstract ihp: myWarehouses.keySet()){
			TimeBasedOrderCarrier shipper = (TimeBasedOrderCarrier)ihp.getOrderShipper();
			shipper.setImmediateTransportFlag(true);
		}
	}
	
	/** Returns a List copy of all the Warehouses currently attached to the network
	 *  This method can be used to assist in building an instance of a class that implements 
	 *  the DemandFillerFinderIfc interface and needs a list of all possible
	 *  Warehouses in the network. For example, this could be useful in developing an
	 *  external demand generator that can probabilistically send to any Warehouse.
	 *  
	 * @return
	 */
	public final List<WarehouseAbstract> getWarehouses(){
		return(new ArrayList<WarehouseAbstract>(myWarehouses.keySet()));
	}
	
}
