/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.afosr.examples;

import jsl.modeling.ModelElement;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.supplychain.afosr.forecasting.DemandArrivalListenerDefault;
import jslx.supplychain.inventorylayer.BackLogQueue;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandFillerFinderIfc;
import jslx.supplychain.inventorylayer.DemandFillerIfc;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.DemandSenderIfc;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryPolicyAbstract;
import jslx.supplychain.inventorylayer.InventoryPolicyReorderPointOrderUpToLevel;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;

/** 
 * This class models a Poisson demand arrival on an inventory that follows a base stock policy
 * and is replenished by a lead time demand filler. After every aggregation level the demand for that
 * time bucket is aggregated.  
 *
 */
public class DemandAggregatorModel extends ModelElement implements DemandFillerFinderIfc {

	private ItemType myItemType;
	private DemandGenerator myDemandGenerator;
	private Inventory myInventory;
    private InventoryPolicyAbstract myInventoryPolicy;
    private LeadTimeDemandFiller myLeadTimeDemandFiller;
    private int myBasestock;
    
	/**
	 * @param parent
	 */
	public DemandAggregatorModel(ModelElement parent, RandomIfc timebtwDemand, 
						    DistributionIfc leadTime, int basestock, int initialLevel) {
		this(parent, timebtwDemand, leadTime, basestock, initialLevel, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public DemandAggregatorModel(ModelElement parent, RandomIfc timebtwDemand, DistributionIfc leadTime, 
	                        int basestock, int initialLevel, String name) {
		super(parent, name);

		// define the type
		myItemType = new ItemType(this, "Type-A");
		
		// define the filler
		myLeadTimeDemandFiller = new LeadTimeDemandFiller(this);
		myLeadTimeDemandFiller.addLeadTime(myItemType,leadTime);
		
		myBasestock = basestock;
		myInventoryPolicy = new InventoryPolicyReorderPointOrderUpToLevel(this, myBasestock, myBasestock+1);		
		myInventory = new Inventory(this, myItemType, myInventoryPolicy, initialLevel);		
		BackLogQueue bpolicy = new BackLogQueue(myInventory, myInventory.getName()+ "-BackLogQ");		
		
	    myInventory.setDemandFillerFinder(this);
	    
	    int agglevel = 1;
	    myInventory.setDemandArrivalListener(new DemandArrivalListenerDefault(this, myInventory, myInventory.getName() + " DemandObserver",agglevel));
	    
	    // define the demand generator
		myDemandGenerator = new DemandGenerator(this, myItemType, timebtwDemand, timebtwDemand);
		myDemandGenerator.setDemandFillerFinder(this);
		
	}

	public DemandFillerIfc findDemandFiller(Demand demand) {
		DemandSenderIfc sender = demand.getSender();
		if (sender == myDemandGenerator)
			return(myInventory);
		else
			return(myLeadTimeDemandFiller);

	}

}
