/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.afosr.examples;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.forecasting.ForecastTechnique;
import jslx.supplychain.afosr.forecasting.DemandAndForecastArrivalListenerDefault;
import jslx.supplychain.afosr.forecasting.ForecastManager;
import jslx.supplychain.afosr.forecasting.UpdatableIfc;
import jslx.supplychain.inventorylayer.BackLogQueue;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandFillerFinderIfc;
import jslx.supplychain.inventorylayer.DemandFillerIfc;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.DemandSenderIfc;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryPolicyAbstract;
import jslx.supplychain.inventorylayer.InventoryPolicyReorderPointOrderUpToLevel;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;

/**
 * This class models a (s,S) inventory model that is optimized by the 
 * an optimization algorithm as in Hoff and Spearman. Subsequently the 
 * policy parameters are updated. 
 * 
 */
public class ForecastBasedUpdatingsSInventoryModel extends SchedulingElement implements DemandFillerFinderIfc, UpdatableIfc {

	protected ItemType myItemType;
	protected DistributionIfc myLeadTime;
	protected DemandGenerator myDemandGenerator;
	protected Inventory myInventory;
	protected LeadTimeDemandFiller myLeadTimeDemandFiller;
	protected InventoryPolicyAbstract myInventoryPolicy;
	protected int myReorderPoint;
	protected int myOrderUptoLevel;

    public DemandAndForecastArrivalListenerDefault myDemandArrivalListener;
    protected int myDemandAggregationAndPolicyUpdatingLevel;
    
    protected ForecastTechnique myForecaster;
    protected DefaultsSInventoryUpdater myUpdater;    
    protected boolean UPDATABLE;
    public double myInitialDemandRateEstimate;
    
    public String myOutput;
	
    private Counter myNumOrderUptoLevelIncrement;
	private Counter myNumOrderUptoLevelDecrement;
	private ResponseVariable mySizeOrderUptoLevelIncrement;
	private ResponseVariable mySizeOrderUptoLevelIncrementVariance;
	private ResponseVariable mySizeOrderUptoLevelDecrement;
	private ResponseVariable mySizeOrderUptoLevelDecrementVariance;
    
	private Counter myNumReorderPointIncrement;
	private Counter myNumReorderPointDecrement;
	private ResponseVariable mySizeReorderPointIncrement;
	private ResponseVariable mySizeReorderPointIncrementVariance;
	private ResponseVariable mySizeReorderPointDecrement;
	private ResponseVariable mySizeReorderPointDecrementVariance;
    
	private Counter myNumOrderQuantityIncrement;
	private Counter myNumOrderQuantityDecrement;
	private ResponseVariable mySizeOrderQuantityIncrement;
	private ResponseVariable mySizeOrderQuantityIncrementVariance;
	private ResponseVariable mySizeOrderQuantityDecrement;
	private ResponseVariable mySizeOrderQuantityDecrementVariance;
	
	private TimeWeighted mySurplusOnHand;
	private TimeWeighted myTimeSpentInSurplusOnHand;
	private ResponseVariable myEndSurplusOnHand;
	private ResponseVariable myEndProbabilityOfSurplusOnHand;
	
	public ForecastBasedUpdatingsSInventoryModel(ModelElement parent, ItemType it, 
			RandomIfc timebtwDemand, RandomIfc timeUntilFirstDemand, DistributionIfc leadTime, 
			int initialLevel, int reorderpoint, int orderuptolevel, int level, 
			double initialDemandRateEstimate, String name,
			ForecastTechnique forecaster, boolean updatable) {
		super(parent, name);
		
		myNumOrderUptoLevelIncrement = new Counter(this, getName()+"NumOrderUptoLevelIncrement");		
		myNumOrderUptoLevelDecrement = new Counter(this, getName()+"NumOrderUptoLevelDecrement");		
		mySizeOrderUptoLevelIncrement = new ResponseVariable(this, getName()+"SizeOrderUptoLevelIncrement");		
		mySizeOrderUptoLevelIncrementVariance = new ResponseVariable(this, getName()+"SizeOrderUptoLevelIncrementVariance");		
		mySizeOrderUptoLevelDecrement = new ResponseVariable(this, getName()+"SizeOrderUptoLevelDecrement");		
		mySizeOrderUptoLevelDecrementVariance = new ResponseVariable(this, getName()+"SizeOrderUptoLevelDecrementVariance");
		
		myNumReorderPointIncrement = new Counter(this, getName()+"NumReorderPointIncrement");		
		myNumReorderPointDecrement = new Counter(this, getName()+"NumReorderPointDecrement");		
		mySizeReorderPointIncrement = new ResponseVariable(this, getName()+"SizeReorderPointIncrement");		
		mySizeReorderPointIncrementVariance = new ResponseVariable(this, getName()+"SizeReorderPointIncrementVariance");		
		mySizeReorderPointDecrement = new ResponseVariable(this, getName()+"SizeReorderPointDecrement");		
		mySizeReorderPointDecrementVariance = new ResponseVariable(this, getName()+"SizeReorderPointDecrementVariance");
		
		myNumOrderQuantityIncrement = new Counter(this, getName()+"NumOrderQuantityIncrement");		
		myNumOrderQuantityDecrement = new Counter(this, getName()+"NumOrderQuantityDecrement");		
		mySizeOrderQuantityIncrement = new ResponseVariable(this, getName()+"SizeOrderQuantityIncrement");		
		mySizeOrderQuantityIncrementVariance = new ResponseVariable(this, getName()+"SizeOrderQuantityIncrementVariance");		
		mySizeOrderQuantityDecrement = new ResponseVariable(this, getName()+"SizeOrderQuantityDecrement");		
		mySizeOrderQuantityDecrementVariance = new ResponseVariable(this, getName()+"SizeOrderQuantityDecrementVariance");
		
		mySurplusOnHand = new TimeWeighted(this, getName() + "-SurplusOnHand");
		myTimeSpentInSurplusOnHand = new TimeWeighted(this, getName() + "-TimeSpentInSurplusOnHand");
		myEndSurplusOnHand = new ResponseVariable(this, getName() + "-EndSurplusOnHand");
		myEndProbabilityOfSurplusOnHand = new ResponseVariable(this, getName() + "-EndProbabilityOfSurplusOnHand");
		
		// define the type
		myItemType = it;
		myLeadTime = leadTime;		
		// define the filler
		myLeadTimeDemandFiller = new LeadTimeDemandFiller(this);
		myLeadTimeDemandFiller.addLeadTime(myItemType,leadTime);		
		
		myReorderPoint = reorderpoint;
		myOrderUptoLevel = orderuptolevel;
		
		myInventoryPolicy = new InventoryPolicyReorderPointOrderUpToLevel(this, myReorderPoint, myOrderUptoLevel);		
		myInventory = new Inventory(this, myItemType, myInventoryPolicy, initialLevel);		
		BackLogQueue bpolicy = new BackLogQueue(myInventory, myInventory.getName()+ "-BackLogQ");
		
	    myInventory.setDemandFillerFinder(this);
	    
	    myDemandGenerator = new DemandGenerator(this, myItemType, timeUntilFirstDemand, timebtwDemand);
		myDemandGenerator.setDemandFillerFinder(this);
		
		myInitialDemandRateEstimate = initialDemandRateEstimate;
		UPDATABLE = updatable;
		
		DefaultsSInventoryUpdater updater = new DefaultsSInventoryUpdater(this, level);		
		setUpdater(updater);
		
		// the listener will listens to demand arrivals and it aggregates and forecasts acco to an aggregation level
		myDemandAggregationAndPolicyUpdatingLevel = level;
		myDemandArrivalListener = new DemandAndForecastArrivalListenerDefault(this, myInventory, this, this+"-DemandAndForecastArrivalListener", myDemandAggregationAndPolicyUpdatingLevel);
		
		//a ses forecaster is created
		myForecaster = forecaster;
		myDemandArrivalListener.setDemandForecaster(new ForecastManager(this, myForecaster, this+" Forecaster "+forecaster.getName()));
		
		try {
	        BufferedWriter out = new BufferedWriter(new FileWriter("FinalReport" + ".csv"));
	        out.write("Replication,Time,Demand,Forecast,s--,S--,OH--,OO--,PO--,BL--,IP--,s,S,OH,OO,PO,BL,IP");
	        out.close();
	    }
		catch (IOException e1){
	    }
	}
	
	public void initialize(){
		myUpdater.initparam();
		System.out.println("updating");
		scheduleEvent(myDemandAggregationAndPolicyUpdatingLevel, 2, 2, "PolicyUpdatingInterval");
	}
	
	protected void handleEvent(JSLEvent event){
		schedulePolicyUpdating(event);
	}
	
	/**
	 * @return Returns the myDemandArrivalListener.
	 */
	public DemandAndForecastArrivalListenerDefault getDemandArrivalListener() {
		return myDemandArrivalListener;
	}
	
	/**
	 * @return Returns the myInventory.
	 */
	private Inventory getInventory() {
		return myInventory;
	}
	
	/**
	 * @return Returns the myLeadTime.
	 */
	public DistributionIfc getLeadTime() {
		return myLeadTime;
	}
	
	/**
	 * @return Returns the myContinuousOptimized.
	 */
	public boolean isUpdatable() {
		return UPDATABLE;
	}
	
	public void setUpdater(DefaultsSInventoryUpdater updater){
		if(updater.getUpdatableInventoryModel()==this){
			myUpdater = updater;
		}		
	}

	public DemandFillerIfc findDemandFiller(Demand demand) {
		DemandSenderIfc sender = demand.getSender();
		if (sender == myDemandGenerator)
			return(myInventory);
		else
			return(myLeadTimeDemandFiller);
	}
	
	protected void schedulePolicyUpdating(JSLEvent event){
		//System.out.println("updating");
		if(this.myDemandArrivalListener.FORECASTINITIALIZED == true){
			double fcst = this.getDemandArrivalListener().getDemandForecaster().getForecast();
			double dem = this.getDemandArrivalListener().getDemandForecaster().getDemand();
			double time = this.getDemandArrivalListener().getDemandForecaster().getCurrentTimeIndex();
			double rep = this.getDemandArrivalListener().getCurrentReplication().getReplicationNumber();		
			myOutput = rep+","+time+","+dem+","+fcst+","+myReorderPoint+","+myOrderUptoLevel
							+","+this.getDemandArrivalListener().getAmountOnHand()
							+","+this.getDemandArrivalListener().getAmountOnOrder()
							+","+this.getDemandArrivalListener().getAmountPending()
							+","+this.getDemandArrivalListener().getAmountBackLogged()
							+","+this.getDemandArrivalListener().getInventoryPosition();			
			myUpdater.optimize(this);
			double t = myDemandAggregationAndPolicyUpdatingLevel;
			scheduleEvent(t, 2, 2, "PolicyUpdatingInterval");
		}
	}

	public void receiveInitialParameters(double [] par) {
		myReorderPoint = (int)par[0];
		myOrderUptoLevel = (int)par[1];
		myInventory.setInventoryPolicyParameters(par);
		myInventory.setInitialOnHand((int)par[1]);
	}

	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.OptimizableIfc#receiveOptimalParameters(java.lang.Object)
	 */
	public void receiveOptimalParameters(double [] par) {
		
		
		double o1 = myInventory.getInventoryPolicyParameters()[1];
		double o2 = myInventory.getInventoryPolicyParameters()[0];
		double o3 = o1-o2;
		
		myReorderPoint = (int)par[0];
		myOrderUptoLevel = (int)par[1];
		myInventory.setInventoryPolicyParameters(par);
		double n1 = myInventory.getInventoryPolicyParameters()[1];
		double n2 = myInventory.getInventoryPolicyParameters()[0];
		double n3 = n1-n2;
		
		int ip = myInventory.getInventoryPosition();
		
		double change1 = o1 - n1;
		if(change1 > 0){
			myNumOrderUptoLevelIncrement.increment();
			mySizeOrderUptoLevelIncrement.setValue(change1);
			if(ip>n1){						
				mySurplusOnHand.setValue(ip-n1);
				myTimeSpentInSurplusOnHand.setValue(1.0);
			}else{
				mySurplusOnHand.setValue(0.0);
				myTimeSpentInSurplusOnHand.setValue(0.0);
			}
		}else if (change1 < 0){
			myNumOrderUptoLevelDecrement.increment();
			mySizeOrderUptoLevelDecrement.setValue(Math.abs(change1));
			mySurplusOnHand.setValue(0.0);
			myTimeSpentInSurplusOnHand.setValue(0.0);
		}
		
		double change2 = o2 - n2;
		if(change2 > 0){
			myNumReorderPointIncrement.increment();
			mySizeReorderPointIncrement.setValue(change2);
		}else if (change2 < 0){
			myNumReorderPointDecrement.increment();
			mySizeReorderPointDecrement.setValue(Math.abs(change2));
		}
		
		double change3 = o3 - n3;
		if(change3 > 0){
			myNumOrderQuantityIncrement.increment();
			mySizeOrderQuantityIncrement.setValue(change3);
		}else if (change3 < 0){
			myNumOrderQuantityDecrement.increment();
			mySizeOrderQuantityDecrement.setValue(Math.abs(change3));
		}
		

		myOutput = myOutput+","+myReorderPoint+","+myOrderUptoLevel+","+this.getDemandArrivalListener().getAmountOnHand()
					+","+this.getDemandArrivalListener().getAmountOnOrder()
					+","+this.getDemandArrivalListener().getAmountPending()
					+","+this.getDemandArrivalListener().getAmountBackLogged()
					+","+this.getDemandArrivalListener().getInventoryPosition();
		//System.out.println(myOutput);
		try {
	        BufferedWriter out = new BufferedWriter(new FileWriter("FinalReport" + ".csv",true));
	        out.newLine();
	        out.write(myOutput);
	        out.close();
	    } 
		catch (IOException e1){	
	    }
		
	}
	
	protected void replicationEnded() {				
		double var_inc_big_S = mySizeOrderUptoLevelIncrement.getWithinReplicationStatistic().getVariance();
		mySizeOrderUptoLevelIncrementVariance.setValue(var_inc_big_S);			
		double var_dec_big_S = mySizeOrderUptoLevelDecrement.getWithinReplicationStatistic().getVariance();
		mySizeOrderUptoLevelDecrementVariance.setValue(var_dec_big_S);
		
		double last_surplus_on_hand = mySurplusOnHand.getWithinReplicationStatistic().getLastValue();
		myEndSurplusOnHand.setValue(last_surplus_on_hand);
		if(last_surplus_on_hand>0){
			myEndProbabilityOfSurplusOnHand.setValue(1.0);
		}else{
			myEndProbabilityOfSurplusOnHand.setValue(0.0);
		}
		
		double var_inc_small_S = mySizeReorderPointIncrement.getWithinReplicationStatistic().getVariance();
		mySizeReorderPointIncrementVariance.setValue(var_inc_small_S);			
		double var_dec_small_S = mySizeReorderPointDecrement.getWithinReplicationStatistic().getVariance();
		mySizeReorderPointDecrementVariance.setValue(var_dec_small_S);
		
		double var_inc_Q = mySizeOrderQuantityIncrement.getWithinReplicationStatistic().getVariance();
		mySizeOrderQuantityIncrementVariance.setValue(var_inc_Q);			
		double var_dec_Q = mySizeOrderQuantityDecrement.getWithinReplicationStatistic().getVariance();
		mySizeOrderQuantityDecrementVariance.setValue(var_dec_Q);
	}
	
	
}
