/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.afosr.examples;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.InventoryRepairingHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;
import jslx.supplychain.transportlayer.TimeBasedDemandCarrier;

/** 
 * This class models a METRIC model
 * 
 */

public class MetricIRHPNetwork extends ModelElement {

	/** Models the lead time to produce the item type at the external supplier
	 */
	private LeadTimeDemandFiller myExternalSupplier;
	
	/** Models the transport time (time to deliver) the item type from the external
	 *  supplier to the connected (top level) IHPs
	 */
	private TimeBasedDemandCarrier myExternalSupplierShipper;

	/** Collects statistics on the aggregate amount backordered
	 */
	private AggregateTimeWeightedVariable myTotalAmtBackOrdered;
	
	/** Collects statistics on the aggregate number waiting in backorder
	 */
	private AggregateTimeWeightedVariable myTotalNumBackOrdered;

	/** Collects statistics on the aggregate inventory
	 */
	private AggregateTimeWeightedVariable myTotalOnHand;

	/** Collects statistics on the aggregate on order
	 */
	private AggregateTimeWeightedVariable myTotalOnOrder;

	/** Collects statistics on the aggregate stockout
	 */
	private AggregateTimeWeightedVariable myTotalStockOutIndicator;

	/** Collects statistics on the aggregate first fill rate
	 */
	private ResponseVariableAverageObserver myAvgFirstFillRate;

	/** Collects statistics on the aggregate customer waiting time
	 */
	private ResponseVariableAverageObserver myAvgCustomerWaitTime;
	
	/** Counts all the replenishments made by the network
	 */
	private AggregateCounter myTotalNumOfReplenishments;
	
	/** Holds the inventory repairing holding points and their children as a list
	 */
	private Map<InventoryRepairingHoldingPoint, List<InventoryRepairingHoldingPoint>> myIHRPs;
		
	/** Holds all the IHRPs that act as customers to other IHRPs in the network
	 */
	private Set<InventoryRepairingHoldingPoint> myCustomerIHRPs;
	
	/** Holds all the IHRPs that are attached to the external supplier
	 */
	private Set<InventoryRepairingHoldingPoint> myIHRPsAttachedToExternalSupplier;
	
	/** Holds the DemandGenerators for an IHRP if they are attached
	 */
	private Map<InventoryRepairingHoldingPoint, List<DemandGenerator>> myDemandGenerators;
	
	/** Indicates whether the network will allow its IHRPs to receive demand
	 *  from any external demand generator.  Any external demand generator
	 *  must be willing to accept a no delay shippment.
	 */
	private boolean myAllowExternalDemandGeneratorsFlag = false;
	
	public MetricIRHPNetwork(ModelElement parent) {
		this(parent, null);
	}
	
	public MetricIRHPNetwork(ModelElement parent, String name) {
		super(parent, name);
				
		myIHRPs = new HashMap<InventoryRepairingHoldingPoint, List<InventoryRepairingHoldingPoint>>();
		myCustomerIHRPs = new HashSet<InventoryRepairingHoldingPoint>();
		myIHRPsAttachedToExternalSupplier = new HashSet<InventoryRepairingHoldingPoint>();
		myDemandGenerators = new HashMap<InventoryRepairingHoldingPoint, List<DemandGenerator>>();
		
		myTotalOnHand = new AggregateTimeWeightedVariable(this, getName() + " Total On Hand");
		myAvgFirstFillRate = new ResponseVariableAverageObserver(this, getName() + " Aggregate Fill Rate");
		myTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, getName() + " Aggregate Number of Stockouts");
		myTotalOnOrder = new AggregateTimeWeightedVariable(this, getName() + " Total On Order");
		myTotalNumOfReplenishments = new AggregateCounter(this, getName() + " NumberOfReplenishments");
		myTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Amt BackOrdered");
		myTotalNumBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Number BackOrdered");
		myAvgCustomerWaitTime = new ResponseVariableAverageObserver(this, getName() + "Avg Customer Wait");
		
		//create external LeadTimeDemandFiller and its own TimeBasedDemandShipper		
		myExternalSupplier = new LeadTimeDemandFiller(this, getName() + "DummyExternalSupplier");
		myExternalSupplierShipper = new TimeBasedDemandCarrier(this, getName() + "ExternalShipper");
		myExternalSupplier.setDemandCarrier(myExternalSupplierShipper);
		
	}
	
	public final ItemType addItemType(String name){
		ItemType type = new ItemType(this, name);
		return(type);
	}

	/** Adds an ItemType and its lead time distribution to the network. The name
	 *  of the item type must be unique to the model
	 * 
	 * @param name
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(String name, RandomIfc leadtime){
		ItemType type = new ItemType(this, name);
		myExternalSupplier.addLeadTime(type, leadtime);
		return(type);
	}
	
	/** Causes an InventoryRepairingHoldingPoint to be created with a uniquely assigned name
	 *  and added to the network.  This only adds the IHRP to the network it does
	 *  not cause the IHRP to be connected to any other IHRPs in the network. The client
	 *  can use the returned reference to the IHRP to connect to other IHRPs.
	 * 
	 * @return the created IHRP
	 */
	public final InventoryRepairingHoldingPoint addInventoryRepairingHoldingPoint(){
		return(addInventoryRepairingHoldingPoint(null));
	}
	
	/** Causes an InventoryRepairingHoldingPoint to be created with the supplied name
	 *  and added to the network.  This only adds the IHRP to the network it does
	 *  not cause the IHRP to be connected to any other IHRPs in the network. The client
	 *  can use the returned reference to the IHRP to connect to other IHRPs.
	 * 
	 * @param name must be unique in the Model
	 * @return the created IHRP
	 */
	public final InventoryRepairingHoldingPoint addInventoryRepairingHoldingPoint(String name){
		
		//create the ihrp and add it to the network so that it may have child ihrps
		InventoryRepairingHoldingPoint ihrp = new InventoryRepairingHoldingPoint(this, name);
		myIHRPs.put(ihrp, new ArrayList<InventoryRepairingHoldingPoint>());
		
		// assign the ihp its own time based shipper for shipping to its customers
		TimeBasedDemandCarrier shipper = new TimeBasedDemandCarrier(this, ihrp.getName() + "TimeBasedShipper");
		ihrp.setDemandCarrier(shipper);
		if (myAllowExternalDemandGeneratorsFlag == true){
			shipper.setImmediateTransportFlag(true);
		}

		ihrp.subscribeToAggregateOnHandInventory(myTotalOnHand);
		ihrp.subscribeToAggregateAmountOnOrder(myTotalOnOrder);
		ihrp.subscribeToAggregateStockOutIndicator(myTotalStockOutIndicator);
		ihrp.subscribeToAggregateAvgFirstFillRate(myAvgFirstFillRate);
		ihrp.subscribeToAggregateReplenishmentDemandCounter(myTotalNumOfReplenishments);		
		ihrp.subscribeToAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		ihrp.subscribeToAggregateNumberBackOrdered(myTotalNumBackOrdered);
		ihrp.subscribeToAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);
				
		return(ihrp);
	}

	public final boolean isAttachedToExternalSuppler(InventoryRepairingHoldingPoint ihrp){
		return(myIHRPsAttachedToExternalSupplier.contains(ihrp));
	}
	
	public final boolean isAttachedAsCustomer(InventoryRepairingHoldingPoint ihrp){
		return(myCustomerIHRPs.contains(ihrp));
	}
	
	public final boolean isCustomer(InventoryRepairingHoldingPoint supplier, InventoryRepairingHoldingPoint customer){
		if (!myIHRPs.containsKey(supplier))
			return(false);
		
		if (!myIHRPs.containsKey(customer))
			return(false);
				
		List<InventoryRepairingHoldingPoint> customers = myIHRPs.get(supplier);
		return(customers.contains(customer) && (customer.getDemandFiller() == supplier));		
	}
	
	/** Attaches the IHP to the external supplier with a no transport delay
	 * 
	 * @param ihp Must not already be attached to the external supplier or to another IHP
	 */
	public final void attachIHRPToExternalSupplier(InventoryRepairingHoldingPoint ihrp){
		attachIHRPToExternalSupplier(ihrp, null);
	}

	
	public final void attachIHRPToExternalSupplier(InventoryRepairingHoldingPoint ihrp, DistributionIfc transportTime){
		if (myIHRPsAttachedToExternalSupplier.contains(ihrp))
			throw new IllegalArgumentException("The supplied IHP is already attached to the external supplier.");
		
		if (myCustomerIHRPs.contains(ihrp))
			throw new IllegalArgumentException("The supplied IHP is already a customer to another IHP");
		
		myIHRPsAttachedToExternalSupplier.add(ihrp);
		if (transportTime != null)
			myExternalSupplierShipper.setTransportTime(ihrp, transportTime);
		else
			myExternalSupplierShipper.setImmediateTransportFlag(true);	

		ihrp.setDemandFiller(myExternalSupplier);
	}
	
	public final void attachIHRPToSupplier(InventoryRepairingHoldingPoint supplier, InventoryRepairingHoldingPoint customer){
		attachIHRPToSupplier(supplier, customer, null);
	}

	public final void attachIHRPToSupplier(InventoryRepairingHoldingPoint supplier, InventoryRepairingHoldingPoint customer, 
			DistributionIfc transportTime){
		if (!myIHRPs.containsKey(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		if (!myIHRPs.containsKey(customer))
			throw new IllegalArgumentException("The provided customer has not be added to the network");

		if (myCustomerIHRPs.contains(customer))
			throw new IllegalArgumentException("The supplied customer IHP is already a customer IHP");
		
		//both in the network and customer is not already a customer, can add it to the supplier's list
		List<InventoryRepairingHoldingPoint> customers = myIHRPs.get(supplier);
		customers.add(customer);
		// designate it as a customer
		myCustomerIHRPs.add(customer);
		// ensure that it uses the supplier for its replensishments
		customer.setDemandFiller(supplier);
		
		if (transportTime != null){
			// ensure that it uses its supplier's shipper
			TimeBasedDemandCarrier shipper = (TimeBasedDemandCarrier)supplier.getDemandCarrier();
			shipper.setTransportTime(customer, transportTime);
		} else {
			// directing an attached IHP to be shipped with no delay time
			// designate the time based shipper to allow no delay shipment
			TimeBasedDemandCarrier shipper = (TimeBasedDemandCarrier)supplier.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}
	}

	public final DemandGenerator addDemandGeneratorToIHRP(ItemType type, DistributionIfc timeUntilNext){
		return(addDemandGeneratorToIHRP(type, timeUntilNext, null));
	}
	
	public final DemandGenerator addDemandGeneratorToIHRP(ItemType type, DistributionIfc timeUntilNext, String name){
		DemandGenerator dg = new DemandGenerator(this, type, timeUntilNext, name);
		return(dg);
	}
	
	public final void attachDemandGeneratorToIHRP(InventoryRepairingHoldingPoint supplier, DemandGenerator generator){
		attachDemandGeneratorToIHRP(supplier, generator, null);
	}

	public final void attachDemandGeneratorToIHRP(InventoryRepairingHoldingPoint supplier, DemandGenerator generator,
			DistributionIfc transportTime){
		if (!myIHRPs.containsKey(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		List<DemandGenerator> generators = myDemandGenerators.get(supplier);
		if (generators == null){
			generators = new ArrayList<DemandGenerator>();
			myDemandGenerators.put(supplier, generators);
		}
		if (generators.contains(generator))
			throw new IllegalArgumentException("The provided generator is already attached to the supplier");
		
		generators.add(generator);
		generator.setDemandFiller(supplier);
		if (transportTime != null){
			// ensure that it uses its supplier's shipper
			TimeBasedDemandCarrier shipper = (TimeBasedDemandCarrier)supplier.getDemandCarrier();
			shipper.setTransportTime(generator, transportTime);
		} else {
			// directing an attached IHP to be shipped with no delay time
			// designate the time based shipper to allow no delay shipment
			TimeBasedDemandCarrier shipper = (TimeBasedDemandCarrier)supplier.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}		
	}
	
	public final void allowExternalDemandGenerators(){
		myAllowExternalDemandGeneratorsFlag = true;
		for (InventoryRepairingHoldingPoint ihp: myIHRPs.keySet()){
			TimeBasedDemandCarrier shipper = (TimeBasedDemandCarrier)ihp.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}
	}
	
	public final List<InventoryRepairingHoldingPoint> getInventoryRepairingHoldingPoints(){
		return(new ArrayList<InventoryRepairingHoldingPoint>(myIHRPs.keySet()));
	}
	
}
