/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.afosr.examples;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.EventGenerator;
import jsl.modeling.elements.EventGeneratorListenerIfc;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.supplychain.inventorylayer.BackLogQueue;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandFillerFinderIfc;
import jslx.supplychain.inventorylayer.DemandFillerIfc;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.DemandSenderIfc;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryPolicyAbstract;
import jslx.supplychain.inventorylayer.InventoryPolicyReorderPointOrderUpToLevel;
import jslx.supplychain.inventorylayer.InventoryPolicyReorderPointReorderQuantity;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;

/**
 * This class models a Poisson demand arrival on an inventory that follows a (r, Q)
 * and is replenished by a lead time demand filler. After every periods the re order point
 * is incremented by 5.0.
 * 
 */
public class RQInventoryModelParameterUpdatable extends ModelElement implements DemandFillerFinderIfc {

	private ItemType myItemType;
	private DemandGenerator myDemandGenerator;
	private Inventory myInventory;
    private InventoryPolicyAbstract myInventoryPolicy;
    private LeadTimeDemandFiller myLeadTimeDemandFiller;
    private EventGenerator myPolicyUpdateGenerator;
    private PolicyUpdateEventListener myPolicyUpdateEventListener = new PolicyUpdateEventListener();
    private int myReorderPt; 
    private int myReorderQty;
    
	/**
	 * @param parent
	 */
	public RQInventoryModelParameterUpdatable(ModelElement parent, DistributionIfc timebtwDemand, 
						    DistributionIfc leadTime, int reorderPt, int reorderQty, int initialLevel) {
		this(parent, timebtwDemand, leadTime, reorderPt, reorderQty, initialLevel, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public RQInventoryModelParameterUpdatable(ModelElement parent, DistributionIfc timebtwDemand, DistributionIfc leadTime, 
	                        int reorderPt, int reorderQty, int initialLevel, String name) {
		super(parent, name);

		// define the type
		myItemType = new ItemType(this, "Type-A");
		
		// define the filler
		myLeadTimeDemandFiller = new LeadTimeDemandFiller(this);
		myLeadTimeDemandFiller.addLeadTime(myItemType,leadTime);
		
		myReorderPt = reorderPt;
		myReorderQty = reorderQty;
		
		myInventoryPolicy = new InventoryPolicyReorderPointOrderUpToLevel(this, myReorderPt, myReorderQty);		
		myInventory = new Inventory(this, myItemType, myInventoryPolicy, initialLevel);		
		BackLogQueue bpolicy = new BackLogQueue(myInventory, myInventory.getName()+ "-BackLogQ");
	    
		myInventory.setDemandFillerFinder(this);
		// define the demand generator
		myDemandGenerator = new DemandGenerator(this, myItemType, timebtwDemand, timebtwDemand);
		myDemandGenerator.setDemandFillerFinder(this);
		myPolicyUpdateGenerator = new EventGenerator(this, myPolicyUpdateEventListener, new Constant(2.0), new Constant(2.0));
		
	}

	/* (non-Javadoc)
	 * @see jsl.supplychain.inventorylayer.DemandFillerFinderIfc#findDemandFiller(jsl.supplychain.inventorylayer.DemandSenderIfc, jsl.supplychain.inventorylayer.ItemType, int)
	 */
	public DemandFillerIfc findDemandFiller(Demand demand) {
		DemandSenderIfc sender = demand.getSender();
		if (sender == myDemandGenerator)
			return(myInventory);
		else
			return(myLeadTimeDemandFiller);

	}
	

	private class PolicyUpdateEventListener implements EventGeneratorListenerIfc{
		public void generate(EventGenerator generator, JSLEvent event) {
			myReorderPt = myReorderPt + 5;
			//TODO would you rather use the inventory class as an access
			int q = myReorderQty;
			int r = myReorderPt;
			double [] par = {r,q};
			RQInventoryModelParameterUpdatable.this.myInventory.setInventoryPolicyParameters(par);			
		}		
	}

}
