/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.afosr.examples;

import java.io.IOException;

import jsl.modeling.Experiment;
import jsl.modeling.Model;
import jsl.observers.textfile.CSVExperimentReport;
import jsl.observers.textfile.CSVReport;
import jsl.utilities.random.distributions.Bernoulli;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.InventoryRepairingHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;


public class TestMetricIRHPNetworkSherbrookeExample {

	/**
	 * @param args
	 */
	
	public static void main(String[] args) {
		System.out.println("SherbrookeExampleOfMetric-T1");        
		testExperiment();        
		System.out.println("Done!");	
	}

	public static Model buildModel(){
		Model m = Model.createModel();
		
		//Create the network
		MetricIRHPNetwork n = new MetricIRHPNetwork(m, "SherbrookeExampleOfMetric-T1");
		
		//Define item types
		ItemType type1 = n.addItemType("Type-1");
		
		//Define Warehouse
		InventoryRepairingHoldingPoint depot = n.addInventoryRepairingHoldingPoint("Depot");
		
		//addInventory to warehouse
		depot.addReorderPointReorderQuantityInventory(type1, 1, 1, 2,"Depot Type-1");
		depot.setRepairStation(type1,new Bernoulli(1.0),new Constant(0.0253));
		
		// Create the retailers which use the warehouse as its filler
		InventoryRepairingHoldingPoint base1 = n.addInventoryRepairingHoldingPoint("Base1");
		InventoryRepairingHoldingPoint base2 = n.addInventoryRepairingHoldingPoint("Base2");
		InventoryRepairingHoldingPoint base3 = n.addInventoryRepairingHoldingPoint("Base3");
		InventoryRepairingHoldingPoint base4 = n.addInventoryRepairingHoldingPoint("Base4");
		InventoryRepairingHoldingPoint base5 = n.addInventoryRepairingHoldingPoint("Base5");
		
		//Create the inventory at each retailer
		base1.addReorderPointReorderQuantityInventory(type1,-1,1,0,"Base1 Type-1");		
		base2.addReorderPointReorderQuantityInventory(type1,-1,1,0,"Base2 Type-1");		
		base3.addReorderPointReorderQuantityInventory(type1,-1,1,0,"Base3 Type-1");		
		base4.addReorderPointReorderQuantityInventory(type1,-1,1,0,"Base4 Type-1");
		base5.addReorderPointReorderQuantityInventory(type1,-1,1,0,"Base5 Type-1");
		
		base1.setRepairStation(type1,new Bernoulli(0.2), new Constant(1));
		base2.setRepairStation(type1,new Bernoulli(0.2), new Constant(1));
		base3.setRepairStation(type1,new Bernoulli(0.2), new Constant(1));
		base4.setRepairStation(type1,new Bernoulli(0.2), new Constant(1));
		base5.setRepairStation(type1,new Bernoulli(0.2), new Constant(1));
		
		// create the structure of the network
		// attach the warehouse to the external supplier
		n.attachIHRPToExternalSupplier(depot);
		
		// attach the retailers to the warehouse, reuse the constant leadtime of 1.0
		DistributionIfc d = new Constant(0.01);
		n.attachIHRPToSupplier(depot, base1, d);
		n.attachIHRPToSupplier(depot, base2, d);
		n.attachIHRPToSupplier(depot, base3, d);
		n.attachIHRPToSupplier(depot, base4, d);
		n.attachIHRPToSupplier(depot, base5, d);
		
		// create the demand generators for the retailers						
		DistributionIfc dA = new Exponential(1.0/23.2);		
		DemandGenerator dg1 = n.addDemandGeneratorToIHRP(type1, dA);		
		DemandGenerator dg2 = n.addDemandGeneratorToIHRP(type1, dA);
		DemandGenerator dg3 = n.addDemandGeneratorToIHRP(type1, dA);
		DemandGenerator dg4 = n.addDemandGeneratorToIHRP(type1, dA);
		DemandGenerator dg5 = n.addDemandGeneratorToIHRP(type1, dA);
		
		n.attachDemandGeneratorToIHRP(base1,dg1);
		n.attachDemandGeneratorToIHRP(base2,dg2);
		n.attachDemandGeneratorToIHRP(base3,dg3);
		n.attachDemandGeneratorToIHRP(base4,dg4);
		n.attachDemandGeneratorToIHRP(base5,dg5);
		
		return(m);
	}
	
	public static void testExperiment(){
		
		Model m = buildModel();	
		
		// create the experiment to run the model
		Experiment e = new Experiment(m,"SherbrookeExampleOfMetric-T1");
		
		try{
        	CSVReport ecsv = new CSVExperimentReport("SherbrookeExampleOfMetric-T1");
        	m.addObserver(ecsv);
        }
        catch(IOException exception){
			System.out.println("I/O exception opening file Output.txt");
			exception.printStackTrace();
        }
        
		// set the parameters of the experiment
		e.setNumberOfReplications(10);
		e.setLengthOfReplication(3650.0);
		e.setLengthOfWarmUp(100.0);
		
		e.turnOnExperimentReport();
		e.turnOnReplicationReport();
		e.turnOnExperimentConsoleOutput(true);
		
		// tell the experiment to run
		e.runAll();
	}
}
