/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.afosr.examples;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.random.RandomIfc;

/** 
 * This class models a network model which is optimized by
 * simple heuristics independent of other locations or items 
 *  
 *
 */
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.forecasting.ForecastTechnique;
import jslx.supplychain.afosr.forecasting.DemandAndForecastArrivalListenerForNetwork;
import jslx.supplychain.afosr.forecasting.ForecastManager;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;
import jslx.supplychain.transportlayer.TimeBasedTypeLocationIndependentDemandCarrier;

/**
 * @author vvarghe
 *
 */
public class UpdatableSSInventoryNetwork extends SchedulingElement {

	/*
	 * It decided how often the demand is aggregated and forecast is made.
	 * It also decides how often the policies are to be updated
	 * 
	 */
	private double myUpdatePeriod;
	
	/*
	 * The supplier to the network warehouse
	 * 
	 */
	private LeadTimeDemandFiller myExternalSupplier;
	
	/*
	 * The carrier for the external supplier. The transport time
	 * is independent of the warehouses it supplies and the items
	 * 
	 */
	private TimeBasedTypeLocationIndependentDemandCarrier myExternalSupplierCarrier;
	
	/*
	 * The warehouse
	 * 
	 */
	private InventoryHoldingPoint myDepot;
	
	/*
	 * The carrier for the warehouse. The transport time
	 * is independent of the warehouses it supplies and the items
	 * 
	 */
	private TimeBasedTypeLocationIndependentDemandCarrier myDepotCarrier;
	
	/*
	 * A set of the bases the network holds
	 * 
	 */
	private Set<InventoryHoldingPoint> myBases;
	
	/*
	 * It holds the IHPs and the demand generators that each serves
	 * 
	 */
	private Map<InventoryHoldingPoint, List<DemandGenerator>> myDemandGenerators;
	
	/*
	 * 
	 */
	private boolean myAllowExternalDemandGeneratorsFlag = false;
	
	/*
	 * maps the inventory and the variables
	 * 
	 * These varibales collects statistics at each inventory level
	 * Note that the inventory already has variables for the 
	 * usual operational performance measures.
	 * Here the declared variables are that for the volatility performance measure
	 * not available currently for the inventory
	 * 
	 */
	private Map <Inventory,Counter> mapInventoryToNumOfTimesOrderUptoLevelIncremented;
	private Map <Inventory,Counter> mapInventoryToNumOfTimesOrderUptoLevelDecremented;
	private Map <Inventory,ResponseVariable> mapInventoryToAverageOfOrderUptoLevelIncrements;
	private Map <Inventory,ResponseVariable> mapInventoryToVarianceOfOrderUptoLevelIncrements;
	private Map <Inventory,ResponseVariable> mapInventoryToAverageOfOrderUptoLevelDecrements;
	private Map <Inventory,ResponseVariable> mapInventoryToVarianceOfOrderUptoLevelDecrements;
	private Map <Inventory,Counter> mapInventoryToNumOfTimesReorderPointIncremented;
	private Map <Inventory,Counter> mapInventoryToNumOfTimesReorderPointDecremented;
	private Map <Inventory,ResponseVariable> mapInventoryToAverageOfReorderPointIncrements;
	private Map <Inventory,ResponseVariable> mapInventoryToVarianceOfReorderPointIncrements;
	private Map <Inventory,ResponseVariable> mapInventoryToAverageOfReorderPointDecrements;
	private Map <Inventory,ResponseVariable> mapInventoryToVarianceOfReorderPointDecrements;
	//TODO	
	private Map <Inventory, TimeWeighted> mapInventoryToSurplusOnHand;
	private Map <Inventory, TimeWeighted> mapInventoryToTimeSpentInSurplusOnHand;
	private Map <Inventory, ResponseVariable> mapInventoryToEndSurplusOnHand;
	private Map <Inventory, ResponseVariable> mapInventoryToEndProbabilityOfSurplusOnHand;
	
	/*
	 * Maps the location and location based aggregators
	 * 
	 * these aggregators collects statistics aggregated at the IHP level
	 * Note that the IHPs already has aggregate variables for the 
	 * usual operational performance measures.
	 * Here the declared variables are that for the volatility performance measure
	 * not available currently for the IHPs
	 * 
	 */
	private Map <InventoryHoldingPoint, AggregateCounter> mapIHPToNumOfTimesOrderUptoLevelIncrementedAtIHP;
	private Map <InventoryHoldingPoint, AggregateCounter> mapIHPToNumOfTimesOrderUptoLevelDecrementedAtIHP;
	private Map <InventoryHoldingPoint, ResponseVariableAverageObserver> mapIHPToAverageOfOrderUptoLevelIncrementsAtIHP;
	private Map <InventoryHoldingPoint, ResponseVariableAverageObserver> mapIHPToVarianceOfOrderUptoLevelIncrementsAtIHP;
	private Map <InventoryHoldingPoint, ResponseVariableAverageObserver> mapIHPToAverageOfOrderUptoLevelDecrementsAtIHP;
	private Map <InventoryHoldingPoint, ResponseVariableAverageObserver> mapIHPToVarianceOfOrderUptoLevelDecrementsAtIHP;
	private Map <InventoryHoldingPoint, AggregateCounter> mapIHPToNumOfTimesReorderPointIncrementedAtIHP;
	private Map <InventoryHoldingPoint, AggregateCounter> mapIHPToNumOfTimesReorderPointDecrementedAtIHP;
	private Map <InventoryHoldingPoint, ResponseVariableAverageObserver> mapIHPToAverageOfReorderPointIncrementsAtIHP;
	private Map <InventoryHoldingPoint, ResponseVariableAverageObserver> mapIHPToVarianceOfReorderPointIncrementsAtIHP;
	private Map <InventoryHoldingPoint, ResponseVariableAverageObserver> mapIHPToAverageOfReorderPointDecrementsAtIHP;
	private Map <InventoryHoldingPoint, ResponseVariableAverageObserver> mapIHPToVarianceOfReorderPointDecrementsAtIHP;
	//TODO
	private Map <InventoryHoldingPoint, AggregateTimeWeightedVariable> mapIHPToSurplusOnHandAtIHP;
	private Map <InventoryHoldingPoint, AggregateTimeWeightedVariable> mapIHPToTimeSpentInSurplusOnHandAtIHP;
	private Map <InventoryHoldingPoint, ResponseVariableAverageObserver> mapIHPToEndSurplusOnHandAtIHP;
	private Map <InventoryHoldingPoint, ResponseVariableAverageObserver> mapIHPToEndProbabilityOfSurplusOnHandAtIHP;
	
	/*
	 * Collects statistics aggregated at the network level
	 * 
	 */	
	private AggregateTimeWeightedVariable myTotalAmtBackOrdered;
	private AggregateTimeWeightedVariable myTotalNumBackOrdered;
	private AggregateTimeWeightedVariable myTotalOnHand;
	private AggregateTimeWeightedVariable myTotalOnOrder;
	private AggregateTimeWeightedVariable myTotalStockOutIndicator;
	private ResponseVariableAverageObserver myAvgFirstFillRate;
	private ResponseVariableAverageObserver myAvgCustomerWaitTime;
	private AggregateCounter myTotalNumOfReplenishments;
	
	private AggregateCounter myNumOfTimesOrderUptoLevelIncrementedInTheNetwork;
	private AggregateCounter myNumOfTimesOrderUptoLevelDecrementedInTheNetwork;
	private ResponseVariableAverageObserver myAverageOfOrderUptoLevelIncrementsInTheNetwork;
	private ResponseVariableAverageObserver myVarianceOfOrderUptoLevelIncrementsInTheNetwork;
	private ResponseVariableAverageObserver myAverageOfOrderUptoLevelDecrementsInTheNetwork;
	private ResponseVariableAverageObserver myVarianceOfOrderUptoLevelDecrementsInTheNetwork;
	private AggregateCounter myNumOfTimesReorderPointIncrementedInTheNetwork;
	private AggregateCounter myNumOfTimesReorderPointDecrementedInTheNetwork;
	private ResponseVariableAverageObserver myAverageOfReorderPointIncrementsInTheNetwork;
	private ResponseVariableAverageObserver myVarianceOfReorderPointIncrementsInTheNetwork;
	private ResponseVariableAverageObserver myAverageOfReorderPointDecrementsInTheNetwork;
	private ResponseVariableAverageObserver myVarianceOfReorderPointDecrementsInTheNetwork;
	//TODO
	private AggregateTimeWeightedVariable mySurplusInventoryInTheNetwork;
	private AggregateTimeWeightedVariable myTimeSpentInSurplusOnHandInTheNetwork;
	private ResponseVariableAverageObserver myEndSurplusInventoryInTheNetwork;
	private ResponseVariableAverageObserver myEndProbabilityOfSurplusOnHandInTheNetwork;
	
	/*
	 * Holds the inventory and its policy updater
	 * 
	 */
	private Map <Inventory, DefaultSSNetworkInventoryUpdater> myInventoryToUpdaterMap;
	
	/*
	 * Holds the inventory and its listeners
	 * 
	 */
	private Map <Inventory, DemandAndForecastArrivalListenerForNetwork> myInventoryToListenerForNetworkMap;	
	
	private static int myItemNumber = 0;
	
	public UpdatableSSInventoryNetwork(ModelElement parent) {
		this(parent, 1, null);
	}
	public UpdatableSSInventoryNetwork(ModelElement parent, double period, String name) {
		super(parent, name);		
		myUpdatePeriod = period;

		//create external LeadTimeDemandFiller and its own TimeBasedItemTypeLocationIndependentDemandShipper
		myExternalSupplier = new LeadTimeDemandFiller(this, getName() + "ExternalSupplier");
		myExternalSupplierCarrier = new TimeBasedTypeLocationIndependentDemandCarrier(this, getName() + "ExternalShipper");
		myExternalSupplier.setDemandCarrier(myExternalSupplierCarrier);

		//create Depot and its own TimeBasedItemTypeLocationIndependentDemandShipper
		myDepot = new InventoryHoldingPoint(this, getName() + "Depot");
		myDepotCarrier = new TimeBasedTypeLocationIndependentDemandCarrier(this, getName() + "DepotShipper");
		myDepot.setDemandCarrier(myDepotCarrier);
		
		myBases = new HashSet<InventoryHoldingPoint>();
		myDemandGenerators = new HashMap<InventoryHoldingPoint, List<DemandGenerator>>();
		
		myInventoryToUpdaterMap = new HashMap <Inventory, DefaultSSNetworkInventoryUpdater>();		
		myInventoryToListenerForNetworkMap = new HashMap <Inventory, DemandAndForecastArrivalListenerForNetwork>();
		
		myExternalSupplierCarrier.setImmediateTransportFlag(true);
		myDepot.setDemandFiller(myExternalSupplier);	
		
		myTotalOnHand = new AggregateTimeWeightedVariable(this, getName() + " Total On Hand");
		myAvgFirstFillRate = new ResponseVariableAverageObserver(this, getName() + " Aggregate Fill Rate");
		myTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, getName() + " Aggregate Number of Stockouts");
		myTotalOnOrder = new AggregateTimeWeightedVariable(this, getName() + " Total On Order");
		myTotalNumOfReplenishments = new AggregateCounter(this, getName() + " NumberOfReplenishments");		
		myTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Amt BackOrdered");
		myTotalNumBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Number BackOrdered");
		myAvgCustomerWaitTime = new ResponseVariableAverageObserver(this, getName() + "Avg Customer Wait");
		
		myDepot.subscribeToAggregateOnHandInventory(myTotalOnHand);
		myDepot.subscribeToAggregateAmountOnOrder(myTotalOnOrder);
		myDepot.subscribeToAggregateStockOutIndicator(myTotalStockOutIndicator);
		myDepot.subscribeToAggregateAvgFirstFillRate(myAvgFirstFillRate);		
		myDepot.subscribeToAggregateReplenishmentDemandCounter(myTotalNumOfReplenishments);		
		myDepot.subscribeToAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		myDepot.subscribeToAggregateNumberBackOrdered(myTotalNumBackOrdered);
		myDepot.subscribeToAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);	
		
		/*
		 * Creates the map that holds the inventory and the variables
		 * 
		 */
		mapInventoryToNumOfTimesOrderUptoLevelIncremented = new HashMap <Inventory,Counter>();
		mapInventoryToNumOfTimesOrderUptoLevelDecremented = new HashMap <Inventory,Counter>();
		mapInventoryToAverageOfOrderUptoLevelIncrements = new HashMap <Inventory,ResponseVariable>();
		mapInventoryToVarianceOfOrderUptoLevelIncrements = new HashMap <Inventory,ResponseVariable>();
		mapInventoryToAverageOfOrderUptoLevelDecrements = new HashMap <Inventory,ResponseVariable>();
		mapInventoryToVarianceOfOrderUptoLevelDecrements = new HashMap <Inventory,ResponseVariable>();
		mapInventoryToNumOfTimesReorderPointIncremented = new HashMap <Inventory,Counter>();
		mapInventoryToNumOfTimesReorderPointDecremented = new HashMap <Inventory,Counter>();
		mapInventoryToAverageOfReorderPointIncrements = new HashMap <Inventory,ResponseVariable>();
		mapInventoryToVarianceOfReorderPointIncrements = new HashMap <Inventory,ResponseVariable>();
		mapInventoryToAverageOfReorderPointDecrements = new HashMap <Inventory,ResponseVariable>();
		mapInventoryToVarianceOfReorderPointDecrements = new HashMap <Inventory,ResponseVariable>();
		//TODO
		mapInventoryToSurplusOnHand = new HashMap <Inventory,TimeWeighted>();
		mapInventoryToTimeSpentInSurplusOnHand = new HashMap <Inventory,TimeWeighted>();
		mapInventoryToEndSurplusOnHand = new HashMap <Inventory,ResponseVariable>();
		mapInventoryToEndProbabilityOfSurplusOnHand = new HashMap <Inventory,ResponseVariable>();
				
		/*
		 * Creates the aggregate statistics for the depot and
		 * map the location to the location based aggregators 
		 * 
		 */
		mapIHPToNumOfTimesOrderUptoLevelIncrementedAtIHP =  new HashMap <InventoryHoldingPoint, AggregateCounter>();
		mapIHPToNumOfTimesOrderUptoLevelDecrementedAtIHP =  new HashMap <InventoryHoldingPoint, AggregateCounter>();
		mapIHPToAverageOfOrderUptoLevelIncrementsAtIHP =  new HashMap <InventoryHoldingPoint, ResponseVariableAverageObserver>();
		mapIHPToVarianceOfOrderUptoLevelIncrementsAtIHP =  new HashMap <InventoryHoldingPoint, ResponseVariableAverageObserver>();
		mapIHPToAverageOfOrderUptoLevelDecrementsAtIHP =  new HashMap <InventoryHoldingPoint, ResponseVariableAverageObserver>();
		mapIHPToVarianceOfOrderUptoLevelDecrementsAtIHP =  new HashMap <InventoryHoldingPoint, ResponseVariableAverageObserver>();
		mapIHPToNumOfTimesReorderPointIncrementedAtIHP =  new HashMap <InventoryHoldingPoint, AggregateCounter>();
		mapIHPToNumOfTimesReorderPointDecrementedAtIHP =  new HashMap <InventoryHoldingPoint, AggregateCounter>();
		mapIHPToAverageOfReorderPointIncrementsAtIHP =  new HashMap <InventoryHoldingPoint, ResponseVariableAverageObserver>();
		mapIHPToVarianceOfReorderPointIncrementsAtIHP =  new HashMap <InventoryHoldingPoint, ResponseVariableAverageObserver>();
		mapIHPToAverageOfReorderPointDecrementsAtIHP =  new HashMap <InventoryHoldingPoint, ResponseVariableAverageObserver>();
		mapIHPToVarianceOfReorderPointDecrementsAtIHP =  new HashMap <InventoryHoldingPoint, ResponseVariableAverageObserver>();
		//TODO
		mapIHPToSurplusOnHandAtIHP = new HashMap <InventoryHoldingPoint, AggregateTimeWeightedVariable>();
		mapIHPToTimeSpentInSurplusOnHandAtIHP = new HashMap <InventoryHoldingPoint, AggregateTimeWeightedVariable>();
		mapIHPToEndSurplusOnHandAtIHP = new HashMap <InventoryHoldingPoint, ResponseVariableAverageObserver>();
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP = new HashMap <InventoryHoldingPoint, ResponseVariableAverageObserver>();
		
		AggregateCounter c1 = new AggregateCounter(this, getName() + "-Depot-NumOfTimesOrderUptoLevelIncrementedAtIHP");
		AggregateCounter c2 = new AggregateCounter(this, getName() + "-Depot-NumOfTimesOrderUptoLevelDecrementedAtIHP");
		ResponseVariableAverageObserver r1 = new ResponseVariableAverageObserver(this, getName() + "-Depot-AverageOfOrderUptoLevelIncrementsAtIHP");
		ResponseVariableAverageObserver r2 = new ResponseVariableAverageObserver(this, getName() + "-Depot-VarianceOfOrderUptoLevelIncrementsAtIHP");
		ResponseVariableAverageObserver r3 = new ResponseVariableAverageObserver(this, getName() + "-Depot-AverageOfOrderUptoLevelDecrementsAtIHP");
		ResponseVariableAverageObserver r4 = new ResponseVariableAverageObserver(this, getName() + "-Depot-VarianceOfOrderUptoLevelDecrementsAtIHP");
		AggregateCounter d1 = new AggregateCounter(this, getName() + "-Depot-NumOfTimesReorderPointIncrementedAtIHP");
		AggregateCounter d2 = new AggregateCounter(this, getName() + "-Depot-NumOfTimesReorderPointDecrementedAtIHP");
		ResponseVariableAverageObserver s1 = new ResponseVariableAverageObserver(this, getName() + "-Depot-AverageOfReorderPointIncrementsAtIHP");
		ResponseVariableAverageObserver s2 = new ResponseVariableAverageObserver(this, getName() + "-Depot-VarianceOfReorderPointIncrementsAtIHP");
		ResponseVariableAverageObserver s3 = new ResponseVariableAverageObserver(this, getName() + "-Depot-AverageOfReorderPointDecrementsAtIHP");
		ResponseVariableAverageObserver s4 = new ResponseVariableAverageObserver(this, getName() + "-Depot-VarianceOfReorderPointDecrementsAtIHP");
		//TODO
		AggregateTimeWeightedVariable t1 = new AggregateTimeWeightedVariable(this, getName() + "-Depot-SurplusOnHandAtIHP");
		AggregateTimeWeightedVariable t2 = new AggregateTimeWeightedVariable(this, getName() + "-Depot-TimeSpentInSurplusOnHandAtIHP");
		ResponseVariableAverageObserver t3 = new ResponseVariableAverageObserver(this, getName() + "-Depot-EndSurplusOnHandAtIHP");
		ResponseVariableAverageObserver t4 = new ResponseVariableAverageObserver(this, getName() + "-Depot-EndProbabilityOfSurplusOnhHandAtIHP");
				
		mapIHPToNumOfTimesOrderUptoLevelIncrementedAtIHP.put(myDepot, c1);
		mapIHPToNumOfTimesOrderUptoLevelDecrementedAtIHP.put(myDepot, c2);
		mapIHPToAverageOfOrderUptoLevelIncrementsAtIHP.put(myDepot, r1);
		mapIHPToVarianceOfOrderUptoLevelIncrementsAtIHP.put(myDepot, r2);
		mapIHPToAverageOfOrderUptoLevelDecrementsAtIHP.put(myDepot, r3);
		mapIHPToVarianceOfOrderUptoLevelDecrementsAtIHP.put(myDepot, r4);
		mapIHPToNumOfTimesReorderPointIncrementedAtIHP.put(myDepot, d1);
		mapIHPToNumOfTimesReorderPointDecrementedAtIHP.put(myDepot, d2);
		mapIHPToAverageOfReorderPointIncrementsAtIHP.put(myDepot, s1);
		mapIHPToVarianceOfReorderPointIncrementsAtIHP.put(myDepot, s2);
		mapIHPToAverageOfReorderPointDecrementsAtIHP.put(myDepot, s3);
		mapIHPToVarianceOfReorderPointDecrementsAtIHP.put(myDepot, s4);
		//TODO
		mapIHPToSurplusOnHandAtIHP.put(myDepot, t1);
		mapIHPToTimeSpentInSurplusOnHandAtIHP.put(myDepot, t2);
		mapIHPToEndSurplusOnHandAtIHP.put(myDepot, t3);
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP.put(myDepot, t4);
		
		/*
		 * Creates the aggregate statistics for whole network
		 * 
		*/		
		myNumOfTimesOrderUptoLevelIncrementedInTheNetwork = new AggregateCounter(this, getName() + "-Network-NumOfTimesOrderUptoLevelIncremented");
		myNumOfTimesOrderUptoLevelDecrementedInTheNetwork = new AggregateCounter(this, getName() + "-Network-NumOfTimesOrderUptoLevelDecremented");
		myAverageOfOrderUptoLevelIncrementsInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-AverageOfOrderUptoLevelIncrements");
		myVarianceOfOrderUptoLevelIncrementsInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-VarianceOfOrderUptoLevelIncrements");
		myAverageOfOrderUptoLevelDecrementsInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-AverageOfOrderUptoLevelDecrements");
		myVarianceOfOrderUptoLevelDecrementsInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-VarianceOfOrderUptoLevelDecrements");
		myNumOfTimesReorderPointIncrementedInTheNetwork = new AggregateCounter(this, getName() + "-Network-NumOfTimesReorderPointIncremented");
		myNumOfTimesReorderPointDecrementedInTheNetwork = new AggregateCounter(this, getName() + "-Network-NumOfTimesReorderPointDecremented");
		myAverageOfReorderPointIncrementsInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-AverageOfReorderPointIncrements");
		myVarianceOfReorderPointIncrementsInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-VarianceOfReorderPointIncrements");
		myAverageOfReorderPointDecrementsInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-AverageOfReorderPointDecrements");
		myVarianceOfReorderPointDecrementsInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-VarianceOfReorderPointDecrements");
		//TODO
		mySurplusInventoryInTheNetwork = new AggregateTimeWeightedVariable(this, getName() + "-Network-SurplusOnHand");
		myTimeSpentInSurplusOnHandInTheNetwork = new AggregateTimeWeightedVariable(this, getName() + "-Network-TimeSpentInSurplusOnHand");
		myEndSurplusInventoryInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-EndSurplusOnHand");
		myEndProbabilityOfSurplusOnHandInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-EndProbabilityOfSurplusOnHand");
		
	}	
	
	public final ItemType addItemType(String name, RandomIfc leadtime){
		ItemType type = new ItemType(this, name);
		myExternalSupplier.addLeadTime(type, leadtime);
		return(type);
	}
	
	public final Inventory addDepotInventory(ItemType type, int reorderPt, int reorderLevel, int initialLevel,
			DistributionIfc transportTime, ForecastTechnique forecaster, double initialforecastestimate, 
			boolean updatable){
		
		Inventory inv = myDepot.addReorderPointOrderUpToLevelInventory(type, reorderPt, reorderLevel, initialLevel, "Inv-Depot-"+type);
		setShippingDelayFromExternalSupplierToDepot(type, transportTime);
		
		
		
		myItemNumber++;		
		DemandAndForecastArrivalListenerForNetwork myListener 
			= new DemandAndForecastArrivalListenerForNetwork(this, inv, type+"-Depot-Listener", myUpdatePeriod, initialforecastestimate, updatable);
		myListener.setForecastManager(new ForecastManager(this, forecaster, type+"Depot-Forecaster"));
		myInventoryToListenerForNetworkMap.put(inv,myListener);		
		DefaultSSNetworkInventoryUpdater updater = new DefaultSSNetworkInventoryUpdater(inv, myListener, 
				initialforecastestimate, transportTime, myUpdatePeriod);		
		myInventoryToUpdaterMap.put(inv,updater);
				
		/*
		 * The volatility variable for each inventory at the depot are attached to the 
		 * corresponding aggregator for the depot 
		 * 
		 */
		Counter c1 = new Counter(this, getName()+"-"+type+"-Depot-NumOfTimesOrderUptoLevelIncrementedAtIHP");
		Counter c2 = new Counter(this, getName()+"-"+type+"-Depot-NumOfTimesOrderUptoLevelDecrementedAtIHP");
		ResponseVariable r1 = new ResponseVariable(this, getName()+"-"+type+"-Depot-AverageOfOrderUptoLevelIncrementsAtIHP");		
		ResponseVariable r2 = new ResponseVariable(this, getName()+"-"+type+"-Depot-VarianceOfOrderUptoLevelIncrementsAtIHP");
		ResponseVariable r3 = new ResponseVariable(this, getName()+"-"+type+"-Depot-AverageOfOrderUptoLevelDecrementsAtIHP");
		ResponseVariable r4 = new ResponseVariable(this, getName()+"-"+type+"-Depot-VarianceOfOrderUptoLevelDecrementsAtIHP");
		Counter d1 = new Counter(this, getName()+"-"+type+"-Depot-NumOfTimesReorderPointIncrementedAtIHP");
		Counter d2 = new Counter(this, getName()+"-"+type+"-Depot-NumOfTimesReorderPointDecrementedAtIHP");
		ResponseVariable s1 = new ResponseVariable(this, getName()+"-"+type+"-Depot-AverageOfReorderPointIncrementsAtIHP");		
		ResponseVariable s2 = new ResponseVariable(this, getName()+"-"+type+"-Depot-VarianceOfReorderPointIncrementsAtIHP");
		ResponseVariable s3 = new ResponseVariable(this, getName()+"-"+type+"-Depot-AverageOfReorderPointDecrementsAtIHP");
		ResponseVariable s4 = new ResponseVariable(this, getName()+"-"+type+"-Depot-VarianceOfReorderPointDecrementsAtIHP");
		
		mapInventoryToNumOfTimesOrderUptoLevelIncremented.put(inv, c1);		
		mapInventoryToNumOfTimesOrderUptoLevelDecremented.put(inv, c2);		
		mapInventoryToAverageOfOrderUptoLevelIncrements.put(inv, r1);		
		mapInventoryToVarianceOfOrderUptoLevelIncrements.put(inv, r2);		
		mapInventoryToAverageOfOrderUptoLevelDecrements.put(inv, r3);		
		mapInventoryToVarianceOfOrderUptoLevelDecrements.put(inv, r4);
		mapInventoryToNumOfTimesReorderPointIncremented.put(inv, d1);		
		mapInventoryToNumOfTimesReorderPointDecremented.put(inv, d2);		
		mapInventoryToAverageOfReorderPointIncrements.put(inv, s1);		
		mapInventoryToVarianceOfReorderPointIncrements.put(inv, s2);		
		mapInventoryToAverageOfReorderPointDecrements.put(inv, s3);		
		mapInventoryToVarianceOfReorderPointDecrements.put(inv, s4);
		
		mapIHPToNumOfTimesOrderUptoLevelIncrementedAtIHP.get(myDepot).subscribeTo(c1);
		mapIHPToNumOfTimesOrderUptoLevelDecrementedAtIHP.get(myDepot).subscribeTo(c2);
		mapIHPToAverageOfOrderUptoLevelIncrementsAtIHP.get(myDepot).subscribeTo(r1);
		mapIHPToVarianceOfOrderUptoLevelIncrementsAtIHP.get(myDepot).subscribeTo(r2);
		mapIHPToAverageOfOrderUptoLevelDecrementsAtIHP.get(myDepot).subscribeTo(r3);
		mapIHPToVarianceOfOrderUptoLevelDecrementsAtIHP.get(myDepot).subscribeTo(r4);
		mapIHPToNumOfTimesReorderPointIncrementedAtIHP.get(myDepot).subscribeTo(d1);
		mapIHPToNumOfTimesReorderPointDecrementedAtIHP.get(myDepot).subscribeTo(d2);
		mapIHPToAverageOfReorderPointIncrementsAtIHP.get(myDepot).subscribeTo(s1);
		mapIHPToVarianceOfReorderPointIncrementsAtIHP.get(myDepot).subscribeTo(s2);
		mapIHPToAverageOfReorderPointDecrementsAtIHP.get(myDepot).subscribeTo(s3);
		mapIHPToVarianceOfReorderPointDecrementsAtIHP.get(myDepot).subscribeTo(s4);	
		
		
		myNumOfTimesOrderUptoLevelIncrementedInTheNetwork.subscribeTo(c1);
		myNumOfTimesOrderUptoLevelDecrementedInTheNetwork.subscribeTo(c2);
		myAverageOfOrderUptoLevelIncrementsInTheNetwork.subscribeTo(r1);
		myVarianceOfOrderUptoLevelIncrementsInTheNetwork.subscribeTo(r2);
		myAverageOfOrderUptoLevelDecrementsInTheNetwork.subscribeTo(r3);
		myVarianceOfOrderUptoLevelDecrementsInTheNetwork.subscribeTo(r4);		
		myNumOfTimesReorderPointIncrementedInTheNetwork.subscribeTo(d1);
		myNumOfTimesReorderPointDecrementedInTheNetwork.subscribeTo(d2);
		myAverageOfReorderPointIncrementsInTheNetwork.subscribeTo(s1);
		myVarianceOfReorderPointIncrementsInTheNetwork.subscribeTo(s2);
		myAverageOfReorderPointDecrementsInTheNetwork.subscribeTo(s3);
		myVarianceOfReorderPointDecrementsInTheNetwork.subscribeTo(s4);		
		
		// TODO
		TimeWeighted t1 = new TimeWeighted(this, getName()+"-"+type+"-Depot-SurplusOnHandAtIHP");
		TimeWeighted t2 = new TimeWeighted(this, getName()+"-"+type+"-Depot-TimeSpentInSurplusOnHandAtIHP");
		ResponseVariable t3 = new ResponseVariable(this, getName()+"-"+type+"-Depot-EndSurplusOnHandAtIHP");
		ResponseVariable t4 = new ResponseVariable(this, getName()+"-"+type+"-Depot-EndProbabilityOfSurplusOnHandAtIHP");
		
		mapInventoryToSurplusOnHand.put(inv, t1);
		mapInventoryToTimeSpentInSurplusOnHand.put(inv, t2);
		mapInventoryToEndSurplusOnHand.put(inv, t3);
		mapInventoryToEndProbabilityOfSurplusOnHand.put(inv, t4);
		
		mapIHPToSurplusOnHandAtIHP.get(myDepot).subscribeTo(t1);
		mapIHPToTimeSpentInSurplusOnHandAtIHP.get(myDepot).subscribeTo(t2);
		mapIHPToEndSurplusOnHandAtIHP.get(myDepot).subscribeTo(t3);
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP.get(myDepot).subscribeTo(t4);
		
		/*
		 * The volatility variable for each inventory at the depot are attached to the 
		 * corresponding aggregator for the network 
		 * 
		 */		
		//TODO
		mySurplusInventoryInTheNetwork.subscribeTo(t1);
		myTimeSpentInSurplusOnHandInTheNetwork.subscribeTo(t2);
		myEndSurplusInventoryInTheNetwork.subscribeTo(t3);
		myEndProbabilityOfSurplusOnHandInTheNetwork.subscribeTo(t4);
		
		return inv;
	}
	
	private final void setShippingDelayFromExternalSupplierToDepot(ItemType type, DistributionIfc transportTime){				
		// TODO
		Demand d = new Demand(type);
		// only if the inventory is added to the depot
		if(!myDepot.canFillItemType(d))
			throw new IllegalArgumentException("The supplied depot doesnt have the itemtype inventory");
		
		if (transportTime != null){
			myExternalSupplierCarrier.setTransportTime(myDepot, type, transportTime);
		}else{
			myExternalSupplierCarrier.setImmediateTransportFlag(true);
		}
	} 
	
	public final boolean isBase(InventoryHoldingPoint irhp){
		return(myBases.contains(irhp));
	}

	/*
	 * Add the base and connect it to the dept and assigned the carrier of the depot
	 * the commom opm aggregator are retrieved and attached to the aggregator of the network
	 * 
	 * 
	 */
	public final InventoryHoldingPoint addBase(String name){
		// create the ihp and add it to the network
		InventoryHoldingPoint ihp = new InventoryHoldingPoint(this, name);
		myBases.add(ihp);
		// assign the ihp its own time based shipper for shipping to its customers
		TimeBasedTypeLocationIndependentDemandCarrier shipper = new TimeBasedTypeLocationIndependentDemandCarrier(this, ihp.getName() + "TimeBasedShipper");
		ihp.setDemandCarrier(shipper);		
		if (myAllowExternalDemandGeneratorsFlag == true){
			shipper.setImmediateTransportFlag(true);
		}
		myDepotCarrier.setImmediateTransportFlag(true);
		
		ihp.setDemandFiller(myDepot);
		
		ihp.subscribeToAggregateOnHandInventory(myTotalOnHand);
		ihp.subscribeToAggregateAmountOnOrder(myTotalOnOrder);
		ihp.subscribeToAggregateStockOutIndicator(myTotalStockOutIndicator);
		ihp.subscribeToAggregateAvgFirstFillRate(myAvgFirstFillRate);
		ihp.subscribeToAggregateReplenishmentDemandCounter(myTotalNumOfReplenishments);		
		ihp.subscribeToAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		ihp.subscribeToAggregateNumberBackOrdered(myTotalNumBackOrdered);
		ihp.subscribeToAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);
		
		AggregateCounter c1 = new AggregateCounter(this, getName() + "-"+name+ "-NumOfTimesOrderUptoLevelIncrementedAtIHP");
		AggregateCounter c2 = new AggregateCounter(this, getName() + "-"+name+ "-AggregateNumOfTimesOrderUptoLevelDecrementedAtIHP");
		ResponseVariableAverageObserver r1 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-AverageOfOrderUptoLevelIncrementsAtIHP");
		ResponseVariableAverageObserver r2 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-VarianceOfOrderUptoLevelIncrementsAtIHP");
		ResponseVariableAverageObserver r3 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-AverageOfOrderUptoLevelDecrementsAtIHP");
		ResponseVariableAverageObserver r4 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-VarianceOfOrderUptoLevelDecrementsAtIHP");		
		AggregateCounter d1 = new AggregateCounter(this, getName() + "-"+name+ "-NumOfTimesReorderPointIncrementedAtIHP");
		AggregateCounter d2 = new AggregateCounter(this, getName() + "-"+name+ "-NumOfTimesReorderPointDecrementedAtIHP");
		ResponseVariableAverageObserver s1 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-AverageOfReorderPointIncrementsAtIHP");
		ResponseVariableAverageObserver s2 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-VarianceOfReorderPointIncrementsAtIHP");
		ResponseVariableAverageObserver s3 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-AverageOfReorderPointDecrementsAtIHP");
		ResponseVariableAverageObserver s4 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-VarianceOfReorderPointDecrementsAtIHP");		
		
		mapIHPToNumOfTimesOrderUptoLevelIncrementedAtIHP.put(ihp, c1);
		mapIHPToNumOfTimesOrderUptoLevelDecrementedAtIHP.put(ihp, c2);
		mapIHPToAverageOfOrderUptoLevelIncrementsAtIHP.put(ihp, r1);
		mapIHPToVarianceOfOrderUptoLevelIncrementsAtIHP.put(ihp, r2);
		mapIHPToAverageOfOrderUptoLevelDecrementsAtIHP.put(ihp, r3);
		mapIHPToVarianceOfOrderUptoLevelDecrementsAtIHP.put(ihp, r4);		
		mapIHPToNumOfTimesReorderPointIncrementedAtIHP.put(ihp, d1);
		mapIHPToNumOfTimesReorderPointDecrementedAtIHP.put(ihp, d2);
		mapIHPToAverageOfReorderPointIncrementsAtIHP.put(ihp, s1);
		mapIHPToVarianceOfReorderPointIncrementsAtIHP.put(ihp, s2);
		mapIHPToAverageOfReorderPointDecrementsAtIHP.put(ihp, s3);
		mapIHPToVarianceOfReorderPointDecrementsAtIHP.put(ihp, s4);
		
		// TODO
		AggregateTimeWeightedVariable t1 = new AggregateTimeWeightedVariable(this, getName() + "-"+name+ "-SurplusInventoryAtIHP");
		AggregateTimeWeightedVariable t2 = new AggregateTimeWeightedVariable(this, getName() + "-"+name+ "-TimeSpentInSurplusInventoryAtIHP");
		ResponseVariableAverageObserver t3 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-EndSurplusInventoryAtIHP");
		ResponseVariableAverageObserver t4 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-EndProbabilityOfSurplusInventoryAtIHP");
		
		mapIHPToSurplusOnHandAtIHP.put(ihp, t1);
		mapIHPToTimeSpentInSurplusOnHandAtIHP.put(ihp, t2);
		mapIHPToEndSurplusOnHandAtIHP.put(ihp, t3);
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP.put(ihp, t4);
		
		return(ihp);
	}

	public final Inventory addBaseInventory(InventoryHoldingPoint base, ItemType type, int reorderPt, int reorderLevel, int initialLevel,
			DistributionIfc transportTime, ForecastTechnique forecaster, double initialforecastestimate, 
			boolean updatable, RandomIfc demandDist){		
		
		Inventory inv = base.addReorderPointOrderUpToLevelInventory(type, reorderPt, reorderLevel, initialLevel, "Inv-"+base.getName()+"-"+type);
		setShippingDelayFromDepotToBase(base, type, transportTime);
		attachDemandGeneratorToBase(base, type, demandDist, new Constant(0.0));
		
		myItemNumber++;
		
		DemandAndForecastArrivalListenerForNetwork myListener 
			= new DemandAndForecastArrivalListenerForNetwork(this, inv, type+"-"+base.getName()+"-Listener", myUpdatePeriod, 
					initialforecastestimate, updatable);
		myListener.setForecastManager(new ForecastManager(this, forecaster, 
				type+"-"+base.getName()+"-Forecaster"));
		myInventoryToListenerForNetworkMap.put(inv,myListener);
		DefaultSSNetworkInventoryUpdater updater = new DefaultSSNetworkInventoryUpdater(inv, myListener, initialforecastestimate, 
				transportTime, myUpdatePeriod);		
		myInventoryToUpdaterMap.put(inv,updater);
		
		Counter c1 = new Counter(this, getName()+"-"+type+"-"+base.getName()+"-NumOfTimesOrderUptoLevelIncremented");
		Counter c2 = new Counter(this, getName()+"-"+type+"-"+base.getName()+"-NumOfTimesOrderUptoLevelDecremented");
		ResponseVariable r1 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-AverageOfOrderUptoLevelIncrements");		
		ResponseVariable r2 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-VarianceOfOrderUptoLevelIncrements");
		ResponseVariable r3 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-AverageOfOrderUptoLevelDecrements");
		ResponseVariable r4 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-VarianceOfOrderUptoLevelDecrements");
		Counter d1 = new Counter(this, getName()+"-"+type+"-"+base.getName()+"-NumOfTimesReorderPointIncremented");
		Counter d2 = new Counter(this, getName()+"-"+type+"-"+base.getName()+"-NumOfTimesReorderPointDecremented");
		ResponseVariable s1 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-AverageOfReorderPointIncrements");		
		ResponseVariable s2 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-VarianceOfReorderPointIncrements");
		ResponseVariable s3 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-AverageOfReorderPointDecrements");
		ResponseVariable s4 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-VarianceOfReorderPointDecrements");
				
		mapInventoryToNumOfTimesOrderUptoLevelIncremented.put(inv, c1);		
		mapInventoryToNumOfTimesOrderUptoLevelDecremented.put(inv, c2);		
		mapInventoryToAverageOfOrderUptoLevelIncrements.put(inv, r1);		
		mapInventoryToVarianceOfOrderUptoLevelIncrements.put(inv, r2);		
		mapInventoryToAverageOfOrderUptoLevelDecrements.put(inv, r3);		
		mapInventoryToVarianceOfOrderUptoLevelDecrements.put(inv, r4);
		mapInventoryToNumOfTimesReorderPointIncremented.put(inv, d1);		
		mapInventoryToNumOfTimesReorderPointDecremented.put(inv, d2);		
		mapInventoryToAverageOfReorderPointIncrements.put(inv, s1);		
		mapInventoryToVarianceOfReorderPointIncrements.put(inv, s2);		
		mapInventoryToAverageOfReorderPointDecrements.put(inv, s3);		
		mapInventoryToVarianceOfReorderPointDecrements.put(inv, s4);
		
		mapIHPToNumOfTimesOrderUptoLevelIncrementedAtIHP.get(base).subscribeTo(c1);
		mapIHPToNumOfTimesOrderUptoLevelDecrementedAtIHP.get(base).subscribeTo(c2);
		mapIHPToAverageOfOrderUptoLevelIncrementsAtIHP.get(base).subscribeTo(r1);
		mapIHPToVarianceOfOrderUptoLevelIncrementsAtIHP.get(base).subscribeTo(r2);
		mapIHPToAverageOfOrderUptoLevelDecrementsAtIHP.get(base).subscribeTo(r3);
		mapIHPToVarianceOfOrderUptoLevelDecrementsAtIHP.get(base).subscribeTo(r4);
		mapIHPToNumOfTimesReorderPointIncrementedAtIHP.get(base).subscribeTo(d1);
		mapIHPToNumOfTimesReorderPointDecrementedAtIHP.get(base).subscribeTo(d2);
		mapIHPToAverageOfReorderPointIncrementsAtIHP.get(base).subscribeTo(s1);
		mapIHPToVarianceOfReorderPointIncrementsAtIHP.get(base).subscribeTo(s2);
		mapIHPToAverageOfReorderPointDecrementsAtIHP.get(base).subscribeTo(s3);
		mapIHPToVarianceOfReorderPointDecrementsAtIHP.get(base).subscribeTo(s4);	
				
		myNumOfTimesOrderUptoLevelIncrementedInTheNetwork.subscribeTo(c1);
		myNumOfTimesOrderUptoLevelDecrementedInTheNetwork.subscribeTo(c2);
		myAverageOfOrderUptoLevelIncrementsInTheNetwork.subscribeTo(r1);
		myVarianceOfOrderUptoLevelIncrementsInTheNetwork.subscribeTo(r2);
		myAverageOfOrderUptoLevelDecrementsInTheNetwork.subscribeTo(r3);
		myVarianceOfOrderUptoLevelDecrementsInTheNetwork.subscribeTo(r4);		
		myNumOfTimesReorderPointIncrementedInTheNetwork.subscribeTo(d1);
		myNumOfTimesReorderPointDecrementedInTheNetwork.subscribeTo(d2);
		myAverageOfReorderPointIncrementsInTheNetwork.subscribeTo(s1);
		myVarianceOfReorderPointIncrementsInTheNetwork.subscribeTo(s2);
		myAverageOfReorderPointDecrementsInTheNetwork.subscribeTo(s3);
		myVarianceOfReorderPointDecrementsInTheNetwork.subscribeTo(s4);	

		
		//TODO
		TimeWeighted t1 = new TimeWeighted(this, getName()+"-"+type+"-"+base.getName()+"-SurplusInventory");
		TimeWeighted t2 = new TimeWeighted(this, getName()+"-"+type+"-"+base.getName()+"-TimeSpentInSurplusInventory");
		ResponseVariable t3 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-EndSurplusInventory");
		ResponseVariable t4 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-EndProbabilityOfSurplusInventory");
		
		mapInventoryToSurplusOnHand.put(inv, t1);
		mapInventoryToTimeSpentInSurplusOnHand.put(inv, t2);
		mapInventoryToEndSurplusOnHand.put(inv, t3);
		mapInventoryToEndProbabilityOfSurplusOnHand.put(inv, t4);		

		mapIHPToSurplusOnHandAtIHP.get(base).subscribeTo(t1);
		mapIHPToTimeSpentInSurplusOnHandAtIHP.get(base).subscribeTo(t2);
		mapIHPToEndSurplusOnHandAtIHP.get(base).subscribeTo(t3);
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP.get(base).subscribeTo(t4);
		
		//TODO
		mySurplusInventoryInTheNetwork.subscribeTo(t1);
		myTimeSpentInSurplusOnHandInTheNetwork.subscribeTo(t2);
		myEndSurplusInventoryInTheNetwork.subscribeTo(t3);
		myEndProbabilityOfSurplusOnHandInTheNetwork.subscribeTo(t4);
		
				
		
		return inv;
	}
	
	
	private final void setShippingDelayFromDepotToBase(InventoryHoldingPoint base, ItemType type, DistributionIfc transportTime){
		if (!myBases.contains(base))
			throw new IllegalArgumentException("The provided base has not be added to the network");
		
		if (transportTime != null){
			// ensure that it uses its supplier's shipper
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)myDepot.getDemandCarrier();
			shipper.setTransportTime(base, type, transportTime);
		} else {
			// directing an attached IHP to be shipped with no delay time
			// designate the time based shipper to allow no delay shipment
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)myDepot.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}
		
	}

	public final DemandGenerator attachDemandGeneratorToBase(InventoryHoldingPoint base,
			ItemType type, RandomIfc timeUntilNext, DistributionIfc shiptimeIHPtoDG){
		return(attachDemandGeneratorToBase(base, type, timeUntilNext, shiptimeIHPtoDG, null));
	}
	public final DemandGenerator attachDemandGeneratorToBase(InventoryHoldingPoint base,
			ItemType type, RandomIfc timeUntilNext, DistributionIfc shiptimeIHPtoDG, String name){
		DemandGenerator dg = new DemandGenerator(this, type, timeUntilNext, name);
		attachDemandGeneratorToBase(base, dg, shiptimeIHPtoDG);
		return(dg);
	}
	public final void attachDemandGeneratorToBase(InventoryHoldingPoint base, DemandGenerator generator){
		attachDemandGeneratorToBase(base, generator, null);
	}
	public final void attachDemandGeneratorToBase(InventoryHoldingPoint base, DemandGenerator generator,
			DistributionIfc transportTime){
		if (!myBases.contains(base))
			throw new IllegalArgumentException("The provided base has not be added to the network");
		
		List<DemandGenerator> generators = myDemandGenerators.get(base);
		if (generators == null){
			generators = new ArrayList<DemandGenerator>();
			myDemandGenerators.put(base, generators);
		}
		if (generators.contains(generator))
			throw new IllegalArgumentException("The provided generator is already attached to the supplier");
		
		generators.add(generator);
		generator.setDemandFiller(base);
		if (transportTime != null){
			// ensure that it uses its supplier's shipper
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)base.getDemandCarrier();
			shipper.setTransportTime(generator, generator.getItemType(), transportTime);
		} else {
			// directing an attached IHP to be shipped with no delay time
			// designate the time based shipper to allow no delay shipment
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)base.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}		
	}
	public final void allowExternalDemandGenerators(){
		myAllowExternalDemandGeneratorsFlag = true;
		for(InventoryHoldingPoint irhp: myBases){
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)irhp.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}
	}
	
	/** 
	 * 
	 * @return the bases
	 */
	public final Set<InventoryHoldingPoint> getBases(){
		return(myBases);
	}
	
	public void initialize(){
		//System.out.println("updating"+this);
		//System.out.println("Time :"+ getTime());
		for(Inventory inv: myInventoryToUpdaterMap.keySet()){
			/*double oldOrderUptoLevel = inv.getInventoryPolicyParameter()[1];
			double oldReorderPoint = inv.getInventoryPolicyParameter()[0];*/
			
			//System.out.println("Before updating of "+inv.getName()+": "+ inv.getInventoryPolicyParameter()[0]+", "+inv.getInventoryPolicyParameter()[1]);
			double [] par = this.myInventoryToUpdaterMap.get(inv).initparam();
			inv.setInitialOnHand((int)par[1]);
			//System.out.println("After updating of "+inv.getName()+": "+ par[0]+", "+par[1]);
			inv.setInventoryPolicyParameters(par);			
			
			/*double newOrderUptoLevel = inv.getInventoryPolicyParameter()[1];
			double changeOrderUptoLevel = oldOrderUptoLevel - newOrderUptoLevel;	
			if(changeOrderUptoLevel > 0){
				myNumOfTimesOrderUptoLevelIncremented.get(inv).increment();
				myAverageOfOrderUptoLevelIncrements.get(inv).setValue(changeOrderUptoLevel);
			}else if (changeOrderUptoLevel < 0){
				myNumOfTimesOrderUptoLevelDecremented.get(inv).increment();
				myAverageOfOrderUptoLevelDecrements.get(inv).setValue(Math.abs(changeOrderUptoLevel));
			}			
			
			double newReorderPoint = inv.getInventoryPolicyParameter()[0];			
			double changeReorderPoint = oldReorderPoint - newReorderPoint;
			if(changeReorderPoint > 0){
				myNumOfTimesReorderPointIncremented.get(inv).increment();
				myAverageOfReorderPointIncrements.get(inv).setValue(changeReorderPoint);
			}else if (changeReorderPoint < 0){
				myNumOfTimesReorderPointDecremented.get(inv).increment();
				myAverageOfReorderPointDecrements.get(inv).setValue(Math.abs(changeReorderPoint));
			}*/
		}
		scheduleEvent(myUpdatePeriod, 2, 2, "PolicyUpdatingInterval");
	}
	
	protected void handleEvent(JSLEvent event){
		schedulePolicyUpdating(event);
	}
	
	protected void schedulePolicyUpdating(JSLEvent event){
		//System.out.println("updating"+this);
		//System.out.println("Time :"+ getTime());
		for(Inventory inv: myInventoryToUpdaterMap.keySet()){
			double oldOrderUptoLevel = inv.getInventoryPolicyParameters()[1];
			double oldReorderPoint = inv.getInventoryPolicyParameters()[0];
			
			if(myInventoryToListenerForNetworkMap.get(inv).getForecastManager().
					FORECASTINITIALIZED==false){				
			}else{	
				
				//System.out.println("Before updating of "+inv.getName()+": "+ inv.getInventoryPolicyParameter()[0]+", "+inv.getInventoryPolicyParameter()[1]);
				double [] par = this.myInventoryToUpdaterMap.get(inv).optimize();
				//System.out.println("After updating of "+inv.getName()+": "+ par[0]+", "+par[1]);
				inv.setInventoryPolicyParameters(par);
								
				double newOrderUptoLevel = inv.getInventoryPolicyParameters()[1];
				double newReorderPoint = inv.getInventoryPolicyParameters()[0];
				
				int ip = inv.getInventoryPosition();				
				//System.out.println(inv.getName()+" (s0, S0) ("+oldReorderPoint+" ,"+oldOrderUptoLevel+") ip "+ip+" (s1, S1) ("+newReorderPoint+", "+newOrderUptoLevel+")");				
				double changeOrderUptoLevel = oldOrderUptoLevel - newOrderUptoLevel;	
				if(changeOrderUptoLevel > 0){
					mapInventoryToNumOfTimesOrderUptoLevelIncremented.get(inv).increment();
					mapInventoryToAverageOfOrderUptoLevelIncrements.get(inv).setValue(changeOrderUptoLevel);					
					if(ip>newOrderUptoLevel){						
						this.mapInventoryToSurplusOnHand.get(inv).setValue(ip-newOrderUptoLevel);
						this.mapInventoryToTimeSpentInSurplusOnHand.get(inv).setValue(1.0);
					}else{
						this.mapInventoryToSurplusOnHand.get(inv).setValue(0.0);
						this.mapInventoryToTimeSpentInSurplusOnHand.get(inv).setValue(0.0);
					}
				}else if (changeOrderUptoLevel < 0){
					mapInventoryToNumOfTimesOrderUptoLevelDecremented.get(inv).increment();
					mapInventoryToAverageOfOrderUptoLevelDecrements.get(inv).setValue(Math.abs(changeOrderUptoLevel));
					this.mapInventoryToSurplusOnHand.get(inv).setValue(0.0);
					this.mapInventoryToTimeSpentInSurplusOnHand.get(inv).setValue(0.0);					
				}				
							
				double changeReorderPoint = oldReorderPoint - newReorderPoint;
				if(changeReorderPoint > 0){
					mapInventoryToNumOfTimesReorderPointIncremented.get(inv).increment();
					mapInventoryToAverageOfReorderPointIncrements.get(inv).setValue(changeReorderPoint);
				}else if (changeReorderPoint < 0){
					mapInventoryToNumOfTimesReorderPointDecremented.get(inv).increment();
					mapInventoryToAverageOfReorderPointDecrements.get(inv).setValue(Math.abs(changeReorderPoint));
				}
			}
		}
		scheduleEvent(myUpdatePeriod, 2, 2, "PolicyUpdatingInterval");		
	}
	

	protected void replicationEnded() {				
		for(Inventory inv: myInventoryToUpdaterMap.keySet()){			
			double var_inc_orderuptolevel = mapInventoryToAverageOfOrderUptoLevelIncrements.get(inv).getWithinReplicationStatistic().getVariance();
			//System.out.println("Var of Increment Depot "+var_inc);
			mapInventoryToVarianceOfOrderUptoLevelIncrements.get(inv).setValue(var_inc_orderuptolevel);			
			double var_dec_orderuptolevel = mapInventoryToAverageOfOrderUptoLevelDecrements.get(inv).getWithinReplicationStatistic().getVariance();
			//System.out.println("Var of Decrement Depot "+var_inc);			
			mapInventoryToVarianceOfOrderUptoLevelDecrements.get(inv).setValue(var_dec_orderuptolevel);
			
			double last_surplus_on_hand = mapInventoryToSurplusOnHand.get(inv).getWithinReplicationStatistic().getLastValue();
			mapInventoryToEndSurplusOnHand.get(inv).setValue(last_surplus_on_hand);
			if(last_surplus_on_hand>0){
				mapInventoryToEndProbabilityOfSurplusOnHand.get(inv).setValue(1.0);
			}else{
				mapInventoryToEndProbabilityOfSurplusOnHand.get(inv).setValue(0.0);
			}			
			
			//System.out.println("$$"+mapInventoryToSurplusOnHand.get(inv).getWithinReplicationStatistic().getLastValue());
			
			double var_inc_reorderpoint = mapInventoryToAverageOfReorderPointIncrements.get(inv).getWithinReplicationStatistic().getVariance();
			//System.out.println("Var of Increment Depot "+var_inc);
			mapInventoryToVarianceOfReorderPointIncrements.get(inv).setValue(var_inc_reorderpoint);			
			double var_dec_reorderpoint = mapInventoryToAverageOfReorderPointDecrements.get(inv).getWithinReplicationStatistic().getVariance();
			//System.out.println("Var of Decrement Depot "+var_inc);			
			mapInventoryToVarianceOfReorderPointDecrements.get(inv).setValue(var_dec_reorderpoint);	
		}		
	}
	
}
