/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.afosr.examples;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Bernoulli;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.inventory.varimetric.VMBaseItem;
import jslx.inventory.varimetric.VMFMarginalAnalysisAlgorithm;
import jslx.inventory.varimetric.VMItem;
import jslx.inventory.varimetric.VariMetricModel;
import jslx.supplychain.afosr.forecasting.DemandArrivalListenerWithForecaster;
import jslx.supplychain.afosr.forecasting.ForecastArrivalListenerDefault;
import jslx.supplychain.afosr.forecasting.ForecastManager;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryRepairingHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;
import jslx.supplychain.transportlayer.TimeBasedTypeLocationIndependentDemandCarrier;

/** 
 * This class models a Varimetric model which is optimized by
 * Marginal Analysis approach
 *  
 *
 */
public class VarimetricNetworkMarginalAnalysis extends SchedulingElement {

	/** 
	 *  Assigned how often the varimetric model must be optimized
	 */
	private double myUpdatePeriod;
	
	/** 
	 *  The reference to the underlying varimetric model
	 *  
	 */
	private VariMetricModel myVariMetricModel;	
	
	/** 
	 *  The counterpart of the inventory object at the depot of the simulation model
	 *  is the VMItem
	 *  
	 */
	private Map <Inventory, VMItem> myInventoryToVMItemMap;	
	
	/** 
	 *  The counterpart of the inventory object at the bases of the simulation model
	 *  is the VMBaseItem
	 *  
	 */
	private Map <Inventory, VMBaseItem> myInventoryToVMBaseItemMap;	
	
	/** 
	 *  Maps the itemtype within the network to the inventory object at the depot 
	 *  
	 */
	private Map <ItemType, Inventory> myItemTypeToDepotInventoryMap;	
	
	/** 
	 *  Each inventory is listened by a listener that can aggregate and
	 *  can notify a demand forecaster attached to it. This network class creates the
	 *  inventory object as well as demand arrival listener and attach the 
	 *  listener to the inventory. However listener is invisible to another class
	 *  Hence the network maps both the objects so that the network can access
	 *  the listener of a certain inventoy.
	 *  
	 */
	private Map <Inventory, DemandArrivalListenerWithForecaster> myInventoryToBaseDemandArrivalListenerWithForecasterMap;	
	
	/** 
	 *  Each inventory is listened by a listener that can aggregate and
	 *  can notify a demand forecaster attached to it. 
	 *  This forecaster is listened by a forecast listener.
	 *  This network class creates the
	 *  inventory object, demand arrival listener, demand forecaster
	 *  and the forecast arrival listener. However forecast listener 
	 *  cannot be accessed through the inventory object.
	 *  Hence the network maps both the objects so that the network can access
	 *  the listener of a certain inventoy.
	 */
	private Map <Inventory, ForecastArrivalListenerDefault> myInventoryToBaseForecastArrivalListenerDefaultMap;	
	
	/** 
	 *  
	 */	
	private static int myItemNumber = 0;
	
	private double budget;
	private double tolerance;
	
	/** Models the lead time to produce the item type at the external supplier
	 */
	private LeadTimeDemandFiller myExternalSupplier;
	
	/** Models the transport time (time to deliver) the item type from the external
	 *  supplier to the connected (top level) IHPs
	 */
	private TimeBasedTypeLocationIndependentDemandCarrier myExternalSupplierShipper;

	/** Models the lead time to produce the item type at the depot
	 */
	private InventoryRepairingHoldingPoint myDepot;
	
	/** Models the transport time (time to deliver) the item type from the depot
	 *  to the connected (top level) IHPs
	 */
	private TimeBasedTypeLocationIndependentDemandCarrier myDepotShipper;
	
	/** Holds the bases within the network
	 */
	private Set<InventoryRepairingHoldingPoint> myBases;
	
	/** Holds the DemandGenerators for an IHP if they are attached
	 */
	//TODO Vijith changed
	private Map<InventoryRepairingHoldingPoint, List<DemandGenerator>> myDemandGenerators;

	/** Indicates whether the network will allow its IHPs to receive demand
	 *  from any external demand generator.  Any external demand generator
	 *  must be willing to accept a no delay shippment.
	 */
	private boolean myAllowExternalDemandGeneratorsFlag = false;
	
	private Map <Inventory,Counter> myNumBaseStockIncrement;
	private Map <Inventory,Counter> myNumBaseStockDecrement;
	private Map <Inventory,ResponseVariable> myBaseStockIncrement;
	private Map <Inventory,ResponseVariable> myBaseStockIncrementVariance;
	private Map <Inventory,ResponseVariable> myBaseStockDecrement;
	private Map <Inventory,ResponseVariable> myBaseStockDecrementVariance;
	//TODO	
	private Map <Inventory, TimeWeighted> mapInventoryToSurplusOnHand;
	private Map <Inventory, TimeWeighted> mapInventoryToTimeSpentInSurplusOnHand;
	private Map <Inventory, ResponseVariable> mapInventoryToEndSurplusOnHand;
	private Map <Inventory, ResponseVariable> mapInventoryToEndProbabilityOfSurplusOnHand;
	
	private Map <InventoryRepairingHoldingPoint, AggregateCounter> myLocTotalNumBaseStockIncrement;
	private Map <InventoryRepairingHoldingPoint, AggregateCounter> myLocTotalNumBaseStockDecrement;
	private Map <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver> myLocTotalBaseStockIncrement;
	private Map <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver> myLocTotalBaseStockIncrementVariance;
	private Map <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver> myLocTotalBaseStockDecrement;
	private Map <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver> myLocTotalBaseStockDecrementVariance;
	//TODO
	private Map <InventoryRepairingHoldingPoint, AggregateTimeWeightedVariable> mapIHPToSurplusOnHandAtIHP;
	private Map <InventoryRepairingHoldingPoint, AggregateTimeWeightedVariable> mapIHPToTimeSpentInSurplusOnHandAtIHP;
	private Map <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver> mapIHPToEndSurplusOnHandAtIHP;
	private Map <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver> mapIHPToEndProbabilityOfSurplusOnHandAtIHP;
	
	private AggregateTimeWeightedVariable myTotalAmtBackOrdered;
	private AggregateTimeWeightedVariable myTotalNumBackOrdered;
	private AggregateTimeWeightedVariable myTotalOnHand;
	private AggregateTimeWeightedVariable myTotalOnOrder;
	private AggregateTimeWeightedVariable myTotalStockOutIndicator;
	private ResponseVariableAverageObserver myAvgFirstFillRate;
	private ResponseVariableAverageObserver myAvgCustomerWaitTime;
	private AggregateCounter myTotalNumOfReplenishments;
	
	private AggregateCounter myTotalNumBaseStockIncrement;
	private AggregateCounter myTotalNumBaseStockDecrement;
	private ResponseVariableAverageObserver myTotalBaseStockIncrement;
	private ResponseVariableAverageObserver myTotalBaseStockIncrementVariance;
	private ResponseVariableAverageObserver myTotalBaseStockDecrement;
	private ResponseVariableAverageObserver myTotalBaseStockDecrementVariance;
	//TODO
	private AggregateTimeWeightedVariable mySurplusInventoryInTheNetwork;
	private AggregateTimeWeightedVariable myTimeSpentInSurplusOnHandInTheNetwork;
	private ResponseVariableAverageObserver myEndSurplusInventoryInTheNetwork;
	private ResponseVariableAverageObserver myEndProbabilityOfSurplusOnHandInTheNetwork;
	
	
	/** Creates a VarimetricNetworkLGEnumerative
	 * 
	 * @param parent
	 */
	public VarimetricNetworkMarginalAnalysis(ModelElement parent) {
		this(parent, 1, null);
	}
	
	/** Creates a VarimetricNetworkLGEnumerative
	 * 
	 * @param parent
	 * @param period
	 * @param name
	 */
	public VarimetricNetworkMarginalAnalysis(ModelElement parent, double period, String name){		
		this(parent, period, name, 10000.0, 0.1);
	}
	
	/** Creates a VarimetricNetworkLGEnumerative
	 * 
	 * @param parent
	 * @param period
	 * @param name
	 * @param budget_
	 * @param minTheta_
	 * @param maxTheta_
	 * @param numPoints_
	 */
	public VarimetricNetworkMarginalAnalysis(ModelElement parent, double period, String name, double budget_,
			double tolerance_) {
		super(parent, name);
		
		myUpdatePeriod = period;
		budget = budget_;
		tolerance = tolerance_;
		
		myBases = new HashSet<InventoryRepairingHoldingPoint>();
		myDemandGenerators = new HashMap<InventoryRepairingHoldingPoint, List<DemandGenerator>>();
		
		myTotalOnHand = new AggregateTimeWeightedVariable(this, getName() + " Total On Hand");
		myAvgFirstFillRate = new ResponseVariableAverageObserver(this, getName() + " Aggregate Fill Rate");
		myTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, getName() + " Aggregate Number of Stockouts");
		myTotalOnOrder = new AggregateTimeWeightedVariable(this, getName() + " Total On Order");
		myTotalNumOfReplenishments = new AggregateCounter(this, getName() + " NumberOfReplenishments");
		
		//create external LeadTimeDemandFiller and its own TimeBasedItemTypeLocationIndependentDemandShipper
		myExternalSupplier = new LeadTimeDemandFiller(this, getName() + "ExternalSupplier");
		myExternalSupplierShipper = new TimeBasedTypeLocationIndependentDemandCarrier(this, getName() + "ExternalShipper");
		myExternalSupplier.setDemandCarrier(myExternalSupplierShipper);
		
		//create Depot and its own TimeBasedItemTypeLocationIndependentDemandShipper
		myDepot = new InventoryRepairingHoldingPoint(this, getName() + "Depot");
		myDepotShipper = new TimeBasedTypeLocationIndependentDemandCarrier(this, getName() + "DepotShipper");
		myDepot.setDemandCarrier(myDepotShipper);		
		
		myInventoryToVMItemMap = new HashMap <Inventory, VMItem>();
		myInventoryToVMBaseItemMap = new HashMap <Inventory, VMBaseItem>();
		myItemTypeToDepotInventoryMap = new HashMap <ItemType, Inventory>();
		myInventoryToBaseDemandArrivalListenerWithForecasterMap = new HashMap <Inventory, DemandArrivalListenerWithForecaster>();
		myInventoryToBaseForecastArrivalListenerDefaultMap = new HashMap <Inventory, ForecastArrivalListenerDefault>();
		

		myTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Amt BackOrdered");
		myTotalNumBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Number BackOrdered");
		myAvgCustomerWaitTime = new ResponseVariableAverageObserver(this, getName() + "Avg Customer Wait");
		
		myDepot.subscribeToAggregateOnHandInventory(myTotalOnHand);
		myDepot.subscribeToAggregateAmountOnOrder(myTotalOnOrder);
		myDepot.subscribeToAggregateStockOutIndicator(myTotalStockOutIndicator);
		myDepot.subscribeToAggregateAvgFirstFillRate(myAvgFirstFillRate);
		myDepot.subscribeToAggregateReplenishmentDemandCounter(myTotalNumOfReplenishments);		
		myDepot.subscribeToAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		myDepot.subscribeToAggregateNumberBackOrdered(myTotalNumBackOrdered);
		myDepot.subscribeToAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);			
		
		
		myNumBaseStockIncrement = new HashMap <Inventory,Counter>();
		myNumBaseStockDecrement = new HashMap <Inventory,Counter>();
		myBaseStockIncrement = new HashMap <Inventory,ResponseVariable>();
		myBaseStockIncrementVariance = new HashMap <Inventory,ResponseVariable>();
		myBaseStockDecrement = new HashMap <Inventory,ResponseVariable>();
		myBaseStockDecrementVariance = new HashMap <Inventory,ResponseVariable>();
		//TODO
		mapInventoryToSurplusOnHand = new HashMap <Inventory,TimeWeighted>();
		mapInventoryToTimeSpentInSurplusOnHand = new HashMap <Inventory,TimeWeighted>();
		mapInventoryToEndSurplusOnHand = new HashMap <Inventory,ResponseVariable>();
		mapInventoryToEndProbabilityOfSurplusOnHand = new HashMap <Inventory,ResponseVariable>();
		
		myLocTotalNumBaseStockIncrement =  new HashMap <InventoryRepairingHoldingPoint, AggregateCounter>();
		myLocTotalNumBaseStockDecrement =  new HashMap <InventoryRepairingHoldingPoint, AggregateCounter>();
		myLocTotalBaseStockIncrement =  new HashMap <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver>();
		myLocTotalBaseStockIncrementVariance =  new HashMap <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver>();
		myLocTotalBaseStockDecrement =  new HashMap <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver>();
		myLocTotalBaseStockDecrementVariance =  new HashMap <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver>();
		//TODO
		mapIHPToSurplusOnHandAtIHP = new HashMap <InventoryRepairingHoldingPoint, AggregateTimeWeightedVariable>();
		mapIHPToTimeSpentInSurplusOnHandAtIHP = new HashMap <InventoryRepairingHoldingPoint, AggregateTimeWeightedVariable>();
		mapIHPToEndSurplusOnHandAtIHP = new HashMap <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver>();
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP = new HashMap <InventoryRepairingHoldingPoint, ResponseVariableAverageObserver>();
		
		/*
		 * Creates the aggregate statistics for each location
		 * 
		*/
		AggregateCounter c1 = new AggregateCounter(this, getName() + "-Depot-AggregateNumBaseStockIncrement");
		AggregateCounter c2 = new AggregateCounter(this, getName() + "-Depot-AggregateNumBaseStockDecrement");
		ResponseVariableAverageObserver r1 = new ResponseVariableAverageObserver(this, getName() + "-Depot-AggregateBaseStockIncrement");
		ResponseVariableAverageObserver r2 = new ResponseVariableAverageObserver(this, getName() + "-Depot-AggregateBaseStockIncrementVariance");
		ResponseVariableAverageObserver r3 = new ResponseVariableAverageObserver(this, getName() + "-Depot-AggregateBaseStockDecrement");
		ResponseVariableAverageObserver r4 = new ResponseVariableAverageObserver(this, getName() + "-Depot-AggregateBaseStockDecrementVariance");
		//TODO
		AggregateTimeWeightedVariable t1 = new AggregateTimeWeightedVariable(this, getName() + "-Depot-SurplusOnHandAtIHP");
		AggregateTimeWeightedVariable t2 = new AggregateTimeWeightedVariable(this, getName() + "-Depot-TimeSpentInSurplusOnHandAtIHP");
		ResponseVariableAverageObserver t3 = new ResponseVariableAverageObserver(this, getName() + "-Depot-EndSurplusOnHandAtIHP");
		ResponseVariableAverageObserver t4 = new ResponseVariableAverageObserver(this, getName() + "-Depot-EndProbabilityOfSurplusOnhHandAtIHP");
		
		myLocTotalNumBaseStockIncrement.put(myDepot, c1);
		myLocTotalNumBaseStockDecrement.put(myDepot, c2);
		myLocTotalBaseStockIncrement.put(myDepot, r1);
		myLocTotalBaseStockIncrementVariance.put(myDepot, r2);
		myLocTotalBaseStockDecrement.put(myDepot, r3);
		myLocTotalBaseStockDecrementVariance.put(myDepot, r4);
		//TODO
		mapIHPToSurplusOnHandAtIHP.put(myDepot, t1);
		mapIHPToTimeSpentInSurplusOnHandAtIHP.put(myDepot, t2);
		mapIHPToEndSurplusOnHandAtIHP.put(myDepot, t3);
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP.put(myDepot, t4);
		
		
		/*
		 * Creates the aggregate statistics for whole network
		 * 
		*/		
		myTotalNumBaseStockIncrement = new AggregateCounter(this, getName() + "-Network-AggregateNumBaseStockIncrement");
		myTotalNumBaseStockDecrement = new AggregateCounter(this, getName() + "-Network-AggregateNumBaseStockDecrement");
		myTotalBaseStockIncrement = new ResponseVariableAverageObserver(this, getName() + "-Network-AggregateBaseStockIncrement");
		myTotalBaseStockIncrementVariance = new ResponseVariableAverageObserver(this, getName() + "-Network-AggregateBaseStockIncrementVariance");
		myTotalBaseStockDecrement = new ResponseVariableAverageObserver(this, getName() + "-Network-AggregateBaseStockDecrement");
		myTotalBaseStockDecrementVariance = new ResponseVariableAverageObserver(this, getName() + "-Network-AggregateBaseStockDecrementVariance");
//		TODO
		mySurplusInventoryInTheNetwork = new AggregateTimeWeightedVariable(this, getName() + "-Network-SurplusOnHand");
		myTimeSpentInSurplusOnHandInTheNetwork = new AggregateTimeWeightedVariable(this, getName() + "-Network-TimeSpentInSurplusOnHand");
		myEndSurplusInventoryInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-EndSurplusOnHand");
		myEndProbabilityOfSurplusOnHandInTheNetwork = new ResponseVariableAverageObserver(this, getName() + "-Network-EndProbabilityOfSurplusOnHand");
		
		
		myExternalSupplierShipper.setImmediateTransportFlag(true);
		myDepot.setDemandFiller(myExternalSupplier);
	}	
	
	/** Adds an ItemType with a given name
	 *  to the network
	 * 
	 * @param name
	 * @return
	 */
	public final ItemType addItemType(String name){
		ItemType type = new ItemType(this, name);
		return(type);
	}
	
	/** Adds an ItemType with a default name and given leadtime distribution
	 *  to the network
	 * 
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(RandomIfc leadtime){
		return(addItemType(null, leadtime));
	}
	
	/** Adds an ItemType and its lead time distribution to the network. The name
	 *  of the item type must be unique to the model
	 * 
	 * @param name
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(String name, RandomIfc leadtime){
		ItemType type = new ItemType(this, name);
		myExternalSupplier.addLeadTime(type, leadtime);
		return(type);
	}
	
	/** Adds inventory to depot
	 * Creates corresponding VMItem
	 * 
	 * @param type
	 * @param repairdelay
	 * @param transportTime
	 * @param vmcost
	 * @return Inventory
	 */	
	public final Inventory addDepotInventory(ItemType type, DistributionIfc repairdelay, DistributionIfc transportTime, double vmcost){
		int default_basestock = 1;
		// TODO int default_basestock = 2;
		Inventory inv = myDepot.addReorderPointReorderQuantityInventory(type,default_basestock-1,1,default_basestock-1,"Inv-Depot-"+type);
		myDepot.setRepairStation(type, new Bernoulli(1.0), repairdelay);
		setShippingDelayFromExternalSupplierToDepot(type, transportTime);
		
		Counter c1 = new Counter(this, getName()+"-"+type+"-Depot-NumBaseStockIncrement");
		myNumBaseStockIncrement.put(inv, c1);
		
		Counter c2 = new Counter(this, getName()+"-"+type+"-Depot-NumBaseStockDecrement");
		myNumBaseStockDecrement.put(inv, c2);
		
		ResponseVariable r1 = new ResponseVariable(this, getName()+"-"+type+"-Depot-BaseStockIncrement");		
		myBaseStockIncrement.put(inv, r1);
		
		ResponseVariable r2 = new ResponseVariable(this, getName()+"-"+type+"-Depot-BaseStockIncrementVariance");
		myBaseStockIncrementVariance.put(inv, r2);
		
		ResponseVariable r3 = new ResponseVariable(this, getName()+"-"+type+"-Depot-BaseStockDecrement");
		myBaseStockDecrement.put(inv, r3);
		
		ResponseVariable r4 = new ResponseVariable(this, getName()+"-"+type+"-Depot-BaseStockDecrementVariance");
		myBaseStockDecrementVariance.put(inv, r4);
		
		myLocTotalNumBaseStockIncrement.get(myDepot).subscribeTo(c1);
		myLocTotalNumBaseStockDecrement.get(myDepot).subscribeTo(c2);
		myLocTotalBaseStockIncrement.get(myDepot).subscribeTo(r1);
		myLocTotalBaseStockIncrementVariance.get(myDepot).subscribeTo(r2);
		myLocTotalBaseStockDecrement.get(myDepot).subscribeTo(r3);
		myLocTotalBaseStockDecrementVariance.get(myDepot).subscribeTo(r4);
		
		myTotalNumBaseStockIncrement.subscribeTo(c1);
		myTotalNumBaseStockDecrement.subscribeTo(c2);
		myTotalBaseStockIncrement.subscribeTo(r1);
		myTotalBaseStockIncrementVariance.subscribeTo(r2);
		myTotalBaseStockDecrement.subscribeTo(r3);
		myTotalBaseStockDecrementVariance.subscribeTo(r4);
		
//		 TODO
		TimeWeighted t1 = new TimeWeighted(this, getName()+"-"+type+"-Depot-SurplusOnHandAtIHP");
		TimeWeighted t2 = new TimeWeighted(this, getName()+"-"+type+"-Depot-TimeSpentInSurplusOnHandAtIHP");
		ResponseVariable t3 = new ResponseVariable(this, getName()+"-"+type+"-Depot-EndSurplusOnHandAtIHP");
		ResponseVariable t4 = new ResponseVariable(this, getName()+"-"+type+"-Depot-EndProbabilityOfSurplusOnHandAtIHP");
		
		mapInventoryToSurplusOnHand.put(inv, t1);
		mapInventoryToTimeSpentInSurplusOnHand.put(inv, t2);
		mapInventoryToEndSurplusOnHand.put(inv, t3);
		mapInventoryToEndProbabilityOfSurplusOnHand.put(inv, t4);
		
		mapIHPToSurplusOnHandAtIHP.get(myDepot).subscribeTo(t1);
		mapIHPToTimeSpentInSurplusOnHandAtIHP.get(myDepot).subscribeTo(t2);
		mapIHPToEndSurplusOnHandAtIHP.get(myDepot).subscribeTo(t3);
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP.get(myDepot).subscribeTo(t4);
		
		/*
		 * The volatility variable for each inventory at the depot are attached to the 
		 * corresponding aggregator for the network 
		 * 
		 */		
		//TODO
		mySurplusInventoryInTheNetwork.subscribeTo(t1);
		myTimeSpentInSurplusOnHandInTheNetwork.subscribeTo(t2);
		myEndSurplusInventoryInTheNetwork.subscribeTo(t3);
		myEndProbabilityOfSurplusOnHandInTheNetwork.subscribeTo(t4);
		
		// TODO
		myItemNumber++;
		VMItem vm = new VMItem(myItemNumber,default_basestock,vmcost,repairdelay.getMean());
		//System.out.println("Added Depot Inventory: "+vm);
		myInventoryToVMItemMap.put(inv,vm);
		myItemTypeToDepotInventoryMap.put(type,inv);
		return inv;
	}
	
	/** Adds a shipping delay from external supplier and the depot
	 * 
	 * @param type
	 * @param transportTime
	 */
	private final void setShippingDelayFromExternalSupplierToDepot(ItemType type, DistributionIfc transportTime){				
		// TODO
		Demand d = new Demand(type);
		// only if the inventory is added to the depot
		if(!myDepot.canFillItemType(d))
			throw new IllegalArgumentException("The supplied depot doesnt have the itemtype inventory");
		
		if (transportTime != null){
			myExternalSupplierShipper.setTransportTime(myDepot, type, transportTime);
		}else{
			myExternalSupplierShipper.setImmediateTransportFlag(true);
		}
	} 
	
	/** Checks if the supplied IRHP is a base
	 * 
	 * @param irhp
	 * @return true if it is a base
	 */
	public final boolean isBase(InventoryRepairingHoldingPoint irhp){
		return(myBases.contains(irhp));
	}

	/** Causes an InventoryRepairingHoldingPoint to be created .
	 * 
	 * @param name must be unique in the Model
	 * @return the created IRHP
	 */
	public final InventoryRepairingHoldingPoint addBase(String name){
		// create the irhp and add it to the network
		InventoryRepairingHoldingPoint irhp = new InventoryRepairingHoldingPoint(this, name);
		myBases.add(irhp);
		// assign the ihp its own time based shipper for shipping to its customers
		TimeBasedTypeLocationIndependentDemandCarrier shipper = new TimeBasedTypeLocationIndependentDemandCarrier(this, irhp.getName() + "TimeBasedShipper");
		irhp.setDemandCarrier(shipper);
		if (myAllowExternalDemandGeneratorsFlag == true){
			shipper.setImmediateTransportFlag(true);
		}
		
		irhp.subscribeToAggregateOnHandInventory(myTotalOnHand);
		irhp.subscribeToAggregateAmountOnOrder(myTotalOnOrder);
		irhp.subscribeToAggregateStockOutIndicator(myTotalStockOutIndicator);
		irhp.subscribeToAggregateAvgFirstFillRate(myAvgFirstFillRate);		
		irhp.subscribeToAggregateReplenishmentDemandCounter(myTotalNumOfReplenishments);		
		irhp.subscribeToAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		irhp.subscribeToAggregateNumberBackOrdered(myTotalNumBackOrdered);
		irhp.subscribeToAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);				

		myDepotShipper.setImmediateTransportFlag(true);
		irhp.setDemandFiller(myDepot);
		
		AggregateCounter c1 = new AggregateCounter(this, getName() + "-"+name+ "-AggregateNumBaseStockIncrement");
		AggregateCounter c2 = new AggregateCounter(this, getName() + "-"+name+ "-AggregateNumBaseStockDecrement");
		ResponseVariableAverageObserver r1 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-AggregateBaseStockIncrement");
		ResponseVariableAverageObserver r2 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-AggregateBaseStockIncrementVariance");
		ResponseVariableAverageObserver r3 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-AggregateBaseStockDecrement");
		ResponseVariableAverageObserver r4 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-AggregateBaseStockDecrementVariance");		
		myLocTotalNumBaseStockIncrement.put(irhp, c1);
		myLocTotalNumBaseStockDecrement.put(irhp, c2);
		myLocTotalBaseStockIncrement.put(irhp, r1);
		myLocTotalBaseStockIncrementVariance.put(irhp, r2);
		myLocTotalBaseStockDecrement.put(irhp, r3);
		myLocTotalBaseStockDecrementVariance.put(irhp, r4);
		
		// TODO
		AggregateTimeWeightedVariable t1 = new AggregateTimeWeightedVariable(this, getName() + "-"+name+ "-SurplusInventoryAtIHP");
		AggregateTimeWeightedVariable t2 = new AggregateTimeWeightedVariable(this, getName() + "-"+name+ "-TimeSpentInSurplusInventoryAtIHP");
		ResponseVariableAverageObserver t3 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-EndSurplusInventoryAtIHP");
		ResponseVariableAverageObserver t4 = new ResponseVariableAverageObserver(this, getName() + "-"+name+ "-EndProbabilityOfSurplusInventoryAtIHP");
		
		mapIHPToSurplusOnHandAtIHP.put(irhp, t1);
		mapIHPToTimeSpentInSurplusOnHandAtIHP.put(irhp, t2);
		mapIHPToEndSurplusOnHandAtIHP.put(irhp, t3);
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP.put(irhp, t4);
		
		return(irhp);
	}

	/** Causes an inventory to be created and added to the base.
	 * Causes a demand arrival listener (of type DemandArrivalListenerWithForecaster) to be created and attached to the inventory.
	 * Causes a forecast arrival listener (of type ForecastArrivalListenerDefault) to be created and attached to the forecaster.
	 * Causes a VMBaseItem to be created and added to the VMItem.
	 * 
	 * @param base
	 * @param type
	 * @param repairprob
	 * @param repairdelay
	 * @param transportTime
	 * @param demandDist
	 * @param agglevel
	 * @param vmbcost
	 * @param forecaster
	 * @return the created inventory
	 */
	public final Inventory addBaseInventory(InventoryRepairingHoldingPoint base, ItemType type, DistributionIfc repairprob,
				DistributionIfc repairdelay, DistributionIfc transportTime, DistributionIfc demandDist, double agglevel, 
				double vmbcost, ForecastManager forecaster, 
				boolean updatable){		
		
		int default_basestock = 1;
		Inventory inv = base.addReorderPointReorderQuantityInventory(type,default_basestock-1,1,default_basestock-1,"Inv-"+base.getName()+"-"+type);
		base.setRepairStation(type, repairprob, repairdelay);
		setShippingDelayFromDepotToBase(base, type, transportTime);
		attachDemandGeneratorToBase(base, type, demandDist, new Constant(0.0));
		
		DemandArrivalListenerWithForecaster myDemandArrivalListener = new DemandArrivalListenerWithForecaster(this, inv, forecaster,
				inv.getName() + " DemandObserver", agglevel, updatable);
		
		ForecastArrivalListenerDefault myForecastArrivalListener = new ForecastArrivalListenerDefault(forecaster, 1/demandDist.getMean(), 
				forecaster+" ForecastArrivalListener");
		
		Counter c1 = new Counter(this, getName()+"-"+type+"-"+base.getName()+"-NumBaseStockIncrement");
		myNumBaseStockIncrement.put(inv, c1);
		
		Counter c2 = new Counter(this, getName()+"-"+type+"-"+base.getName()+"-NumBaseStockDecrement");
		myNumBaseStockDecrement.put(inv, c2);
		
		ResponseVariable r1 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-BaseStockIncrement");		
		myBaseStockIncrement.put(inv, r1);
		
		ResponseVariable r2 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-BaseStockIncrementVariance");
		myBaseStockIncrementVariance.put(inv, r2);
		
		ResponseVariable r3 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-BaseStockDecrement");
		myBaseStockDecrement.put(inv, r3);
		
		ResponseVariable r4 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-BaseStockDecrementVariance");
		myBaseStockDecrementVariance.put(inv, r4);
		
		myLocTotalNumBaseStockIncrement.get(base).subscribeTo(c1);
		myLocTotalNumBaseStockDecrement.get(base).subscribeTo(c2);
		myLocTotalBaseStockIncrement.get(base).subscribeTo(r1);
		myLocTotalBaseStockIncrementVariance.get(base).subscribeTo(r2);
		myLocTotalBaseStockDecrement.get(base).subscribeTo(r3);
		myLocTotalBaseStockDecrementVariance.get(base).subscribeTo(r4);
		
		myTotalNumBaseStockIncrement.subscribeTo(c1);
		myTotalNumBaseStockDecrement.subscribeTo(c2);
		myTotalBaseStockIncrement.subscribeTo(r1);
		myTotalBaseStockIncrementVariance.subscribeTo(r2);
		myTotalBaseStockDecrement.subscribeTo(r3);
		myTotalBaseStockDecrementVariance.subscribeTo(r4);

		
		//TODO
		TimeWeighted t1 = new TimeWeighted(this, getName()+"-"+type+"-"+base.getName()+"-SurplusInventory");
		TimeWeighted t2 = new TimeWeighted(this, getName()+"-"+type+"-"+base.getName()+"-TimeSpentInSurplusInventory");
		ResponseVariable t3 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-EndSurplusInventory");
		ResponseVariable t4 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-EndProbabilityOfSurplusInventory");
		
		mapInventoryToSurplusOnHand.put(inv, t1);
		mapInventoryToTimeSpentInSurplusOnHand.put(inv, t2);
		mapInventoryToEndSurplusOnHand.put(inv, t3);
		mapInventoryToEndProbabilityOfSurplusOnHand.put(inv, t4);		

		mapIHPToSurplusOnHandAtIHP.get(base).subscribeTo(t1);
		mapIHPToTimeSpentInSurplusOnHandAtIHP.get(base).subscribeTo(t2);
		mapIHPToEndSurplusOnHandAtIHP.get(base).subscribeTo(t3);
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP.get(base).subscribeTo(t4);
		
		//TODO
		mySurplusInventoryInTheNetwork.subscribeTo(t1);
		myTimeSpentInSurplusOnHandInTheNetwork.subscribeTo(t2);
		myEndSurplusInventoryInTheNetwork.subscribeTo(t3);
		myEndProbabilityOfSurplusOnHandInTheNetwork.subscribeTo(t4);
		
		myItemNumber++;
		VMItem vm = (VMItem)myInventoryToVMItemMap.get(myItemTypeToDepotInventoryMap.get(type));
		VMBaseItem vmb = (VMBaseItem)vm.addBaseItem(myItemNumber,1/demandDist.getMean(),
				repairdelay.getMean(),repairprob.getMean(),transportTime.getMean(),default_basestock,vmbcost);
		//System.out.println("Added Base Inventory: "+vmb+" to the Base Inventory "+vm);
		myInventoryToVMBaseItemMap.put(inv,vmb);
		myInventoryToBaseDemandArrivalListenerWithForecasterMap.put(inv,myDemandArrivalListener);
		myInventoryToBaseForecastArrivalListenerDefaultMap.put(inv,myForecastArrivalListener);
		return inv;
	}
	
	/** Causes an inventory to be created and added to the base.
	 * Causes a demand arrival listener (of type DemandArrivalListenerWithForecaster) to be created and attached to the inventory.
	 * Causes a forecast arrival listener (of type ForecastArrivalListenerDefault) to be created and attached to the forecaster.
	 * Causes a VMBaseItem to be created and added to the VMItem.
	 * 
	 * @param base
	 * @param type
	 * @param repairprob
	 * @param repairdelay
	 * @param transportTime
	 * @param demandDist
	 * @param initialdemandrate
	 * @param agglevel
	 * @param vmbcost
	 * @param forecaster
	 * @return the created inventory
	 */
	public final Inventory addBaseInventory(InventoryRepairingHoldingPoint base, ItemType type, DistributionIfc repairprob,
			DistributionIfc repairdelay, DistributionIfc transportTime, RandomIfc demandDist, double initialdemandrate, double agglevel, 
			double vmbcost, ForecastManager forecaster, 
			boolean updatable){
	
		int default_basestock = 1;
		Inventory inv = base.addReorderPointReorderQuantityInventory(type,default_basestock-1,1,default_basestock-1,"Inv-"+base.getName()+"-"+type);
		base.setRepairStation(type, repairprob, repairdelay);
		setShippingDelayFromDepotToBase(base, type, transportTime);
		attachDemandGeneratorToBase(base, type, demandDist, new Constant(0.0));
	
		DemandArrivalListenerWithForecaster myDemandArrivalListener = new DemandArrivalListenerWithForecaster(this, inv, forecaster,
				inv.getName() + " DemandObserver", agglevel, updatable);
		
		ForecastArrivalListenerDefault myForecastArrivalListener = new ForecastArrivalListenerDefault(forecaster, initialdemandrate, 
				forecaster+" ForecastArrivalListener");
		
		Counter c1 = new Counter(this, getName()+"-"+type+"-"+base.getName()+"-NumBaseStockIncrement");
		myNumBaseStockIncrement.put(inv, c1);
		
		Counter c2 = new Counter(this, getName()+"-"+type+"-"+base.getName()+"-NumBaseStockDecrement");
		myNumBaseStockDecrement.put(inv, c2);
		
		ResponseVariable r1 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-BaseStockIncrement");		
		myBaseStockIncrement.put(inv, r1);
		
		ResponseVariable r2 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-BaseStockIncrementVariance");
		myBaseStockIncrementVariance.put(inv, r2);
		
		ResponseVariable r3 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-BaseStockDecrement");
		myBaseStockDecrement.put(inv, r3);
		
		ResponseVariable r4 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-BaseStockDecrementVariance");
		myBaseStockDecrementVariance.put(inv, r4);
		
		myLocTotalNumBaseStockIncrement.get(base).subscribeTo(c1);
		myLocTotalNumBaseStockDecrement.get(base).subscribeTo(c2);
		myLocTotalBaseStockIncrement.get(base).subscribeTo(r1);
		myLocTotalBaseStockIncrementVariance.get(base).subscribeTo(r2);
		myLocTotalBaseStockDecrement.get(base).subscribeTo(r3);
		myLocTotalBaseStockDecrementVariance.get(base).subscribeTo(r4);
		
		myTotalNumBaseStockIncrement.subscribeTo(c1);
		myTotalNumBaseStockDecrement.subscribeTo(c2);
		myTotalBaseStockIncrement.subscribeTo(r1);
		myTotalBaseStockIncrementVariance.subscribeTo(r2);
		myTotalBaseStockDecrement.subscribeTo(r3);
		myTotalBaseStockDecrementVariance.subscribeTo(r4);

		
		//TODO
		TimeWeighted t1 = new TimeWeighted(this, getName()+"-"+type+"-"+base.getName()+"-SurplusInventory");
		TimeWeighted t2 = new TimeWeighted(this, getName()+"-"+type+"-"+base.getName()+"-TimeSpentInSurplusInventory");
		ResponseVariable t3 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-EndSurplusInventory");
		ResponseVariable t4 = new ResponseVariable(this, getName()+"-"+type+"-"+base.getName()+"-EndProbabilityOfSurplusInventory");
		
		mapInventoryToSurplusOnHand.put(inv, t1);
		mapInventoryToTimeSpentInSurplusOnHand.put(inv, t2);
		mapInventoryToEndSurplusOnHand.put(inv, t3);
		mapInventoryToEndProbabilityOfSurplusOnHand.put(inv, t4);		

		mapIHPToSurplusOnHandAtIHP.get(base).subscribeTo(t1);
		mapIHPToTimeSpentInSurplusOnHandAtIHP.get(base).subscribeTo(t2);
		mapIHPToEndSurplusOnHandAtIHP.get(base).subscribeTo(t3);
		mapIHPToEndProbabilityOfSurplusOnHandAtIHP.get(base).subscribeTo(t4);
		
		//TODO
		mySurplusInventoryInTheNetwork.subscribeTo(t1);
		myTimeSpentInSurplusOnHandInTheNetwork.subscribeTo(t2);
		myEndSurplusInventoryInTheNetwork.subscribeTo(t3);
		myEndProbabilityOfSurplusOnHandInTheNetwork.subscribeTo(t4);
		
		myItemNumber++;
		VMItem vm = (VMItem)myInventoryToVMItemMap.get(myItemTypeToDepotInventoryMap.get(type));
		VMBaseItem vmb = (VMBaseItem)vm.addBaseItem(myItemNumber, initialdemandrate,
				repairdelay.getMean(),repairprob.getMean(),transportTime.getMean(),default_basestock,vmbcost);
		//System.out.println("Added Base Inventory: "+vmb+" to the Base Inventory "+vm);
		myInventoryToVMBaseItemMap.put(inv,vmb);
		myInventoryToBaseDemandArrivalListenerWithForecasterMap.put(inv,myDemandArrivalListener);
		myInventoryToBaseForecastArrivalListenerDefaultMap.put(inv,myForecastArrivalListener);
		return inv;
	}

	/** Adds a shipping delay from depot and the base
	 * 
	 * @param base
	 * @param type
	 * @param transportTime
	 */
	private final void setShippingDelayFromDepotToBase(InventoryRepairingHoldingPoint base, ItemType type, DistributionIfc transportTime){
		if (!myBases.contains(base))
			throw new IllegalArgumentException("The provided base has not be added to the network");
		
		if (transportTime != null){
			// ensure that it uses its supplier's shipper
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)myDepot.getDemandCarrier();
			shipper.setTransportTime(base, type, transportTime);
		} else {
			// directing an attached IHP to be shipped with no delay time
			// designate the time based shipper to allow no delay shipment
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)myDepot.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}
		
	}

	/** Creates and attaches a demand generator to the IRHP
	 * 
	 * @param supplier
	 * @param type
	 * @param timeUntilNext
	 * @return
	 */
	public final DemandGenerator attachDemandGeneratorToBase(InventoryRepairingHoldingPoint base,
			ItemType type, RandomIfc timeUntilNext, DistributionIfc shiptimeIHPtoDG){
		return(attachDemandGeneratorToBase(base, type, timeUntilNext, shiptimeIHPtoDG, null));
	}
	
	/** Creates and attaches a demand generator to the IRHP
	 * 
	 * @param supplier
	 * @param type
	 * @param timeUntilNext
	 * @param name
	 * @return
	 */
	public final DemandGenerator attachDemandGeneratorToBase(InventoryRepairingHoldingPoint base,
			ItemType type, RandomIfc timeUntilNext, DistributionIfc shiptimeIHPtoDG, String name){
		DemandGenerator dg = new DemandGenerator(this, type, timeUntilNext, name);
		attachDemandGeneratorToBase(base, dg, shiptimeIHPtoDG);
		return(dg);
	}
	
	/** Attaches the supplied demand generator as customer on the given IRHP (supplier)
	 *  with no transport delay
	 *  
	 * @param supplier Must be in the network
	 * @param generator Must not have been already attached to the given supplier
	 */
	public final void attachDemandGeneratorToBase(InventoryRepairingHoldingPoint base, DemandGenerator generator){
		attachDemandGeneratorToBase(base, generator, null);
	}

	/** Attaches the supplied demand generator as a customer on the given IRHP (supplier)
	 *  with the given transport delay
	 *  
	 * @param supplier Must be in the network
	 * @param generator Must not have been already attached to the given supplier
	 * @param transportTime
	 */
	public final void attachDemandGeneratorToBase(InventoryRepairingHoldingPoint base, DemandGenerator generator,
			DistributionIfc transportTime){
		if (!myBases.contains(base))
			throw new IllegalArgumentException("The provided base has not be added to the network");
		
		List<DemandGenerator> generators = myDemandGenerators.get(base);
		if (generators == null){
			generators = new ArrayList<DemandGenerator>();
			myDemandGenerators.put(base, generators);
		}
		if (generators.contains(generator))
			throw new IllegalArgumentException("The provided generator is already attached to the supplier");
		
		generators.add(generator);
		generator.setDemandFiller(base);
		if (transportTime != null){
			// ensure that it uses its supplier's shipper
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)base.getDemandCarrier();
			shipper.setTransportTime(generator, generator.getItemType(), transportTime);
		} else {
			// directing an attached IHP to be shipped with no delay time
			// designate the time based shipper to allow no delay shipment
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)base.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}		
	}
	
	/** Causes the IRHPs in the network to be able to deliver completed demands from 
	 *  any external demand generator with a no delay transport time.  Any IHPs attached
	 *  after this option has been set will also have this capability within their attached shipper
	 */
	public final void allowExternalDemandGenerators(){
		myAllowExternalDemandGeneratorsFlag = true;
		for(InventoryRepairingHoldingPoint irhp: myBases){
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)irhp.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}
	}
	
	/** 
	 * 
	 * @return the bases
	 */
	public final Set<InventoryRepairingHoldingPoint> getBases(){
		return(myBases);
	}
	
	/** 
	 * Initialize the optimization.
	 * Varimetric model is created and optimized and the policy is updated
	 * The next event is scheduled for next update period
	 * 
	 */
	public void initialize(){
		myVariMetricModel = new VariMetricModel();
		for(Inventory inv: myInventoryToVMItemMap.keySet()){
			myVariMetricModel.addVMItem(myInventoryToVMItemMap.get(inv));
		}
//		System.out.println("");
//		System.out.println("");
//		System.out.println("**********************************************************");
//		System.out.println("********* VARIMETRIC MODEL BEFORE FIRST OPTIM ************");
//		System.out.println("**********************************************************");
//		System.out.println("**********************************************************");
		//System.out.println(myVariMetricModel.toString());
//		System.out.println("**********************************************************");
//		System.out.println("**********************************************************");
//		System.out.println("**********************************************************");
//		System.out.println("**********************************************************");
//		System.out.println("");
//		System.out.println("");
//		System.out.println("INITIALIZING THE FIRST OPTIM ............................ ");
//		System.out.println("......................................................... ");
		VMFMarginalAnalysisAlgorithm myAlgorithm =new VMFMarginalAnalysisAlgorithm(myVariMetricModel,budget,tolerance);
		myAlgorithm.optimize();
		// boolean satisfied = myAlgorithm.optimize();
//		System.out.println("INITIALIZED THE FIRST OPTIM!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
//		System.out.println("");
//		System.out.println("");
//		System.out.println("**********************************************************");
//		System.out.println("********* VARIMETRIC MODEL AFTER FIRST OPTIM *************");
//		System.out.println("**********************************************************");
//		System.out.println("**********************************************************");
		//System.out.println(myVariMetricModel.toString());
//		System.out.println("**********************************************************");
//		System.out.println("**********************************************************");
//		System.out.println("**********************************************************");
//		System.out.println("**********************************************************");
//		System.out.println("");
//		System.out.println("");
//		System.out.println("----------------------------------------------------------");		
//		System.out.println("STARTING THE FIRST POLICY UPDATE ........................ ");
		initializePolicy();
//		System.out.println("FINISHED THE FIRST POLICY UPDATE !!!!!!!!!!!!!!!!!!!!!!!!!");
//		System.out.println(" ... ");
//		System.out.println(" ... ");
//		System.out.println(" ... ");
//		System.out.println(" ... ");
		scheduleEvent(myUpdatePeriod, 2, 2, "PolicyUpdatingInterval");		
	}
	
	protected void handleEvent(JSLEvent event){
		schedulePolicyUpdating(event);
	}
	
	/** 
	 * Handles the policy updating for the current update perriod
	 * A Varimetric model is created and optimized and the policy is updated
	 * The next event is scheduled for next update period
	 * 
	 */
	protected void schedulePolicyUpdating(JSLEvent event){
		//System.out.println("");
		//System.out.println("");
		//System.out.println("**********************************************************");
		System.out.println("BEGIN THE NEXT OPTIMIZE AND UPDATE "+getTime()+" ......... ");
		//System.out.println("**********************************************************");
		//System.out.println("**********************************************************");		
		for(Inventory invbase: myInventoryToVMBaseItemMap.keySet()){
			DemandArrivalListenerWithForecaster demlistener = this.myInventoryToBaseDemandArrivalListenerWithForecasterMap.get(invbase);
			//System.out.println(" ... ");
			//System.out.println(" ... ");
			//System.out.println("	RESETTING THE VARIMETRIC FOR NEXT OPTIM AT TIME "+getTime()+" ......... ");
			double agglevel = this.myInventoryToBaseDemandArrivalListenerWithForecasterMap.get(invbase).getAggregationLevel();
			double forecast_estimate_at_agg_level = this.myInventoryToBaseForecastArrivalListenerDefaultMap.get(invbase).getCurrentForecastEstimate();
			double demandrate = forecast_estimate_at_agg_level/agglevel;
			//System.out.println("		"+invbase.getName()+" "+getTime()+" months : "+demandrate);				
			myInventoryToVMBaseItemMap.get(invbase).setDemandRate(demandrate);
			//System.out.println("	RESET THE VARIMETRIC FOR THE CURRENT OPTIM AT TIME "+getTime()+" ......... ");
			//System.out.println(" ... ");
			//System.out.println(" ... ");
		}		
//		System.out.println("");
//		System.out.println("");
//		System.out.println("**********************************************************");
//		System.out.println("************* VARIMETRIC MODEL BEFORE OPTIM **************");
//		System.out.println("**********************************************************");
		//System.out.println(myVariMetricModel.toString());
//		System.out.println("**********************************************************");
//		System.out.println("**********************************************************");
//		System.out.println("");
//		System.out.println("");
//		System.out.println("**********************************************************");
//		System.out.println("BEGIN THE NEXT OPTIM AT TIME "+getTime()+" .............. ");
//		System.out.println("**********************************************************");		
		VMFMarginalAnalysisAlgorithm myAlgorithm =new VMFMarginalAnalysisAlgorithm(myVariMetricModel,budget,tolerance);
		myAlgorithm.optimize();
		// boolean satisfied = myAlgorithm.optimize();
//		System.out.println("");
//		System.out.println("");
//		System.out.println("**********************************************************");
//		System.out.println("************** VARIMETRIC MODEL AFTER OPTIM **************");
//		System.out.println("**********************************************************");
		//System.out.println(myVariMetricModel.toString());
//		System.out.println("**********************************************************");
//		System.out.println("**********************************************************");	
//		System.out.println("");
//		System.out.println("");
//		System.out.println("**********************************************************");
//		System.out.println("	FINISHED THE CURRENT OPTIM AT TIME "+getTime()+" ......... ");
//		System.out.println("**********************************************************");
//		System.out.println("");
//		System.out.println("");
		
		//System.out.println("*************************************************************");
		//System.out.println("************** UPDATING THE POLICY AFTER OPTIM **************");
		//System.out.println("*************************************************************");
		updatePolicy();
//		System.out.println("*************************************************************");
//		System.out.println("************** UPDATED THE POLICY AFTER OPTIM  **************");
//		System.out.println("*************************************************************");
//		System.out.println(" ... ");
//		System.out.println(" ... ");
//		System.out.println(" ... ");
//		System.out.println(" ... ");
		double t = myUpdatePeriod;
		scheduleEvent(t, 2, 2, "PolicyUpdatingInterval");				
		
	}
	
	/** 
	 * Initial policy updating
	 * 
	 */
	protected void initializePolicy(){
		for(Inventory inv: myInventoryToVMItemMap.keySet()){
			int default_basestock = myInventoryToVMItemMap.get(inv).getStockLevel();
			double [] param = {default_basestock-1,1};
			//System.out.println("	UPDATING THE POLICY FOR: "+inv.getName()+" to "+(param[0]+1)+" ......... ");
			inv.setInventoryPolicyParameters(param);
			inv.setInitialOnHand((int)param[1]);			
			//System.out.println("	UPDATED THE POLICY FOR: "+inv.getName());
			//System.out.println("	... ");
		}
		for(Inventory inv: myInventoryToVMBaseItemMap.keySet()){
			int default_basestock = myInventoryToVMBaseItemMap.get(inv).getStockLevel();
			double [] param = {default_basestock-1,1};
			//System.out.println("	UPDATING THE POLICY FOR: "+inv.getName()+" to "+(param[0]+1)+" ......... ");
			inv.setInventoryPolicyParameters(param);
			inv.setInitialOnHand((int)param[1]);			
			//System.out.println("	UPDATED THE POLICY FOR: "+inv.getName());
			//System.out.println("	... ");
		}
	}
	
	/** 
	 * Handles the policy updating for the current update period
	 * 
	 */
	protected void updatePolicy(){
		for(Inventory inv: myInventoryToVMItemMap.keySet()){
			double original = inv.getInventoryPolicyParameters()[0]+1;
			int default_basestock = myInventoryToVMItemMap.get(inv).getStockLevel();
			
			double change = original - default_basestock;			
			
			double [] param = {default_basestock-1,1};
			//System.out.println("	UPDATING THE POLICY FOR: "+inv.getName()+" to "+(param[0]+1)+" ......... ");
			inv.setInventoryPolicyParameters(param);
			
			int ip = inv.getInventoryPosition();
			
			if(change > 0){
				myNumBaseStockIncrement.get(inv).increment();
				myBaseStockIncrement.get(inv).setValue(change);				
				if(ip>default_basestock+1){						
					this.mapInventoryToSurplusOnHand.get(inv).setValue(ip-default_basestock+1);
					this.mapInventoryToTimeSpentInSurplusOnHand.get(inv).setValue(1.0);
				}else{
					this.mapInventoryToSurplusOnHand.get(inv).setValue(0.0);
					this.mapInventoryToTimeSpentInSurplusOnHand.get(inv).setValue(0.0);
				}
			}else if (change < 0){
				myNumBaseStockDecrement.get(inv).increment();
				myBaseStockDecrement.get(inv).setValue(Math.abs(change));
				this.mapInventoryToSurplusOnHand.get(inv).setValue(0.0);
				this.mapInventoryToTimeSpentInSurplusOnHand.get(inv).setValue(0.0);
			}
			
			//System.out.println("	UPDATED THE POLICY FOR: "+inv.getName());
			//System.out.println("	... ");
		}
		for(Inventory inv: myInventoryToVMBaseItemMap.keySet()){
			double original = inv.getInventoryPolicyParameters()[0]+1;
			int default_basestock = myInventoryToVMBaseItemMap.get(inv).getStockLevel();
			
			double change = original - default_basestock;
			
			double [] param = {default_basestock-1,1};
			//System.out.println("	UPDATING THE POLICY FOR: "+inv.getName()+" to "+(param[0]+1)+" ......... ");
			inv.setInventoryPolicyParameters(param);
			
			int ip = inv.getInventoryPosition();
			
			if(change > 0){
				myNumBaseStockIncrement.get(inv).increment();
				myBaseStockIncrement.get(inv).setValue(change);					
				if(ip>default_basestock+1){						
					this.mapInventoryToSurplusOnHand.get(inv).setValue(ip-default_basestock+1);
					this.mapInventoryToTimeSpentInSurplusOnHand.get(inv).setValue(1.0);
				}else{
					this.mapInventoryToSurplusOnHand.get(inv).setValue(0.0);
					this.mapInventoryToTimeSpentInSurplusOnHand.get(inv).setValue(0.0);
				}
			}else if (change < 0){
				myNumBaseStockDecrement.get(inv).increment();
				myBaseStockDecrement.get(inv).setValue(Math.abs(change));
				this.mapInventoryToSurplusOnHand.get(inv).setValue(0.0);
				this.mapInventoryToTimeSpentInSurplusOnHand.get(inv).setValue(0.0);
			}
			
			
			//System.out.println("	UPDATED THE POLICY FOR: "+inv.getName());
			//System.out.println("	... ");
		}
	}
	
	protected void replicationEnded() {		
		
		for(Inventory inv: myInventoryToVMItemMap.keySet()){
			
			double var_inc = myBaseStockIncrement.get(inv).getWithinReplicationStatistic().getVariance();
			//System.out.println("Var of Increment Depot "+var_inc);
			myBaseStockIncrementVariance.get(inv).setValue(var_inc);			
			
			
			double last_surplus_on_hand = mapInventoryToSurplusOnHand.get(inv).getWithinReplicationStatistic().getLastValue();
			mapInventoryToEndSurplusOnHand.get(inv).setValue(last_surplus_on_hand);
			if(last_surplus_on_hand>0){
				mapInventoryToEndProbabilityOfSurplusOnHand.get(inv).setValue(1.0);
			}else{
				mapInventoryToEndProbabilityOfSurplusOnHand.get(inv).setValue(0.0);
			}			
			
			double var_dec = myBaseStockDecrement.get(inv).getWithinReplicationStatistic().getVariance();
			//System.out.println("Var of Decrement Depot "+var_inc);			
			myBaseStockDecrementVariance.get(inv).setValue(var_dec);			
			
		}
		
		for(Inventory inv: myInventoryToVMBaseItemMap.keySet()){
			
			double var_inc = myBaseStockIncrement.get(inv).getWithinReplicationStatistic().getVariance();
			//System.out.println("Var of Increment Base "+var_inc);
			myBaseStockIncrementVariance.get(inv).setValue(var_inc);			
			
			double last_surplus_on_hand = mapInventoryToSurplusOnHand.get(inv).getWithinReplicationStatistic().getLastValue();
			mapInventoryToEndSurplusOnHand.get(inv).setValue(last_surplus_on_hand);
			if(last_surplus_on_hand>0){
				mapInventoryToEndProbabilityOfSurplusOnHand.get(inv).setValue(1.0);
			}else{
				mapInventoryToEndProbabilityOfSurplusOnHand.get(inv).setValue(0.0);
			}			
			
			
			double var_dec = myBaseStockDecrement.get(inv).getWithinReplicationStatistic().getVariance();
			//System.out.println("Var of Decrement Base "+var_inc);			
			myBaseStockDecrementVariance.get(inv).setValue(var_dec);
			
		}
		
	}
	
}