/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.afosr.varimetricsimulator;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jsl.modeling.Experiment;
import jsl.modeling.Model;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.nhpp.NHPPTimeBtwEventRV;
import jsl.modeling.elements.variable.nhpp.PiecewiseLinearRateFunction;
import jsl.modeling.elements.variable.nhpp.PiecewiseRateFunction;
import jsl.observers.textfile.CSVExperimentReport;
import jsl.observers.textfile.CSVReport;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Bernoulli;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jslx.forecasting.Croston;
import jslx.forecasting.ForecastTechnique;
import jslx.supplychain.afosr.examples.MetricIRHPNetwork;
import jslx.supplychain.afosr.forecasting.DemandArrivalListenerWithForecaster;
import jslx.supplychain.afosr.forecasting.ForecastArrivalListenerDefault;
import jslx.supplychain.afosr.forecasting.ForecastManager;
import jslx.supplychain.afosr.generation.BItem;
import jslx.supplychain.afosr.generation.DItem;
import jslx.supplychain.afosr.generation.NetworkConfiguration;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryRepairingHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;
import jslx.supplychain.transportlayer.TimeBasedDemandCarrier;
import jslx.supplychain.transportlayer.TimeBasedTypeLocationIndependentDemandCarrier;

/** 
 * This class models a METRIC model
 * 
 */

public class SimVMModel extends ModelElement {
	
	private LeadTimeDemandFiller myExternalSupplier;
	private TimeBasedTypeLocationIndependentDemandCarrier myExternalSupplierCarrier;
	private InventoryRepairingHoldingPoint myDepot;
	private TimeBasedTypeLocationIndependentDemandCarrier myDepotCarrier;
	
	private Set <InventoryRepairingHoldingPoint> myBases;
	private Set <String> myBaseIDs;	
	private Set <ItemType> myItems;
	private Set <String> myItemIDs;	
	private Map <InventoryRepairingHoldingPoint, List<DemandGenerator>> myDemandGenerators;
		
	private AggregateTimeWeightedVariable myTotalAmtBackOrdered;
	private AggregateTimeWeightedVariable myTotalNumBackOrdered;
	private AggregateTimeWeightedVariable myTotalOnHand;
	private AggregateTimeWeightedVariable myTotalOnOrder;
	private AggregateTimeWeightedVariable myTotalStockOutIndicator;
	private ResponseVariableAverageObserver myAvgFirstFillRate;
	private ResponseVariableAverageObserver myAvgCustomerWaitTime;
	private AggregateCounter myTotalNumOfReplenishments;
	
	private boolean myAllowExternalDemandGeneratorsFlag = false;
	
	public SimVMModel(ModelElement parent) {
		this(parent, null);
	}
	
	public SimVMModel(ModelElement parent, String name) {
		super(parent, name);
				
		myBases = new HashSet<InventoryRepairingHoldingPoint>();
		myBaseIDs = new HashSet<String>();
		myItems = new HashSet<ItemType>();
		myItemIDs = new HashSet<String>();
		myDemandGenerators = new HashMap<InventoryRepairingHoldingPoint, List<DemandGenerator>>();
		
		myTotalOnHand = new AggregateTimeWeightedVariable(this, getName() + " Total On Hand");
		myAvgFirstFillRate = new ResponseVariableAverageObserver(this, getName() + " Aggregate Fill Rate");
		myTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, getName() + " Aggregate Number of Stockouts");
		myTotalOnOrder = new AggregateTimeWeightedVariable(this, getName() + " Total On Order");
		myTotalNumOfReplenishments = new AggregateCounter(this, getName() + " NumberOfReplenishments");
		myTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Amt BackOrdered");
		myTotalNumBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Number BackOrdered");
		myAvgCustomerWaitTime = new ResponseVariableAverageObserver(this, getName() + "Avg Customer Wait");
		
		//create external LeadTimeDemandFiller and its own TimeBasedDemandShipper		
		myExternalSupplier = new LeadTimeDemandFiller(this, getName() + "ExternalSupplier");
		myExternalSupplierCarrier = new TimeBasedTypeLocationIndependentDemandCarrier(this, getName() + "ExternalShipper");
		myExternalSupplier.setDemandCarrier(myExternalSupplierCarrier);		
		myDepot = new InventoryRepairingHoldingPoint(this, getName() + "Depot");
		myDepotCarrier = new TimeBasedTypeLocationIndependentDemandCarrier(this, getName() + "DepotCarrier");
		myDepot.setDemandCarrier(myDepotCarrier);		
		myExternalSupplierCarrier.setImmediateTransportFlag(true);
		myDepot.setDemandFiller(myExternalSupplier);

	}
	
	public final ItemType addItemType(String name){
		if(myItemIDs.contains(name))
			throw new IllegalArgumentException("The item has been already added");		
		myItemIDs.add(name);
		
		ItemType type = new ItemType(this, name);
		myItems.add(type);
		
		return(type);
	}
	
	/** Adds an ItemType and its lead time distribution to the network. The name
	 *  of the item type must be unique to the model
	 * 
	 * @param name
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(String name, RandomIfc leadtime){
		if(myItemIDs.contains(name))
			throw new IllegalArgumentException("The item has been already added");		
		myItemIDs.add(name);
		
		ItemType type = new ItemType(this, name);
		myItems.add(type);
		myExternalSupplier.addLeadTime(type, leadtime);
		
		return(type);
	}
	
	/** Causes an InventoryRepairingHoldingPoint to be created with a uniquely assigned name
	 *  and added to the network.  This only adds the IHRP to the network it does
	 *  not cause the IHRP to be connected to any other IHRPs in the network. The client
	 *  can use the returned reference to the IHRP to connect to other IHRPs.
	 * 
	 * @return the created IHRP
	 */
	public final InventoryRepairingHoldingPoint addBase(){
		return(addBase(null));
	}
	
	/** Causes an InventoryRepairingHoldingPoint to be created with the supplied name
	 *  and added to the network.  This only adds the IHRP to the network it does
	 *  not cause the IHRP to be connected to any other IHRPs in the network. The client
	 *  can use the returned reference to the IHRP to connect to other IHRPs.
	 * 
	 * @param name must be unique in the Model
	 * @return the created IHRP
	 */
	public final InventoryRepairingHoldingPoint addBase(String name){
		if(myBaseIDs.contains(name))
			throw new IllegalArgumentException("The base has been already added");
		
		myBaseIDs.add(name);	
		
		//create the ihrp and add it to the network so that it may have child ihrps
		InventoryRepairingHoldingPoint ihrp = new InventoryRepairingHoldingPoint(this, name);
		myBases.add(ihrp);
		
		// assign the ihp its own time based shipper for shipping to its customers
		TimeBasedTypeLocationIndependentDemandCarrier shipper = new TimeBasedTypeLocationIndependentDemandCarrier(this, ihrp.getName() + "TimeBasedGenericCarrier");
		ihrp.setDemandCarrier(shipper);
		if (myAllowExternalDemandGeneratorsFlag == true){
			shipper.setImmediateTransportFlag(true);
		}

		ihrp.subscribeToAggregateOnHandInventory(myTotalOnHand);
		ihrp.subscribeToAggregateAmountOnOrder(myTotalOnOrder);
		ihrp.subscribeToAggregateStockOutIndicator(myTotalStockOutIndicator);
		ihrp.subscribeToAggregateAvgFirstFillRate(myAvgFirstFillRate);
		ihrp.subscribeToAggregateReplenishmentDemandCounter(myTotalNumOfReplenishments);		
		ihrp.subscribeToAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		ihrp.subscribeToAggregateNumberBackOrdered(myTotalNumBackOrdered);
		ihrp.subscribeToAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);			
		
		attachBaseToDepot(ihrp);
		myBases.add(ihrp);
		
		return(ihrp);
	}
	
	public final Inventory addDepotInventory(ItemType type, 
			int basestock, int initiallevel, DistributionIfc repairdelay, DistributionIfc transportTime){
		Inventory inv = myDepot.addReorderPointReorderQuantityInventory(type, basestock-1, 1, initiallevel,
				"Inv-"+myDepot.getName()+"-"+type);
		myDepot.setRepairStation(type, new Bernoulli(1.0), repairdelay);
		setShippingDelayFromExternalSupplierToDepot(type, transportTime);
		
		return inv;
	}
	
	private final void setShippingDelayFromExternalSupplierToDepot(ItemType type, DistributionIfc transportTime){				
		Demand d = new Demand(type);
		// assign the depot correctly
		if(!myDepot.canFillItemType(d))
			throw new IllegalArgumentException("The supplied depot doesnt have the itemtype inventory");		
		if (transportTime != null){
			TimeBasedTypeLocationIndependentDemandCarrier c = (TimeBasedTypeLocationIndependentDemandCarrier) getExternalSupplier().getDemandCarrier();
			c.setTransportTime(myDepot, type, transportTime);
		}else{
			TimeBasedTypeLocationIndependentDemandCarrier c = (TimeBasedTypeLocationIndependentDemandCarrier) getExternalSupplier().getDemandCarrier();
			c.setImmediateTransportFlag(true);
		}
	} 

	public final Inventory addBaseInventory(InventoryRepairingHoldingPoint base, ItemType type, 
		DistributionIfc repairprob, DistributionIfc repairdelay, DistributionIfc transportTime, 
		RandomIfc demandDist, double initialdemandrate, int basestocklevel, int initiallevel){
	
		Inventory inv = base.addReorderPointReorderQuantityInventory(type,basestocklevel-1,1,initiallevel,"Inv-"+base.getName()+"-"+type);
		base.setRepairStation(type, repairprob, repairdelay);
		setShippingDelayFromDepotToBase(base, type, transportTime); 
		
		DemandGenerator dg = addDemandGeneratorToBase(type, demandDist);                
		attachDemandGeneratorToBase(base, dg, new Constant(0.0));
		
		return inv;
	}

	/** Adds a shipping delay from depot and the base
	 * 
	 * @param base
	 * @param type
	 * @param transportTime
	 */
	private final void setShippingDelayFromDepotToBase(InventoryRepairingHoldingPoint base, ItemType type, DistributionIfc transportTime){
		// assign the list of bases correctly and then correct below
		if (!getBases().contains(base))
			throw new IllegalArgumentException("The provided base has not be added to the network");		
		if (transportTime != null){
			TimeBasedTypeLocationIndependentDemandCarrier c = (TimeBasedTypeLocationIndependentDemandCarrier) myDepot.getDemandCarrier();
			c.setTransportTime(base, type, transportTime);
		}else{
			TimeBasedTypeLocationIndependentDemandCarrier c = (TimeBasedTypeLocationIndependentDemandCarrier) myDepot.getDemandCarrier();
			c.setImmediateTransportFlag(true);
		}		
	}

	public final boolean isDepot(InventoryRepairingHoldingPoint ihrp){
		return(ihrp==myDepot);
	}
	
	public final boolean isBase(InventoryRepairingHoldingPoint ihrp){
		return(myBases.contains(ihrp));
	}
	
	public final void attachBaseToDepot(InventoryRepairingHoldingPoint customer){
		
		myBases.add(customer);
		// ensure that it uses the supplier for its replenishments
		customer.setDemandFiller(myDepot);
	}
	
	public final DemandGenerator addDemandGeneratorToBase(ItemType type, RandomIfc timeUntilNext){
		return(addDemandGeneratorToBase(type, timeUntilNext, null));
	}
	
	public final DemandGenerator addDemandGeneratorToBase(ItemType type, RandomIfc timeUntilNext, String name){
		DemandGenerator dg = new DemandGenerator(this, type, timeUntilNext, name);
		return(dg);
	}
	
	public final void attachDemandGeneratorToBase(InventoryRepairingHoldingPoint supplier, DemandGenerator generator){
		attachDemandGeneratorToBase(supplier, generator, null);
	}

	public final void attachDemandGeneratorToBase(InventoryRepairingHoldingPoint supplier, DemandGenerator generator,
			DistributionIfc transportTime){
		if (!myBases.contains(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		List<DemandGenerator> generators = myDemandGenerators.get(supplier);
		if (generators == null){
			generators = new ArrayList<DemandGenerator>();
			myDemandGenerators.put(supplier, generators);
		}
		if (generators.contains(generator))
			throw new IllegalArgumentException("The provided generator is already attached to the supplier");
		
		generators.add(generator);
		generator.setDemandFiller(supplier);
		if (transportTime != null){
			// ensure that it uses its supplier's shipper
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)supplier.getDemandCarrier();
			shipper.setTransportTime(generator, generator.getItemType(), transportTime);
		} else {
			// directing an attached IHP to be shipped with no delay time
			// designate the time based shipper to allow no delay shipment
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)supplier.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}		
	}
	
	public final void allowExternalDemandGenerators(){
		myAllowExternalDemandGeneratorsFlag = true;
		for (InventoryRepairingHoldingPoint ihp: myBases){
			TimeBasedDemandCarrier shipper = (TimeBasedDemandCarrier)ihp.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}
	}
	
	public final Set<InventoryRepairingHoldingPoint> getBases(){
		return myBases;
	}

	public LeadTimeDemandFiller getExternalSupplier() {
		return myExternalSupplier;
	}

	public InventoryRepairingHoldingPoint getDepot() {
		return myDepot;
	}
	
	public String getModelStringFormat(){
		String s = "";
		s = s + "External Supplier: "+this.myExternalSupplier.getName()+"\t"+"External Carrier: "+this.myExternalSupplierCarrier.getName();
		s = s + "\n";
		s = s + "**************************************************************";
		s = s + "\n";
		s = s + "Depot: "+this.myDepot.getName()+"\t"+"Depot Carrier: "+this.myDepotCarrier.getName();
		s = s + "\n";
		s = s + "**************************************************************";
		s = s + "\n";
		s = s + "ItemTypes";
		s = s + "\n";
		for(ItemType type: this.myItems){
			s = s + type.getName() + "\n";			
		}
		s = s + "**************************************************************";
		s = s + "\n";
		s = s + "Depot Inventory";
		s = s + "\n";		
		for(ItemType type: myDepot.getItemTypes()){
			s = s + type.getName();
			s = s + "\t";
			s = s + "Prob \t";
			Inventory inv = myDepot.getInventory(type);
			/*double [] prob = myDepot.getRepairProbability().get(type).getParameters();			
			for(int i = 0; i < prob.length; i++){
				s = s + prob[i] +"\t";				
			}*/
			double prob = myDepot.getRepairProbabilityParameters(type);
			s = s + prob +"\t";
			
			s = s + "Delay \t";
			/*double [] rep = myDepot.getLocalRepairingPoint().getLeadTimes().get(type).getParameters();
			for(int i = 0; i < rep.length; i++){
				s = s + rep[i] +"\t";				
			}*/
			double delay = myDepot.getRepairingDelay(type);
			s = s + delay +"\t";			
			s = s + "\n";		
		}
		s = s + "**************************************************************";
		s = s + "\n";		
		s = s + "Base Inventory";
		s = s + "\n";		
		for(InventoryRepairingHoldingPoint ihrp: myBases){		
			s = s + "Base "+ihrp.getName();
			s = s + "\n";		
			for(ItemType type: ihrp.getItemTypes()){
				s = s + type.getName();
				s = s + "\t";
				Inventory inv = ihrp.getInventory(type);
				
				s = s + "Prob \t";				
				double prob = ihrp.getRepairProbabilityParameters(type);
				s = s + prob +"\t";
				
				s = s + "Delay \t";
				double repairdelay = ihrp.getRepairingDelay(type);
				s = s + repairdelay +"\t";
				
				s = s + "Ship Delay \t";
				double shipdelay = myDepotCarrier.getTransportTimes(type, ihrp);
				s = s + shipdelay +"\t";
				
				s = s + "\n";
			}			
		}
		s = s + "**************************************************************";
		s = s + "\n";		
		return s;		
	}
	
	public static void main(String []args){
		Model m = Model.createModel();
		
		//Create the network
		SimVMModel n = new SimVMModel(m, "SherbrookeExampleOfMetric");
		
		//Define item types
		ItemType type1 = n.addItemType("Type-1");
		
		//addInventory to depot
		int basestock = 0;// corresponding r, Q policy is (2, 1)
		int initiallevel = 0;
		n.addDepotInventory(type1, basestock, initiallevel, new Constant(0.02531), new Constant(0.0));
		
		// Create the retailers which use the warehouse as its filler
		InventoryRepairingHoldingPoint base1 = n.addBase("Base1");
		InventoryRepairingHoldingPoint base2 = n.addBase("Base2");
		InventoryRepairingHoldingPoint base3 = n.addBase("Base3");
		InventoryRepairingHoldingPoint base4 = n.addBase("Base4");
		InventoryRepairingHoldingPoint base5 = n.addBase("Base5");
		
		//Create the inventory at each retailer
		n.addBaseInventory(base1, type1, new Bernoulli(0.2), new Constant(0.01), new Constant(0.01), new Exponential(1.0/23.2), 23.2, 0, 0);
		n.addBaseInventory(base2, type1, new Bernoulli(0.2), new Constant(0.01), new Constant(0.01), new Exponential(1.0/23.2), 23.2, 0, 0);
		n.addBaseInventory(base3, type1, new Bernoulli(0.2), new Constant(0.01), new Constant(0.01), new Exponential(1.0/23.2), 23.2, 0, 0);
		n.addBaseInventory(base4, type1, new Bernoulli(0.2), new Constant(0.01), new Constant(0.01), new Exponential(1.0/23.2), 23.2, 0, 0);
		n.addBaseInventory(base5, type1, new Bernoulli(0.2), new Constant(0.01), new Constant(0.01), new Exponential(1.0/23.2), 23.2, 0, 0);
		
		System.out.println(n.getModelStringFormat());
		
		// create the experiment to run the model
		Experiment e = new Experiment(m,"SherbrookeExampleOfMetric");
		
		try{
        	CSVReport ecsv = new CSVExperimentReport("SherbrookeExampleOfMetric");
        	m.addObserver(ecsv);
        }
        catch(IOException exception){
			System.out.println("I/O exception opening file Output.txt");
			exception.printStackTrace();
        }
        
		// set the parameters of the experiment
		e.setNumberOfReplications(10);
		e.setLengthOfReplication(3650.0);
		e.setLengthOfWarmUp(100.0);
		
		e.turnOnExperimentReport();
		e.turnOnReplicationReport();
		e.turnOnExperimentConsoleOutput(true);
		
		// tell the experiment to run
		e.runAll();
		
	}
	
	
}
