/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.afosr.varimetricsimulator;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import jsl.modeling.Experiment;
import jsl.modeling.Model;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.nhpp.NHPPTimeBtwEventRV;
import jsl.modeling.elements.variable.nhpp.PiecewiseLinearRateFunction;
import jsl.modeling.elements.variable.nhpp.PiecewiseRateFunction;
import jsl.observers.textfile.CSVExperimentReport;
import jsl.observers.textfile.CSVReport;
import jsl.utilities.random.distributions.Bernoulli;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DUniform;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.random.RandomIfc;
import jslx.supplychain.afosr.generation.BItem;
import jslx.supplychain.afosr.generation.DItem;
import jslx.supplychain.afosr.generation.NetworkConfiguration;
import jslx.supplychain.afosr.generation.NetworkGenerator;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryRepairingHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;

/** 
 * This class models a METRIC model
 * 
 */

public class SimVMModelNCModel extends ModelElement {
	
	protected SimVMModel mySimVMModel;
	
	protected Set <String> myBaseIDs;	
		
	protected Map <DItem, ItemType> mapDItemToItemType;	
	protected Map <DItem, Inventory> mapDItemToInventory;
	protected Map <BItem, Inventory> mapBItemToInventory;
	protected Map <BItem, InventoryRepairingHoldingPoint> mapBItemToIRHP;
	protected Map <String, InventoryRepairingHoldingPoint> mapBaseIDToIRHP;	
	
	public SimVMModelNCModel(ModelElement parent, NetworkConfiguration nc) {
		this(parent, nc, null);
	}
	
	public SimVMModelNCModel(ModelElement parent, NetworkConfiguration nc, String name) {
		super(parent, name);
		
		mySimVMModel = new SimVMModel(parent, name+"-SimVMModel");		
		mapDItemToItemType = new HashMap<DItem, ItemType>();
		mapDItemToInventory = new HashMap<DItem, Inventory>();
		mapBItemToInventory = new HashMap<BItem, Inventory>();	
		mapBItemToIRHP = new HashMap<BItem, InventoryRepairingHoldingPoint>();	
		mapBaseIDToIRHP = new HashMap<String, InventoryRepairingHoldingPoint>();
		myBaseIDs = new HashSet<String>();
		
		addItemTypes(nc);
		addBases(nc);
		setInventoryAtEachLocation(nc);
	}
	
	protected void addItemTypes(NetworkConfiguration nc){
		for(DItem di:nc.myItems){
			int id = di.myItemID;
			String sid = "Type-"+id;			
			ItemType type = mySimVMModel.addItemType(sid);			
			mapDItemToItemType.put(di, type);			
		}
	}
	
	protected void addBases(NetworkConfiguration nc){
		for(DItem di:nc.myItems){
			for(BItem bi:di.myBItems){
				int id = bi.myBaseID;
				String bid = "Base-"+id;
				System.out.println(bid+" >> "+myBaseIDs.contains(bid));
				if(!myBaseIDs.contains(bid)){
					InventoryRepairingHoldingPoint base = mySimVMModel.addBase(bid);
					myBaseIDs.add(bid);
					mapBItemToIRHP.put(bi, base);
					mapBaseIDToIRHP.put(bid, base);
				}else{
					mapBItemToIRHP.put(bi, mapBaseIDToIRHP.get(bid));
				}
				
			}			
		}
	}
		
	protected void setInventoryAtEachLocation(NetworkConfiguration nc) {
		//loop through the DItems and add the depot inventory
		for(DItem di: nc.myItems){
			int locationid = 0;// location id
	        String name = di.myName;// item name
	        int itemid = di.myItemID;// item id
	        double cost = di.myCost;// item cost
	        int costlevel = di.myCostLevel;// item cost level
	        int basestocklevel = di.myBaseStockLevel; // item stock level
	        double lambda = di.myLambda;// average demand
	        double repairdelay = di.myRepairDelay;// repair delay
	        double reparability = 1.0;// reparability
	        double shipdelay = 0.0;// ship delay
	        double dibudgetnw = di.getDItemBudget();// item budget at location
	        double dibudgetloc = di.getBudget();// total item level budget
	        int numbasesstocked = di.myNumStockedBases;// num stocked level
	        
	        ItemType type = this.mapDItemToItemType.get(di);
	        Inventory inv = addDepotInventory(type, basestocklevel, basestocklevel, new Exponential(repairdelay), new Constant(shipdelay));
	        mapDItemToInventory.put(di, inv);	        
	        
	        for(BItem bi: di.myBItems){
				int locationid_ = 0;// location id
		        String name_ = bi.myName;// item name
		        int itemid_ = bi.myItemID;// item id
		        double cost_ = bi.myCost;// item cost
		        int costlevel_ = bi.myCostLevel;// item cost level
		        int basestocklevel_ = bi.myBaseStockLevel; // item stock level
		        double lambda_ = bi.myLambda;// average demand
		        double repairdelay_ = bi.myRepairDelay;// repair delay
		        double reparability_ = bi.myReparability;// reparability
		        double shipdelay_ = bi.myShipDelay;// ship delay
		        double bibudgetnw_ = bi.getBudget();// item budget at location
		        double bibudgetloc_ = di.getBudget();// total item level budget
                int plevel = bi.myPatternLevel;
                double tp = bi.myHorizonSpan;
                double td = bi.myDeploymentSpan;
                double ts = bi.mySustainmentSpan;
                double tr = bi.myReconfigurationSpan;
                double lMIN = bi.myLambdaMin;
                double lMAX = bi.myLambdaMax;
		        int numbasesstocked_ = 0;// num stocked level
		        
		        InventoryRepairingHoldingPoint base = this.mapBItemToIRHP.get(bi);
		        
		        PiecewiseRateFunction f = new PiecewiseLinearRateFunction(lMIN, td, lMAX);
		        f.addRateSegment(ts,lMAX);
				f.addRateSegment(tr,lMIN);		
				NHPPTimeBtwEventRV myNHPPTimeBtwEventRV = new NHPPTimeBtwEventRV(this, f);
				
				Inventory inv_ = addBaseInventory(base, type, new Bernoulli(reparability_), 
		        		new Exponential(repairdelay_), new Exponential(shipdelay_), 
		        		myNHPPTimeBtwEventRV, lMIN, basestocklevel_, basestocklevel);
		        
		        mapBItemToInventory.put(bi, inv_);			
			}			
		}	
		//loop through the BItems and add the base inventory	
	}

	private final Inventory addDepotInventory(ItemType type, 
			int basestock, int initiallevel, DistributionIfc repairdelay, DistributionIfc transportTime){
		Inventory inv = mySimVMModel.addDepotInventory(type, basestock, initiallevel, repairdelay, transportTime);
		
		return inv;
	} 

	private final Inventory addBaseInventory(InventoryRepairingHoldingPoint base, ItemType type, 
		DistributionIfc repairprob, DistributionIfc repairdelay, DistributionIfc transportTime, 
		RandomIfc demandDist, double initialdemandrate, int basestocklevel, int initiallevel){
	
		Inventory inv = mySimVMModel.addBaseInventory(base, type, repairprob, repairdelay, transportTime, 
				demandDist, initialdemandrate, basestocklevel, initiallevel);
		
		return inv;
	}
	
	public final ItemType addItemType(String name){
		ItemType type = mySimVMModel.addItemType(name);
		
		return(type);
	}
	
	/** Adds an ItemType and its lead time distribution to the network. The name
	 *  of the item type must be unique to the model
	 * 
	 * @param name
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(String name, RandomIfc leadtime){
		
		ItemType type = mySimVMModel.addItemType(name, leadtime);
		
		return(type);
	}
	
	/** Causes an InventoryRepairingHoldingPoint to be created with a uniquely assigned name
	 *  and added to the network.  This only adds the IHRP to the network it does
	 *  not cause the IHRP to be connected to any other IHRPs in the network. The client
	 *  can use the returned reference to the IHRP to connect to other IHRPs.
	 * 
	 * @return the created IHRP
	 */
	public final InventoryRepairingHoldingPoint addBase(){
		return(addBase(null));
	}
	
	/** Causes an InventoryRepairingHoldingPoint to be created with the supplied name
	 *  and added to the network.  This only adds the IHRP to the network it does
	 *  not cause the IHRP to be connected to any other IHRPs in the network. The client
	 *  can use the returned reference to the IHRP to connect to other IHRPs.
	 * 
	 * @param name must be unique in the Model
	 * @return the created IHRP
	 */
	public final InventoryRepairingHoldingPoint addBase(String name){
		
		InventoryRepairingHoldingPoint ihrp = mySimVMModel.addBase(name);
		
		return(ihrp);
	}
	
	public final boolean isDepot(InventoryRepairingHoldingPoint ihrp){
		return(mySimVMModel.isDepot(ihrp));
	}
	
	public final boolean isBase(InventoryRepairingHoldingPoint ihrp){
		return(mySimVMModel.isBase(ihrp));
	}

	public Map<BItem, InventoryRepairingHoldingPoint> getMapBItemToIRHP() {
		return mapBItemToIRHP;
	}

	public Map<DItem, ItemType> getMapDItemToItemType() {
		return mapDItemToItemType;
	}

	public Map<DItem, Inventory> getMapDItemToInventory() {
		return mapDItemToInventory;
	}

	public Map<BItem, Inventory> getMapBItemToInventory() {
		return mapBItemToInventory;
	}
	
	public final Set<InventoryRepairingHoldingPoint> getBases(){
		return mySimVMModel.getBases();
	}

	public LeadTimeDemandFiller getExternalSupplier() {
		return mySimVMModel.getExternalSupplier();
	}

	public InventoryRepairingHoldingPoint getDepot() {
		return mySimVMModel.getDepot();
	}
	
	public static void main(String[]args){
        /*double [] cratio = {0.6, 0.1};
        double tp = 500;
        double [] pratio = {0.1, 0.1, 0.1};
        DUniform items = new DUniform(1, 5);
        DUniform bases = new DUniform(1, 10);
        double budgetconstant = 0.7;*/
        double [] cratio = {0.8, 0.1};
        double tp = 500;
        double [] pratio = {0.1, 0.1, 0.1};
        DUniform items = new DUniform(3, 5);
        DUniform bases = new DUniform(3, 10);
        double budgetconstant = 0.5;
        
        NetworkGenerator ng = new NetworkGenerator(items, bases, budgetconstant, 
                cratio, pratio, tp);
        
        NetworkConfiguration nc =  ng.getNetwork();
        
        System.out.println(nc);
        
		Model m = Model.createModel();
		
		//Create the network
		SimVMModelNCModel n = new SimVMModelNCModel(m, nc, "SimVMModelNCModel-Sept10-T1");
        
		System.out.println(n.mySimVMModel.getModelStringFormat());
		
		// create the experiment to run the model
		Experiment e = new Experiment(m,"SimVMModelNCModel-T1");
		
		try{
        	CSVReport ecsv = new CSVExperimentReport("SimVMModelNCModel-T1");
        	m.addObserver(ecsv);
        }
        catch(IOException exception){
			System.out.println("I/O exception opening file Output.txt");
			exception.printStackTrace();
        }
        
		// set the parameters of the experiment
		e.setNumberOfReplications(1);
		e.setLengthOfReplication(3650.0);
		e.setLengthOfWarmUp(100.0);
		
		e.turnOnExperimentReport();
		e.turnOnReplicationReport();
		e.turnOnExperimentConsoleOutput(true);
		
		// tell the experiment to run
		e.runAll();
		
		
	}
	
	
}
