/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.afosr.varimetricsimulator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.nhpp.NHPPTimeBtwEventRV;
import jsl.modeling.elements.variable.nhpp.PiecewiseLinearRateFunction;
import jsl.modeling.elements.variable.nhpp.PiecewiseRateFunction;
import jsl.utilities.random.distributions.Bernoulli;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.random.RandomIfc;
import jslx.forecasting.Croston;
import jslx.forecasting.ForecastTechnique;
import jslx.supplychain.afosr.forecasting.DemandArrivalListenerWithForecaster;
import jslx.supplychain.afosr.forecasting.ForecastArrivalListenerDefault;
import jslx.supplychain.afosr.forecasting.ForecastManager;
import jslx.supplychain.afosr.generation.BItem;
import jslx.supplychain.afosr.generation.DItem;
import jslx.supplychain.afosr.generation.NetworkConfiguration;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryRepairingHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;
import jslx.supplychain.transportlayer.TimeBasedDemandCarrier;
import jslx.supplychain.transportlayer.TimeBasedTypeLocationIndependentDemandCarrier;

/** 
 * This class models a METRIC model
 * 
 */

public class SimVMModelOriginal extends ModelElement {
	
	private LeadTimeDemandFiller myExternalSupplier;
	private TimeBasedTypeLocationIndependentDemandCarrier myExternalSupplierCarrier;
	private InventoryRepairingHoldingPoint myDepot;
	private TimeBasedTypeLocationIndependentDemandCarrier myDepotCarrier;
	
	private Set <InventoryRepairingHoldingPoint> myBases;
	private Set <String> myBaseIDs;	
	private Set <ItemType> myItems;
	private Set <String> myItemIDs;	
	private Map <InventoryRepairingHoldingPoint, List<DemandGenerator>> myDemandGenerators;
	private Map <Inventory, DemandArrivalListenerWithForecaster> mapInventoryToDemandArrivalListener;
	private Map <Inventory, ForecastArrivalListenerDefault> mapInventoryToForecastArrivalListener;
	private ForecastTechnique myForecastTechnique;
		
	private Map <DItem, ItemType> mapDItemToItemType;	
	private Map <DItem, Inventory> mapDItemToInventory;
	private Map <BItem, Inventory> mapBItemToInventory;
	private Map <BItem, InventoryRepairingHoldingPoint> mapBItemToIRHP;	
	
	private AggregateTimeWeightedVariable myTotalAmtBackOrdered;
	private AggregateTimeWeightedVariable myTotalNumBackOrdered;
	private AggregateTimeWeightedVariable myTotalOnHand;
	private AggregateTimeWeightedVariable myTotalOnOrder;
	private AggregateTimeWeightedVariable myTotalStockOutIndicator;
	private ResponseVariableAverageObserver myAvgFirstFillRate;
	private ResponseVariableAverageObserver myAvgCustomerWaitTime;
	private AggregateCounter myTotalNumOfReplenishments;
	
	private boolean myAllowExternalDemandGeneratorsFlag = false;
	
	public SimVMModelOriginal(ModelElement parent) {
		this(parent, null);
	}
	
	public SimVMModelOriginal(ModelElement parent, String name) {
		super(parent, name);
				
		myBases = new HashSet<InventoryRepairingHoldingPoint>();
		myBaseIDs = new HashSet<String>();
		myItems = new HashSet<ItemType>();
		myItemIDs = new HashSet<String>();
		myDemandGenerators = new HashMap<InventoryRepairingHoldingPoint, List<DemandGenerator>>();
		
		mapDItemToItemType = new HashMap<DItem, ItemType>();
		mapDItemToInventory = new HashMap<DItem, Inventory>();
		mapBItemToInventory = new HashMap<BItem, Inventory>();	
		mapInventoryToDemandArrivalListener = new HashMap<Inventory, DemandArrivalListenerWithForecaster>();
		mapInventoryToForecastArrivalListener = new HashMap<Inventory, ForecastArrivalListenerDefault>();
		mapBItemToIRHP = new HashMap<BItem, InventoryRepairingHoldingPoint>();
				
		myTotalOnHand = new AggregateTimeWeightedVariable(this, getName() + " Total On Hand");
		myAvgFirstFillRate = new ResponseVariableAverageObserver(this, getName() + " Aggregate Fill Rate");
		myTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, getName() + " Aggregate Number of Stockouts");
		myTotalOnOrder = new AggregateTimeWeightedVariable(this, getName() + " Total On Order");
		myTotalNumOfReplenishments = new AggregateCounter(this, getName() + " NumberOfReplenishments");
		myTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Amt BackOrdered");
		myTotalNumBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Number BackOrdered");
		myAvgCustomerWaitTime = new ResponseVariableAverageObserver(this, getName() + "Avg Customer Wait");
		
		//create external LeadTimeDemandFiller and its own TimeBasedDemandShipper		
		myExternalSupplier = new LeadTimeDemandFiller(this, getName() + "ExternalSupplier");
		myExternalSupplierCarrier = new TimeBasedTypeLocationIndependentDemandCarrier(this, getName() + "ExternalShipper");
		myExternalSupplier.setDemandCarrier(myExternalSupplierCarrier);		
		myDepot = new InventoryRepairingHoldingPoint(this, getName() + "Depot");
		myDepotCarrier = new TimeBasedTypeLocationIndependentDemandCarrier(this, getName() + "DepotCarrier");
		myDepot.setDemandCarrier(myDepotCarrier);		
		myExternalSupplierCarrier.setImmediateTransportFlag(true);
		myDepot.setDemandFiller(myExternalSupplier);

	}
	
	public static SimVMModelOriginal getSimVMModel(ModelElement parent, String name, NetworkConfiguration nc, double period){
		
		SimVMModelOriginal svmmodel = new SimVMModelOriginal(parent, name);		

		svmmodel.myForecastTechnique = new Croston();//TODO
		svmmodel.addItemTypes(nc);
		svmmodel.addBases(nc);
		svmmodel.setInventoryAtEachLocation(nc, period);
		
		return svmmodel;
	}
	

	
	public static SimVMModelOriginal getSimVMModel(ModelElement parent, String name, NetworkConfiguration nc, 
			ForecastTechnique ft, double period){
		
		SimVMModelOriginal svmmodel = new SimVMModelOriginal(parent, name);
		
		svmmodel.myForecastTechnique = ft;
		svmmodel.addItemTypes(nc);
		svmmodel.addBases(nc);
		svmmodel.setInventoryAtEachLocation(nc, period);
		
		return svmmodel;
	}
	
	private void addItemTypes(NetworkConfiguration nc){
		for(DItem di:nc.myItems){
			int id = di.myItemID;
			String sid = "Type-"+id;			
			ItemType type = addItemType(sid);			
			mapDItemToItemType.put(di, type);			
		}
	}
	
	private void addBases(NetworkConfiguration nc){
		for(DItem di:nc.myItems){
			for(BItem bi:di.myBItems){
				int id = bi.myBaseID;
				String bid = "Base-"+id;
				if(!myBaseIDs.contains(bid)){
					InventoryRepairingHoldingPoint base = addBase(bid);
					mapBItemToIRHP.put(bi, base);
				}				
			}			
		}
	}
		
	private void setInventoryAtEachLocation(NetworkConfiguration nc, double myUpdatePeriod) {
		//loop through the DItems and add the depot inventory
		for(DItem di: nc.myItems){
			int locationid = 0;// location id
	        String name = di.myName;// item name
	        int itemid = di.myItemID;// item id
	        double cost = di.myCost;// item cost
	        int costlevel = di.myCostLevel;// item cost level
	        int basestocklevel = di.myBaseStockLevel; // item stock level
	        double lambda = di.myLambda;// average demand
	        double repairdelay = di.myRepairDelay;// repair delay
	        double reparability = 1.0;// reparability
	        double shipdelay = 0.0;// ship delay
	        double dibudgetnw = di.getDItemBudget();// item budget at location
	        double dibudgetloc = di.getBudget();// total item level budget
	        int numbasesstocked = di.myNumStockedBases;// num stocked level
	        
	        ItemType type = this.mapDItemToItemType.get(di);	        
	        Inventory inv = addDepotInventory(type, basestocklevel, new Exponential(repairdelay), new Constant(shipdelay), cost);
	        mapDItemToInventory.put(di, inv);	        
	        
	        for(BItem bi: di.myBItems){
				int locationid_ = 0;// location id
		        String name_ = bi.myName;// item name
		        int itemid_ = bi.myItemID;// item id
		        double cost_ = bi.myCost;// item cost
		        int costlevel_ = bi.myCostLevel;// item cost level
		        int basestocklevel_ = bi.myBaseStockLevel; // item stock level
		        double lambda_ = bi.myLambda;// average demand
		        double repairdelay_ = bi.myRepairDelay;// repair delay
		        double reparability_ = bi.myReparability;// reparability
		        double shipdelay_ = bi.myShipDelay;// ship delay
		        double bibudgetnw_ = bi.getBudget();// item budget at location
		        double bibudgetloc_ = di.getBudget();// total item level budget
                int plevel = bi.myPatternLevel;
                double tp = bi.myHorizonSpan;
                double td = bi.myDeploymentSpan;
                double ts = bi.mySustainmentSpan;
                double tr = bi.myReconfigurationSpan;
                double lMIN = bi.myLambdaMin;
                double lMAX = bi.myLambdaMax;
		        int numbasesstocked_ = 0;// num stocked level
		        
		        InventoryRepairingHoldingPoint base = this.mapBItemToIRHP.get(bi);	
		        
		        PiecewiseRateFunction f = new PiecewiseLinearRateFunction(lMIN, td, lMAX);
		        f.addRateSegment(ts,lMAX);
				f.addRateSegment(tr,lMIN);		
				NHPPTimeBtwEventRV myNHPPTimeBtwEventRV = new NHPPTimeBtwEventRV(this, f);
				
				ForecastTechnique ft = myForecastTechnique;
				ft.setName("FT"+base.getName()+"-"+type.getName());
				
				ForecastManager fm = new ForecastManager(base, ft, "-FM-"+ft.getName());		        
		        Inventory inv_ = addBaseInventory(base, type, new Bernoulli(reparability_), 
		        		new Exponential(repairdelay_), new Exponential(shipdelay_), 
		        		myNHPPTimeBtwEventRV, lMIN, myUpdatePeriod, cost_, 
		        		fm, basestocklevel_);
		        
		        mapBItemToInventory.put(bi, inv_);			
			}			
		}	
		//loop through the BItems and add the base inventory	
	}
	
	public final ItemType addItemType(String name){
		if(myItemIDs.contains(name))
			throw new IllegalArgumentException("The item has been already added");		
		myItemIDs.add(name);
		
		ItemType type = new ItemType(this, name);
		myItems.add(type);
		
		return(type);
	}
	
	/** Adds an ItemType and its lead time distribution to the network. The name
	 *  of the item type must be unique to the model
	 * 
	 * @param name
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(String name, RandomIfc leadtime){
		if(myItemIDs.contains(name))
			throw new IllegalArgumentException("The item has been already added");		
		myItemIDs.add(name);
		
		ItemType type = new ItemType(this, name);
		myItems.add(type);
		myExternalSupplier.addLeadTime(type, leadtime);
		
		return(type);
	}
	
	/** Causes an InventoryRepairingHoldingPoint to be created with a uniquely assigned name
	 *  and added to the network.  This only adds the IHRP to the network it does
	 *  not cause the IHRP to be connected to any other IHRPs in the network. The client
	 *  can use the returned reference to the IHRP to connect to other IHRPs.
	 * 
	 * @return the created IHRP
	 */
	public final InventoryRepairingHoldingPoint addBase(){
		return(addBase(null));
	}
	
	/** Causes an InventoryRepairingHoldingPoint to be created with the supplied name
	 *  and added to the network.  This only adds the IHRP to the network it does
	 *  not cause the IHRP to be connected to any other IHRPs in the network. The client
	 *  can use the returned reference to the IHRP to connect to other IHRPs.
	 * 
	 * @param name must be unique in the Model
	 * @return the created IHRP
	 */
	public final InventoryRepairingHoldingPoint addBase(String name){
		if(myBaseIDs.contains(name))
			throw new IllegalArgumentException("The base has been already added");
		
		myBaseIDs.add(name);	
		
		//create the ihrp and add it to the network so that it may have child ihrps
		InventoryRepairingHoldingPoint ihrp = new InventoryRepairingHoldingPoint(this, name);
		myBases.add(ihrp);
		
		// assign the ihp its own time based shipper for shipping to its customers
		TimeBasedTypeLocationIndependentDemandCarrier shipper = new TimeBasedTypeLocationIndependentDemandCarrier(this, ihrp.getName() + "TimeBasedGenericCarrier");
		ihrp.setDemandCarrier(shipper);
		if (myAllowExternalDemandGeneratorsFlag == true){
			shipper.setImmediateTransportFlag(true);
		}

		ihrp.subscribeToAggregateOnHandInventory(myTotalOnHand);
		ihrp.subscribeToAggregateAmountOnOrder(myTotalOnOrder);
		ihrp.subscribeToAggregateStockOutIndicator(myTotalStockOutIndicator);
		ihrp.subscribeToAggregateAvgFirstFillRate(myAvgFirstFillRate);
		ihrp.subscribeToAggregateReplenishmentDemandCounter(myTotalNumOfReplenishments);		
		ihrp.subscribeToAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		ihrp.subscribeToAggregateNumberBackOrdered(myTotalNumBackOrdered);
		ihrp.subscribeToAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);			
		
		attachBaseToDepot(ihrp);
		myBases.add(ihrp);
		
		return(ihrp);
	}
	
	public Map<BItem, InventoryRepairingHoldingPoint> getMapBItemToIRHP() {
		return mapBItemToIRHP;
	}

	public final Inventory addDepotInventory(ItemType type, 
			int basestock, DistributionIfc repairdelay, DistributionIfc transportTime, double vmcost){
		Inventory inv = myDepot.addReorderPointReorderQuantityInventory(type,basestock-1,1,basestock-1,
				"Inv-"+myDepot.getName()+"-"+type);
		myDepot.setRepairStation(type, new Bernoulli(1.0), repairdelay);
		setShippingDelayFromExternalSupplierToDepot(type, transportTime);
		
		return inv;
	}
	
	private final void setShippingDelayFromExternalSupplierToDepot(ItemType type, DistributionIfc transportTime){				
		Demand d = new Demand(type);
		// assign the depot correctly
		if(!myDepot.canFillItemType(d))
			throw new IllegalArgumentException("The supplied depot doesnt have the itemtype inventory");		
		if (transportTime != null){
			TimeBasedTypeLocationIndependentDemandCarrier c = (TimeBasedTypeLocationIndependentDemandCarrier) getExternalSupplier().getDemandCarrier();
			c.setTransportTime(myDepot, type, transportTime);
		}else{
			TimeBasedTypeLocationIndependentDemandCarrier c = (TimeBasedTypeLocationIndependentDemandCarrier) getExternalSupplier().getDemandCarrier();
			c.setImmediateTransportFlag(true);
		}
	} 

	public final Inventory addBaseInventory(InventoryRepairingHoldingPoint base, ItemType type, 
		DistributionIfc repairprob, DistributionIfc repairdelay, DistributionIfc transportTime, 
		RandomIfc demandDist, double initialdemandrate, double agglevel, double vmbcost, 
		ForecastManager forecaster, int basestocklevel){
	
		Inventory inv = base.addReorderPointReorderQuantityInventory(type,basestocklevel-1,1,basestocklevel-1,"Inv-"+base.getName()+"-"+type);
		base.setRepairStation(type, repairprob, repairdelay);
		setShippingDelayFromDepotToBase(base, type, transportTime); 
		
		DemandGenerator dg = addDemandGeneratorToIHRP(type, demandDist);                
		attachDemandGeneratorToIHRP(base, dg, new Constant(0.0));
		
		DemandArrivalListenerWithForecaster myDemandArrivalListener;
		
		if(new Double(agglevel).isInfinite()){
			myDemandArrivalListener = new DemandArrivalListenerWithForecaster(this, inv, forecaster,
					inv.getName() + " DemandObserver", agglevel, true);
		}else{
			myDemandArrivalListener = new DemandArrivalListenerWithForecaster(this, inv, forecaster,
					inv.getName() + " DemandObserver", agglevel, false);
		}
		
		ForecastArrivalListenerDefault myForecastArrivalListener = new ForecastArrivalListenerDefault(forecaster, initialdemandrate, 
				forecaster+" ForecastArrivalListener");	

		mapInventoryToDemandArrivalListener.put(inv,myDemandArrivalListener);
		mapInventoryToForecastArrivalListener.put(inv,myForecastArrivalListener);
		
		return inv;
	}

	/** Adds a shipping delay from depot and the base
	 * 
	 * @param base
	 * @param type
	 * @param transportTime
	 */
	private final void setShippingDelayFromDepotToBase(InventoryRepairingHoldingPoint base, ItemType type, DistributionIfc transportTime){
		// assign the list of bases correctly and then correct below
		if (!getBases().contains(base))
			throw new IllegalArgumentException("The provided base has not be added to the network");		
		if (transportTime != null){
			TimeBasedTypeLocationIndependentDemandCarrier c = (TimeBasedTypeLocationIndependentDemandCarrier) myDepot.getDemandCarrier();
			c.setTransportTime(base, type, transportTime);
		}else{
			TimeBasedTypeLocationIndependentDemandCarrier c = (TimeBasedTypeLocationIndependentDemandCarrier) myDepot.getDemandCarrier();
			c.setImmediateTransportFlag(true);
		}		
	}

	public final boolean isDepot(InventoryRepairingHoldingPoint ihrp){
		return(ihrp==myDepot);
	}
	
	public final boolean isBase(InventoryRepairingHoldingPoint ihrp){
		return(myBases.contains(ihrp));
	}
	
	public final void attachBaseToDepot(InventoryRepairingHoldingPoint customer){
		attachBaseToDepot(customer, null);
	}

	public final void attachBaseToDepot(InventoryRepairingHoldingPoint customer, 
			DistributionIfc transportTime){
		
		if (myBases.contains(customer))
			throw new IllegalArgumentException("The supplied customer IHP is already a customer IHP");
		
		myBases.add(customer);
		// ensure that it uses the supplier for its replenishments
		customer.setDemandFiller(myDepot);
	}

	public DemandArrivalListenerWithForecaster getDemandArrivalListener(Inventory inv) {
		return mapInventoryToDemandArrivalListener.get(inv);
	}

	public ForecastArrivalListenerDefault getForecastArrivalListener(Inventory inv) {
		return mapInventoryToForecastArrivalListener.get(inv);
	}

	public Map<DItem, ItemType> getMapDItemToItemType() {
		return mapDItemToItemType;
	}

	public Map<DItem, Inventory> getMapDItemToInventory() {
		return mapDItemToInventory;
	}

	public final DemandGenerator addDemandGeneratorToIHRP(ItemType type, RandomIfc timeUntilNext){
		return(addDemandGeneratorToIHRP(type, timeUntilNext, null));
	}
	
	public Map<BItem, Inventory> getMapBItemToInventory() {
		return mapBItemToInventory;
	}

	public final DemandGenerator addDemandGeneratorToIHRP(ItemType type, RandomIfc timeUntilNext, String name){
		DemandGenerator dg = new DemandGenerator(this, type, timeUntilNext, name);
		return(dg);
	}
	
	public final void attachDemandGeneratorToIHRP(InventoryRepairingHoldingPoint supplier, DemandGenerator generator){
		attachDemandGeneratorToIHRP(supplier, generator, null);
	}

	public final void attachDemandGeneratorToIHRP(InventoryRepairingHoldingPoint supplier, DemandGenerator generator,
			DistributionIfc transportTime){
		if (!myBases.contains(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		List<DemandGenerator> generators = myDemandGenerators.get(supplier);
		if (generators == null){
			generators = new ArrayList<DemandGenerator>();
			myDemandGenerators.put(supplier, generators);
		}
		if (generators.contains(generator))
			throw new IllegalArgumentException("The provided generator is already attached to the supplier");
		
		generators.add(generator);
		generator.setDemandFiller(supplier);
		if (transportTime != null){
			// ensure that it uses its supplier's shipper
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)supplier.getDemandCarrier();
			shipper.setTransportTime(generator, generator.getItemType(), transportTime);
		} else {
			// directing an attached IHP to be shipped with no delay time
			// designate the time based shipper to allow no delay shipment
			TimeBasedTypeLocationIndependentDemandCarrier shipper = (TimeBasedTypeLocationIndependentDemandCarrier)supplier.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}		
	}
	
	public final void allowExternalDemandGenerators(){
		myAllowExternalDemandGeneratorsFlag = true;
		for (InventoryRepairingHoldingPoint ihp: myBases){
			TimeBasedDemandCarrier shipper = (TimeBasedDemandCarrier)ihp.getDemandCarrier();
			shipper.setImmediateTransportFlag(true);
		}
	}
	
	public final Set<InventoryRepairingHoldingPoint> getBases(){
		return myBases;
	}

	public LeadTimeDemandFiller getExternalSupplier() {
		return myExternalSupplier;
	}

	public InventoryRepairingHoldingPoint getDepot() {
		return myDepot;
	}
	
	
}
