/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import java.util.Observer;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.ResponseStatisticsIfc;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.TimeWeighted;

/** An abstract base class for handling backlogging of demands for the Inventory class
 * 
 *
 */
public abstract class BackLogPolicyAbstract extends ModelElement implements BackLogInfoIfc {

	/** Tracks the amount of individual inventory units
	 *  required to fill back ordered demands
	 * 
	 */
	protected TimeWeighted myAmtBackLogged;
		
	/** Refers to the inventory that this policy is
	 *  controlling
	 * 
	 */
	protected Inventory myInventory;
        
	/** Creates a BackLogPolicy
	 * @param parent, the policy's parent model element
	 */
	public BackLogPolicyAbstract(Inventory inventory) {
		this(inventory, null);
	}
	
	/** Creates a BackLogPolicy
	 * @param parent the policy's parent model element
	 * @param name the policy's name
	 */
	public BackLogPolicyAbstract(Inventory inventory, String name) {
		super(inventory, name);
		inventory.setBackLogPolicy(this);
		myAmtBackLogged = new TimeWeighted(this, getName() + "-AmtBackLogged");
	}

	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================


	/* (non-Javadoc)
	 * @see jsl.supplychain.inventorylayer.BackLogInfoIfc#getAmountBackLogged()
	 */
	public final int getAmountBackLogged() {
		return (int)myAmtBackLogged.getValue();
	}
	
	/** Allows an observer to attach to the amount backlogged
	 * 
	 * @param observer
	 */
	public final void addAmountBackLoggedObserver(Observer observer){
		myAmtBackLogged.addObserver(observer);
	}
	
	/** Attaches an AggregateTimeWeightedVariable to the amount backlogged.  Whenever the
	 *  amount backlogged changes, the variable is updated
	 * 
	 * @param variable
	 */
	public final void attachAggregateTimeWeightedVariableToAmountBackLogged(AggregateTimeWeightedVariable variable){
		variable.subscribeTo(myAmtBackLogged);
	}

	/** Allows an observer to be removed from the amount backlogged
	 * 
	 * @param observer
	 */
	public final void removeAmountBackLoggedObserver(Observer observer){
		myAmtBackLogged.deleteObserver(observer);
	}
	
	/** Detaches an AggregateTimeWeightedVariable from the amount backlogged.  
	 * 
	 * @param variable
	 */
	public final void detachAggregateTimeWeightedVariableFromAmountBackLogged(AggregateTimeWeightedVariable variable){
		variable.unsubscribeFrom(myAmtBackLogged);
	}

	/** Allows an observer to attach to the waiting time in the backlog queue
	 * 
	 * @param observer
	 */
	abstract public void addBackLoggedQueueTimeObserver(Observer observer);

	/** Allows an observer to be removed from the waiting time in the backlog queue
	 * 
	 * @param observer
	 */
	abstract public void removeBackLoggedQueueTimeObserver(Observer observer);

	/** Allows an observer to attach to the number in the backlog queue
	 * 
	 * @param observer
	 */
	abstract public void addNumInBackLoggedQueueObserver(Observer observer);

	/** Allows an observer to be removed from the number in the backlog queue
	 * 
	 * @param observer
	 */
	abstract public void removeNumInBackLoggedQueueObserver(Observer observer);

	/** Attaches an AggregateTimeWeightedVariable to the number of demands in backlog queue
	 * 
	 * @param variable
	 */
	abstract public void attachAggregateTimeWeightedVariableToNumInBackLoggedQueue(AggregateTimeWeightedVariable variable);

	/** Detaches an AggregateTimeWeightedVariable from the number of demands in backlog queue
	 * 
	 * @param variable
	 */
	abstract public void detachAggregateTimeWeightedVariableFromNumInBackLoggedQueue(AggregateTimeWeightedVariable variable);

	/** Attaches an ResponseVariableAverageObserver to the time spent in queue
	 * 
	 * @param variable
	 */
	abstract public void attachResponseVariableAverageObserverToTimeInQueue(ResponseVariableAverageObserver variable);

	/** Detaches an ResponseVariableAverageObserver from the time spent in queue
	 * 
	 * @param variable
	 */
	abstract public void detachResponseVariableAverageObserverFromTimeInQueue(ResponseVariableAverageObserver variable);

	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================
		
	/** Asks the backlogger to backlog the demand
	 * 
	 * @param demand Must not be null
	 */
	abstract protected void backlog(Demand demand);
	
	/** Asks the backlogger to try to fill
	 *  backlog demands
	 */
	abstract protected void fillBackLogs();
		
	/** Returns the Inventory that is using this BackLogPolicy
	 * 
	 * @return Returns the Inventory
	 */
	protected final Inventory getInventory() {
		return myInventory;
	}
	
	protected final int allocateInventory(Demand demand){
		return(myInventory.allocateInventory(demand));
	}
	
	protected final void fillBackLog(Demand demand, int amount){
		myAmtBackLogged.decrement(amount);
		myInventory.fillBackLog(demand, amount);
	}
	
	protected void removedFromModel(){
		super.removedFromModel();
		myAmtBackLogged = null;
		myInventory = null;
	}

	//  ===========================================
	//      PRIVATE METHODS
	//  ===========================================

	/**
	 * @param inventory The inventory to set.
	 */
	final void setInventory(Inventory inventory) {
        if (inventory == null)
            throw new IllegalArgumentException("The Inventory must be non-null");
        myInventory = inventory;
	}

	/** Get the response statistics for the amount of back logged demand
	 * 
	 * @return the AmtBackLogged
	 */
	public final ResponseStatisticsIfc getAmtBackLoggedResponseStatistics() {
		return myAmtBackLogged;
	}

	//  ===========================================
	//      PRIVATE METHODS
	//  ===========================================


}
