/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import java.util.Observer;

import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.queue.QueueDiscipline;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.ResponseStatisticsIfc;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;

/** A concrete implementation of a backlog policy which uses a queue for
 *  managing the backlog
 *  
 *
 */
public class BackLogQueue extends BackLogPolicyAbstract {

	/** Holds the back ordered demand
	 * 
	 */
	protected Queue myBackLogQ;

	/** Creates a backlog queue for processing backlogs with a FIFO discipline.
	 * @param parent
	 */
	public BackLogQueue(Inventory inventory) {
		this(inventory, Queue.FIFO, null);
	}

	/** Creates a backlog queue for processing backlogs
	 * 
	 * @param parent
	 * @param discipline
	 */
	public BackLogQueue(Inventory inventory, QueueDiscipline discipline) {
		this(inventory, discipline, null);
	}

	/**
	 * @param parent
	 * @param discipline
	 * @param name
	 */
	public BackLogQueue(Inventory inventory, String name) {
		this(inventory, Queue.FIFO, name);
	}
	
	/**
	 * @param parent
	 * @param discipline
	 * @param name
	 */
	public BackLogQueue(Inventory inventory, QueueDiscipline discipline, String name) {
		super(inventory, name);
		setBackLoggedDemandQueue(new Queue(this, getName() + "-DemandBackLogQ", discipline));
	}

	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================


	/* (non-Javadoc)
	 * @see jsl.supplychain.inventorylayer.BackLogInfoIfc#getNumberOfDemandsBackLogged()
	 */
	public final int getNumberOfDemandsBackLogged() {
		return myBackLogQ.size();
	}

	/** Allows an observer to attach to the waiting time in the backlog queue
	 * 
	 * @param observer
	 */
	public final void addBackLoggedQueueTimeObserver(Observer observer){
		myBackLogQ.addTimeInQueueObserver(observer);
	}

	/** Allows an observer to be removed from the waiting time in the backlog queue
	 * 
	 * @param observer
	 */
	public final void removeBackLoggedQueueTimeObserver(Observer observer){
		myBackLogQ.removeTimeInQueueObserver(observer);
	}

	/** Allows an observer to attach to the number in the backlog queue
	 * 
	 * @param observer
	 */
	public final void addNumInBackLoggedQueueObserver(Observer observer){
		myBackLogQ.addNumberInQueueObserver(observer);
	}

	/** Allows an observer to be removed from the number in the backlog queue
	 * 
	 * @param observer
	 */
	public final void removeNumInBackLoggedQueueObserver(Observer observer){
		myBackLogQ.removeNumberInQueueObserver(observer);
	}

	@Override
	public void attachAggregateTimeWeightedVariableToNumInBackLoggedQueue(AggregateTimeWeightedVariable variable) {
		myBackLogQ.subscribe(variable);
	}

	public void detachAggregateTimeWeightedVariableFromNumInBackLoggedQueue(AggregateTimeWeightedVariable variable) {
		myBackLogQ.unsubscribe(variable);
	}

	@Override
	public void attachResponseVariableAverageObserverToTimeInQueue(ResponseVariableAverageObserver variable) {
		myBackLogQ.subscribe(variable);
	}

	public void detachResponseVariableAverageObserverFromTimeInQueue(ResponseVariableAverageObserver variable) {
		myBackLogQ.unsubscribe(variable);
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	protected final void backlog(Demand demand) {	
		demand.backlog();
		myAmtBackLogged.increment(demand.getRemainingDemand());
		// just put the demand in the queue
		myBackLogQ.enqueue(demand);	
		//System.out.println(getTime() + "> demand " + demand.getId() + " was backlogged");
		//System.out.println(getInventory());
	}

	protected final void fillBackLogs(){
		//System.out.println(getTime() + " > starting to fill backlogs ");
		// process the back logs as long as there are backlogs
		// that can be filled
		boolean stopFlag = false;
		while(hasNextBackLog() && (stopFlag == false)){
			// get the next backlog to process
			Demand d = nextBackLog();
			// process the backlog, determine how much 
			// can be supplied to the demand by the inventory
			int canSupply = allocateInventory(d);
			if (canSupply > 0){ // something can be supplied
				if (canSupply == d.getRemainingDemand()){// amount to supply == amount needed
					// this demand will be filled in full
					// remove it from the queue
					removeNextBackLog();
				} else {
					// the selected demand could not 
					// be supplied in full, stop processing further demands
					stopFlag = true;										
				}
				
				// go ahead and let the selected backlogged demand have
				// all of what can be supplied
				fillBackLog(d, canSupply);
			} else // nothing could be supplied to the next demand, must stop processing
				stopFlag = true;
		}
		//System.out.println(getTime() + " > completed filling backlogs ");	
	}

	/**
	 * @return Returns the myBackLoggedDemands.
	 */
	protected final Queue getBackLoggedDemandQueue() {
		return myBackLogQ;
	}
	
	//  ===========================================
	//      PRIVATE METHODS
	//  ===========================================

	protected final boolean hasNextBackLog() {
		return !myBackLogQ.isEmpty();
	}

	protected final Demand nextBackLog() {
		return((Demand)myBackLogQ.peekNext());
	}

	protected final void removeNextBackLog() {
		myBackLogQ.removeNext();
	}

	/**
	 * @param queue The queue to set.
	 */
	private final void setBackLoggedDemandQueue(Queue queue) {
        if (queue == null)
            throw new IllegalArgumentException("The Queue must be non-null");
		myBackLogQ = queue;
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.queue.Queue#getNumInQResponseStatistics()
	 */
	public final ResponseStatisticsIfc getNumInBackLogQResponseStatistics() {
		return myBackLogQ.getNumInQResponseStatistics();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.queue.Queue#getTimeInQResponseStatistics()
	 */
	public final ResponseStatisticsIfc getTimeInBackLogQResponseStatistics() {
		return myBackLogQ.getTimeInQResponseStatistics();
	}

}
