/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import jsl.modeling.ModelElement;
import jsl.utilities.random.RandomIfc;

public class DefaultOrderGenerator extends OrderGenerator {

	/**
	 * 
	 * @param parent
	 */
	public DefaultOrderGenerator(ModelElement parent) {
		this(parent, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * 
	 * @param parent
	 * @param name
	 */
	public DefaultOrderGenerator(ModelElement parent, String name) {
		this(parent, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * 
	 * @param parent
	 * @param timeUntilNext
	 */
	public DefaultOrderGenerator(ModelElement parent, RandomIfc timeUntilNext) {
		this(parent, null, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}
	
	/**
	 * 
	 * @param parent
	 * @param timeUntilNext
	 * @param name
	 */
	public DefaultOrderGenerator(ModelElement parent, RandomIfc timeUntilNext, String name) {
		this(parent, null, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * 
	 * @param parent
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 */
	public DefaultOrderGenerator(ModelElement parent, RandomIfc timeUntilFirst, RandomIfc timeUntilNext) {
		this(parent, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}
	
	/**
	 * 
	 * @param parent
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param name
	 */
	public DefaultOrderGenerator(ModelElement parent, RandomIfc timeUntilFirst, 
			RandomIfc timeUntilNext, String name) {
		this(parent, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * 
	 * @param parent
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 * @param name
	 */
	public DefaultOrderGenerator(ModelElement parent, RandomIfc timeUntilFirst, 
			RandomIfc timeUntilNext, long maxNum, String name) {
		this(parent, timeUntilFirst, timeUntilNext, maxNum, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * 
	 * @param parent
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 * @param timeUntilLast
	 * @param name
	 */
	public DefaultOrderGenerator(ModelElement parent, RandomIfc timeUntilFirst, 
			RandomIfc timeUntilNext, long maxNum, double timeUntilLast, String name) {
		super(parent, null, timeUntilFirst,timeUntilNext, maxNum, timeUntilLast, name);
		setOrderCreator(new RandomOrderCreator(this));
	}

	/** Adds the ItemType to the order creator with default amount of 1 unit
	 *  and probability of 1.0 that it will be included on the order
	 * 
	 * @param type The type of item
	 */
	public void addItemTypeDistribution(ItemType type){
		addItemTypeDistribution(type, 1.0, RandomOrderCreator.ONE);
	}
	
	/** Adds the ItemType to the order creator with default amount of 1 unit
	 * 
	 * @param type The type of item
	 * @param prob The probability that the item should be included on the order
	 */
	public void addItemTypeDistribution(ItemType type, double prob){
		addItemTypeDistribution(type, prob, RandomOrderCreator.ONE);
	}
	
	/** Adds the ItemType to the order creator
	 * 
	 * @param type The type of item
	 * @param prob The probability that the item should be included on the order
	 * @param distribution The distribution for the amount of demand if the item is included on the order. 
	 *  Must not produce a zero.  The amount must be >=1.0.  For non integral amounts, it will be truncated to
	 *  an integer
	 */
	public void addItemTypeDistribution(ItemType type, double prob, RandomIfc distribution){
		RandomOrderCreator oc = (RandomOrderCreator)myOrderCreator;
		oc.addItemTypeDistribution(type, prob, distribution);
	}
	
	/** Returns the order creator for this generator
	 * 
	 * @return
	 */
	public final RandomOrderCreator getOrderCreator(){
		return((RandomOrderCreator)myOrderCreator);
	}
}
