/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import jsl.modeling.IllegalStateException;
import jsl.modeling.elements.queue.QObject;
import jsl.utilities.reporting.JSL;
import jslx.supplychain.transportlayer.NoDelayDemandCarrier;

/** A Demand represents a request for units of inventory.  The initial 
 *  status of the demand is NO_STATUS.  The initial state of the demand
 *  is IN_PREPARATION
 *  
 *  //TODO explain the legal state transitions
 * 
 *
 */
public class Demand extends QObject {
	
	/** A instance of DemandShipperIfc that is used by demands by default when the
	 *  client does not attach a DemandListenerFilledIfc to handle what happens
	 *  after the demand is filled
	 */
	public static final NoDelayDemandCarrier DefaultNoDelayDemandShipper = new NoDelayDemandCarrier();
	
	/** Used to indicate that no status has been set for the demand
	 */
	public static final int NO_STATUS = JSL.getNextEnumConstant();

	/** Used to indicate that the demand will be filled in full, immediately
	 */
	public static final int IMMEDIATE_FILL = JSL.getNextEnumConstant();

	/** Used to indicate that the demand will be filled in full, immediately
	 */
	public static final int WILL_BE_BACKLOGGED = JSL.getNextEnumConstant();	

	/** Used to indicate the demand's status when rejecting or negotiating
	 *  Indicates that the filler was unavailable to 
	 */
	public static final int FILLER_UNAVAILABLE = JSL.getNextEnumConstant();

	/** Used to indicate the demand's status when rejecting or negotiating
	 *  Indicates that there was a mismatch between the item type of the demand
	 *  and the receiver's ability to process the item type.
	 */
	public static final int ITEMTYPE_MISMATCH = JSL.getNextEnumConstant();
	
	/** A non partial fill demand was sent to a partial filling receiver
	 */
	public static final int NPFD_SENT_TO_PFR = JSL.getNextEnumConstant();
	
	/** A non backloggable demand was sent to a backlogging receiver and
	 *  could not be immediately filled.
	 */
	public static final int NBLD_SENT_TO_BLR_NIF = JSL.getNextEnumConstant();

	/** A non backloggable demand was sent to a non backlogging receiver and
	 *  could not be immediately filled.
	 */
	public static final int NBLD_SENT_TO_NBLR_NIF = JSL.getNextEnumConstant();

	/** A backloggable demand was sent to a non backlogging receiver and
	 *  could not be immediately filled.
	 */
	public static final int BLD_SENT_TO_NBLR_NIF = JSL.getNextEnumConstant();

	/** A demand was placed on an Order and the order was rejected. When
	 *  the order that a demand is on is rejected, all the demands on the order
	 *  might be rejected (even though the status might have been non-rejection).
	 */
	public static final int ORDER_REJECTED = JSL.getNextEnumConstant();

	/** The demand is being prepared to be sent or negotiated
	 */
	public static final DemandState IN_PREPARATION = new DemandStatePreparing("IN_PREPARATION");

	/** The demand has been sent to a filler
	 */
	public static final DemandState SENT = new DemandStateSent("SENT");

	/** The demand is ready to be processed or negotiated
	 */
	public static final DemandState RECEIVED = new DemandStateReceived("RECEIVED");
	
	/** The demand is being filled
	 */
	public static final DemandState IN_PROCESS = new DemandStateInProcess("IN_PROCESS");

	/** The demand is in the backlog state
	 */
	public static final DemandState BACKLOGGED = new DemandStateBackLogged("BACKLOGGED");

	/** The demand was rejected after being received
	 */
	public static final DemandState REJECTED = new DemandStateRejected("REJECTED");
	
	/** The demand was cancelled after being in process
	 */
	public static final DemandState CANCELLED = new DemandStateCancelled("CANCELLED");
	
	/** The demand has been filled
	 */
	public static final DemandState FILLED = new DemandStateFilled("FILLED");

	/** The demand has been delivered
	 */
	public static final DemandState SHIPPED = new DemandStateShipped("SHIPPED");

	/** The demand has been delivered
	 */
	public static final DemandState DELIVERED = new DemandStateDelivered("DELIVERED");

	/** The demand is being negotiated
	 */
	public static final DemandState NEGOTIATING = new DemandStateNegotiating("NEGOTIATING");
	
	/** Used to indicate the demand's status when rejecting or negotiating
	 */
	private int myStatus;
	
	/** The object that has agreed to fill the demand
	 */
	private DemandFillerIfc myFiller;
	
	/** The object that actually received the demand
	 */
	private DemandFillerIfc myReceiver;
	
	/** The object that sent the demand to be filled
	 */
	private DemandSenderIfc mySender;
		
	/** Notified when the demand is recieved
	 */
	private DemandListenerReceivedIfc myReceivedListener;

	/** Notified when the demand is rejected
	 */
	private DemandListenerRejectedIfc myRejectedListener;

	/** Notified when the demand is processed
	 */
	private DemandListenerInProcessIfc myInProcessListener;
	
	/** Notified when the demand is cancelled
	 */
	private DemandListenerCancelledIfc myCancelledListener;
	
	/** Notified when the demand is delivered
	 */
	private DemandListenerDeliveredIfc myDeliveredListener;
	
	/** Notified when the demand is filled, this presents an opportunity to cause the
	 *  demand to be shipped
	 */
	private DemandListenerFilledIfc myFilledListener = DefaultNoDelayDemandShipper;
	
	/** Notified when the demand is told to ship
	 */
	private DemandListenerShippedIfc myShippedListener;
		
	/** The type of item required by the demand
	 */
	private ItemType myItemType;
	
	/** The amount of demand still needed
	 */
	private int myAmountNeeded;
	
	/** The amount of demand originally needed
	 */
	private int myOriginalAmtDemanded;
	
	/** The amount of the original demand that has been
	 *  filled.  When this equals the original amount
	 *  the demand has been filled.
	 */
	private int myAmountFilled;

	/** The order containing this demand
	 */
	private Order myOrder;
	
	/** The where this demand is in the order
	 */
	private int myOrderIndex;

	/** Marks when the demand was received
	 */
	private double myTimeReceived;
	
	/** Marks when the demand entered the backlog state
	 */
	private double myTimeEnteredBackLog;

	/** Marks when the demand entered the in process state
	 */
	private double myTimeEnteredInProcess;

	/** Marks when the demand completed filling
	 */
	private double myTimeFillingEnded;
	
	/** Marks the time that the demand was shipped
	 */
	private double myTimeShipped;

	/** Marks when the demand was delivered
	 */
	private double myTimeDelivered;

	/** Whether or not the demand is willing
	 *  to wait in a backlog (true means willing)
	 *  The default is true.
	 */
	private boolean myAllowBackLoggingFlag;
	
	/** Whether or not the demand requires all
	 *  original units demanded to be filled at one
	 *  time. (true means that it is allowed)
	 */
	private boolean myAllowPartialFillingFlag;

	/** Whether or not the order will allow itself to be cancelled
	 */
	private boolean myAllowCancellingFlag;

	/** The current state of the demand
	 */
	private DemandState myDemandState;

	/** Once the demand has been filled this flag is set
	 */
	private boolean myFilledFlag;

    /** Can be used to record the total amount of demand received
     *  to date for the filler of the demand when the demand
     *  arrived to be filled
     *
     */
    private long myTotalArrivedDemand = 0;
	
	/** Creates a demand, with defaults:
	 *   - Allow backlogging
	 *   - Allow partial filling
	 *   - amount demanded = 1
	 *   
	 * @param itemType
	 */
	public Demand(ItemType itemType) {
		this(itemType, 1, null);
	}

	/** Creates a demand, with defaults:
	 *   - Allow backlogging
	 *   - Allow partial filling
	 * @param itemType
	 * @param amtDemanded
	 */
	public Demand(ItemType itemType, int amtDemanded) {
		this(itemType, amtDemanded, null);
	}

	/** Creates a demand, with defaults:
	 *   - Allow backlogging
	 *   - Allow partial filling
	 *   
	 * @param itemType
	 * @param amtDemanded
	 * @param name
	 */
	public Demand(ItemType itemType, int amtDemanded, String name) {
		super(name);
		myStatus = NO_STATUS;
		myTimeReceived = Double.NaN;
		myTimeEnteredBackLog = Double.NaN;
		myTimeEnteredInProcess = Double.NaN;
		myTimeFillingEnded = Double.NaN;
		myTimeDelivered = Double.NaN;
		myTimeShipped = Double.NaN;
		myFilledFlag = false;
		myAllowBackLoggingFlag = true;
		myAllowPartialFillingFlag = true;
		setDemandState(IN_PREPARATION);
		setItemType(itemType);
		setOriginalAmountDemanded(amtDemanded);
	}

	/** Gets the item type for the demand
	 * @return the ItemType for this demand
	 */
	public final ItemType getItemType(){
		return(myItemType);
	}

	/** Return the amount that has been supplied
	 *  to the orginal demand amount.  When this reaches
	 *  the original amount demanded then the demand
	 *  is considered "filled"
	 *  
	 * @return Returns the amountFilled.
	 */
	public final int getAmountFilled() {
		return myAmountFilled;
	}

    /** Returns whatever was set for the total arrived
     *  demand
     *
     * @return
     */
    public final long getTotalArrivedDemand() {
        return myTotalArrivedDemand;
    }

    /** Sets the total arrived demand value
     *
     * @param amount
     */
    public final void setTotalArrivedDemand(long amount) {
        if (amount < 0)
            throw new IllegalArgumentException("The amount must be >= 0");
        
        myTotalArrivedDemand = amount;
    }
	
	/** Returns the weight of the amount supplied
	 *  to this demand
	 * 
	 * @return, the weight as a double
	 */
	public final double getWeight(){
		double w = 0.0;
		if (getItemType() != null)
			w = getItemType().getWeight();
		return(w*getAmountFilled());
	}
	
	/** Returns the cube of the amount supplied
	 *  to this demand
	 * 
	 * @return, the cube as a double
	 */
	public final double getCube(){
		double c = 0.0;
		if (getItemType() != null)
			c = getItemType().getCube();
		return(c*getAmountFilled());		
	}

	/** The line number of the demand if is in an order
	 * 
	 * @return an integer representing where the demand is in the order
	 */
	public final int getOrderIndex() {
		return myOrderIndex;
	}

	/** Returns true if the demand is part of an order
	 * 
	 * @return
	 */
	public final boolean isPartOfAnOrder(){
		return(myOrder != null);
	}
	
	/** Returns the order if the demand is part of an order or null
	 * 
	 * @return the order if the demand is part of an order or null
	 */
	public final Order getOrder() {
		return myOrder;
	}
	
	/** The original amount of demand requested.  Once
	 *  this quantity is supplied to the demand then the 
	 *  demand is considered "filled"
	 * 
	 * @return the amount remaining to be filled
	 */
	public final int getOriginalAmountDemanded() {
		return myOriginalAmtDemanded;
	}

	/** Gets the amount of demand still remaining to be filled
	 *  When this reaches zero the demand will be "filled"
	 *  
	 * @return the amount remaining to be filled
	 */
	public final int getRemainingDemand() {
		return myAmountNeeded;
	}
	
	/** Returns when the demand was received or Double.NaN if never received
	 * 
	 * @return when the demand was received or Double.NaN if never received
	 */
	public final double getTimeReceived(){
		return(myTimeReceived);
	}

	/** Returns the time that filling ended
	 * @return Returns the time that filling ended or Double.NaN if not yet filled
	 */
	public final double getTimeFillingEnded() {
		return myTimeFillingEnded;
	}

	/** Returns the time that the demand entered backlog
	 * @return Returns the time that the demand entered backlog or Double.NaN if never backlogged
	 */
	public final double getTimeEnteredBackLog() {
		return myTimeEnteredBackLog;
	}

	/** Returns when the demand started to be filled.  This assumes
	 *  that as soon as the demand is in process at a filler, the filling
	 *  begins
	 * 
	 * @return the time the demand began in process
	 */
	public final double getTimeEnteredInProcess(){
		return(myTimeEnteredInProcess);
	}

	/** The time that it took to fill the demand
	 * 
	 * @return the time to fille the demand
	 */
	public final double getFillingTime(){
		return(myTimeFillingEnded - myTimeEnteredInProcess);
	}

	/** The time that the demand was shipped
	 * 
	 * @return the time that the demand was shipped
	 */
	public final double getTimeShipped(){
		return(myTimeShipped);
	}
	
	/** The time that the demand was delivered
	 * 
	 * @return the time that the demand was delivered
	 */
	public final double getTimeDelivered(){
		return(myTimeDelivered);
	}
	
	/** True indicates that this demand can be backordered
	 * 
	 * @return Returns the allowBackLoggingFlag.
	 */
	public final boolean getAllowBackLoggingFlag() {
		return myAllowBackLoggingFlag;
	}
	
	/** True indicates that this demand can be backordered. This method can only
	 *  be called when the demand is in the READY state
	 * 
	 * @param allowBackLoggingFlag The allowBackLoggingFlag to set.
	 */
	public final void setAllowBackLoggingFlag(boolean allowBackLoggingFlag) {
		myDemandState.setAllowBackLoggingFlag(this, allowBackLoggingFlag);
	}

	/**True indicates that this demand allows partial filling
	 * 
	 * @return Returns the myAllowPartialFillingFlag.
	 */
	public final boolean getAllowPartialFillingFlag() {
		return myAllowPartialFillingFlag;
	}
	
	/** True indicates that this demand allows partial filling. This method can only
	 *  be called when the demand is in the IN_PREPARATION state
	 * @param allowPartialFillingFlag The allowPartialFillingFlag to set.
	 */
	public final void setAllowPartialFillingFlag(boolean allowPartialFillingFlag) {
		myDemandState.setAllowPartialFillingFlag(this, allowPartialFillingFlag);
	}

	/** Indicates if the demand is allowed to be cancelled.  This method can only
	 *  be called when the demand is in the IN_PREPARATION state
	 * 
	 * @param flag
	 */
	public final void setAllowCancellingFlag(boolean flag){
		myDemandState.setAllowCancellingFlag(this, flag);
	}
	
	/** Indicates if the demand can be cancelled
	 * 
	 * @return true means it can be cancelled
	 */
	public boolean isCancellingAllowed(){
		return(myAllowCancellingFlag);
	}

	/** Returns whether or not this demand has ever been filled
	 * 
	 * @return
	 */
	public final boolean isFilled(){
		return(myFilledFlag);
	}
		
	/** Checks if the demand is in the given state
	 * 
	 * @param state
	 * @return true if the demand is in the given state
	 */
	public final boolean isState(DemandState state){
		return(myDemandState == state);
	}
	
	/** Gets the filler associated with this demand
	 * 
	 * @return May be null
	 */
	public final DemandFillerIfc getFiller(){//TODO
		return(myFiller);
	}

	public final DemandFillerIfc getReceiver(){//TODO
		return(myReceiver);
	}
	
	/** Gets the DemandSenderIfc associated with the demand
	 * 
	 * @return
	 */
	public final DemandSenderIfc getSender(){
		return(mySender);
	}
	
	/** Sets the listener that will react when the demand is filled.  Setting this
	 *  listener only makes sense prior to the demand being filled
	 * 
	 * @param listener
	 */
	public final void setFilledListener(DemandListenerFilledIfc listener) {
		myFilledListener = listener;
	}

	/** Sets the listener that will react when the demand is shipped.  Setting this
	 *  listener only makes sense prior to the demand being shipped
	 * 
	 * @param listener
	 */
	public final void setShippedListener(DemandListenerShippedIfc listener) {
		myShippedListener = listener;
	}

	/** Sets the DemandSenderIfc for the demand.  This is only allowed when
	 *  the demand is in the IN_PREPARATION state
	 * 
	 * @param sender
	 */
	public final void setSender(DemandSenderIfc sender){
		myDemandState.setSender(this, sender);
	}
	
	/** Sets the integer status for the demand, see the integer status variables defined
	 *  as public final static for this class.  Setting the status is only allowd from the NEGOTIATING or SENT states
	 * 
	 * @param status
	 */
	public final void setDemandStatus(int status) {
		myDemandState.setStatus(this, status);
	}
	
	/** Returns an integer representing the status of the demand.
	 *   NO_STATUS 	 			Used to indicate that no status has been set for the demand
	 *   IMMEDIATE_FILL  		Used to indicate that the demand will be filled in full, immediately
	 *   WILL_BE_BACKLOGGED 	Used to indicate that the demand will be filled in full, immediately
	 *   FILLER_UNAVAILABLE 	Used to indicate the demand's status when rejecting or negotiating
	 *   						Indicates that the filler was unavailable to 
	 *   ITEMTYPE_MISMATCH  	Used to indicate the demand's status when rejecting or negotiating
	 *   						Indicates that there was a mismatch between the item type of the demand
	 *   						and the receiver's ability to process the item type.
	 *   NPFD_SENT_TO_PFR 		A non partial fill demand was sent to a partial filling receiver
	 *   NBLD_SENT_TO_BLR_NIF 	A non backloggable demand was sent to a backlogging receiver and
	 *   						could not be immediately filled.
	 *   NBLD_SENT_TO_NBLR_NIF 	A non backloggable demand was sent to a non backlogging receiver and
	 *   						could not be immediately filled
	 *   BLD_SENT_TO_NBLR_NIF 	A backloggable demand was sent to a non backlogging receiver and 
	 *   						could not be immediately filled.
	 *   ORDER_REJECTED 		A demand was placed on an Order and the order was rejected. When 
	 *   						the order that a demand is on is rejected, all the demands on the order 
	 *   						might be rejected (even though the status might have been non-rejection).
	 * @return
	 */
	public final int getStatus() {
		return myStatus;
	}
	
	public String toString(){
		StringBuilder sb = new StringBuilder();

		sb.append("Demand " + getId() + " AmtRequested=");
		sb.append(getOriginalAmountDemanded() + " AmtNeeded="+ getRemainingDemand());
		sb.append(" ItemType= " + getItemType().getName() + "\n");

		return(sb.toString());
	}
	
	/** Returns the string equivalent for getStatus()
	 * 
	 * @return
	 */
	public String getStatusAsString(){
		if (myStatus == NO_STATUS)
			return("NO_STATUS");
		if (myStatus == IMMEDIATE_FILL)
			return("IMMEDIATE_FILL");
		if (myStatus == WILL_BE_BACKLOGGED)
			return("WILL_BE_BACKLOGGED");
		if (myStatus == FILLER_UNAVAILABLE)
			return("FILLER_UNAVAILABLE");
		if (myStatus == ITEMTYPE_MISMATCH)
			return("ITEMTYPE_MISMATCH");
		if (myStatus == NPFD_SENT_TO_PFR)
			return("NPFD_SENT_TO_PFR");
		if (myStatus == NBLD_SENT_TO_BLR_NIF)
			return("NBLD_SENT_TO_BLR_NIF");
		if (myStatus == NBLD_SENT_TO_NBLR_NIF)
			return("NBLD_SENT_TO_NBLR_NIF");
		if (myStatus == BLD_SENT_TO_NBLR_NIF)
			return("BLD_SENT_TO_NBLR_NIF");
		return(null);
	}

	/** Sets the listener for when the demand is received at a filler.  Only
	 *  makes sense to set prior to being received
	 * 
	 * @param listener
	 */
	public final void setReceivedListener(DemandListenerReceivedIfc listener) {
		myReceivedListener = listener;
	}
	
	/** If there is a demand received listener it will be notified
	 */
	protected final void notifyDemandReceiptListener(){
		if (myReceivedListener != null)
			myReceivedListener.demandReceived(this);
	}
	
	/** Sets the listener that reacts to when the demand is rejected.
	 * 
	 * @param listener
	 */
	public final void setRejectedListener(DemandListenerRejectedIfc listener) {
		myRejectedListener = listener;
	}
		
	/** If there is a demand rejected listener, it will be notified
	 */
	protected final void notifyDemandRejectedListener(){
		if (myRejectedListener != null)
			myRejectedListener.demandRejected(this);
	}
	
	/** Sets the in process listener to be notified when the demand is placed in the
	 *  in process state
	 * 
	 * @param listener
	 */
	public final void setInProcessListener(DemandListenerInProcessIfc listener) {
		myInProcessListener = listener;
	}
	
	/** Sets the delivered listener to be notified when the demand is placed in the
	 *  in delivered state
	 * 
	 * @param listener
	 */
	public final void setDeliveredListener(DemandListenerDeliveredIfc listener) {
		myDeliveredListener = listener;
	}

	
	/** Sets the cancelled listener to be notified when the demand is placed in the
	 *  cancelled state
	 * 
	 * @param listener
	 */
	public final void setCancelledListener(DemandListenerCancelledIfc listener) {
		myCancelledListener = listener;
	}
	
	/** Sets the filler for this demand, can only be done while demand is
	 *  in the preparation state
	 * 
	 * @param filler
	 */
	public final void setFiller(DemandFillerIfc filler){
		myDemandState.setFiller(this, filler);
	}

	/* (non-Javadoc)
	 * @see jsl.modeling.elements.queue.QObject#nullify()
	 */
	public void nullify(){
		super.nullify();
		myFiller = null;
		myReceiver = null;
		mySender = null;
		myReceivedListener = null;
		myRejectedListener = null;
		myInProcessListener = null;
		myCancelledListener = null;
		myDeliveredListener = null;
		myFilledListener = null;
		myShippedListener = null;
		myItemType = null;
		myOrder = null;
		myDemandState = null;
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	protected void setDemandState(DemandState state){
//		if (myDemandState != null){
//			JSL.out.println("------------------------------------------------");
//			JSL.out.println("In Demand.setDemandState(DemandState state)");
//			JSL.out.println(getTime() + " > " + "Demand: " + getId() + " is transitioning from " + myDemandState.toString() + " to state " + state.toString());
//			JSL.out.println("------------------------------------------------");
//		}
		myDemandState = state;
		
		if (isState(Demand.IN_PREPARATION))
			return;
		
		if (isState(Demand.NEGOTIATING))
			return;

		if (isState(Demand.SENT))
			return;
		
		// leaves received, inprocess, backlogged, rejected, filled, shipped, cancelled, delivered
		if (isState(Demand.DELIVERED)){
			if (myDeliveredListener != null)
				myDeliveredListener.demandDelivered(this);
			return;
		}

		if (isState(Demand.BACKLOGGED)){
			if (myOrder != null)
				myOrder.demandBackLogged(this);
			return;
		}

		if (isState(Demand.FILLED)){
			if (myFilledListener != null)
				myFilledListener.demandFilled(this);
			return;
		}

		if (isState(Demand.RECEIVED)){
			// if the demand is not part of an order, try to use it's own DemandListenerReceivedIfc
			if (myOrder == null){
				if (myReceivedListener != null)
					myReceivedListener.demandReceived(this);
				return;				
			} else { // demand is part of an order, use the order to react to demand receipt
				myOrder.demandReceived(this);
			}
		}

		if (isState(Demand.REJECTED)){
			if (myOrder == null){
				if (myRejectedListener != null)
					myRejectedListener.demandRejected(this);
				return;				
			} else {
				myOrder.demandRejected(this);
			}
		}
		
		if (isState(Demand.IN_PROCESS)){
			if (myInProcessListener != null)
				myInProcessListener.demandInProcess(this);
			return;
		}

		if (isState(Demand.SHIPPED)){
			if (myShippedListener != null)
				myShippedListener.demandShipped(this);
			return;
		}
		
		if (isState(Demand.CANCELLED)){
			if (myCancelledListener != null)
				myCancelledListener.demandCancelled(this);
			return;
		}
	}
	
	/** Sets the order.  Used when the demand is added to an order
	 * @param order
	 */
	protected final void setOrder(Order order) {
		myOrder = order;
	}

	/** Indicates the index in the order
	 * @param i
	 */
	protected final void setOrderIndex(int i) {
		myOrderIndex = i;
	}

	/** Sets the item type for the demand
	 * @param the ItemType for this demand
	 */
	protected final void setItemType(ItemType itemType){
		if (itemType == null)
			throw new IllegalArgumentException("Invalid argument ItemType itemType was null");

		myItemType = itemType;
	}	
	
	/** Sets the original amount demanded.  The amount that the demand needed when
	 *  created
	 * @param amtDemanded, The amtDemanded to set, must be > 0
	 */
	protected final void setOriginalAmountDemanded(int amtDemanded) {
		if (amtDemanded <= 0)
			throw new IllegalArgumentException(
					"Invalid argument. Amount demanded must be > 0");

		myAmountNeeded = amtDemanded;
		myOriginalAmtDemanded = amtDemanded;
		myAmountFilled = 0;
	}

	protected final void setSender_(DemandSenderIfc sender){
		mySender = sender;
	}

	protected final void setFiller_(DemandFillerIfc filler){
		myFiller = filler;
	}

	protected final void setAllowPartialFillingFlag_(boolean allowPartialFillingFlag) {
		myAllowPartialFillingFlag = allowPartialFillingFlag;
	}

	protected final void setAllowCancellingFlag_(boolean flag){
		myAllowCancellingFlag = flag;
	}

	protected final void setAllowBackLoggingFlag_(boolean allowBackLoggingFlag) {
		myAllowBackLoggingFlag = allowBackLoggingFlag;
	}
	
	//  ===========================================
	//     PACKAGE SCOPE
	//  ===========================================

	final void setFillStateToFilled(){
		myTimeFillingEnded = Demand.getTime();
		myFilledFlag = true;
		setDemandState(FILLED);
	}
	
	final void setTimeReceived(){
		myTimeReceived = Demand.getTime();
	}
	
	final void setTimeEnteredInProcess(){
		if (Demand.getTime() != myTimeReceived)
			throw new IllegalStateException("A time delay occurred between receipt and attempting to place the demand in process");
		
		myTimeEnteredInProcess = Demand.getTime();
	}

	final void setTimeEnteredBackLog(){
		myTimeEnteredBackLog = Demand.getTime();
	}

	final void setTimeDelivered(){
		myTimeDelivered = Demand.getTime();
	}

	final void setTimeShipped(){
		myTimeShipped = Demand.getTime();
	}
	
	final void fill_(int amtSupplied) {
		if (amtSupplied < 0)
			throw new IllegalArgumentException(
					"Invalid argument. Amount filled must be >= 0");
		
		if (myAllowPartialFillingFlag == false){
			if (amtSupplied != myOriginalAmtDemanded)
				throw new IllegalArgumentException(
				"Invalid argument. Amount supplied must be equal to " +
				"the amount demanded if partial filling is not allowed");
		}
		
		int amtUsed = 0;
		if (amtSupplied > myAmountNeeded)
			amtUsed = myAmountNeeded;
		else
			amtUsed = amtSupplied;
		myAmountNeeded = myAmountNeeded - amtUsed;
		myAmountFilled = myAmountFilled + amtUsed;
		if (myAmountFilled == myOriginalAmtDemanded){
			setFillStateToFilled();				
		}

	}

	final void setStatus(int status){
		myStatus = status;
	}
	
	/** Sets the receiver for this demand
	 * 
	 * @param receiver
	 */
	final void setReceiver(DemandFillerIfc receiver){
		if (receiver == null)
			throw new IllegalArgumentException("The receiver was null.");
		myReceiver = receiver;
	}
	
	/**
	 * If the amount supplied is < 0 an illegal argument exception is thrown. If
	 * the amount supplied is > amount demanded then only the amount needed will
	 * be used from that supplied.  This method can only
	 *  be called when the demand is in the INPROCESS state
	 * 
	 * If the amount supplied is not equal to the amount originally demanded
	 * and partial order filling is not permitted then an illegal argument exception
	 * will be thrown.
	 *
	 * @param amtSupplied   The amount to supply to the demand.
	 */
	public final void fill(int amtSupplied) {
		myDemandState.fill(this, amtSupplied);
	}
	
	/** Places the demand in the SENT state, can only be called from the IN_PREPARATION state
	 */
	public final void sent(){
		myDemandState.sent(this);
	}

	/** Places the demand in the IN_PREPARATION state, can only be called
	 *  from the NEGOTIATING and REJECTED states
	 */
	public final void prepare(){
		myDemandState.prepare(this);
	}
	
	/** Places the demand in the RECEIVED state, can only be called
	 *  from the SENT state
	 *  @param receiver  The DemandFillerIfc that is receiving the demand
	 */
	public final void receive(DemandFillerIfc receiver){
		setReceiver(receiver);
		myDemandState.receive(this);
	}
	
	/** Places the demand in the IN_PROCESS state, can only be called
	 *  from the RECEIVED state
	 *  @param filler  The DemandFillerIfc that will fill the demand after it is in the IN_PROCESS state
	 *   this must be the same as the filler that received the demand
	 */
	public final void process(DemandFillerIfc filler){
		if (getReceiver() != filler)
			throw new IllegalStateException("The demand was received by a different filler");
		myDemandState.process(this);
	}
	
	/** Places the demand in the BACKLOGGED state, can only be called
	 *  from the IN_PROCESS state
	 */
	public final void backlog(){
		myDemandState.backlog(this);
	}
		
	/** Places the demand in the REJECTED state, can only be called
	 *  from the RECEIVED state
	 */
	public final void reject(){
		myDemandState.reject(this);	
	}
	
	/** Places the demand in the DELIVERED state, can only be called
	 *  from the SHIPPED state
	 */
	public final void deliver(){
		myDemandState.deliver(this);
	}
	
	/** Places the demand in the CANCELLED state, can only be called
	 *  from the RECEIVED and IN_PROCESS states if the demand is cancellable
	 */
	public final void cancel(){
		myDemandState.cancel(this);
	}
	
	/** Places the demand in the SHIPPED state, can only be called
	 *  from the FILLED state
	 */
	public final void ship(){
		myDemandState.ship(this);
	}
	
	/** Places the demand in the NEGOTIATING state, can only be called
	 *  from the IN_PREPARATION state
	 */
	public final void negotiate(){
		myDemandState.negotiate(this);
	}
	
}
