/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import java.util.Collection;

import jsl.utilities.IdentityIfc;
import jslx.supplychain.transportlayer.DemandCarrierIfc;

/** Objects that implement the DemandFillerIfc can fill demands
 *  The implication is that the receiver can or will handle the Demand in
 *  such a way that the demand will eventually be filled. 
 * 
 *
 */
public interface DemandFillerIfc extends IdentityIfc, AvailabilityIfc {
	
	/** This method allows the filler to receive Demands from clients. Once an
	 *  demand has been successfully received by a filler, it cannot be sent
	 *  to another filler. The demand should only be put in the received state when it
	 *  is ready to be processed by the filler. The sender of the demand is responsible
	 *  for setting the appropriate listeners on the demand.
	 *  
	 *  If the sent demand cannot be received by this DemandFiller
	 *  then the DemandFiller should reject the demand otherwise it
	 *  should receive the demand.
	 *  
	 *  The filler *may* indicate on the demand
	 *  what conditions caused the rejection; however, the filler does not promise
	 *  to do this.  It is up to the client to properly react to changes of state by the demand
	 *  after calling this method. 
	 *  
	 *  If the filler receives the demand, then the demand is placed in the received state. If the sender
	 *  attached a DemandListenerReceivedIfc to the demand then it will be notified and the sender
	 *  can decide to ask the filler to fill the demand.
	 *  
	 *  If the filler rejects the demand, then the demand is placed in the rejected state.  If the sender
	 *  attached a DemandListenerRejectedIfc to the demand then it will be notified of the rejection.  The sender
	 *  can then decide how to react to the rejection.
	 *  
	 *  An demand that does not permit backlogging can still be sent to a filler that
	 *  may backlog orders.  If the demand (that does not permit backlogging)
	 *  can be processed without backlogging then it should be received. If it cannot
	 *  be processed without backlogging, then it should be rejected.
	 *  
	 *  A backloggable demand can still be sent to a filler that does not permit backlogging.
	 *  If the backloggable demand can be processed without backlogging, then it should be;
	 *  otherwise it should be rejected.
	 *  
	 *  A demand that permits partial filling sent to a filler that does not partially fill is still valid.
	 *  The demand simply will not be partially filled and the client cannot expect it to be partially filled.
	 *  
	 *  An demand that is not allowed to be partially filled sent to a filler that partially fills demands 
	 *  can still be valid.  The filler may decide to reject the demand in this case.
	 *  If the filler accepts the demand, then the filler is responsible for not partially filling the demand.	
	 *  
	 * @param d
	 */
	public void receive(Demand demand);
	
	/** This method allows the filler to fill Demands from clients. Once a
	 *  demand has been successfully received by a filler, it can only be filled
	 *  by the filler that received it.  This method asks the filler to fill a demand
	 *  that has already been received by the filler.  A received demand must be placed in process
	 *  at the same simulation time that it was received.  In otherwords, no simulation time must elapse
	 *  between receipt and beginning the fill process.
	 *  
	 * @param demand
	 */
	public void fillDemand(Demand demand);
			
	/** Asks the filler to negotiate on the demand.  The returned message
	 *  should contain the results of the negotiation.  If the filler
	 *  cannot negotiate then this method should return null.
	 *  
	 *  Attempting to negotiate with an unavailable filler may or may not
	 *  return a null message. This depends upon the implementation of the client.
	 * 
	 * @param demand
	 * @return
	 */
	public DemandMessageIfc negotiate(Demand demand);
	
	/** Indicates whether or not the filler can fill the given ItemType of the Demand
	 * 
	 * @param demand
	 * @return
	 */
	public boolean canFillItemType(Demand demand);

	/** Indicates whether or not the filler can fill the given ItemType
	 * 
	 * @param type
	 * @return
	 */
	public boolean canFillItemType(ItemType type);
	
	/** Gets a list of the item types that can be filled by this filler
	 * 
	 * @return
	 */
	public Collection<ItemType> getItemTypes();

	/** Given a demand, this method will return an integer that represents
	 *  what will happened to the demand if it is requested to be filled at
	 *  the current time.  The status is only valid at the current simulation time.
	 * 
	 * @param demand
	 * @return
	 */
	public int determineRequestStatus(Demand demand);
	
	/** Given a demand, this method indicates whether or not the filler
	 *  will reject the demand if it is requested to be filled at the current time.
	 *  True means that it will be rejected, false means that it will be received w/o rejection
	 * @param demand
	 * @return
	 */
	public boolean willReject(Demand demand);

	/** Sets the carrier that this filler will use when sending demands
	 *  to customers of its inventory.
	 *  
	 * @param carrier
	 */
	public void setDemandCarrier(DemandCarrierIfc carrier);

	/** Gets the carrier for this filler
	 * 
	 * @return
	 */
	public DemandCarrierIfc getDemandCarrier();

}
