/*
 * Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
 *
 * Contact:
 *	Manuel D. Rossetti, Ph.D., P.E.
 *	Department of Industrial Engineering
 *	University of Arkansas
 *	4207 Bell Engineering Center
 *	Fayetteville, AR 72701
 *	Phone: (479) 575-6756
 *	Email: rossetti@uark.edu
 *	Web: www.uark.edu/~rossetti
 *
 * This file is part of the JSL (a Java Simulation Library). The JSL is a framework
 * of Java classes that permit the easy development and execution of discrete event
 * simulation programs.
 *
 * The JSL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * The JSL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the JSL (see file COPYING in the distribution);
 * if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA, or see www.fsf.org
 *
 */
package jslx.supplychain.inventorylayer;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.*;
import jsl.modeling.elements.EventGenerator;
import jsl.utilities.random.RandomIfc;

/** A DemandGenerator is an EventGenerator that creates and sends demands to its
 *  associated filler at each event action.  The client must specify the ItemType
 *  of the demands generated and my specify an distribution for the amount of
 *  demand of the ItemType to be requested from the DemandFiller.  The user can specify
 *  whether or not the demands allow partial filling and/or if they permit backlogging.  The
 *  user is responsible for ensuring that compatible demands are sent to appropriate demand
 *  fillers (or recognizing/handling any rejections that occur).
 *
 *  The user can specify whether or not the demand generator only sends
 *  unit demands.  If the unit demand only flag is on
 *  and the amount generated for the demand is > 1, say d, then
 *  d individual demands of 1 unit each will be sent. This permits
 *  the demands to be filled individually (as if like partial filling).
 *  The unit demand only flag is false by default.
 *
 *  The user MUST either directly set the demand filler for the generator via the
 *  setDemandFiller() method or provide an instance of a DemandFillerFinderIfc which
 *  will be responsible for finding an appropriate filler for the generator.  If no filler
 *  can be found upon the demand event creation, then a NoFillerFoundException will be thrown.
 *  
 *  The DemandGenerator uses an instance of DemandRejectionListener to handle the rejection
 *  of any demands.  The default behavior is to throw an exception when a demand is rejected. This
 *  behavior can be modified by supplying a class that implements DemandListenerRejectedIfc
 *  
 *
 */
public class DemandGenerator extends EventGenerator implements DemandSenderIfc {

    /** Indicates whether or not the demand generator only sends
     *  unit demands.  If the unit demand only flag is on
     *  an the amount generated for the demand is > 1, say d, then
     *  d individual demands of 1 unit each will be sent. This permits
     *  the demands to be filled individually (as if like partial filling)
     * 
     */
    protected boolean myUnitDemandOnlyFlag;

    /** Indicated whether or not the DemandGenerator will
     *  permit backlogging of its demand requests
     */
    protected boolean myPermitBackLoggingFlag;

    /** Indicated whether or not the DemandGenerator will
     *  permit partial filling of its demand requests
     */
    protected boolean myPermitPartialFillingFlag;

    /** Used to determine the amount of demand at each creation
     */
    protected RandomIfc myAmtDistribution;

    /** Used to determine the amount of demand at each creation
     */
    protected RandomVariable myAmtRV;

    /** Responsible for finding an appropriate filler for the demand generator
     */
    protected DemandFillerFinderIfc myDemandFillerFinder;

    /** This attribute can be used to directly set the filler for this demand generator
     */
    protected DemandFillerIfc myDemandFiller;

    /** used to identify the type of demand that is being generated
     */
    protected ItemType myItemType;

    /** Acts as a listener for when the inventory sends demand out for replenishment
     */
    protected DemandListener myDemandListener = new DemandListener();

    /** Acts as a listener for when the inventory sends demand out for replenishment
     */
    protected DemandListenerRejectedIfc myDemandRejectedListener = new DemandRejectionListener();

    /**
     * @param parent
     * @param demandFillerFinder an object that finds a DemandFillerIfc
     */
    public DemandGenerator(ModelElement parent, ItemType type) {
        this(parent, type, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
    }

    /**
     * @param parent
     * @param demandFillerFinder an object that finds a DemandFillerIfc
     * @param name
     */
    public DemandGenerator(ModelElement parent, ItemType type, String name) {
        this(parent, type, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
    }

    /**
     * @param parent
     * @param demandFillerFinder an object that finds a DemandFillerIfc
     * @param name
     * @param timeUntilFirst
     */
    public DemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilNext) {
        this(parent, type, null, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
    }

    /**
     * @param parent
     * @param demandFillerFinder an object that finds a DemandFillerIfc
     * @param name
     * @param timeUntilFirst
     */
    public DemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilNext, String name) {
        this(parent, type, null, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
    }

    /**
     * @param parent
     * @param name
     * @param timeUntilFirst
     * @param timeUntilNext
     */
    public DemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilFirst, RandomIfc timeUntilNext) {
        this(parent, type, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
    }

    /**
     * @param parent
     * @param name
     * @param timeUntilFirst
     * @param timeUntilNext
     */
    public DemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilFirst,
            RandomIfc timeUntilNext, String name) {
        this(parent, type, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
    }

    /**
     * @param parent
     * @param name
     * @param timeUntilFirst
     * @param timeUntilNext
     * @param maxNum
     */
    public DemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilFirst,
            RandomIfc timeUntilNext, long maxNum, String name) {
        this(parent, type, timeUntilFirst, timeUntilNext, maxNum, Double.POSITIVE_INFINITY, name);
    }

    /**
     * @param parent
     * @param name
     * @param timeUntilFirst
     * @param timeUntilNext
     * @param maxNum
     * @param timeUntilLast
     */
    public DemandGenerator(ModelElement parent, ItemType type, RandomIfc timeUntilFirst,
            RandomIfc timeUntilNext, long maxNum, double timeUntilLast, String name) {
        super(parent, null, timeUntilFirst, timeUntilNext, maxNum, timeUntilLast, name);
        
        myUnitDemandOnlyFlag = false;
        myPermitBackLoggingFlag = true;

        myPermitPartialFillingFlag = true;

        setItemType(type);

    }

    /**
     *  If the unit demand only flag is on
     *  an the amount generated for the demand is > 1, say d, then
     *  d individual demands of 1 unit each will be sent. This permits
     *  the demands to be filled individually (as if like partial filling)
     * @return
     */
    public boolean getUnitDemandOnlyFlag() {
        return myUnitDemandOnlyFlag;
    }

    /** If the unit demand only flag is on (true)
     *  an the amount generated for the demand is > 1, say d, then
     *  d individual demands of 1 unit each will be sent. This permits
     *  the demands to be filled individually (as if like partial filling)
     *
     * @param flag
     */
    public void setUnitDemandOnlyFlag(boolean flag) {
        myUnitDemandOnlyFlag = flag;
    }
    
    /**
     * @return Returns the distribution governing the amount of demand generated
     */
    public final RandomIfc getAmountDistribution() {
        return myAmtDistribution;
    }

    /**
     * @param myAmtDistribution The myAmtDistribution to set.
     */
    public final void setAmountDistributionInitialRandomSource(RandomIfc distribution) {
        if (distribution == null) {
            throw new IllegalArgumentException("Invalid argument. RandomIfc was null");
        }

        if (myAmtRV == null) {
            myAmtRV = new RandomVariable(this, distribution);
        } else {
            myAmtRV.setInitialRandomSource(distribution);
        }

        myAmtDistribution = distribution;
    }

    /** The type of item to generate
     *
     * @return Returns the itemType, may be null
     */
    public ItemType getItemType() {
        return myItemType;
    }

    /** The supplied finder is responsible for finding an appropriate filler
     *  so that the generator's demand requests can be fulfilled.
     *
     * @param finder Must not be null
     */
    public final void setDemandFillerFinder(DemandFillerFinderIfc finder) {
        if (finder == null) {
            throw new IllegalArgumentException("Invalid argument. DemandFillerFinder must be non-null.");
        }
        myDemandFillerFinder = finder;
    }

    /**  Allows access the generator's filler finder
     *
     * @return
     */
    public final DemandFillerFinderIfc getDemandFillerFinder() {
        return (myDemandFillerFinder);
    }

    /** Can be used to directly set the filler for this generator, rather than relying
     *  on the filler finder
     *
     * @param fillerFinder Must not be null
     */
    public final void setDemandFiller(DemandFillerIfc filler) {
        myDemandFiller = filler;
    }

    /**  Allows access the generator's filler
     *
     * @return
     */
    public final DemandFillerIfc getDemandFiller() {
        return (myDemandFiller);
    }

    /* (non-Javadoc)
     * @see jsl.supplychainNew.inventorylayer.DemandSenderIfc#mightRequest(jsl.supplychainNew.inventorylayer.ItemType)
     */
    public final boolean mightRequest(ItemType type) {
        return myItemType != type;
    }

    /**
     * @return
     */
    public final boolean mayRequestPartialFilling() {
        return myPermitPartialFillingFlag;
    }

    /**
     * @return
     */
    public final boolean mayRequestBackLoggingOfDemands() {
        return myPermitBackLoggingFlag;
    }

    /**
     * @param permitBackLoggingFlag
     */
    public final void setPermitBackLoggingFlag(boolean permitBackLoggingFlag) {
        myPermitBackLoggingFlag = permitBackLoggingFlag;
    }

    /**
     * @param permitPartialFillingFlag
     */
    public final void setPermitPartialFillingFlag(boolean permitPartialFillingFlag) {
        myPermitPartialFillingFlag = permitPartialFillingFlag;
    }

    /** Sets the listener that will be attached to demands when they are generated
     *  to listen for rejections.  This may be null. In which case, no reactions to
     *  rejections will occur.
     *
     * @param listener The DemandRejectedListener to set.
     */
    public final void setDemandRejectedListener(DemandListenerRejectedIfc listener) {
        myDemandRejectedListener = listener;
    }

    //  ===========================================
    //      PROTECTED METODS
    //  ===========================================
    /** Sets the ItemType that will be generated by this generator
     *
     * @param type
     */
    protected void setItemType(ItemType type) {
        if (type == null) {
            throw new IllegalArgumentException("Invalid argument. ItemType was null");
        }
        myItemType = type;
    }

    /** If there is no filler available then a NoDemandFillerFoundException exception is thrown
     *  First generates the amount of the demand, then creates it, and sends it to the filler
     *  If the amount generated is zero (or negative) then no demand is created or sent.
     * @param event
     */
    protected void generate(JSLEvent event) {
        // get the amount of the demand
        int amt = generateAmount();
//		System.out.println(getTime()+"> generating demand = " + amt);
        if (amt > 0) {
            if (getUnitDemandOnlyFlag() == true){
                for(int i=1;i<=amt;i++){
                    Demand demand = createDemand(1);
                    // set the filler
                    setFillerOnDemand(demand);
                    // set the listeners
                    setDemandListeners(demand);
                    // send the demand to the filler
                    sendDemand(demand);
                }
            } else {
                // create the demand
                Demand demand = createDemand(amt);
                // set the filler
                setFillerOnDemand(demand);
                // set the listeners
                setDemandListeners(demand);
                // send the demand to the filler
                sendDemand(demand);
            }
        }
    }

    /** If an amount distribution is supplied, then
     *  it is used to generate the amount of the demand.
     *  If no amount distribution is supplied, then
     *  the default demand amount is 1 unit. Sub-classes
     *  can override this method to provide any general
     *  amount generation scheme.
     * @return the amount for the demand at the generation event
     */
    protected int generateAmount() {
        // get the amount of the demand
        int amt = 0;
        if (myAmtRV == null) {
            amt = 1;
        } else {
            amt = (int) myAmtRV.getValue();
        }
        return amt;
    }

    /** Creates a demand for the specified amount with this generators
     *  requirements
     *
     * @param amt
     * @return
     */
    protected Demand createDemand(int amt) {
        // create the demand, with the requirements
        Demand demand = new Demand(getItemType(), amt);
        demand.setAllowBackLoggingFlag(mayRequestBackLoggingOfDemands());
        demand.setAllowPartialFillingFlag(mayRequestPartialFilling());
        demand.setSender(this);
        return (demand);
    }

    /** Determines the filler for the demand for this generator
     *
     * @param demand
     */
    protected void setFillerOnDemand(Demand demand) {
        // if the filler is set, use it, otherwise try to find a filler using the filler finder
        DemandFillerIfc filler;
        if (myDemandFiller == null) {
            filler = myDemandFillerFinder.findDemandFiller(demand);
            if (filler == null) {
                throw new NoDemandFillerFoundException("No demand filler was found for demand: " + demand);
            }
        } else {
            filler = myDemandFiller;
        }

        // set the demand filler
        demand.setFiller(filler);

    }

    /** Sents the received, rejected, and delivered listeners for the demand
     *  prior to sending the demand
     *
     * @param demand
     */
    protected void setDemandListeners(Demand demand) {
        demand.setReceivedListener(myDemandListener);
        demand.setRejectedListener(myDemandRejectedListener);
        demand.setDeliveredListener(myDemandListener);
    }

    /** Asks the appropriate demand filler to receive the demand
     *
     * @param demand
     */
    protected void sendDemand(Demand demand) {
        //System.out.println(getTime() + " > DemandGenerator " + getId() + " sending demand " + demand);
        // place the demand in the SENT state
        demand.sent();
        // get the demand's filler
        DemandFillerIfc filler = demand.getFiller();
        // tell the filler to receive the demand
        filler.receive(demand);
    }

    /**  When the generator's demand is received by a filler
     *   this method is called.  This method calls the filler and asks the filler
     *   to fill the demand.
     *
     * @param demand
     */
    protected void demandReceived(Demand demand) {
        DemandFillerIfc filler = demand.getFiller();
        //System.out.println(getTime() + " > DemandGenerator " + getId() + " asking " + filler.getName() + " to fill " + demand);
        filler.fillDemand(demand);
    }

    /** This method is called when a demand from the generator
     *  has been delivered.
     *
     * @param demand
     */
    protected void demandDelivered(Demand demand) {
        // The demand has been delivered to this generator
        demand.nullify();
    }

    protected void removedFromModel() {
        super.removedFromModel();
        myAmtDistribution = null;
        myAmtRV = null;
        myDemandFillerFinder = null;
        myDemandFiller = null;
        myItemType = null;
        myDemandListener = null;
        myDemandRejectedListener = null;
    }

    //  ===========================================
    //      PRIVATE INNER CLASSES
    //  ===========================================
    /** Provides default listeners for the receipt and delivery of demands
     */
    protected class DemandListener implements DemandListenerReceivedIfc, DemandListenerDeliveredIfc {

        public void demandReceived(Demand demand) {
            if (!demand.isPartOfAnOrder()) {
                DemandGenerator.this.demandReceived(demand);
            }
        }

        public void demandDelivered(Demand demand) {
            DemandGenerator.this.demandDelivered(demand);
        }
    }
}
