/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

/** A DemandMessageIfc provides information from a DemandFillerIfc in regards to what 
 *  will happen to a demand that is sent to the filler at the current time.  A DemandFillerIfc
 *  will return a DemandMessageIfc from its negotiate() method if it can negotiate; otherwise, the 
 *  the message will be null.  The message might be used by a DemandSenderIfc to decide on an
 *  appropriate course of action before sending the demand to a DemandFillerIfc.  The information
 *  returned in the message is only about the demand that was given to the filler via the negotiate()
 *  method.  All responses are to be interpreted relative to that demand instance.
 * 
 *
 */
public interface DemandMessageIfc {
	
	/** The DemandFillerIfc associated with this message.
	 * 
	 * @return  the filler
	 */
	public DemandFillerIfc getDemandFiller();

	/** Allows clients to check if the ItemType can be received.  This can
	 *  be used before calling fillDemand(Demand demand)
	 * 
	 * @return True if the filler is willing to fill demands that have the given item type
	 */
	public boolean canFillItemType();

	/** Returns an integer that represents what will happen to the demand
	 * 
	 * @return the potential status of the demand
	 */
	public int getRequestStatus();
	
	/** Returns how many units that the demand will be allocated if it is
	 *  processed by the filler at this time
	 * 
	 * @return the amt to be allocated
	 */
	public int getRequestFillAmount();
	
	/** An InventoryIfc that can be used to check the state of the inventory
	 * 
	 * @return This may be null
	 */
	public InventoryIfc getInventory();

	/** Indicates whether or not the demand filler may partially fill demands.  This
	 *  does not necessarily imply that the demands WILL be partially filled by the filler.  It only
	 *  implies that the filler may partially attempt to partially fill demands during its processing of the demand.
	 *  
	 *  A demand that permits partial filling sent to a filler that does not partially fill is still valid.
	 *  The demand simply will not be partially filled and the client cannot expect it to be partially filled.
	 *  
	 *  An demand that is not allowed to be partially filled sent to a filler that partially fills demands 
	 *  can still be valid.  The filler may decide to reject the demand in this case.
	 *  If the filler accepts the demand, then filler is responsible for not partially filling the demand.	
	 * 
	 * @return
	 */
	public boolean mayPartiallyFillDemands();
	
	/** Indicates whether or not the filler can fill demands that permit
	 *  backlogging.  This does not necessarily imply that the demands WILL be backlogged
	 *  by the filler.  It only implies that the filler might backlog incoming orders. 
	 *  The sender must be willing to accept the fact that the demands might be backlogged.
	 *  
	 *  An demand that does not permit backlogging can still be sent to a filler that
	 *  may backlog orders.  If the demand (that does not permit backlogging)
	 *  can be processed without backlogging then it should be received. If it cannot
	 *  be processed without backlogging, then it should be rejected.
	 *  
	 * @return
	 */
	public boolean mayBackLogDemands();
		
	/** Indicates when the message was created.  The time that it was valid.
	 * 
	 * @return The time that the message was negotiated.
	 */
	public double getTimeStamp();
	
}
