/*
 * Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
 *
 * Contact:
 *	Manuel D. Rossetti, Ph.D., P.E.
 *	Department of Industrial Engineering
 *	University of Arkansas
 *	4207 Bell Engineering Center
 *	Fayetteville, AR 72701
 *	Phone: (479) 575-6756
 *	Email: rossetti@uark.edu
 *	Web: www.uark.edu/~rossetti
 *
 * This file is part of the JSL (a Java Simulation Library). The JSL is a framework
 * of Java classes that permit the easy development and execution of discrete event
 * simulation programs.
 *
 * The JSL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * The JSL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the JSL (see file COPYING in the distribution);
 * if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA, or see www.fsf.org
 *
 */
package jslx.supplychain.inventorylayer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Observer;

import jsl.modeling.IllegalStateException;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AcrossReplicationStatisticIfc;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.ResponseStatisticsIfc;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.reporting.JSL;

/**
 * 
 */
public class Inventory extends DemandFillerAbstract implements
        InventoryIfc, DemandSenderIfc {

    //  ===========================================
    //      OBJECT VARIABLES
    //  ===========================================

    /** A reference to the type of item for this
     *  inventory provider
     */
    private ItemType myItemType;

    /** Tracks the amount of inventory actual
     *  on hand and available for supply
     */
    protected TimeWeighted myOnHand;

    /** Tracks when the net inventory is <= 0
     *  net inventory = myOnHand - myAmtBackLogged
     */
    protected TimeWeighted myStockOutIndicator;

    /** Tracks the amount of individual inventory units
     *  that have been ordered to replenish the stock
     */
    protected TimeWeighted myOnOrder;

    /** Tracks the amount of individual inventory units
     *  that will be on order when the replenishment request is
     *  received by a filler
     */
    protected int myAmountPending;

    /** Tracks whether or not a demand is filled
     *  directly from onhand on first attempt
     */
    protected ResponseVariable myFirstFillRate;

    /** Counts the number of replenishment requests made by the inventory
     */
    protected Counter myReplenishmentRequestCounter;

    /** Counts the total amount of demand filled
     *  during a replication
     */
    protected long myDemandArrivalCounter;

    /** Indicated whether or not the inventory will
     *  permit backlogging of its replenishment requests
     */
    private boolean myPermitBackLoggingFlag;

    /** Indicated whether or not the inventory will
     *  permit partial filling of its replenishment requests
     */
    private boolean myPermitPartialFillingFlag;

    /** Controls how replenishment occurs
     */
    private InventoryPolicyAbstract myInventoryPolicy;

    /** Controls how backlogs are handled
     */
    private BackLogPolicyAbstract myBackLogPolicy;

    /** Indicates whether or not this inventory may backlog
     *  demand requests
     */
    private boolean myAllowBackLoggingFlag;

    /** Responsible for finding an appropriate filler for the generator
     */
    private DemandFillerFinderIfc myDemandFillerFinder;

    /** This attribute can be used to directly set the filler for this inventory
     */
    private DemandFillerIfc myDemandFiller;

    /** Indicates whether or not this filler may partially fill
     *  demand requests
     */
    private boolean myMayPartiallyFillFlag;

    /** If set, the inventory will use this to request replenishments instead
     *  of its normal procedures
     */
    private ReplenishmentRequesterIfc myReplenishmentRequester;

    /** Listens for Received, Cancelled, InProcess, and Delivered
     */
    private ReplenishmentDemandListener myDemandListener = new ReplenishmentDemandListener();
    
    /** Listener for rejection
     */
    private DemandListenerRejectedIfc myDemandRejectedListener = new DemandRejectionListener();

    /** Can be attached to listen for demand arrivals for filling
     *  Called after processing the arrival
     */
    protected InventoryDemandArrivalListenerIfc myDemandArrivalListener;

    /** Can be attached to listen for the arrival of replenshment
     *  Called after to processing the replenishment
     */
    protected InventoryDemandArrivalListenerIfc myReplenishmentArrivalListener;

    /** Creates an Inventory to fill demands
     *
     * @param parent
     * @param type The type of item this inventory fills
     * @param policy Controls the inventory's replenishment ordering
     * @param initialLevel The amount of inventory on hand at the beginning of each replication, default = 0
     */
    public Inventory(ModelElement parent, ItemType type, InventoryPolicyAbstract policy, int initialLevel) {
        this(parent, type, policy, initialLevel, true, null);
    }

    /** Creates an Inventory to fill demands
     *
     * @param parent
     * @param type The type of item this inventory fills
     * @param policy Controls the inventory's replenishment ordering
     * @param backLogPolicy  Allows the inventory to backlog demands, if not supplied then the inventory does not backlog
     */
    public Inventory(ModelElement parent, ItemType type, InventoryPolicyAbstract policy) {
        this(parent, type, policy, 0, true, null);
    }

    /** Creates an Inventory to fill demands
     *
     * @param parent
     * @param type The type of item this inventory fills
     * @param policy Controls the inventory's replenishment ordering
     * @param initialLevel The amount of inventory on hand at the beginning of each replication
     * @param name
     */
    public Inventory(ModelElement parent, ItemType type, InventoryPolicyAbstract policy, int initialLevel, String name) {
        this(parent, type, policy, initialLevel, true, name);
    }

    /** Creates an Inventory to fill demands
     *
     * @param parent
     * @param type The type of item this inventory fills
     * @param policy Controls the inventory's replenishment ordering
     * @param initialLevel The amount of inventory on hand at the beginning of each replication, default = 0
     * @param partialFillFlag Whether or not the inventory may partially fill demand requests, default = true
     */
    public Inventory(ModelElement parent, ItemType type, InventoryPolicyAbstract policy, int initialLevel,
            boolean partialFillFlag) {
        this(parent, type, policy, initialLevel, partialFillFlag, null);
    }

    /** Creates an Inventory to fill demands
     *
     * @param parent
     * @param type The type of item this inventory fills
     * @param policy Controls the inventory's replenishment ordering
     * @param initialLevel The amount of inventory on hand at the beginning of each replication, default = 0
     * @param partialFillFlag Whether or not the inventory may partially fill demand requests, default = true
     */
    public Inventory(ModelElement parent, ItemType type, InventoryPolicyAbstract policy, int initialLevel,
            boolean partialFillFlag, boolean cancelDemandFlag) {
        this(parent, type, policy, initialLevel, partialFillFlag, null);
    }

    /** Creates an Inventory to fill demands
     *
     * @param parent
     * @param type The type of item this inventory fills
     * @param policy Controls the inventory's replenishment ordering
     * @param initialLevel The amount of inventory on hand at the beginning of each replication, default = 0
     * @param partialFillFlag Whether or not the inventory may partially fill demand requests, default = true
     * @param name
     */
    public Inventory(ModelElement parent, ItemType type, InventoryPolicyAbstract policy, int initialLevel,
            boolean partialFillFlag, String name) {
        super(parent, true, name);

        setItemType(type);
        setInventoryPolicy(policy);
        setMayPartiallyFillFlag(partialFillFlag);
        setPermitBackLoggingFlag(true);

        myStockOutIndicator = new TimeWeighted(this, getName() + "-StockoutIndicator");
        myOnHand = new TimeWeighted(this, initialLevel, 0, getName() + "-OnHand");
        myOnOrder = new TimeWeighted(this, getName() + "-OnOrder");
        myAmountPending = 0;
//		myAmountPending = new TimeWeighted(this, getName() + "-Pending");
        myFirstFillRate = new ResponseVariable(this, getName() + "-FirstFillRate");
        myReplenishmentRequestCounter = new Counter(this, getName() + "-#ReplenishmentRequests");

    }

    //  ===========================================
    //      PUBLIC METHODS
    //  ===========================================
    /** Creates an inventory with an (reorder point,reorder quantity) policy that uses a backlog queue
     *  Initial level is zero.
     *
     * @param parent
     * @param type
     * @param reorderPt
     * @param reorderQty
     * @return
     */
    public final static Inventory createReorderPointReorderQuantityInventory(ModelElement parent, ItemType type, int reorderPt, int reorderQty) {
        return (createReorderPointReorderQuantityInventory(parent, type, reorderPt, reorderQty, 0, null));
    }

    /** Creates an inventory with an (reorder point,reorder quantity) policy that uses a backlog queue
     *
     * @param parent
     * @param type
     * @param reorderPt
     * @param reorderQty
     * @param initialLevel
     * @return
     */
    public final static Inventory createReorderPointReorderQuantityInventory(ModelElement parent, ItemType type, int reorderPt, int reorderQty, int initialLevel) {
        return (createReorderPointReorderQuantityInventory(parent, type, reorderPt, reorderQty, initialLevel, null));
    }

    /** Creates an inventory with an (reorder point,reorder quantity) policy that uses a backlog queue
     *
     * @param parent
     * @param type
     * @param reorderPt
     * @param reorderQty
     * @param initialLevel
     * @param name
     * @return
     */
    public final static Inventory createReorderPointReorderQuantityInventory(ModelElement parent, ItemType type, int reorderPt, int reorderQty, int initialLevel, String name) {
        InventoryPolicyReorderPointReorderQuantity rQPolicy = new InventoryPolicyReorderPointReorderQuantity(parent, reorderPt, reorderQty);
        Inventory inventory = new Inventory(parent, type, rQPolicy, initialLevel, name);
//		BackLogQueue bpolicy = new BackLogQueue(parent, inventory.getName() + "-BackLogQ");
        BackLogQueue bpolicy = new BackLogQueue(inventory, inventory.getName() + "-BackLogQ");
//		inventory.setBackLogPolicy(bpolicy);
        return (inventory);
    }

    /** Creates an inventory with an (reorder point, order up to level) policy that uses a backlog queue
     *  Initial level is zero.
     *
     * @param parent
     * @param type
     * @param reorderPt
     * @param reorderLevel
     * @return
     */
    public final static Inventory createReorderPointOrderUpToLevelInventory(ModelElement parent, ItemType type, int reorderPt, int reorderLevel) {
        return (createReorderPointOrderUpToLevelInventory(parent, type, reorderPt, reorderLevel, 0, null));
    }

    /** Creates an inventory with an (reorder point, order up to level) policy that uses a backlog queue
     *
     * @param parent
     * @param type
     * @param reorderPt
     * @param reorderLevel
     * @param initialLevel
     * @return
     */
    public final static Inventory createReorderPointOrderUpToLevelInventory(ModelElement parent, ItemType type, int reorderPt, int reorderLevel, int initialLevel) {
        return (createReorderPointOrderUpToLevelInventory(parent, type, reorderPt, reorderLevel, initialLevel, null));
    }

    /** Creates an inventory with an (reorder point, order up to level) policy that uses a backlog queue
     *
     * @param parent
     * @param type
     * @param reorderPt
     * @param reorderLevel
     * @param initialLevel
     * @param name
     * @return
     */
    public final static Inventory createReorderPointOrderUpToLevelInventory(ModelElement parent, ItemType type, int reorderPt, int reorderLevel, int initialLevel, String name) {
        InventoryPolicyReorderPointOrderUpToLevel policy = new InventoryPolicyReorderPointOrderUpToLevel(parent, reorderPt, reorderLevel);
        Inventory inventory = new Inventory(parent, type, policy, initialLevel, name);
        BackLogQueue bpolicy = new BackLogQueue(inventory, inventory.getName() + "-BackLogQ");
//		inventory.setBackLogPolicy(bpolicy);
        return (inventory);
    }

    /* (non-Javadoc)
     * @see jsl.supplychain.inventorylayer.DemandSenderIfc#mightRequest(jsl.supplychain.inventorylayer.ItemType)
     */
    public final boolean mightRequest(ItemType type) {
        return myItemType == type;
    }

    /** Returns whether or not this inventory will allow its
     *  replenshment requests to be partially filled
     *
     * @return true means partial filling is allowed
     */
    public final boolean mayRequestPartialFilling() {
        return myPermitPartialFillingFlag;
    }

    /** Returns whether or not this inventory will allow its replenishment
     *  requests to be backlogged
     *
     * @return true means that backlogging is allowed
     */
    public final boolean mayRequestBackLoggingOfDemands() {
        return myPermitBackLoggingFlag;
    }

    /** Controls whether replenishment demands are allowed to be backlogged.  Changing this flag,
     *  affect how future replenishment demands are made. It does not change any already made demands.
     *  This flag may affect whether or not demands can be filled since an appropriate demand filler
     *  may not be able to handle the requests and demands will be rejected.  The inventory is
     *  responsible for reacting to any rejected demands.
     *
     * @param permitBackLoggingFlag
     */
    public final void setPermitBackLoggingFlag(boolean permitBackLoggingFlag) {
        myPermitBackLoggingFlag = permitBackLoggingFlag;
    }

    /** Controls whether replenishment demands are allowed to be partial filled.  Changing this flag,
     *  affect how future replenishment demands are made. It does not change any already made demands.
     *  This flag may affect whether or not demands can be filled since an appropriate demand filler
     *  may not be able to handle the requests and demands will be rejected.  The inventory is
     *  responsible for reacting to any rejected demands.
     *
     * @param permitPartialFillingFlag
     */
    public final void setPermitPartialFillingFlag(boolean permitPartialFillingFlag) {
        myPermitPartialFillingFlag = permitPartialFillingFlag;
    }

    /* (non-Javadoc)
     * @see jsl.supplychain.inventorylayer.DemandFillerIfc#canFillItemType(jsl.supplychain.inventorylayer.ItemType)
     */
    public final boolean canFillItemType(ItemType type) {
        return myItemType == type;
    }

    /* (non-Javadoc)
     * @see jsl.research.supplychain.inventorylayer.DemandFillerIfc#getItemTypes()
     */
    public final Collection<ItemType> getItemTypes() {
        ArrayList<ItemType> t = new ArrayList<ItemType>();
        t.add(myItemType);
        return t;
    }

    /* (non-Javadoc)
     * @see jsl.supplychain.inventorylayer.InventoryIfc#getAmountOnHand()
     */
    public final int getAmountOnHand() {
        return ((int) myOnHand.getValue());
    }

    /** Allows an observer to be attached to the amount on hand.
     *
     * @param observer
     */
    public final void addOnHandObserver(Observer observer) {
        myOnHand.addObserver(observer);
    }

    /** Removes the supplied observer from the on hand if attached
     *
     * @param observer
     */
    public final void removeOnHandObserver(Observer observer) {
        myOnHand.deleteObserver(observer);
    }

    /** Gets the total amount of demand that has arrived
     *  since the beginning of a replication.
     * 
     * @return
     */
    public final long getTotalArrivedDemand(){
        return myDemandArrivalCounter;
    }

    /** Attaches an AggregateTimeWeightedVariable to the on hand
     *  for this inventory.  Whenever the on hand gets updated
     *  the aggregate variable is updated.
     *
     * @param variable
     */
    public final void attachAggregateTimeWeightedVariableToOnHand(AggregateTimeWeightedVariable variable) {
        variable.subscribeTo(myOnHand);
    }

    /** Detaches the supplied variable from the on-hand as an aggregate
     *
     * @param variable
     */
    public final void detachAggregateTimeWeightedVariableFromOnHand(AggregateTimeWeightedVariable variable) {
        variable.unsubscribeFrom(myOnHand);
    }

    /** Allows an observer to be attached to the counter for the number
     *  of replenishment demand made by the inventory
     *
     * @param observer
     */
    public final void addReplishmentRequestCounterObserver(Observer observer) {
        myReplenishmentRequestCounter.addObserver(observer);
    }

    /** Attaches an AggregateCounter to the ReplenishmentRequestCounter
     *  for this inventory.  Whenever the replenishment request counter gets updated
     *  the aggregate counter is updated.
     *
     * @param counter
     */
    public final void attachAggregateCounterToReplenishmentRequestCounter(AggregateCounter counter) {
        counter.subscribeTo(myReplenishmentRequestCounter);
    }

    /** Removes the supplied observer from the replenishment request counter if attached
     *
     * @param observer
     */
    public final void removeReplishmentRequestCounterObserver(Observer observer) {
        myReplenishmentRequestCounter.deleteObserver(observer);
    }

    /** Detaches the supplied variable from the replenishment request counter as an aggregate
     *
     * @param variable
     */
    public final void detachAggregateCounterFromReplenishmentRequestCounter(AggregateCounter counter) {
        counter.unsubscribeFrom(myReplenishmentRequestCounter);
    }

    /** Allows an observer to be attached to the amount on order
     *
     * @param observer
     */
    public final void addOnOrderObserver(Observer observer) {
        myOnOrder.addObserver(observer);
    }

    /** Attaches an AggregateTimeWeightedVariable to the amount on order
     *  for this inventory.  Whenever the amount on order gets updated
     *  the aggregate variable is updated.
     *
     * @param variable
     */
    public final void attachAggregateTimeWeightedVariableToOnOrder(AggregateTimeWeightedVariable variable) {
        variable.subscribeTo(myOnOrder);
    }

    /** Removes the supplied observer from the on order if attached
     *
     * @param observer
     */
    public final void removeOnOrderObserver(Observer observer) {
        myOnOrder.deleteObserver(observer);
    }

    /** Detaches the supplied variable from the on-order as an aggregate
     *
     * @param variable
     */
    public final void detachAggregateTimeWeightedVariableFromOnOrder(AggregateTimeWeightedVariable variable) {
        variable.unsubscribeFrom(myOnOrder);
    }

    /** Allows an observer to be attached to the amount on hand.
     *
     * @param observer
     */
    public final void addStockOutIndicatorObserver(Observer observer) {
        myStockOutIndicator.addObserver(observer);
    }

    /** Attaches an AggregateTimeWeightedVariable to the stock out indicator
     *  for this inventory.  Whenever the stock out indicator gets updated
     *  the aggregate variable is updated.
     *
     * @param variable
     */
    public final void attachAggregateTimeWeightedVariableToStockOutIndicator(AggregateTimeWeightedVariable variable) {
        variable.subscribeTo(myStockOutIndicator);
    }

    /** Removes the supplied observer from the stock out indicator if attached
     *
     * @param observer
     */
    public final void removeStockOutIndicatorObserver(Observer observer) {
        myStockOutIndicator.deleteObserver(observer);
    }

    /** Detaches the supplied variable from the stock out indicator as an aggregate
     *
     * @param variable
     */
    public final void detachAggregateTimeWeightedVariableFromStockOutIndicator(AggregateTimeWeightedVariable variable) {
        variable.unsubscribeFrom(myStockOutIndicator);
    }

    /** Attaches an ResponseVariableAverageObserver to the first fill
     *  for this inventory.
     *
     * @param variable
     */
    public final void attachResponseVariableAverageObserverToFirstFillRate(ResponseVariableAverageObserver variable) {
        variable.subscribeTo(myFirstFillRate);
    }

    /** Allows an observer to be attached to the first fill opportunities
     *
     * @param observer
     */
    public final void addFirstFillRateObserver(Observer observer) {
        myFirstFillRate.addObserver(observer);
    }

    /** Removes the supplied observer from the first fill indicator if attached
     *
     * @param observer
     */
    public final void removeFillRateObserver(Observer observer) {
        myFirstFillRate.deleteObserver(observer);
    }

    /** Detaches the supplied variable from the first fill indicator as an aggregate
     *
     * @param variable
     */
    public final void detachResponseVariableAverageObserverFromFirstFillRate(ResponseVariableAverageObserver variable) {
        variable.unsubscribeFrom(myFirstFillRate);
    }

    /** Allows an observer to be attached to the amount backlogged if a backlog policy is used
     *
     * @param observer
     */
    public final void addAmountBackLoggedObserver(Observer observer) {
        if (myBackLogPolicy != null) {
            myBackLogPolicy.addAmountBackLoggedObserver(observer);
        }
    }

    /** Attaches an AggregateTimeWeightedVariable to the on hand
     *  for this inventory.  Whenever the on hand gets updated
     *  the aggregate variable is updated.
     *
     * @param variable
     */
    public final void attachAggregateTimeWeightedVariableToAmountBackLogged(AggregateTimeWeightedVariable variable) {
        if (myBackLogPolicy != null) {
            myBackLogPolicy.attachAggregateTimeWeightedVariableToAmountBackLogged(variable);
        }
    }

    /** Removes the supplied observer from the amount back logged if attached
     *
     * @param observer
     */
    public final void removeAmountBackLoggedObserver(Observer observer) {
        if (myBackLogPolicy != null) {
            myBackLogPolicy.removeAmountBackLoggedObserver(observer);
        }
    }

    /** Detaches the supplied variable from the amount back logged as an aggregate
     *
     * @param variable
     */
    public final void detachAggregateTimeWeightedVariableFromAmountBackLogged(AggregateTimeWeightedVariable variable) {
        if (myBackLogPolicy != null) {
            myBackLogPolicy.detachAggregateTimeWeightedVariableFromAmountBackLogged(variable);
        }
    }

    /** Allows an observer to be attached to the waiting time in the backlog queue
     *
     * @param observer
     */
    public final void addBackLoggedQueueTimeObserver(Observer observer) {
        if (myBackLogPolicy != null) {
            myBackLogPolicy.addBackLoggedQueueTimeObserver(observer);
        }
    }

    /** Allows an observer to be attached to the number in the backlog queue
     *
     * @param observer
     */
    public final void addNumInBackLoggedQueueObserver(Observer observer) {
        if (myBackLogPolicy != null) {
            myBackLogPolicy.addNumInBackLoggedQueueObserver(observer);
        }
    }

    /* (non-Javadoc)
     * @see jsl.supplychain.inventorylayer.InventoryIfc#getAmountOnOrder()
     */
    public final int getAmountOnOrder() {
        return (int) myOnOrder.getValue();
    }

    /** Returns the amount of inventory units that will be on order
     *  once the replenishment demand is received by the filler. If
     *  there is no time delay in the receipt of the request by the filler
     *  then the statistics on the amount pending will always be zero.
     *
     * @return
     */
    public final int getAmountPending() {
        return myAmountPending;
//		return((int)myAmountPending.getValue());
    }

    /** Gets the initial value of the onhand
     * @return The initial value.
     */
    public final int getInitialOnHand() {
        return ((int) myOnHand.getInitialValue());
    }

    /** Sets the initial on hand of the inventory. Only relevant prior to each
     * replication.  Changing during a replication has no effect until the next replication.
     * @param value The initial value for the inventory.
     */
    public final void setInitialOnHand(int value) {
        if (value < 0) {
            throw new IllegalArgumentException("Invalid argument. Cannot have a negative on hand");
        }

        myOnHand.setInitialValue(value);
        if (value == 0) {
            myStockOutIndicator.setInitialValue(1.0);
        } else {
            myStockOutIndicator.setInitialValue(0.0);
        }
    }

    /* (non-Javadoc)
     * @see jsl.supplychain.inventorylayer.InventoryIfc#getBackLogInfo()
     */
    public final BackLogInfoIfc getBackLogInfo() {
        return (myBackLogPolicy);
    }

    /** Returns the inventory position of the InventoryProvider
     *  InventoryPosition should be Onhand + Amt Pending + OnOrder - BackLogged
     *
     * @return The current inventory position of the ItemType
     */
    public final int getInventoryPosition() {
        int bl = 0;
        if (myBackLogPolicy != null) {
            bl = myBackLogPolicy.getAmountBackLogged();
        }

        int ip = getAmountOnHand() + getAmountOnOrder() + getAmountPending() - bl;
        return (ip);
    }

    /** Sets the inventory policy that this inventory is using.  Changing the policy
     *  during the simulation is possible.  Care must be taken to consider unintended side-effects
     *
     * @param policy
     */
    public final void setInventoryPolicy(InventoryPolicyAbstract policy) {
        if (policy == null) {
            throw new IllegalArgumentException("Invalid argument. InventoryPolicyAbstract policy must be non-null.");
        }
        myInventoryPolicy = policy;
        myInventoryPolicy.setInventory(this);
    }

    /** Null implies that the inventory does not backlog demand.
     *  Once the backlog policy has been set to a non-null value
     *  it cannot be changed, i.e. once the backlog policy is not null
     *  this method does nothing.
     * @param backLogPolicy The backLogPolicy to set.
     */
    protected final void setBackLogPolicy(BackLogPolicyAbstract backLogPolicy) {

        // if the backlog policy has already been set, do not allow it to change
        if (myBackLogPolicy != null) {
            return;
        }

        if (backLogPolicy != null) {
            myAllowBackLoggingFlag = true;
            myBackLogPolicy = backLogPolicy;
            myBackLogPolicy.setInventory(this);
        }
    }

    /** The supplied finder is responsible for finding an appropriate filler
     *  so that the inventory's replenishment requests can be fulfilled.
     *
     * @param finder Must not be null
     */
    public final void setDemandFillerFinder(DemandFillerFinderIfc finder) {
        if (finder == null) {
            throw new IllegalArgumentException("Invalid argument. DemandFillerFinder must be non-null.");
        }
        myDemandFillerFinder = finder;
    }

    /**  Allows access the inventory's filler finder
     *
     * @return
     */
    public final DemandFillerFinderIfc getDemandFillerFinder() {
        return (myDemandFillerFinder);
    }

    /** Can be used to directly set the filler for this inventory, rather than relying
     *  on the filler finder
     *
     * @param fillerFinder Must not be null
     */
    public final void setDemandFiller(DemandFillerIfc filler) {
        if (filler == null) {
            throw new IllegalArgumentException("Invalid argument. DemandFillerIfc must be non-null.");
        }

        myDemandFiller = filler;
    }

    /**  Allows access the inventory's filler
     *
     * @return
     */
    public final DemandFillerIfc getDemandFiller() {
        return (myDemandFiller);
    }

    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuilder sb = new StringBuilder();
        int oh = getAmountOnHand();
        int oo = getAmountOnOrder();
        int po = getAmountPending();
        int ip = getInventoryPosition();
        int bl = 0;
        if (myBackLogPolicy != null) {
            bl = myBackLogPolicy.getAmountBackLogged();
        }
        sb.append("Inventory:\t");
        sb.append(getName());
        sb.append("\t Type: ");
        sb.append(getItemType());
        sb.append("\t OH: ");
        sb.append(oh);
        sb.append("\t OO: ");
        sb.append(oo);
        sb.append("\t PO: ");
        sb.append(po);
        sb.append("\t BL: ");
        sb.append(bl);
        sb.append("\t IP: ");
        sb.append(ip);
        sb.append("\n");
        return (sb.toString());
    }

    /** Returns true if this inventory may backlog demands that it receives.
     * @returns
     */
    public final boolean mayBackLogDemands() {
        return myAllowBackLoggingFlag;
    }

    /** The item type that this inventory fills
     *
     * @return
     */
    public final ItemType getItemType() {
        return myItemType;
    }

    /* (non-Javadoc)
     * @see jsl.supplychain.inventorylayer.DemandFillerIfc#receiveDemand(jsl.supplychain.inventorylayer.Demand)
     */
    public void receive(Demand demand) {

        // determine whether to reject or process the demand

        if (!isAvailable()) {
            demand.setDemandStatus(Demand.FILLER_UNAVAILABLE);
            demand.reject();
            return;
        }

        if (!canFillItemType(demand.getItemType())) {
            demand.setDemandStatus(Demand.ITEMTYPE_MISMATCH);
            demand.reject();
            return;
        }

        int onHand = getAmountOnHand();
        int amtNeeded = demand.getRemainingDemand();
        //JSL.out.println(getTime() + " > Inventory " + getId() + " received request for Demand: " + demand.getId() + " for " + amtNeeded);
        //JSL.out.println(this);
        if (onHand >= amtNeeded) {
            demand.setDemandStatus(Demand.IMMEDIATE_FILL);
            demand.receive(this);
        } else { // cannot fill all the demand, amtNeeded > onhand, the demand may have to wait
            // check if demand allows backlogging
            if (demand.getAllowBackLoggingFlag() == false) { //NBLD, demand does not allow backlogging
                // the demand is unwilling to wait
                // some or all of the request cannot be satisfied
                if (mayBackLogDemands() == true) { //BLR, the receiver allows backlogging
                    demand.setDemandStatus(Demand.NBLD_SENT_TO_BLR_NIF);
                    demand.reject();
                } else { // NBLR, the receiver does not allow backlogging
                    demand.setDemandStatus(Demand.NBLD_SENT_TO_NBLR_NIF);
                    demand.reject();
                }
            } else {// willing to wait for the request to be filled, BLD, the demand allows backlogging
                if (mayBackLogDemands() == true) { //BLR, the reciever allows backlogging
                    demand.setDemandStatus(Demand.WILL_BE_BACKLOGGED);
                    demand.receive(this);
                } else { // NBLR, the receiver does not allow backlogging
                    demand.setDemandStatus(Demand.BLD_SENT_TO_NBLR_NIF);
                    demand.reject();
                }
            }
        }

    }

    /* (non-Javadoc)
     * @see jsl.supplychain.inventorylayer.DemandFillerIfc#fillDemand(jsl.supplychain.inventorylayer.Demand)
     */
    public void fillDemand(Demand demand) {

        // count the demand
        myDemandArrivalCounter = myDemandArrivalCounter + demand.getOriginalAmountDemanded();

        // place the demand in process
        demand.process(this);

        if (demand.getStatus() == Demand.IMMEDIATE_FILL) {
            //JSL.out.println(getTime() + " > in fillDemand() Demand: " + demand.getId() + " immediate fill");
            fillImmediately(demand);
        } else { // cannot fill all the demand, amtNeeded > onhand, the demand may have to wait
            //JSL.out.println(getTime() + " > in fillDemand() Demand: " + demand.getId() + " to be backlogged");
            backLogInitialDemand(demand);
        }

        //notify demand arrival listeners about the arrival
        if (myDemandArrivalListener != null) {
            myDemandArrivalListener.demandArrived(this, demand);
        }

    }

    /* (non-Javadoc)
     * @see jsl.supplychain.inventorylayer.DemandFillerIfc#negotiate(jsl.supplychain.inventorylayer.Demand)
     */
    public DemandMessageIfc negotiate(Demand demand) {
        demand.negotiate();
        boolean itemTypeFlag = (getItemType() == demand.getItemType());
        int status = determineRequestStatus(demand);
        int amt = allocateInventory(demand);

        DemandMessageIfc dm = new DemandMessage(this, itemTypeFlag, status, amt, this,
                mayPartiallyFillDemands(), mayBackLogDemands());
        return dm;
    }

    /** Given a demand, this method will return an integer that represents
     *  what will happend to the demand if it is requested to be filled at
     *  the current time.  The status is only valid at the current simulation time.
     *
     * @param demand
     * @return
     */
    public int determineRequestStatus(Demand demand) {

        int status = Demand.NO_STATUS;

        if (!isAvailable()) {
            status = Demand.FILLER_UNAVAILABLE;
            return (status);
        }

        if (!canFillItemType(demand.getItemType())) {
            status = Demand.ITEMTYPE_MISMATCH;
            return (status);
        }

        int onHand = getAmountOnHand();
        int amtNeeded = demand.getRemainingDemand();

        if (onHand >= amtNeeded) {
            status = Demand.IMMEDIATE_FILL;
        } else { // cannot fill all the demand, amtNeeded > onhand, the demand may have to wait
            // check if demand allows backlogging
            if (demand.getAllowBackLoggingFlag() == false) { //NBLD, demand does not allow backlogging
                // the demand is unwilling to wait
                // some or all of the request cannot be satisfied
                if (mayBackLogDemands() == true) { //BLR, the receiver allows backlogging
                    status = Demand.NBLD_SENT_TO_BLR_NIF;
                } else { // NBLR, the receiver does not allow backlogging
                    status = Demand.NBLD_SENT_TO_NBLR_NIF;
                }
            } else {// willing to wait for the request to be filled, BLD, the demand allows backlogging
                if (mayBackLogDemands() == true) { //BLR, the reciever allows backlogging
                    status = Demand.WILL_BE_BACKLOGGED;
                } else { // NBLR, the receiver does not allow backlogging
                    status = Demand.BLD_SENT_TO_NBLR_NIF;
                }
            }
        }
        return status;
    }

    /** Given a demand, this method indicates whether or not the filler
     *  will reject the demand if it is requested to be filled at the current time.
     *  True means that it will be rejected, false means that it will be received w/o rejection
     * @param demand
     * @return
     */
    public final boolean willReject(Demand demand) {
        // determine whether to reject or process the demand

        if (!isAvailable()) {
            return true;
        }

        if (!canFillItemType(demand.getItemType())) {
            return true;
        }

        int onHand = getAmountOnHand();
        int amtNeeded = demand.getRemainingDemand();

        if (onHand >= amtNeeded) {
            return false;
        } else { // cannot fill all the demand, amtNeeded > onhand, the demand may have to wait
            // check if demand allows backlogging
            if (demand.getAllowBackLoggingFlag() == false) { //NBLD, demand does not allow backlogging
                return true;
            } else {// willing to wait for the request to be filled, BLD, the demand allows backlogging
                if (mayBackLogDemands() == true) { //BLR, the reciever allows backlogging
                    return false;
                } else { // NBLR, the receiver does not allow backlogging
                    return true;
                }
            }
        }
    }

    /**
     * @return
     */
    public final boolean mayPartiallyFillDemands() {
        return myMayPartiallyFillFlag;
    }

    /* (non-Javadoc)
     * @see jsl.supplychainNew.inventorylayer.DemandFillerIfc#canFillItemType(jsl.supplychainNew.inventorylayer.Demand)
     */
    public final boolean canFillItemType(Demand demand) {
        return (demand.getItemType() == myItemType);
    }

    //  ===========================================
    //      PROTECTED METHODS
    //  ===========================================
    /** Sets the type of item associated with this inventory
     * @param type
     */
    protected final void setItemType(ItemType type) {
        if (type == null) {
            throw new IllegalArgumentException("Invalid argument. ItemType type was null.");
        }
        myItemType = type;
    }

    /** Allows subclasses to access the inventory's backlog policy
     *
     * @return
     */
    protected final BackLogPolicyAbstract getBackLogPolicy() {
        return myBackLogPolicy;
    }

    /**
     *
     * @return inventory policy parameters
     */
    // TODO Vijith added it
    public final double[] getInventoryPolicyParameters() {
        return myInventoryPolicy.getPolicyParameters();
    }

    /** Allows subclasses to access the inventory's inventory policy
     *
     * @return
     */
    protected final InventoryPolicyAbstract getInventoryPolicy() {
        return myInventoryPolicy;
    }

    /** Allows subclasses to access the on hand
     *
     * @return
     */
    protected final TimeWeighted getOnHand() {
        return myOnHand;
    }

    /** Allows subclasses to access the amount on order
     *
     * @return
     */
    protected final TimeWeighted getOnOrder() {
        return myOnOrder;
    }

    /** Asks the inventory policy to check the inventory
     *
     *
     */
    protected final void checkInventory() {
        getInventoryPolicy().checkInventory();
    }

    /** Initializes the inventory prior to a replication
     *  Does a check of the inventory to see if an replenishment
     *  is needed.
     */
    protected void initialize() {
        super.initialize();
        myAmountPending = 0;
        myDemandArrivalCounter = 0;
        
//		System.out.println("Initializing for replication: " + this.getModel().getReplication().getReplicationNumber());
//		System.out.println(this);
        checkInventory();
    }

    protected void removedFromModel() {
        super.removedFromModel();
        myItemType = null;
        myOnHand = null;
        myStockOutIndicator = null;
        myOnOrder = null;
//		myAmountPending = null;
        myFirstFillRate = null;
        myReplenishmentRequestCounter = null;
        myInventoryPolicy = null;
        myBackLogPolicy = null;
        myDemandFillerFinder = null;
        myReplenishmentRequester = null;
        myDemandListener = null;
        myDemandRejectedListener = null;
    }

    /** Allows subclasses to directly decrement the on hand. This method
     *  does not automatically check the inventory position.
     *
     * @param amount The amount to decrement by
     */
    protected final void decrementOnHand(int amt) {
        myOnHand.decrement(amt);

        if (myOnHand.getValue() < 0) {
            throw new IllegalStateException("The amount on hand was set to a negative value: " + myOnHand.getValue());
        }

        if (myOnHand.getValue() < 1.0) {
            myStockOutIndicator.setValue(1.0);
        }

    }

    /** Allows subclasses to directly increment the on hand
     * @param amount The amount to increment by
     */
    protected final void incrementOnHand(int amt) {
        myOnHand.increment(amt);
        if (myOnHand.getValue() > 0.0) {
            myStockOutIndicator.setValue(0.0);
        }
    }

    /** Decrements the amount on order
     * @param amount The amount to decrement by
     */
    protected final void decrementOnOrder(int amount) {
        myOnOrder.decrement(amount);
        if (myOnOrder.getValue() < 0) {
            throw new IllegalStateException("The amount on order was set to a negative value: " + myOnOrder.getValue());
        }
    }

    /** Increments the amount on order
     *
     * @param amount
     */
    protected final void incrementOnOrder(int amount) {
        myOnOrder.increment(amount);
    }

    /** Decrements the amount pending
     * @param amount The amount to decrement by
     */
    protected final void decrementAmountPending(int amount) {
        myAmountPending = myAmountPending - amount;
        if (myAmountPending < 0) {
            throw new IllegalStateException("The amount pending was set to a negative value: " + myAmountPending);
        }

//		myAmountPending.decrement(amount);
//		if (myAmountPending.getValue() < 0)
//			throw new IllegalStateException("The amount pending was set to a negative value: " + myAmountPending.getValue());
    }

    /** Increments the amount pending
     *
     * @param amount
     */
    protected final void incrementAmountPending(int amount) {
        myAmountPending = myAmountPending + amount;
//		myAmountPending.increment(amount);
    }

    /** When an demand is first attempted to be filled, it can either
     *  be filled immediately in full, or not.  This method ensures the capture
     *  of statistics on the probability of first fill.
     *
     * @param flag
     */
    protected final void setFirstFillFlag(boolean flag) {
        myFirstFillRate.setValue(flag);
    }

    /** Gets the requester that assists with replenishment demands for internal inventory
     * @return the replenishment requester
     */
    protected final ReplenishmentRequesterIfc getReplenishmentRequester() {
        return myReplenishmentRequester;
    }

    /** Sets the requester that assists with replenishment demands for internal inventory
     * @param requester the ReplenishmentRequesterAbstract to set
     */
    protected final void setReplenishmentRequester(ReplenishmentRequesterIfc requester) {
        myReplenishmentRequester = requester;
    }

    /** A key method that subclasses may want to override.  This method is
     *  responsible for requesting the supplied amount for a replenishment.
     *  The current implementation simply gets a DemandFillerIfc from the
     *  replenishment finder and then creates a demand to be sent to the filler.
     *  Alternatively, the user can supply a DemandRequesterAbstract to handle
     *  the generation of the replenishment request.
     *
     * @param replenishmentQty
     */
    protected void requestReplenishment(int replenishmentQty) {

        myReplenishmentRequestCounter.increment();

        // create the demand
        Demand demand = createDemand(replenishmentQty);
        
        demand.setTotalArrivedDemand(myDemandArrivalCounter);

        demand.setDeliveredListener(myDemandListener);//TODO
        
        // assume that inventory will use its own listener
        // these might be changed by the requester
        setDemandListeners(demand);//TODO

        if (myReplenishmentRequester != null) {
            myReplenishmentRequester.requestReplenishment(this, demand);
        } else {
            sendDemand(demand);
        }

    }

    /** Ensures the creation of a replenishment demand for the proper amount
     *  and requirements for this inventory
     *
     * @param replenishmentQty
     * @return
     */
    protected Demand createDemand(int replenishmentQty) {
        // create the demand
        Demand demand = new Demand(getItemType(), replenishmentQty);
        //JSL.out.println(getTime() + " > Demand: " + demand.getId() + " requesting replenishment inventory: " + replenishmentQty);
        //JSL.out.println(this);
        incrementAmountPending(replenishmentQty);
        // set the requirements of the demand
        demand.setAllowBackLoggingFlag(mayRequestBackLoggingOfDemands());
        demand.setAllowPartialFillingFlag(mayRequestPartialFilling());
        return (demand);
    }

    /** Ensures that the replenishment filler (if found) is called to receive the demand
     *
     * @param demand
     */
    protected void sendDemand(Demand demand) {
        demand.setSender(this);
        // set the filler
        setFillerOnDemand(demand);
        // set the listeners
        setDemandListeners(demand);
        // place the demand in the SENT state
        demand.sent();
        // get the demand's filler
        DemandFillerIfc filler = demand.getFiller();
        // tell the filler to receive the demand
        filler.receive(demand);
    }

    /** Determines an appropriate replenishment filler for this inventory
     *  May throw a NoDemandFillerFoundException if no filler can be found.
     *
     * @param demand
     */
    protected void setFillerOnDemand(Demand demand) {
        // if the filler is set, use it, otherwise try to find a filler using the filler finder
        DemandFillerIfc filler;
        if (myDemandFiller == null) {
            filler = myDemandFillerFinder.findDemandFiller(demand);
            if (filler == null) {
                throw new NoDemandFillerFoundException("No demand filler was found for demand: " + demand);
            }
        } else {
            filler = myDemandFiller;
        }

        // set the demand filler
        demand.setFiller(filler);

    }

    /** This method sets the listeners (delivery, cancelled, received, rejected) to
     *  the inventory's internal listeners so that the inventory handles
     *  these state changes itself.
     *
     * @param demand
     */
    protected void setDemandListeners(Demand demand) {
        demand.setReceivedListener(myDemandListener);//TODO
        demand.setCancelledListener(myDemandListener);
        demand.setRejectedListener(myDemandRejectedListener);
    }

    protected final DemandListenerReceivedIfc getDemandReceivedListener() {
        return (myDemandListener);
    }

    protected final DemandListenerDeliveredIfc getDemandDeliveredListener() {
        return (myDemandListener);
    }

    protected final DemandListenerCancelledIfc getDemandCancelledListener() {
        return (myDemandListener);
    }

    protected final DemandListenerRejectedIfc getDemandRejectedListener() {
        return (myDemandRejectedListener);
    }

    /** Called internally by fillDemand() when a demand can be filled immediately
     *
     * @param demand
     */
    protected void fillImmediately(Demand demand) {
        //demand.process();
        // initial request can be filled immediately
        // set the fill indicator to indicate a successful first fill
        setFirstFillFlag(true);
        // give the request everything it asked for
        int amt = demand.getOriginalAmountDemanded();
        // fill by the amount supplied
        decrementOnHand(amt);
        // give the units to the demand
        demand.fill(amt);
        // check inventory after decrement
        checkInventory();
    }

    /** Called internally by fillDemand() when a new demand must
     *  be backlogged.
     *
     * @param demand
     */
    protected void backLogInitialDemand(Demand demand) {
        //demand.process();
        // initial request cannot be filled immediately
        // set the fill indicator to indicate no successful first fill
        setFirstFillFlag(false);

        // determine how much to give the demand
        int amt = allocateInventory(demand);
        if (amt > 0) {
            // fill by the amount supplied
            decrementOnHand(amt);
            // give the units to the demand
            demand.fill(amt);
        }
        // ask the backlog policy to backlog the demand
        myBackLogPolicy.backlog(demand);
        // check inventory after decrement/backlog
        checkInventory();

    }

    /** Can be used to determine how much inventory to give to the demand
     *  This method does not give the units.
     *
     * @param demand
     * @return
     */
    protected int allocateInventory(Demand demand) {
        int onHand = getAmountOnHand();
        int amtNeeded = demand.getRemainingDemand();

        // determine the amount that can be supplied
        int amtSupplied = 0;
        // check if amount on hand can satisfy all the demand
        if (onHand >= amtNeeded) {
            amtSupplied = amtNeeded;
        } else {// cannot satisfy all of the request
            if (demand.getAllowPartialFillingFlag() == true) {
                // part of the demand can be filled
                // since demand > onhand, we can supply all remaining on hand
                amtSupplied = onHand;
            }
        }
        return amtSupplied;
    }

    /** This method is called when a replenishment demand from the inventory
     *  has been delivered.
     *
     * @param demand
     */
    protected void replenishmentDemandDelivered(Demand demand) {

        int replenishmentQty = demand.getAmountFilled();

        decrementOnOrder(replenishmentQty);
        incrementOnHand(replenishmentQty);

//	    JSL.out.println("##############################################");
//	    JSL.out.println("In Inventory.replenishmentDemandDelivered(Demand d)");
//	    JSL.out.println(getTime() + " > replenishment delivered " + replenishmentQty);
//		JSL.out.println(this);
//	    JSL.out.println("##############################################");
        
        if (mayBackLogDemands()) {
            myBackLogPolicy.fillBackLogs();
        }
        
        checkInventory();

        // notify replenishment arrival listener
        if (myReplenishmentArrivalListener != null){
            myReplenishmentArrivalListener.demandArrived(this, demand);
        }

        demand.nullify();
    }

    /**  When the inventory's replenishment request is received by a filler
     *   this method is called to ask the filler to fill the demand.
     *
     * @param demand
     */
    protected void demandRequestReceivedByFiller(Demand demand) {
//	    JSL.out.println("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^");
//	    JSL.out.println("In Inventory.demandRequestReceivedByFiller(Demand demand)");
//		JSL.out.println(getTime() + " > demandRequestReceivedByFiller ");
//		JSL.out.println(this);
//	    JSL.out.println("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^");
        int qty = demand.getOriginalAmountDemanded(); // argument could be int qty
        decrementAmountPending(qty);
        incrementOnOrder(qty);
//		JSL.out.println(getTime() + "> replenishment ordered " + qty);
//		JSL.out.println(this);
        if (!demand.isPartOfAnOrder()) {
            DemandFillerIfc filler = demand.getFiller();
            filler.fillDemand(demand);
        }
    }

    /** If the inventory's replenishment request is cancelled, then this method is called
     *  This would allow the inventory to react to a cancellation of one of its requests.
     *  The default behavior is to throw an illegal state exception
     *
     * @param demand
     */
    protected void replenishmentDemandCancelled(Demand demand) {
        throw new IllegalStateException("The inventory's replenishment request was cancelled and not handled");
    }

    /** Used by the backlog policy to cause a backlogged demand
     *  to be filled with the given amount.
     *
     * @param demand
     * @param amount
     */
    protected final void fillBackLog(Demand demand, int amount) {
        //JSL.out.println(getTime() + " > filled backlog for demand " + demand.getId() + " for amount " + amount);
        // fill by the amount supplied
        decrementOnHand(amount);
        // give the units to the demand
        demand.fill(amount);
        // check inventory after decrement
        checkInventory();

    }

    /**
     * @param mayPartiallyFillFlag The mayPartiallyFillFlag to set.
     */
    protected final void setMayPartiallyFillFlag(boolean mayPartiallyFillFlag) {
        myMayPartiallyFillFlag = mayPartiallyFillFlag;
    }

    //  ===========================================
    //      PRIVATE METHODS
    //  ===========================================
    //  ===========================================
    //      PRIVATE INNER CLASSES
    //  ===========================================
    
    private class ReplenishmentDemandListener implements DemandListenerReceivedIfc, DemandListenerDeliveredIfc,
            DemandListenerCancelledIfc {

        public void demandReceived(Demand demand) {
            Inventory.this.demandRequestReceivedByFiller(demand);
        }

        public void demandDelivered(Demand demand) {
            Inventory.this.replenishmentDemandDelivered(demand);
        }

        public void demandCancelled(Demand demand) {
            Inventory.this.replenishmentDemandCancelled(demand);
        }
    }

    /** Sets a listener to be called when a demand arrives
     *
     * @param demarrivallistener
     */
    public final void setDemandArrivalListener(InventoryDemandArrivalListenerIfc demarrivallistener) {
        myDemandArrivalListener = demarrivallistener;
    }

    /** Sets a listener to be called after a replenishment demand arrives
     *
     * @param demarrivallistener
     */
    public final void setReplenishmentDemandArrivalListener(InventoryDemandArrivalListenerIfc demarrivallistener) {
        myReplenishmentArrivalListener = demarrivallistener;
    }

    /** Changes/sets the current inventory policy parameters.  Immediately checks
     *  the policy for replenishment
     *
     * @param parameter
     * set the inventory policy parameter
     */
    public void setInventoryPolicyParameters(double[] parameter) {
        //System.out.println(">>>>>>>>>>>>>>>>>>>");
        //System.out.println("Before Updating the policy to "+parameter[0]+", "+parameter[1]);
        //System.out.println("Time: "+getTime());
        //System.out.println(toString());
        myInventoryPolicy.setPolicyParameters(parameter);
        myInventoryPolicy.checkInventory();
    //System.out.println(">>>>>>>>>>>>>>>>>>>");
    //System.out.println("");
    //System.out.print("At time: "+getTime());
    //System.out.println(" after Updating the policy to "+parameter[0]+", "+parameter[1]+" and Replenishment if required");
    //System.out.println(""+toString());
    //System.out.println(">>>>>>>>>>>>>>>>>>>");
    }

    /** Changes/sets the policy parameters used at the beginning of each replication
     * Changing these parameters affects the next replication to be run
     * @param initialPolicyParameters
     */
    public void setInitialPolicyParameters(double[] initialPolicyParameters) {
        myInventoryPolicy.setInitialPolicyParameters(initialPolicyParameters);
    }

    /**
     * @return
     * @see jslx.supplychain.inventorylayer.InventoryPolicyAbstract#getResetInitialParametersFlag()
     */
    public final boolean getResetInitialParametersFlag() {
        return myInventoryPolicy.getResetInitialParametersFlag();
    }

    /**
     * @return
     * @see jslx.supplychain.inventorylayer.InventoryPolicyAbstract#getResetInitialParametersWarningFlag()
     */
    public final boolean getResetInitialParametersWarningFlag() {
        return myInventoryPolicy.getResetInitialParametersWarningFlag();
    }

    /**
     * @param option
     * @see jslx.supplychain.inventorylayer.InventoryPolicyAbstract#setResetInitialParametersFlag(boolean)
     */
    public final void setResetInitialParametersFlag(boolean option) {
        myInventoryPolicy.setResetInitialParametersFlag(option);
    }

    /**
     * @param option
     * @see jslx.supplychain.inventorylayer.InventoryPolicyAbstract#setResetInitialParametersWarningFlag(boolean)
     */
    public final void setResetInitialParametersWarningFlag(boolean option) {
        myInventoryPolicy.setResetInitialParametersWarningFlag(option);
    }

    /** Response statistics for on hand inventory
     *
     * @return the OnHand
     */
    public final ResponseStatisticsIfc getOnHandResponseStatistics() {
        return myOnHand;
    }

    /** Response statistics for the stock out indicator
     *
     * @return the StockOutIndicator
     */
    public final ResponseStatisticsIfc getStockOutIndicatorResponseStatistics() {
        return myStockOutIndicator;
    }

    /** Response statistics for the amount on order
     *
     * @return the OnOrder
     */
    public final ResponseStatisticsIfc getOnOrderResponseStatistics() {
        return myOnOrder;
    }

    /** Response statistics for the first fill rate
     *
     * @return the FirstFillRate
     */
    public final ResponseStatisticsIfc getFirstFillRateResponseStatistics() {
        return myFirstFillRate;
    }

    /** Across replication statistics for the count of the number of replishment requests
     *
     * @return the ReplenishmentRequestCounter
     */
    public final AcrossReplicationStatisticIfc getReplenishmentRequestCounterStatistics() {
        return myReplenishmentRequestCounter;
    }
}
