/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;

/** An InventoryHolderAbstract is a subclass of DemandFillerAbstract.  It holds inventory
 *  by ItemType.  It automatically collects aggregate performance on the usage of its 
 *  internal inventory (total (on-hand, amount back ordered, number of back orders, on-order,
 *  percentage of time without stock, probability of any demand being filled upon a first
 *  request, average customer wait time).  It provides behavior to ensure that a supplied
 *  instance of a ReplenishmentRequesterAbstract will be used by all internal inventory to
 *  manage their requests for replenishment.  Thus, subclasses must supply an instance of
 *  ReplenishmentRequesterAbstract (preferably within their constructors) to ensure the proper
 *  replenishment of internal inventory.  No inventory can be added to a holder that does
 *  not have an instance of ReplenishmentRequesterAbstract.
 *  
 *  An InventoryHolderAbstract provides the basic capability of adding inventory for different
 *  ItemTypes
 *
 */
abstract public class InventoryHolderAbstract extends DemandFillerAbstract implements DemandSenderIfc, AvailabilityIfc, AggregateInventoryIfc {

	//  ===========================================
	//      OBJECT VARIABLES
	//  ===========================================

	/** A Map to hold the inventory
	 *  key = ItemType, object = Inventory
	 */	
	protected Map<ItemType, Inventory> myInventory;

	/** Collects statistics on the aggregate amount backordered
	 */
	protected AggregateTimeWeightedVariable myTotalAmtBackOrdered;

	/** Collects statistics on the aggregate number waiting in backorder
	 */
	protected AggregateTimeWeightedVariable myTotalNumBackOrdered;

	/** Collects statistics on the aggregate inventory
	 */
	protected AggregateTimeWeightedVariable myTotalOnHand;

	/** Collects statistics on the aggregate on order
	 */
	protected AggregateTimeWeightedVariable myTotalOnOrder;

	/** Collects statistics on the aggregate stockout
	 */
	protected AggregateTimeWeightedVariable myTotalStockOutIndicator;

	/** Collects statistics on the aggregate first fill rate
	 */
	protected ResponseVariableAverageObserver myAvgFirstFillRate;

	/** Collects statistics on the aggregate customer waiting time
	 */
	protected ResponseVariableAverageObserver myAvgCustomerWaitTime;

	/** Counts the total number of replenishment orders made by the holder
	 */
	protected AggregateCounter myTotalNumReplenishmentDemands;

	/** Used by internal inventory to send replenishment requests (demands) to the 
	 *  outside.  Subclasses must supply an instance of this prior to adding any inventory
	 */
	private ReplenishmentRequesterIfc myReplenishmentRequester;

	/** Creates an InventoryHolderAbstract with initial availability set to true
	 * 
	 * @param parent  The parent ModelElement
	 */
	public InventoryHolderAbstract(ModelElement parent) {
		this(parent, true, null);
	}

	/** Creates an InventoryHolderAbstract with initial availability set to true
	 * 
	 * @param parent  The parent ModelElement
	 * @param name  The name of the holder
	 */
	public InventoryHolderAbstract(ModelElement parent, String name) {
		this(parent, true, name);
	}

	/** Creates an InventoryHolderAbstract
	 * 
	 * @param parent  The parent ModelElement
	 * @param initialAvailability Whether or not the holder is available at the start of a replication
	 * @param name  The name of the holder
	 */
	public InventoryHolderAbstract(ModelElement parent, boolean initialAvailability, String name) {
		super(parent, initialAvailability, name);
		myInventory = new HashMap<ItemType, Inventory>();
		myTotalOnHand = new AggregateTimeWeightedVariable(this, getName() + " Total On Hand");
		myAvgFirstFillRate = new ResponseVariableAverageObserver(this, getName() + " Aggregate Fill Rate");
		myTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, getName() + " Aggregate Number of Stockouts");
		myTotalOnOrder = new AggregateTimeWeightedVariable(this, getName() + " Total On Order");
		myTotalNumReplenishmentDemands = new AggregateCounter(this, getName() + " Aggregate#ReplenishmentRequests");
		myTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "TotalAmtBackordered");
		myTotalNumBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "TotalNumBackordered");
		myAvgCustomerWaitTime = new ResponseVariableAverageObserver(this, getName() + "AvgCustWaitTime");
	}

	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================

	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append(getName() + "\n");
		for(Map.Entry<ItemType,Inventory> entry: myInventory.entrySet()){
			sb.append(entry.getValue());
		}
		return(sb.toString());
	}
	
	/** Returns an InventoryIfc that can be used to get information concerning
	 *  the state of the inventory associated with the supplied item type
	 *  
	 *  If the supplied item type is null, this method returns null
	 *  If the supplied item type is not carried by this filler, then this
	 *  method returns null.
	 * 
	 * @param type
	 * @return
	 */
	public InventoryIfc getInventoryInfo(ItemType type){
		return myInventory.get(type);
	}

	/** Returns an Inventory representing the item type at this holding point
	 *  
	 *  If the supplied item type is null, this method returns null
	 *  If the supplied item type is not carried by this filler, then this
	 *  method returns null.
	 * 
	 * @param type
	 * @return
	 */
	public Inventory getInventory(ItemType type){
		return myInventory.get(type);
	}
	
	/** Gets the number of item types carried
	 * 
	 * @return the number of item types
	 */
	public final int getNumberOfItemTypes() {
		return myInventory.size();
	}
	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.DemandFillerIfc#getItemTypes()
	 */
	public final Set<ItemType> getItemTypes(){
		return Collections.unmodifiableSet(myInventory.keySet());
	}
	
	/** Adds an inventory. The ItemType of the supplied
	 *  inventory must not have already been added
	 * 
	 * @param inventory Must not be null
	 */
	public final void addInventory(Inventory inventory){
		if (inventory == null)
			throw new IllegalArgumentException("The Inventory inventory must be non-null.");
		
		ItemType type = inventory.getItemType();
		
		if (myInventory.containsKey(type))
			throw new IllegalArgumentException("The IHP already contains an inventory for type: " + type.getName());
		
		// setup collection of aggregate statistics
		inventory.attachAggregateTimeWeightedVariableToOnHand(myTotalOnHand);
		inventory.attachResponseVariableAverageObserverToFirstFillRate(myAvgFirstFillRate);
		inventory.attachAggregateTimeWeightedVariableToStockOutIndicator(myTotalStockOutIndicator);
		inventory.attachAggregateTimeWeightedVariableToOnOrder(myTotalOnOrder);
		inventory.attachAggregateCounterToReplenishmentRequestCounter(myTotalNumReplenishmentDemands);
		
		BackLogPolicyAbstract bpolicy = inventory.getBackLogPolicy();
		if (bpolicy != null){
			bpolicy.attachAggregateTimeWeightedVariableToAmountBackLogged(myTotalAmtBackOrdered);
			bpolicy.attachAggregateTimeWeightedVariableToNumInBackLoggedQueue(myTotalNumBackOrdered);
			bpolicy.attachResponseVariableAverageObserverToTimeInQueue(myAvgCustomerWaitTime);		
		}
		
		// tell the inventory to go through the IHP		
		if (myReplenishmentRequester == null)
			throw new IllegalStateException("No replenishment requestor: a ReplenishmentRequesterAbstract must be supplied to the inventory holder");
		else
			inventory.setReplenishmentRequester(myReplenishmentRequester);
		
		myInventory.put(type, inventory);		
	}
	
	/** Removes the inventory associated with the supplied type from the holder
	 *  The holder must contain the type and the supplied type must not be null
	 *  This causes the inventory and the resulting type to no longer
	 *   be stocked (held) by this holder.
	 *   
	 *   Does not cause the removed inventory to be removed from the model
	 * 
	 * @param type
	 * @return
	 */
	public final Inventory removeInventory(ItemType type){
		return removeInventory(type, false);
	}
	
	/** Removes the inventory associated with the supplied type from the holder
	 *  The holder must contain the type and the supplied type must not be null
	 *  This causes the inventory and the resulting type to no longer
	 *   be stocked (held) by this holder.
	 * 
	 * @param type
	 * @param fromModelFlag indicates whether the associated inventory should be removed from the model
	 * @return
	 */
	public final Inventory removeInventory(ItemType type, boolean fromModelFlag){
		if (type == null)
			throw new IllegalArgumentException("The ItemType must be non-null.");
		
		if (!myInventory.containsKey(type))
			throw new IllegalArgumentException("The IHP does not hold " + type.getName());

		Inventory i = myInventory.get(type);

		removeInventory(i, fromModelFlag);
		
		return i;
		
	}
	
	/** Causes the supplied inventory to be removed from the holder
	 *  The inventory must not be null and must be held by the holder
	 *  This causes the inventory and the resulting type to no longer
	 *   be stocked (held) by this holder.
	 *   
	 *   Does not cause the removed inventory to be removed from the model
	 * 
	 * @param type
	 * @return
	 */
	public final void removeInventory(Inventory inventory){
		removeInventory(inventory, false);
	}
	
	/** Causes the supplied inventory to be removed from the holder
	 *  The inventory must not be null and must be held by the holder
	 *  This causes the inventory and the resulting type to no longer
	 *   be stocked (held) by this holder.
	 * 
	 * @param inventory
	 * @param fromModelFlag indicates whether the associated inventory should be removed from the model
	 */
	public final void removeInventory(Inventory inventory, boolean fromModelFlag){
		if (inventory == null)
			throw new IllegalArgumentException("The Inventory inventory must be non-null.");

		ItemType type = inventory.getItemType();
		
		Inventory i = myInventory.get(type);
		
		if (i == null)
			throw new IllegalArgumentException("The IHP does not hold " + type.getName());
		
		if (i != inventory)
			throw new IllegalArgumentException("The IHP does not hold the referenced inventory");
		
		myInventory.remove(type);
		
		// detach from aggregate statistics
		inventory.detachAggregateTimeWeightedVariableFromOnHand(myTotalOnHand);
		inventory.detachResponseVariableAverageObserverFromFirstFillRate(myAvgFirstFillRate);
		inventory.detachAggregateTimeWeightedVariableFromStockOutIndicator(myTotalStockOutIndicator);
		inventory.detachAggregateTimeWeightedVariableFromOnOrder(myTotalOnOrder);
		inventory.detachAggregateCounterFromReplenishmentRequestCounter(myTotalNumReplenishmentDemands);
		
		BackLogPolicyAbstract bpolicy = inventory.getBackLogPolicy();
		if (bpolicy != null){
			bpolicy.detachAggregateTimeWeightedVariableFromAmountBackLogged(myTotalAmtBackOrdered);
			bpolicy.detachAggregateTimeWeightedVariableFromNumInBackLoggedQueue(myTotalNumBackOrdered);
			bpolicy.detachResponseVariableAverageObserverFromTimeInQueue(myAvgCustomerWaitTime);		
		}
		
		inventory.setReplenishmentRequester(null);
		
		if (fromModelFlag == true){
			inventory.removeFromModel(); 
		    // bpolicy will be removed since it is a child
//			if (bpolicy != null)
//				bpolicy.removeFromModel();
		}
	}

	/** Adds a (reorder point, reorder quantity) based inventory that permits backlogging,
	 *  with no initial inventory
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderQty
	 * @return
	 */
	public final Inventory addReorderPointReorderQuantityInventory(ItemType type, int reorderPt, int reorderQty){
		return(addReorderPointReorderQuantityInventory(type, reorderPt, reorderQty, 0, null));
	}
	
	/** Adds a (reorder point, reorder quantity) based inventory that permits backlogging
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderQty
	 * @param initialLevel
	 * @return
	 */
	public final Inventory addReorderPointReorderQuantityInventory(ItemType type, int reorderPt, int reorderQty, int initialLevel){
		return(addReorderPointReorderQuantityInventory(type, reorderPt, reorderQty, initialLevel, null));
	}
	
	/** Adds a (reorder point, reorder quantity) based inventory that permits backlogging
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderQty
	 * @param initialLevel
	 * @param name
	 * @return
	 */
	public final Inventory addReorderPointReorderQuantityInventory(ItemType type, int reorderPt, int reorderQty, int initialLevel, String name){
		if (name == null)
			name = getName() + "-Inventory-" + type.getName();
		Inventory inventory = Inventory.createReorderPointReorderQuantityInventory(this, type, reorderPt, reorderQty, initialLevel, name);
		addInventory(inventory);
		return(inventory);
	}

	/** Adds a (reorder point, reorder level) based inventory that permits backlogging,
	 *  with no initial inventory
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderLevel
	 * @return
	 */
	public final Inventory addReorderPointOrderUpToLevelInventory(ItemType type, int reorderPt, int reorderLevel){
		return(addReorderPointOrderUpToLevelInventory(type, reorderPt, reorderLevel, 0, null));
	}
	
	/** Adds a (reorder point, reorder level) based inventory that permits backlogging
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderLevel
	 * @param initialLevel
	 * @return
	 */
	public final Inventory addReorderPointOrderUpToLevelInventory(ItemType type, int reorderPt, int reorderLevel, int initialLevel){
		return(addReorderPointOrderUpToLevelInventory(type, reorderPt, reorderLevel, initialLevel, null));
	}
	
	/** Adds a (reorder point, reorder level) based inventory that permits backlogging
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderLevel
	 * @param initialLevel
	 * @param name
	 * @return
	 */
	public final Inventory addReorderPointOrderUpToLevelInventory(ItemType type, int reorderPt, int reorderLevel, int initialLevel, String name){
		if (name == null)
			name = getName() + "-" + type.getName();
		Inventory inventory = Inventory.createReorderPointOrderUpToLevelInventory(this, type, reorderPt, reorderLevel, initialLevel, name);
		addInventory(inventory);
		return(inventory);
	}

	//TODO create some convenience methods to allow the creation and addition of different types of Inventory with other policies

	/* (non-Javadoc)
	 * @see jsl.supplychain.inventorylayer.DemandSenderIfc#mightRequest(jsl.supplychain.inventorylayer.ItemType)
	 */
	public final boolean mightRequest(ItemType type) {
		return myInventory.containsKey(type);
	}

	/* (non-Javadoc)
	 * @see jsl.supplychain.inventorylayer.DemandFillerIfc#negotiate(jsl.supplychain.inventorylayer.Demand)
	 */
	public DemandMessageIfc negotiate(Demand demand) {
		
		if (!isAvailable()){
			return(null);			
		}	
				
		Inventory inventory = myInventory.get(demand.getItemType());

		if (inventory == null)
			return null;
		else
			return inventory.negotiate(demand);
	}
	
	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.DemandFillerIfc#canFillItemType(jsl.supplychainNew.inventorylayer.Demand)
	 */
	public boolean canFillItemType(Demand demand){
		return(myInventory.containsKey(demand.getItemType()));
	}
	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.DemandFillerIfc#canFillItemType(jsl.research.supplychain.inventorylayer.ItemType)
	 */
	public boolean canFillItemType(ItemType type){
		return(myInventory.containsKey(type));
	}
	
	/** Given a demand, this method will return an integer that represents
	 *  what will happened to the demand if it is requested to be filled at
	 *  the current time.  The status is only valid at the current simulation time.
	 * 
	 * @param demand
	 * @return
	 */
	public int determineRequestStatus(Demand demand){
		Inventory inventory = myInventory.get(demand.getItemType());
		if (inventory == null)
			return Demand.ITEMTYPE_MISMATCH;
		else
			return (inventory.determineRequestStatus(demand));
	}
	
	/** Given a demand, this method indicates whether or not the filler
	 *  will reject the demand if it is requested to be filled at the current time.
	 *  True means that it will be rejected, false means that it will be received w/o rejection
	 * @param demand
	 * @return
	 */
	public boolean willReject(Demand demand){
		Inventory inventory = myInventory.get(demand.getItemType());
		if (inventory == null)
			return true;
		else
			return (inventory.willReject(demand));
	}
	
	/**
	 * @return Returns the Aggregate Avg Customer Wait Time ResponseVariable.
	 */
	public final ResponseVariableAverageObserver getAggregateAvgCustomerWaitTimeResponseVariable() {
		return myAvgCustomerWaitTime;
	}
	
	/**
	 * @return Returns the get Aggregate Avg First FillRate ResponseVariable.
	 */
	public final ResponseVariableAverageObserver getAggregateAvgFirstFillRateResponseVariable() {
		return myAvgFirstFillRate;
	}
	
	/**
	 * @return Returns the Aggregate Amount BackOrdered ResponseVariable.
	 */
	public final AggregateTimeWeightedVariable getAggregateAmountBackOrderedResponseVariable() {
		return myTotalAmtBackOrdered;
	}
	
	/**
	 * @return Returns the Aggregate Number BackOrdered ResponseVariable.
	 */
	public final AggregateTimeWeightedVariable getAggregateNumberBackOrderedResponseVariable() {
		return myTotalNumBackOrdered;
	}
	
	/**
	 * @return Returns the Aggregate OnHand Inventory ResponseVariable.
	 */
	public final AggregateTimeWeightedVariable getAggregateOnHandInventoryResponseVariable() {
		return myTotalOnHand;
	}
	
	/**
	 * @return Returns the Aggregate OnOrder ResponseVariable.
	 */
	public final AggregateTimeWeightedVariable getAggregateAmountOnOrderResponseVariable() {
		return myTotalOnOrder;
	}
	
	/** 
	 * 
	 * @return Returns a counter for the total number of replenishment orders made at this IHP
	 */
	public final AggregateCounter getAggregateNumberOfReplenishmentDemandsCounter(){
		return(myTotalNumReplenishmentDemands);
	}
	
	/**
	 * @return Returns the Aggregate StockOut Percentage ResponseVariable.
	 */
	public final AggregateTimeWeightedVariable getAggregateStockOutPercentageResponseVariable() {
		return myTotalStockOutIndicator;
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#attachAggregateAvgCustomerWaitTime(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void subscribeToAggregateAvgCustomerWaitTime(ResponseVariableAverageObserver aggregate) {
		aggregate.subscribeTo(myAvgCustomerWaitTime);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#attachAggregateAvgFirstFillRate(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void subscribeToAggregateAvgFirstFillRate(ResponseVariableAverageObserver aggregate) {
		aggregate.subscribeTo(myAvgFirstFillRate);
	}
	
	/**
	 * @param aggregate The aggregate time weighted variable to attach
	 */
	public void subscribeToAggregateAmountBackOrdered(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myTotalAmtBackOrdered);
	}

	/**
	 * @param aggregate The aggregate time weighted variable to attach
	 */
	public void subscribeToAggregateNumberBackOrdered(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myTotalNumBackOrdered);
	}

	/**
	 * @param aggregate The aggregate time weighted variable to attach
	 */
	public void subscribeToAggregateOnHandInventory(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myTotalOnHand);
	}

	/**
	 * @param aggregate The aggregate time weighted variable to attach
	 */
	public void subscribeToAggregateAmountOnOrder(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myTotalOnOrder);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#attachAggregateReplenishmentDemandCounter(jsl.modeling.elements.variable.AggregateCounter)
	 */
	public void subscribeToAggregateReplenishmentDemandCounter(AggregateCounter counter) {
		counter.subscribeTo(myTotalNumReplenishmentDemands);
	}

	/**
	 * @param aggregate The aggregate time weighted variable to attach
	 */
	public void subscribeToAggregateStockOutIndicator(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myTotalStockOutIndicator);
	}
	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateAvgCustomerWaitTime(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void unsubscribeFromAggregateAvgCustomerWaitTime(ResponseVariableAverageObserver aggregate) {
		aggregate.unsubscribeFrom(myAvgCustomerWaitTime);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateAvgFirstFillRate(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void unsubscribeFromAggregateAvgFirstFillRate(ResponseVariableAverageObserver aggregate) {
		aggregate.unsubscribeFrom(myAvgFirstFillRate);
	}
	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateAmountBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateAmountBackOrdered(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myTotalAmtBackOrdered);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateNumberBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateNumberBackOrdered(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myTotalNumBackOrdered);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateOnHandInventory(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateOnHandInventory(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myTotalOnHand);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateAmountOnOrder(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateAmountOnOrder(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myTotalOnOrder);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateReplenishmentDemandCounter(jsl.modeling.elements.variable.AggregateCounter)
	 */
	public void unsubscribeFromAggregateReplenishmentDemandCounter(AggregateCounter counter) {
		counter.unsubscribeFrom(myTotalNumReplenishmentDemands);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateStockOutIndicator(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateStockOutIndicator(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myTotalStockOutIndicator);
	}

	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	/** Gets the requester that assists with replenishment demands for internal inventory
	 * @return the replenishment requester
	 */
	protected final ReplenishmentRequesterIfc getReplenishmentRequester() {
		return myReplenishmentRequester;
	}

	/** Sets the requester that assists with replenishment demands for internal inventory
	 * @param requester the ReplenishmentRequesterAbstract to set
	 */
	protected final void setReplenishmentRequester(ReplenishmentRequesterIfc requester) {//TODO maybe make this abstract
		myReplenishmentRequester = requester;
	}

	//  ===========================================
	//      PRIVATE INNER CLASSES
	//  ===========================================
	

}
