/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import jsl.modeling.ModelElement;

/** An InventoryHoldingPoint is a subclass of InventoryHolderAbstract. 
 *  An InventoryHoldingPoint (IHP) holds instances of Inventory. It acts as a DemandFiller
 *  by delegating incoming demand requests to its internal inventory. 
 *  
 *  It provides behavior that allows the internal inventory to be replenished through the IHP.  
 *  This is accomplished by the user supplying either an instance of a DemandFillerFinderIfc that 
 *  every internal inventory will use to find an appropriate DemandFillerIfc for replensishment or by providing 
 *  an instance of a DemandFillerIfc directly.  Once a DemandFillerIfc is directly supplied
 *  then any DemandFillerFinder that may have been supplied WILL NOT be used (thereby bypassing
 *  any search for a filler).  
 *  
 *  The IHP also listens for when incoming demands are filled by its
 *  internal inventory so that it can either use a DemandCarrierIfc (if supplied) to ensure the
 *  transport of the demand back to the sender, or if no DemandCarrierIfc is supplied, 
 *  it directly transports the demand back to the customer with no delay.  A side effect of this
 *  behavior is that the IHP becomes the fill listener for any demands that are sent to it.
 *
 *
 */
public class InventoryHoldingPoint extends InventoryHolderAbstract {

	/** Listens for the demands to be filled by internal inventory and then reacts by shipping them
	 */
	protected DemandListenerFilledIfc myDemandFilledListener = new DemandFilledListener();

	/** Responsible for finding an appropriate (replenishment) filler for the holder.  
	 *  Finds (replenishment) fillers for the inventory within this holder
	 */
	private DemandFillerFinderIfc myDemandFillerFinder;
	
	/** This attribute can be used to directly set the filler for the holder
	 *  This is the filler for the inventory within this holder
	 */
	private DemandFillerIfc myDemandFiller;

	/** Creates the InventoryHoldingPoint with initial availability set to true
	 * 
	 * @param parent The parent ModelElement
	 */
	public InventoryHoldingPoint(ModelElement parent) {
		this(parent, true, null);
	}

	/** Creates the InventoryHoldingPoint with initial availability set to true
	 * 
	 * @param parent The parent ModelElement
	 * @param name The name for the IHP
	 */
	public InventoryHoldingPoint(ModelElement parent, String name) {
		this(parent, true, name);
	}
		
	/** Creates the InventoryHoldingPoint with initial availability set to true
	 * 
	 * @param parent The parent ModelElement
	 * @param initialAvailability Indicates whether or not the IHP will be available at the start of each replication
	 * @param name The name for the IHP
	 */
	public InventoryHoldingPoint(ModelElement parent, boolean initialAvailability, String name) {
		super(parent, initialAvailability, name);
		// sets the replenishment requester that prepares demands to be sent
		// out for replenishment for internal inventory
		setReplenishmentRequester(new IHPReplenishmentRequester());//TODO make a better design for sub-classes
	}

	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================

	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.DemandFillerIfc#receive(jsl.supplychainNew.inventorylayer.Demand)
	 */
	public final void receive(Demand demand) {

		// determine whether to reject or process the demand
		
		if (!isAvailable()){
			demand.setDemandStatus(Demand.FILLER_UNAVAILABLE);
			demand.reject();
			return;			
		}
		
		Inventory inventory = myInventory.get(demand.getItemType());
		
		if (inventory == null){
			demand.setDemandStatus(Demand.ITEMTYPE_MISMATCH);
			demand.reject();
			return;
		}
		// now ask the inventory to receive the demand
		inventory.receive(demand);
		
	}

	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.DemandFillerIfc#fillDemand(jsl.supplychainNew.inventorylayer.Demand)
	 */
	public final void fillDemand(Demand demand) {
		// set the IHP as the fill listener for the demand
		demand.setFilledListener(myDemandFilledListener);

		/*NOTE We could assign inventory as the filler in receive
		 	this would save the look up from the hash map; however, the fillDemand()
		 	method on the IHP would never be called since the fillDemand() method
		 	on the internal inventory would instead be called.  This is left this way
		 	because the inventory doesn't become the actual filler until the fillDemand()
		 	method is called
		 */
		Inventory inventory = myInventory.get(demand.getItemType());
		inventory.fillDemand(demand);
	}

	/** The supplied finder is responsible for finding an appropriate filler
	 *  so that the holders's demand requests can be fulfilled.
	 * 
	 * @param finder Must not be null
	 */
	public final void setDemandFillerFinder(DemandFillerFinderIfc finder){
		if (finder == null)
			throw new IllegalArgumentException("Invalid argument. DemandFillerFinder must be non-null.");
		myDemandFillerFinder = finder;
	}
	
	/**  Allows access the holder's filler finder
	 * 
	 * @return
	 */
	public final DemandFillerFinderIfc getDemandFillerFinder(){
		return(myDemandFillerFinder);
	}
	
	/** Can be used to directly set the filler for this holder, rather than relying
	 *  on the filler finder. Once set the DemandFillerFinder WILL NOT and CANNOT be
	 *  used.  Thus, this filler must be used and can only be changed through this method
	 * 
	 * @param fillerFinder Must not be null
	 */
	public final void setDemandFiller(DemandFillerIfc filler) {
		if (filler == null)
			throw new IllegalArgumentException("Invalid argument. DemandFillerIfc must be non-null.");
		
		myDemandFiller= filler;
	}
	
	/** Gets the IHP's demand filler if it has been directly set. A null
	 *  returned filler does not mean that the IHP does not have a filler. It
	 *  only means that it has not be directly set, since the IHP may be getting
	 *  its filler through the use of its DemandFillerFinderIfc
	 * 
	 * @return  may be null
	 */
	public final DemandFillerIfc getDemandFiller(){
		return(myDemandFiller);
	}
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================


	//  ===========================================
	//      PRIVATE INNER CLASSES
	//  ===========================================

	/** Listens for incoming demands to be filled by internal inventory
	 *  When the demand is filled, the demand is either shipped by
	 *  a supply DemandCarrierIfc or is shipped and delivered immediately
	 */
	private final class DemandFilledListener implements DemandListenerFilledIfc {

		public void demandFilled(Demand demand) {

			if (myExternalDemandCarrier != null)
				myExternalDemandCarrier.transportDemand(demand);
			else {
				demand.ship();
				demand.deliver();
			}				
		}
		
	}
	
	/** Used by all the internal inventory within the holder to send
	 *  demand (replenishment) requests to the outside
	 *
	 */
	private final class IHPReplenishmentRequester implements ReplenishmentRequesterIfc {

		public void requestReplenishment(Inventory inventory, Demand demand) {
			// An internal inventory is calling to be replenished
			// go out to the IHP's filler and use it
			
			// set the sender of the demand to the holder so that external fillers
			// use this holder as the sender
			demand.setSender(InventoryHoldingPoint.this);
			
			// if the filler is set, use it, otherwise try to find a filler using the filler finder
			DemandFillerIfc filler;
			if (myDemandFiller == null){
				filler = myDemandFillerFinder.findDemandFiller(demand);
				if (filler == null)
					throw new NoDemandFillerFoundException("No demand filler was found for demand: " + demand);
			} else
				filler = myDemandFiller;
			
			// set the filler on the demand so it uses the one suggested by the IHP
			demand.setFiller(filler);
			//set the demand listeners, have the internal inventory supply the listeners
			inventory.setDemandListeners(demand);
			// place the demand in the SENT state
			demand.sent();
			// tell the filler to receive the demand
			filler.receive(demand);			
		}
		
	}
	
}
