/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.utilities.reporting.JSL;

/**
 *
 */
public abstract class InventoryPolicyAbstract extends SchedulingElement {
	
	private Inventory myInventory;
	
    /** The initial values of the parameters of the policy
     *  These are remembered to allow the policy
     *  to be reset to the initial parameters at the beginning of 
     *  each replication (if they were changed during the replication)
     * 
     */
	protected double[] myInitialPolicyParameters;

    /** During a replication, the client may change the parameters of
     *  the policy.  This flag indicates whether or not
     *  the original parameters of the policy should be used
     *  to reset the parameters of the policy to the originals.
     *  The default option is true (it will reset the parameters).
     *  This is only applicable if the parameters of the policy changes. 
     * 
     */
    protected boolean myResetInitialParametersFlag = true;
    
    /** If the parameters of the policy were changed during the replication
     *  and reset initial parameters flag is false, then a warning message will
     *  occur.  This flag turns off that default warning message.
     * 
     */
    protected boolean myResetInitialParametersWarningFlag = true;
    
    /** Indicates if the policy parameters were changed at least once
     *  during a replication
     * 
     */
    protected boolean myPolicyChangedDuringRepFlag = false;

	/**
	 * @param parent
	 */
	public InventoryPolicyAbstract(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public InventoryPolicyAbstract(ModelElement parent, String name) {
		super(parent, name);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public InventoryPolicyAbstract(ModelElement parent, double[] initialPolicyParameters, String name) {
		super(parent, name);
		setInitialPolicyParameters(initialPolicyParameters);
	}
	
	/**
	 * @return
	 */
	protected final Inventory getInventory() {
		return myInventory;
	}

	/**
	 * @param inventory
	 */
	protected final void setInventory(Inventory inventory) {
		if (inventory == null)
			throw new IllegalArgumentException("Inventory inventory was null!");    

		myInventory = inventory;
	}

	/**
	 * @return
	 */
	protected final int getInventoryPosition() {
		return myInventory.getInventoryPosition();
	}
	
	/**
	 * @param replenishmentQty
	 */
	protected final void requestReplenishment(int replenishmentQty) {
		myInventory.requestReplenishment(replenishmentQty);
	}

    /** Returns whether the reset initial parameters flag is true/false
     * 
     * @return
     */
    public final boolean getResetInitialParametersFlag(){
    	return myResetInitialParametersFlag;
    }
    
    /** Sets the option to reset initial parameters to original values
     *  for the policy
     * 
     * @param option
     */
    public final void setResetInitialParametersFlag(boolean option){
    	myResetInitialParametersFlag = option;
    }
   
    /** Returns whether the warning message is on 
     * 
     * @return
     */
    public final boolean getResetInitialParametersWarningFlag(){
    	return myResetInitialParametersWarningFlag;
    }
    
    /** Sets the option have a warning message if resetting the initial parameters to original values
     *  for the policy
     * 
     * @param option
     */
    public final void setResetInitialParametersWarningFlag(boolean option){
    	myResetInitialParametersWarningFlag = option;
    }
    
	/** Returns the initial policy parameters as an array
	 * 
	 * @return
	 */
	public double[] getInitialPolicyParameters(){
		double[] p = new double[myInitialPolicyParameters.length];
		System.arraycopy(myInitialPolicyParameters, 0, p, 0, myInitialPolicyParameters.length);
		return p;
	}

	/** Sets the initial values for the policy parameters for each replication. Changing
	 *  these parameters affects the next replication to be run
	 * 
	 * @param initialPolicyParameters
	 */
	abstract public void setInitialPolicyParameters(double[] initialPolicyParameters);
	
	/** Implements the behavior to check the inventory to determine if an order is necessary
	 * 
	 */
	abstract protected void checkInventory();

	/** Allows for the policy parameters to be changed during a replication in a generic
	 *  manner.
	 * @param parameters
	 */
	abstract protected void setPolicyParameters(double [] parameters);
	
	/** Gets the current parameter values of the policy as an array
	 * 
	 * @return
	 */
	abstract public double[] getPolicyParameters();
	
	protected void beforeReplication(){
		if (myResetInitialParametersFlag)
			setPolicyParameters(myInitialPolicyParameters);
		else {
			if (myPolicyChangedDuringRepFlag == true){
				// reset was off and policy change occurred during replication, make warning
       			if (myResetInitialParametersWarningFlag){
    	   			JSL.LOGGER.warning("The parameters for the policy " + getName() + " were changed during the replication");       				
    	   			JSL.LOGGER.warning("The parameters were not reset to the values available from the last call to setInitialPolicyParameters() for the next replication.");        				
    			}
       			myPolicyChangedDuringRepFlag = false;
			}
		}
	}
	
	protected void removedFromModel(){
		super.removedFromModel();
		myInventory = null;
		myInitialPolicyParameters = null;
	}

}
