/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import jsl.modeling.ModelElement;

/**
 *
 */
public class InventoryPolicyReorderPointOrderUpToLevel extends InventoryPolicyAbstract{

	/** The reorder point.  When this is reached, an amount of inventory is ordered to
	 *  get back up to the order up to point. Must be < order up to level
	 * 
	 */
	protected int myReorderPoint;
	
	/** The maximum level of inventory.  Must be >= 1
	 * 
	 */
	protected int myOrderUpToPoint;
	
	/**
	 * @param parent
	 */
	public InventoryPolicyReorderPointOrderUpToLevel(ModelElement parent) {
		this(parent, 0, 1, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public InventoryPolicyReorderPointOrderUpToLevel(ModelElement parent, int reorderPoint, int orderUpToPoint) {
		this(parent, reorderPoint, orderUpToPoint, null);
	}

	/**
	 * parameters[0] reorderPoint the reorder point must be < order up to point
	 * parameters[1] orderUpToPoint the orderUpToPoint must be >= 1
	 * 
	 * @param parent
	 * @param parameters
	 * @param name
	 */
	public InventoryPolicyReorderPointOrderUpToLevel(ModelElement parent, double[] parameters, String name){
		this(parent, (int)parameters[0], (int)parameters[1], name);
	}

	/**
	 * parameters[0] reorderPoint the reorder point must be < order up to point
	 * parameters[1] orderUpToPoint the orderUpToPoint must be >= 1
	 * 
	 * @param parent
	 * @param parameters
	 */
	public InventoryPolicyReorderPointOrderUpToLevel(ModelElement parent, double[] parameters){
		this(parent, (int)parameters[0], (int)parameters[1], null);
	}
	/**
	 * @param parent
	 * @param name
	 */
	public InventoryPolicyReorderPointOrderUpToLevel(ModelElement parent, int reorderPoint, int orderUpToPoint, String name) {
		super(parent, name);
		setInitialPolicyParameters(reorderPoint, orderUpToPoint);
		myReorderPoint = reorderPoint;
		myOrderUpToPoint = orderUpToPoint;		
	}

	/* (non-Javadoc)
	 * @see jsl.examples.inventory.InventoryPolicyAbstract#checkInventory(jsl.examples.inventory.Inventory)
	 */
	protected void checkInventory() {
		
		if (getInventoryPosition()<= myReorderPoint){
			int orderSize = (int)(myOrderUpToPoint - getInventoryPosition());
			requestReplenishment(orderSize);
		}	

	}

	/**
	 * @return Returns the reorderPoint.
	 */
	public final int getReorderPoint() {
		return myReorderPoint;
	}

	/**
	 * @return Returns the myOrderUpToPoint.
	 */
	public final int getOrderUpToPoint() {
		return myOrderUpToPoint;
	}
	
	/** Sets the policy parameters
	 * 
	 * @param reorderPoint the reorder point must be < order up to point
	 * @param orderUpToPoint the  must be >= 1
	 */
	protected final void setPolicyParameters(int reorderPoint, int orderUpToPoint) {
		if (orderUpToPoint < 1)
			throw new IllegalArgumentException("The order up to point must be >=1");

		if (reorderPoint >= orderUpToPoint)
			throw new IllegalArgumentException("The reorder point must be < order up to point");

		myReorderPoint = reorderPoint;
		myOrderUpToPoint = orderUpToPoint;
		
		if (isCurrentReplicationRunning())
			myPolicyChangedDuringRepFlag = true;

	}

	/** Sets the initial policy parameters for the next replication
	 * 
	 * parameters[0] reorderPoint the reorder point must be < order up to point
	 * parameters[1] orderUpToPoint the orderUpToPoint must be >= 1

	 * @param parameters must not be null
	 */
	public final void setInitialPolicyParameters(double[] parameters) {
		setInitialPolicyParameters((int)parameters[0], (int)parameters[1]);		
	}

	/** Sets the initial policy parameters
	 * 
	 * @param reorderPoint the reorder point must be < order up to point
	 * @param orderUpToPoint the  must be >= 1
	 */
	public final void setInitialPolicyParameters(int reorderPoint, int orderUpToPoint) {
		if (orderUpToPoint < 1)
			throw new IllegalArgumentException("The order up to point must be >=1");

		if (reorderPoint >= orderUpToPoint)
			throw new IllegalArgumentException("The reorder point must be < order up to point");

		myInitialPolicyParameters = new double[2];
		myInitialPolicyParameters[0] = reorderPoint;
		myInitialPolicyParameters[1] = orderUpToPoint;
	}
	
	/** Gets the current values of the policy parameters as an array
	 * parameters[0] reorderPoint
	 * parameters[1] orderUpToPoint
	 * 
	 * @return 
	 */
	public double[] getPolicyParameters() {
		double [] para = {this.getReorderPoint(),this.getOrderUpToPoint()};
		return para;
	}

	/** Sets the policy parameters for the current replication
	 * 
	 * parameters[0] reorderPoint the reorder point must be < order up to point
	 * parameters[1] orderUpToPoint the orderUpToPoint must be >= 1
	 * 
	 * @param parameters must not be null
	 */
	protected void setPolicyParameters(double[] parameter) {		
		setPolicyParameters((int)parameter[0],(int)parameter[1]);	
	}
}
