/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.supplychain.inventorylayer;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;

/**
 * @author rossetti
 *
 */
public class InventoryPolicyReorderPointOrderUpToLevelPeriodic extends	InventoryPolicyAbstract  {

	/** The reorder point.  When this is reached, an amount of inventory is ordered to
	 *  get back up to the order up to point. Must be < order up to level
	 * 
	 */
	protected int myReorderPoint;
	
	/** The maximum level of inventory.  Must be >= 1
	 * 
	 */
	protected int myOrderUpToPoint;

	/** The time between reviews, must be > 0
	 * 
	 */
	protected double myReviewPeriod;
	
	/** The time of the initial review, must be >= 0
	 * 
	 */
	protected double myInitialReviewTime;

	/** Creates a periodic (reorder point, reorder up to) policy
	 * initialReviewTime The time of the first review = 0.0
	 * reviewPeriod The time between reviews = 1.0
	 * reorderPoint the reorder point = 0
	 * orderUpToPoint = 1

	 * @param parent
	 */
	public InventoryPolicyReorderPointOrderUpToLevelPeriodic(ModelElement parent) {
		this(parent, 0, 1, 1.0, 0.0, null);
	}

	/**
	 * parameters[0] reorderPoint the reorder point must be < order up to point
	 * parameters[1] orderUpToPoint the orderUpToPoint must be >= 1
	 *  parameters[2] = reviewPeriod > 0.0
	 *  parameters[3] = initialReviewTime >= 0.0 
	 * 
	 * @param parent
	 * @param parameters
	 * @param name
	 */
	public InventoryPolicyReorderPointOrderUpToLevelPeriodic(ModelElement parent, double[] parameters, String name){
		this(parent, (int)parameters[0], (int)parameters[1], parameters[2], parameters[3], name);
	}

	/**
	 * parameters[0] reorderPoint the reorder point must be < order up to point
	 * parameters[1] orderUpToPoint the orderUpToPoint must be >= 1
	 *  parameters[2] = reviewPeriod > 0.0
	 *  parameters[3] = initialReviewTime >= 0.0 
	 * 
	 * @param parent
	 * @param parameters
	 */
	public InventoryPolicyReorderPointOrderUpToLevelPeriodic(ModelElement parent, double[] parameters){
		this(parent, (int)parameters[0], (int)parameters[1], parameters[2], parameters[3], null);
	}

	/** Creates a periodic (reorder point, reorder up to) policy
	 * initialReviewTime The time of the first review = 0.0
	 * reviewPeriod The time between reviews = 1.0
	 * 
	 * @param parent
	 * @param reorderPoint the reorder point, must be < the order up to level
	 * @param orderUpToPoint
	 */
	public InventoryPolicyReorderPointOrderUpToLevelPeriodic(ModelElement parent, int reorderPoint, int orderUpToPoint) {
		this(parent, reorderPoint, orderUpToPoint, 1.0, 0.0, null);
	}

	/** Creates a periodic (reorder point, reorder up to) policy
	 * initialReviewTime The time of the first review = 0.0
	 * 
	 * @param parent
	 * @param reorderPoint the reorder point, must be < the order up to level
	 * @param orderUpToPoint
	 * @param reviewPeriod The time between reviews, must be > 0
	 */
	public InventoryPolicyReorderPointOrderUpToLevelPeriodic(ModelElement parent, int reorderPoint, int orderUpToPoint, double reviewPeriod) {
		this(parent, reorderPoint, orderUpToPoint, reviewPeriod, 0.0, null);
	}

	/** Creates a periodic (reorder point, reorder up to) policy
	 * initialReviewTime The time of the first review = 0.0
	 * 
	 * @param parent
	 * @param reorderPoint the reorder point, must be < the order up to level
	 * @param orderUpToPoint
	 * @param reviewPeriod The time between reviews, must be > 0
	 * @param name
	 */
	public InventoryPolicyReorderPointOrderUpToLevelPeriodic(ModelElement parent, int reorderPoint, int orderUpToPoint, double reviewPeriod, String name) {
		this(parent, reorderPoint, orderUpToPoint, reviewPeriod, 0.0, name);
	}
	
	/** Creates a periodic (reorder point, reorder up to) policy
	 * 
	 * @param parent
	 * @param reorderPoint the reorder point, must be < the order up to level
	 * @param orderUpToPoint
	 * @param reviewPeriod The time between reviews, must be > 0
	 * @param initialReviewTime The time of the first review, must be >= 0
	 * @param name
	 */
	public InventoryPolicyReorderPointOrderUpToLevelPeriodic(ModelElement parent, int reorderPoint, int orderUpToPoint, 
			double reviewPeriod, double initialReviewTime, String name) {
		super(parent, name);
		setInitialPolicyParameters(reorderPoint, orderUpToPoint, reviewPeriod, initialReviewTime);
		myReorderPoint = reorderPoint;
		myOrderUpToPoint = orderUpToPoint;
		myReviewPeriod = reviewPeriod;
		myInitialReviewTime = initialReviewTime;		
	}

	@Override
	public void setInitialPolicyParameters(double[] parameters) {
		if (parameters.length == 2)
			setInitialPolicyParameters((int)parameters[0],(int)parameters[1], myInitialPolicyParameters[2], myInitialPolicyParameters[3]);	
		else if (parameters.length == 3)	
			setInitialPolicyParameters((int)parameters[0],(int)parameters[1], parameters[2], myInitialPolicyParameters[3]);	
		else
			setInitialPolicyParameters((int)parameters[0],(int)parameters[1], parameters[2], parameters[3]);			
	}
	
	/** Sets the initial policy parameters
	 * 
	 * @param reorderPoint the reorder point must be < order up to point
	 * @param orderUpToPoint the  must be >= 1
	 * @param reviewPeriod must be > 0.0
	 * @param initialReviewTime must be >= 0.0
	 */
	public final void setInitialPolicyParameters(int reorderPoint, int orderUpToPoint, double reviewPeriod, double initialReviewTime) {
		if (orderUpToPoint < 1)
			throw new IllegalArgumentException("The order up to point must be >=1");

		if (reorderPoint >= orderUpToPoint)
			throw new IllegalArgumentException("The reorder point must be < order up to point");

		if (reviewPeriod <= 0)
			throw new IllegalArgumentException("The review period must be > 0.0");

		if (initialReviewTime < 0)
			throw new IllegalArgumentException("The initial review time must be >= 0.0");

		myInitialPolicyParameters = new double[4];
		myInitialPolicyParameters[0] = reorderPoint;
		myInitialPolicyParameters[1] = orderUpToPoint;
		myInitialPolicyParameters[2] = reviewPeriod;
		myInitialPolicyParameters[3] = initialReviewTime;
	
	}

	public final void setPolicyParameters(int reorderPoint, int orderUpToPoint) {
		setPolicyParameters(reorderPoint, orderUpToPoint, myReviewPeriod, myInitialReviewTime);
	}

	public final void setPolicyParameters(int reorderPoint, int orderUpToPoint, double reviewPeriod) {
		setPolicyParameters(reorderPoint, orderUpToPoint, reviewPeriod, myInitialReviewTime);
	}

	public final void setPolicyParameters(int reorderPoint, int orderUpToPoint, double reviewPeriod, double initialReviewTime) {
		if (orderUpToPoint < 1)
			throw new IllegalArgumentException("The order up to point must be >=1");

		if (reorderPoint >= orderUpToPoint)
			throw new IllegalArgumentException("The reorder point must be < order up to point");

		if (reviewPeriod <= 0)
			throw new IllegalArgumentException("The review period must be > 0.0");

		if (initialReviewTime < 0)
			throw new IllegalArgumentException("The initial review time must be >= 0.0");

		myReorderPoint = reorderPoint;
		myOrderUpToPoint = orderUpToPoint;
		myReviewPeriod = reviewPeriod;
		myInitialReviewTime = initialReviewTime;
	
	}

	/** Gets the current values of the policy parameters as an array
	 *	parameters[0] = reorderPoint
	 *	parameters[1] = orderUpToPoint
	 *  parameters[2] = reviewPeriod
	 *  parameters[3] = initialReviewTime
	 * 
	 * @return 
	 */
	public double[] getPolicyParameters() {
		double [] para = {myReorderPoint, myOrderUpToPoint, myReviewPeriod, myInitialReviewTime};
		return para;
	}

	/** Sets the policy parameters for the current replication
	 * 
	 * parameters[0] reorderPoint the reorder point must be < order up to point
	 * parameters[1] orderUpToPoint the orderUpToPoint must be >= 1
	 *  parameters[2] = reviewPeriod > 0.0
	 *  parameters[3] = initialReviewTime >= 0.0 
	 *   
	 * @param parameters must not be null
	 */
	protected void setPolicyParameters(double[] parameters) {		
		if (parameters.length == 2)
			setPolicyParameters((int)parameters[0],(int)parameters[1]);	
		else if (parameters.length == 3)	
			setPolicyParameters((int)parameters[0],(int)parameters[1], parameters[2]);	
		else
			setPolicyParameters((int)parameters[0],(int)parameters[1], parameters[2], parameters[3]);	
	}

	/* (non-Javadoc)
	 * @see jsl.examples.inventory.InventoryPolicyAbstract#checkInventory(jsl.examples.inventory.Inventory)
	 */
	protected void checkInventory() {
		
	}	

	protected final void setReviewPeriod(double reviewPeriod){
		if (reviewPeriod <= 0)
			throw new IllegalArgumentException("The review period must be > 0.0");
		myReviewPeriod = reviewPeriod;
	}

	protected final void setInitialReviewTime(double time){
		if (time < 0)
			throw new IllegalArgumentException("The initial review time must be >= 0.0");
		myInitialReviewTime = time;		
	}
	
	protected void initialize(){
		scheduleEvent(myInitialReviewTime);		
	}
	
	protected void handleEvent(JSLEvent event){
		if (getInventoryPosition()<= myReorderPoint){
			int orderSize = (int)(myOrderUpToPoint - getInventoryPosition());
			requestReplenishment(orderSize);
		}
		scheduleEvent(myReviewPeriod);
	}

	/**
	 * @return Returns the reorderPoint.
	 */
	public final int getReorderPoint() {
		return myReorderPoint;
	}

	/**
	 * @return Returns the orderUpToPoint.
	 */
	public final int getOrderUpToPoint() {
		return myOrderUpToPoint;
	}
	
	/**
	 * 
	 * @return
	 */
	public final double getReviewPeriod(){
		return(myReviewPeriod);
	}
	
	/**
	 * 
	 * @return
	 */
	public final double getInitialReviewTime(){
		return(myInitialReviewTime);
	}

}
