/*
 * Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
 *
 * Contact:
 *	Manuel D. Rossetti, Ph.D., P.E.
 *	Department of Industrial Engineering
 *	University of Arkansas
 *	4207 Bell Engineering Center
 *	Fayetteville, AR 72701
 *	Phone: (479) 575-6756
 *	Email: rossetti@uark.edu
 *	Web: www.uark.edu/~rossetti
 *
 * This file is part of the JSL (a Java Simulation Library). The JSL is a framework
 * of Java classes that permit the easy development and execution of discrete event
 * simulation programs.
 *
 * The JSL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * The JSL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the JSL (see file COPYING in the distribution);
 * if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA, or see www.fsf.org
 *
 */
package jslx.supplychain.inventorylayer;

import jsl.modeling.ModelElement;

/** An (r,Q) inventory policy orders the reorder quantity when the inventory position
 *  reaches the reorder point.  In the situation of the inventory position going
 *  significantly below the reorder point, such that a replenishment order of 
 *  size Q will not enable the inventory position to get above r, multiples of Q are 
 *  ordered.  In this situation, the flag getSeparateBatchOrdersOption() controls
 *  how the ordering takes place.  The ordering process can occur in two ways:
 *  1) 1 order of size n*Q where n is the number of batches to order such
 *     that the reorder point is exceeded,
 *  2) n separate orders of size Q are placed (at the same time) such that the
 *     reorder point is exceeded.
 *  If getSeparateBatchOrdersOption() is true, then the second case occurs.  The
 *  default behavior is case 1.  In case two, the n separate orders are counted
 *  and treated separately (within statistics, etc)
 *
 *
 *  
 *
 */
public class InventoryPolicyReorderPointReorderQuantity extends InventoryPolicyAbstract {


    private boolean mySeparateBatchOrdersOption;

    /** the reorder point must be >= -reorderQty
     */
    private int myReorderPoint;

    /** reorderQty the reorder quantity must be >= 1
     */
    private int myReorderQty;

    /** Constructs an (r,Q) inventory control policy with the supplied reorder point = 0 and reorder
     *  quantity = 1
     *
     * @param parent
     */
    public InventoryPolicyReorderPointReorderQuantity(ModelElement parent) {
        this(parent, 0, 1, null);
    }

    /** Constructs an (r,Q) inventory control policy with the supplied reorder point and reorder
     *  quantity.
     *
     * @param parent
     * @param reorderPoint
     * @param reorderQty
     */
    public InventoryPolicyReorderPointReorderQuantity(ModelElement parent, int reorderPoint, int reorderQty) {
        this(parent, reorderPoint, reorderQty, null);
    }

    /**
     *
     * @param parent
     * @param parameters
     * @param name
     */
    public InventoryPolicyReorderPointReorderQuantity(ModelElement parent, double[] parameters, String name) {
        this(parent, (int) parameters[0], (int) parameters[1], name);
    }

    /**
     *
     * @param parent
     * @param parameters
     */
    public InventoryPolicyReorderPointReorderQuantity(ModelElement parent, double[] parameters) {
        this(parent, (int) parameters[0], (int) parameters[1], null);
    }

    /** Constructs an (r,Q) inventory control policy with the supplied reorder point and reorder
     *  quantity.
     *
     * @param parent
     * @param reorderPoint
     * @param reorderQty
     * @param name
     */
    public InventoryPolicyReorderPointReorderQuantity(ModelElement parent, int reorderPoint, int reorderQty, String name) {
        super(parent, name);
        setInitialPolicyParameters(reorderPoint, reorderQty);
        myReorderPoint = reorderPoint;
        myReorderQty = reorderQty;
        mySeparateBatchOrdersOption = false;
    }

    /** If true, then n separate orders of size Q are placed (at the same time) such that the
     *  reorder point is exceeded where n is the number of orders necessary to get
     *  the inventory position above the reorder point
     *
     * @return
     */
    public boolean getSeparateBatchOrdersOption(){
        return mySeparateBatchOrdersOption;
    }

    /** If true, then n separate orders of size Q are placed (at the same time) such that the
     *  reorder point is exceeded where n is the number of orders necessary to get
     *  the inventory position above the reorder point
     *
     * @param flag
     */
    public void setSeparateBatchOrdersOption(boolean flag){
        mySeparateBatchOrdersOption = flag;
    }

    /* (non-Javadoc)
     * @see jsl.supplychain.inventorylayer.InventoryPolicyAbstract#checkInventory()
     */
    protected void checkInventory() {
        int ip = getInventoryPosition();
//		if( this.getModel().getReplication().getReplicationNumber() == 3.0){
//			System.out.println(getTime() + " > Checking inventory position: " + ip + " vs reorder point " + myReorderPoint);
//			System.out.println(this.getInventory());
//		}

        //System.out.println(getTime() + " > Checking inventory position: " + ip + " vs reorder point " + myReorderPoint);
        //System.out.println(this.getInventory());
        if (ip <= myReorderPoint) {
            // need to place an order, figure out the amount below reorder point

            if (ip == myReorderPoint){// hit reorder point exactly
                requestReplenishment(myReorderQty);
            } else {
                // find number of batches to order
                double gap = myReorderPoint - ip;
                int n = (int)Math.ceil(gap/myReorderQty);
                if (n == 0)
                    throw new IllegalStateException("number of batches was zero");
                if (getSeparateBatchOrdersOption()){
                    for(int i=1;i<=n;i++)
                      requestReplenishment(myReorderQty);
                } else
                   requestReplenishment(n*myReorderQty);
            }
        }
    }

    /**
     * @return Returns the reorderPoint.
     */
    public final int getReorderPoint() {
        return myReorderPoint;
    }

    /**
     * @return Returns the myReorderQty.
     */
    public final int getReorderQty() {
        return myReorderQty;
    }

    /** Sets the initial policy parameters for the next replication
     *
     * parameters[0] reorderPoint the reorder point must be >= -reorderQty
     * parameters[1] reorderQty the reorder quantity must be >= 1
     *
     * @param parameters must not be null
     */
    public final void setInitialPolicyParameters(double[] parameters) {
        setInitialPolicyParameters((int) parameters[0], (int) parameters[1]);
    }

    /** Sets the policy parameters
     *
     * @param reorderPoint the reorder point must be >= -reorderQty
     * @param reorderQty the reorder quantity must be >= 1
     */
    public final void setInitialPolicyParameters(int reorderPoint, int reorderQty) {
        if (reorderQty < 1) {
            throw new IllegalArgumentException("The reorder quantity must be >=1. The value was: " + reorderQty);
        }

        if (reorderPoint < -reorderQty) {
            throw new IllegalArgumentException("The reorder point must be >= -reorderQty. The value was: " + reorderPoint);
        }

        myInitialPolicyParameters = new double[2];
        myInitialPolicyParameters[0] = reorderPoint;
        myInitialPolicyParameters[1] = reorderQty;
    }

    /** Gets the current values of the policy parameters as an array
     *
     * @return
     */
    public double[] getPolicyParameters() {
        double[] para = {getReorderPoint(), getReorderQty()};
        return para;
    }

    /** Sets the values of the current policy parameters
     *
     * @param reorderPoint the reorder point must be >= -reorderQty
     * @param reorderQty the reorder quantity must be >= 1
     */
    protected final void setPolicyParameters(int reorderPoint, int reorderQty) {
        if (reorderQty < 1) {
            throw new IllegalArgumentException("The reorder quantity must be >=1. It was: " + reorderQty);
        }

        if (reorderPoint < -reorderQty) {
            throw new IllegalArgumentException("The reorder point quantity must be >= -reorderQty. It was: " + reorderPoint);
        }

        myReorderPoint = reorderPoint;
        myReorderQty = reorderQty;

        if (isCurrentReplicationRunning()) {
            myPolicyChangedDuringRepFlag = true;
        }
    }

    /** Sets the policy parameters for the current replication
     *
     * parameters[0] reorderPoint the reorder point must be >= -reorderQty
     * parameters[1] reorderQty the reorder quantity must be >= 1
     *
     * @param parameters must not be null
     */
    protected void setPolicyParameters(double[] parameter) {
        setPolicyParameters((int) parameter[0], (int) parameter[1]);
    }
}
