/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import jsl.modeling.ModelElement;
import jsl.utilities.random.RandomIfc;

/** Models ItemTypes within inventory systems.  An item type is like a stock keeping
 *  unit for a given type of product, typically at a particular location.  The class
 *  Inventory holds amounts of a given ItemType for usage in a supply network
 *  
 */
public class ItemType extends ModelElement {

	/** indicates the weight of the item
	 * 
	 */
	private double myWeight;
	
	/** indicates the cube (length x width x height) of the item
	 * 
	 */
	private double myCube;
	
	/** Item types are stocked a locations, they may be stocked
	 *  at a location that has a lead time for supplying the item type
	 *  This field can be used to determine the lead time for the
	 *  item type if it is set.  It may be null.
	 * 
	 */
	private RandomIfc myLeadTime;
	
	/**
	 * @param parent
	 */
	public ItemType(ModelElement parent) {
		this(parent, null, 1.0, 1.0, null);
	}
	
	/**
	 * @param parent
	 * @param name
	 */
	public ItemType(ModelElement parent, String name) {
		this(parent, name, 1.0, 1.0, null);
	}

	/**
	 * 
	 * @param parent
	 * @param weight
	 * @param cube
	 */
	public ItemType(ModelElement parent, double weight, double cube) {
		this(parent, null, weight, cube, null);
	}
	
	/**
	 * 
	 * @param parent
	 * @param name
	 * @param weight
	 * @param cube
	 */
	public ItemType(ModelElement parent, String name, double weight, double cube) {
		this(parent, name, weight, cube, null);
	}

	/**
	 * 
	 * @param parent
	 * @param name
	 * @param weight
	 * @param cube
	 */
	public ItemType(ModelElement parent, String name, double weight, double cube, RandomIfc leadTime) {
		super(parent, name);
		setWeight(weight);
		setCube(cube);
		setLeadTime(leadTime);
	}
	
	public String toString(){
		return(getName());
	}
	
	/** The cube of the item
	 * @return
	 */
	public final double getCube() {
		return myCube;
	}

	/** The weight of the item
	 * @return
	 */
	public final double getWeight() {
		return myWeight;
	}

	/**
	 * @return the leadTime
	 */
	public RandomIfc getLeadTime() {
		return myLeadTime;
	}

	/**
	 * @param myLeadTime the myLeadTime to set
	 */
	protected void setLeadTime(RandomIfc leadTime) {
		myLeadTime = leadTime;
	}

	/** Sets the cube of the item, must be >=0
	 * @param cube, the cube of the item
	 */
	protected final void setCube(double cube) {
		if (cube < 0)
			throw new IllegalArgumentException("Cube must be >= 0");
		myCube = cube;
	}

	/** Sets the weight of the item, must be >=0
	 * @param weight, the weight of the item
	 */
	protected final void setWeight(double weight) {
		if (weight <= 0)
			throw new IllegalArgumentException("Weight must be >= 0");
		myWeight = weight;
	}

	protected void removedFromModel(){
		super.removedFromModel();
		myLeadTime = null;
	}
	
}
