/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import java.util.HashMap;
import java.util.Map;
import java.util.Collections;
import java.util.Set;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.reporting.JSL;

/** A LeadTimeDemandFiller is a subclass of DemandFillerAbstract which can
 *  satisfy any demands that are set to it for which it has a leadtime defined.
 *  The demand will be filled after a time delay specified by the leadtime. The
 *  client may supply an instance of a class that implements the DemandCarrierIfc to
 *  allow transport from the LeadTimeDemandFiller.  If no DemandCarrierIfc is supplied
 *  the transport occurs with no time delay back to the customer.  In this way, the
 *  leadtime can conceptually contain the transport time, explicitly or implicitly.
 * 
 *
 */
public class LeadTimeDemandFiller extends DemandFillerAbstract {
	
	/** A Map to hold the leadtimes by ItemType
	 *  key = ItemType, object = RandomVariable representing the leadtime
	 */	
	protected Map<ItemType, RandomVariable> myLeadTimes;
	
	/** Listens for the demand to be filled after the lead time
	 * 
	 */
	protected DemandListenerFilledIfc myDemandFilledListener = new DemandFilledListener();
	
	/** Maps a item type in the LTDF to the statistics of time weighted average
	 * of number in delay 
	 * 
	 */
	protected Map<ItemType, TimeWeighted> myMapTWRVNumOfDelayedItems;
	
	/** time weighted average
	 * of number in delay across the item types
	 * 
	 */
	protected TimeWeighted myAggregateTWRVNumOfDelayedItems;
	
	/** Maps a item type in the LTDF to the statistics of average
	 * of delay time
	 * 
	 */
	protected Map<ItemType, ResponseVariable> myMapTWRVDelayTimeOfItems;	
	/** average of delay time
	 * 
	 */
	protected ResponseVariable myAggregateTWRVDelayTimeOfItems;
	
	/**
	 * @param parent
	 */
	public LeadTimeDemandFiller(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public LeadTimeDemandFiller(ModelElement parent, String name) {
		super(parent, name);
		myLeadTimes = new HashMap<ItemType, RandomVariable>();
		myMapTWRVNumOfDelayedItems = new HashMap<ItemType, TimeWeighted>();
		myAggregateTWRVNumOfDelayedItems = new TimeWeighted(this, getName() + " Num In Delay Aggregate Value");
		myMapTWRVDelayTimeOfItems = new HashMap<ItemType, ResponseVariable>();
		myAggregateTWRVDelayTimeOfItems = new ResponseVariable(this, getName() + " Delay Time Aggregate Value");
	}

	/** Adds a leadtime distribution for the given ItemType
	 * 
	 * @param itemType Must not be null
	 * @param distribution Must not be null
	 */
	public final void addLeadTime(ItemType itemType, RandomIfc distribution){
		if (itemType == null)
			throw new IllegalArgumentException("The ItemType type must be non-null.");

		if (distribution == null)
			throw new IllegalArgumentException("The DistributionIfc distribution must be non-null.");
		
		if (myLeadTimes.containsKey(itemType)){ // item type has already been added
			// need only to set the distribution associated with the type
			RandomVariable rv = myLeadTimes.get(itemType);
			rv.setInitialRandomSource(distribution);		//TODO should this be setInitialRandomSource()?	
		} else {
			// new item type being entered
			// create the random variable
			RandomVariable rv = new RandomVariable(this, distribution);
			myLeadTimes.put(itemType, rv);

			myMapTWRVNumOfDelayedItems.put(itemType, new TimeWeighted(this, this.getName()+ itemType.getName() +"Num In Delay"));
			myMapTWRVDelayTimeOfItems.put(itemType, new ResponseVariable(this, this.getName()+ itemType.getName() +"Time of Delay"));
		
		}
	}
	
	/** Removes the item type and its lead time characteristics from the LeadTimeDemandFiller
	 *  NOTE: Also removes any associated model elements from the model, via ModelElement.removeFromModel();
	 *  
	 *  WARNING: Removal of an item type from a filler may cause demand to be rejected if demand for the
	 *  item type is still in progress.  In addition, if a fill event is scheduled for the item type, this
	 *  event is not canceled.  This may cause errors if used during a replication.  It is best to
	 *  use this method when a replication is not running.
	 * 
	 * @param itemType
	 */
	public final void removeLeadTime(ItemType itemType){
		if (itemType == null)
			throw new IllegalArgumentException("The ItemType type must be non-null.");
		
		RandomVariable rv = myLeadTimes.remove(itemType);
		
		if (rv == null)
			return;
		// item type responses need to be removed
		TimeWeighted tw = myMapTWRVNumOfDelayedItems.remove(itemType);
		ResponseVariable rs = myMapTWRVDelayTimeOfItems.remove(itemType);
		rv.removeFromModel();
		tw.removeFromModel();
		rs.removeFromModel();
	}
	
	/* (non-Javadoc)
	 * @see jsl.supplychain.inventorylayer.DemandFillerIfc#fillDemand(jsl.supplychain.inventorylayer.Demand)
	 */
	public void fillDemand(Demand demand) {
		
		// place the demand in process
		demand.process(this);
		demand.setFilledListener(myDemandFilledListener);

		// get the type of the demand
		ItemType type = demand.getItemType();
		// get the associated leadtime random variable
		RandomVariable rv = myLeadTimes.get(type);
		// generate the leadtime
		double leadtime = rv.getValue();

		myAggregateTWRVNumOfDelayedItems.increment();
		myAggregateTWRVDelayTimeOfItems.setValue(leadtime);
		myMapTWRVNumOfDelayedItems.get(type).increment();
		//TODO this really should be captured after the lead time
		myMapTWRVDelayTimeOfItems.get(type).setValue(leadtime);
		
		JSLEvent e = scheduleEvent(leadtime);
		e.setMessage(demand);

	}

	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.DemandFillerIfc#canFillItemType(jsl.supplychainNew.inventorylayer.Demand)
	 */
	public final boolean canFillItemType(Demand demand) {
		return myLeadTimes.containsKey(demand.getItemType());
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.DemandFillerIfc#canFillItemType(jsl.research.supplychain.inventorylayer.ItemType)
	 */
	public final boolean canFillItemType(ItemType type) {
		return myLeadTimes.containsKey(type);
	}
	
	/** Returns an unmodifiable set of the item types handled by this filler
	 * 
	 * @return
	 */
	public final Set<ItemType> getItemTypes(){
		return Collections.unmodifiableSet(myLeadTimes.keySet());
	}
	
	/* (non-Javadoc)
	 * @see jsl.supplychain.inventorylayer.DemandFillerIfc#negotiate(jsl.supplychain.inventorylayer.Demand)
	 */
	public DemandMessageIfc negotiate(Demand demand) {
		demand.negotiate();
		boolean itemTypeFlag = canFillItemType(demand);
		int status = Demand.IMMEDIATE_FILL;
		int amt = demand.getRemainingDemand();		
		DemandMessageIfc dm = new DemandMessage(this, itemTypeFlag, status, amt, null, false, false);
		return dm;
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.DemandFillerIfc#determineRequestStatus(jsl.research.supplychain.inventorylayer.Demand)
	 */
	public int determineRequestStatus(Demand demand) {
		
		if (!isAvailable()){
			return (Demand.FILLER_UNAVAILABLE);		
		}
		
		if (!canFillItemType(demand)){
			return (Demand.ITEMTYPE_MISMATCH);
		}

		return Demand.IMMEDIATE_FILL;
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.DemandFillerIfc#willReject(jsl.research.supplychain.inventorylayer.Demand)
	 */
	public boolean willReject(Demand demand) {
		
		if (!isAvailable()){
			return true;			
		}
		
		if (!canFillItemType(demand)){
			return true;
		}
		
		return false;
	}
	
	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.DemandFillerIfc#receive(jsl.supplychainNew.inventorylayer.Demand)
	 */
	public void receive(Demand demand) {

		// determine whether to reject or process the demand
		
		if (!isAvailable()){
			demand.setDemandStatus(Demand.FILLER_UNAVAILABLE);
			demand.reject();
			return;			
		}
		
		if (!canFillItemType(demand)){
			demand.setDemandStatus(Demand.ITEMTYPE_MISMATCH);
			demand.reject();
			return;
		}
//		JSL.out.println("-------------------------------------");
//		JSL.out.println("In LeadTimeDemandFiller.receive(Demand demand)");
//		JSL.out.println(getTime() + " > LeadTimeDemandFiller received Demand: " + demand.getId() + " for " + demand.getOriginalAmountDemanded());
//		JSL.out.println("-------------------------------------");
		
		demand.setDemandStatus(Demand.IMMEDIATE_FILL);
		demand.receive(this);	

	}
	
	protected void handleEvent(JSLEvent event){
		Demand demand = (Demand)event.getMessage();
//		JSL.out.println("-------------------------------------");
//		JSL.out.println("In LeadTimeDemandFiller.handleEvent(JSLEvent event) - lead time done");
//		JSL.out.println(getTime() + " > LeadTimeDemandFiller filled Demand: " + demand.getId() + " for " + demand.getOriginalAmountDemanded());
//		JSL.out.println("-------------------------------------");
		demand.fill(demand.getOriginalAmountDemanded());

		myAggregateTWRVNumOfDelayedItems.decrement();
		TimeWeighted tw = myMapTWRVNumOfDelayedItems.get(demand.getItemType());
		if (tw != null)
			tw.decrement();
	}
	
	/** Used internally to cause an incoming demand to be shipped to
	 *  its sender 
	 * 
	 * @param demand
	 */
	protected final void shipDemand(Demand demand){
		if (myExternalDemandCarrier != null)
			myExternalDemandCarrier.transportDemand(demand);
		else {
			demand.ship();
			demand.deliver();
		}		
	}

	/** Can be overridden by subclasses to provide behavior for when the demand is filled.
	 *  Current behavior is to call shipDemand()
	 * 
	 * @param demand
	 */
	protected void demandFilled(Demand demand){
		shipDemand(demand);
	}
	
	public RandomIfc getLeadTime(ItemType type) {
		return myLeadTimes.get(type).getRandomSource();
	}	
	
	protected final class DemandFilledListener implements DemandListenerFilledIfc {
		public final void demandFilled(Demand demand) {
			LeadTimeDemandFiller.this.demandFilled(demand);
		}
	}

}
