package jslx.supplychain.inventorylayer;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.reporting.JSL;
import jslx.supplychain.transportlayer.OrderCarrierIfc;

public class LeadTimeOrderFiller extends OrderFillerAbstract {

	/** The shipper that sends filled order to this filler's customers
	 *  If supplied, it will be used to ensure that a filled order
	 *  is shipped back to the customer.  Thus, this attribute may be null.
	 *  The default behavior is to immediately ship and deliver the order
	 *  with zero time delay.
	 *  
	 */
	protected OrderCarrierIfc myOrderShipper;
	
	/** Collects statistics on time average number on order
	 */
	protected TimeWeighted myNumOnOrder;
	
	/** Serves as the demand filler for demands on the incoming orders
	 */
	protected InnerDemandFiller myLeadTimeDemandFiller;

	/** When all the demands on an order have been filled, this listener
	 *  is called
	 */
	protected InnerOrderFilledListener myOrderFilledListener = new InnerOrderFilledListener();
	
	public LeadTimeOrderFiller(ModelElement parent) {
		this(parent, true, null);
	}

	public LeadTimeOrderFiller(ModelElement parent, String name) {
		this(parent, true, name);
	}

	public LeadTimeOrderFiller(ModelElement parent, boolean availabilityFlag,
			String name) {
		super(parent, availabilityFlag, name);
		// create a demand filler for the demands within the orders
		myLeadTimeDemandFiller = new InnerDemandFiller(this,getName()+"InnerLTDF");
		// ensure that demands within orders will be able to find a filler
		setDemandFillerFinder(myLeadTimeDemandFiller);
		myNumOnOrder = new TimeWeighted(this, getName() + "#OnOrder");
	}

	public String toString(){
		return(getName() + " Number on Order: " + (int)myNumOnOrder.getValue());
	}
	
	/** Adds a leadtime distribution for the given ItemType
	 * 
	 * @param type Must not be null
	 * @param distribution Must not be null
	 */
	public final void addLeadTime(ItemType type, RandomIfc distribution){
		myLeadTimeDemandFiller.addLeadTime(type, distribution);
	}

	/** Sets the shipper that this warehouse will use when sending filled orders
	 *  to customers of its inventory.
	 *  
	 * @param shipper
	 */
	public final void setOrderShipper(OrderCarrierIfc shipper) {
		myOrderShipper = shipper;
	}
	
	public final boolean canFillItemType(Demand demand) {
		return myLeadTimeDemandFiller.canFillItemType(demand);
	}

	public void fill(Order order) {
		myNumOnOrder.increment();
//		JSL.out.println("*********************************************************");
//		JSL.out.println("In LeadTimeOrderFiller.fill(Order order)");
//		JSL.out.println(getTime() +"> " + getName() + " has been asked to fill order " + order.getId() + " order# " + order.getName());
//		JSL.out.println("Number on order: "+ (int)myNumOnOrder.getValue());
//		JSL.out.println("*********************************************************");
		order.setFilledListener(myOrderFilledListener);
		placeOrderInProcessState(order);
		askOrderToBeginProcessingDemands(order);
	}

	public OrderMessageIfc negotiate(Order order) {
		// TODO Auto-generated method stub
		return null;
	}
	
	protected void afterReplication(){
		super.afterReplication();
//		JSL.out.println("After replication -------------:");
//		JSL.out.println(this);
	}

	/** Gets the shipper for this warehouse
	 * 
	 * @return
	 */
	protected final OrderCarrierIfc getOrderShipper() {
		return myOrderShipper;
	}
	
	protected class InnerOrderFilledListener implements OrderListenerFilledIfc {

		public void orderFilled(Order order) {
			myNumOnOrder.decrement();
			// TODO Auto-generated method stub
			// need to package the order in a Load and ship it
//			JSL.out.println("*********************************************************");
//			JSL.out.println("In LeadTimeOrderFiller.InnerOrderFilledListener.orderFilled(Order order)");
//			JSL.out.println(getTime() +"> " + getName() + " has FILLED order " + order.getId() + " order# " + order.getName());
//			JSL.out.println("Number on order: "+ (int)myNumOnOrder.getValue());
//			JSL.out.println("*********************************************************");
			if (myOrderShipper != null)
				myOrderShipper.transportOrder(order);
			else {
				order.ship();
				order.deliver();
			}
		}
		
	}
	
	protected class InnerDemandFiller extends LeadTimeDemandFiller implements DemandFillerFinderIfc {

		public InnerDemandFiller(ModelElement parent, String name) {
			super(parent, name);
		}

		public void fillDemand(Demand demand) {
			// place the demand in process
			demand.process(this);
			// get the type of the demand
			ItemType type = demand.getItemType();
			// get the associated leadtime random variable
			RandomVariable rv = myLeadTimes.get(type);
			// generate the leadtime
			double leadtime = rv.getValue();
			JSLEvent e = scheduleEvent(leadtime);
			e.setMessage(demand);
		}
				
		public DemandFillerIfc findDemandFiller(Demand demand) {
			return this;
		}
		
	}
}
