/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import jsl.modeling.elements.queue.QObject;
import jsl.utilities.reporting.JSL;

/**
 *
 */
public class Order extends QObject {

	//  ===========================================
	//     PUBLIC STATIC ATTRIBUTES
	//  ===========================================
	
	// Keeps track of the current status of the Order
	/** Used to indicate that no status has been set for the Order
	 */
	public static final int NO_STATUS = JSL.getNextEnumConstant();

	/** Used to indicate that the order was rejected because a 
	 * demand on the order was rejected
	 */
	public static final int DEMAND_REJECTED = JSL.getNextEnumConstant();

	/** Used to indicate that the order will be filled in full, immediately
	 *  That is, all the demands on the order are immediate fills
	 */
	public static final int IMMEDIATE_FILL = JSL.getNextEnumConstant();
	
	/** Used to indicate that the order will not be filled in full, immediately
	 *  That is, at least one demand on the order will be backlogged
	 */
	public static final int WILL_BE_BACKLOGGED = JSL.getNextEnumConstant();	

	/** Used to indicate the order's status when rejecting or negotiating
	 *  Indicates that the filler was unavailable to receive the order
	 */
	public static final int FILLER_UNAVAILABLE = JSL.getNextEnumConstant();

	//  ===========================================
	//     PROTECTED STATIC ATTRIBUTES
	//  ===========================================

	/** The order contains no demands, and has just been created
	 */
	protected static final OrderStateCreated CREATED = new OrderStateCreated("ORDER_CREATED");

	/** The order contains at least one unfilled demand and has
	 *  not yet been sent for filling
	 */
	protected static final OrderStatePreparing IN_PREPARATION = new OrderStatePreparing("ORDER_IN_PREPARATION");

	/** The order has been sent to a filler
	 */
	protected static final OrderState SENT = new OrderStateSent("ORDER_SENT");
	
	/** The order is ready to be processed or negotiate
	 */
	protected static final OrderState RECEIVED = new OrderStateReceived("ORDER_RECEIVED");
	
	/** The order is being filled
	 */
	protected static final OrderState IN_PROCESS = new OrderStateInProcess("ORDER_IN_PROCESS");

	/** The order is being backlogged
	 */
	protected static final OrderState BACKLOGGED = new OrderStateBackLogged("ORDER_BACKLOGGED");
	
	/** The order was rejected after being received
	 */
	protected static final OrderState REJECTED = new OrderStateRejected("ORDER_REJECTED");
	
	/** The order was cancelled after being in process
	 */
	protected static final OrderState CANCELLED = new OrderStateCancelled("ORDER_CANCELLED");
	
	/** The order contains only filled demands
	 */
	protected static final OrderState FILLED = new OrderStateFilled("ORDER_FILLED");
	
	/** The order has been shipped
	 */
	protected static final OrderState SHIPPED = new OrderStateShipped("ORDER_SHIPPED");
	
	/** The order has been delivered
	 */
	protected static final OrderState DELIVERED = new OrderStateDelivered("ORDER_DELIVERED");

	/** The order is under negotiation
	 */
	protected static final OrderState NEGOTIATING = new OrderStateNegotiating("ORDER_NEGOTIATING");

	//  ===========================================
	//     PACKAGE ATTRIBUTES
	//  ===========================================

	//PACKAGE so that OrderState can access
	/** Whether or not the Order allows itself to be partially shipped
	 */
	boolean myAllowPartialShippingFlag;  
	
	/** Whether or not the order allows itself to be backlogged
	 */
	boolean myAllowBackLoggingFlag;
	
	/** Whether or not the order will allow itself to be cancelled
	 */
	boolean myAllowCancellingFlag;

	/** Used to indicate if the order has been filled
	 */
	boolean myFilledFlag;

	//  ===========================================
	//     PROTECTED ATTRIBUTES
	//  ===========================================
		

	//  ===========================================
	//     PRIVATE ATTRIBUTES
	//  ===========================================

	/** Indicates that the order has a demand that was rejected
	 */
	private boolean myRejectionFlag;

	/** The object that sent the order to be filled
	 */
	private OrderSenderIfc mySender;
	
	/** The object currently filling the order
	 */
	private OrderFillerIfc myFiller;
		
	/** Keeps track of the current state of the Order
	 */
	private OrderState myOrderState;
	
	/** Used to indicate the order's status when rejecting or negotiating
	 * 
	 */
	private int myStatus = NO_STATUS;
	
	/** Used to count the number of demands on an order successfully received
	 */
	private int myNumReceived = 0;
	
	/** Used to count the number of demands on an order that were rejected
	 */
	private int myNumRejected = 0;
	
	/** Used to count the number of demands asked to be received
	 *  myNumIterated = myNumReceived + myNumRejected
	 *  When myNumIterated equals the number of demands in the order
	 *  then all demands have been received/processed
	 */
	private int myNumIterated = 0;
	
	/** Notified whenever a demand on this order is received
	 *  Notified whenever a demand on this order is rejected
	 *  Notified whenever a demand on this order is filled
	 *  Notified whenever a demand on this order is backlogged
	 *  Notified whenever the order is shipped to provide default
	 *  behavior that automatically causes each demand to also be shipped
	 */
	private DemandListener myDemandListener = new DemandListener();
		
	/** Can be attached to listen for when a demand inside the order
	 *  is filled
	 */
	private DemandListenerFilledIfc myDemandOnOrderFilledListener;
	
	/** The total weight that will be associated with this order
	 *  if all the demands are filled.
	 */
	private double myRequiredWeight;

	/** The total cube that will be associated with this order
	 *  if all the demands are filled.
	 */
	private double myRequiredCube;
	
	/** Holds the demands associated with this order
	 *  The assumption is that the demands are held in the sequence
	 *  in which they were added to the order.
	 */
	private List<Demand> myDemands;
			
	/** Holds the ItemTypes associated with this order
	 *  The assumption is that the ItemTypes are held in
	 *  no particular order.
	 */
	private Set<ItemType> myItemTypes;
	
	/** Represents the number of demands that have been filled
	 */
	private int myNumFilledDemands;

	/** Can be used to notify listeners on this order that the
	 * order has been backlogged
	 */
	private OrderListenerBackLoggedIfc myBackLogListener;
	
	/** Can be used to notify listeners on this order that the
	 * order has been cancelled
	 */
	private OrderListenerCancelledIfc myCancelListener;
	
	/** Can be used to notify listeners on this order that the
	 * order has been delivered
	 */
	private OrderListenerDeliveredIfc myDeliveryListener;
	
	/** Can be used to notify listeners on this order that the
	 * order has been filled
	 */
	private OrderListenerFilledIfc myFilledListener;
	
	/** Can be used to notify listeners on this order that the
	 * order has been placed in process
	 */
	private OrderListenerInProcessIfc myInProcessListener;
	
	/** Can be used to notify listeners on this order that the
	 * order has been received
	 */
	private OrderListenerReceivedIfc myReceivedListener;
	
	/** Can be used to notify listeners on this order that the
	 * order has been rejected
	 */
	private OrderListenerRejectedIfc myRejectedListener;
	
	/** Can be used to notify listeners on this order that the
	 * order has been shipped
	 */
	private OrderListenerShippedIfc myShippedListener;

	/** Creates an order with defaults:
	 *  1) Allows backlogging
	 *  2) No partial shipping
	 *  3) No cancelling allowed
	 * 
	 */
	public Order() {
		this(true, false, false, null);
	}

	/** Creates an order with defaults:
	 *  1) Allows backlogging
	 *  2) No partial shipping
	 *  3) No cancelling allowed
	 *  
	 * @param name
	 */
	public Order(String name) {
		this(true, false, false, name);
	}

	/** Creates an order with defaults:
	 *  1) No partial shipping
	 *  2) No cancelling allowed
	 * 
	 * @param allowBackLogFlag
	 * @param name
	 */
	public Order(boolean allowBackLogFlag, String name){
		this(allowBackLogFlag, false, false, name);
	}

	/** Creates an order with defaults:
	 *  1) No partial shipping
	 *  2) No cancelling allowed
	 * 
	 * @param allowBackLogFlag
	 */
	public Order(boolean allowBackLogFlag){
		this(allowBackLogFlag, false, false, null);
	}

	/** Creates an order with defaults:
	 *  1) No cancelling allowed
	 * 
	 * @param allowBackLogFlag
	 * @param partialShipFlag
	 */
	public Order(boolean allowBackLogFlag, boolean partialShipFlag){
		this(allowBackLogFlag, partialShipFlag, false, null);
	}
	
	/** Creates an order with defaults:
	 *  1) No cancelling allowed
	 * 
	 * @param allowBackLogFlag
	 * @param partialShipFlag
	 * @param name
	 */
	public Order(boolean allowBackLogFlag, boolean partialShipFlag, String name){
		this(allowBackLogFlag, partialShipFlag, false, name);
	}

	/** Creates an order with the given options
	 * 
	 * @param allowBackLogFlag
	 * @param partialShipFlag
	 * @param allowCancellingFlag
	 * @param name
	 */
	public Order(boolean allowBackLogFlag, boolean partialShipFlag, boolean allowCancellingFlag, String name){
		super(name);
		myAllowPartialShippingFlag= partialShipFlag;
		myAllowBackLoggingFlag = allowBackLogFlag;
		myAllowCancellingFlag = allowCancellingFlag;
		myNumFilledDemands = 0;
		myFilledFlag = false;
		myDemands = new ArrayList<Demand>();
		myItemTypes = new HashSet<ItemType>();
		myShippedListener = myDemandListener;
		setOrderState(CREATED);
	}

	//  ===========================================
	//     PUBLIC METHODS
	//  ===========================================

	/** Indicates whether or not partial shipping is allowed for the order
	 * @param flag
	 */
	public final void setPartialShippingFlag(boolean flag){
		myOrderState.setPartialShippingFlag(this,flag);
	}
	
	public final boolean isPartialShippingAllowed(){
		return(myAllowPartialShippingFlag);
	}
	
	public final void setBackLoggingFlag(boolean flag){
		myOrderState.setAllowBackLoggingFlag(this, flag);
	}
	
	public final boolean isBackLoggingAllowed(){
		return(myAllowBackLoggingFlag);
	}
	
	public final void setAllowCancelingFlag(boolean flag){
		myOrderState.setAllowCancellingFlag(this, flag);
	}
	
	public final boolean isCancellingAllowed(){
		return(myAllowCancellingFlag);
	}

	public final int getOrderStatus(){
		return(myStatus);
	}
	
	/** An order consists of a list of demands that must be filled
	 *  This method adds a demand to the order
	 * 
	 * @param demand, must not be null and must not have already been added
	 *    to the order.
	 */
	public final void addDemand(Demand demand) {
		if (demand == null)
			throw new IllegalArgumentException("Demand demand was null!");

		if (demand.getAllowBackLoggingFlag() != isBackLoggingAllowed())
			throw new IllegalArgumentException("Demand's backlogging was incompatible with order's backlogging.");
		
		if (myItemTypes.contains(demand.getItemType()))
			throw new IllegalArgumentException("Item type was already in the order.");

		if (myDemands.contains(demand))
			throw new IllegalArgumentException("Demand was already in the order.");
		
		myOrderState.addDemand(this, demand);
	}
	
	public final void removeDemand(Demand demand){
		if (demand == null)
			throw new IllegalArgumentException("Demand demand was null!");

		if (demand.getOrder() != this)
			throw new IllegalArgumentException("The demand was not part of this order!");
		
		myOrderState.removeDemand(this, demand);
	}

	/** Tells each demand to change its state to shipped.
	 *  This can only be invoked if the order is already in
	 *  the shipped state
	 */
	public final void setDemandStateToShipped(){
		myOrderState.setDemandStateToShipped(this);
	}

	/** Tells each demand to change its state to the delivered
	 *  state.  This should only be invoked if the order is already
	 *  in the DELIVERED state.  In other words, the demands on an
	 *  orderd cannot be delivered unless the order has been delivered
	 */
	public final void setDemandStateToDelivered(){
		myOrderState.setDemandStateToDelivered(this);
	}
	
	/** Gets the number of demands on the order
	 * 
	 * @return
	 */
	public final int getNumberOfDemands(){
		return(myDemands.size());
	}
	
	/** Checks if the order has any demands
	 * 
	 * @return
	 */
	public final boolean isEmpty(){
		return(myDemands.size() == 0);
	}
	
	/** Gets the number of demands that have 
	 *  been filled on the order at the current time
	 * 
	 * @return
	 */
	public final int getNumberOfFilledDemands(){
		return(myNumFilledDemands);
	}
	
	/** Returns a reference to the demand at supplied index
	 *  The first demand is at index = 0, the last is at
	 *  index = size() -1.  Throws index out of bounds exception 
	 *  if the supplied index is outside this range
	 * 
	 * @param index
	 * @return
	 */
	public final Demand getDemand(int index){
		return(myDemands.get(index));
	}
	
	/** Returns whether or not the order has been filled.
	 *  This is not the same as the FILLED state.  After the order
	 *  is placed in the FILLED state, this method will return true
	 * 
	 * @return
	 */
	public final boolean isFilled(){
		return(myFilledFlag);
	}
	
	/** Checks whether or not the order is in the supplied state
	 * 
	 * @param state
	 * @return true if the order is in the supplied state
	 */
	public final boolean isState(OrderState state){
		return(myOrderState == state);
	}
	
	/** Sets the filler for this order
	 * 
	 * @param filler
	 */
	public final void setFiller(OrderFillerIfc filler){
		myOrderState.setFiller(this, filler);
	}
	
	/** Returns the filler for this order
	 * 
	 * @return
	 */
	public final OrderFillerIfc getFiller(){
		return(myFiller);
	}
	
	/**
	 * @return Returns the myBackLogListener.
	 */
	public final OrderListenerBackLoggedIfc getbackLogListener() {
		return myBackLogListener;
	}
	
	/**
	 * @param listener The backLogListener to set.
	 */
	public final void setBackLogListener(OrderListenerBackLoggedIfc listener) {
		myBackLogListener = listener;
	}
	
	/**
	 * @return Returns the myCancelListener.
	 */
	public final OrderListenerCancelledIfc getCancelListener() {
		return myCancelListener;
	}
	
	/**
	 * @param listener The cancelListener to set.
	 */
	public final void setCancelListener(OrderListenerCancelledIfc listener) {
		myCancelListener = listener;
	}
	
	/**
	 * @return Returns the deliveryListener.
	 */
	public final OrderListenerDeliveredIfc getDeliveryListener() {
		return myDeliveryListener;
	}
	
	/**
	 * @param listener The deliveryListener to set.
	 */
	public final void setDeliveryListener(OrderListenerDeliveredIfc listener) {
		myDeliveryListener = listener;
	}
	
	/**
	 * @return Returns the filledListener.
	 */
	public final OrderListenerFilledIfc getFilledListener() {
		return myFilledListener;
	}
	
	/**
	 * @param listener The filledListener to set.
	 */
	public final void setFilledListener(OrderListenerFilledIfc listener) {
		myFilledListener = listener;
	}
	
	/**
	 * @return Returns the inProcessListener.
	 */
	public final OrderListenerInProcessIfc getInProcessListener() {
		return myInProcessListener;
	}
	
	/**
	 * @param listener The OrderListenerInProcessIfc to set.
	 */
	public final void setInProcessListener(OrderListenerInProcessIfc listener) {
		myInProcessListener = listener;
	}
	
	/**
	 * @return Returns the receivedListener.
	 */
	public final OrderListenerReceivedIfc getReceivedListener() {
		return myReceivedListener;
	}
	
	/**
	 * @param listener The OrderListenerReceivedIfc to set.
	 */
	public final void setReceivedListener(OrderListenerReceivedIfc listener) {
		myReceivedListener = listener;
	}
	
	/**
	 * @return Returns the rejectedListener.
	 */
	public final OrderListenerRejectedIfc getRejectedListener() {
		return myRejectedListener;
	}
	
	/**
	 * @param listener The OrderListenerRejectedIfc to set.
	 */
	public final void setRejectedListener(OrderListenerRejectedIfc listener) {
		myRejectedListener = listener;
	}
	
	/**
	 * @return Returns the shippedListener.
	 */
	public final OrderListenerShippedIfc getShippedListener() {
		return myShippedListener;
	}
	
	/**
	 * @param listener The OrderListenerShippedIfc to set.
	 */
	public final void setShippedListener(OrderListenerShippedIfc listener) {
		myShippedListener = listener;
	}
	
	public final DemandListenerFilledIfc getDemandOnOrderFilledListener() {
		return myDemandOnOrderFilledListener;
	}

	public final void setDemandOnOrderFilledListener(DemandListenerFilledIfc listener) {
		myDemandOnOrderFilledListener = listener;
	}

	public OrderSenderIfc getSender() {
		return mySender;
	}
	
	protected void setSender(OrderSenderIfc sender){
		mySender = sender;
	}
	
	public String toString(){
		StringBuffer sb = new StringBuffer();
		sb.append(" Order ");
		sb.append(getId());
		sb.append(" " + getName());
		sb.append("\n");
		for(Demand d: myDemands)
			sb.append(d);
		
		return(sb.toString());
	}
	
	//  ===========================================
	//     PROTECTED METHODS
	//  ===========================================

	protected void setOrderState(OrderState state){
//		if (myOrderState != null){
//			JSL.out.println("------------------------------------------------");
//			JSL.out.println("In Order.setOrderState(OrderState state)");
//			JSL.out.println(getTime() + " > " + "Order: " + getId() + " is transitioning from " + myOrderState.toString() + " to state " + state.toString());
//			JSL.out.println("------------------------------------------------");
//		}
		myOrderState = state;
		
		if (isState(Order.CREATED))
			return;
		
		if (isState(Order.SENT))
			return;
		
		if (isState(Order.NEGOTIATING))
			return;
		
		if (isState(Order.RECEIVED)){
			if (myReceivedListener != null)
				myReceivedListener.orderReceived(this);
			return;
		}

		if (isState(Order.REJECTED)){
			if (myRejectedListener != null)
				myRejectedListener.orderRejected(this);
			return;
		}

		if (isState(Order.IN_PROCESS)){
			if (myInProcessListener != null)
				myInProcessListener.orderInProcess(this);
			return;
		}

		if (isState(Order.FILLED)){
			if (myFilledListener != null)
				myFilledListener.orderFilled(this);
			return;
		}

		if (isState(Order.BACKLOGGED)){
			if (myBackLogListener != null)
				myBackLogListener.orderBackLogged(this);
			return;
		}
		
		if (isState(Order.SHIPPED)){
			if (myShippedListener != null)
				myShippedListener.orderShipped(this);
			return;
		}

		if (isState(Order.DELIVERED)){
			if (myDeliveryListener != null)
				myDeliveryListener.orderDelivered(this);
			return;
		}
	}

	protected void addDemand_(Demand demand){
		myDemands.add(demand);
		demand.setOrder(this);
		demand.setOrderIndex(myDemands.indexOf(demand));
		if (demand.isFilled())//Should we permit filled demands to be added?
			myNumFilledDemands++;
		myRequiredWeight = myRequiredWeight + demand.getWeight();
		myRequiredCube = myRequiredCube + demand.getCube();
		myItemTypes.add(demand.getItemType());
		//Add Order demand listeners on the demand
		demand.setFilledListener(myDemandListener);
	}

	protected void removeDemand_(Demand demand){
		myDemands.remove(demand);
		demand.setOrder(null);
		demand.setOrderIndex(-1);
		if (demand.isFilled())
			myNumFilledDemands--;
		myRequiredWeight = myRequiredWeight - demand.getWeight();
		myRequiredCube = myRequiredCube + demand.getCube();
		myItemTypes.remove(demand.getItemType());
		//remove Order demand listeners from the demand
		demand.setFilledListener(null);
	}

	/** Sets the filler for this order
	 * 
	 * @param filler
	 */
	protected final void setFiller_(OrderFillerIfc filler){
		myFiller = filler;
	}
	
	protected final List<Demand> getDemands(){
		return(myDemands);
	}
		
	protected void beginReceivingDemands_(){
		myRejectionFlag = false;
		myNumIterated = 0;
		myNumReceived = 0;
		myNumRejected = 0;
		for(Demand d: myDemands){
			DemandFillerIfc filler = d.getFiller();
			filler.receive(d);			
		}
	}
	
	protected void beginProcessingDemands_(){
		myNumFilledDemands = 0;
		for(Demand d: myDemands){
			DemandFillerIfc filler = d.getFiller();
			filler.fillDemand(d);			
		}
	}
	
	protected void demandReceived(Demand demand) {
		// a demand on the order was received
		// increment the number received
		myNumReceived++;
		// increment the number asked to be received
		myNumIterated++;
		if (myNumIterated == myDemands.size()){
			if (myRejectionFlag){
				// the order must be rejected, because at least one demand was rejected
				setOrderStatus(Order.DEMAND_REJECTED);
				reject();						
			} else {// no demands were rejected, receive the order
				receive();
			}
		}
	}
	
	protected void demandRejected(Demand demand) {
//		JSL.out.println("Demand rejected on an order.\n");
//		JSL.out.println("Demand status " + demand.getStatusAsString());
		
		// the demand was rejected during the attempt to receive by its filler
		// indicate that a rejection occurred
		myRejectionFlag = true;
		// increment the number rejected on the order
		myNumRejected++;
		// increment the number asked to be received
		myNumIterated++;
		// check if this was the last demand to receive/reject
		if (myNumIterated == myDemands.size()){
			// the order must be rejected, because at least one demand was rejected
			setOrderStatus(Order.DEMAND_REJECTED);
			reject();				
		}
	}

	protected void demandBackLogged(Demand demand) {
		// the demand was backlogged during the attempt to fill by its filler
		// if the order isn't already backlogged, place it in the backlog state
		if (!isState(Order.BACKLOGGED))
			backLog();	
	}

	protected void notifyDemandListenersOfOrderReceipt(){
		for(Demand d: myDemands){
			d.notifyDemandReceiptListener();
		}		
	}

	protected void notifyDemandListenersOfOrderRejection(){
		for(Demand d: myDemands){
			d.notifyDemandRejectedListener();
		}		
	}
	
	protected void setDemandStateToShipped_(){
		for(Demand d: myDemands){
			d.ship();		
		}		
	}
	
	protected void setDemandStateToDelivered_(){
		for(Demand d: myDemands){
			d.deliver();		
		}		
	}
	
	protected void placeDemandsInSentState(){
		for(Demand d: myDemands){
			d.sent();		
		}		
	}
	
	/** Places the order in the SENT state, can only be called from the IN_PREPARATION state
	 */
	public final void sent(){
		myOrderState.sent(this);
	}

	/** Places the order in the REJECTED state, can only be called from the SENT state
	 */
	public final void reject() {
		myOrderState.reject(this);
	}

	/** Places the order in the IN_PROCESS state, can only be called from the RECEIVED state
	 */
	public final void process(){
		myOrderState.process(this);
	}

	/** Places the order in the RECEIVED state, can only be called from the SENT state
	 */
	public final void receive(){
		myOrderState.receive(this);
	}
	
	/** Places the order in the FILLED state, can be called from the BACKLOGGED and IN_PROCESS states
	 */
	public final void fill(){
		myOrderState.fill(this);
	}
	
	/** Places the order in the BACKLOGGED state, can only be called from the IN_PROCESS state
	 */
	public final void backLog(){
		myOrderState.backlog(this);
	}

	/** Places the order in the SHIPPED state, can only be called
	 *  from the FILLED state
	 */
	public final void ship(){
		myOrderState.ship(this);
	}
	
	/** Places the order in the DELIVERED state, can only be called
	 *  from the SHIPPED state
	 */
	public final void deliver(){
		myOrderState.deliver(this);
	}
	
	/**
	 * @param status
	 */
	final void setOrderStatus(int status) {
		myStatus = status;
	}
		
	//  ===========================================
	//     INNER CLASSES
	//  ===========================================
		
	class DemandListener implements DemandListenerFilledIfc, OrderListenerShippedIfc {

		public void demandFilled(Demand demand) {
//			JSL.out.println(getTime() + "> Demand Filled " + demand.getId());//TODO
			// increment the number of filled demands on the order
			myNumFilledDemands++;
			if(myNumFilledDemands == myDemands.size()) // place the order in the filled state
				fill();
			else {
				if (myDemandOnOrderFilledListener != null)
					myDemandOnOrderFilledListener.demandFilled(demand);
			}
		}

		public void orderShipped(Order order) {
			order.setDemandStateToShipped();
		}		
	}

}
