/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import java.util.List;

import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.utilities.reporting.JSL;

/**
 *
 */
public abstract class OrderFillerAbstract extends SchedulingElement implements OrderFillerIfc {

	/** Indicates whether or not this filler is available
	 */
	private boolean myAvailabilityFlag;
	
	/** Used by the filler to find DemandFillers to fill the demands on the incoming orders
	 */
	protected DemandFillerFinderIfc myDemandFillerFinder;
	
	/**
	 * @param parent
	 */
	public OrderFillerAbstract(ModelElement parent) {
		this(parent, true, null);
	}

	/**
	 * 
	 * @param parent
	 * @param name
	 */
	public OrderFillerAbstract(ModelElement parent, String name) {
		this(parent, true, name);
	}
	
	/**
	 * 
	 * @param parent
	 * @param availabilityFlag
	 * @param name
	 */
	public OrderFillerAbstract(ModelElement parent, boolean availabilityFlag, String name) {
		super(parent, name);
		setAvailability(availabilityFlag);
	}
	
	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.OrderFillerIfc#willFillItemTypes(jsl.supplychainNew.inventorylayer.Order)
	 */
	public final boolean canFillItemTypes(Order order){
		boolean flag = false;
		for(Demand d:getDemandsOnOrder(order))
			flag = canFillItemType(d);
		return(flag);
	}
	
	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.OrderFillerIfc#receive(jsl.supplychainNew.inventorylayer.Order)
	 */
	public void receive(Order order) {
//		JSL.out.println("*********************************************************");
//		JSL.out.println("In LeadTimeOrderFiller.receive(Order order)");
//		JSL.out.println(getTime() +"> " + getName() + " has been asked to fill order " + order.getId() + " order# " + order.getName());
//		JSL.out.println("*********************************************************");

		// determine whether to reject or to try to receive the order
		// check if the order filler is available
		if (!isAvailable()){
			rejectOrder(order, Order.FILLER_UNAVAILABLE);
			return;			
		}
		
		// set up the demand fillers for the demands on the order
		setUpDemandFillersOnOrder(order);
		
		// check demands for rejection, determine order status
		boolean rejected = determineOrderStatus(order);
		if (rejected){
			rejectOrder(order, order.getOrderStatus());
			return;
		}

		// the order will either be backlogged or filled immediately
		// after its demands have been received, its state has not yet
		// changed from SENT
		
		// ask the order to begin trying to have the demands received
		askOrderToBeginReceivingDemands(order);
	}
		
	protected boolean determineOrderStatus(Order order){
		boolean rejected = false;
		boolean immediateFill = true;
		for(Demand d: order.getDemands()){
			DemandFillerIfc f = d.getFiller();
			int status = f.determineRequestStatus(d);
			d.setDemandStatus(status);
			if (f.willReject(d)){ // demand will be rejected
				rejected = true;
			} else { // demand will not be rejected
				if (status == Demand.WILL_BE_BACKLOGGED)
					immediateFill = false;			
			}
		}
		if (rejected)
			order.setOrderStatus(Order.DEMAND_REJECTED);
		else {
			if (immediateFill)
				order.setOrderStatus(Order.IMMEDIATE_FILL);
			else
				order.setOrderStatus(Order.WILL_BE_BACKLOGGED);
		}

		return(rejected);
	}
	
	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.OrderFillerIfc#isAvailable()
	 */
	public final boolean isAvailable() {
		return myAvailabilityFlag;
	}
		
	/**
	 * @param availabilityFlag The availabilityFlag to set.
	 */
	protected final void setAvailability(boolean availabilityFlag) {
		myAvailabilityFlag = availabilityFlag;
	}
		
	/**
	 * @param finder The demandFillerFinder to set.
	 */
	protected final void setDemandFillerFinder(DemandFillerFinderIfc finder) {
		if (finder == null)
			throw new IllegalArgumentException("The supplied finder was null.");
		
		myDemandFillerFinder = finder;
	}

	protected final DemandFillerFinderIfc getDemandFillerFinder() {
		return myDemandFillerFinder;
	}

	/** Allows subclasses of OrderFillerAbstract to place the order in the in process
	 *  state
	 * 
	 * @param order
	 */
	protected final void placeOrderInProcessState(Order order){
		order.process();
	}

	/** Allows subclasses of OrderFillerAbstract to reject the order with
	 *  the given status
	 * 
	 * @param order
	 * @param status
	 */
	protected final void rejectOrder(Order order, int status){
		order.setOrderStatus(status);
		order.reject();		
	}
	
	/** Allows subclasses to setup the demand fillers on the supplied order
	 * 
	 * @param order
	 */
	protected void setUpDemandFillersOnOrder(Order order){
		for(Demand d: order.getDemands()){
			DemandFillerIfc filler = myDemandFillerFinder.findDemandFiller(d);
			if (filler == null)
				throw new NoDemandFillerFoundException("No filler found for demand " + d + " in order " + this);
			d.setFiller(filler);
		}
	}
	
	/** Allows subclasses to ask the order to begin receiving demands.
	 *  This can be overridden in subclasses to allow the order filler to
	 *  re-sequence the demands on the order.  The default is to receive
	 *  the demands in the sequence listed on the order.
	 * 
	 * @param order
	 */
	protected void askOrderToBeginReceivingDemands(Order order){
		order.beginReceivingDemands_();
	}

	/** Allows subclasses to ask the order to begin processing demands.
	 *  This can be overridden in subclasses to allow the order filler to
	 *  re-sequence the demands on the order.  The default is to process
	 *  the demands in the sequence listed on the order.
	 * 
	 * @param order
	 */
	protected void askOrderToBeginProcessingDemands(Order order){
		order.beginProcessingDemands_();
	}
	
	/** Allows subclasses to get the list of demands on an order
	 * 
	 * @param order
	 * @return
	 */
	protected List<Demand> getDemandsOnOrder(Order order){
		return(order.getDemands());
	}
	

	//  ===========================================
	//     INNER CLASSES
	//  ===========================================


}
