/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import jsl.utilities.IdentityIfc;

/** Objects that implement the OrderFillerIfc can receive orders
 *  The implication is that the receiver can or will handle the order in
 *  such a way that the order will eventually be filled. 
 * 
 *
 */
public interface OrderFillerIfc extends IdentityIfc, AvailabilityIfc {
	
	/** This method allows the filler to receive Orders from clients. Once an
	 *  order has been successfully received by a filler, it cannot be sent
	 *  to another order filler.
	 *  
	 *  If the received order cannot be received by this OrderFiller
	 *  then the OrderFiller should return false and should act as if
	 *  it has not received the order.  The order should be rejected (placed in
	 *  the rejected state).  The filler *may* indicate on the Order
	 *  what conditions caused the rejection; however, the filler does not promise
	 *  to do this.  It is up to the client to properly interpret the order
	 *  after calling this method. 
	 *  
	 *  An order that does not permit backlogging can still be sent to a filler that
	 *  may backlog orders.  If the order (that does not permit backlogging)
	 *  can be processed without backlogging then it should be received. If it cannot
	 *  be processed without backlogging, then it should be rejected.
	 *  
	 *  A backloggable order can still be sent to a filler that does not permit backlogging.
	 *  If the backloggable order can be processed without backlogging, then it should be;
	 *  otherwise it should be rejected.
	 *  
	 *  A partial shipment order sent to a filler that does not partially ship orders is still valid.
	 *  The order simply will not be partially shipped and the customer cannot expect it to be partially shipped
	 *  
	 *  An order that is not allowed to be partially shipped sent to a filler that partially ships orders 
	 *  can still be valid.  The filler may decide to reject the order in this case.
	 *  If the filler accepts the order, then filler is responsible for not partially shipping the order.	
	 *  
	 * @param order
	 * @return true if the OrderFiller received and will process the Order
	 */
	public void receive(Order order);

	/** This method allows the filler to fill orders from clients. Once an
	 *  order has been successfully received by a filler, it can only be filled
	 *  by the filler that received it.  This method asks the filler to fill an order
	 *  that has already been received by the filler.  A received order must be placed in process
	 *  at the same simulation time that it was received.  In otherwords, no simulation time must elapse
	 *  between receipt and beginning the fill process.
	 *  
	 * @param demand
	 */
	public void fill(Order order);
	
	/** Allows clients to check if the ItemTypes on the order can be received.  This can
	 *  be used before calling fillOrder(Order order).  This only promises to
	 *  check if every ItemType associated with the order can be filled by
	 *  this filler.  This does not check backlogging/partial filling requirements.
	 * 
	 * @param order
	 * @return True if the filler is willing to fill the order's item types
	 */
	public boolean canFillItemTypes(Order order);
	
	/** Allows clients to check if the supplied ItemType can be filled.  This can
	 *  be used before calling fillOrder(Order order)
	 * 
	 * @param demand
	 * @return True if the filler is willing to fill the ItemType of the given demand
	 */
	public boolean canFillItemType(Demand demand);
			
	/** Asks the filler to negotiate on the order.  The returned message
	 *  should contain the results of the negotiation.  If the filler
	 *  cannot negotiate then this method should return null.
	 *  
	 *  Attempting to negotiate with an unavailable filler should indicate that the filler is unavailable
	 * 
	 * @param order
	 * @return
	 */
	public OrderMessageIfc negotiate(Order order);
		
}
