/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.EventGenerator;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.reporting.JSL;

public class OrderGenerator extends EventGenerator implements OrderSenderIfc {

	/** Responsible for creating an appropriately configured order
	 */
	protected OrderCreatorIfc myOrderCreator;

	/** Responsible for finding an appropriate filler for the order generator
	 */
	private OrderFillerFinderIfc myOrderFillerFinder;
	
	/** This attribute can be used to directly set the filler for this order generator
	 */
	private OrderFillerIfc myOrderFiller;

	/** Listens for and reacts to rejections on the orders
	 *  Listens for and reacts to when the order is received
	 */
	protected OrderListener myOrderListener =  new OrderListener();

	/**
	 * 
	 * @param parent
	 */
	public OrderGenerator(ModelElement parent) {
		this(parent, null, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * 
	 * @param parent
	 * @param orderCreator
	 */
	public OrderGenerator(ModelElement parent, String name) {
		this(parent, null, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * 
	 * @param parent
	 * @param orderCreator
	 */
	public OrderGenerator(ModelElement parent, OrderCreatorIfc orderCreator) {
		this(parent, orderCreator, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * 
	 * @param parent
	 * @param orderCreator
	 * @param name
	 */
	public OrderGenerator(ModelElement parent, OrderCreatorIfc orderCreator, String name) {
		this(parent, orderCreator, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * 
	 * @param parent
	 * @param orderCreator
	 * @param timeUntilNext
	 */
	public OrderGenerator(ModelElement parent, OrderCreatorIfc orderCreator, RandomIfc timeUntilNext) {
		this(parent, orderCreator, null, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}
	
	/**
	 * 
	 * @param parent
	 * @param orderCreator
	 * @param timeUntilNext
	 * @param name
	 */
	public OrderGenerator(ModelElement parent, OrderCreatorIfc orderCreator, RandomIfc timeUntilNext, String name) {
		this(parent, orderCreator, null, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * 
	 * @param parent
	 * @param orderCreator
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 */
	public OrderGenerator(ModelElement parent, OrderCreatorIfc orderCreator, RandomIfc timeUntilFirst, RandomIfc timeUntilNext) {
		this(parent, orderCreator, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}
	
	/**
	 * 
	 * @param parent
	 * @param orderCreator
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param name
	 */
	public OrderGenerator(ModelElement parent, OrderCreatorIfc orderCreator, RandomIfc timeUntilFirst, 
			RandomIfc timeUntilNext, String name) {
		this(parent, orderCreator, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * 
	 * @param parent
	 * @param orderCreator
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 * @param name
	 */
	public OrderGenerator(ModelElement parent, OrderCreatorIfc orderCreator, RandomIfc timeUntilFirst, 
			RandomIfc timeUntilNext, long maxNum, String name) {
		this(parent, orderCreator, timeUntilFirst, timeUntilNext, maxNum, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * 
	 * @param parent
	 * @param orderCreator
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 * @param timeUntilLast
	 * @param name
	 */
	public OrderGenerator(ModelElement parent, OrderCreatorIfc orderCreator, RandomIfc timeUntilFirst, 
			RandomIfc timeUntilNext, long maxNum, double timeUntilLast, String name) {
		super(parent, null, timeUntilFirst,timeUntilNext, maxNum, timeUntilLast, name);
		setOrderCreator(orderCreator);
	}

	/** Sets the order creator, which knows how to properly create the orders
	 *  If null, then no order will be created when the event takes place.
	 * 
	 * @param orderCreator
	 */
	public final void setOrderCreator(OrderCreatorIfc orderCreator) {
		myOrderCreator = orderCreator;
	}

	/** Returns the order creator for this generator
	 * 
	 * @return
	 */
	public OrderCreatorIfc getOrderCreator(){
		return(myOrderCreator);
	}

	/** Returns true if this OrderGenerator may request (create/send)
	 *  orders containing the specified item type
	 * 
	 * @param type
	 * @return
	 */
	public boolean mightRequest(ItemType type) {
		return(myOrderCreator.createsType(type));
	}
	
	/**
	 * @return the OrderFiller
	 */
	public final OrderFillerIfc getOrderFiller() {
		return myOrderFiller;
	}

	/**
	 * @param filler the OrderFiller to set
	 */
	public final void setOrderFiller(OrderFillerIfc filler) {
		if (filler == null)
			throw new IllegalArgumentException("Invalid argument. OrderFiller must be non-null.");
		myOrderFiller = filler;
	}

	/**
	 * @return the OrderFillerFinder
	 */
	public final OrderFillerFinderIfc getOrderFillerFinder() {
		return myOrderFillerFinder;
	}

	/**
	 * @param finder the OrderFillerFinder to set
	 */
	public final void setOrderFillerFinder(OrderFillerFinderIfc finder) {
		if (finder == null)
			throw new IllegalArgumentException("Invalid argument. OrderFillerFinder must be non-null.");
		myOrderFillerFinder = finder;
	}
	
	//  ===========================================
	//      PROTECTED METODS
	//  ===========================================

	/** If there is no filler available then an NoDemandFillerFoundException exception is thrown
	 *  
	 * @param event
	 */
	protected void generate(JSLEvent event){

		if (myOrderCreator != null){
			// create the order
			Order order = myOrderCreator.createOrder();
			order.setName(getId() + "-" + order.getId() + "-" + this.getNumberOfEventsGenerated());
			// set the order filler for the order
			setFillerForOrder(order);
			// set the listeners for the order
			setOrderListeners(order);
			// send the order to the filler
//			JSL.out.println("------------------------------------------------------------------------------");
//			JSL.out.println("In OrderGenerator: generate(JSLEvent event)");
//			JSL.out.println(getTime() + ">" + " Sending order " + order.getId() + " to " + order.getFiller().getName() + " from " + getName());
//		    JSL.out.println(order);
//			JSL.out.println("------------------------------------------------------------------------------");
			sendOrder(order);

		}
		
	}
	
	protected void setFillerForOrder(Order order) {
		// if the filler is set, use it, otherwise try to find a filler using the filler finder
		OrderFillerIfc filler;
		if (myOrderFiller == null){
			filler = myOrderFillerFinder.findOrderFiller(order);
			if (filler == null)
				throw new NoDemandFillerFoundException("No demand filler was found for order: " + order);
		} else
			filler = myOrderFiller;
		
		// set the order filler
		order.setFiller(filler);		
	}
	
	protected void setOrderListeners(Order order) {
		order.setReceivedListener(myOrderListener);
		order.setRejectedListener(myOrderListener);
		order.setDeliveryListener(myOrderListener);
	}
	
	protected void sendOrder(Order order) {
		// indicate the sender of the order
		order.setSender(this);
		// place the order in the SENT state
		order.sent();
		// tell the filler to receive the order
		OrderFillerIfc filler = order.getFiller();
		filler.receive(order);		
	}

	/**  When the generator's order is received by a filler
	 *   this method is called
	 * 
	 * @param order
	 */
	protected void orderReceived(Order order){
//		JSL.out.println("------------------------------------------------------------------------------");
//		JSL.out.println("In OrderGenerator: generate(JSLEvent event)");
//		JSL.out.println("Order " + order.getId() + " was received by " + getName());
//		JSL.out.println("------------------------------------------------------------------------------");
		OrderFillerIfc filler = order.getFiller();
		filler.fill(order);		
	}

	/**  When the generator's order is rejected by a filler
	 *   this method is called.
	 * 
	 * @param order
	 */
	protected void orderRejected(Order order){
		throw new IllegalStateException("The order was rejected");		
	}
	
	/** Can be implemented to react to the order being delivered
	 * 
	 * @param order
	 */
	protected void orderDelivered(Order order){
		
	}
	
	//  ===========================================
	//      PRIVATE INNER CLASSES
	//  ===========================================
 
	private final class OrderListener implements OrderListenerReceivedIfc, OrderListenerRejectedIfc, OrderListenerDeliveredIfc {

		public void orderReceived(Order order) {
			OrderGenerator.this.orderReceived(order);
		}
		
		public void orderRejected(Order order) {
			OrderGenerator.this.orderRejected(order);
		}

		public void orderDelivered(Order order) {
			OrderGenerator.this.orderDelivered(order);
		}
	}



}
