/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import java.util.HashMap;
import java.util.Map;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Bernoulli;
import jsl.utilities.random.distributions.Constant;

public class RandomOrderCreator extends ModelElement implements OrderCreatorIfc {

	/** Indicated whether or not the orders permit backlogging 
	 */
	private boolean myPermitBackLoggingFlag;

	protected final static Constant ONE = new Constant(1.0);

	protected Map<ItemType, ItemTypeInfo> myItemTypes;
	
	/**
	 * @param parent
	 */
	public RandomOrderCreator(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public RandomOrderCreator(ModelElement parent, String name) {
		super(parent, name);
		myPermitBackLoggingFlag = true;
		myItemTypes = new HashMap<ItemType, ItemTypeInfo>();
	}

	/** Adds the ItemType to the order creator with default amount of 1 unit
	 *  and probability of 1.0 that it will be included on the order
	 * 
	 * @param type The type of item
	 */
	public void addItemTypeDistribution(ItemType type){
		addItemTypeDistribution(type, 1.0, ONE);
	}
	
	/** Adds the ItemType to the order creator with default amount of 1 unit
	 * 
	 * @param type The type of item
	 * @param prob The probability that the item should be included on the order
	 */
	public void addItemTypeDistribution(ItemType type, double prob){
		addItemTypeDistribution(type, prob, ONE);
	}

	/** Adds the ItemType to the order creator
	 * 
	 * @param type The type of item
	 * @param prob The probability that the item should be included on the order
	 * @param distribution The distribution for the amount of demand if the item is included on the order. 
	 *  Must not produce a zero.  The amount must be >=1.0.  For non integral amounts, it will be truncated to
	 *  an integer
	 */
	public void addItemTypeDistribution(ItemType type, double prob, RandomIfc distribution){
		if (type == null)
			throw new IllegalArgumentException("The ItemType cannot be null");
		
	    if ((prob < 0.0) || (prob > 1.0))
	            throw new IllegalArgumentException("Probability must be [0,1]");
		
		if (distribution == null)
			throw new IllegalArgumentException("The RandomIfc cannot be null");
			
		if (myItemTypes.containsKey(type))
			throw new IllegalArgumentException("The ItemType has already been added.");
		
		ItemTypeInfo x = new ItemTypeInfo();
		x.amtRV = new RandomVariable(this, distribution, 1, Double.POSITIVE_INFINITY);
		x.includeRV = new RandomVariable(this, new Bernoulli(prob));
		myItemTypes.put(type, x);
	}
	
	/** Returns true if this OrderCreator may create the given type
	 * 
	 * @param type
	 * @return
	 */
	public final boolean createsType(ItemType type){
		return(myItemTypes.containsKey(type));
	}

	/**
	 * @return
	 */
	public final boolean getPermitBackLoggingFlag() {
		return myPermitBackLoggingFlag;
	}
	
	/**
	 * @param permitBackLoggingFlag
	 */
	public final void setPermitBackLoggingFlag(boolean permitBackLoggingFlag) {
		myPermitBackLoggingFlag = permitBackLoggingFlag;
	}
	
	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.OrderCreatorIfc#createOrder()
	 */
	public Order createOrder() {
		Order order = null;;
		for(ItemType type: myItemTypes.keySet()){
			ItemTypeInfo x = myItemTypes.get(type);
			if (x.includeRV.getValue() > 0){
				if (order == null){
					order = new Order();
					order.setBackLoggingFlag(myPermitBackLoggingFlag);
				}
				Demand d = new Demand(type, (int)x.amtRV.getValue());
				d.setAllowBackLoggingFlag(myPermitBackLoggingFlag);
				order.addDemand(d);
			}
		}

		return order;
	}
	
	protected class ItemTypeInfo {
		protected RandomVariable amtRV;
		protected RandomVariable includeRV;
	}
}
