/*
* Created on Mar 7, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.utilities.reporting.JSL;

/** A concrete implementation of WarehouseAbstract which holds incoming orders in a 
 *  queue until all demands on the order are filled.  When an order is filled, this 
 *  class looks for an OrderShipper, if not found, it immediately ships and delivers
 *  the order (with no time delay).
 *
 */
public class Warehouse extends WarehouseAbstract {

	/** Holds the Orders that are to be processed
	 */
	protected Queue myOrderQ;
	
	/** Holds the inventory for the warehouse
	 */
	protected WarehouseInventoryHolder myInventoryHolder; 

	/**
	 * @param parent
	 */
	public Warehouse(ModelElement parent) {
		this(parent, true, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public Warehouse(ModelElement parent, String name) {
		this(parent, true, name);
	}

	/**
	 * @param parent
	 * @param availabilityFlag
	 * @param name
	 */
	public Warehouse(ModelElement parent, boolean availabilityFlag, String name) {
		super(parent, availabilityFlag, name);
		myOrderQ = new Queue(this, getName() + " OrderQ");
		myInventoryHolder = new WarehouseInventoryHolder(this, getName() + "InventoryHolder");
		setDemandFillerFinder(myInventoryHolder);
	}

	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================

	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append(getName() + "\n");
		sb.append("Number of Orders in Process " + myOrderQ.size() + "\n");
		sb.append(myInventoryHolder);
		return(sb.toString());
	}
	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.OrderSenderIfc#mightRequest(jsl.research.supplychain.inventorylayer.ItemType)
	 */
	public final boolean mightRequest(ItemType type) {
		return myInventoryHolder.mightRequest(type);
	}
	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.WarehouseAbstract#fill(jsl.research.supplychain.inventorylayer.Order)
	 */
	public void fill(Order order) {
		if (order == null)
			throw new IllegalArgumentException("The order was null");

		placeOrderInProcessState(order);

//		JSL.out.println("------------------------------------------------------------------------------");
//		JSL.out.println("In Warehouse: fill(Order order)");
//		JSL.out.println(order.getFiller().getName() + " was asked to fill order " + order.getId());
//	    JSL.out.println(order.getFiller());
//	    JSL.out.println(this);	    
//		JSL.out.println("------------------------------------------------------------------------------");
	    
		// enqueue the order during processing
		enqueueOrder(order);
		
		askOrderToBeginProcessingDemands(order);
	}

	public final void addInventory(Inventory inventory){
		myInventoryHolder.addInventory(inventory);
	}

	/** Adds a (reorder point, reorder quantity) based inventory that permits backlogging,
	 *  with no initial inventory
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderQty
	 * @return
	 */
	public final Inventory addReorderPointReorderQuantityInventory(ItemType type, int reorderPt, int reorderQty){
		return(myInventoryHolder.addReorderPointReorderQuantityInventory(type, reorderPt, reorderQty, 0, null));
	}
	
	/** Adds a (reorder point, reorder quantity) based inventory that permits backlogging
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderQty
	 * @param initialLevel
	 * @return
	 */
	public final Inventory addReorderPointReorderQuantityInventory(ItemType type, int reorderPt, int reorderQty, int initialLevel){
		return(myInventoryHolder.addReorderPointReorderQuantityInventory(type, reorderPt, reorderQty, initialLevel, null));
	}
	
	/** Adds a (reorder point, reorder quantity) based inventory that permits backlogging
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderQty
	 * @param initialLevel
	 * @param name
	 * @return
	 */
	public final Inventory addReorderPointReorderQuantityInventory(ItemType type, int reorderPt, int reorderQty, int initialLevel, String name){
		return(myInventoryHolder.addReorderPointReorderQuantityInventory(type, reorderPt, reorderQty, initialLevel, name));
	}

	/** Adds a (reorder point, reorder level) based inventory that permits backlogging,
	 *  with no initial inventory
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderLevel
	 * @return
	 */
	public final Inventory addReorderPointOrderUpToLevelInventory(ItemType type, int reorderPt, int reorderLevel){
		return(myInventoryHolder.addReorderPointOrderUpToLevelInventory(type, reorderPt, reorderLevel, 0, null));
	}
	
	/** Adds a (reorder point, reorder level) based inventory that permits backlogging
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderLevel
	 * @param initialLevel
	 * @return
	 */
	public final Inventory addReorderPointOrderUpToLevelInventory(ItemType type, int reorderPt, int reorderLevel, int initialLevel){
		return(myInventoryHolder.addReorderPointOrderUpToLevelInventory(type, reorderPt, reorderLevel, initialLevel, null));
	}
	
	/** Adds a (reorder point, reorder level) based inventory that permits backlogging
	 * 
	 * @param type
	 * @param reorderPt
	 * @param reorderLevel
	 * @param initialLevel
	 * @param name
	 * @return
	 */
	public final Inventory addReorderPointOrderUpToLevelInventory(ItemType type, int reorderPt, int reorderLevel, int initialLevel, String name){
		return(myInventoryHolder.addReorderPointOrderUpToLevelInventory(type, reorderPt, reorderLevel, initialLevel, name));
	}
	
	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.AggregateInventoryIfc#getAggregateAmountBackOrderedResponseVariable()
	 */
	public AggregateTimeWeightedVariable getAggregateAmountBackOrderedResponseVariable() {
		return myInventoryHolder.getAggregateAmountBackOrderedResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.AggregateInventoryIfc#getAggregateAvgCustomerWaitTimeResponseVariable()
	 */
	public ResponseVariableAverageObserver getAggregateAvgCustomerWaitTimeResponseVariable() {
		return myInventoryHolder.getAggregateAvgCustomerWaitTimeResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.AggregateInventoryIfc#getAggregateAvgFirstFillRateResponseVariable()
	 */
	public ResponseVariableAverageObserver getAggregateAvgFirstFillRateResponseVariable() {
		return myInventoryHolder.getAggregateAvgFirstFillRateResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.AggregateInventoryIfc#getAggregateNumberBackOrderedResponseVariable()
	 */
	public AggregateTimeWeightedVariable getAggregateNumberBackOrderedResponseVariable() {
		return myInventoryHolder.getAggregateNumberBackOrderedResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.AggregateInventoryIfc#getAggregateNumberOfReplenishmentDemandsCounter()
	 */
	public AggregateCounter getAggregateNumberOfReplenishmentDemandsCounter() {
		return myInventoryHolder.getAggregateNumberOfReplenishmentDemandsCounter();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.AggregateInventoryIfc#getAggregateOnHandInventoryResponseVariable()
	 */
	public AggregateTimeWeightedVariable getAggregateOnHandInventoryResponseVariable() {
		return myInventoryHolder.getAggregateOnHandInventoryResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.AggregateInventoryIfc#getAggregateAmountOnOrderResponseVariable()
	 */
	public AggregateTimeWeightedVariable getAggregateAmountOnOrderResponseVariable() {
		return myInventoryHolder.getAggregateAmountOnOrderResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.inventorylayer.AggregateInventoryIfc#getAggregateStockOutPercentageResponseVariable()
	 */
	public AggregateTimeWeightedVariable getAggregateStockOutPercentageResponseVariable() {
		return myInventoryHolder.getAggregateStockOutPercentageResponseVariable();
	}

	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#attachAggregateAvgCustomerWaitTime(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void subscribeToAggregateAvgCustomerWaitTime(ResponseVariableAverageObserver aggregate) {
		myInventoryHolder.subscribeToAggregateAvgCustomerWaitTime(aggregate);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#attachAggregateAvgFirstFillRate(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void subscribeToAggregateAvgFirstFillRate(ResponseVariableAverageObserver aggregate) {
		myInventoryHolder.subscribeToAggregateAvgFirstFillRate(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#subscribeToAggregateAmountBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateAmountBackOrdered(AggregateTimeWeightedVariable aggregate) {
		myInventoryHolder.subscribeToAggregateAmountBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#subscribeToAggregateAmountOnOrder(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateAmountOnOrder(AggregateTimeWeightedVariable aggregate) {
		myInventoryHolder.subscribeToAggregateAmountOnOrder(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#subscribeToAggregateNumberBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateNumberBackOrdered(AggregateTimeWeightedVariable aggregate) {
		myInventoryHolder.subscribeToAggregateNumberBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#subscribeToAggregateOnHandInventory(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateOnHandInventory(AggregateTimeWeightedVariable aggregate) {
		myInventoryHolder.subscribeToAggregateOnHandInventory(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#subscribeToAggregateStockOutIndicator(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateStockOutIndicator(AggregateTimeWeightedVariable aggregate) {
		myInventoryHolder.subscribeToAggregateStockOutIndicator(aggregate);
	}

	/**
	 * @param counter
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#subscribeToAggregateReplenishmentDemandCounter(jsl.modeling.elements.variable.AggregateCounter)
	 */
	public void subscribeToAggregateReplenishmentDemandCounter(AggregateCounter counter) {
		myInventoryHolder.subscribeToAggregateReplenishmentDemandCounter(counter);
	}

	
	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#unsubscribeFromAggregateAmountBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateAmountBackOrdered(
			AggregateTimeWeightedVariable aggregate) {
		myInventoryHolder.unsubscribeFromAggregateAmountBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#unsubscribeFromAggregateAmountOnOrder(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateAmountOnOrder(
			AggregateTimeWeightedVariable aggregate) {
		myInventoryHolder.unsubscribeFromAggregateAmountOnOrder(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#unsubscribeFromAggregateAvgCustomerWaitTime(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void unsubscribeFromAggregateAvgCustomerWaitTime(
			ResponseVariableAverageObserver aggregate) {
		myInventoryHolder.unsubscribeFromAggregateAvgCustomerWaitTime(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#unsubscribeFromAggregateAvgFirstFillRate(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void unsubscribeFromAggregateAvgFirstFillRate(
			ResponseVariableAverageObserver aggregate) {
		myInventoryHolder.unsubscribeFromAggregateAvgFirstFillRate(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#unsubscribeFromAggregateNumberBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateNumberBackOrdered(
			AggregateTimeWeightedVariable aggregate) {
		myInventoryHolder.unsubscribeFromAggregateNumberBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#unsubscribeFromAggregateOnHandInventory(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateOnHandInventory(
			AggregateTimeWeightedVariable aggregate) {
		myInventoryHolder.unsubscribeFromAggregateOnHandInventory(aggregate);
	}

	/**
	 * @param counter
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#unsubscribeFromAggregateReplenishmentDemandCounter(jsl.modeling.elements.variable.AggregateCounter)
	 */
	public void unsubscribeFromAggregateReplenishmentDemandCounter(
			AggregateCounter counter) {
		myInventoryHolder.unsubscribeFromAggregateReplenishmentDemandCounter(counter);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.inventorylayer.InventoryHolderAbstract#unsubscribeFromAggregateStockOutIndicator(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateStockOutIndicator(
			AggregateTimeWeightedVariable aggregate) {
		myInventoryHolder.unsubscribeFromAggregateStockOutIndicator(aggregate);
	}

	@Override
	public void attachAggregateResponseVariableToOrderWaitingTime(ResponseVariableAverageObserver aggregate) {
		myOrderQ.subscribe(aggregate);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.WarehouseAbstract#attachNumberOfOrdersWaitingObserver(java.util.Observer)
	 */
	public void attachAggregateTimeWeightedVariableToNumberOfOrdersWaiting(AggregateTimeWeightedVariable aggregate) {
		myOrderQ.subscribe(aggregate);
	}

	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.WarehouseAbstract#orderFilled(jsl.research.supplychain.inventorylayer.Order)
	 */
	protected void orderFilled(Order order) {
		// dequeue the order
		dequeueOrder(order);
		if (myOrderShipper != null){
			myOrderShipper.transportOrder(order);
		}
		else {
			order.ship();
			order.deliver();
		}
	}
	
    /** Enqueues the order, sets the order's filled listener to the warehouse,
     *  and ensures that the demand on order filled listener is set
     * 
     * @param order
     */
	protected void enqueueOrder(Order order){
		// enqueue the order for processing
		myOrderQ.enqueue(order);
		// tell the order that the warehouse is the fill listener
		order.setFilledListener(myOrderListener);
		order.setDemandOnOrderFilledListener(myOrderListener);   	
    }

	/** Dequeus the order, sets the order's filled listener and demand on order
	 *  listener to null
	 * 
	 * @param order
	 */
	protected void dequeueOrder(Order order){
		myOrderQ.remove(order);
		order.setFilledListener(null);
		order.setDemandOnOrderFilledListener(null);
	}

	protected void replenishmentOrderDelivered(Order order) {
		// TODO Auto-generated method stub
		// the order has been delivered, ensure that
		// each demand on the order is placed in its delivered state
		order.setDemandStateToDelivered();
	}
	
	//  ===========================================
	//      INNER CLASSES
	//  ===========================================

	/** Warehouse inner class used to develop a specialized InventoryHolderAbstract for
	 *  internal inventory within the warehouse.  Acts to ensure the filling of demands from
	 *  the outside and ensures that internal inventory can eventually be replenished.
	 */
	private final class WarehouseInventoryHolder extends InventoryHolderAbstract implements DemandFillerFinderIfc,
		ReplenishmentRequesterIfc {

		public WarehouseInventoryHolder(ModelElement parent) {
			this(parent, null);
		}
		
		public WarehouseInventoryHolder(ModelElement parent, String name) {
			super(parent, name);
			setReplenishmentRequester(this);
		}

		public void fillDemand(Demand demand) {
			DemandFillerIfc inventory = demand.getFiller();
			inventory.fillDemand(demand);
		}

		public void receive(Demand demand) {
			// determine whether to reject or process the demand
			
			if (!isAvailable()){
				demand.setDemandStatus(Demand.FILLER_UNAVAILABLE);
				demand.reject();
				return;			
			}
			
			Inventory inventory = myInventory.get(demand.getItemType());
			
			if (inventory == null){
				demand.setDemandStatus(Demand.ITEMTYPE_MISMATCH);
				demand.reject();
				return;
			}
			// set the filler for the demand, the order listens for fills
			demand.setFiller(inventory); //TODO
			// now ask the inventory to receive the demand
			inventory.receive(demand);			
		}

		public final DemandFillerIfc findDemandFiller(Demand demand) {
			return myInventory.get(demand.getItemType());
		}

		/* (non-Javadoc)
		 * @see jsl.research.supplychain.inventorylayer.ReplenishmentRequesterIfc#requestDemand(jsl.research.supplychain.inventorylayer.Inventory, jsl.research.supplychain.inventorylayer.Demand)
		 */
		public final void requestReplenishment(Inventory inventory, Demand demand) {
			Warehouse.this.requestReplenishment(inventory, demand);
		}

		/** Returns null
		 * 
		 */
		public DemandFillerIfc getDemandFiller() {
			return null;
		}

		/** Throws UnsupportedOperationException
		 * 
		 */
		public void setDemandFiller(DemandFillerIfc filler) {
			throw new UnsupportedOperationException("The method is not supported");
			
		}
		
		/** Returns itself
		 * 
		 */
		public DemandFillerFinderIfc getDemandFillerFinder() {
			return this;
		}

		/** Throws UnsupportedOperationException
		 * 
		 */
		public void setDemandFillerFinder(DemandFillerFinderIfc finder) {
			throw new UnsupportedOperationException("The method is not supported");
			
		}

	}

}
