/*
* Created on Mar 7, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer;

import java.util.Observer;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.utilities.reporting.JSL;
import jslx.supplychain.transportlayer.OrderCarrierIfc;

/**  A warehouse holds inventory for the purpose of filling orders. It extends
 *   the OrderFillerAbstract class so as to act like an order filler (OrderFillerIfc)
 *   Sub-classes of this abstract base class must provide mechanisms for holding 
 *   the inventory and filling the received orders from the inventory. Since orders
 *   may be filled from local stocks, a warehouse is also responsible for ensuring
 *   that the orders are shipped to their customers after they are filled.  This
 *   can be accomplished by supplying an OrderShipperIfc (see field comments).
 *   Since warehouse extends OrderFillerAbstract, sub-classes must provide an
 *   instance of a class that implements the DemandFillerFinderIfc to find
 *   a DemandFillerIfc for the demands on the orders. Since we assume that a
 *   warehouse also may hold local inventory sub-classes must ensure that the
 *   internal inventory can be replenished.
 *
 */
abstract public class WarehouseAbstract extends OrderFillerAbstract implements OrderSenderIfc, AggregateInventoryIfc{

	/** Indicated whether or not the warehouse's replenishment orders permit backlogging 
	 *  The default is true.
	 */
	private boolean myPermitBackLoggingFlag;

	/** The shipper that sends filled order to this warehouse's customers
	 *  If supplied, can be used to ensure that a filled order
	 *  is shipped back to the customer.  Thus, this attribute may be null.
	 *  If supplied, subclasses may use the shipper when implementing the
	 *  orderFilled() method to ensure shipping after order filling.  Thus,
	 *  sub-classes are responsible for providing shipping behavior after
	 *  order filling.
	 *  
	 */
	protected OrderCarrierIfc myOrderShipper;

	/** Listens for when an order is filled.  A default OrderListener
	 *  is supplied which ensures that 1) the warehouse's orderFilled() method
	 *  is called when an order is filled and ensures that demandOnOrderFilled() or 
	 *  a supplied DemandOnOrderFilledListenerIfc is called when individual
	 *  demands on an order are filled.
	 */
	protected OrderListener myOrderListener = new OrderListener();

	/** Can be supplied to add behavior to react to when a demand inside an order 
	 *  is filled.  For example, partial shipping.  Subclasses can directly supply
	 *  this or allow it to be set
	 */
	protected DemandOnOrderFilledListenerIfc myDemandOnOrderFilledListener;

	/** Counts the number of replenishment orders made by the warehouse
	 */
	private Counter myCountReplenishmentOrders;

	/** Can be supplied to add behavior to build orders (consisting of demands)
	 *  that are required to be sent out for replenishment of inventory at the warehouse
	 * 
	 */
	protected ReplenishmentRequesterIfc myReplenishmentRequester;
	
	/** Responsible for finding an appropriate replenishment order filler for the warehouse
	 */
	protected OrderFillerFinderIfc myReplenishmentOrderFillerFinder;
	
	/** This attribute can be used to directly set the filler for replenishment orders for the warehouse
	 */
	protected OrderFillerIfc myReplenishmentOrderFiller;
	
	/** Listens for and reacts to rejections on the replenishment orders
	 *  Listens for and reacts to when the replenishment order is received
	 */
	protected ReplenishmentOrderListener myReplenishmentOrderListener =  new ReplenishmentOrderListener();
	
	/**
	 * @param parent
	 */
	public WarehouseAbstract(ModelElement parent) {
		this(parent, true, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public WarehouseAbstract(ModelElement parent, String name) {
		this(parent, true, name);
	}

	/**
	 * @param parent
	 * @param availabilityFlag
	 * @param name
	 */
	public WarehouseAbstract(ModelElement parent, boolean availabilityFlag, String name) {
		super(parent, availabilityFlag, name);
		myPermitBackLoggingFlag = true;
		myCountReplenishmentOrders = new Counter(this, getName() + "-#ReplenishmentOrders");
	}

	//  ===========================================
	//      PUBLIC METHODS
	//  ===========================================
	
	/** Sets the shipper that this warehouse will use when sending filled orders
	 *  to customers of its inventory.
	 *  
	 * @param shipper
	 */
	public final void setOrderShipper(OrderCarrierIfc shipper) {
		myOrderShipper = shipper;
	}

	/** Gets the replenishment order filler for this warehouse.  This
	 *  may be null if the warehouse uses an OrderFillerFinder
	 *  
	 * @return the OrderFiller
	 */
	public final OrderFillerIfc getReplenishmentOrderFiller() {
		return myReplenishmentOrderFiller;
	}

	/** Gets the object that will find an appropriate order filler to satisfy
	 *  replenishment orders for the warehouse
	 *  
	 * @return the OrderFillerFinder
	 */
	public final OrderFillerFinderIfc getReplenishmentOrderFillerFinder() {
		return myReplenishmentOrderFillerFinder;
	}

	/** Sets the object that will satisfy replenishment orders for the warehouse.
	 *  Once set, the OrderFillerFinder will not be used
	 * 
	 * @param filler the OrderFiller to set, must not be null
	 */
	public final void setReplenishmentOrderFiller(OrderFillerIfc filler) {
		if (filler == null)
			throw new IllegalArgumentException("Invalid argument. OrderFiller must be non-null.");
		myReplenishmentOrderFiller = filler;
	}

	/** Sets the object that will find an appropriate order filler to satisfy
	 *  replenishment orders for the warehouse
	 *  
	 * @param finder the OrderFillerFinder to set, must not be null
	 */
	public final void setReplenismentOrderFillerFinder(OrderFillerFinderIfc finder) {
		if (finder == null)
			throw new IllegalArgumentException("Invalid argument. OrderFillerFinder must be non-null.");
		 myReplenishmentOrderFillerFinder = finder;
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.WarehouseAbstract#canFillItemType(jsl.research.supplychain.inventorylayer.Demand)
	 */
	public boolean canFillItemType(Demand demand) {
		DemandFillerIfc filler = myDemandFillerFinder.findDemandFiller(demand);
		if (filler == null)
			return (false);
		else
			return(filler.canFillItemType(demand));
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.WarehouseAbstract#negotiate(jsl.research.supplychain.inventorylayer.Order)
	 */
	public OrderMessageIfc negotiate(Order order) {
		OrderMessage om = new OrderMessage(this);
		for(Demand d: order.getDemands()){
			DemandFillerIfc filler = myDemandFillerFinder.findDemandFiller(d);
			if (filler != null){
				DemandMessageIfc dm = filler.negotiate(d);
				om.add(dm);
			} else
				om.add(null);
		}
		return om;
	}

	/**
	 * @return
	 */
	public final boolean getPermitBackLoggingFlag() {
		return myPermitBackLoggingFlag;
	}
	
	/**
	 * @param permitBackLoggingFlag
	 */
	public final void setPermitBackLoggingFlag(boolean permitBackLoggingFlag) {
		myPermitBackLoggingFlag = permitBackLoggingFlag;
	}
	
	/** Attaches an observer to be updated when the counter that counts
	 *  the number of orders sent for replenisment is updated
	 * 
	 * @param observer
	 */
	public final void addReplenishmentOrderCounterObserver(Observer observer) {
		myCountReplenishmentOrders.addObserver(observer);
	}
	
	/** Attaches an AggregateCounter to the replenishment order counter
	 * 
	 * @param counter
	 */
	public final void attachAggregateReplenishmentOrderCounter(AggregateCounter counter){
		counter.subscribeTo(myCountReplenishmentOrders);
	}
	
	//  ===========================================
	//      ABSTRACT PUBLIC METHODS
	//  ===========================================

	abstract public void fill(Order order);

	/** Attaches an AggregateTimeWeightedVariable to be updated when the number of waiting orders changes
	 *
	 * @param aggregate
	 */
	 abstract public void attachAggregateTimeWeightedVariableToNumberOfOrdersWaiting(AggregateTimeWeightedVariable aggregate);
	
	/** Attaches an observer to be updated to observe order wait times
	 * 
	 * @param observer
	 */
	 abstract public void attachAggregateResponseVariableToOrderWaitingTime(ResponseVariableAverageObserver aggregate);
	 	
	//  ===========================================
	//      ABSTRACT PROTECTED METHODS
	//  ===========================================
	
	/** Called when all the demands on the order are filled.  Subclasses need
	 *  to implement this method to provide behavior to react to an that the
	 *  warehouse is processing being filled (e.g. shipping the order)
	 * 
	 * @param order
	 */
	abstract protected void orderFilled(Order order);
	
	/** Called when the replenishment order has been delivered.  Subclasses need
	 *  to implement this method to provide behavior to react to an order being
	 *  delivered (e.g. ensuring that the inventory is notified of the replenishment
	 *  of the associated demands on the order.)
	 * 
	 * @param order
	 */
	abstract protected void replenishmentOrderDelivered(Order order);
	
	//  ===========================================
	//      PROTECTED METHODS
	//  ===========================================

	protected void afterReplication(){
		super.afterReplication();
//		JSL.out.println("After replication -------------:");
//		JSL.out.println(this);
	}
	
	/** This method should take in the supplied demand and ensure that it gets processed/placed on
	 *  an appropriate outgoing order so that the warehouse's internal inventory will get replensished
	 *  If an instance of ReplenishmentRequester is supplied it will be used; otherwise, the
	 *  default behavior is to form an order with a single demand and send it.  Alternatively,
	 *  this method can be overridden in sub-classes to perform the necessary actions.
	 *  
	 *  Note:  Subclasses need to ensure that internal inventory uses this method
	 *  
	 * @param demand
	 */
	protected void requestReplenishment(Inventory inventory, Demand demand){
		
		if (myReplenishmentRequester != null)
			myReplenishmentRequester.requestReplenishment(inventory, demand);
		else {
			// create the order
//			JSL.out.println("------------------------------------");
//			JSL.out.println("In WarehouseAbstract.requestReplenishment(Inventory inventory, Demand demand)");
//			JSL.out.println("*** The " + getName() + " is creating a replenishment order.");
//			JSL.out.println("------------------------------------");
			Order order = new Order();
			order.setBackLoggingFlag(getPermitBackLoggingFlag());
			// make the demand consistent with the order's backlogging
			demand.setAllowBackLoggingFlag(getPermitBackLoggingFlag());
			// add the demand to the order
			order.addDemand(demand);
			// tell the order who the filler will be
			setFillerForReplenishmentOrder(order);
			// set up to be the receive and rejection listeners for the order
			setOrderListeners(order);
			// ask the order filler to receive the order
			sendOrder(order);
		}
	}
	
	/** Sets the order listeners for the replenishment order to this warehouse's listeners
	 *  for receipt and rejection
	 * 
	 * @param order
	 */
	protected void setOrderListeners(Order order) {
		order.setReceivedListener(myReplenishmentOrderListener);
		order.setRejectedListener(myReplenishmentOrderListener);
		order.setDeliveryListener(myReplenishmentOrderListener);
	}
	
	/** Asks the replenishment order filler to receive the replenishment order
	 * 
	 * @param order
	 */
	protected void sendOrder(Order order) {
		myCountReplenishmentOrders.increment();
		int i = (int)myCountReplenishmentOrders.getValue();
		order.setName(getId() + "-" + order.getId() + "-" + i);
//		JSL.out.println("------------------------------------");
//		JSL.out.println("In WarehouseAbstract.sendOrder(Order order)");
//		JSL.out.println(getTime() + ">" + " Sending order " + order.getId() + " to " + order.getFiller().getName() + " from " + getName());
//	    JSL.out.println(order);
//		JSL.out.println("------------------------------------");
		// indicate the sender of the order
		order.setSender(this);
		// place the order in the SENT state
		order.sent();
		// tell the filler to receive the order
		OrderFillerIfc filler = order.getFiller();
		filler.receive(order);		
	}
	
	/** Subclasses can override this method to provide behavior for when an individual
	 *  demand on an order is filled.  This could be useful for implementing partial shipping.
	 *  Alternatively clients can provide a DemandOnOrderFilledListenerIfc via 
	 *  setDemandOnOrderFilledListener().  If a DemandOnOrderFilledListenerIfc is provided, then
	 *  this method is not called.  The default behavior for this method is to do nothing.
	 * 
	 * @param demand
	 */
	protected void demandOnOrderFilled(Demand demand) {
		
	}

	/** Gets the shipper for this warehouse
	 * 
	 * @return
	 */
	public final OrderCarrierIfc getOrderShipper() {
		return myOrderShipper;
	}

	/**  When the replenishment order is received by an order filler
	 *   this method is called
	 * 
	 * @param order
	 */
	protected void replenishmentOrderReceived(Order order){
		OrderFillerIfc filler = order.getFiller();
		filler.fill(order);		
	}

	/**  When the replenishment order is rejected by an order filler
	 *   this method is called.
	 * 
	 * @param order
	 */
	protected void replenishmentOrderRejected(Order order){
		throw new IllegalStateException("The order was rejected");		
	}
	
	/** Sets the filler for a replenishment order based on the replenishment order filler finder
	 *  or the direct filler for the facility
	 * 
	 * @param order
	 */
	protected final void setFillerForReplenishmentOrder(Order order) {
		// if the filler is set, use it, otherwise try to find a filler using the filler finder
		OrderFillerIfc filler;
		if (myReplenishmentOrderFiller == null){
			filler = myReplenishmentOrderFillerFinder.findOrderFiller(order);
			if (filler == null)
				throw new NoDemandFillerFoundException("No demand filler was found for order: " + order);
		} else
			filler = myReplenishmentOrderFiller;
		
		// set the order filler
		order.setFiller(filler);		
	}
	
	//  ===========================================
	//      INNER CLASSES
	//  ===========================================

	/** An inner class that listens on orders to ensure the reaction of the warehouse
	 *  to orders being filled or individual demands on the orders being filled.
	 */
	private final class OrderListener implements OrderListenerFilledIfc, DemandListenerFilledIfc {

		public void orderFilled(Order order) {
			WarehouseAbstract.this.orderFilled(order);
		}
		
		public void demandFilled(Demand demand) {
			if (myDemandOnOrderFilledListener == null)
				demandOnOrderFilled(demand);
			else
				myDemandOnOrderFilledListener.demandFilled(WarehouseAbstract.this, demand);	
		}

	}
	
	/** An inner class that listens on replenishment order (outgoing) to ensure the reaction of
	 *  the warehouse to them being received or rejected.
	 *
	 */
	private final class ReplenishmentOrderListener implements OrderListenerReceivedIfc, OrderListenerRejectedIfc,
		OrderListenerDeliveredIfc {

		public void orderReceived(Order order) {
			WarehouseAbstract.this.replenishmentOrderReceived(order);
		}
		
		public void orderRejected(Order order) {
			WarehouseAbstract.this.replenishmentOrderRejected(order);
		}

		public void orderDelivered(Order order) {
			WarehouseAbstract.this.replenishmentOrderDelivered(order);
		}
	}

}
