/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer.testing;

import jsl.modeling.Experiment;
import jsl.modeling.Model;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jslx.supplychain.TimeBasedShippingMultiEchelonIHPNetwork;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;

public class TestMultiEchelonInventoryNetworkExampleSS {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		System.out.println("MultiEchelonInventoryNetwork Test");
        
//		JSL.LOGGER.setLevel(Level.INFO);
   
//		testReplication();
		testExperiment();
        
		System.out.println("Done!");	
	}

	public static Model buildModel(){
		Model m = Model.createModel();
		
		//Create the network
		TimeBasedShippingMultiEchelonIHPNetwork n = new TimeBasedShippingMultiEchelonIHPNetwork(m, "ME-Inventory-Network");
		
		//Define item types with associated leadtime for production
		DistributionIfc ltp1 = new Exponential(1.0);
		DistributionIfc ltp2 = new Exponential(0.5);
		DistributionIfc ltp3 = new Exponential(1.5);
		DistributionIfc ltp4 = new Exponential(2.0);
		
		ItemType type1 = n.addItemType("Type-1", ltp1);
		ItemType type2 = n.addItemType("Type-2", ltp2);
		ItemType type3 = n.addItemType("Type-3", ltp3);
		ItemType type4 = n.addItemType("Type-4", ltp4);
				
		//Define Warehouse
		InventoryHoldingPoint warehouse = n.addInventoryHoldingPoint("Warehouse");
		
		//addInventory to warehouse
		warehouse.addReorderPointReorderQuantityInventory(type1,4,1,20);
		warehouse.addReorderPointReorderQuantityInventory(type2,5,1,20);
		warehouse.addReorderPointReorderQuantityInventory(type3,3,2,20);
		warehouse.addReorderPointReorderQuantityInventory(type4,4,2,20);
				
		// Create the retailers which use the warehouse as its filler
		InventoryHoldingPoint retailer1 = n.addInventoryHoldingPoint("R1");
		InventoryHoldingPoint retailer2 = n.addInventoryHoldingPoint("R2");
		InventoryHoldingPoint retailer3 = n.addInventoryHoldingPoint("R3");
		InventoryHoldingPoint retailer4 = n.addInventoryHoldingPoint("R4");
		InventoryHoldingPoint retailer5 = n.addInventoryHoldingPoint("R5");
		
		//Create the inventory at each retailer
		retailer1.addReorderPointOrderUpToLevelInventory(type1,2,3,10);
		retailer1.addReorderPointOrderUpToLevelInventory(type2,1,2,10);
		retailer1.addReorderPointOrderUpToLevelInventory(type3,2,4,10);
		retailer1.addReorderPointOrderUpToLevelInventory(type4,3,6,10);
				
		retailer2.addReorderPointOrderUpToLevelInventory(type1,1,3,10);
		retailer2.addReorderPointOrderUpToLevelInventory(type2,2,4,10);
		retailer2.addReorderPointOrderUpToLevelInventory(type3,2,5,10);
		retailer2.addReorderPointOrderUpToLevelInventory(type4,2,3,10);
		
		retailer3.addReorderPointOrderUpToLevelInventory(type1,2,4,10);
		retailer3.addReorderPointOrderUpToLevelInventory(type2,1,2,10);
		retailer3.addReorderPointOrderUpToLevelInventory(type3,2,3,10);
		retailer3.addReorderPointOrderUpToLevelInventory(type4,2,3,10);
		
		retailer4.addReorderPointOrderUpToLevelInventory(type1,3,6,10);
		retailer4.addReorderPointOrderUpToLevelInventory(type2,3,4,10);
		retailer4.addReorderPointOrderUpToLevelInventory(type3,1,2,10);
		retailer4.addReorderPointOrderUpToLevelInventory(type4,2,3,10);
		
		retailer5.addReorderPointOrderUpToLevelInventory(type1,2,3,10);
		retailer5.addReorderPointOrderUpToLevelInventory(type2,0,1,10);
		retailer5.addReorderPointOrderUpToLevelInventory(type3,3,6,10);
		retailer5.addReorderPointOrderUpToLevelInventory(type4,1,2,10);
		
		// create the structure of the network
		// attach the warehouse to the external supplier with a specified tranpostation time of 3.0
		DistributionIfc t1 = new Constant(3.0);
		n.attachIHPToExternalSupplier(warehouse,t1);
		
		// attach the retailers to the warehouse, reuse the constant transportation time of 1.0
		DistributionIfc t2 = new Constant(1.0);
		n.attachIHPToSupplier(warehouse, retailer1, t2);
		n.attachIHPToSupplier(warehouse, retailer2, t2);
		n.attachIHPToSupplier(warehouse, retailer3, t2);
		n.attachIHPToSupplier(warehouse, retailer4, t2);
		n.attachIHPToSupplier(warehouse, retailer5, t2);
		
		// create the demand generators for the retailers						
		DistributionIfc dA11 = new Exponential(2.0);
		DistributionIfc dA12 = new Exponential(1.0);	
		DistributionIfc dA13 = new Exponential(1.5);	
		DistributionIfc dA14 = new Exponential(3.0);	
		DistributionIfc dA21 = new Exponential(1.0);
		DistributionIfc dA22 = new Exponential(2.0);	
		DistributionIfc dA23 = new Exponential(2.5);	
		DistributionIfc dA24 = new Exponential(1.5);	
		DistributionIfc dA31 = new Exponential(2.5);
		DistributionIfc dA32 = new Exponential(1.5);	
		DistributionIfc dA33 = new Exponential(2.0);	
		DistributionIfc dA34 = new Exponential(2.0);	
		DistributionIfc dA41 = new Exponential(3.0);
		DistributionIfc dA42 = new Exponential(2.5);	
		DistributionIfc dA43 = new Exponential(1.0);	
		DistributionIfc dA44 = new Exponential(2.5);	
		DistributionIfc dA51 = new Exponential(1.5);
		DistributionIfc dA52 = new Exponential(0.5);	
		DistributionIfc dA53 = new Exponential(3.0);	
		DistributionIfc dA54 = new Exponential(0.5);
		
		n.attachDemandGeneratorToIHP(retailer1, type1, dA11);
		n.attachDemandGeneratorToIHP(retailer1, type2, dA12);
		n.attachDemandGeneratorToIHP(retailer1, type3, dA13);
		n.attachDemandGeneratorToIHP(retailer1, type4, dA14);
		
		n.attachDemandGeneratorToIHP(retailer2, type1, dA21);
		n.attachDemandGeneratorToIHP(retailer2, type2, dA22);
		n.attachDemandGeneratorToIHP(retailer2, type3, dA23);
		n.attachDemandGeneratorToIHP(retailer2, type4, dA24);
		
		n.attachDemandGeneratorToIHP(retailer3, type1, dA31);
		n.attachDemandGeneratorToIHP(retailer3, type2, dA32);
		n.attachDemandGeneratorToIHP(retailer3, type3, dA33);
		n.attachDemandGeneratorToIHP(retailer3, type4, dA34);
		
		n.attachDemandGeneratorToIHP(retailer4, type1, dA41);
		n.attachDemandGeneratorToIHP(retailer4, type2, dA42);
		n.attachDemandGeneratorToIHP(retailer4, type3, dA43);
		n.attachDemandGeneratorToIHP(retailer4, type4, dA44);
		
		n.attachDemandGeneratorToIHP(retailer5, type1, dA51);
		n.attachDemandGeneratorToIHP(retailer5, type2, dA52);
		n.attachDemandGeneratorToIHP(retailer5, type3, dA53);
		n.attachDemandGeneratorToIHP(retailer5, type4, dA54);
		
		
		
		return(m);
	}
	
	public static void testExperiment(){
		
		Model m = buildModel();	
		
		m.turnOnLogReport();
		
		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
		// set the parameters of the experiment
		e.setNumberOfReplications(30);
		e.setLengthOfReplication(5400.0);
		e.setLengthOfWarmUp(1800.0);
		
		// turn on the desired reporting
//	    e.turnOnBatchReport(); 
		e.turnOnExperimentReport("ME-Inventory-NetworkSS-EXP");
		e.turnOnReplicationReport("ME-Inventory-NetworkSS-REP");
		e.turnOnExperimentConsoleOutput(true);
		
		// tell the experiment to run
		e.runAll();
	}
}
