/*
* Created on Mar 6, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer.testing;

import jsl.modeling.Experiment;
import jsl.modeling.Model;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.reporting.JSL;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeOrderFiller;
import jslx.supplychain.inventorylayer.OrderGenerator;
import jslx.supplychain.inventorylayer.RandomOrderCreator;
import jslx.supplychain.inventorylayer.Warehouse;
import jslx.supplychain.transportlayer.TimeBasedOrderCarrier;

/**
 * @author rossetti
 *
 */
public class TestSOPNetwork {

	/**
	 * @param args
	 */
	public static void main(String[] args) {

//		JSL.out.OUTPUT_ON = false;
		System.out.println("Running the model");
		runModel(buildModel());
		System.out.println("Done!");
		
	}
	
	public static Model buildModel(){
		Model m = Model.createModel("Test SOP Network");
		
		//Define item types
		ItemType type1=new ItemType(m,"Type-1");
		ItemType type2=new ItemType(m,"Type-2");
		ItemType type3=new ItemType(m,"Type-3");
		
		// Define an external supplier for the network
		// leadtimes include shipping time from external supplier to DC
		DistributionIfc ltd = Constant.ONE;
		LeadTimeOrderFiller ltof = new LeadTimeOrderFiller(m, "LT Order filler");
		ltof.addLeadTime(type1, ltd);
		ltof.addLeadTime(type2, ltd);
		ltof.addLeadTime(type3, ltd);
		
		// create the distribution center
		Warehouse dc = new Warehouse(m, "DC");
		
		// need to add inventory to the dc

		dc.addReorderPointReorderQuantityInventory(type1,4,1,20);
		dc.addReorderPointReorderQuantityInventory(type2,4,1,20);
		dc.addReorderPointReorderQuantityInventory(type3,4,1,20);

		// create the first "retailer"
		Warehouse r1 = new Warehouse(m, "R1");
		
		// add inventory to r1
		r1.addReorderPointReorderQuantityInventory(type1,2,1,10);
		r1.addReorderPointReorderQuantityInventory(type2,2,1,10);
		
		// create the 2nd "retailer"
		Warehouse r2 = new Warehouse(m, "R2");
		
		// add inventory to r2
		r2.addReorderPointReorderQuantityInventory(type1,2,1,10);
		r2.addReorderPointReorderQuantityInventory(type2,2,1,10);
		r2.addReorderPointReorderQuantityInventory(type3,2,1,10);

		// create the 3rd "retailer"
		Warehouse r3 = new Warehouse(m, "R3");
		
		// add inventory to r2
		r3.addReorderPointReorderQuantityInventory(type1,2,1,10);
		r3.addReorderPointReorderQuantityInventory(type2,2,1,10);
		r3.addReorderPointReorderQuantityInventory(type3,2,1,10);

		// create the 4th "retailer"
		Warehouse r4 = new Warehouse(m, "R4");
		
		// add inventory to r4
		r4.addReorderPointReorderQuantityInventory(type1,2,1,10);
		r4.addReorderPointReorderQuantityInventory(type2,2,1,10);

		
		// attach dc to ltof, r1 to dc, and r2 to dc, etc
		dc.setReplenishmentOrderFiller(ltof);// dc is filled by ltof
		r1.setReplenishmentOrderFiller(dc); // r1 is filled by dc
		r2.setReplenishmentOrderFiller(dc); // r2 is filled by dc
		r3.setReplenishmentOrderFiller(dc); // r3 is filled by dc
		r4.setReplenishmentOrderFiller(dc); // r4 is filled by dc

		// set up shipping from dc-r1, dc-r2
		TimeBasedOrderCarrier dcToRetailerShipper = new TimeBasedOrderCarrier(m, "ShipperFromDCtoRetailers");
		dcToRetailerShipper.setTransportTime(r1, Constant.ONE);
		dcToRetailerShipper.setTransportTime(r2, Constant.ONE);
		dcToRetailerShipper.setTransportTime(r3, Constant.ONE);
		dcToRetailerShipper.setTransportTime(r4, Constant.ONE);
		dc.setOrderShipper(dcToRetailerShipper);
		
		// create order generator for r1
		RandomOrderCreator roc1 = new RandomOrderCreator(m, "R1-OrderCreator");
		roc1.addItemTypeDistribution(type1);
		roc1.addItemTypeDistribution(type2);

		DistributionIfc tbo1 = new Exponential(1.0);
		OrderGenerator og1 = new OrderGenerator(m,"R1-OrderGenerator");
		og1.setTimeBetweenEventsForReplications(tbo1);
		og1.setOrderCreator(roc1);
		og1.setOrderFiller(r1);
		
		// create order generator for r2
		RandomOrderCreator roc2 = new RandomOrderCreator(m, "R2-OrderCreator");
		roc2.addItemTypeDistribution(type1);
		roc2.addItemTypeDistribution(type2);
		roc2.addItemTypeDistribution(type3);

		DistributionIfc tbo2 = new Exponential(1.0);
		OrderGenerator og2 = new OrderGenerator(m,"R2-OrderGenerator");
		og2.setTimeBetweenEventsForReplications(tbo2);
		og2.setOrderCreator(roc2);
		og2.setOrderFiller(r2);

		// create order generator for r3
		RandomOrderCreator roc3 = new RandomOrderCreator(m, "R3-OrderCreator");
		roc3.addItemTypeDistribution(type1);
		roc3.addItemTypeDistribution(type2);
		roc3.addItemTypeDistribution(type3);

		DistributionIfc tbo3 = new Exponential(1.0);
		OrderGenerator og3 = new OrderGenerator(m,"R3-OrderGenerator");
		og3.setTimeBetweenEventsForReplications(tbo3);
		og3.setOrderCreator(roc3);
		og3.setOrderFiller(r3);

		// create order generator for r4
		RandomOrderCreator roc4 = new RandomOrderCreator(m, "R4-OrderCreator");
		roc4.addItemTypeDistribution(type1);
		roc4.addItemTypeDistribution(type2);

		DistributionIfc tbo4 = new Exponential(1.0);
		OrderGenerator og4 = new OrderGenerator(m,"R4-OrderGenerator");
		og4.setTimeBetweenEventsForReplications(tbo4);
		og4.setOrderCreator(roc4);
		og4.setOrderFiller(r4);

		return(m);
	}
	
	public static void runModel(Model m){
		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
		// set the parameters of the experiment
		// set the parameters of the experiment
		e.setNumberOfReplications(10);
		e.setLengthOfReplication(3650.0);
		e.setLengthOfWarmUp(100.0);

		e.turnOnReplicationConsoleOutput();
		e.turnOnExperimentReport();
		e.turnOnReplicationReport();
		
		// tell the experiment to run
		e.runAll();
	}

}
