/*
* Created on Mar 6, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.inventorylayer.testing;

import jsl.modeling.Experiment;
import jsl.modeling.Model;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.reporting.JSL;
import jslx.supplychain.TimeBasedShippingMultiEchelonWarehouseNetwork;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeOrderFiller;
import jslx.supplychain.inventorylayer.OrderGenerator;
import jslx.supplychain.inventorylayer.RandomOrderCreator;
import jslx.supplychain.inventorylayer.Warehouse;
import jslx.supplychain.transportlayer.TimeBasedOrderCarrier;

/**
 * @author rossetti
 *
 */
public class TestTimeBasedShippingMultiEchelonWarehouseNetwork {

	/**
	 * @param args
	 */
	public static void main(String[] args) {

//		JSL.out.OUTPUT_ON = false;
		System.out.println("Running the model");
		runModel(buildModel());
		System.out.println("Done!");
		
	}
	
	public static Model buildModel(){
		Model m = Model.createModel("Test MEWN Network");
		
		TimeBasedShippingMultiEchelonWarehouseNetwork n = new TimeBasedShippingMultiEchelonWarehouseNetwork(m, "MEWN-");
		
		//Define item types
		RandomIfc ltd = Constant.ONE;
		ItemType type1 = n.addItemType("Type-1", ltd);
		ItemType type2 = n.addItemType("Type-2", ltd);
		ItemType type3 = n.addItemType("Type-3", ltd);
		
		// create the distribution center
		Warehouse dc = n.addWarehouse("DC");
		
		// need to add inventory to the dc

		dc.addReorderPointReorderQuantityInventory(type1,4,1,20);
		dc.addReorderPointReorderQuantityInventory(type2,4,1,20);
		dc.addReorderPointReorderQuantityInventory(type3,4,1,20);

		// create the first "retailer"
		Warehouse r1 = n.addWarehouse("R1");
		
		// add inventory to r1
		r1.addReorderPointReorderQuantityInventory(type1,2,1,10);
		r1.addReorderPointReorderQuantityInventory(type2,2,1,10);
		
		// create the 2nd "retailer"
		Warehouse r2 = n.addWarehouse("R2");
		
		// add inventory to r2
		r2.addReorderPointReorderQuantityInventory(type1,2,1,10);
		r2.addReorderPointReorderQuantityInventory(type2,2,1,10);
		r2.addReorderPointReorderQuantityInventory(type3,2,1,10);

		// create the 3rd "retailer"
		Warehouse r3 = n.addWarehouse("R3");
		
		// add inventory to r2
		r3.addReorderPointReorderQuantityInventory(type1,2,1,10);
		r3.addReorderPointReorderQuantityInventory(type2,2,1,10);
		r3.addReorderPointReorderQuantityInventory(type3,2,1,10);

		// create the 4th "retailer"
		Warehouse r4 = n.addWarehouse("R4");
		
		// add inventory to r4
		r4.addReorderPointReorderQuantityInventory(type1,2,1,10);
		r4.addReorderPointReorderQuantityInventory(type2,2,1,10);

		// create the structure of the network
		// attach the dc to the external supplier
		n.attachWarehouseToExternalSupplier(dc);
		
		// attach the retailers to the dc, reuse the constant leadtime of 1.0
		n.attachWarehouseToSupplier(dc, r1,  Constant.ONE);
		n.attachWarehouseToSupplier(dc, r2,  Constant.ONE);
		n.attachWarehouseToSupplier(dc, r3,  Constant.ONE);
		n.attachWarehouseToSupplier(dc, r4,  Constant.ONE);
		
		// create order generator for r1
		RandomOrderCreator roc1 = new RandomOrderCreator(m, "R1-OrderCreator");
		roc1.addItemTypeDistribution(type1);
		roc1.addItemTypeDistribution(type2);

		DistributionIfc tbo1 = new Exponential(1.0);
		OrderGenerator og1 = new OrderGenerator(m,"R1-OrderGenerator");
		og1.setTimeBetweenEventsForReplications(tbo1);
		og1.setOrderCreator(roc1);
		
		// create order generator for r2
		RandomOrderCreator roc2 = new RandomOrderCreator(m, "R2-OrderCreator");
		roc2.addItemTypeDistribution(type1);
		roc2.addItemTypeDistribution(type2);
		roc2.addItemTypeDistribution(type3);

		DistributionIfc tbo2 = new Exponential(1.0);
		OrderGenerator og2 = new OrderGenerator(m,"R2-OrderGenerator");
		og2.setTimeBetweenEventsForReplications(tbo2);
		og2.setOrderCreator(roc2);

		// create order generator for r3
		RandomOrderCreator roc3 = new RandomOrderCreator(m, "R3-OrderCreator");
		roc3.addItemTypeDistribution(type1);
		roc3.addItemTypeDistribution(type2);
		roc3.addItemTypeDistribution(type3);

		DistributionIfc tbo3 = new Exponential(1.0);
		OrderGenerator og3 = new OrderGenerator(m,"R3-OrderGenerator");
		og3.setTimeBetweenEventsForReplications(tbo3);
		og3.setOrderCreator(roc3);

		// create order generator for r4
		RandomOrderCreator roc4 = new RandomOrderCreator(m, "R4-OrderCreator");
		roc4.addItemTypeDistribution(type1);
		roc4.addItemTypeDistribution(type2);

		DistributionIfc tbo4 = new Exponential(1.0);
		OrderGenerator og4 = new OrderGenerator(m,"R4-OrderGenerator");
		og4.setTimeBetweenEventsForReplications(tbo4);
		og4.setOrderCreator(roc4);

		// attach order generators to the retailers
		n.attachOrderGeneratorToWarehouse(r1, og1);
		n.attachOrderGeneratorToWarehouse(r2, og2);
		n.attachOrderGeneratorToWarehouse(r3, og3);
		n.attachOrderGeneratorToWarehouse(r4, og4);

		return(m);
	}
	
	public static void runModel(Model m){
		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
		// set the parameters of the experiment
		// set the parameters of the experiment
		e.setNumberOfReplications(10);
		e.setLengthOfReplication(3650.0);
		e.setLengthOfWarmUp(100.0);

		e.turnOnReplicationConsoleOutput();
		e.turnOnExperimentReport();
		e.turnOnReplicationReport();
		
		// tell the experiment to run
		e.runAll();
	}

}
