/*
* Created on Aug 27, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.multiechelon;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.supplychain.inventorylayer.AggregateInventoryIfc;
import jslx.supplychain.inventorylayer.DemandFillerIfc;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.DemandSenderIfc;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;

/** A MultiEchelonIHPNetwork models an abhorescent "tree" network of inventory holding points (IHPs).
 *  In this type of network, each IHP can have one parent IHP that fills its replenishment requests.  It
 *  can have many child IHPs to which it fills replenshments.  No checking is done to ensure
 *  that the IHPs are compatible in terms of their ability to supply certain ItemTypes.  Any demands
 *  sent to an IHP that cannot be handled by the IHP will be rejected.  The default rejection
 *  behavior for an IHP is to throw an exception.  To override this behavior the client must supply
 *  a DemandListenerRejectedIfc for the IHP or DemandGenerator.
 *  
 *  The top level of the network is modeled with a LeadTimeDemandFiller and
 *  is considered an external supplier to the network.  The external supplier (LeadTimeDemandFiller)
 *  has an infinite supply of the ItemTypes added to the network, which can be produced and delivered
 *  after a given leadtime, which may depend on the ItemType.  Since the top level supplier is "external"
 *  to the network the leadtime represents the time to produce the item.  If no item types are added 
 *  then all demands sent to the external supplier will be rejected.
 *  
 *  The generation of external demand to the IHPs is modeled through the use of instances of 
 *  the DemandGenerator class.  Each IHP may have zero or more DemandGenerators attached to it, and
 *  each DemandGenerator may be attached to one or more IHPs.  If no DemandGenerators are attached to any IHP (or to the
 *  network) then nothing will happen, since IHPs must have demands sent to them to begin processing.
 *  
 *  In addition, external DemandGenerators may be attached to the network by calling the 
 *  allowExternalDemandGenerators() method which allows the IHPs to ship/deliver to the external
 *  demand generators with zero delay.  The external demand generators must know where to
 *  send their demands (typically through a DemandFillerFinderIfc that they have access to).
 *  
 *  NOTE:  It is up to clients to provide transport between locations.  This class does not explicitly model
 *  the transport between locations.  The filled demands are automatically shipped and delivered after being filled.
 *  To model transport provide a DemandCarrierIfc to the IHP's
 *
 */
public class MultiEchelonInventoryNetwork extends SchedulingElement implements AggregateInventoryIfc {

	/** Models the lead time to produce the item type at the external supplier
	 */
	protected LeadTimeDemandFiller myExternalSupplier;

	/** Collects statistics on the aggregate amount backordered
	 */
	protected AggregateTimeWeightedVariable myTotalAmtBackOrdered;
	
	/** Collects statistics on the aggregate number waiting in backorder
	 */
	protected AggregateTimeWeightedVariable myTotalNumBackOrdered;

	/** Collects statistics on the aggregate inventory
	 */
	protected AggregateTimeWeightedVariable myTotalOnHand;

	/** Collects statistics on the aggregate on order
	 */
	protected AggregateTimeWeightedVariable myTotalOnOrder;

	/** Collects statistics on the aggregate stockout
	 */
	protected AggregateTimeWeightedVariable myTotalStockOutIndicator;

	/** Collects statistics on the aggregate first fill rate
	 */
	protected ResponseVariableAverageObserver myAvgFirstFillRate;

	/** Collects statistics on the aggregate customer waiting time
	 */
	protected ResponseVariableAverageObserver myAvgCustomerWaitTime;
	
	/** Counts all the replenishments made by the network
	 */
	protected AggregateCounter myTotalNumOfReplenishments;
	
	/** A Map that holds the inventory holding points and their children as a list
	 */
	protected Map<InventoryHoldingPoint, List<InventoryHoldingPoint>> myIHPs;
	
	/** Holds all the IHPs that act as customers to other IHPs in the network
	 */
	protected Set<InventoryHoldingPoint> myCustomerIHPs;
	
	/** Holds all the IHPs that are attached to the external supplier
	 */
	protected Set<InventoryHoldingPoint> myIHPsAttachedToExternalSupplier;
	
	/** Holds the demand senders for an IHP if they are attached
	 */
	protected Map<InventoryHoldingPoint, List<DemandSenderIfc>> myDemandSenders;
	
	/**
	 * @param parent
	 */
	public MultiEchelonInventoryNetwork(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public MultiEchelonInventoryNetwork(ModelElement parent, String name) {
		super(parent, name);
		myIHPs = new HashMap<InventoryHoldingPoint, List<InventoryHoldingPoint>>();
		myCustomerIHPs = new HashSet<InventoryHoldingPoint>();
		myIHPsAttachedToExternalSupplier = new HashSet<InventoryHoldingPoint>();
		myDemandSenders = new HashMap<InventoryHoldingPoint, List<DemandSenderIfc>>();
		
		myTotalOnHand = new AggregateTimeWeightedVariable(this, getName() + " Total On Hand");
		myAvgFirstFillRate = new ResponseVariableAverageObserver(this, getName() + " Aggregate Fill Rate");
		myTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, getName() + " Aggregate Number of Stockouts");
		myTotalOnOrder = new AggregateTimeWeightedVariable(this, getName() + " Total On Order");
		myTotalNumOfReplenishments = new AggregateCounter(this, getName() + " NumberOfReplenishments");
		myTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Amt BackOrdered");
		myTotalNumBackOrdered = new AggregateTimeWeightedVariable(this, getName() + "Aggregate Number BackOrdered");
		myAvgCustomerWaitTime = new ResponseVariableAverageObserver(this, getName() + "Avg Customer Wait");
		
		//create external LeadTimeDemandFiller and its own TimeBasedDemandShipper
		myExternalSupplier = new LeadTimeDemandFiller(this, getName() + "ExternalSupplier");

	}

	/** Adds an ItemType with the given leadtime distribution
	 *  to the network
	 * 
	 * @param itemType
	 * @param leadtime
	 */
	public final void addItemType(ItemType itemType, RandomIfc leadtime){
		myExternalSupplier.addLeadTime(itemType, leadtime);
	}
	
	/** Adds an ItemType with the given leadtime distribution
	 *  to the network.  Uses the lead time associated with the ItemType
	 * 
	 * @param itemType
	 */
	public final void addItemType(ItemType itemType){
		myExternalSupplier.addLeadTime(itemType, itemType.getLeadTime());
	}
	
	/** Adds an ItemType with a default name and given leadtime distribution
	 *  to the network
	 * 
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(RandomIfc leadtime){
		ItemType itemType = new ItemType(this);
		addItemType(itemType, leadtime);
		return itemType;
	}
	
	/** Adds an ItemType and its lead time distribution to the network. The name
	 *  of the item type must be unique to the model
	 * 
	 * @param name
	 * @param leadtime
	 * @return
	 */
	public final ItemType addItemType(String name, RandomIfc leadtime){
		ItemType itemType = new ItemType(this, name);
		addItemType(itemType, leadtime);
		return(itemType);
	}
	
	/** 
	 * 	NOTE: Also removes any associated model elements from the model, via ModelElement.removeFromModel();
	 * @param itemType
	 * @see jslx.supplychain.inventorylayer.LeadTimeDemandFiller#removeLeadTime(jslx.supplychain.inventorylayer.ItemType)
	 */
	public final void removeLeadTime(ItemType itemType) {
		myExternalSupplier.removeLeadTime(itemType);
	}

	/** Returns an unmodifiable set of the item types added to the network
	 * 
	 * @return
	 */
	public final Set<ItemType> getItemTypes(){
		return myExternalSupplier.getItemTypes();
	}
	
	/** Removes the supplied itemType from the network and from any
	 *  ihp within the network.  
	 *  NOTE: Also removes any associated model elements from the model, via ModelElement.removeFromModel();  
	 *  WARNING: If events are scheduled for the item type then errors will occur
	 *  It is best to use this method when the replication is not running.
	 * 
	 * @param itemType
	 */
	public final void removeItemTypeFromNetwork(ItemType itemType){
		myExternalSupplier.removeLeadTime(itemType);
		Set<InventoryHoldingPoint> ihps = myIHPs.keySet();
		for(InventoryHoldingPoint ihp: ihps){
			if (ihp.canFillItemType(itemType))
				ihp.removeInventory(itemType, true);
		}
	}
	
	/** Checks of the ihp is part of the network
	 * 
	 * @param ihp
	 * @return
	 */
	public boolean containsInventoryHoldingPoint(InventoryHoldingPoint ihp) {
		return myIHPs.containsKey(ihp);
	}

	/** Causes an InventoryHoldingPoint to be created with the supplied name
	 *  and added to the network.  This only adds the IHP to the network it does
	 *  not cause the IHP to be connected to any other IHPs in the network. The client
	 *  can use the returned reference to the IHP to connect to other IHPs.
	 * 
	 * @return the created IHP
	 */
	public final InventoryHoldingPoint addInventoryHoldingPoint(){
		// create the ihp and add it to the network so that it may have child ihps
		InventoryHoldingPoint ihp = new InventoryHoldingPoint(this);		
		addInventoryHoldingPoint(ihp);
		return(ihp);
	}

	/** Causes an InventoryHoldingPoint to be created with the supplied name
	 *  and added to the network.  This only adds the IHP to the network it does
	 *  not cause the IHP to be connected to any other IHPs in the network. The client
	 *  can use the returned reference to the IHP to connect to other IHPs.
	 * 
	 * @param name must be unique in the Model
	 * @return the created IHP
	 */
	public final InventoryHoldingPoint addInventoryHoldingPoint(String name){
		// create the ihp and add it to the network so that it may have child ihps
		InventoryHoldingPoint ihp = new InventoryHoldingPoint(this, name);	
		addInventoryHoldingPoint(ihp);
		return(ihp);
	}
	
	/** Adds an IHP to the network. This only adds the IHP to the network it does
	 *  not cause the IHP to be connected to any other IHPs in the network. The client
	 *  can use the returned reference to the IHP to connect to other IHPs.
	 * 
	 * 
	 * @param ihp Must not be null, must not already be in the network
	 */
	public final void addInventoryHoldingPoint(InventoryHoldingPoint ihp){
		if (ihp == null)
			throw new IllegalArgumentException("The supplied IHP was null");
		if (myIHPs.containsKey(ihp))
			throw new IllegalArgumentException("The supplied IHP is already in the network.");
		
		// add it to the network so that it may have child ihps
	
		myIHPs.put(ihp, new ArrayList<InventoryHoldingPoint>());
		
		// attach aggregate response variable observers to each ihp, to get network performance
		ihp.subscribeToAggregateOnHandInventory(myTotalOnHand);
		ihp.subscribeToAggregateAmountOnOrder(myTotalOnOrder);
		ihp.subscribeToAggregateStockOutIndicator(myTotalStockOutIndicator);
		ihp.subscribeToAggregateAvgFirstFillRate(myAvgFirstFillRate);
		ihp.subscribeToAggregateReplenishmentDemandCounter(myTotalNumOfReplenishments);		
		ihp.subscribeToAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		ihp.subscribeToAggregateNumberBackOrdered(myTotalNumBackOrdered);
		ihp.subscribeToAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);
		 
	}

	/** This removes the supplied IHP from the network if and only if
	 *  the IHP has no children and it is not attached as a customer to any supplier
	 * 
	 * @param ihp
	 */
	public final void removeInventoryHoldingPoint(InventoryHoldingPoint ihp){
		if (ihp == null)
			throw new IllegalArgumentException("The supplied IHP was null");
		
		if (!myIHPs.containsKey(ihp))
			throw new IllegalArgumentException("The supplied IHP is not in the network.");
		
		// check if it has children
		List<InventoryHoldingPoint> children = myIHPs.get(ihp);
		if (!children.isEmpty())
			throw new IllegalStateException("Cannot remove an IHP that has customers from the network");
		
		if (isAttachedAsCustomer(ihp))
			throw new IllegalStateException("Cannot remove an IHP that is attached as a customer");

		if (isAttachedToExternalSupplier(ihp))
			throw new IllegalArgumentException("Cannont remove an IHP that is attached to the external supplier.");

		// in the network, no children, not a customer
		myIHPs.remove(ihp);

		// detach aggregate response variable observers to each ihp, to get network performance
		ihp.unsubscribeFromAggregateOnHandInventory(myTotalOnHand);
		ihp.unsubscribeFromAggregateAmountOnOrder(myTotalOnOrder);
		ihp.unsubscribeFromAggregateStockOutIndicator(myTotalStockOutIndicator);
		ihp.unsubscribeFromAggregateAvgFirstFillRate(myAvgFirstFillRate);
		ihp.unsubscribeFromAggregateReplenishmentDemandCounter(myTotalNumOfReplenishments);		
		ihp.unsubscribeFromAggregateAmountBackOrdered(myTotalAmtBackOrdered);
		ihp.unsubscribeFromAggregateNumberBackOrdered(myTotalNumBackOrdered);
		ihp.unsubscribeFromAggregateAvgCustomerWaitTime(myAvgCustomerWaitTime);
		
	}
	
	/** Checks if the supplied IHP is attached to the external supplier of the network
	 * 
	 * @param ihp
	 * @return true if attached
	 */
	public final boolean isAttachedToExternalSupplier(InventoryHoldingPoint ihp){
		return(myIHPsAttachedToExternalSupplier.contains(ihp));
	}
	
	/** Checks if the supplied IHP is attached as a customer in the network.
	 *  An IHP may only be attached to one IHP as a customer
	 * 
	 * @param ihp
	 * @return true if it a customer somewhere in the network
	 */
	public final boolean isAttachedAsCustomer(InventoryHoldingPoint ihp){
		return(myCustomerIHPs.contains(ihp));
	}
	
	/** Checks if the supplied IHP is attached as a customer for the supplied supplier.
	 *  An IHP may only be attached to one IHP as a customer
	 * 
	 * @param supplier
	 * @param customer
	 * @return true if it is a customer of the provided supplier
	 */
	public final boolean isCustomer(InventoryHoldingPoint supplier, InventoryHoldingPoint customer){
		if (!myIHPs.containsKey(supplier))
			return(false);
		
		if (!myIHPs.containsKey(customer))
			return(false);
				
		List<InventoryHoldingPoint> customers = myIHPs.get(supplier);
		return(customers.contains(customer) && (customer.getDemandFiller() == supplier));		
	}

	/** Returns an unmodifiable set of the IHPs that are customers of the external supplier
	 * 
	 * @param supplier
	 * @return
	 */
	public final Set<InventoryHoldingPoint> getExternalSupplierCustomers(){		
		return(Collections.unmodifiableSet(myIHPsAttachedToExternalSupplier));
	}
	
	/** Returns an unmodifiable list of the IHPs that are customers for the given supplier
	 * 
	 * @param supplier
	 * @return
	 */
	public final List<InventoryHoldingPoint> getCustomers(InventoryHoldingPoint supplier){
		if (!myIHPs.containsKey(supplier))
			throw new IllegalArgumentException("The supplied IHP is not in the network.");
		
		return(Collections.unmodifiableList(myIHPs.get(supplier)));
	}
	
	/** Returns the number of customers for the indicated IHP
	 * 
	 * @param supplier
	 * @return
	 */
	public final int getNumberOfCustomers(InventoryHoldingPoint supplier){
		return myIHPs.get(supplier).size();
	}
	
	/** Gets the DemandFillerIfc that is serving as the external supplier to the network
	 *  This reference can be used to establish transport between the external supplier
	 *  and the IHP's that are attached to it through a DemandCarrierIfc
	 *  
	 * @return the externalSupplier
	 */
	public DemandFillerIfc getExternalSupplier() {
		return myExternalSupplier;
	}

	/** Attaches the IHP to the external supplier 
	 * 
	 * @param ihp Must not already be attached to the external supplier or to another IHP
	 * @param transportTime
	 */
	public final void attachIHPToExternalSupplier(InventoryHoldingPoint ihp){
		if (ihp == null)
			throw new IllegalArgumentException("The supplied IHP was null");
		
		if (!myIHPs.containsKey(ihp))
			throw new IllegalArgumentException("The supplied IHP is not in the network.");
		
		if (myIHPsAttachedToExternalSupplier.contains(ihp))
			throw new IllegalArgumentException("The supplied IHP is already attached to the external supplier.");
		
		if (myCustomerIHPs.contains(ihp))
			throw new IllegalArgumentException("The supplied IHP is already a customer to another IHP");
		
		myIHPsAttachedToExternalSupplier.add(ihp);

		// designate its demand filler as the external supplier
		ihp.setDemandFiller(myExternalSupplier);
	}
	
	/** Detaches the supplied IHP from the external supplier.  The detached IHP is not
	 *  removed from the network. It is only detached from the external supplier. It no longer has a demand filler.
	 *  The supplied IHP must not be null, must be in the network, must not have any customers,
	 *  and must already be attached to the external supplier
	 *  
	 * @param ihp
	 */
	public final void detachIHPFromExternalSupplier(InventoryHoldingPoint ihp){
		if (ihp == null)
			throw new IllegalArgumentException("The supplied IHP was null");
		
		if (!myIHPs.containsKey(ihp))
			throw new IllegalArgumentException("The supplied IHP is not in the network.");
		
		// check if it has children
		List<InventoryHoldingPoint> children = myIHPs.get(ihp);
		if (!children.isEmpty())
			throw new IllegalStateException("Cannot detach an IHP that has customers");
		
		if (!isAttachedToExternalSupplier(ihp))
			throw new IllegalArgumentException("The supplied IHP is not attached to the external supplier.");
		
		myIHPsAttachedToExternalSupplier.remove(ihp);
		ihp.setDemandFiller(null);
		
	}

	/** Attaches the IHP (customer) to the given IHP (supplier)
	 * 
	 * @param supplier Must be in the network
	 * @param customer Must be in the network and must not be a customer already on another IHP
	 */
	public final void attachIHPToSupplier(InventoryHoldingPoint supplier, InventoryHoldingPoint customer){
		if (!myIHPs.containsKey(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		if (!myIHPs.containsKey(customer))
			throw new IllegalArgumentException("The provided customer has not be added to the network");

		if (myCustomerIHPs.contains(customer))
			throw new IllegalArgumentException("The supplied customer IHP is already a customer IHP");
		
		//both in the network and customer is not already a customer, can add it to the supplier's list
		List<InventoryHoldingPoint> customers = myIHPs.get(supplier);
		customers.add(customer);
		// designate it as a customer
		myCustomerIHPs.add(customer);
		// ensure that it uses the supplier for its replensishments
		customer.setDemandFiller(supplier);
	
	}
	
	/** The detached IHP is not removed from the network. It is only detached from the supplier. 
	 *  It no longer has a demand filler.
	 *  The supplied supplier IHP must not be null, must be in the network
	 *  The supplied customer IHP must not be null, must be in the network, must not have any customers,
	 *  and must already be attached to the supplier
	 * 
	 * @param supplier
	 * @param customer
	 */
	public final void detachIHPFromSupplier(InventoryHoldingPoint supplier, InventoryHoldingPoint customer){
		if (supplier == null)
			throw new IllegalArgumentException("The provided supplier was null");

		if (customer == null)
			throw new IllegalArgumentException("The provided customer was null");

		if (!myIHPs.containsKey(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		if (!myIHPs.containsKey(customer))
			throw new IllegalArgumentException("The provided customer has not be added to the network");
		
		if (!isCustomer(supplier, customer))
			throw new IllegalArgumentException("The provided customer is not a customer of the provided supplier");

		// check if customer has children
		List<InventoryHoldingPoint> children = myIHPs.get(customer);
		if (!children.isEmpty())
			throw new IllegalStateException("Cannot detach an IHP that has customers");

		//both in the network and customer is attached, and customer has no children
		List<InventoryHoldingPoint> customers = myIHPs.get(supplier);
		customers.remove(customer);
		myCustomerIHPs.remove(customer);
		customer.setDemandFiller(null);
		
	}

	/** Creates and attaches a demand generator to the IHP
	 * 
	 * @param supplier
	 * @param type
	 * @param timeUntilNext
	 * @return
	 */
	public final DemandGenerator attachDemandGeneratorToIHP(InventoryHoldingPoint supplier,
			ItemType type, DistributionIfc timeUntilNext){
		return(attachDemandGeneratorToIHP(supplier, type, timeUntilNext, null));
	}
	
	/** Creates and attaches a demand generator to the IHP
	 * 
	 * @param supplier
	 * @param type
	 * @param timeUntilNext
	 * @param name
	 * @return
	 */
	public final DemandGenerator attachDemandGeneratorToIHP(InventoryHoldingPoint supplier,
			ItemType type, DistributionIfc timeUntilNext, String name){
		DemandGenerator dg = new DemandGenerator(this, type, timeUntilNext, name);
		attachDemandSenderToIHP(supplier, dg);
		return(dg);
	}

	/** Attaches the supplied demand sender as a customer on the given IHP (supplier)
	 *  
	 * @param supplier Must be in the network
	 * @param sender Must not have been already attached to the given supplier
	 */
	public final void attachDemandSenderToIHP(InventoryHoldingPoint supplier, DemandSenderIfc sender){
		if (!myIHPs.containsKey(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");
		
		List<DemandSenderIfc> senders = myDemandSenders.get(supplier);
		if (senders == null){
			senders = new ArrayList<DemandSenderIfc>();
			myDemandSenders.put(supplier, senders);
		}
		if (senders.contains(sender))
			throw new IllegalArgumentException("The provided demand sender is already attached to the supplier");
		
		senders.add(sender);
		sender.setDemandFiller(supplier);	
	}
	
	/** Detaches the supplied demand sender from the supplier.
	 *  The provided supplier and sender must not be null.  They both must already be in the network.
	 *  The sender must already be attached to the supplier. After detaching, the sender
	 *  will not have a demand filler.
	 * 
	 * @param supplier
	 * @param sender
	 */
	public final void detachDemandSenderFromIHP(InventoryHoldingPoint supplier, DemandSenderIfc sender){
		if (supplier == null)
			throw new IllegalArgumentException("The provided supplier was null");

		if (sender == null)
			throw new IllegalArgumentException("The provided generator was null");
	
		if (!myIHPs.containsKey(supplier))
			throw new IllegalArgumentException("The provided supplier has not be added to the network");

		List<DemandSenderIfc> senders = myDemandSenders.get(supplier);
		if (senders == null)
			throw new IllegalArgumentException("The provided supplier does not have any demand senders");
		
		if (!senders.contains(sender))
			throw new IllegalArgumentException("The provided demand sender is not attached to the provided supplier.");
		senders.remove(sender);
		sender.setDemandFiller(null);
	}

	
	/** Returns an unmodifiable list of the demand senders that are customers for the given supplier
	 * 
	 * @param supplier
	 * @return
	 */
	public final List<DemandSenderIfc> getDemandSenders(InventoryHoldingPoint supplier){
		if (supplier == null)
			throw new IllegalArgumentException("The provided supplier was null");
		
		if (!myIHPs.containsKey(supplier))
			throw new IllegalArgumentException("The supplied IHP is not in the network.");
		
		return(Collections.unmodifiableList(myDemandSenders.get(supplier)));
	}

	/** Returns a List copy of all the IHPs currently attached to the network
	 *  This method can be used to assist in building an instance of a class that implements 
	 *  the DemandFillerFinderIfc interface and needs a list of all possible
	 *  IHPs in the network. For example, this could be useful in developing an
	 *  external demand generator that can probabilistically send to any IHP.
	 *  
	 * @return
	 */
	public final List<InventoryHoldingPoint> getInventoryHoldingPoints(){
		return(new ArrayList<InventoryHoldingPoint>(myIHPs.keySet()));
	}

	/**
	 * @return Returns the Aggregate Avg Customer Wait Time ResponseVariable.
	 */
	public final ResponseVariableAverageObserver getAggregateAvgCustomerWaitTimeResponseVariable() {
		return myAvgCustomerWaitTime;
	}
	
	/**
	 * @return Returns the get Aggregate Avg First FillRate ResponseVariable.
	 */
	public final ResponseVariableAverageObserver getAggregateAvgFirstFillRateResponseVariable() {
		return myAvgFirstFillRate;
	}
	
	/**
	 * @return Returns the Aggregate Amount BackOrdered ResponseVariable.
	 */
	public final AggregateTimeWeightedVariable getAggregateAmountBackOrderedResponseVariable() {
		return myTotalAmtBackOrdered;
	}
	
	/**
	 * @return Returns the Aggregate Number BackOrdered ResponseVariable.
	 */
	public final AggregateTimeWeightedVariable getAggregateNumberBackOrderedResponseVariable() {
		return myTotalNumBackOrdered;
	}
	
	/**
	 * @return Returns the Aggregate OnHand Inventory ResponseVariable.
	 */
	public final AggregateTimeWeightedVariable getAggregateOnHandInventoryResponseVariable() {
		return myTotalOnHand;
	}
	
	/**
	 * @return Returns the Aggregate OnOrder ResponseVariable.
	 */
	public final AggregateTimeWeightedVariable getAggregateAmountOnOrderResponseVariable() {
		return myTotalOnOrder;
	}
	
	/** 
	 * 
	 * @return Returns a counter for the total number of replenishment orders made at this IHP
	 */
	public final AggregateCounter getAggregateNumberOfReplenishmentDemandsCounter(){
		return(myTotalNumOfReplenishments);
	}
	
	/**
	 * @return Returns the Aggregate StockOut Percentage ResponseVariable.
	 */
	public final AggregateTimeWeightedVariable getAggregateStockOutPercentageResponseVariable() {
		return myTotalStockOutIndicator;
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#attachAggregateAvgCustomerWaitTime(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void subscribeToAggregateAvgCustomerWaitTime(ResponseVariableAverageObserver aggregate) {
		aggregate.subscribeTo(myAvgCustomerWaitTime);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#attachAggregateAvgFirstFillRate(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void subscribeToAggregateAvgFirstFillRate(ResponseVariableAverageObserver aggregate) {
		aggregate.subscribeTo(myAvgFirstFillRate);
	}
	
	/**
	 * @param aggregate The aggregate time weighted variable to attach
	 */
	public void subscribeToAggregateAmountBackOrdered(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myTotalAmtBackOrdered);
	}

	/**
	 * @param aggregate The aggregate time weighted variable to attach
	 */
	public void subscribeToAggregateNumberBackOrdered(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myTotalNumBackOrdered);
	}

	/**
	 * @param aggregate The aggregate time weighted variable to attach
	 */
	public void subscribeToAggregateOnHandInventory(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myTotalOnHand);
	}

	/**
	 * @param aggregate The aggregate time weighted variable to attach
	 */
	public void subscribeToAggregateAmountOnOrder(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myTotalOnOrder);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#attachAggregateReplenishmentDemandCounter(jsl.modeling.elements.variable.AggregateCounter)
	 */
	public void subscribeToAggregateReplenishmentDemandCounter(AggregateCounter counter) {
		counter.subscribeTo( myTotalNumOfReplenishments);
	}

	/**
	 * @param aggregate The aggregate time weighted variable to attach
	 */
	public void subscribeToAggregateStockOutIndicator(AggregateTimeWeightedVariable aggregate){
		aggregate.subscribeTo(myTotalStockOutIndicator);
	}
	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateAvgCustomerWaitTime(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void unsubscribeFromAggregateAvgCustomerWaitTime(ResponseVariableAverageObserver aggregate) {
		aggregate.unsubscribeFrom(myAvgCustomerWaitTime);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateAvgFirstFillRate(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void unsubscribeFromAggregateAvgFirstFillRate(ResponseVariableAverageObserver aggregate) {
		aggregate.unsubscribeFrom(myAvgFirstFillRate);
	}
	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateAmountBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateAmountBackOrdered(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myTotalAmtBackOrdered);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateNumberBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateNumberBackOrdered(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myTotalNumBackOrdered);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateOnHandInventory(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateOnHandInventory(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myTotalOnHand);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateAmountOnOrder(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateAmountOnOrder(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myTotalOnOrder);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateReplenishmentDemandCounter(jsl.modeling.elements.variable.AggregateCounter)
	 */
	public void unsubscribeFromAggregateReplenishmentDemandCounter(AggregateCounter counter) {
		counter.unsubscribeFrom(myTotalNumOfReplenishments);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.inventorylayer.AggregateInventoryIfc#detachAggregateStockOutIndicator(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateStockOutIndicator(AggregateTimeWeightedVariable aggregate){
		aggregate.unsubscribeFrom(myTotalStockOutIndicator);
	}

}
