/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.multiechelon;

import java.util.List;
import java.util.Set;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.supplychain.inventorylayer.AggregateInventoryIfc;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.DemandSenderIfc;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;

/**
 * @author rossetti
 *
 */
public class OneWarehouseNRetailers extends ModelElement implements AggregateInventoryIfc {

	protected MultiEchelonInventoryNetwork myNetwork;
	
	protected InventoryHoldingPoint myWarehouse;
	
	/**
	 * @param parent
	 */
	public OneWarehouseNRetailers(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public OneWarehouseNRetailers(ModelElement parent, String name) {
		super(parent, name);
		// TODO Auto-generated constructor stub
		myNetwork = new MultiEchelonInventoryNetwork(this);
		myWarehouse = myNetwork.addInventoryHoldingPoint(getName()+"-Warehouse");
		myNetwork.attachIHPToExternalSupplier(myWarehouse);
	}

	/**
	 * @param leadtime
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#addItemType(jsl.utilities.random.RandomIfc)
	 */
	public ItemType addItemType(RandomIfc leadtime) {
		return myNetwork.addItemType(leadtime);
	}
	
	/**
	 * @param itemType
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#addItemType(jslx.supplychain.inventorylayer.ItemType)
	 */
	public final void addItemType(ItemType itemType) {
		myNetwork.addItemType(itemType);
	}

	/**
	 * @param itemType
	 * @param leadtime
	 */
	public void addItemType(ItemType itemType, RandomIfc leadtime) {
		myNetwork.addItemType(itemType, leadtime);
	}
	
	/**
	 * @param name
	 * @param leadtime
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#addItemType(java.lang.String, jsl.utilities.random.RandomIfc)
	 */
	public ItemType addItemType(String name, RandomIfc leadtime) {
		return myNetwork.addItemType(name, leadtime);
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#getItemTypes()
	 */
	public final Set<ItemType> getItemTypes() {
		return myNetwork.getItemTypes();
	}

	/**
	 * @param itemType
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#removeItemTypeFromNetwork(jslx.supplychain.inventorylayer.ItemType)
	 */
	public final void removeItemTypeFromNetwork(ItemType itemType) {
		myNetwork.removeItemTypeFromNetwork(itemType);
	}

	/** Returns the warehouse
	 * 
	 * @return
	 */
	public final InventoryHoldingPoint getWarehouse(){
		return myWarehouse;
	}

	/**  Checks to see if all the retailers can fill the item type
	 * 
	 * @param itemType
	 * @return
	 */
	public boolean stockedAtAllRetailers(ItemType itemType){
		List<InventoryHoldingPoint> retailers = getRetailers();
		for(InventoryHoldingPoint r: retailers){
			if (!r.canFillItemType(itemType))
				return false;	
		}
		return true;
	}

	/** Attaches a retailer to the warehouse
	 * 
	 * @param name
	 * @return
	 */
	public final InventoryHoldingPoint addRetailer(){
		return addRetailer(null);
	}

	/** Attaches a retailer to the warehouse
	 * 
	 * @param name
	 * @return
	 */
	public InventoryHoldingPoint addRetailer(String name){
		InventoryHoldingPoint ihp = myNetwork.addInventoryHoldingPoint(name);
		myNetwork.attachIHPToSupplier(myWarehouse, ihp);
		return (ihp);
	}
	
	/** The retailers supplied by the warehouse
	 * 
	 * @return
	 */
	public final List<InventoryHoldingPoint> getRetailers(){
		return myNetwork.getCustomers(myWarehouse);
	}

	/** Returns a reference to the IHP (retailer) with the provided name or null
	 *  if no such retailer
	 * 
	 * @param retailerName
	 * @return
	 */
	public final InventoryHoldingPoint getRetailer(String retailerName){
		return (InventoryHoldingPoint)getModel().getModelElement(retailerName);//TODO not the best way, but ok for now
	}
	
	/**
	 * @param retailer
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#containsInventoryHoldingPoint(jslx.supplychain.inventorylayer.InventoryHoldingPoint)
	 */
	public boolean containsRetailer(InventoryHoldingPoint retailer) {
		return myNetwork.containsInventoryHoldingPoint(retailer);
	}

	/**
	 * @param retailer
	 * @param type
	 * @param timeUntilNext
	 * @param name
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#attachDemandGeneratorToIHP(jslx.supplychain.inventorylayer.InventoryHoldingPoint, jslx.supplychain.inventorylayer.ItemType, jsl.utilities.random.distributions.DistributionIfc, java.lang.String)
	 */
	public final DemandGenerator attachDemandGeneratorToRetailer(
			InventoryHoldingPoint retailer, ItemType type,
			DistributionIfc timeUntilNext, String name) {
		return myNetwork.attachDemandGeneratorToIHP(retailer, type,
				timeUntilNext, name);
	}

	/**
	 * @param retailer
	 * @param type
	 * @param timeUntilNext
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#attachDemandGeneratorToIHP(jslx.supplychain.inventorylayer.InventoryHoldingPoint, jslx.supplychain.inventorylayer.ItemType, jsl.utilities.random.distributions.DistributionIfc)
	 */
	public final DemandGenerator attachDemandGeneratorToRetailer(
			InventoryHoldingPoint retailer, ItemType type,
			DistributionIfc timeUntilNext) {
		return myNetwork.attachDemandGeneratorToIHP(retailer, type, timeUntilNext);
	}

	/**
	 * @param retailer
	 * @param sender
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#attachDemandSenderToIHP(jslx.supplychain.inventorylayer.InventoryHoldingPoint, jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public final void attachDemandSenderToRetailer(InventoryHoldingPoint retailer,
			DemandSenderIfc sender) {
		myNetwork.attachDemandSenderToIHP(retailer, sender);
	}

	/**
	 * @param retailer
	 * @param sender
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#detachDemandSenderFromIHP(jslx.supplychain.inventorylayer.InventoryHoldingPoint, jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public final void detachDemandSenderFromRetailer(InventoryHoldingPoint retailer, DemandSenderIfc sender) {
		myNetwork.detachDemandSenderFromIHP(retailer, sender);
	}

	/**
	 * @param supplier
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#getDemandSenders(jslx.supplychain.inventorylayer.InventoryHoldingPoint)
	 */
	public final List<DemandSenderIfc> getDemandSenders(InventoryHoldingPoint supplier) {
		return myNetwork.getDemandSenders(supplier);
	}

	/**
	 * @param ihp
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#isAttachedAsCustomer(jslx.supplychain.inventorylayer.InventoryHoldingPoint)
	 */
	public final boolean isAttachedAsCustomer(InventoryHoldingPoint ihp) {
		return myNetwork.isAttachedAsCustomer(ihp);
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#getAggregateAmountBackOrderedResponseVariable()
	 */
	public final AggregateTimeWeightedVariable getAggregateAmountBackOrderedResponseVariable() {
		return myNetwork.getAggregateAmountBackOrderedResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#getAggregateAmountOnOrderResponseVariable()
	 */
	public final AggregateTimeWeightedVariable getAggregateAmountOnOrderResponseVariable() {
		return myNetwork.getAggregateAmountOnOrderResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#getAggregateAvgCustomerWaitTimeResponseVariable()
	 */
	public final ResponseVariableAverageObserver getAggregateAvgCustomerWaitTimeResponseVariable() {
		return myNetwork.getAggregateAvgCustomerWaitTimeResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#getAggregateAvgFirstFillRateResponseVariable()
	 */
	public final ResponseVariableAverageObserver getAggregateAvgFirstFillRateResponseVariable() {
		return myNetwork.getAggregateAvgFirstFillRateResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#getAggregateNumberBackOrderedResponseVariable()
	 */
	public final AggregateTimeWeightedVariable getAggregateNumberBackOrderedResponseVariable() {
		return myNetwork.getAggregateNumberBackOrderedResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#getAggregateNumberOfReplenishmentDemandsCounter()
	 */
	public final AggregateCounter getAggregateNumberOfReplenishmentDemandsCounter() {
		return myNetwork.getAggregateNumberOfReplenishmentDemandsCounter();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#getAggregateOnHandInventoryResponseVariable()
	 */
	public final AggregateTimeWeightedVariable getAggregateOnHandInventoryResponseVariable() {
		return myNetwork.getAggregateOnHandInventoryResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#getAggregateStockOutPercentageResponseVariable()
	 */
	public final AggregateTimeWeightedVariable getAggregateStockOutPercentageResponseVariable() {
		return myNetwork.getAggregateStockOutPercentageResponseVariable();
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#subscribeToAggregateNumberBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateNumberBackOrdered(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.subscribeToAggregateNumberBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#subscribeToAggregateAmountBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateAmountBackOrdered(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.subscribeToAggregateAmountBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#subscribeToAggregateAmountOnOrder(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateAmountOnOrder(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.subscribeToAggregateAmountOnOrder(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#subscribeToAggregateAvgCustomerWaitTime(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void subscribeToAggregateAvgCustomerWaitTime(
			ResponseVariableAverageObserver aggregate) {
		myNetwork.subscribeToAggregateAvgCustomerWaitTime(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#subscribeToAggregateAvgFirstFillRate(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void subscribeToAggregateAvgFirstFillRate(
			ResponseVariableAverageObserver aggregate) {
		myNetwork.subscribeToAggregateAvgFirstFillRate(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#subscribeToAggregateOnHandInventory(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateOnHandInventory(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.subscribeToAggregateOnHandInventory(aggregate);
	}

	/**
	 * @param counter
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#subscribeToAggregateReplenishmentDemandCounter(jsl.modeling.elements.variable.AggregateCounter)
	 */
	public void subscribeToAggregateReplenishmentDemandCounter(
			AggregateCounter counter) {
		myNetwork.subscribeToAggregateReplenishmentDemandCounter(counter);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#subscribeToAggregateStockOutIndicator(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateStockOutIndicator(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.subscribeToAggregateStockOutIndicator(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#unsubscribeFromAggregateAmountBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateAmountBackOrdered(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.unsubscribeFromAggregateAmountBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#unsubscribeFromAggregateAmountOnOrder(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateAmountOnOrder(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.unsubscribeFromAggregateAmountOnOrder(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#unsubscribeFromAggregateAvgCustomerWaitTime(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void unsubscribeFromAggregateAvgCustomerWaitTime(
			ResponseVariableAverageObserver aggregate) {
		myNetwork.unsubscribeFromAggregateAvgCustomerWaitTime(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#unsubscribeFromAggregateAvgFirstFillRate(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void unsubscribeFromAggregateAvgFirstFillRate(
			ResponseVariableAverageObserver aggregate) {
		myNetwork.unsubscribeFromAggregateAvgFirstFillRate(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#unsubscribeFromAggregateNumberBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateNumberBackOrdered(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.unsubscribeFromAggregateNumberBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#unsubscribeFromAggregateOnHandInventory(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateOnHandInventory(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.unsubscribeFromAggregateOnHandInventory(aggregate);
	}

	/**
	 * @param counter
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#unsubscribeFromAggregateReplenishmentDemandCounter(jsl.modeling.elements.variable.AggregateCounter)
	 */
	public void unsubscribeFromAggregateReplenishmentDemandCounter(
			AggregateCounter counter) {
		myNetwork.unsubscribeFromAggregateReplenishmentDemandCounter(counter);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.MultiEchelonInventoryNetwork#unsubscribeFromAggregateStockOutIndicator(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateStockOutIndicator(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.unsubscribeFromAggregateStockOutIndicator(aggregate);
	}


	
	
}
