/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.supplychain.multiechelon;

import java.util.List;
import java.util.Set;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.queue.QueueDiscipline;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.supplychain.inventorylayer.AggregateInventoryIfc;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.DemandSenderIfc;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.transportlayer.DemandLoadFormingRuleIfc;
import jslx.supplychain.transportlayer.TimeBasedLoadCarrier;
import jslx.supplychain.transportlayer.DemandLoadBuilder.LoadFormingOption;

/**
 * @author rossetti
 *
 */
public class OneWarehouseNRetailersWithLoadBuildingCarrier extends ModelElement implements AggregateInventoryIfc{

	protected TimeBasedLoadCarrier myCarrier; 
	
	protected OneWarehouseNRetailers myNetwork;
		
	/**
	 * @param parent
	 */
	public OneWarehouseNRetailersWithLoadBuildingCarrier(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public OneWarehouseNRetailersWithLoadBuildingCarrier(ModelElement parent, String name) {
		super(parent, name);
		// TODO Auto-generated constructor stub
		myNetwork = new OneWarehouseNRetailers(this);
		myCarrier = new TimeBasedLoadCarrier(this);
		myNetwork.getWarehouse().setDemandCarrier(myCarrier);
	}

	/**
	 * @param flag
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#setReactToLoadBuildersFlag(boolean)
	 */
	public void setReactToLoadBuildersFlag(boolean flag) {
		myCarrier.setReactToLoadBuildersFlag(flag);
	}

	/**
	 * @return
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#getShipmentCounter()
	 */
	public final AggregateCounter getShipmentCounter() {
		return myCarrier.getShipmentCounter();
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#getShipmentCounter(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public Counter getShipmentCounter(DemandSenderIfc destination) {
		return myCarrier.getShipmentCounter(destination);
	}

	/** Sets the transport time from the external supplier to the warehouse
	 * @param distribution
	 * @see jslx.supplychain.transportlayer.TimeBasedDemandCarrier#setTransportTime(jslx.supplychain.inventorylayer.DemandSenderIfc, jsl.utilities.random.RandomIfc)
	 */
	public final void setTransportTimeToWarehouse(RandomIfc distribution) {
		myCarrier.setTransportTime(myNetwork.getWarehouse(), distribution);
	}

	/**
	 * @param itemType
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#stockedAtAllRetailers(jslx.supplychain.inventorylayer.ItemType)
	 */
	public boolean stockedAtAllRetailers(ItemType itemType) {
		return myNetwork.stockedAtAllRetailers(itemType);
	}

	/**
	 * @param leadtime
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#addItemType(jsl.utilities.random.RandomIfc)
	 */
	public ItemType addItemType(RandomIfc leadtime) {
		String name = null;
		return addItemType(name, leadtime);
	}

	/**
	 * @param itemType
	 * @param leadtime
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#addItemType(jslx.supplychain.inventorylayer.ItemType, jsl.utilities.random.RandomIfc)
	 */
	public void addItemType(ItemType itemType, RandomIfc leadtime) {
		myNetwork.addItemType(itemType, leadtime);
	}

	/**
	 * @param itemType
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#addItemType(jslx.supplychain.inventorylayer.ItemType)
	 */
	public final void addItemType(ItemType itemType) {
		myNetwork.addItemType(itemType);
	}

	/**
	 * @param name
	 * @param leadtime
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#addItemType(java.lang.String, jsl.utilities.random.RandomIfc)
	 */
	public final ItemType addItemType(String name, RandomIfc leadtime) {
		return myNetwork.addItemType(name, leadtime);
	}

	
	/**
	 * @param itemType
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#removeItemTypeFromNetwork(jslx.supplychain.inventorylayer.ItemType)
	 */
	public final void removeItemTypeFromNetwork(ItemType itemType) {
		myNetwork.removeItemTypeFromNetwork(itemType);
	}

	/**
	 * @param name
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#addRetailer(java.lang.String)
	 */
	public InventoryHoldingPoint addRetailer(String name, RandomIfc transportTime) {
		InventoryHoldingPoint ihp = myNetwork.addRetailer(name);
		myCarrier.setTransportTime(ihp, transportTime);
		myCarrier.assignLoadBuilder(ihp);
		return ihp;
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#addRetailer(java.lang.String)
	 */
	public InventoryHoldingPoint addRetailer(RandomIfc transportTime) {
		return addRetailer(null, transportTime);
	}
	
	/**
	 * @param retailer
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#containsRetailer(jslx.supplychain.inventorylayer.InventoryHoldingPoint)
	 */
	public boolean containsRetailer(InventoryHoldingPoint retailer) {
		return myNetwork.containsRetailer(retailer);
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getWarehouse()
	 */
	public final InventoryHoldingPoint getWarehouse() {
		return myNetwork.getWarehouse();
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#geMinCubeLimit(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public double geMinCubeLimit(DemandSenderIfc destination) {
		return myCarrier.geMinCubeLimit(destination);
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#geMinWeightLimit(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public double geMinWeightLimit(DemandSenderIfc destination) {
		return myCarrier.geMinWeightLimit(destination);
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#getCountLimit(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public int getCountLimit(DemandSenderIfc destination) {
		return myCarrier.getCountLimit(destination);
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#getDefaultLoadFormingOption(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public LoadFormingOption getDefaultLoadFormingOption(
			DemandSenderIfc destination) {
		return myCarrier.getDefaultLoadFormingOption(destination);
	}

	/**
	 * @return
	 * @see jslx.supplychain.transportlayer.TimeBasedDemandCarrier#getImmediateTransportFlag()
	 */
	public boolean getImmediateTransportFlag() {
		return myCarrier.getImmediateTransportFlag();
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#getMaxCubeLimit(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public double getMaxCubeLimit(DemandSenderIfc destination) {
		return myCarrier.getMaxCubeLimit(destination);
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#getMaxWeightLimit(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public double getMaxWeightLimit(DemandSenderIfc destination) {
		return myCarrier.getMaxWeightLimit(destination);
	}

	/**
	 * @return
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#getReactToLoadBuildersFlag()
	 */
	public boolean getReactToLoadBuildersFlag() {
		return myCarrier.getReactToLoadBuildersFlag();
	}

	/**
	 * @param destination
	 * @param countLimit
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#setCountLimit(jslx.supplychain.inventorylayer.DemandSenderIfc, int)
	 */
	public void setCountLimit(DemandSenderIfc destination, int countLimit) {
		myCarrier.setCountLimit(destination, countLimit);
	}

	/**
	 * @param destination
	 * @param minCube
	 * @param maxCube
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#setCubeFormingLimits(jslx.supplychain.inventorylayer.DemandSenderIfc, double, double)
	 */
	public void setCubeFormingLimits(DemandSenderIfc destination,
			double minCube, double maxCube) {
		myCarrier.setCubeFormingLimits(destination, minCube, maxCube);
	}

	/**
	 * @param destination
	 * @param option
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#setDefaultLoadFormingOption(jslx.supplychain.inventorylayer.DemandSenderIfc, jslx.supplychain.transportlayer.DemandLoadBuilder.LoadFormingOption)
	 */
	public void setDefaultLoadFormingOption(DemandSenderIfc destination,
			LoadFormingOption option) {
		myCarrier.setDefaultLoadFormingOption(destination, option);
	}

	/**
	 * @param destination
	 * @param discipline
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#setDemandQueueInitialDiscipline(jslx.supplychain.inventorylayer.DemandSenderIfc, jsl.modeling.elements.queue.QueueDiscipline)
	 */
	public void setDemandQueueInitialDiscipline(DemandSenderIfc destination,
			QueueDiscipline discipline) {
		myCarrier.setDemandQueueInitialDiscipline(destination, discipline);
	}

	/**
	 * @param destination
	 * @param rule
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#setLoadFormingRule(jslx.supplychain.inventorylayer.DemandSenderIfc, jslx.supplychain.transportlayer.DemandLoadFormingRuleIfc)
	 */
	public void setLoadFormingRule(DemandSenderIfc destination,
			DemandLoadFormingRuleIfc rule) {
		myCarrier.setLoadFormingRule(destination, rule);
	}

	/**
	 * @param destination
	 * @param discipline
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#setLoadQueueInitialDiscipline(jslx.supplychain.inventorylayer.DemandSenderIfc, jsl.modeling.elements.queue.QueueDiscipline)
	 */
	public void setLoadQueueInitialDiscipline(DemandSenderIfc destination,
			QueueDiscipline discipline) {
		myCarrier.setLoadQueueInitialDiscipline(destination, discipline);
	}

	/**
	 * @param destination
	 * @param minWeight
	 * @param maxWeight
	 * @see jslx.supplychain.transportlayer.TimeBasedLoadCarrier#setWeightFormingLimits(jslx.supplychain.inventorylayer.DemandSenderIfc, double, double)
	 */
	public void setWeightFormingLimits(DemandSenderIfc destination,
			double minWeight, double maxWeight) {
		myCarrier.setWeightFormingLimits(destination, minWeight, maxWeight);
	}

	/**
	 * @param retailer
	 * @param type
	 * @param timeUntilNext
	 * @param name
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#attachDemandGeneratorToRetailer(jslx.supplychain.inventorylayer.InventoryHoldingPoint, jslx.supplychain.inventorylayer.ItemType, jsl.utilities.random.distributions.DistributionIfc, java.lang.String)
	 */
	public final DemandGenerator attachDemandGeneratorToRetailer(
			InventoryHoldingPoint retailer, ItemType type,
			DistributionIfc timeUntilNext, String name) {
		return myNetwork.attachDemandGeneratorToRetailer(retailer, type,
				timeUntilNext, name);
	}

	/**
	 * @param retailer
	 * @param type
	 * @param timeUntilNext
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#attachDemandGeneratorToRetailer(jslx.supplychain.inventorylayer.InventoryHoldingPoint, jslx.supplychain.inventorylayer.ItemType, jsl.utilities.random.distributions.DistributionIfc)
	 */
	public final DemandGenerator attachDemandGeneratorToRetailer(InventoryHoldingPoint retailer, ItemType type, DistributionIfc timeUntilNext) {
		return myNetwork.attachDemandGeneratorToRetailer(retailer, type, timeUntilNext);
	}

	/**
	 * @param retailer
	 * @param sender
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#attachDemandSenderToRetailer(jslx.supplychain.inventorylayer.InventoryHoldingPoint, jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public final void attachDemandSenderToRetailer(InventoryHoldingPoint retailer, DemandSenderIfc sender) {
		myNetwork.attachDemandSenderToRetailer(retailer, sender);
	}

	/**
	 * @param retailer
	 * @param sender
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#detachDemandSenderFromRetailer(jslx.supplychain.inventorylayer.InventoryHoldingPoint, jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public final void detachDemandSenderFromRetailer(InventoryHoldingPoint retailer, DemandSenderIfc sender) {
		myNetwork.detachDemandSenderFromRetailer(retailer, sender);
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getAggregateAmountBackOrderedResponseVariable()
	 */
	public final AggregateTimeWeightedVariable getAggregateAmountBackOrderedResponseVariable() {
		return myNetwork.getAggregateAmountBackOrderedResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getAggregateAmountOnOrderResponseVariable()
	 */
	public final AggregateTimeWeightedVariable getAggregateAmountOnOrderResponseVariable() {
		return myNetwork.getAggregateAmountOnOrderResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getAggregateAvgCustomerWaitTimeResponseVariable()
	 */
	public final ResponseVariableAverageObserver getAggregateAvgCustomerWaitTimeResponseVariable() {
		return myNetwork.getAggregateAvgCustomerWaitTimeResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getAggregateAvgFirstFillRateResponseVariable()
	 */
	public final ResponseVariableAverageObserver getAggregateAvgFirstFillRateResponseVariable() {
		return myNetwork.getAggregateAvgFirstFillRateResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getAggregateNumberBackOrderedResponseVariable()
	 */
	public final AggregateTimeWeightedVariable getAggregateNumberBackOrderedResponseVariable() {
		return myNetwork.getAggregateNumberBackOrderedResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getAggregateNumberOfReplenishmentDemandsCounter()
	 */
	public final AggregateCounter getAggregateNumberOfReplenishmentDemandsCounter() {
		return myNetwork.getAggregateNumberOfReplenishmentDemandsCounter();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getAggregateOnHandInventoryResponseVariable()
	 */
	public final AggregateTimeWeightedVariable getAggregateOnHandInventoryResponseVariable() {
		return myNetwork.getAggregateOnHandInventoryResponseVariable();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getAggregateStockOutPercentageResponseVariable()
	 */
	public final AggregateTimeWeightedVariable getAggregateStockOutPercentageResponseVariable() {
		return myNetwork.getAggregateStockOutPercentageResponseVariable();
	}

	/**
	 * @param supplier
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getDemandSenders(jslx.supplychain.inventorylayer.InventoryHoldingPoint)
	 */
	public final List<DemandSenderIfc> getDemandSenders(
			InventoryHoldingPoint supplier) {
		return myNetwork.getDemandSenders(supplier);
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getItemTypes()
	 */
	public final Set<ItemType> getItemTypes() {
		return myNetwork.getItemTypes();
	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#getRetailers()
	 */
	public final List<InventoryHoldingPoint> getRetailers() {
		return myNetwork.getRetailers();
	}

	/**
	 * @param ihp
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#isAttachedAsCustomer(jslx.supplychain.inventorylayer.InventoryHoldingPoint)
	 */
	public final boolean isAttachedAsCustomer(InventoryHoldingPoint ihp) {
		return myNetwork.isAttachedAsCustomer(ihp);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#subscribeToAggregateNumberBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateNumberBackOrdered(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.subscribeToAggregateNumberBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#subscribeToAggregateAmountBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateAmountBackOrdered(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.subscribeToAggregateAmountBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#subscribeToAggregateAmountOnOrder(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateAmountOnOrder(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.subscribeToAggregateAmountOnOrder(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#subscribeToAggregateAvgCustomerWaitTime(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void subscribeToAggregateAvgCustomerWaitTime(
			ResponseVariableAverageObserver aggregate) {
		myNetwork.subscribeToAggregateAvgCustomerWaitTime(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#subscribeToAggregateAvgFirstFillRate(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void subscribeToAggregateAvgFirstFillRate(
			ResponseVariableAverageObserver aggregate) {
		myNetwork.subscribeToAggregateAvgFirstFillRate(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#subscribeToAggregateOnHandInventory(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateOnHandInventory(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.subscribeToAggregateOnHandInventory(aggregate);
	}

	/**
	 * @param counter
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#subscribeToAggregateReplenishmentDemandCounter(jsl.modeling.elements.variable.AggregateCounter)
	 */
	public void subscribeToAggregateReplenishmentDemandCounter(
			AggregateCounter counter) {
		myNetwork.subscribeToAggregateReplenishmentDemandCounter(counter);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#subscribeToAggregateStockOutIndicator(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void subscribeToAggregateStockOutIndicator(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.subscribeToAggregateStockOutIndicator(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#unsubscribeFromAggregateAmountBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateAmountBackOrdered(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.unsubscribeFromAggregateAmountBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#unsubscribeFromAggregateAmountOnOrder(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateAmountOnOrder(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.unsubscribeFromAggregateAmountOnOrder(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#unsubscribeFromAggregateAvgCustomerWaitTime(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void unsubscribeFromAggregateAvgCustomerWaitTime(
			ResponseVariableAverageObserver aggregate) {
		myNetwork.unsubscribeFromAggregateAvgCustomerWaitTime(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#unsubscribeFromAggregateAvgFirstFillRate(jsl.modeling.elements.variable.ResponseVariableAverageObserver)
	 */
	public void unsubscribeFromAggregateAvgFirstFillRate(
			ResponseVariableAverageObserver aggregate) {
		myNetwork.unsubscribeFromAggregateAvgFirstFillRate(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#unsubscribeFromAggregateNumberBackOrdered(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateNumberBackOrdered(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.unsubscribeFromAggregateNumberBackOrdered(aggregate);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#unsubscribeFromAggregateOnHandInventory(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateOnHandInventory(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.unsubscribeFromAggregateOnHandInventory(aggregate);
	}

	/**
	 * @param counter
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#unsubscribeFromAggregateReplenishmentDemandCounter(jsl.modeling.elements.variable.AggregateCounter)
	 */
	public void unsubscribeFromAggregateReplenishmentDemandCounter(
			AggregateCounter counter) {
		myNetwork.unsubscribeFromAggregateReplenishmentDemandCounter(counter);
	}

	/**
	 * @param aggregate
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailers#unsubscribeFromAggregateStockOutIndicator(jsl.modeling.elements.variable.AggregateTimeWeightedVariable)
	 */
	public void unsubscribeFromAggregateStockOutIndicator(
			AggregateTimeWeightedVariable aggregate) {
		myNetwork.unsubscribeFromAggregateStockOutIndicator(aggregate);
	}
	
	

}
