/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.supplychain.proliferationYC;

import java.util.ArrayList;
import java.util.List;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.RandomList;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.reporting.JSL;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;

/** An adoption proliferator must be set up and turned on whenever a new item type is
 *  added to a category at a particular location
 *  
 *  Its initial list of candidate locations for adoption are those locations
 *  that do not stock the item.
 *  
 *  As more locations get selected to stock the item, the candidate list should
 *  be reduced.  When there are no more candidate locations the generator should
 *  be turned off.
 *  
 *  It does not setup or manage demand generators. It only sends proliferation
 *  messages to its associated category
 *  
 */
public class AdoptionProliferator extends ProliferatorAbstract {

	/** The item type associated with this proliferator.
	 * 
	 */
	protected CItemType myItemType;

	/** The initial locations to be considered during proliferation. These
	 *  locations may or may not already stock the item type 
	 */
	protected List<InventoryHoldingPoint> myInitialLocations;

	/** The stocking locations of the item type. These
	 *  locations should already stock the item type. 
	 */
	protected List<InventoryHoldingPoint> myStockingLocations;
	
	/** Facilitates the random selection of a candidate location
	 * 
	 */
	protected RandomList<InventoryHoldingPoint> myCandidateLocations;

	/** The base mean time between arrivals of the proliferation events
	 * 
	 */
	protected double myBaseMTBA;
	
	/**
	 * @param parent
	 */
	public AdoptionProliferator(ModelElement parent, double mtba, CItemType itemType) {
		this(parent, mtba, itemType, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public AdoptionProliferator(ModelElement parent, double mtba, CItemType itemType, String name) {
		super(parent, name);
		if (mtba <= 0.0)
			throw new IllegalArgumentException("The base proliferation rate must be > 0.0");
		myBaseMTBA = mtba;
		setTimeBetweenEvents(new Exponential(mtba));
		setStartOnInitializeFlag(false); // generator does not start on model initialization
		
		setItemType(itemType);	
		
		myInitialLocations = new ArrayList<InventoryHoldingPoint>();
		myCandidateLocations = new RandomList<InventoryHoldingPoint>(this);
		myCandidateLocations.setInitializationOption(false);// will be initialized via initializeLocations() method
		myStockingLocations = new ArrayList<InventoryHoldingPoint>();

	}

	/** Initialize the set of locations to be considered for proliferation
	 * 
	 * @param locations  All possible stocking locations to be considered
	 */
	public void initializeLocations(List<InventoryHoldingPoint> locations){
//		JSL.out.println("Clearing the stocking and candidate lists");
		myStockingLocations.clear();
		myCandidateLocations.clear();
		for(InventoryHoldingPoint ihp: locations){
			if (!ihp.canFillItemType(myItemType)){
//				JSL.out.println("Adding " + ihp.getName() + " to candidate list");
				myCandidateLocations.add(ihp);
			}
			else {
//				JSL.out.println("Adding " + ihp.getName() + " to stocked at list");
				myStockingLocations.add(ihp);
			}
		}
		// need to set mtba based on # items
		int n = myCandidateLocations.size();
		if (n > 0){
			Exponential e = (Exponential)this.getTimeBetweenEvents();
			e.setMean(myBaseMTBA/n);
		}
	}
	
	/** Knows how to select the location from the current candidate location
	 *  list to indicate where to add the item type
	 * 
	 * @return
	 */
	protected InventoryHoldingPoint selectLocation(){
		InventoryHoldingPoint ihp = myCandidateLocations.getRandomElement();
		myCandidateLocations.remove(ihp);
		myStockingLocations.add(ihp);
		int n = myCandidateLocations.size();
		if (n > 0){
			Exponential e = (Exponential)this.getTimeBetweenEvents();
			e.setMean(myBaseMTBA/n);
		} else
			suspend();

		return ihp;
	}
	
	protected void setItemType(CItemType itemType){
		if (itemType == null)
			throw new IllegalArgumentException("The supplied item type was null");
		myItemType = itemType;
	}

	@Override
	protected void proliferationAction(JSLEvent event) {
		// select the ihp from the candidate list
		InventoryHoldingPoint ihp = selectLocation();
		ProliferationMessage m = new ProliferationMessage();
		m.proliferator = this;
		m.category = myItemType.getCategory();
		m.itemType = myItemType;
		m.ihp = ihp;
		m.status = ProliferationMessage.Status.ADOPTION;
		notifyProliferationListeners(m);
	}
	
	protected void removedFromModel(){
		super.removedFromModel();
		myItemType = null;
		myInitialLocations.clear();
		myInitialLocations = null;
		myStockingLocations.clear();
		myStockingLocations = null;
	}
	
}
