/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.proliferationYC;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Constant;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;

/** This class controls the proliferation rate for item types within
 *  a category at a particular location.  It's sole function
 *  is to ask a category to proliferate. 
 * 
 *
 */
public class CategoryLocationProliferator extends ProliferatorAbstract {

	/** The IHP associated with this proliferator.
	 * 
	 */
	protected InventoryHoldingPoint myIHP;

	/** The category for item types for this proliferator
	 * 
	 */
	protected ItemCategoryAbstract myItemCategory;

	/**
	 * @param parent
	 */
	public CategoryLocationProliferator(ModelElement parent) {
		this(parent, Constant.ZERO, Constant.POSITIVE_INFINITY, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public CategoryLocationProliferator(ModelElement parent, String name) {
		this(parent, Constant.ZERO, Constant.POSITIVE_INFINITY, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * @param parent
	 * @param timeUntilFirst
	 */
	public CategoryLocationProliferator(ModelElement parent,
			RandomIfc timeUntilFirst) {
		this(parent, timeUntilFirst, Constant.POSITIVE_INFINITY, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * @param parent
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 */
	public CategoryLocationProliferator(ModelElement parent,
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext) {
		this(parent, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * @param parent
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param name
	 */
	public CategoryLocationProliferator(ModelElement parent,
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext, String name) {
		this(parent, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/**
	 * @param parent
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 */
	public CategoryLocationProliferator(ModelElement parent,
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext, long maxNum) {
		this(parent, timeUntilFirst, timeUntilNext, maxNum, Double.POSITIVE_INFINITY, null);
	}

	/**
	 * @param parent
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 * @param timeUntilLast
	 * @param name
	 */
	public CategoryLocationProliferator(ModelElement parent,
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext, long maxNum,
			double timeUntilLast, String name) {
		super(parent, timeUntilFirst, timeUntilNext, maxNum, timeUntilLast, name);
	}
	
	public void setCategoryLocation(ItemCategoryAbstract category, InventoryHoldingPoint ihp){
		setInventoryHoldingPoint(ihp);
		setItemCategory(category);
	}
	
	public void setInventoryHoldingPoint(InventoryHoldingPoint ihp){
		if (ihp == null)
			throw new IllegalArgumentException("The supplied IHP was null");
		myIHP = ihp;
	}
	
	protected void setItemCategory(ItemCategoryAbstract category){
		if (category == null)
			throw new IllegalArgumentException("The supplied item category was null");
		
		if (myItemCategory != null)
			throw new IllegalArgumentException("The item category has already been set.");
		
		myItemCategory = category;
	}
	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.proliferation.ProliferatorAbstract#proliferationAction(jsl.modeling.JSLEvent)
	 */
	@Override
	protected void proliferationAction(JSLEvent event) {
		CItemType itemType = myItemCategory.generateItemType();
		myItemCategory.addItemType(itemType);
		ProliferationMessage m = new ProliferationMessage();
		m.proliferator = this;
		m.category = myItemCategory;
		m.itemType = itemType;
		m.ihp = myIHP;
		m.status = ProliferationMessage.Status.ACCEPTANCE;
		notifyProliferationListeners(m);
	}
	
}
