/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.proliferationYC;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.Constant;

/** Any item types added to this category have similar
 *  weight, cube, and lead time characteristics
 * 
 *
 */
public class DefaultItemCategory extends ItemCategoryAbstract {
		
	/** A random variable representing the weight of the items
	 *  in the category
	 * 
	 */
	protected RandomVariable myWeightRV;
	
	/** A random variable representing the cube of the items
	 *  in the category
	 * 
	 */
	protected RandomVariable myCubeRV;
	
	/** Item types are stocked a locations, they may be stocked
	 *  at a location that has a lead time for supplying the item type
	 *  This field can be used to determine the lead time for the
	 *  item types if it is set.  
	 * 
	 */
	protected RandomIfc myItemTypeLeadTime = Constant.ONE;

	/**
	 * @param parent
	 */
	public DefaultItemCategory(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public DefaultItemCategory(ModelElement parent, String name) {
		super(parent, name);
		
		myWeightRV = new RandomVariable(this, Constant.ONE);
		myCubeRV = new RandomVariable(this, Constant.ONE);
	}

	/**
	 * @param weightCDF the CDF to set
	 */
	public void setWeightCDFInitialRandomSource(RandomIfc weightCDF) {
		myWeightRV.setInitialRandomSource(weightCDF);
	}

	/**
	 * @param cubeCDF the CubeCDF to set
	 */
	public void setCubeCDFInitialRandomSource(RandomIfc cubeCDF) {
		myCubeRV.setInitialRandomSource(cubeCDF);
	}

	/**
	 * @param leadTime the leadTime to set
	 */
	public void setLeadTime(RandomIfc leadTime) {
		if (leadTime == null)
			throw new IllegalArgumentException("The lead time was null");
		myItemTypeLeadTime = leadTime;
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.proliferation.ItemCategoryAbstract#determineCube()
	 */
	@Override
	protected double determineCube() {
		return myCubeRV.getValue();
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.proliferation.ItemCategoryAbstract#determineWeight()
	 */
	@Override
	protected double determineWeight() {
		return myWeightRV.getValue();
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.proliferation.ItemCategoryAbstract#determineLeadTimeDistribution(jsl.research.supplychain.inventorylayer.ItemType)
	 */
	@Override
	protected RandomIfc determineLeadTimeDistribution() {
		return myItemTypeLeadTime;
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.proliferation.ItemCategoryAbstract#generateItemType()
	 */
	@Override
	public CItemType generateItemType() {
		CItemType itemType = new CItemType(this, determineItemTypeName(), determineWeight(), determineCube(), determineLeadTimeDistribution());
		return itemType;
	}

}
