/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.proliferationYC;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import jsl.modeling.ModelElement;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.reporting.JSL;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;

/** An ItemCategory is a set or list of item types that are grouped
 *  together.  For example, they may be grouped together because they
 *  are functionally equivalent
 *
 *
 */
abstract public class ItemCategoryAbstract extends ModelElement {
	
	/** Indicates that an initial item type has been added to the category
	 * 
	 */
	public static final int INITIAL_ITEM_TYPE_ADDED = JSL.getNextEnumConstant();
	
	/** Indicates that an initial item type has been added to the category
	 * 
	 */
	public static final int INITIAL_ITEM_TYPE_REMOVED = JSL.getNextEnumConstant();

	/** Indicates that an item type has been added only for use during the replication
	 * 
	 */
	public static final int ITEM_TYPE_ADDED = JSL.getNextEnumConstant();

	/** Indicates that an item type has been removed that was only used during a replication
	 * 
	 */
	public static final int ITEM_TYPE_REMOVED = JSL.getNextEnumConstant();
	
	/** A reference to the last item type that was added or removed
	 * 
	 */
	protected CItemType myLastItemType;

	/** The initial list of item types to which this
	 *  model element will be reset at the beginning
	 *  of each replication, added via the addInitialItemType() method
	 * 
	 */
	protected List<CItemType> myInitialItemTypes;
	
	/** All item types currently associated with this category
	 * 
	 */
	protected List<CItemType> myItemTypes;

	/** Proliferated item types, all item types added via the addItemType() method
	 * 
	 */
	protected List<CItemType> myProliferatedItemTypes;

	/** Listens for addition/removals of item types in the category
	 * 
	 */
	protected List<ItemCategoryListenerIfc> myItemCategoryChangeListeners;
	
	/**
	 * @param parent
	 */
	public ItemCategoryAbstract(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public ItemCategoryAbstract(ModelElement parent, String name) {
		super(parent, name);

		myInitialItemTypes = new ArrayList<CItemType>();
		myItemTypes = new ArrayList<CItemType>();
		myItemCategoryChangeListeners = new ArrayList<ItemCategoryListenerIfc>();
		myProliferatedItemTypes = new ArrayList<CItemType>();
		
	}

	/** Returns a new list representing the inventory at the supplied
	 *  ihp that are within this category
	 * 	A convenience method, since the ihp itself can do this.
	 * 
	 * @param ihp
	 * @param list the list to fill
	 * @return
	 */
	public void getInventory(InventoryHoldingPoint ihp, List<Inventory> list){
		 if (ihp == null)
			 throw new IllegalArgumentException("The ihp was null");
		 
		 if (list == null)
			 throw new IllegalArgumentException("The list was null");
		
		for(ItemType itemType: myItemTypes){
			Inventory i = ihp.getInventory(itemType);
			if (i != null)
				list.add(i);
		}
	}
	 
	/** Returns a new list representing the inventory at the supplied
	 *  ihp that are within this category
	 *  A convenience method, since the ihp itself can do this.
	 * 
	 * @param ihp
	 * @return
	 */
	public List<Inventory> getInventory(InventoryHoldingPoint ihp){
		List<Inventory> list = new ArrayList<Inventory>();
		getInventory(ihp, list);
		return list;
	}
	
	/** Only the initial item types (unmodifiable)
	 * 
	 * @return the InitialItemTypes
	 */
	public List<CItemType> getInitialItemTypes() {
		return Collections.unmodifiableList(myInitialItemTypes);
	}

	/** All item types (initial or proliferated), unmodifiable
	 * 
	 * @return the ItemTypes
	 */
	public List<CItemType> getItemTypes() {
		return Collections.unmodifiableList(myItemTypes);
	}
	
	/** Returns the number of item types in the category
	 * 
	 * @return
	 */
	public int getNumberOfItemTypes(){
		return myItemTypes.size();
	}

	/** Checks if the supplied item type is within this category
	 * @param o
	 * @return
	 */
	public boolean containsItemType(CItemType type) {
		return myItemTypes.contains(type);
	}

	/** The item types that were added via addItemTypes(), unmodifiable
	 * 
	 * @return the ProliferatedItemTypes
	 */
	public List<CItemType> getProliferatedItemTypes() {
		return Collections.unmodifiableList(myProliferatedItemTypes);
	}

	/** Creates and adds an item type to the list of initial item
	 *  types for this category
	 * 
	 * @return
	 */
	public CItemType addInitialItemType(){
		CItemType itemType = generateItemType();
		addInitialItemType(itemType);
		return itemType;
	}

	/**
	 * @param itemType must not already have been added
	 */
	public void addInitialItemType(CItemType itemType) {
		if (myInitialItemTypes.contains(itemType))
				throw new IllegalArgumentException("The ItemType has already been added");
		
		myInitialItemTypes.add(itemType);
		myItemTypes.add(itemType);
		itemType.setCategory(this);
		myLastItemType = itemType;
		notifyItemCategoryListeners(INITIAL_ITEM_TYPE_ADDED);
	}

	/** Creates and adds the item type to the category
	 * 	
	 *  Note: The item type is considered proliferated, i.e. not part of the initial item types for 
	 *  the category
	 *  
	 * @return
	 */
	public CItemType addItemType(){
		CItemType type = generateItemType();
		addItemType(type);
		return type;
	}
		
	/** Adds the item type to the category
	 *  
	 *  Note: The item type is considered proliferated, i.e. not part of the initial item types for 
	 *  the category
	 *  
	 * @param itemType if already added, it throws an exception
	 */
	public void addItemType(CItemType itemType) {
		if (myItemTypes.contains(itemType))
			throw new IllegalArgumentException("The ItemType has already been added");

		myItemTypes.add(itemType);
		myProliferatedItemTypes.add(itemType);			
		itemType.setCategory(this);			
		myLastItemType = itemType;
		notifyItemCategoryListeners(ITEM_TYPE_ADDED);
	}

	/** Returns the last item type either removed or added
	 * 
	 * @return
	 */
	public CItemType getLastItemType(){
		return myLastItemType;
	}
	
	/**
	 * @param o
	 * @return
	 */
	public boolean attachItemCategoryChangeListener(ItemCategoryListenerIfc o) {
		if (o == null)
			return false;
		
		if (myItemCategoryChangeListeners == null)
			myItemCategoryChangeListeners = new ArrayList<ItemCategoryListenerIfc>();

		return myItemCategoryChangeListeners.add(o);
	}

	/**
	 * @param o
	 * @return
	 */
	public boolean containsItemCategoryChangeListener(ItemCategoryListenerIfc o) {
		if (o == null)
			return false;
		return myItemCategoryChangeListeners.contains(o);
	}

	/**
	 * @param o
	 * @return
	 */
	public boolean detachItemCategoryChangeListener(ItemCategoryListenerIfc o) {
		if (o == null)
			return false;

		return myItemCategoryChangeListeners.remove(o);
	}

	/** Tells the category to notify its item category listeners with the supplied message
	 * 
	 * @param message
	 */
	protected void notifyItemCategoryListeners(int status){
//		JSL.out.println("In ItemCategoryAbstract notifyItemCategoryListeners(int status)");
//		JSL.out.println("At time " + getTime() + " notifying itemCategoryChangeListeners");
//		JSL.out.println("Status = "+status);
		
		for(ItemCategoryListenerIfc p: myItemCategoryChangeListeners){
			p.itemCategoryChanged(this, status);
		}
	}
	
	/** Knows how to determine the lead time distribution for ItemTypes
	 *  that are proliferated
	 * @return
	 */
	abstract protected RandomIfc determineLeadTimeDistribution();

	/** Knows how to generate an ItemType to be added to the associated IHP
	 * 
	 * @return
	 */
	abstract public CItemType generateItemType();

	/** Knows how to determine the weight for proliferated item types
	 * 
	 * @return
	 */
	abstract protected double determineWeight();
	
	/** Knows how to determine the cube for proliferated item types
	 * 
	 * @return
	 */
	abstract protected double determineCube();

	/** This method is called before a replication to remove 
	 *  any item types that need to be removed from the previous replication
	 * 
	 */
	protected void removeProliferatedItemTypes(){
//		JSL.out.println("-----------------------------------------------------");
//		JSL.out.println(getName() + " In ItemCategoryAbstract removeProliferatedItemTypes()");
//		JSL.out.println("^^^^^^^^^^^^^^^  # of items prior to removal, myInitialItemTypes : " + myInitialItemTypes.size());
//		JSL.out.println("^^^^^^^^^^^^^^^  # of items prior to removal: myItemTypes: " + myItemTypes.size());
//		JSL.out.println("^^^^^^^^^^^^^^^  # of items prior to removal: myProliferatedItemTypes: " + myProliferatedItemTypes.size());		
		
//		JSL.out.println("All item types prior to removal of proliferated item types: ");
//		JSL.out.println(myItemTypes);
		Iterator<CItemType> i = myProliferatedItemTypes.listIterator();
		while(i.hasNext()){
			CItemType itemType= i.next();
			itemType.removeFromModel();
			i.remove();
			myItemTypes.remove(itemType);
			myLastItemType = itemType;
			notifyItemCategoryListeners(ITEM_TYPE_REMOVED);
		}
//		JSL.out.println("^^^^^^^^^^^^^^^  # of items after removal, myInitialItemTypes : " + myInitialItemTypes.size());
//		JSL.out.println("^^^^^^^^^^^^^^^  # of items after removal: myItemTypes: " + myItemTypes.size());
//		JSL.out.println("^^^^^^^^^^^^^^^  # of items after removal: myProliferatedItemTypes: " + myProliferatedItemTypes.size());		
//		JSL.out.println("All item types after removal of proliferated item types: ");
//		JSL.out.println(myItemTypes);
//		JSL.out.println("-----------------------------------------------------");
	}

	protected String determineItemTypeName(){
		int id = this.myItemTypes.size() + 1;
//		JSL.out.println("In ItemCategoryAbstract determineItemTypeName()");
//		JSL.out.println(getName() + "-T" + id);
		return (getName()+"-T"+id);
	}

	protected void beforeReplication() { //TODO this maybe should be in afterReplication()
//		JSL.out.println("Before replication = " + this.getCurrentReplication().getReplicationNumber());
		removeProliferatedItemTypes();
	}
	
	protected void initialize(){
		//TODOD
	}

}
