/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.supplychain.proliferationYC;


import java.util.*;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.queue.QueueDiscipline;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.reporting.JSL;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.DemandSenderIfc;
import jslx.supplychain.inventorylayer.Inventory;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier;
import jslx.supplychain.transportlayer.DemandLoadFormingRuleIfc;
import jslx.supplychain.transportlayer.DemandLoadBuilder.LoadFormingOption;

/**
 * @author rossetti
 *
 */
public class Proliferation1WNRLoadBuildingRQNetwork extends ModelElement {
	
	/** Listens for category changes
	 * 
	 */
	protected ItemCategoryListener myItemCategoryListener = new ItemCategoryListener();
 
	/** Used to listen for acceptance proliferation messages
	 * 
	 */
	protected AcceptanceListener myAcceptanceListener = new AcceptanceListener();

	/** Used to listen for acceptance proliferation messages
	 * 
	 */
	protected AdoptionListener myAdoptionListener = new AdoptionListener();

	/** the 1 warehouse, n retailer load building network
	 * 
	 */
	protected OneWarehouseNRetailersWithLoadBuildingCarrier myNetwork;
	
	/** Used to hold category location proliferators
	 * 
	 */
	protected List<CategoryLocationProliferator> myCategoryLocationProliferators;
	
	/** Used to hold category, retail specific information
	 * 
	 */
	protected Map<AdoptableItemCategory, Map<InventoryHoldingPoint, CatLocInfo>> myCatLocInfo;
	
	/** The list of adoption proliferators added during the replication
	 * 
	 */
	protected List<AdoptionProliferator> myAdoptionProliferators;
	
	/** The categories registered with the network
	 * 
	 */
	protected Set<AdoptableItemCategory> myCategories;
	
	protected Map<ItemCategoryAbstract, ResponseVariableAverageObserver> myWarehouseCategoryAvgFirstFillRates;
	protected Map<ItemCategoryAbstract, AggregateTimeWeightedVariable> myWarehouseCategoryTotalOnHand;
	protected Map<ItemCategoryAbstract, AggregateTimeWeightedVariable> myWarehouseCategoryTotalOnOrder;
	protected Map<ItemCategoryAbstract, AggregateTimeWeightedVariable> myWarehouseCategoryTotalStockOutIndicator;
	protected Map<ItemCategoryAbstract, AggregateTimeWeightedVariable> myWarehouseCategoryTotalAmtBackOrdered; 
	
	/**
	 * @param parent
	 */
	public Proliferation1WNRLoadBuildingRQNetwork(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public Proliferation1WNRLoadBuildingRQNetwork(ModelElement parent, String name) {
		super(parent, name);
		myNetwork = new OneWarehouseNRetailersWithLoadBuildingCarrier(this);
		myCategories = new HashSet<AdoptableItemCategory>();
		myCategoryLocationProliferators = new ArrayList<CategoryLocationProliferator>();
		myCatLocInfo = new HashMap<AdoptableItemCategory, Map<InventoryHoldingPoint, CatLocInfo>>();
		myAdoptionProliferators = new ArrayList<AdoptionProliferator>();
		
		myWarehouseCategoryAvgFirstFillRates = new HashMap<ItemCategoryAbstract, ResponseVariableAverageObserver>();
		myWarehouseCategoryTotalOnHand = new HashMap<ItemCategoryAbstract, AggregateTimeWeightedVariable>();
		myWarehouseCategoryTotalOnOrder = new HashMap<ItemCategoryAbstract, AggregateTimeWeightedVariable>();
		myWarehouseCategoryTotalStockOutIndicator = new HashMap<ItemCategoryAbstract, AggregateTimeWeightedVariable>();
		myWarehouseCategoryTotalAmtBackOrdered = new HashMap<ItemCategoryAbstract, AggregateTimeWeightedVariable>();
	}
	
	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#getReactToLoadBuildersFlag()
	 */
	public boolean getReactToLoadBuildersFlag() {
		return myNetwork.getReactToLoadBuildersFlag();
	}

	/**
	 * @param flag
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#setReactToLoadBuildersFlag(boolean)
	 */
	public void setReactToLoadBuildersFlag(boolean flag) {
		myNetwork.setReactToLoadBuildersFlag(flag);
	}

	/**
	 * @param itemType
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#addItemType(jslx.supplychain.inventorylayer.ItemType)
	 */
	public void addInitialItemType(CItemType itemType) {
		if (!containsItemCategory(itemType.getCategory()))
			throw new IllegalArgumentException("The item type is not part of a category registered with the network");
		
		myNetwork.addItemType(itemType);
	}
	
	/** Adds the item types in the category as initial item types in the network
	 * 
	 * @param category
	 */
	public void addInitialItemTypesFromCategory(AdoptableItemCategory category){
		List<CItemType> itemTypes = category.getInitialItemTypes();
		for(CItemType itemType: itemTypes)
			addInitialItemType(itemType);
	}
	
	/** Adds the item types of each category as initial item types in the network
	 * 
	 * @param categories
	 */
	public void addInitialItemTypesFromCategories(Collection<AdoptableItemCategory> categories){
		for(AdoptableItemCategory c: categories){
			addInitialItemTypesFromCategory(c);
		}
	}
			
	/** Returns a new list representing the inventory at the supplied
	 *  ihp that are within this category
	 * 
	 * @param ihp
	 * @return
	 * 
	 * @param category
	 * @param ihp
	 * @return
	 */
	protected List<Inventory> getInventoryForCategoryAtIHP(ItemCategoryAbstract category, InventoryHoldingPoint ihp){
		if (category == null)
			throw new IllegalArgumentException("The supplied item category was null");
		if (ihp == null)
			throw new IllegalArgumentException("The supplied IHP was null");
		return category.getInventory(ihp);
	}

	/** Adds an (r, Q) inventory to the warehouse for the item type
	 * 
	 * @param itemType
	 * @param r
	 * @param q
	 * @param initialLevel
	 */
	public void addItemTypeToWarehouse(CItemType itemType, int r, int q, int initialLevel){
		if (!containsItemCategory(itemType.getCategory()))
			throw new IllegalArgumentException("The item type is not part of a category registered with the network");
		
		InventoryHoldingPoint w = myNetwork.getWarehouse();
		Inventory inv;
		inv = w.addReorderPointReorderQuantityInventory(itemType, r, q, initialLevel);
		
		ResponseVariableAverageObserver warehouseFillRate = myWarehouseCategoryAvgFirstFillRates.get(itemType.getCategory());
		if(warehouseFillRate == null){
			warehouseFillRate = new ResponseVariableAverageObserver(this, "WarehouseCat "+itemType.getCategory().getName() + " Aggregate Fill Rate");
			myWarehouseCategoryAvgFirstFillRates.put(itemType.getCategory(), warehouseFillRate);
		}
		inv.attachResponseVariableAverageObserverToFirstFillRate(warehouseFillRate);
		
		AggregateTimeWeightedVariable warehouseTotalOnHand = myWarehouseCategoryTotalOnHand.get(itemType.getCategory());
		if(warehouseTotalOnHand == null){
			warehouseTotalOnHand = new AggregateTimeWeightedVariable(this, "WarehouseCat "+itemType.getCategory().getName() + " Total On Hand");
			myWarehouseCategoryTotalOnHand.put(itemType.getCategory(), warehouseTotalOnHand);
		}
		inv.attachAggregateTimeWeightedVariableToOnHand(warehouseTotalOnHand);
		
		AggregateTimeWeightedVariable warehouseTotalOnOrder = myWarehouseCategoryTotalOnOrder.get(itemType.getCategory());
		if(warehouseTotalOnOrder == null){
			warehouseTotalOnOrder = new AggregateTimeWeightedVariable(this, "WarehouseCat "+itemType.getCategory().getName() + " Total On Order");
			myWarehouseCategoryTotalOnOrder.put(itemType.getCategory(), warehouseTotalOnOrder);
		}
		inv.attachAggregateTimeWeightedVariableToOnOrder(warehouseTotalOnOrder);
		
		AggregateTimeWeightedVariable warehouseTotalStockOutIndicator = myWarehouseCategoryTotalStockOutIndicator.get(itemType.getCategory());
		if(warehouseTotalStockOutIndicator == null){
			warehouseTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, "WarehouseCat "+itemType.getCategory().getName() + " Total StockOutIndicator");
			myWarehouseCategoryTotalStockOutIndicator.put(itemType.getCategory(), warehouseTotalStockOutIndicator);
		}
		inv.attachAggregateTimeWeightedVariableToStockOutIndicator(warehouseTotalStockOutIndicator);
		
		AggregateTimeWeightedVariable warehouseTotalAmtBackOrdered = myWarehouseCategoryTotalAmtBackOrdered.get(itemType.getCategory());
		if(warehouseTotalAmtBackOrdered == null){
			warehouseTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, "WarehouseCat "+itemType.getCategory().getName() + " Total AmtBackOrdered");
			myWarehouseCategoryTotalAmtBackOrdered.put(itemType.getCategory(), warehouseTotalAmtBackOrdered);
		}
		inv.attachAggregateTimeWeightedVariableToAmountBackLogged(warehouseTotalAmtBackOrdered);
	}
	
	/** Sets up an (r, Q) inventory at the retailer with a Poisson process 
	 * 
	 * @param itemType
	 * @param retailer
	 * @param r
	 * @param q
	 * @param initialLevel
	 * @param mtbd
	 */
	public void addInitialItemTypeToRetailer(CItemType itemType, InventoryHoldingPoint retailer, int r, int q, int initialLevel, double mtbd){
		if (!containsItemCategory(itemType.getCategory()))
			throw new IllegalArgumentException("The item type is not part of a category registered with the network");
	
		if (!myNetwork.containsRetailer(retailer))
			throw new IllegalArgumentException("The retailer is not part of the network");
		
		//create and add the inventory for the item type
		Inventory inv;
		inv = retailer.addReorderPointReorderQuantityInventory(itemType, r, q, initialLevel);
		
		// need to set up the category, location information
		AdoptableItemCategory category = (AdoptableItemCategory) itemType.getCategory();
		
		// see if category has been added already
		Map<InventoryHoldingPoint,CatLocInfo> m = myCatLocInfo.get(category);
		if (m == null){
			// category has not been added yet, create the map for the retailer
			m = new HashMap<InventoryHoldingPoint, CatLocInfo>();
			myCatLocInfo.put(category, m);	
		}
		// check if retailer has been added
		CatLocInfo cl = m.get(retailer);
		if (cl == null){
			// retailer has not already been added, need to create CatLocInfo
			// create the CatLocInfo for the category, retailer combination
			cl = new CatLocInfo();
			cl.myCategory = category;
			cl.myLocation = retailer;
			
			cl.myTotalOnHand = new AggregateTimeWeightedVariable(this, "CatLoc "+cl.myCategory.getName() + cl.myLocation.getName() + " Total On Hand");
			cl.myAvgFirstFillRate = new ResponseVariableAverageObserver(this, "CatLoc "+cl.myCategory.getName() + cl.myLocation.getName() + " Aggregate Fill Rate");
			cl.myTotalStockOutIndicator = new AggregateTimeWeightedVariable(this, "CatLoc "+cl.myCategory.getName() + cl.myLocation.getName() + " Aggregate Number of Stockouts");
			cl.myTotalOnOrder = new AggregateTimeWeightedVariable(this, "CatLoc "+cl.myCategory.getName() + cl.myLocation.getName() + " Total On Order");
			cl.myTotalAmtBackOrdered = new AggregateTimeWeightedVariable(this, "CatLoc "+cl.myCategory.getName() + cl.myLocation.getName() + " Aggregate Amt BackOrdered");
			
			m.put(retailer, cl);				
		}
		
		//use int r, int q, double mtbd to update CatLocInfo
		cl.updateCatLocInfo(r, q, 1.0/mtbd);
		
		inv.attachAggregateTimeWeightedVariableToAmountBackLogged(cl.myTotalAmtBackOrdered);
		inv.attachAggregateTimeWeightedVariableToOnHand(cl.myTotalOnHand);
		inv.attachAggregateTimeWeightedVariableToOnOrder(cl.myTotalOnOrder);
		inv.attachAggregateTimeWeightedVariableToStockOutIndicator(cl.myTotalStockOutIndicator);
		inv.attachResponseVariableAverageObserverToFirstFillRate(cl.myAvgFirstFillRate);
		
		// create the initial demand generator for the item type at the retailer
		DistributionIfc timeUntilNext = new Exponential(mtbd);
		String name = "DG-"+retailer.getName()+"-"+itemType.getName();
		DemandGenerator dg = myNetwork.attachDemandGeneratorToRetailer(retailer, itemType, timeUntilNext, name);
		//dg.setMaximumNumberOfEventsForReplications(2);//TODO only for testing purposes
		
		//need to remember the demand generators for the initial item types
		cl.myInitialDGs.add(dg);
		cl.myDGs.add(dg);
		
	}

	public void setUpRetailerCategoryProliferation(AdoptableItemCategory category, InventoryHoldingPoint retailer, 
			double mtbProliferations) {
		
		String name = "CLP-"+category.getName()+"-"+retailer.getName();
		setUpRetailerCategoryProliferation(category, retailer, mtbProliferations, name);
	}

	
	/**
	 * 
	 * @param category The category to proliferate at the retailer
	 * @param retailer The location of the proliferation
	 * @param mtbProliferations The mean time between proliferation events for this category at the retailer
	 */
	public void setUpRetailerCategoryProliferation(AdoptableItemCategory category, InventoryHoldingPoint retailer, 
			double mtbProliferations, String name){
		
		CatLocInfo cl = getCategoryLocationInfo(category, retailer);
		if (cl == null)
			throw new IllegalArgumentException("No item types for the category have been added to the retailer for the network, see addInitialItemTypeToRetailer()");
		
		if (cl.myProliferator != null)
			throw new IllegalArgumentException("The category, retailer combination has already been setup");
		
		Exponential e = new Exponential(mtbProliferations);
		CategoryLocationProliferator p = new CategoryLocationProliferator(this, name);
		p.setCategoryLocation(category, retailer);
		cl.myProliferator = p;// indicate that it is set up
		
		// proliferation DOES NOT start on initialization, it is supposed to start after warm up
		p.setStartOnInitializeFlag(false);
		// need to set both because of the false for start on initialize
		p.setTimeBetweenEventsForReplications(e);
		p.setTimeBetweenEvents(e);
//		JSL.out.println("In Proliferation1WNRLoadBuildingRQNetwork setUpRetailerCategoryProliferation()");
//		JSL.out.println("setMaximumNumberOfEventsForReplications(2)");
//		p.setMaximumNumberOfEventsForReplications(2);//TODO only for testing purposes
	
		// list can be used to start proliferators
		myCategoryLocationProliferators.add(p);
		// listen for proliferation events
		p.attachProliferationListener(myAcceptanceListener);

	}
	
	/** Returns the category location information for the provided category, retailer combination
	 *  May return null if no information is available for the provided combination
	 * @param category
	 * @param retailer
	 * @return
	 */
	protected CatLocInfo getCategoryLocationInfo(ItemCategoryAbstract category, InventoryHoldingPoint retailer){
		Map<InventoryHoldingPoint,CatLocInfo> m = myCatLocInfo.get(category);
		if (m != null){
			return m.get(retailer);
		} 
		
		return null;
	}
		
	/** Registers the category with the network
	 * 
	 * @param category
	 *
	 */
	public void registerItemCategory(AdoptableItemCategory category){
		if (category == null)
			throw new IllegalArgumentException("The supplied item category was null");
		
		if (myCategories.contains(category))
			throw new IllegalArgumentException("The item category has already been registered.");
		
		myCategories.add(category);
		category.attachItemCategoryChangeListener(myItemCategoryListener);
		
	}

	/** Checks if the category has been registered
	 * 
	 * @param category
	 * @return
	 */
	public boolean containsItemCategory(ItemCategoryAbstract category){
		return myCategories.contains(category);
	}
	
	/** Returns an unmodifiable set of the registered categories
	 * 
	 * @return
	 */
	public Set<AdoptableItemCategory> getItemCategories(){
		return Collections.unmodifiableSet(myCategories);
	}

	/**
	 * @param transportTime
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#addRetailer(jsl.utilities.random.RandomIfc)
	 */
	public InventoryHoldingPoint addRetailer(RandomIfc transportTime) {
		return myNetwork.addRetailer(transportTime);
	}

	/**
	 * @param name
	 * @param transportTime
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#addRetailer(java.lang.String, jsl.utilities.random.RandomIfc)
	 */
	public InventoryHoldingPoint addRetailer(String name, RandomIfc transportTime) {
		return myNetwork.addRetailer(name, transportTime);
	}	

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#geMinCubeLimit(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public double geMinCubeLimit(DemandSenderIfc destination) {
		return myNetwork.geMinCubeLimit(destination);
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#geMinWeightLimit(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public double geMinWeightLimit(DemandSenderIfc destination) {
		return myNetwork.geMinWeightLimit(destination);
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#getCountLimit(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public int getCountLimit(DemandSenderIfc destination) {
		return myNetwork.getCountLimit(destination);
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#getDefaultLoadFormingOption(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public LoadFormingOption getDefaultLoadFormingOption(
			DemandSenderIfc destination) {
		return myNetwork.getDefaultLoadFormingOption(destination);
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#getMaxCubeLimit(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public double getMaxCubeLimit(DemandSenderIfc destination) {
		return myNetwork.getMaxCubeLimit(destination);
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#getMaxWeightLimit(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public double getMaxWeightLimit(DemandSenderIfc destination) {
		return myNetwork.getMaxWeightLimit(destination);
	}

	/**
	 * @param destination
	 * @param countLimit
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#setCountLimit(jslx.supplychain.inventorylayer.DemandSenderIfc, int)
	 */
	public void setCountLimit(DemandSenderIfc destination, int countLimit) {
		myNetwork.setCountLimit(destination, countLimit);
	}

	/**
	 * @param destination
	 * @param minCube
	 * @param maxCube
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#setCubeFormingLimits(jslx.supplychain.inventorylayer.DemandSenderIfc, double, double)
	 */
	public void setCubeFormingLimits(DemandSenderIfc destination,
			double minCube, double maxCube) {
		myNetwork.setCubeFormingLimits(destination, minCube, maxCube);
	}

	/**
	 * @param destination
	 * @param option
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#setDefaultLoadFormingOption(jslx.supplychain.inventorylayer.DemandSenderIfc, jslx.supplychain.transportlayer.DemandLoadBuilder.LoadFormingOption)
	 */
	public void setDefaultLoadFormingOption(DemandSenderIfc destination,
			LoadFormingOption option) {
		myNetwork.setDefaultLoadFormingOption(destination, option);
	}

	/**
	 * @param destination
	 * @param discipline
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#setDemandQueueInitialDiscipline(jslx.supplychain.inventorylayer.DemandSenderIfc, jsl.modeling.elements.queue.QueueDiscipline)
	 */
	public void setDemandQueueInitialDiscipline(DemandSenderIfc destination,
			QueueDiscipline discipline) {
		myNetwork.setDemandQueueInitialDiscipline(destination, discipline);
	}

	/**
	 * @param destination
	 * @param rule
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#setLoadFormingRule(jslx.supplychain.inventorylayer.DemandSenderIfc, jslx.supplychain.transportlayer.DemandLoadFormingRuleIfc)
	 */
	public void setLoadFormingRule(DemandSenderIfc destination,
			DemandLoadFormingRuleIfc rule) {
		myNetwork.setLoadFormingRule(destination, rule);
	}

	/**
	 * @param destination
	 * @param discipline
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#setLoadQueueInitialDiscipline(jslx.supplychain.inventorylayer.DemandSenderIfc, jsl.modeling.elements.queue.QueueDiscipline)
	 */
	public void setLoadQueueInitialDiscipline(DemandSenderIfc destination,
			QueueDiscipline discipline) {
		myNetwork.setLoadQueueInitialDiscipline(destination, discipline);
	}

	/**
	 * @param distribution
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#setTransportTimeToWarehouse(jsl.utilities.random.RandomIfc)
	 */
	public final void setTransportTimeToWarehouse(RandomIfc distribution) {
		myNetwork.setTransportTimeToWarehouse(distribution);
	}

	/**
	 * @param destination
	 * @param minWeight
	 * @param maxWeight
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#setWeightFormingLimits(jslx.supplychain.inventorylayer.DemandSenderIfc, double, double)
	 */
	public void setWeightFormingLimits(DemandSenderIfc destination,
			double minWeight, double maxWeight) {
		myNetwork.setWeightFormingLimits(destination, minWeight, maxWeight);
	}
	
	
	protected void adjustInventoryPolicyAtWarehouse(CItemType itemType, int r, int q){//TODO
		InventoryHoldingPoint w = myNetwork.getWarehouse();
		Inventory i = w.getInventory(itemType);
		double[] p = new double[2];
		p[0] = r;
		p[1] = q;
		i.setInventoryPolicyParameters(p);
	}
	
	
	/** Can be used to start the category location proliferators
	 * 
	 */
	protected void startCategoryLocationProliferators(){
		for(CategoryLocationProliferator p: myCategoryLocationProliferators){
			p.turnOnGenerator();
		}
	}
	
	protected void startInitialAdoptionProliferators(){
		for(AdoptionProliferator p: myAdoptionProliferators){
			p.turnOnGenerator();
		}
	}
	
	protected void setUpInitialAdoptionProliferators(){
		for(AdoptableItemCategory c: myCategories){
			for(CItemType t: c.getItemTypes()){
				if (!stockedAtAllRetailers(t)){
					// if it is not stocked at all retailers, then it can proliferate through adoption
					double mtba = c.getMeanTimeBetweenAdoption();
					setUpAdoptionProliferator(mtba, t);
				}
			}
		}		
	}
	
	/**
	 * @param itemType
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#stockedAtAllRetailers(jslx.supplychain.inventorylayer.ItemType)
	 */
	public boolean stockedAtAllRetailers(ItemType itemType) {
		return myNetwork.stockedAtAllRetailers(itemType);
	}

	protected AdoptionProliferator setUpAdoptionProliferator(double mtba, CItemType itemType) {

		String name = "AP-" + itemType.getName();
		AdoptionProliferator p = new AdoptionProliferator(this, mtba, itemType, name);
//		JSL.out.println("Setting up adoption proliferator: " + p.getName() + " for item type " + itemType.getName());
//		JSL.out.println("mtb proliferations: = " + mtba);

//		JSL.out.println("In Proliferation1WNRLoadBuildingRQNetwork setUpAdoptionProliferator");
//		JSL.out.println("setMaximumNumberOfEventsForReplication(2)");
//		p.setMaximumNumberOfEventsForReplications(2);//TODO only for testing purposes
		myAdoptionProliferators.add(p);
		p.attachProliferationListener(myAdoptionListener);
		p.initializeLocations(myNetwork.getRetailers());
		return p;
	}

	
	protected void warmUp(){
//		JSL.out.println("###################################################");
//		JSL.out.println("In Proliferation1WNRLoadBuildingRQNNetwork warmUp()");
//		JSL.out.println("Setting up and starting category location proliferators");
		// turn on category location proliferators
		startCategoryLocationProliferators();
		// set up adoption proliferators
//		JSL.out.println("Setting up initial adoption proliferators");
		setUpInitialAdoptionProliferators();
		// turn on adoption proliferators
//		JSL.out.println("Starting initial adoption proliferators");
		startInitialAdoptionProliferators();
//		JSL.out.println("###################################################");
	}
	
	protected void afterReplication(){
		
		// remove any proliferated demand generators, from network and model
		List<InventoryHoldingPoint> retailers = myNetwork.getRetailers();
		for(ItemCategoryAbstract c: myCategories){
			for(InventoryHoldingPoint r: retailers){
				CatLocInfo cl = getCategoryLocationInfo(c, r);
				if (cl != null){
					cl.removeProliferatedDemandGenerators();
					cl.resetDemandGenerators();
				}
			}
		}

		// remove adoption proliferators added during replication
		for(AdoptionProliferator p: myAdoptionProliferators){
			p.removeFromModel();
		}
		myAdoptionProliferators.clear();
		
	}
	
	protected class CatLocInfo {
		
		/** Holds the initial demand generators for the category at the location
		 * 
		 */
		protected List<DemandGenerator> myInitialDGs = new ArrayList<DemandGenerator>();

		/** Holds all demand generators for the category at the location
		 * 
		 */
		protected List<DemandGenerator> myDGs = new ArrayList<DemandGenerator>();

		/** Holds proliferated demand generators for the category at the location
		 * 
		 */
		protected List<DemandGenerator> myProliferatedDGs = new ArrayList<DemandGenerator>();
		
		/** The associated item category
		 * 
		 */
		protected AdoptableItemCategory myCategory;
		
		/** The associated inventory holding point
		 * 
		 */
		protected InventoryHoldingPoint myLocation;
		
		/** The proliferator for this category, location combination
		 * 
		 */
		protected CategoryLocationProliferator myProliferator;
				
		/** 
		 *  This is the overall demand for the category. The mean
		 *  total demand for all the item types in the category
		 * 
		 */
		protected double mySumD;
		
		/** The overall reorder point for the category at
		 *  the particular location
		 * 
		 */
		protected int mySumR;
		
		/** The overall reorder quantity for the category
		 *  at the particular location
		 * 
		 */
		protected int mySumQ;
		
		/** Counts the number of updates
		 * 
		 */
		protected int myCnt;
		
		protected ResponseVariableAverageObserver myAvgFirstFillRate;
		protected AggregateTimeWeightedVariable myTotalOnHand;
		protected AggregateTimeWeightedVariable myTotalOnOrder;
		protected AggregateTimeWeightedVariable myTotalStockOutIndicator;
		protected AggregateTimeWeightedVariable myTotalAmtBackOrdered;

		protected void resetDemandGenerators(){
			myDGs.clear();
			myProliferatedDGs.clear();
			myDGs.addAll(myInitialDGs);
		}
		
		protected void removeProliferatedDemandGenerators(){
			for(DemandGenerator dg: myProliferatedDGs){
				myNetwork.detachDemandSenderFromRetailer(myLocation, dg);
				dg.removeFromModel();
			}
		}

		/** Updates the totals for this category, location combination
		 * 
		 * @param r
		 * @param q
		 * @param d
		 */
		protected void updateCatLocInfo(int r, int q, double d){
			mySumR = mySumR + r;
			mySumQ = mySumQ + q;
			mySumD = mySumD + d;
			myCnt++;
		}
		
		protected void adjustRetailerRQParameters(){
			// this method should adjust the r, Q values for all the inventory
			// at the retailer within the provided category
			List<Inventory> iList = getInventoryForCategoryAtIHP(myCategory, myLocation);
			if (iList.isEmpty())
				return;
			
			double n = iList.size(); // number of items in the category
			// all items get the same allocation
			double[] p = new double[2];
			p[0] = Math.ceil(mySumR/n);
			p[1] = Math.ceil(mySumQ/n);
//			JSL.out.println("In Proliferation1WNRLoadBuildingRQNetwork adjustRetailerRQParameters()");
//			JSL.out.println("Adjusting retailer r, Q parameters");
//			JSL.out.println("n = " + n + "\t r = " + p[0] + "\t Q = " + p[1]);
			for(Inventory i: iList){
				i.setInventoryPolicyParameters(p);
			}
		}
		
		protected void adjustRetailerDemandGenerators(){
			// this method should adjust the demand rates for
			// each demand generator associated with the category at the particular retailer
			if (myDGs.isEmpty())	
				return;
			
			int n = myDGs.size(); // the number of demand generators
			double[] mtbd = new double[1];
			mtbd[0] = n/mySumD;
			for(DemandGenerator dg: myDGs){
				RandomIfc r = dg.getTimeBetweenEvents();
				r.setParameters(mtbd);
			}
		}
		
		public String toString(){
			StringBuilder sb = new StringBuilder();
			sb.append("CatLocInfor \n");
			sb.append("category = " + myCategory.getName() + " \n");
			sb.append("location = " + myLocation.getName() + " \n");
			sb.append("sum of demand = " + mySumD + " \n");
			sb.append("sum of r = " + mySumR + " \n");
			sb.append("sum of q = " + mySumQ + " \n");
			sb.append("count = " + myCnt + " \n");
			return sb.toString();
		}

	}

	
	protected class ItemCategoryListener implements ItemCategoryListenerIfc {
		public void itemCategoryChanged(ItemCategoryAbstract category, int status) {
			if (status == ItemCategoryAbstract.ITEM_TYPE_REMOVED){
				// remove the item type from the network
				CItemType itemType = category.getLastItemType();
				myNetwork.removeItemTypeFromNetwork(itemType);				
			}
		}
	}
	
	protected class AcceptanceListener implements ProliferationListenerIfc {
		
		public void handleProliferation(ProliferatorAbstract proliferator, ProliferationMessage m) {
			// The message indicates that an item type in a particular category was proliferated at a particular retailer
			
			// add the item type to the network
			myNetwork.addItemType(m.itemType);
			
			// add the item type to the warehouse, need to determine parameters
			proliferateItemTypeAtWarehouse(m.itemType);//TODO is this okay *before* adding to retailer?

			// look up the category, location information
			CatLocInfo cl = getCategoryLocationInfo(m.category, m.ihp);
			
			//create and add the inventory for the item type, parameters don't matter here
			Inventory inv;
			inv = m.ihp.addReorderPointReorderQuantityInventory(m.itemType, 0, 1, 1);//TODO how to set initial level
			
			inv.attachAggregateTimeWeightedVariableToAmountBackLogged(cl.myTotalAmtBackOrdered);
			inv.attachAggregateTimeWeightedVariableToOnHand(cl.myTotalOnHand);
			inv.attachAggregateTimeWeightedVariableToOnOrder(cl.myTotalOnOrder);
			inv.attachAggregateTimeWeightedVariableToStockOutIndicator(cl.myTotalStockOutIndicator);
			inv.attachResponseVariableAverageObserverToFirstFillRate(cl.myAvgFirstFillRate);
			
			// create the demand generator for the item type at the retailer
			DistributionIfc timeUntilNext = new Exponential();// mtbd doesn't matter here
			DemandGenerator dg = myNetwork.attachDemandGeneratorToRetailer(m.ihp, m.itemType, timeUntilNext);
//			dg.setMaximumNumberOfEventsForReplications(2);//TODO only for testing purposes
			
			// remember the demand generator
			cl.myDGs.add(dg);
			cl.myProliferatedDGs.add(dg);
			
			// adjust the inventory policies
			cl.adjustRetailerRQParameters();
			
			// adjust the demand generators
			cl.adjustRetailerDemandGenerators();

			//TODO  set up the adoption proliferator
			AdoptionProliferator p = setUpAdoptionProliferator(cl.myCategory.getMeanTimeBetweenAdoption(), m.itemType);
			p.turnOnGenerator();
			
		}
		
		protected void proliferateItemTypeAtWarehouse(CItemType itemType){
			ItemCategoryAbstract category = itemType.getCategory();
			// get category location information for each location
			Map<InventoryHoldingPoint,CatLocInfo> m = myCatLocInfo.get(category);
			
			int rtot = 0;
			int qtot = 0;
			for(InventoryHoldingPoint retailer: m.keySet()){
				CatLocInfo cl = m.get(retailer);
				rtot = rtot + cl.mySumR;
				qtot = qtot + cl.mySumQ;
			}
			double n = category.getNumberOfItemTypes();
			int rw = (int)Math.ceil(rtot/n);
			int qw = (int)Math.ceil(qtot/n);
//			JSL.out.println("In proliferation1WNRLoadBuildingRQNetwork proliferateItemTypeAtWarehouse()");
//			JSL.out.println("Proliferating " + itemType.getName() + " at warehouse with rw = " + rw + " qw = " + qw);
			addItemTypeToWarehouse(itemType, rw, qw, 1);//TODO how to set initial level
		}

	}

	protected class AdoptionListener implements ProliferationListenerIfc {
		public void handleProliferation(ProliferatorAbstract proliferator, ProliferationMessage m) {
			// The message indicates that an item type in a particular category was proliferated at a particular retailer
			
			//create and add the inventory for the item type, parameters don't matter here
			Inventory inv;
			inv = m.ihp.addReorderPointReorderQuantityInventory(m.itemType, 0, 1, 1);//TODO how to set initial level
			
			// create the demand generator for the item type at the retailer
			DistributionIfc timeUntilNext = new Exponential();// mtbd doesn't matter here
			DemandGenerator dg = myNetwork.attachDemandGeneratorToRetailer(m.ihp, m.itemType, timeUntilNext);
//			dg.setMaximumNumberOfEventsForReplications(2);//TODO only for testing purposes

			// look up the category, location information
			CatLocInfo cl = getCategoryLocationInfo(m.category, m.ihp);
			
			if (cl == null) {
				String s = "An adoption event was attempted at a location for which no category exists\n";
				s = s + "The initial network state must have at least one item type from every category at every location.\n";
				throw new IllegalStateException(s);
			}

			// remember the demand generator
			cl.myDGs.add(dg);
			cl.myProliferatedDGs.add(dg);
			
			// adjust the inventory policies
			cl.adjustRetailerRQParameters();
			
			// adjust the demand generators
			cl.adjustRetailerDemandGenerators();
			
			inv.attachResponseVariableAverageObserverToFirstFillRate(cl.myAvgFirstFillRate);
			inv.attachAggregateTimeWeightedVariableToOnHand(cl.myTotalOnHand);
			inv.attachAggregateTimeWeightedVariableToOnOrder(cl.myTotalOnOrder);
			inv.attachAggregateTimeWeightedVariableToStockOutIndicator(cl.myTotalStockOutIndicator);
			inv.attachAggregateTimeWeightedVariableToAmountBackLogged(cl.myTotalAmtBackOrdered);
			
			// no need to add to warehouse 
			
			//TODO Do we need to adjust inventory parameters at the warehouse for the item type? No, since CatLocInfo didn't change??
		
		}

	}

	/**
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#getShipmentCounter()
	 */
	public final AggregateCounter getShipmentCounter() {
		return myNetwork.getShipmentCounter();
	}

	/**
	 * @param destination
	 * @return
	 * @see jslx.supplychain.multiechelon.OneWarehouseNRetailersWithLoadBuildingCarrier#getShipmentCounter(jslx.supplychain.inventorylayer.DemandSenderIfc)
	 */
	public Counter getShipmentCounter(DemandSenderIfc destination) {
		return myNetwork.getShipmentCounter(destination);
	}


}
