/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.proliferationYC;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import jsl.modeling.Experiment;
import jsl.modeling.Model;
import jsl.observers.textfile.*;

import jsl.utilities.random.*;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.Triangular;
import jslx.supplychain.inventorylayer.*;
import jslx.supplychain.transportlayer.DemandLoadBuilder.LoadFormingOption;

/**
 * @author rossetti
 *
 */
public class ThesisTest1 {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		System.out.println("Starting experiment.");
		Model m = buildModel();

		try{
//        	CSVReport rcsv = new CSVReplicationReport("Rep-TestProliferationNetworkYC1");
        	CSVReport bcsv = new CSVExperimentReport_YC("Experiment-ThesisTest1");
        	CSVReport bcsvf = new CSVExperimentReport("ExperimentFull-ThesisTest1");
//        	m.addObserver(rcsv);
        	m.addObserver(bcsv);
        	m.addObserver(bcsvf);
        } catch(IOException e){
			System.out.println("I/O exception opening file Output.txt");
			e.printStackTrace();
        }

		runExperiment(m);
		
		System.out.println("Done!");
	}
	
	public static Model buildModel(){
		
		Model m = Model.createModel();
		
		//STEP 1: Create the network structure
        //-------------------------------------
		Proliferation1WNRLoadBuildingRQNetwork n = new Proliferation1WNRLoadBuildingRQNetwork(m,"Network");
		
		n.setTransportTimeToWarehouse(new Constant(1));
		
		InventoryHoldingPoint r1 = n.addRetailer("AR", new Constant(3.3/24));
		InventoryHoldingPoint r2 = n.addRetailer("MO", new Constant(3.4/24));
		InventoryHoldingPoint r3 = n.addRetailer("OK", new Constant(4.5/24));
		InventoryHoldingPoint r4 = n.addRetailer("KS", new Constant(2.0/24));
		
		//Step 2: Set up the load building options
        //-------------------------------------
		//TODO use whatever load forming criteria you want here
		n.setDefaultLoadFormingOption(r1, LoadFormingOption.CUBE);
		n.setDefaultLoadFormingOption(r2, LoadFormingOption.CUBE);
		n.setDefaultLoadFormingOption(r3, LoadFormingOption.CUBE);
		n.setDefaultLoadFormingOption(r4, LoadFormingOption.CUBE);
			
		n.setCubeFormingLimits(r1, 20, 50);
		n.setCubeFormingLimits(r2, 20, 50);
		n.setCubeFormingLimits(r3, 20, 50);
		n.setCubeFormingLimits(r4, 20, 50);
		
		//Step 3: Create the basic categories (only makes them)
        //-------------------------------------
		
		List<AdoptableItemCategory> categories = new ArrayList<AdoptableItemCategory>();
		
		AdoptableItemCategory c1 = new AdoptableItemCategory(m, "C1");
		c1.setCubeCDFInitialRandomSource(new Triangular(1,3,6));
		c1.setLeadTime(new Triangular(7,11,15));
		c1.setMeanTimeBetweenAdoption(1000.0);
		categories.add(c1);
		
		AdoptableItemCategory c2 = new AdoptableItemCategory(m, "C2");
		c2.setCubeCDFInitialRandomSource(new Triangular(6,9,12));
		c2.setLeadTime(new Triangular(7,11,15));
		c2.setMeanTimeBetweenAdoption(1000.0);
		categories.add(c2);
		
		AdoptableItemCategory c3 = new AdoptableItemCategory(m, "C3");
		c3.setCubeCDFInitialRandomSource(new Triangular(1,3,6));
		c3.setLeadTime(new Triangular(5,7,9));
		c3.setMeanTimeBetweenAdoption(1000.0);
		categories.add(c3);
		
		AdoptableItemCategory c4 = new AdoptableItemCategory(m, "C4");
		c4.setCubeCDFInitialRandomSource(new Triangular(6,9,12));
		c4.setLeadTime(new Triangular(5,7,9));
		c4.setMeanTimeBetweenAdoption(1000.0);
		categories.add(c4);
		
		AdoptableItemCategory c5 = new AdoptableItemCategory(m, "C5");
		c5.setCubeCDFInitialRandomSource(new Triangular(1,3,6));
		c5.setLeadTime(new Triangular(2,4,6));
		c5.setMeanTimeBetweenAdoption(1000.0);
		categories.add(c5);
		
		AdoptableItemCategory c6 = new AdoptableItemCategory(m, "C6");
		c6.setCubeCDFInitialRandomSource(new Triangular(6,9,12));
		c6.setLeadTime(new Triangular(2,4,6));
		c6.setMeanTimeBetweenAdoption(1000.0);
		categories.add(c6);
		
		//TODO create other categories here

		//Step 4: Register the categories on the network
        //-------------------------------------
		for(AdoptableItemCategory c: categories)
			n.registerItemCategory(c);

		//Step 5: Setup the categories with initial item types
        //-------------------------------------
		// this only creates and adds the item type to the category
		// it does not notify category listeners
		CItemType t1c1 = c1.addInitialItemType();
		CItemType t1c2 = c2.addInitialItemType();
		CItemType t1c3 = c3.addInitialItemType();
		CItemType t1c4 = c4.addInitialItemType();
		CItemType t1c5 = c5.addInitialItemType();
		CItemType t1c6 = c6.addInitialItemType();
		

		//TODO create other initial item types in each category here

		//Step 6: Setup the initial item types in the network
        //-------------------------------------
		// set lead time for item type for external supplier in network, based on category
		n.addInitialItemTypesFromCategories(categories);
		
		//Step 7: Setup the initial item types at the warehouse
        //-------------------------------------
		n.addItemTypeToWarehouse(t1c1, 8, 16, 24);
		n.addItemTypeToWarehouse(t1c2, 8, 16, 24);
		n.addItemTypeToWarehouse(t1c3, 40, 80, 120);
		n.addItemTypeToWarehouse(t1c4, 40, 80, 120);
		n.addItemTypeToWarehouse(t1c5, 120, 180, 300);
		n.addItemTypeToWarehouse(t1c6, 120, 180, 300);

		//TODO setup other item types at warehouse here
		
		//Step 8: Setup the initial item types at each retailer and add initial demand generators
        //-------------------------------------
		n.addInitialItemTypeToRetailer(t1c1, r1, 2, 4, 6, 0.1);
		n.addInitialItemTypeToRetailer(t1c1, r2, 2, 4, 6, 0.1);
		n.addInitialItemTypeToRetailer(t1c1, r3, 2, 4, 6, 0.1);
		n.addInitialItemTypeToRetailer(t1c1, r4, 2, 4, 6, 0.1);
		
		n.addInitialItemTypeToRetailer(t1c2, r1, 2, 4, 6, 0.1);
		n.addInitialItemTypeToRetailer(t1c2, r2, 2, 4, 6, 0.1);
		n.addInitialItemTypeToRetailer(t1c2, r3, 2, 4, 6, 0.1);
		n.addInitialItemTypeToRetailer(t1c2, r4, 2, 4, 6, 0.1);
		
		n.addInitialItemTypeToRetailer(t1c3, r1, 10, 20, 30, 1);
		n.addInitialItemTypeToRetailer(t1c3, r2, 10, 20, 30, 1);
		n.addInitialItemTypeToRetailer(t1c3, r3, 10, 20, 30, 1);
		n.addInitialItemTypeToRetailer(t1c3, r4, 10, 20, 30, 1);
		
		n.addInitialItemTypeToRetailer(t1c4, r1, 10, 20, 30, 1);
		n.addInitialItemTypeToRetailer(t1c4, r2, 10, 20, 30, 1);
		n.addInitialItemTypeToRetailer(t1c4, r3, 10, 20, 30, 1);
		n.addInitialItemTypeToRetailer(t1c4, r4, 10, 20, 30, 1);
		
		n.addInitialItemTypeToRetailer(t1c5, r1, 30, 60, 90, 10);
		n.addInitialItemTypeToRetailer(t1c5, r2, 30, 60, 90, 10);
		n.addInitialItemTypeToRetailer(t1c5, r3, 30, 60, 90, 10);
		n.addInitialItemTypeToRetailer(t1c5, r4, 30, 60, 90, 10);
		
		n.addInitialItemTypeToRetailer(t1c6, r1, 30, 60, 90, 10);
		n.addInitialItemTypeToRetailer(t1c6, r2, 30, 60, 90, 10);
		n.addInitialItemTypeToRetailer(t1c6, r3, 30, 60, 90, 10);
		n.addInitialItemTypeToRetailer(t1c6, r4, 30, 60, 90, 10);
		//TODO setup other item types at each retailer here


		//Step 9: Setup the category/location proliferation
        //-------------------------------------
		double mtbProliferations = 1000.0;
		n.setUpRetailerCategoryProliferation(c1, r1, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c1, r2, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c1, r3, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c1, r4, mtbProliferations);
		
		n.setUpRetailerCategoryProliferation(c2, r1, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c2, r2, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c2, r3, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c2, r4, mtbProliferations);
		
		n.setUpRetailerCategoryProliferation(c3, r1, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c3, r2, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c3, r3, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c3, r4, mtbProliferations);
		
		n.setUpRetailerCategoryProliferation(c4, r1, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c4, r2, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c4, r3, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c4, r4, mtbProliferations);
		
		n.setUpRetailerCategoryProliferation(c5, r1, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c5, r2, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c5, r3, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c5, r4, mtbProliferations);
		
		n.setUpRetailerCategoryProliferation(c6, r1, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c6, r2, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c6, r3, mtbProliferations);
		n.setUpRetailerCategoryProliferation(c6, r4, mtbProliferations);
		
		//TODO set up other category/location proliferation here

		// at this point the initial network should be ready
		// proliferation should be started after the warm up event
		
		categories.clear();
		m.setElementAddedAfterReplicationStartedWarningOption(false);
		
		return m;
	}

	public static void runExperiment(Model m){
		// create the experiment to run the model
		Experiment e = new Experiment(m);

		// set the parameters of the experiment
		e.setNumberOfReplications(5);
		e.setLengthOfWarmUp(50.0);
		e.setLengthOfReplication(1825.0); // 5 years

		e.turnOnReplicationConsoleOutput();
		e.turnOnExperimentReport();
//		e.turnOnReplicationReport();
		
		// tell the experiment to run
		e.runAll();
	}
}
