/*
* Created on Aug 1, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.transportlayer;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import jsl.modeling.elements.queue.QObject;

import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandSenderIfc;

/**
 * @author rossetti
 *
 */
public class DemandLoad extends QObject {

	/** The current weight of the load
	 */
	protected double myWeight;
	
	/** The current cube of the load
	 */
	protected double myCube;
	
	/** The demands on the load
	 */
	protected Set<Demand> myDemands;
	
	/** The sender of the demands on the load is the destination
	 */
	protected DemandSenderIfc myDestination;
	
	/**
	 * 
	 */
	public DemandLoad() {
		this(null);
	}

	/**
	 * @param name
	 */
	public DemandLoad(String name) {
		super(name);
		myDemands = new HashSet<Demand>();
		
	}
	
	/* (non-Javadoc)
	 * @see jsl.modeling.elements.queue.QObject#nullify()
	 */
	public void nullify(){
		super.nullify();
		myDemands.clear();
		myDemands = null;
		myDestination = null;
	}
	
	public void addDemand(Demand demand){
		if (demand == null)
			throw new IllegalArgumentException("The supplied demand was null!");

		if (myDemands.contains(demand))
			throw new IllegalArgumentException("The supplied demand is already in the load");

		DemandSenderIfc destination = demand.getSender();

		if (destination == null)
			throw new IllegalArgumentException("The sender on the demand was null!");
		
		if (myDestination == null)
			myDestination = destination;
		else {
			if (destination != myDestination)
				throw new IllegalArgumentException("The sender (destinatioin) on the demand was different than the destination for the load!");		
		}
		
		myWeight = myWeight + demand.getWeight();
		myCube = myCube + demand.getCube();
		myDemands.add(demand);
	}

	/**
	 * @return the weight
	 */
	public double getWeight() {
		return myWeight;
	}

	/**
	 * @return the cube
	 */
	public double getCube() {
		return myCube;
	}

	/**
	 * @return the destination
	 */
	public DemandSenderIfc getDestination() {
		return myDestination;
	}
	
	/** Returns an iterator to the demands on the load
	 * 
	 * @return
	 */
	public Iterator<Demand> getDemandIterator(){
		return myDemands.iterator();
	}
	
	/** Causes all of the demands on the load to be delivered
	 *  All the demands must be in the shipped state
	 */
	public void deliver(){
		for(Demand d: myDemands)
			d.deliver();
	}
	
	/** Causes all of the demands on the load to be shipped.
	 *  All the demands must be in the filled state
	 */
	public void ship(){
		for(Demand d: myDemands)
			d.ship();
	}
	
	/** Returns the demands on the load as a collection
	 * 
	 * @return
	 */
	protected Collection<Demand> getDemands(){
		return myDemands;
	}

}
