/*
* Created on Aug 1, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.transportlayer;

import java.util.Iterator;

import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.queue.QueueDiscipline;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.reporting.JSL;
import jslx.supplychain.inventorylayer.Demand;

/** This class holds demands in a queue to form loads.
 *  Loads can be formed automatically by using the COUNT, WEIGHT, or CUBE options.  
 *  The NONE option turns off automatic load forming.  The user can also supply their
 *  own instance of a DemandLoadFormingRuleIfc which will be used regardless of the
 *  load forming option. If a load is formed automatically, then the load is placed in 
 *  the load queue and a DemandLoadFormedListenerIfc will be called if supplied.  
 *  Clients can also use the methods of this class to directly form loads and to process 
 *  the load queue (e.g. via a schedule)
 *
 */
public class DemandLoadBuilder extends SchedulingElement {

	/** Specifies how a load is to be formed upon the arrival of a new demand
	 *  NONE - do not try to form a load whenever a new demand arrives, leaves demand in
	 *         the demand queue until they are removed by another means
	 *  COUNT - wait until the count limit is reached to form a load
	 *  WEIGHT - wait until the min weight is reached before forming a load < max weight
	 *  CUBE - wait until the min cube is reached before forming a load < max cube
	 *  RULE - use a user supplied rule
	 *  ALWAYS - always form an immediate load for the arriving demand.
	 */
	public static enum LoadFormingOption {NONE, COUNT, WEIGHT, CUBE, RULE, ALWAYS};
	
	/** Controls how the loads are formed by default.  The default setting is
	 *  ALWAYS
	 * 
	 */
	protected LoadFormingOption myDefaultLoadFormingOption = LoadFormingOption.ALWAYS;
	
	/** Can be used to form loads by weight
	 */
	protected double myMinWeight = 1.0;

	/** Can be used to form loads by weight
	 */
	protected double myMaxWeight = 1.0;
	
	/** Can be used to form loads by cube
	 */
	protected double myMinCube = 1.0;

	/** Can be used to form loads by cube
	 */
	protected double myMaxCube = 1.0;

	/** Can be used to form loads by count
	 */
	protected int myCountLimit = 1;
	
	/** Represents the total weight waiting to be formed
	 *  into a load
	 */
	protected TimeWeighted myTotalWeight;
	
	/** Represents the total cube waiting to be formed
	 *  into a load
	 */
	protected TimeWeighted myTotalCube;
	
	/** Holds the demands waiting to be formed into loads
	 */
	protected Queue myDemandQ;
	
	/** Holds the loads after they have been formed
	 */
	protected Queue myLoadQ;
	
	/** Should be called whenever a load is formed
	 */
	protected DemandLoadFormedListenerIfc myLoadFormedListener;
	
	/** Can be supplied to provide mechanism to form loads upon
	 *  the arrival of a new demand rather than overriding the formLoad() method
	 */
	protected DemandLoadFormingRuleIfc myLoadFormingRule;
	
	/**
	 * @param parent
	 */
	public DemandLoadBuilder(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public DemandLoadBuilder(ModelElement parent, String name) {
		super(parent, name);
		myTotalWeight = new TimeWeighted(this, getName() + "Weight");
		myTotalCube = new TimeWeighted(this, getName() + "Cube");
		myDemandQ = new Queue(this, getName() + "DemandQ");
		myLoadQ = new Queue(this, getName() + "LoadQ");
	}
	
	/** Sets the queue discipline used for waiting demands to be formed into loads
	 * 
	 * @param discipline
	 */
	public void setDemandQueueInitialDiscipline(QueueDiscipline discipline){
		myDemandQ.setInitialDiscipline(discipline);
	}

	/** Sets the queue discipline used for waiting loads after being formed
	 * 
	 * @param discipline
	 */
	public void setLoadQueueInitialDiscipline(QueueDiscipline discipline){
		myLoadQ.setInitialDiscipline(discipline);
	}
	
	/** Takes in the supplied demand and enqueue it as a waiting demand
	 *  until a load can be formed
	 * 
	 * @param demand
	 */
	public void receiveDemand(Demand demand){
//		JSL.out.println(getTime() +" > " + "received a demand in DemandLoadBuilder " + getName());
		if (demand == null)
			throw new IllegalArgumentException("The supplied demand was null!");
		// always enqueue the newly arriving demand
		myTotalWeight.increment(demand.getWeight());
		myTotalCube.increment(demand.getCube());
		myDemandQ.enqueue(demand);
		if (myDefaultLoadFormingOption != LoadFormingOption.NONE){
			if (formLoad() == true){
//				JSL.out.println(getTime() +" > " + "formed a load in DemandLoadBuilder " + getName());
				if (myLoadFormedListener != null)
					myLoadFormedListener.loadFormed(this);
			}			
		}

	}
	
	/** Returns the queue that holds the demands waiting to be formed into loads
	 * 
	 * @return
	 */
	public Queue getDemandQueue(){
		return myDemandQ;
	}
	
	/** Returns a the Queue that holds the DemandLoads that have been formed
	 * 
	 * @return
	 */
	public Queue getLoadQueue(){
		return myLoadQ;
	}
	
	/**
	 * @return the defaultLoadFormingOption
	 */
	public LoadFormingOption getDefaultLoadFormingOption() {
		return myDefaultLoadFormingOption;
	}

	/** Sets the load forming rule.
	 * 
	 *  If the load forming option is RULE, then the user
	 *  must supply a rule via setLoadFormingRule() otherwise
	 *  an exception will result.  Use the method setLoadFormingRule()
	 *  to have the RULE option used.
	 * @param option
	 */
	public void setDefaultLoadFormingOption(LoadFormingOption option) {
		myDefaultLoadFormingOption = option;
	}

	/** Takes demands from the waiting queue and forms them into a load
	 *  by whatever internal load forming method is used by the LoadBuilder
	 * 
	 * @return true if a load was formed and placed in the load queue
	 */
	public boolean formLoad(){
		return formLoad(myDefaultLoadFormingOption);
	}
	
	/** Takes demands from the waiting queue and forms them into a load
	 *  by whatever internal load forming method is used by the LoadBuilder
	 * 
	 * @return true if a load was formed and placed in the load queue
	 */
	public boolean formLoad(LoadFormingOption option){

		if (option == LoadFormingOption.ALWAYS)
			return(true);
		if (option == LoadFormingOption.RULE)
			return(myLoadFormingRule.formLoad(this));	
		else if (option == LoadFormingOption.COUNT)
			return formLoadByCount(myCountLimit);
		else if (option == LoadFormingOption.CUBE)
			return formLoadByCube(myMinCube, myMaxCube);
		else if (option == LoadFormingOption.WEIGHT)
			return formLoadByWeight(myMinWeight, myMaxWeight);
		else
			return false;
	
	}
	
	/** Takes the first numDemands from the waiting queue
	 *  and forms a load, the load if formed should be placed in
	 *  the outgoing load queue.  If there are not numDemands
	 *  waiting then no load should be formed
	 * 
	 * @param numDemands, must be >0
	 * @return true if a load was formed and placed in the load queue
	 */
	public boolean formLoadByCount(int numDemands){
		if (numDemands <= 0)
			throw new IllegalArgumentException("The number of required demands must be > 0");
		
		if (myDemandQ.size() < numDemands)
			return (false);
		else {
			// make the load out of first numDemands
			DemandLoad load = new DemandLoad();
			for(int i=1;i<=numDemands;i++){
				Demand d = (Demand)myDemandQ.removeNext();
				myTotalWeight.decrement(d.getWeight());
				myTotalCube.decrement(d.getCube());
				load.addDemand(d);
			}
			// place load in outgoing load queue
			myLoadQ.enqueue(load);
			return (true);
		}
	}
	
	/** Takes the demands from the waiting queue
	 *  and forms a load up to maxWeight. The load if formed should be placed in
	 *  the outgoing load queue.  If there are not enough demands to have minWeight
	 *  then no load should be formed
	 * 
	 * @param minWeight The load should be at least this weight
	 * @param maxWeight The load cannot be more than this weight
	 * @return true if a load was formed and placed in the load queue
	 */
	public boolean formLoadByWeight(double minWeight, double maxWeight){
		if (minWeight <= 0)
			throw new IllegalArgumentException("The minimum weight must be > 0");

		if (maxWeight < minWeight)
			throw new IllegalArgumentException("The maximum weight must be >= the minimum weight");
	
		if (myDemandQ.isEmpty())
			return (false);
		
		if (myTotalWeight.getValue() < minWeight)
			return (false);
		
		DemandLoad load = new DemandLoad();
		// first find the demands and build the load
		int n = myDemandQ.size();
		for(int i=0; i < n; i++){
			Demand d = (Demand)myDemandQ.peekAt(i);		
			if (d.getWeight() + load.getWeight() <= maxWeight)
				load.addDemand(d);								
		}
		// now remove the found demands from the demand Q
		removeInBoundDemands(load.getDemandIterator());
		
		// place load in outgoing load queue
		myLoadQ.enqueue(load);
		return (true);
		
	}
	
	/** Removes demands from the in bound demand waiting queue that are presented by the iterator
	 * 
	 * @param load
	 */
	protected void removeInBoundDemands(Iterator<Demand> itr){
		while (itr.hasNext()){
			Demand d = itr.next();
			if (myDemandQ.remove(d)) {
				myTotalWeight.decrement(d.getWeight());
				myTotalCube.decrement(d.getCube());				
			}
		}
	}
	
	/** Takes the demands from the waiting queue
	 *  and forms a load up to maxCube.  The load if formed should be placed in
	 *  the outgoing load queue.  If there are not enough demands to have minCube
	 *  then no load should be formed
	 * 
	 * @param minCube The load should be at least this cube
	 * @param maxCube The load cannot be more than this cube
	 * @return true if a load was formed and placed in the load queue
	 */
	 public boolean formLoadByCube(double minCube, double maxCube){
		 
		if (minCube <= 0)
			throw new IllegalArgumentException("The minimum cube must be > 0");

		if (maxCube < minCube)
			throw new IllegalArgumentException("The maximum cube must be >= the minimum cube");
			
		if (myDemandQ.isEmpty())
			return (false);
		
		if (myTotalCube.getValue() < minCube)
			return (false);
		
		DemandLoad load = new DemandLoad();
		// first find the demands and build the load
		int n = myDemandQ.size();
		for(int i=0; i < n; i++){
			Demand d = (Demand)myDemandQ.peekAt(i);		
			if (d.getCube() + load.getCube() <= maxCube)
				load.addDemand(d);								
		}
		// now remove the found demands from the demand Q
		removeInBoundDemands(load.getDemandIterator());
		
		// place load in outgoing load queue
		myLoadQ.enqueue(load);
		return (true);
	}

	/** The lower limit on the cube size for loads if formed
	 *  by cube
	 *  
	 * @return the myMinCube
	 */
	public double geMinCubeLimit() {
		return myMinCube;
	}

	/** The upper limit on the cube size for loads if formed
	 *  by cube
	 *  
	 * @return the myMaxCube
	 */
	public double getMaxCubeLimit() {
		return myMaxCube;
	}

	/** Changing the limts only effects the next load to be formed, when it is formed
	 * @param myMaxCube the myMaxCube to set
	 */
	public void setCubeFormingLimits(double minCube, double maxCube) {
		
		if (minCube <= 0)
			throw new IllegalArgumentException("The minimum cube must be > 0");

		if (maxCube < minCube)
			throw new IllegalArgumentException("The maximum cube must be >= the minimum cube");
		
		myMaxCube = maxCube;
		myMinCube = minCube;
	}

	/** The lower limit on the weight size for loads if formed
	 *  by weight
	 *  
	 * @return the myMinWeight
	 */
	public double geMinWeightLimit() {
		return myMinWeight;
	}

	/** The upper limit on the weight size for loads if formed
	 *  by weight
	 *  
	 * @return the maxWeight
	 */
	public double getMaxWeightLimit() {
		return myMaxWeight;
	}

	/** Changing the limtis only effects the next load to be formed, when it is formed
	 * 
	 * @param minWeight
	 * @param maxWeight
	 */
	public void setWeightFormingLimits(double minWeight, double maxWeight) {
		
		if (minWeight <= 0)
			throw new IllegalArgumentException("The minimum weight must be > 0");

		if (maxWeight < minWeight)
			throw new IllegalArgumentException("The maximum weight must be >= the minimum weight");
		
		myMaxWeight = maxWeight;
		myMinWeight = minWeight;
	}

	/** The minimum number of demands that should be in a load
	 *  if forming the loads by count
	 *  
	 * @return the myCountLimit
	 */
	public int getCountLimit() {
		return myCountLimit;
	}

	/** The minimum number of demands that should be in a load
	 *  if forming the loads by count.  Changing the limit only effects the next load
	 *  to be formed
	 *  
	 * @param countLimit the CountLimit to set
	 */
	public void setCountLimit(int countLimit) {
		if (countLimit <= 0)
			throw new IllegalArgumentException("The number of required demands must be > 0");
		myCountLimit = countLimit;
	}

	/** The reference to the object currently listening for automatic load formation
	 * 
	 * @return the listener
	 */
	protected DemandLoadFormedListenerIfc getLoadFormedListener() {
		return myLoadFormedListener;
	}

	/** Sets the reference for the object that reacts to load formations
	 * @param listener the listener to set
	 */
	public void setLoadFormedListener(DemandLoadFormedListenerIfc listener) {
		myLoadFormedListener = listener;
	}

	/** The reference to the object that handles forming a load
	 * @return the loadBuildingRule
	 */
	protected DemandLoadFormingRuleIfc getLoadFormingRule() {
		return myLoadFormingRule;
	}

	/** If the rule is null, the default load forming option becomes NONE
	 *  If the rule is not null, the default load forming option becomes RULE
	 * @param rule the LoadBuildingRule to set
	 */
	public void setLoadFormingRule(DemandLoadFormingRuleIfc rule) {
		if (rule != null)
			myDefaultLoadFormingOption = LoadFormingOption.RULE;
		else
			myDefaultLoadFormingOption = LoadFormingOption.NONE;
			
		myLoadFormingRule = rule;
	}

}
