/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.transportlayer;

import java.util.HashMap;
import java.util.Map;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.RandomIfc;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandSenderIfc;

/** A TimeBasedDemandCarrier allows for a random transport delay when
 *  transporting the supplied demand.  The transport
 *  times are provided by a class that implements RandomIfc.  An example application of this
 *  is to have a DemandFillerIfc (e.g. InventoryHoldingPoint) use an instance
 *  of a TimeBasedDemandCarrier to allow time based delays to transport to its
 *  customers.  This can be accomplished by attaching an instance of a
 *  DemandListenerFilledIfc to the demand and calling an appropriate transport
 *  option from the listener when the demand is filled.
 * 
 */
public class TimeBasedDemandCarrier extends DemandCarrierAbstract {

	/** Holds the transport times to each of the senders as random variables
	 */
	protected Map<DemandSenderIfc, RandomVariable> myTransportTimes;
	
	/** Allows the time based carrier to have a no delay (immediate) transport
	 *  for any demand that needs transport.  By default this is false.
	 */
	private boolean myImmediateTransportFlag;
	
	/** Creates the TimeBasedDemandCarrier with the given model element as
	 *  a parent and a default name
	 * 
	 * @param parent
	 */
	public TimeBasedDemandCarrier(ModelElement parent) {
		this(parent, null);
	}

	/** Creates the TimeBasedDemandCarrier with the given model element as
	 *  a parent and the given name
	 * 
	 * @param parent
	 * @param name
	 */
	public TimeBasedDemandCarrier(ModelElement parent, String name) {
		super(parent, name);
		myTransportTimes = new HashMap<DemandSenderIfc, RandomVariable>();
		myImmediateTransportFlag = false;
	}

	/** Checks if the carrier ships to the supplied sender
	 *  demand sender.
	 * 
	 * @param sender
	 * @return
	 */
	public final boolean shipsTo(DemandSenderIfc sender){
		return(myTransportTimes.containsKey(sender));
	}
	
	/** If the immediate transport flag is true then if the sender is not found
	 *  with a corresponding shipping time, a direct no delay transport will occur.
	 *  If this flag is not true and the sender is not found, then
	 *  a NoCarrierOptionException exception will be thrown. The default is
	 *  to have the immediate transport flag set to false.  By setting the flag
	 *  to true, the TimeBasedCarrier can act as a NoDelayDemandCarrier for
	 *  ANY DemandSenderIfc attached to demands sent to it.
	 * 
	 * @param flag
	 */
	public final void setImmediateTransportFlag(boolean flag){
		myImmediateTransportFlag = flag;
	}
	
	/** If the immediate transport flag is true then if the sender is not found
	 *  with a corresponding shipping time, a direct no delay shipment will
	 *  be used.  If this flag is not true and the sender is not found, then
	 *  a NoCarrierOptionException exception will be thrown. The default is
	 *  to have the immediate transport flag set to false.
	 * 
	 * @return
	 */
	public boolean getImmediateTransportFlag(){
		return myImmediateTransportFlag;
	}
	
	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.DemandShipperIfc#shipDemand(jsl.supplychainNew.inventorylayer.Demand)
	 */
	public void transportDemand(Demand d) {
		
		if (d == null)
			throw new IllegalArgumentException("The supplied demand was null!");
		
		DemandSenderIfc sender = d.getSender();

		if (sender == null)
			throw new IllegalArgumentException("The sender on the demand was null!");

		RandomVariable shippingTime = myTransportTimes.get(sender);
		
		if (shippingTime == null){
			if (myImmediateTransportFlag == false)
				throw new NoCarrierOptionException("The sender is not valid for this shipper");
			d.ship();
			d.deliver();
		} else {
			scheduleEvent(shippingTime, d);
			d.ship();
		}

	}
	
	/** Sets the transport time distribution for the provided demand sender
	 * 
	 * @param sender
	 * @param distribution
	 */
	public final void setTransportTime(DemandSenderIfc sender, RandomIfc distribution){
		if (sender == null)
			throw new IllegalArgumentException("The sender must be non-null.");

		if (distribution == null)
			throw new IllegalArgumentException("The DistributionIfc distribution must be non-null.");
		
		if (myTransportTimes.containsKey(sender)){ // the sender has already been added
			// need only to set the distribution associated with the sender 
			RandomVariable rv = myTransportTimes.get(sender);
			rv.setInitialRandomSource(distribution);			
		} else {
			// new sender being entered
			// create the random variable
			RandomVariable rv = new RandomVariable(this, distribution);
			myTransportTimes.put(sender, rv);
		}
	}
	
	protected void handleEvent(JSLEvent event){
		Demand d = (Demand)event.getMessage();
		d.deliver();
	}
}
