/*
* Created on Aug 13, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.transportlayer;

import java.util.HashMap;
import java.util.Map;

import jsl.modeling.ActionListenerIfc;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.queue.QueueDiscipline;
import jsl.modeling.elements.variable.AggregateCounter;
import jsl.modeling.elements.variable.Counter;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.reporting.JSL;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandSenderIfc;
import jslx.supplychain.transportlayer.DemandLoadBuilder.LoadFormingOption;
/**
 * @author rossetti
 *
 */
public class TimeBasedLoadCarrier extends TimeBasedDemandCarrier {

	/** Holds the load builders for each destination
	 */
	protected Map<DemandSenderIfc, DemandLoadBuilder> myLoadBuilders;
	
	/** Counts the number of loads shipped by destination
	 */
	protected Map<DemandSenderIfc, Counter> myShipmentCounters;
	
	/** Counts the total number of shipments regardless of destination
	 */
	protected AggregateCounter myShipmentCounter;
	
	/** Should be called whenever a load is formed by the a LoadBuilder
	 */
	protected DemandLoadFormedListenerIfc myLoadFormedListener = new LoadBuiltListener();
	
	/** Determines whether or not the shipper will react when a load builder
	 *  forms a load.  By default this is false
	 * 
	 */
	protected boolean myReactToLoadBuildersFlag = false;
	
	/** Used to invoke the shipping action from the shipping schedule
	 */
	protected ShippingAction myShippingAction = new ShippingAction();
		
	/**
	 * @param parent
	 */
	public TimeBasedLoadCarrier(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public TimeBasedLoadCarrier(ModelElement parent, String name) {
		super(parent, name);
		
		myLoadBuilders = new HashMap<DemandSenderIfc, DemandLoadBuilder>();
		myShipmentCounters = new HashMap<DemandSenderIfc, Counter>();
		myShipmentCounter = new AggregateCounter(this, getName() + " Total Shipments");
	}

	/** Returns whether or not this shipper will react to load formation by builders
	 *  using its loadFormed() method
	 * 
	 * @return
	 */
	public boolean getReactToLoadBuildersFlag(){
		return myReactToLoadBuildersFlag;
	}
	
	/** Sets whether or not this shipper will react to load formation by builders
	 *  using its loadFormed() method
	 * 
	 * @param flag
	 */
	public void setReactToLoadBuildersFlag(boolean flag){
		myReactToLoadBuildersFlag = flag;
	}
	
	/* (non-Javadoc)
	 * @see jsl.research.supplychain.shippinglayer.TimeBasedDemandShipper#shipDemand(jsl.research.supplychain.inventorylayer.Demand)
	 */
	@Override
	public void transportDemand(Demand d) {
		
//		JSL.out.println(getTime() +" > " + "asked to transport demand in " + getName());

		if (d == null)
			throw new IllegalArgumentException("The supplied demand was null!");
		
		DemandSenderIfc sender = d.getSender();

		if (sender == null)
			throw new IllegalArgumentException("The sender on the demand was null!");

		DemandLoadBuilder loadBuilder = myLoadBuilders.get(sender);
		
		if (loadBuilder == null)
			throw new IllegalArgumentException("There is no load builder for the demand's sender");
		
		loadBuilder.receiveDemand(d);
		
	}

	/** Creates and assigns a default load builder for the given destination.  The returned
	 *  LoadBuilder can be used to set specific load building characteristics
	 * 
	 * @param destination
	 * @return
	 */
	public DemandLoadBuilder assignLoadBuilder(DemandSenderIfc destination){
		DemandLoadBuilder loadBuilder = new DemandLoadBuilder(this);
		assignLoadBuilder(destination, loadBuilder);
		return(loadBuilder);
	}
	
	/** Creates and assigns a default load builder for the given destination.  The returned
	 *  LoadBuilder can be used to set specific load building characteristics
	 * 
	 * @param destination
	 * @param name The name for the load builder, must be unique in the model
	 * @return
	 */
	public DemandLoadBuilder assignLoadBuilder(DemandSenderIfc destination, String name){
		DemandLoadBuilder loadBuilder = new DemandLoadBuilder(this, name);
		assignLoadBuilder(destination, loadBuilder);
		return(loadBuilder);
	}
	
	/** Assigns the supplied LoadBuilder to the given destination.  Attaches a 
	 * LoadFormedListenerIfc to the LoadBuilder
	 * 
	 * @param destination
	 * @param loadBuilder
	 */
	public void assignLoadBuilder(DemandSenderIfc destination, DemandLoadBuilder loadBuilder){
		if (destination== null)
			throw new IllegalArgumentException("The sender must be non-null.");

		if (loadBuilder == null)
			throw new IllegalArgumentException("The load builder must be non-null.");
		
		if (myLoadBuilders.containsKey(destination))
			throw new IllegalArgumentException("The supplied destination has already been assigned a load builder");
		
		myLoadBuilders.put(destination, loadBuilder);
		loadBuilder.setLoadFormedListener(myLoadFormedListener);
		
		Counter c = new Counter(this, destination.getName() + " #Shipments");
		myShipmentCounters.put(destination, c);
		myShipmentCounter.subscribeTo(c);
	}
	
	/** Returns an aggregate counter for all shipments made by the carrier
	 * @return the shipmentCounter
	 */
	public final AggregateCounter getShipmentCounter() {
		return myShipmentCounter;
	}
	
	/** Returns a counter for shipments to the destination
	 * 
	 * @param destination
	 * @return
	 */
	public Counter getShipmentCounter(DemandSenderIfc destination) {
		return myShipmentCounters.get(destination);
	}

	/** Returns the load builder for the destination
	 * 
	 * @param destination  cannot be null, must have been assigned a load builder
	 * @return
	 */
	public DemandLoadBuilder getLoadBuilder(DemandSenderIfc destination){
		if (destination== null)
			throw new IllegalArgumentException("The sender must be non-null.");
		
		if (!myLoadBuilders.containsKey(destination))
			throw new IllegalArgumentException("The supplied destination has not been assigned a load builder");

		return myLoadBuilders.get(destination);
	}
	
	/** Indicates whether a load builder for the destination has been set up
	 * 
	 * @param destination
	 * @return
	 */
	public boolean containsLoadBuilder(DemandSenderIfc destination){
		return myLoadBuilders.containsKey(destination);
	}

	/**
	 * @param numDemands
	 * @return
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#formLoadByCount(int)
	 */
	public boolean formLoadByCount(DemandSenderIfc destination, int numDemands) {
		return getLoadBuilder(destination).formLoadByCount(numDemands);
	}

	/**
	 * @param minCube
	 * @param maxCube
	 * @return
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#formLoadByCube(double, double)
	 */
	public boolean formLoadByCube(DemandSenderIfc destination, double minCube, double maxCube) {
		return getLoadBuilder(destination).formLoadByCube(minCube, maxCube);
	}

	/**
	 * @param minWeight
	 * @param maxWeight
	 * @return
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#formLoadByWeight(double, double)
	 */
	public boolean formLoadByWeight(DemandSenderIfc destination, double minWeight, double maxWeight) {
		return getLoadBuilder(destination).formLoadByWeight(minWeight, maxWeight);
	}

	/**
	 * @return
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#geMinCubeLimit()
	 */
	public double geMinCubeLimit(DemandSenderIfc destination) {
		return getLoadBuilder(destination).geMinCubeLimit();
	}

	/**
	 * @return
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#geMinWeightLimit()
	 */
	public double geMinWeightLimit(DemandSenderIfc destination) {
		return getLoadBuilder(destination).geMinWeightLimit();
	}

	/**
	 * @return
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#getCountLimit()
	 */
	public int getCountLimit(DemandSenderIfc destination) {
		return getLoadBuilder(destination).getCountLimit();
	}

	/**
	 * @return
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#getDefaultLoadFormingOption()
	 */
	public LoadFormingOption getDefaultLoadFormingOption(DemandSenderIfc destination) {
		return getLoadBuilder(destination).getDefaultLoadFormingOption();
	}

	/**
	 * @return
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#getMaxCubeLimit()
	 */
	public double getMaxCubeLimit(DemandSenderIfc destination) {
		return getLoadBuilder(destination).getMaxCubeLimit();
	}

	/**
	 * @return
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#getMaxWeightLimit()
	 */
	public double getMaxWeightLimit(DemandSenderIfc destination) {
		return getLoadBuilder(destination).getMaxWeightLimit();
	}

	/**
	 * @param countLimit
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#setCountLimit(int)
	 */
	public void setCountLimit(DemandSenderIfc destination, int countLimit) {
		getLoadBuilder(destination).setCountLimit(countLimit);
	}

	/**
	 * @param minCube
	 * @param maxCube
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#setCubeFormingLimits(double, double)
	 */
	public void setCubeFormingLimits(DemandSenderIfc destination, double minCube, double maxCube) {
		getLoadBuilder(destination).setCubeFormingLimits(minCube, maxCube);
	}

	/**
	 * @param option
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#setDefaultLoadFormingOption(jslx.supplychain.transportlayer.DemandLoadBuilder.LoadFormingOption)
	 */
	public void setDefaultLoadFormingOption(DemandSenderIfc destination, LoadFormingOption option) {
		getLoadBuilder(destination).setDefaultLoadFormingOption(option);
	}

	/**
	 * @param discipline
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#setDemandQueueInitialDiscipline(jsl.modeling.elements.queue.QueueDiscipline)
	 */
	public void setDemandQueueInitialDiscipline(DemandSenderIfc destination, QueueDiscipline discipline) {
		getLoadBuilder(destination).setDemandQueueInitialDiscipline(discipline);
	}

	/**
	 * @param rule
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#setLoadFormingRule(jslx.supplychain.transportlayer.DemandLoadFormingRuleIfc)
	 */
	public void setLoadFormingRule(DemandSenderIfc destination, DemandLoadFormingRuleIfc rule) {
		getLoadBuilder(destination).setLoadFormingRule(rule);
	}

	/**
	 * @param discipline
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#setLoadQueueInitialDiscipline(jsl.modeling.elements.queue.QueueDiscipline)
	 */
	public void setLoadQueueInitialDiscipline(DemandSenderIfc destination, QueueDiscipline discipline) {
		getLoadBuilder(destination).setLoadQueueInitialDiscipline(discipline);
	}

	/**
	 * @param minWeight
	 * @param maxWeight
	 * @see jslx.supplychain.transportlayer.DemandLoadBuilder#setWeightFormingLimits(double, double)
	 */
	public void setWeightFormingLimits(DemandSenderIfc destination, double minWeight, double maxWeight) {
		getLoadBuilder(destination).setWeightFormingLimits(minWeight, maxWeight);
	}

	/* (non-Javadoc)
	 * @see jsl.research.supplychain.shippinglayer.TimeBasedDemandShipper#handleEvent(jsl.modeling.JSLEvent)
	 */
	@Override
	protected void handleEvent(JSLEvent event) {
		// a load has been delivered to the receiver
		DemandLoad load = (DemandLoad)event.getMessage();
		DemandSenderIfc destination = load.getDestination();
		Counter c = myShipmentCounters.get(destination);
		c.increment();
		load.deliver();
//		JSL.out.println(getTime() +" > " + "delivered load: " + load);
		load.nullify();
	}
	
	/** This method can be overridden to provide shipping behavior for when a load is formed
	 *  by a LoadBuilder.  This method is called when the provided LoadBuilder has
	 *  formed a load and notified its attached LoadFormedListenerIfc
	 * 
	 * @param loadBuilder
	 */
	protected void loadFormed(DemandLoadBuilder loadBuilder){
//		JSL.out.println(getTime() +" > " + loadBuilder.getName() + " has formed a load for " + getName());
		// one or more loads have been formed, remove them and ship them
		shipFromLoadBuilder(loadBuilder);
	}
	
	/** Processes the load builder's load queue, removing any loads and
	 *  causing them to be shipped to their destination
	 * 
	 * @param loadBuilder
	 */
	protected void shipFromLoadBuilder(DemandLoadBuilder loadBuilder){
		Queue loadQ = loadBuilder.getLoadQueue();
		while(loadQ.isNotEmpty()){
			DemandLoad load = (DemandLoad)loadQ.removeNext();
			DemandSenderIfc destination = load.getDestination();
			RandomVariable shippingTime = myTransportTimes.get(destination);
			if (shippingTime == null){
				if (getImmediateTransportFlag() == false)
					throw new NoCarrierOptionException("The sender is not valid for this shipper");
				load.ship();
				load.deliver();
				load.nullify();
			} else {
//				JSL.out.println(getTime() +" > " + "shipping load: " + load);
				scheduleEvent(shippingTime, load);
				load.ship();
			}
		}		
	}
	
	/** The default action is to process all load builders causing
	 *  any formed loads to be shipped
	 */
	protected void shipLoads(){
		
		for(DemandLoadBuilder dlb: myLoadBuilders.values()){
			shipFromLoadBuilder(dlb);
		}
		
	}
	
	protected class LoadBuiltListener implements DemandLoadFormedListenerIfc {
		public void loadFormed(DemandLoadBuilder loadBuilder) {
//			JSL.out.println("----> " + "in LoadBuiltListener for " + getName());
			if (myReactToLoadBuildersFlag == true)
				TimeBasedLoadCarrier.this.loadFormed(loadBuilder);
		}
	}
	
	protected class ShippingAction implements ActionListenerIfc {
		public void action(JSLEvent evt) {
			shipLoads();	
		}	
	}

}
