/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.transportlayer;

import java.util.HashMap;
import java.util.Map;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandFillerIfc;
import jslx.supplychain.inventorylayer.DemandSenderIfc;

/** TimeBasedNetworkDemandCarrier handles the connection/transport between DemandFillerIfc
 *  and DemandSenerIfc via a specified time delay distribution.  Each specified
 *  DemandFillerIfc may have many DemandSenderIfcs to which it must transport.  DemandSenderIfcs
 *  can be associated with multiple DemandFillerIfcs, but only one link between a particular
 *  DemandFillerIfc and a particular DemandSenderIfc may be defined.
 *  
 *
 */
public class TimeBasedNetworkDemandCarrier extends DemandCarrierAbstract {

	/** Allows easy modeling of no delay shipping between a filler and a sender
	 */
	protected static final DistributionIfc NoDelayCDF = Constant.ZERO;

	/** Holds the transport times from each filler to each of the senders as random variables
	 */
	private Map<DemandFillerIfc, Map<DemandSenderIfc, RandomVariable>> myTransportTimes;

	/** Allows the time based carrier to immediately transport
	 *  any demand that needs shipping.  By default this is false.
	 */
	private boolean myImmediateTransportFlag;
	
	/** Allows sharing of the RV for no delay transport between a filler and a sender
	 */
	protected RandomVariable myNoDelayRV;

	/**
	 * @param parent
	 */
	public TimeBasedNetworkDemandCarrier(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public TimeBasedNetworkDemandCarrier(ModelElement parent, String name) {
		super(parent, name);
		myTransportTimes = new HashMap<DemandFillerIfc, Map<DemandSenderIfc, RandomVariable>>();
		myNoDelayRV = new RandomVariable(this, NoDelayCDF);
		myImmediateTransportFlag = false;
	}
	/** Sets the transport time distribution between the filler and the sender with a zero
	 *  delay distribution
	 * 
	 * @param filler
	 * @param sender
	 */
	public final void setTransportTime(DemandFillerIfc filler, DemandSenderIfc sender){
		setTransportTime(filler, sender, NoDelayCDF);
	}
	
	/** Sets the shipping time distribution between the filler and the sender
	 * 
	 * @param filler
	 * @param sender
	 * @param distribution
	 */
	public final void setTransportTime(DemandFillerIfc filler, DemandSenderIfc sender, DistributionIfc distribution){

		if (filler == null)
			throw new IllegalArgumentException("The filler must be non-null.");

		if (sender == null)
			throw new IllegalArgumentException("The sender must be non-null.");

		if (distribution == null)
			throw new IllegalArgumentException("The DistributionIfc distribution must be non-null.");
		
		// get the shipping time map for the filler
		Map<DemandSenderIfc, RandomVariable> shippingTimes;
		if (!myTransportTimes.containsKey(filler)){// filler not added yet, set up shipping times
			shippingTimes = new HashMap<DemandSenderIfc, RandomVariable>();
			myTransportTimes.put(filler, shippingTimes);
		} else {// filler already added, just get shipping times
			shippingTimes = myTransportTimes.get(filler);
		}

		if (shippingTimes.containsKey(sender)){ // the sender has already been added
			// need only to change the distribution associated with the sender 
			// get the random variable
			RandomVariable rv = shippingTimes.get(sender);
			if ((myNoDelayRV == rv) && (distribution != NoDelayCDF)){
				// the rv is the no delay rv and incoming distribution is not the no delay CDF
				// need to change the random variable and set the distribution
				shippingTimes.put(sender, new RandomVariable(this, distribution));
			} else // just changing the distribution
				rv.setInitialRandomSource(distribution);			
		} else {
			// new sender being entered
			if (distribution == NoDelayCDF)// designated a the no delay CDF
				shippingTimes.put(sender, myNoDelayRV);// just reuse the no delay RV
			else {// not the no delay CDF
				// create a new random variable, for the distribution
				RandomVariable rv = new RandomVariable(this, distribution);
				// add it to the shipping times
				shippingTimes.put(sender, rv);				
			}
		}
	}

	/** Attaches the supplied sender to all currently included demand fillers
	 *  with an immediate transport option
	 * 
	 * @param sender
	 */
	public final void attachDemandSenderToAllFillers(DemandSenderIfc sender){
		for(DemandFillerIfc filler: myTransportTimes.keySet()){
			setTransportTime(filler, sender);			
		}
	}
	/** If the immediate transport flag is true then if the sender is not found
	 *  with a corresponding shipping time, a direct no delay shipment will
	 *  be used.  If this flag is not true and the filler/sender pair is not found, then
	 *  a NoCarrierOptionException exception will be thrown. The default is
	 *  to have the immediate transport flag set to false.  By setting the flag
	 *  to true, the TimeBasedCarrier can act as a NoDelayDemandCarrier for
	 *  ANY DemandFiller/DemandSenderIfc pair that have an interaction when the
	 *  DemandFillerIfc uses this as the carrier.
	 * 
	 * @param flag
	 */
	public final void setImmediateTransportFlag(boolean flag){
		myImmediateTransportFlag = flag;
	}
	
	/* (non-Javadoc)
	 * @see jsl.supplychainNew.inventorylayer.DemandShipperIfc#shipDemand(jsl.supplychainNew.inventorylayer.Demand)
	 */
	public void transportDemand(Demand d) {
		
		if (d == null)
			throw new IllegalArgumentException("The supplied demand was null!");

		DemandFillerIfc filler = d.getFiller();

		if (filler == null)
			throw new IllegalArgumentException("The filler on the demand was null!");

		DemandSenderIfc sender = d.getSender();

		if (sender == null)
			throw new IllegalArgumentException("The sender on the demand was null!");

		Map<DemandSenderIfc, RandomVariable> shippingTimes = myTransportTimes.get(filler);
		if (shippingTimes == null){
			if (myImmediateTransportFlag == false)
				throw new NoCarrierOptionException("The filler is not registered with the shipper");
			d.ship();
			d.deliver();
		} else {
			// the filler is registered, check for sender
			RandomVariable shippingTime = shippingTimes.get(sender);
			if (shippingTime == null){
				if (myImmediateTransportFlag == false)
					throw new NoCarrierOptionException("The sender is not valid for this shipper");
				d.ship();
				d.deliver();
			} else {
				double st = shippingTime.getValue();
				if (st == 0.0){
					d.ship();
					d.deliver();					
				} else {
					scheduleEvent(st, d);
					d.ship();
				}
			}			
		}
	}

	protected void handleEvent(JSLEvent event){
		Demand d = (Demand)event.getMessage();
		d.deliver();
	}

	/** Checks if the carrier ships to the supplied sender
	 *  demand sender.  If any sender is found with a defined transport time
	 *  then this will return true
	 * 
	 * @param sender
	 * @return
	 */
	public boolean shipsTo(DemandSenderIfc sender) {
		if (sender == null)
			throw new IllegalArgumentException("The sender on the demand was null!");
		
		for(Map<DemandSenderIfc, RandomVariable> map: myTransportTimes.values()){
			if (map.containsKey(sender))
				return true;
		}

		return false;
	}

}
