/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.transportlayer;

import java.util.HashMap;
import java.util.Map;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.RandomIfc;
import jslx.supplychain.inventorylayer.Order;
import jslx.supplychain.inventorylayer.OrderSenderIfc;

/** A TimeBasedOrderCarrier allows for a random transport delay when
 *  transporting to the supplied order senders (customers).  The transport
 *  times are provided by an instance of RandomIfc.  An example application of this
 *  is to have a OrderFillerIfc (e.g. Warehouse) use an instance
 *  of a TimeBasedOrderCarrier to allow time based delays to transport to its
 *  customers.  This can be accomplished by attaching an instance of a
 *  OrderListenerFilledIfc to the order and calling an appropriate transport
 *  option from the listener when the order is filled.
 */
public class TimeBasedOrderCarrier extends SchedulingElement implements OrderCarrierIfc {

	/** Holds the shipping times to each of the senders as random variables
	 */
	private Map<OrderSenderIfc, RandomVariable> myTransportTimes;
	
	/** Allows the time based carrier to have a no delay (immediate) transport
	 *  for any demand that needs transport.  By default this is false.
	 */
	private boolean myImmediateTransportFlag;
	
	/** Creates the TimeBasedOrderCarrier with the given model element as
	 *  a parent and a default name
	 * 
	 * @param parent
	 */
	public TimeBasedOrderCarrier(ModelElement parent) {
		this(parent, null);
	}

	/** Creates the TimeBasedDemandCarrier with the given model element as
	 *  a parent and the given name
	 * 
	 * @param parent
	 * @param name
	 */
	public TimeBasedOrderCarrier(ModelElement parent, String name) {
		super(parent, name);
		myTransportTimes = new HashMap<OrderSenderIfc, RandomVariable>();
		myImmediateTransportFlag = false;
	}

	/** Checks if this carrier has a transport time defined for the supplied
	 *  demand sender.
	 * 
	 * @param sender
	 * @return
	 */
	public final boolean contains(OrderSenderIfc sender){
		return(myTransportTimes.containsKey(sender));
	}

	/** If the immediate transport flag is true then if the sender is not found
	 *  with a corresponding shipping time, a direct no delay shipment will
	 *  be used.  If this flag is not true and the sender is not found, then
	 *  a NoCarrierOptionException exception will be thrown. The default is
	 *  to have the immediate transport flag set to false.  By setting the flag
	 *  to true, the TimeBasedCarrier can act as a NoDelayOrderCarrier for
	 *  ANY OrderSenderIfc attached to demands sent to it.
	 * 
	 * @param flag
	 */
	public final void setImmediateTransportFlag(boolean flag){
		myImmediateTransportFlag = flag;
	}

	public void transportOrder(Order order) {
		
		if (order == null)
			throw new IllegalArgumentException("The supplied order was null!");
		
		OrderSenderIfc sender = order.getSender();

		if (sender == null)
			throw new IllegalArgumentException("The sender on the order was null!");

		RandomVariable shippingTime = myTransportTimes.get(sender);
		
		if (shippingTime == null){
			if (myImmediateTransportFlag == false)
				throw new NoCarrierOptionException("The sender is not valid for this shipper");
			order.ship();
			order.deliver();
		} else {
			scheduleEvent(shippingTime.getValue(), order);
			order.ship();
		}

	}
	
	/** Sets the transport time distribution for the provided order sender
	 * 
	 * @param sender
	 * @param distribution
	 */
	public final void setTransportTime(OrderSenderIfc sender, RandomIfc distribution){
		if (sender == null)
			throw new IllegalArgumentException("The sender must be non-null.");

		if (distribution == null)
			throw new IllegalArgumentException("The RandomIfc must be non-null.");
		
		if (myTransportTimes.containsKey(sender)){ // the sender has already been added
			// need only to set the distribution associated with the sender 
			RandomVariable rv = myTransportTimes.get(sender);
			rv.setInitialRandomSource(distribution);			
		} else {
			// new sender being entered
			// create the random variable
			RandomVariable rv = new RandomVariable(this, distribution);
			myTransportTimes.put(sender, rv);
		}
	}
	
	protected void handleEvent(JSLEvent event){
		Order order = (Order)event.getMessage();
		order.deliver();
	}
}
