/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.transportlayer;

import java.util.HashMap;
import java.util.Map;

import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.RandomIfc;
import jslx.supplychain.inventorylayer.Demand;
import jslx.supplychain.inventorylayer.DemandSenderIfc;
import jslx.supplychain.inventorylayer.ItemType;

/** A TimeBasedTypeLocationIndependentDemandCarrier allows for a random transport delay when
 *  shipping a particular item type to the supplied demand senders (customers).  
 *  Each item type delay can be thus made independent to each other within a particular
 *  shipping line. And the item type delay may be independent of the shipping line The item types are The shipping
 *  times are provided by a class that implements RandomIfc.  An example application of this
 *  is to have a DemandFillerIfc (e.g. InventoryHoldingPoint) use an instance
 *  of a TimeBasedDemandShipper to allow time based delays to ship to its
 *  customers.  This can be accomplished by attaching an instance of a
 *  DemandListenerFilledIfc to the demand and calling an appropriate shipping
 *  option from the listener when the demand is filled.
 * 
 */
public class TimeBasedTypeLocationIndependentDemandCarrier extends DemandCarrierAbstract {

	private Map<DemandSenderIfc, Map<ItemType, RandomVariable>> myTransportTimes;
	
	private boolean myImmediateTransportFlag;
	
	public TimeBasedTypeLocationIndependentDemandCarrier(ModelElement parent) {
		this(parent, null);
	}

	public TimeBasedTypeLocationIndependentDemandCarrier(ModelElement parent, String name) {
		super(parent, name);
		myTransportTimes = new HashMap<DemandSenderIfc, Map<ItemType, RandomVariable>>();
		myImmediateTransportFlag = false;
	}

	public final boolean contains(DemandSenderIfc sender){
		return(myTransportTimes.containsKey(sender));
	}
	
	public final boolean contains(DemandSenderIfc sender, ItemType type){
		if(contains(sender)){			
			return(myTransportTimes.get(sender).containsKey(type));
		}else{
			return false;
		}
	}
	
	
	public final void setImmediateTransportFlag(boolean flag){
		myImmediateTransportFlag = flag;
	}

	public boolean getImmediateTransportFlag(){
		return myImmediateTransportFlag;
	}
	
	public void transportDemand(Demand d) {
		
		if (d == null)
			throw new IllegalArgumentException("The supplied demand was null!");
		
		DemandSenderIfc sender = d.getSender();

		if (sender == null)
			throw new IllegalArgumentException("The sender on the demand was null!");
		
		ItemType type = d.getItemType();

		if (type == null)
			throw new IllegalArgumentException("The type on the demand was null!");		
		
		RandomVariable shippingTime = myTransportTimes.get(sender).get(type);
		
		if (shippingTime == null){
			if (myImmediateTransportFlag == false)
				throw new NoCarrierOptionException("The sender is not valid for this shipper");
			d.ship();
			d.deliver();
		} else {
			scheduleEvent(shippingTime.getValue(), d);
			d.ship();
		}

	}
	
	/** Sets the shipping time distribution for the provided demand sender
	 * 
	 * @param sender
	 * @param distribution
	 */
	public final void setTransportTime(DemandSenderIfc sender, ItemType type, RandomIfc distribution){
		if (sender == null)
			throw new IllegalArgumentException("The sender must be non-null.");
		
		if (type == null)
			throw new IllegalArgumentException("The type must be non-null.");

		if (distribution == null)
			throw new IllegalArgumentException("The DistributionIfc distribution must be non-null.");
		
		if (myTransportTimes.containsKey(sender)){ // the sender has already been added
			// need to reset the map associated with the sender
			Map<ItemType, RandomVariable> typetodist = myTransportTimes.get(sender);
			if (typetodist.containsKey(type)){ // the type has already been added
				//need only to set the distribution associated with the type
				RandomVariable rv = (RandomVariable)typetodist.get(type);
				rv.setInitialRandomSource(distribution);			
			} else {
				// new type being entered
				// create the random variable
				RandomVariable rv = new RandomVariable(this, distribution);
				typetodist.put(type, rv);
			}			
		} else {
			// new sender being entered
			// create the map
			Map<ItemType, RandomVariable> typetodist = new HashMap <ItemType, RandomVariable>();
			RandomVariable rv = new RandomVariable(this, distribution);
			typetodist.put(type,rv);
			myTransportTimes.put(sender, typetodist);
		}
	}
	
	protected void handleEvent(JSLEvent event){
		Demand d = (Demand)event.getMessage();
		d.deliver();
	}

	public boolean shipsTo(DemandSenderIfc sender) {
		if (sender == null)
			throw new IllegalArgumentException("The sender on the demand was null!");
		
		return myTransportTimes.containsKey(sender);
	}

	public double getTransportTimes(ItemType type, DemandSenderIfc sender) {		
		return myTransportTimes.get(sender).get(type).getParameters()[0];
	}

	
}
