/*
* Created on Aug 21, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.supplychain.transportlayer.testing;

import jsl.modeling.Experiment;
import jsl.modeling.Model;
import jsl.modeling.Replication;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jslx.supplychain.inventorylayer.DemandGenerator;
import jslx.supplychain.inventorylayer.InventoryHoldingPoint;
import jslx.supplychain.inventorylayer.ItemType;
import jslx.supplychain.inventorylayer.LeadTimeDemandFiller;
import jslx.supplychain.transportlayer.DemandLoadBuilder;
import jslx.supplychain.transportlayer.TimeBasedLoadCarrier;

/**
 * @author rossetti
 *
 */
public class DemandLoadBuildingNetworkTest1 {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// TODO Auto-generated method stub
		System.out.println("Load Building Network Test 1");
        
//		JSL.LOGGER.setLevel(Level.INFO);
   
		testReplication();
//		testExperiment();
        
		System.out.println("Done!");	
	}

	public static Model buildModel(){
		// create the containing model
		Model m = Model.createModel();
		
		// build the network
		
		//Define item types
		ItemType type1=new ItemType(m,"Type-1");
		ItemType type2=new ItemType(m,"Type-2");		

		//Define external supplier
		LeadTimeDemandFiller supplier = new LeadTimeDemandFiller(m);
		
		//Add manufacturing lead time at the supplier
		supplier.addLeadTime(type1, Constant.ONE);
		supplier.addLeadTime(type2, Constant.ONE);
		
		// make the warehouse
		InventoryHoldingPoint warehouse = new InventoryHoldingPoint(m, "warehouse");
		// hook the warehouse to the supplier
		warehouse.setDemandFiller(supplier);
		
		//add inventory to warehouse
		warehouse.addReorderPointReorderQuantityInventory(type1,4,1,20);
		warehouse.addReorderPointReorderQuantityInventory(type2,4,1,20);

		// make the retailer
		InventoryHoldingPoint retailer = new InventoryHoldingPoint(m, "R1");
		// hook the retailer to the warehouse
		retailer.setDemandFiller(warehouse);

		//add inventory to retailer
		retailer.addReorderPointReorderQuantityInventory(type1,2,1,10);
		retailer.addReorderPointReorderQuantityInventory(type2,2,1,10);

		// Define the demand generators
		DistributionIfc d1 = new Exponential(1.0);				
		DemandGenerator dg1 = new DemandGenerator(m, type1, d1, d1);
		// hook the generator to the retailer
		dg1.setDemandFiller(retailer);
		
		DistributionIfc d2 = new Exponential(1.0);				
		DemandGenerator dg2 = new DemandGenerator(m, type2, d2, d2);
		// hook the generator to the retailer
		dg2.setDemandFiller(retailer);
		
		// connect the load builders and demand carriers
		// note that we can use a separate DemandCarrierIfc to transport
		// between supplier and warehouse, warehouse to retailer or we can use
		// the same carrier as long as the correct interpretation of the transport times are made
		// in other words the carrier can be attached to both the supplier and the warehouse
		
		TimeBasedLoadCarrier carrier = new TimeBasedLoadCarrier(m, "Carrier");
		
		// load builder for supplier to warehouse
		DemandLoadBuilder lbsw = carrier.assignLoadBuilder(warehouse, "LB S to W");
		// reference to load builder allows specification of the load building process
		// alternatively a specific load builder could have been supplier to the carrier.

		// use count to form loads
		lbsw.setCountLimit(5);

		// load builder for warehouse to retailer
		DemandLoadBuilder lbwr = carrier.assignLoadBuilder(retailer, "LB W to R1");
		// reference to load builder allows specification of the load building process
		// alternatively a specific load builder could have been supplier to the carrier.

		// use count to form loads
		lbwr.setCountLimit(2);
		
		// give transport times, these can be distributions
		
		carrier.setTransportTime(warehouse, Constant.TWO);
		carrier.setTransportTime(retailer, Constant.ONE);
		
		// set it so that the carrier reacts to load formation, if this
		// is not true then the carrier is responsible for causing loads to be shipped
		carrier.setReactToLoadBuildersFlag(true);
		
		// attach carriers
		supplier.setDemandCarrier(carrier);
		warehouse.setDemandCarrier(carrier);
		
		return m;
	}
	
	public static void testExperiment(){
		
		Model m = buildModel();
		
		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
		// set the parameters of the experiment
		e.setNumberOfReplications(10);
		e.setLengthOfReplication(3650.0);
		e.setLengthOfWarmUp(100.0);
		
		// turn on the desired reporting
		e.turnOnExperimentConsoleOutput();
		e.turnOnExperimentReport();
		e.turnOnReplicationReport();
		
		// tell the experiment to run
		e.runAll();
	}
	
	public static void testReplication(){
		Model m = buildModel();
		
        // create a replication to run the model
		Replication r = new Replication(m);

		// set the running parameters of the replication
		r.setLengthOfReplication(50.0);
//		r.setLengthOfWarmUp(5.0);
		
		// turn on the desired reporting
	//	r.turnOnBatchReport();
		
		// tell the replication to run
		r.runAll();
	}
}
