/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.transportlayer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.spatial.SpatialResource;
import jslx.transportlayer.DriverSelectionRuleIfc;

/** Dispatcher will receive load from the Transportation Network. Dispatcher is responsible for searching idle trucks, drivers and trailers
 *  for the transport of the load. If the idle resouces are not available at the origin of the load, the dispatcher will reallocate the 
 *  resources in the network. This will be done by creating several tasks eg, driver will be asked to move empty (without load) without trailer 
 *  or move empty with trailer to make the processing of the load possible. The driver will be asked to perform the task. After the task 
 *  is completed the dispatcher will be notified and will attempt to reallocate the resources in the network.
 *  
 *  Default behaviors:
 *  receivesLoad(load)- Receives load from the TransportationNetwork
 *  taskCompleted()- After the task assigned by the dispatcher to the driver is completed
 *  
 *
 */
abstract public class AbstractDispatcherFTL extends AbstractDispatcher {

	//  ===========================================
	//  Data structures to hold the resources of the network
	//  ===========================================
	
	/** List of all Drivers in the network of the type Driver 
	 */
	private List<Driver> myDrivers;
	
	/** List of all Trailers in the network of the type Spatial Resource 
	 */
	private List<SpatialResource> myTrailers;
	
	/** List of all Trucks in the network of the type Spatial Resource 
	 */
	private List<SpatialResource> myTrucks;
	
	/** For each location, there is a list of initially assigned drivers
	 */
	private Map<TransportLocationIfc, List<Driver>> myDriverLocationAssignments;
	
	/** For each location, there is a list of initiallly assigned trucks
	 */
	private Map<TransportLocationIfc, List <SpatialResource>> myTruckLocationAssignments;
	
	/** For each location, there is a list of initially assigned trailers
	 */
	private Map<TransportLocationIfc, List<SpatialResource>> myTrailerLocationAssignments;
	
	/** For each location, there is a list of idle drivers
	 */
	private Map<TransportLocationIfc, List<Driver>> myIdleDriversByLocation;
	
	/** For each location, there is a list of idle trucks
	 */
	private Map<TransportLocationIfc, List <SpatialResource>> myIdleTrucksByLocation;
	
	/** For each location, there is a list of idle trailers
	 */
	private Map<TransportLocationIfc, List<SpatialResource>> myIdleTrailersByLocation;
	
	/** An instance of DriverSelectionRuleIfc 
	  */
	private DriverSelectionRuleIfc myDriverSelectionRule;

	/** An instance of DriverSelectionRuleIfc 
	  */
	private TruckSelectionRuleIfc myTruckSelectionRule;

	/** An instance of DriverSelectionRuleIfc 
	  */
	private TrailerSelectionRuleIfc myTrailerSelectionRule;
	
	private Comparator<TransportLocationIfc> myTrailerLocationComparator;
	
	private Comparator<TransportLocationIfc> myDriverTruckTrailerLocationComparator;
	
	private Comparator<TransportLocationIfc> myDriverTruckLocationComparator;
	
	/** The maximum default distance for considering deadhead or bobtails
	 * 
	 */
	private double myMaxDHBTDistance;

	/** For each location, there is a list of idle trailers
	 */
	private Map<TransportLocationIfc, Double> myMaxDHBTDistancesByLocation;

	public AbstractDispatcherFTL(ModelElement parent, TransportationNetwork network, double distance) {
		this(parent, network, distance, null);
	}

	public AbstractDispatcherFTL(ModelElement parent, TransportationNetwork network, double distance, String name) {
		super(parent, network, name);

		// set up the resource data structures
		myDrivers=new ArrayList<Driver>();
		myTrailers=new ArrayList<SpatialResource>();
		myTrucks=new ArrayList<SpatialResource>();
		myDriverLocationAssignments = new HashMap<TransportLocationIfc, List<Driver>>();
		myTruckLocationAssignments = new HashMap<TransportLocationIfc, List<SpatialResource>>();
		myTrailerLocationAssignments = new HashMap<TransportLocationIfc, List<SpatialResource>>();
		myIdleDriversByLocation = new HashMap<TransportLocationIfc, List<Driver>>();
		myIdleTrucksByLocation = new HashMap<TransportLocationIfc, List<SpatialResource>>();
		myIdleTrailersByLocation = new HashMap<TransportLocationIfc, List<SpatialResource>>();
		
		// fill up the distances
		//TODO need to worry about client adding locations after creating the dispatcher.
		setMaxDeadHeadBobTailDistance(distance);
		myMaxDHBTDistancesByLocation = new HashMap<TransportLocationIfc, Double>();
		Collection<TransportLocationIfc> locations = network.getOrigins();
		for(TransportLocationIfc location: locations){
			myMaxDHBTDistancesByLocation.put(location, new Double(distance));
		}

	}
	
	/** Assigns a driver to the supplied location and adds the driver to the network
	 * 
	 * @param location
	 */
	public final void addDriver(TransportLocationIfc location){
		if (location == null)
			throw new IllegalArgumentException("The supplied location was null.");

		if (!getNetwork().containsLocation(location))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		// create the driver
		Driver driver = new Driver(this, null, location.getCurrentPosition());
		
		// will hold the list of drivers for this location
		List<Driver> drivers = null;
		
		// get driver assignment list for this location
		if (!myDriverLocationAssignments.containsKey(location)){
			drivers = new ArrayList<Driver>();
			myDriverLocationAssignments.put(location, drivers);
		} else
			drivers = myDriverLocationAssignments.get(location);
		// add driver to assignment list
		drivers.add(driver);
		
		// get idle driver list by location
		if (!myIdleDriversByLocation.containsKey(location)){
			drivers = new ArrayList<Driver>();
			myIdleDriversByLocation.put(location, drivers);
		} else
			drivers = myIdleDriversByLocation.get(location);
		// add driver to idle driver list
		drivers.add(driver);
		// add driver to overall list of drivers
		myDrivers.add(driver);
		System.out.println("Driver added");
	}

	/** Assigns a trailer to the supplied location and adds the trailer to the network
	 * 
	 * @param location
	 */
	public final void addTrailer(TransportLocationIfc location){
		if (location == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!getNetwork().containsLocation(location))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		// create the trailer
		SpatialResource trailer = new SpatialResource(this, location.getCurrentPosition());
		
		// will hold the list of trailers for this location
		List<SpatialResource> trailers = null;
		
		// get trailer assignment list for this location
		if (!myTrailerLocationAssignments.containsKey(location)){
			trailers = new ArrayList<SpatialResource>();
			myTrailerLocationAssignments.put(location, trailers);
		} else
			trailers = myTrailerLocationAssignments.get(location);
		// add trailer to assignment list
		trailers.add(trailer);
		
		// get idle trailer list by location
		if (!myIdleTrailersByLocation.containsKey(location)){
			trailers = new ArrayList<SpatialResource>();
			myIdleTrailersByLocation.put(location, trailers);
		} else
			trailers = myIdleTrailersByLocation.get(location);
		// add trailer to idle trailer list
		trailers.add(trailer);
		// add trailer to overall list of trailers
		myTrailers.add(trailer);
		System.out.println("Trailer added");
	}

	/** Assigns a truck to the supplied location and adds the truck to the network
	 * 
	 * @param location
	 */
	public final void addTruck(TransportLocationIfc location){
		if (location == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!getNetwork().containsLocation(location))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		// create the truck
		SpatialResource truck = new SpatialResource(this, location.getCurrentPosition());
		
		// will hold the list of trucks for this location
		List<SpatialResource> trucks = null;
		
		// get truck assignment list for this location
		if (!myTruckLocationAssignments.containsKey(location)){
			trucks = new ArrayList<SpatialResource>();
			myTruckLocationAssignments.put(location, trucks);
		} else
			trucks = myTruckLocationAssignments.get(location);
		// add truck to assignment list
		trucks.add(truck);
		
		// get idle truck list by location
		if (!myIdleTrucksByLocation.containsKey(location)){
			trucks = new ArrayList<SpatialResource>();
			myIdleTrucksByLocation.put(location, trucks);
		} else
			trucks = myIdleTrucksByLocation.get(location);
		// add truck to idle truck list
		trucks.add(truck);
		// add truck to overall list of trucks
		myTrucks.add(truck);
		System.out.println("Truck added");
	}

	public final TrailerSelectionRuleIfc getTrailerSelectionRule() {
		return myTrailerSelectionRule;
	}
	
	public final void setTrailerSelectionRule(TrailerSelectionRuleIfc trailerSelectionRule) {
		myTrailerSelectionRule = trailerSelectionRule;
	}
	
	public final TruckSelectionRuleIfc getTruckSelectionRule() {
		return myTruckSelectionRule;
	}
	
	public final void setTruckSelectionRule(TruckSelectionRuleIfc truckSelectionRule) {
		myTruckSelectionRule = truckSelectionRule;
	}

	/**
	 * @return Returns the myDriverSelectionRule.
	 */
	public final DriverSelectionRuleIfc getDriverSelectionRule() {
		return myDriverSelectionRule;
	}

	/**
	 * @param myDriverSelectionRule The myDriverSelectionRule to set.
	 */
	public final void setDriverSelectionRule(DriverSelectionRuleIfc driverSelectionRule) {
		myDriverSelectionRule = driverSelectionRule;
	}
	
	public final boolean driverIsIdleAt(TransportLocationIfc location){
		List<Driver> drivers = myIdleDriversByLocation.get(location);
		if (drivers == null)
			return(false);
		else
			return(drivers.isEmpty());
	}

	public final boolean truckIsIdleAt(TransportLocationIfc location){
		List<SpatialResource> trucks = myIdleTrucksByLocation.get(location);
		if (trucks == null)
			return(false);
		else
			return(trucks.isEmpty());
	}

	public final boolean trailerIsIdleAt(TransportLocationIfc location){
		List<SpatialResource> trailers = myIdleTrailersByLocation.get(location);
		if (trailers == null)
			return(false);
		else
			return(trailers.isEmpty());
	}
	
	public final int getNumIdleDrivers(TransportLocationIfc location){
		return(myIdleDriversByLocation.get(location).size());
	}

	public final int getNumIdleTrucks(TransportLocationIfc location){
		return(myIdleTrucksByLocation.get(location).size());
	}
	
	public final int getNumIdleTrailers(TransportLocationIfc location){
		return(myIdleTrailersByLocation.get(location).size());
	}

	/** Given a collection of locations, this method finds the location that
	 *  has most idle drivers and trucks in usable combination. If no location
	 *  is found that has a usable (driver, truck) pair, then this method returns
	 *  null. The number of usable (driver, truck) pairs at a location is the 
	 *  min(#idle drivers, #idle trucks). This method is useful in identifying
	 *  a location that can be used as a place to initiate a bobtail
	 * 
	 * @param locations
	 * @return
	 */
	public TransportLocationIfc findLocationWithMostIdleDriversAndTrucks(Collection<TransportLocationIfc> locations){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(TransportLocationIfc location: locations){
			int nd = getNumIdleDrivers(location);
			int nt = getNumIdleTrucks(location);
			int nr = Math.min(nd, nt);
			if (nr > max){
				max = nr;
				mLoc = location;
			}
		}
		return(mLoc);
	}
	
	/** Given a collection of locations, this method returns a sorted set of locations
	 *  that have idle drivers and trucks in usable combination. If no location
	 *  is found that has a usable (driver, truck) pair, then this method returns
	 *  an empty set. The number of usable (driver, truck) pairs at a location is the 
	 *  min(#idle drivers, #idle trucks). This method is useful in identifying
	 *  locations that can be used as a place to initiate a bobtail
	 *  
	 * @param locations
	 * @return SortedSet<TransportLocationIfc>
	 */
	public SortedSet<TransportLocationIfc> getLocationsSortedByMostIdleDriversAndTrucks(
			Collection<TransportLocationIfc> locations){
		SortedSet<TransportLocationIfc> set = new TreeSet<TransportLocationIfc>(getDriverTruckLocationComparator());
		for(TransportLocationIfc loc: locations){
			if (getNumIdleTrailers(loc) > 0)
				set.add(loc);
		}
		return(set);
	}
	
	/** From a given origin, return a destination on its incoming lanes that has the most idle
	 *  driver, truck pairs within the specified distance If no destination has an idle pair then null is returned
	 *  The natural use of this method is to supply a list of incoming lanes.  The origin on an incoming
	 *  lane is the location that the resources might be sent from.
	 * 
	 * @param origin The location that wants to identify the resources
	 * @param distance
	 * @return location with most (driver, truck) resources
	 */
	public TransportLocationIfc getDestinationWithMostIdleDriversAndTrucks(TransportLocationIfc origin, double distance){
		return(getLocationWithMostIdleDriversAndTrucks(getNetwork().getIncomingLanes(origin), distance));
	}
	
	/** Given a collection of lanes find the origin that has the most usable idle driver, truck
	 *  pairs within the specified distance. If no lane origin has idle pairs then, null is returned
	 *  The natural use of this method is to supply a list of incoming lanes.  The origin on an incoming
	 *  lane is the location that the resources might be sent from.
	 *  
	 * @param lanes
	 * @param distance
	 * @return location with most (driver, truck) resources
	 * @return
	 */
	public TransportLocationIfc getLocationWithMostIdleDriversAndTrucks(Collection<LaneIfc> lanes, double distance){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc origin = lane.getOrigin();
				int nd = getNumIdleDrivers(origin);
				int nt = getNumIdleTrucks(origin);
				int nr = Math.min(nd, nt);
				if (nr > max){
					max = nr;
					mLoc = origin;
				}				
			}
		}
		return(mLoc);
	}

	/** Return the closest location within the given distance that has an idle (driver, truck) combination
	 * 
	 * @param origin
	 * @param distance
	 * @return
	 */
	public TransportLocationIfc getFirstLocationWithIdleDriversAndTrucks(TransportLocationIfc origin, double distance){
		return(getFirstLocationWithIdleDriversAndTrucks(getNetwork().getIncomingLanes(origin), distance));
	}
	
	/** Given a collection of lanes return the origin of the lane that has an idle (driver, truck) combination
	 *  within the specified distance. If no lane origin has idle pairs then, null is returned
	 *  The natural use of this method is to supply a sorted list of incoming lanes by distance.  The origin on an incoming
	 *  lane is the location that the resources might be sent from. The sort causes this method to 
	 *  return the closest location that has an idle (driver, truck) combination.
	 *  	
	 * @param lanes
	 * @param distance
	 * @return
	 *
	 */
	public TransportLocationIfc getFirstLocationWithIdleDriversAndTrucks(Collection<LaneIfc> lanes, double distance){
		TransportLocationIfc mLoc = null;
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc origin = lane.getOrigin();
				if (driverIsIdleAt(origin) && truckIsIdleAt(origin))
					return(origin);				
			}
		}
		return(mLoc);
	}
	
	/** Given a collection of locations, this method finds the location that
	 *  has the most idle trailers. If no location
	 *  is found that has idle trailer, then this method returns
	 *  null. This method is useful in identifying
	 *  a location that can be used as a place to send a bobtail.  The location
	 *  may or may not have loads to pick up.
	 * 
	 * @param locations
	 * @return
	 */
	public TransportLocationIfc findLocationWithMostIdleTrailers(Collection<TransportLocationIfc> locations){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(TransportLocationIfc location: locations){
			int nt = getNumIdleTrailers(location);
			if (nt > max){
				max = nt;
				mLoc = location;
			}
		}
		return(mLoc);
	}
	
	/** Given a collection of locations, this method returns a sorted set of locations
	 *  that have trailers. The set is sorted from most idle to least idle. If no location
	 *  is found that has idle trailers, then this method returns
	 *  an empty set.  This method is useful in identifying
	 *  locations that may be useful to send a bobtail to.  The locations may or may
	 *  not have loads waiting.
	 *  
	 * @param locations
	 * @return SortedSet<TransportLocationIfc>
	 */
	public SortedSet<TransportLocationIfc> getLocationsSortedByMostIdleTrailers(
			Collection<TransportLocationIfc> locations){
		SortedSet<TransportLocationIfc> set = new TreeSet<TransportLocationIfc>(getTrailerLocationComparator());
		for(TransportLocationIfc loc: locations){
			if (getNumIdleTrailers(loc) > 0)
				set.add(loc);
		}
		return(set);
	}

	/** From a given origin, return a destination on its outgoing lanes that has the most idle
	 *  trailers.  If no destination has an idle trailer then null is returned. This method
	 *  is useful in looking for an idle trailer to match with a location that might
	 *  be a bobtail initiation site. Do not consider locations further than the supplied distance.
	 * 
	 * @param origin
	 * @param distance
	 * @return
	 */
	public TransportLocationIfc getDestinationWithMostIdleTrailers(TransportLocationIfc origin, double distance){
		return(getLocationWithMostIdleTrailers(getNetwork().getOutGoingLanes(origin), distance));
	}

	/** From a collection of lanes return a destination that has the most idle
	 *  trailers.  If no destination has an idle trailer then null is returned. This method
	 *  is useful in looking for an idle trailer to match with a location that might
	 *  be a bobtail initiation site.  Do not consider locations further than the supplied distance
	 * 
	 * @param lanes
	 * @param distance
	 * @return
	 */
	public TransportLocationIfc getLocationWithMostIdleTrailers(Collection<LaneIfc> lanes, double distance){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc dest = lane.getDestination();
				int nt = getNumIdleTrailers(dest);
				if (nt > max){
					max = nt;
					mLoc = dest;
				}				
			}
		}
		return(mLoc);
	}
	
	/** Given a collection of locations, this method finds the location that
	 *  has most idle drivers, trucks, trailers in usable combination. If no location
	 *  is found, then this method returns null. The number of usable resources at a location is the 
	 *  min(#idle drivers, #idle trucks, #idle trailers). This method is useful in identifying
	 *  a location that can be used as a place to initiate a deadhead
	 * 
	 * @param locations
	 * @return
	 */
	public TransportLocationIfc findLocationWithMostIdleResources(Collection<TransportLocationIfc> locations){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(TransportLocationIfc location: locations){
			int nd = getNumIdleDrivers(location);
			int nt = getNumIdleTrucks(location);
			int nr = Math.min(nd, nt);
			nr = Math.min(nr, getNumIdleTrailers(location));
			if (nr > max){
				max = nr;
				mLoc = location;
			}
		}
		return(mLoc);
	}
	
	/** From a given location, return a origin on its incoming lanes that has the most idle
	 *  resources.  If no origin location has usable idle resources then null is returned. This method
	 *  is useful in looking for idle resources to match with a location may need a deadhead
	 *  The natural use of this method supplies a list of incoming lanes to the given origin.  
	 *  The origin on an incoming lane is the location that the resources might be sent from. 
	 * 
	 * @param origin The location that wants to identify the resources
	 * @param distance Only consider those locations within this distance
	 * @return
	 */
	public TransportLocationIfc getDestinationWithMostIdleResources(TransportLocationIfc origin, double distance){
		return(getLocationWithMostIdleResources(getNetwork().getIncomingLanes(origin), distance));
	}
	
	/** Return from a collection of lanes the origin location that has the most idle
	 *  resources.  If no destination has usable idle resources then null is returned. This method
	 *  is useful in looking for idle resources to match with a location may need a deadhead
	 * 
	 * @param lanes
	 * @param distance Only consider those locations within this distance
	 * @return
	 */
	public TransportLocationIfc getLocationWithMostIdleResources(Collection<LaneIfc> lanes, double distance){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc origin = lane.getOrigin();
				int nd = getNumIdleDrivers(origin);
				int nt = getNumIdleTrucks(origin);
				int nr = Math.min(nd, nt);
				nr = Math.min(nr, getNumIdleTrailers(origin));
				if (nr > max){
					max = nr;
					mLoc = origin;
				}				
			}
		}
		return(mLoc);
	}
	
	/** Given a collection of locations, this method finds a sorted set of location that
	 *  have the most idle drivers, trucks, trailers in usable combination. If no location
	 *  is found, then this method returns and emtpy set. The number of usable resources at a location is the 
	 *  min(#idle drivers, #idle trucks, #idle trailers). This method is useful in identifying
	 *  locations that can be used as a place to initiate a deadhead
	 * 
	 * @param locations
	 * @return
	 */
	public SortedSet<TransportLocationIfc> getLocationsSortedByMostIdleResources(
			Collection<TransportLocationIfc> locations){
		SortedSet<TransportLocationIfc> set = new TreeSet<TransportLocationIfc>(getDriverTruckTrailerLocationComparator());
		for(TransportLocationIfc loc: locations){
			if (getNumIdleTrailers(loc) > 0)
				set.add(loc);
		}		
		return(set);
	}

	/** Return the closest location within the given distance that has an idle (driver, truck, trailer) combination
	 * 
	 * @param origin
	 * @param distance
	 * @return
	 */
	public TransportLocationIfc getFirstLocationWithIdleResources(TransportLocationIfc origin, double distance){
		return(getFirstLocationWithIdleResources(getNetwork().getIncomingLanes(origin), distance));
	}
	
	/** Given a collection of lanes return the origin of the lane that has an idle (driver, truck, trailer) combination
	 *  within the specified distance. If no lane origin has idle resources then, null is returned
	 *  The natural use of this method is to supply a sorted list of incoming lanes by distance.  The origin on an incoming
	 *  lane is the location that the resources might be sent from. The sort causes this method to 
	 *  return the closest location that has an idle (driver, truck, trailer) combination.
	 *  	
	 * @param lanes
	 * @param distance
	 * @return
	 *
	 */
	public TransportLocationIfc getFirstLocationWithIdleResources(Collection<LaneIfc> lanes, double distance){
		TransportLocationIfc mLoc = null;
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc origin = lane.getOrigin();
				if (driverIsIdleAt(origin) && truckIsIdleAt(origin) && trailerIsIdleAt(origin))
					return(origin);				
			}
		}
		return(mLoc);
	}
	
	public String toString(){
		StringBuffer s = new StringBuffer();
		
		s.append("----- Driver Assignments ----- \n");
		
		for(TransportLocationIfc location: myDriverLocationAssignments.keySet()){
			int i = getNetwork().getIndexOf(location);
			s.append("Location" + i + "\t" + location.getName() + " : \t");
			for(Driver d: myDriverLocationAssignments.get(location)){
				s.append(d.getId() + " " + d.getName() + " ");
			}
			s.append("\n");
		}

		s.append("----- Truck Assignments ----- \n");
		
		for(TransportLocationIfc location: myTruckLocationAssignments.keySet()){
			int i = getNetwork().getIndexOf(location);
			s.append("Location" + i + "\t" + location.getName() + " : \t");
			for(SpatialResource r: myTruckLocationAssignments.get(location)){
				s.append(r.getId() + " " + r.getName() + " ");
			}
			s.append("\n");
		}

		s.append("----- Trailer Assignments ----- \n");
		
		for(TransportLocationIfc location: myTrailerLocationAssignments.keySet()){
			int i = getNetwork().getIndexOf(location);
			s.append("Location" + i + "\t" + location.getName() + " : \t");
			for(SpatialResource r: myTrailerLocationAssignments.get(location)){
				s.append(r.getId() + " " + r.getName() + " ");
			}
			s.append("\n");
		}

		return(s.toString());
	}

	//  ===========================================
	//  PROTECTED METHODS
	//  ===========================================

	/**
	 * Returns the list of Drivers 
	 * @return 
	 */
	protected final List<Driver> getDrivers() {
		return myDrivers;
	}
	
	/**
	 * Returns the list of Trailers
	 * @return   
	 */
	protected final List<SpatialResource> getTrailers() {
		return myTrailers;
	}
	
	/**
	 * Returns the list of Trucks
	 * @return  
	 */
	protected final List<SpatialResource> getTrucks() {
		return myTrucks;
	}
	
	/**
	 * @return Returns the maxDHBTDistance.
	 */
	public final double getMaxDeadHeadBobTailDistance() {
		return myMaxDHBTDistance;
	}

	/**
	 * @param myMaxDHBTDistance The maxDHBTDistance to set.
	 */
	protected final void setMaxDeadHeadBobTailDistance(double maxDHBTDistance) {
		//TODO maybe check to see if it is >= than smallest lane and <= than largest lane
		if (maxDHBTDistance <= 0)
			throw new IllegalArgumentException("The supply distance must be > 0");
		
		myMaxDHBTDistance = maxDHBTDistance;
	}

	protected final void setMaxDeadHeadBobTailDistance(TransportLocationIfc location, double distance){
		//TODO maybe check to see if it is >= than smallest incoming lane and <= than largest incoming lane
		myMaxDHBTDistancesByLocation.put(location, new Double(distance));
	}

	protected final double getMaxDeadHeadBobTailDistance(TransportLocationIfc location){
		return(Double.valueOf(myMaxDHBTDistancesByLocation.get(location)));
	}
	
	protected final List<Driver> getAssignedDrivers(TransportLocationIfc location){
		return(myDriverLocationAssignments.get(location));
	}
	
	protected final List<SpatialResource> getAssignedTrucks(TransportLocationIfc location){
		return(myTruckLocationAssignments.get(location));
	}
	
	protected final List<SpatialResource> getAssignTrailers(TransportLocationIfc location){
		return(myTrailerLocationAssignments.get(location));
	}

	protected final List<Driver> getIdleDrivers(TransportLocationIfc location){
		return(myIdleDriversByLocation.get(location));
	}
		
	protected final List<SpatialResource> getIdleTrucks(TransportLocationIfc location){
		return(myIdleTrucksByLocation.get(location));
	}

	protected final List<SpatialResource> getIdleTrailers(TransportLocationIfc location){
		return(myIdleTrailersByLocation.get(location));
	}
	
	protected Driver getIdleDriver(TransportLocationIfc location){
		
		if(myDriverSelectionRule != null){
			return(myDriverSelectionRule.selectNextDriver(this, location));
		}
		else{
			List<Driver> drivers = myIdleDriversByLocation.get(location);
			if(drivers.isEmpty())
				return (null);
			else {
				Driver d = drivers.remove(0);
				return(d);
			}
		}
	}
		
	protected void releaseDriver(DispatchTask task){
		task.releaseDriver();
		List<Driver> drivers = myIdleDriversByLocation.get(task.getLoadDestination());
		drivers.add(task.getDriver());
	}

	protected SpatialResource getIdleTruck(TransportLocationIfc location){
		
		if(myTruckSelectionRule != null){
			return(myTruckSelectionRule.selectNextTruck(this, location));
		}
		else{
			List<SpatialResource> trucks = myIdleTrucksByLocation.get(location);
			if(trucks.isEmpty())
				return (null);
			else {
				SpatialResource truck = trucks.remove(0);
				return(truck);
			}
		}
	}
	
	protected void releaseTruck(DispatchTask task){
		task.releaseTruck();
		List<SpatialResource> trucks = myIdleTrucksByLocation.get(task.getLoadDestination());
		trucks.add(task.getTruck());
	}
	
	protected SpatialResource getIdleTrailer(TransportLocationIfc location){
		
		if(myTrailerSelectionRule != null){
			return(myTrailerSelectionRule.selectNextTrailer(this, location));
		}
		else{
			List<SpatialResource> trailers = myIdleTrailersByLocation.get(location);
			if(trailers.isEmpty())
				return (null);
			else {
				SpatialResource trailer = trailers.remove(0);
				return(trailer);
			}
		}
	}
	
	protected void releaseTrailer(DispatchTask task){
		task.releaseTrailer();
		List<SpatialResource> trailers = myIdleTrailersByLocation.get(task.getLoadDestination());
		trailers.add(task.getTrailer());
	}

	/** Trys to find resources at the task's origin that will permit
	 *  a transport to occur.  Modifies the resources of the task
	 *  if all the resources are available for a transport.
	 * 
	 * @param task
	 * @return True if transport resources have been identified
	 */
	protected boolean transportAvailable(DispatchTask task){
		TransportLocationIfc location = task.getLoadOrigin();
		Driver driver = getIdleDriver(location);
		if (driver == null)
			return false;
		SpatialResource truck = getIdleTruck(location);
		if (truck == null)
			return false;
		SpatialResource trailer = getIdleTrailer(location);
		if (trailer == null)
			return false;
		// if here, then a driver, truck and trailer has been found
		task.setDriver(driver);
		task.setTruck(truck);
		task.setTrailer(trailer);
		return true;
	}

	/** When the DispatchTask is completed then this method can be called back
	 *  to indicate the completion of the task
	 * 
	 * @param task
	 */
	abstract protected void dispatchCompleted(DispatchTask task);

	//  ===========================================
	//  INNER CLASSES
	//  ===========================================

	/** Compares two locations.  The location with the most usuable (idle)
	 *  (driver, truck) resources is considered bigger. 
	 *  Ties go to lower location number
	 *  
	 *  Equal is only via object equality
	 *
	 * Note:  This class may have a natural ordering that is inconsistent
	 * with equals.
	 * 
	 * @param loc1
	 * @param loc2
	 * @return Returns a negative integer, zero, or a positive integer
	 *
	 */ 
	public class DriverTruckLocationComparator implements Comparator<TransportLocationIfc>{

		public int compare(TransportLocationIfc loc1, TransportLocationIfc loc2) {
	        if (loc1.equals(loc2))// objects are equal with zero
	            return(0);
	        
	        List<Driver> drivers1 = getIdleDrivers(loc1);
	        List<SpatialResource> trucks1 = getIdleTrucks(loc1);
	        List<Driver> drivers2 = getIdleDrivers(loc2);
	        List<SpatialResource> trucks2 = getIdleTrucks(loc2);
	        // the number of usable resources is constrained by the smaller of the two
	        int numIdle1 = Math.min(drivers1.size(), trucks1.size());
	        int numIdle2 = Math.min(drivers2.size(), trucks2.size());
	        
	        if (numIdle1 > numIdle2)
	        	return(1);
	        else if (numIdle1 < numIdle2)
        		return(-1);
	        else {
				long id1 = loc1.getId();
				long id2 = loc2.getId();
		        	if (id1 < id2)
		        		return(-1);
		        	else
		        		return(1);	        	
			}	        	
		}	
	}

	/** Compares two locations.  The location with the most usuable (idle)
	 *  (driver, truck, trailer) resources is considered bigger. 
	 *  Ties go to lower location number
	 *  
	 *  Equal is only via object equality
	 *
	 * Note:  This class may have a natural ordering that is inconsistent
	 * with equals.
	 * 
	 * @param loc1
	 * @param loc2
	 * @return Returns a negative integer, zero, or a positive integer
	 *
	 */ 
	public class DriverTruckTrailerLocationComparator implements Comparator<TransportLocationIfc>{

		public int compare(TransportLocationIfc loc1, TransportLocationIfc loc2) {
	        if (loc1.equals(loc2))// objects are equal with zero
	            return(0);
	        
	        List<Driver> drivers1 = getIdleDrivers(loc1);
	        List<SpatialResource> trucks1 = getIdleTrucks(loc1);
	        List<SpatialResource> trailers1 = getIdleTrailers(loc1);
	        List<Driver> drivers2 = getIdleDrivers(loc2);
	        List<SpatialResource> trucks2 = getIdleTrucks(loc2);
	        List<SpatialResource> trailers2 = getIdleTrailers(loc2);
	        // the number of usable resources is constrained by the smaller of the two
	        int numIdle1 = Math.min(drivers1.size(), trucks1.size());
	        numIdle1 = Math.min(numIdle1, trailers1.size());
	        int numIdle2 = Math.min(drivers2.size(), trucks2.size());
	        numIdle2 = Math.min(numIdle2, trailers2.size());
	        
	        if (numIdle1 > numIdle2)
	        	return(1);
	        else if (numIdle1 < numIdle2)
        		return(-1);
	        else {
				long id1 = loc1.getId();
				long id2 = loc2.getId();
		        	if (id1 < id2)
		        		return(-1);
		        	else
		        		return(1);	        	
			}			
		}
	}

	/** Compares two locations.  The location with the most usuable (idle)
	 *  (trailer) resources is considered bigger. Ties go to lower location number
	 *  
	 *  Equal is only via object equality
	 *
	 * Note:  This class may have a natural ordering that is inconsistent
	 * with equals.
	 * 
	 * @param loc1
	 * @param loc2
	 * @return Returns a negative integer, zero, or a positive integer
	 *
	 */ 
	public class TrailerLocationComparator implements Comparator<TransportLocationIfc>{

		public int compare(TransportLocationIfc loc1, TransportLocationIfc loc2) {
	        if (loc1.equals(loc2))// objects are equal with zero
	            return(0);
	        
	        List<SpatialResource> trailers1 = getIdleTrailers(loc1);
	        List<SpatialResource> trailers2 = getIdleTrailers(loc2);
	        int n1 = trailers1.size();
	        int n2 = trailers2.size();
	        
	        if (n1 > n2)
	        	return(1);
	        else if (n1 < n2)
        		return(-1);
	        else {
				long id1 = loc1.getId();
				long id2 = loc2.getId();
		        	if (id1 < id2)
		        		return(-1);
		        	else
		        		return(1);	        	
			}			
        	
		}	
	}
	
	protected final Comparator<TransportLocationIfc> getDriverTruckLocationComparator() {
		if (myDriverTruckLocationComparator == null)
			myDriverTruckLocationComparator = new DriverTruckLocationComparator();
		return myDriverTruckLocationComparator;
	}
	
	protected final void setDriverTruckLocationComparator(
			Comparator<TransportLocationIfc> driverTruckLocationComparator) {
		myDriverTruckLocationComparator = driverTruckLocationComparator;
	}
	
	protected final Comparator<TransportLocationIfc> getDriverTruckTrailerLocationComparator() {
		if (myDriverTruckTrailerLocationComparator == null)
			myDriverTruckTrailerLocationComparator = new DriverTruckTrailerLocationComparator();
		return myDriverTruckTrailerLocationComparator;
	}
	
	protected final void setDriverTruckTrailerLocationComparator(
			Comparator<TransportLocationIfc> driverTruckTrailerLocationComparator) {
		myDriverTruckTrailerLocationComparator = driverTruckTrailerLocationComparator;
	}
	
	protected final Comparator<TransportLocationIfc> getTrailerLocationComparator() {
		if (myTrailerLocationComparator == null)
			myTrailerLocationComparator = new TrailerLocationComparator();
		return myTrailerLocationComparator;
	}
	
	protected final void setTrailerLocationComparator(
			Comparator<TransportLocationIfc> trailerLocationComparator) {
		myTrailerLocationComparator = trailerLocationComparator;
	}
	
}
