/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.transportlayer;

import jsl.modeling.elements.queue.QObject;
import jsl.modeling.elements.resource.Request;
import jsl.modeling.elements.spatial.SpatialResource;
import jsl.utilities.reporting.JSL;

/** The dispatcher uses DispatchTasks to make moves within the network
 * 
 *
 */
public class DispatchTask extends QObject{

	//  ===========================================
	//  CLASS ATTRIBUTES
	//  ===========================================
	
	/** Constant to represent that the DispatchTask is a bob tail
	 *  that has been dispatched to pick up a load
	 */
	public static final int BOBTAIL_TO_LOAD = JSL.getNextEnumConstant();

	/** Constant to represent that the DispatchTask is a dead head 
	 *  that has been dispatched to pick up a load
	 */
	public static final int DEADHEAD_TO_LOAD = JSL.getNextEnumConstant();

	/** Constant to represent that the DispatchTask is a bob tail
	 *  that has been dispatched but there is no load to pick up
	 *  There can be a no load bob tail when a driver and truck
	 *  have simply been told to move to another location
	 */
	public static final int BOBTAIL_NO_LOAD = JSL.getNextEnumConstant();

	/** Constant to represent that the DispatchTask is a dead head that
	 *  has been dispatched with out a corresponding load.  The deadhead
	 *  is simply to move the driver, truck, trailer to another location
	 *  w/o picking up a load when they arrive
	 */
	public static final int DEADHEAD_NO_LOAD = JSL.getNextEnumConstant();

	/** Constant to represent that the DispatchTask is a transport
	 */
	public static final int TRANSPORT = JSL.getNextEnumConstant();

	//  ===========================================
	//  OBJECT ATTRIBUTES
	//  ===========================================
		
	/** Represents dispatcher of the task
	 */
	private AbstractDispatcher myDispatcher;
		
	/** Represents the destination of the task 
	 */
	private TransportLocationIfc myDestination;

	/** Represents the destination of the task 
	 */
	private TransportLocationIfc myOrigin;

	/** Represents the driver associated with the dispatch task 
	 */
	private Driver myDriver;
	
	/** The request make for the driver resource
	 */
	private Request myDriverRequest;
	
	/** Represents the truck associated with the dispatch task 
	 */
	private SpatialResource myTruck;
	
	/** The request made on the truck resource
	 */
	private Request myTruckRequest;

	/** Represents the load associated with task if any 
	 */
	private Load myLoad;

	/** Represents the trailer associated with the load 
	 */
	private SpatialResource myTrailer;

	/** Represents the request for the trailer resource
	 */
	private Request myTrailerRequest;
	
	/** Represents the type (BOBTAIL_TO_LOAD, DEADHEAD_TO_LOAD, 
	 * BOBTAIL_NO_LOAD, DEADHEAD_NO_LOAD, TRANSPORT)
	 */
	private int myTaskType;

	//  ===========================================
	//  CONSTRUCTORS
	//  ===========================================
	
	/** A DispatchTask is created by dispatchers to carry out moves in the network
	 * 
	 * @param dispatcher
	 * @param destination
	 */
	public DispatchTask(AbstractDispatcher dispatcher, Load load){
		setDispatcher(dispatcher);
		setLoad(load);
		
		System.out.println("Task created at "+getTime());
	}
	
	//  ===========================================
	//  PUBLIC METHODS
	//  ===========================================
	
	/** Gets destination of the task 
	 * 
	 * @return Destination
	 */
	public final TransportLocationIfc getLoadDestination(){
		return myLoad.getDestination();
	}

	/** Gets origin of the task 
	 * 
	 * @return Destination
	 */
	public final TransportLocationIfc getLoadOrigin(){
		return myLoad.getOrigin();
	}

	public final boolean hasDriver(){
		return(myDriver != null);
	}
	
	public final boolean hasTruck(){
		return(myTruck != null);
	}
	
	public final boolean hasTrailer() {
		return(myTrailer != null);
	}

	public final boolean hasDriverAllocated(){
		return(myDriverRequest != null);
	}

	public final boolean hasTruckAllocated(){
		return(myTruckRequest != null);
	}

	public final boolean hasTrailerAllocated(){
		return(myTrailerRequest != null);
	}

	/**
	 * @return Returns the myLoad.
	 */
	public final Load getLoad() {
		return myLoad;
	}

	public final void setTypeToBobTailToLoad(){
		myTaskType = BOBTAIL_TO_LOAD;
	}
	public final void setTypeToDeadHeadToLoad(){
		myTaskType = DEADHEAD_TO_LOAD;
	}

	public final void setTypeToBobTailNoLoad(){
		myTaskType = BOBTAIL_NO_LOAD;
	}
	public final void setTypeToDeadHeadNoLoad(){
		myTaskType = DEADHEAD_NO_LOAD;
	}

	public final void setTypeToTransport(){
		setOrigin(getLoadOrigin());
		myTaskType = TRANSPORT;
	}
	
	public final boolean areOriginsSame(){
		return(getOrigin() == getLoadOrigin());
	}
	
	public final boolean isBobTailToLoad(){
		return(myTaskType == BOBTAIL_TO_LOAD);
	}
	
	public final boolean isDeadHeadToLoad(){
		return(myTaskType == DEADHEAD_TO_LOAD);
	}

	public final boolean isBobTailNoLoad(){
		return(myTaskType == BOBTAIL_NO_LOAD);
	}
	
	public final boolean isDeadHeadNoLoad(){
		return(myTaskType == DEADHEAD_NO_LOAD);
	}

	public final boolean isTransport(){
		return(myTaskType == TRANSPORT);	
	}
	
	public final int getTaskType(){
		return(myTaskType);
	}
			
	//  ===========================================
	//  PROTECTED METHODS
	//  ===========================================
	
	/** Sets the dispatcher for the task
	 * 
	 * @param dispatcher The dispatcher to set, must not be null
	 */
	protected final void setDispatcher(AbstractDispatcher dispatcher){
		if (dispatcher == null)
			throw new IllegalArgumentException("The dispatcher was null!");
		myDispatcher = dispatcher;
	}

	/** Gets the dispatcher for this task
	 * @return
	 */
	protected final AbstractDispatcher getDispatcher(){
		return myDispatcher;
	}	

	protected final void setDriver(Driver driver) {
		myDriver = driver;
	}

	protected final Driver getDriver() {
		return myDriver;
	}

	protected final void setDriverRequest(Request driverRequest) {
		myDriverRequest = driverRequest;
	}

	protected final Request getDriverRequest() {
		return myDriverRequest;
	}

	protected final void setTruck(SpatialResource truck) {
		myTruck = truck;
	}
	
	protected final SpatialResource getTruck() {
		return myTruck;
	}
	
	protected final void setTruckRequest(Request truckRequest) {
		myTruckRequest = truckRequest;
	}

	protected final Request getTruckRequest() {
		return myTruckRequest;
	}

	protected final SpatialResource getTrailer() {
		return myTrailer;
	}
	
	protected final void setTrailer(SpatialResource trailer) {
		myTrailer = trailer;
	}

	protected final Request getTrailerRequest() {
		return myTrailerRequest;
	}

	protected final void setTrailerRequest(Request trailerRequest) {
		myTrailerRequest = trailerRequest;
	}

	/** Sets Load for the Task
	 * @param myLoad 
	 */
	protected final void setLoad(Load load) {
		if (load == null)
			throw new IllegalArgumentException("The load cannot be null!");
		myLoad = load;
	}

	/** Sets destination of the task
	 * 
	 * @param load
	 */
	protected final void setDestination(TransportLocationIfc destination){
		if (destination == null)
			throw new IllegalArgumentException("The destination was null!");
		myDestination= destination;
	}

	/**
	 * @return
	 */
	protected final TransportLocationIfc getDestination() {
		return myDestination;
	}

	/**
	 * @return
	 */
	protected final TransportLocationIfc getOrigin() {
		return myOrigin;
	}

	/** Sets destination of the task
	 * 
	 * @param load
	 */
	protected final void setOrigin(TransportLocationIfc origin){
		if (origin == null)
			throw new IllegalArgumentException("The destination was null!");
		myOrigin = origin;
	}
	
	protected void waitingForTransport(){
		//TODO probably should check if it is a transport task and that it has its resources??
		// tell the driver to perform the task
		myDriver.performTask(this);
	}
	
	protected final void allocateDriver(){
		myDriverRequest = myDriver.allocate();		
	}
	
	protected final void allocateTruck(){
		myTruckRequest = myTruck.allocate();		
	}
	
	protected final void allocateTrailer(){
		myTrailerRequest = myTrailer.allocate();		
	}
	
	protected void releaseDriver(){
		myDriver.release(myDriverRequest);
	}
	
	protected void releaseTruck(){
		myTruck.release(myTruckRequest);
	}
	
	protected void releaseTrailer(){
		myTrailer.release(myTrailerRequest);
	}
		
	protected void taskCompleted(){
		myDispatcher.dispatchCompleted(this);
	}
/*	
	protected void executeTransport(){
		// make the task a transport type
		setTypeToTransport();
		// the dispatch load has identified the required resources
		// tell the task to allocate its resources
		allocateDriver();
		allocateTruck();
		allocateTrailer();
		// tell the driver to perform the task
		myDriver.performTask(this);
	}
*/		
}
