/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.transportlayer;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import jsl.modeling.ModelElement;
import jsl.modeling.elements.queue.Queue;

/**
 *
 */
public class DispatcherFTL extends AbstractDispatcherFTL {

	/** Queue for holding the loads waiting to be dispatched by location
	 */
	protected Map<TransportLocationIfc, Queue> myDispatchQByLocation; //TODO need a way to specify these

	//TODO need to add methods for easily enqueuing and dequeing from these queues
	/** Queue for holding the bobtail loads leaving by location
	 */
	protected Map<TransportLocationIfc, Queue> myBobTailPickupQByLocation; //TODO need a way to specify these

	/** Queue for holding the deadhead loads leaving by location
	 */
	protected Map<TransportLocationIfc, Queue> myDeadHeadPickupQByLocation; //TODO need a way to specify these

	/** Queue for holding the transport loads leaving by location
	 */
	protected Map<TransportLocationIfc, Queue> myTransportPickupQByLocation; //TODO need a way to specify these

	/** Queue for holding the transport loads inbound by location
	 */
	protected Map<TransportLocationIfc, Queue> myIncomingLoadsByLocation; //TODO need a way to specify these

	/** Holds loads to be dispatched
	 */
	protected List<DispatchTask> myDispatchQ;
		
	public DispatcherFTL(ModelElement parent, TransportationNetwork network, double distance) {
		this(parent, network, distance, null);
	}
	
	public DispatcherFTL(ModelElement parent, TransportationNetwork network, double distance, String name) {
		super(parent, network, distance, name);
		
		myDispatchQ = new LinkedList<DispatchTask>();
		
		// make all the queues for each possible origin location
		//TODO need to worry about client adding locations after creating the dispatcher.
		myDispatchQByLocation = new HashMap<TransportLocationIfc, Queue>();
		Collection<TransportLocationIfc> locations = network.getOrigins();
		for(TransportLocationIfc location: locations){
			Queue queue = new Queue(this, location.getId() + " " + location.getName() + " DispatchQ " + network.getIndexOf(location));
			myDispatchQByLocation.put(location, queue);

			queue = new Queue(this, location.getId() + " " + location.getName() + " BobTailPickupQ " + network.getIndexOf(location));
			myBobTailPickupQByLocation.put(location, queue);

			queue = new Queue(this, location.getId() + " " + location.getName() + " DeadHeadPickupQ " + network.getIndexOf(location));
			myDeadHeadPickupQByLocation.put(location, queue);

			queue = new Queue(this, location.getId() + " " + location.getName() + " TransportPickupQ " + network.getIndexOf(location));
			myTransportPickupQByLocation.put(location, queue);

			queue = new Queue(this, location.getId() + " " + location.getName() + " IncomingLoadQ " + network.getIndexOf(location));
			myIncomingLoadsByLocation.put(location, queue);
		}		
	}

	//  ===========================================
	//  PUBLIC METHODS
	//  ===========================================

	/** Returns true if there is an inbound transport load coming to the location
	 * 
	 * @param location
	 * @return
	 */
	public boolean isLoadInbound(TransportLocationIfc location){
		Queue queue = myIncomingLoadsByLocation.get(location);
		if (queue == null)
			return(false);
		else {
			return(!queue.isEmpty());
		}
	}
	
	/** Searches all the dispatch queues by location to find the location of the queue
	 *  that has the largest number of waiting tasks. The search is performed
	 *  in order of locations and ties with the max are ignored (the first max 
	 *  found is the location returned). If you need all tied locations, use 
	 *  findListOfLocationsWithLargestDispatchQ()
	 * 
	 * @return
	 */
	public TransportLocationIfc findLocationOfLargestDispatchQ(){
		TransportLocationIfc locmaxQ = null;
		int max = -1;
		for (TransportLocationIfc location: myDispatchQByLocation.keySet()){
			 Queue q = myDispatchQByLocation.get(location);
			 if (q.size() > max){
			 	max = q.size();
			 	locmaxQ = location;
			 }
		}
		return(locmaxQ);
	}
	
	/** Searches all the dispatch queues by location to find the locations of the queues
	 *  that have the largest number of waiting tasks. The search is performed
	 *  in order of locations.
	 * 
	 * @return The list of queues with the maximum number of tasks
	 */
	public List<TransportLocationIfc> findListOfLocationsWithLargestDispatchQ(){
		List<TransportLocationIfc> list = new LinkedList<TransportLocationIfc>();
		int max = -1;
		for (TransportLocationIfc location: myDispatchQByLocation.keySet()){
			 Queue q = myDispatchQByLocation.get(location);
			 if (q.size() > max){
			 	// a new max was found, clear the list
			 	list.clear();
			 	// add the new location
			 	list.add(location);
			 	// remember the new maximum
			 	max = q.size();
			 } else {
			 	if (q.size() == max){ // ties the max
			 		list.add(location);
			 	}
			 }
		}
		return(list);
	}

	public TransportLocationIfc findLocationNeedingBobTail(TransportLocationIfc location){
		double d = this.getMaxDeadHeadBobTailDistance(location);
		return(findLocationNeedingBobTail(location, d));
	}

	public TransportLocationIfc findLocationNeedingBobTail(TransportLocationIfc location, double distance){

		Collection<LaneIfc> lanes = getNetwork().getOutGoingLanes(location);
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc dest = lane.getDestination();
				if (isBobTailNeededAt(dest))
					return dest;				
			}
		}

		return(null);
	}

	public TransportLocationIfc findLocationNeedingDeadHead(TransportLocationIfc location){
		double d = this.getMaxDeadHeadBobTailDistance(location);
		return(findLocationNeedingDeadHead(location, d));
	}

	public TransportLocationIfc findLocationNeedingDeadHead(TransportLocationIfc location, double distance){

		Collection<LaneIfc> lanes = getNetwork().getOutGoingLanes(location);
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc dest = lane.getDestination();
				if (isDeadHeadNeededAt(dest))
					return dest;				
			}
		}

		return(null);
	}
	
	public boolean isBobTailNeededAt(TransportLocationIfc location){
		boolean needed = false;
		
		if (isLoadWaitingAt(location)){ // load is waiting
			if (!isLoadInbound(location)){ // no transport inbound
				if (trailerIsIdleAt(location)) // trailer is idle
					needed = true;
			}
		}
		return(needed);
	}
	
	public boolean isDeadHeadNeededAt(TransportLocationIfc location){
		boolean needed = false;
		
		if (isLoadWaitingAt(location)){ // load is waiting
			if (!isLoadInbound(location)){ // no transport inbound
				needed = true;
			}
		}
		return(needed);
	}
	
	public boolean isLoadWaitingAt(TransportLocationIfc location){
		Queue queue = myDispatchQByLocation.get(location);
		if (queue == null)
			return(false);
		else {
			return(queue.isEmpty());
		}
	}
	
	public int getNumberWaitingAt(TransportLocationIfc location){
		Queue queue = myDispatchQByLocation.get(location);
		if (queue == null)
			return(0);
		else {
			return(queue.size());
		}
	}

	//  ===========================================
	//  PROTECTED METHODS
	//  ===========================================

	protected void initialize(){
		myDispatchQ.clear();
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.AbstractDispatcher#receive(transportlayer.Load)
	 */
	protected void receive(Load load){
		
		System.out.println("Dispatcher received load");
		
		// create a task for the load
		DispatchTask task = new DispatchTask(this, load);
		
		// place the task in the task queues
		// to ensure collection of waiting time
		enqueue(task);
		
		// allow allocation mechanism to work 
		allocateResourcesAfterIncomingLoad(task);
		
	}
	
	protected void enqueue(DispatchTask task){
		// add the task to overall queue
		myDispatchQ.add(task);
		// add the task by location to the queue at the load's origin
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myDispatchQByLocation.get(location);
		queue.enqueue(task);
	}
	
	protected void dequeue(DispatchTask task){
		// remove it from overall queue
		myDispatchQ.remove(task);
		// remove it by location
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myDispatchQByLocation.get(location);
		queue.remove(task);
	}
			
	protected void dispatchTransport(DispatchTask task){
		// make the task a transport type
		task.setTypeToTransport();
		
		//TODO need to place the task in the transport pick up queue
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myTransportPickupQByLocation.get(location);
		queue.enqueue(task);

		// the dispatch load has identified the required resources
		// tell the task to allocate its resources
		task.allocateDriver();
		task.allocateTruck();
		task.allocateTrailer();
		// tell the task to execute
		task.waitingForTransport();
	}
	
	/** This method needs to be called when the driver completes
	 *  the loading of the task
	 * 
	 * @param task
	 */
	protected void loadingCompleted(DispatchTask task){
		//TODO The driver has completed loading, 
		// the task needs to be removed from the queue that it is waiting in
		// and placed in the incoming load queue for its destination
		
		// get the queue that the task is currently waiting in
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myTransportPickupQByLocation.get(location);
		// now remove it
		queue.remove(task); 
		// get the load's destination
		location = task.getLoadDestination();
		// get the incoming load queue for the destination
		queue = myIncomingLoadsByLocation.get(location);
		// put the task in the incoming load queue
		queue.enqueue(task);
	}
		
	/** This allocation method assumes that if all of the resources are available
	 *  for the incoming task, then it should be immediately dispatched. If
	 *  all the resources for the load are not immediately available the load task
	 *  waits in the dispatch queues.
	 *  
	 *  This behavior may be overridden.  For example, the dispatcher may want to
	 *  wait for some condition before dispatching the task even if resources
	 *  are immediately available
	 * 
	 * @param loadTask
	 */
	protected void allocateResourcesAfterIncomingLoad(DispatchTask incomingTask) {
		
		// This allocation mechanism assumes that if the resources
		// are available for the load at its location, then it
		// should be immediately dispatched.
				
		// check if the incoming load can be dispatched
		if (transportAvailable(incomingTask)){
			// remove the task and dispatch it
			dequeue(incomingTask);
			// all resources have been set, so dispatch transport task
			dispatchTransport(incomingTask);
		}else {
			TransportLocationIfc origin = incomingTask.getLoadOrigin();
			if(trailerIsIdleAt(origin)){
				if (!isLoadInbound(origin))// only ask for bobtail if no inbound loads
					bobTailNeededAfterIncomingLoad(incomingTask);
			}else {
				if (!isLoadInbound(origin)) // only ask for bobtail if no inbound loads
					deadHeadNeededAfterIncomingLoad(incomingTask);
			}
		}
	}
		
	/** If not all resources are available for the incoming task,
	 *  then if a trailer is idle at the location, the dispatcher
	 *  might want to consider having a bob tail for the incoming
	 *  task or some other task at the task's location.  This method
	 *  can be used to initiate a bob tail
	 * 
	 * @param incomingTask
	 */
	protected void bobTailNeededAfterIncomingLoad(DispatchTask incomingTask){
		// try to identify a nearby location that can serve as a bobtail
		// if found, remove from dispatch queue, place in bobtail pickup queue
		// dispatch a bobtail to load task
		TransportLocationIfc origin = incomingTask.getOrigin();
		
		double allowedDistance = getMaxDeadHeadBobTailDistance(origin);
		TransportLocationIfc bobTailLocation = getFirstLocationWithIdleDriversAndTrucks(origin, allowedDistance);
		if (bobTailLocation != null){
			// found a location, make the bobtail dispatch
			dispatchBobTail(incomingTask, bobTailLocation);
		}
	}
	
	/** If not all resources are available for the incoming task,
	 *  the dispatcher might want to consider having a deadhead for the incoming
	 *  task or some other task at the task's location.  This method
	 *  can be used to initiate a deadhead in this situation
	 * 
	 * @param incomingTask
	 */
	protected void deadHeadNeededAfterIncomingLoad(DispatchTask incomingTask){

		// try to identify a nearby location that can serve as a deadhead
		// if found, remove from dispatch queue, place in deadhead pickup queue
		// dispatch a deadhead to load task
		TransportLocationIfc origin = incomingTask.getOrigin();
		
		double allowedDistance = getMaxDeadHeadBobTailDistance(origin);
		TransportLocationIfc deadHeadLocation = getFirstLocationWithIdleResources(origin, allowedDistance);
		if (deadHeadLocation != null){
			// found a location, make the deadhead dispatch
			dispatchDeadHead(incomingTask, deadHeadLocation);
		}
	}
		
	/** Selects the next DispatchLoad to dispatch. The default behavior
	 *  of this method is to search the dispatch queue (without regard to location) 
	 *  in order of arrival and to return the first load task that can be dispatched.  If no
	 *  load task is found, this method returns null.  This method does not
	 *  remove the item to be dispatched
	 * 
	 *  Assumptions:
	 *  1) A DispatchLoad can be dispatched, if the required resources can be found
	 *     at its current location
	 *  2) If no DispatchLoad meets this requirement then this method
	 *     should return null
	 * 
	 * @return
	 */
	protected DispatchTask selectNextTask(){
		
		DispatchTask loadTask = null;
		if (myLoadSelectionRule != null)
			loadTask = myLoadSelectionRule.selectNextTransportTask(this, Collections.unmodifiableList(myDispatchQ));
		else{
			// find the first DispatchLoad that can be dispatched
			for(DispatchTask task: myDispatchQ){
				if (transportAvailable(task)){
					loadTask = task;
					break;
				}
			}
		}
		return(loadTask);
	}

	/** Selects the next task at the given location, returns null if none
	 *  Does not remove the task from the queue.
	 *  
	 * @param location
	 * @return
	 */
	protected DispatchTask selectNextTask(TransportLocationIfc location){
		Queue queue = myDispatchQByLocation.get(location);
		if (queue == null)
			return(null);
		else {
			return((DispatchTask)queue.peekNext());
		}
	}
	
	/** Processes all waiting tasks and if available resources can be found
	 *  at the task's origin then the task is dispatched.
	 */
	protected void dispatchLoads(){
		
		boolean done = false;
		while((!myDispatchQ.isEmpty()) && (!done)){
			DispatchTask loadTask = selectNextTask();
			if (loadTask == null)
				done = true;
			else {
				dequeue(loadTask);
				// all resources have been set, so dispatch transport task				
				dispatchTransport(loadTask);
			}
		}
	}
	
	/** Called when the task has completed.  This method delegates the
	 *  work to transportCompleted(), deadHeadToLoadCompleted(), bobTailToLoadCompleted()
	 *  deadHeadNoLoadCompleted(), bobTailNoLoadCompleted() depending on the type of task.
	 *  
	 * @param task
	 */
	protected void dispatchCompleted(DispatchTask task) {
		
		if (task.isTransport())
			transportCompleted(task);
		else if (task.isDeadHeadToLoad())
			deadHeadToLoadCompleted(task);
		else if (task.isBobTailToLoad())
			bobTailToLoadCompleted(task);
		else if (task.isBobTailNoLoad())
			bobTailNoLoadCompleted(task);
		else if (task.isDeadHeadNoLoad())
			deadHeadNoLoadCompleted(task);
			
	}

	/**
	 * @param task
	 */
	protected void bobTailToLoadCompleted(DispatchTask task) {
		// task already has driver, truck, trailer
		// when the bob tail was dispatched it had to have seized the trailer
		// need to proceed to load's destination
		task.setTypeToTransport(); // this makes it a transport

		// get the queue that the task is currently waiting in
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myBobTailPickupQByLocation.get(location);
		// now remove it
		queue.remove(task); 

		//place the task in the transport pick up queue
		queue = myTransportPickupQByLocation.get(location);
		queue.enqueue(task);

		task.waitingForTransport();
		// it is important that loading complete be called by the driver
	}
	
	/**
	 * @param task
	 */
	protected void deadHeadToLoadCompleted(DispatchTask task) {
		// task already has driver, truck, and trailer
		// need to proceed to load's destination
		task.setTypeToTransport(); // this makes it a transport
		
		// get the queue that the task is currently waiting in
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myDeadHeadPickupQByLocation.get(location);
		// now remove it
		queue.remove(task); 

		//place the task in the transport pick up queue
		queue = myTransportPickupQByLocation.get(location);
		queue.enqueue(task);
		// indicate that the task is waiting for transport
		task.waitingForTransport();		
		// it is important that loading complete be called by the driver
	}

	/**
	 * @param task
	 */
	protected void transportCompleted(DispatchTask task) {
		// the task has arrived at it destination
		// it should no longer be in the incoming load queue for the destination
		Queue q = task.getQueue(); // should be the incoming load queue
		q.remove(task);
		// now release the resources held by the task
		releaseDriver(task);
		releaseTruck(task);
		releaseTrailer(task);
		TransportLocationIfc location = task.getLoadDestination();
		allocateResourcesAfterTransportCompletedAt(location);
	}	

	/**
	 * @param task
	 */
	protected void deadHeadNoLoadCompleted(DispatchTask task) {
		
	}

	/**
	 * @param task
	 */
	protected void bobTailNoLoadCompleted(DispatchTask task) {
		
	}

	/** This allocation method is invoked after a successful transport. The
	 *  default behavior is to check if there is a load waiting at the
	 *  dropoff location.  If so, the next load at the dropoff location,
	 *  is selected for dispatching. If there are no loads waiting at this location,
	 *  then there is an opportunity for deadheading.
	 *  
	 *  This behavior may be overridden.  For example, the dispatcher may want to
	 *  wait for some condition before dispatching the task even if resources
	 *  are immediately available

	 * @param location
	 */
	protected void allocateResourcesAfterTransportCompletedAt(TransportLocationIfc location) {
		
		// get the queue at the location
		Queue queue = myDispatchQByLocation.get(location);

		// select next task at the drop off location
		DispatchTask nextTask = (DispatchTask)queue.peekNext();

		// if there is a task, then dispatch it
		if (nextTask != null){			
			// check if the task can be dispatched
			if (transportAvailable(nextTask)){
				// remove the task and dispatch it
				dequeue(nextTask);
				// all resources have been set, so dispatch transport task
				dispatchTransport(nextTask);
			}				
		} else
			resourcesAvailableAt(location);
	}
	
	protected DispatchTask findTaskNeedingBobTail(TransportLocationIfc location){
		TransportLocationIfc needsTransport = findLocationNeedingBobTail(location);
		DispatchTask task = null;
		if (needsTransport != null){
			task = selectNextTask(needsTransport);
		}
		return(task);
	}

	protected DispatchTask findTaskNeedingDeadHead(TransportLocationIfc location){
		TransportLocationIfc needsTransport = findLocationNeedingDeadHead(location);
		DispatchTask task = null;
		if (needsTransport != null){
			task = selectNextTask(needsTransport);
		}
		return(task);
	}
	
	/** A driver, truck, and trailer were just released at the given
	 *  location and there were no loads available for pick up.  The
	 *  dispatcher might want to allocate the resources to waiting loads
	 *  at other locations
	 * 
	 * @param location
	 */
	protected void resourcesAvailableAt(TransportLocationIfc location){
		
		// check for needed bobtails
		DispatchTask task = findTaskNeedingBobTail(location);
		if (task != null){
			// dispatch bobtail task
			dispatchBobTail(task, location);
		} else{
			task = findTaskNeedingDeadHead(location);
			if (task != null){
				// dispatch dead head
				dispatchDeadHead(task, location);
			}
		}
		
	}
	
	/** Changes the task to a bobtail and initiates a pick up from the supplied location
	 * 
	 * @param task
	 * @param fromLoc The location that the bobtail is to come from
	 */
	protected void dispatchBobTail(DispatchTask task, TransportLocationIfc fromLoc){
		task.setTypeToBobTailToLoad();
		TransportLocationIfc origin = task.getLoadOrigin();
		dequeue(task);
		Queue queue = myBobTailPickupQByLocation.get(origin);
		queue.enqueue(task);
		task.setDriver(getIdleDriver(fromLoc));
		task.setTruck(getIdleTruck(fromLoc));
		task.setTrailer(getIdleTrailer(origin));
		task.allocateDriver();
		task.allocateTruck();
		task.allocateTrailer();
		task.waitingForTransport();	
	}
	
	/** Initiates a deadhead for the task from the supplied location
	 * 
	 * @param task
	 * @param fromLoc The location that the deadhead is to come from
	 */
	protected void dispatchDeadHead(DispatchTask task, TransportLocationIfc fromLoc){
		task.setTypeToDeadHeadToLoad();
		TransportLocationIfc origin = task.getLoadOrigin();
		dequeue(task);
		Queue queue = myDeadHeadPickupQByLocation.get(origin);
		queue.enqueue(task);
		task.setDriver(getIdleDriver(fromLoc));
		task.setTruck(getIdleTruck(fromLoc));
		task.setTrailer(getIdleTrailer(fromLoc));
		task.allocateDriver();
		task.allocateTruck();
		task.allocateTrailer();
		task.waitingForTransport();
	}
}
