/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.transportlayer;

import java.util.HashMap;
import java.util.Map;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.EventGenerator;
import jsl.modeling.elements.EventGeneratorIfc;
import jsl.modeling.elements.EventGeneratorListenerIfc;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.RandomIfc;
import jsl.utilities.random.distributions.DEmpirical;

/**
 * LoadGenerator extends ModelElment. It generates load and sets the load origin probability distribution i.e., 
 * the probability with which the load is generated from a  certain origin. It also sets the destination 
 * probability distribution given the origin it has arrived from
 * 
 */
public class LoadGenerator extends ModelElement implements LoadSenderIfc, EventGeneratorIfc {
	
	/** An instance of EventGenerator sets its LoadGeneratorListener which has a refernce to the 
	 * private class LoadGeneraorListener 
	 */
	private EventGenerator myEventGenerator;
	
	/** An instance of LoadGeneratorListener implements the generate() method of the EventGeneratorListenerIfc
	 * 
	 */
	private LoadGeneratorListener myLoadGeneratorListener;
	
	/**An instance of Transportation Network 
	 */
	private TransportationNetwork myNetwork;
	
	/** A Random Variable for origin 
	 */
	protected RandomVariable myOriginRV;
	
	/** Discrete Empirical Probability Distribution of Origins 
	 */
	protected DEmpirical myOriginPDF;
	
	/** A hash map of Origin-Destination random variables with origin as the key and value
	 *  as the random distribution of the origin 
	 */
	protected Map<TransportLocationIfc, RandomVariable> myODRVs;
	
	/** Creates LoadGenerator
	 * 
	 * @param parent
	 * @param transportationNetwork
	 */
	public LoadGenerator(ModelElement parent,TransportationNetwork transportationNetwork) {
		this(parent,transportationNetwork, null, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/** Creates LoadGenerator
	 * 
	 * @param parent
	 * @param transportationNetwork
	 * @param timeUntilFirst
	 */
	public LoadGenerator(ModelElement parent, TransportationNetwork transportationNetwork,  
			RandomIfc timeUntilFirst) {
		this(parent,transportationNetwork, timeUntilFirst, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}
	
	/** Creates LoadGenerator
	 * 
	 * @param parent
	 * @param transportationNetwork
	 * @param timeUntilFirst
	 * @param name
	 */
	public LoadGenerator(ModelElement parent, TransportationNetwork transportationNetwork, 
			RandomIfc timeUntilFirst, String name) {
		this(parent,transportationNetwork, timeUntilFirst, null, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}
	
	/** Creates LoadGenerator
	 * 
	 * @param parent
	 * @param transportationNetwork
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 */
	public LoadGenerator(ModelElement parent, TransportationNetwork transportationNetwork, 
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext) {
		this(parent,transportationNetwork, timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, null);
	}

	/** Creates LoadGenerator
	 * 
	 * @param parent
	 * @param transportationNetwork
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param name
	 */
	 public LoadGenerator(ModelElement parent,  TransportationNetwork transportationNetwork,
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext, String name) {
		this(parent, transportationNetwork,timeUntilFirst, timeUntilNext, Long.MAX_VALUE, Double.POSITIVE_INFINITY, name);
	}

	/** Creates LoadGenerator
	 * 
	 * @param parent
	 * @param transportationNetwork
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 * @param name
	 */
	 public LoadGenerator(ModelElement parent,TransportationNetwork transportationNetwork,  
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext,
			long maxNum, String name) {
		this(parent, transportationNetwork,timeUntilFirst, timeUntilNext, maxNum, Double.POSITIVE_INFINITY, name);
	}

	/** Creates LoadGenerator with parameters as transportation network, a RandomIfc object that 
	 *  supplies the time until the first event,a RandomIfc object that supplies the time between the load 
	 *  arrivals,a RandomIfc object that suuplies the maximum number load arrivals, a RandomIfc object
	 *  that supplies the time at which the last load arrives
	 *   
	 * @param parent
	 * @param transportationNetwork
	 * @param timeUntilFirst
	 * @param timeUntilNext
	 * @param maxNum
	 * @param timeUntilLast
	 * @param name
	 */
	public LoadGenerator(ModelElement parent, TransportationNetwork transportationNetwork, 
			RandomIfc timeUntilFirst, RandomIfc timeUntilNext,
			long maxNum, double timeUntilLast, String name) {
		super(parent, name);
		
		setTransportationNetwork(transportationNetwork);
		myLoadGeneratorListener = new LoadGeneratorListener();
		myEventGenerator = new EventGenerator(this, myLoadGeneratorListener, timeUntilFirst,
				timeUntilNext, maxNum, timeUntilLast, getName() + "Demand Generator");
		myOriginPDF = new DEmpirical();
		myOriginRV = new RandomVariable(this, myOriginPDF);
		myODRVs = new HashMap<TransportLocationIfc, RandomVariable>();
    }
	
	public final double getEndingTime() {
		return myEventGenerator.getEndingTime();
	}
	
	public final long getMaxNumActions() {
		return myEventGenerator.getMaximumNumberOfEvents();
	}
	
	public final boolean isGeneratorDone() {
		return myEventGenerator.isGeneratorDone();
	}
					
	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getGenerationEndingTimeForReplications()
	 */
	public double getGenerationEndingTimeForReplications() {
		return myEventGenerator.getGenerationEndingTimeForReplications();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getMaximumNumberOfEvents()
	 */
	public long getMaximumNumberOfEvents() {
		return myEventGenerator.getMaximumNumberOfEvents();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getMaximumNumberOfEventsForReplications()
	 */
	public long getMaximumNumberOfEventsForReplications() {
		return myEventGenerator.getMaximumNumberOfEventsForReplications();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getNumberOfEventsGenerated()
	 */
	public long getNumberOfEventsGenerated() {
		return myEventGenerator.getNumberOfEventsGenerated();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getTimeBetweenEventsForReplications()
	 */
	public RandomIfc getTimeBetweenEventsForReplications() {
		return myEventGenerator.getTimeBetweenEventsForReplications();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#getTimeUntilFirstEventForReplications()
	 */
	public RandomIfc getTimeUntilFirstEventForReplications() {
		return myEventGenerator.getTimeUntilFirstEventForReplications();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGeneratorIfc#isSuspended()
	 */
	public boolean isSuspended() {
		return myEventGenerator.isSuspended();
	}

	/**
	 * 
	 * @see jsl.modeling.elements.EventGeneratorIfc#resume()
	 */
	public void resume() {
		myEventGenerator.resume();
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGenerator#getStartOnInitializeFlag()
	 */
	public final boolean getStartOnInitializeFlag() {
		return myEventGenerator.getStartOnInitializeFlag();
	}

	/**
	 * @param flag
	 * @see jsl.modeling.elements.EventGenerator#setStartOnInitializeFlag(boolean)
	 */
	public final void setStartOnInitializeFlag(boolean flag) {
		myEventGenerator.setStartOnInitializeFlag(flag);
	}

	/**
	 * 
	 * @see jsl.modeling.elements.EventGenerator#turnOnGenerator()
	 */
	public final void turnOnGenerator() {
		myEventGenerator.turnOnGenerator();
	}

	/**
	 * @param t
	 * @see jsl.modeling.elements.EventGenerator#turnOnGenerator(double)
	 */
	public final void turnOnGenerator(double t) {
		myEventGenerator.turnOnGenerator(t);
	}

	/**
	 * @param r
	 * @see jsl.modeling.elements.EventGenerator#turnOnGenerator(jsl.utilities.random.RandomIfc)
	 */
	public void turnOnGenerator(RandomIfc r) {
		myEventGenerator.turnOnGenerator(r);
	}

	/**
	 * @param endingTime
	 * @see jsl.modeling.elements.EventGeneratorIfc#setEndingTime(double)
	 */
	public void setEndingTime(double endingTime) {
		myEventGenerator.setEndingTime(endingTime);
	}

	/**
	 * @param endingTime
	 * @see jsl.modeling.elements.EventGeneratorIfc#setGenerationEndingTimeForReplications(double)
	 */
	public void setGenerationEndingTimeForReplications(double endingTime) {
		myEventGenerator.setGenerationEndingTimeForReplications(endingTime);
	}

	/**
	 * @param maxNum
	 * @see jsl.modeling.elements.EventGeneratorIfc#setMaximumNumberOfEvents(long)
	 */
	public void setMaximumNumberOfEvents(long maxNum) {
		myEventGenerator.setMaximumNumberOfEvents(maxNum);
	}

	/**
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setMaximumNumberOfEventsForReplications(long)
	 */
	public void setMaximumNumberOfEventsForReplications(long maxNumEvents) {
		myEventGenerator.setMaximumNumberOfEventsForReplications(maxNumEvents);
	}

	/**
	 * @param timeBtwEvents
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEvents(jsl.utilities.random.RandomIfc, long)
	 */
	public void setTimeBetweenEvents(RandomIfc timeBtwEvents, long maxNumEvents) {
		myEventGenerator.setTimeBetweenEvents(timeBtwEvents, maxNumEvents);
	}

	/**
	 * @param timeUntilNext
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEvents(jsl.utilities.random.RandomIfc)
	 */
	public void setTimeBetweenEvents(RandomIfc timeUntilNext) {
		myEventGenerator.setTimeBetweenEvents(timeUntilNext);
	}

	/**
	 * @return
	 * @see jsl.modeling.elements.EventGenerator#getTimeBetweenEvents()
	 */
	public final RandomIfc getTimeBetweenEvents() {
		return myEventGenerator.getTimeBetweenEvents();
	}

	/**
	 * @param timeBtwEvents
	 * @param maxNumEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEventsForReplications(jsl.utilities.random.RandomIfc, long)
	 */
	public void setTimeBetweenEventsForReplications(RandomIfc timeBtwEvents, long maxNumEvents) {
		myEventGenerator.setTimeBetweenEventsForReplications(timeBtwEvents, maxNumEvents);
	}

	/**
	 * @param timeBtwEvents
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeBetweenEventsForReplications(jsl.utilities.random.RandomIfc)
	 */
	public void setTimeBetweenEventsForReplications(RandomIfc timeBtwEvents) {
		myEventGenerator.setTimeBetweenEventsForReplications(timeBtwEvents);
	}

	/**
	 * @param timeUntilFirst
	 * @see jsl.modeling.elements.EventGeneratorIfc#setTimeUntilFirstEventForReplications(jsl.utilities.random.RandomIfc)
	 */
	public void setTimeUntilFirstEventForReplications(RandomIfc timeUntilFirst) {
		myEventGenerator.setTimeUntilFirstEventForReplications(timeUntilFirst);
	}

	/**
	 * 
	 * @see jsl.modeling.elements.EventGeneratorIfc#suspend()
	 */
	public void suspend() {
		myEventGenerator.suspend();
	}

	public final void turnOffGenerator() {
		myEventGenerator.turnOffGenerator();
	}
	
	/** Sets the origin probability as a discrete empirical probability distribution myOriginPDF
	 *  Creates a hash map with the random origin as the key and the distribution of the random origin
	 * @param location
	 * @param p
	 */
	public void setOriginProbability(TransportLocationIfc location, double p ){
		if(!myNetwork.isOrigin(location))	
			new IllegalArgumentException("The supplied location is not a valid origin.");

		if (!myODRVs.containsKey(location)){
			RandomVariable rv = new RandomVariable(this, new DEmpirical());
			myODRVs.put(location, rv);
		}
		int i = myNetwork.getIndexOf(location);
		myOriginPDF.addProbabilityPoint(i, p);
	}
	
	/** Sets the origin probability for the last last origin
	 * 
	 * @param location
	 */
	public void setLastOriginProbability(TransportLocationIfc location){
		
		if(!myNetwork.isOrigin(location))	
			new IllegalArgumentException("The supplied location is not a valid origin.");
		
		if (!myODRVs.containsKey(location)){
			RandomVariable rv = new RandomVariable(this, new DEmpirical());
			myODRVs.put(location, rv);
		}		
		int i = myNetwork.getIndexOf(location);
		myOriginPDF.addLastProbabilityPoint(i);
	}
		
	/** Sets the loads destination probability for a given origin and destination
	 *  as a discrete empirical distribution
	 * 
	 * @param lane
	 * @param p
	 */
	public void setDestinationProbability(LaneIfc lane, double p){
		
		if(!myNetwork.isLane(lane))	
			new IllegalArgumentException("The supplied lane is not a valid lane.");

		TransportLocationIfc origin = lane.getOrigin();
		TransportLocationIfc destination = lane.getDestination();
		int i = myNetwork.getIndexOf(destination);
		RandomVariable rv = myODRVs.get(origin);
		DEmpirical d = (DEmpirical) rv.getRandomSource();
		d.addProbabilityPoint(i,p);
		
	}

	/** Sets the load's destination probability as discrete empirical probability distribution
	 *  for the last destination
	 * 
	 * @param lane
	 */
	public void setLastDestinationProbability(LaneIfc lane){
		
		if(!myNetwork.isLane(lane))	
			new IllegalArgumentException("The supplied lane is not a valid lane.");

		TransportLocationIfc origin = lane.getOrigin();
		TransportLocationIfc destination = lane.getDestination();
		int i = myNetwork.getIndexOf(destination);		
		RandomVariable rv = myODRVs.get(origin);		
		DEmpirical d = (DEmpirical) rv.getRandomSource();
		d.addLastProbabilityPoint(i);
	}

	//  ===========================================
	//  PROTECTED METHODS
	//  ===========================================

	/** Sets the transportation network
	 * 
	 * @param network
	 */
	protected void setTransportationNetwork(TransportationNetwork network){
	   	 if(network == null)
	   		 throw new IllegalArgumentException("The Transportation Network was Null!");		 
	   	 myNetwork = network;
	}

	/** Returns random origin from the potential origins
	 * 
	 * @return
	 */
	protected TransportLocationIfc getRandomOrigin(){
		int i = (int)myOriginRV.getValue();
		return(myNetwork.getLocation(i));
	}
	
	/** Returns random destination from the list of the destinations of the origin 
	 *  given that origin
	 * 
	 * @param origin
	 * @return
	 */
	protected TransportLocationIfc getRandomDestination(TransportLocationIfc origin){
		RandomVariable rv = myODRVs.get(origin);
		int i = (int)rv.getValue();
		return(myNetwork.getLocation(i));
	}
	
	/** The load is created with a random origin and random detsination given that origin
	 *  The LoadGenerator sends the load to the TransportationNetwork
	 * 
	 * @param generator
	 * @param event
	 */
	protected void generateLoad(EventGenerator generator, JSLEvent event){
		
		TransportLocationIfc origin = getRandomOrigin();
		System.out.println();
		System.out.println("Origin is :"+origin.getName());
		TransportLocationIfc destination = getRandomDestination(origin);
		System.out.println("Destination is :"+destination.getName());

		//TODO What about assigning a duedate
		Load load = myNetwork.createLoad(this, origin, destination);
		if (load != null)
			myNetwork.receive(load);
		else
			invalidLoadGenerated(this, origin, destination);
		
	}
	
    /** LoadGeneratorListener implements the generate()method of the EventGeneratorListenerIfc
     *  It calls the generateLoad()in this class  
     *
     */
	private class LoadGeneratorListener implements EventGeneratorListenerIfc{

		public void generate(EventGenerator generator, JSLEvent event){
			
			generateLoad(generator, event);
		}
	}

	protected void invalidLoadGenerated(LoadSenderIfc sender, TransportLocationIfc origin, TransportLocationIfc destination) {
		String s = "An invalid load was generated for origin: " + origin + " and destination : " + destination;
		throw new IllegalStateException(s);
	}

	/* (non-Javadoc)
	 * @see transportlayer.LoadSenderIfc#reachedDestination(transportlayer.Load)
	 */
	public void reachedDestination(Load load) {
		
	}
}