/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.transportlayer;

import java.util.Collection;
import java.util.List;

import jsl.modeling.ModelElement;
import jsl.modeling.SchedulingElement;
import jsl.modeling.elements.spatial.SpatialModel;

/** A base class for constructing transportation networks
 * 
 *
 */
public class TransportationNetwork extends SchedulingElement implements NetworkIfc, LoadReceiverIfc {

	/** An instance of DispatcherAbstract class
	 */
	private AbstractDispatcher myDispatcher;
	
	/** Holds the network of locations and lanes
	 */
	protected NetworkIfc myNetwork;

	//  ===========================================
	//  CONSTRUCTORS
	//  ===========================================

	/** Creates the Transportation Network with the same spatial model as that of its parent
	 *  using a default SimpleDispatcher
	 *  
	 * @param parent
	 * @param spatialModel
	 */
	public TransportationNetwork(ModelElement parent, SpatialModel spatialModel) {
		this(parent, spatialModel, null);		
	}
	
	/**  Creates the Transportation Network with the same spatial model as that of its parent
	 *   using the supplied dispatcher
	 *   
	 * @param parent
	 * @param spatialModel
	 * @param dispatcher
	 * @param name
	 */
	public TransportationNetwork(ModelElement parent, SpatialModel spatialModel, String name) {
		super(parent, name);
		setSpatialModel(spatialModel);
		myNetwork = new Network();
	}

	/** This method should be used to create loads to be sent into the network.
	 *  
	 *  Conditions: 
	 *   1) The origin must be in this network
	 *   2) The destination must be in this network
	 *   3) The origin and destination must be different
	 * 
	 * @param sender
	 * @param origin
	 * @param destination
	 * @return
	 */
	public Load createLoad(LoadSenderIfc sender, TransportLocationIfc origin, TransportLocationIfc destination){
		//TODO need to supply default due date
		return(createLoad(sender, origin, destination, 0.0));
	}

	/** This method should be used to create loads to be sent into the network.
	 *  
	 *  Conditions: 
	 *   1) The origin must be in this network
	 *   2) The destination must be in this network
	 *   3) The origin and destination must be different
	 *  
	 * @param sender
	 * @param origin
	 * @param destination
	 * @param dueDate
	 * @return Returns null if the destination is not reachable from the origin
	 */
	public Load createLoad(LoadSenderIfc sender, TransportLocationIfc origin, TransportLocationIfc destination, double dueDate){
		if (!containsLocation(origin))
			throw new IllegalArgumentException("The supplied origin is not in this network");

		if (!containsLocation(destination))
			throw new IllegalArgumentException("The supplied destination is not in this network");

		if (destination == origin)
			throw new IllegalArgumentException("The origin and destination must not be the same.");

		Load load = null;
		if(isAllReachable()){
			load = new Load(this, sender, origin, destination, dueDate);
		} else {
			if (isReachable(origin, destination)){
				load = new Load(this, sender, origin, destination, dueDate);
			} 
		}
		return(load);
	}
	
	/** Receives a load into the transportation network and
	 *  passes it on to the dispatcher for dispatching.
	 *  
	 *  Conditions: 
	 *  1) The load must not be null
	 *  2) The load must have been made by this network
	 *  
	 * @param load
	 */
	public void receive(Load load) {

		if (load == null)
			throw new IllegalArgumentException("The load was null!");

		if (load.getNetwork() != this)
			throw new IllegalArgumentException("The load was not made by this network!");

		if (myDispatcher != null)
			myDispatcher.receive(load);

	}
	
	/**
	 * @return Returns the dispatcher.
	 */
	protected final AbstractDispatcher getDispatcher() {
		return myDispatcher;
	}
	
	/**
	 * @param dispatcher The dispatcher to set.
	 */
	protected final void setDispatcher(AbstractDispatcher dispatcher) {
		myDispatcher = dispatcher;
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#addLane(transportlayer.LaneIfc)
	 */
	public final boolean addLane(LaneIfc lane) {
		return myNetwork.addLane(lane);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#containsLocation(transportlayer.TransportLocationIfc)
	 */
	public final boolean containsLocation(TransportLocationIfc location) {
		return myNetwork.containsLocation(location);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getDestinations()
	 */
	public final Collection<TransportLocationIfc> getDestinations() {
		return myNetwork.getDestinations();
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getIncomingLanes(transportlayer.TransportLocationIfc)
	 */
	public final Collection<LaneIfc> getIncomingLanes(TransportLocationIfc destination) {
		return myNetwork.getIncomingLanes(destination);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getIndexOf(transportlayer.TransportLocationIfc)
	 */
	public final int getIndexOf(TransportLocationIfc location) {
		return myNetwork.getIndexOf(location);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getLane(transportlayer.TransportLocationIfc, transportlayer.TransportLocationIfc)
	 */
	public final LaneIfc getLane(TransportLocationIfc origin,
			TransportLocationIfc destination) {
		return myNetwork.getLane(origin, destination);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getLanes()
	 */
	public final Collection<LaneIfc> getLanes() {
		return myNetwork.getLanes();
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getLocation(int)
	 */
	public final TransportLocationIfc getLocation(int index) {
		return myNetwork.getLocation(index);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getLocations()
	 */
	public final List<TransportLocationIfc> getLocations() {
		return myNetwork.getLocations();
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getNumberOfIncomingLanes(transportlayer.TransportLocationIfc)
	 */
	public final int getNumberOfIncomingLanes(TransportLocationIfc destination) {
		return myNetwork.getNumberOfIncomingLanes(destination);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getNumberOfLanes()
	 */
	public final int getNumberOfLanes() {
		return myNetwork.getNumberOfLanes();
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getNumberOfLocations()
	 */
	public final int getNumberOfLocations() {
		return myNetwork.getNumberOfLocations();
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getNumberOfOutGoingLanes(transportlayer.TransportLocationIfc)
	 */
	public final int getNumberOfOutGoingLanes(TransportLocationIfc origin) {
		return myNetwork.getNumberOfOutGoingLanes(origin);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getOrigins()
	 */
	public final Collection<TransportLocationIfc> getOrigins() {
		return myNetwork.getOrigins();
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getOutGoingLanes(transportlayer.TransportLocationIfc)
	 */
	public final Collection<LaneIfc> getOutGoingLanes(TransportLocationIfc origin) {
		return myNetwork.getOutGoingLanes(origin);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getIncomingLocations(transportlayer.TransportLocationIfc)
	 */
	public Collection<TransportLocationIfc> getIncomingLocations(TransportLocationIfc destination) {
		return myNetwork.getIncomingLocations(destination);
	}

	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#getOutGoingLocations(transportlayer.TransportLocationIfc)
	 */
	public Collection<TransportLocationIfc> getOutGoingLocations(TransportLocationIfc origin) {
		return myNetwork.getOutGoingLocations(origin);
	}

	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#isDestination(transportlayer.TransportLocationIfc)
	 */
	public final boolean isDestination(TransportLocationIfc location) {
		return myNetwork.isDestination(location);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#isLane(transportlayer.LaneIfc)
	 */
	public final boolean isLane(LaneIfc lane) {
		return myNetwork.isLane(lane);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#isLane(transportlayer.TransportLocationIfc, transportlayer.TransportLocationIfc)
	 */
	public final boolean isLane(TransportLocationIfc origin,
			TransportLocationIfc destination) {
		return myNetwork.isLane(origin, destination);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#isOrigin(transportlayer.TransportLocationIfc)
	 */
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#isOrigin(transportlayer.TransportLocationIfc)
	 */
	public final boolean isOrigin(TransportLocationIfc location) {
		return myNetwork.isOrigin(location);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#isReachable(transportlayer.TransportLocationIfc, transportlayer.TransportLocationIfc)
	 */
	public final boolean isReachable(TransportLocationIfc origin,
			TransportLocationIfc destination) {
		return myNetwork.isReachable(origin, destination);
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#isAllReachable()
	 */
	public boolean isAllReachable() {
		return myNetwork.isAllReachable();
	}

	/* (non-Javadoc)
	 * @see transportlayer.NetworkIfc#updateLane(transportlayer.LaneIfc)
	 */
	public final void updateLane(LaneIfc lane) {
		myNetwork.updateLane(lane);
	}

	public String toString() {
		return myNetwork.toString();
	}
}
