package jslx.transportlayerFinal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.spatial.SpatialResource;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.utilities.random.distributions.DistributionIfc;
import jslx.transportlayerFinal.DriverSelectionRuleIfc;

/** AbstractDispatcherFTL is a sub class of AbstractDispatcher and more specific to Full Truckload Transportation
 *  Network.It holds the lists of all drivers, trailers and trucks in the Network and also by locations and also 
 *  the list of these resources at the locations they were originally assigned to.It has methods to add driver,
 *  trailer and trucks at a location and checking if there are idle resources available at a location. It is 
 *  helpful in searching the closest location within the user defined range of distance with most idle trucks and trailers 
 *  to the one where the load arrives incase there are no idle trucks and drivers available at the origin of the load. Similarly, it searches the 
 *  closest location with all the three idle resources to facilitate transportation of load at the origin incase 
 *  there is no idle trailer in addition to no idle truck and driver at the origin.
 *  
 *  @author snangia
 *
 */

abstract public class AbstractDispatcherFTL extends AbstractDispatcher {

	//  ===========================================
	//  Data structures to hold the resources of the network
	//  ===========================================
	
	/** List of all Drivers in the network of the type Driver 
	 */
	private List<Driver> myDrivers;
	
	/** List of all Trailers in the network of the type Spatial Resource 
	 */
	private List<SpatialResource> myTrailers;
	
	/** List of all Trucks in the network of the type Spatial Resource 
	 */
	private List<SpatialResource> myTrucks;
	
	/** For each location, there is a list of initially assigned drivers
	 */
	private Map<TransportLocationIfc, List<Driver>> myDriverLocationAssignments;
	
	/** For each location, there is a list of initiallly assigned trucks
	 */
	private Map<TransportLocationIfc, List <SpatialResource>> myTruckLocationAssignments;
	
	/** For each location, there is a list of initially assigned trailers
	 */
	private Map<TransportLocationIfc, List<SpatialResource>> myTrailerLocationAssignments;
	
	/** For each location, there is a list of idle drivers
	 */
	protected Map<TransportLocationIfc, List<Driver>> myIdleDriversByLocation;
	
	/** For each location, there is a list of idle trucks
	 */
	protected Map<TransportLocationIfc, List <SpatialResource>> myIdleTrucksByLocation;
	
	/** For each location, there is a list of idle trailers
	 */
	protected Map<TransportLocationIfc, List<SpatialResource>> myIdleTrailersByLocation;
	
	/** An instance of DriverSelectionRuleIfc 
	  */
	private DriverSelectionRuleIfc myDriverSelectionRule;

	/** An instance of TruckSelectionRuleIfc 
	  */
	private TruckSelectionRuleIfc myTruckSelectionRule;

	/** An instance of TrailerSelectionRuleIfc 
	  */
	private TrailerSelectionRuleIfc myTrailerSelectionRule;
	
	/** Comparator which sorts location in the
	 *  ascending order of most usable trailers available
	 *  at the locations 
	 */
	private Comparator<TransportLocationIfc> myTrailerLocationComparator;
	
	/** Comparator which sorts location in the
	 *  ascending order of most usable resources 
	 *  available at the locations 
	 * 
	 */
	private Comparator<TransportLocationIfc> myDriverTruckTrailerLocationComparator;
	
	/** Comparator which sorts location in the
	 *  ascending order of most usable (driver,truck) pairs
	 *  available at the locations 
	 * 
	 */
	private Comparator<TransportLocationIfc> myDriverTruckLocationComparator;
	
	/** The maximum default distance for considering deadhead or bobtails
	 * 
	 */
	private double myMaxDHBTDistance;

	/** The maximum default distance for considering deadhead or bobtails for a location
	 * 
	 */
	private Map<TransportLocationIfc, Double> myMaxDHBTDistancesByLocation;
	
	protected TimeWeighted myNumBusyDrivers;
	
	protected TimeWeighted myNumIdleDrivers;
	
	protected TimeWeighted myNumOffDutyDrivers;
	
	protected TimeWeighted myNumOffDutyWithTaskDrivers;
	
	protected TimeWeighted myNumOffDutyWithoutTaskDrivers;
	
	protected TimeWeighted myNumBusyTrucks;
	
	protected TimeWeighted myNumBusyTrailers;
	
	protected ResponseVariable myAverageDeadHeadMiles;
	
	protected ResponseVariable myAverageBobTailMiles;
	
	protected ResponseVariable myAverageTransportMiles;
	
	protected TimeWeighted myAvgNumDeadHeadMiles;
	
	protected TimeWeighted myAvgNumBobTailMiles;
	
	protected TimeWeighted myAvgNumTransportMiles;
	
	protected double myTotalNumberOfDrivers;
	
	protected Collection<TransportLocationIfc> myLocations;
	
	//  ===========================================
	//  CONSTRUCTORS
	//  ===========================================
	
	public AbstractDispatcherFTL(ModelElement parent, TransportationNetwork network, double distance) {
		this(parent, network, distance, null);
	}

	/** Creates AbstractDispatcherFTL by supplying transportation network sending it the load
	 *  for processing and the max distance for deadheads or bobtails
	 *  Creates new lists and maps for all the resources in the network, idle resources, assigned resources 
	 *  at all the locations
	 *  
	 * @param parent
	 * @param network
	 * @param distance
	 * @param name
	 */
	public AbstractDispatcherFTL(ModelElement parent, TransportationNetwork network, double distance, String name) {
		super(parent, network, name);
		// set up the resource data structures
		myDrivers = new ArrayList<Driver>();
		myTrailers = new ArrayList<SpatialResource>();
		myTrucks = new ArrayList<SpatialResource>();
		
		myDriverLocationAssignments = new HashMap<TransportLocationIfc, List<Driver>>();
		myTruckLocationAssignments = new HashMap<TransportLocationIfc, List<SpatialResource>>();
		myTrailerLocationAssignments = new HashMap<TransportLocationIfc, List<SpatialResource>>();
		myIdleDriversByLocation = new HashMap<TransportLocationIfc, List<Driver>>();
		myIdleTrucksByLocation = new HashMap<TransportLocationIfc, List<SpatialResource>>();
		myIdleTrailersByLocation = new HashMap<TransportLocationIfc, List<SpatialResource>>();
	
		myNumBusyDrivers = new TimeWeighted(this,0.0,"Total Number of Busy Drivers");
		
		myNumIdleDrivers = new TimeWeighted(this,0.0,"Total Number of Idle Drivers");
		
		myNumOffDutyDrivers=new TimeWeighted(this,0.0,"Total Number of Drivers Off Duty");
		
		myNumOffDutyWithTaskDrivers=new TimeWeighted(this,0.0,"Total Number of Drivers Off Duty With Task");
		
		myNumOffDutyWithoutTaskDrivers=new TimeWeighted(this,0.0,"Total Number of Drivers Off Duty Without Task");
		
		myNumBusyTrucks = new TimeWeighted(this,0.0,"Total Number of Busy Trucks");
		
		myNumBusyTrailers = new TimeWeighted(this,0.0,"Total Number of Busy Trailers");
		
		myAvgNumDeadHeadMiles = new TimeWeighted(this,0.0,"Average Number of DeadHead Miles");
		
		myAvgNumBobTailMiles = new TimeWeighted(this,0.0,"Average Number of BobTail Miles");
		
		myAvgNumTransportMiles = new TimeWeighted(this,0.0,"Average Number of Transport Miles");
		
		myAverageDeadHeadMiles = new ResponseVariable(this,"Statistic Response: Average Number of DeadHead Miles");//TODO
		
		myAverageBobTailMiles = new ResponseVariable(this,"Statistic Response: Average Number of BobTail Miles"); //TODO
		
		myAverageTransportMiles = new ResponseVariable(this,"Statistic Response: Average Number of Transport Miles");	//TODO	
		
		Collection<TransportLocationIfc> locations = network.getLocations();
		
		myLocations  = locations;
		
		setMaxDeadHeadBobTailDistance(distance);
		
		myMaxDHBTDistancesByLocation = new HashMap<TransportLocationIfc, Double>();
		
		Collection<TransportLocationIfc> origins = network.getOrigins();
		
		for(TransportLocationIfc location: origins){			
			// will hold the list of drivers for this location
			myIdleDriversByLocation.put(location,new ArrayList<Driver>());
			myDriverLocationAssignments.put(location,new ArrayList<Driver>());
			
			myIdleTrucksByLocation.put(location,new ArrayList<SpatialResource>());
			myTruckLocationAssignments.put(location,new ArrayList<SpatialResource>());
			
			myIdleTrailersByLocation.put(location,new ArrayList<SpatialResource>());
			myTrailerLocationAssignments.put(location,new ArrayList<SpatialResource>());
			
			myMaxDHBTDistancesByLocation.put(location, new Double(distance));
		}
}
	
	/** Assigns a driver to the supplied location and adds the driver to the network
	 *  Also adds driver to the overall driver list in the network, idle lists and 
	 *  assigned lists at the location supplied 
	 *  @param location
	 */
	public final void addDriver(TransportLocationIfc location, double onDutyLimit,double drivingLimit,double longOnDutyTimeLimit, 
			DistributionIfc restTimeDistributionForFourteenHourRule,DistributionIfc restTimeDistributionForTenHourRule,
			DistributionIfc restTimeDistributionForThirtyFourHourRule,DistributionIfc velocityDistribution){
		
		if (location == null)
			throw new IllegalArgumentException("The supplied location was null.");

		if (!getNetwork().containsLocation(location))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		Driver driver =new Driver(this,location.getCurrentPosition(),location,onDutyLimit,drivingLimit,longOnDutyTimeLimit,restTimeDistributionForFourteenHourRule,
				restTimeDistributionForTenHourRule,restTimeDistributionForThirtyFourHourRule,velocityDistribution);
		
		// will hold the list of drivers for this location
		List<Driver> drivers = null;
		
		// get driver assignment list for this location
		if (!myDriverLocationAssignments.containsKey(location)){
			drivers = new ArrayList<Driver>();
			myDriverLocationAssignments.put(location, drivers);
		} else
			drivers = myDriverLocationAssignments.get(location);
		// add driver to assignment list
		drivers.add(driver);
		
		// get idle driver list by location
		if (!myIdleDriversByLocation.containsKey(location)){
			drivers = new ArrayList<Driver>();
			myIdleDriversByLocation.put(location, drivers);
		} else
			drivers = myIdleDriversByLocation.get(location);
		// add driver to idle driver list
		drivers.add(driver);
		// add driver to overall list of drivers
		myDrivers.add(driver);
		//System.out.println("At t> "+getTime()+" Driver "+driver.getId()+" added at "+location.getName());
		driver.setNetwork(getNetwork());
		myNumIdleDrivers.increment();
		setTotalNumberOfDrivers(myNumIdleDrivers.getValue());
	}

	/** Assigns a trailer to the supplied location and adds the trailer to the network
	 *  Also adds trailer to the overall trailer list in the network, idle lists and 
	 *  assigned lists at the supplied location
	 *  @param location
	 */
	public final void addTrailer(TransportLocationIfc location){
		if (location == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!getNetwork().containsLocation(location))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		// create the trailer
		SpatialResource trailer = new SpatialResource(this, location.getCurrentPosition());
		
		// will hold the list of trailers for this location
		List<SpatialResource> trailers = null;
		
		// get trailer assignment list for this location
		if (!myTrailerLocationAssignments.containsKey(location)){
			trailers = new ArrayList<SpatialResource>();
			myTrailerLocationAssignments.put(location, trailers);
		} else
			trailers = myTrailerLocationAssignments.get(location);
		// add trailer to assignment list
		trailers.add(trailer);
		
		// get idle trailer list by location
		if (!myIdleTrailersByLocation.containsKey(location)){
			trailers = new ArrayList<SpatialResource>();
			myIdleTrailersByLocation.put(location, trailers);
		} else
			trailers = myIdleTrailersByLocation.get(location);
		// add trailer to idle trailer list
		trailers.add(trailer);
		// add trailer to overall list of trailers
		myTrailers.add(trailer);
	//	System.out.println("At t> "+getTime()+" Trailer "+trailer.getId()+" added at "+location.getName());
	}

	/** Assigns a truck to the supplied location and adds the truck to the network
	 *  Also adds truck to the overall truck list in the network, idle lists and 
	 *  assigned lists at the supplied location
	 * @param location
	 */
	public final void addTruck(TransportLocationIfc location){
		if (location == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!getNetwork().containsLocation(location))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		// create the truck
		SpatialResource truck = new SpatialResource(this, location.getCurrentPosition());
		
		// will hold the list of trucks for this location
		List<SpatialResource> trucks = null;
		
		// get truck assignment list for this location
		if (!myTruckLocationAssignments.containsKey(location)){
			trucks = new ArrayList<SpatialResource>();
			myTruckLocationAssignments.put(location, trucks);
		} else
			trucks = myTruckLocationAssignments.get(location);
		// add truck to assignment list
		trucks.add(truck);
		
		// get idle truck list by location
		if (!myIdleTrucksByLocation.containsKey(location)){
			trucks = new ArrayList<SpatialResource>();
			myIdleTrucksByLocation.put(location, trucks);
		} else
			trucks = myIdleTrucksByLocation.get(location);
		// add truck to idle truck list
		trucks.add(truck);
		// add truck to overall list of trucks
		myTrucks.add(truck);
		//System.out.println("At t> "+getTime()+" Truck "+truck.getId()+" added at "+location.getName());
	}

	/**
	 * 
	 * @return trailer selection rule
	 */
	public final TrailerSelectionRuleIfc getTrailerSelectionRule() {
		return myTrailerSelectionRule;
	}
	
	/** sets the trailer selection rule
	 * 
	 * @param trailerSelectionRule
	 */
	public final void setTrailerSelectionRule(TrailerSelectionRuleIfc trailerSelectionRule) {
		myTrailerSelectionRule = trailerSelectionRule;
	}
	
	/**
	 * 
	 * @return truck selection rule
	 */
	public final TruckSelectionRuleIfc getTruckSelectionRule() {
		return myTruckSelectionRule;
	}
	
	/** sets truck selection rule
	 * 
	 * @param truckSelectionRule
	 */
	public final void setTruckSelectionRule(TruckSelectionRuleIfc truckSelectionRule) {
		myTruckSelectionRule = truckSelectionRule;
	}

	/**
	 * @return Returns DriverSelectionRule.
	 */
	public final DriverSelectionRuleIfc getDriverSelectionRule() {
		return myDriverSelectionRule;
	}

	// TODO this is not implemented properly
	/** sets driver selection rule to the rule supplied 
	 * @param myDriverSelectionRule 
	 */
	public final void setDriverSelectionRule(DriverSelectionRuleIfc driverSelectionRule) {
		
		if(driverSelectionRule==null){
		myDriverSelectionRule = new DriverSelectionRule();
		myDriverSelectionRule.setTypeToSelectWithMinDeliverTime();
		}
		else 
			myDriverSelectionRule =driverSelectionRule;
	}
	
	/**
	 * 
	 * @param location
	 * @return true if an idle driver is available at the location supplied
	 *         false if the idle list is empty
	 */
	public final boolean driverIsIdleAt(TransportLocationIfc location){
		List<Driver> drivers = myIdleDriversByLocation.get(location);
		if (drivers == null)
			return(false);
		else
			return(!drivers.isEmpty());
		   
	}

	/**
	 * 
	 * @param location
	 * @return true if an idle truck is available at the location supplied
	 *         false if the idle list is empty
	 */
	public final boolean truckIsIdleAt(TransportLocationIfc location){
		List<SpatialResource> trucks = myIdleTrucksByLocation.get(location);
		if (trucks == null)
			return(false);
		else
			return(!trucks.isEmpty());
		}

	/**
	 * 
	 * @param location
	 * @return true if an idle trailer is available at the location supplied
	 *         false if the idle list is empty
	 */
	 public final boolean trailerIsIdleAt(TransportLocationIfc location){
		List<SpatialResource> trailers = myIdleTrailersByLocation.get(location);
		if (trailers == null)
			return(false);
		else
			return(!trailers.isEmpty());
	}
	
	/**
	 * 
	 * @param location
	 * @return number of idle drivers at the location supplied
	 */
	 public final int getNumIdleDrivers(TransportLocationIfc location){
		return(myIdleDriversByLocation.get(location).size());
	}

	/**
	 * 
	 * @param location
	 * @return number of idle trucks at the location supplied
	 */
	 public final int getNumIdleTrucks(TransportLocationIfc location){
		return(myIdleTrucksByLocation.get(location).size());
	}
	
	/**
	 * 
	 * @param location
	 * @return number of idle trailers at the location supplied
	 */
	 public final int getNumIdleTrailers(TransportLocationIfc location){
		return(myIdleTrailersByLocation.get(location).size());
	}

	/** Given a collection of locations, this method finds the location that
	 *  has most idle drivers and trucks in usable combination. If no location
	 *  is found that has a usable (driver, truck) pair, then this method returns
	 *  null. The number of usable (driver, truck) pairs at a location is the 
	 *  min(#idle drivers, #idle trucks). This method is useful in identifying
	 *  a location that can be used as a place to initiate a bobtail
	 * 
	 * @param locations
	 * @return
	 */
	public TransportLocationIfc findLocationWithMostIdleDriversAndTrucks(Collection<TransportLocationIfc> locations){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(TransportLocationIfc location: locations){
			int nd = getNumIdleDrivers(location);
			int nt = getNumIdleTrucks(location);
			int nr = Math.min(nd, nt);
			if (nr > max){
				max = nr;
				mLoc = location;
			}
		}
		return(mLoc);
	}
	
	/** Given a collection of locations, this method returns a sorted set of locations
	 *  that have idle drivers and trucks in usable combination. If no location
	 *  is found that has a usable (driver, truck) pair, then this method returns
	 *  an empty set. The number of usable (driver, truck) pairs at a location is the 
	 *  min(#idle drivers, #idle trucks). This method is useful in identifying
	 *  locations that can be used as a place to initiate a bobtail
	 *  
	 * @param locations
	 * @return SortedSet<TransportLocationIfc>
	 */
	public SortedSet<TransportLocationIfc> getLocationsSortedByMostIdleDriversAndTrucks(
			Collection<TransportLocationIfc> locations){
		SortedSet<TransportLocationIfc> set = new TreeSet<TransportLocationIfc>(getDriverTruckLocationComparator());
		for(TransportLocationIfc loc: locations){
			if (getNumIdleTrailers(loc) > 0)
				set.add(loc);
		}
		return(set);
	}
	
	/** From a given origin, return a destination on its incoming lanes that has the most idle
	 *  driver, truck pairs within the specified distance If no destination has an idle pair then null is returned
	 *  The natural use of this method is to supply a list of incoming lanes.  The origin on an incoming
	 *  lane is the location that the resources might be sent from.
	 * 
	 * @param origin The location that wants to identify the resources
	 * @param distance
	 * @return location with most (driver, truck) resources
	 */
	public TransportLocationIfc getDestinationWithMostIdleDriversAndTrucks(TransportLocationIfc origin, double distance){
		return(getLocationWithMostIdleDriversAndTrucks(getNetwork().getIncomingLanes(origin), distance));
	}
	
	/** Given a collection of lanes find the origin that has the most usable idle driver, truck
	 *  pairs within the specified distance. If no lane origin has idle pairs then, null is returned
	 *  The natural use of this method is to supply a list of incoming lanes.  The origin on an incoming
	 *  lane is the location that the resources might be sent from.
	 *  
	 * @param lanes
	 * @param distance
	 * @return location with most (driver, truck) resources
	 * @return
	 */
	public TransportLocationIfc getLocationWithMostIdleDriversAndTrucks(Collection<LaneIfc> lanes, double distance){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc origin = lane.getOrigin();
				int nd = getNumIdleDrivers(origin);
				int nt = getNumIdleTrucks(origin);
				int nr = Math.min(nd, nt);
				if (nr > max){
					max = nr;
					mLoc = origin;
				}				
			}
		}
		return(mLoc);
	}

	/** Return the closest location within the given distance that has an idle (driver, truck) combination
	 * 
	 * @param origin
	 * @param distance
	 * @return
	 */
	public TransportLocationIfc getFirstLocationWithIdleDriversAndTrucks(TransportLocationIfc origin, double distance){
		//System.out.println(" "+getNetwork().getIncomingLanes(origin));
		return(getFirstLocationWithIdleDriversAndTrucks(getNetwork().getIncomingLanes(origin), distance));
	}
	
	/** Given a collection of lanes return the origin of the lane that has an idle (driver, truck) combination
	 *  within the specified distance. If no lane origin has idle pairs then, null is returned
	 *  The natural use of this method is to supply a sorted list of incoming lanes by distance.  The origin on an incoming
	 *  lane is the location that the resources might be sent from. The sort causes this method to 
	 *  return the closest location that has an idle (driver, truck) combination.
	 *  	
	 * @param lanes
	 * @param distance
	 * @return
	 *
	 */
	public TransportLocationIfc getFirstLocationWithIdleDriversAndTrucks(Collection<LaneIfc> lanes, double distance){
		TransportLocationIfc mLoc = null;
		for(LaneIfc lane: lanes){
		//	System.out.println(lane.getOrigin().getName());
			if (lane.getDistance() <= distance){
			//	System.out.println("I came here after distance checking");
				 TransportLocationIfc origin = lane.getOrigin();
				 if (driverIsIdleAt(origin) && truckIsIdleAt(origin)){
				//	  System.out.println("I came here at origin:"+origin.getName());
					  return(origin);	
				  }
				}
			}
		return(mLoc);
	}
	
	
	/** Given a collection of locations, this method finds the location that
	 *  has the most idle trailers. If no location
	 *  is found that has idle trailer, then this method returns
	 *  null. This method is useful in identifying
	 *  a location that can be used as a place to send a bobtail.  The location
	 *  may or may not have loads to pick up.
	 * 
	 * @param locations
	 * @return
	 */
	public TransportLocationIfc findLocationWithMostIdleTrailers(Collection<TransportLocationIfc> locations){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(TransportLocationIfc location: locations){
			int nt = getNumIdleTrailers(location);
			if (nt > max){
				max = nt;
				mLoc = location;
			}
		}
		return(mLoc);
	}
	
	/** Given a collection of locations, this method returns a sorted set of locations
	 *  that have trailers. The set is sorted from most idle to least idle. If no location
	 *  is found that has idle trailers, then this method returns
	 *  an empty set.  This method is useful in identifying
	 *  locations that may be useful to send a bobtail to.  The locations may or may
	 *  not have loads waiting.
	 *  
	 * @param locations
	 * @return SortedSet<TransportLocationIfc>
	 */
	public SortedSet<TransportLocationIfc> getLocationsSortedByMostIdleTrailers(
			Collection<TransportLocationIfc> locations){
		SortedSet<TransportLocationIfc> set = new TreeSet<TransportLocationIfc>(getTrailerLocationComparator());
		for(TransportLocationIfc loc: locations){
			if (getNumIdleTrailers(loc) > 0)
				set.add(loc);
		}
		return(set);
	}

	/** From a given origin, return a destination on its outgoing lanes that has the most idle
	 *  trailers.  If no destination has an idle trailer then null is returned. This method
	 *  is useful in looking for an idle trailer to match with a location that might
	 *  be a bobtail initiation site. Do not consider locations further than the supplied distance.
	 * 
	 * @param origin
	 * @param distance
	 * @return
	 */
	public TransportLocationIfc getDestinationWithMostIdleTrailers(TransportLocationIfc origin, double distance){
		return(getLocationWithMostIdleTrailers(getNetwork().getOutGoingLanes(origin), distance));
	}

	/** From a collection of lanes return a destination that has the most idle
	 *  trailers.  If no destination has an idle trailer then null is returned. This method
	 *  is useful in looking for an idle trailer to match with a location that might
	 *  be a bobtail initiation site.  Do not consider locations further than the supplied distance
	 * 
	 * @param lanes
	 * @param distance
	 * @return
	 */
	public TransportLocationIfc getLocationWithMostIdleTrailers(Collection<LaneIfc> lanes, double distance){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc dest = lane.getDestination();
				int nt = getNumIdleTrailers(dest);
				if (nt > max){
					max = nt;
					mLoc = dest;
				}				
			}
		}
		return(mLoc);
	}
	
	/** Given a collection of locations, this method finds the location that
	 *  has most idle drivers, trucks, trailers in usable combination. If no location
	 *  is found, then this method returns null. The number of usable resources at a location is the 
	 *  min(#idle drivers, #idle trucks, #idle trailers). This method is useful in identifying
	 *  a location that can be used as a place to initiate a deadhead
	 * 
	 * @param locations
	 * @return
	 */
	public TransportLocationIfc findLocationWithMostIdleResources(Collection<TransportLocationIfc> locations){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(TransportLocationIfc location: locations){
			int nd = getNumIdleDrivers(location);
			int nt = getNumIdleTrucks(location);
			int nr = Math.min(nd, nt);
			nr = Math.min(nr, getNumIdleTrailers(location));
			if (nr > max){
				max = nr;
				mLoc = location;
			}
		}
		return(mLoc);
	}
	
	/** From a given location, return a origin on its incoming lanes that has the most idle
	 *  resources.  If no origin location has usable idle resources then null is returned. This method
	 *  is useful in looking for idle resources to match with a location may need a deadhead
	 *  The natural use of this method supplies a list of incoming lanes to the given origin.  
	 *  The origin on an incoming lane is the location that the resources might be sent from. 
	 * 
	 * @param origin The location that wants to identify the resources
	 * @param distance Only consider those locations within this distance
	 * @return
	 */
	public TransportLocationIfc getDestinationWithMostIdleResources(TransportLocationIfc origin, double distance){
		return(getLocationWithMostIdleResources(getNetwork().getIncomingLanes(origin), distance));
	}
	
	/** Return from a collection of lanes the origin location that has the most idle
	 *  resources.  If no destination has usable idle resources then null is returned. This method
	 *  is useful in looking for idle resources to match with a location may need a deadhead
	 * 
	 * @param lanes
	 * @param distance Only consider those locations within this distance
	 * @return
	 */
	public TransportLocationIfc getLocationWithMostIdleResources(Collection<LaneIfc> lanes, double distance){
		TransportLocationIfc mLoc = null;
		int max = 0;
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc origin = lane.getOrigin();
				int nd = getNumIdleDrivers(origin);
				int nt = getNumIdleTrucks(origin);
				int nr = Math.min(nd, nt);
				nr = Math.min(nr, getNumIdleTrailers(origin));
				if (nr > max){
					max = nr;
					mLoc = origin;
				}				
			}
		}
		return(mLoc);
	}
	
	/** Given a collection of locations, this method finds a sorted set of location that
	 *  have the most idle drivers, trucks, trailers in usable combination. If no location
	 *  is found, then this method returns and emtpy set. The number of usable resources at a location is the 
	 *  min(#idle drivers, #idle trucks, #idle trailers). This method is useful in identifying
	 *  locations that can be used as a place to initiate a deadhead
	 * 
	 * @param locations
	 * @return
	 */
	public SortedSet<TransportLocationIfc> getLocationsSortedByMostIdleResources(
			Collection<TransportLocationIfc> locations){
		SortedSet<TransportLocationIfc> set = new TreeSet<TransportLocationIfc>(getDriverTruckTrailerLocationComparator());
		for(TransportLocationIfc loc: locations){
			if (getNumIdleTrailers(loc) > 0)
				set.add(loc);
		}		
		return(set);
	}

	/** Return the closest location within the given distance that has an idle (driver, truck, trailer) combination
	 * 
	 * @param origin
	 * @param distance
	 * @return
	 */
	public TransportLocationIfc getFirstLocationWithIdleResources(TransportLocationIfc origin, double distance){
		return(getFirstLocationWithIdleResources(getNetwork().getIncomingLanes(origin), distance));
	}
	
	/** Given a collection of lanes return the origin of the lane that has an idle (driver, truck, trailer) combination
	 *  within the specified distance. If no lane origin has idle resources then, null is returned
	 *  The natural use of this method is to supply a sorted list of incoming lanes by distance.  The origin on an incoming
	 *  lane is the location that the resources might be sent from. The sort causes this method to 
	 *  return the closest location that has an idle (driver, truck, trailer) combination.
	 *  	
	 * @param lanes
	 * @param distance
	 * @return
	 *
	 */
	public TransportLocationIfc getFirstLocationWithIdleResources(Collection<LaneIfc> lanes, double distance){
		TransportLocationIfc mLoc = null;
		for(LaneIfc lane: lanes){
			//System.out.println("Lane with Destination as "+lane.getDestination().getName()+" "+lane);
			if (lane.getDistance() <= distance){
				TransportLocationIfc origin = lane.getOrigin();
				if (driverIsIdleAt(origin) && truckIsIdleAt(origin) && trailerIsIdleAt(origin))
					return(origin);		
				}
		}
		return(mLoc);
	}
	
	public String toString(){
		StringBuffer s = new StringBuffer();
		
		s.append("----- Driver Assignments ----- \n");
		
		for(TransportLocationIfc location: myDriverLocationAssignments.keySet()){
			int i = getNetwork().getIndexOf(location);
			s.append("Location" + i + "\t" + location.getName() + " : \t");
			for(Driver d: myDriverLocationAssignments.get(location)){
				s.append(d.getId() + " " + d.getName() + " ");
			}
			s.append("\n");
		}

		s.append("----- Truck Assignments ----- \n");
		
		for(TransportLocationIfc location: myTruckLocationAssignments.keySet()){
			int i = getNetwork().getIndexOf(location);
			s.append("Location" + i + "\t" + location.getName() + " : \t");
			for(SpatialResource r: myTruckLocationAssignments.get(location)){
				s.append(r.getId() + " " + r.getName() + " ");
			}
			s.append("\n");
		}

		s.append("----- Trailer Assignments ----- \n");
		
		for(TransportLocationIfc location: myTrailerLocationAssignments.keySet()){
			int i = getNetwork().getIndexOf(location);
			s.append("Location" + i + "\t" + location.getName() + " : \t");
			for(SpatialResource r: myTrailerLocationAssignments.get(location)){
				s.append(r.getId() + " " + r.getName() + " ");
			}
			s.append("\n");
		}

		return(s.toString());
	}

	//  ===========================================
	//  PROTECTED METHODS
	//  ===========================================

	/**
	 * Returns the list of Drivers 
	 * @return 
	 */
	protected final List<Driver> getDrivers() {
		return myDrivers;
	}
	
	/**
	 * Returns the list of Trailers
	 * @return   
	 */
	protected final List<SpatialResource> getTrailers() {
		return myTrailers;
	}
	
	/**
	 * Returns the list of Trucks
	 * @return  
	 */
	protected final List<SpatialResource> getTrucks() {
		return myTrucks;
	}
	
	/**
	 * @return Returns the maxDHBTDistance.
	 */
	public final double getMaxDeadHeadBobTailDistance() {
		return myMaxDHBTDistance;
	}

	/**
	 * @param myMaxDHBTDistance The maxDHBTDistance to set.
	 */
	protected final void setMaxDeadHeadBobTailDistance(double maxDHBTDistance) {
		//TODO maybe check to see if it is >= than smallest lane and <= than largest lane
		if (maxDHBTDistance <= 0)
			throw new IllegalArgumentException("The supply distance must be > 0");
		
		myMaxDHBTDistance = maxDHBTDistance;
	//	System.out.println("Max DHBT distance has been set as "+myMaxDHBTDistance);
	}

	protected final void setMaxDeadHeadBobTailDistance(TransportLocationIfc location, double distance){
		//TODO maybe check to see if it is >= than smallest incoming lane and <= than largest incoming lane
		myMaxDHBTDistancesByLocation.put(location, new Double(distance));
	//	System.out.println("Max DHBT distance has been set as "+myMaxDHBTDistance+" for Location "+location.getName());
	}

	/** 
	 * @param location
	 * @return max distance allowed for bobtails and deadheads
	 */
	protected final double getMaxDeadHeadBobTailDistance(TransportLocationIfc location){
		return(Double.valueOf(myMaxDHBTDistancesByLocation.get(location)));
	}
	
	/**
	 * @param location
	 * @return list of drivers assigned at the location supplied
	 */
	protected final List<Driver> getAssignedDrivers(TransportLocationIfc location){
		return(myDriverLocationAssignments.get(location));
	}
	
	/**
	 * @param location
	 * @return list of trucks assigned at the location supplied
	 */
	protected final List<SpatialResource> getAssignedTrucks(TransportLocationIfc location){
		return(myTruckLocationAssignments.get(location));
	}
	
	/**
	 * @param location
	 * @return list of trailers assigned at the location supplied
	 */
	protected final List<SpatialResource> getAssignTrailers(TransportLocationIfc location){
		return(myTrailerLocationAssignments.get(location));
	}

	/**
	 * @param location
	 * @return list of idle drivers at the location supplied
	 */
	protected final List<Driver> getIdleDrivers(TransportLocationIfc location){
		return(myIdleDriversByLocation.get(location));
	}
		
	/**
	 *  @param location
	 * @return list of idle trucks at the location supplied
	 */
	protected final List<SpatialResource> getIdleTrucks(TransportLocationIfc location){
		return(myIdleTrucksByLocation.get(location));
	}

	/**
	 * @param location
	 * @return list of idle trailers at the location supplied
	 */
	protected final List<SpatialResource> getIdleTrailers(TransportLocationIfc location){
		return(myIdleTrailersByLocation.get(location));
	}
	
	// I have set the driver selection rule, dont think its very efficient
	/** Returns the idle driver from the idle driver list based on the 
	 *  driver selection rule at the supplied location
	 *  Default rule is select based on mim delivery time taken
	 *  Implements selectNextDriver() method
	 *  Returns null if idle driver list is empty
	 *   
	 * @param location
	 * @param task
	 * @return
	 */
	protected Driver getIdleDriver(TransportLocationIfc location,DispatchTask task){
		
		Driver driver=null;		
		List<Driver> drivers =getIdleDrivers(location);
		
		if(drivers.isEmpty()){
			return null;
		}
		else {	
		//	System.out.println(" "+drivers);
			double tmin=(drivers.get(0).getExpectedTotalTaskTime(task,location));
			driver=drivers.get(0);
			for(Driver d : drivers){
				double t=d.getExpectedTotalTaskTime(task,location);
			//	System.out.println("Driver "+d.getId()+" at "+task.getOrigin().getName()+" takes delivery task time of: "+t);
				if(t<=tmin){	
					tmin=t;
					driver=d;
					}
				}
			}
		
	//	System.out.println("Driver "+driver.getId()+" at "+task.getOrigin().getName()+" is selected "+" as it takes minimum delivery task time of: "+driver.getExpectedTotalTaskTime(task,location));
		return(driver);
	//	else {
		//	driver = drivers.get(0);		//TODO changed from remove(0) to get(0)
		//	driver = drivers.remove(0);
	//	}
	//	DriverSelectionRuleIfc driverSelectionRule = new DriverSelectionRule();
	//	driverSelectionRule.setTypeToSelectWithMinDeliverTime();
	//	 driverSelectionRule.setTypeToSelectFirst();
		
	//	driver=driverSelectionRule.selectNextDriver(this,location,task,Collections.unmodifiableList(drivers));
	//	return (driver);
	}
	
	//original
	protected Driver getIdleDriver1(TransportLocationIfc location,DispatchTask task){//TODO changes done 
		
		Driver driver=null;		
		List<Driver> drivers =getIdleDrivers(location);
		
		if(drivers.isEmpty()){
			return null;
		}
			
	//	else {
		//	driver = drivers.get(0);		//TODO changed from remove(0) to get(0)
		//	driver = drivers.remove(0);
	//	}
		DriverSelectionRuleIfc driverSelectionRule = new DriverSelectionRule();
		driverSelectionRule.setTypeToSelectWithMinDeliverTime();// option 1
	//	 driverSelectionRule.setTypeToSelectFirst(); // option 2
		
		driver=driverSelectionRule.selectNextDriver(this,location,task,Collections.unmodifiableList(drivers));
		return (driver);
	}
	
	/**
	 *  Dispatcher will ask the task to allocate driver 
	 *  Driver is removed from the idle list of the drivers at its current location
	 *  Number of busy drivers is incremented 
	 *  Number of idle drivers is decremented
	 *  @param task
	 */
	protected void allocateDriver(DispatchTask task){
		task.allocateDriver();
		List<Driver> drivers = myIdleDriversByLocation.get(task.getOrigin());
		//System.out.println("t> "+getTime()+" Task "+task.getId()+" is allocated Driver:"+task.getDriver().getId()+" from idle drivers list of Location "+task.getOrigin().getName());
		drivers.remove(task.getDriver());
		myNumBusyDrivers.increment();
		myNumIdleDrivers.decrement();
	}
	
	/** Dispatcher will ask the task to release the driver 
	 *  Driver is added to the idle list of drivers at its current location
	 *  Number of busy drivers is decremented 
	 *  Number of idle drivers is incremented
	 * 
	 * @param task
	 */
	protected void releaseDriver(DispatchTask task){
		//System.out.println("t> "+getTime()+" Task "+task.getId()+" releases Driver "+task.getDriver().getId()+" at Location "+task.getLoadDestination().getName());
		task.releaseDriver();
		List<Driver> drivers = myIdleDriversByLocation.get(task.getLoadDestination());
		drivers.add(task.getDriver());
		myNumBusyDrivers.decrement();
		myNumIdleDrivers.increment();
	}
	
	/** Returns the idle truck from the idle truck list based on the 
	 *  truck selection rule at the supplied location
	 *  Default rule is to select the first idle truck in the list
	 *  Returns null if idle driver list is empty
	 * 
	 * @param location
	 * @return
	 */
	protected SpatialResource getIdleTruck(TransportLocationIfc location){
		
		if(myTruckSelectionRule != null){
			return(myTruckSelectionRule.selectNextTruck(this, location));
		}
		else{
			List<SpatialResource> trucks = myIdleTrucksByLocation.get(location);
			if(trucks.isEmpty())
				return (null);
			else {
			//	SpatialResource truck = trucks.remove(0);
				SpatialResource truck = trucks.get(0);
				return(truck);
			}
		}
	}
	
	/** Dispatcher will ask the task to allocate truck
	 *  Driver is removed from the idle list of the trucks at its current location
	 *  Number of busy trucks is incremented 
	 *  
	 * @param task
	 */
	protected void allocateTruck(DispatchTask task){
		task.allocateTruck();
		List<SpatialResource> trucks = myIdleTrucksByLocation.get(task.getOrigin());
	//	System.out.println("t> "+getTime()+" Task "+task.getId()+" is allocated Truck:"+task.getTruck().getId()+" from idle trucks list of Location "+task.getOrigin().getName());
		trucks.remove(task.getTruck());
		myNumBusyTrucks.increment();
	}
	
	/** Dispatcher will ask the task to release the truck
	 *  Truck is added to the idle list of trucks at its current location
	 *  Number of busy trucks is decremented 
	 *   
	 * @param task
	 */
	protected void releaseTruck(DispatchTask task){
		task.releaseTruck();
		List<SpatialResource> trucks = myIdleTrucksByLocation.get(task.getLoadDestination());
	//	System.out.println("t> "+getTime()+" Task "+task.getId()+" releases Truck:"+task.getTruck().getId()+" at Location "+task.getLoadDestination().getName());
		trucks.add(task.getTruck());
		myNumBusyTrucks.decrement();
	}
	
	/** Returns the idle trailer from the idle trailer list based on the 
	 *  trailer selection rule at the supplied location
	 *  Default rule is to select the first idle trailer in the list
	 *  Returns null if idle trailer list is empty
	 * 
	 * @param location
	 * @return
	 */
	protected SpatialResource getIdleTrailer(TransportLocationIfc location){
		
	if(myTrailerSelectionRule != null){
			return(myTrailerSelectionRule.selectNextTrailer(this, location));
		}
		else{
			List<SpatialResource> trailers = myIdleTrailersByLocation.get(location);
			if(trailers.isEmpty())
				return (null);
			else {
			//	SpatialResource trailer = trailers.remove(0);
				SpatialResource trailer = trailers.get(0);
				return(trailer);
			}
		}
	}
	
	/** Dispatcher will ask the task to allocate trailer
	 *  Trailer is removed from the idle list of the trailers at its current location
	 *  Number of busy trailers is incremented 
	 *  This method is used in case of transport or bobtail task
	 * @param task
	 */
	protected void allocateTrailer(DispatchTask task){
		task.allocateTrailer();
		List<SpatialResource> trailers = myIdleTrailersByLocation.get(task.getOrigin());
	//	System.out.println("t> "+getTime()+" Task "+task.getId()+" is allocated Trailer:"+task.getTrailer().getId()+" from idle trailers list of Location "+task.getOrigin().getName());
		trailers.remove(task.getTrailer());
		myNumBusyTrailers.increment();
	}
	
	//TODO added new 
	/** Dispatcher will ask the task to allocate trailer
	 *  Trailer is removed from the idle list of the trailers at its current location
	 *  Number of busy trailers is incremented 
	 *  This method is used in case of deadhead task
	 *  
	 * @param task
	 * @param origin
	 */
	protected void allocateTrailer(DispatchTask task,TransportLocationIfc origin){
		task.allocateTrailer();
		List<SpatialResource> trailers = myIdleTrailersByLocation.get(origin);
	//	System.out.println("t> "+getTime()+" Task "+task.getId()+" is allocated Trailer:"+task.getTrailer().getId()+" from idle trailers list of Location "+origin.getName());
		trailers.remove(task.getTrailer());
		myNumBusyTrailers.increment();
	}
	
	/** Dispatcher will ask the task to release the trailer
	 *  Trailer is added to the idle list of trailers at its current location
	 *  Number of busy trailers is decremented 
	 *  
	 * @param task
	 */
	protected void releaseTrailer(DispatchTask task){
		task.releaseTrailer();
		List<SpatialResource> trailers = myIdleTrailersByLocation.get(task.getLoadDestination());
	//	System.out.println("t> "+getTime()+" Task "+task.getId()+" releases Trailer:"+task.getTrailer().getId()+" at Location "+task.getLoadDestination().getName());
		trailers.add(task.getTrailer());
		myNumBusyTrailers.decrement();
	}

	// TODO changes here are the task is set to transport type when used in 
	// getIdleDriver so that the expected time is calculated for the transport task
	/** Trys to find resources at the task's origin that will permit
	 *  a transport to occur.  Modifies the resources of the task
	 *  if all the resources are available for a transport.
	 * 
	 * @param task
	 * @return True if transport resources have been identified
	 */
	protected boolean transportAvailable(DispatchTask task){
		
		TransportLocationIfc location = task.getLoadOrigin();
		
	//	System.out.println("t> "+getTime()+" Dispatch Task "+task.getId()+" is checked for availability of Transport at location: "+task.getLoadOrigin().getName());
		
		task.setTypeToTransport();
		
		Driver driver = getIdleDriver(location,task);
		if (driver == null)
			return false;
		SpatialResource truck = getIdleTruck(location);
		if (truck == null)
			return false;
		SpatialResource trailer = getIdleTrailer(location);
		if (trailer == null)
			return false;
		// if here, then a driver, truck and trailer has been found
		
		task.setDriver(driver);
		task.setTruck(truck);
		task.setTrailer(trailer);
		
	//	System.out.println("t> "+getTime()+" Dispatch Task "+task.getId()+" has Transport available at location "+task.getLoadOrigin().getName());
		
		return true;
	}

	/** When the DispatchTask is completed then this method can be called back
	 *  to indicate the completion of the task
	 * 
	 * @param task
	 */
	abstract protected void dispatchCompleted(DispatchTask task);

	//  ===========================================
	//  INNER CLASSES
	//  ===========================================

	/** Compares two locations.  The location with the most usuable (idle)
	 *  (driver, truck) resources is considered bigger. 
	 *  Ties go to lower location number
	 *  
	 *  Equal is only via object equality
	 *
	 * Note:  This class may have a natural ordering that is inconsistent
	 * with equals.
	 * 
	 * @param loc1
	 * @param loc2
	 * @return Returns a negative integer, zero, or a positive integer
	 *
	 */ 
	public class DriverTruckLocationComparator implements Comparator<TransportLocationIfc>{

		public int compare(TransportLocationIfc loc1, TransportLocationIfc loc2) {
	        if (loc1.equals(loc2))// objects are equal with zero
	            return(0);
	        
	        List<Driver> drivers1 = getIdleDrivers(loc1);
	        List<SpatialResource> trucks1 = getIdleTrucks(loc1);
	        List<Driver> drivers2 = getIdleDrivers(loc2);
	        List<SpatialResource> trucks2 = getIdleTrucks(loc2);
	        // the number of usable resources is constrained by the smaller of the two
	        int numIdle1 = Math.min(drivers1.size(), trucks1.size());
	        int numIdle2 = Math.min(drivers2.size(), trucks2.size());
	        
	        if (numIdle1 > numIdle2)
	        	return(1);
	        else if (numIdle1 < numIdle2)
        		return(-1);
	        else {
				long id1 = loc1.getId();
				long id2 = loc2.getId();
		        	if (id1 < id2)
		        		return(-1);
		        	else
		        		return(1);	        	
			}	        	
		}	
	}

	/** Compares two locations.  The location with the most usuable (idle)
	 *  (driver, truck, trailer) resources is considered bigger. 
	 *  Ties go to lower location number
	 *  
	 *  Equal is only via object equality
	 *
	 * Note:  This class may have a natural ordering that is inconsistent
	 * with equals.
	 * 
	 * @param loc1
	 * @param loc2
	 * @return Returns a negative integer, zero, or a positive integer
	 *
	 */ 
	public class DriverTruckTrailerLocationComparator implements Comparator<TransportLocationIfc>{

		public int compare(TransportLocationIfc loc1, TransportLocationIfc loc2) {
	        if (loc1.equals(loc2))// objects are equal with zero
	            return(0);
	        
	        List<Driver> drivers1 = getIdleDrivers(loc1);
	        List<SpatialResource> trucks1 = getIdleTrucks(loc1);
	        List<SpatialResource> trailers1 = getIdleTrailers(loc1);
	        List<Driver> drivers2 = getIdleDrivers(loc2);
	        List<SpatialResource> trucks2 = getIdleTrucks(loc2);
	        List<SpatialResource> trailers2 = getIdleTrailers(loc2);
	        // the number of usable resources is constrained by the smaller of the two
	        int numIdle1 = Math.min(drivers1.size(), trucks1.size());
	        numIdle1 = Math.min(numIdle1, trailers1.size());
	        int numIdle2 = Math.min(drivers2.size(), trucks2.size());
	        numIdle2 = Math.min(numIdle2, trailers2.size());
	        
	        if (numIdle1 > numIdle2)
	        	return(1);
	        else if (numIdle1 < numIdle2)
        		return(-1);
	        else {
				long id1 = loc1.getId();
				long id2 = loc2.getId();
		        	if (id1 < id2)
		        		return(-1);
		        	else
		        		return(1);	        	
			}			
		}
	}

	/** Compares two locations.  The location with the most usuable (idle)
	 *  (trailer) resources is considered bigger. Ties go to lower location number
	 *  
	 *  Equal is only via object equality
	 *
	 * Note:  This class may have a natural ordering that is inconsistent
	 * with equals.
	 * 
	 * @param loc1
	 * @param loc2
	 * @return Returns a negative integer, zero, or a positive integer
	 *
	 */ 
	public class TrailerLocationComparator implements Comparator<TransportLocationIfc>{

		public int compare(TransportLocationIfc loc1, TransportLocationIfc loc2) {
	        if (loc1.equals(loc2))// objects are equal with zero
	            return(0);
	        
	        List<SpatialResource> trailers1 = getIdleTrailers(loc1);
	        List<SpatialResource> trailers2 = getIdleTrailers(loc2);
	        int n1 = trailers1.size();
	        int n2 = trailers2.size();
	        
	        if (n1 > n2)
	        	return(1);
	        else if (n1 < n2)
        		return(-1);
	        else {
				long id1 = loc1.getId();
				long id2 = loc2.getId();
		        	if (id1 < id2)
		        		return(-1);
		        	else
		        		return(1);	        	
			}			
        	
		}	
	}
	
	/**
	 * @return DriverTruckLocationComparator
	 */
	protected final Comparator<TransportLocationIfc> getDriverTruckLocationComparator() {
		if (myDriverTruckLocationComparator == null)
			myDriverTruckLocationComparator = new DriverTruckLocationComparator();
		return myDriverTruckLocationComparator;
	}
	
	/**
	 *  Sets DriverTruckLocationComparator
	 * @param driverTruckLocationComparator
	 */
	protected final void setDriverTruckLocationComparator(
			Comparator<TransportLocationIfc> driverTruckLocationComparator) {
		myDriverTruckLocationComparator = driverTruckLocationComparator;
	}
	
	/**
	 * 
	 * @return DriverTruckTrailerLocationComparator
	 */
	protected final Comparator<TransportLocationIfc> getDriverTruckTrailerLocationComparator() {
		if (myDriverTruckTrailerLocationComparator == null)
			myDriverTruckTrailerLocationComparator = new DriverTruckTrailerLocationComparator();
		return myDriverTruckTrailerLocationComparator;
	}
	
	/**
	 * Sets DriverTruckTrailerLocationComparator
	 * @param driverTruckTrailerLocationComparator
	 */
	protected final void setDriverTruckTrailerLocationComparator(
			Comparator<TransportLocationIfc> driverTruckTrailerLocationComparator) {
		myDriverTruckTrailerLocationComparator = driverTruckTrailerLocationComparator;
	}
	
	/**
	 * 
	 * @return TrailerLocationComparator
	 */
	protected final Comparator<TransportLocationIfc> getTrailerLocationComparator() {
		if (myTrailerLocationComparator == null)
			myTrailerLocationComparator = new TrailerLocationComparator();
		return myTrailerLocationComparator;
	}
	
	/**
	 * Sets TrailerLocationComparator
	 * @param trailerLocationComparator
	 */
	protected final void setTrailerLocationComparator(
		Comparator<TransportLocationIfc> trailerLocationComparator) {
		myTrailerLocationComparator = trailerLocationComparator;
	}
    

	/** Iterate through all the locations and get the 
	 *  idle driver, truck, trailer lists;Clear the lists 
	 *  Get the resources assigned list of those locations and 
	 *  add them to their idle list of location 
	 */
	protected void initialize(){
		super.initialize();
	//	System.out.println("IN ABSTRACTDISPATCHER initialize()");
		setInitialValues(); 		 
		 
		for(TransportLocationIfc location: myLocations){
			
			 getIdleDrivers(location).clear();
			 getIdleTrucks(location).clear();
			 getIdleTrailers(location).clear();
			 
			List<Driver> drivers = myDriverLocationAssignments.get(location);
			for (Driver d : drivers){
					getIdleDrivers(location).add(d);
				}
			
			List<SpatialResource> trucks = myTruckLocationAssignments.get(location);
			for (SpatialResource t : trucks){
					getIdleTrucks(location).add(t);
				}
						
		    List<SpatialResource> trailers = myTrailerLocationAssignments.get(location);
			
			for (SpatialResource r : trailers){
					getIdleTrailers(location).add(r);
				}
			}
	}
	
	protected void setInitialValues(){
		
		myNumBusyDrivers.setValue(0.0);
		
		myNumIdleDrivers.setValue(myTotalNumberOfDrivers);
		
		myNumOffDutyDrivers.setValue(0.0);
		
		myNumBusyTrucks.setValue(0.0);
		
		myNumBusyTrailers.setValue(0.0);
		
		myAvgNumDeadHeadMiles.setValue(0.0);
		
		myAvgNumBobTailMiles.setValue(0.0);
		
		myAvgNumTransportMiles.setValue(0.0);
	}
	
	//TODO added new 
	/** Dispatcher is notified that driver is off duty 
	 *  It has not been assigned any task
	 *  Driver is removed from the idle list at its rest location
	 *  so that it is not assigned by the dispatcher to a task
	 *  Increment number of off duty drivers
	 *  Decrement number of idle drivers
	 * @param driver
	 */
	protected void driverOffDuty(Driver driver){
		List<Driver> drivers = getIdleDrivers(driver.getCurrentLocation());
	//	System.out.println("t> "+getTime()+" Driver:"+driver.getId()+" is now OFF DUTY WITHOUT A TASK at Location "+driver.getCurrentLocation().getName());
		drivers.remove(driver);
		myNumOffDutyDrivers.increment();
		myNumIdleDrivers.decrement();
		myNumOffDutyWithoutTaskDrivers.increment();
	}
	
	//	TODO added new 
	/** Dispatcher is notified that driver is on duty 
	 *  It has not been assigned any task
	 *  Driver is added to the idle list at its rest location
	 *  Decrement number of off duty drivers
	 *  Increment number of idle drivers
	 *  
	 * @param driver
	 */
	protected void driverOnDuty(Driver driver){
		List<Driver> drivers = getIdleDrivers(driver.getCurrentLocation());
	//	System.out.println("t> "+getTime()+" Driver:"+driver.getId()+" is now ON DUTY WITHOUT A TASK at Location "+driver.getCurrentLocation().getName());
		drivers.add(driver);
		myNumIdleDrivers.increment();
		myNumOffDutyDrivers.decrement();
		myNumOffDutyWithoutTaskDrivers.decrement();
	}
	
	//	TODO added new 
	/** Dispatcher is notified that driver is off duty 
	 *  while performing a task
	 *  Increment number of off duty drivers
	 *  Decrement number of busy drivers
	 *  @param driver
	 */
	protected void driverOffDutyWithTask(Driver driver){
		//System.out.println("t> "+getTime()+" Driver:"+driver.getId()+" is now OFF DUTY WHILE ON A TASK at Location "+driver.getCurrentLocation().getName());
		myNumOffDutyDrivers.increment();
		myNumBusyDrivers.decrement();
		myNumOffDutyWithTaskDrivers.increment();
	}

	//	TODO added new 	
	/** Dispatcher is notified that driver is on duty 
	 *  while performing a task
	 *  Decrement number of off duty drivers
	 *  Increment number of busy drivers
	 * @param driver
	 */
	protected void driverOnDutyWithTask(Driver driver){
		//System.out.println("t> "+getTime()+" Driver:"+driver.getId()+" is now ON DUTY WHILE ON A TASK at Location "+driver.getCurrentLocation().getName());
		myNumBusyDrivers.increment();
		myNumOffDutyDrivers.decrement();
		myNumOffDutyWithTaskDrivers.decrement();
	}
	
	//TODO new method added
	/** Returns the location within the given distance that has an idle (driver, truck, trailer) combination
	 *  based on the minimum delivery time taken by the driver
	 * 
	 */
	public TransportLocationIfc getLocationWithIdleResourcesWithMinDeliveryTime(TransportLocationIfc origin,DispatchTask task){
		return(getLocationWithIdleResourcesWithMinDeliveryTime(getNetwork().getIncomingLanes(origin), task));
	}
	
	//	TODO new method added
	/**
	 * Returns the location within the given distance that has an idle (driver, truck) combination
	 *  based on the minimum delivery time taken by the driver
	 */
	public TransportLocationIfc getLocationWithIdleDriverandTruckWithMinDeliveryTime(TransportLocationIfc origin,DispatchTask task){
		return(getLocationWithIdleDriverandTruckWithMinDeliveryTime(getNetwork().getIncomingLanes(origin), task));
	}
	
	//	TODO new method added
	/** Given a collection of lanes return the origin of the lane that has an idle (driver, truck, trailer) combination
	 *  within the specified distance based on the minimum delivery time of the driver. If no lane origin has idle resources then, 
	 *  null is returned.Iterate through a list of incoming lanes and check for the location across all the lanes which has
	 *  idle driver that takes minimum time for completion of the task supplied
	 *  return that location for deadhead
	 * 
	 */
	public TransportLocationIfc  getLocationWithIdleResourcesWithMinDeliveryTime(Collection<LaneIfc> lanes,DispatchTask task){
		
		TransportLocationIfc mloc=null;
		
		task.setTypeToDeadHeadToLoad();
		
		List<Driver> driversAtLocation=new ArrayList<Driver>();
		
		for(LaneIfc l:lanes){
			TransportLocationIfc origin = l.getOrigin();
			if (driverIsIdleAt(origin) && truckIsIdleAt(origin) && trailerIsIdleAt(origin)){
				task.setOrigin(l.getOrigin());
				Driver d=getIdleDriverWithMinDeliveryTimeAtLocation(origin,task);
				driversAtLocation.add(d);
			}
		}
		
		if(driversAtLocation.isEmpty())
			return null;
		
		Driver driver = driversAtLocation.get(0);
		mloc = driver.getCurrentLocation();
		double tmin=driver.getExpectedTotalTaskTime(task,mloc);
		
		for(Driver d:driversAtLocation){
			TransportLocationIfc loc = d.getCurrentLocation();
			double t=d.getExpectedTotalTaskTime(task,loc);
		//	System.out.println("Driver "+d.getId()+" at "+task.getOrigin().getName()+" takes delivery task time of: "+t);
			if(t<=tmin){
				tmin=t;
				mloc=loc;
			}
		}
		//System.out.println("Location: "+mloc.getName()+" is selected for DeadHead of Task: "+task.getId());
		return (mloc);
	}
	
	//	TODO new method added
	/**
	 *  Given a collection of lanes return the origin of the lane that has an idle (driver, truck) combination
	 *  within the specified distance based on the minimum delivery time of the driver. If no lane origin has idle resources then, 
	 *  null is returned.Iterate through a list of incoming lanes and check for the location across all the lanes which has
	 *  idle driver that takes minimum time for completion of the task supplied
	 *  return that location for bobtail
	 */
	public TransportLocationIfc  getLocationWithIdleDriverandTruckWithMinDeliveryTime(Collection<LaneIfc> lanes,DispatchTask task){
		
		task.setTypeToBobTailToLoad();
		
		TransportLocationIfc mloc=null;
		
		List<Driver> driversAtLocation=new ArrayList<Driver>();
		
		for(LaneIfc l:lanes){
			TransportLocationIfc origin = l.getOrigin();
			if (driverIsIdleAt(origin) && truckIsIdleAt(origin)){
				task.setOrigin(l.getOrigin());
				Driver d=getIdleDriverWithMinDeliveryTimeAtLocation(origin,task);
				driversAtLocation.add(d);
			}
		}
		
		if(driversAtLocation.isEmpty())
			return null;
		Driver driver = driversAtLocation.get(0);
		mloc = driver.getCurrentLocation();
		double tmin=driver.getExpectedTotalTaskTime(task,mloc);
		
		for(Driver d:driversAtLocation){
			TransportLocationIfc loc = d.getCurrentLocation();
			double t=d.getExpectedTotalTaskTime(task,loc);
			if(t<=tmin){
				tmin=t;
				mloc=loc;
			}
		}
	//	System.out.println("Location: "+mloc.getName()+" is selected for BobTail of Task: "+task.getId());
		return (mloc);
	}
	
	/**
	 * Returns the location within the given distance that has an idle (driver, truck) combination
	 *  based on the minimum delivery time taken by the driver
	 *  
	 * @param location
	 * @param task
	 * @return
	 */
	public Driver getIdleDriverWithMinDeliveryTimeAtLocation(TransportLocationIfc location, DispatchTask task){
		
		List<Driver> drivers = getIdleDrivers(location);
		Driver driver=null;
		if(drivers.isEmpty())
			return (driver);
		else{			
			double tmin=(drivers.get(0).getExpectedTotalTaskTime(task,location));
			driver = drivers.get(0);		
			for(Driver d:drivers){
				double t=d.getExpectedTotalTaskTime(task,location);
				if(t<=tmin){
					tmin=t;
					driver=d;
				}
			}
		//	System.out.println("Driver "+driver.getId()+" at "+task.getOrigin().getName()+" is selected at that location ");
			return(driver);
		}
	}

	/**
	 * 
	 * @return average number of deadhead miles covered 
	 */
	protected TimeWeighted getAvgNumDeadHeadMiles() {
		return myAvgNumDeadHeadMiles;
	}

	/**
	 * 
	 * @return average number of bobtail miles covered 
	 */
	protected TimeWeighted getAvgNumBobTailMiles() {
		return myAvgNumBobTailMiles;
	}

	/**
	 * 
	 * @return average number of transport miles covered 
	 */
	protected TimeWeighted getAvgNumTransportMiles() {
		return myAvgNumTransportMiles;
	}

	protected double getTotalNumberOfDrivers() {
		return myTotalNumberOfDrivers;
	}

	protected void setTotalNumberOfDrivers(double totalNumberOfDrivers) {
		myTotalNumberOfDrivers = totalNumberOfDrivers;
	}

}
