package jslx.transportlayerFinal;

import jsl.modeling.elements.queue.QObject;
import jsl.modeling.elements.resource.Request;
import jsl.modeling.elements.spatial.SpatialResource;
import jsl.utilities.reporting.JSL;

/** The dispatcher uses DispatchTasks to make moves within the network
 * 
 */
public class DispatchTask extends QObject{

	//  ===========================================
	//  CLASS ATTRIBUTES
	//  ===========================================
	
	/** Constant to represent that the DispatchTask is a bob tail
	 *  that has been dispatched to pick up a load
	 */
	public static final int BOBTAIL_TO_LOAD = JSL.getNextEnumConstant();

	/** Constant to represent that the DispatchTask is a dead head 
	 *  that has been dispatched to pick up a load
	 */
	public static final int DEADHEAD_TO_LOAD = JSL.getNextEnumConstant();

	/** Constant to represent that the DispatchTask is a bob tail
	 *  that has been dispatched but there is no load to pick up
	 *  There can be a no load bob tail when a driver and truck
	 *  have simply been told to move to another location
	 */
	public static final int BOBTAIL_NO_LOAD = JSL.getNextEnumConstant();

	/** Constant to represent that the DispatchTask is a dead head that
	 *  has been dispatched with out a corresponding load.  The deadhead
	 *  is simply to move the driver, truck, trailer to another location
	 *  w/o picking up a load when they arrive
	 */
	public static final int DEADHEAD_NO_LOAD = JSL.getNextEnumConstant();

	/** Constant to represent that the DispatchTask is a transport
	 */
	public static final int TRANSPORT = JSL.getNextEnumConstant();

	//  ===========================================
	//  OBJECT ATTRIBUTES
	//  ===========================================
		
	/** Represents dispatcher of the task
	 */
	private AbstractDispatcher myDispatcher;
		
	/** Represents the destination of the task 
	 */
	private TransportLocationIfc myDestination;

	/** Represents the destination of the task 
	 */
	private TransportLocationIfc myOrigin;

	/** Represents the driver associated with the dispatch task 
	 */
	private Driver myDriver;
	
	/** The request make for the driver resource
	 */
	private Request myDriverRequest;
	
	/** Represents the truck associated with the dispatch task 
	 */
	private SpatialResource myTruck;
	
	/** The request made on the truck resource
	 */
	private Request myTruckRequest;

	/** Represents the load associated with task if any 
	 */
	private Load myLoad;

	/** Represents the trailer associated with the load 
	 */
	private SpatialResource myTrailer;

	/** Represents the request for the trailer resource
	 */
	private Request myTrailerRequest;
	
	/** Represents the type (BOBTAIL_TO_LOAD, DEADHEAD_TO_LOAD, 
	 * BOBTAIL_NO_LOAD, DEADHEAD_NO_LOAD, TRANSPORT)
	 */
	private int myTaskType;
	
	/** Represnts time taken for the transport of this task
	 * 
	 */
	private double myTaskTime;
	
	/** Represents the expected time taken for the transport of this task
	 * 
	 */
	private double myExpectedTaskTime;
	
	//  ===========================================
	//  CONSTRUCTORS
	//  ===========================================
	
	/** A DispatchTask is created by dispatchers to carry out moves in the network
	 * 
	 * @param dispatcher
	 * @param destination
	 */
	public DispatchTask(AbstractDispatcher dispatcher, Load load){
		setDispatcher(dispatcher);
		setLoad(load);
	}
	
	
	//  ===========================================
	//  PUBLIC METHODS
	//  ===========================================
	
	/** Gets destination of this task 
	 * 
	 * @return Destination
	 */
	public final TransportLocationIfc getLoadDestination(){
		return myLoad.getDestination();
	}

	/** Gets origin of this task 
	 * 
	 * @return Destination
	 */
	public final TransportLocationIfc getLoadOrigin(){
		return myLoad.getOrigin();
	}

	/** Checks if this task has assigned driver
	 * 
	 * @return
	 */
	public final boolean hasDriver(){
		return(myDriver != null);
	}
	
	/** Checks if the task has assigned truck
	 * 
	 * @return
	 */
	public final boolean hasTruck(){
		return(myTruck != null);
	}
	
	/** Checks if this task has assigned trailer
	 * 
	 * @return
	 */
	public final boolean hasTrailer() {
		return(myTrailer != null);
	}

	/** Checks if this task has a driver allocated
	 * 
	 * @return
	 */
	public final boolean hasDriverAllocated(){
		return(myDriverRequest != null);
	}

	/** Checks if this task has a truck allocated
	 * 
	 * @return
	 */
	public final boolean hasTruckAllocated(){
		return(myTruckRequest != null);
	}

	/** Checks if this task has a trailer allocated
	 * 
	 * @return
	 */
	public final boolean hasTrailerAllocated(){
		return(myTrailerRequest != null);
	}

	/**
	 * 
	 * @return Returns the load of this task
	 */
	public final Load getLoad() {
		return myLoad;
	}

	/** sets this task to BobTailToLoad type
	 * 
	 */
	public final void setTypeToBobTailToLoad(){
		setDestination(getLoadOrigin());
	//	System.out.println("Task is set as Bobtail To Load Task");
		myTaskType = BOBTAIL_TO_LOAD;
	}
	
	/** sets this task to DeadHeadToLoad type 
	 *
	 */
	public final void setTypeToDeadHeadToLoad(){
		setDestination(getLoadOrigin());
	//	System.out.println("Task is set as Deadhead To Load Task");
		myTaskType = DEADHEAD_TO_LOAD;
	}
	
	/** sets this task to BobTailNoLoad type
	 * 
	 */
	public final void setTypeToBobTailNoLoad(){
	//	System.out.println("Task is set as Bobtail No Load Task");
		myTaskType = BOBTAIL_NO_LOAD;
	}
	
	/** sets this task to DeadHeadNoLoad type 
	 *
	 */
	public final void setTypeToDeadHeadNoLoad(){
	//	System.out.println("Task is set as Deadhead No Load Task");
		myTaskType = DEADHEAD_NO_LOAD;
	}

	/** sets this task to Transport type 
	 *
	 */
	public final void setTypeToTransport(){
		setOrigin(getLoadOrigin());
		setDestination(getLoadDestination());
	//	System.out.println("Task is set as Transport Task");
		myTaskType = TRANSPORT;
	}
	
	/** checks if the origin of this task is same as the 
	 *  origin of its load
	 * 
	 * @return
	 */
	public final boolean areOriginsSame(){
		return(getOrigin() == getLoadOrigin());
	}
	
	/** checks if this task is of BobTailToLoad type
	 * 
	 * @return
	 */
	public final boolean isBobTailToLoad(){
		return(myTaskType == BOBTAIL_TO_LOAD);
	}
	
	/** checks if this task is of DeadHeadToLoad type
	 * 
	 * @return
	 */
	public final boolean isDeadHeadToLoad(){
		return(myTaskType == DEADHEAD_TO_LOAD);
	}

	/** checks if this task is of BobTailNoLoad type
	 * 
	 * @return
	 */
	public final boolean isBobTailNoLoad(){
		return(myTaskType == BOBTAIL_NO_LOAD);
	}
	
	/** checks if this task is of DeadHeadNoLoad type
	 * 
	 * @return
	 */
	public final boolean isDeadHeadNoLoad(){
		return(myTaskType == DEADHEAD_NO_LOAD);
	}

	/** checks if this task is of Transport type
	 * 
	 * @return
	 */
	public final boolean isTransport(){
		return(myTaskType == TRANSPORT);	
	}
	
	/** 
	 * 
	 * @return the Type of this Task
	 */
	public final int getTaskType(){
		return(myTaskType);
	}
			
	//  ===========================================
	//  PROTECTED METHODS
	//  ===========================================
	
	/** Sets the dispatcher for this task
	 * 
	 * @param dispatcher The dispatcher to set, must not be null
	 */
	protected final void setDispatcher(AbstractDispatcher dispatcher){
		if (dispatcher == null)
			throw new IllegalArgumentException("The dispatcher was null!");
		myDispatcher = dispatcher;
	}

	/**  
	 * @return dispatcher for this task
	 */
	protected final AbstractDispatcher getDispatcher(){
		return myDispatcher;
	}	

	/** Sets the driver assigned for this task
	 * 
	 * @param driver
	 */
	protected final void setDriver(Driver driver) {
		myDriver = driver;
	}

	/** 
	 * @return the driver assigned for this task
	 */
	 protected final Driver getDriver() {
		return myDriver;
	}

	/** Sets the driver request for this task
	 * 
	 * @param driverRequest
	 */
	 protected final void setDriverRequest(Request driverRequest) {
		myDriverRequest = driverRequest;
	}

	/** 
	 * 
	 * @return the driver request for this task
	 */
	 protected final Request getDriverRequest() {
		return myDriverRequest;
	}

	/** Sets the truck assigned for this task 
	 * 
	 * @param truck
	 */
	 protected final void setTruck(SpatialResource truck) {
		myTruck = truck;
	}
	
	/** 
	 * @return the truck assigned for this task
	 */
	 protected final SpatialResource getTruck() {
		return myTruck;
	}
	
	/** Sets the truck request for this task
	 * 
	 * @param truckRequest
	 */
	 protected final void setTruckRequest(Request truckRequest) {
		myTruckRequest = truckRequest;
	}

	/**
	 * 
	 * @return truck request for this task
	 */
	 protected final Request getTruckRequest() {
		return myTruckRequest;
	}

	/**
	 * 
	 * @return trailer assigned for the task
	 */
	 protected final SpatialResource getTrailer() {
		return myTrailer;
	}
	
	/** sets the trailer assigned for this task
	 * 
	 * @param trailer
	 */
	 protected final void setTrailer(SpatialResource trailer) {
		myTrailer = trailer;
	}

	/**
	 * @return the trailer request of the task
	 */
	 protected final Request getTrailerRequest() {
		return myTrailerRequest;
	}

	/** Sets the Trailer request for this task
	 * 
	 * @param trailerRequest
	 */
	protected final void setTrailerRequest(Request trailerRequest) {
		myTrailerRequest = trailerRequest;
	}

	/** Sets Load for this Task
	 * @param myLoad 
	 */
	protected final void setLoad(Load load) {
		if (load == null)
			throw new IllegalArgumentException("The load cannot be null!");
		myLoad = load;
	}

	/** Sets destination of this task
	 * 
	 * @param load
	 */
	protected final void setDestination(TransportLocationIfc destination){
		if (destination == null)
			throw new IllegalArgumentException("The destination was null!");
		myDestination= destination;
	}

	/** 
	 * @return destination of this task
	 */
	protected final TransportLocationIfc getDestination() {
		return myDestination;
	}

	/**
	 * @return origin of this task
	 */
	protected final TransportLocationIfc getOrigin() {
		return myOrigin;
	}

	/** Sets origin of this task
	 * 
	 * @param load
	 */
	protected final void setOrigin(TransportLocationIfc origin){
		if (origin == null)
			throw new IllegalArgumentException("The origin was null!");
		myOrigin = origin;
	}
	
	/**
	 *  Task is waiting to be executed 
	 *  Task will ask the driver to execute
	 *
	 */
	protected void waitingForTransport(){
	//	System.out.println("t>"+getTime()+" Driver "+myDriver.getId()+" will be asked to perform the Task "+getId());
		myDriver.performTask(this);
	}
	
	/** 
	 * Allocates driver to this task
	 */
	protected final void allocateDriver(){
		myDriverRequest = myDriver.allocate();			
	}
	
	/**
	 * Allocates truck to this task
	 */
	protected final void allocateTruck(){
		myTruckRequest = myTruck.allocate();		
	}
	
	/**
	 * Allocates trailer to this task
	 */
	protected final void allocateTrailer(){
		myTrailerRequest = myTrailer.allocate();		
	}
	
	/**
	 * Releases driver allocated for this task
	 *
	 */
	protected void releaseDriver(){
		myDriver.release(myDriverRequest);
	}
	
	/**
	 * Releases truck allocated for this task
	 *
	 */
	protected void releaseTruck(){
		myTruck.release(myTruckRequest);
	}
	
	/**
	 * Releases trailer allocated for this task
	 *
	 */
	protected void releaseTrailer(){
		myTrailer.release(myTrailerRequest);
	}
		
	/**
	 *  Task notifies dispatcher that it has been executed
	 *
	 */
	protected void taskCompleted(){
		myDispatcher.dispatchCompleted(this);
	}
	
	//TODO added new 
	/** 
	 * 
	 * @return Time taken to execute the transport of this task
	 */
	protected double getTransportTime(Driver driver){
		Driver d = driver;
		if(d==null)
			throw new IllegalArgumentException("The Driver was null");
		LaneIfc l = getDispatcher().getNetwork().getLane(getOrigin(),getDestination());
		double t=l.getDistance();
		double m = d.getRandomVelocity();
		myTaskTime = (t/m);
		return myTaskTime;
	}
	
	// TODO added new 
	/**
	 * 
	 * @return Expected time taken to execute the transport of this task
	 */
	protected double getExpectedTransportTime(Driver driver){
		Driver d = driver;
		if(d==null)
			throw new IllegalArgumentException("The Driver was null");
		LaneIfc l = getDispatcher().getNetwork().getLane(getOrigin(),getDestination());
		double t = l.getDistance();
		double m = d.getVelocityDistribution().getMean();
		myExpectedTaskTime=(t/m);
		return myExpectedTaskTime;
	}
 }
