package jslx.transportlayerFinal;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.queue.Queue;
import jsl.modeling.elements.variable.ResponseVariable;
import jsl.modeling.elements.variable.ResponseVariableAverageObserver;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.utilities.statistic.Statistic;

/** DispatcherFTL is a subclass of AbstractDispatcherFTL. It has the data structures which holds queue of all
 *  the loads in the network waiting to be dispatched, queues of the load at their respective origins, queues of the 
 *  loads at their origins waiting for a bobtail at their origin, queues of the loads at their origins 
 *  waiting for a deadhead, queues of the loads at their origins waiting to be transported to their destinations.
 *  It also has methods which facilitate in retrieving the locations which could require bobtail or deadhead.
 *  DispatcherFTL receives load from the TransportationNetwork. It checks to see if the load has an idle truck, 
 *  trailer and driver at its origin to facilitate its transportation to its destination. If there happens to be an 
 *  idle trailer then the DispatcherFTL creates a bobtail task, on the other hand if there are no idle resources then
 *  it creates a deadhead task. If still there are no idle resources available at the locations within the reachable 
 *  range of the loads location, the load will remain in the queue and the next load will be processed. If all the 
 *  resources are available then they are allocated to the load and a transport task is created notifying the driver
 *  to perform the task. The DispatcherFTL is notified upon completion of the task which gives it an opportunity to 
 *  reallocate the resources in the network in order to process other tasks pending in the network.  
 *  
 *  @author rossetti
 */

public class DispatcherFTL extends AbstractDispatcherFTL {

	/** Queue for holding the loads waiting to be dispatched by location
	 */
	protected Map<TransportLocationIfc, Queue> myDispatchQByLocation; 

	//TODO need to add methods for easily enqueuing and dequeing from these queues
	/** Queue for holding the bobtail loads leaving by location
	 */
	protected Map<TransportLocationIfc, Queue> myBobTailPickupQByLocation; 

	/** Queue for holding the deadhead loads leaving by location
	 */
	protected Map<TransportLocationIfc, Queue> myDeadHeadPickupQByLocation; 

	/** Queue for holding the transport loads leaving by location
	 */
	protected Map<TransportLocationIfc, Queue> myTransportPickupQByLocation; 

	/** Queue for holding the transport loads inbound by location
	 */
	protected Map<TransportLocationIfc, Queue> myIncomingLoadsByLocation; 

	/** Holds loads to be dispatched
	 */
	protected List<DispatchTask> myDispatchQ;
	
	protected ResponseVariable loadSystemTime;
	
	protected AggregateTimeWeightedVariable myTotalNumInDispatchPickUpQObserver;
	
	protected AggregateTimeWeightedVariable myTotalNumInBobTailPickUpQObserver;
	
	protected AggregateTimeWeightedVariable myTotalNumInDeadHeadPickUpQObserver;
	
	protected AggregateTimeWeightedVariable myTotalNumInIncomingLoadsQObserver;
	
	protected AggregateTimeWeightedVariable myTotalNumInTransportPickUpQObserver;
	
	protected ResponseVariableAverageObserver myAvgTimeInDispatchPickUpQObserver;
	
	protected ResponseVariableAverageObserver myAvgTimeInBobTailPickUpQObserver;
	
	protected ResponseVariableAverageObserver myAvgTimeInDeadHeadPickUpQObserver;
	
	protected ResponseVariableAverageObserver myAvgTimeInTransportPickUpQObserver;
	
	protected ResponseVariableAverageObserver myAvgTimeInIncomingLoadsQObserver;
	
	protected Statistic myDeadHeadMiles;
	
	protected Statistic myBobTailMiles;
	
	protected Statistic myTransportMiles;
	
	protected Statistic myNumTasks;
		
	/**
	 * @param parent
	 * @param network
	 * @param distance
	 */
	public DispatcherFTL(ModelElement parent, TransportationNetwork network, double distance) {
		this(parent, network, distance, null);
	}
	
	/**
	 * @param parent
	 * @param network
	 * @param distance
	 * @param name
	 */
	 public DispatcherFTL(ModelElement parent, TransportationNetwork network, double distance, String name) {
		super(parent, network, distance, name);
		
		myDispatchQ = new LinkedList<DispatchTask>();
		
		myDispatchQByLocation = new HashMap<TransportLocationIfc, Queue>();
		
		myBobTailPickupQByLocation = new HashMap<TransportLocationIfc, Queue>();
		
		myDeadHeadPickupQByLocation = new HashMap<TransportLocationIfc, Queue>();
		
		myTransportPickupQByLocation = new HashMap<TransportLocationIfc, Queue>();
		
		myIncomingLoadsByLocation = new HashMap<TransportLocationIfc, Queue>();
		
		loadSystemTime = new ResponseVariable(this,"Average Load System Time");
		
		myNumTasks = new Statistic("Total Number of Loads Processed");
		
		myDeadHeadMiles =new Statistic ("DeadHead Miles Statistic");
		
		myBobTailMiles = new Statistic("BobTail Miles Statistic");
		
		myTransportMiles = new Statistic ("Transport Miles Statistic");
		
		myTotalNumInDispatchPickUpQObserver = new AggregateTimeWeightedVariable(this, getName()+" Total Num in DispatchPickUpQ"); 
		
		myTotalNumInTransportPickUpQObserver = new AggregateTimeWeightedVariable(this, getName()+" Total Num in TransportPickUpQ");
		
		myTotalNumInBobTailPickUpQObserver = new AggregateTimeWeightedVariable(this, getName()+" Total Num in BobTailPickUpQ");
		
		myTotalNumInDeadHeadPickUpQObserver = new AggregateTimeWeightedVariable(this, getName()+" Total Num in DeadHeadPickUpQ");
		
		myTotalNumInIncomingLoadsQObserver = new AggregateTimeWeightedVariable(this, getName()+" Total Num in IncomingLoadsQ");
		
		myAvgTimeInDispatchPickUpQObserver = new ResponseVariableAverageObserver(this, getName()+" Average Time in DispatchPickUpQ");
		
		myAvgTimeInBobTailPickUpQObserver = new ResponseVariableAverageObserver(this, getName()+" Average Time in BobTailPickUpQ");
		
		myAvgTimeInDeadHeadPickUpQObserver = new ResponseVariableAverageObserver(this, getName()+" Average Time in DeadHeadPickUpQ");
		
		myAvgTimeInTransportPickUpQObserver = new ResponseVariableAverageObserver(this, getName()+" Average Time in TransportPickUpQ");
		
		myAvgTimeInIncomingLoadsQObserver = new ResponseVariableAverageObserver(this, getName()+" Average Time in IncomingLoadsQ");
		
		Collection<TransportLocationIfc> locations = network.getOrigins(); 
	
		for(TransportLocationIfc location: locations){
			
			Queue queue = new Queue(this, location.getId() + " " + location.getName() + " DispatchPickUpQ " + network.getIndexOf(location));
			myDispatchQByLocation.put(location, queue);
			
			queue.subscribe(myAvgTimeInDispatchPickUpQObserver);
			queue.subscribe(myTotalNumInDispatchPickUpQObserver);
									
			queue = new Queue(this, location.getId() + " " + location.getName() + " BobTailPickupQ " + network.getIndexOf(location));
			myBobTailPickupQByLocation.put(location, queue);
			
			queue.subscribe(myAvgTimeInBobTailPickUpQObserver);
			queue.subscribe(myTotalNumInBobTailPickUpQObserver);

			queue = new Queue(this, location.getId() + " " + location.getName() + " DeadHeadPickupQ " + network.getIndexOf(location));
			myDeadHeadPickupQByLocation.put(location, queue);
						
			queue.subscribe(myAvgTimeInDeadHeadPickUpQObserver);
			queue.subscribe(myTotalNumInDeadHeadPickUpQObserver);
			
			queue = new Queue(this, location.getId() + " " + location.getName() + " TransportPickupQ " + network.getIndexOf(location));
			myTransportPickupQByLocation.put(location, queue);
						
			queue.subscribe(myAvgTimeInTransportPickUpQObserver);
			queue.subscribe(myTotalNumInTransportPickUpQObserver);
						
		}		
		
		Collection<TransportLocationIfc> destinations = network.getDestinations();
		
		for(TransportLocationIfc destination: destinations){
		
			Queue queue = new Queue(this, destination.getId() + " " + destination.getName() + " IncomingLoadQ " + network.getIndexOf(destination));
			
			myIncomingLoadsByLocation.put(destination, queue);
			
			queue.subscribe(myAvgTimeInIncomingLoadsQObserver);
			queue.subscribe(myTotalNumInIncomingLoadsQObserver);
									
		}
		
	}

	//  ===========================================
	//  PUBLIC METHODS
	//  ===========================================

	/** Returns true if there is an inbound transport load coming to the location
	 * 
	 * @param location
	 * @return
	 */
	public boolean isLoadInbound(TransportLocationIfc location){
		Queue queue = myIncomingLoadsByLocation.get(location);
		if (queue == null)
			return(false);
		else {
			return(!queue.isEmpty());
		}
	}
	
	/** Searches all the dispatch queues by location to find the location of the queue
	 *  that has the largest number of waiting tasks. The search is performed
	 *  in order of locations and ties with the max are ignored (the first max 
	 *  found is the location returned). If you need all tied locations, use 
	 *  findListOfLocationsWithLargestDispatchQ()
	 * 
	 * @return
	 */
	public TransportLocationIfc findLocationOfLargestDispatchQ(){
		TransportLocationIfc locmaxQ = null;
		int max = -1;
		for (TransportLocationIfc location: myDispatchQByLocation.keySet()){
			 Queue q = myDispatchQByLocation.get(location);
			 if (q.size() > max){
			 	max = q.size();
			 	locmaxQ = location;
			 }
		}
		return(locmaxQ);
	}
	
	/** Searches all the dispatch queues by location to find the locations of the queues
	 *  that have the largest number of waiting tasks. The search is performed
	 *  in order of locations.
	 * 
	 * @return The list of queues with the maximum number of tasks
	 */
	public List<TransportLocationIfc> findListOfLocationsWithLargestDispatchQ(){
		List<TransportLocationIfc> list = new LinkedList<TransportLocationIfc>();
		int max = -1;
		for (TransportLocationIfc location: myDispatchQByLocation.keySet()){
			 Queue q = myDispatchQByLocation.get(location);
			 if (q.size() > max){
			 	// a new max was found, clear the list
			 	list.clear();
			 	// add the new location
			 	list.add(location);
			 	// remember the new maximum
			 	max = q.size();
			 } else {
			 	if (q.size() == max){ // ties the max
			 		list.add(location);
			 	}
			 }
		}
		return(list);
	}

	public TransportLocationIfc findLocationNeedingBobTail(TransportLocationIfc location){
		double d = this.getMaxDeadHeadBobTailDistance(location);
		return(findLocationNeedingBobTail(location, d));
	}

	public TransportLocationIfc findLocationNeedingBobTail(TransportLocationIfc location, double distance){

		Collection<LaneIfc> lanes = getNetwork().getOutGoingLanes(location);
		
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc dest = lane.getDestination();
				if (isBobTailNeededAt(dest))
					return dest;				
			}
		}
		return(null);
	}

	/** Returns the location needing deadhead 
	 *  generally called during reallocation of resources
	 * 
	 * @param location
	 * @return
	 */
	public TransportLocationIfc findLocationNeedingDeadHead(TransportLocationIfc location){
		double d = this.getMaxDeadHeadBobTailDistance(location);
		return(findLocationNeedingDeadHead(location, d));
	}

	/** Returns the location needing deadhead 
	 *  Uses outGoingLanes of the location supplied to check for a destination
	 *  requiring deadhead
	 * 
	 * @param location
	 * @param distance
	 * @return
	 */
	public TransportLocationIfc findLocationNeedingDeadHead(TransportLocationIfc location, double distance){

		Collection<LaneIfc> lanes = getNetwork().getOutGoingLanes(location);
		for(LaneIfc lane: lanes){
			if (lane.getDistance() <= distance){
				TransportLocationIfc dest = lane.getDestination();
				if (isDeadHeadNeededAt(dest))
					return dest;				
			}
		}
		return(null);
	}
	
	/** checks if bobtail is needed at any of the destinations 
	 *  on the outgoing lanes of the supplied location
	 *  Checks in the following order:
	 *  1) If a load/task is waiting at any destination
	 *  2) The load/task found is not inbound
	 *  3) Idle trailer is available at that destination
	 * @param location
	 * @return
	 */
	public boolean isBobTailNeededAt(TransportLocationIfc location){
		boolean needed = false;
		
		if (isLoadWaitingAt(location)){ // load is waiting
			if (!isLoadInbound(location)){ // no transport inbound
				if (trailerIsIdleAt(location)) // trailer is idle
					needed = true;
			}
		}
		return(needed);
	}
	
	/** checks if deadhead is needed at any of the destinations 
	 *  on the outgoing lanes of the supplied location
	 *  Checks in the following order:
	 *  1) If a load/task is waiting at any destination
	 *  2) The load/task found is not inbound
	 * @param location
	 * @return
	 */
	public boolean isDeadHeadNeededAt(TransportLocationIfc location){
		boolean needed = false;
		
		if (isLoadWaitingAt(location)){ // load is waiting
			if (!isLoadInbound(location)){ // no transport inbound
				needed = true;
			}
		}
		return(needed);
	}
	
	/** checks if a load requires to be processes at a location
	 *  by checking its dispatch queue
	 *  returns null if the queue is empty
	 * @param location
	 * @return
	 */
	public boolean isLoadWaitingAt(TransportLocationIfc location){
		Queue queue = myDispatchQByLocation.get(location);
		if (queue == null)
			return(false);
		else {
			return(!queue.isEmpty());
		}
	}
	
	/**
	 * 
	 * @param location
	 * @return number of tasks waiting in the dispatch queue of a location
	 */
	public int getNumberWaitingAt(TransportLocationIfc location){
		Queue queue = myDispatchQByLocation.get(location);
		if (queue == null)
			return(0);
		else {
			return(queue.size());
		}
	}
	
	//  ===========================================
	//  PROTECTED METHODS
	//  ===========================================

	protected void initialize(){
		super.initialize();
		//System.out.println("DISPATCHER FTL INITIALIZE METHOD");
		myDispatchQ.clear();	
	}
	
	/* (non-Javadoc)
	 * @see transportlayer.AbstractDispatcher#receive(transportlayer.Load)
	 */
	protected void receive(Load load){
		
	//	System.out.println("t> "+getTime()+" Load "+load.getId()+" sent by the Network is received by "+getName());
		
		// create a task for the load
		DispatchTask task = new DispatchTask(this, load);
		// place the task in the task queues
		// to ensure collection of waiting time
		task.setOrigin(load.getOrigin());
		task.setDestination(load.getDestination());
	//	System.out.println("t> "+getTime()+" Dispatch Task "+task.getId()+" is created for Load "+load.getId());
		
		enqueue(task);
		// allow allocation mechanism to work 
		allocateResourcesAfterIncomingLoad(task);
	}
	
	/**
	 *  The task supplied is enqueud in the overall dispatch list of the network
	 *  and the dispatch queue at its origin
	 * 
	 * @param task
	 */
	protected void enqueue(DispatchTask task){
		
		//System.out.println("t> "+getTime()+" Task "+task.getId()+" is enqueued in the Overall DispatchQ");
		// add the task to overall queue
		myDispatchQ.add(task);
		// add the task by location to the queue at the load's origin
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myDispatchQByLocation.get(location);
		queue.enqueue(task);	
		//System.out.println("t> "+getTime()+" Task "+task.getId()+" is enqueued in the DispatchQ of Location "+task.getLoadOrigin().getName());
	}
	
	/**
	 * The task supplied is removed from the overall dispatch list of the network
	 *  and the dispatch queue at its origin
	 * 
	 * @param task
	 */
	protected void dequeue(DispatchTask task){
		
	//	System.out.println("t> "+getTime()+" Task "+task.getId()+" is dequeued from the Overall DispatchQ");
		// remove it from overall queue
		myDispatchQ.remove(task);
		// remove it by location
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myDispatchQByLocation.get(location);
	//	System.out.println("t> "+getTime()+" Task "+task.getId()+" is dequeued from the DispatchQ of Location "+task.getLoadOrigin().getName());
		queue.remove(task);
		
	}
			
	/** All the resources are avilable to execute the transport task
	 *  The task is enqueued in the transport pick up queue of its origin
	 *  The resources are allocated to the task
	 *  The task is asked to send message for its execution
	 * @param task
	 */
	protected void dispatchTransport(DispatchTask task){
		
		// make the task a transport type
		task.setTypeToTransport();
		//TODO need to place the task in the transport pick up queue
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myTransportPickupQByLocation.get(location);
	//	System.out.println("t> "+getTime()+" Transport task "+task.getId()+" is enqueued in the Transport PickUpQ of Location "+task.getLoadOrigin().getName());
		queue.enqueue(task);
		// the dispatch load has identified the required resources
		// tell the task to allocate its resources
	//	System.out.println(" "+task.getDriver().myLastEventTime+" for Task "+task.getId());
		
		allocateDriver(task);
		allocateTruck(task);
		allocateTrailer(task);
		
		// tell the task to execute
		task.waitingForTransport();
	}
	
	/** This method needs to be called when the driver completes
	 *  the loading of the task
	 *  The ransport task is removed from transport queue and enqueud 
	 *  in the incoming load queue at its destination
	 * 
	 * @param task
	 */
	protected void loadingCompleted(DispatchTask task){
		if(task==null)
			throw new IllegalArgumentException("The task was null!");
		//TODO The driver has completed loading, 
		// the task needs to be removed from the queue that it is waiting in
		// and placed in the incoming load queue for its destination
		// get the queue that the task is currently waiting in
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myTransportPickupQByLocation.get(location);
		// now remove it
		queue.remove(task); 
		// get the load's destination
		location = task.getLoadDestination();
		// get the incoming load queue for the destination
		queue = myIncomingLoadsByLocation.get(location);
		// put the task in the incoming load queue
		queue.enqueue(task);
	}
			
	/** This allocation method assumes that if all of the resources are available
	 *  for the incoming task, then it should be immediately dispatched. If
	 *  all the resources for the load are not immediately available the load task
	 *  waits in the dispatch queues.
	 *  
	 *  This behavior may be overridden.  For example, the dispatcher may want to
	 *  wait for some condition before dispatching the task even if resources
	 *  are immediately available
	 * 
	 * @param loadTask
	 */
	protected void allocateResourcesAfterIncomingLoad(DispatchTask incomingTask) {
		
	//	System.out.println("t> "+getTime()+" Check if Incoming Dispatch Task "+incomingTask.getId()+" can be allocated resources at Origin: "+incomingTask.getOrigin().getName());
		// This allocation mechanism assumes that if the resources
		// are available for the load at its location, then it
		// should be immediately dispatched.
				
		// check if the incoming load can be dispatched
	if (transportAvailable(incomingTask)){
		// remove the task and dispatch it
			dequeue(incomingTask);
			// all resources have been set, so dispatch transport task
			dispatchTransport(incomingTask);
		}
		 else {
			TransportLocationIfc origin = incomingTask.getLoadOrigin();
		//	System.out.println("t> "+getTime()+" Transport is not available for Incoming Dispatch Task "+incomingTask.getId()+" at Origin: "+incomingTask.getOrigin().getName());
		
			if(trailerIsIdleAt(origin)){
				if (!isLoadInbound(origin))// only ask for bobtail if no inbound loads
				//	System.out.println("t> "+getTime()+" Incoming Task "+incomingTask.getId()+" needs BOBTAIL at "+origin.getName());
					 bobTailNeededAfterIncomingLoad(incomingTask);
			}else {
				if (!isLoadInbound(origin)) // only ask for deadhead if no inbound loads
				//	System.out.println("t> "+getTime()+" Incoming Task "+incomingTask.getId()+" needs DEADHEAD at "+origin.getName());
					deadHeadNeededAfterIncomingLoad(incomingTask);
			}
		}
	}
		
	/** If not all resources are available for the incoming task,
	 *  then if a trailer is idle at the location, the dispatcher
	 *  might want to consider having a bob tail for the incoming
	 *  task or some other task at the task's location.  This method
	 *  can be used to initiate a bob tail
	 * 
	 * @param incomingTask
	 */
	protected void bobTailNeededAfterIncomingLoad(DispatchTask incomingTask){
		// try to identify a nearby location that can serve as a bobtail
		// if found, remove from dispatch queue, place in bobtail pickup queue
		// dispatch a bobtail to load task
		
	//	System.out.println("Did I come here bobTailNeededAfterIncomingLoad");
		
		TransportLocationIfc origin = incomingTask.getLoadOrigin();
		
	//	System.out.println("Origin of the Incoming Task: "+origin.getName());
	
		double allowedDistance = getMaxDeadHeadBobTailDistance(origin);
		
		TransportLocationIfc bobTailLocation = getFirstLocationWithIdleDriversAndTrucks(origin, allowedDistance);
	
	//	TransportLocationIfc bobTailLocation = getLocationWithIdleDriverandTruckWithMinDeliveryTime(origin,incomingTask);
		
		if (bobTailLocation != null){
			System.out.println("t> "+getTime()+" Incoming Task "+incomingTask.getId()+" gets BOBTAIL from "+bobTailLocation.getName());
			// found a location, make the bobtail dispatch
			dispatchBobTail(incomingTask, bobTailLocation);
		}
		else
			System.out.println("Incoming Task "+incomingTask.getId()+" cannot get BOBTAIL");
	}
	
	/** If not all resources are available for the incoming task,
	 *  the dispatcher might want to consider having a deadhead for the incoming
	 *  task or some other task at the task's location.  This method
	 *  can be used to initiate a deadhead in this situation
	 * 
	 * @param incomingTask
	 */
	protected void deadHeadNeededAfterIncomingLoad(DispatchTask incomingTask){
		
		// try to identify a nearby location that can serve as a deadhead
		// if found, remove from dispatch queue, place in deadhead pickup queue
		// dispatch a deadhead to load task
		TransportLocationIfc origin = incomingTask.getLoadOrigin();
		
		double allowedDistance = getMaxDeadHeadBobTailDistance(origin);
		TransportLocationIfc deadHeadLocation = getFirstLocationWithIdleResources(origin, allowedDistance);
	//	TransportLocationIfc deadHeadLocation = getLocationWithIdleResourcesWithMinDeliveryTime(origin,incomingTask);
		if (deadHeadLocation != null){
			System.out.println("t> "+getTime()+" Incoming Task "+incomingTask.getId()+" gets DEADHEAD from "+deadHeadLocation.getName());
			// found a location, make the deadhead dispatch
			dispatchDeadHead(incomingTask, deadHeadLocation);
		}
		else
			System.out.println("Incoming Task "+incomingTask.getId()+" cannot get DEADHEAD");
	}
		
	/** Selects the next DispatchLoad to dispatch. The default behavior
	 *  of this method is to search the dispatch queue (without regard to location) 
	 *  in order of arrival and to return the first load task that can be dispatched.  If no
	 *  load task is found, this method returns null.  This method does not
	 *  remove the item to be dispatched
	 * 
	 *  Assumptions:
	 *  1) A DispatchLoad can be dispatched, if the required resources can be found
	 *     at its current location
	 *  2) If no DispatchLoad meets this requirement then this method
	 *     should return null
	 * 
	 * @return
	 */
	protected DispatchTask selectNextTask(){
		
		DispatchTask loadTask = null;
		if (myLoadSelectionRule != null)
			loadTask = myLoadSelectionRule.selectNextTransportTask(this, Collections.unmodifiableList(myDispatchQ));
		else{
			// find the first DispatchLoad that can be dispatched
			for(DispatchTask task: myDispatchQ){
				if (transportAvailable(task)){
					loadTask = task;
					break;
				}
			}
		}
		return(loadTask);
	}

	/** Selects the next task at the given location, returns null if none
	 *  Does not remove the task from the queue.
	 *  
	 * @param location
	 * @return
	 */
	protected DispatchTask selectNextTask(TransportLocationIfc location){
		Queue queue = myDispatchQByLocation.get(location);
		if (queue == null)
			return(null);
		else {
			return((DispatchTask)queue.peekNext());
		}
	}
	
	/** Processes all waiting tasks and if available resources can be found
	 *  at the task's origin then the task is dispatched.
	 */
	protected void dispatchLoads(){
		
		boolean done = false;
		while((!myDispatchQ.isEmpty()) && (!done)){
			DispatchTask loadTask = selectNextTask();
			if (loadTask == null)
				done = true;
			else {
				dequeue(loadTask);
				// all resources have been set, so dispatch transport task				
				dispatchTransport(loadTask);
			}
		}
	}
	
	/** Called when the task has completed.  This method delegates the
	 *  work to transportCompleted(), deadHeadToLoadCompleted(), bobTailToLoadCompleted()
	 *  deadHeadNoLoadCompleted(), bobTailNoLoadCompleted() depending on the type of task.
	 *  
	 * @param task
	 */
	protected void dispatchCompleted(DispatchTask task) {
		
		if (task.isTransport())
			transportCompleted(task);
		else if (task.isDeadHeadToLoad())
			deadHeadToLoadCompleted(task);
		else if (task.isBobTailToLoad())
			bobTailToLoadCompleted(task);
		else if (task.isBobTailNoLoad())
			bobTailNoLoadCompleted(task);
		else if (task.isDeadHeadNoLoad())
			deadHeadNoLoadCompleted(task);
			
	}

	/**
	 * @param task
	 */
	protected void bobTailToLoadCompleted(DispatchTask task) {
		
	//	System.out.println("t>"+getTime()+"Driver "+task.getDriver().getId()+" completed BobTail to Load Task "+task.getId()+" at "+task.getLoadOrigin().getName());
		// task already has driver, truck, trailer
		// when the bob tail was dispatched it had to have seized the trailer
		// need to proceed to load's destination
		task.setTypeToTransport(); // this makes it a transport
	    myBobTailMiles.getAverage();
		// get the queue that the task is currently waiting in
		TransportLocationIfc location = task.getLoadOrigin();
		Queue queue = myBobTailPickupQByLocation.get(location);
		// now remove it
		queue.remove(task); 
	//	System.out.println("t>"+getTime()+" Task "+task.getId()+" is removed from the BobTail PickUp Queue at Location "+location.getName());
		//place the task in the transport pick up queue
		queue = myTransportPickupQByLocation.get(location);
		queue.enqueue(task);
	//	System.out.println("t>"+getTime()+" Task "+task.getId()+" is enqueued in Transport PickUp Queue at Location "+location.getName());
		
		task.waitingForTransport();
		// it is important that loading complete be called by the driver*/
	}
	
	/**
	 * @param task
	 */
	protected void deadHeadToLoadCompleted(DispatchTask task) {
			
		//System.out.println("t>"+getTime()+"Driver "+task.getDriver().getId()+" completed DeadHead to Load Task "+task.getId()+" at "+task.getLoadOrigin().getName());
		// task already has driver, truck, and trailer
		// need to proceed to load's destination
	//	System.out.println("Dead head to load completed by driver "+task.getDriver().getId());
		task.setTypeToTransport(); // this makes it a transport
		
	//	myDeadHeadMiles.getAverage();
		// get the queue that the task is currently waiting in
		TransportLocationIfc location = task.getLoadOrigin();
		
		Queue queue = myDeadHeadPickupQByLocation.get(location);
		// now remove it
		queue.remove(task); 
	//	System.out.println("t>"+getTime()+" Task "+task.getId()+" is removed from the DeadHead PickUp Queue at Location "+location.getName());
		//place the task in the transport pick up queue
		queue = myTransportPickupQByLocation.get(location);
		queue.enqueue(task);
	//	System.out.println("t>"+getTime()+" Task "+task.getId()+" is enqueued in the Transport PickUp Queue at Location "+location.getName());
		
		// indicate that the task is waiting for transport
		task.waitingForTransport();		
		// it is important that loading complete be called by the driver
	}

	/** A transport task is completed by the driver
	 *  The dispatcher will ask the task to release all the resources 
	 *  Dispatcher will try to reallocated resources to the pending tasks in teh network
	 * @param task
	 */
	protected void transportCompleted(DispatchTask task) {
		
		if(task==null)
			throw new IllegalArgumentException("Task was null!");
		if(task.getQueue()==null)
			throw new IllegalArgumentException("Task Q was null!");
		// the task has arrived at it destination
		// it should no longer be in the incoming load queue for the destination
		Queue q = task.getQueue(); // should be the incoming load queue
	//	System.out.println("t>"+getTime()+" Task "+task.getId()+" is removed from the Incoming Load Queue");
		q.remove(task);
		
		double tnow=getTime();
		loadSystemTime.setValue(tnow-(task.getCreateTime()));
		myNumTasks.collect(1);
		Driver d=task.getDriver();
		
		releaseTruck(task);
		releaseTrailer(task);
		releaseDriver(task);
				
		if(d.isUnloadingFlag()){
			d.setUnloadingFlag(false);
			d.scheduleRestAfterUnloading();
		}
		
		TransportLocationIfc location = task.getLoadDestination();
		
		d.setCurrentTask(null);
	//	System.out.println("t>"+getTime()+"  "+getName()+" is REALLOCATING RESOURCES at "+location.getName());
    //    System.out.println("");
        allocateResourcesAfterTransportCompletedAt(location);
       }
	
	/**
	 * @param task
	 */
	protected void deadHeadNoLoadCompleted(DispatchTask task) {
	//	System.out.println("t>"+getTime()+" Driver "+task.getDriver().getId()+"Completed DeadHead no Load Task "+task.getId()+" at "+task.getLoadDestination().getName());
		// TODO Auto-generated method stub
		
	}

	/**
	 * @param task
	 */
	protected void bobTailNoLoadCompleted(DispatchTask task) {
		// TODO Auto-generated method stub
				
	}

	/** This allocation method is invoked after a successful transport. The
	 *  default behavior is to check if there is a load waiting at the
	 *  dropoff location.  If so, the next load at the dropoff location,
	 *  is selected for dispatching. If there are no loads waiting at this location,
	 *  then there is an opportunity for deadheading.
	 *  
	 *  This behavior may be overridden.  For example, the dispatcher may want to
	 *  wait for some condition before dispatching the task even if resources
	 *  are immediately available

	 * @param location
	 */
	protected void allocateResourcesAfterTransportCompletedAt(TransportLocationIfc location) {
	   
	    // get the queue at the location
		Queue queue = myDispatchQByLocation.get(location);
   //     System.out.println("t>"+getTime()+" Check if there is a Next Task at "+location.getName());
		// select next task at the drop off location
		DispatchTask nextTask = (DispatchTask)queue.peekNext();

		// if there is a task, then dispatch it
		if (nextTask != null){			
		//	System.out.println("There is Next Task "+nextTask.getId()+" available at "+location.getName());
			// check if the task can be dispatched
			if (transportAvailable(nextTask)){
				// remove the task and dispatch it
				dequeue(nextTask);
				// all resources have been set, so dispatch transport task
				dispatchTransport(nextTask);
			}
		} else{			   
				resourcesAvailableAt(location);
		}
	}
	
	/**
	 *  Returns task needing bobtail by supplying the location
	 *  from where the idle resources can be assigned for that bobtail
	 *  task 
	 * @param location
	 * @return
	 */
	protected DispatchTask findTaskNeedingBobTail(TransportLocationIfc location){
		TransportLocationIfc needsTransport = findLocationNeedingBobTail(location);
		DispatchTask task = null;
		if (needsTransport != null){
			task = selectNextTask(needsTransport);
		}
		return(task);
	}

	/**
	 * Returns task needing deadhead by supplying the location
	 * from where the idle resources can be assigned for that deadhead
	 *  task 
	 * @param location
	 * @return
	 */
	protected DispatchTask findTaskNeedingDeadHead(TransportLocationIfc location){
		TransportLocationIfc needsTransport = findLocationNeedingDeadHead(location);
		DispatchTask task = null;
		if (needsTransport != null){
			task = selectNextTask(needsTransport);
		}
		return(task);
	}
	
	/** A driver, truck, and trailer were just released at the given
	 *  location and there were no loads available for pick up.  The
	 *  dispatcher might want to allocate the resources to waiting loads
	 *  at other locations
	 * 
	 * @param location
	 */
	protected void resourcesAvailableAtOld(TransportLocationIfc location){
		
	//	System.out.println("t>"+getTime()+" Resources are Available at "+location.getName());
	//	System.out.println("Checking if there are any BobTail Task Available around "+location.getName());
		// check for needed bobtails
		DispatchTask task = findTaskNeedingBobTail(location);
		
		if (task != null){
			// dispatch bobtail task
		//	System.out.println("t>"+getTime()+" BobTail Task "+task.getId()+" is Available at "+location.getName());
			dispatchBobTail(task, location);
		} else{
		//	System.out.println("There is no BobTail Task Available around "+location.getName());
		//	System.out.println("Checking if there are any DeadHead Task Available around "+location.getName());
			task = findTaskNeedingDeadHead(location);
			if (task != null){
			//	System.out.println("t>"+getTime()+" DeadHead Task "+task.getId()+" is Available at "+location.getName());
				// dispatch dead head
				dispatchDeadHead(task, location);
			}
		//	System.out.println("There is no DeadHead Task Available around "+location.getName());
		//	System.out.println(" ");
		}
	}
	
	/** A driver, truck, and trailer were just released at the given
	 *  location and there were no loads available for pick up.  The
	 *  dispatcher might want to allocate the resources to waiting loads
	 *  at other locations
	 * 
	 * @param location
	 */
	protected void resourcesAvailableAt(TransportLocationIfc location){
		
		// if here, then a driver or truck and trailer has been found
	//	System.out.println("Checking if there are any BobTail Task Available around "+location.getName());
		// check for needed bobtails
		DispatchTask task = findTaskNeedingBobTail(location);
		
		if (task != null){
		//	System.out.println("t>"+getTime()+" BobTail Task "+task.getId()+" is Available at "+location.getName());
		
			if (driverIsIdleAt(location) && truckIsIdleAt(location))
				dispatchBobTail(task, location);
			}
			else{
		//	System.out.println("There is no BobTail Task Available around "+location.getName());
		//	System.out.println("Checking if there are any DeadHead Task Available around "+location.getName());
			task = findTaskNeedingDeadHead(location);
		
			if (task != null){
				System.out.println("t>"+getTime()+" DeadHead Task "+task.getId()+" is Available at "+location.getName());
				if (driverIsIdleAt(location) && truckIsIdleAt(location) && trailerIsIdleAt(location)){	
					dispatchDeadHead(task, location);
				}
			}
			else
				System.out.println("There is no DeadHead Task Available around "+location.getName());
		}
	}
	
	/** Changes the task to a bobtail and initiates a pick up from the supplied location
	 * 
	 * @param task
	 * @param fromLoc The location that the bobtail is to come from
	 */
	protected void dispatchBobTail(DispatchTask task, TransportLocationIfc fromLoc){
		
		TransportLocationIfc origin = task.getLoadOrigin();
	/*	dequeue(task);
		
		Queue queue = myBobTailPickupQByLocation.get(origin);
		queue.enqueue(task);
		System.out.println("t>"+getTime()+" Task "+task.getId()+" enqueud in BobTail PickUpQ at Location "+origin.getName());
	*/	
		task.setOrigin(fromLoc);
		task.setTypeToBobTailToLoad();
	
		task.setDriver(getIdleDriver(fromLoc,task));
		task.setTruck(getIdleTruck(fromLoc));
		task.setTrailer(getIdleTrailer(origin));
		
		task.setOrigin(fromLoc);
		task.setDestination(task.getLoadOrigin());
		
		task.setTypeToBobTailToLoad();
		dequeue(task);
		
		Queue queue = myBobTailPickupQByLocation.get(origin);
		queue.enqueue(task);
	//	System.out.println("t>"+getTime()+" Task "+task.getId()+" enqueud in BobTail PickUpQ at Location "+origin.getName());
		
		allocateDriver(task);
		allocateTruck(task);
		allocateTrailer(task,origin);
		
		task.waitingForTransport();	
	}
	
	/** Initiates a deadhead for the task from the supplied location
	 * 
	 * @param task
	 * @param fromLoc The location that the deadhead is to come from
	 */
	protected void dispatchDeadHead(DispatchTask task, TransportLocationIfc fromLoc){
		
		TransportLocationIfc origin = task.getLoadOrigin();
		
		task.setOrigin(fromLoc);
		task.setTypeToDeadHeadToLoad();
		task.setDriver(getIdleDriver(fromLoc,task));
		task.setTruck(getIdleTruck(fromLoc));
		task.setTrailer(getIdleTrailer(fromLoc));
		
		task.setOrigin(fromLoc);
		task.setDestination(task.getLoadOrigin());
		task.setTypeToDeadHeadToLoad();
		
		dequeue(task);
		Queue queue = myDeadHeadPickupQByLocation.get(origin);
	//	System.out.println("t>"+getTime()+" Task "+task.getId()+" enqueud in DeadHead PickUpQ at Location "+origin.getName());
		queue.enqueue(task);
		
		allocateDriver(task);
		allocateTruck(task);
		allocateTrailer(task);
		
		task.waitingForTransport();
	}
	
	//TODO not implemeted anywhere
	/**
	 * sets delivery time on the load
	 * @param t
	 * @param d
	 * @return
	 */
	protected double setDeliveryTime(DispatchTask t, Driver d){
		return 0.0;
	}

	/**
	 * This method is redundant. Dr. Rossetti I realized much later that the way i am counting 
	 *  number of loads processed could be wrong. Please can you check it in this method //TODO
	 * 
	 */
	protected void afterReplication(){
		
		System.out.println("DeadHead Miles: "+myDeadHeadMiles.getAverage());
		System.out.println("BobTail Miles: "+myBobTailMiles.getAverage());
		System.out.println("Transport Miles: "+myTransportMiles.getAverage());
		System.out.println("Total Number of Loads Processed: "+myNumTasks.getCount());
		
		myDispatchQ.clear();
		for(TransportLocationIfc location:myDispatchQByLocation.keySet()){
			Queue q = myDispatchQByLocation.get(location);
			q.clear();
		}
		
		for(TransportLocationIfc location:myDeadHeadPickupQByLocation.keySet()){
			Queue q = myDeadHeadPickupQByLocation.get(location);
			q.clear();
		}
		
		for(TransportLocationIfc location:myBobTailPickupQByLocation.keySet()){
			Queue q = myBobTailPickupQByLocation.get(location);
			q.clear();
		}
		
		for(TransportLocationIfc location:myTransportPickupQByLocation.keySet()){
			Queue q = myTransportPickupQByLocation.get(location);
			q.clear();
		}
		
		for(TransportLocationIfc location:myIncomingLoadsByLocation.keySet()){
			Queue q = myIncomingLoadsByLocation.get(location);
			q.clear();
		}
	}

}
