package jslx.transportlayerFinal;

import jsl.modeling.ActionListenerIfc;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.spatial.CoordinateIfc;
import jsl.modeling.elements.spatial.SpatialResource;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.distributions.DistributionFunctionIfc;
import jsl.utilities.random.distributions.DistributionIfc;

/**
 * Driver extends SpatialResource.Driver is the resource which is responsible for completing the task assigned to it by the dispatcher. 
 * It is responsible for transporting load from its origin to its destination in case of the transport task or just
 * completing transport in case of the bobtail and the deadhead task. The Driver is modeled based on discrete event scheduling.
 * The event scheduling is in accordance with the Department of Transportation regulations (2006).
 * We have assumed here that the if the driver does not have sufficient hours to complete loading, then driver will immediately take 
 * rest and then load. Secondly, if the driver does not have sufficient time left for unloading, then the driver will still perform
 * unloading and take its due rest immediately after unloading.
 * 
 * @author snangia
 *
 */
public class Driver extends SpatialResource {
	
	/** Flag to indicate if loading was delayed due to rest 
	 *  True if the loading is still due
	 */
	protected boolean myLoadingFlag;
	
	/** Flag to indicate if unloading was scheduled before due rest
	 *  True if the driver has to rest immediately after unloading
	 */
	protected boolean myUnloadingFlag;
	
	/** Flag to indicate if transport has to be completed after rest
	 *  True if the driver has to transport after rest
	 */
	protected boolean myIncompleteTransportFlag;
	
	/** True if driver has to take 10 hour rest while performing task (before the task gets completed)
	 */
	protected boolean myTaskAttachedWith10HourRestFlag;
	
	/** True if driver has to take 34 hour rest while performing task (before the task gets completed)
	 */
	protected boolean myTaskAttachedWith34HourRestFlag;
	
	/**
	 *  True if no task is assigned to this driver since it was
	 *  added to the network or if it hasnt completed its first
	 *  14 hour on duty cycle since it was added to the network
	 */
	protected boolean myNoTaskFlag;
	
	/** True if 10 hour rest is due after unloading 
	 */
	protected boolean myTenHourRestFlag;
	
	/** True if 34 hour rest is due after unloading 
	 */
	protected boolean myThirtyFourHourRestFlag;
	
	/** True if driver doesnt have to rest immediately after unloading
	 */
	protected boolean myIdleFlag;
	
	/** True if the driver has already been on a task atleast once or if it has completed its first
	 *  14 hour on duty cycle since it was added to the network
	 */
	protected boolean myNextTaskFlag;
	
	/** The current task this driver is performing
	 */
	protected DispatchTask myCurrentTask;    
	
	/** Current location of this driver 
	 */
	protected TransportLocationIfc myCurrentLocation;
	
	/** Initial location of the driver where it was assigned 
	 */
	protected TransportLocationIfc myInitialLocation;
	
	/** Current coordinates of the drivers locations 
	 */
	protected CoordinateIfc myCurrentCoordinate;
	
	/** Initial coordinates of the location of the driver where it was assigned  
	 */
	protected CoordinateIfc myInitialCoordinate;
	
	/** The probability distribution for 10 hour rest following the 14 hour on duty cycle 
	 */
	protected DistributionIfc myFourteenHourRuleRestTimeDistribution;
	
	/** Random variable for 10 hour rest time distribution following the 14 hour on duty cycle  
	 */
	protected RandomVariable myFourteenHourRuleRestTimeRV;
	
	/** The probability distribution for 10 hour rest 
	 */
	protected DistributionIfc myTenHourRuleRestTimeDistribution;
	
	/** Random variable for 10 hour rest time distribution  
	 */
	protected RandomVariable myTenHourRuleRestTimeRV;
	
	/** The probability distribution for 34 hour rest  
	 */
	protected DistributionIfc myThirtyFourHourRuleRestTimeDistribution;
	
	/** Random variable for 34 hour rest time distribution  
	 */
	protected RandomVariable myThirtyFourHourRuleRestTimeRV;
	
	/** The probability distribution for 34 hour rest after driver has accumulated 70 hours towards its on duty  
	 */
	protected DistributionIfc mySeventyHourRuleRestTimeDistribution;
	
	/** Random variable for 34 hour rest time distribution after driver has accumulated 70 hours towards its on duty  
	 */
	protected RandomVariable mySeventyHourRuleRestTimeRV;
	
	/**
	 * Schedules Rest of 14 Hour Rule at the end of every 14 hour cycle of the driver
	 */
	protected FourteenHourRuleRestingListener myFourteenHourRuleRestingListener;
	
	/**
	 * Schedules Rest of 10 Hours in case the transport cannot be completed in the driving time available
	 * 
	 */
	protected TenHourRestingListener myTenHourRestingListener;
	
	/**
	 * Schedules begining of 10 Hour Rest of driver when it completed the 14 Hour Cycle
	 */
	protected TenHourRestingOfFourteenHourRuleListener myTenHourHourRestingOfFourteenHourRuleListener;
	
	/**
	 * Schedules Rest of 70 Hour Rule when the driver accumulated 70 hours on duty hours
	 */
	protected SeventyHourRuleRestingListener mySeventyHourRuleRestingListener;
	
	/**
	 * Schedules Rest of 34 Hours in case the transport cannot be completed in the driving time available
	 * 
	 */
	protected ThirtyFourHourRestingListener myThirtyFourHourRestingListener;
	
	/**
	 * Schedules begining of 34 Hour Rest of driver 
	 */
	protected ThirtyFourHourRestingOfSeventyHourRuleListener myThirtyFourHourRestingOfSeventyHourRuleListener;
	
	/** Schedules end of incomplete transport  
	 */
	protected IncompleteTransportListener myIncompleteTransportListener;
	
	/** Schedules end of complete transport  
	 */
	protected CompleteTransportListener myCompleteTransportListener;
	
	/** Last Event that occured 
	 */
	protected JSLEvent myLastEvent;
	
	/** Event scheduled at the end of 14 hours on duty cycle 
	 */
	protected JSLEvent myFourteenHourRuleRestEvent;
	
	/** Event scheduled at the end of 14 hour on duty cycle while performing task  (10 hour rest)
	 */
	protected JSLEvent myTenHourRestWithTaskAttachedEvent;
	
	/** Event scheduled at the begining of 10 hour rest 
	 */
	protected JSLEvent myTenHourRestOfFourteenHourRuleEvent;
	
	/** Event scheduled when 70 hours on duty are accumulated 
	 */
	protected JSLEvent mySeventyHourRuleRestEvent;
	
	/** Event scheduled at the end of 70 hour on duty cycle during while performing task (34 hour rest)
	 */
	protected JSLEvent myThirtyFourHourRestWithTaskAttachedEvent;
	
	/** Event schedules end of 34 hour rest
	 */
	protected JSLEvent myThirtyFourHourRestOfSeventyHourRuleEvent;
	
	/** Incomplete Transport Event scheduled when sufficient driving time is not available
	 *  to complete the transport
	 */
	protected JSLEvent IncompleteTransportEvent;
	
	/** Complete Transport Event scheduled when sufficient driving time is available
	 *  to complete the transport 
	 */
	protected JSLEvent CompleteTransportEvent;
	
	/** Represents maximum driving time available with the driver in a 14 hour cycle 
	 */
	protected double myDrivingTimeLimit;
	
	/** Represents maximum on duty time available with the driver in a 14 hour cycle 
	 */
	protected double myOnDutyTimeLimit;
	
	/** Represents maximum on duty time available with the driver in a 70 hour cycle 
	 */
	protected double myCumulativeOnDutyTimeLimit;
	
	/**
	 *  Time accumulated towards cumulative on duty cycle
	 */
	protected double myTimeAccumulatedTowardsCumulativeOnDutyCycle;
	
	/** Represents loading time  
	 */
	protected double myLoadingTime;
	
	/** Represents unloading time  
	 */
	protected double myUnloadingTime;
	
	/** Represents transport time in a task 
	 */
	protected double myTransportTime;
	
	/** Represents expected transport time in a task 
	 */
	protected double myExpectedTransportTime;
	
	/** Represents expected time taken for a transport task 
	 */
	protected double myExpectedTransportTaskTime;
	
	/** Represents expected time taken for a deadhead part of the task 
	 */
	protected double myExpectedDeadHeadTaskTime;
	
	/** Represents expected time taken for a bobtail part of the task 
	 */
	protected double myExpectedBobTailTaskTime;
	
	/** Represents expected time taken for loading 
	 */
	protected double myExpectedLoadingTime;
	
	/** Represents expected time taken for unloading 
	 */
	protected double myExpectedUnloadingTime;
	
	/** Represents expected time taken for a task 
	 */
	protected double myExpectedTotalTaskTime;
	
	/** Represents driving available with the driver towards its 14 hour cycle  
	 */
	protected double myDrivingTimeAvailable;
	
	/** Represents on duty available with the driver towards its 14 hour cycle  
	 */
	protected double myOnDutyTimeAvailable;
	
	/** Represents Total on duty time accumulated by the driver so far towards 70 hour cycle 
	 */
	protected double myOnDutyTimeAccumulatedTowards70Hours;
	
	/** Represents Total driving time accumulated by the driver so far towards 70 hour cycle 
	 */
	protected double myDrivingTimeAccumulatedTowards70Hours;
	
	/** Represents Total on duty time accumulated by the driver so far towards 14 hour cycle 
	 */
	protected double myOnDutyTimeAccumulatedTowards14Hours;
	
	/** Represents Total on duty time accumulated by the driver so far towards 14 hour cycle 
	 */
	protected double myDrivingTimeAccumulatedTowards14Hours;
	
	/** Represents Total Expeceted on duty time accumulated by the driver so far towards 14 hour cycle 
	 */
	protected double myExpectedDrivingTimeAvailable;
	
	/** Represents Total Expected on duty time accumulated by the driver so far towards 14 hour cycle 
	 */
	protected double myExpectedOnDutyTimeAvailable;
	
	/** Represents Total Expected on duty time accumulated by the driver so far towards 70 hour cycle 
	 */
	protected double myExpectedOnDutyTimeAccumulatedTowards70Hours;
	
	/** Represents Total Expected on duty time accumulated by the driver so far towards 14 hour cycle 
	 */
	protected double myExpectedOnDutyTimeAccumulatedTowards14Hours;
	
	/** Represents transport time of the task remaining to be completed  
	 */
	protected double myTransportTimeRemaining;
	
	/** Represents time remaining towards accumulation of 70 hours of on duty 
	 */
	protected double myTimeRemainingTowards70Hours;
	
	/** Represents the simulation time at which the event started 
	 */
	protected double myEventStartTime;
	
	/** Represents the simulation time at which the last event ended 
	 */
	protected double myLastEventTime;
	
	/** Represents the simulation time at which the task started 
	 */
	protected double myTaskBeginTime;
	
	/** Represents the simulation time at which the task ended 
	 */
	protected double myTaskEndTime;
	
	/** Represents the network to which it is added 
	 */
	protected TransportationNetwork myNetwork;
	
	/** Represents the dispatcher whom whenever it goes off duty and on duty 
	 */
	protected AbstractDispatcherFTL myDispatcher;
	
	/** Represents velocity distribution as it covers miles 
	 */
	protected DistributionIfc myVelocityDistribution;
	
	/** Random Variable for the velocity distribution 
	 */
	protected RandomVariable myVelocityRV;
	
	/**
	 *  Cumulative on duty Limit (70 hours here)
	 */
	protected double myCumulativeOnDutyLimit;
	//  ===========================================
	//  CONSTRUCTORS
	//  ===========================================
	
	/**
	 * Creates Driver
	 * @param parent
	 * @param coordinate
	 * @param location
	 * @param onDutyTimeLimit
	 * @param drivingTimeLimit
	 * @param cumulativeOnDutyTimeLimit
	 * @param restTimeDistributionForFourteenHourRule
	 * @param restTimeDistributionForTenHourRule
	 * @param restTimeDistributionForThirtyFourHourRule
	 * @param velocityDistribution
	 */
	public Driver(ModelElement parent,CoordinateIfc coordinate, TransportLocationIfc location,double onDutyTimeLimit, double drivingTimeLimit,
			double cumulativeOnDutyTimeLimit,DistributionIfc restTimeDistributionForFourteenHourRule,DistributionIfc restTimeDistributionForTenHourRule,
			DistributionIfc restTimeDistributionForThirtyFourHourRule,DistributionIfc velocityDistribution) {
			this(parent,null,coordinate,location,onDutyTimeLimit,drivingTimeLimit,cumulativeOnDutyTimeLimit,restTimeDistributionForFourteenHourRule,restTimeDistributionForTenHourRule,restTimeDistributionForThirtyFourHourRule,
				velocityDistribution);
	}
	
	/**
	 * Creates Driver
	 * @param parent
	 * @param name
	 * @param coordinate
	 * @param location
	 * @param onDutyTimeLimit
	 * @param drivingTimeLimit
	 * @param cumulativeOnDutyTimeLimit
	 * @param restTimeDistributionForFourteenHourRule
	 * @param restTimeDistributionForTenHourRule
	 * @param restTimeDistributionForThirtyFourHourRule
	 * @param velocityDistribution
	 */
	public Driver(ModelElement parent, String name, CoordinateIfc coordinate, TransportLocationIfc location, double onDutyTimeLimit,double drivingTimeLimit,
			double cumulativeOnDutyTimeLimit, DistributionIfc restTimeDistributionForFourteenHourRule,DistributionIfc restTimeDistributionForTenHourRule,
			DistributionIfc restTimeDistributionForThirtyFourHourRule,DistributionIfc velocityDistribution) {
		super(parent,1,name,null,coordinate);
		myFourteenHourRuleRestingListener = new FourteenHourRuleRestingListener();
		myTenHourRestingListener = new TenHourRestingListener();
		myTenHourHourRestingOfFourteenHourRuleListener = new TenHourRestingOfFourteenHourRuleListener();
		mySeventyHourRuleRestingListener = new SeventyHourRuleRestingListener();
		myThirtyFourHourRestingListener = new ThirtyFourHourRestingListener();
		myThirtyFourHourRestingOfSeventyHourRuleListener = new ThirtyFourHourRestingOfSeventyHourRuleListener();
		myIncompleteTransportListener = new IncompleteTransportListener();
		myCompleteTransportListener = new CompleteTransportListener();
		setOnDutyTimeLimit(onDutyTimeLimit);
		setDrivingTimeLimit(drivingTimeLimit);
		setCumulativeOnDutyTimeLimit(cumulativeOnDutyTimeLimit);
		setRestTimeDistributionForFourteenHourRule(restTimeDistributionForFourteenHourRule);
		setRestTimeDistributionForTenHourRule(restTimeDistributionForTenHourRule);
		setRestTimeDistributionForThirtyFourHourRule(restTimeDistributionForThirtyFourHourRule);
		setVelocityDistribution(velocityDistribution);
		setCurrentLocation(location);
		setCurrentCoordinate(location);
		setInitialLocation(location);
		setInitialCoordinate(location);
	}
	
	/**
	 *  Following things happen when driver is initialized:
	 * 	1)On Duty Time Available at the beginning of 14 hour cycle as myOnDutyTimeLimit(14)
	 *  2)Driving Time Available at the beginning of 14 hour cycle as myDrivingTimeLimit(11)
	 *  3)Driving time accumulated towards 14 hour cycle as zero
	 *  4)On duty time accumulated towards 14 hour cycle as zero
	 *  5)On duty time accumulated towards 70 hour cycle as zero
	 *  6)Driving time accumulated towards 70 hours cycle as zero 
	 *  7)Driver assigned location is set as its initial and current location
	 *  8)Schedule the end of 14 hour on duty cycle followed by scheduling 10 hour off duty 
	 *  9)Drivers current task is set to null
	 */
	protected void initialize(){
		super.initialize();
		if(myCurrentTask!=null){
		//	System.out.println("Driver: "+getId()+"task: "+myCurrentTask.getId());
		//	System.out.println("Task origin: "+myCurrentTask.getOrigin().getName()+"task desination: "+myCurrentTask.getDestination().getName());
			throw new IllegalArgumentException("Driver "+getId()+" still has a task: "+myCurrentTask.getId()+" at: "+myCurrentLocation.getName());
		}
		resetOnDutyTimeAvailable(myOnDutyTimeLimit);
		resetDrivingTimeAvailable(myDrivingTimeLimit);
		mySeventyHourRuleRestEvent=null;
		myFourteenHourRuleRestEvent=null;
		myTenHourRestWithTaskAttachedEvent=null;
		myTenHourRestOfFourteenHourRuleEvent=null;
		mySeventyHourRuleRestEvent=null;
		myThirtyFourHourRestWithTaskAttachedEvent=null;
		myThirtyFourHourRestOfSeventyHourRuleEvent=null;
		IncompleteTransportEvent=null;
		CompleteTransportEvent=null;
		resetOnDutyTimeAccumulatedTowards14Hours(0.0);
		resetDrivingTimeAccumulatedTowards14Hours(0.0);
		resetOnDutyTimeAccumulatedTowards70Hours(0.0);
		resetDrivingTimeAccumulatedTowards70Hours(0.0);
		collectOnDutyTimeAccumulatedTowards70Hours(myOnDutyTimeAccumulatedTowards70Hours);
		collectDrivingTimeAccumulatedTowards70Hours(myDrivingTimeAccumulatedTowards70Hours);
		collectOnDutyTimeAccumulatedTowards14Hours(myOnDutyTimeAccumulatedTowards14Hours);
		collectDrivingTimeAccumulatedTowards14Hours(myDrivingTimeAccumulatedTowards14Hours);
		schedule14HourRuleRest();
		myLastEventTime=getTime();
		setCurrentLocation(myInitialLocation);
		setCurrentCoordinate(myInitialLocation);
		setCurrentTask(null);
		setLoadingFlag(false); 
		setUnloadingFlag(false);
		setIncompleteTransportFlag(false); 
		setTaskAttachedWith10HourRestFlag(false); 
		setTaskAttachedWith34HourRestFlag(false);
		setNoTaskFlag(true); // the driver has no task
		setTenHourRestFlag(false); 
		setThirtyFourHourRestFlag(false);
		setIdleFlag(false);
		setNextTaskFlag(false);
		}
	
	/** The driver receives the task from the dispatcher.It sets its current task as the task supplied.
	 *  The diver sets its current location and coordinates as that of the task's origin. The hours of service are updated.
	 *  Task's begining time is set as the current simulation time. This method delegates the work to performTransport(), 
	 *  performDeadHeadToLoad(), performBobTailToLoad(),performDeadHeadNoLoad(),performBobTailNoLoadCompleted() depending on the type of task.
	 * 
	 * 	@param task
	 */
	protected void performTask(DispatchTask task){
		
		if(task==null)
			throw new IllegalArgumentException("Task is null!");
		setCurrentTask(task);
	//	System.out.println("t> "+getTime()+" Driver is assigned task "+task.getId());
		setCurrentLocation(task.getOrigin());
		setCurrentCoordinate(task.getOrigin());
		setNextTaskFlag(false);
		setNoTaskFlag(false); // driver started performing a task
		collectOnDutyTimeAccumulatedTowards70Hours(getTime()- myLastEventTime);
		collectOnDutyTimeAccumulatedTowards14Hours(getTime()- myLastEventTime);
		setDrivingTimeAvailable(myDrivingTimeAccumulatedTowards14Hours);
		setOnDutyTimeAvailable(myOnDutyTimeAccumulatedTowards14Hours);
		
//		double time = getExpectedTotalTaskTime(task, myCurrentLocation);
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" is ready to perform the Task "+ task.getId());
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" is Expected to take "+time+" hours to Complete the Total Task "+ task.getId());
	//	System.out.println("The Task is Expected to get Completed at "+(time+getTime())+" from the Current Simulation Time: "+getTime());
		
		double taskBeginTime=getTime();
		setTaskBeginTime(taskBeginTime);
		
		if (task.isTransport())
			performTransport(task);
		else if (task.isDeadHeadToLoad())
			performDeadHeadToLoad(task);
		else if (task.isBobTailToLoad())
			performBobTailToLoad(task);
		else if (task.isBobTailNoLoad())
			performBobTailNoLoad(task);
		else if (task.isDeadHeadNoLoad())
			performDeadHeadNoLoad(task);
	}
	
	/** The driver performs the Transport task.Here the driver checks if the driver has sufficient on duty time available 
	 *  to complete the "Loading". The driver gets the required "loading time" from the location of the task via getLoadingTime(task)
	 *  and sets its as its loading time.If the loading time is less than the on duty time available then the driver schedules loading.
	 *  If the loading time is greater than the on duty time available, then the driver will cancel any rest (10 hour or 34 hour rest)
	 *  that was scheduled in the future and rescedule it at the current simulation time.
	 * 
	 * @param task
	 */
	protected void performTransport(DispatchTask task){
		
		setCurrentLocation(task.getOrigin());
		setCurrentCoordinate(task.getOrigin());
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" is ready to perform the Transport of Task "+ task.getId()+" at "+myCurrentLocation.getName());
//		double time = getExpectedTotalTaskTime(task,myCurrentLocation);
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" is Expected to take  "+time+" hrs to complete the Transport of Task "+ task.getId());
		double loadingTime = getLoadingTime(task);
		setLoadingTime(loadingTime);
	//	System.out.println("t> "+getTime()+"Loading will take  "+loadingTime+" hours for Task "+task.getId()+" at "+myCurrentLocation.getName());
		
		if(loadingTime < myOnDutyTimeAvailable){	
			System.out.println("Loading Time < On Duty Time Available");
			System.out.println("Schedule Loading of Task "+task.getId());
			scheduleLoading(task);
		}
		else{
			setLoadingFlag(true); // Loading Flag is set true implying that loading is required in future
			System.out.println("Loading Time > = On Duty Time Available");
			System.out.println("Schedule rest first and then resume Loading of Task "+task.getId());
			// if 10 hour rest of 14 hour rule is scheduled
			if(myFourteenHourRuleRestEvent!=null){
				if(myFourteenHourRuleRestEvent.isScheduled()){
					cancelEvent(myFourteenHourRuleRestEvent);
				//	System.out.println("Fourteen hour rule rest event is scheduled");
					schedule10HourRestWithTaskAttached(task);				
				}
			}
			// if 34 hour rest of 70 hour rule is scheduled
			if(mySeventyHourRuleRestEvent!=null){
				if(mySeventyHourRuleRestEvent.isScheduled()){
					cancelEvent(mySeventyHourRuleRestEvent);
				//	System.out.println("Seventy hour rule rest event is scheduled");
					schedule34HourRestWithTaskAttached(task);
				}
			}
		}
	}
	
	/** The driver performs the BobTailToLoad task.The current loaction and the coordinates of the driver are set
	 *  as the task's origin.
	 * 
	 * @param task
	 */
	protected void performBobTailToLoad(DispatchTask task){
		setCurrentLocation(task.getOrigin());
		setCurrentCoordinate(task.getOrigin());
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" is ready to perform the BobTailToLoad part of Task"+ task.getId()+" at "+myCurrentLocation.getName());
		double time = getExpectedTotalTaskTime(task,myCurrentLocation);
		System.out.println("t>"+getTime()+" Driver "+getId()+" is expected to take  "+time+" hrs to complete the BobTail part of Task "+ task.getId());
	//	System.out.println("The destination of the current BobTail Task is: "+task.getDestination().getName());
		scheduleTransport(task);
	}
	
	/** The driver performs the DeadHeadToLoad task.The current location and the coordinates of the driver are set
	 *  as the task's origin.
	 * 
	 * @param task
	 */
	protected void performDeadHeadToLoad(DispatchTask task){
		setCurrentLocation(task.getOrigin());
		setCurrentCoordinate(task.getOrigin());
		double time = getExpectedTotalTaskTime(task,myCurrentLocation);
		System.out.println("t>"+getTime()+" Driver "+getId()+" is expected to take  "+time+" hrs to complete the DeadHead part of Task "+ task.getId());
		//System.out.println("t>"+getTime()+" Driver "+getId()+" is ready to perform the DeadHeadToLoad part of Task "+ task.getId()+" at "+myCurrentLocation.getName());
		//System.out.println("The destination of the current DeadHead Task is: "+task.getDestination().getName());
		scheduleTransport(task);
	}
	
	/** Driver schedules loading at the origin of the load. The location is asked to schedule loading by 
	 *  supplying the task and the loading time
	 * 
	 * @param task
	 */
	protected void scheduleLoading(DispatchTask task){
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" starts Loading for Task "+task.getId()+" at "+myCurrentLocation.getName());
		Location l = (Location)task.getOrigin();
		double loadingTime = getLoadingTime();
		l.loading(task,loadingTime);
	}
	
	/** Upon completion of loading, the dispatcher is notified.The On Duty time towards 14 and 70 hours are updated.
	 *  The driving and on duty time available are updated after loading. The driver next schedules transport.
	 *  
	 * @param task
	 */
	protected void loadingCompleted(DispatchTask task){
		
		DispatcherFTL dispatcher = (DispatcherFTL)myNetwork.getDispatcher();
		dispatcher.loadingCompleted(task);
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" completes Loading for Task "+task.getId()+" at "+myCurrentLocation.getName());
		myLastEventTime=getTime();
		double loadingTime = getLoadingTime();
		collectOnDutyTimeAccumulatedTowards70Hours(loadingTime);
		collectOnDutyTimeAccumulatedTowards14Hours(loadingTime);
		setDrivingTimeAvailable(myDrivingTimeAccumulatedTowards14Hours);
		setOnDutyTimeAvailable(myOnDutyTimeAccumulatedTowards14Hours);
		scheduleTransport(task);
	}
	
	/** This method schedules tranport for all types of task:transport, bobtail or deadhead. The driver gets the the time taken to
	 *  transport from the task via getTransportTime(task)and sets it as its current tasks transport time. The driver sets the transport
	 *  time as its current "transport time remaining". If the transport time taken is greater than the driving time available with the driver
	 *  then the driver will drive for the driving time available and take rest at the end of the transport. "myIncompleteFlag" is turned on.
	 *  The 14 and 70 hour rest which were scheduled will be cancelled and 14 and 10 hour rest with task will be scheduled at the same time.
	 *  At the the end of incomplete transport; one of these two rests will be scheduled. If "myTaskAttachedWith10HourRestFlag" is true then 
	 *  10 hour rest will be scheduled, on the other hand, if "myTaskAttachedWith34HourRestFlag" is true then 34 hour rest will be scheduled.
	 *  If the driver has sufficient driving time available to complete the task then a complete transport event is scehduled for 
	 *  the transport time.
	 * 
	 * @param task
	 */
	protected void scheduleTransport(DispatchTask task){
		
		setCurrentLocation(task.getOrigin());
		setCurrentCoordinate(task.getOrigin());
//		System.out.println("t>"+getTime()+" Driver "+task.getDriver().getId()+" is scheduled for Transport of Task"+task.getId()+
//				" from Origin : "+myCurrentLocation.getName()+" to the Destination : "+task.getDestination().getName());
		double transportTime=getTransportTime(task);
		setTransportTime(transportTime);
		setTransportTimeRemaining(transportTime);
//		System.out.println("t>"+getTime()+" Driver "+getId()+" has "+myDrivingTimeAvailable+" hours of Driving Time available");		
		
		if(myTransportTimeRemaining>= myDrivingTimeAvailable){
	//		System.out.println("t>"+getTime()+" Task "+task.getId()+" takes more time than the driving hours available with the Driver "+getId());
			double t = myDrivingTimeAvailable;
			setIncompleteTransportFlag(true);
			
			if(myFourteenHourRuleRestEvent!=null){	
				if(myFourteenHourRuleRestEvent.isScheduled()){
					if(myFourteenHourRuleRestEvent.isScheduled()){
						cancelEvent(myFourteenHourRuleRestEvent);
					//	System.out.println("t>"+getTime()+" Driver "+getId()+" will drive for  "+t+" hours and then take a Rest of 10 hours");
						setTaskAttachedWith10HourRestFlag(true);
					}
				}
			}
			
			if(mySeventyHourRuleRestEvent!=null){
				if(mySeventyHourRuleRestEvent.isScheduled()){
					cancelEvent(mySeventyHourRuleRestEvent);
				//	System.out.println("Driver "+getId()+" will drive for  "+t+" hours and then take a Rest of 34 hours");
					setTaskAttachedWith34HourRestFlag(true);
				}
			}
			
			myEventStartTime=getTime();
		//	System.out.println("t>"+myEventStartTime+" Incomplete Transport Event for starts");
			IncompleteTransportEvent = scheduleEvent(myIncompleteTransportListener,t,"Incomplete Transport",task);
		}
		else{
		//	System.out.println("Task "+task.getId()+" takes less time than the driving hours available "+ 
		//			"with the driver "+getId());
			double t=myTransportTimeRemaining;
		//	System.out.println("Transport Time Remaining: "+t);
			myEventStartTime=getTime();
		//	System.out.println("t> "+myEventStartTime+" Complete Transport Event starts");
			CompleteTransportEvent = scheduleEvent(myCompleteTransportListener,t,"Transport",task);
		}
	}
	
	/** The driver comes to this method if it still has some transport time remaining for its current task. Again the driver will
	 *  check if it has sufficient driving time available to complete the transport. If sufficient driving time is not available
	 *  then it will schedule incomplete transport again otherwise it will schedule complete transport event.
	 * 
	 * @param task
	 */
	protected void scheduleRemainingTransport(DispatchTask task){
		
	//	System.out.println("t>"+getTime()+" Driver"+getId()+" is scheduled for Remaining Transport of Task"+task.getId()+
	//			" from Origin : "+myCurrentLocation.getName()+" to the Destination : "+task.getDestination().getName());
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" has "+myTransportTimeRemaining+" hours of Transport for Task"+task.getId()+"left");
	//	System.out.println("t> "+getTime()+"Driver "+getId()+" has "+myDrivingTimeAvailable+" hours of driving time available");		
		
		if(myTransportTimeRemaining>=myDrivingTimeAvailable){
		//	System.out.println("Task "+task.getId()+" takes more time than the driving hours available "+ 
		//			"with the driver "+getId());
			
			double t = myDrivingTimeAvailable;
			setIncompleteTransportFlag(true);
			
			if(myFourteenHourRuleRestEvent!=null){
				if(myFourteenHourRuleRestEvent.isScheduled()){
					cancelEvent(myFourteenHourRuleRestEvent);
				//	System.out.println("Driver "+getId()+" will drive for  "+t+" hours and then take a Rest of 10 hours");
					setTaskAttachedWith10HourRestFlag(true);
				}
			}
			
			if(mySeventyHourRuleRestEvent!=null){
				if(mySeventyHourRuleRestEvent.isScheduled()){
					cancelEvent(mySeventyHourRuleRestEvent);
				//	System.out.println("Driver "+getId()+" will drive for  "+t+" hours and then take a Rest of 34 hours");
					setTaskAttachedWith34HourRestFlag(true);
				}
			}
			myEventStartTime=getTime();
		//	System.out.println("t> "+myEventStartTime+" Incomplete Transport Event starts at "+myCurrentLocation.getName());
			IncompleteTransportEvent = scheduleEvent(myIncompleteTransportListener,t,"Incomplete Transport",task);
		}
		else{
		//	System.out.println("Task "+task.getId()+" takes less time than the driving hours available "+"with the driver "+getId());
			double t=myTransportTimeRemaining;
			myEventStartTime=getTime();
		//	System.out.println("t> "+myEventStartTime+" Complete Transport Event starts at "+myCurrentLocation.getName());
			CompleteTransportEvent = scheduleEvent(myCompleteTransportListener,t,"Complete Transport",task);
		}
	}
	
	/** Private class for the incomplete transport listener.At the end of the incomplete transport event the time remaining for 
	 *  the transport of the task is updated. The on duty and driving time accumulated towards 14 and 70 hours are updated.The driving 
	 *  and on duty time available are updated. Since at the end of incomplete transport, a 10 hour or 34 hour rest has to be scheduled.  
	 *  Depending on whether "myTaskAttachedWith10HourRestFlag" or "myTaskAttachedWith34HourRestFlag" is turned on, the appropriate 
	 *  rest while driver is on a task is scheduled.
	 * 
	 * @author snangia
	 *
	 */
	class IncompleteTransportListener implements ActionListenerIfc{
		public void action(JSLEvent event){
			
			DispatchTask task = (DispatchTask)event.getMessage();  
		//	System.out.println("t> "+getTime()+" Incomplete Transport Ends for Driver "+getId());
			
			double t=(getTime()- myEventStartTime);
		//	System.out.println("Time taken by the Incomplete Transport Event:"+t);
			setTransportTimeRemaining(myTransportTime-t);
			collectDrivingTimeAccumulatedTowards14Hours(t);
			collectOnDutyTimeAccumulatedTowards14Hours(t);
			collectDrivingTimeAccumulatedTowards70Hours(t);
			collectOnDutyTimeAccumulatedTowards70Hours(t);
			
			setDrivingTimeAvailable(myDrivingTimeAccumulatedTowards14Hours);
			setOnDutyTimeAvailable(myOnDutyTimeAccumulatedTowards14Hours);
			
			if(myTaskAttachedWith10HourRestFlag){
				setTaskAttachedWith10HourRestFlag(false);
			//	System.out.println("t> "+getTime()+" Driver "+getId()+" is scheduled for 10 hours rest now");
				schedule10HourRestWithTaskAttached(task);	
			}
			
			if(myTaskAttachedWith34HourRestFlag){
				setTaskAttachedWith34HourRestFlag(false);
			//	System.out.println("t> "+getTime()+" Driver "+getId()+" is scheduled for 34 hours rest now");
				schedule34HourRestWithTaskAttached(task);	
			}
		}
	}
	
	/** At the end of the complete transport, the driver's current location is set as the destination of the task. 
	 *  The on duty and driving time accumulated towards 14 and 70 hours are updated.This method delegates the
	 *  work to transportTaskCompleted(), deadHeadToLoadTaskCompleted(), bobTailToLoadTaskCompleted()
	 *  deadHeadNoLoadTaskCompleted(), bobTailNoLoadTaskCompleted() depending on the type of task.
	 *  @author snangia
	 *
	 */
	class CompleteTransportListener implements ActionListenerIfc{
		public void action(JSLEvent event){
			
			DispatchTask task = (DispatchTask)event.getMessage();  
		//	System.out.println("t>"+getTime()+" Driver "+getId()+" Completed Transport ");
			
			setCurrentLocation(task.getDestination());
			setCurrentCoordinate(task.getDestination());
			
		//	System.out.println("Event Start Time: "+myEventStartTime);		
			double t=(getTime()- myEventStartTime);
		//	System.out.println("Time Taken for the Last Transport:"+t);
			
			collectDrivingTimeAccumulatedTowards14Hours(t);
			collectOnDutyTimeAccumulatedTowards14Hours(t);
			collectDrivingTimeAccumulatedTowards70Hours(t);
			collectOnDutyTimeAccumulatedTowards70Hours(t);
			setDrivingTimeAvailable(myDrivingTimeAccumulatedTowards14Hours);
			setOnDutyTimeAvailable(myOnDutyTimeAccumulatedTowards14Hours);		
			
			if (task.isTransport())
				transportTaskCompleted(task);
			else if (task.isDeadHeadToLoad())
				deadHeadToLoadTaskCompleted(task);
			else if (task.isBobTailToLoad())
				bobTailToLoadTaskCompleted(task);
			else if (task.isBobTailNoLoad())
				bobTailNoLoadTaskCompleted(task);
			else if (task.isDeadHeadNoLoad())
				deadHeadNoLoadTaskCompleted(task);
		}
	}
	
	/** When the transport of the Transport task gets completed, then the driver schedules unloading. The driver gets the required unloading
	 *  time from the location via getUnloadingTime(task). Irrespective of the fact whether sufficient on duty time is available with the driver 
	 *  or not, the driver will schedule unloading immediately after transport. If the sufficient time is not available, the turn on
	 *  "myUnloadingFlag" indicating that the driver has to take immediately after it has completed unloading.Any 14 or 70 hour rule which was scheduled
	 *   will be delayed by the unloading time in this case.
	 * @param task
	 */
	protected void transportTaskCompleted(DispatchTask task){
	 
		double unloadingTime = getUnloadingTime(task);
		setUnloadingTime(unloadingTime);
		if(unloadingTime>=myOnDutyTimeAvailable){
			System.out.println("Unloading Time>=Driving Time Available");
			setUnloadingFlag(true);
		
			if(myFourteenHourRuleRestEvent!=null){
				if(myFourteenHourRuleRestEvent.isScheduled()){
					cancelEvent(myFourteenHourRuleRestEvent);
					setTenHourRestFlag(true);
				}
			}
			
			if(mySeventyHourRuleRestEvent!=null){
				if(mySeventyHourRuleRestEvent.isScheduled()){
					cancelEvent(mySeventyHourRuleRestEvent);
					setThirtyFourHourRestFlag(true);
				}
			}
			scheduleUnloading(task);
		}else{
		//	System.out.println("Unloading Time<Driving Time Available");
			scheduleUnloading(task);
		}
	}
	
	
	/** When the transport of the deadHeadToLoad task gets completed, the driver accumulates the deadhead miles it has covered so far.
	 *  The driver will notify the task via taskCompleted() that it has executed the dead head part of the transport.
	 * 
	 * @param task
	 */
	protected void deadHeadToLoadTaskCompleted(DispatchTask task){
		//		System.out.println("t>"+getTime()+" Driver "+getId()+" completes DeadHead to Load of Task "+task.getId()+" at "+myCurrentLocation.getName());
		//		System.out.println("t>"+getTime()+" Actual On Duty time accumulated towards 70 hours by the driver "+getId()+" at the end of deadhead to load task "+task.getId()+" is "+myOnDutyTimeAccumulatedTowards70Hours);
		//		System.out.println("t>"+getTime()+" Actual Driving time accumulated towards 70 hours by the driver "+getId()+" at the end of deadhead to load task "+task.getId()+" is "+myDrivingTimeAccumulatedTowards70Hours);
		myLastEventTime=getTime();
		double taskEndTime=getTime();
		setTaskEndTime(taskEndTime);
		double td=(myTaskEndTime-myTaskBeginTime);
		double d= myNetwork.getLane(task.getOrigin(),task.getDestination()).getDistance();
		myDispatcher.myAverageDeadHeadMiles.setValue(d);
		DispatcherFTL dispatcher = (DispatcherFTL)myDispatcher;
		dispatcher.myDeadHeadMiles.collect(d);
		myDispatcher.myAvgNumDeadHeadMiles.increment(d);
		System.out.println("myDispatcher.myAvgNumDeadHeadMiles:"+myDispatcher.myAvgNumDeadHeadMiles.getValue());
		System.out.println("Actual Time taken to complete the DeadHead Part of Task "+task.getId()+" is "+td);
		task.taskCompleted();	
	}
	
	/**	When the transport of the bobTailToLoad task gets completed, the driver accumulates the bobtail miles it has covered so far.
	 *  The driver will notify the task via taskCompleted() that it has executed the bob tail part of the transport.
	 *  
	 * @param task
	 */
	protected void bobTailToLoadTaskCompleted(DispatchTask task){
		//		System.out.println("t>"+getTime()+" Driver "+getId()+" completes BobTail to Load of Task "+task.getId()+" at "+myCurrentLocation.getName());
		//		System.out.println("t>"+getTime()+" Actual On Duty time accumulated towards 70 hours by the driver "+getId()+" at the end of bobtail to load task "+task.getId()+" is "+myOnDutyTimeAccumulatedTowards70Hours);
		//		System.out.println("t>"+getTime()+" Actual Driving time accumulated towards 70 hours by the driver"+getId()+" at the end of bobtail to load task "+task.getId()+" is "+myDrivingTimeAccumulatedTowards70Hours);
		myLastEventTime=getTime();
		double taskEndTime=getTime();
		setTaskEndTime(taskEndTime);
		double tb=(taskEndTime-myTaskBeginTime);
		double d= myNetwork.getLane(task.getOrigin(),task.getDestination()).getDistance();
		myDispatcher.myAverageBobTailMiles.setValue(d);
		DispatcherFTL dispatcher = (DispatcherFTL)myDispatcher;
		dispatcher.myAvgNumBobTailMiles.increment(d);
		dispatcher.myBobTailMiles.collect(d);
		System.out.println("myDispatcher.myAvgNumBobTailMiles:"+myDispatcher.myAvgNumBobTailMiles.getValue());
		System.out.println("Actual Time taken to complete the Bobtail Part of Task "+task.getId()+" is "+tb);
		task.taskCompleted();
	}
	
	/** When the transport of the bobTailNoLoad task gets completed, the driver accumulates the bobtail miles it has covered so far.
	 *  The driver will notify the task via taskCompleted() that it has executed the task.
	 *  
	 * @param task
	 */
	protected void bobTailNoLoadTaskCompleted(DispatchTask task){
		//		System.out.println("t>"+getTime()+" Driver "+getId()+" completes BobTail no Load of Task "+task.getId()+" at "+myCurrentLocation.getName());
		//		System.out.println("t>"+getTime()+" Actual On Duty time accumulated towards 70 hours by the driver "+getId()+" at the end of bobtail no load task "+task.getId()+" is "+myOnDutyTimeAccumulatedTowards70Hours);
		//		System.out.println("t>"+getTime()+" Actual On Duty time accumulated towards 70 hours by the driver "+getId()+" at the end of bobtail no load task "+task.getId()+" is "+myDrivingTimeAccumulatedTowards70Hours);
		myLastEventTime=getTime();
		double taskEndTime=getTime();
		setTaskEndTime(taskEndTime);
		double tb=(taskEndTime-myTaskBeginTime);
		System.out.println("Actual Time taken to complete the Bobtail Task "+task.getId()+" is "+tb);
		double d= myNetwork.getLane(task.getOrigin(),task.getDestination()).getDistance();
		myDispatcher.myAvgNumBobTailMiles.increment(d);
		task.taskCompleted();		
	}
	
	/** When the transport of the deahHeadNoLoad task gets completed, the driver accumulates the deadhead miles it has covered so far.
	 *  The driver will notify the task via taskCompleted() that it has executed the task.
	 * 
	 * @param task
	 */
	protected void deadHeadNoLoadTaskCompleted(DispatchTask task){
		//		System.out.println("t>"+getTime()+" Driver "+getId()+" completes DeadHead no Load of Task "+task.getId()+" at "+myCurrentLocation.getName());
		//		System.out.println("t>"+getTime()+" Actual On Duty time accumulated towards 70 hours by the driver "+getId()+" at the end of deadhead no load task "+task.getId()+" is "+myOnDutyTimeAccumulatedTowards70Hours);
		//		System.out.println("t>"+getTime()+" Actual On Duty time accumulated towards 70 hours by the driver "+getId()+" at the end of deadhead no load task "+task.getId()+" is "+myDrivingTimeAccumulatedTowards70Hours);
		myLastEventTime=getTime();
		double taskEndTime=getTime();
		setTaskEndTime(taskEndTime);
		double td=(taskEndTime-myTaskBeginTime);
		System.out.println("Actual Time taken to complete the DeadHead Task "+task.getId()+" is "+td);
		double d= myNetwork.getLane(task.getOrigin(),task.getDestination()).getDistance();
		myDispatcher.myAvgNumDeadHeadMiles.increment(d); 
		task.taskCompleted();	
	}
	
	
	/**
	 *  Schedules 10 hour rest while driver is performing a task 
	 * @param task
	 */
	protected void schedule10HourRestWithTaskAttached(DispatchTask task){
		
		myDispatcher.driverOffDutyWithTask(this);
		double t = myTenHourRuleRestTimeRV.getValue();
		myTenHourRestWithTaskAttachedEvent = scheduleEvent(myTenHourRestingListener,t,"10 Hour Resting",task);	
	}
	
	/**
	 *  After the completion of 10 hour rest, the driver schedules next rest which is either 10 hour rest 
	 *  or 34 hour rest whichever comes earlier. The driver will next resume either loading or remaining transport whichever of 
	 *  two he/she has been performing
	 * 
	 */
	class TenHourRestingListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
		//	System.out.println("Driver: "+getId()+" Last Event Time: "+myLastEventTime);
			DispatchTask task = (DispatchTask)event.getMessage();  
		//	System.out.println("t>"+getTime()+" Driver "+getId()+" completed 10 hour resting and will continue performing remaining Task "+task.getId());
			myDispatcher.driverOnDutyWithTask(Driver.this);
			resetDriverStatusAfter10HourRest();
			
			//	System.out.println("The total on duty time accumulated towards 70 hours by driver "+getId()+" so far is "+myOnDutyTimeAccumulatedTowards70Hours+" hours");
			//	System.out.println("The total driving time accumulated towards 70 hours by driver "+getId()+" so far is "+myDrivingTimeAccumulatedTowards70Hours+" hours");
			
			double t= myOnDutyTimeAccumulatedTowards70Hours + myOnDutyTimeLimit;
			System.out.println("On Duty Time Accumulated Towards 70Hours + On Duty Time Limit is: "+t);
			// if the driver will accumulate more than 70 hours in the next 14 hour on duty cycle then schedule 34 hour rest when required
			if(t>=myCumulativeOnDutyTimeLimit){
				myTimeRemainingTowards70Hours=myCumulativeOnDutyTimeLimit-myOnDutyTimeAccumulatedTowards70Hours;
			//	System.out.println("Which is Greater than 70.0, schedule 34 hour rest of 70 Hour Rule after "+(myTimeRemainingTowards70Hours)+" hours");
			//	schedule70HourRuleRest(myTimeRemainingTowards70Hours);
				myOnDutyTimeAvailable=myTimeRemainingTowards70Hours;
				myDrivingTimeAvailable=myTimeRemainingTowards70Hours;
				if(myDrivingTimeAvailable>=myDrivingTimeLimit)// if the driving time available towards 70 hours is more than the driving limit
					myDrivingTimeAvailable=myDrivingTimeLimit;// the driving time available is equal to the driving time limit
			//	System.out.println("On Duty Time Available: "+myOnDutyTimeAvailable+" hours");
			//	System.out.println("Driving Time Available: "+myDrivingTimeAvailable+" hours");
				}
			else{
			//	System.out.println("Which is Less than 70.0, we do not schedule 34 hour rest of 70 Hour Rule yet!");
				schedule14HourRuleRest();
			}
			
			if(myLoadingFlag){
				setLoadingFlag(false);
				performTransport(task);
			}
			
			if(myIncompleteTransportFlag){
				setIncompleteTransportFlag(false);
				scheduleRemainingTransport(task);
			}
		}
	}
	
	/**
	 * Schedules 34 hour rest while performing a task
	 * @param task
	 */
	 protected void schedule34HourRestWithTaskAttached(DispatchTask task){
		myDispatcher.driverOffDutyWithTask(this);
		double t = myThirtyFourHourRuleRestTimeRV.getValue();
		myThirtyFourHourRestWithTaskAttachedEvent =	scheduleEvent(myThirtyFourHourRestingListener,t, "34 Hour Resting",task);	
	}
	
	/**
	 * 
	 *  After the completion of 34 hour rest, the driver schedules next rest which is either 10 hour rest 
	 *  or 34 hour rest whichever comes earlier. The driver will next resume either loading or remaining transport whichever of 
	 *  two he/she has been performing
	 * 
	 */
	 class ThirtyFourHourRestingListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			
			DispatchTask task = (DispatchTask)event.getMessage();  
		//	System.out.println("t>"+getTime()+" Driver "+getId()+" completed 34 hour resting and will continue performing remaining Task "+task.getId());
			
			myDispatcher.driverOnDutyWithTask(Driver.this);
			resetDriverStatusAfter34HourRest();
			
			//		System.out.println("The total on duty time accumulated towards 70 hours by driver "+getId()+" so far is "+myOnDutyTimeAccumulatedTowards70Hours+" hours");
			//		System.out.println("The total driving time accumulated towards 70 hours by driver "+getId()+" so far is "+myDrivingTimeAccumulatedTowards70Hours+" hours");
			
			double t= myOnDutyTimeAccumulatedTowards70Hours + myOnDutyTimeLimit;
		//	System.out.println("On Duty Time Accumulated Towards 70Hours + On Duty Time Limit is: "+t);
			
			if(t>=myCumulativeOnDutyTimeLimit){
								
				myTimeRemainingTowards70Hours=myCumulativeOnDutyTimeLimit-myOnDutyTimeAccumulatedTowards70Hours;
		//		System.out.println("Which is Greater than 70.0, schedule 34 hour rest of 70 Hour Rule after "+(myTimeRemainingTowards70Hours)+" hours");
				schedule70HourRuleRest(myTimeRemainingTowards70Hours);
				
				myOnDutyTimeAvailable=myTimeRemainingTowards70Hours;
				myDrivingTimeAvailable=myTimeRemainingTowards70Hours;
				if(myDrivingTimeAvailable>=myDrivingTimeLimit)
					myDrivingTimeAvailable=myDrivingTimeLimit;
		//		System.out.println("On Duty Time Available: "+myOnDutyTimeAvailable+" hours");
		//		System.out.println("Driving Time Available: "+myDrivingTimeAvailable+" hours");
			}
			else{
		//		System.out.println("Which is Less than 70.0, we do not schedule 34 hour rest of 70 Hour Rule yet!");
				schedule14HourRuleRest();	
			}
			
			if(myLoadingFlag){
				setLoadingFlag(false);
				performTransport(task);
			}
			
			if(myIncompleteTransportFlag){
				setIncompleteTransportFlag(false);
				scheduleRemainingTransport(task);
			}
		}
	}
	
	/**
	 *  Schedules 10 hour rest of 14 hour rule 
	 *
	 */
	 protected void schedule14HourRuleRest(){
		setNextTaskFlag(false);
		setNoTaskFlag(false);				
		double t = myFourteenHourRuleRestTimeRV.getValue();
		myFourteenHourRuleRestEvent = scheduleEvent(myFourteenHourRuleRestingListener, t, "Scheduling 14 Hour Rule Resting");	
	}
	
	class FourteenHourRuleRestingListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			
		//	System.out.println("t>"+getTime()+" Driver "+getId()+" Begins 10 Hours Resting");
		//	System.out.println("myLastEventTime: "+myLastEventTime);
		//	System.out.println("myCurrentEventTime: "+getTime());
			//	System.out.println("If the last event happens to unloading, then there is accumulation of on duty time otherwise no");
			collectOnDutyTimeAccumulatedTowards70Hours(getTime()- myLastEventTime);
			collectOnDutyTimeAccumulatedTowards14Hours(getTime()- myLastEventTime);
		//	System.out.println("t> "+"The total on duty time accumulated towards 70 hours by Driver "+getId()+" so far is "+myOnDutyTimeAccumulatedTowards70Hours+" hours");
		//	System.out.println("t> "+"The total on duty time accumulated towards 14 hours by Driver "+getId()+" so far is "+myOnDutyTimeAccumulatedTowards14Hours+" hours");
			schedule10HourRestOfFourteenHourRule();
		}
	}
	
	/**
	 * Schedule 34 hour rest of 70 Hour Rule
	 * @param time
	 */
	protected void schedule70HourRuleRest(double time){
		setNextTaskFlag(false);
		if(time<0.0)
			throw new IllegalArgumentException("Time was less than zero");
		double t = time;
		mySeventyHourRuleRestEvent=scheduleEvent(mySeventyHourRuleRestingListener, t, "Scheduling 70 Hour Rule Resting");			
	}
	
	class SeventyHourRuleRestingListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			
	//		System.out.println("t>"+getTime()+" Driver "+getId()+" Begins 34 Hours Resting");
	//		System.out.println("myLastEventTime: "+myLastEventTime);
	//		System.out.println("myCurrentEventTime: "+getTime());
	//		System.out.println("If the last event happens to unloading, then there is accumulation of on duty time otherwise no");
			collectOnDutyTimeAccumulatedTowards70Hours(getTime()- myLastEventTime);//this will not be zero in case of unloading, in other cases it will be zero
			collectOnDutyTimeAccumulatedTowards14Hours(getTime()- myLastEventTime);
	//		System.out.println("The total on duty time accumulated towards 70 hours by driver "+getId()+" so far is "+myOnDutyTimeAccumulatedTowards70Hours+" hours");
	//		System.out.println("The total on duty time accumulated towards 14 hours by driver "+getId()+" so far is "+myOnDutyTimeAccumulatedTowards14Hours+" hours");
	//		System.out.println("The total driving time accumulated towards 70 hours by driver "+getId()+" so far is "+myDrivingTimeAccumulatedTowards70Hours+" hours");
			schedule34HourRestOfSeventyHourRule();
		}
	}
	
	/**
	 * Schedules 10 hour rest of 14 Hour Rule
	 *
	 */
	protected void schedule10HourRestOfFourteenHourRule(){
	//	System.out.println("Driver: "+getId()+" Last Event Time: "+myLastEventTime);
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" Begins 10 Hours Resting");
		myDispatcher.driverOffDuty(this);
		double t = myTenHourRuleRestTimeRV.getValue();
		myTenHourRestOfFourteenHourRuleEvent = scheduleEvent(myTenHourHourRestingOfFourteenHourRuleListener,t, "10 Hour Resting");	
	}
	
	class TenHourRestingOfFourteenHourRuleListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			
			myDispatcher.driverOnDuty(Driver.this);
		//	System.out.println("t>"+getTime()+" Driver "+Driver.this.getId()+" completed 10 hour resting");
			resetDriverStatusAfter10HourRest();
		//	System.out.println("t>"+getTime()+" Driver "+getId()+" has accumulated "+(myOnDutyTimeAccumulatedTowards70Hours)+" hours towards 70 hours");
			
			double t= myOnDutyTimeAccumulatedTowards70Hours + myOnDutyTimeLimit;
			System.out.println("On Duty Time Accumulated Towards 70Hours+ On Duty Time Limit is: "+t);
			
			if(t>=myCumulativeOnDutyTimeLimit){
				myTimeRemainingTowards70Hours=myCumulativeOnDutyTimeLimit-myOnDutyTimeAccumulatedTowards70Hours;
			//	System.out.println("Which is Greater than 70.0, schedule 34 hour rest of 70 Hour Rule after "+(myTimeRemainingTowards70Hours)+" hours");
				schedule70HourRuleRest(myTimeRemainingTowards70Hours);
				
				myOnDutyTimeAvailable=myTimeRemainingTowards70Hours;
				myDrivingTimeAvailable=myTimeRemainingTowards70Hours;
				if(myDrivingTimeAvailable>=myDrivingTimeLimit)
					myDrivingTimeAvailable=myDrivingTimeLimit;
			//	System.out.println("On Duty Time Available: "+myOnDutyTimeAvailable+" hours");
			//	System.out.println("Driving Time Available: "+myDrivingTimeAvailable+" hours");
				}
			else{
				
			//	System.out.println("Which is Less than 70.0, we do not schedule 34 hour rest of 70 Hour Rule yet!");
				schedule14HourRuleRest();
			}
			TransportLocationIfc location = Driver.this.myCurrentLocation;
			// True if the driver is reallacated after it takes rest after it has delivered load
			if(myIdleFlag){
				setIdleFlag(false);
				DispatcherFTL d=(DispatcherFTL)myDispatcher;
			//	System.out.println("t>"+getTime()+"  "+d.getId()+" is Reallocating Resouces after Driver "+Driver.this.getId()+" is On Duty at "+location.getName());
			//	System.out.println("");
				d.allocateResourcesAfterTransportCompletedAt(location);
			}
		}
	}
	
	/**
	 * The driver goes off duty after it has accumulated 70 hour of on duty. The driver is without
	 * a task. 
	 *
	 */
	protected void schedule34HourRestOfSeventyHourRule(){
		myDispatcher.driverOffDuty(this);
		double t = myThirtyFourHourRuleRestTimeRV.getValue();
		myThirtyFourHourRestOfSeventyHourRuleEvent = scheduleEvent(myThirtyFourHourRestingOfSeventyHourRuleListener, t, "34 Hour Resting");	
	}
	
	class ThirtyFourHourRestingOfSeventyHourRuleListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			
			myDispatcher.driverOnDuty(Driver.this);
		//	System.out.println("t>"+getTime()+" Driver "+getId()+" completed 34 hour resting of Task ");
			resetDriverStatusAfter34HourRest();
			
		//	System.out.println("The total on duty time accumulated towards 70 hours by driver "+getId()+" so far is "+myOnDutyTimeAccumulatedTowards70Hours+" hours");
		//	System.out.println("The total driving time accumulated towards 70 hours by driver "+getId()+" so far is "+myDrivingTimeAccumulatedTowards70Hours+" hours");
			
			double t= myOnDutyTimeAccumulatedTowards70Hours + myOnDutyTimeLimit;
		//	System.out.println("On Duty Time Accumulated Towards 70Hours + On Duty Time Limit is: "+t);
			
			if(t>=myCumulativeOnDutyTimeLimit){
				myTimeRemainingTowards70Hours=myCumulativeOnDutyTimeLimit-myOnDutyTimeAccumulatedTowards70Hours;
			//	System.out.println("Which is Greater than 70.0, schedule 34 hour rest of 70 Hour Rule after "+(myTimeRemainingTowards70Hours)+" hours");
				schedule70HourRuleRest(myTimeRemainingTowards70Hours);
				myOnDutyTimeAvailable=myTimeRemainingTowards70Hours;
				myDrivingTimeAvailable=myTimeRemainingTowards70Hours;
				if(myDrivingTimeAvailable>=myDrivingTimeLimit)
					myDrivingTimeAvailable=myDrivingTimeLimit;
			//	System.out.println("On Duty Time Available: "+myOnDutyTimeAvailable+" hours");
			//	System.out.println("Driving Time Available: "+myDrivingTimeAvailable+" hours");
			}
			else{
			//	System.out.println("Which is Less than 70.0, we do not schedule 34 hour rest of 70 Hour Rule yet!");
				schedule14HourRuleRest();
				}
			
			TransportLocationIfc location = Driver.this.getCurrentLocation();
			if(myIdleFlag){
				setIdleFlag(false);
				DispatcherFTL d=(DispatcherFTL)myDispatcher;
			//	System.out.println("t>"+getTime()+"  "+d.getId()+" is Reallocating Resouces after Driver "+Driver.this.getId()+" is On Duty at "+location.getName());
				d.allocateResourcesAfterTransportCompletedAt(location);
			}
		}
	}
	
	/**
	 * Checks the next event scheduled for the driver
	 *
	 */
	protected void checkNextEvent(){
		if(mySeventyHourRuleRestEvent!=null){
			if(mySeventyHourRuleRestEvent.isScheduled()){
				System.out.println("SEVENTY HOUR RULE REST IS SCHEDULED FOR DRIVER"+getId());}
		}
		if(myFourteenHourRuleRestEvent!=null){
			if(myFourteenHourRuleRestEvent.isScheduled()){
				System.out.println("FOURTEEN HOUR RULE REST IS SCHEDULED FOR DRIVER"+getId());}
		}
		if(myTenHourRestOfFourteenHourRuleEvent!=null){
			if(myTenHourRestOfFourteenHourRuleEvent.isScheduled()){
				System.out.println("TEN HOUR REST OF FOURTEEN HOUR RULE REST IS SCHEDULED FOR DRIVER"+getId());}
		}
		if(myTenHourRestWithTaskAttachedEvent!=null){
			if(myTenHourRestWithTaskAttachedEvent.isScheduled()){
				System.out.println("TEN HOUR REST WITH TASK ATTACHED IS SCHEDULED FOR DRIVER"+getId());}
		}
		if(myThirtyFourHourRestOfSeventyHourRuleEvent!=null){
			if(myThirtyFourHourRestOfSeventyHourRuleEvent.isScheduled()){
				System.out.println("THIRTY FOUR HOUR REST OF SEVENTY HOUR RULE IS SCHEDULED FOR DRIVER"+getId());}
		}
		if(myThirtyFourHourRestWithTaskAttachedEvent!=null){
			if(myThirtyFourHourRestWithTaskAttachedEvent.isScheduled()){
				System.out.println("THIRTY FOUR REST WITH TASK ATTACHED IS SCHEDULED FOR DRIVER"+getId());}
		}
		
	}
	
	/**
	 * Schedules unloading by notifying the location
	 * @param task
	 */
	protected void scheduleUnloading(DispatchTask task){
		
		double unloadingTime=getUnloadingTime();
	//	System.out.println("Unloding Time: "+unloadingTime);
	//	System.out.println("t>"+getTime()+" Task "+task.getId()+" is scheduled for Unloading at "+task.getDestination().getName());
		Location l= (Location)task.getDestination();
		l.unloading(task,unloadingTime);
	}
	
	/**
	 * Unloading is completed. The statistics are collected and the hours of service is updated
	 * The task is notified to complete that the transport task is completed
	 * @param task
	 */
	protected void unloadingCompleted(DispatchTask task){
		
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" Completes Unloading for Task "+task.getId()+" at "+task.getDestination().getName());
		double unloadingTime=getUnloadingTime();
		collectOnDutyTimeAccumulatedTowards70Hours(unloadingTime);
		collectOnDutyTimeAccumulatedTowards14Hours(unloadingTime);
	//	System.out.println("t>"+getTime()+" Actual on duty time accumulated towards 70 hours by driver "+getId()+" at the end of Transport Task "+task.getId()+" is "+myOnDutyTimeAccumulatedTowards70Hours);
	//	System.out.println("t>"+getTime()+" Actual driving time accumulated towards 70 hours by driver "+getId()+" at the end of Transport Task "+task.getId()+" is "+myDrivingTimeAccumulatedTowards70Hours);
		setDrivingTimeAvailable(myDrivingTimeAccumulatedTowards14Hours);
		setOnDutyTimeAvailable(myOnDutyTimeAccumulatedTowards14Hours);
		myLastEventTime=getTime();
		double taskEndTime=getTime();
		setTaskEndTime(taskEndTime);
	//	double t=(taskEndTime-myTaskBeginTime);
	//	System.out.println("Actual Time taken to complete the Loading, Transport and Unloading of Task "+task.getId()+" is "+t);
		double d= myNetwork.getLane(task.getLoadOrigin(),task.getLoadDestination()).getDistance();
		myDispatcher.myAverageTransportMiles.setValue(d);
		DispatcherFTL dispatcher = (DispatcherFTL)myDispatcher;
		dispatcher.myTransportMiles.collect(d);
		myDispatcher.myAvgNumTransportMiles.increment(d);
		System.out.println("myDispatcher.myAvgNumTransportMiles:"+myDispatcher.myAvgNumTransportMiles.getValue());
	//	System.out.println("t>"+getTime()+" Driver "+getId()+" completes Transport to Load of Task "+task.getId()+" at "+myCurrentLocation.getName());
		setNextTaskFlag(true);
		task.taskCompleted();
	}
	
	/**
	 * If the driver was supposed to take rest before unloading; the driver is scheduled for rest after unloading
	 *
	 */
	protected void scheduleRestAfterUnloading(){
		
		setIdleFlag(true);
	//	System.out.println("The total on duty time accumulated towards 70 hours by driver "+getId()+" so far is "+myOnDutyTimeAccumulatedTowards70Hours+" hours");
		double t=myOnDutyTimeAccumulatedTowards70Hours;
		
		if(t>=70.0){
			setTenHourRestFlag(false);
			setThirtyFourHourRestFlag(true);
		//	System.out.println("t> "+" Driver"+getId()+"will take rest of 34 hours now after unloading");
		}
		
		if(myTenHourRestFlag){
			setTenHourRestFlag(false);
		//	System.out.println("t> "+" Driver "+getId()+" will take rest of 10 hours now after unloading");
			schedule10HourRestOfFourteenHourRule();
		}
		
		if(myThirtyFourHourRestFlag){
			setThirtyFourHourRestFlag(false);
		//	System.out.println("t> "+" Driver"+getId()+"will take rest of 34 hours now after unloading");
			schedule34HourRestOfSeventyHourRule();
		}
	}
	
	/**
	 * Returns the loading time 
	 * @param task
	 * @return
	 */
	protected double getLoadingTime(DispatchTask task){
		Location l=(Location)task.getOrigin();
		double t=l.getLoadingTime(task);
		return (t);
	}
	
	/**
	 *  Returns the unloading time
	 * @param task
	 * @return
	 */
	protected double getUnloadingTime(DispatchTask task){
		Location l=(Location)task.getDestination();
		double ut=l.getUnloadingTime(task);
		return (ut);
	}
	
	/**
	 * Returns the time taken in travel from the origin to the destination
	 * @param task
	 * @return
	 */
	protected double getTransportTime(DispatchTask task) {
		double t=task.getTransportTime(this);
		return(t);
	}
	
	/**
	 * Resets driver's hours of service after it has taken 10 hour rest
	 *
	 */
	protected void resetDriverStatusAfter10HourRest(){
		myLastEventTime=getTime();
		resetDrivingTimeAvailable(myDrivingTimeLimit);
		resetOnDutyTimeAvailable(myOnDutyTimeLimit);
		resetOnDutyTimeAccumulatedTowards14Hours(0.0);
		resetDrivingTimeAccumulatedTowards14Hours(0.0);
	}
	
	/**
	 * Resets driver's hours of service after it has taken 34 hour rest
	 *
	 */
	protected void resetDriverStatusAfter34HourRest(){
		myLastEventTime=getTime();
		resetDrivingTimeAvailable(myDrivingTimeLimit);
		resetOnDutyTimeAvailable(myOnDutyTimeLimit);
		resetOnDutyTimeAccumulatedTowards14Hours(0.0);
		resetDrivingTimeAccumulatedTowards14Hours(0.0);
		resetOnDutyTimeAccumulatedTowards70Hours(0.0);
		resetDrivingTimeAccumulatedTowards70Hours(0.0);
	}
	
	/**
	 *  performs bobtail with no load involved
	 * @param task
	 */
	protected void performBobTailNoLoad(DispatchTask task){
		
	//	System.out.println("t>"+getTime()+"Driver"+task.getDriver().getId()+" starts performing BobTail no Load Task "+task.getId());
	//	System.out.println("The destination of the current BobTail no Load Task is: "+task.getDestination().getName());
	}
	
	/**
	 * performs deadhead with no load involved
	 * @param task
	 */
	protected void performDeadHeadNoLoad(DispatchTask task){
		
	//	System.out.println("t>"+getTime()+" Driver"+task.getDriver().getId()+" starts performing DeadHead no Load Task "+task.getId());
	//	System.out.println("The destination of the current DeadHead no Load Task is: "+task.getDestination().getName());
	}
	
	/**
	 * Updates driving time available
	 * @param drivingTimeLimit
	 */
	protected void resetDrivingTimeAvailable(double drivingTimeLimit){
		
		if(drivingTimeLimit<0.0)
			throw new IllegalArgumentException(" Driving Time limit was less than zero!");
		
		myDrivingTimeAvailable=drivingTimeLimit;
		//	System.out.println("t>"+getTime()+" Driving Time Available with the Driver "+getId()+" is "+myDrivingTimeAvailable);
	}
	
	/**
	 * Updates on duty time available
	 * @param onDutyTimeLimit
	 */
	protected void resetOnDutyTimeAvailable(double onDutyTimeLimit){
		
		if(onDutyTimeLimit<0.0)
			throw new IllegalArgumentException("On Duty Time limit was less than zero!");
		
		myOnDutyTimeAvailable=onDutyTimeLimit;
		
		//	System.out.println("t>"+getTime()+" On Duty Time Available with the Driver "+getId()+" is "+myOnDutyTimeAvailable);
	}
	
	/**
	 * Updates driving time accumulated towards cumulative on duty time limit(70 hours here)
	 * @param drivingTimeAccumulatedTowards70Hours
	 */
	protected void resetDrivingTimeAccumulatedTowards70Hours(double drivingTimeAccumulatedTowards70Hours){
		
		if(drivingTimeAccumulatedTowards70Hours<0.0)
			throw new IllegalArgumentException(" On Duty Time Accumulated Reset was less than zero!");
		
		myDrivingTimeAccumulatedTowards70Hours=drivingTimeAccumulatedTowards70Hours;
		//	System.out.println("t>"+getTime()+" Driving Time accumulated with the Driver "+getId()+" towards 70 hours is "+myDrivingTimeAccumulatedTowards70Hours);
	}
	
	/**
	 * Updates on duty time accumulated towards cumulative on duty time limit(70 hours here)
	 * @param onDutyTimeAccumulatedTowards70Hours
	 */
	protected void resetOnDutyTimeAccumulatedTowards70Hours(double onDutyTimeAccumulatedTowards70Hours){
		
		if(onDutyTimeAccumulatedTowards70Hours<0.0)
			throw new IllegalArgumentException(" On Duty Time Accumulated Reset was less than zero!");
		myOnDutyTimeAccumulatedTowards70Hours=onDutyTimeAccumulatedTowards70Hours;
		//	System.out.println("t>"+getTime()+" On Duty Time accumulated with the Driver "+getId()+" towards 70 hours is "+myOnDutyTimeAccumulatedTowards70Hours);
	}
	
	/**
	 * Updates driving time accumulated towards 14 hours on duty time limit
	 * @param drivingTimeAccumulatedTowards14Hours
	 */
	protected void resetDrivingTimeAccumulatedTowards14Hours(double drivingTimeAccumulatedTowards14Hours){
		
		if(drivingTimeAccumulatedTowards14Hours<0.0)
			throw new IllegalArgumentException(" On Duty Time Accumulated was less than zero!");
		myDrivingTimeAccumulatedTowards14Hours=drivingTimeAccumulatedTowards14Hours;
		//	System.out.println("t>"+getTime()+" Driving Time accumulated with the Driver "+getId()+" towards 14 hours is "+myDrivingTimeAccumulatedTowards14Hours);
	}
	
	/**
	 * Updates on duty time accumulated towards towards 14 hours on duty time limit
	 * @param onDutyTimeAccumulatedTowards14Hours
	 */
	protected void resetOnDutyTimeAccumulatedTowards14Hours(double onDutyTimeAccumulatedTowards14Hours){
		
		if(onDutyTimeAccumulatedTowards14Hours<0.0)
			throw new IllegalArgumentException(" On Duty Time Accumulated was less than zero!");
		myOnDutyTimeAccumulatedTowards14Hours=onDutyTimeAccumulatedTowards14Hours;
		//	System.out.println("t>"+getTime()+" On Duty Time accumulated with the Driver "+getId()+" towards 14 hours currently is "+myOnDutyTimeAccumulatedTowards14Hours);
	}
	
	/**
	 * Returns driving time available before it is scheduled for rest
	 * @return
	 */
	protected double getDrivingTimeAvailable() {
		
		if(myNoTaskFlag){
			double td=myDrivingTimeLimit;
			double to=getOnDutyTimeAvailable();
			if(to<td){
				td=to;
			}
			return td;
		}
		
		return myDrivingTimeAvailable;
	}
	
	/**
	 * Computes actual driving time available before it is scheduled for rest
	 * @param drivingTimeAccumulatedTowards14Hours
	 */
	protected void setDrivingTimeAvailable(double drivingTimeAccumulatedTowards14Hours) {
		
		if(drivingTimeAccumulatedTowards14Hours<0.0)
			throw new IllegalArgumentException("The driving time is less than zero!");
		
		myDrivingTimeAvailable = myDrivingTimeLimit-drivingTimeAccumulatedTowards14Hours;
		
		if(mySeventyHourRuleRestEvent!=null){
			if(mySeventyHourRuleRestEvent.isScheduled()){
				myDrivingTimeAvailable = myTimeRemainingTowards70Hours-drivingTimeAccumulatedTowards14Hours;}
		}
		
		if(myDrivingTimeAvailable<0.0)
			myDrivingTimeAvailable=0.0;
	}
	
	/**
	 * Computes expected driving time available before it is scheduled for rest
	 * @param drivingTimeAccumulatedTowards14Hours
	 */
	protected void setExpectedDrivingTimeAvailable(double drivingTimeAccumulatedTowards14Hours) {
		
		 if(drivingTimeAccumulatedTowards14Hours<0.0)
			throw new IllegalArgumentException("The driving time is less than zero!");
		
		myExpectedDrivingTimeAvailable = myDrivingTimeLimit-drivingTimeAccumulatedTowards14Hours;
		
		if(mySeventyHourRuleRestEvent!=null){
			if(mySeventyHourRuleRestEvent.isScheduled()){
				myExpectedDrivingTimeAvailable = myTimeRemainingTowards70Hours-drivingTimeAccumulatedTowards14Hours;}
		}
		
		if(myExpectedDrivingTimeAvailable<0.0)
			myExpectedDrivingTimeAvailable=0.0;
	}
	
	/**
	 * returns actual on duty time available
	 * @return
	 */
	protected double getOnDutyTimeAvailable() {
		
		if(myNoTaskFlag){
			double to=myOnDutyTimeLimit-(getTime()-myLastEventTime);
			if(to<0.0)
				to=0.0;
			return to;
		}
		
		return myOnDutyTimeAvailable;
	}
	
	/**
	 * Computes actual on duty time available
	 * @param onDutyTimeAccumulatedTowards14Hours
	 */
	protected void setOnDutyTimeAvailable(double onDutyTimeAccumulatedTowards14Hours) {
		
		if(onDutyTimeAccumulatedTowards14Hours<0.0)
			throw new IllegalArgumentException("The on duty time is less than zero!");
		
		myOnDutyTimeAvailable = myOnDutyTimeLimit-onDutyTimeAccumulatedTowards14Hours;
				
		if(myFourteenHourRuleRestEvent!=null){
			if(myFourteenHourRuleRestEvent.isScheduled()){
			//	System.out.println("Fourteen Hour rule is scheduled for the driver"+getId());
				}
		}
		
		if(mySeventyHourRuleRestEvent!=null){
			if(mySeventyHourRuleRestEvent.isScheduled()){
			//	System.out.println("I COME HERE ");
				myOnDutyTimeAvailable = myTimeRemainingTowards70Hours-onDutyTimeAccumulatedTowards14Hours;
			}
		}
		
		if(myOnDutyTimeAvailable<0.0)
			myOnDutyTimeAvailable=0.0;
		
	//	System.out.println("The On Duty Time available with the driver "+getId()+" is "+myOnDutyTimeAvailable+" hours");
		
		if(myOnDutyTimeAvailable<=myDrivingTimeAvailable){
			myDrivingTimeAvailable=myOnDutyTimeAvailable;
		//	System.out.println("The Driving Time available with the driver "+getId()+" is "+myDrivingTimeAvailable+" hours");
		}
		else {
			System.out.println("The Driving Time available with the driver "+getId()+" is "+myDrivingTimeAvailable+" hours");
		}
	}
	
	/**
	 * computes expected on duty time available
	 * @param onDutyTimeAccumulatedTowards14Hours
	 */
	protected void setExpectedOnDutyTimeAvailable(double onDutyTimeAccumulatedTowards14Hours) {
		
		if(onDutyTimeAccumulatedTowards14Hours<0.0)
			throw new IllegalArgumentException("The on duty time is less than zero!");
		
		myExpectedOnDutyTimeAvailable = myOnDutyTimeLimit-onDutyTimeAccumulatedTowards14Hours;
		
		if(mySeventyHourRuleRestEvent!=null){
			if(mySeventyHourRuleRestEvent.isScheduled()){
				myExpectedOnDutyTimeAvailable = myTimeRemainingTowards70Hours-onDutyTimeAccumulatedTowards14Hours;
			}
		}
		
		if(myExpectedOnDutyTimeAvailable<0.0)
			myExpectedOnDutyTimeAvailable=0.0;
		
		//	System.out.println("The Expected On Duty Time available with the driver "+getId()+" is "+myExpectedOnDutyTimeAvailable+" hours");
		
		if(myExpectedOnDutyTimeAvailable<=myExpectedDrivingTimeAvailable){
			myExpectedDrivingTimeAvailable=myExpectedOnDutyTimeAvailable;
			//	System.out.println("The Expected Driving Time available with the driver "+getId()+" is "+myExpectedDrivingTimeAvailable+" hours");
		}
		else {
			//	System.out.println("The Expected Driving Time available with the driver "+getId()+" is "+myExpectedDrivingTimeAvailable+" hours");
		}
	}
	
	/**
	 * 
	 * @return driving time limit
	 */
	protected double getDrivingTimeLimit() {
		return myDrivingTimeLimit;
	}
	
	/**
	 * 
	 * @param drivingTimeLimit
	 */
	public void setDrivingTimeLimit(double drivingTimeLimit) {
		if(drivingTimeLimit<0.0)
			throw new IllegalArgumentException("The driving time limit was set less than zero!");
		myDrivingTimeLimit = drivingTimeLimit;
	}
	
	/**
	 * 
	 * @param onDutyTimeLimit
	 */
	protected void setOnDutyTimeLimit(double onDutyTimeLimit) {
		if(onDutyTimeLimit<0.0)
			throw new IllegalArgumentException("The driving time limit was set less than zero!");
		myOnDutyTimeLimit =onDutyTimeLimit;
	}
	
	/**
	 * 
	 * @return on duty time limit
	 */
	protected double getOnDutyTimeLimit() {
		return myOnDutyTimeLimit;
	}
	
	/**
	 * computes on duty time accumulated towards 70 hour cycle
	 * @return
	 */
	protected double getOnDutyTimeAccumulatedTowards70Hours(){
		
		double a=getTime()- myLastEventTime;
		
		if(myNoTaskFlag){
			return(getTime()- myLastEventTime);
		}
		
		if(myNextTaskFlag){
			return(myOnDutyTimeAccumulatedTowards70Hours+a);
		}
		return myOnDutyTimeAccumulatedTowards70Hours;
	}
	
	/**
	 * collects on duty time accumulated towards 70 hour cycle
	 * @param time
	 */
	protected void collectOnDutyTimeAccumulatedTowards70Hours(double time) {
		if(time<0.0)
			throw new IllegalArgumentException("The on duty time accumulated was less than zero!");
		myOnDutyTimeAccumulatedTowards70Hours = myOnDutyTimeAccumulatedTowards70Hours+time;
		//System.out.println("The total on duty time accumulated towards 70 hours by driver "+getId()+" so far is "+myOnDutyTimeAccumulatedTowards70Hours+" hours");
	}
	
	/**
	 * 
	 * @return driving time accumulated towards 70 hour cycle
	 */
	protected double getDrivingTimeAccumulatedTowards70Hours() {
		return myDrivingTimeAccumulatedTowards70Hours;
	}
	
	/**
	 * Collecets driving time accumulated towards 70 hour cycle
	 * @param time
	 */
	protected void collectDrivingTimeAccumulatedTowards70Hours(double time) {
		if(time<0.0)
			throw new IllegalArgumentException("The driving time accumulated was less than zero!");
		myDrivingTimeAccumulatedTowards70Hours = myDrivingTimeAccumulatedTowards70Hours+time;
		//System.out.println("The total driving time accumulated towards 70 hours by driver "+getId()+" so far is "+myDrivingTimeAccumulatedTowards70Hours+" hours");
	}
	
	/**
	 * 
	 * @return on duty time accumulated towards 14 hour cycle
	 */
	protected double getOnDutyTimeAccumulatedTowards14Hours(){
		
		double a = getTime()- myLastEventTime;
		if(myNoTaskFlag){
			return(getTime()- myLastEventTime);
		}
		if(myNextTaskFlag){
			return(myOnDutyTimeAccumulatedTowards14Hours+a);
		}
			return myOnDutyTimeAccumulatedTowards14Hours;
	}
	
	/**
	 * collects on duty time accumulated towards 14 hour cycle
	 * @param time
	 */
	protected void collectOnDutyTimeAccumulatedTowards14Hours(double time) {
		if(time<0.0)
			throw new IllegalArgumentException("The on duty time accumulated was less than zero!");
		myOnDutyTimeAccumulatedTowards14Hours = myOnDutyTimeAccumulatedTowards14Hours+time;
		//	System.out.println("The total on duty time accumulated towards 14 hours by driver "+getId()+" so far is "+myOnDutyTimeAccumulatedTowards14Hours+" hours");
	}
	
	/**
	 * 
	 * @return driving time accumulated towards 14 hours
	 */
	protected double getDrivingTimeAccumulatedTowards14Hours() {
		return myDrivingTimeAccumulatedTowards14Hours;
	}
	
	/**
	 * Collects driving time accumulated towards 14 hour cycle
	 */
	 protected void collectDrivingTimeAccumulatedTowards14Hours(double time) {
		if(time<0.0)
			throw new IllegalArgumentException("The driving time accumulated was less than zero!");
		
		myDrivingTimeAccumulatedTowards14Hours = myDrivingTimeAccumulatedTowards14Hours+time;
		//	System.out.println("The total driving time accumulated towards 14 hours by driver "+getId()+" so far is "+myDrivingTimeAccumulatedTowards14Hours+" hours");
	}
	
	/**
	 * 
	 * @return travel time remaining for the driver
	 */
	 protected double getTransportTimeRemaining() {
		return myTransportTimeRemaining;
	}
	
	/**
	 * 
	 * @param transportTimeRemaining
	 */
	 protected void setTransportTimeRemaining(double transportTimeRemaining) {
		if(transportTimeRemaining<0.0)
			throw new IllegalArgumentException("The driving time accumulated was less than zero!");
		myTransportTimeRemaining = transportTimeRemaining;
		
		//	System.out.println("At t> "+getTime()+" Driver "+getId()+" has "+myTransportTimeRemaining+" hours of Transport for Task "+task.getId()+" left");
	}
	
	/**
	 * 
	 * @return drivers network
	 */
	 protected TransportationNetwork getNetwork() {
		return myNetwork;
	}
	
	/**
	 * Set driver's network and dispatcher
	 * @param network
	 */
	 protected void setNetwork(TransportationNetwork network) {
		if (network == null)
			throw new IllegalArgumentException("The network must not be null");
		
		myNetwork = network;
		setDispatcher(network);
	}
	
	
	protected void setDispatcher(TransportationNetwork network) {
		if (network == null)
			throw new IllegalArgumentException("The network must not be null");
		
		AbstractDispatcherFTL dispatcher = (AbstractDispatcherFTL)myNetwork.getDispatcher();
		
		if(dispatcher==null)
			throw new IllegalArgumentException("The dispatcher was null");
		
		myDispatcher = dispatcher;
		
	}
	
	/**
	 * Sets probability distribution for time duration before driver takes rest in a 14 hour cycle - here it is 14 hours
	 * @param d
	 */
	protected void setRestTimeDistributionForFourteenHourRule(DistributionIfc d) {
		
		if (d == null)
			throw new IllegalArgumentException("Rest Time Distribution was null!");
		
		myFourteenHourRuleRestTimeDistribution = d;
		
		if (myFourteenHourRuleRestTimeRV == null){// not made yet
			myFourteenHourRuleRestTimeRV = new RandomVariable(this, myFourteenHourRuleRestTimeDistribution);
		} else { // already had been made, and added to model
			// just change the distribution
			myFourteenHourRuleRestTimeRV.setInitialRandomSource(myFourteenHourRuleRestTimeDistribution);
		}
	}
	
	/**
	 * Sets probability distribution for rest time in short on duty cycle - here it is 10 hours
	 * @param d
	 */
	protected void setRestTimeDistributionForTenHourRule(DistributionIfc d) {
		
		if (d == null)
			throw new IllegalArgumentException("Rest Time Distribution was null!");
		
		myTenHourRuleRestTimeDistribution = d;
		
		if (myTenHourRuleRestTimeRV == null){// not made yet
			myTenHourRuleRestTimeRV = new RandomVariable(this, myTenHourRuleRestTimeDistribution);
		} else { // already had been made, and added to model
			// just change the distribution
			myTenHourRuleRestTimeRV.setInitialRandomSource(myTenHourRuleRestTimeDistribution);
		}
	}
	
	/**
	 * Sets probability distribution for time duration before driver takes rest in a 70 hour cycle - here it is 70 hours
	 * @param d
	 */
	protected void setRestTimeDistributionForSeventyHourRule(DistributionIfc d) {
		
		if (d == null)
			throw new IllegalArgumentException("Rest Time Distribution was null!");
		
		mySeventyHourRuleRestTimeDistribution = d;
		
		if (mySeventyHourRuleRestTimeRV == null){// not made yet
			mySeventyHourRuleRestTimeRV = new RandomVariable(this, mySeventyHourRuleRestTimeDistribution);
		} else { // already had been made, and added to model
			// just change the distribution
			mySeventyHourRuleRestTimeRV.setInitialRandomSource(mySeventyHourRuleRestTimeDistribution);
		}
	}
	
	/**
	 * Sets probability distribution for rest time in cumulative on duty cycle - here it is 34 hours
	 * @param d
	 */
	protected void setRestTimeDistributionForThirtyFourHourRule(DistributionIfc d) {
		
		if (d == null)
			throw new IllegalArgumentException("Rest Time Distribution was null!");
		
		myThirtyFourHourRuleRestTimeDistribution = d;
		
		if (myThirtyFourHourRuleRestTimeRV == null){// not made yet
			myThirtyFourHourRuleRestTimeRV = new RandomVariable(this, myThirtyFourHourRuleRestTimeDistribution);
		} else { // already had been made, and added to model
			// just change the distribution
			myThirtyFourHourRuleRestTimeRV.setInitialRandomSource(myThirtyFourHourRuleRestTimeDistribution);
		}
	}
	
	/**
	 * Sets velocity distribution of driver
	 * @param d
	 */
	protected void setVelocityDistribution(DistributionIfc d) {
		
		if (d == null)
			throw new IllegalArgumentException("Rest Time Distribution was null!");
		
		myVelocityDistribution = d;
		
		if (myVelocityRV == null){// not made yet
			myVelocityRV = new RandomVariable(this, myVelocityDistribution);
		} else { // already had been made, and added to model
			// just change the distribution
			myVelocityRV.setInitialRandomSource(myVelocityDistribution);
		}
	}
	
	/**
	 * 
	 * @param task
	 * @return expected loading time at location
	 */
	protected double getExpectedLoadingTime(DispatchTask task){
		Location l=(Location)task.getOrigin();
		double t=l.getExpectedLoadingTime(task);
		return t;
	}
	
	/**
	 * 
	 * @param task
	 * @return expected unloading time at location
	 */
	protected double getExpectedUnloadingTime(DispatchTask task){
		Location l=(Location)task.getOrigin();
		double t=l.getExpectedUnloadingTime(task);
		return t;
	}
	
	/**
	 * 
	 * @param task
	 * @return expected travel time between origin and destination
	 */
	protected double getExpectedTransportTime(DispatchTask task) {
		myExpectedTransportTime = task.getExpectedTransportTime(this);
		return myExpectedTransportTime;
	}
	
	/**
	 * 
	 * @param nextTask
	 * @param location
	 * @return expected total task time depending on the type of task give to the driver - tranport, bobtail with load,
	 * deadhead with load, bobtail no load, deadhead no load
	 */
	protected double getExpectedTotalTaskTime(DispatchTask nextTask,TransportLocationIfc location) {
		
		myExpectedDrivingTimeAvailable = getDrivingTimeAvailable();
		myExpectedOnDutyTimeAvailable = getOnDutyTimeAvailable();
		myExpectedOnDutyTimeAccumulatedTowards70Hours = getOnDutyTimeAccumulatedTowards70Hours();
		myExpectedOnDutyTimeAccumulatedTowards14Hours = getOnDutyTimeAccumulatedTowards14Hours();
		
		if(myNextTaskFlag){
			setExpectedDrivingTimeAvailable(myDrivingTimeAccumulatedTowards14Hours);
			setExpectedOnDutyTimeAvailable(myExpectedOnDutyTimeAccumulatedTowards14Hours);
		}
		
		//	System.out.println("t> "+getTime()+" Expected On Duty Time Available with driver "+getId()+" is "+myExpectedOnDutyTimeAvailable);
		//	System.out.println("t> "+getTime()+" Expected Driving Time Available with driver "+getId()+" is "+myExpectedDrivingTimeAvailable);
		
		if (nextTask.isTransport()){
			myExpectedTotalTaskTime=getExpectedTaskTimeForTransportTask(nextTask);
			//		System.out.println("t>"+getTime()+" Driver "+getId()+" is Expected to take  "+myExpectedTotalTaskTime+" hrs to complete the Transport Task "+ task.getId());
			//		System.out.println("Expected Time Required by the Driver"+getId()+" for the Completion of the Total Task "+task.getId()+" is: "+myExpectedTotalTaskTime);
		}
		
		else if (nextTask.isDeadHeadToLoad()){
			double td=getExpectedTaskTimeForDeadHeadTask(nextTask,location);
			//	System.out.println("t>"+getTime()+" Driver "+getId()+" is expected to take  "+td+" hrs to complete the DeadHead part of Task "+ task.getId());
			//	System.out.println("The destination of the current DeadHeadToLoad Task is: "+task.getDestination().getName());
			double tt=getExpectedTaskTimeForTransportTask(nextTask);
			nextTask.setOrigin(location);
			nextTask.setDestination(nextTask.getLoadOrigin());
			myExpectedTotalTaskTime=td+tt;
			//	System.out.println("Expected Time Required by the Driver"+getId()+" for the Completion of the Total Task "+task.getId()+" is: "+myExpectedTotalTaskTime);
		}
		
		else if (nextTask.isBobTailToLoad()){
			nextTask.setOrigin(location);
			double tb=getExpectedTaskTimeForBobTailTask(nextTask,location);
			//	System.out.println("t>"+getTime()+" Driver "+getId()+" is expected to take  "+tb+" hrs to complete the BobTail part of Task "+ task.getId());
			//	System.out.println("The destination of the current BobTail Task is: "+task.getDestination().getName());
			double tt=getExpectedTaskTimeForTransportTask(nextTask);
			nextTask.setOrigin(location);
			nextTask.setDestination(nextTask.getLoadOrigin());
			myExpectedTotalTaskTime=tb+tt;
			//	System.out.println("Expected Time Required by the Driver"+getId()+" for the Completion of the Total Task "+task.getId()+" is: "+myExpectedTotalTaskTime);
		}
		
		else if (nextTask.isBobTailNoLoad()){
			myExpectedTotalTaskTime=getExpectedTaskTimeForBobTailTask(nextTask,location);
			//	System.out.println("Expected Time Required by the Driver"+getId()+" for the Completion of the Total Task "+task.getId()+" is: "+myExpectedTotalTaskTime);
		}
		
		else if (nextTask.isDeadHeadNoLoad()){
			myExpectedTotalTaskTime=getExpectedTaskTimeForDeadHeadTask(nextTask,location);
			//	System.out.println("Expected Time Required by the Driver"+getId()+" for the Completion of the Total Task "+task.getId()+" is: "+myExpectedTotalTaskTime);
		}
		
		return myExpectedTotalTaskTime;
	}
	
	/**
	 * 
	 * @param task
	 * @param location
	 * @return expected task time for bobtail with load task
	 */
	protected double getExpectedTaskTimeForBobTailTask(DispatchTask task,TransportLocationIfc location){
		
		task.setOrigin(location);
		
		double t=getExpectedTransportTime(task);
		
		if((myExpectedOnDutyTimeAccumulatedTowards70Hours+getExpectedTransportTime(task))>70.0){
			
			myExpectedBobTailTaskTime=t+34;
			double d=(t-myExpectedDrivingTimeAvailable);
			myExpectedDrivingTimeAvailable=(myDrivingTimeLimit-d);
			myExpectedOnDutyTimeAvailable=(myOnDutyTimeLimit-d);
			myExpectedOnDutyTimeAccumulatedTowards70Hours=d;
			return myExpectedBobTailTaskTime;
		}
		
		if(t>myExpectedDrivingTimeAvailable){
			double n=(double)Math.round(t/10);
			t=t+(10*n);
			myExpectedBobTailTaskTime = t;
			double d=(t-myExpectedDrivingTimeAvailable-(10*n));
			myExpectedDrivingTimeAvailable=(myDrivingTimeLimit-d);
			myExpectedOnDutyTimeAvailable=(myOnDutyTimeLimit-d);
			myExpectedOnDutyTimeAccumulatedTowards70Hours=d;			
		}
		else{
			myExpectedBobTailTaskTime=t;
			double d=(myExpectedDrivingTimeAvailable-t);
			myExpectedDrivingTimeAvailable=d;
			myExpectedOnDutyTimeAvailable=d;
			myExpectedOnDutyTimeAccumulatedTowards70Hours=myExpectedOnDutyTimeAccumulatedTowards70Hours+t;
		}
		
		return myExpectedBobTailTaskTime;
	}
	
	/**
	 * 
	 * @param task
	 * @param location
	 * @return expected task time for deadhead with load task
	 */
	protected double getExpectedTaskTimeForDeadHeadTask(DispatchTask task, TransportLocationIfc location){
		
		task.setOrigin(location);
		double t=getExpectedTransportTime(task);
		
		if((myExpectedOnDutyTimeAccumulatedTowards70Hours+getExpectedTransportTime(task))>70.0){
			
			myExpectedDeadHeadTaskTime=t+34;
			double d=(t-myExpectedDrivingTimeAvailable);
			myExpectedDrivingTimeAvailable=(myDrivingTimeLimit-d);
			myExpectedOnDutyTimeAvailable=(myOnDutyTimeLimit-d);
			myExpectedOnDutyTimeAccumulatedTowards70Hours=d;
			
			return myExpectedDeadHeadTaskTime;
		}
		
		if(t>myExpectedDrivingTimeAvailable){
			double n=(double)Math.round(t/10);
			t=t+(10*n);
			myExpectedDeadHeadTaskTime = t;
			double d=(t-myExpectedDrivingTimeAvailable-(10*n));
			myExpectedDrivingTimeAvailable=(myDrivingTimeLimit-d);
			myExpectedOnDutyTimeAvailable=(myOnDutyTimeLimit-d);
			myExpectedOnDutyTimeAccumulatedTowards70Hours=d;			
		}
		else{
			myExpectedDeadHeadTaskTime=t;
			double d=(myExpectedDrivingTimeAvailable-t);
			myExpectedDrivingTimeAvailable=d;
			myExpectedOnDutyTimeAvailable=d;
			myExpectedOnDutyTimeAccumulatedTowards70Hours=myExpectedOnDutyTimeAccumulatedTowards70Hours+t;
		}
			return myExpectedDeadHeadTaskTime;
	}
	
	/**
	 * returns expected task time for transport task - loading, unloading and transport
	 * @param task
	 * @return
	 */
	protected double getExpectedTaskTimeForTransportTask(DispatchTask task){
		
		task.setOrigin(task.getLoadOrigin());
		task.setDestination(task.getLoadDestination());
		
		double l= getExpectedLoadingTime(task);
		double ul= getExpectedUnloadingTime(task);
		double t= getExpectedTransportTime(task);
		
		if((myExpectedOnDutyTimeAccumulatedTowards70Hours+getExpectedLoadingTime(task))>70.0){
			l=l+34;
			//     myExpectedOnDutyTimeAvailable=myOnDutyTimeLimit-getExpectedLoadingTime(task);
			myExpectedOnDutyTimeAvailable=myOnDutyTimeLimit;
			myExpectedDrivingTimeAvailable=myDrivingTimeLimit;
			myExpectedOnDutyTimeAccumulatedTowards70Hours=getExpectedLoadingTime(task);
		}else{		
		if(getExpectedLoadingTime(task)>myExpectedOnDutyTimeAvailable){
			l=l+10;
			//		myExpectedOnDutyTimeAvailable=myOnDutyTimeLimit-getExpectedLoadingTime(task);
			myExpectedOnDutyTimeAvailable=myOnDutyTimeLimit;
			myExpectedDrivingTimeAvailable=myDrivingTimeLimit;
			myExpectedOnDutyTimeAccumulatedTowards70Hours=getExpectedLoadingTime(task);
		}
		}
		
		if((myExpectedOnDutyTimeAccumulatedTowards70Hours+getExpectedTransportTime(task))>70.0){
			t=t+34;
			myExpectedOnDutyTimeAvailable=myOnDutyTimeLimit;
			myExpectedDrivingTimeAvailable=myDrivingTimeLimit;
		}else{		
		if(getExpectedTransportTime(task)>myExpectedDrivingTimeAvailable){
			double n=(double)Math.round(t/10);
			t=t+(10*n);
		}
		}
		
		myExpectedTransportTaskTime=(l+ul+t);
		
//		System.out.println("Expected Loading Time: "+l+" by Driver "+getId());
//		System.out.println("Expected Unloading Time: "+ul+" by Driver "+getId());
//		System.out.println("Expected Tranporting Time: "+t+" by Driver "+getId());
//		System.out.println("Expected Transport Time: "+myExpectedTransportTaskTime+" by Driver "+getId());
		
		return myExpectedTransportTaskTime;
	}
	
	/**
	 * Same as getExpectedTaskTimeForTransportTask() with variables used instead of actual values 14 and 34 on duty limit
	 * @param task
	 * @return
	 */
	protected double getExpectedTaskTimeForTransportTask1(DispatchTask task){
		task.setOrigin(task.getLoadOrigin());
		task.setDestination(task.getLoadDestination());
		double l= getExpectedLoadingTime(task);
		double ul= getExpectedUnloadingTime(task);
		double t= getExpectedTransportTime(task);
		if((myTimeAccumulatedTowardsCumulativeOnDutyCycle+
				getExpectedLoadingTime(task))> myCumulativeOnDutyLimit){
			l=l+34;
		}else{
			if(getExpectedLoadingTime(task)>myOnDutyTimeAvailable){
				l=l+10;
			}
		}
		if((myTimeAccumulatedTowardsCumulativeOnDutyCycle+
				getExpectedTransportTime(task))> myCumulativeOnDutyLimit){
				t=t+34;
		}else{
			if(getExpectedTransportTime(task)>myDrivingTimeAvailable){
				double n=(double)Math.round(t/10);
				t=t+(10*n);
			}
		}
		myExpectedTransportTaskTime=(l+ul+t);
		return (myExpectedTransportTaskTime);
	}
	
	protected double getRandomVelocity(){
		double v = (double)myVelocityRV.getValue();
		return v;
	}
	
	public DistributionFunctionIfc getVelocityDistribution(){
		return myVelocityDistribution;
	}
	
	protected TransportLocationIfc getCurrentLocation() {
		return myCurrentLocation;
	}
	
	protected void setCurrentLocation(TransportLocationIfc currentLocation) {
		if(currentLocation==null)
			throw new IllegalArgumentException("Current Location was null!");
		myCurrentLocation = currentLocation;
	//	System.out.println("Driver "+getId()+" is now at location: "+myCurrentLocation.getName());
	}
	
	/**
	 * 
	 * @return initial location of driver
	 */
	protected TransportLocationIfc getInitialLocation() {
		return myInitialLocation;
	}
	
	protected void setInitialLocation(TransportLocationIfc initialLocation) {
		if(initialLocation==null)
			throw new IllegalArgumentException("Initial Location was null!");
		myInitialLocation = initialLocation;
	}
	
	/**
	 * 
	 * @return current coordinates of driver
	 */
	protected CoordinateIfc getCurrentCoordinate() {
		return myCurrentCoordinate;
	}
	
	protected void setCurrentCoordinate(TransportLocationIfc location) {
		if(location==null)
			throw new IllegalArgumentException("Location was null!");
		myCurrentCoordinate = location.getCurrentPosition();
	}
	
	/**
	 * 
	 * @return initial coordinates of driver
	 */
	protected CoordinateIfc getInitialCoordinate() {
		return myInitialCoordinate;
	}
	
	protected void setInitialCoordinate(TransportLocationIfc location) {
		if(location==null)
			throw new IllegalArgumentException("Location was null!");
		myInitialCoordinate=location.getInitialPosition();	
	}
	
	protected boolean isLoadingFlag() {
		return myLoadingFlag;
	}
	
	protected void setLoadingFlag(boolean loadingFlag) {
		myLoadingFlag = loadingFlag;
	}
	
	protected boolean isUnloadingFlag() {
		return myUnloadingFlag;
	}
	
	protected void setUnloadingFlag(boolean unloadingFlag) {
		myUnloadingFlag = unloadingFlag;
	}
	
	protected boolean isIncompleteTransportFlag() {
		return myIncompleteTransportFlag;
	}
	
	protected void setIncompleteTransportFlag(boolean incompleteTransportFlag) {
		myIncompleteTransportFlag = incompleteTransportFlag;
	}
	
	protected boolean isTaskAttachedWith10HourRestFlag() {
		return myTaskAttachedWith10HourRestFlag;
	}
	
	protected void setTaskAttachedWith10HourRestFlag(boolean taskAttachedWith10HourRestFlag) {
		myTaskAttachedWith10HourRestFlag = taskAttachedWith10HourRestFlag;
	}
	
	protected boolean isTaskAttachedWith34HourRestFlag() {
		return myTaskAttachedWith34HourRestFlag;
	}
	
	protected void setTaskAttachedWith34HourRestFlag(
			boolean taskAttachedWith34HourRestFlag) {
		myTaskAttachedWith34HourRestFlag = taskAttachedWith34HourRestFlag;
	}
	
	protected boolean isNoTaskFlag() {
		return myNoTaskFlag;
	}
	
	protected void setNoTaskFlag(boolean noTaskFlag) {
		myNoTaskFlag = noTaskFlag;
	}
	
	protected boolean isTenHourRestFlag() {
		return myTenHourRestFlag;
	}
	
	protected void setTenHourRestFlag(boolean tenHourRestFlag) {
		myTenHourRestFlag =tenHourRestFlag;
	}
	
	protected boolean isThirtyFourHourRestFlag(){
		return myThirtyFourHourRestFlag;
	}
	
	protected void setThirtyFourHourRestFlag(boolean thirtyFourHourRestFlag) {
		myThirtyFourHourRestFlag = thirtyFourHourRestFlag;
	}
	
	protected boolean isIdleFlag() {
		return myIdleFlag;
	}
	
	protected void setIdleFlag(boolean idleFlag) {
		myIdleFlag = idleFlag;
	}
	
	/**
	 * @return current task of the driver
	 */
	protected DispatchTask getCurrentTask() {
		return myCurrentTask;
	}
	
	/**
	 * Sets the current task of the driver
	 * @param currentTask
	 */
	protected void setCurrentTask(DispatchTask currentTask) {
		myCurrentTask = currentTask;              
	}
	
	protected boolean isNextTaskFlag() {
		return myNextTaskFlag;
	}
	
	protected void setNextTaskFlag(boolean nextTaskFlag) {
		myNextTaskFlag = nextTaskFlag;
	}
	
	protected void afterReplication(){
		setCurrentTask(null);
	}
	
	protected double getLoadingTime() {
		return myLoadingTime;
	}
	
	protected void setLoadingTime(double loadingTime) {
		myLoadingTime = loadingTime;
	}
	
	protected double getUnloadingTime() {
		return myUnloadingTime;
	}
	
	protected void setUnloadingTime(double unloadingTime) {
		myUnloadingTime = unloadingTime;
	}
	
	protected double getTransportTime() {
		return myTransportTime;
	}
	
	protected void setTransportTime(double transportTime) {
		myTransportTime = transportTime;
	}
	
	protected double getTaskBeginTime() {
		return myTaskBeginTime;
	}
	
	protected void setTaskBeginTime(double taskBeginTime) {
		myTaskBeginTime = taskBeginTime;
	}
	
	protected double getTaskEndTime() {
		return myTaskEndTime;
	}
	
	protected void setTaskEndTime(double taskEndTime) {
		myTaskEndTime = taskEndTime;
	}

	protected double getCumulativeOnDutyTimeLimit() {
		return myCumulativeOnDutyTimeLimit;
	}

	protected void setCumulativeOnDutyTimeLimit(double cumulativeOnDutyTimeLimit) {
		myCumulativeOnDutyTimeLimit = cumulativeOnDutyTimeLimit;
	}
	
}
