package jslx.transportlayerFinal;

/** Lane class is for creating lane given origin and destination. The given origin and destination must be 
 *  different otherwise an exception will be thrown 
 *
 */
public class Lane implements LaneIfc {
	
	//  =================================
	// 		 Attributes of a Lane
	//  =================================

	/** The transportation network for the lane
	 */
	private NetworkIfc myNetwork;
	
	/** Represents origin of the lane 
	 */
	private TransportLocationIfc myOrigin;
	
	/** Represents destination of the lane 
	 */
	private TransportLocationIfc myDestination;
	
	/** Represents the distance between the origin and destination
	 */
	private double myDistance;

	/** Creates Lane with the given origin and destination as parameters
	 * 
	 * @param origin
	 * @param destination
	 */
	public Lane(NetworkIfc network, TransportLocationIfc origin, TransportLocationIfc destination) {
		this(network, origin, destination, 1.0);
	}
	
	/** Creates Lane with the given origin and destination as parameters
	 * 
	 * @param origin
	 * @param destination
	 */
	public Lane(NetworkIfc network, TransportLocationIfc origin, TransportLocationIfc destination, double distance) {
		setNetwork(network);
		setOriginAndDestination(origin, destination);
		setDistance(distance);
	}
	
	/** Returns destination of the lane
	 * 
	 * @return destination
	 */
	public final TransportLocationIfc getDestination() {
		return myDestination;
	}

	/** Returns origin of the lane
	 * 
	 * @return origin
	 */
	public final TransportLocationIfc getOrigin() {
		return myOrigin;
	}
	
	/** Returns distance of the lane
	 * 
	 * @return double
	 */
	public final double getDistance(){
		return(myDistance);
	}

	/** Returns the lane's transportation network
	 * 
	 * @return
	 */
	public final NetworkIfc getNetwork(){
		return(myNetwork);
	}

	public String toString(){
		String s = "From: " + myOrigin.getName() + " To: " + myDestination.getName() + " Distance = " + getDistance();
		return(s);
	}
	
	//  ===========================================
	//  PROTECTED METHODS
	//  ===========================================

	/** Ensures that the distance is set between the origin and the destination
	 *
	 */
	protected void setDistance(double d){
		myDistance = d;
	}

	/** Sets origin and destination of the lane and attached the lane
	 *  as a position observer on them.  Ensures that the distance between
	 *  the origin and destination is initialized.
	 *  
	 * @param origin The origin of the lane
	 * @param destination The destination of the lane
	 */
	protected void setOriginAndDestination(TransportLocationIfc origin, TransportLocationIfc destination) {
		if (destination == null)
			throw new IllegalArgumentException("The Location destination must not be null");
		if (origin == null)
			throw new IllegalArgumentException("The Location origin must not be null");
		if (destination == origin)
			throw new IllegalArgumentException("The origin and destination must not be the same.");
		myDestination = destination;
		myOrigin = origin;
	}	
	
	/** Sets the network for the lane
	 * 
	 * @param network
	 */
	protected final void setNetwork(NetworkIfc network){
		if (network == null)
			throw new IllegalArgumentException("The TransportationNetwork must not be null");
		myNetwork = network;
	}

}
