package jslx.transportlayerFinal;
	
import jsl.modeling.ActionListenerIfc;
import jsl.modeling.JSLEvent;
import jsl.modeling.ModelElement;
import jsl.modeling.elements.spatial.CoordinateIfc;
import jsl.modeling.elements.spatial.SpatialElementIfc;
import jsl.modeling.elements.spatial.SpatialModel;
import jsl.modeling.elements.spatial.SpatialModelElement;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.utilities.random.distributions.DistributionFunctionIfc;
import jsl.utilities.random.distributions.DistributionIfc;
	
/** Location extends SpatialModelElement and implements TransportLocationIfc
 *  Network consists of many Locations. The Network requires atleast two locations 
 *  to build the network.Location is responsible for setting the loading and unloading 
 *  distribuwtions for the transport task. It also schedules loading 
 *  and unloading by notifying loading and unloading listeners
 */
public class Location extends SpatialModelElement implements TransportLocationIfc {

	//  =================================
	// 		 Attributes of a Lane
	//  =================================
	
		/** Probability distribution for loading times
		 */
		protected DistributionIfc myLoadingTimeDistribution;
		
		/** Probability distribution for unloading times 
		 */
		protected DistributionIfc myUnloadingTimeDistribution;
		
		/** A Random Variable for Loading 
		 */ 
		protected RandomVariable  myLoadingTimeRV;
		
		/** A Random Variable for Unloading 
		 */
		protected RandomVariable  myUnloadingTimeRV;
		
		/** Listener for scheduling loading at the location 
		 */
		private LoadingListener   myLoadingListener;
		
		/** Listener for scheduling unloading at the location 
		 */
		private UnloadingListener myUnloadingListener;
		
		/**
		 *  JSL Event representing loading
		 */
		protected JSLEvent myLoadingEvent;
		
		/**
		 *  JSL Event representing unloading
		 */
		protected JSLEvent myUnloadingEvent;
		
		//  ===========================================
		//  CONSTRUCTORS
		//  ===========================================
		
		/** Creates Location
		 * @param parent
		 */
		public Location(ModelElement parent) {
			this(parent,null,null,null);
		}
		
		/** Creates Location with supplied parent and name
		 * 
		 * @param parent
		 * @param name
		 */
		public Location(ModelElement parent, String name) {
			this(parent,name,null,null);
			
		}
		
		/** Creates Location with supplied parent and coordinates
		 * 
		 * @param parent
		 * @param position
		 */
		public Location(ModelElement parent, CoordinateIfc position) {
			this(parent, null,null,position);
			
		}
		
		/** Creates Location with supplied parent and position
		 * 
		 * @param parent
		 * @param position
		 */
		public Location(ModelElement parent, SpatialElementIfc position) {
			this(parent,null,null, position.getCurrentPosition());
			
		}
		
		/** Creates Location with supplied parent, name, and spatial model 
		 * 
		 * @param parent
		 * @param name
		 * @param spatialModel
		 */
		public Location(ModelElement parent, String name, SpatialModel spatialModel) {
			this(parent, name, spatialModel,null);
			
		}
		
		/** Creates Location with supplied parent and spatial model
		 * 
		 * @param parent
		 * @param spatialModel
		 */
		public Location(ModelElement parent, SpatialModel spatialModel) {
			this(parent,null, spatialModel,null);
			
		}
		
		/** Creates Location with supplied parent, spatial model, and spatial element
		 * 
		 * @param parent
		 * @param spatialModel
		 * @param element
		 */
		public Location(ModelElement parent, SpatialModel spatialModel,	SpatialElementIfc element) {
			this(parent, null,spatialModel, element.getCurrentPosition());
			
		}
		
		/** Creates Location with supplied parent, name, spatial model, and coordinates
		 *  Creates the loading and unloading listeners
		 * @param parent
		 * @param name
		 * @param spatialModel
		 * @param coordinate
		 */
		public Location(ModelElement parent, String name,SpatialModel spatialModel, CoordinateIfc coordinate) {
			super(parent, name, spatialModel, coordinate);
			
			myLoadingListener = new LoadingListener();
			myUnloadingListener = new UnloadingListener();
		}
		
		/**
		 * This method is redundant
		 */
		protected void initialize(){
			super.initialize();
			myLoadingEvent=null;
			myUnloadingEvent=null;
		}
			
		//TODO public methods should be protected
		//  ===========================================
		//  PUBLIC METHODS
		//  ===========================================
		
		/** Receives transport task
		 * 
		 */
		public void receiveTransportTask(DispatchTask task) {
			
		}
		
		/** Sets loading time distribution at that location 
		 *  Requires the distribution to be passed as parameter
		 *  @param d
		 */
		public void setLoadingTimeDistribution(DistributionIfc d) {
			
			if (d == null)
				throw new IllegalArgumentException("Loading Time Distribution was null!");
			
			myLoadingTimeDistribution = d;
			
			if (myLoadingTimeRV == null){// not made yet
				myLoadingTimeRV = new RandomVariable(this, myLoadingTimeDistribution);
			} else { // already had been made, and added to model
				// just change the distribution
				myLoadingTimeRV.setInitialRandomSource(myLoadingTimeDistribution);
			}
		}
		
		/** Sets unloading time distribution at that location 
		 *  Requires the distribution to be passed as parameter
		 * @param d
		 */
		public void setUnloadingTimeDistribution(DistributionIfc d) {
			
			if (d == null)
				throw new IllegalArgumentException("Loading Time Distribution was null!");
			
			myUnloadingTimeDistribution = d;
			
			if (myUnloadingTimeRV == null){// not made yet
				myUnloadingTimeRV = new RandomVariable(this, myUnloadingTimeDistribution);
			} else { // already had been made, and added to model
				// just change the distribution
				myUnloadingTimeRV.setInitialRandomSource(myUnloadingTimeDistribution);
			}
		}
		
		/** returns probability distribution of loading at this location
		 * @return
		 */
		public DistributionFunctionIfc getLoadingTimeDistribution() {
			return(myLoadingTimeDistribution);
		}
		
		/** returns probability distribution of unloading at this location 
		 * @return
		 */
		public DistributionFunctionIfc getUnloadingTimeDistribution() {
			return(myUnloadingTimeDistribution);
		}
		
		/** returns random variable of loading distribution
		 * 
		 * @return
		 */
		public RandomVariable getLoadingTimeRV(){
			return(myLoadingTimeRV);
		}
		
		/** returns random variable of unloading distribution
		 * 
		 * @return
		 */
		public RandomVariable getUnloadingTimeRV(){
			return(myUnloadingTimeRV);
		}
		
		/** schedules loading by generating random value of myLoadingTimeRV
		 * 
		 * @param task
		 */
		protected void loading(DispatchTask task){
			double t = myLoadingTimeRV.getValue();
			scheduleEvent(myLoadingListener, t, "Loading", task); 
		}
		
		// TODO this method is called by the driver  
		/** schedules loading for the supplied loading time for a task
		 * 
		 * @param task
		 * @param loadingTime
		 */
		protected void loading(DispatchTask task,double loadingTime){
		//	System.out.println("t> "+getTime()+" Task in loading: "+task.getId()+" for driver "+task.getDriver());
			double t = loadingTime;
			myUnloadingEvent=scheduleEvent(myLoadingListener, t, "Loading", task); 
		}
		
		/** Loading listener class 
		 *  It send driver the message that the loading is completed
		 * 
		 * @author snangia
		 *
		 */
		class LoadingListener implements ActionListenerIfc {
			public void action(JSLEvent event) {
				DispatchTask task = (DispatchTask)event.getMessage(); 
		//		System.out.println("t> "+getTime()+" Loading Listener of Task: "+task.getId()+" for driver "+task.getDriver());
				task.getDriver().loadingCompleted(task);
			}
		}
		
		/** schedules unloading by generating random value of myUnloadingTimeRV
		 * 
		 * @param task
		 */
		protected void unloading(DispatchTask task){
			double t = myUnloadingTimeRV.getValue();
			scheduleEvent(myUnloadingListener, t, "Unloading",task);			
		}
		
		/** schedules unloading for the supplied unloading time for a task
		 * 
		 * @param task
		 * @param unloadingTime
		 */
		protected void unloading(DispatchTask task,double unloadingTime){
			double t = unloadingTime;
			myUnloadingEvent=scheduleEvent(myUnloadingListener, t, "Loading", task); 
		}
		
		/** Unloading listener class 
		 *  It send driver the message that the unloading is completed
		 * 
		 * @author snangia
		 *
		 */
		class UnloadingListener implements ActionListenerIfc {
			public void action(JSLEvent event) {
				DispatchTask task = (DispatchTask)event.getMessage();  
				task.getDriver().unloadingCompleted(task);
			}
		}
		
		/** returns random loading time
		 * 
		 * @param task
		 * @return
		 */
		public double getLoadingTime(DispatchTask task){
			double t = myLoadingTimeRV.getValue();
			return t;
		}
		
		/** returns random unloading time 
		 * 
		 * @param task
		 * @return
		 */
		public double getUnloadingTime(DispatchTask task){
			double t = myUnloadingTimeRV.getValue();
			return t;
		}
		
		/** returns the mean loading time for the task
		 * 
		 * @param task
		 * @return
		 */
		public double getExpectedLoadingTime(DispatchTask task){
			double t = myLoadingTimeDistribution.getMean();
			return t;
		}
		
		/** returns the mean unloading time for the task
		 * 
		 * @param task
		 * @return
		 */
		public double getExpectedUnloadingTime(DispatchTask task){
			double t = myUnloadingTimeDistribution.getMean();
			return t;
		}
		
		/**
		 * This method is redundant
		 * 
		 */
		protected void afterReplication(){
		
			if(myLoadingEvent!=null){
			if(myLoadingEvent.isScheduled())
				cancelEvent(myLoadingEvent);
			}
			if(myUnloadingEvent!=null){
			if(myUnloadingEvent.isScheduled())
				cancelEvent(myUnloadingEvent);
			}
			myLoadingEvent=null;
			myUnloadingEvent=null;
		}
		
	}
