/*
 * Created on Jul 8, 2006
 *
 *	@author rossetti
 */
package jslx.transportlayerFinal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;

import jslx.transportlayer.graph.AdjMatrixGraph;


/**
 * The Network class represents the locations within
 * a network that can serve as origins and/or destinations.
 * The transportation network is represented as 
 * a directed graph with locations serving as the vertices and lanes serving as
 * the edges.  A Lane is a directed edge connecting two locations. 
 * One location serves as the origin and the other location serves as a destination.  
 * The network is built by adding Lanes.  The origin of a lane cannot be
 * the same as the destination of a lane.   Thus, a network with only one
 * location cannot be constructed.
 *
 */
public class Network implements NetworkIfc {
	
	//  =============================================
	//  Data structures to hold the network structure
	//  =============================================
	
	/** This data structure allows the lane associated with a given origin
	 *  and a given destination to be easily found.
	 *  Represents a map with origin as the key and value as another 
	 *  map whose key is a destination and value is a lane
	 */
	protected Map<TransportLocationIfc, Map<TransportLocationIfc, LaneIfc>> myODLanes;
	
	/** This data structure holds the list of incident (outgoing) lanes connected
	 *  to the given location. A map with the origin location as the key and the list of associated
	 *  lanes as the value
	 * 
	 */
	protected Map<TransportLocationIfc, Collection<LaneIfc>> myOutGoingLanes;
	
	/** This data structure holds the list of incident (outgoing) locations connected
	 *  to the given location. A map with the origin location as the key and the list of associated
	 *  locations as the value
	 * 
	 */
	protected Map<TransportLocationIfc, List<TransportLocationIfc>> myOutGoingLocations;
	
	/** This data structure holds the list of incident (incoming) lanes connected
	 *  to the given location. A map with the destination location as the key and the list of associated
	 *  lanes as the value
	 * 
	 */
	protected Map<TransportLocationIfc, Collection<LaneIfc>> myIncomingLanes;

	/** This data structure holds the list of incident (incoming) locations connected
	 *  to the given location. A map with the destination location as the key and the list of associated
	 *  coming locations as the value
	 * 
	 */
	protected Map<TransportLocationIfc, List<TransportLocationIfc>> myIncomingLocations;
	
	/** A list of all the locations in the network that can be visited
	 */
	protected List<TransportLocationIfc> myTransportLocations;
	
	/** The set of locations designated as destinations
	 */
	protected Collection<TransportLocationIfc> myDestinations;
	
	/** A list of all the transport lanes in the network
	 */
	protected Collection<LaneIfc> myLanes;
		
	/** Rather than representing the transitive closure network
	 *  as a full network, this matrix of booleans represents
	 *  the adjacency matrix for the transitive closure network
	 */
	protected AdjMatrixGraph myTCNMatrix;
	
	/** A flag to indicate that a lane has been added
	 *  since the last construction of the transitive
	 *  closure network
	 */
	protected boolean myLaneAddedFlag = false;
	
	/** This flag indicates whether or not the network has all
	 *  reachable locations.
	 */
	protected boolean myReachabilityFlag = false;
	
	/** A flag to indicate whether isAllReachable() has been called
	 */
	protected boolean myCalledReachable = false;
	
	/** The comparator for comparing lanes Returns a negative integer or a positive integer
	 * if this lane's distance is less than or greater than the
	 * specified lane's distance. 
	 * 
	 * If the distances are equal, the id of the origin and
	 * destinations are checked. If the origin's id is smaller 
	 * than the destination's then the lane is considered "smaller" 
	 * than the supplied lane
	 * 
	 * Equality occurs only if the object references are equal.  This
	 * allows the checking of existence in sets.
	 * 
	 */
	protected LaneComparator myLaneComparator;
	
	//  ===========================================
	//  CONSTRUCTORS
	//  ===========================================

	/**  Creates the Network 
	 */
	public Network() {
		this(null);
	}
	
	/**  Creates the Network 
	 */
	public Network(LaneComparator laneComparator) {
		setLaneComparator(laneComparator);
		// set up the network data structures
		myODLanes = new HashMap<TransportLocationIfc, Map<TransportLocationIfc, LaneIfc>>();
		myOutGoingLanes = new HashMap<TransportLocationIfc, Collection<LaneIfc>>();
		myOutGoingLocations = new HashMap<TransportLocationIfc, List<TransportLocationIfc>>();
		myIncomingLanes =new HashMap<TransportLocationIfc, Collection<LaneIfc>>();
		myIncomingLocations = new HashMap<TransportLocationIfc, List<TransportLocationIfc>>();
		myTransportLocations = new ArrayList<TransportLocationIfc>();
		myLanes = new TreeSet<LaneIfc>(myLaneComparator);
		myDestinations = new HashSet<TransportLocationIfc>();
	}
	
	//  ===========================================
	//  PUBLIC METHODS
	//  ===========================================
		
	/** Adds a directed edge (Lane) between two locations in the network.  
	 *  A lane between an origin location and a destination location implies
	 *  that a move can be made from the origin to the destination.
	 * 
	 *  Requirements:
	 *  1) The locations must be different (origin cannot equal destination)
	 *  2) Only one lane can exist between a given origin/destination pair
	 *  
	 *  Issues:
	 *  1) The network developed by this method does not ensure reachability. Checking
	 *     for reachability is not performed by this method.
	 *  2) No error is thrown if a user tries to add a lane between an origin
	 *     and a destination that already exists in the network.  The already existing
	 *     lane is simply returned.
	 * 
	 * @param lane The lane to add
	 * @return True if the lane was added, false otherwise
	 */
	public boolean addLane(LaneIfc lane) {

		if (lane == null)
			throw new IllegalArgumentException("The supplied LaneIfc was null.");

		if (myLanes.contains(lane))
			throw new IllegalArgumentException("The supplied LaneIfc is already in the network.");
			
		TransportLocationIfc origin = lane.getOrigin();
		if (origin == null)
			throw new IllegalArgumentException("The supplied LaneIfc did not have an origin");

		TransportLocationIfc destination = lane.getDestination();
		if (destination == null)
			throw new IllegalArgumentException("The supplied LaneIfc did not have a destination");

		if (origin == destination)
			throw new IllegalArgumentException("The supplied LaneIfc's origin and destination were the same");
		
		// the actual lane object is not in the network, but the client
		// may be trying to add a different lane object to the network
		// with the same origin, destination pair
		// need to ensure only one lane btw given origin/destination pair
		
		// this map holds the lanes by destination
		Map<TransportLocationIfc, LaneIfc> lanesByDest = null;
		
		// need to insert the lane between the origin and the destination		
		// check if origin is already in OD lanes
		if (myODLanes.containsKey(origin)){
			// get the map of lanes for the origin
			lanesByDest = myODLanes.get(origin);
			if (lanesByDest.containsKey(destination)){
				// origin already in network and destination already in network
				// do not add this lane
				return(false);
			}
			else {
				myDestinations.add(destination);
				lanesByDest.put(destination, lane);
				myLanes.add(lane);
			}
		} else { // no map for destinations exists, create the map
			lanesByDest = new HashMap<TransportLocationIfc, LaneIfc>();
			myODLanes.put(origin, lanesByDest);
			myDestinations.add(destination);
			lanesByDest.put(destination,lane);
			myLanes.add(lane);
		}
		
		myLaneAddedFlag = true;
		myCalledReachable = false;
		
		// the lane has been added, add the O and D to the list of locations
		// if they have not already been added
		if (!myTransportLocations.contains(origin))
			myTransportLocations.add(origin);
		if (!myTransportLocations.contains(destination))
			myTransportLocations.add(destination);
		
		// a lane has been created, need to place it in 
		// the outGoing and inComing lane lists for the origin and destination
		
		// this list will reference the outgoing lanes from the origin
		Collection<LaneIfc> lanes = myOutGoingLanes.get(origin);
		if (lanes == null){
			lanes = new TreeSet<LaneIfc>(myLaneComparator);
			myOutGoingLanes.put(origin, lanes);
		}
		lanes.add(lane);
		
		List<TransportLocationIfc> outgoing = myOutGoingLocations.get(origin);
		if (outgoing == null){
			outgoing = new LinkedList<TransportLocationIfc>();
			myOutGoingLocations.put(origin, outgoing);
		}
		outgoing.add(destination);
		
		// this list will reference the incoming lanes to the destination
		lanes = myIncomingLanes.get(destination);
		if (lanes == null){
			lanes = new TreeSet<LaneIfc>(myLaneComparator);
			myIncomingLanes.put(destination, lanes);
		}
		lanes.add(lane);

		List<TransportLocationIfc> incoming = myOutGoingLocations.get(destination);
		if (incoming == null){
			incoming = new LinkedList<TransportLocationIfc>();
			myIncomingLocations.put(destination, incoming);
		}
		incoming.add(origin);

		return(true);
	}
	
	/** Returns the list of locations in the network
	 * @return
	 */
	public List<TransportLocationIfc> getLocations() {
		return (Collections.unmodifiableList(myTransportLocations));
	}
	
	/** Returns the number of location in the network
	 * 
	 * @return
	 */
	public int getNumberOfLocations(){
		return(myTransportLocations.size());
	}
	
	/** Checks if the supplied location is already in the network
	 * 
	 * @param location
	 * @return
	 */
	public boolean containsLocation(TransportLocationIfc location) {
		return(myTransportLocations.contains(location));
	}
	
	/** Returns an index 0 to getNumberOfLocations() - 1, that can 
	 *  be used to identify the location in the network 
	 * 
	 * @param location
	 * @return
	 */
	public int getIndexOf(TransportLocationIfc location){
		return(myTransportLocations.indexOf(location));
	}

	/** Returns the location at the supplied index
	 * 
	 * @param index Must be 0 to getNumberOfLocations() - 1
	 * @return
	 */
	public TransportLocationIfc getLocation(int index){
		return(myTransportLocations.get(index));
	}
	
	/** Returns the set of locations that have been designated as Origins
	 * @return
	 */
	public Collection<TransportLocationIfc> getOrigins(){
		return(Collections.unmodifiableCollection(myODLanes.keySet()));
	}
	
	/** Checks if the supplied location has been designated as an origin
	 * 
	 * @param location
	 * @return
	 */
	public boolean isOrigin(TransportLocationIfc location){
		return(myODLanes.containsKey(location));
	}
	
	/** Checks if the supplied location has been designated as a destination
	 * 
	 * @param location
	 * @return
	 */
	public boolean isDestination(TransportLocationIfc location){
		return(myDestinations.contains(location));
	}
	
	/** Returns the set of locations that have been designated as Origins
	 * @return
	 */
	public Collection<TransportLocationIfc> getDestinations(){
		return(Collections.unmodifiableCollection(myDestinations));
	}
	
	/** Each location may be act as the origin for many other locations (destinations)
	 *  Returns the list of outgoing lanes for the given origin
	 *  @param origin
	 * 	@return If no lanes exist then null is returned
	 */
	public Collection<LaneIfc> getOutGoingLanes(TransportLocationIfc origin) {
		if (origin == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!myTransportLocations.contains(origin))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		if (myOutGoingLanes.get(origin)==null)
			throw new IllegalArgumentException("The outgoing lane list was null for:"+origin.getName()+"!");
		
		return(Collections.unmodifiableCollection(myOutGoingLanes.get(origin)));
	}

	/** Each location may be act as the origin for many other locations (destinations)
	 *  Returns the list of outgoing locations for the given origin
	 *  @param origin
	 * 	@return If no lanes exist then null is returned
	 */
	public Collection<TransportLocationIfc> getOutGoingLocations(TransportLocationIfc origin) {
		if (origin == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!myTransportLocations.contains(origin))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		if (myOutGoingLocations.get(origin)==null)
			throw new IllegalArgumentException("The outgoing location list was null for:"+origin.getName()+"!");
		
		return(Collections.unmodifiableCollection(myOutGoingLocations.get(origin)));
	}
	
	/** Returns the number of outgoing lanes from the supplied origin
	 * 
	 * @param origin
	 * @return
	 */
	public int getNumberOfOutGoingLanes(TransportLocationIfc origin){
		if (origin == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!myTransportLocations.contains(origin))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		if (myOutGoingLanes.get(origin)==null)
			throw new IllegalArgumentException("The outgoing lane list was null for:"+origin.getName()+"!");
		
		return(myOutGoingLanes.get(origin).size());
	}
	
	/** Each location may be act as the destination for many other locations (origins)
	 *  Returns the list of incoming lanes for the given destination
	 *  @param origin
	 * 	@return If no lanes exist then null is returned
	 */
	public Collection<LaneIfc> getIncomingLanes(TransportLocationIfc destination) {
		if (destination == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!myTransportLocations.contains(destination))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		if (myIncomingLanes.get(destination)==null)
			throw new IllegalArgumentException("The incoming lane list was null for:"+destination.getName()+"!");
			
		return(Collections.unmodifiableCollection(myIncomingLanes.get(destination)));
	}
	
	/** Each location may be act as the destination for many other locations (origins)
	 *  Returns the list of incoming locations for the given destination
	 *  
	 *  @param destination
	 * 	@return If no locations exist then null is returned
	 */
	public Collection<TransportLocationIfc> getIncomingLocations(TransportLocationIfc destination){
		if (destination == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!myTransportLocations.contains(destination))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		if (myIncomingLanes.get(destination)==null)
			throw new IllegalArgumentException("The incoming location list was null for:"+destination.getName()+"!");
		
		return(Collections.unmodifiableCollection(myIncomingLocations.get(destination)));
	}
	
	/** Returns the number of incoming lanes to the supplied destination
	 * 
	 * @param destination
	 * @return
	 */
	public int getNumberOfIncomingLanes(TransportLocationIfc destination){
		if (destination == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!myTransportLocations.contains(destination))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		if (myIncomingLanes.get(destination)==null)
			throw new IllegalArgumentException("The incoming lane list was null for:"+destination.getName()+"!");
		
		return(myIncomingLanes.get(destination).size());
	}
	
	/** Returns the lane associated with the given origin/destination locations.  If
	 *  no lane exists, null will be returned.
	 * 
	 * @param origin
	 * @param destination
	 * @return
	 */
	public LaneIfc getLane(TransportLocationIfc origin, TransportLocationIfc destination) {
		if (origin == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!myTransportLocations.contains(origin))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		if (destination == null)
			throw new IllegalArgumentException("The supplied location was null.");
		
		if (!myTransportLocations.contains(destination))
			throw new IllegalArgumentException("The supplied location is not in the network.");
		
		if (origin == destination)
			throw new IllegalArgumentException("The supplied origin and destination must different");
		
		Map<TransportLocationIfc, LaneIfc> lanes = myODLanes.get(origin);
		
		if (lanes == null)
			return (null);
		
		return(lanes.get(destination));
	}
	
	/** Returns the list of Lanes. The list is sorted from smallest to largest
	 * @return
	 */
	public Collection<LaneIfc> getLanes() {
		return(Collections.unmodifiableCollection(myLanes));
	}
	
	/** Returns the number of lanes (directed edges) in the network
	 * 
	 * @return
	 */
	public int getNumberOfLanes(){
		return(myLanes.size());
	}
	
	/** Checks if there is a lane (directed edge) between
	 *  the given origin and the given destination
	 * 
	 * @param origin
	 * @param destination
	 * @return True means there is a lane
	 */
	public boolean isLane(TransportLocationIfc origin, TransportLocationIfc destination){
		if (origin == destination)
			return(false);
		else
			return(getLane(origin, destination) != null);
	}
	
	/** Checks if the lane is in the network
	 * 
	 * @param lane
	 * @return
	 */
	public boolean isLane(LaneIfc lane){
		return(myLanes.contains(lane));
	}
	
	/** Converts the network to a string representation
	 * 
	 */
	public String toString(){
		StringBuffer s = new StringBuffer();
		
		s.append("----- Network Structure ----- \n");
		for(TransportLocationIfc location: myTransportLocations){
			int i = getIndexOf(location);
			s.append("Vertex = " + i + "\t" + location.getName() + " : \t");
			Collection<LaneIfc> outLanes = getOutGoingLanes(location);
			for(LaneIfc lane: outLanes){
				int j = getIndexOf(lane.getDestination());
				s.append(j + " ");
			}
			s.append("\n");
		}
		
		s.append("----- Transitive Closure Reachability -----\n");
		
		if (isAllReachable())
			s.append("All vertices are reachable \n");
		else
			s.append("All vertices are not reachable \n");
			
	    s.append(getTransitiveClosureAdjMatrixGraph());
		s.append("\n");
		
		s.append("----- Network Lanes -----\n");
	    for(LaneIfc lane : myLanes){
	    		s.append(lane + "\n");
	    }
		return(s.toString());
	}
	
	/** Indicates whether or not the destination is reachable from the origin.
	 *  If a path exists from the origin back to itself, it will be indicated as reachable.
	 * 
	 * @param origin
	 * @param destination
	 * @return
	 */
	public boolean isReachable(TransportLocationIfc origin, TransportLocationIfc destination){
		int v = myTransportLocations.indexOf(origin);
		int w = myTransportLocations.indexOf(destination);			
		return(getTransitiveClosureAdjMatrixGraph().edge(v, w));
	}
	
	/** Using the reachability matrix of the network, this method checks if
	 *  all locations are reachable from each other. All locations are reachable
	 *  if there exists a directed path from each location to all other locations.
	 *  We assume that a location is not self-reachable, unless a cycle exists
	 *  such that it can get back to itself.
	 * 
	 * @return True means that all the locations can be reached
	 */
	public boolean isAllReachable(){
		if (myCalledReachable == false){
			AdjMatrixGraph tcg = getTransitiveClosureAdjMatrixGraph();
			myReachabilityFlag = tcg.isAllConnected();
			myCalledReachable = true;
		}
		return(myReachabilityFlag);
	}

	/** This method is called whenever a location associated with a lane changes
	 *  position.  This method updates all the distances of the lanes and re-sorts
	 *  the adjacency lists by distance.
	 *  
	 *  The updateNetworkAlgorithms() method is called by this method.
	 * 
	 * @param lane
	 */
	public void updateLane(LaneIfc lane) {
		// The lane distance changed, causing sorted sets to be invalid
		// remove the lane from each of its sets and put it back to correct the ordering
		myLanes.remove(lane);
		myLanes.add(lane);
		
		Collection<LaneIfc> lanes = getOutGoingLanes(lane.getOrigin());
		lanes.remove(lane);
		lanes.add(lane);
		
		lanes = getIncomingLanes(lane.getDestination());
		lanes.remove(lane);
		lanes.add(lane);
		
		updateNetworkAlgorithms();
	}
	
	//  =============================================
	//  PROTECTED METHODS
	//  =============================================
				
	/** Creates and returns a new transitive closure network for this network
	 *  A transitive closure network allows for easy queries on reachability
	 *  for a network.  See pg. 169 of Sedgewick's Algorithms in Java
	 *  Part 5 Graph Algorithms, 3rd edition
	 *  
	 *  This representation does not show self-loops, since a network (as defined
	 *  here) does not permit a lane (edge) between a location and itself
	 *  
	 * @return
	 */
	protected NetworkIfc createTransitiveClosureNetwork(){
		NetworkIfc tcn = new Network();
		// starting at every location, do the recursive DFS
		for(TransportLocationIfc location: myTransportLocations){
			tcRecursiveDFS(tcn, location, location);
		}
		return(tcn);
	}
	
	/** Creates and returns a new transitive closure network representation for this network
	 *  A transitive closure network allows for easy queries on reachability
	 *  for a network.  See pg. 169 of Sedgewick's Algorithms in Java
	 *  Part 5 Graph Algorithms, 3rd edition
	 *  
	 * 
	 * @return
	 */
	protected AdjMatrixGraph getTransitiveClosureAdjMatrixGraph(){
		if (myLaneAddedFlag){// a lane has been added since the last call
			myLaneAddedFlag = false;
			// need to recreate the transitive closure network
			myTCNMatrix = createTransitiveClosureAdjMatrixGraph();
		}
		return(myTCNMatrix);
	}
	
	/** The AdjMatrixGraph that is produced indicates the reachability
	 *  of the locations in the network.  Self-loops are shown as reachable
	 * 
	 * @return
	 */
	protected AdjMatrixGraph createTransitiveClosureAdjMatrixGraph(){
		
		AdjMatrixGraph tcg = new AdjMatrixGraph(myTransportLocations.size());
		// starting at every location, do the recursive DFS
		for(TransportLocationIfc location: myTransportLocations){
			tcRecursiveDFS(tcg, location, location);
		}
		
		// the tcg that results will have locations reachable
		// from themselves if they are on a cycle
		// we require that no location serve as its own destination
		// remove any self-loops, so that locations are not reachable from themselves
//		int n = tcg.getNumVertices();
//		for(int i=0;i<n;i++)
//			tcg.removeEdge(i,i);
			
		return(tcg);
	}
		
	//  =============================================
	//  PRIVATE METHODS
	//  =============================================
	
	private void tcRecursiveDFS(NetworkIfc tcn, TransportLocationIfc startingLocation, 
			TransportLocationIfc visitingLocation){
		// get the set of outgoing lanes for this location
		Collection<LaneIfc> outLanes = myOutGoingLanes.get(visitingLocation);
		for(LaneIfc lane: outLanes){
			TransportLocationIfc d = lane.getDestination();
			if(startingLocation == d) // if it reachs itself go to next lane
				continue;
			if (!tcn.isLane(startingLocation, d)){// startingLocation and d are not in tcn
				// create a new lane
				LaneIfc nlane = new Lane(this, startingLocation, d);
				// add the new lane to the tcn
				tcn.addLane(nlane);
				// go "down" through the network by destination
				tcRecursiveDFS(tcn, startingLocation, d);
			}
		}
	}
	
	private void tcRecursiveDFS(AdjMatrixGraph tcg, TransportLocationIfc startingLocation, 
			TransportLocationIfc visitingLocation){
		// get the indices of the locations for mapping to matrix
		int s = myTransportLocations.indexOf(startingLocation);
		
		// get the set of outgoing lanes for this location
		Collection<LaneIfc> outLanes = myOutGoingLanes.get(visitingLocation);
		for(LaneIfc lane: outLanes){
			TransportLocationIfc d = lane.getDestination();
			int v = myTransportLocations.indexOf(d);
//			if (s == v) // if it reached itself, continue with next lane
//				continue;
			if (!tcg.edge(s, v)){// startingLocation and d are not in tcg
				// create an edge to represent the path from starting location to d
				tcg.insertEdge(s, v);
				// go "down" through the network by destination
				tcRecursiveDFS(tcg, startingLocation, d);
			}
		}
	}
	
	/** Can be used to ensure that any network algorithms that depend on the order of the
	 *  outgoing lane, incoming lane, and lanes list get updated.  Also any algorithms that
	 *  depend upon the distance of a lane
	 */
	protected void updateNetworkAlgorithms(){
		
	}
	
	/** Returns a negative integer or a positive integer
	 * if this lane's distance is less than or greater than the
	 * specified lane's distance. 
	 * 
	 * If the distances are equal, the id of the origin and
	 * destinations are checked. If the origin's id is smaller 
	 * than the destination's then the lane is considered "smaller" 
	 * than the supplied lane
	 * 
	 * Equality occurs only if the object references are equal.  This
	 * allows the checking of existence in sets.
	 *
	 * Note:  This class may have a natural ordering that is inconsistent
	 * with equals.
	 * 
	 * @param lane The lane to compare to
	 * @return Returns a negative integer, zero, or a positive integer
	 *
	 */ 
	protected class LaneComparator implements Comparator<LaneIfc>{

		public int compare(LaneIfc lane1, LaneIfc lane2) {
	        if (lane1.equals(lane2))// object are equal with zero
	            return(0);
			
	        // check distances
			double d1 = lane1.getDistance();
			double d2 = lane2.getDistance();
			if (d1 > d2)
				return(1);
			else if (d1 < d2)
				return(-1);
			else { // distances are equal, use id to distinguish
				long id1 = lane1.getOrigin().getId();
				long id2 = lane2.getOrigin().getId();
		        	if (id1 < id2)
		        		return(-1);
		        	else
		        		return(1);	        	
			}			
		}
		
	}

	/** If null is passed, then the default comparator is used
	 * @param myLaneComparator The laneComparator to set.
	 */
	private final void setLaneComparator(LaneComparator laneComparator) {
		if (laneComparator == null)
			myLaneComparator = new LaneComparator();
		else
			myLaneComparator = laneComparator;
	}
	
}
