/*
 * Created on Jul 9, 2006
 *
 */
package jslx.transportlayerFinal;

import java.util.Collection;
import java.util.List;

/**
 * @author rossetti
 *
 */
public interface NetworkIfc {
	
	/** Adds a directed edge (Lane) between two locations in the network.  
	 *  A lane between an origin location and a destination location implies
	 *  that a move can be made from the origin to the destination.
	 * 
	 *  Requirements:
	 *  1) The locations must be different (origin cannot equal destination)
	 *  2) Only one lane can exist between a given origin/destination pair
	 *  
	 *  Issues:
	 *  1) The network developed by this method does not ensure reachability. Checking
	 *     for reachability is not performed by this method.
	 *  2) No error is thrown if a user tries to add a lane between an origin
	 *     and a destination that already exists in the network.  The already existing
	 *     lane is simply returned.
	 * 
	 * @param lane The lane to add
	 * @return True if the lane was added, false otherwise
	 */
	public abstract boolean addLane(LaneIfc lane);

	/** Returns the list of locations in the network
	 * @return
	 */
	public abstract List<TransportLocationIfc> getLocations();

	/** Returns the number of location in the network
	 * 
	 * @return
	 */
	public abstract int getNumberOfLocations();

	/** Checks if the supplied location is already in the network
	 * 
	 * @param location
	 * @return
	 */
	public abstract boolean containsLocation(TransportLocationIfc location);

	/** Returns an index 0 to getNumberOfLocations() - 1, that can 
	 *  be used to identify the location in the network 
	 * 
	 * @param location
	 * @return
	 */
	public abstract int getIndexOf(TransportLocationIfc location);

	/** Returns the location at the supplied index
	 * 
	 * @param index Must be 0 to getNumberOfLocations() - 1
	 * @return
	 */
	public abstract TransportLocationIfc getLocation(int index);

	/** Returns the set of locations that have been designated as Origins
	 * @return
	 */
	public abstract Collection<TransportLocationIfc> getOrigins();

	/** Checks if the supplied location has been designated as an origin
	 * 
	 * @param location
	 * @return
	 */
	public abstract boolean isOrigin(TransportLocationIfc location);

	/** Checks if the supplied location has been designated as a destination
	 * 
	 * @param location
	 * @return
	 */
	public abstract boolean isDestination(TransportLocationIfc location);

	/** Returns the set of locations that have been designated as Origins
	 * @return
	 */
	public abstract Collection<TransportLocationIfc> getDestinations();

	/** Each location may be act as the origin for many other locations (destinations)
	 *  Returns the list of outgoing lanes for the given origin
	 *  @param origin
	 * 	@return If no lanes exist then null is returned
	 */
	public abstract Collection<LaneIfc> getOutGoingLanes(
			TransportLocationIfc origin);

	/** Returns the number of outgoing lanes from the supplied origin
	 * 
	 * @param origin
	 * @return
	 */
	public abstract int getNumberOfOutGoingLanes(TransportLocationIfc origin);

	/** Each location may be act as the destination for many other locations (origins)
	 *  Returns the list of incoming lanes for the given destination
	 *  @param origin
	 * 	@return If no lanes exist then null is returned
	 */
	public abstract Collection<LaneIfc> getIncomingLanes(
			TransportLocationIfc destination);

	/** Returns the number of incoming lanes to the supplied destination
	 * 
	 * @param destination
	 * @return
	 */
	public abstract int getNumberOfIncomingLanes(
			TransportLocationIfc destination);

	/** Each location may be act as the origin for many other locations (destinations)
	 *  Returns the list of outgoing locations for the given origin
	 *  @param origin
	 * 	@return If no lanes exist then null is returned
	 */
	public abstract Collection<TransportLocationIfc> getOutGoingLocations(TransportLocationIfc origin);

	/** Each location may be act as the destination for many other locations (origins)
	 *  Returns the list of incoming locations for the given destination
	 *  
	 *  @param destination
	 * 	@return If no locations exist then null is returned
	 */
	public abstract Collection<TransportLocationIfc> getIncomingLocations(TransportLocationIfc destination);

	/** Returns the lane associated with the given origin/destination locations.  If
	 *  no lane exists, null will be returned.
	 * 
	 * @param origin
	 * @param destination
	 * @return
	 */
	public abstract LaneIfc getLane(TransportLocationIfc origin,
			TransportLocationIfc destination);

	/** Returns the list of Lanes. The list is sorted from smallest to largest
	 * @return
	 */
	public abstract Collection<LaneIfc> getLanes();

	/** Returns the number of lanes (directed edges) in the network
	 * 
	 * @return
	 */
	public abstract int getNumberOfLanes();

	/** Checks if there is a lane (directed edge) between
	 *  the given origin and the given destination
	 * 
	 * @param origin
	 * @param destination
	 * @return
	 */
	public abstract boolean isLane(TransportLocationIfc origin,
			TransportLocationIfc destination);

	/** Checks if the lane is in the network
	 * 
	 * @param lane
	 * @return
	 */
	public abstract boolean isLane(LaneIfc lane);
	
	/** Indicates whether or not the destination is reachable from the origin
	 * 
	 * @param origin
	 * @param destination
	 * @return
	 */
	public abstract boolean isReachable(TransportLocationIfc origin, TransportLocationIfc destination);
	
	/** Using the transitive closure of the network, this method checks if
	 *  all locations are reachable from each other. All locations are reachable
	 *  if there exists a directed path from each location to all other locations
	 * 
	 * @return True means that all the locations can be reached
	 */
	public abstract boolean isAllReachable();
	
	/** This method is called whenever a location associated with a lane changes
	 *  position.  This method updates all the distances of the lanes and re-sorts
	 *  the adjacency lists by distance.
	 *  
	 *  The updateNetworkAlgorithms() method is called by this method.
	 * 
	 * @param lane
	 */
	public abstract void updateLane(LaneIfc lane);

}