/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package jslx.utilities.queueing;



/**
 *
 */
public class GGcWhittApproximation {
    
    private int myNumServers;
    private double myRho;
    private double myMTBA;
    private double myMST;
    private double mySCVA;
    private double mySCVS;
    
    /** Creates a new instance of GGcWhittApproximation */
    public GGcWhittApproximation() {
        this(2.0, 1.0, 1.0, 1.0, 1);
    }
    
    public GGcWhittApproximation(double mtba, double scva, double mst, double scvs, int servers) {
        setParameters(mtba, scva, mst, scvs, servers);
    }
    
    public void setParameters(double[] parameter){
        if (parameter.length < 5)
            throw new IllegalArgumentException("Not enough parameters passed.");
        setParameters(parameter[0], parameter[1], parameter[2], parameter[3], (int)parameter[4]);
    }
        
    public void setParameters(double mtba, double scva, double mst, double scvs, int servers){
        setUtilization(mtba, mst, servers);
        setArrivalSquaredCoeffOfVariation(scva);
        setServiceSquaredCoeffOfVariation(scvs);
    }
    
    public void setParameters(QParameterSet qps){
        setParameters(qps.mtba, qps.c2va, qps.mst, qps.c2va, qps.numServers);
    }
    
    private void setMeanTimeBtwArrivals(double mtba){
        if (mtba <= 0)
            throw new IllegalArgumentException("The mean time between arrivals must be > 0");
        
        myMTBA = mtba;
    }
    
    private void setMeanServiceTime(double mst){
        if (mst <= 0)
            throw new IllegalArgumentException("The mean service time must be > 0");
        
        myMST = mst;
    }
    
    private void setNumberOfServers(int servers){
        if (servers <= 0)
            throw new IllegalArgumentException("Number of servers must be > 0");
        
        myNumServers = servers;
    }
    
    public void setUtilization(double mtba, double mst, int servers){
        setMeanTimeBtwArrivals(mtba);
        setMeanServiceTime(mst);
        setNumberOfServers(servers);
        double rho = myMST/(myMTBA*myNumServers);
        if ((rho <= 0.0) || (rho >= 1.0))
            throw new IllegalArgumentException("Utilization must be (0,1)");
        
        myRho = rho;
    }
    
    public void setArrivalSquaredCoeffOfVariation(double scva){
        if (scva < 0)
            throw new IllegalArgumentException("Squared Coefficient of Variation must be > 0");
        
        mySCVA = scva;
    }
    
    public void setServiceSquaredCoeffOfVariation(double scvs){
        if (scvs < 0)
            throw new IllegalArgumentException("Squared Coefficient of Variation must be > 0");
        
        mySCVS = scvs;
    }
    
    public double getApproxWaitInQ(){
        double ewqMMm = MMcWq(myRho, 1.0, myNumServers);
        double scv = (mySCVA + mySCVS)/2.0;
        double phi = phi(myRho, mySCVA, mySCVS, myNumServers);
        double wq = phi*scv*ewqMMm;
        return(myMST*wq);
    }
    
    public double getApproxNumInQ(){
        double wq = getApproxWaitInQ();
        return(wq/myMTBA);
    }
    
    public double getApproxNumInSys(){
        double ws = getApproxWaitInSys();
        return(ws/myMTBA);
    }
    
    public double getApproxWaitInSys(){
        double wq = getApproxWaitInQ();
        return(wq+myMST);
    }
    
    public double getExpectedNumInService(){
        return(myMST/myMTBA);
    }
    
    public String toString(){
        String s;
        
        s = "Whitt Approximation \n";
        s = s + "Number of servers " + myNumServers + "\n";
        s = s + "Arrival rate " + 1.0/myMTBA + "\n";
        s = s + "Mean time btw arrivals " + myMTBA + "\n";
        s = s + "Service rate " + 1.0/myMST + "\n";
        s = s + "Mean Service time " + myMST + "\n";
        s = s + "Arrival Squared coeff of variation " + mySCVA + "\n";
        s = s + "Service Squared coeff of variation " + mySCVS + "\n";
        s = s + "Utilization (rho) " + myRho + "\n";
        s = s + "Approx wait in queue " + getApproxWaitInQ() + "\n";
        s = s + "Approx wait in system " + getApproxWaitInSys() + "\n";
        s = s + "Approx num in queue " + getApproxNumInQ() + "\n";
        s = s + "Approx num in system " + getApproxNumInSys() + "\n";
        s = s + "Expected num in service" + getExpectedNumInService() + "\n";
        
        double ewqMMm = MMcWq(myRho, 1.0/myMST, myNumServers);
        
        s = s + "MMc wait in queue " + ewqMMm + "\n";
        s = s + "MMc num in queue " + ewqMMm/myMTBA + "\n";
        s = s + "MMc wait in system " + ewqMMm + myMST + "\n";
        s = s + "MMc num in system " + (ewqMMm + myMST)/myMTBA + "\n";
        return(s);
    }
    
    public static double MMcWq(double rho, double mu, int c){
        if ((rho <= 0.0) || (rho >= 1.0))
            throw new IllegalArgumentException("Utilization must be (0,1)");
        
        if (mu <= 0.0)
            throw new IllegalArgumentException("Service rate must be > 0");
        
        if (c <= 0)
            throw new IllegalArgumentException("Number of servers must be > 0");
        
        double a = c*rho;
        double csa = erlangC(c, a);
        double cmwq = 1.0/((1.0 - rho)*c*mu); // conditional mean wait in queue
        return(csa*cmwq);
    }
    
    public static double MMcLq(double rho, double mu, int c){
        double wq = MMcWq(rho, mu, c);
        
        double lambda = rho*c*mu;
        
        return(lambda*wq);
    }
    
    public static double erlangB(int m, double a){
        double b = 1.0;
        double d = 1.0;
        double t = 1.0;
        
        for(int i=1; i<=m; i++){
            t = a*b;
            d = i + t;
            b = t/d;
        }
        
        return(b);
    }
    
    public static double erlangC(int m, double a){
        double b = erlangB(m,a);
        double t = m*b;
        double d = m - a*(1.0-b);
        return(t/d);
    }
    
    private double phiOne(int m, double rho){
        double phiOne = 1.0 + gamma(m, rho);
        System.out.println("phiOne = " + phiOne);
        return(phiOne);
    }
    
    private double phiTwo(int m, double rho){
        double phiTwo = 1.0 - 4.0*gamma(m, rho);
        System.out.println("phiTwo = " + phiTwo);
        return(phiTwo);
    }
    
    private double phiThree(int m, double rho){
        double x = Math.exp(-2.0*(1.0-rho)/(3.0*rho));
        double phiThree = phiTwo(m, rho)*x;
        System.out.println("phiThree = " + phiThree);
        return(phiThree);
    }
    
    private double phiFour(int m, double rho){
        double x = (phiOne(m, rho) + phiThree(m, rho))/2.0;
        double phiFour = Math.min(1.0, x);
        System.out.println("phiFour = " + phiFour);
        return(phiFour);
    }
    
    private double psi(double scv, int m, double rho){
        double psi;
        if (scv >= 1.0)
            psi = 1.0;
        else {
            double x = 2.0*(1.0 - scv);
            psi = Math.pow(phiFour(m, rho), x);
        }
        System.out.println("psi = " + psi);
        return(psi);
    }
    
    private double phi(double rho, double scva, double scvs, int m){
        double phi;
        double scv = (scva + scvs)/2.0;
        if (scva == scvs){ // this case should reduce directly to psi
            phi = psi(scva, m, rho);
        } else if (scva > scvs){
            double d = 4.0*scva - 3.0*scvs;
            double a = (4.0*(scva - scvs))/d;
            double b = scvs/d;
            phi = a*phiOne(m, rho) + b*psi(scv, m, rho);
        } else {// scva < scvs
            if (scva == 0.0) // handle special case where equation 2.25 doesn't reduce
                phi = phiThree(m,rho);
            else {
                double d = 2.0*(scva + scvs);
                double a = (scvs - scva)/d;
                double b = (scvs + 3.0*scva)/d;
                phi = a*phiThree(m, rho) + b*psi(scv, m, rho);
            }
        }
        System.out.println("phi = " + phi);
        return(phi);
    }
    
    private double gamma(int m, double rho){
        double x;
        x = (1.0-rho)*(m-1.0)*(Math.sqrt(4 +5*m) - 2.0)/(16.0*m*rho);
        double gamma = Math.min(0.24,x);
        System.out.println("gamma = " + gamma);
        return(gamma);
    }
    
    //  ===========================================
    //      TEST METHODS
    //  ===========================================
    //  For individual classes use main() as test
    
    public static void main( String[] args ){
        System.out.println("GGC Approximation Test");
        
        GGcWhittApproximation g = new GGcWhittApproximation();
        
/*        double rho = 0.70;
        double scvs = 0.5;
        double mst = 1.0;
        int m = 20;
        double scva = 0.5;
 */
        
        double rho = 0.95;
        double scvs = 1.0;
        double mst = 1.0;
        int m = 20;
        double scva = 0.0;
        
        double mtba = mst/(rho*m);
        
        g.setParameters(mtba, scva, mst, scvs, m);
        
        System.out.println(g);
        
        System.out.println("Done!");
        
    }
}
