/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/


package jslx.utilities.queueing;

import java.util.ArrayList;

import jsl.utilities.random.*;
import jsl.utilities.random.distributions.DUniform;
import jsl.utilities.random.distributions.Distribution;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.random.distributions.Gamma;
import jsl.utilities.random.distributions.Lognormal;
import jsl.utilities.random.distributions.Triangular;
import jsl.utilities.random.distributions.Uniform;
import jsl.utilities.random.distributions.Weibull;
import jsl.utilities.reporting.*;
import java.io.IOException;


import jsl.utilities.rootfinding.*;
import jsl.utilities.math.*;

/**
 *
 * @author  rossetti
 */
public class QParameterSetGenerator {
    
    private int myMaxNumServers;
    private double myMaxMTBA;
    private double myMaxC2v;
    
    private Uniform myMTBACDF;
    private DUniform myServerCDF;
    private Uniform myRhoCDF;
    private Uniform myC2vCDF;
    
    private DUniform myTriangCasesCDF;
    private DUniform myDCDF;
    
    private ArrayList<DistributionIfc> myArrivalCDFs;
    private ArrayList<DistributionIfc> myServiceCDFs;
    
    private WFunction myWFunction;
    private double myAlphaNeg;
    private double myAlphaPos;
//    private BisectionZeroFinder myZeroFinder;
    private BisectionRootFinder myZeroFinder;
    
    private final int EXPONENTIAL = 1;
    private final int UNIFORM = 2;
    private final int GAMMA = 3;
    private final int WEIBULL = 4;
    private final int TRIANGULAR = 5;
    private final int LOGNORMAL = 6;
    
    private final int ARRIVAL = 1;
    private final int SERVICE = 2;
    
    /** Creates a new instance of QParameterSetGenerator */
    public QParameterSetGenerator(){
        this(150, 100.0, 100.0);
    }
    
    public QParameterSetGenerator(int maxNumServers, double maxMTBA, double maxC2v) {
        setMaxMTBA(maxMTBA);
        setMaxNumServers(maxNumServers);
        setMaxSqCoeffOfVariation(maxC2v);
        myC2vCDF = new Uniform(0.0, myMaxC2v);
        myRhoCDF = new Uniform();
        
        myTriangCasesCDF = new DUniform(1,3);
        
        // distribution information
        myDCDF = new DUniform(1, 6);
        myArrivalCDFs = new ArrayList<DistributionIfc>();
        myServiceCDFs = new ArrayList<DistributionIfc>();
        
        myArrivalCDFs.add(new Exponential());
        myArrivalCDFs.add(new Uniform());
        myArrivalCDFs.add(new Gamma());
        myArrivalCDFs.add(new Weibull());
        myArrivalCDFs.add(new Triangular());
        myArrivalCDFs.add(new Lognormal());
        
        myServiceCDFs.add(new Exponential());
        myServiceCDFs.add(new Uniform());
        myServiceCDFs.add(new Gamma());
        myServiceCDFs.add(new Weibull());
        myServiceCDFs.add(new Triangular());
        myServiceCDFs.add(new Lognormal());
        
        myWFunction = new WFunction();
//        myZeroFinder = new BisectionZeroFinder(myWFunction);
        myZeroFinder = new BisectionRootFinder();
        
    }
    
    public void setMaxMTBA(double maxMTBA){
        if (maxMTBA <= 0.0)
            throw new IllegalArgumentException("Maximum mean time btw arrivals must be > 0");
        myMaxMTBA = maxMTBA;
        if (myMTBACDF == null)
            myMTBACDF = new Uniform(0.0, myMaxMTBA);
        else
            myMTBACDF.setRange(0.0, myMaxMTBA);
    }
    
    public void setMaxNumServers(int maxNumServers){
        if (maxNumServers < 1)
            throw new IllegalArgumentException("Number of servers must be >=1");
        myMaxNumServers = maxNumServers;
        if (myServerCDF == null)
            myServerCDF = new DUniform(1, myMaxNumServers);
        else
            myServerCDF.setRange(1, myMaxNumServers);
    }
    
    public void setMaxSqCoeffOfVariation(double maxC2v){
        if (maxC2v <= 0.0)
            throw new IllegalArgumentException("Squared Coeff. of Variation must be > 0.0");
        myMaxC2v = maxC2v;
    }
    
    public QParameterSet generateQParameterSet(){
        QParameterSet q = new QParameterSet();
        
        q.mtba = myMTBACDF.getValue();
        q.numServers = (int)myServerCDF.getValue();
        q.rho = myRhoCDF.getValue();
        q.mst = q.rho*q.mtba*q.numServers;
        
        matchDistribution(q, ARRIVAL);
        matchDistribution(q, SERVICE);
        
        return(q);
    }
    
    private void matchDistribution(QParameterSet q, int processType){
        double mean = 0.0;
        double p1 = 0.0;
        double p2 = 0.0;
        double p3 = 0.0;
        double[] param = new double[3];
        
        Distribution d;
        
        int type = (int)myDCDF.getValue();
        
        if (processType == ARRIVAL){
            q.acdfType = type;
            q.acdf = (Distribution)myArrivalCDFs.get(q.acdfType - 1);
            d = q.acdf;
            mean = q.mtba;
        } else {
            q.scdfType = type;
            q.scdf = (Distribution)myServiceCDFs.get(q.scdfType - 1);
            d = q.scdf;
            mean = q.mst;
        }
        
        switch (type){
            case EXPONENTIAL:
                matchExponential(mean, (Exponential)d);
                param = d.getParameters();
                p1 = param[0];
                break;
            case UNIFORM:
                matchUniform(mean, (Uniform)d);
                param = d.getParameters();
                p1 = param[0];
                p2 = param[1];
                break;
            case GAMMA:
                matchGamma(mean, (Gamma)d);
                param = d.getParameters();
                p1 = param[0];
                p2 = param[1];
                break;
            case WEIBULL:
                matchWeibull(mean, (Weibull)d);
                param = d.getParameters();
                p1 = param[0];
                p2 = param[1];
                break;
            case TRIANGULAR:
                matchTriangular(mean, (Triangular)d);
                param = d.getParameters();
                p1 = param[0];
                p2 = param[1];
                p3 = param[2];
                break;
            case LOGNORMAL:
                matchLognormal(mean, (Lognormal)d);
                param = d.getParameters();
                p1 = param[0];
                p2 = param[1];
                break;
        }
        
        double var = d.getVariance();
        double cv2 = var/(mean*mean);
        if (processType == ARRIVAL){
            q.c2va = cv2;
            q.aP1 = p1;
            q.aP2 = p2;
            q.aP3 = p3;
        }
        else{
            q.c2vs = cv2;
            q.sP1 = p1;
            q.sP2 = p2;
            q.sP3 = p3;
        }
    }
    
    private void matchExponential(double mean, Exponential e){
        e.setMean(mean);
    }
    
    private void matchUniform(double mean, Uniform u){
        myC2vCDF.setRange(0.0, 1.0/3.0);
        double c2v = myC2vCDF.getValue();
        double var = c2v*mean*mean;
        double a = mean - Math.sqrt(3.0*var);
        double b = mean + Math.sqrt(3.0*var);
        u.setRange(a, b);
    }
    
    private void matchGamma(double mean, Gamma g){
        myC2vCDF.setRange(0.0, myMaxC2v);
        double c2v = myC2vCDF.getValue();
        double var = c2v*mean*mean;
        
        double shape = mean*mean/var;
        double scale = var/mean;
        
        g.setShape(shape);
        g.setScale(scale);
    }
    
    private void matchWeibull(double mean, Weibull w){
        myC2vCDF.setRange(0.0, myMaxC2v);
        boolean flag = true;
        do {
            double c2v = myC2vCDF.getValue();
            myWFunction.setC2v(c2v);
            findBracket();
            myZeroFinder.setInterval(myWFunction, myAlphaNeg, myAlphaPos);
            myZeroFinder.setInitialPoint((myAlphaNeg + myAlphaPos)/2.0);
            
            myZeroFinder.evaluate();
            double shape = myZeroFinder.getResult();
            double scale = mean/Gamma.gammaFunction(1.0 + 1.0/shape);
            w.setScale(scale);
            w.setShape(shape);
            if (myZeroFinder.hasConverged())
                flag = true;
            else
                flag = false;
        } while (flag == false);
    }
    
    private void findBracket(){
        
        if (myWFunction.fx(1.0) > 0.0){
            myAlphaPos = 1.0;
            double alpha = 1.0;
            do {
                alpha = alpha*2.0;
            } while (myWFunction.fx(alpha) >= 0.0);  // when f(a) < 0, then it drops out of the while loop
            myAlphaNeg = alpha;
        } else {
            myAlphaNeg = 1.0;
            double alpha = 1.0;
            do {
                alpha = alpha/2.0;
            } while (myWFunction.fx(alpha) <= 0.0);  // when f(a) > 0, then it drops out of the while loop
            myAlphaPos = alpha;
        }
    }
    
    private class WFunction implements FunctionIfc{
        
        private double myC2v = 1.0;
        private double c0;
        
        public void setC2v(double c2v){
            myC2v = c2v;
            c0 = Math.log(myC2v+1.0);
        }
        
        /** Returns the value of the function for the specified variable value.
         */
        public double fx(double alpha) {
            double x = Gamma.logGammaFunction(1.0 + 2.0/alpha);
            double y = Gamma.logGammaFunction(1.0 + 1.0/alpha);
            double f = x - 2.0*y - c0;
            return(f);
        }
        
    }
    
    private void matchLognormal(double mean, Lognormal ln){
        myC2vCDF.setRange(0.0, myMaxC2v);
        double c2v = myC2vCDF.getValue();
        double var = c2v*mean*mean;
        ln.setParameters(mean, var);
    }
    
    private void matchTriangular(double mean, Triangular t){
        double c2v = 0.0;
        double var = 0.0;
        int i = (int)myTriangCasesCDF.getValue();
        
        switch (i){
            case 1: // symmetric
                myC2vCDF.setRange(0.0, 1.0/6.0);
                c2v = myC2vCDF.getValue();
                var = c2v*mean*mean;
                triangSymmetric(mean, var, t);
                break;
            case 2: // positive skew
                myC2vCDF.setRange(0.0, 1.0/8.0);
                c2v = myC2vCDF.getValue();
                var = c2v*mean*mean;
                triangPositiveSkew(mean, var, t);
                break;
            case 3: //negative skew
                myC2vCDF.setRange(0.0, 1.0/2.0);
                c2v = myC2vCDF.getValue();
                var = c2v*mean*mean;
                triangNegativeSkew(mean, var, t);
                break;
        }
    }
    
    private void triangSymmetric(double mean, double var, Triangular t){
        double a = mean - Math.sqrt(6.0*var);
        double b = mean + Math.sqrt(6.0*var);
        double c = mean;
        t.setParameters(a, c, b);
    }
    
    private void triangNegativeSkew(double mean, double var, Triangular t){
        double a = mean - Math.sqrt(2.0*var);
        double c = a;
        double b = mean + 2.0*Math.sqrt(2.0*var);
        t.setParameters(a, c, b);
    }
    
    private void triangPositiveSkew(double mean, double var, Triangular t){
        double b = mean + Math.sqrt(2.0*var);
        double c = b;
        double a = mean - 2.0*Math.sqrt(2.0*var);
        t.setParameters(a, c, b);
    }
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        
        QParameterSetGenerator g = new QParameterSetGenerator(150, 100.0, 100.0);
        
        System.out.println("Starting generation");
        try{
            TextReport r;
            r = new TextReport("qdata");
            r.addLines("mtba \t c2va \t mst \t c2vs \t numservers \t rho \t acdf \t scdf \t ap1 \t ap2 \t ap3 \t sp1 \t sp2 \t sp3");
            
            QParameterSet q;
            for(int i=1; i<=2000; i++){
                q = g.generateQParameterSet();
                r.addLines(q);
//                r.addLines(q.mtba + "\t" + q.c2va + "\t" + q.mst + "\t" + q.c2vs + "\t" + q.numServers + "\t" + q.rho + "\t" + q.acdfType + "\t" + q.scdfType);
            }
        }
        catch(IOException e){
            System.out.println("I/O exception opening ");
            System.exit(1);
        }
        System.out.println("Completed generation");
    }
    
}
