/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 *
 * Copyright (c) Manuel D. Rossetti (rossetti@uark.edu)
 *
 * Contact:
 *	Manuel D. Rossetti, Ph.D., P.E.
 *	Department of Industrial Engineering
 *	University of Arkansas
 *	4207 Bell Engineering Center
 *	Fayetteville, AR 72701
 *	Phone: (479) 575-6756
 *	Email: rossetti@uark.edu
 *	Web: www.uark.edu/~rossetti
 *
 * This file is part of the JSL (a Java Simulation Library). The JSL is a framework
 * of Java classes that permit the easy development and execution of discrete event
 * simulation programs.
 *
 * The JSL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * The JSL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the JSL (see file COPYING in the distribution);
 * if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA, or see www.fsf.org
 *
 */
package jslx.utilities.random;

import jsl.utilities.random.rng.RandomStreamIfc;
import jsl.utilities.random.rng.RngIfc;
import jsl.utilities.random.rng.RngStream;

/** Generates multi-variate lognormal random variates based on
 *  an underlying mulit-variate normal.  This procedure
 *  depends upon the Cholesky decomposition of the var/cov
 *  matrix of the underlying multi-variate normal.  Because of
 *  this it is recommend that you check if the var/cov matrix of the underlying mvn
 *  is positive symmetric definite.
 *
 *  The class takes as input the mean and var/cov matrix
 *  of the underlying multi-variate normal as parameters
 *
 *  The static method getMVNormalParametersFromMVLognormal() can
 *  be used to get the multi-variate normal (mvn) mean and var/cov matrix that 
 *  is needed for a given multi-variate lognormal mean and var/cov matrix.
 *  The resulting mvn var/cov matrix should be checked if it is
 *  positive symmetric definite
 *
 *  The static method getMVLogNormalParametersFromMVNormal() can
 *  be used to get the multi-variate log normal mean and variances
 *  from a mean and var/cov matrix representing the multi-variate normal
 *
 * @author rossetti
 */
public class MVLogNormal implements RandomStreamIfc {

    protected MVNormal myMVN;

    /** Creates a multi-variate lognormal, based on
     *  standard multi-variate normal with identity var/cov matrix
     *
     * @param n  dimension of the multi-variate random variable
     */
    public MVLogNormal(int n) {
        this(n, new RngStream());
    }

    /** Creates a multi-variate lognormal, based on
     *  standard multi-variate normal with identity var/cov matrix
     *
     * @param n  dimension of the multi-variate random variable
     * @param rng
     */
    public MVLogNormal(int n, RngIfc rng) {
        myMVN = new MVNormal(n, rng);
    }

    /** Creates a multi-variate lognormal based on the
     *  supplied means and covariances
     *
     * @param mu  means for the normal
     * @param cov var/cov matrix of the normals
     */
    public MVLogNormal(double[] mu, double[][] cov) {
        this(mu, cov, new RngStream());
    }

    /** Creates a MVLogNormal based on the mean
     *  and var/cov matrix of of the multi-variate lognormal
     *
     * @param mu mean of multi-variate normals
     * @param cov var/cov of multi-variate normals
     * @param rng
     */
    public MVLogNormal(double[] mu, double[][] cov, RngIfc rng) {
        myMVN = new MVNormal(mu, cov, rng);
    }

    /** Sets the parameters of the underlying multi-variate
     *  normal
     *
     * @param mu mean of the underlying multi-variate normal
     * @param cov var/cov matrix of the underlying multi-variate normal
     */
    public void setParameters(double[] mu, double[][] cov) {
        myMVN.setParameters(mu, cov);
    }

    /** Computes the multi-variate normal mean and var/cov
     *  from the mean and var/cov of the multi-variate lognormal
     *
     * @param mu  mean of log normal
     * @param cov  var/cov of log normal
     * @param m   mean of normal, values are changed by call
     * @param cv  var/cov of normal, values are changed by call
     */
    public static void getMVNormalParametersFromMVLognormal(double[] mu, double[][] cov, double[] m, double[][] cv) {
        for (int i = 0; i < mu.length; i++) {
            m[i] = Math.log(mu[i] * mu[i] / Math.signum(mu[i] * mu[i] + cov[i][i]));
            for (int j = 0; j < mu.length; j++) {
                cv[i][j] = Math.log(1.0 + (cov[i][j] / Math.abs(mu[i] * mu[j])));
            }
        }
    }

    /** Computes the multi-variate lognormal mean and var/cov
     *  from the mean and var/cov of the multi-variate normal
     *
     * @param mu  mean of log normal, values are changed by call
     * @param cov  var/cov of log normal, values are changed by call
     * @param m   mean of normal, values
     * @param cv  var/cov of normal
     */
    public static void getMVLogNormalParametersFromMVNormal(double[] mu, double[][] cov, double[] m, double[][] cv) {
        for (int i = 0; i < mu.length; i++) {
            mu[i] = Math.exp(m[i] + cv[i][i] / 2.0);
            for (int j = 0; j < mu.length; j++) {
                cov[i][j] = (Math.exp(cv[i][j]) - 1.0) * Math.exp(m[i] + m[j] + ((cv[i][i] + cv[j][j]) / 2.0));
            }
        }
    }

    /** Gets the means of the lognormals
     *
     * @return
     */
    public double[] getMeans() {
        int n = getDimension();
        double[] mu = new double[n];
        for (int i = 0; i < n; i++) {
            mu[i] = getMean(i);
        }
        return mu;
    }

    /** Gets the ith mean of the lognormal
     *
     * @param i
     * @return
     */
    public double getMean(int i) {
        double m = myMVN.getMean(i);
        double v = myMVN.getVariance(i);
        return Math.exp(m + v / 2.0);
    }

    /** Gets the covariances of the lognormals
     *
     * @return
     */
    public double[][] getCovariances() {
        int n = getDimension();
        double[][] cov = new double[n][n];
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                cov[i][j] = getCovariance(i, j);
            }
        }
        return cov;
    }

    /** Gets the ith, jth covariances of the lognormals
     *
     * @param i
     * @param j
     * @return
     */
    public double getCovariance(int i, int j) {
        double mi = myMVN.getMean(i);
        double mj = myMVN.getMean(j);
        double vi = myMVN.getVariance(i);
        double vj = myMVN.getVariance(j);
        double cvij = myMVN.getCovariance(i, j);
        return (Math.exp(cvij) - 1.0) * Math.exp(mi + mj + ((vi + vj) / 2.0));
    }

    /** Gets the ith variance
     *  Indexing is 0 based
     *
     * @param i
     * @return
     */
    public double getVariance(int i) {
        return getCovariance(i, i);
    }

    /** Gets the ith, jth correlation
     *  Indexing is 0 based
     *
     * @param i
     * @param j
     * @return
     */
    public double getCorrelation(int i, int j) {
        double vi = getVariance(i);
        double vj = getVariance(j);
        return (getCovariance(i, j) / Math.sqrt(vi * vj));
    }

    /** Gets an array representing the correlations
     *
     * @return
     */
    public double[][] getCorrelations() {
        int n = getDimension();
        double[][] rho = new double[n][n];
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                rho[i][j] = getCorrelation(i, j);
            }
        }
        return rho;
    }

    public final int getDimension() {
        return myMVN.getDimension();
    }

    public void setRandomNumberGenerator(RngIfc rng) {
        myMVN.setRandomNumberGenerator(rng);
    }

    public RngIfc getRandomNumberGenerator() {
        return myMVN.getRandomNumberGenerator();
    }

    public void setAntithetic(boolean flag) {
        myMVN.setAntithetic(flag);
    }

    public void resetStartSubstream() {
        myMVN.resetStartSubstream();
    }

    public void resetStartStream() {
        myMVN.resetStartStream();
    }

    public void resetNextSubstream() {
        myMVN.resetNextSubstream();
    }

    public double[] getValues() {
        return getValues(new double[getDimension()]);
    }

    public double[] getValues(double[] x) {
        x = myMVN.getValues(x);
        for (int i = 0; i < x.length; i++) {
            x[i] = Math.exp(x[i]);
        }
        return x;
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("Multi-variate lognormal\n");
        sb.append("Mean parameter vector\n");
        sb.append("{");
        int n = getDimension();
        for (int j = 0; j < n; j++) {
            sb.append(getMean(j));
            if (j < n - 1) {
                sb.append(",");
            }
        }
        sb.append("}\n");
        sb.append("\nVariance-Covariance Matrix\n");
        sb.append("{\n");
        for (int i = 0; i < n; i++) {
            sb.append("{");
            for (int j = 0; j < n; j++) {
                sb.append(getCovariance(i, j));
                if (j < n - 1) {
                    sb.append(",");
                }
            }
            sb.append("}\n");
        }
        sb.append("}\n");
        sb.append("Underlying MVNormal\n");
        sb.append(myMVN);
        return sb.toString();
    }

    public static void main(String args[]) {

        //test1();

    //test2();

           test4();
    }

    public static void test3(){
        double[] m = {
            0.561852,
            0.098666,
            0.569733,
            0.292096,
            0.81027,
            0.668751
        };

        double[][] cv = {
            {0.06061, 0.10232, 0.01393, 0.0487, -0.02716, -0.01813},
            {0.10232, 0.21135, 0.02063, 0.07194, -0.04022, -0.02444},
            {0.01393, 0.02063, 0.03077, 0.05071, -0.03301, -0.05279},
            {0.0487, 0.07194, 0.05071, 0.13336, -0.06497, -0.06917},
            {-0.02716, -0.04022, -0.03301, -0.06497, 0.04847, 0.06989},
            {-0.01813, -0.02444, -0.05279, -0.06917, 0.06989, 0.16012}
        };

        MVLogNormal mvln = new MVLogNormal(m, cv);

        System.out.println(mvln);

        System.out.println();

        System.out.println("Generating values");
        double[] x = new double[m.length];

        for (int i = 1; i <= 1000; i++) {
            x = mvln.getValues(x);
            for (int j = 0; j < x.length; j++) {
                if (x[0] <= 1.0) {
                    System.out.print(x[j]);
                    if (j < x.length - 1) {
                        System.out.print(",");
                    }
                }
            }
            if (x[0] <= 1.0)
                 System.out.println();
        }
    }

       public static void test4(){
           double[] m = {
               -0.3692,
               0.0987,
               0.1814,
               0.6688,
               -0.3085,
               0.2921};

           double[][] cv = {
               {0.30996, 0.23520, -0.12874, -0.04590, 0.07736, 0.11429},
               {0.23520, 0.21135, -0.08015, -0.02444, 0.04524, 0.07194},
               {-0.12874, -0.08015, 0.18524, 0.13703, -0.14026, -0.12632},
               {-0.04590, -0.02444, 0.13703, 0.16012, -0.11959, -0.06917},
               {0.07736, 0.04524, -0.14026, -0.11959, 0.15766, 0.11555},
               {0.11429, 0.07194, -0.12632, -0.06917, 0.11555, 0.13336}
           };

        MVLogNormal mvln = new MVLogNormal(m, cv);

        System.out.println(mvln);

        System.out.println();

        System.out.println("Generating values");
        double[] x = new double[m.length];

        for (int i = 1; i <= 1000; i++) {
            x = mvln.getValues(x);
            for (int j = 0; j < x.length; j++) {
                if (x[0] <= 1.0) {
                    System.out.print(x[j]);
                    if (j < x.length - 1) {
                        System.out.print(",");
                    }
                }
            }
            if (x[0] <= 1.0)
                 System.out.println();
        }
    }

    public static void test1() {
        // define the mvn parameters
        double[] m = {6.588, 2.974, 5.552, 2.026};
        double[][] cv = {
            {0.40434286, 0.09376327, 0.30328980, 0.04909388},
            {0.09376327, 0.10400408, 0.07137959, 0.04762857},
            {0.30328980, 0.07137959, 0.30458776, 0.04882449},
            {0.04909388, 0.04762857, 0.04882449, 0.07543265}
        };

        MVNormal mvn = new MVNormal(m, cv);
        System.out.println(mvn);

        System.out.println();
        System.out.println();

        MVLogNormal mvln = new MVLogNormal(m, cv);

        System.out.println(mvln);

        System.out.println();

        System.out.println("Generating values");
        double[] x = new double[m.length];

        for (int i = 1; i <= 10; i++) {
            x = mvln.getValues(x);
            for (int j = 0; j < x.length; j++) {
                System.out.print(x[j]);
                if (j < x.length - 1) {
                    System.out.print(",");
                }
            }
            System.out.println();
        }
    }

    public static void test2() {
        // define the mvn parameters
        double[] m = {6.588, 2.974, 5.552, 2.026};
        double[][] cv = {
            {0.40434286, -0.09376327, 0.30328980, 0.04909388},
            {-0.09376327, 0.10400408, 0.07137959, 0.04762857},
            {0.30328980, 0.07137959, 0.30458776, 0.04882449},
            {0.04909388, 0.04762857, 0.04882449, 0.07543265}
        };

        System.out.println("Test if the matrix is symmetric positive definite");

        if (!MVNormal.isSymmetricPositiveDefinite(cv)) {
            System.out.println("The var/cov matrix is not symmetric positive definite");
        } else {
            MVNormal mvn = new MVNormal(m, cv);
            System.out.println(mvn);

            // define the arrays for the MVLN
            double[] mu = new double[m.length];
            double[][] cov = new double[m.length][m.length];

            MVLogNormal.getMVLogNormalParametersFromMVNormal(mu, cov, m, cv);
            MVLogNormal mvln = new MVLogNormal(mu, cov);

            System.out.println(mvln);

            System.out.println("Generating values");
            double[] x = new double[mu.length];

            for (int i = 1; i <= 10; i++) {
                x = mvln.getValues(x);
                for (int j = 0; j < x.length; j++) {
                    System.out.print(x[j]);
                    if (j < x.length - 1) {
                        System.out.print(",");
                    }
                }
                System.out.println();
            }
        }


    }
}
