/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 *
 * Copyright (c) Manuel D. Rossetti (rossetti@uark.edu)
 *
 * Contact:
 *	Manuel D. Rossetti, Ph.D., P.E.
 *	Department of Industrial Engineering
 *	University of Arkansas
 *	4207 Bell Engineering Center
 *	Fayetteville, AR 72701
 *	Phone: (479) 575-6756
 *	Email: rossetti@uark.edu
 *	Web: www.uark.edu/~rossetti
 *
 * This file is part of the JSL (a Java Simulation Library). The JSL is a framework
 * of Java classes that permit the easy development and execution of discrete event
 * simulation programs.
 *
 * The JSL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * The JSL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the JSL (see file COPYING in the distribution);
 * if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA, or see www.fsf.org
 *
 */
package jslx.utilities.random;


import cern.colt.matrix.DoubleFactory1D;
import cern.colt.matrix.DoubleFactory2D;
import cern.colt.matrix.DoubleMatrix1D;
import cern.colt.matrix.DoubleMatrix2D;

import cern.colt.matrix.linalg.Algebra;
import cern.colt.matrix.linalg.CholeskyDecomposition;
import cern.jet.math.Functions;
import jsl.utilities.random.distributions.*;

import jsl.utilities.random.rng.RandomStreamIfc;
import jsl.utilities.random.rng.RngIfc;

import jsl.utilities.random.rng.RngStream;


/** Allows for the generation of multi-variate normal
 *  random variables.
 *
 *  This procedure depends upon the Cholesky decomposition of the var/cov
 *  matrix of the underlying multi-variate normal.  Because of
 *  this it is recommend that you check if the var/cov matrix of the underlying mvn
 *  is positive symmetric definite.
 *
 * @author rossetti
 */
public class MVNormal implements RandomStreamIfc {

    protected RngIfc myRNG;
    protected DoubleMatrix1D myMu;
    protected DoubleMatrix2D myCov;
    protected DoubleMatrix2D myL;
    protected DoubleMatrix1D myZ;

    /** Creates a standard MVN(0,I) of size n
     *
     * @param n
     */
    public MVNormal(int n){
        this(n, new RngStream());
    }

    /** Creates a standard MVN(0, I) of size n
     *
     * @param n
     * @param rng
     */
    public MVNormal(int n, RngIfc rng){
        if (n <= 0)
            throw new IllegalArgumentException("n must be > 0");

        double[] mu = new double[n];
        double[][] cov = new double[n][n];
        for(int i=0;i<n;i++){
            mu[i] = 0;
            for(int j=0;j<n;j++){
                if(i==j)
                    cov[i][j] = 1.0;
                else
                    cov[i][j] = 0.0;
            }
        }
        setParameters(mu, cov);
        setRandomNumberGenerator(rng);
    }

    /** Creates a multi-variate normal with given mean
     *  vector and variance/covariance matrix
     *
     * @param mu
     * @param cov
     */
    public MVNormal(double[] mu, double[][] cov){
        this(mu, cov, new RngStream());
    }

    /** Creates a multi-variate normal with given mean
     *  vector and variance/covariance matrix
     *
     * @param mu
     * @param cov
     * @param rng
     */
    public MVNormal(double[] mu, double[][] cov, RngIfc rng) {
        setParameters(mu, cov);
        setRandomNumberGenerator(rng);
    }

    /** Returns the multi-variate dimension
     *
     * @return
     */
    public int getDimension(){
        return myMu.size();
    }

    /** Sets the parameters of the multi-variate normal
     *  The mu.length must be equal to cov.length
     *  cov must be square. cov is assumed to be symmetric and
     *  positive definite. Only the lower half of cov is used
     *  in setting up the MVNormal
     *  The arrays are copied.
     *
     *  Throws an IllegalArgumentException if 
     *  - the covariance matrix is not symmetric positive definite
     *  - the length of the mean vector is not the same as
     *    the variance/covariance matrix
     *
     * @param mu a 1-d array of the mean values
     * @param cov a 2-d array representing the variance/covariance matrix
     */
    public void setParameters(double[] mu, double[][] cov) {
        if (mu.length != cov.length) {
            throw new IllegalArgumentException("The size of the mean array is not the same as the cov matrix");
        }

        myCov = DoubleFactory2D.dense.make(mu.length, mu.length);

        // guarantee symmetry
        for (int i = 0; i < cov.length; i++) {
            for (int j = 0; j < i + 1; j++) {
                myCov.set(i, j, cov[i][j]);
                myCov.set(j, i, cov[i][j]);
            }
        }

        // do the cholesky decomposition
        CholeskyDecomposition cd = new CholeskyDecomposition(myCov);

        if (!cd.isSymmetricPositiveDefinite()) {
            throw new IllegalArgumentException("The variance/covariance matrix was not symmetric positive definite");
        }

        myL = cd.getL();
        myMu = DoubleFactory1D.dense.make(mu);
        myZ = DoubleFactory1D.dense.make(mu);

    }

    /** Can be used to check if the supplied array (interpreted as a matrix)
     *  is  symmetric positive definite
     *
     * @param m
     * @return
     */
    public static boolean isSymmetricPositiveDefinite(double[][] m){
        DoubleMatrix2D c = DoubleFactory2D.dense.make(m);
        CholeskyDecomposition cd = new CholeskyDecomposition(c);
        return (cd.isSymmetricPositiveDefinite());
    }

    public double[] getValues(double[] x) {
        if (x.length < myMu.size()) {
            throw new IllegalArgumentException("The array must have at least " + myMu.size() + " elements");
        }

       for(int i=0;i<x.length;i++){
            myZ.set(i, Normal.stdNormalInvCDF(myRNG.randU01()));
        }

        Functions F = Functions.functions;
        DoubleMatrix1D y = Algebra.DEFAULT.mult(myL, myZ);
        y.assign(myMu, F.plus);
        y.toArray(x);
        return x;
    }

    public double[] getValues(){
        return (getValues(new double[myMu.size()]));
    }
    
    public void setAntithetic(boolean flag) {
        myRNG.setAntithetic(flag);
    }

    public void resetStartSubstream() {
        myRNG.resetStartSubstream();
    }

    public void resetStartStream() {
        myRNG.resetStartStream();
    }

    public void resetNextSubstream() {
        myRNG.resetNextSubstream();
    }

    /** Returns the means of the mvn
     *
     * @return
     */
    public double[] getMeans(){
        return myMu.toArray();
    }

    /** Gets the mean at index n
     *  The indexing is zero based,
     *  i.e. 0 is the first mean
     *
     * @param n
     * @return
     */
    public double getMean(int n){
        return myMu.get(n);
    }

    /** Returns the var/cov matrix of the mvn
     *
     * @return
     */
    public double[][] getCovariances(){
        return myCov.toArray();
    }

    /** Get the ith, jth covariance
     *  Indexing is 0 based
     *
     * @param i
     * @param j
     * @return
     */
    public double getCovariance(int i, int j){
        return myCov.get(i, j);
    }

    /** Gets the ith variance
     *  Indexing is 0 based
     *
     * @param i
     * @return
     */
    public double getVariance(int i){
        return getCovariance(i,i);
    }

    /** Gets the ith, jth correlation
     *  Indexing is 0 based
     *
     * @param i
     * @param j
     * @return
     */
    public double getCorrelation(int i, int j){
        double vi = getVariance(i);
        double vj = getVariance(j);
        return (getCovariance(i,j)/Math.sqrt(vi*vj));
    }

    /** Gets an array representing the correlations
     *
     * @return
     */
    public double[][] getCorrelations(){
        int n = getDimension();
        double[][] rho = new double[n][n];
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < n; j++) {
                rho[i][j] = getCorrelation(i,j);
            }
        }
        return rho;
    }

    /** Returns the distributions underlying random number generator
     *
     * @return
     */
    public RngIfc getRandomNumberGenerator() {
        return (myRNG);
    }

    /** Sets the underlying random number generator for the distribution
     * Throws a NullPointerException if rng is null
     * @param rng the reference to the random number generator
     */
    public void setRandomNumberGenerator(RngIfc rng) {
        if (rng == null) {
            throw new NullPointerException("RngIfc rng must be non-null");
        }
        myRNG = rng;
    }

    public String toString(){
         StringBuilder sb = new StringBuilder();
         sb.append("Multi-variate Normal\n");
         sb.append("Mean parameter vector\n");
         sb.append(myMu);
         sb.append("\nVariance-Covariance Matrix\n");
         sb.append(myCov);
         sb.append("\nCholesky Lower Triangular Matrix\n");
         sb.append(myL);
         return sb.toString();
    }

    public static void main(String args[]) {
        // define the parameters
        double[] mu = {6.588, 2.974, 5.552, 2.026};
        double[][] cov = {
            {0.40434286, 0.09376327, 0.30328980, 0.04909388},
            {0.09376327, 0.10400408, 0.07137959, 0.04762857},
            {0.30328980, 0.07137959, 0.30458776, 0.04882449},
            {0.04909388, 0.04762857, 0.04882449, 0.07543265}
        };

        MVNormal mvn = new MVNormal(mu, cov);
        System.out.println(mvn);
        System.out.println("Generating values");

        double[] x = new double[mu.length];

        for(int i=1;i<=10;i++){
            x = mvn.getValues(x);
            for(int j=0;j<x.length;j++){
                System.out.print(x[j]);
                if (j < x.length - 1)
                    System.out.print(",");
            }
            System.out.println();
        }
    }
}
