/*
* Created on Feb 16, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.vrnetwork;

import java.util.Iterator;
import java.util.LinkedHashMap;

import jsl.modeling.elements.spatial.CoordinateIfc;
import jsl.modeling.elements.spatial.SpatialElementIfc;

/**
 * @author rossetti
 *
 */
public class Route {
		
	private LinkedHashMap<SpatialElementIfc, VRTaskIfc> myTasks;
	
	private SpatialElementIfc myStartingLocation;
	private SpatialElementIfc myLastLocationAdded;
	private boolean myClosedFlag;
	private double myDistance;
	
	public Route(SpatialElementIfc startingLocation){
		setStartingLocation(startingLocation);
		myTasks = new LinkedHashMap<SpatialElementIfc, VRTaskIfc>();
		clear();
	}
	
	public void addTask(VRTaskIfc task){
		if (task == null)
			throw new IllegalArgumentException("The supplied task was null");
		if(myClosedFlag == true)
			throw new IllegalArgumentException("The route has been closed (back) to the starting location");
		// get the location associated with the task
		SpatialElementIfc location = task.getLocation();
		// check if trying to add last location
		if (location == myLastLocationAdded)
			throw new IllegalArgumentException("The location has already been added to the Route");
		// check if trying to add any previous location
		if (myTasks.containsKey(location))
			throw new IllegalArgumentException("The location has already been added to the Route.");
		myTasks.put(myLastLocationAdded, task);
		myDistance = myDistance + myLastLocationAdded.distanceTo(location);
		myLastLocationAdded = location;
		if (location == myStartingLocation)
			myClosedFlag = true;
	}
	
	public double getDistance(){
		return(myDistance);
	}
	
	public boolean isClosed(){
		return(myClosedFlag);
	}
	
	public CoordinateIfc getStartingLocationCoordinate(){
		return(myStartingLocation.getCurrentPosition());
	}
	
	public boolean isStartingPositionEqualTo(CoordinateIfc coordinate){
		return(myStartingLocation.isPositionEqualTo(coordinate));
	}
	
	public boolean isStartingLocation(SpatialElementIfc location){
		return(myStartingLocation == location);
	}
	
	public boolean isLastLocationAdded(SpatialElementIfc location){
		return(myLastLocationAdded == location);
	}
	
	public boolean isLinked(SpatialElementIfc origin, SpatialElementIfc destination){
		VRTaskIfc task = myTasks.get(origin);
		if (task == null)
			return false;
		if (destination == task.getLocation())
			return(true);
		else
			return(false);
	}
	
	public VRTaskIfc getTask(SpatialElementIfc location){
		return(myTasks.get(location));
	}
	
	public SpatialElementIfc getDestination(SpatialElementIfc origin){
		VRTaskIfc task = myTasks.get(origin);
		if (task != null)
			return(task.getLocation());
		else
			return(null);
	}
	
	public Iterator<VRTaskIfc> getTaskIterator(){
		return(myTasks.values().iterator());
	}

	public Iterator<SpatialElementIfc> getOriginIterator(){
		return(myTasks.keySet().iterator());
	}

	public boolean contains(SpatialElementIfc location) {
		return myTasks.containsKey(location);
	}
	
	public boolean contains(VRTaskIfc task){
		return(myTasks.containsValue(task));
	}

	public boolean isEmpty() {
		return myTasks.isEmpty();
	}

	public int getNumberOfTasks() {
		return myTasks.size();
	}
	
	public void clear(){
		myLastLocationAdded = myStartingLocation;
		myClosedFlag = false;
		myDistance = 0.0;
		myTasks.clear();
	}

	protected void setStartingLocation(SpatialElementIfc startingLocation){
		if (startingLocation == null)
			throw new IllegalArgumentException("The starting location was null for the Route.");
		myStartingLocation = startingLocation;
	}
	
	
}
