/*
* Created on Feb 16, 2007
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/
package jslx.vrnetwork;

import java.util.Iterator;

import jsl.modeling.elements.spatial.AbstractMover;
import jsl.modeling.elements.spatial.SpatialElementIfc;

/**
 * @author rossetti
 *
 */
public class Vehicle extends AbstractMover {
	
	//TODO need vehicle space limit
	//TODO need vehicle weight limit
	//TODO need to keep track of what is on the vehicle
	
	protected AbstractVRNetwork myVRNetwork;
	
	protected Route myRoute;
	
	protected VRTaskIfc myCurrentTask;
	
	protected Iterator<VRTaskIfc> myRouteIterator;
	
	protected LoadingCompleteListenerIfc myLoadingCompleteListener = new LoadingCompleted();
	
	protected UnloadingCompleteListenerIfc myUnloadingCompleteListener = new UnloadingCompleted();

	/**
	 * @param vrNetwork
	 * @param position
	 * @param name
	 */
	public Vehicle(AbstractVRNetwork vrNetwork, SpatialElementIfc position, String name) {
		super(vrNetwork, position, name);
		myVRNetwork = vrNetwork;
		// TODO Auto-generated constructor stub
	}
	
	/** Assigns a route to vehicle
	 * 
	 * @param route
	 */
	public void assignRoute(Route route){
		myRoute = route;
	}
	
	/** Tells the vehicle to start travelling the route
	 * 
	 *
	 */
	public void startRoute(){
	
		if (!isPositionEqualTo(myRoute.getStartingLocationCoordinate()))//TODO concerned about this check
			throw new IllegalStateException("The vehicle is not at the starting position of the route.");
		
		if (myRouteIterator != null)
			throw new IllegalStateException("The vehicle has already started its route.");
		myRouteIterator = myRoute.getTaskIterator();
		continueRoute();
	}
	
	protected void continueRoute(){
		if (myRouteIterator.hasNext()){
			myCurrentTask = myRouteIterator.next();
			SpatialElementIfc destination = myCurrentTask.getLocation();
			// starts a trip to the destination
			moveTo(destination);
		} else {// no more tasks on route, route is completed
			myVRNetwork.routeCompleted(this);
			myRoute = null;
			myCurrentTask = null;
			myCurrentTask = null;
		}
	}
	
	/**
	 * Use to invoke logic after a trip ends
	 */
	protected void afterTripEnds() {
		// get the load handler from the current task
		LoadHandlerIfc handler = myCurrentTask.getLoadHandler();
		if (handler !=null){// there is a handler
			VRLoadIfc dropOff = myCurrentTask.getDropOffLoad();
			if (dropOff != null) // do unloads first
				handler.unload(dropOff, myUnloadingCompleteListener);
			else {// no unloads, check for pick ups
				VRLoadIfc pickUp = myCurrentTask.getPickUpLoad();
				if (pickUp != null){
					handler.load(pickUp, myLoadingCompleteListener);
				} else {// no drop off or pick up
					continueRoute();// just continue route
				}
			}
		} else { // no load handler
			// there still might be drop offs and pick ups
			VRLoadIfc dropOff = myCurrentTask.getDropOffLoad();
			if (dropOff != null) // do unloads first
				dropOffLoad(dropOff);
			VRLoadIfc pickUp = myCurrentTask.getPickUpLoad();
			if (pickUp != null)
				pickUpLoad(pickUp);
			continueRoute();
		}
	}
	
	protected void dropOffLoad(VRLoadIfc load){
		//TODO update state of vehicle to reflect drop off
	}
	
	protected void pickUpLoad(VRLoadIfc load){
		//TODO update state of vehicle to reflect pick up
	}

	protected class UnloadingCompleted implements UnloadingCompleteListenerIfc {

		public void unloadingComplete() {
			LoadHandlerIfc handler = myCurrentTask.getLoadHandler();
			VRLoadIfc dropOff = myCurrentTask.getDropOffLoad();
			dropOffLoad(dropOff);
			VRLoadIfc pickUp = myCurrentTask.getPickUpLoad();
			if (pickUp != null){
				handler.load(pickUp, myLoadingCompleteListener);
			} else {// no pick up
				continueRoute();// just continue route
			}
		}
		
	}
	
	protected class LoadingCompleted implements LoadingCompleteListenerIfc {

		public void loadingComplete() {
			VRLoadIfc pickUp = myCurrentTask.getPickUpLoad();
			pickUpLoad(pickUp);
			continueRoute();
		}
		
	}

}
