/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package testing.modeling;

import java.util.HashMap;

import jsl.modeling.*;
import jsl.modeling.elements.variable.*;
import jsl.modeling.elements.queue.*;
import jsl.utilities.ControllableIfc;
import jsl.utilities.Controls;
import jsl.utilities.ResponseAndControlIfc;
import jsl.utilities.ResponseMakerIfc;
import jsl.utilities.Responses;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.statistic.StatisticAccessorIfc;


public class GGcModelWithControls extends SchedulingElement implements ResponseAndControlIfc {
    
    private Queue myWaitingQ;
    private int myNumServers;
    private RandomIfc myServiceDistribution;
    private RandomIfc myArrivalDistribution;
    private RandomVariable myServiceRV;
    private RandomVariable myArrivalRV;
    private TimeWeighted myNumBusy;
    private ResponseVariable myWs;
        
    private ArrivalListener myArrivalListener;
    private EndServiceListener myEndServiceListener;
        
    public GGcModelWithControls(ModelElement parent) {
        this(parent, 1, new Exponential(1.0), new Exponential(0.5));
    }
    
    public GGcModelWithControls(ModelElement parent, int numServers, RandomIfc ad, RandomIfc sd) {
    	super(parent);
        setNumberOfServers(numServers);
        setServiceDistributionInitialRandomSource(sd);
        setArrivalDistributionInitialRandomSource(ad);
        
        myWaitingQ = new Queue(this, "GGC Q1");
        myNumBusy = new TimeWeighted(this, 0.0, "NumBusy");
        myWs = new ResponseVariable(this, "WS");
        
        myArrivalListener = new ArrivalListener();
        myEndServiceListener = new EndServiceListener();
    }
    
    public int getNumberOfServers() {
        return(myNumServers);
    }
	
    protected void setNumberOfServers(int n) {
        if (n < 0)
            throw new IllegalArgumentException();
        myNumServers = n;
    }
    
    protected void setServiceDistributionInitialRandomSource(RandomIfc d) {
        if (d == null)
            throw new IllegalArgumentException("Service Time Distribution was null!");

		myServiceDistribution = d;
        
        if (myServiceRV == null){// not made yet
        	myServiceRV = new RandomVariable(this, myServiceDistribution,"Service RV");
        } else { // already had been made, and added to model
        	// just change the distribution
        	myServiceRV.setInitialRandomSource(myServiceDistribution);
        }
                
    }
    
    protected void setArrivalDistributionInitialRandomSource(RandomIfc d) {
		if (d == null)
			throw new IllegalArgumentException("Arrival Time Distribution was null!");

		myArrivalDistribution = d;
        
		if (myArrivalRV == null){// not made yet
			myArrivalRV = new RandomVariable(this, myArrivalDistribution, "Arrival RV");
		} else { // already had been made, and added to model
			// just change the distribution
			myArrivalRV.setInitialRandomSource(myArrivalDistribution);
		}
    }
    
    protected void initialize() {
        super.initialize();
        // start the arrivals
        scheduleArrival();
    }
    
    private void scheduleArrival() {
        scheduleEvent(myArrivalListener, myArrivalRV.getValue(), "Arrival");
    }
    
    private void serveNextCustomer(){
        myNumBusy.increment();
        QObject c = myWaitingQ.removeNext();
        scheduleEndService(c);
    }
    
    private void scheduleEndService(QObject c) {
    	scheduleEvent(myEndServiceListener, myServiceRV.getValue(), "End Service", c);
    }
    
    class ArrivalListener implements ActionListenerIfc {
        public void action(JSLEvent event) {
            myWaitingQ.enqueue(new Double(getTime()));
            if (myNumBusy.getValue() < myNumServers){
                serveNextCustomer();
            }
            scheduleArrival();
        }
    }
    
    class EndServiceListener implements ActionListenerIfc {
        public void action(JSLEvent event) {
            myNumBusy.decrement();
            QObject c = (QObject)event.getMessage();
            double ws = getTime() - c.getTimeEnteredQueue();
            myWs.setValue(ws);
            if (myWaitingQ.size() > 0 ) {
                serveNextCustomer();
            }
        }
    }

    class GGcModelControls extends Controls {
		protected GGcModelControls(ControllableIfc controllable) {
			super(controllable);
			//create the necessary maps
			myIntegerControls = new HashMap<String, Integer>();
			myRandomIfcControls = new HashMap<String, RandomIfc>();
			// fill the maps with the control values
			myIntegerControls.put("Num Servers", myNumServers);
			myRandomIfcControls.put("ArrivalCDF", myArrivalDistribution);
			myRandomIfcControls.put("ServiceCDF", myServiceDistribution);
		}
    }
    
	public Controls makeControls() {
		return new GGcModelControls(this);
	}

	class GGcResponse extends Responses {
		protected GGcResponse(ResponseMakerIfc responseMaker) {
			super(responseMaker);
			//create the maps
			myStatisticAccessorResponses = new HashMap<String, StatisticAccessorIfc>();
			// fill the maps
			myStatisticAccessorResponses.put("SystemTime", myWs.getAcrossReplicationStatistic());
			myStatisticAccessorResponses.put("NumBusy", myNumBusy.getAcrossReplicationStatistic());
		}
	}
		
	protected void useControls(){
		setNumberOfServers(getControls().getIntegerControl("Num Servers"));
		setArrivalDistributionInitialRandomSource(getControls().getRandomIfcControl("ArrivalCDF"));
		setServiceDistributionInitialRandomSource(getControls().getRandomIfcControl("ServiceCDF"));

		System.out.println("Setting up experiment");
		System.out.println("Controls:");
		System.out.println(getControls());		
	}
		
	public Responses makeResponses(){
		System.out.println("Making responses");
		return (new GGcResponse(this));		
	}

	public static void main(String[] args) {
		
		System.out.println("GGcWSModel Test");
        
		runExperiment();
        
		System.out.println("Done!");
        
	}
	
	public static void runExperiment(){
		// create the containing model
		Model m = Model.createModel();
		
		// create the model element and attach it to the main model
		GGcModelWithControls ggc = new GGcModelWithControls(m);
		
		Controls c = ggc.makeControls();
		System.out.println(c);
		
		c.changeIntegerControl("Num Servers", 1);
		
		ggc.setControls(c);
		
		// create the experiment to run the model
		Experiment e = new Experiment(m);
		
		// set the parameters of the experiment
		e.setNumberOfReplications(10);
		e.setLengthOfReplication(20000.0);
		e.setLengthOfWarmUp(5000.0);
//		e.turnOnExperimentReport();
//		e.turnOnReplicationReport();
		e.turnOnExperimentConsoleOutput();

		e.runAll();
		
		Responses r = ggc.makeResponses();
		
		System.out.println(r);

	}
}
