/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package testing.modeling;

import java.util.logging.Level;

import jsl.modeling.elements.processview.description.MacroCommand;
import jsl.modeling.elements.processview.description.ProcessDescription;
import jsl.modeling.elements.processview.description.commands.*;
import jsl.modeling.elements.variable.*;
import jsl.modeling.*;
import jsl.utilities.reporting.*;
import jsl.modeling.elements.resource.*;
import jsl.modeling.elements.queue.*;

/**
 *
 */
public class TestProcessView {

	public static void main(String[] args) {
		
		// turns on the level of logging
		// watch out FINEST is very verbose, e.g. every event
		JSL.LOGGER.setLevel(Level.FINEST);
		
		testProcessDescription();

//		testSubProcessDescription();

//		testTerminate();
		
//		testSeizeDelayRelease();
		
		System.out.println("done!");
	}

	public static void testSeizeDelayRelease(){
		
		// create the containing model
		Model m = Model.createModel();

		// create the process description, so that the model contains it
		
		ProcessDescription pd = new ProcessDescription(m, "TestDelays");

		// create the model elements used in the process description
		
		Variable v1 = new Variable(pd, 10.0, "delay1");
		Variable v2 = new Variable(pd, 15.0, "delay2");
		Variable v3 = new Variable(pd, 1.0, "amt R1");
		Variable v4 = new Variable(pd, 2.0, "amt R2");
		
		Resource r1 = new Resource(pd, "R1");
		Resource r2 = new Resource(pd, 2, "R2");
		Queue q1 = new Queue(pd, "R1 Queue");
		Queue q2 = new Queue(pd, "R2 Queue");
		
		// create the commands used in the process description

		Delay d1 = new Delay(pd, v1, "delay 1");

		Delay d2 = new Delay(pd, v2, "delay 2");
		
		Seize s1 = new Seize(pd, v3, r1, q1, "seize 1");

		Seize s2 = new Seize(pd, v4, r2, q2, "seize 2");
		
		Release rl1 = new Release(pd, r1, q1, "release 1");
		
		Release rl2	= new Release(pd, r2, q2, "release 2");
				
		// add the commands to the process description
		// order matters when adding commands
		
		pd.addProcessCommand(s1);
		pd.addProcessCommand(d1);
		pd.addProcessCommand(rl1);
		pd.addProcessCommand(s2);
		pd.addProcessCommand(d2);
		pd.addProcessCommand(rl2);	
			
		// turn on the auto start option for the process description
		// automatically creates 1 process executor and starts it
		// at the beginning of a replication, primarily for testing purposes
		pd.turnOnAutoStart();
		
		m.turnOnLogReport();
      
		// create a replication to run the model
		Replication r = new Replication(m);

		// set the running parameters of the replication
		r.setLengthOfReplication(500.0);
		
		// turn on the desired reporting, this gets scheduler trace report
		r.turnOnDefaultTraceReport();

		// no need for batching for this simple example
		r.setDefaultBatchingOption(false);
		
		// tell the replication to run
		r.runAll();
 
	}

	public static void testProcessDescription(){

		// create the containing model
		Model m = Model.createModel();

		// create the process description, so that the model contains it
		
		ProcessDescription pd = new ProcessDescription(m, "TestDelays");

		// create the model elements used in the process description
		
		Variable v1 = new Variable(pd, 10.0, "vdelay 10.0");
		Variable v2 = new Variable(pd, 15.0, "vdelay 15.0");
		Variable v3 = new Variable(pd, 20.0, "vdelay 20.0");
		Variable v4 = new Variable(pd, 25.0, "vdelay 25.0");
		Variable v5 = new Variable(pd, 30.0, "vdelay 30.0");
		Variable v6 = new Variable(pd, 35.0, "vdelay 35.0"); 
		
		// create the commands used in the process description

		Delay d1 = new Delay(pd, v1, "Delay 1");
		Delay d2 = new Delay(pd, v2, "Delay 2");
		Delay d3 = new Delay(pd, v3, "Delay 3");
		Delay d4 = new Delay(pd, v4, "Delay 4");
		Delay d5 = new Delay(pd, v5, "Delay 5");
		Delay d6 = new Delay(pd, v6, "Delay 6");
		
		// add the commands to the process description
		// order matters when adding commands
		
		pd.addProcessCommand(d1);
		pd.addProcessCommand(d2);
		pd.addProcessCommand(d3);
		pd.addProcessCommand(d4);
		pd.addProcessCommand(d5);
		pd.addProcessCommand(d6);
//		pd.addProcessCommand(new Terminate(pd));
		
		// turn on the auto start option for the process description
		// automatically creates 1 process executor and starts it
		// at the beginning of a replication, primarily for testing purposes
		pd.turnOnAutoStart();
		
		m.turnOnLogReport();
      
		// create a replication to run the model
		Replication r = new Replication(m);

		// set the running parameters of the replication
//		r.setLengthOfReplication(500.0);
		
		// turn on the desired reporting, this gets scheduler trace report
		r.turnOnDefaultTraceReport();

		// no need for batching for this simple example
		r.setDefaultBatchingOption(false);
		
		// tell the replication to run
		r.runAll();
 
	}


	public static void testSubProcessDescription(){
		
		// create the containing model
		Model m = Model.createModel();

		// create the process description, so that the model contains it
		
		ProcessDescription pd = new ProcessDescription(m, "TestDelays");
		
		// create the model elements used in the process descriptions
		
		Variable v1 = new Variable(pd, 10.0, "delay1");
		Variable v2 = new Variable(pd, 15.0, "delay2");
		Variable v3 = new Variable(pd, 20.0, "delay3");
		
		// create the commands used in the process descriptions

		Delay d1 = new Delay(pd, v1);
		Delay d2 = new Delay(pd, v2);
		Delay d3 = new Delay(pd, v3);
		
		// create a macro command
		
		MacroCommand mc = new MacroCommand(pd, "DelayMacro");

		// create the variable and command used in the macrocommand
		Variable v4 = new Variable(mc, 25.0, "delay4");
		Variable v5 = new Variable(mc, 30.0, "delay5");
		Variable v6 = new Variable(mc, 35.0, "delay6"); 

		Delay d4 = new Delay(mc, v4);
		Delay d5 = new Delay(mc, v5);
		Delay d6 = new Delay(mc, v6);

		// add the subcommands to the macro
		mc.addSubCommand(d4);
		mc.addSubCommand(d5);
		mc.addSubCommand(d6);
		
		// add the commands to the process description
		// order matters when adding commands
		
		pd.addProcessCommand(d1);
		pd.addProcessCommand(mc);
		pd.addProcessCommand(d2);
		pd.addProcessCommand(d3);
		
		// turn on the auto start option for the process description
		// automatically creates 1 process executor and starts it
		// at the beginning of a replication, primarily for testing purposes
		pd.turnOnAutoStart();
		
		m.turnOnLogReport();
      
		// create a replication to run the model
		Replication r = new Replication(m);

		// set the running parameters of the replication
		r.setLengthOfReplication(500.0);
		
		// turn on the desired reporting, this gets scheduler trace report
		r.turnOnDefaultTraceReport();

		// no need for batching for this simple example
		r.setDefaultBatchingOption(false);
		
		// tell the replication to run
		r.runAll();
 
	}

	public static void testTerminate(){
		
		// create the containing model
		Model m = Model.createModel();

		// create the process description, so that the model contains it
		
		ProcessDescription pd = new ProcessDescription(m, "TestDelays");

		// create the model elements used in the process description
		
		Variable v1 = new Variable(pd, 10.0, "delay1");
		Variable v2 = new Variable(pd, 15.0, "delay2");
		Variable v3 = new Variable(pd, 20.0, "delay3");

		
		// create the commands used in the process description

		Delay d1 = new Delay(pd, v1);
		Delay d2 = new Delay(pd, v2);
		Delay d3 = new Delay(pd, v3);

		Terminate t2 = new Terminate(pd);
		
		MacroCommand mc = new MacroCommand(pd, "DelayMacro");

		Variable v4 = new Variable(mc, 25.0, "delay4");
		Variable v5 = new Variable(mc, 30.0, "delay5");
		Variable v6 = new Variable(mc, 35.0, "delay6"); 
		
		Delay d4 = new Delay(mc, v4);
		Delay d5 = new Delay(mc, v5);
		Delay d6 = new Delay(mc, v6);

		Terminate t1 = new Terminate(mc);
		
		// add the subcommands to the macro
		mc.addSubCommand(d4);
		mc.addSubCommand(d5);
		mc.addSubCommand(t1);
		mc.addSubCommand(d6);
		
		// add the commands to the process description
		// order matters when adding commands
		
		pd.addProcessCommand(d1);
		pd.addProcessCommand(mc);
		pd.addProcessCommand(d2);
		pd.addProcessCommand(t2);
		pd.addProcessCommand(d3);
		
		// turn on the auto start option for the process description
		// automatically creates 1 process executor and starts it
		// at the beginning of a replication, primarily for testing purposes
		pd.turnOnAutoStart();
		
		m.turnOnLogReport();
      
		// create a replication to run the model
		Replication r = new Replication(m);

		// set the running parameters of the replication
		r.setLengthOfReplication(500.0);
		
		// turn on the desired reporting, this gets scheduler trace report
		r.turnOnDefaultTraceReport();

		// no need for batching for this simple example
		r.setDefaultBatchingOption(false);
		
		// tell the replication to run
		r.runAll();
 
	}
	
}
