package testing.parameters;


import jsl.modeling.*;
import jsl.modeling.elements.variable.*;
import testing.parameters.ExperimentableIfc;
import testing.parameters.Responses;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;

public class DriverLicenseBureauParameterized extends SchedulingElement implements ExperimentableIfc{

	private int myNumServers;
	private DistributionIfc myServiceDistribution;
	private DistributionIfc myArrivalDistribution;
	private RandomVariable myServiceRV;
	private RandomVariable myArrivalRV;
	private TimeWeighted myNumBusy;
	private TimeWeighted myNQ;
	private TimeWeighted myNS;
	
	private ArrivalListener myArrivalListener;
	private EndServiceListener myEndServiceListener;
	private Parameters p;
	
	public DriverLicenseBureauParameterized(ModelElement parent) {
		this(parent,new Exponential(1.0), new Exponential(0.5));
	}

	public DriverLicenseBureauParameterized(ModelElement parent,DistributionIfc ad, DistributionIfc sd) {
		super(parent);
		setServiceDistributionInitialRandomSource(sd);
		setArrivalDistributionInitialRandomSource(ad);

		myNumBusy = new TimeWeighted(this, 0.0, "NumBusy");

		myNQ = new TimeWeighted(this, 0.0, "NQ");

		myNS = new TimeWeighted(this, 0.0, "NS");

		myArrivalListener = new ArrivalListener();
		myEndServiceListener = new EndServiceListener();
		p=new DriversLicenseBureauParameters(this);
		//this.setParameters(parameters);
	}

	public int getNumberOfServers() {
		return p.getInt("numServers");
	}

	public void setServiceDistributionInitialRandomSource(DistributionIfc d) {

		if (d == null)
			throw new IllegalArgumentException("Service Time Distribution was null!");

		myServiceDistribution = d;

		if (myServiceRV == null) { // not made yet
			myServiceRV = new RandomVariable(this, myServiceDistribution, "Service RV");
		} else { // already had been made, and added to model
			// just change the distribution
			myServiceRV.setInitialRandomSource(myServiceDistribution);
		}

	}

	public void setArrivalDistributionInitialRandomSource(DistributionIfc d) {

		if (d == null)
			throw new IllegalArgumentException("Arrival Time Distribution was null!");

		myArrivalDistribution = d;

		if (myArrivalRV == null) { // not made yet
			myArrivalRV = new RandomVariable(this, myArrivalDistribution, "Arrival RV");
		} else { // already had been made, and added to model
			// just change the distribution
			myArrivalRV.setInitialRandomSource(myArrivalDistribution);
		}
	}
	
	protected void initialize() {
		super.initialize();
		myNumServers=p.getInt("numServers");
		// start the arrivals
		scheduleEvent(myArrivalListener, myArrivalRV.getValue(), "Arrival");
	}
	
	class ArrivalListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			
			myNS.increment(); // new customer arrived
			if (myNumBusy.getValue() < myNumServers) { // server available
				myNumBusy.increment(); // make server busy
				//	schedule end of service
				scheduleEvent(myEndServiceListener, myServiceRV.getValue(), "End Service"); 
			} else { // no server available
				myNQ.increment(); // place customer in queue
			}
			
			//	always schedule the next arrival
			scheduleEvent(myArrivalListener, myArrivalRV.getValue(), "Arrival"); 
		}
	}

	class EndServiceListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
			myNS.decrement(); // customer departed
			myNumBusy.decrement(); // customer is leaving server is freed

			if (myNQ.getValue() > 0) { // queue is not empty
				myNQ.decrement(); // remove from queue
				myNumBusy.increment(); // make server busy
				//	schedule end of service
				scheduleEvent(myEndServiceListener, myServiceRV.getValue(), "End Service");
			}
		}
	}

	public Parameters getParameters() {

		return p;
	}

	public void setParameters(Parameters parameters) {
		if(parameters.getParameterized()!=this)
			throw new IllegalArgumentException("parameters does not belong to "+this.getName());
		if(parameters==null)
			throw new IllegalArgumentException("parameters was null");
		p.setInt("numServers",parameters.getInt("numServers"));
		this.myNumServers = parameters.getInt("numServers");
	}
	
	public void setResponses(String[] values){
		//this.getModel().getModelElement(id)
	}
	
	public double[] getResponses(String[] responseNames) {
		double[] responses=new double[responseNames.length];
		int i=0;
		for(String responseName:responseNames)
			responses[i++]=((ResponseVariable)this.getModel().getModelElement(responseName)).getAcrossReplicationAverage();
		return responses;
	}
	
	public static void main(String[] args) {
		Model m=Model.createModel();
		DriverLicenseBureauParameterized model=new DriverLicenseBureauParameterized(m);
		DriversLicenseBureauParameters params=new DriversLicenseBureauParameters(model);
		Experiment e = new Experiment(m);
		ResponseVariable rv=(ResponseVariable)m.getModelElement("NQ");
		// set the parameters of the experiment
		e.setNumberOfReplications(2);
		e.setLengthOfReplication(20000.0);
		e.setLengthOfWarmUp(5000.0);
		for(int i=1;i<=10;i++){
			params.setInt("numServers", i);
			model.setParameters(params);
			e.runAll();
			System.out.println(i+"\t"+rv.getAcrossReplicationAverage());
		}
	}

	public Responses getResponses() {
		// TODO Auto-generated method stub
		return null;
	}

	public void runModel() {
		// TODO Auto-generated method stub
		
	}
}
