/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package testing.parameters;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Collections;

/** This class acts like a Map to allow named parameters and
 *  their associated values to be viewed and set.  This is similar
 *  in concept to Java's Preferences class
 *  
 *  Implementors of ParameterizedIfc are responsible for making
 *  sub-classes of this class that are filled appropriately
 * 
 *
 */
public abstract class Parameters {

	/** The ParameterizedIfc that uses the Parameters
	 * 
	 */
	protected ParameterizedIfc myParameterized;
	
	/** The Map that hold the parameters as string pairs
	 *  key = name of parameter
	 *  value = value of the parameter as a String
	 * 
	 */
	protected Map<String, String> myParameters;
	
	/**
	 * 
	 */
	public Parameters(ParameterizedIfc parameterized) {
		if (parameterized == null)
			throw new IllegalArgumentException("The ParameterizedIfc was null");
		myParameterized = parameterized;
		myParameters = new HashMap<String, String>();
	}
	
	/**
	 * 
	 * @return
	 */
	public ParameterizedIfc getParameterized(){
		return myParameterized;
	}

	/** Gets the value associated with the supplied key as a String.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @return
	 */
	public String getString(String key) {
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");

		return myParameters.get(key);
	}

	/** Sets the value of the key to the supplied value.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @param value
	 */
	public String setString(String key, String value) {
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");
		
		return myParameters.put(key, value);
	}

	/** Gets the value associated with the supplied key as a double.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @return
	 */
	public double getDouble(String key){

		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");

		return Double.parseDouble(myParameters.get(key));
	}
	
	/** Sets the value of the key to the supplied value.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key key with which the string form of value is to be associated
	 * @param value value whose string form is to be associated with key
	 */
	public void setDouble(String key, double value){
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");
	
		myParameters.put(key, Double.toString(value));
	}

	/** Gets the value associated with the supplied key. The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @return
	 */
	public int getInt(String key){
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");

		return Integer.parseInt(myParameters.get(key));
	}
	
	/** Sets the value of the key to the supplied value.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @param value
	 */
	public void setInt(String key, int value){
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");
		
		myParameters.put(key, Integer.toString(value));
	}
	
	/** Gets the value associated with the supplied key.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @return
	 */
	public long getLong(String key){
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");

		return Long.parseLong(myParameters.get(key));
	}
	
	/** Sets the value of the key to the supplied value.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @param value
	 */
	public void setLong(String key, long value){
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");
		
		myParameters.put(key, Long.toString(value));
	}

	/** Gets the value associated with the supplied key.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @return
	 */
	public float getFloat(String key){
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");

		return Float.parseFloat(myParameters.get(key));
	}
	
	/** Sets the value of the key to the supplied value.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @param value
	 */
	public void setFloat(String key, float value){
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");
		
		myParameters.put(key, Float.toString(value));
	}

	/** Gets the value associated with the supplied key.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @return
	 */
	public boolean getBoolean(String key){
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");

		return Boolean.parseBoolean(myParameters.get(key));
	}
	
	/** Sets the value of the key to the supplied value.  The key must already
	 *  exist in the parameters and cannot be null.
	 * 
	 * @param key
	 * @param value
	 */
	public void setBoolean(String key, boolean value){
		if (key == null)
			throw new IllegalArgumentException("The supplied key cannot be null");
		
		if (!myParameters.containsKey(key))
			throw new IllegalArgumentException("The supplied key is not associated with a parameter value");
		
		myParameters.put(key, Boolean.toString(value));
	}
	
	/** Checks if the supplied key is contained in the Parameters
	 * 
	 * @param key
	 * @return
	 */
	public boolean containsKey(String key){
		return myParameters.containsKey(key);
	}
	
	/** Returns an unmodifiable Set of the Parameter's keys
	 * 
	 * @return
	 */
	public Set<String> keySet(){
		return Collections.unmodifiableSet(myParameters.keySet());
	}
	
	public String toString(){
		StringBuilder sb = new StringBuilder();
		sb.append("ID \t " + myParameterized.getId() + "\n");
		sb.append("Name \t " + myParameterized.getName() + "\n");
		sb.append(myParameters.toString());
		return sb.toString();
	}

}
