/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package testing.variable;

import java.util.logging.Level;

import jsl.modeling.*;
import jsl.modeling.elements.variable.AggregateTimeWeightedVariable;
import jsl.modeling.elements.variable.RandomVariable;
import jsl.modeling.elements.variable.TimeWeighted;
import jsl.modeling.elements.variable.Variable;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.Bernoulli;
import jsl.utilities.random.distributions.Constant;
import jsl.utilities.random.distributions.DUniform;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.reporting.JSL;

/**
 */
public class TestAggregation extends SchedulingElement {

	private Variable myTBA1;
	private Variable myTBA2;
	private Variable myTBA3;
	private Variable myC1;
	private Variable myC2;
	private Variable myC3;
	private RandomVariable mySign;
	
	private TimeWeighted myTW1;
	private TimeWeighted myTW2;
	private TimeWeighted myTW3;
	
	private AggregateTimeWeightedVariable myATW1;
	private AggregateTimeWeightedVariable myATW2;
	
	/**
	 * @param parent
	 */
	public TestAggregation(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public TestAggregation(ModelElement parent, String name) {
		super(parent, name);
		
		//DistributionIfc a1 = new Exponential(1.0);
		DistributionIfc a1 = new Constant(1.0);
		myTBA1 = new RandomVariable(this, a1, "tba1");
		
		//DistributionIfc a2 = new Exponential(1.0);
		DistributionIfc a2 = new Constant(2.0);
		myTBA2 = new RandomVariable(this, a2, "tba2");

		//DistributionIfc a3 = new Exponential(1.0);
		DistributionIfc a3 = new Constant(2.5);
		myTBA3 = new RandomVariable(this, a3, "tba3");
		
		DistributionIfc a4 = new DUniform(1, 5);
		myC1 = new RandomVariable(this, a4, "c1");
		myC2 = new RandomVariable(this, a4, "c2");
		myC3 = new RandomVariable(this, a4, "c3");
		mySign = new RandomVariable(this, new Bernoulli(1.0), "sign");

	    myTW1 = new TimeWeighted(this, "tw1");
	    myTW2 = new TimeWeighted(this, "tw2");
	    myTW3 = new TimeWeighted(this, "tw3");
//	    myTW1.turnOnTrace();

	    myATW1 = new AggregateTimeWeightedVariable(this, "atw1");
	    myATW1.subscribeTo(myTW1);
	    myATW1.subscribeTo(myTW2);
	    
	    myATW2 = new AggregateTimeWeightedVariable(this, "atw2");
	    myATW2.subscribeTo(myTW1);
	    myATW2.subscribeTo(myTW3);
	    
		//myCountArrivals.setTimedUpdateInterval(10.0);
	}
	
	protected void w1Change(){
		double x = myC1.getValue();
		if (mySign.getValue() < 1.0){
			x = -x;
		}
		myTW1.setValue(x);
//	    System.out.println("time: " + getTime() + " tw1 = " + myTW1.getValue());	       
	}
	
	protected void w2Change(){
		double x = myC2.getValue();
		if (mySign.getValue() < 1.0){
			x = -x;
		}
		myTW2.setValue(x);
//	    System.out.println("time: " + getTime() + " tw2 = " + myTW2.getValue());
	}

	protected void w3Change(){
		double x = myC3.getValue();
		if (mySign.getValue() < 1.0){
			x = -x;
		}
		myTW3.setValue(x);
//	    System.out.println("time: " + getTime() + " tw3 = " + myTW3.getValue());
	}
	
	public void beforeExperiment(){
//	    System.out.println("t \t tw1 \t tw2 \t tw3 \t atw1 \t atw2");
	}
/*	
	public void afterExperiment(){
		double w1 = myTW1.getAcrossReplicationAverage();
		double w2 = myTW2.getAcrossReplicationAverage();
		double w3 = myTW3.getAcrossReplicationAverage();
		double aw1 = myATW1.getAcrossReplicationAverage();
		double aw2 = myATW2.getAcrossReplicationAverage();

	}
*/
	public void initialize(){
		scheduleEvent(myTBA1, 1);
		scheduleEvent(myTBA2, 2);
		scheduleEvent(myTBA3, 3);
	}
	
	protected void handleEvent(JSLEvent e){
		int type = e.getType();
		if (type == 1){
			w1Change();
			rescheduleEvent(e, myTBA1);
		} else if (type == 2){
			w2Change();
			rescheduleEvent(e, myTBA2);			
		} else {
			w3Change();
			rescheduleEvent(e, myTBA3);			
		}	
//	    System.out.println(getTime() + " \t " + myTW1.getValue() + " \t " + myTW2.getValue() + " \t " + myTW3.getValue() + " \t " + myATW1.getValue() + " \t "  + myATW2.getValue());
//	    System.out.println("time: " + getTime() + " atw1 = " + myATW1.getValue());
//	    System.out.println("time: " + getTime() + " atw2 = " + myATW2.getValue());
 	}
	
	public static void main(String[] args) {
		System.out.println("TestVariables test");
        
//		JSL.LOGGER.setLevel(Level.FINE);
   
//		testReplication();
       testExperiment();
        
		System.out.println("Done!");
	}

	public static void testReplication(){
		// create the containing model
		Model m = Model.createModel();
		
		// create the model element and attach it to the main model
		TestAggregation tv = new TestAggregation(m);
		        
		// create a replication to run the model
		Replication r = new Replication(m);

		// set the running parameters of the replication
		r.setLengthOfReplication(50.0);
//		r.setLengthOfWarmUp(10.0);
		
		// turn on the desired reporting
//		r.turnOnBatchReport();
		
				
		// tell the replication to run
		r.runAll();
 
	}

	public static void testExperiment(){
		// create the containing model
		Model m = Model.createModel();
		
		// create the model element and attach it to the main model
		TestAggregation tv = new TestAggregation(m);
		        
		// create a replication to run the model
		Experiment e = new Experiment(m);

		// set the running parameters of the experiment
		e.setNumberOfReplications(5);
		e.setLengthOfReplication(50.0);
		e.setLengthOfWarmUp(10.0);
//		e.setDefaultBatchingOption(true);

		// turn on the desired reporting
//		e.turnOnBatchReport();
		e.turnOnExperimentReport();
		e.turnOnExperimentConsoleOutput();
		e.turnOnReplicationReport();
				
		// tell the replication to run
		e.runAll();
 
	}
	

}
