/*
* Copyright (c) 2007, Manuel D. Rossetti (rossetti@uark.edu)
*
* Contact:
*	Manuel D. Rossetti, Ph.D., P.E. 
*	Department of Industrial Engineering 
*	University of Arkansas 
*	4207 Bell Engineering Center 
*	Fayetteville, AR 72701 
*	Phone: (479) 575-6756 
*	Email: rossetti@uark.edu 
*	Web: www.uark.edu/~rossetti
*
* This file is part of the JSL (a Java Simulation Library). The JSL is a framework
* of Java classes that permit the easy development and execution of discrete event
* simulation programs.
*
* The JSL is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* The JSL is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with the JSL (see file COPYING in the distribution); 
* if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, 
* Boston, MA  02110-1301  USA, or see www.fsf.org
* 
*/

package testing.variable;

import java.io.IOException;
import java.util.logging.Level;

import jsl.modeling.*;
import jsl.modeling.elements.variable.*;
import jsl.observers.textfile.CSVExperimentReport;
import jsl.observers.textfile.CSVReplicationReport;
import jsl.observers.textfile.CSVReport;
import jsl.utilities.random.*;
import jsl.utilities.random.distributions.Bernoulli;
import jsl.utilities.random.distributions.DistributionIfc;
import jsl.utilities.random.distributions.Exponential;
import jsl.utilities.reporting.JSL;

/**
 */
public class TestUpDownComponent extends SchedulingElement {

	private Variable myHeight;
	private RandomVariable myUpTime;
	private RandomVariable myDownTime;
	private RandomVariable myInitialState;
	private TimeWeighted myState;
	private ResponseVariable myCycleLength;
	private Counter myCountFailures;
	private TimeWeighted myCountF;
	private UpChangeListener myUpChangeListener = new UpChangeListener();
	private DownChangeListener myDownChangeListener = new DownChangeListener();
	private double myTimeLastUp;
	private double myCount;
//	private ResponseVariable myTest;
	
	/**
	 * @param parent
	 */
	public TestUpDownComponent(ModelElement parent) {
		this(parent, null);
	}

	/**
	 * @param parent
	 * @param name
	 */
	public TestUpDownComponent(ModelElement parent, String name) {
		super(parent, name);

		myHeight = new Variable(this, 1.0, "height");
		DistributionIfc utd = new Exponential(1.0);
		DistributionIfc dtd = new Exponential(2.0);
		myUpTime = new RandomVariable(this, utd, "up time");
		myDownTime = new RandomVariable(this, dtd, "down time");
		myState = new TimeWeighted(this, "state");
		myCycleLength = new ResponseVariable(this,"cycle length");

		myCountFailures = new Counter(this, "count failures");
		myCountF = new TimeWeighted(this, "count failures TW shadow");
//		myTest = new ResponseVariable(this, "test RV");
		
//		myCountFailures.setTimedUpdateInterval(10.0);
//		myCountFailures.setLimit(10.0);
//		myCountFailures.addStoppingAction();
		// Let's get fancy and randomly determine the initial state
		double meanUpTime = utd.getMean();
		double meanDownTime = dtd.getMean();

		// chance of being in up state
		double p = meanUpTime/(meanUpTime + meanDownTime);
//		System.out.println("p = " + p);
		myInitialState = new RandomVariable(this, new Bernoulli(p), "initial state");

	}

	public void initialize(){
		myTimeLastUp = 0.0; // assume 0.0 but test
		myCount = 0.0;
		if(myInitialState.getValue() > 0.0){ // up state
			myState.setValue(myHeight.getValue());
			// schedule the down state change after the uptime
			scheduleEvent(myDownChangeListener, myUpTime.getValue(), "Down");
		} else { // down state
			myState.setValue(0.0);
			// schedule the up state change after the down time
			scheduleEvent(myUpChangeListener, myDownTime.getValue(), "Up");
		}
	//	scheduleEvent(500.0);
	}
	
	public void handleEvent(JSLEvent e){
		System.out.println("After replication in handle event:");
		double c = myCountFailures.getValue();
		System.out.println("The count was : " + c);
//		myTest.setValue(c);		
	}

	public void replicationEnded(){
//		System.out.println("Replication Ended:");
		double c = myCountFailures.getValue();
//		System.out.println("The count was : " + c);
//		myTest.setValue(c);
	}
	
	class UpChangeListener implements ActionListenerIfc {
		public void action(JSLEvent event) {
            // record the cycle length, the time btw up states
            if (myTimeLastUp > 0.0) // it has been set at least once, record the cycle
            	myCycleLength.setValue(getTime() - myTimeLastUp);
            
            // component has just gone up, change its state value
            
            myState.setValue(myHeight.getValue());
            
            // record the time it went up
			myTimeLastUp = getTime();
			
            // schedule when it goes down, reuse the event
            event.setListener(myDownChangeListener); // change the listener
            event.setName("Down");
			rescheduleEvent(event, myUpTime.getValue());// after the up time elapse, state goes to down
		}
	}

	class DownChangeListener implements ActionListenerIfc {
		public void action(JSLEvent event) {

			// component has just gone down, change its state value
			myCountFailures.increment();
			myCountF.increment();
			myState.setValue(0.0);
			myCount++;
//            System.out.println("time: " + getTime() + " num failures " + myCount);
			// schedule when it goes up, reuse the event
			event.setListener(myUpChangeListener); // change the listener
			event.setName("Up");
			rescheduleEvent(event, myDownTime.getValue());// after the down time elapse, state goes to up

		}
	}
	
	public static void main(String[] args) {
		System.out.println("TestVariables test");
        
		JSL.LOGGER.setLevel(Level.FINE);
   
//		testReplication();
        testExperiment();
        
		System.out.println("Done!");
        
	}
	
	public static void testReplication(){
		// create the containing model
		Model m = Model.createModel();
		
		// create the model element and attach it to the main model
		TestUpDownComponent tv = new TestUpDownComponent(m);
		        
		// create a replication to run the model
		Replication r = new Replication(m);

		// set the running parameters of the replication
		r.setLengthOfReplication(50.0);

		// turn on the desired reporting
//		r.turnOnBatchReport();
				
		// tell the replication to run
		r.runAll();
 
	}

	public static void testExperiment(){
		// create the containing model
		Model m = Model.createModel("UpDown");
		
		// create the model element and attach it to the main model
		TestUpDownComponent tv = new TestUpDownComponent(m);

        try{
        	CSVReport bcsv = new CSVExperimentReport("UpDownCSVExampleExp");
        	m.addObserver(bcsv);
        	CSVReport rcsv = new CSVReplicationReport("UpDownCSVExampleRep");
        	m.addObserver(rcsv);
        }
        catch(IOException e){
			System.out.println("I/O exception opening file.");
			e.printStackTrace();
        }
      
//		m.turnOnLogReport("TestAfterReplication-2");
		// create a replication to run the model
		Experiment e = new Experiment(m);

		// set the running parameters of the experiment
		e.setNumberOfReplications(10);
		e.setLengthOfReplication(500.0);

		// turn on the desired reporting
		e.turnOnExperimentReport("Test-Exp-Report");
		e.turnOnReplicationReport("Test-Rep-Report");
//		e.turnOnReplicationConsoleOutput();
		e.turnOnExperimentConsoleOutput();
				
		// tell the replication to run
		e.runAll();
 
	}
	
	
}
